/*********************************************************************/
/* File:   l2hofespace.cpp                                         */
/* Author: Start                                                     */
/* Date:   24. Feb. 2003                                             */
/*********************************************************************/

/**
   High Order Finite Element Space for L2
*/

/* ***********************************************
To do: *Internal External Dofs (eliminate internal) 
       *Flag for low-order dofs eliminated ...   
************************* */ 
#include <comp.hpp>
#include <multigrid.hpp>

#include <parallelngs.hpp>


#ifdef PARALLEL
extern MPI_Group MPI_HIGHORDER_WORLD;
extern MPI_Comm MPI_HIGHORDER_COMM;
#endif

using namespace ngmg;

namespace ngcomp
{
  using namespace ngcomp;
  using namespace ngparallel;

  L2HighOrderFESpace ::  
  L2HighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FESpace (ama, flags)
  {
    name="L2HighOrderFESpace(l2ho)";
    
    // defined flags
    DefineNumFlag("relorder");
    DefineDefineFlag("print"); 
    DefineDefineFlag("l2ho");
    
    if (parseflags) ParseFlags(flags);
    
    var_order = 0; 
    
    print = flags.GetDefineFlag("print"); 
    if (flags.NumFlagDefined("order"))
      order =  int (flags.GetNumFlag("order",0));
    else 
      {
	if(flags.NumFlagDefined("relorder"))
	  {
	    order=0; 
	    var_order = 1; 
	    rel_order = int (flags.GetNumFlag("relorder",0));
	  }
	else 
	  order = 0; 
      }
    
    if(flags.GetDefineFlag("variableorder") )
      {
	throw Exception ("Flag 'variableorder' for l2ho is obsolete. \n  Either choose uniform order by -order= .. \n -relorder=.. for relative mesh order "); 
      }
    
    if (ma.GetDimension() == 2)
      {
	evaluator = 
	  new MassIntegrator<2> (new ConstantCoefficientFunction(1));
      }
    else
      {
	evaluator = 
	  new MassIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      {
	evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
      }

#ifndef PARALLEL
     low_order_space = new  ElementFESpace (ma, 0, dimension, iscomplex);
#else
     low_order_space = new ParallelElementFESpace ( ma, 0, dimension, iscomplex);
#endif

     prol = new ElementProlongation (*static_cast<ElementFESpace*> (low_order_space));
  }


  L2HighOrderFESpace :: ~L2HighOrderFESpace ()
  {
    ;
  }


  FESpace * L2HighOrderFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    int order = int(flags.GetNumFlag ("order", 0));
    int dim = int(flags.GetNumFlag ("dim", 1));
    bool iscomplex = flags.GetDefineFlag ("complex");

    if (order == 0)
      return new ElementFESpace (ma, order, dim, iscomplex);
    else
      return new L2HighOrderFESpace (ma, flags, true);
  }  

  
  void L2HighOrderFESpace :: Update()
  {
    if(low_order_space) low_order_space -> Update();
                  
    nel = ma.GetNE();
    order_inner.SetSize(nel); 
 
 
    int p  = (var_order ? 0 : order); 
    order_inner = INT<3>(p,p,p); 
    
    int dim = ma.GetDimension(); 

    if(var_order) 
      for(int i=0;i<nel;i++) 
	{
	  INT<3> el_orders = ma.GetElOrders(i);
	  	  
	  for(int j=0;j<dim;j++)
	    order_inner[i][j] =  int(max2(el_orders[j]+rel_order,0)); 
	} 

    if(print) 
      *testout << " order_inner (l2ho) " << order_inner << endl; 

    ndof = nel;
    
    UpdateDofTables();

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;

    if(low_order_space) prol->Update();

#ifdef PARALLEL
    UpdateParallelDofs();
#endif
  } 

  void L2HighOrderFESpace :: UpdateDofTables()
  {
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	INT<3> pi = order_inner[i]; 
	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    ndof += (pi[0]+1)*(pi[0]+2)/2 ;
	    break;
	  case ET_QUAD:
	    ndof += (pi[0]+1)*(pi[1]+1);
	    break;
	  case ET_TET:
	    ndof += (pi[0]+1)*(pi[0]+2)*(pi[0]+3)/6;
	    break;
	  case ET_PRISM:
	    ndof += (pi[0]+1)*(pi[0]+2)*(pi[2]+1)/2;
	    break;
	  case ET_PYRAMID:
	    ndof += 5 + 8*(pi[0]-1) + 2*(pi[0]-1)*(pi[0]-2) + (pi[0]-1)*(pi[0]-1) 
	      + (pi[0]-1)*(pi[0]-2)*(2*pi[0]-3)/6;
	    break;
	  case ET_HEX:
	    ndof += (pi[0]+1)*(pi[1]+1)*(pi[2]+1);
	    break;
	  }
	ndof--; // subtract constant 
      }
    first_element_dof[nel] = ndof;


    if(print) 
      *testout << " first_element dof (l2hofe) " << first_element_dof << endl;  

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;

    prol->Update();

  }

  const FiniteElement & L2HighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    try
      { 
	FiniteElement * fe;
	
	typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
	
	switch (ma.GetElType(elnr))
	  {
	  case ET_TET:
	    { 
//               fe = new (lh.Alloc (sizeof(H1HighOrderTet<T_ORTHOPOL>)))  H1HighOrderTet<T_ORTHOPOL> (order);
              fe = new (lh.Alloc (sizeof(L2HighOrderTet)))  L2HighOrderTet (order);
	      break;
	    }
	  case ET_PYRAMID:
	    {
	      fe = new (lh.Alloc (sizeof(H1HighOrderPyramid<T_ORTHOPOL>)))  H1HighOrderPyramid<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_PRISM:
	    {
// 	      fe = new (lh.Alloc (sizeof(H1HighOrderPrism<T_ORTHOPOL>)))  H1HighOrderPrism<T_ORTHOPOL> (order);
              fe = new (lh.Alloc (sizeof(L2HighOrderPrism)))  L2HighOrderPrism (order);
              break;
	    }
	  case ET_HEX:
	    {
// 	      fe = new (lh.Alloc (sizeof(H1HighOrderHex<T_ORTHOPOL>)))  H1HighOrderHex<T_ORTHOPOL> (order);
              fe = new (lh.Alloc (sizeof(L2HighOrderHex)))  L2HighOrderHex (order);
              break;
	    }
	  case ET_TRIG:
	    { 
	      fe = new (lh.Alloc (sizeof(L2HighOrderTrig)))  L2HighOrderTrig(order);
	      break;
	    }
	  case ET_QUAD:
	    {
	      fe = new (lh.Alloc (sizeof(L2HighOrderQuad)))  L2HighOrderQuad(order);
	      break;
	    }
	  default:
	    fe = 0; 
	  }
	
	if (!fe)
	  {
	    stringstream str;
	    str << "L2HighOrderFESpace " << GetClassName() 
		<< ", undefined eltype " 
		<< ElementTopology::GetElementName(ma.GetElType(elnr))
		<< ", order = " << order << endl;
	    throw Exception (str.str());
	  }
	

	ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
	ma.GetElVertices(elnr, vnums);
        //         if( ma.GetElType(elnr) == ET_TRIG || ma.GetElType(elnr) == ET_QUAD) // all other elements h1
        if( ma.GetElType(elnr) == ET_TRIG || 
            ma.GetElType(elnr) == ET_QUAD || 
            ma.GetElType(elnr) == ET_TET  ||  
            ma.GetElType(elnr) == ET_PRISM || 
            ma.GetElType(elnr) == ET_HEX)
	  { 
	    L2HighOrderFiniteElement * hofe = dynamic_cast<L2HighOrderFiniteElement* > (fe);
// 	    hofe-> SetVertexNumbers (vnums); 
	    hofe-> SetOrderInner(order_inner[elnr]); 
	    hofe-> ComputeNDof(); 
	  } 
	else 
	  {
	    if(var_order) 
	      throw Exception("L2HighOrderFESpace not working for 3D and variable order!"); 
	    
	    H1HighOrderFiniteElement * hofe = dynamic_cast<H1HighOrderFiniteElement* > (fe);
	    hofe-> SetVertexNumbers (vnums,lh); 
	    // ATTENTION for var_order this is not correct !!! 
	    // on edges and faces order is still used !!! 
	    hofe-> SetOrderInner(order_inner[elnr]); 
	    hofe-> ComputeNDof(); 
	  } 
	return *fe;
      } 
    catch (Exception & e)
      {
	e.Append ("in L2HoFESpace::GetElement");
	e.Append ("\n");
	throw; 
      }
  }
 
  const FiniteElement & L2HighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
	fe = new FE_SegmDummy; break;
      case ET_TRIG:
	fe = new FE_SegmDummy ; break;
      case ET_QUAD:
	fe = new FE_SegmDummy; break;
      default:
	fe = 0;
      }
    
    ARRAY<int> vnums;
    ma.GetSElVertices(elnr, vnums);

    if (!fe)
      {
	stringstream str;
	str << "FESpace " << GetClassName() 
	    << ", undefined surface eltype " << ma.GetSElType(elnr) 
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }
    /*
    ArrayMem<int, 8> vnums;
    hofe -> SetOrderInner (0); 
    */
    /*
    dynamic_cast<H1HighOrderFiniteElement*> (fe)
      -> SetVertexNumbers (vnums);
    */
    return *fe;
  }
 

  int L2HighOrderFESpace :: GetNDof () const
  {
    return ndof;
  }

  int L2HighOrderFESpace :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }


  void L2HighOrderFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    dnums.Append (elnr); // lowest_order 
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;

    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;

    // (*testout) << "l2ho, dnums = " << dnums << endl;
  }
  
  void  L2HighOrderFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }
    dnums.SetSize(0);
  }


  void L2HighOrderFESpace :: 
  GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
  
  Table<int> * L2HighOrderFESpace :: 
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, first;
    ARRAY<int> cnt(nel);
    cnt = 0;
    for (i = 0; i < nel; i++)
      cnt[i] = first_element_dof[i+1]-first_element_dof[i];
	
    Table<int> & table = *new Table<int> (cnt);
    
    for (i = 0; i < nel; i++)
      {
	first = first_element_dof[i];
	for (j = 0; j < cnt[i]; j++)
	  table[i][j] = first+j;
      }
    return &table;
  }

  void  L2HighOrderFESpace :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpace ::GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpace ::GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpace ::GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs ( elnr, dnums ); return; }

  L2SurfaceHighOrderFESpace ::  
  L2SurfaceHighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FESpace (ama, flags)
  {
    name="L2SurfaceHighOrderFESpace(l2surf)";
    // defined flags 
    DefineDefineFlag("l2surf");

    if(parseflags) ParseFlags(flags);
    
    
    if(flags.NumFlagDefined("relorder"))
      throw Exception("Variable order not implemented for L2SurfaceHighOrderFESpace"); 
    
    segm = new H1HighOrderSegm<IntegratedLegendreMonomialExt> (order);
    trig = new H1HighOrderTrig<IntegratedLegendreMonomialExt> (order);
    quad = new H1HighOrderQuad<IntegratedLegendreMonomialExt> (order);

    if (ma.GetDimension() == 2)
      {
	boundary_evaluator = 
	  new RobinIntegrator<2> (new ConstantCoefficientFunction(1));
      }
    else
      {
	boundary_evaluator = 
	  new RobinIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      {
	boundary_evaluator = 
	  new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
      }
  }

  L2SurfaceHighOrderFESpace :: ~L2SurfaceHighOrderFESpace ()
  {
    ;
  }

  FESpace * L2SurfaceHighOrderFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new L2SurfaceHighOrderFESpace (ma, flags, true);
  }

  void L2SurfaceHighOrderFESpace :: Update()
  {
    nel = ma.GetNSE();

    ndof = 0;
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	switch (ma.GetSElType(i))
	  {
	  case ET_SEGM:
	    ndof += order+1;
	    break;
	  case ET_TRIG:
	    ndof += (order+1)*(order+2)/2;
	    break;
	  case ET_QUAD:
	    ndof += (order+1)*(order+1);
	    break;
	  default:
	    ;
	  }
      }
    first_element_dof[nel] = ndof;
  }

  const FiniteElement & L2SurfaceHighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
	fe = segm; break;
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0;
      }
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetSElVertices(elnr, vnums);

    if (!fe)
      {
	stringstream str;
	str << "L2SurfaceHighOrderFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    dynamic_cast<H1HighOrderFiniteElement*> (fe) -> SetVertexNumbers (vnums,lh);

    return *fe;
  }
 
  const FiniteElement & L2SurfaceHighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    throw Exception ("Volume elements not available for L2SurfaceHighOrderFESpace");
  }
 
  int L2SurfaceHighOrderFESpace :: GetNDof () const
  {
    return ndof;
  }

  void L2SurfaceHighOrderFESpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetSElIndex (elnr)))
      dnums = -1;
  }
  
  void L2SurfaceHighOrderFESpace :: 
  GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
  
  Table<int> * L2SurfaceHighOrderFESpace :: 
  CreateSmoothingBlocks ( int type) const
  {
    int i, j, first;
    ARRAY<int> cnt(nel);
    cnt = 0;
    for (i = 0; i < nel; i++)
      cnt[i] = first_element_dof[i+1]-first_element_dof[i];
	
    Table<int> & table = *new Table<int> (cnt);
    
    for (i = 0; i < nel; i++)
      {
	first = first_element_dof[i];
	for (j = 0; j < cnt[i]; j++)
	  table[i][j] = first+j;
      }
    return &table;
  }


  void  L2SurfaceHighOrderFESpace :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2SurfaceHighOrderFESpace ::GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2SurfaceHighOrderFESpace ::GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { GetDofNrs ( fanr, dnums ); return; }
  
  void  L2SurfaceHighOrderFESpace ::GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs ( elnr, dnums ); return; }
    

#ifdef PARALLEL

void L2HighOrderFESpace :: UpdateParallelDofs_hoproc()
  {
    
     // ******************************
     // update exchange dofs 
     // ******************************
    *testout << "L2HOFESpace::UpdateParallelDofs_hoproc" << endl;
    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks);
    nexdof = 0;

    const MeshAccess & ma = (*this). GetMeshAccess();

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

    for ( int i = 0; i < ndof; i++ )
      {
	paralleldofs->ClearExchangeDof(i);
	for ( int dest = 0; dest < ntasks; dest++ )
	  paralleldofs->ClearExchangeDof(dest, i);
      }


    int * ndof_dist = new int[ntasks];
    MPI_Allgather ( &ndof, 1, MPI_INT, &ndof_dist[0], 
		    1, MPI_INT, MPI_COMM_WORLD);
    for ( int dest = 0; dest < ntasks; dest++ )
      paralleldofs -> SetDistNDof( dest, ndof_dist[dest]) ;

    delete [] ndof_dist;


  }

  void L2HighOrderFESpace :: UpdateParallelDofs_loproc()
  {
    *testout << "L2HOFESpace::UpdateParallelDofs_loproc" << endl;

    const MeshAccess & ma = (*this). GetMeshAccess();

    int ndof = GetNDof();

    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks); 
    nexdof = 0;

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

 
  }
#endif // PARALLEL



  // register FESpaces
  namespace
#ifdef MACOS
  l2hofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("l2", L2HighOrderFESpace::Create);
      GetFESpaceClasses().AddFESpace ("l2ho", L2HighOrderFESpace::CreateHO);
      GetFESpaceClasses().AddFESpace ("l2surf", L2SurfaceHighOrderFESpace::Create);
    }
    
    Init init;
  }
}
