#include <comp.hpp>


#include <parallelngs.hpp>

namespace ngcomp
{
  using namespace ngcomp;
  using namespace ngparallel;

  LinearForm ::
  LinearForm (const FESpace & afespace, 
	      const string & aname,
	      const Flags & flags)
    : NGS_Object(afespace.GetMeshAccess(), aname), fespace(afespace)
  {
    independent = false;
    print = flags.GetDefineFlag ("print");
    printelvec = flags.GetDefineFlag ("printelvec");
    assembled = false;
    initialassembling = true;
    cacheblocksize = 1;
  }

  LinearForm :: ~LinearForm ()
  { 
    for (int i = 0; i < parts.Size(); i++)
      if ( parts_deletable[i]) delete parts[i];
  }

  void LinearForm :: AddIntegrator (LinearFormIntegrator * lfi, const bool deletable)
  {
    parts.Append (lfi);
    parts_deletable.Append(deletable);
  }

  void LinearForm :: PrintReport (ostream & ost)
  {
    ost << "on space " << GetFESpace().GetName() << endl
	<< "integrators: " << endl;
    for (int i = 0; i < parts.Size(); i++)
      ost << "  " << parts[i]->Name() << endl;
  }

  void LinearForm :: MemoryUsage (ARRAY<MemoryUsageStruct*> & mu) const
  {
    if (&GetVector())  
      {
	int olds = mu.Size();
	GetVector().MemoryUsage (mu);
	for (int i = olds; i < mu.Size(); i++)
	  mu[i]->AddName (string(" lf ")+GetName());
      }
  }

  bool LinearForm :: InitialAssembling (void) {return initialassembling; }
  void LinearForm :: SetNoInitialAssembling (void) {initialassembling =  false; }

  bool LinearForm :: IsAssembled (void) { return assembled; }

  template <class SCAL>
  BaseVector & T_LinearForm<SCAL> :: GetVector () const 
  { 
#ifdef PARALLEL
    const FESpace & afespace = LinearForm :: GetFESpace();
    vec -> SetParallelDofs ( &afespace.GetParallelDofs() );
    if ( &(afespace.GetParallelDofs()) )
      vec -> SetStatus (DISTRIBUTED);
    else
      vec -> SetStatus ( NOT_PARALLEL );
#endif
    return *vec; 
  }

  template <class SCAL>
  void S_LinearForm<SCAL> :: Assemble (LocalHeap & lh)
  {
    static int vectimer = NgProfiler::CreateTimer ("Vector assembling");
    NgProfiler::RegionTimer reg (vectimer);

    assembled = true;

    if (independent)
      {
	AssembleIndependent(lh);
	return;
      }
    
    try
      {
	ma.PushStatus ("Assemble Vector");
	int i, j;

	AllocateVector();

	ARRAY<int> dnums;
	ElementTransformation eltrans;
	
	int ne = ma.GetNE();
	int nse = ma.GetNSE();

	bool hasbound = false;
	bool hasinner = false;
	
	for (j = 0; j < parts.Size(); j++)
	  {
	    if (parts[j] -> BoundaryForm())
	      hasbound = true;
	    else if (!parts[j] -> IntegrationAlongCurve())
	      hasinner = true;
	  }
	
        if (hasinner)
	  {
	    
	    for (i = 0; i < ne; i++)
	      {
		if (i % 10 == 0)
		  cout << "\rassemble element " << i << "/" << ne << flush;
		ma.SetThreadPercentage ( 100.0*(i+1) / (ne+nse) );

		lh.CleanUp();
		
		if ( ma.IsGhostEl( i ) ) continue;

		const FiniteElement & fel = fespace.GetFE (i, lh);
		ma.GetElementTransformation (i, eltrans, lh);
		      
		fespace.GetDofNrs (i, dnums);
		
		for (j = 0; j < parts.Size(); j++)
		  {
		    if (parts[j] -> BoundaryForm()) continue;
		    if (!parts[j] -> DefinedOn (ma.GetElIndex (i))) continue;
		    if (parts[j] -> IntegrationAlongCurve()) continue;

		    FlatVector<TSCAL> elvec;
		  
		    parts[j] -> AssembleElementVector (fel, eltrans, elvec, lh);
                 
		    if (printelvec)
		      {
			testout->precision(8);

			(*testout) << "elnum= " << i << endl;
			(*testout) << "integrator " << parts[j]->Name() << endl;
			(*testout) << "dnums = " << endl << dnums << endl;
			(*testout) << "element-index = " << eltrans.GetElementIndex() << endl;
			(*testout) << "elvec = " << endl << elvec << endl;
		      }
		
		    fespace.TransformVec (i, false, elvec, TRANSFORM_RHS);
		
		    AddElementVector (dnums, elvec, parts[j]->CacheComp()-1);
		
		    
		  }
	      }
	    cout << "\rassemble element " << ne << "/" << ne << endl;
	  }


	if (hasbound)
	  {
	    for (i = 0; i < nse; i++)
	      {
		if (i % 10 == 0)
		  cout << "\rassemble surface element " << i << "/" << nse << flush;
		ma.SetThreadPercentage ( 100.0*(ne+i+1) / (ne+nse) );

		if ( ma.IsGhostSEl ( i ) ) continue;

		lh.CleanUp();
	      
		const FiniteElement & fel = fespace.GetSFE (i, lh);
	      
		ma.GetSurfaceElementTransformation (i, eltrans, lh);
		fespace.GetSDofNrs (i, dnums);
	      
		for (j = 0; j < parts.Size(); j++)
		  {
		    if (!parts[j] -> BoundaryForm()) continue;
		    if (!parts[j] -> DefinedOn (ma.GetSElIndex (i))) continue;
		    if (parts[j] -> IntegrationAlongCurve()) continue;		    
		  
		    FlatVector<TSCAL> elvec;
		    parts[j] -> AssembleElementVector (fel, eltrans, elvec, lh);
		    if (printelvec)
		      {
			testout->precision(8);

			(*testout) << "surface-elnum= " << i << endl;
			(*testout) << "integrator " << parts[j]->Name() << endl;
			(*testout) << "dnums = " << endl << dnums << endl;
			(*testout) << "element-index = " << eltrans.GetElementIndex() << endl;
			(*testout) << "elvec = " << endl << elvec << endl;
		      }



		    fespace.TransformVec (i, true, elvec, TRANSFORM_RHS);
		    
		    
		    AddElementVector (dnums, elvec, parts[j]->CacheComp()-1);
		    
		  }
	      }
	    cout << "\rassemble surface element " << nse << "/" << nse << endl;	  
	  }



	for ( j=0; j<parts.Size(); j++ )
	  {
	    if (!(parts[j] -> IntegrationAlongCurve())) continue;
	    
	    const FiniteElement * fel;
	    int oldelement;
	    int element;
	    double oldlength;
	    

	    for(int nc = 0; nc < parts[j]->GetNumCurveParts(); nc++)
	      {
		oldelement = -1;
		oldlength = 0;

		for(i = parts[j]->GetStartOfCurve(nc); i < parts[j]->GetEndOfCurve(nc); i++)
		  {
		    
		    if (i%500 == 0)
		      {
			cout << "\rassemble curvepoint " << i << "/" << parts[j]->NumCurvePoints() << flush;
			ma.SetThreadPercentage(100.*i/parts[j]->NumCurvePoints());
		      }
		    
		    FlatVector<TSCAL> elvec;
		    IntegrationPoint ip;
		    element = ma.FindElementOfPoint(parts[j]->CurvePoint(i),ip,true);
		    if(element < 0)
		      throw Exception("element for curvepoint not found");
		    
		    if(element != oldelement)
		      { 
			lh.CleanUp();
			fel = &fespace.GetFE(element,lh);
			ma.GetElementTransformation(element,eltrans,lh);
			fespace.GetDofNrs(element,dnums);
		      }
		    
		    void * heapp = lh.GetPointer();
		    
		    FlatVector<double> tangent(parts[j]->CurvePoint(0).Size(),lh);
		    tangent = parts[j]->CurvePointTangent(i);
		    double length = L2Norm(tangent);
		    if(length < 1e-5)
		      {
			int i1,i2;
			i1 = i-1; i2=i+1;
			if(i1 < parts[j]->GetStartOfCurve(nc)) i1=parts[j]->GetStartOfCurve(nc);
			if(i2 >= parts[j]->GetEndOfCurve(nc)) i2 = parts[j]->GetEndOfCurve(nc)-1;
			
			for(int k=0; k<tangent.Size(); k++)
			  tangent[k] = (parts[j]->CurvePoint(i2))[k]-(parts[j]->CurvePoint(i1))[k];
			
			length = L2Norm(tangent);
		      }
		    
		    tangent *= 1./length;
		    
		    length = 0;
		    if(i < parts[j]->GetEndOfCurve(nc)-1)
		      for(int k=0; k<tangent.Size(); k++)
			length += pow(parts[j]->CurvePoint(i+1)[k]-parts[j]->CurvePoint(i)[k],2);
		    length = 0.5*sqrt(length);
		    
		    
		    if (eltrans.SpaceDim() == 3)
		      {
			SpecificIntegrationPoint<1,3> s_sip(ip,eltrans,lh);
			SpecificIntegrationPoint<3,3> g_sip(ip,eltrans,lh);
			Vec<3> tv;
			tv(0) = tangent(0); tv(1) = tangent(1); tv(2) = tangent(2);
			s_sip.SetTV(tv);
			parts[j]->AssembleElementVectorIndependent(*fel,
								   s_sip,
								   g_sip,
								   elvec,lh,true);
		      }
		    else if (eltrans.SpaceDim() == 2)
		      {
			SpecificIntegrationPoint<1,2> s_sip(ip,eltrans,lh);
			SpecificIntegrationPoint<2,2> g_sip(ip,eltrans,lh);
			Vec<2> tv;
			tv(0) = tangent(0); tv(1) = tangent(1);
			s_sip.SetTV(tv);
			parts[j]->AssembleElementVectorIndependent(*fel,
								   s_sip,
								   g_sip,
								   elvec,lh,true);
		      }
		    fespace.TransformVec (element, false, elvec, TRANSFORM_RHS);

		    elvec *= (oldlength+length); // Des is richtig
		    
		    
		    AddElementVector (dnums, elvec, parts[j]->CacheComp()-1);
		    
		    
		    
		    
		    lh.CleanUp(heapp);
		    
		    oldlength = length;      
		    
		    oldelement = element;
		  }
	      }



	    cout << "\rassemble curvepoint " << parts[j]->NumCurvePoints() << "/" 
		 << parts[j]->NumCurvePoints() << endl;
	    lh.CleanUp();

	  }
	
	




	if (print)
	  {
	    (*testout) << "Linearform " << GetName() << ": " << endl;
	    (*testout) << GetVector() << endl;
	  }

	ma.PopStatus ();
	// (*testout) << "Linearform, vec = " << endl << GetVector() << endl;


      }
    catch (Exception & e)
      {
	stringstream ost;
	ost << "in Assemble LinearForm" << endl;
	e.Append (ost.str());
	throw;
      }
    catch (exception & e)
      {
	throw (Exception (string(e.what()) +
			  string("\n in Assemble LinearForm\n")));
      }
  }








  template <class SCAL>
  void S_LinearForm<SCAL> :: AssembleIndependent (LocalHeap & lh)
  {
    assembled = true;

    try
      {
	int i, j, k;
	
	AllocateVector();

	// int ne = ma.GetNE();
	int nse = ma.GetNSE();
	
	
	ARRAY<int> dnums;
	ElementTransformation seltrans, geltrans;

	for (i = 0; i < nse; i++)
	  {
	    lh.CleanUp();
	    
	    const FiniteElement & sfel = fespace.GetSFE (i, lh);
	    ma.GetSurfaceElementTransformation (i, seltrans, lh);
	      	
	    // (*testout) << "el = " << i << ", ind = " << ma.GetSElIndex(i) << endl;
	    if (!parts[0]->DefinedOn (ma.GetSElIndex(i))) continue;
	    // (*testout) << "integrate surf el " << endl;
	    
	    const IntegrationRule & ir =
	      GetIntegrationRules().SelectIntegrationRule (sfel.ElementType(), 5);
	    
	    for (j = 0; j < ir.GetNIP(); j++)
	      {
		const IntegrationPoint & ip = ir.GetIP(j);
		SpecificIntegrationPoint<2,3> sip(ip, seltrans, lh);
		
		// (*testout) << "point = " << sip.GetPoint() << endl;
		
		IntegrationPoint gip;
		int elnr;
		elnr = ma.FindElementOfPoint (FlatVector<>(sip.GetPoint()), gip, 1);
		
		// (*testout) << "elnr = " << elnr << endl;
		if (elnr == -1) continue;
		
		const FiniteElement & gfel = fespace.GetFE (elnr, lh);
		ma.GetElementTransformation (elnr, geltrans, lh);
		SpecificIntegrationPoint<3,3> gsip(gip, geltrans, lh);
		
		// (*testout) << " =?= p = " << gsip.GetPoint() << endl;

		fespace.GetDofNrs (elnr, dnums);
		
		for (k = 0; k < parts.Size(); k++)
		  {
		    FlatVector<TSCAL> elvec;
		    
		    if(geltrans.SpaceDim() == 3)
		      {
			SpecificIntegrationPoint<2,3> s_sip(ip,seltrans,lh);
			SpecificIntegrationPoint<3,3> g_sip(gip,geltrans,lh);
			parts[k] -> AssembleElementVectorIndependent
			  (gfel,s_sip,g_sip,elvec,lh);
		      }
		    else if(geltrans.SpaceDim() == 2)
		      {
			SpecificIntegrationPoint<1,2> s_sip(ip,seltrans,lh);
			SpecificIntegrationPoint<2,2> g_sip(gip,geltrans,lh);
			parts[k] -> AssembleElementVectorIndependent
			  (gfel,s_sip,g_sip,elvec,lh);
		      }

		    // (*testout) << "elvec, 1 = " << elvec << endl;

		    elvec *= fabs (sip.GetJacobiDet()) * ip.Weight();
		    fespace.TransformVec (elnr, 0, elvec, TRANSFORM_RHS);

		    // (*testout) << "final vec = " << elvec << endl;
		    // (*testout) << "dnums = " << dnums << endl;
		    AddElementVector (dnums, elvec);
		  }
	      }
	  }
	// (*testout) << "Assembled vector = " << endl << GetVector() << endl;
      }
    
    catch (Exception & e)
      {
	stringstream ost;
	ost << "in Assemble LinearForm Independent" << endl;
	e.Append (ost.str());
	throw;
      }
    catch (exception & e)
      {
	throw (Exception (string(e.what()) +
			  string("\n in Assemble LinearForm Independent\n")));
      }
  }














  template class S_LinearForm<double>;
  template class S_LinearForm<Complex>;




  template <typename TV>
  T_LinearForm<TV> ::
  T_LinearForm (const FESpace & afespace, const string & aname, const Flags & flags)
    : S_LinearForm<TSCAL> (afespace, aname, flags), vec(0) 
  { 
    ; 
  }
  
  template <typename TV>
  T_LinearForm<TV> :: ~T_LinearForm ()
  {
    delete vec;
  }

#ifndef PARALLEL
  template <typename TV>
  void T_LinearForm<TV> :: AllocateVector ()
  {
    delete vec;
    vec = new ngla::VVector<TV> (this->fespace.GetNDof());
    (*vec) = TSCAL(0);
  }
#else
  template <typename TV>
  void T_LinearForm<TV> :: AllocateVector ()
  {
    delete vec;
    const FESpace & afespace = this->fespace;
    if ( &afespace.GetParallelDofs() == 0 )
      vec = new VVector<TV> (afespace.GetNDof());
    else
      vec = new ParallelVVector<TV> ( afespace.GetNDof(),& afespace.GetParallelDofs());
    (*vec) = TSCAL(0);
  }
#endif


  template <typename TV>
  void T_LinearForm<TV> :: CleanUpLevel ()
  {
    delete vec;
    vec = 0;
  }


  template <typename TV>
  void T_LinearForm<TV> ::
  AddElementVector (const ARRAY<int> & dnums,
		    const FlatVector<TSCAL> & elvec,
		    const int cachecomp) 
  {
    FlatVector<TV> fv = vec->FV();

    if(cachecomp < 0)
      {
	for (int k = 0; k < dnums.Size(); k++)
	  if (dnums[k] != -1)
	    for (int j = 0; j < HEIGHT; j++)
	      fv(dnums[k])(j) += elvec(k*HEIGHT+j);
      }
    else
      {
	for (int k = 0; k < dnums.Size(); k++)
	  if (dnums[k] != -1)
	    fv(dnums[k])(cachecomp) += elvec(k);
      }
  }
  
  template <> void T_LinearForm<double>::
  AddElementVector (const ARRAY<int> & dnums,
		    const FlatVector<double> & elvec,
		    const int cachecomp) 
  {
    FlatVector<double> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	fv(dnums[k]) += elvec(k);
  }
  
  template <> void T_LinearForm<Complex>::
  AddElementVector (const ARRAY<int> & dnums,
		    const FlatVector<Complex> & elvec,
		    const int cachecomp) 
  {
    FlatVector<Complex> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	fv(dnums[k]) += elvec(k);
  }
  




  template <typename TV>
  void T_LinearForm<TV> ::
  SetElementVector (const ARRAY<int> & dnums,
		    const FlatVector<TSCAL> & elvec) 
  {
    FlatVector<TV> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	for (int j = 0; j < HEIGHT; j++)
	  fv(dnums[k])(j) = elvec(k*HEIGHT+j);
  }
  
  template <> void T_LinearForm<double>::
  SetElementVector (const ARRAY<int> & dnums,
		    const FlatVector<double> & elvec) 
  {
    FlatVector<double> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	fv(dnums[k]) = elvec(k);
  }
  
  template <> void T_LinearForm<Complex>::
  SetElementVector (const ARRAY<int> & dnums,
		    const FlatVector<Complex> & elvec) 
  {
    FlatVector<Complex> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	fv(dnums[k]) = elvec(k);
  }
  




  template <typename TV>
  void T_LinearForm<TV> ::
  GetElementVector (const ARRAY<int> & dnums,
		    FlatVector<TSCAL> & elvec) const
  {
    FlatVector<TV> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	for (int j = 0; j < HEIGHT; j++)
	  elvec(k*HEIGHT+j) = fv(dnums[k])(j);
  }
  
  template <> void T_LinearForm<double>::
  GetElementVector (const ARRAY<int> & dnums,
		    FlatVector<double> & elvec) const
  {
    FlatVector<double> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	elvec(k)= fv(dnums[k]);
  }
  
  template <> void T_LinearForm<Complex>::
  GetElementVector (const ARRAY<int> & dnums,
		    FlatVector<Complex> & elvec) const
  {
    FlatVector<Complex> fv = vec->FV();
    for (int k = 0; k < dnums.Size(); k++)
      if (dnums[k] != -1)
	elvec(k)= fv(dnums[k]);
  }
  







  LinearForm * CreateLinearForm (const FESpace * space,
				 const string & name, const Flags & flags)
  {
    /*
    LinearForm * lf;
    CreateVecObject2 (lf, T_LinearForm, 
		      space->GetDimension(), space->IsComplex(),   
		      *space, name);

    lf->SetIndependent (flags.GetDefineFlag ("independent"));
    return lf;
    */

    LinearForm * lf = 
      CreateVecObject  <T_LinearForm, LinearForm, const FESpace, const string, const Flags>
      (space->GetDimension() * int(flags.GetNumFlag("cacheblocksize",1)), space->IsComplex(), *space, name, flags);
    
    lf->SetIndependent (flags.GetDefineFlag ("independent"));

    if (flags.GetDefineFlag ( "noinitialassembling" )) lf->SetNoInitialAssembling();
    
    lf->SetCacheBlockSize(int(flags.GetNumFlag("cacheblocksize",1)));

    return lf;
  }

}
