#ifdef FAST

#include <fem.hpp>
  
namespace ngfem
{
  using namespace ngfem;
  
  

  class FastMassIntegratorTet : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;
    BilinearFormIntegrator * orig;

  public:
    FastMassIntegratorTet (CoefficientFunction * acoef)
      : BilinearFormIntegrator (), coef(acoef)
    { 
      orig = new MassIntegrator<3> (acoef); 
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new FastMassIntegratorTet (coeffs[0]);
    }

    virtual bool BoundaryForm () const { return 0; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const;

    
    virtual string Name () const { return "fast mass tet"; }
  };


  void FastMassIntegratorTet ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    FlatMatrix<> mat1;
    if (checkfast)
      {
	orig -> AssembleElementMatrix (bfel, eltrans, mat1, lh); 
	// (*testout) << "original: " << mat1 << endl;
      }

    try
      {
	static int maintimer = NgProfiler::CreateTimer ("FastMassIntegrator::Assemble");

	NgProfiler::StartTimer (maintimer);

	
	NgProfiler::StartTimer (101);



	const H1HighOrderTetFast & fel = dynamic_cast<const H1HighOrderTetFast&> (bfel);
	int ndof = fel.GetNDof();

	elmat.AssignMemory (ndof, ndof, lh);
	elmat = 0;

	
	int intorder = 2*fel.Order() + 6;
	if (common_integration_order >= 0)
	  intorder = common_integration_order;

	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, intorder);

	TensorProductIntegrationRule ir (fel, ir1d);
	

	FlatVector<> dmats(ir.GetNIP(), lh);

	
	void * heapp = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<3,3> sip(ir[i], eltrans, lh);

	    Mat<1> dmat;
	    Vec<1> hv1, hv2;

	    dmats(i) = coef -> Evaluate (sip);
	    
	    dmats(i) *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    lh.CleanUp (heapp);
	  }


	int order = fel.Order();

	FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    double x = ir.GetIPX(ix);
	    double * hp = &legxbub(ix, 2);
	    // LegendrePolynomial (order-2, 2*x-1, hp);
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);

	    legxbub(ix, 0) = x;
	    legxbub(ix, 1) = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      legxbub(ix, j+2) *= x * (1-x);
	  }

	FlatMatrix<> legybub(ir.GetNIPY(), order+1, lh);
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double y = ir.GetIPY(iy);
	    double * hp = &legybub(iy, 2);
	    // LegendrePolynomial (order-2, 2*y-1, hp);
	    JacobiPolynomial (order-2, 2*y-1, 1, 1, hp);
	    for (int j = 2; j <= order; j++)
	      legybub(iy, j) *= y*(1-y);
	    legybub(iy,0) = y;
	    legybub(iy,1) = 1-y;
	  }

	FlatMatrix<> legybubT(order+1, ir.GetNIPY(), lh);
	legybubT = Trans (legybub);

	FlatMatrix<> legzbub(ir.GetNIPZ(), order+1, lh);
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    double z = ir.GetIPZ(iz);
	    double * hp = &legzbub(iz, 2);
	    // LegendrePolynomial (order-2, 2*z-1, hp);
	    JacobiPolynomial (order-2, 2*z-1, 1, 1, hp);
	    for (int j = 2; j <= order; j++)
	      legzbub(iz, j) *= z*(1-z);
	    legzbub(iz,0) = z;
	    legzbub(iz,1) = 1-z;
	  }

	FlatMatrix<> legzbubT(order+1, ir.GetNIPZ(), lh);
	legzbubT = Trans (legzbub);


	FlatMatrix<> powzm(ir.GetNIPZ(), 2*order+2, lh);
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    double fac = 1-ir.GetIPZ(iz);
	    double val = 1;
	    for (int j = 0; j < 2*order+2; j++, val *= fac)
	      powzm(iz, j) = val;
	  }

	FlatMatrix<> powym(ir.GetNIPZ(), 2*order+2, lh);
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double fac = 1-ir.GetIPY(iy);
	    double val = 1;
	    for (int j = 0; j < 2*order+2; j++, val *= fac)
	      powym(iy, j) = val;
	  }


	ARRAY<FlatMatrix<> > jaczbTs (order+1);
	FlatMatrix<> jaczb(ir.GetNIPZ(), order+1, lh);

	for (int k = 0; k <= order; k++)
	  {
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      {
		double * hp = &jaczb(iz, 1);
		JacobiPolynomial (order-1, 2*ir.GetIPZ(iz)-1, 2*k-3, 1, hp);
		for (int j = 1; j <= order; j++)
		  jaczb(iz, j) *= ir.GetIPZ(iz) * powzm(iz, k);
		jaczb(iz, 0) = powzm(iz, k);
	      }
	    
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    jaczbTs[k] = Trans (jaczb);
	  }
	jaczbTs[0] = Trans (legzbub);
	jaczbTs[1] = Trans (legzbub);




	/*
	(*testout) << "Generic pols: " << endl;
	for (int k = 0; k <= order; k++)
	  (*testout) << *generic_jaczb[k] << endl;
	*/


	ARRAY<FlatMatrix<> > jacybTs (order+1);
	FlatMatrix<> jacyb(ir.GetNIPY(), order+1, lh);

	for (int k = 0; k <= order; k++)
	  {
	    for (int iy = 0; iy < ir.GetNIPY(); iy++)
	      {
		double * hp = &jacyb(iy, 1);
		JacobiPolynomial (order-1, 2*ir.GetIPY(iy)-1, 2*k-3, 1, hp);
		for (int j = 1; j <= order; j++)
		  jacyb(iy, j) *= ir.GetIPY(iy) * powym(iy, k);
		jacyb(iy, 0) = powym(iy, k);
	      }
	    
	    jacybTs[k].AssignMemory (order+1, ir.GetNIPY(), lh);
	    jacybTs[k] = Trans (jacyb);
	  }
	jacybTs[0] = Trans (legybub);
	jacybTs[1] = Trans (legybub);

	NgProfiler::StopTimer (101);







	NgProfiler::StartTimer (102);

	ARRAY<RecPol*> generic_jaczb(order+1);
	generic_jaczb[0] = new RecPolJacobi (order+1, 1, 1);
	generic_jaczb[0] -> MultLinear (0.5, 0.5);
	generic_jaczb[0] -> MultLinear (0.5, -0.5);

	generic_jaczb[1] = new RecPolJacobi (order+1, 1, 1);
	generic_jaczb[1] -> MultLinear (0.5, 0.5);
	generic_jaczb[1] -> MultLinear (0.5, -0.5);

	for (int k = 2; k <= order; k++)
	  {
	    generic_jaczb[k] = new RecPolJacobi (order+1, 2*k-3, 1);
	    generic_jaczb[k] -> MultLinear (0.5, 0.5);
	  }
	
	ARRAY<FlatMatrix<> > generic_jaczb_vals(order+1);
	FlatMatrix<> hgenmat(ir.GetNIPZ(), order+1, lh);
	for (int k = 0; k <= order; k++)
	  {
	    for (int i = 0; i < ir.GetNIPZ(); i++)
	      generic_jaczb[k]->Evaluate (order, 2*ir.GetIPZ(i)-1, &hgenmat(i,0)); 

	    generic_jaczb_vals[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    generic_jaczb_vals[k] = Trans (hgenmat);
	  }

	NgProfiler::StopTimer (102);





	NgProfiler::StartTimer (103);

    
	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (fel.GetVNums()[sort[j]] > fel.GetVNums()[sort[j+1]]) swap (sort[j], sort[j+1]);
    
	int isort[4];
	for (int i = 0; i < 4; i++) isort[sort[i]] = i;



	// precompute 2D matrices

	int ndof2d = (order+1)*(order+2)/2;
	int ndof1d = order+1;
	int ngrid2d = ir.GetNIPX()*ir.GetNIPY();


	ARRAY<FlatMatrix<> > mat2d(ir.GetNIPZ());

	ARRAY<FlatMatrix<> > mat1d(ir.GetNIPY());
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  mat1d[iy].AssignMemory (ndof1d, ndof1d, lh);

	FlatVector<> coefy(ir.GetNIPY(), lh);
	FlatVector<> coefyb1(ir.GetNIPY(), lh);


	FlatArray<int> quad2trig((order+1)*(order+1), lh);
	quad2trig = -1;
	quad2trig[0] = 0;
	quad2trig[order+1] = 0;
	quad2trig[1] = 1;
	quad2trig[order+2] = 2;
	
	int pos = 3;
	for (int ky = 2; ky <= order; ky++)
	  {
	    quad2trig[ky] = pos++;
	    quad2trig[ky+order+1] = pos++;
	  }
	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = pos++;


	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    mat2d[iz].AssignMemory (ndof2d, ndof2d, lh);
	    mat2d[iz] = 0.0;


	    // precompute 1D matrices mat1d[iy]
	    for (int iy = 0; iy < ir.GetNIPY(); iy++)
	      {
		int first_dmati = iz*ngrid2d+iy*ir.GetNIPX();
		for (int i = 0; i < order+1; i++)
		  for (int j = 0; j < order+1; j++)
		    {
		      double sum = 0;
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			sum += legxbub(ix, i) * dmats[first_dmati+ix] * legxbub(ix, j);
		      mat1d[iy](i,j) = sum;
		    }
	      }

	    for (int kx1 = 0; kx1 <= order; kx1++)
	      for (int kx2 = 0; kx2 <= order; kx2++)
		{
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    coefy(iy) = mat1d[iy](kx1,kx2);

		  FlatMatrix<> base1( jacybTs[kx1] );
		  FlatMatrix<> base2( jacybTs[kx2] );

		  for (int i = 0; i < order+1; i++)
		    {
		      int mi = quad2trig[kx1*(order+1)+i];
		      if (mi != -1)
			{
			  for (int k = 0; k < ir.GetNIPY(); k++)
			    coefyb1(k) = base1(i,k) * coefy(k);
			  
			  for (int j = 0; j < order+1; j++)
			    {
			      int mj = quad2trig[kx2*(order+1)+j];
			      if (mj != -1)
				{
				  double sum = 0;
				  for (int k = 0; k < ir.GetNIPY(); k++)
				    sum += coefyb1(k) * base2(j,k);
				  mat2d[iz](mi, mj) += sum;
				}
			    }
			}

		    }
		}
	  }
	
	
	NgProfiler::StopTimer (103);	
	
	NgProfiler::StartTimer (104);	
	
	
	
	

	// scaling with (1-z)^m
	
	FlatArray<int> powerz(ndof2d, lh);
	for (int k = 0; k < 3; k++)
	  powerz[k] = 0;
	
	pos = 3;
	for (int k = 2; k <= order; k++)
	  {
	    powerz[pos++] = k;
	    powerz[pos++] = k;
	  }

	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    powerz[pos++] = kx+ky;


	
	

	// transform prism ---> tet

	ARRAY<int> prism2tet;
	fel.GetPrismTetReordering (isort, prism2tet);

	// build 3D - tensor product matrix:

	FlatVector<> coefz(ir.GetNIPZ(), lh);
	FlatVector<> coefzb1(ir.GetNIPZ(), lh);
	FlatVector<> pts_z(ir.GetNIPZ(), lh);
	FlatMatrix<> mat1dz(order+1, lh);
	      
	elmat = 0.0;

	NgProfiler::StopTimer (104);	
	NgProfiler::StartTimer (105);
	
	for (int kxy = 0; kxy < ndof2d; kxy++)
	  for (int kxy2 = 0; kxy2 < ndof2d; kxy2++)
	    {
	      // get coefficients in z-direction
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		coefz(iz) = mat2d[iz](kxy,kxy2);

	      /*
		// p^3 algorithm for 1D matrix
	      FlatMatrix<> base1( jaczbTs[powerz[kxy]] );
	      FlatMatrix<> base2( jaczbTs[powerz[kxy2]] );

	      for (int i = 0; i < order+1 - powerz[kxy]; i++)
		{
		  int mi = prism2tet[kxy*(order+1)+i];
		  if (mi != -1)
		    {
		      for (int k = 0; k < ir.GetNIPZ(); k++)
			coefzb1(k) = base1(i,k) * coefz(k);

		      for (int j = 0; j < order+1 - powerz[kxy2]; j++)
			{
			  int mj = prism2tet[kxy2*(order+1)+j];
			  if (mj != -1)
			    {
			      double sum = 0;
			      for (int k = 0; k < ir.GetNIPZ(); k++)
				sum += coefzb1(k) * base2(j,k);
			      elmat(mi, mj) += sum;
			    }
			}
		    }
		}
	      */

	      // (*testout) << "basis " << powerz[kxy] << " - " <<  powerz[kxy] << endl;

	      /*
	      // the un-optimized version:
	      mat1dz = 0;
	      for (int k = 0; k < ir.GetNIPZ(); k++)
		for (int i = 0; i < order+1; i++)
		  for (int j = 0; j < order+1; j++)
		    mat1dz(i,j) += coefz(k) * base1(i,k) * base2(j,k);
	      (*testout) << "mat1d = " << endl << mat1dz << endl;
	      */



	      for (int k = 0; k < ir.GetNIPZ(); k++)
		coefz(k) *= powzm(k, powerz[kxy]+powerz[kxy2]);
	      
	      int h = order+1-powerz[kxy];
	      int w = order+1-powerz[kxy2];

	      FlatMatrix<> hmat1dz (h, w, &mat1dz(0,0));
	      hmat1dz = 0.0;


	      hmat1dz = 0.0;
 	      GenerateMatrix (*generic_jaczb[powerz[kxy]],  generic_jaczb_vals[powerz[kxy]],
 			      *generic_jaczb[powerz[kxy2]], generic_jaczb_vals[powerz[kxy2]],
 			      coefz, hmat1dz);

	      if (powerz[kxy2] == 0)
		for (int i = 0; i < h; i++)
		  hmat1dz(i,0) -= hmat1dz(i,1);

	      if (powerz[kxy] == 0)
		for (int i = 0; i < w; i++)
		  hmat1dz(0, i) -= hmat1dz(1,i);

	      // (*testout) << "generic2, mat1d = " << endl << hmat1dz << endl;

	      for (int i = 0; i < h; i++)
		{
		  int teti = prism2tet[kxy*(order+1)+i];
		  for (int j = 0; j < w; j++)
		    {
		      int tetj = prism2tet[kxy2*(order+1)+j];
		      elmat(teti, tetj) += hmat1dz(i,j);
		    }
		}
	    }

	NgProfiler::StopTimer (105);


	if (checkfast)
	  {
	    double sum = 0;
	    for (int i = 0; i < mat1.Height()*mat1.Width(); i++)
	      sum += sqr (mat1(i)-elmat(i));
	    cout << "diff between std and fast: " << sqrt (sum) << endl;
	    (*testout) << "old mat " << endl << mat1 << endl;
	    (*testout) << "new mat " << endl << elmat << endl;
	    (*testout) << "diff mat " << endl << elmat-mat1 << endl;
	  }

	for (int i = 0; i < generic_jaczb.Size(); i++)
	  delete generic_jaczb[i];

	NgProfiler::StopTimer (maintimer);
      }
    catch (...)
      {
	orig -> AssembleElementMatrix (bfel, eltrans, elmat, lh); 	
	return;
      }
  }

  
  namespace {
    class Init
    { 
    public: 
      Init ();
    };        
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("fastmass", 3, 1,
					FastMassIntegratorTet::Create);
    }


    Init init;
    
  }
}



#endif
