#include <comp.hpp>

#include <multigrid.hpp>
using namespace ngmg;

#include <solve.hpp>



namespace ngcomp
{
  using namespace ngcomp;
  

  class RefElementASMMatrix : public BaseMatrix
  {
    const H1HighOrderFESpace & fes;
    const BilinearForm & bfa;
    const SparseMatrixSymmetric<double> & mat;
    Vector<double> diag;

    Matrix<> inv_element;
    Matrix<> inv_face;
    Matrix<> inv_edge;
    SparseCholesky<double> * sparse_inv_element;

    ARRAY<Matrix<>*> extensions;
    ARRAY<Matrix<>*> edge2face_extensions;
    ARRAY<Matrix<>*> edge2element_extensions;

    ARRAY<SparseMatrix<double>*> sparse_extensions;
    ARRAY<Matrix<>*> sparse_edge2element_extensions;


    
    BaseMatrix * coarse_inverse;
    int ncoarse;

    ARRAY<int> n_elements_on_edge;
    ARRAY<int> n_elements_on_face;

  public:
    RefElementASMMatrix (const H1HighOrderFESpace & afes,
			 const BilinearForm & abfa,
			 const SparseMatrixSymmetric<double> & amat)
      : fes(afes), bfa(abfa), mat(amat), diag(amat.Height()), 
	extensions(4), edge2face_extensions(3), edge2element_extensions(6),
	sparse_extensions(4), sparse_edge2element_extensions(6)
    {
      static int maintimer = NgProfiler::CreateTimer ("RefElementASM::Setup");
      NgProfiler::RegionTimer reg(maintimer);


      const MeshAccess & ma = fes.GetMeshAccess ();

      int h = mat.Height();
      for (int i = 0; i < h; i++)
	diag(i) = 1.0 / mat(i,i);


      LocalHeap lh(300000000);


      int p = fes.GetOrder();

      //      const FiniteElement & tet = fes.GetFE(0, lh);

      H1HighOrderTetFast tet(fes.GetOrder());
      ArrayMem<int,12> vnums; 
      ma.GetElVertices(0, vnums);
      tet.SetVertexNumbers (vnums);


      int n = tet.GetNDof();

      ConstantCoefficientFunction coef(1);
      FastLaplaceIntegrator<3> laplace(&coef);

      ElementTransformation eltrans;
      fes.GetMeshAccess().GetElementTransformation (0, eltrans, lh);

      cout << "prepare precond" << endl;

      HashTable<INT<2>, double> hash_elmat(100*tet.GetNDof());
      laplace.AssembleElementMatrix (tet, eltrans, hash_elmat, lh);

      cout << "has hash elmat" << endl;

      


      int augmented = 0; // fes.GetAugmented();
      int n_vert = 4;
      if (augmented == 1) n_vert = 8;
      int n_edge = (p-1);
      int n_inner = (p-1)*(p-2)*(p-3) / 6;
      int n_face = (p-1)*(p-2)/2;

      int first_edge = n_vert;
      int first_face = first_edge+6*n_edge;
      int first_inner = first_face+4*n_face;








      ARRAY<int> cntnz(n_inner);
      cntnz = 0;
      int cnt_all = 0;

      for (int i = 0; i < hash_elmat.Size(); i++)
	for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	  {
	    INT<2> i2;
	    double val;
	    hash_elmat.GetData (i, j, i2, val);
	    if (i2[0] >= first_inner && i2[1] >= first_inner)
	      { cntnz[i2[0]-first_inner]++; cnt_all++; }
	  }

      /*
      for (int i = 0; i < n_inner; i++)
	for (int j = 0; j <= i; j++)
	  if (element_inner(i,j) != 0.0) 
	    { cntnz[i]++; cnt_all++; }
      */

      cout << "n = " << n << ", ni = " << n_inner << ", nb = " << n-n_inner << endl;
      cout << 100.0 * cnt_all / (n_inner * n_inner) << " % non-zero inner" << endl;
      

      SparseMatrixSymmetric<double> sparse_element_inner (cntnz);

      for (int i = 0; i < hash_elmat.Size(); i++)
	for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	  {
	    INT<2> i2;
	    double val;
	    hash_elmat.GetData (i, j, i2, val);
	    if (i2[0] >= first_inner && i2[1] >= first_inner)
	      sparse_element_inner.CreatePosition(i2[0]-first_inner,
						  i2[1]-first_inner);
	  }
      /*
	for (int i = 0; i < n_inner; i++)
	for (int j = 0; j <= i; j++)
	  if (element_inner(i,j) != 0.0) 
	  sparse_element_inner.CreatePosition(i,j);
      */
      
      
      for (int i = 0; i < hash_elmat.Size(); i++)
	for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	  {
	    INT<2> i2;
	    double val;
	    hash_elmat.GetData (i, j, i2, val);
	    if (i2[0] >= first_inner && i2[1] >= first_inner)
	      sparse_element_inner(i2[0]-first_inner, i2[1]-first_inner) = val;
	  }

      /*
      for (int i = 0; i < n_inner; i++)
	for (int j = 0; j <= i; j++)
	  if (element_inner(i,j) != 0.0) 
	    sparse_element_inner(i,j) = element_inner(i,j);
      */

      cout << "compute sparse inverse" << endl;
      sparse_inv_element = new SparseCholesky<double> (sparse_element_inner);
      cout << "done" << endl;





      int sort[4] = { 0, 1, 2, 3 };
      for (int i = 0; i < 4; i++)
	for (int j = 0; j < 3; j++)
	  if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
      
      int isort[4];
      for (int i = 0; i < 4; i++) isort[sort[i]] = i;


















      Matrix<> el_if(n_inner, n_face);
      Matrix<> schur_face(n_face);
      VVector<> hv1(n_inner), hv2(n_inner);

      for (int fi = 0; fi < 4; fi++)
	{
	  extensions[isort[fi]] = new Matrix<> (n_inner, n_face);
	  
	  cout << "face " << flush;

	  int first_col = first_face+fi*n_face;

	  /*
	  for (int i = 0; i < n_inner; i++)
	    for (int j = 0; j < n_face; j++)
	      el_if(i, j) = elmat_lap(first_inner+i, first_col+j);
	  */

	  el_if = 0.0;
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] >= first_col && i2[1] < first_col+n_face)
		  el_if(i2[0]-first_inner, i2[1]-first_col) = val;
	      }


	  /*
	  sparse_extensions[isort[fi]] = new Matrix<> (n_inner, n_face);
	  *sparse_extensions[isort[fi]] = -el_if;
	  */

	  


	  ARRAY<int> cntnz(n_inner);
	  cntnz = 0;
	  int cntall = 0;
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] >= first_col && i2[1] < first_col+n_face)
		  {
		    cntnz[i2[0]-first_inner]++; 
		    cntall++;
		  }
	      }

	  cout << 100.0 * cntall / (n_inner * n_face) << " % non-zero " << flush;
	  sparse_extensions[isort[fi]] = new SparseMatrix<double> (cntnz);
	  SparseMatrix<double> & sparseext = *sparse_extensions[isort[fi]];

	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] >= first_col && i2[1] < first_col+n_face)
		  sparseext.CreatePosition(i2[0]-first_inner, i2[1]-first_col);
	      }
	  
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] >= first_col && i2[1] < first_col+n_face)
		  sparseext(i2[0]-first_inner, i2[1]-first_col) = -val;
	      }
	  
	  if (fi != 3) continue;

	  // *extensions[isort[fi]] = -(inv_element * el_if);
	  Matrix<> & ext = *extensions[isort[fi]];
	  for (int i = 0; i < n_face; i++)
	    {
	      for (int j = 0; j < n_inner; j++)
		hv1(j) = el_if(j,i);
	      hv2 = (*sparse_inv_element) * hv1;
	      for (int j = 0; j < n_inner; j++)
		ext(j,i) = -hv2(j);
	    }


	  schur_face = Trans (el_if) * (*extensions[isort[fi]]);
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_col && i2[0] < first_col+n_face && 
		    i2[1] >= first_col && i2[1] < first_col+n_face)
		  {
		    schur_face(i2[0]-first_col, i2[1]-first_col) += val;
		    if (i2[0] != i2[1])
		      schur_face(i2[1]-first_col, i2[0]-first_col) += val;
		  }
	      }
	  /*
	  for (int i = 0; i < n_face; i++)
	    for (int j = 0; j < n_face; j++)
	      schur_face(i, j) += elmat_lap(first_face+fi*n_face+i, first_face+fi*n_face+j);
	  */
	}

      cout << "compute face inv " << flush;

      inv_face.SetSize (n_face);
      CalcInverse (schur_face, inv_face);

      // (*testout) << "schur = " << endl << schur_face << endl;
      // (*testout) << "inv_face = " << endl << inv_face << endl;

      cout << "done" << endl;










      const EDGE * edges = ElementTopology::GetEdges (ET_TET);
      // prepare edge2element
      Matrix<> el_ei(n_inner, n_edge);
      for (int ei = 0; ei < 6; ei++)
	{
	  int vn1 = isort[edges[ei][0]], vn2 = isort[edges[ei][1]];
	  int enr = ElementTopology::GetEdgeNr (ET_TET, vn1, vn2);

	  edge2element_extensions[enr] = new Matrix<> (n_inner, n_edge);
	  sparse_edge2element_extensions[enr] = new Matrix<> (n_inner, n_edge);

	  /*	  
	  for (int i = 0; i < n_inner; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_ei(i, j) = elmat_lap(first_inner+i, first_edge+ei*n_edge+j);
	  */

	  int first_col = first_edge+ei*n_edge;

	  el_ei = 0.0;
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] >= first_col && i2[1] < first_col+n_edge)
		  el_ei(i2[0]-first_inner, i2[1]-first_col) = val;
	      }

	  *sparse_edge2element_extensions[enr] = -el_ei;


	  // *edge2element_extensions[enr] = (-inv_element) * el_ei;
	  Matrix<> & ext = *edge2element_extensions[enr];
	  for (int i = 0; i < n_edge; i++)
	    {
	      for (int j = 0; j < n_inner; j++)
		hv1(j) = el_ei(j,i);
	      hv2 = (*sparse_inv_element) * hv1;
	      for (int j = 0; j < n_inner; j++)
		ext(j,i) = -hv2(j);
	    }
	}


      // (*testout) << "schur = " << endl << schur_face << endl;
      // (*testout) << "inv_face = " << endl << inv_face << endl;


      cout << "has edge2element" << endl;

     


      /*

      // FlatMatrix<double> elmat_lap;
      // laplace.AssembleElementMatrix (tet, eltrans, elmat_lap, lh);
      Matrix<double> elmat_lap(n);
      elmat_lap = 0.0;
      for (int i = 0; i < hash_elmat.Size(); i++)
	for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	  {
	    INT<2> i2;
	    double val;
	    hash_elmat.GetData (i, j, i2, val);
	    elmat_lap(i2[0], i2[1]) = val;
	    elmat_lap(i2[1], i2[0]) = val;
	  }
      
      cout << "has dense elmat" << endl;

      Matrix<> element_inner(n_inner);
      inv_element.SetSize (n_inner);

      for (int i = 0; i < n_inner; i++)
	for (int j = 0; j < n_inner; j++)
	  element_inner(i,j) = elmat_lap(first_inner+i, first_inner+j);

      // CalcInverse (element_inner, inv_element);
      inv_element = element_inner;
      if (p >= 4)
	LapackInverse (inv_element);
      
      cout << "has inverse" << endl;
      */






      Matrix<> schur_bound(first_inner);
      schur_bound = 0.0;
      for (int k = 0; k < first_inner; k++)
	{
	  if (k % 100 == 0) cout << "." << flush;

	  hv1 = 0.0;
	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] == k)
		  hv1(i2[0]-first_inner) += val;
	      }

	  hv2 = (*sparse_inv_element) * hv1;

	  for (int i = 0; i < hash_elmat.Size(); i++)
	    for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	      {
		INT<2> i2;
		double val;
		hash_elmat.GetData (i, j, i2, val);
		if (i2[0] >= first_inner && i2[1] < first_inner)
		  schur_bound(k, i2[1]) -= val * hv2(i2[0]-first_inner);
	      }
	}
      for (int i = 0; i < hash_elmat.Size(); i++)
	for (int j = 0; j < hash_elmat.EntrySize(i); j++)
	  {
	    INT<2> i2;
	    double val;
	    hash_elmat.GetData (i, j, i2, val);
	    if (i2[0] < first_inner && i2[1] < first_inner)
	      {
		schur_bound(i2[0], i2[1]) += val;
		if (i2[0] != i2[1])
		  schur_bound(i2[1], i2[0]) += val;
	      }
	  }
      






      BitArray used(first_inner);
      used.Clear();
      int e2fsort[3];

      const FACE * faces = ElementTopology::GetFaces (ET_TET);
      // first face is 1,2,3

      int sortf[3] = { 1, 2, 3, };
      for (int i = 0; i < 3; i++)
	for (int j = 0; j < 2; j++)
	  if (vnums[sortf[j]] > vnums[sortf[j+1]]) swap (sortf[j], sortf[j+1]);
      

      for (int i = 0; i < n_face; i++)
	used.Set(first_face+i);

      int cnt = 0;
      for (int i = 0; i < 6; i++)
	if (edges[i][0] != 0 && edges[i][1] != 0)
	  {
	    for (int j = 0; j < n_edge; j++)
	      used.Set(first_edge+n_edge*i+j);

	    for (int j = 0; j < 3; j++)
	      if (edges[i][0] != sortf[j] && edges[i][1] != sortf[j])
		e2fsort[cnt++] = j;

	  }
	else
	  for (int j = 0; j < n_edge; j++)
	    schur_bound(first_edge+n_edge*i+j,first_edge+n_edge*i+j) += 1e26;
	  

      for (int i = 0; i < first_edge; i++)
      	schur_bound(i,i) += 1e26;
      // for (int i = first_face+n_face; i <= first_inner; i++)
      //      	elmat_lap(i,i) += 1e26;

      cout << "call schur" << endl;
      
      Matrix<> faceedges(n_face+3*n_edge);
      CalcSchurComplement (schur_bound, faceedges, used, lh);

      cout << "schur done" << endl;

      // (*testout) << "faceedge-schur = " << endl << faceedges << endl;

      Matrix<> el_fe(n_face, n_edge);
      Matrix<> inv_face(n_face);
      Matrix<> schur_edge(n_edge);


      for (int i = 0; i < n_face; i++)
	for (int j = 0; j < n_face; j++)
	  inv_face(i,j) = faceedges(3*n_edge+i, 3*n_edge+j);
      LapackInverse (inv_face);      

      for (int ei = 0; ei < 3; ei++)
	{
	  edge2face_extensions[e2fsort[ei]] = new Matrix<> (n_face, n_edge);
	  
	  for (int i = 0; i < n_face; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_fe(i, j) = faceedges(3*n_edge+i, ei*n_edge+j);

	  *edge2face_extensions[e2fsort[ei]] = (-inv_face) * el_fe;
	  schur_edge = Trans (el_fe) * (*edge2face_extensions[e2fsort[ei]]);

	  for (int i = 0; i < n_edge; i++)
	    for (int j = 0; j < n_edge; j++)
	      schur_edge(i, j) += faceedges(ei*n_edge+i, ei*n_edge+j);
	  
	  cout << "has edge matrix " << ei << endl;
	  // (*testout) << "e2f-extension = " << endl << *edge2face_extensions[e2fsort[ei]] << endl;
	  // (*testout) << "schur_edge = " << schur_edge << endl;
	}
      
      inv_edge.SetSize (n_edge);
      CalcInverse (schur_edge, inv_edge);

      //      (*testout) << "inv_edge = " << endl << inv_edge << endl;

      cout << "4" << endl;












#ifdef ABC
      // precompute edge2face extensions

      BitArray used(n);
      used.Clear();
      int e2fsort[3];

      const FACE * faces = ElementTopology::GetFaces (ET_TET);
      // first face is 1,2,3

      int sortf[3] = { 1, 2, 3, };
      for (int i = 0; i < 3; i++)
	for (int j = 0; j < 2; j++)
	  if (vnums[sortf[j]] > vnums[sortf[j+1]]) swap (sortf[j], sortf[j+1]);
      

      for (int i = 0; i < n_face; i++)
	used.Set(first_face+i);

      int cnt = 0;
      for (int i = 0; i < 6; i++)
	if (edges[i][0] != 0 && edges[i][1] != 0)
	  {
	    for (int j = 0; j < n_edge; j++)
	      used.Set(first_edge+n_edge*i+j);

	    for (int j = 0; j < 3; j++)
	      if (edges[i][0] != sortf[j] && edges[i][1] != sortf[j])
		e2fsort[cnt++] = j;

	  }
	else
	  for (int j = 0; j < n_edge; j++)
	    elmat_lap(first_edge+n_edge*i+j,first_edge+n_edge*i+j) += 1e26;
	  

      for (int i = 0; i < first_edge; i++)
      	elmat_lap(i,i) += 1e26;
      // for (int i = first_face+n_face; i <= first_inner; i++)
      //      	elmat_lap(i,i) += 1e26;

      cout << "call schur" << endl;
      
      Matrix<> faceedges(n_face+3*n_edge);
      CalcSchurComplement (elmat_lap, faceedges, used, lh);

      cout << "schur done" << endl;

      // (*testout) << "faceedge-schur = " << endl << faceedges << endl;

      Matrix<> el_fe(n_face, n_edge);
      Matrix<> inv_face(n_face);
      Matrix<> schur_edge(n_edge);


      for (int i = 0; i < n_face; i++)
	for (int j = 0; j < n_face; j++)
	  inv_face(i,j) = faceedges(3*n_edge+i, 3*n_edge+j);
      LapackInverse (inv_face);      

      for (int ei = 0; ei < 3; ei++)
	{
	  edge2face_extensions[e2fsort[ei]] = new Matrix<> (n_face, n_edge);
	  
	  for (int i = 0; i < n_face; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_fe(i, j) = faceedges(3*n_edge+i, ei*n_edge+j);

	  *edge2face_extensions[e2fsort[ei]] = (-inv_face) * el_fe;
	  schur_edge = Trans (el_fe) * (*edge2face_extensions[e2fsort[ei]]);

	  for (int i = 0; i < n_edge; i++)
	    for (int j = 0; j < n_edge; j++)
	      schur_edge(i, j) += faceedges(ei*n_edge+i, ei*n_edge+j);
	  
	  cout << "has edge matrix " << ei << endl;
	  // (*testout) << "e2f-extension = " << endl << *edge2face_extensions[e2fsort[ei]] << endl;
	  // (*testout) << "schur_edge = " << schur_edge << endl;
	}
      
      inv_edge.SetSize (n_edge);
      CalcInverse (schur_edge, inv_edge);

      //      (*testout) << "inv_edge = " << endl << inv_edge << endl;

      cout << "4" << endl;
#endif





      /*
      coarse_inverse = 
	dynamic_cast<const BaseSparseMatrix&> 
	(bfa.GetLowOrderBilinearForm().GetMatrix())  . InverseMatrix();
      */

      coarse_inverse = 
	dynamic_cast<const BaseSparseMatrix&> 
	(bfa.GetMatrix())  . InverseMatrix();
      
      ncoarse = coarse_inverse->Height();
      // (*testout) << "coarse mat = " << endl << (*coarse_inverse) << endl;

      cout << "5" << endl;


      ARRAY<int> enums, fnums;

      int ne = ma.GetNE();
      int nfa = ma.GetNFaces();
      int ned = ma.GetNEdges();
      
      n_elements_on_face.SetSize (nfa);
      n_elements_on_face = 0;
      n_elements_on_edge.SetSize (nfa);
      n_elements_on_edge = 0;

      for (int i = 0; i < ne; i++)
	{
	  ma.GetElFaces (i, fnums);
	  for (int j = 0; j < fnums.Size(); j++)
	    n_elements_on_face[fnums[j]]++;
	  ma.GetElEdges (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    n_elements_on_edge[enums[j]]++;
	}
    }


    virtual int VHeight() const { return fes.GetNDof(); }
    virtual int VWidth() const { return fes.GetNDof(); }

    virtual BaseVector * CreateVector () const
    {
      return new VVector<double> (fes.GetNDof());
    }




    virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const
    {
      static int maintimer = NgProfiler::CreateTimer ("RefElementASM::Mult");
      NgProfiler::RegionTimer reg(maintimer);

      FlatVector<double> vx = x.FVDouble();
      FlatVector<double> vy = y.FVDouble();

      Vector<> hv(vx.Size());
      Vector<> hv2(vx.Size());
      


      const MeshAccess & ma = fes.GetMeshAccess();
      ARRAY<int> vnums, enums, fnums;

      int ne = ma.GetNE();
      int nfa = ma.GetNFaces();
      int ned = ma.GetNEdges();

      static int celltimer = NgProfiler::CreateTimer ("RefElementASM::Mult, cells");
      {
	NgProfiler::RegionTimer reg(celltimer);
	hv = 0;
	for (int i = 0; i < ne; i++)
	  {
	    int first = fes.GetFirstElementDof(i);
	    int next = fes.GetFirstElementDof(i+1);
	    
	    // FlatVector<double> elx  (next-first, &vx[first]);
	    //	    FlatVector<double> ely  (next-first, &vy[first]);
	    
	    // ely += s * (inv_element * elx);

	    VFlatVector<double> elx  (next-first, &vx[first]);
	    VFlatVector<double> ely  (next-first, &hv[first]);
	    ely = *sparse_inv_element * elx;
	  }
	y.FVDouble() += s * hv;
      }
      



      // Restriction

      static int resttimer = NgProfiler::CreateTimer ("RefElementASM::Mult, Restriction");
      {
      NgProfiler::RegionTimer reg(resttimer);



      int firste = fes.GetFirstEdgeDof(0);
      int firstf = fes.GetFirstFaceDof(0);
      int firsti = fes.GetFirstElementDof(0);
      int h = vx.Size();

      //      for (int i = firstf; i < h; i++)
      //	hv(i) = vx(i);

      hv = vx;

      for (int i = 0; i < ne; i++)
	{
	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  VFlatVector<double> elx  (next-first, &vx[first]);
	  VFlatVector<double> ely  (next-first, &hv[first]);
	  ely = *sparse_inv_element * elx;
	}


      for (int i = 0; i < ne; i++)
	{
	  ma.GetElVertices(i, vnums);
	  ma.GetElFaces(i, fnums);

	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  VFlatVector<double> elx  (next-first, &hv[first]);


	  int sort[4] = { 0, 1, 2, 3 };
	  for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 3; j++)
	      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	  
	  int isort[4];
	  for (int i = 0; i < 4; i++) isort[sort[i]] = i;
	  


	  for (int fi = 0; fi < 4; fi++)
	    {
	      int first_f = fes.GetFirstFaceDof(fnums[fi]);
	      int next_f = fes.GetFirstFaceDof(fnums[fi]+1);

	      VFlatVector<double> face_x  (next_f-first_f, &hv[first_f]);

	      face_x += Transpose (*sparse_extensions[isort[fi]])  * elx;
	    }
	}




     // edge 2 element restriction
      for (int i = 0; i < ne; i++)
	{
	  ma.GetElVertices(i, vnums);
	  ma.GetElEdges(i, enums);

	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  FlatVector<double> elx  (next-first, &hv[first]);


	  int sort[4] = { 0, 1, 2, 3 };
	  for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 3; j++)
	      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	  
	  int isort[4];
	  for (int i = 0; i < 4; i++) isort[sort[i]] = i;
	  
	  
	  const EDGE * edges = ElementTopology::GetEdges (ET_TET);
	  for (int ei = 0; ei < 6; ei++)
	    {
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv[first_e]);

	      int v1 = isort[edges[ei][0]], v2 = isort[edges[ei][1]];
	      int enr = ElementTopology::GetEdgeNr (ET_TET, v1, v2);

	      // elx += *edge2element_extensions[enr]  * face_x;
	      edge_x += Trans (*sparse_edge2element_extensions[enr]) * elx;
	    }
	}




      // edge 2 face restriction
      for (int i = 0; i < nfa; i++)
	{
	  ma.GetFacePNums(i, vnums);
	  ma.GetFaceEdges(i, enums);

	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);
	  FlatVector<double> elx  (next-first, &hv[first]);

	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (vnums[k] > vnums[k+1]) swap (vnums[k], vnums[k+1]);
	  
	  // (*testout) << "face extension, face = " << i << ", vnums = " << vnums << endl;

	  for (int ei = 0; ei < 3; ei++)
	    {
	      int ev1, ev2;
	      ma.GetEdgePNums (enums[ei], ev1, ev2);

	      // (*testout) << "has edge " << enums[ei] << " with vertices " << ev1 << " and " << ev2 << endl;
	      
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv[first_e]);

	      for (int j = 0; j < 3; j++)
		if (ev1 != vnums[j] && ev2 != vnums[j])
		  // elx += *edge2face_extensions[j] * edge_x;
		  edge_x += Trans (*edge2face_extensions[j]) * elx;
	    }
	}
      }







      hv2 = 0;
      for (int i = 0; i < nfa; i++)
	{
	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);

	  if (next-first == inv_face.Height())
	    {
	      FlatVector<double> elx  (next-first, &hv[first]);
	      FlatVector<double> ely  (next-first, &hv2[first]);
	      
	      ely = inv_face * elx;
	      ely *= 1.0 / n_elements_on_face[i];
	    }
	}

      for (int i = 0; i < ned; i++)
	{
	  int first = fes.GetFirstEdgeDof(i);
	  int next = fes.GetFirstEdgeDof(i+1);

	  if (next-first == inv_edge.Height())
	    {
	      FlatVector<double> elx  (next-first, &hv[first]);
	      FlatVector<double> ely  (next-first, &hv2[first]);
	      
	      ely = inv_edge * elx;
	      ely *= 1.0 / n_elements_on_edge[i];
	    }
	}



      // edge 2 face extension
      for (int i = 0; i < nfa; i++)
	{
	  ma.GetFacePNums(i, vnums);
	  ma.GetFaceEdges(i, enums);

	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);
	  FlatVector<double> elx  (next-first, &hv2[first]);

	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (vnums[k] > vnums[k+1]) swap (vnums[k], vnums[k+1]);
	  
	  for (int ei = 0; ei < 3; ei++)
	    {
	      int ev1, ev2;
	      ma.GetEdgePNums (enums[ei], ev1, ev2);

	      
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv2[first_e]);

	      for (int j = 0; j < 3; j++)
		if (ev1 != vnums[j] && ev2 != vnums[j])
		  elx += *edge2face_extensions[j] * edge_x;
	    }
	}


      // edge 2 element extension
      for (int i = 0; i < ne; i++)
	{
	  ma.GetElVertices(i, vnums);
	  ma.GetElEdges(i, enums);

	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  FlatVector<double> elx  (next-first, &hv2[first]);


	  int sort[4] = { 0, 1, 2, 3 };
	  for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 3; j++)
	      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	  
	  int isort[4];
	  for (int i = 0; i < 4; i++) isort[sort[i]] = i;
	  
	  
	  const EDGE * edges = ElementTopology::GetEdges (ET_TET);
	  for (int ei = 0; ei < 6; ei++)
	    {
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv2[first_e]);

	      int v1 = isort[edges[ei][0]], v2 = isort[edges[ei][1]];
	      int enr = ElementTopology::GetEdgeNr (ET_TET, v1, v2);

	      elx += *sparse_edge2element_extensions[enr]  * edge_x;
	    }
	}




      // face 2 element extension
      for (int i = 0; i < ne; i++)
	{
	  ma.GetElVertices(i, vnums);
	  ma.GetElFaces(i, fnums);

	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  VFlatVector<double> elx  (next-first, &hv2[first]);


	  int sort[4] = { 0, 1, 2, 3 };
	  for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 3; j++)
	      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	  
	  int isort[4];
	  for (int i = 0; i < 4; i++) isort[sort[i]] = i;
	  


	  for (int fi = 0; fi < 4; fi++)
	    {
	      int first_f = fes.GetFirstFaceDof(fnums[fi]);
	      int next_f = fes.GetFirstFaceDof(fnums[fi]+1);

	      VFlatVector<double> face_x  (next_f-first_f, &hv2[first_f]);

	      elx += *sparse_extensions[isort[fi]]  * face_x;
	    }
	}


      hv = hv2;
      for (int i = 0; i < ne; i++)
	{
	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  VFlatVector<double> elx  (next-first, &hv2[first]);
	  VFlatVector<double> ely  (next-first, &hv[first]);
	  ely = *sparse_inv_element * elx;
	}
      

      vy += s * hv;



      //      for (int i = 0; i < firste; i++)
      //	vy(i) += s * diag(i) * vx(i);


      VFlatVector<> coarse_x (ncoarse, &vx(0));
      VFlatVector<> coarse_hv (ncoarse, &hv(0));
      VFlatVector<> coarse_y (ncoarse, &vy(0));

      coarse_hv = (*coarse_inverse) * coarse_x;
      coarse_y += s * coarse_hv;
    }


    virtual ostream & Print (ostream & ost) const
    {
      ost << "RefElementASMMatrix:" << endl;
      return ost;
    }
  };









  
  
  
  class RefElementPreconditioner : public Preconditioner
  {
  private:
    const BilinearForm * bfa;
    const H1HighOrderFESpace * fes;
    BaseMatrix * inverse;
  public:

    RefElementPreconditioner (const PDE & pde, const Flags & flags)
      : Preconditioner(flags)
    {
      bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", NULL));
      fes = dynamic_cast<const H1HighOrderFESpace*> (pde.GetFESpace (flags.GetStringFlag ("fespace", NULL)));
      inverse = NULL;
    }
    
    virtual ~RefElementPreconditioner()
    {
      delete inverse;
    }
    
    static Preconditioner * Create (const PDE & pde, const Flags & flags)
    {
      return new RefElementPreconditioner (pde, flags);
    }

    virtual void Update ()
    {
      delete inverse;
      inverse = new RefElementASMMatrix (*fes, *bfa,
      					 dynamic_cast<const SparseMatrixSymmetric<double>&>(bfa->GetMatrix()));
      if (test) Test();
    }

    virtual const BaseMatrix & GetMatrix() const
    {
      return *inverse;
    }

    virtual const BaseMatrix & GetAMatrix() const
    {
      return bfa->GetMatrix(); 
    }
    
    virtual const char * ClassName() const
    { 
      return "ReferenceElement Preconditioner"; 
    }
  };


  



  
  // register FESpaces
  namespace {
    
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetPreconditionerClasses().AddPreconditioner ("refelementasm", RefElementPreconditioner::Create);
    }
    
    Init init;
  }
  
}
