#ifndef FILE_TOPOLOGY
#define FILE_TOPOLOGY

/*********************************************************************/
/* File:   topology.hpp                                              */
/* Author: Joachim Schoeberl                                         */
/* Date:   25. Mar. 2000                                             */
/*********************************************************************/

/*
    Toplogy of reference elements
*/


/**
   Geometry of element.
   Choices are ET_SEGM, ET_TRIG, ET_QUAD, ET_TET, ET_PYRAMID, ET_PRISM, ET_HEX
 */
enum ELEMENT_TYPE { ET_SEGM = 1,
		    ET_TRIG = 10, ET_QUAD = 11, 
		    ET_TET = 20, ET_PYRAMID = 21, ET_PRISM = 22, ET_HEX = 24 };

/// point coordinates
typedef double POINT3D[3];  

/// initial point, end point
typedef int EDGE[2];      

/// points, last one is -1 for trig
typedef int FACE[4];      

/// normal vector
typedef double NORMAL[3];


/// Topology and coordinate information of master element:
class ElementTopology
{
public:
  /// returns name of element type
  static const char * GetElementName (ELEMENT_TYPE et);

  /// returns space dimension of element type
  static int GetSpaceDim (ELEMENT_TYPE et)
  {
    switch (et)
      {
      case ET_SEGM: return 1;
      case ET_TRIG: return 2;
      case ET_QUAD: return 2;
      case ET_TET: return 3;
      case ET_PYRAMID: return 3;
      case ET_PRISM: return 3;
      case ET_HEX: return 3;
      }
    return 0;
  }

  /// returns number of vertices
  static int GetNVertices (ELEMENT_TYPE et)
  {
    switch (et)
      {
      case ET_SEGM: return 2;
      case ET_TRIG: return 3;
      case ET_QUAD: return 4;
      case ET_TET: return 4;
      case ET_PYRAMID: return 5;
      case ET_PRISM: return 6;
      case ET_HEX: return 8;
      }
    return 0;
  }

  /// returns number of edges
  static int GetNEdges (ELEMENT_TYPE et)
  { 
    switch (et)
      {
      case ET_SEGM: return 1;
      case ET_TRIG: return 3;
      case ET_QUAD: return 4;
      case ET_TET: return 6;
      case ET_PYRAMID: return 8;
      case ET_PRISM: return 9;
      case ET_HEX: return 12;
      }
    return 0;
  }


  /// returns number of faces
  static int GetNFaces (ELEMENT_TYPE et)
  {
    switch (et)
      {
      case ET_SEGM: return 0;
      case ET_TRIG: return 1;
      case ET_QUAD: return 1;
      case ET_TET: return 4;
      case ET_PYRAMID: return 5;
      case ET_PRISM: return 5;
      case ET_HEX: return 6;
      }  
    return 0;
  }

  /// returns number of facets: == GetNFaces in 3D, GetNEdges in 2D
  static int GetNFacets (ELEMENT_TYPE et)
  {
    switch (et)
    {
      case ET_SEGM: return 0;
      case ET_TRIG: return 3;
      case ET_QUAD: return 4;
      case ET_TET: return 4;
      case ET_PYRAMID: return 5;
      case ET_PRISM: return 5;
      case ET_HEX: return 6;
    }  
    return 0;
  }

  /// returns number of facets: == GetNFaces in 3D, GetNEdges in 2D
  static ELEMENT_TYPE GetFacetType (ELEMENT_TYPE et, int k)
  {
    switch (et)
    {
      case ET_TRIG: return ET_SEGM;
      case ET_QUAD: return ET_SEGM;
      case ET_TET: return ET_TRIG;
      case ET_PYRAMID: return (k<4 ? ET_TRIG : ET_QUAD); 
      case ET_PRISM: return (k<2 ? ET_TRIG : ET_QUAD);
      case ET_HEX: return ET_QUAD;
      default:
        cerr << "*** error in GetFacetType: Unhandled Elementtype" << endl;
        return ET_SEGM;
    }  
  }

  /// returns vertex coordinates (as 3D points)
  static const POINT3D * GetVertices (ELEMENT_TYPE et);

  /// returns edges of elements. zero-based pairs of integers
  static const EDGE * GetEdges (ELEMENT_TYPE et)
  {
    static int segm_edges[1][2] =
      { { 0, 1 }};
    

    static int trig_edges[3][2] =
      { { 2, 0 },
	{ 1, 2 },
	{ 0, 1 }};

    static int quad_edges[4][2] =
      { { 0, 1 },
	{ 2, 3 },
	{ 3, 0 },
	{ 1, 2 }};
    
    static int tet_edges[6][2] =
      { { 3, 0 },
	{ 3, 1 },
	{ 3, 2 }, 
	{ 0, 1 }, 
	{ 0, 2 },
	{ 1, 2 }};
    
    static int prism_edges[9][2] =
      { { 2, 0 },
	{ 0, 1 },
	{ 2, 1 },
	{ 5, 3 },
	{ 3, 4 },
	{ 5, 4 },
	{ 2, 5 },
	{ 0, 3 },
	{ 1, 4 }};

    static int pyramid_edges[8][2] =
      { { 0, 1 },
	{ 1, 2 },
	{ 0, 3 },
	{ 3, 2 },
	{ 0, 4 },
	{ 1, 4 },
	{ 2, 4 },
	{ 3, 4 }};

    static int hex_edges[12][2] =
      {
        { 0, 1 },
        { 2, 3 },
        { 3, 0 },
        { 1, 2 },
        { 4, 5 },
        { 6, 7 },
        { 7, 4 },
        { 5, 6 },
        { 0, 4 },
        { 1, 5 },
        { 2, 6 },
        { 3, 7 },
      };
    
    switch (et)
      {
      case ET_SEGM: return segm_edges;
      case ET_TRIG: return trig_edges;
      case ET_QUAD: return quad_edges;
      case ET_TET:  return tet_edges;
      case ET_PYRAMID: return pyramid_edges;
      case ET_PRISM: return prism_edges;
      case ET_HEX: return hex_edges;
      default:
	break;
      }
    cerr << "Ng_GetEdges, illegal element type " << et << endl;
    return 0;  
  }

  /// returns faces of elements. zero-based array of 4 integers, last one is -1 for triangles
  static const FACE * GetFaces (ELEMENT_TYPE et)
  {
    static int tet_faces[4][4] =
      { { 3, 1, 2, -1 },
	{ 3, 2, 0, -1 },
	{ 3, 0, 1, -1 },
        { 0, 2, 1, -1 } }; // all faces point into interior!
  
    static int prism_faces[5][4] =
      {
	{ 0, 2, 1, -1 },
	{ 3, 4, 5, -1 },
	{ 2, 0, 3, 5 },
	{ 0, 1, 4, 3 },
	{ 1, 2, 5, 4 } 
      };

    static int pyramid_faces[5][4] =
      {
	{ 0, 1, 4, -1 },
	{ 1, 2, 4, -1 },
	{ 2, 3, 4, -1 },
	{ 3, 0, 4, -1 },
        { 0, 1, 2, 3 } // points into interior!
      };
  
    static int hex_faces[6][4] =
      {
        { 0, 3, 2, 1 },
        { 4, 5, 6, 7 },
        { 0, 1, 5, 4 },
        { 1, 2, 6, 5 },
        { 2, 3, 7, 6 },
        { 3, 0, 4, 7 }
    };
    
    static int trig_faces[1][4] = 
      {
	{ 0, 1, 2, -1 },
      };
    
    static int quad_faces[1][4] = 
      {
	{ 0, 1, 2, 3 },
      };
    
    switch (et)
      {
      case ET_TET: return tet_faces;
      case ET_PRISM: return prism_faces;
      case ET_PYRAMID: return pyramid_faces;
      case ET_HEX: return hex_faces;

      case ET_TRIG: return trig_faces;
      case ET_QUAD: return quad_faces;
        
      case ET_SEGM:
      default:
	break;
      }
    
    cerr << "Ng_GetFaces, illegal element type " << et << endl;
    return 0;
  }

  /// return normals on facets
  static const NORMAL * GetNormals(ELEMENT_TYPE et);
  
  /// returns number of edge from vertex v1 to vertex v2
  static int GetEdgeNr (ELEMENT_TYPE et, int v1, int v2);

  /// returns number of face containing vertices v1, v2, v3.  (trig only ?)
  static int GetFaceNr (ELEMENT_TYPE et, int v1, int v2, int v3);
};


#endif
