/*********************************************************************/
/* File:   h1hofe.cpp                                                */
/* Author: Start                                                     */
/* Date:   1. June 2007                                              */
/*********************************************************************/


/*

High order integrators in tensor product form

*/

// #define NOPROFILE


#include <fem.hpp>
namespace ngfem
{
  using namespace ngfem;




  // ******************************** source integrator **************************************



  
  template <int D>
  class SourceIntegratorTP : public SourceIntegrator<D>
  {
  public:
    SourceIntegratorTP (CoefficientFunction * acoef)
      : SourceIntegrator<D> (acoef)
    { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new SourceIntegratorTP (coeffs[0]);
    }

    virtual void
    AssembleElementVector (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatVector<double> & elvec,
			   LocalHeap & lh) const
    {
      const H1HighOrderTetTP * fel_tet = dynamic_cast<const H1HighOrderTetTP*> (&bfel);      
      if (fel_tet)
        {
          TAssembleElementVector (*fel_tet, eltrans, elvec, lh);
          return;
        }

      const H1HighOrderPrismTP * fel_prism = dynamic_cast<const H1HighOrderPrismTP*> (&bfel);      
      if (fel_prism)
        if (fel_prism->Sort(0)+3 == fel_prism->Sort(3))  // same singular vertex at bottom and top
          {
            TAssembleElementVector (*fel_prism, eltrans, elvec, lh);
            return;
          }
      
      const H1HighOrderHexTP * fel_hex = dynamic_cast<const H1HighOrderHexTP*> (&bfel);      
      if (fel_hex)
        {
          TAssembleElementVector (*fel_hex, eltrans, elvec, lh);
          return;
        }

      SourceIntegrator<D>::AssembleElementVector(bfel, eltrans, elvec, lh);
    }




    template <class FEL>
    void TAssembleElementVector (const FEL & fel,
                                 const ElementTransformation & eltrans, 
                                 FlatVector<double> & elvec,
                                 LocalHeap & lh) const
    {
      int ndof = fel.GetNDof();
      
      elvec.AssignMemory (ndof, lh);
      elvec = 0;

      HeapReset hr (lh);

      IntegrationRuleTP irtp(fel, eltrans, 2*fel.Order(), lh);
          
      const IntegrationRule & irx = irtp.GetIRX();
      const IntegrationRule & iry = irtp.GetIRY();
      const IntegrationRule & irz = irtp.GetIRZ();
      
      int nipx = irx.GetNIP();
      int nipy = iry.GetNIP();
      int nipz = irz.GetNIP();
      int nip = nipx * nipy * nipz;
      
      FlatVector<Vec<1> > dvecs(nip, lh);
      
      for (int ii = 0; ii < nip; ii++)
        {
          IntegrationPoint ip(irtp.GetXi(ii), irtp.GetWeight(ii));
          SpecificIntegrationPoint<3,3> sip(ip, eltrans, irtp.GetPoint(ii), irtp.GetJacobian(ii), lh);
          
          (this->dvecop).GenerateVector (fel, sip, dvecs[ii], lh);
          dvecs[ii] *= fabs (sip.GetJacobiDet()) * ip.Weight();
        }
      
      FlatArray<int> map2dto1d = fel.Get2dTo1dMapping ();
      FlatArray<int> map3dto2d = fel.Get3dTo2dMapping ();
      
      int ndof1d = fel.GetNDof1d();
      int ndof2d = fel.GetNDof2d();
      
      FlatMatrix<> fac_x(ndof,   nipx, lh); 
      FlatMatrix<> fac_y(ndof2d, nipy, lh); 
      FlatMatrix<> fac_z(ndof1d, nipz, lh); 
      
      for (int i1 = 0; i1 < nipz; i1++)
        fel.CalcZFactor (irz[i1](0), fac_z.Col(i1), lh);

      for (int i1 = 0; i1 < nipy; i1++)
        fel.CalcYFactor (iry[i1](0), fac_y.Col(i1), lh);

      for (int i1 = 0; i1 < nipx; i1++)
        fel.CalcXFactor (irx[i1](0), fac_x.Col(i1), lh);

      FlatVector<> sum2d(ndof2d, lh);
      FlatVector<> sum1d(ndof1d, lh);
      
      for (int i1 = 0, ii = 0; i1 < nipx; i1++)
        {
          sum2d = 0;
        
          for (int i2 = 0; i2 < nipy; i2++)
            {
              sum1d = 0;
              for (int i3 = 0; i3 < nipz; i3++, ii++)
                {
                  for (int i = 0; i < ndof1d; i++)
                    sum1d(i) += dvecs[ii](0) * fac_z(i, i3);
                }
            
              for (int i = 0; i < ndof2d; i++)
                sum2d(i) += sum1d(map2dto1d[i]) * fac_y(i, i2);
            }
        
          for (int i = 0; i < ndof; i++)
            elvec(i) += sum2d(map3dto2d[i]) * fac_x(i, i1);
        }
    }
  };














  // ******************************** Laplace **************************************



  class TrafoGradientZ
  {
    const AutoDiff<1> & adz;
  public:

    TrafoGradientZ (const AutoDiff<1> & hadz) : adz(hadz) { ; }

    template <typename TIN, typename TOUT>
    void Transform (const TIN & in, TOUT & out)
    {
      out(0) = adz.Value() * in(0);
      out(1) = adz.Value() * in(1);
      out(2) = adz.DValue(0) * in(2);
    }

    template <typename TIN, typename TOUT>
    void TransformAdd (const TIN & in, TOUT & out)
    {
      out(0) += adz.Value() * in(0);
      out(1) += adz.Value() * in(1);
      out(2) += adz.DValue(0) * in(2);
    }


    template <typename TIN, typename TOUT>
    void TransformAddT (const TIN & in, TOUT & out)
    {
      out(0) += adz.Value() * in(0);
      out(1) += adz.Value() * in(1);
      out(2) += adz.DValue(0) * in(2);
    }

  };


  class TrafoGradientY
  {
    const AutoDiff<1> & ady;
  public:
    
    TrafoGradientY (const AutoDiff<1> & hady) : ady(hady) { ; }
    
    template <typename TIN, typename TOUT>
    void Transform (const TIN & in, TOUT & out)
    {
      out(0) = ady.Value() * in(2) + ady.DValue(0) * in(1);
      out(1) = ady.Value() * in(0);
    }
    
    template <typename TIN, typename TOUT>
    void TransformAdd (const TIN & in, TOUT & out)
    {
      out(0) += ady.Value() * in(2) + ady.DValue(0) * in(1);
      out(1) += ady.Value() * in(0);
    }

    template <typename TIN, typename TOUT>
    void TransformAddT (const TIN & in, TOUT & out)
    {
      out(0) += ady.Value() * in(1);
      out(1) += ady.DValue(0) * in(0);
      out(2) += ady.Value() * in(0);
    }

  };


  class TrafoGradientX
  {
    const AutoDiff<1> & adx;
  public:
    
    TrafoGradientX (const AutoDiff<1> & hx) : adx(hx) { ; }
    
    template <typename TIN, typename TOUT>
    void Transform (const TIN & in, TOUT & out)
    {
      out = adx.Value() * in(0) + adx.DValue(0) * in(1);
    }
    
    template <typename TIN, typename TOUT>
    void TransformAdd (const TIN & in, TOUT & out)
    {
      out += adx.Value() * in(0) + adx.DValue(0) * in(1);
    }

    template <typename TIN, typename TOUT>
    void TransformAddT (const TIN & in, TOUT & out)
    {
      out(0) += adx.Value() * in;
      out(1) += adx.DValue(0) * in;
    }

  };









  
  template <int D>
  class LaplaceIntegratorTP : public LaplaceIntegrator<D>
  {

  public:
    LaplaceIntegratorTP (CoefficientFunction * acoef)
      : LaplaceIntegrator<D> (acoef)
    { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new LaplaceIntegratorTP (coeffs[0]);
    }

    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {

      const H1HighOrderTetTP * fel_tet = dynamic_cast<const H1HighOrderTetTP*> (&bfel);      
      const H1HighOrderPrismTP * fel_prism = dynamic_cast<const H1HighOrderPrismTP*> (&bfel);      
      const H1HighOrderHexTP * fel_hex = dynamic_cast<const H1HighOrderHexTP*> (&bfel);      


      if (fel_tet)
        {
          TAssembleElementMatrix (*fel_tet, eltrans, elmat, lh);
          return;
        }

      if (fel_prism)
        if (fel_prism->Sort(0)+3 == fel_prism->Sort(3))  // same singular vertex at bottom and top
          {
            TAssembleElementMatrix (*fel_prism, eltrans, elmat, lh);
            return;
          }
      
      if (fel_hex)
        {
          TAssembleElementMatrix (*fel_hex, eltrans, elmat, lh);
          return;
        }

      LaplaceIntegrator<D>::AssembleElementMatrix(bfel, eltrans, elmat, lh);
    }




    virtual void
    ApplyElementMatrix (const FiniteElement & bfel,
                        const ElementTransformation & eltrans, 
                        const FlatVector<double> & elx, 
                        FlatVector<double> & ely,
                        void * precomputed,
                        LocalHeap & lh) const
    {
      
      const H1HighOrderTetTP * fel_tet = dynamic_cast<const H1HighOrderTetTP*> (&bfel);      
      const H1HighOrderPrismTP * fel_prism = dynamic_cast<const H1HighOrderPrismTP*> (&bfel);      
      const H1HighOrderHexTP * fel_hex = dynamic_cast<const H1HighOrderHexTP*> (&bfel);      


      if (fel_tet)
        {
          TApplyElementMatrix (*fel_tet, eltrans, elx, ely, precomputed, lh);
          return;
        }

      
      if (fel_prism)
        if (fel_prism->Sort(0)+3 == fel_prism->Sort(3))  // same singular vertex at bottom and top
          {
            TApplyElementMatrix (*fel_prism, eltrans, elx, ely, precomputed, lh);
            return;
          }
      
      if (fel_hex)
        {
          TApplyElementMatrix (*fel_hex, eltrans, elx, ely, precomputed, lh);
          return;
        }

      LaplaceIntegrator<D>::ApplyElementMatrix(bfel, eltrans, elx, ely, precomputed, lh);
    }






    template <class FEL>
    void TAssembleElementMatrix (const FEL & fel,
                                 const ElementTransformation & eltrans, 
                                 FlatMatrix<double> & elmat,
                                 LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("Fast assembling - Laplace");
      static int timerg = NgProfiler::CreateTimer ("Fast assembling - geom");
      static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
      static int timerc = NgProfiler::CreateTimer ("Fast assembling - compute");

      NgProfiler::RegionTimer reg (timer);
      NgProfiler::StartTimer (timerg);      


      int ndof = fel.GetNDof();
      elmat.AssignMemory (ndof, ndof, lh);

      HeapReset hr (lh);

      int intorder = 2 * fel.Order();
      if (fel.ElementType() == ET_TET) intorder -= 2;

      IntegrationRuleTP irtp(fel, eltrans, intorder, lh);
      
      const IntegrationRule & irx = irtp.GetIRX();
      const IntegrationRule & iry = irtp.GetIRY();
      const IntegrationRule & irz = irtp.GetIRZ();
      
      int nipx = irx.GetNIP();
      int nipy = iry.GetNIP();
      int nipz = irz.GetNIP();
      int nip = nipx * nipy * nipz;
      
      FlatVector<Mat<3> > tdmat(nip, lh);      // transformed D-matrix
      
      // geometry calculation
      for (int ii = 0; ii < nip; ii++)
        {
          IntegrationPoint ip(irtp.GetXi(ii), irtp.GetWeight(ii));
          SpecificIntegrationPoint<3,3> sip(ip, eltrans, irtp.GetPoint(ii), irtp.GetJacobian(ii), lh);
          
          Mat<3> dmat, hmat, trans;
          this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
          dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();
          
          trans = irtp.GetDuffyJacobian(ii) * sip.GetJacobianInverse();
          hmat = trans * dmat;
          tdmat[ii] = hmat * Trans (trans);
        }

      NgProfiler::StopTimer (timerg);      

      NgProfiler::StartTimer (timeri);

      int ndof2d = fel.GetNDof2d();
      int ndof1d = fel.GetNDof1d();

      FlatArray<int> pmap1dto2d = fel.Get1dTo2dMappingPointer ();
      FlatArray<int> map1dto2d  = fel.Get1dTo2dMapping ();

      FlatArray<int> pmap2dto3d = fel.Get2dTo3dMappingPointer ();
      FlatArray<int> map2dto3d  = fel.Get2dTo3dMapping ();


      
      FlatMatrix<AutoDiff<1> > fac_xdx(ndof,   nipx, lh);    
      FlatMatrix<AutoDiff<1> > fac_ydy(ndof2d, nipy, lh);    
      FlatMatrix<AutoDiff<1> > fac_zdz(ndof1d, nipz, lh);    
      
      for (int i1 = 0; i1 < nipz; i1++)
        {
          AutoDiff<1> z (irz[i1](0), 0);
          fel.CalcZFactor (z, fac_zdz.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipy; i1++)
        {
          AutoDiff<1> y (iry[i1](0), 0);
          fel.CalcYFactor (y, fac_ydy.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipx; i1++)
        {
          AutoDiff<1> x (irx[i1](0), 0);
          fel.CalcXFactor (x, fac_xdx.Col(i1), lh);
        }


      int nipxy = nipx *nipy;
      FlatMatrix<Mat<2> > allsum2d(nipx, lh);
      FlatVector<Mat<3> > allsum1d(nipxy, lh);
      FlatVector<Vec<2> > mult3d(nipx, lh);
      FlatVector<Mat<3,2> > mult2d(nipxy, lh);
      FlatVector<Mat<3> > mult1d(nip, lh);

      NgProfiler::StopTimer (timeri);      



      NgProfiler::StartTimer (timerc);

      for (int kz = 0; kz < ndof1d; kz++)
        {
          for (int ixy = 0, ii = 0; ixy < nipxy; ixy++)
            for (int iz = 0; iz < nipz; iz++, ii++)
              for (int l = 0; l < 3; l++)
                TrafoGradientZ(fac_zdz(kz,iz)) . Transform(tdmat[ii].Col(l), mult1d(ii).Col(l));
          
          for (int jz = 0; jz <= kz; jz++)
            {
              for (int ixy = 0, ii = 0; ixy < nipxy; ixy++)
                {
                  Mat<3> sum = 0.0;

                  for (int iz = 0; iz < nipz; iz++, ii++)
                    for (int l = 0; l < 3; l++)
                      TrafoGradientZ(fac_zdz(jz,iz)) . TransformAdd(mult1d(ii).Row(l), sum.Row(l));
                  
                  allsum1d(ixy) = sum;
                }
              
              for (int ky = pmap1dto2d[kz]; ky < pmap1dto2d[kz+1]; ky++)
                {
                  int kyz = map1dto2d[ky];
                  for (int ix = 0, ixy=0; ix < nipx; ix++)
                    for (int iy = 0; iy < nipy; iy++, ixy++)
                      for (int l = 0; l < 3; l++)
                        TrafoGradientY (fac_ydy(kyz,iy)) . Transform(allsum1d(ixy).Col(l), mult2d(ixy).Row(l));


                  for (int jy = pmap1dto2d[jz]; jy < pmap1dto2d[jz+1]; jy++)
                    {
                      if (jy > ky) continue;
                      int jyz = map1dto2d[jy];

                      for (int ix = 0; ix < nipx; ix++)
                        {
                          Mat<2> hsum2d = 0.0;
                        
                          for (int iy = 0; iy < nipy; iy++)
                            for (int l = 0; l < 2; l++)
                              TrafoGradientY(fac_ydy(jyz,iy)) . TransformAdd(mult2d(ix*nipy+iy).Col(l), hsum2d.Row(l));
                        
                          allsum2d(ix) = Trans(hsum2d);
                        }
                    
                      for (int kx = pmap2dto3d[kyz]; kx < pmap2dto3d[kyz+1]; kx++)
                        {
                          int kk = map2dto3d[kx];

                          for (int ix = 0; ix < nipx; ix++)
                            for (int l = 0; l < 2; l++)
                              TrafoGradientX (fac_xdx(kk,ix)) . Transform(allsum2d(ix).Row(l), mult3d(ix)(l));

                          for (int jx = pmap2dto3d[jyz]; jx < pmap2dto3d[jyz+1]; jx++)
                            {
                              int jj = map2dto3d[jx];

                              double sum = 0;
                              for (int ix = 0; ix < nipx; ix++)
                                TrafoGradientX (fac_xdx(jj,ix)) . TransformAdd(mult3d(ix), sum);
                            
                              elmat(kk,jj) = sum;
                              elmat(jj,kk) = sum;
                            }
                        }
                    }
                }
            }
        }

      NgProfiler::StopTimer (timerc);


      NgProfiler::AddFlops (timerc, 9*nip*ndof1d);
      NgProfiler::AddFlops (timerc, 9*nip*ndof1d*(ndof1d+1)/2);

      NgProfiler::AddFlops (timerc, 9*nipxy*ndof1d*ndof2d);
      NgProfiler::AddFlops (timerc, 3*nipxy*ndof2d*(ndof2d+1));

      NgProfiler::AddFlops (timerc, 4*ndof2d*ndof*nipx);
      NgProfiler::AddFlops (timerc, ndof*(ndof+1)*nipx);
    }










//  #define OLDVERSION
#ifdef OLDVERSION


    template <class FEL>
    void TAssembleElementMatrix (const FEL & fel,
                                 const ElementTransformation & eltrans, 
                                 FlatMatrix<double> & elmat,
                                 LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("Fast assembling");

      static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
      static int timerg1 = NgProfiler::CreateTimer ("Fast assembling - geom1");
      static int timerg2 = NgProfiler::CreateTimer ("Fast assembling - geom2");


      NgProfiler::RegionTimer reg (timer);


      NgProfiler::StartTimer (timerg1);      

      int ndof = fel.GetNDof();
      elmat.AssignMemory (ndof, ndof, lh);

      HeapReset hr (lh);

      
      int order = fel.Order();

      FlatArray<int> map2dto1d = fel.Get2dTo1dMapping ();
      FlatArray<int> map3dto2d = fel.Get3dTo2dMapping ();

      int ndof2d = fel.GetNDof2d();
      int ndof1d = fel.GetNDof1d();
          
      int intorder = 2 * order;
      if (fel.ElementType() == ET_TET) intorder -= 2;
      
      IntegrationRuleTP irtp(fel, eltrans, intorder, lh);
      
      const IntegrationRule & irx = irtp.GetIRX();
      const IntegrationRule & iry = irtp.GetIRY();
      const IntegrationRule & irz = irtp.GetIRZ();
      
      int nipx = irx.GetNIP();
      int nipy = iry.GetNIP();
      int nipz = irz.GetNIP();
      int nip = nipx * nipy * nipz;
      
      NgProfiler::StopTimer (timerg1);            
      NgProfiler::StartTimer (timerg2);      

      FlatVector<Mat<3> > tdmat(nip, lh);      // transformed D-matrix
      
      // geometry calculation
      for (int ii = 0; ii < nip; ii++)
        {
          IntegrationPoint ip(irtp.GetXi(ii), irtp.GetWeight(ii));
          SpecificIntegrationPoint<3,3> sip(ip, eltrans, irtp.GetPoint(ii), irtp.GetJacobian(ii), lh);
          
          Mat<3> dmat, hmat, trans;
          this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
          dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();
          
          trans = irtp.GetDuffyJacobian(ii) * sip.GetJacobianInverse();
          hmat = trans * dmat;
          tdmat[ii] = hmat * Trans (trans);
        }

      NgProfiler::StartTimer (timeri);
      NgProfiler::StopTimer (timerg2);      


      
      FlatMatrix<AutoDiff<1> > fac_xdx(ndof, nipx, lh);    // ndof   x nipx
      FlatMatrix<AutoDiff<1> > fac_ydy(ndof2d, nipy, lh);    // ndof2d x nipy
      FlatMatrix<AutoDiff<1> > fac_zdz(ndof1d, nipz, lh);    // ndof1d x nipz

      
      for (int i1 = 0; i1 < nipz; i1++)
        {
          AutoDiff<1> z (irz[i1](0), 0);
          fel.CalcZFactor (z, fac_zdz.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipy; i1++)
        {
          AutoDiff<1> y (iry[i1](0), 0);
          fel.CalcYFactor (y, fac_ydy.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipx; i1++)
        {
          AutoDiff<1> x (irx[i1](0), 0);
          fel.CalcXFactor (x, fac_xdx.Col(i1), lh);
        }

      NgProfiler::StopTimer (timeri);      


        
      FlatMatrix<Vec<2> > hfac_xdx(ndof, nipx, lh);    // ndof   x nipx
      for (int i1 = 0; i1 < nipx; i1++)
        for (int i = 0; i < ndof; i++)
          {
            hfac_xdx(i, i1)(0) = fac_xdx(i, i1).Value();
            hfac_xdx(i, i1)(1) = fac_xdx(i, i1).DValue(0);
          }



      static int timerz = NgProfiler::CreateTimer ("Fast assembling - transform z");
      static int timery = NgProfiler::CreateTimer ("Fast assembling - transform y");
      static int timerx = NgProfiler::CreateTimer ("Fast assembling - transform x");



      FlatMatrix<Mat<2> > sum2d(ndof2d, lh);
      FlatMatrix<Mat<2> > allsum2d(ndof2d*ndof2d, nipx, lh);
      FlatMatrix<Mat<3> > allsum1d(ndof1d*ndof1d, nipx*nipy, lh);
          
      FlatMatrix<Mat<3> > sum1d(ndof1d, lh);

      NgProfiler::StartTimer (timerz);


      NgProfiler::AddFlops (timerz, 9*nipx*nipy*nipz*ndof1d);
      NgProfiler::AddFlops (timerz, 9*nipx*nipy*nipz*ndof1d*ndof1d/2);


      // transform z
      FlatVector<Mat<3> > mult1d(ndof1d, lh);
      for (int i1 = 0, ii = 0; i1 < nipx; i1++)
        for (int i2 = 0; i2 < nipy; i2++)
          {
            sum1d = 0.0;
                
            for (int i3 = 0; i3 < nipz; i3++, ii++)
              {
                for (int j = 0; j < ndof1d; j++)
                  for (int l = 0; l < 3; l++)
                    {
                      mult1d(j)(0,l) = tdmat[ii](0,l) * fac_zdz(j, i3).Value();
                      mult1d(j)(1,l) = tdmat[ii](1,l) * fac_zdz(j, i3).Value();
                      mult1d(j)(2,l) = tdmat[ii](2,l) * fac_zdz(j, i3).DValue(0);
                    }
                    
                for (int j = 0; j < ndof1d; j++)
                  for (int k = 0; k <= j; k++)
                    for (int l = 0; l < 3; l++)
                      {
                        sum1d(j,k)(l,0) += mult1d(j)(l,0) * fac_zdz (k, i3).Value();   
                        sum1d(j,k)(l,1) += mult1d(j)(l,1) * fac_zdz (k, i3).Value();   
                        sum1d(j,k)(l,2) += mult1d(j)(l,2) * fac_zdz (k, i3).DValue(0);  
                      }
              }
                
            for (int j = 0; j < ndof1d; j++)
              {
                for (int k = 0; k <= j; k++)
                  allsum1d(j*ndof1d+k, i1*nipy+i2) = sum1d(j,k);
                for (int k = j+1; k < ndof1d; k++)
                  allsum1d(j*ndof1d+k, i1*nipy+i2) = Trans (sum1d(k,j));
              }
          }

      NgProfiler::StopTimer (timerz);


      // transform y

      NgProfiler::AddFlops (timery, 9*nipx*nipy*ndof1d*ndof2d);
      NgProfiler::AddFlops (timery, 3*nipx*nipy*ndof2d*ndof2d);

      int nipxy = nipx *nipy;
      FlatMatrix<Vec<6> > mult2d(ndof1d, nipxy, lh);

      for (int jj = 0; jj < ndof2d; jj++)
        {
          int j1d = map2dto1d[jj];
          
          for (int k1d = 0; k1d < ndof1d; k1d++)
            for (int i1 = 0; i1 < nipx; i1++)
              for (int i2 = 0; i2 < nipy; i2++)
                {
                  double fy = fac_ydy(jj, i2).Value();
                  double fdy = fac_ydy(jj, i2).DValue(0);

                  int i12 = i1 *nipy+i2;
                  const Mat<3> & hsum1d = allsum1d(j1d*ndof1d+k1d, i12);
                  
                  mult2d(k1d,i12)(0) = fy * hsum1d(0,0);
                  mult2d(k1d,i12)(1) = fy * hsum1d(0,1);
                  mult2d(k1d,i12)(2) = fy * hsum1d(0,2);
                  mult2d(k1d,i12)(3) = fy * hsum1d(2,0) + fdy * hsum1d(1,0);
                  mult2d(k1d,i12)(4) = fy * hsum1d(2,1) + fdy * hsum1d(1,1);
                  mult2d(k1d,i12)(5) = fy * hsum1d(2,2) + fdy * hsum1d(1,2);
                }
          
          for (int kk = jj; kk < ndof2d; kk++)
            {
              int k1d = map2dto1d[kk];
              
              for (int i1 = 0; i1 < nipx; i1++)
                {
                  Mat<2> hsum2d (0.0);
              
                  for (int i2 = 0; i2 < nipy; i2++)
                    {
                      double fy = fac_ydy(kk, i2).Value();
                      double fdy = fac_ydy(kk, i2).DValue(0);
                      
                      const Vec<6> & hmult2d = mult2d(k1d,i1*nipy+i2);
                      
                      hsum2d(0,0) += fy * hmult2d(5) + fdy * hmult2d(4);
                      hsum2d(0,1) += fy * hmult2d(2) + fdy * hmult2d(1);
                      hsum2d(1,0) += fy * hmult2d(3);
                      hsum2d(1,1) += fy * hmult2d(0);
                    }
                  
                  allsum2d(jj*ndof2d+kk, i1) = Trans (hsum2d);
                  allsum2d(kk*ndof2d+jj, i1) = hsum2d;
                }
            }
        }

      NgProfiler::StopTimer (timery);



      NgProfiler::StartTimer (timerx);



      // transform x

      FlatMatrix<Vec<2> > help(ndof, nipx, lh);
      FlatArray<int> map2dto3d(ndof2d * (order+1), lh);
      FlatArray<int> nmap2dto3d(ndof2d, lh);
      nmap2dto3d = 0;
      for (int j = 0; j < ndof; j++)
        {
          int j2d = map3dto2d[j];
          map2dto3d[(order+1)*j2d + nmap2dto3d[j2d]++] = j;
        }

      NgProfiler::AddFlops (timerx, 4*ndof2d*ndof*nipx);
      NgProfiler::AddFlops (timerx, ndof*ndof*nipx);

      for (int k2d = 0; k2d < ndof2d; k2d++)
        {
          for (int jj = 0; jj < ndof; jj++)
            {
              int j2d = map3dto2d[jj];
              int ind = k2d*ndof2d+j2d;

              for (int i1 = 0; i1 < nipx; i1++)
                help(jj,i1) = allsum2d(ind, i1) * hfac_xdx(jj, i1);
            }
              
          for (int k1d = 0; k1d < nmap2dto3d[k2d]; k1d++)
            {
              int kk = map2dto3d[k2d*(order+1)+k1d];

              for (int j = 0; j <= kk; j++)
                {
                  double sum = InnerProduct (help.Row(j), hfac_xdx.Row(kk));

                  elmat(kk*ndof+j) = sum;
                  elmat(j*ndof+kk) = sum;
                }
            }
        }              

      NgProfiler::StopTimer (timerx);
    }

#endif

    















    template <class FEL>
    void TApplyElementMatrix (const FEL & fel,
                              const ElementTransformation & eltrans, 
                              const FlatVector<double> & elx, 
                              FlatVector<double> & ely,
                              void * precomputed,
                              LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("Fast apply");


      NgProfiler::RegionTimer reg (timer);

      int ndof = fel.GetNDof();

      HeapReset hr (lh);

      
      int order = fel.Order();

      FlatArray<int> map2dto1d = fel.Get2dTo1dMapping ();
      FlatArray<int> map3dto2d = fel.Get3dTo2dMapping ();

      int ndof2d = fel.GetNDof2d();
      int ndof1d = fel.GetNDof1d();
          
      int intorder = 2 * order;
      if (fel.ElementType() == ET_TET) intorder -= 2;
      
      IntegrationRuleTP irtp(fel, eltrans, intorder, lh);
      
      const IntegrationRule & irx = irtp.GetIRX();
      const IntegrationRule & iry = irtp.GetIRY();
      const IntegrationRule & irz = irtp.GetIRZ();
      
      int nipx = irx.GetNIP();
      int nipy = iry.GetNIP();
      int nipz = irz.GetNIP();
      int nip = nipx * nipy * nipz;
      
      FlatVector<Mat<3> > tdmat(nip, lh);      // transformed D-matrix
      
      // geometry calculation
      for (int ii = 0; ii < nip; ii++)
        {
          IntegrationPoint ip(irtp.GetXi(ii), irtp.GetWeight(ii));
          SpecificIntegrationPoint<3,3> sip(ip, eltrans, irtp.GetPoint(ii), irtp.GetJacobian(ii), lh);
          
          Mat<3> dmat, hmat, trans;
          this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
          dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();
          
          trans = irtp.GetDuffyJacobian(ii) * sip.GetJacobianInverse();
          hmat = trans * dmat;
          tdmat[ii] = hmat * Trans (trans);
        }

      
      FlatMatrix<AutoDiff<1> > fac_xdx(ndof, nipx, lh);    // ndof   x nipx
      FlatMatrix<AutoDiff<1> > fac_ydy(ndof2d, nipy, lh);    // ndof2d x nipy
      FlatMatrix<AutoDiff<1> > fac_zdz(ndof1d, nipz, lh);    // ndof1d x nipz

      
      for (int i1 = 0; i1 < nipz; i1++)
        {
          AutoDiff<1> z (irz[i1](0), 0);
          fel.CalcZFactor (z, fac_zdz.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipy; i1++)
        {
          AutoDiff<1> y (iry[i1](0), 0);
          fel.CalcYFactor (y, fac_ydy.Col(i1), lh);
        }
      
      for (int i1 = 0; i1 < nipx; i1++)
        {
          AutoDiff<1> x (irx[i1](0), 0);
          fel.CalcXFactor (x, fac_xdx.Col(i1), lh);
        }


      static int timerc = NgProfiler::CreateTimer ("Fast apply - compute");
      NgProfiler::StartTimer (timerc);
      
      int nipxy = nipx *nipy;
      
      FlatMatrix<Mat<2> > allsum2d(nipx, lh);
      FlatVector<Mat<3> > allsum1d(nipxy, lh);

      FlatVector<Vec<2> > mult3d(nipx, lh);
      FlatVector<Mat<3,2> > mult2d(nipxy, lh);
      FlatVector<Mat<3> > mult1d(nip, lh);

      FlatMatrix<Vec<2> > hmat1(ndof2d, nipx, lh);
      FlatMatrix<Vec<3> > hmat2(ndof1d, nipx*nipy, lh);
      FlatVector<Vec<3> > hmat3(nip, lh);

      /*
        cout << "ndof = " << ndof << endl;
        int flops = 2 * (ndof*nipx*2 +  ndof2d*nipx*nipy*3 + ndof1d*nip*3 ) +  nip*9;
        cout << "flops = " << flops << ", flops/n = " << flops/ndof << endl;
      */

      // for (int cnt = 0; cnt < ndof; cnt++)
      {

        hmat1 = 0.0;
        for (int j = 0; j < ndof; j++)
          {
            int j2d = map3dto2d[j];
            for (int ix = 0; ix < nipx; ix++)
              TrafoGradientX(fac_xdx(j, ix)) . TransformAddT (elx(j), hmat1(j2d, ix));
          }
        NgProfiler::AddFlops (timerc, ndof*nipx*2);

        hmat2 = 0.0;
        for (int j = 0; j < ndof2d; j++)
          {
            int j1d = map2dto1d[j];
            for (int ix = 0; ix < nipx; ix++)
              for (int iy = 0; iy < nipy; iy++)
                TrafoGradientY(fac_ydy(j, iy)) . TransformAddT (hmat1(j, ix), hmat2(j1d, ix*nipy+iy));
          }
        NgProfiler::AddFlops (timerc, ndof2d*nipx*nipy*3);

        hmat3 = 0.0;
        for (int j = 0; j < ndof1d; j++)
          for (int ixy = 0, ii = 0; ixy < nipxy; ixy++)
            for (int iz = 0; iz < nipz; iz++, ii++)
              TrafoGradientZ(fac_zdz(j, iz)) . TransformAddT (hmat2(j, ixy), hmat3(ii));


        NgProfiler::AddFlops (timerc, ndof1d*nip*3);

        for (int ii = 0; ii < nip; ii++)
          {
            Vec<3> hv = hmat3(ii);
            hmat3(ii) = tdmat[ii] * hv;
          }
        NgProfiler::AddFlops (timerc, nip*9);

        hmat2 = 0.0;
        for (int j = 0; j < ndof1d; j++)
          for (int ixy = 0, ii = 0; ixy < nipxy; ixy++)
            for (int iz = 0; iz < nipz; iz++, ii++)
              TrafoGradientZ(fac_zdz(j, iz)) . TransformAdd (hmat3(ii), hmat2(j, ixy));

        NgProfiler::AddFlops (timerc, ndof1d*nip*3);

        hmat1 = 0.0;
        for (int j = 0; j < ndof2d; j++)
          {
            int j1d = map2dto1d[j];
            for (int ix = 0; ix < nipx; ix++)
              for (int iy = 0; iy < nipy; iy++)
                TrafoGradientY(fac_ydy(j, iy)) . TransformAdd (hmat2(j1d, ix*nipy+iy), hmat1(j, ix));
          }
        NgProfiler::AddFlops (timerc, ndof2d*nipx*nipy*3);

        ely = 0.0;
        for (int j = 0; j < ndof; j++)
          {
            int j2d = map3dto2d[j];
            for (int ix = 0; ix < nipx; ix++)
              TrafoGradientX(fac_xdx(j, ix)) . TransformAdd (hmat1(j2d, ix), ely(j));
          }
        NgProfiler::AddFlops (timerc, ndof*nipx*2);
      }
      

      NgProfiler::StopTimer (timerc);
      
    }

  };









  
  
  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };        
    
    Init::Init()
    {
      GetIntegrators().AddLFIntegrator ("sourcetp", 3, 1,
                                        SourceIntegratorTP<3>::Create);
      GetIntegrators().AddBFIntegrator ("laplacetp", 3, 1,
                                        LaplaceIntegratorTP<3>::Create);

    }
    
    Init init;
  }

}
