
#include "../include/solve.hpp"

namespace ngsolve
{
  using namespace ngsolve;
  using namespace ngbla;
  using namespace ngla;




  class NumProcLinDynamics : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    BilinearForm * bfm;
    BilinearForm * bfalin;
    ///
    LinearForm * lff;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    int maxsteps;
    ///
    double prec;
    ///
    double beta;
    ///
    double gamma;
    ///
    double tau;
    ///
    int timesteps;

  public:
    ///
    NumProcLinDynamics (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcLinDynamics();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcLinDynamics (pde, flags);
    }


    ///
    virtual void Do();
    ///
    virtual string GetClassName () const
    {
      return "LinDyna";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Linear-form   = " << lff->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl
	  << "precision     = " << prec << endl
	  << "maxsteps      = " << maxsteps << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcLinDynamics :: NumProcLinDynamics (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bfa", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bfm", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);
    //    maxsteps = int(flags.GetNumFlag ("maxsteps", 20));
    prec = flags.GetNumFlag ("prec", 1e-12);
    beta = flags.GetNumFlag ("beta", 0.25);
    gamma = flags.GetNumFlag ("gamma", 0.5);
    tau = flags.GetNumFlag ("tau", 0.1);
    timesteps = int(flags.GetNumFlag ("timesteps", 10000));
  }

  NumProcLinDynamics :: ~NumProcLinDynamics()
  {
    ;
  }

  void NumProcLinDynamics :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc LinDynamics:\n" \
      "------------\n" \
      "Solves the nonlinear system resulting from a boundary value problem\n\n" \
      "Required flags:\n" 
      "-bilinearform=<bfname>\n" 
      "    bilinear-form providing the matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "\nOptional flags:\n"
      "-predoncitioner=<prename>\n"
      "-maxsteps=n\n"
      "-prec=eps\n"
      "-qmr\n"
      "-print\n"
      "    write matrices and vectors into logfile\n"
	<< endl;
  }


  void NumProcLinDynamics :: Do()
  {
    cout << "Solve linear dynamics" << endl;

    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();
    const BaseMatrix & premat = pre->GetMatrix();

    BaseVector & uold = *vecu.CreateVector();
    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();

    BaseVector & u1 = *vecu.CreateVector();
    BaseVector & u2 = *vecu.CreateVector();
    BaseVector & u1old = *vecu.CreateVector();
    BaseVector & u2old = *vecu.CreateVector();
    BaseVector & F = *vecf.CreateVector(); 


    BilinearFormApplication applya(bfa);
    BilinearFormApplication applym(bfm);

    /*
    LinearizedBilinearFormApplication applyalin(&bfa, &uk);
    CGSolver<double> inva (applyalin, c);
    inva.SetMaxSteps (500);
    */

    // const BaseMatrix & inva = premat;

    // stationary solver:
      
    double err, errold, err0;
    double energy, energyold;

    LocalHeap lh (1000000);

    double t = 0;

    

    bfa -> AssembleLinearization (vecu, lh);
    bfm -> AssembleLinearization (vecu, lh);

    const BaseMatrix & mata = bfa->GetMatrix();	 
    const BaseMatrix & matm = bfm->GetMatrix();

    BaseMatrix & matma = *mata.CreateMatrix();
    //    BaseMatrix & matma = *const_cast<BaseMatrix&>(mata).CreateMatrix();

    matma.AsVector() = (tau*tau*beta)*mata.AsVector() + matm.AsVector();
    
    BaseMatrix & invma = *dynamic_cast<BaseSparseMatrix&>(matma).InverseMatrix();
	//    BaseMatrix & invma = matma.InverseMatrix();

    for (int j = 1; j <= timesteps; j++)
      {
	u2 = (uold + tau*u1old + tau*tau*0.5*(1-2*beta)*u2old); 
	F = vecf - mata*u2;
	u2 = invma * F; 

	u1 = u1old + tau*((1-gamma)*u2old + gamma*u2);
	vecu = uold + tau*u1old + tau*tau*0.5*((1-2*beta)*u2old + 2*beta*u2);
	
	energy = bfa->Energy(vecu) - bfm->Energy(u1) - InnerProduct (vecf, vecu);
	cout << "energy = " << energy << endl;

	uold = vecu;
	u1old = u1;
	u2old = u2;

	gfu->Visualize (gfu->GetName());
      }
  }






  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("lindyna", NumProcLinDynamics::Create, NumProcLinDynamics::PrintDoc);
    }

    
    Init init;
  }

}
