
#include "../include/solve.hpp"

namespace ngsolve
{
  using namespace ngsolve;
  using namespace ngbla;
  using namespace ngla;



  class NumProcNonlinDynamics : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    BilinearForm * bfm;
    BilinearForm * bfalin;
    ///
    LinearForm * lff;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    int maxsteps;
    ///
    double prec;
    ///
    double beta;
    ///
    double gamma;
    ///
    double tau;
    ///
    int timesteps;

  public:
    ///
    NumProcNonlinDynamics (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcNonlinDynamics();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcNonlinDynamics (pde, flags);
    }


    ///
    virtual void Do();
    ///
    virtual string GetClassName () const
    {
      return "NonlinElast";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Linear-form   = " << lff->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl
	  << "precision     = " << prec << endl
	  << "maxsteps      = " << maxsteps << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcNonlinDynamics :: NumProcNonlinDynamics (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bfa", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bfm", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);
    maxsteps = int(flags.GetNumFlag ("maxsteps", 20));
    prec = flags.GetNumFlag ("prec", 1e-12);
    beta = flags.GetNumFlag ("beta", 0.25);
    gamma = flags.GetNumFlag ("gamma", 0.5);
    tau = flags.GetNumFlag ("tau", 0.1);
    timesteps = int(flags.GetNumFlag ("timesteps", 10000));
  }

  NumProcNonlinDynamics :: ~NumProcNonlinDynamics()
  {
    ;
  }

  void NumProcNonlinDynamics :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc NonlinDynamics:\n" \
      "------------\n" \
      "Solves the nonlinear system resulting from a boundary value problem\n\n" \
      "Required flags:\n" 
      "-bilinearform=<bfname>\n" 
      "    bilinear-form providing the matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "\nOptional flags:\n"
      "-predoncitioner=<prename>\n"
      "-maxsteps=n\n"
      "-prec=eps\n"
      "-qmr\n"
      "-print\n"
      "    write matrices and vectors into logfile\n"
	<< endl;
  }


  void NumProcNonlinDynamics :: Do()
  {
    cout << "Solve nonlinear dynamics" << endl;

    const BaseVector & vecf = lff->GetVector();
    BaseVector & uj = gfu->GetVector();
    const BaseMatrix & premat = pre->GetMatrix();

    BaseVector & d = *uj.CreateVector();
    BaseVector & w = *uj.CreateVector();

    BaseVector & u1j = *uj.CreateVector();
    BaseVector & u2j = *uj.CreateVector();

    BaseVector & nu = *uj.CreateVector();
    BaseVector & nu1 = *uj.CreateVector();
    BaseVector & nu2 = *uj.CreateVector();

    BaseVector & nuold = *uj.CreateVector();
    BaseVector & nu1old = *uj.CreateVector();
    BaseVector & nu2old = *uj.CreateVector();

    BaseVector & help = *uj.CreateVector();

    BaseVector & F = *vecf.CreateVector(); 


    BilinearFormApplication applya(bfa);
    BilinearFormApplication applym(bfm);

    /*
    LinearizedBilinearFormApplication applyalin(&bfa, &uk);
    CGSolver<double> inva (applyalin, c);
    inva.SetMaxSteps (500);
    */

    // const BaseMatrix & inva = premat;

    // stationary solver:
      
    double err, errold, err0;
    double energy, energyold;

    LocalHeap lh (1000000);

    double t = 0;

    uj = 0;
    u1j = 0;
    u2j = 0;
  

    cout << " vor timeiteration " << endl; 

    bfm -> AssembleLinearization (uj, lh);

    //    const BaseMatrix & mata = bfa->GetMatrix();	 
    const BaseMatrix & matm = bfm->GetMatrix();

    BaseMatrix & matma = *matm.CreateMatrix();

    for (int j = 1; j <= timesteps; j++)
      {
	help = uj + tau*u1j + tau*tau*0.5*(1-2*beta)*u2j; 
	F = tau*tau*beta*vecf + matm*help;

	nuold = uj;
	nu1old = u1j;
	nu2old = u2j;
	cout << " timestep " << j << " time " << tau * j << endl; 

	for (int i = 1; i <= maxsteps; i++)
	  {

	    bfa -> AssembleLinearization (uj, lh);
	    const BaseMatrix & mata = bfa->GetMatrix();	

	    matma.AsVector() = (tau*tau*beta)*mata.AsVector() + matm.AsVector();
	    
	    BaseMatrix & invma = 
	      *dynamic_cast<const BaseSparseMatrix&>(matma).InverseMatrix();
	    
	    d = F - tau*tau*beta*(applya * nuold) - matm * nuold;

	    //	    u2 = vecu - uold - tau*u1old + tau*tau*0.5*(1-2*beta)*u2old;
	    //	    u2 /= (beta*tau*tau);  

	    //	    u1 = u1old + tau*((1-gamma)*u2old + gamma*u2);
	    
	    errold = L2Norm(d);
	    if (i == 1) err0 = err;
	    
	    energyold = bfa->Energy(nuold) + bfm->Energy(nu1) - InnerProduct (vecf, nuold); 
	    

	    cout << "newton it " << i;
	    cout << " err = " << errold/err0;
	    cout << " energy = " << energyold << endl;
	    	    
	    w = invma * d; // newton direction
	   
	    int lin_its = 0;
	    double theta = 2;
	    do
	      {
		nu = nuold + w;
		d = F - tau*tau*beta*(applya * nu) - matm * nu;

		w *= 0.5;
		theta *= 0.5;
		err = L2Norm(d);

		nu2 = nu - uj - tau*u1j + tau*tau*0.5*(1-2*beta)*u2j;
		nu2 /= (beta*tau*tau);  

		nu1 = u1j + tau*((1-gamma)*u2j + gamma*nu2);

		energy = bfa->Energy(nu) + bfm->Energy(nu1) - InnerProduct (vecf, nu);

		nuold= nu;
			
		cout << "theta = " << theta;
		cout << " energy = " << energy << endl;
		cout << " linsearch, err = " << err/err0 << endl;
		cout << " ww " << InnerProduct(w,w) << endl; 
	      }
	    // while (err > errold);
	    while (energy > energyold && lin_its++ < 30 && err > 1e-7*err0);
	    
	    if (err < 1e-7*err0) break;
	  }
	
	uj = nu; 
	u1j = nu1;
	u2j = nu2; 
      }

  }






  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("nonlindyna", NumProcNonlinDynamics::Create, NumProcNonlinDynamics::PrintDoc);
    }

    
    Init init;
  }

}
