
#include "../include/solve.hpp"
#include "mechsolve.hpp"


namespace ngsolve
{
  using namespace ngsolve;


  class NumProcNonlinElast : public NumProc
  {
  protected:
    BilinearForm * bfa;
    LinearForm * lff;
    GridFunction * gfu;
    Preconditioner * pre;

    int maxsteps;
    double prec;

  public:
    NumProcNonlinElast (PDE & apde, const Flags & flags)
      : NumProc (apde)
    {
      bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
      lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
      gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
      pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);
      maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
      prec = flags.GetNumFlag ("prec", 1e-12);
    }
    
    virtual ~NumProcNonlinElast()
    {
      ;
    }

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcNonlinElast (pde, flags);
    }

    virtual void Do(LocalHeap & lh);

    virtual string GetClassName () const
    {
      return "NonlinElast";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Linear-form   = " << lff->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl
	  << "precision     = " << prec << endl
	  << "maxsteps      = " << maxsteps << endl;
    }

    static void PrintDoc (ostream & ost)
    {
      ost <<
	"\n\nNumproc NonlinElast:\n" \
	"------------\n" \
	"Solves the nonlinear system resulting from a boundary value problem\n\n" \
	"Required flags:\n" 
	"-bilinearform=<bfname>\n" 
	"    bilinear-form providing the matrix\n" \
	"-linearform=<lfname>\n" \
	"    linear-form providing the right hand side\n" \
	"-gridfunction=<gfname>\n" \
	"    grid-function to store the solution vector\n" 
	"\nOptional flags:\n"
	"-predoncitioner=<prename>\n"
	"-maxsteps=n\n"
	"-prec=eps\n"
	"-qmr\n"
	"-print\n"
	"    write matrices and vectors into logfile\n"
	  << endl;
    }
    
  };





  void NumProcNonlinElast :: Do(LocalHeap & lh)
  {
    cout << "Solve nonlinear elasticity" << endl;

    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();
    const BaseMatrix & premat = pre->GetMatrix();


    BaseVector & uold = *vecu.CreateVector();
    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();
    BilinearFormApplication applya(bfa);

    double err, errold, err0;
    double energy, energyold;

    /*
    const SparseMatrix<Mat<3,3> > & mata = dynamic_cast<const SparseMatrix<Mat<3,3> > & > (bfa->GetMatrix());	
    SparseCholesky<Mat<3,3> > inva (mata, 0, 0, 1);
    */

    const SparseMatrix<double> & mata = dynamic_cast<const SparseMatrix<double> & > (bfa->GetMatrix());	
    SparseCholesky<double> inva (mata, 0, 0, 1);

    for (int i = 1; i <= maxsteps; i++)
      {
	bfa -> AssembleLinearization (vecu, lh);
	inva.FactorNew (mata);

	d = vecf - applya * vecu;

	err = L2Norm(d);
	if (i == 1) err0 = err;

	energy = bfa->Energy(vecu) - InnerProduct (vecf, vecu);

	cout << "newton it " << i;
	cout << " err = " << err/err0;
	cout << " energy = " << energy << endl;

	errold = err;
	energyold = energy;

	w = inva * d;
	uold = vecu;
	int lin_its = 0;
	double tau = 2;

	do
	  {
	    vecu = uold + w;
	    d = vecf - applya * vecu;
	    w *= 0.5;
	    tau *= 0.5;
	    err = L2Norm(d);

	    energy = bfa->Energy(vecu) - InnerProduct (vecf, vecu);

	    cout << "tau = " << tau;
	    cout << " energy = " << energy << endl;
	    cout << " linsearch, err = " << err/err0 << endl;
	  }
	while (energy > energyold && lin_its++ < 30 && err > 1e-7*err0);

	if (err < 1e-7*err0) break;
      }
  }






  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("nonlinelast", NumProcNonlinElast::Create, NumProcNonlinElast::PrintDoc);

    }

    
    Init init;
  }

}
