#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


#include <ngsapps.hpp>

// maxwell equation, mixed formulation
//   
//   curl E   -  i omega mu H = 0
//   curl H   +  i omega eps E = j
//
// boundary conditions
//    omega mu H_tau = abs(k) E_tau    on GammaR

// hybridized system for E, H (discont. HCurl), 
//                       E_F, H_F (Lagrange parameters for E_tau, H_tau
//
//  ( E, curl h )_T  - (E_F, h_tau )_F  - (i omega mu H, h ) = 0
//  ( i omega eps E, e)  + ( curl H, e )_T  = ( j, e)
//  ( H_tau, e_F )_F  - (e_F , abs(k)/omega mu E_F )_GammaR   = 0
//  ( E_tau - E_F, e_tau - e_F )_F = 0
//  ( H_tau - H_F, h_tau - h_F )_F = 0
//


// HybridMaxwellIntegrator: Integrator for mixed formulation,


// using:

// HybridMaxwellFESpace
// [ hcurl(discont), hcurl(discont), vectorfacet, vectorfacet ]
// for [ E, H, E_F, H_F ]

// HybridMaxwellFiniteElement


namespace hybrid_maxwell
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace hybrid_maxwell;


  // hybrid maxwell integrator for domain integrals, mixed formulation
  // parts:
  // + (E, curl h)            
  // - (i omega mu H, h )     
  // + (i omega eps E, e)
  // + (curl H, e)
  //
  // needed matrices
  //   mat_e
  //   scalmat_e
  //   mat_h
  //   scalmat_h
  //   mat_curle
  //   mat_curlh


  HybridMaxwellIntegrator :: HybridMaxwellIntegrator ( ARRAY<CoefficientFunction*> & coeffs )
    : BilinearFormIntegrator ()
  {
    coeff_omega = coeffs[0];
    coeff_eps_real = coeffs[1];
    coeff_mu_real = coeffs[2];
    if (coeffs.Size() >= 4)
      coeff_eps_imag = coeffs[3];
    else
      coeff_eps_imag = 0;
    if ( coeffs.Size() >= 5 )
      coeff_mu_imag = coeffs[4];
    else
      coeff_mu_imag = 0;

  }

  Integrator * HybridMaxwellIntegrator :: 
  Create (ARRAY<CoefficientFunction*> & coeffs )
  {
    return new HybridMaxwellIntegrator (coeffs);
  }


  void HybridMaxwellIntegrator :: 
  AssembleElementMatrix (const FiniteElement & fel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & lh) const
  {
    int elnum = eltrans.GetElementNr();
    static int timer = NgProfiler :: CreateTimer ("hybridmaxwell-domains");
    NgProfiler :: RegionTimer reg (timer);

    const HybridMaxwellFiniteElement & cfel =
      dynamic_cast<const HybridMaxwellFiniteElement&> (fel);

    const HCurlFiniteElementD<D> & fel_e = 
      dynamic_cast<const HCurlFiniteElementD<D> & > (cfel[0]);
    const HCurlFiniteElementD<D> & fel_h = 
      dynamic_cast<const HCurlFiniteElementD<D> & > (cfel[1]);

    int intorder = max2 (fel_e.Order(), fel_h.Order());

    ELEMENT_TYPE eltype = fel_e.ElementType();

    int nd = cfel.GetNDof();
    int nd_h = fel_h.GetNDof();
    int nd_e = fel_e.GetNDof();

    int base_e = 0;
    int base_h = nd_e;

    elmat.AssignMemory(nd, nd, lh);
    elmat = 0.0;

    FlatMatrixFixHeight<D> mat_e(nd_e, lh); 
    FlatMatrixFixHeight<D> mat_h(nd_h, lh);
    FlatMatrixFixHeight<D, Complex> scalmat_e(nd_e, lh), scalmat_h(nd_h, lh);
    FlatMatrixFixHeight<D> mat_curle(nd_e, lh), mat_curlh(nd_h, lh);

    const IntegrationRule & ir =
      SelectIntegrationRule (eltype, 2*intorder);

    void * heappointer = lh.GetPointer();
    for ( int l = 0; l < ir.GetNIP(); l++ )
      {
	const SpecificIntegrationPoint<D,D> sip( ir[l], eltrans, lh);
	double weightdet = sip.GetJacobiDet() * ir[l].Weight();

	double omega = coeff_omega->Evaluate(sip);
	double mu_real = coeff_mu_real->Evaluate(sip);
	double eps_real = coeff_eps_real->Evaluate(sip);
	double mu_imag = (coeff_mu_imag) ? coeff_mu_imag->Evaluate(sip) : 0.0;
	double eps_imag = (coeff_eps_imag) ? coeff_eps_imag->Evaluate(sip) : 0.0;

	Complex iomegaeps (-eps_imag*omega*weightdet, eps_real*omega*weightdet);
	Complex iomegamu  (-mu_imag*omega*weightdet,  mu_real*omega*weightdet);

	DiffOpIdEdge<D>::GenerateMatrix (fel_e, sip, mat_e, lh);
	DiffOpIdEdge<D>::GenerateMatrix (fel_h, sip, mat_h, lh);

	DiffOpCurlEdge<D> :: GenerateMatrix (fel_e, sip, mat_curle, lh);
	DiffOpCurlEdge<D> :: GenerateMatrix (fel_h, sip, mat_curlh, lh);

	scalmat_e = iomegaeps * mat_e;
	scalmat_h = iomegamu * mat_h;

	// ( i omega eps E, e )
	for ( int i = base_e; i < nd_e; i++ )
	  for ( int j = base_e; j < nd_e; j++ )
	    for ( int d = 0; d < D; d++ )
	      elmat(i,j) += scalmat_e(d,i-base_e) * mat_e(d,j-base_e);

	// - ( i omega mu H, h)
	for ( int i = base_h; i < base_h+nd_h; i++ )
	  for ( int j = base_h; j < base_h+nd_h; j++ )
	    for ( int d = 0; d < D; d++ )
	      elmat(i,j) -=  scalmat_h(d,i-base_h) * mat_h(d,j-base_h);

	// + ( curlH, e ) + ( E, curlh )
	for ( int i = base_e; i < base_e+nd_e; i++ )
	  for ( int j = base_h; j < base_h+nd_h; j++ )
	    {
	      double val_ij = 0;
	      for ( int d = 0; d < D; d++ )
		val_ij += mat_curlh(d,j-base_h) * mat_e(d,i-base_e);
	      val_ij *= weightdet;

	      real(elmat(i,j)) += val_ij;
	      real(elmat(j,i)) += val_ij;
	    }

	lh.CleanUp(heappointer);
    
      }

  }



  HybridMaxwellFacetIntegrator :: HybridMaxwellFacetIntegrator ()
    : BilinearFormIntegrator ()
  {
    ;
  }

  Integrator * HybridMaxwellFacetIntegrator :: 
  Create (ARRAY<CoefficientFunction*> & coeffs )
  {
    return new HybridMaxwellFacetIntegrator ();
  }


  void HybridMaxwellFacetIntegrator :: AssembleElementMatrix ( const FiniteElement & fel,
							       const ElementTransformation & eltrans,
							       FlatMatrix<Complex> & elmat,
							       LocalHeap & lh) const
  {
    int elnum = eltrans.GetElementNr();

    static int timer = NgProfiler :: CreateTimer ( "hybridmaxwell-facets" );

    NgProfiler :: RegionTimer reg(timer);

    const HybridMaxwellFiniteElement & cfel =
      dynamic_cast<const HybridMaxwellFiniteElement&> (fel);

    const HCurlFiniteElementD<D> & fel_e = 
      dynamic_cast<const HCurlFiniteElementD<D> & > (cfel[0]);
    const HCurlFiniteElementD<D> & fel_h = 
      dynamic_cast<const HCurlFiniteElementD<D> & > (cfel[1]);

    const VectorFacetVolumeFiniteElement & fel_ef =
      dynamic_cast<const VectorFacetVolumeFiniteElement & > (cfel[2]);
    const VectorFacetVolumeFiniteElement & fel_hf =
      dynamic_cast<const VectorFacetVolumeFiniteElement & > (cfel[3]);

    ELEMENT_TYPE eltype = cfel.ElementType();


    int nd = cfel.GetNDof();
    int nd_e = fel_e.GetNDof();
    int nd_h = fel_h.GetNDof();
    int nd_ef = fel_ef.GetNDof();
    int nd_hf = fel_hf.GetNDof();

    int base_e = 0;
    int base_h = base_e + nd_e;
    int base_ef = base_h + nd_h;
    int base_hf = base_ef + nd_ef;

    int integrationorder = max2(fel_e.Order(), fel_h.Order());

    elmat.AssignMemory (nd, nd, lh);
    elmat = 0.0;

    FlatMatrixFixHeight<D> mat_etau(nd_e, lh), mat_htau(nd_h, lh), mat_etau_ref(nd_e, lh);
    FlatMatrixFixWidth<D> mat_effw(nd_ef, lh), mat_hffw(nd_hf, lh);
    FlatMatrixFixHeight<D> mat_ef(nd_ef, lh), mat_hf(nd_hf, lh);

    FlatMatrix<> mat_ef1 (mat_effw);
    FlatMatrix<> mat_hf1 (mat_hffw);

    // D matrix
    // for B-matrix 
    // ( mat_etau )
    // ( mat_htau )
    // ( mat_ef   )
    // ( mat_hf   )

    Mat<4> dmat;

    const POINT3D * points = ElementTopology::GetVertices (eltype);
    const FACE * faces = ElementTopology::GetFaces (eltype);
    int nfa = ElementTopology :: GetNFaces(eltype);

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    void * heappointer = lh.GetPointer();


    for ( int k = 0; k < nfa; k++ )
      {
	ELEMENT_TYPE etfacet = ElementTopology :: GetFacetType (eltype, k);
	Vec<D> normal_ref, normal;
	for ( int i = 0; i < D; i++ )
	  normal_ref(i) = normals[k][i];
	double len_ref = L2Norm(normal_ref);

	const IntegrationRule & irfacet =
	  SelectIntegrationRule (etfacet, 2*integrationorder);

	for ( int l = 0; l < irfacet.GetNIP(); l++ )
	  {
	    IntegrationPoint ip = transform(k, irfacet[l]);
	    SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

	    double det = sip.GetJacobiDet();
	    normal = det * Trans(sip.GetJacobianInverse()) * normal_ref;
	    double len = L2Norm(normal);
	    normal /= len;

	    double weightdet = len * irfacet[l].Weight();

	    dmat = 0.0;
	    
	    dmat(1,2) = dmat(2,1) = - weightdet;
	    
	    dmat(0,2) = dmat(2,0) =  weightdet;
	    dmat(0,0) = dmat(2,2) = -  weightdet;

	    dmat(1,3) = dmat(3,1) = - weightdet;
	    dmat(1,1) = dmat(3,3) =   weightdet;

	    DiffOpIdEdge<D>::GenerateMatrix ( fel_e, sip, mat_etau, lh);

	    DiffOpIdEdge<D>::GenerateMatrix ( fel_h, sip, mat_htau, lh);

// 	    DiffOpIdEdge<D>::GenerateMatrix ( fel_hf, sip, mat_hf, lh);
// 	    DiffOpIdEdge<D>::GenerateMatrix ( fel_ef, sip, mat_ef, lh);

	    fel_hf.CalcShape (sip.IP(), k, mat_hf1);
	    fel_ef.CalcShape (sip.IP(), k, mat_ef1);


	    mat_ef = Trans(sip.GetJacobianInverse()) * Trans(mat_ef1);
	    mat_hf = Trans(sip.GetJacobianInverse()) * Trans(mat_hf1);

 	    CrossNormal ( mat_etau, normal );
 	    CrossNormal ( mat_ef, normal );

	    int first_fdof_ef, first_fdof_hf;
	    int next_fdof_ef, next_fdof_hf;
	    first_fdof_ef = fel_ef.GetFirstFacetDof (k);
	    first_fdof_hf = fel_hf.GetFirstFacetDof (k);
	    next_fdof_ef = fel_ef.GetFirstFacetDof (k+1);
	    next_fdof_hf = fel_hf.GetFirstFacetDof (k+1);

	    // consistency
	    
	    for ( int d = 0; d < D; d++ )
	      {
		// - ( Ef, htau )
		for ( int i = first_fdof_ef; i < next_fdof_ef; i++ )
		  for ( int j = 0; j < nd_h; j++ )
		    {
		      real(elmat( i+base_ef, j+base_h )) += dmat(1,2) * mat_ef(d,i) * mat_htau(d,j);
		      real(elmat( j+base_h, i+base_ef )) += dmat(2,1) * mat_ef(d,i) * mat_htau(d,j);
		    }
	      }


	    CrossNormal ( mat_htau, normal );
	    CrossNormal ( mat_hf, normal );

	    // stability terms

	    for ( int d = 0; d < D; d++ )
	      {

		// ( Ef - Etau, ef - etau )
		for ( int i = 0; i < nd_e; i++ )
		  for ( int j = 0; j < nd_e; j++ )
		    {
		      real(elmat(i+base_e, j+base_e )) += dmat(0,0) * mat_etau(d,i) * mat_etau(d,j);
		    }

		for ( int i = 0; i < nd_e; i++ )
		  for ( int j = first_fdof_ef; j < next_fdof_ef; j++ )
		    {
		      real(elmat(i+base_e, j+base_ef)) += dmat(0,2) * mat_etau(d,i) * mat_ef(d,j);
		      real(elmat(j+base_ef, i+base_e)) += dmat(2,0) * mat_etau(d,i) * mat_ef(d,j);
		    }

		for ( int i = first_fdof_ef; i < next_fdof_ef; i++ )
		  for ( int j = first_fdof_ef; j < next_fdof_ef; j++ )
		    {
		      real(elmat(i+base_ef, j+base_ef )) += dmat(2,2) * mat_ef(d,i) * mat_ef(d,j);
		    }

		// ( Hf - Htau, hf -htau )
		for ( int i = 0; i < nd_h; i++ )
		  for ( int j = 0; j < nd_h; j++ )
		    {
		      real(elmat(i+base_h, j+base_h )) += dmat(1,1) * mat_htau(d,i) * mat_htau(d,j);
		    }

		for ( int i = 0; i < nd_h; i++ )
		  for ( int j = first_fdof_hf; j < next_fdof_hf; j++ )
		    {
		      real(elmat(i+base_h, j+base_hf)) += dmat(1,3) * mat_htau(d,i) * mat_hf(d,j);
		      real(elmat(j+base_hf, i+base_h)) += dmat(3,1) * mat_htau(d,i) * mat_hf(d,j);
		    }

		for ( int i = first_fdof_hf; i < next_fdof_hf; i++ )
		  for ( int j = first_fdof_hf; j < next_fdof_hf; j++ )
		    {
		      real(elmat(i+base_hf, j+base_hf )) += dmat(3,3) * mat_hf(d,i) * mat_hf(d,j);
		    }
		
	      } 

	    lh.CleanUp(heappointer);


	  }



      }

  }



  void CrossNormal ( FlatMatrixFixHeight<3> shape, Vec<3> & normal )
  {
    Vec<3> help;
    for ( int i = 0; i < shape.Width(); i++ )
      {
	help = shape.Col(i);
	shape.Col(i) = Cross( help, normal );
      }

  }



  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetIntegrators().AddBFIntegrator ("hybridmaxwellreal", 3, 3,
                                      HybridMaxwellIntegrator::Create);

    GetIntegrators().AddBFIntegrator ("hybridmaxwell", 3, 4,
                                      HybridMaxwellIntegrator::Create);

    GetIntegrators().AddBFIntegrator ("hybridmaxwellcomplex", 3, 5,
                                      HybridMaxwellIntegrator::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridmaxwellfacet", 3, 0,
                                      HybridMaxwellFacetIntegrator::Create);
  
  }

  Init init;


}
