#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>
#include <parallelngs.hpp>

#include <ngsapps.hpp>

// global pde, needed for coefficient functions kappa, eta in finite elements
extern PDE * pde;

namespace uwvf
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace uwvf;

  using namespace ngparallel;

  UWVFFESpace :: UWVFFESpace ( const MeshAccess & ama, const Flags & aflags)
    : FESpace (ama, aflags)
  {
    // no low-order space for two-level method
    low_order_space = 0;

    int nel = ma.GetNE();

    // by default, all elements are plane-wave
    element_type.SetSize(nel);
    element_type = PW_EL;

    // set orders of elements, determine by flag in pdefile
    order_element.SetSize(nel);
    int order = int( aflags.GetNumFlag ( "order", -1) );
    int pw_order = int( aflags.GetNumFlag ( "waveorder", order ) );
    int poly_order = int ( aflags.GetNumFlag("polyorder", order) );

    // elements with distance to zero lower than dist are polynomial 
    // (whyever, for testing reasons)
    double dist = aflags.GetNumFlag("dist", 0.);
    if ( order == -1 )
      {
	order = max2(pw_order, poly_order);
	if ( pw_order == -1 )
	  pw_order = order;
	if ( poly_order == -1 )
	  poly_order = order;
      }

    order_element = pw_order;
    for ( int i = 0; i < nel; i++ )
      {
	Vec<3> point;
	ARRAY<int> vnums;
	ma.GetElPNums(i, vnums);
	point = 0;
	ma.GetPoint(vnums[0], &point(0));

	if ( L2Norm(point) < dist ) 
	  {
	    element_type[i] = RT_EL;
	    order_element[i] = poly_order;
	  }

	if ( i % 2 ) order_element[i]++;
      }
    ;
  }

  FESpace * UWVFFESpace :: Create (const MeshAccess & ma, const Flags & flags )
  {
    return new UWVFFESpace ( ma, flags );
  }

  ///
  void UWVFFESpace :: Update()
  {
    int nel = ma.GetNE();
    int nfa = ma.GetNFacets();

    first_element_dof.SetSize(nel+1);
    surffacet.SetSize(nfa);
    surffacet = -1;

    face_tets.SetSize(nfa);
    face_tets = INT<2> (-1, -1);

    // find elements matching face ( --> face_tets )
    ARRAY<int> elfacets;
    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElFacets ( el, elfacets );

	for ( int i = 0; i < elfacets.Size(); i++ )
	  {
	    if ( face_tets[elfacets[i]][0] == -1 )
	      face_tets[elfacets[i]][0] = el;
	    else
	      face_tets[elfacets[i]][1] = el;
	  }

      }


    // find surface elements matching facets ( --> surffacet )
    for ( int sel = 0; sel < ma.GetNSE(); sel++ )
      {
	ma.GetSElFacets (sel, elfacets);
	for ( int i = 0; i < elfacets.Size(); i++ )
	  surffacet[elfacets[i]] = sel;
      }

    // --------------- HH

    // count number of degrees of freedom
    int dof = 0;
    
    for ( int el = 0; el < nel; el++ )
      {
	first_element_dof[el] = dof;
	switch( element_type[el] )
	  { 
	  case PW_EL:
	    dof += order_element[el];
	    break;
	  case RT_EL:
	    dof += UWVF_RT_FE<2> :: GetNDof(ma.GetElType(el), order_element[el] );
	    break;
	  case PWJ_EL:
	    dof += order_element[el] *  ElementTopology :: GetNFacets(ma.GetElType(el));
	    break;
	  }
      }
    
    first_element_dof[nel] = dof;
    ndof = dof;



  }
  ///
  int  UWVFFESpace :: GetNDof () const
  {
    return ndof;
  }
  ///
  void  UWVFFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int next = first_element_dof[elnr+1];
    for ( int i = first; i < next; i++ )
      dnums.Append(i);


    ELEMENT_TYPE eltype = ma.GetElType(elnr);
    int nfacets = ElementTopology :: GetNFacets(eltype);

    ARRAY<int> elfaces;
    ma.GetElFacets (elnr, elfaces);

    for ( int i = 0; i < nfacets; i++ )
      {
	INT<2> faceelements = face_tets[elfaces[i]];
	for ( int j = 0; j < 2; j++ )
	  if ( faceelements[j] != elnr && faceelements[j] != -1 )
	    {
	      first = first_element_dof[faceelements[j]];
	      next = first_element_dof[faceelements[j]+1];
	      for ( int i = first; i < next; i++ )
		dnums.Append(i);
	    }
      }
  }
  ///
  void  UWVFFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    GetDofNrs (elnr, dnums);
  }
  ///
  void  UWVFFESpace :: GetElementDofNrs ( int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int next = first_element_dof[elnr+1];
    for ( int i = first; i < next; i++ )
      dnums.Append(i);
  }


  

  // ------------------ HH
  const FiniteElement &  UWVFFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    ELEMENT_TYPE eltype = ma.GetElType(elnr);
    int order = order_element[elnr];

    if ( element_type[elnr] == PW_EL )
      {
	switch ( eltype )
          {
          case ET_TET:
            fe = new (lh.Alloc (sizeof(UWVF_PW_Tet))) UWVF_PW_Tet(order);
            break;
          case ET_TRIG:
            fe =  new (lh.Alloc (sizeof(UWVF_PW_Trig))) UWVF_PW_Trig(order);
            //fe = trig; 
            break;
          default:
            fe = 0;
          }
      }
    else if ( element_type[elnr] == RT_EL )
      {
	switch (ma.GetDimension () )
          {
          case 3:
            fe = new (lh.Alloc (sizeof(UWVF_RT_FE<3>))) UWVF_RT_FE<3>(eltype, order);
            break;
          case 2:
            fe = new (lh.Alloc (sizeof(UWVF_RT_FE<2>))) UWVF_RT_FE<2>(eltype, order);
            //fe = trig; 
            break;
          default:
            fe = 0;
          }
      }
    else // eltype PWJ_EL  // TODO
      {
	switch (ma.GetDimension () )
          {
          case 3:
            fe = new (lh.Alloc (sizeof(UWVF_PWJ_Tet))) UWVF_PWJ_Tet(order);
            break;
          case 2:
            fe = new (lh.Alloc (sizeof(UWVF_PWJ_Trig))) UWVF_PWJ_Trig(order);
            //fe = trig; 
            break;
          default:
            fe = 0;
          }
	;
      }

    if ( ma.GetDimension() == 2 )
      {
	UWVF_FE<2> & uwvffe = dynamic_cast < UWVF_FE<2> & > (*fe);
	ARRAY<int> vnums;
	ma.GetElVertices ( elnr, vnums );

	uwvffe.SetUWVFOrder ( order_element[elnr]);
	uwvffe.SetVNums ( vnums );
	uwvffe.SetFESpace( this );

	ARRAY<int> elfacets;
	ma.GetElFacets ( elnr, elfacets );

	int nf = elfacets.Size();
	ARRAY<int> neighbors(nf), elorders(nf), distndof(nf);
	for ( int i = 0; i < elfacets.Size(); i++ )
	  {
	    INT<2> faceels;
	    GetFaceEls(elfacets[i], faceels);
	    neighbors[i] = -1;
	    elorders[i] = -1;
	    distndof[i] = 0;
	    for ( int j = 0; j < 2; j++ )
	      {
		if ( faceels[j] == elnr || faceels[j] == -1 ) continue;
		neighbors[i] = faceels[j];
		elorders[i] = order_element[faceels[j]];
		switch (element_type[neighbors[i]])
		  {
		  case RT_EL:
		    distndof[i] = UWVF_RT_FE<2> :: GetNDof(ma.GetElType(faceels[j]), elorders[i]);
		    break;
		  case PW_EL:
		    distndof[i] = (elorders[i]);
		    break;
		  case PWJ_EL:
		    int nfacets_neighbor = ElementTopology :: GetNFacets(ma.GetElType(neighbors[i]));
		    distndof[i] = elorders[i] * nfacets_neighbor;
		    break;
		  }
	      }

	  }
	uwvffe.SetNeighbors ( neighbors );
	uwvffe.SetNeighborNDof ( distndof );
	uwvffe.SetNeighborOrders ( elorders );
	uwvffe.ComputeNDof();
      }
    else
      {
	UWVF_FE<3> & uwvffe = dynamic_cast < UWVF_FE<3> & > (*fe);
	ARRAY<int> vnums;
	ma.GetElVertices ( elnr, vnums );

	uwvffe.SetUWVFOrder ( order_element[elnr]);
	uwvffe.SetVNums ( vnums );
	uwvffe.SetFESpace( this );
	ARRAY<int> elfacets;
	ma.GetElFacets ( elnr, elfacets );
	int nf = elfacets.Size();
	ARRAY<int>  neighbors(nf), elorders(nf), distndof(nf);
	for ( int i = 0; i < elfacets.Size(); i++ )
	  {
	    INT<2> faceels; 
	    GetFaceEls(elfacets[i], faceels);
	    for ( int j = 0; j < 2; j++ )
	      {
		if ( faceels[j] == elnr || faceels[j] == -1 ) continue;
		neighbors[i] = faceels[j];
		elorders[i] = order_element[faceels[j]];
		switch (element_type[neighbors[i]])
		  {
		  case RT_EL:
		    distndof[i] = UWVF_RT_FE<2> :: GetNDof(ma.GetElType(faceels[j]), elorders[i]);
		    break;
		  case PW_EL:
		    distndof[i] = (elorders[i]);
		    break;
		  case PWJ_EL:
		    int nfacets_neighbor = ElementTopology :: GetNFacets(ma.GetElType(neighbors[i]));
		    distndof[i] = elorders[i] * nfacets_neighbor;
		    break;
		  }
	      }

	  }
	uwvffe.SetNeighbors ( neighbors );
	uwvffe.SetNeighborNDof ( distndof );
	uwvffe.SetNeighborOrders ( elorders );
	uwvffe.ComputeNDof();
      }

    return *fe;
  }
  ///
  const FiniteElement & UWVFFESpace :: GetSFE (int selnr, LocalHeap & lh) const
  {
    FiniteElement * fe = new FE_SegmDummy;

    return *fe;
  }


  // TODO!!!!!!
  void UWVF_PW_Trig :: CalcShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()
    IntegrationPoint ip1d;  
    double x=sip.IP()(0), y=sip.IP()(1);

    Vec<2> point;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

    ELEMENT_TYPE eltype = this->ElementType();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    int face = -1;
    if (fabs(y)<1e-12) 
      {
	face = 0;
      }
    else if (fabs(x)<1e-12)
      {
	face = 1;
      }
    else if (fabs(1-x-y)<1e-12)
      {
	face = 2;
      }
    
    Vec<D> normal_ref;
    for (int i=0; i<D; i++)
      normal_ref(i) = normals[face][i];
	  
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    Vec<D> normal = det * Trans (inv_jac) * normal_ref;
    double len = L2Norm(normal);
    normal /= len;

    for ( int j = 0; j < wavenumber; j++ )
      {
	Vec<2> dj;
	double angle = 2.*M_PI/wavenumber*j ;
	dj(0) = cos(angle );
	dj(1) = sin(angle );

	double djx = InnerProduct(dj, point);
	double djn = InnerProduct(dj, normal);
	shape(j) = (djn + eta) * exp( Complex(0., kappa) * djx );
      }
    lh.CleanUp(heapp);
  }


  void UWVF_PW_Trig :: CalcFShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()
    double x=sip.IP()(0), y=sip.IP()(1);

    Vec<2> point;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

    ELEMENT_TYPE eltype = this->ElementType();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    int face = -1;
    if (fabs(y)<1e-12) 
      {
	face = 0;
      }
    else if (fabs(x)<1e-12)
      {
	face = 1;
      }
    else if (fabs(1-x-y)<1e-12)
      {
	face = 2;
      }
    
    Vec<D> normal_ref;
    for (int i=0; i<D; i++)
      normal_ref(i) = normals[face][i];
	  
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    Vec<D> normal = det * Trans (inv_jac) * normal_ref;
    double len = L2Norm(normal);
    normal /= len;

    for ( int j = 0; j < wavenumber; j++ )
      {
	Vec<2> dj;
	double angle = 2.*M_PI/wavenumber*j ;
	dj(0) = cos(angle );
	dj(1) = sin(angle );

	double djx = InnerProduct(dj, point);
	double djn = InnerProduct(dj, normal);
	shape(j) = (-djn + eta) * exp( Complex(0., kappa) * djx );
      }
    lh.CleanUp(heapp);
  }


  void UWVF_PW_Trig :: CalcInnerShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()

    Vec<2> point;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);

    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();

    shape = 0.0;

    for ( int j = 0; j < wavenumber; j++ )
      {
	Vec<2> dj;
	double angle = 2.*M_PI/wavenumber*j ;
	dj(0) = cos(angle );
	dj(1) = sin(angle );

	double djx = InnerProduct(dj, point);
	shape(j) = exp( Complex(0., kappa) * djx );
      }
    lh.CleanUp(heapp);
  }


  void UWVF_PW_Trig :: CalcSigmaShape ( const SpecificIntegrationPoint<2,2> & sip, 
					FlatMatrixFixHeight<2,Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()

    Vec<2> point;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);

    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();

    shape = 0.0;

    for ( int j = 0; j < wavenumber; j++ )
      {
	Vec<2> dj;
	double angle = 2.*M_PI/wavenumber*j ;
	dj(0) = cos(angle );
	dj(1) = sin(angle );

	double djx = InnerProduct(dj, point);
	for ( int d = 0; d < D; d++ )
	  shape(d,j) = dj(d) * exp( Complex(0., kappa) * djx );
      }
    lh.CleanUp(heapp);
  }

  void UWVF_PW_Trig :: AddFacetElementMatrix (int face, const ElementTransformation & eltrans, 
					      FlatMatrix<Complex> & elmat, LocalHeap & lh) const
  {
    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    int nd_loc_facet = GetNDofLocal();
    int nd_dist_facet = this->neighbor_ndof[face];
    *testout << "local " << nd_loc_facet << ", neighbor " << face << " " << nd_dist_facet << endl;
    int start = nd_loc_facet;
    for ( int i = 0; i < face; i++ )
      start += neighbor_ndof[i];

    void * heapp_start = lh.GetPointer();

    ELEMENT_TYPE etfacet = ElementTopology :: GetFacetType (eltype, face);
    Vec<D> normal_ref, normal, tang;
    for ( int i = 0; i < D; i++ )
      normal_ref(i) = normals[face][i];

    const IntegrationRule & irfacet =
      SelectIntegrationRule (etfacet, 0);
    IntegrationPoint ip = transform(face, irfacet[0]);
    SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
    
    double det = sip.GetJacobiDet();
    normal = det * Trans(sip.GetJacobianInverse()) * normal_ref;
    tang(0) = -normal(1);
    tang(1) = normal(0);
    double len = L2Norm(normal);
    normal /= len;

    Vec<2> point;
    eltrans.CalcPoint(sip.IP(), point, lh);

    int nd_max = max2(nd_loc_facet, nd_dist_facet);

    FlatArray<Vec<2> > d_loc(nd_max, lh), d_dist(nd_max, lh);
    FlatArray<Complex> edx_loc(nd_max, lh);
    FlatArray<double> dn_loc(nd_max, lh), dt_loc(nd_max, lh);
    FlatArray<Complex> edx_dist(nd_max, lh);
    FlatArray<double>  dn_dist(nd_max, lh), dt_dist(nd_max, lh);

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

    for ( int j = 0; j < nd_loc_facet; j++ )
      {
	double angle = 2.*M_PI/wavenumber*j ;
	d_loc[j](0) = cos(angle );
	d_loc[j](1) = sin(angle );
	edx_loc[j] = exp( Complex(0., kappa) * InnerProduct(d_loc[j], point) );
	dn_loc[j] = InnerProduct(d_loc[j], normal);
	dt_loc[j] = InnerProduct(d_loc[j], tang);
      }

  if ( nd_dist_facet == nd_loc_facet )
    {
      edx_dist = edx_loc;
      dn_dist = dn_loc;
      dt_dist = dt_loc;
    }
  else
    {
    for ( int j = 0; j < nd_dist_facet; j++ )
      {
	double angle = 2.*M_PI/nd_dist_facet*j ;
	d_dist[j](0) = cos(angle );
	d_dist[j](1) = sin(angle );
	edx_dist[j] = exp( Complex(0., kappa) * InnerProduct(d_dist[j], point) );
	dn_dist[j] = InnerProduct(d_dist[j], normal);
	dt_dist[j] = InnerProduct(d_dist[j], tang);
      }
    }

    Complex I = Complex(0., 1.);
    for ( int i = 0; i < nd_loc_facet; i++ )
      {
	for ( int j = 0; j < nd_loc_facet; j++ )
	  {
	    double alpha = kappa * 0.5 * (-dt_loc[i] + dt_loc[j]);
	    double sinc = (fabs(alpha) < 1e-10) ? 1 : sin(alpha) / alpha;
	    elmat(i,j) += len * (dn_loc[i]+eta) * (dn_loc[j]+eta) *
	      edx_loc[j] / edx_loc[i] * sinc;
// 	      exp( I * kappa * (-dx_loc[i] + dx_loc[j])) * sinc ;
	  }
	
	for ( int k = 0; k < nd_dist_facet; k++ )
	  {
	    double alpha = kappa * 0.5 * (-dt_loc[i] + dt_dist[k]) ;
	    double sinc = (fabs(alpha) < 1e-10) ? 1 : sin(alpha) / alpha;
	    elmat(i,start + k) -= len * (-dn_dist[k]+eta) * (-dn_loc[i]+eta) *
	      edx_dist[k] / edx_loc[i] * sinc;
// 	      exp ( I * kappa * (dx_dist[k] - dx_loc[i])) * sinc;
	  }

	
      }
  
    lh.CleanUp(heapp_start);
    ;
  }

  void UWVF_PW_Tet :: AddFacetElementMatrix (int facet, const ElementTransformation & eltrans, 
					     FlatMatrix<Complex> & elmat, LocalHeap & lh) const
  {
    ;
  }


  // TODO!!!!!!
  void UWVF_PW_Tet :: CalcShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PW_Tet :: CalcFShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PW_Tet :: CalcInnerShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PW_Tet :: CalcSigmaShape ( const SpecificIntegrationPoint<3,3> & ip, 
				       FlatMatrixFixHeight<3,Complex> shape, LocalHeap & lh) const
  {
    ;
  }


  // *****************************************

  void UWVF_PWJ_Trig :: CalcShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
    IntegrationPoint ip1d;  
    double x=sip.IP()(0), y=sip.IP()(1);

    Vec<2> point, ps;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);
    IntegrationPoint ip0 (0.3, 0.3);
    eltrans.CalcPoint(ip0, ps, lh);

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

    ELEMENT_TYPE eltype = this->ElementType();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    int face = -1;
    if (fabs(y)<1e-12) 
      {
	face = 0;
      }
    else if (fabs(x)<1e-12)
      {
	face = 1;
      }
    else if (fabs(1-x-y)<1e-12)
      {
	face = 2;
      }
    
    Vec<D> mynormal_ref;
    for (int i=0; i<D; i++)
      mynormal_ref(i) = normals[face][i];
	  
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    Vec<D> mynormal = det * Trans (inv_jac) * mynormal_ref;
    double mylen = L2Norm(mynormal);
    mynormal /= mylen;

    int minorder = -wavenumber / 2;
    int maxorder = minorder + wavenumber;

    Complex I = Complex(0.,1.);
    int ii = 0;
    for ( int facei = 0; facei < 3; facei++ )
      {
	Vec<D> normal_ref;
	for (int i=0; i<D; i++)
	  normal_ref(i) = normals[facei][i];
	  
	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
	double len = L2Norm(normal);
	normal /= len;
	Vec<D> tau (-normal(1), normal(0));

	for ( int j = minorder; j < maxorder; j++ )
	  {
	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
	    double kt = fac * (2 * M_PI * j / len);
	    double disc = kappa*kappa-kt*kt;
	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
	    Vec<2, Complex> kvec = kt * tau - kn * normal;
	    
	    Complex kjx = InnerProduct(kvec, point-ps);
	    Complex djn = InnerProduct(kvec, mynormal) / kappa;
	    shape(ii++) = (djn + eta) * exp( I * kjx );
	  }
      }
    lh.CleanUp(heapp);
  }


  void UWVF_PWJ_Trig :: CalcFShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()
    double x=sip.IP()(0), y=sip.IP()(1);

    Vec<2> point, ps;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);
    IntegrationPoint ip0 (0.3, 0.3);
    eltrans.CalcPoint(ip0, ps, lh);

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

    ELEMENT_TYPE eltype = this->ElementType();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    int face = -1;
    if (fabs(y)<1e-12) 
      {
	face = 0;
      }
    else if (fabs(x)<1e-12)
      {
	face = 1;
      }
    else if (fabs(1-x-y)<1e-12)
      {
	face = 2;
      }
    
    Vec<D> mynormal_ref;
    for (int i=0; i<D; i++)
      mynormal_ref(i) = normals[face][i];
	  
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    Vec<D> mynormal = det * Trans (inv_jac) * mynormal_ref;
    double mylen = L2Norm(mynormal);
    mynormal /= mylen;

    int minorder = -wavenumber / 2;
    int maxorder = minorder + wavenumber;

    int ii = 0;
    Complex I = Complex(0.,1.);

    for ( int facei = 0; facei < 3; facei++ )
      {
	Vec<D> normal_ref;
	for (int i=0; i<D; i++)
	  normal_ref(i) = normals[facei][i];
	  
	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
	double len = L2Norm(normal);
	normal /= len;
	Vec<D> tau (-normal(1), normal(0));


	for ( int j = minorder; j < maxorder; j++ )
	  {
	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
	    double kt = fac * (2 * M_PI * j / len);
	    double disc = kappa*kappa-kt*kt;
	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
	    Vec<2, Complex> kvec = kt * tau - kn * normal;
	    
	    Complex kjx = InnerProduct(kvec, point-ps);
	    Complex djn = InnerProduct(kvec, mynormal) / kappa;
	    shape(ii++) = (-djn + eta) * exp( I * kjx );
	  }
      }

    lh.CleanUp(heapp);
  }


  void UWVF_PWJ_Trig :: CalcInnerShape ( const SpecificIntegrationPoint<2,2> & sip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()

    Vec<2> point, ps;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);
    IntegrationPoint ip0 (0.3, 0.3);
    eltrans.CalcPoint(ip0, ps, lh);
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    int minorder = -wavenumber / 2;
    int maxorder = minorder + wavenumber;

    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    Complex I(0.,1.);

    int ii = 0;
    for ( int facei = 0; facei < 3; facei++ )
      {
	Vec<D> normal_ref;
	for (int i=0; i<D; i++)
	  normal_ref(i) = normals[facei][i];
	  
	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
	double len = L2Norm(normal);
	normal /= len;
	Vec<D> tau (-normal(1), normal(0));

	for ( int j = minorder; j < maxorder; j++ )
	  {
	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
	    double kt = fac * (2 * M_PI * j / len);
	    double disc = kappa*kappa-kt*kt;
	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
	    Vec<2, Complex> kvec = kt * tau - kn * normal;

	    Complex kjx = InnerProduct(kvec, point-ps);

	    shape(ii++) = exp( I*kjx );
	  }
      }

    lh.CleanUp(heapp);
  }


  void UWVF_PWJ_Trig :: CalcSigmaShape ( const SpecificIntegrationPoint<2,2> & sip, 
					FlatMatrixFixHeight<2,Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
//   topology: points = 0:()

    Vec<2> point, ps;
    const ElementTransformation & eltrans = sip.GetTransformation();
    eltrans.CalcPoint(sip.IP(), point, lh);
    IntegrationPoint ip0 (0.3, 0.3);
    eltrans.CalcPoint(ip0, ps, lh);

    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    shape = 0.0;

    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = sip.GetJacobiDet();

    int minorder = -wavenumber / 2;
    int maxorder = minorder + wavenumber;

    Complex I(0.,1.);
    int ii = 0;
    for ( int facei = 0; facei < 3; facei++ )
      {
	Vec<D> normal_ref;
	for (int i=0; i<D; i++)
	  normal_ref(i) = normals[facei][i];
	  
	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
	double len = L2Norm(normal);
	normal /= len;
	Vec<D> tau (-normal(1), normal(0));

	for ( int j = minorder; j < maxorder; j++ )
	  {
	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
	    double kt = fac * (2 * M_PI * j / len);
	    double disc = kappa*kappa-kt*kt;
	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
	    Vec<2, Complex> kvec = kt * tau - kn * normal;
	    
	    Complex kjx = InnerProduct(kvec, point-ps);
	    for ( int d = 0; d < D; d++ )
	      shape(d, ii) = kvec(d) * exp( I * kjx );
	    ii++;
	  }
      }
    lh.CleanUp(heapp);
  }

  void UWVF_PWJ_Trig :: AddFacetElementMatrix (int face, const ElementTransformation & eltrans, 
					      FlatMatrix<Complex> & elmat, LocalHeap & lh) const
  {
//     Facet2ElementTrafo transform(eltype);
//     const NORMAL * normals = ElementTopology::GetNormals(eltype);

//     int nd_loc_facet = GetNDofLocal();
//     int nd_dist_facet = this->neighbor_ndof[face];
//     int start = nd_loc_facet;
//     for ( int i = 0; i < face; i++ )
//       start += neighbor_ndof[i];

//     void * heapp_start = lh.GetPointer();

//     ELEMENT_TYPE etfacet = ElementTopology :: GetFacetType (eltype, face);
//     Vec<D> mynormal_ref, mynormal, mytang;
//     for ( int i = 0; i < D; i++ )
//       mynormal_ref(i) = normals[face][i];

//     const IntegrationRule & irfacet =
//       SelectIntegrationRule (etfacet, 0);
//     IntegrationPoint ip = transform(face, irfacet[0]);
//     SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
//     Integration point ip0 (0.3, 0.3);

//     double det = sip.GetJacobiDet();
//     mynormal = det * Trans(sip.GetJacobianInverse()) * mynormal_ref;
//     mytang(0) = -mynormal(1);
//     mytang(1) = mynormal(0);
//     double mylen = L2Norm(mynormal);
//     mynormal /= mylen;

//     Vec<2> point, ps;
//     eltrans.CalcPoint(sip.IP(), point, lh);
//     eltrans.CalcPoint(ip0, ps, lh);

//     int nd_max = max2(nd_loc_facet, nd_dist_facet);

//     FlatArray<Vec<2> > d_loc(nd_max, lh), d_dist(nd_max, lh);
//     FlatArray<Complex> edx_loc(nd_max, lh);
//     FlatArray<Complex> dn_loc(nd_max, lh), dt_loc(nd_max, lh);
//     FlatArray<Complex> edx_dist(nd_max, lh);
//     FlatArray<Complex>  dn_dist(nd_max, lh), dt_dist(nd_max, lh);

//     const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
//     const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
//     double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
//     double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();

//     Mat<D> inv_jac = sip.GetJacobianInverse();
//     double det = sip.GetJacobiDet();

//     int minorder = -wavenumber / 2;
//     int maxorder = minorder + wavenumber;

//     Complex I(0.,1.);
//     int ii = 0;
//     for ( int facei = 0; facei < 3; facei++ )
//       {
// 	Vec<D> normal_ref;
// 	for (int i=0; i<D; i++)
// 	  normal_ref(i) = normals[facei][i];
	  
// 	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
// 	double len = L2Norm(normal);
// 	normal /= len;
// 	Vec<D> tau (-normal(1), normal(0));

// 	for ( int j = minorder; j < maxorder; j++ )
// 	  {
// 	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
// 	    double kt = fac * (2 * M_PI * j / len);
// 	    double disc = kappa*kappa-kt*kt;
// 	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
// 	    Vec<2, Complex> kvec = kt * tau - kn * normal;
	    
// 	    d_loc[ii] = 1.0 / kappa * kvec;
// 	    Complex kjx = InnerProduct(kvec, point-ps);
// 	    edx_loc[ii] = exp( I * kjx );
// 	    dn_loc[ii] = InnerProduct(d_loc[ii], mynormal);
// 	    dt_loc[ii] = InnerProduct(d_loc[ii], mytang);
// 	    ii++;
// 	  }
//       }

//     int wavenumber_dist = nd_dist_facet / 3;
//     minorder = -wavenumber_dist / 2;
//     maxorder = minorder + wavenumber_dist;
//     ElementTransformation disteltrans;
//     this->fespace->GetTransformation(neighbor[face], disteltrans, lh);
// SpecificIntegrationPoint
//     for ( int facei = 0; facei < 3; facei++ )
//       {
// 	Vec<D> normal_ref;
// 	for (int i=0; i<D; i++)
// 	  normal_ref(i) = normals[facei][i];
	  
// 	Vec<D> normal = det * Trans (inv_jac) * normal_ref;
// 	double len = L2Norm(normal);
// 	normal /= len;
// 	Vec<D> tau (-normal(1), normal(0));

// 	for ( int j = minorder; j < maxorder; j++ )
// 	  {
// 	    double fac = (wavenumber < 4) ? 0.5 : 1.0 - 2.0/wavenumber;
	    
// 	    double kt = fac * (2 * M_PI * j / len);
// 	    double disc = kappa*kappa-kt*kt;
// 	    Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));
	    
// 	    Vec<2, Complex> kvec = kt * tau - kn * normal;
	    
// 	    d_loc[ii] = 1.0 / kappa * kvec;
// 	    Complex kjx = InnerProduct(kvec, point-ps);
// 	    edx_loc[ii] = exp( I * kjx );
// 	    dn_loc[ii] = kn / kappa;
// 	    dt_loc[ii] = kt / kappa;
// 	    ii++;
// 	  }
//       }

//     Complex I = Complex(0., 1.);
//     for ( int i = 0; i < nd_loc_facet; i++ )
//       {
// 	for ( int j = 0; j < nd_loc_facet; j++ )
// 	  {
// 	    double alpha = kappa * 0.5 * (-dt_loc[i] + dt_loc[j]);
// 	    double sinc = (fabs(alpha) < 1e-10) ? 1 : sin(alpha) / alpha;
// 	    elmat(i,j) += len * (dn_loc[i]+eta) * (dn_loc[j]+eta) * 
// 	      edx_loc[j] / edx_loc[i] * sinc;
// // 	      exp( I * kappa * (-dx_loc[i] + dx_loc[j])) * sinc ;
// 	  }
	
// 	for ( int k = 0; k < nd_dist_facet; k++ )
// 	  {
// 	    double alpha = kappa * 0.5 * (dt_dist[i] - dt_loc[k]) ;
// 	    double sinc = (fabs(alpha) < 1e-10) ? 1 : sin(alpha) / alpha;
// 	    elmat(i,start + k) -= len * (-dn_dist[k]+eta) * (-dn_loc[i]+eta) *
// 	      edx_dist[k] / edx_loc[i] * sinc;
// // 	      exp ( I * kappa * (dx_dist[k] - dx_loc[i])) * sinc;
// 	  }

	
//       }
  
//     lh.CleanUp(heapp_start);
//     ;
  }

  void UWVF_PWJ_Tet :: AddFacetElementMatrix (int facet, const ElementTransformation & eltrans, 
					     FlatMatrix<Complex> & elmat, LocalHeap & lh) const
  {
    ;
  }


  // TODO!!!!!!
  void UWVF_PWJ_Tet :: CalcShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PWJ_Tet :: CalcFShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PWJ_Tet :: CalcInnerShape ( const SpecificIntegrationPoint<3,3> & ip, 
				  FlatVector<Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // TODO!!!!!!
  void UWVF_PWJ_Tet :: CalcSigmaShape ( const SpecificIntegrationPoint<3,3> & ip, 
				       FlatMatrixFixHeight<3,Complex> shape, LocalHeap & lh) const
  {
    ;
  }

  // *******************************************************

  template <int DIM>
  UWVF_RT_FE<DIM> :: UWVF_RT_FE  (ELEMENT_TYPE aeltype, int aorder)
    : UWVF_FE<DIM> (aeltype, aorder),
      rt_order(aorder),
      hdivfel(0),
      l2fel(0)
  {
    SetElements();
  }

  template<>
  void UWVF_RT_FE<2> :: SetElements ()
  {
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

    switch ( eltype )
      {
      case ET_TRIG :
	hdivfel = new HDivHighOrderTrig<T_ORTHOPOL>(rt_order);
	l2fel = new L2HighOrderTrig(rt_order);
	facetfel = new FacetVolumeTrig();
	break;
      default:
	cout << "ERROR in SetElements2" << endl;
      }


  }



  template<>
  void UWVF_RT_FE<3> :: SetElements ()
  {
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

    switch ( eltype )
      {
      case ET_TET :
	hdivfel = new HDivHighOrderTet<T_ORTHOPOL>(rt_order);
	l2fel = new L2HighOrderTet(rt_order);
	facetfel = new FacetVolumeTet();
	break;
      default:
	cout << "ERROR in SetElements3" << endl;
      }


  }

  template <int D>
  void UWVF_RT_FE<D> :: SetVNums (FlatArray<int> & avnums)
  {
    for ( int i = 0; i < avnums.Size(); i++ )
      this->vnums[i] = avnums[i];

    hdivfel -> SetVertexNumbers(avnums);
    l2fel -> SetVertexNumbers ( avnums );
    facetfel -> SetVertexNumbers ( avnums );
  }


  template <int D>
  void UWVF_RT_FE<D> :: SetRTOrder( int artorder )
  {
    rt_order = artorder;
    ARRAY<int> order_edge (12);
    ARRAY<int> order_face(6);
    INT<3> order_i = INT<3> (artorder, artorder, artorder);
    order_edge = artorder;
    order_face = artorder;
    hdivfel -> SetOrderEdge (order_edge);
    hdivfel -> SetOrderFace (order_face);
    hdivfel -> SetOrderInner (artorder);
    l2fel -> SetOrderInner ( order_i);
    l2fel -> SetOrder (artorder);
    facetfel -> SetOrder (artorder);

  }

  template <int D>
  void UWVF_RT_FE<D> :: CalcRTShape (const SpecificIntegrationPoint<D,D> & sip, 
				     FlatMatrixFixWidth<D> shape, LocalHeap & lh) const
  {
    shape = 0.0;
    int start, next;
    if ( D == 2 )
      {
	start = 0;//(rt_order+1)*(rt_order+2);
	next = (rt_order+1)*(rt_order+2);
	if ( rt_order == 0 ) next = 3;
      }
    else
      {
	start = 0;//(rt_order+1)*(rt_order+2)*(rt_order+3)/2;
	next = (rt_order+1)*(rt_order+2)*(rt_order+3)/2;
	if ( rt_order == 0 ) next = 4;
      }
    FlatMatrixFixHeight<D> scalshape (next - start, &shape(start));
    DiffOpIdHDiv<D> :: GenerateMatrix ( *hdivfel, sip, scalshape, lh);
    //     hdivfel -> CalcShape ( ip, shape);
  }

  template <int D>
  void UWVF_RT_FE<D> :: CalcScalarShape (const IntegrationPoint & ip, 
					 FlatVector<> shape) const
  {
    shape = 0.0;
    int start, next;
    if ( D == 2 )
      {
	start = (rt_order+1)*(rt_order+2);
	if ( rt_order == 0 )
	  start = 3;
	next = (rt_order+1)*(rt_order+2)*3/2;
      }
    else
      {
	start = (rt_order+1)*(rt_order+2)*(rt_order+3)/2;
	if ( rt_order == 0 )
	  start = 4;
	next = (rt_order+1)*(rt_order+2)*(rt_order+3)*2/3;
      }
    FlatVector <> scalshape (next - start, &shape(start));
    l2fel -> CalcShape (ip, scalshape );
  }

  template <int D>
  void UWVF_RT_FE<D> :: CalcFacetShape (const SpecificIntegrationPoint<D,D> & sip, 
					FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();
    shape = 0.0;
    int start, next;
    start = 0;
    if ( D == 2 )
      {
	next = (rt_order+1)*3;
      }
    else
      {
	next = (rt_order+1)*(rt_order+2)*2;
      }
    FlatVector <> scalshape (next - start, lh);
    facetfel -> CalcShape (sip.IP(), scalshape );
    for ( int i = start; i < next; i++ )
      shape(i) = Complex(scalshape(i-start), 0.);

    lh.CleanUp(heapp);
  }

//   template <int D>  
//   void UWVF_RT_FE<D> :: CalcShape (const IntegrationPoint & ip, 
// 				   FlatVector<> shape) const
//   {
//     l2fel -> CalcShape ( ip, shape );
//   }



  template <int D>  
  void UWVF_FE<D> :: CalcDistFacetShape (int facet, const SpecificIntegrationPoint<D,D> & sip, 
					    FlatVector<Complex> shape, LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();

    IntegrationPoint ip = sip.IP();
    shape = 0.0;
    int start = 0, next = 0;
    switch ( this->eltype )
      {
      case ET_TRIG:
	start = GetNDofLocal();
	for ( int i = 0; i < facet; i++ )
	  if ( this->neighbors[i] >= 0 )
	    start += this->neighbor_ndof[i]; //(this->neighbor_orders[i]+1)*3;
	next = start +  this->neighbor_ndof[facet]; // ( (this->neighbor_orders[facet]+1)*3 );
	break;
      case ET_TET:
	start = GetNDofLocal();
	for ( int i = 0; i < facet; i++ )
	  if ( this->neighbors[i] >= 0 )
	    start += this->neighbor_ndof[i]; //(this->neighbor_orders[i]+1)*3;
	next = start +  this->neighbor_ndof[facet]; // ( (this->neighbor_orders[facet]+1)*3 );
// 	for ( int i = 0; i < facet; i++ )
// 	  if ( this->neighbors[i] >= 0 )
// 	    start += (this->neighbor_orders[i]+1)*(this->neighbor_orders[i]+2)*2;

// 	next = start +  ( (this->neighbor_orders[facet]+1) * (this->neighbor_orders[facet]+2) * 2 );
	break;
      default:
	cout << "ERROR uwvf -- element type in CalcDistFacetShape not implemented" << endl;
      }

    Vec<D> point, point_dist;

    ElementTransformation eltrans = sip.GetTransformation();
    eltrans.CalcPoint ( sip.IP(), point, lh);
//     cout << "local point " << point << endl;
    int distnd = next - start;
    FlatVector<Complex> distshape(distnd, lh); // &shape(start));

    // HERE!!!!!
    int distelnum = this->neighbors[facet];
    const UWVF_FE<D> & distfe = dynamic_cast<const UWVF_FE<D> & > 
      (this->fespace -> GetFE(distelnum, lh) );
    const MeshAccess & ma = this->fespace -> GetMeshAccess();

    int nv = ElementTopology :: GetNVertices ( this->eltype );
    ARRAY<int> distvnums(nv);
    ARRAY<double> loclami(nv), distlami(nv);
    IntegrationPoint * distip;
    distlami = 0.0;
    if ( D == 3 )
      {
	loclami[0] = ip(0); loclami[1] = ip(1); loclami[2] = ip(2); loclami[3] = 1-ip(0)-ip(1)-ip(2);
      }
    else
      {
	loclami[0] = ip(0); loclami[1] = ip(1); loclami[2] =  1-ip(0)-ip(1);
      }
    ma.GetElVertices(distelnum, distvnums);
    for ( int i = 0; i < nv; i++ )
      {
	for ( int j = 0; j < nv; j++ )
	  if ( vnums[i] == distvnums[j] )
	    {
	      distlami[j] = loclami[i];
	    }
      }

    if ( D == 2 )
      distip = new IntegrationPoint (distlami[0], distlami[1]);
    else 
      distip = new IntegrationPoint (distlami[0], distlami[1], distlami[2]);

    ElementTransformation disteltrans;
    ma.GetElementTransformation(distelnum, disteltrans, lh);

    SpecificIntegrationPoint<D,D> distsip(*distip, disteltrans, lh);
//     disteltrans.CalcPoint ( *distip, point_dist, lh);
//     cout << "distant point " << point_dist << endl;
//     cout << "***********************" << endl;
//     const FacetVolumeFiniteElement & distfacetfel = distfe . GetFacetFE();
    distfe . CalcFacetShape ( distsip, distshape, lh );

    for ( int i = start; i < next; i++ )
      shape(i) = distshape(i-start);

    delete distip;
    lh.CleanUp(heapp);

  }


  template <int D>
  void UWVF_RT_FE<D> :: CalcFMatrix ( const ElementTransformation & eltrans, 
				      LocalHeap & lh )
  {
    // Assemble element matrix A
    UWVF_RT_FE<D> & uwvffe = *this;

    const HDivHighOrderFiniteElement<D> & fel_hdiv = 
      dynamic_cast<const HDivHighOrderFiniteElement<D> & > (uwvffe.GetRTFE() );
    const L2HighOrderFiniteElement & fel_l2 = 
      dynamic_cast<const L2HighOrderFiniteElement &> (uwvffe.GetL2FE() );
    const FacetVolumeFiniteElement & fel_facet =
      dynamic_cast<const FacetVolumeFiniteElement &> (uwvffe.GetFacetFE() );
    ELEMENT_TYPE eltype = this->ElementType();
    int nfacets = ElementTopology :: GetNFacets ( eltype );

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    int nd_loc_facet = uwvffe.GetNDof(eltype, uwvffe.GetRTOrder());
    int nd_hdiv = fel_hdiv.GetNDof();
    int nd_l2 = fel_l2.GetNDof();
    int nd_loc = nd_hdiv + nd_l2;

    fmatrix.AssignMemory(nd_loc, nd_loc_facet, lh);
    void * heapp_start = lh.GetPointer();

    FlatVector<> mat_l2(nd_l2, lh), scalmat_l2(nd_l2, lh);
    FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
    FlatVector<> mat_hdiv_n(nd_hdiv, lh);
    FlatVector<> mat_facet(nd_loc_facet, lh);
      
    FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh);
    FlatVector<> div_sigma(nd_hdiv, lh);
    FlatMatrixFixHeight<D> scalbmat_sigma(nd_hdiv, lh);
    FlatVector<> scaldiv_sigma(nd_hdiv, lh);
      
    FlatMatrix<Complex> mass_bd_u(nd_loc, nd_loc_facet, lh);
    FlatMatrix<Complex> amat (nd_loc, nd_loc, lh);
    FlatMatrix<Complex> inv_amat (nd_loc, nd_loc, lh);
    amat = 0.0;
    mass_bd_u = 0.0;

    const CoefficientFunction * coeff_eta = pde->GetCoefficientFunction("eta");
    const CoefficientFunction * coeff_kappa = pde->GetCoefficientFunction("kappa");
    double kappa = const_cast<CoefficientFunction&> (*coeff_kappa).EvaluateConst();
    double eta = const_cast<CoefficientFunction&> (*coeff_eta).EvaluateConst();
          
    const IntegrationRule & irDd =
      SelectIntegrationRule (eltype, 2*max2(fel_l2.Order(), fel_hdiv.Order()));
      
      
    
    for (int l=0; l<irDd.GetNIP(); l++)
      {
	void * heapp = lh.GetPointer();
    
	const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
        
	double factor = sip.GetJacobiDet() * irDd[l].Weight();
        
	bmat_sigma = (1.0/sip.GetJacobiDet()) 
	  * Trans (fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian()));
	
	div_sigma = 1.0/sip.GetJacobiDet() *
	  fel_hdiv.GetDivShape(sip.IP(), lh);
	
        
	fel_l2.CalcShape(sip.IP(), mat_l2);
        
	scalmat_l2 = kappa * factor * mat_l2;
	scalbmat_sigma = kappa * factor * bmat_sigma;
	scaldiv_sigma = factor * div_sigma;
	
	for ( int i = 0; i < nd_hdiv; i++ )
	  {
	    for ( int j = 0; j < nd_hdiv; j++ )
	      for ( int k = 0; k < D; k++ )
		{
		  imag(amat(i,j)) -= scalbmat_sigma(k,i) * bmat_sigma(k,j);
		}
	    for ( int j = nd_hdiv; j < nd_loc; j++ )
	      {
		amat(i,j) += scaldiv_sigma(i) * mat_l2(j - nd_hdiv);
		amat(j,i) += scaldiv_sigma(i) * mat_l2(j - nd_hdiv);
	      }
	  }
	for ( int i = nd_hdiv; i < nd_loc; i++ )
	  for ( int j = nd_hdiv; j < nd_loc; j++ )
	    imag(amat(i,j)) += scalmat_l2(i-nd_hdiv) * mat_l2(j-nd_hdiv);

	lh.CleanUp(heapp);
      }
    

    ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

    for (int k = 0; k < nfacets; k++)
      {
	Vec<D> normal_ref;
	for (int i=0; i<D; i++)
	  normal_ref(i) = normals[k][i];
	  
	const IntegrationRule & ir1d =
	  SelectIntegrationRule (etfacet, 2*max (fel_hdiv.Order(),fel_l2.Order()));
	  
	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    IntegrationPoint ip = transform(k, ir1d[l]);
	    SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	      
	    Mat<D> jac = sip.GetJacobian();
	    Mat<D> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    Vec<D> normal = det * Trans (inv_jac) * normal_ref;
	    double len = L2Norm(normal);
	    normal /= len;

              
	    Vec<D> nt = normal_ref; // 
	    nt /= len;

	    fel_l2.CalcShape(sip.IP(), mat_l2);

	    bmat_sigma = (1.0/sip.GetJacobiDet()) 
	      * Trans (fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian()));

	    mat_hdiv_n = Trans(bmat_sigma) * normal;

	    fel_facet.CalcShape(sip.IP(), mat_facet);

	    double factor = len * ir1d[l].Weight() / eta;

	    for ( int i = 0; i < nd_hdiv; i++ )
	      {
		for ( int j = 0; j < nd_hdiv; j++ )
		  {
		    amat(i,j)     -= factor * mat_hdiv_n(i) * mat_hdiv_n(j);
		  }
		for ( int j = 0; j < nd_loc_facet; j++ )
		  {
		    mass_bd_u(i,j) += factor * mat_hdiv_n(i) * mat_facet(j);
		  }
	      }


	    lh.CleanUp(heapp);
	  }
      }
      
    CalcInverse(amat, inv_amat);
    fmatrix = inv_amat * mass_bd_u;

    lh.CleanUp(heapp_start);
//     *testout << "amatrix .. " << endl << amat << endl;
//     *testout << "invamatrix yy " << endl << inv_amat << endl;
//     *testout << "mass_boundary " << endl << mass_bd_u << endl;
  }

  template <int D>
  void UWVF_RT_FE<D> :: ApplyFOperator ( const FlatVector<Complex> & mat_y, 
					 FlatVector<Complex> & mat_Fy, 
					 const SpecificIntegrationPoint<D,D> & sip,
					 int face,
					 LocalHeap & lh) const
  {
    const UWVF_RT_FE<D> & uwvffe = *this;

    const HDivHighOrderFiniteElement<D> & fel_hdiv = 
      dynamic_cast<const HDivHighOrderFiniteElement<D> & > (uwvffe.GetRTFE() );
    const L2HighOrderFiniteElement & fel_l2 = 
      dynamic_cast<const L2HighOrderFiniteElement &> (uwvffe.GetL2FE() );

    ELEMENT_TYPE eltype = this->ElementType();

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    Vec<D> normal_ref, normal;
    for ( int i = 0; i < D; i++ )
      normal_ref(i) = normals[face][i];

    int nd_hdiv = fel_hdiv.GetNDof();
    int nd_l2 = fel_l2.GetNDof();
    int nd_loc = nd_hdiv + nd_l2;


    double det = sip.GetJacobiDet();
    normal = det * Trans(sip.GetJacobianInverse()) * normal_ref;
    double len = L2Norm(normal);
    normal /= len;
    
    void * heapp = lh.GetPointer();

    FlatVector<> mat_uy(nd_loc, lh);
    FlatMatrixFixWidth<D> mat_sy(nd_loc, lh);

    FlatVector <> mat_usigman(nd_loc, lh);

    CalcRTShape ( sip, mat_sy, lh );
    
    FlatVector<> mat_sigman(nd_hdiv, &mat_usigman(0));
    FlatVector<> mat_u(nd_l2, &mat_usigman(nd_hdiv));
    mat_sigman = 2 * mat_sy * normal;
    mat_u = 0.0; //  mat_uy;
    
    mat_Fy = mat_y + Trans(fmatrix) * mat_usigman;

    lh.CleanUp(heapp);
  }

  template <int D>
  void UWVF_RT_FE<D> :: ComputeNDof()
  {
    switch ( this ->eltype )
      {
      case ET_TET:
	this -> ndof = (rt_order+1)*(rt_order+2)*2; 
	for ( int i = 0; i < this->nneighbors; i++ )
	  {
	    if ( this -> neighbors[i] < 0 ) continue;
	    this->ndof += this->neighbor_ndof[i];
	    if (dynamic_cast<const UWVFFESpace*> (this->fespace)->IsPW(this->neighbors[i]) )
	      this -> order = 100;
	  }
	break;
      case ET_TRIG:
	this->ndof = ( (rt_order+1)*3 );
	
	for ( int i = 0; i <  this->nneighbors; i++ )
	  {
	    if ( this -> neighbors[i] < 0 ) continue;

	    this->ndof += this->neighbor_ndof[i];

	    if (dynamic_cast<const UWVFFESpace*> (this->fespace)->IsPW(this->neighbors[i]) )
	      this -> order = 100;

	  }
	break;
      default:
	cout << "ERROR uwvf -- elementtype not implemented" << endl;
      }

  }

  template <int D>
  void UWVF_PW_FE<D> :: CalcFacetShape ( const SpecificIntegrationPoint<D,D> & sip, 
					 FlatVector<Complex> shape, LocalHeap & lh ) const
  { CalcShape(sip, shape, lh); }

//   // TODO!!!!!!
//   template <int D>
//   void UWVF_PW_FE<D> :: CalcDistFacetShape (int facet, const SpecificIntegrationPoint<D,D> & ip, 
// 					    FlatVector<Complex> shape, LocalHeap & lh) const

//   {
//     ;
//   }

  template <int D>
  void UWVF_PW_FE<D> :: CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh )
  {
    ;
  }

  template <int D>
  void UWVF_PW_FE<D> :: ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<D,D> & sip,
				int face,
				LocalHeap & lh) const
  {
    CalcFShape (sip, mat_Fy, lh);
  }

  template <int D>
  const FlatMatrix<Complex> & UWVF_PW_FE<D> :: GetFMatrix() const
  { ; }


  // Plane wave element, Joachim

  template <int D>
  void UWVF_PWJ_FE<D> :: CalcFacetShape ( const SpecificIntegrationPoint<D,D> & sip, 
					 FlatVector<Complex> shape, LocalHeap & lh ) const
  { CalcShape(sip, shape, lh); }

  template <int D>
  void UWVF_PWJ_FE<D> :: CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh )
  {
    ;
  }

  template <int D>
  void UWVF_PWJ_FE<D> :: ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<D,D> & sip,
				int face,
				LocalHeap & lh) const
  {
    CalcFShape (sip, mat_Fy, lh);
  }

  template <int D>
  const FlatMatrix<Complex> & UWVF_PWJ_FE<D> :: GetFMatrix() const
  { ; }


  template class UWVF_FE<2>;
  template class UWVF_FE<3>;

  template class UWVF_RT_FE<2>;
  template class UWVF_RT_FE<3>;

  template class UWVF_PW_FE<2>;
  template class UWVF_PW_FE<3>;

  template class UWVF_PWJ_FE<2>;
  template class UWVF_PWJ_FE<3>;



  // --------------------------------------------------------
  //      INTEGRATOR
  // --------------------------------------------------------


  template <int D>
  HelmholtzUWVFIntegrator<D> :: HelmholtzUWVFIntegrator ( ARRAY<CoefficientFunction*> & coeffs )
    : BilinearFormIntegrator ()
  {
    coeff_kappa = coeffs[0];
    coeff_eta = coeffs[1];
    coeff_Q = coeffs[2];
  }

  template<int D>
  Integrator * HelmholtzUWVFIntegrator<D> :: 
  Create (ARRAY<CoefficientFunction*> & coeffs )
  {
    return new HelmholtzUWVFIntegrator<D> (coeffs);
  }

  template <int D>
  void HelmholtzUWVFIntegrator<D> :: 
  AssembleElementMatrix (const FiniteElement & fel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & lh) const
  {
    static int timer = NgProfiler :: CreateTimer ( "uwvf-assemble" );

    int elnum = eltrans.GetElementNr();

    NgProfiler :: RegionTimer reg(timer);

    const UWVF_FE<D> & uwvffe =
      dynamic_cast<const UWVF_FE<D> &> (fel);

    const UWVFFESpace  &space = dynamic_cast<const UWVFFESpace&> (*uwvffe.GetFESpace());

    ELEMENT_TYPE eltype = uwvffe.ElementType();
    int nfacets = ElementTopology :: GetNFacets ( eltype );

    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);

    int nd = uwvffe.GetNDof();
    int nd_loc_facet = uwvffe.GetNDofLocal();

    elmat.AssignMemory (nd, nd, lh);
    elmat = 0.0;

    void * heapp_start = lh.GetPointer();

    FlatVector <Complex> mat_x(nd, lh);
    FlatVector <Complex> mat_y(nd_loc_facet, lh);
    FlatVector <Complex> mat_Fy(nd_loc_facet, lh);

    const_cast<UWVF_FE<D> &> (uwvffe).CalcFMatrix (eltrans, lh);

    ARRAY<int> neighbors;
    uwvffe.GetNeighbors(neighbors);
    Complex norm1 = 0, norm2 = 0;

    bool isinner = true;
    bool iscurved = eltrans.IsCurvedElement();
    bool pwpw_formula;

    for ( int face = 0; face < nfacets; face++ )
      {

	ELEMENT_TYPE etfacet = ElementTopology :: GetFacetType (eltype, face);
	Vec<D> normal_ref, normal;
	for ( int i = 0; i < D; i++ )
	  normal_ref(i) = normals[face][i];

	pwpw_formula = (neighbors[face] >= 0 );
	pwpw_formula = pwpw_formula && (space.ElementType(elnum) == PW_EL) && 
	  (space.ElementType(neighbors[face]) == PW_EL) && (!iscurved);

	if ( pwpw_formula )
	  {
	    dynamic_cast<const UWVF_PW_FE<D> & > 
	      (uwvffe) . AddFacetElementMatrix(face, eltrans, elmat, lh);
	  }
	else
	  {
	    int integrationorder = 2*(uwvffe.Order());
	    
	    
	    const IntegrationRule & irfacet =
	      SelectIntegrationRule (etfacet, integrationorder);
	    
	    for ( int l = 0; l < irfacet.GetNIP(); l++ )
	      {
		IntegrationPoint ip = transform(face, irfacet[l]);
		SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
		
		double det = sip.GetJacobiDet();
		normal = det * Trans(sip.GetJacobianInverse()) * normal_ref;
		double len = L2Norm(normal);
		normal /= len;
		double weightdet = len * irfacet[l].Weight();
		
		if ( neighbors[face] == -1 )
		  {
		    isinner = false;
		    ARRAY<int> faces;
		    space.GetMeshAccess().GetElFacets ( elnum, faces);
		    int surfel = space.SurfFacet(faces[face]);
		    int selindex = space.GetMeshAccess().GetSElIndex(surfel);
		    double Q = dynamic_cast<DomainConstantCoefficientFunction&>
		      (*coeff_Q)[selindex];
		    uwvffe.CalcFacetShape (sip, mat_x, lh);
		    mat_x *= Q;
// 		    if (  == 0 )
// 		      {
// 			uwvffe.CalcFacetShape (sip, mat_x, lh);
// 			mat_x *= -1.;
// 		      }
// 		    else
// 		      mat_x = 0.0;
		  }
		else
		  {
		    uwvffe.CalcDistFacetShape (face, sip, mat_x, lh);
		  }
		
		uwvffe.CalcFacetShape ( sip, mat_y, lh);
		uwvffe.ApplyFOperator (mat_y, mat_Fy, sip, face, lh);
		
		for ( int i = 0; i < nd_loc_facet; i++ )
		  {
		    for ( int k = 0; k < nd; k++ )
		      elmat(i,k) -= conj(mat_Fy(i)) * weightdet * (mat_x(k));
		    
		    for ( int j = 0; j < nd_loc_facet; j++ )
		      elmat(i,j) += conj(mat_y(i)) * (mat_y(j)) * weightdet;
		  }
		
		
	      }
	  }
	
      }
    
    lh.CleanUp(heapp_start);
  }

  template <int D>
  void HelmholtzUWVFIntegrator<D> ::
  CalcFlux (const FiniteElement & fel,
	    const BaseSpecificIntegrationPoint & bsip,
	    const FlatVector<Complex> & elx, 
	    FlatVector<Complex> & flux,
	    bool applyd,
	    LocalHeap & lh) const
  {
//     const UWVF_RT_FE<D> & uwvffe = 
//       dynamic_cast<const UWVF_RT_FE<D> & > (fel);

//     const HDivHighOrderFiniteElement<D> & fel_hdiv = 
//       dynamic_cast<const HDivHighOrderFiniteElement<D> & > (uwvffe.GetRTFE() );
//     const L2HighOrderFiniteElement & fel_l2 = 
//       dynamic_cast<const L2HighOrderFiniteElement &> (uwvffe.GetL2FE() );

//     int nd = fel.GetNDof();
//     int nd_loc_facet = uwvffe.GetNDofLocal();
//     int nd_hdiv = fel_hdiv.GetNDof();
//     int nd_l2 = fel_l2.GetNDof();
//     int nd_loc = nd_hdiv + nd_l2;

//     FlatVector<> mat_uy(nd_loc, lh);
//     FlatMatrixFixWidth<D> mat_sy(nd_loc, lh);

//     FlatVector <> mat_x(nd, lh);
//     FlatVector <> mat_y(nd_loc_facet, lh);
//     FlatVector <Complex> mat_Fy(nd_loc_facet, lh);
//     FlatVector <> mat_usigman(nd_loc, lh);

// //     FlatMatrix<Complex> fmatrix (nd_loc, nd_loc_facet, lh);
// //     fmatrix = 0;

//     const ElementTransformation &eltrans = bsip.GetTransformation();
//     uwvffe.CalcFMatrix(eltrans, lh);

//     uwvffe.CalcScalarShape ( bsip.IP(), mat_uy );

//     FlatVector<> shape(nd, lh);
//     uwvffe.CalcFacetShape (bsip.IP(), shape);

//     flux = 0;
//     for ( int i = 0; i < nd_loc_facet; i++ )
//       for ( int j = nd_hdiv; j < nd_loc; j++ )
// 	flux(0) += fmatrix(j,i) * mat_uy(j) * elx(i);
  }



  template class HelmholtzUWVFIntegrator<2>;
  template class HelmholtzUWVFIntegrator<3>;

  template <int D>
  void HelmholtzUWVFNeumannIntegrator<D> :: 
  AssembleElementVector (const FiniteElement & fel,
			 const ElementTransformation & eltrans, 
			 FlatVector<Complex> & elvec,
			 LocalHeap & lh) const
  {


    int elnum = eltrans.GetElementNr();

    const UWVF_FE<D> & uwvffe = dynamic_cast<const UWVF_FE<D> & > (fel);
    const UWVFFESpace & fespace = dynamic_cast<const UWVFFESpace&> ( *uwvffe.GetFESpace() );
    const MeshAccess & ma = fespace . GetMeshAccess();
  
    ARRAY<int> neighbors, facets;
    uwvffe.GetNeighbors(neighbors);
    ma.GetElFacets(elnum, facets);
  
    bool issurfel = false;
    for ( int i = 0; i < neighbors.Size(); i++ )
      if ( fespace.SurfFacet(facets[i] ) != -1 ) issurfel = true;
  
    ELEMENT_TYPE eltype = uwvffe.ElementType();
    int nfacets = ElementTopology :: GetNFacets ( eltype );
  
    Facet2ElementTrafo transform(eltype);
    const NORMAL * normals = ElementTopology::GetNormals(eltype);
  
    int nd = uwvffe.GetNDof();
    int nd_loc_facet = uwvffe.GetNDofLocal();

    elvec.AssignMemory (nd, lh);
    elvec = 0.0;
  
    void * heapp_start = lh.GetPointer();

    FlatVector <Complex> mat_y(nd_loc_facet, lh);
    FlatVector <Complex> mat_Fy(nd_loc_facet, lh);

    if ( !issurfel ) return;

    int integrationorder = 2*uwvffe.Order();

    const_cast<UWVF_FE<D> &> (uwvffe).CalcFMatrix(eltrans, lh);

    for ( int face = 0; face < nfacets; face++ )
      {
	int sel = fespace.SurfFacet(facets[face]);
	if ( sel == -1 ) continue;
	int index = ma.GetSElIndex(sel);
	double coeff = dynamic_cast<DomainConstantCoefficientFunction&> (*coeff_rhs)[index];


	ELEMENT_TYPE etfacet = ElementTopology :: GetFacetType (eltype, face);
	Vec<D> normal_ref, normal;
	for ( int i = 0; i < D; i++ )
	  normal_ref(i) = normals[face][i];

	const IntegrationRule & irfacet =
	  SelectIntegrationRule (etfacet, integrationorder);
      
	for ( int l = 0; l < irfacet.GetNIP(); l++ )
	  {
	    IntegrationPoint ip = transform(face, irfacet[l]);
	    SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	  
	    double det = sip.GetJacobiDet();
	    normal = det * Trans(sip.GetJacobianInverse()) * normal_ref;
	    double len = L2Norm(normal);
	    normal /= len;
	  
	    double weightdet = len * irfacet[l].Weight();
	  
	    uwvffe.CalcFacetShape ( sip, mat_y, lh);
	    uwvffe.ApplyFOperator (mat_y, mat_Fy, sip, face, lh);

	    for ( int i = 0; i < nd_loc_facet; i++ )
	      {
		elvec(i) += conj(mat_Fy(i))  * weightdet * coeff;
	      }
	  
	  }



      }
    lh.CleanUp(heapp_start);
  }


  template class HelmholtzUWVFNeumannIntegrator<2>;
  template class HelmholtzUWVFNeumannIntegrator<3>;

  template <int D>
  class HelmholtzIntegrator 
    : public T_BDBIntegrator<DiffOpIdHelmholtz<D>, DiagDMat<1>, FiniteElement>
  {
  public:
    HelmholtzIntegrator (CoefficientFunction * coef)
      :  T_BDBIntegrator<DiffOpIdHelmholtz<D>, DiagDMat<1>, FiniteElement>
    (DiagDMat<1> (coef))
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HelmholtzIntegrator (coeffs[0]);
    }

    virtual string Name () const { return "HDivSym-Robin"; }
    virtual bool BoundaryForm () const { return 0; }
  };

  template class HelmholtzIntegrator<2>;
  template class HelmholtzIntegrator<3>;


  template <int D>
  class HelmholtzSigmaIntegrator 
    : public T_BDBIntegrator<DiffOpSigmaHelmholtz<D>, DiagDMat<D>, FiniteElement>
  {
  public:
    HelmholtzSigmaIntegrator (CoefficientFunction * coef)
      :  T_BDBIntegrator<DiffOpSigmaHelmholtz<D>, DiagDMat<D>, FiniteElement>
    (DiagDMat<D> (coef))
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HelmholtzSigmaIntegrator (coeffs[0]);
    }

    virtual string Name () const { return "HDivSym-Robin"; }
    virtual bool BoundaryForm () const { return 0; }
  };

  template class HelmholtzSigmaIntegrator<2>;
  template class HelmholtzSigmaIntegrator<3>;


  namespace
#ifdef MACOS
  uwvf_cpp
#endif
  {

    class Init
    {
    public:
      Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("uwvf", UWVFFESpace::Create);
      //       GetFESpaceClasses().AddFESpace ("uwvf_helmholtz", HelmholtzUWVFFESpace::Create);


      GetIntegrators().AddBFIntegrator ("helmholtzuwvf", 2, 3,
					HelmholtzUWVFIntegrator<2>::Create);

      GetIntegrators().AddBFIntegrator ("helmholtzuwvf", 3, 3,
					HelmholtzUWVFIntegrator<3>::Create);

      GetIntegrators().AddBFIntegrator ("helmholtzvis", 2, 1,
					HelmholtzIntegrator<2>::Create);

      GetIntegrators().AddBFIntegrator ("helmholtzvis", 3, 1,
					HelmholtzIntegrator<3>::Create);

      GetIntegrators().AddBFIntegrator ("helmholtzvissigma", 2, 1,
					HelmholtzSigmaIntegrator<2>::Create);

      GetIntegrators().AddBFIntegrator ("helmholtzvissigma", 3, 1,
					HelmholtzSigmaIntegrator<3>::Create);


      GetIntegrators().AddLFIntegrator ("helmholtzuwvf_neumann", 2, 3,
					HelmholtzUWVFNeumannIntegrator<2>::Create);

      GetIntegrators().AddLFIntegrator ("helmholtzuwvf_neumann", 3, 3,
					HelmholtzUWVFNeumannIntegrator<3>::Create);
    }

    Init init;
  }


}
