#ifndef FILE_UWVF__
#define FILE_UWVF__

/*

FESpace for Ultra-Weak Variational Fomulation
discontinuous, element-dofs are dofs from this element, plus dofs from neighbouring elements

BitArray is_poly stores if or not the corresponding element is polynomial Raviart-Thomas


*/

enum ELTYPE { RT_EL = 1, PW_EL = 2, PWJ_EL = 3 };

class UWVFFESpace : public FESpace
{
protected:
  // Number of degrees of freedom
  int ndof;

  // contains first local elment-degree of freedom
  ARRAY<int> first_element_dof;
  // contains adjacent elements of one facet
  ARRAY<INT<2> > face_tets;

  // ------HH
  // is Paviart-Thomas element?
  ARRAY<ELTYPE> element_type;
  // order of element
  ARRAY<int> order_element;
  // is facet facet on surface? (for assembling of right hand side
  ARRAY<int> surffacet;

public:
  // Constructor
  UWVFFESpace ( const MeshAccess & ama, const Flags & flags);
  // destructor
  virtual ~UWVFFESpace () { ; }
  // creator function
  static FESpace * Create (const MeshAccess & ma, const Flags & flags );

  virtual string GetClassName() const
  {
    return "UWVFFESpace";
  }

  /// 
  void Update();
  ///
  virtual int GetNDof () const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetElementDofNrs ( int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    return;
  }
  ///


  // ------------ HH
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int selnr, LocalHeap & lh) const;

  virtual void GetFaceEls ( int fanr, INT<2> & els ) const
  {
    els = face_tets[fanr];
  }

  virtual int SurfFacet (int facetnum ) const
  { return surffacet[facetnum]; }

  virtual bool IsPW (int elnum ) const
  { return (! element_type[elnum] == RT_EL); }

  virtual ELTYPE ElementType (int elnum) const
  { return element_type[elnum]; }

};

class UWVF_RT_Trig;
class UWVF_RT_Tet;

class UWVF_PW_Trig;
class UWVF_PW_Tet;

class UWVF_PWJ_Trig;
class UWVF_PWJ_Tet;



template <int D> class UWVF_RT_FE;
template <int D> class UWVF_PW_FE;
template <int D> class UWVF_PWJ_FE;


// Base finite element for UWVF
// has pointer to FESpace, such that it can access global information,
// as neighboring elements...

template <int DIM>
class UWVF_FE : public FiniteElement
{
protected:
  // global vertex numbers
  int vnums[8];
  // UWVFFESpace
  const FESpace * fespace;
  // global element number
  int elnum;
  // number of neighboring elements
  int nneighbors;
  // neighbouring element, ordered by local facet ordering
  int  neighbors[6];
  // order of neighboring element
  int neighbor_orders[6];
  // ndof for neighbouring element
  int neighbor_ndof[6];

public:
  // constructor
  UWVF_FE (ELEMENT_TYPE aeltype, int aorder)
    : FiniteElement (ElementTopology::GetSpaceDim (aeltype), aeltype, -1, aorder),
      fespace(0),
      elnum(-1)
  {
    ;
  }

  // destructor
  ~UWVF_FE ()
  { ; }


  virtual void SetVNums (FlatArray<int> & avnums)
  {
    for ( int i = 0; i < avnums.Size(); i++ )
      vnums[i] = avnums[i];
  }

  void SetUWVFOrder ( int aorder )
  {
    if ( dynamic_cast < const UWVF_RT_FE<DIM>* > (this) )
      {
	dynamic_cast < UWVF_RT_FE<DIM>* > (this) -> SetRTOrder(aorder);
      }
    if ( dynamic_cast < const UWVF_PW_FE<DIM>* > (this) )
      {
	dynamic_cast < UWVF_PW_FE<DIM>* > (this) -> SetWaveNumber(aorder);
      }
    if ( dynamic_cast < const UWVF_PWJ_FE<DIM>* > (this) )
      {
	dynamic_cast < UWVF_PWJ_FE<DIM>* > (this) -> SetWaveNumber(aorder);
      }

  }

  virtual void ComputeNDof() = 0;

  void SetElNum ( int aelnum )
  { 
    elnum = aelnum;
  }

  void SetNeighbors ( ARRAY<int> & elnums )
  {
    nneighbors = elnums.Size();
    for ( int i = 0; i < elnums.Size(); i++ )
      neighbors[i] = elnums[i];
  }

  void GetNeighbors ( ARRAY<int> & elnums ) const
  {
    elnums.SetSize(nneighbors);
    for ( int i = 0; i < nneighbors; i++ )
      elnums[i] = neighbors[i];
  }

  void SetNeighborOrders ( ARRAY<int> & elnums )
  {
    for ( int i = 0; i < elnums.Size(); i++ )
      {
	neighbor_orders[i] = elnums[i];
	if ( this->order < elnums[i] ) this->order = elnums[i];
      }
  }

  void SetNeighborNDof ( ARRAY<int> & elnums )
  {
    for ( int i = 0; i < elnums.Size(); i++ )
      neighbor_ndof[i] = elnums[i];
  }

  int GetNeighborNDof ( int neighbor )
  { return neighbor_ndof[neighbor]; }

  void SetFESpace ( const FESpace * afespace )
  {
    fespace = afespace;
  }

  const FESpace * GetFESpace () const
  { return fespace; }

  // calculates shape function for \mathcal X
  virtual void CalcFacetShape ( const SpecificIntegrationPoint<DIM,DIM> & sip, 
				FlatVector<Complex> shape, LocalHeap & lh ) const = 0;

  // calculates value of \Pi \mathcal X
  virtual void CalcDistFacetShape (int facet, const SpecificIntegrationPoint<DIM,DIM> & ip, 
				   FlatVector<Complex> shape, LocalHeap & lh) const;

  // for RT-Elements, the matrix mapping Y --> (v, u) is computed
  virtual void CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh ) = 0;

  virtual const FlatMatrix<Complex> & GetFMatrix() const = 0;

  // operator F is applied to mat_y in specific integration point
  virtual void ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<DIM,DIM> & sip,
				int face,
				LocalHeap & lh) const = 0;

  // number of degrees of freedom for this element only, neighbors are not included
  virtual int GetNDofLocal () const = 0;
};


// plane wave element for UWVF
// contains wavenumber many plane wave functions

template <int DIM>
class UWVF_PW_FE : public UWVF_FE<DIM>
{

protected:
  int wavenumber;

public:
  UWVF_PW_FE (ELEMENT_TYPE aeltype, int aorder)
    : UWVF_FE<DIM> (aeltype, aorder) ,
      wavenumber(aorder)
  {
    ;
  }


  ~UWVF_PW_FE ()
  { ; }

  // calculate (d_k n + eta)e^(id_k x) on facet
  virtual void CalcShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			  FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate (-d_k n + eta)e^(id_k x) on facet
  virtual void CalcFShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate e^(i d_k x)
  virtual void CalcInnerShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate -d_k e^(i d_k x)
  virtual void CalcSigmaShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			       FlatMatrixFixHeight<DIM,Complex> shape, LocalHeap & lh) const = 0;

  // calculate (d_k n + eta)e^(id_k x) on facet
  virtual void CalcFacetShape ( const SpecificIntegrationPoint<DIM,DIM> & sip, 
				FlatVector<Complex> shape, LocalHeap & lh ) const;

//   virtual void CalcDistFacetShape (int facet, const SpecificIntegrationPoint<DIM,DIM> & ip, 
// 				   FlatVector<Complex> shape, LocalHeap & lh) const;


  virtual int GetWaveNumber() const 
  {
    return wavenumber;
  }

  virtual void SetWaveNumber( int artorder )
  {
    wavenumber = artorder;
  }

  virtual void ComputeNDof() = 0;


  // does nothing, as application of F trivial
  virtual void CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh );

  virtual void ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<DIM,DIM> & sip,
				int face,
				LocalHeap & lh) const;

  virtual int GetNDofLocal () const
  { return wavenumber; }

  virtual const FlatMatrix<Complex> & GetFMatrix() const;

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const = 0;
};


// RaviartThomas Finite Element for UWVF
// contains utilities to solve rt-equation on element

template <int DIM>
class UWVF_RT_FE : public UWVF_FE<DIM>
{
protected:
  // order of Raviart-Thomas space
  int rt_order;

  typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

  // finite elements needed internally
  HDivHighOrderFiniteElement<DIM> * hdivfel;
  L2HighOrderFiniteElement * l2fel;
  FacetVolumeFiniteElement * facetfel;

  // matrix X --> (v, u)
  FlatMatrix<Complex> fmatrix;
public:
  // constructor
  UWVF_RT_FE  (ELEMENT_TYPE aeltype, int aorder);

  // destructor
  ~UWVF_RT_FE ()
  { 
    if ( hdivfel ) delete hdivfel;
    if ( l2fel ) delete l2fel; 
    if ( facetfel ) delete facetfel;
  }

  // sets hdivfel, l2fel, facetfel
  void SetElements ();

  virtual void SetVNums (FlatArray<int> & avnums);

  static int GetNDof ( ELEMENT_TYPE myeltype, int order ) 
  {
    switch (myeltype)
      {
      case ET_TRIG:
	return ( (order+1) * 3 );
      case ET_TET:
	return ( (order+1)*(order+2)*2 );
      default:
	cout << "error in static int GetNDof" << endl;
      }
    return 0;
  }

  virtual int GetNDofLocal () const
  { return facetfel -> GetNDof(); }

  virtual void GetInternalDofs ( ARRAY<int> & idofs ) const
  { idofs.SetSize(0); }

  // shape for RT-Function v
  virtual void CalcRTShape (const SpecificIntegrationPoint<DIM, DIM> & sip, 
			    FlatMatrixFixWidth<DIM> shape,
			    LocalHeap & lh) const;

  // shape for scalar function u
  virtual void CalcScalarShape (const IntegrationPoint & ip, 
				FlatVector<> shape) const;

  // shape for facet function X
  virtual void CalcFacetShape ( const SpecificIntegrationPoint<DIM,DIM> & sip, 
				FlatVector<Complex> shape, LocalHeap & lh ) const;

  virtual int GetRTOrder() const 
  {
    return rt_order;
  }

  void SetRTOrder( int artorder );

  virtual void CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh );

  virtual void ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<DIM,DIM> & sip,
				int face,
				LocalHeap & lh) const;

  virtual void ComputeNDof();

  const FacetVolumeFiniteElement & GetFacetFE () const
  { return * facetfel; }

  FacetVolumeFiniteElement & GetFacetFE () 
  { return * facetfel; }

  const HDivHighOrderFiniteElement<DIM> & GetRTFE () const
  { return * hdivfel; }

  HDivHighOrderFiniteElement<DIM> & GetRTFE ()
  { return * hdivfel; }

  L2HighOrderFiniteElement & GetL2FE ()
  { return *l2fel; }

  const L2HighOrderFiniteElement & GetL2FE () const
  { return *l2fel; }

  virtual const FlatMatrix<Complex> & GetFMatrix() const
  {
    return fmatrix;
  }

};



// plane wave element for UWVF, Joachim style
// contains wavenumber*#facets many plane wave functions

template <int DIM>
class UWVF_PWJ_FE : public UWVF_FE<DIM>
{

protected:
  int wavenumber;

public:
  UWVF_PWJ_FE (ELEMENT_TYPE aeltype, int aorder)
    : UWVF_FE<DIM> (aeltype, aorder) ,
      wavenumber(aorder)
  {
    ;
  }


  ~UWVF_PWJ_FE ()
  { ; }

  // calculate (d_k n + eta)e^(id_k x) on facet
  virtual void CalcShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			  FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate (-d_k n + eta)e^(id_k x) on facet
  virtual void CalcFShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate e^(i d_k x)
  virtual void CalcInnerShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const = 0;

  // calculate -d_k e^(i d_k x)
  virtual void CalcSigmaShape (const SpecificIntegrationPoint<DIM,DIM> & sip, 
			       FlatMatrixFixHeight<DIM,Complex> shape, LocalHeap & lh) const = 0;

  // calculate (d_k n + eta)e^(id_k x) on facet
  virtual void CalcFacetShape ( const SpecificIntegrationPoint<DIM,DIM> & sip, 
				FlatVector<Complex> shape, LocalHeap & lh ) const;

//   virtual void CalcDistFacetShape (int facet, const SpecificIntegrationPoint<DIM,DIM> & ip, 
// 				   FlatVector<Complex> shape, LocalHeap & lh) const;


  virtual int GetWaveNumber() const 
  {
    return wavenumber;
  }

  virtual void SetWaveNumber( int artorder )
  {
    wavenumber = artorder;
  }

  virtual void ComputeNDof() = 0;


  // does nothing, as application of F trivial
  virtual void CalcFMatrix ( const ElementTransformation & eltrans, 
			     LocalHeap & lh );

  virtual void ApplyFOperator ( const FlatVector<Complex> & mat_y, 
				FlatVector<Complex> & mat_Fy, 
				const SpecificIntegrationPoint<DIM,DIM> & sip,
				int face,
				LocalHeap & lh) const;

  virtual int GetNDofLocal () const
  { return ElementTopology :: GetNFacets(this->eltype) *wavenumber; }

  virtual const FlatMatrix<Complex> & GetFMatrix() const;

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const = 0;
};

/// **********************************************************************************
/// ELEMENTS for PW

class UWVF_PW_Trig : public UWVF_PW_FE<2>
{
  enum { D = 2 };
public:
  UWVF_PW_Trig (int aorder)
    : UWVF_PW_FE<2>(ET_TRIG, order)
  { ; }

  ~UWVF_PW_Trig()
  { ; }

  virtual void GetInternalDofs (ARRAY<int> & idofs ) const
  { idofs.SetSize(0); return; }

  virtual void CalcShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcFShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcInnerShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcSigmaShape (const SpecificIntegrationPoint<2,2> & sip, 
			       FlatMatrixFixHeight<2,Complex> shape, LocalHeap & lh) const;

  virtual void ComputeNDof()
  {
    ndof = wavenumber;
    order = 100;
    for ( int i = 0; i < nneighbors; i++ )
      {
	int p = neighbor_ndof[i];
	if ( p > 0 )	ndof += p;
      }

  }

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const;

};


class UWVF_PW_Tet : public UWVF_PW_FE<3>
{
public:
  UWVF_PW_Tet (int aorder)
    : UWVF_PW_FE<3>(ET_TET, order)
  { ; }

  ~UWVF_PW_Tet()
  { ; }

  virtual void GetInternalDofs (ARRAY<int> & idofs ) const
  { idofs.SetSize(0); return; }

  virtual void CalcShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcFShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcInnerShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcSigmaShape (const SpecificIntegrationPoint<3,3> & sip, 
			       FlatMatrixFixHeight<3,Complex> shape, LocalHeap & lh) const;

  virtual void ComputeNDof()
  {
    ndof = wavenumber;
    for ( int i = 0; i < nneighbors; i++ )
      {
	int p = neighbor_ndof[i];
	ndof += p;
      }


  }

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const;

};


/// **********************************************************************************
/// ELEMENTS for PW, Joachim

class UWVF_PWJ_Trig : public UWVF_PWJ_FE<2>
{
  enum { D = 2 };
public:
  UWVF_PWJ_Trig (int aorder)
    : UWVF_PWJ_FE<2>(ET_TRIG, order)
  { ; }

  ~UWVF_PWJ_Trig()
  { ; }

  virtual void GetInternalDofs (ARRAY<int> & idofs ) const
  { idofs.SetSize(0); return; }

  virtual void CalcShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcFShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcInnerShape ( const SpecificIntegrationPoint<2,2> & ip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcSigmaShape (const SpecificIntegrationPoint<2,2> & sip, 
			       FlatMatrixFixHeight<2,Complex> shape, LocalHeap & lh) const;

  virtual void ComputeNDof()
  {
    ndof = 3*wavenumber;
    order = 100;
    for ( int i = 0; i < nneighbors; i++ )
      {
	int p = neighbor_ndof[i];
	if ( p > 0 )	ndof += p;
      }
  }

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const;

};


class UWVF_PWJ_Tet : public UWVF_PWJ_FE<3>
{
public:
  UWVF_PWJ_Tet (int aorder)
    : UWVF_PWJ_FE<3>(ET_TET, order)
  { ; }

  ~UWVF_PWJ_Tet()
  { ; }

  virtual void GetInternalDofs (ARRAY<int> & idofs ) const
  { idofs.SetSize(0); return; }

  virtual void CalcShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcFShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcInnerShape ( const SpecificIntegrationPoint<3,3> & sip, 
			   FlatVector<Complex> shape, LocalHeap & lh) const;

  virtual void CalcSigmaShape (const SpecificIntegrationPoint<3,3> & sip, 
			       FlatMatrixFixHeight<3,Complex> shape, LocalHeap & lh) const;

  virtual void ComputeNDof()
  {
    ndof = 4*wavenumber;
    for ( int i = 0; i < nneighbors; i++ )
      {
	int p = neighbor_ndof[i];
	ndof += p;
      }


  }

  virtual void AddFacetElementMatrix (int facet, const ElementTransformation & eltrans,
				      FlatMatrix<Complex> & elmat, LocalHeap & lh) const;

};


// INTEGRATOR for UWVF
// does not (so far) differ between RT or PW finite elements, 
// computation of Fy is done by virtual functions of element

template <int D>
class HelmholtzUWVFIntegrator : public BilinearFormIntegrator
{
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = 1 };


protected:
  CoefficientFunction * coeff_kappa;
  CoefficientFunction * coeff_eta;
  CoefficientFunction * coeff_Q;

public:
  // constructor
  HelmholtzUWVFIntegrator ( ARRAY<CoefficientFunction*> & coeffs );
  // destructor
  ~HelmholtzUWVFIntegrator () { ; }
  // creator function
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs );

  // is integrator on domain
  virtual bool BoundaryForm () const { return 0; }

  // assembles real valued element matrix --> useless
  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<double> & elmat,
				      LocalHeap & lh) const
  {
    throw Exception ("HybridMaxwellIntegrator::AssembleElementMatrix for double called, equation is complex");
  }
  

  // assembles complex valued elmat
  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<Complex> & elmat,
				      LocalHeap & lh) const;
  
  // currently out of order --> todo
  virtual void
  CalcFlux (const FiniteElement & fel,
	    const BaseSpecificIntegrationPoint & bsip,
	    const FlatVector<Complex> & elx, 
	    FlatVector<Complex> & flux,
	    bool applyd,
	    LocalHeap & lh) const;


};


// Integrator for right hand side
// (F^* g, Y)

template <int D>
class HelmholtzUWVFNeumannIntegrator : public LinearFormIntegrator
{
  CoefficientFunction * coeff_rhs;
  CoefficientFunction * coeff_kappa;
  CoefficientFunction *coeff_eta;
public:

  // constructor
  HelmholtzUWVFNeumannIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
    : LinearFormIntegrator()
  { 
    coeff_rhs = coeffs[0];
    coeff_kappa = coeffs[1];
    coeff_eta = coeffs[2];
  }
  
  // destructor
  virtual ~HelmholtzUWVFNeumannIntegrator () { ; }
  
  // creator function
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HelmholtzUWVFNeumannIntegrator (coeffs);
  }
  
  // is integrator on domain, loop over elements, only surface-facets are assembled
  // (do better?)
  virtual bool BoundaryForm () const 
  { return 0; }
  
  
  // real valued 
  virtual void AssembleElementVector (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
                                        FlatVector<double> & elvec,
				      LocalHeap & lh) const
  {
    cout << "HelmholtzUWVFNeumannIntegrator:: no real right hand side" << endl;
  }

  // complex valued
  virtual void AssembleElementVector (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
                                        FlatVector<Complex> & elvec,
				      LocalHeap & lh) const;
};

// Differential operator calculating u from element vector for X
// needs coefficient functions "kappa", "eta" defined in global pde!!!
// for polynomials, matrix for F has to be computed --> VERY slow

template <int D>
class DiffOpIdHelmholtz : public DiffOp<DiffOpIdHelmholtz<D> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    ;
  }

  template <typename FEL, typename SIP>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      FlatMatrixFixHeight<1, Complex> & mat, LocalHeap & lh)
  {
    void * heapp = lh.GetPointer();


    if ( dynamic_cast<const UWVF_RT_FE<D> * > (&bfel ) )
      {
	// *** RAVIART-THOMAS CASE
	const UWVF_RT_FE<D> & uwvffe = 
	  dynamic_cast<const UWVF_RT_FE<D> & > (bfel);
	
	const HDivHighOrderFiniteElement<D> & fel_hdiv = 
	  dynamic_cast<const HDivHighOrderFiniteElement<D> & > (uwvffe.GetRTFE() );
	const L2HighOrderFiniteElement & fel_l2 = 
	  dynamic_cast<const L2HighOrderFiniteElement &> (uwvffe.GetL2FE() );
	
	// number of dofs, containing neighbors
	int nd = bfel.GetNDof();
	// ndof of facets on this element only
	int nd_loc_facet = uwvffe.GetNDof(bfel.ElementType(), uwvffe.GetRTOrder());
	// ndof for hdiv-element
	int nd_hdiv = fel_hdiv.GetNDof();
	// ndof for l2-element
	int nd_l2 = fel_l2.GetNDof();
	int nd_loc = nd_hdiv + nd_l2;
	
	// matrix storing shape functions for u, (0, u)
	FlatVector<> mat_uy(nd_loc, lh);
	// matrix storing shape functions for sigma, (sigma, 0)
	FlatMatrixFixWidth<D> mat_sy(nd_loc, lh);
	
	// shapes for Pi X
	FlatVector <> mat_x(nd, lh);
	// shapes for Y
	FlatVector <> mat_y(nd_loc_facet, lh);
	// values for Fy
	FlatVector <Complex> mat_Fy(nd_loc, lh);
	// values for vector (sigma_n, u)
	FlatVector <> mat_usigman(nd_loc, lh);
	
	const ElementTransformation &eltrans = sip.GetTransformation();
	const_cast<UWVF_RT_FE<D> &> (uwvffe).CalcFMatrix(eltrans, lh);
	const FlatMatrix<Complex> & fmatrix = uwvffe.GetFMatrix();
	
	uwvffe.CalcScalarShape ( sip.IP(), mat_uy );
	
	mat = 0.0;
	
	for ( int i = 0; i < nd_loc_facet; i++ )
	  for ( int j = nd_hdiv; j < nd_loc; j++ )
	    mat(0,i) += (fmatrix(j,i)) * mat_uy(j);

	
      }
    else if (  dynamic_cast<const UWVF_PW_FE<D> * > (&bfel) )
      {
	// *** PLANE WAVE ELEMENT
	const UWVF_PW_FE<D> & uwvffe = 
	  dynamic_cast<const UWVF_PW_FE<D> & > (bfel);
	
	mat = 0.0;
	
	int nd_loc_facet = uwvffe.GetNDofLocal();

	FlatVector<Complex> shape(nd_loc_facet, lh);
	uwvffe.CalcInnerShape(sip, shape, lh);

	for ( int i = 0; i < nd_loc_facet; i++ )
	  mat(0,i) += shape(i);

      }
    else if (  dynamic_cast<const UWVF_PWJ_FE<D> * > (&bfel) )
      {
	// *** PLANE WAVE ELEMENT
	const UWVF_PWJ_FE<D> & uwvffe = 
	  dynamic_cast<const UWVF_PWJ_FE<D> & > (bfel);
	
	mat = 0.0;
	
	int nd_loc_facet = uwvffe.GetNDofLocal();

	FlatVector<Complex> shape(nd_loc_facet, lh);
	uwvffe.CalcInnerShape(sip, shape, lh);

	for ( int i = 0; i < nd_loc_facet; i++ )
	  mat(0,i) += shape(i);

      }

    lh.CleanUp(heapp);
  }
};

// diff op calculating sigma = v from element vector for X

template <int D>
class DiffOpSigmaHelmholtz : public DiffOp<DiffOpSigmaHelmholtz<D> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = D };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    ;
  }

  template <typename FEL, typename SIP>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      FlatMatrixFixHeight<D, Complex> & mat, LocalHeap & lh)
  {
    if ( dynamic_cast<const UWVF_RT_FE<D>*> (&bfel) )
      {
    const UWVF_RT_FE<D> & uwvffe = 
      dynamic_cast<const UWVF_RT_FE<D> & > (bfel);

    const HDivHighOrderFiniteElement<D> & fel_hdiv = 
      dynamic_cast<const HDivHighOrderFiniteElement<D> & > (uwvffe.GetRTFE() );
    const L2HighOrderFiniteElement & fel_l2 = 
      dynamic_cast<const L2HighOrderFiniteElement &> (uwvffe.GetL2FE() );

    int nd = bfel.GetNDof();
    int nd_loc_facet = uwvffe.GetNDof(bfel.ElementType(), uwvffe.GetRTOrder());
    int nd_hdiv = fel_hdiv.GetNDof();
    int nd_l2 = fel_l2.GetNDof();
    int nd_loc = nd_hdiv + nd_l2;

    FlatVector<> mat_uy(nd_loc, lh);
    FlatMatrixFixWidth<D> mat_sy(nd_loc, lh);

    FlatVector <> mat_x(nd, lh);
    FlatVector <> mat_y(nd_loc_facet, lh);
    FlatVector <Complex> mat_Fy(nd_loc, lh);
    FlatVector <> mat_usigman(nd_loc, lh);

    const ElementTransformation &eltrans = sip.GetTransformation();
    const_cast<UWVF_RT_FE<D> &> (uwvffe).CalcFMatrix(eltrans, lh);
    const FlatMatrix<Complex> & fmatrix = uwvffe.GetFMatrix();

    uwvffe.CalcRTShape ( sip, mat_sy, lh );

    mat = 0.0;

    for ( int i = 0; i < nd_loc_facet; i++ )
      for ( int j = 0; j < nd_hdiv; j++ )
	for ( int n = 0; n < D; n++ )
	  mat(n,i) -= (fmatrix(j,i)) * mat_sy(j,n);
      }

    else if ( dynamic_cast<const UWVF_PW_FE<D> *> (&bfel) )

      {
	const UWVF_PW_FE<D> & uwvffe = 
	  dynamic_cast<const UWVF_PW_FE<D> & > (bfel);

	mat = 0.0;

	int nd_loc_facet = uwvffe.GetNDofLocal();

	FlatMatrixFixHeight<D,Complex> shape(nd_loc_facet, &mat(0,0));
	uwvffe.CalcSigmaShape(sip, shape, lh);

      }

    else if ( dynamic_cast<const UWVF_PWJ_FE<D> *> (&bfel) )

      {
	const UWVF_PWJ_FE<D> & uwvffe = 
	  dynamic_cast<const UWVF_PWJ_FE<D> & > (bfel);

	mat = 0.0;

	int nd_loc_facet = uwvffe.GetNDofLocal();

	FlatMatrixFixHeight<D,Complex> shape(nd_loc_facet, &mat(0,0));
	uwvffe.CalcSigmaShape(sip, shape, lh);

      }
  }
};



#endif
