#include <solve.hpp>
#include <comp.hpp>


namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;






  /* *************************** Numproc BVP ********************** */


  ///
  class NumProcBVPTest : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa, * bfu,  * bfsigma;
    ///
    LinearForm * lff;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    int maxsteps, minsteps;
    ///
    double prec;
    ///
    bool print;
    ///
    enum SOLVER { CG, GMRES, QMR, NCG };
    ///
    SOLVER solver;

    string number;
  public:
    ///
    NumProcBVPTest (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcBVPTest();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcBVPTest (pde, flags);
    }


    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "Boundary Value Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Linear-form   = " << lff->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl
	  << "solver        = " << ((solver==CG) ? "CG" : "QMR") << endl
	  << "precision     = " << prec << endl
	  << "maxsteps      = " << maxsteps << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcBVPTest :: NumProcBVPTest (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    bfsigma = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformsigma", ""));
    bfu = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformu", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    prec = flags.GetNumFlag ("prec", 1e-12);
    solver = CG;
    if (flags.GetDefineFlag ("qmr")) solver = QMR;
    if (flags.GetDefineFlag ("gmres")) solver = GMRES;
    if (flags.GetDefineFlag ("ncg")) solver = NCG;
    print = flags.GetDefineFlag ("print");

    number = flags.GetStringFlag("number","");
  }

  NumProcBVPTest :: ~NumProcBVPTest()
  {
    ;
  }

  void NumProcBVPTest :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc BVP:\n" \
      "------------\n" \
      "Solves the linear system resulting from a boundary value problem\n\n" \
      "Required flags:\n" 
      "-bilinearform=<bfname>\n" 
      "    bilinear-form providing the matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "\nOptional flags:\n"
      "-preconditioner=<prename>\n"
      "-maxsteps=n\n"
      "-prec=eps\n"
      "-qmr\n"
      "-gmres\n"
      "-print\n"
      "    write matrices and vectors into logfile\n"
	<< endl;
  }


  void NumProcBVPTest :: Do(LocalHeap & lh)
  {
    static int solvetimer = NgProfiler::CreateTimer ("Equation solving");
    NgProfiler::RegionTimer reg (solvetimer);



    if (!lff->IsAssembled()) lff->Assemble(lh);

    cout << "solve bvp" << endl;

    const BaseMatrix & mat = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();


    if (print)
      {
	(*testout) << "MatrixHeight = " << endl << mat.VHeight() << endl;
	(*testout) << "MatrixWidth = " << endl << mat.VWidth() << endl;
	(*testout) << "Matrix = " << endl << mat << endl;
	(*testout) << "RHS-Vector = " << endl << vecf << endl;
      }

    const BaseMatrix * premat = NULL;
    if (pre)  premat = &(pre->GetMatrix());
    KrylovSpaceSolver * invmat;


    if (!bfa->GetFESpace().IsComplex())
      {
	switch (solver)
	  {
	  case CG:
	    cout << "cg solve for real problem" << endl;
	    invmat = new CGSolver<double>(mat, *premat);
	    break;
	  case QMR:
	    invmat = new QMRSolver<double>(mat, *premat);
	    break;
	  case GMRES:
	    invmat = new GMRESSolver<double>(mat, *premat);
	    break;
	  }
      }
    else
      {
	switch (solver)
	  {
	  case CG:
	    invmat = new CGSolver<Complex>(mat, *premat);
	    break;
	  case QMR:
	    invmat = new QMRSolver<Complex>(mat, *premat);
	    break;
	  case GMRES:
	    invmat = new GMRESSolver<Complex>(mat, *premat);
	    break;
	  }
      }

    ma.PushStatus ("Iterative solver");

    invmat->SetMaxSteps (maxsteps);
    invmat->SetPrecision (prec);
    invmat->SetPrintRates ();
    invmat->SetInitialize (0);
    invmat->SetStatusHandler(ma);

    clock_t starttime, endtime, soltime;
    starttime = clock();

    invmat->Mult (vecf, vecu);

    ma.PopStatus ();
    
    if (print)
      (*testout) << "Solution = " << endl << vecu << endl;

    endtime = clock();
    cout << "Solution time = " << double(endtime - starttime)/CLOCKS_PER_SEC << endl;
    cout << "Iterations: " << invmat->GetSteps() << endl;

    delete invmat;

    bfa -> ComputeInternal (vecu, lh);

    if (print)
      (*testout) << "Solution, with internal = " << endl << vecu << endl;

    string file = "/home/astrid/solution_sigma";
    file += number;
    file += ".m";
    string fileu = "/home/astrid/solution_u";
    fileu += number;
    fileu += ".m";

    ofstream astridout(fileu.c_str()); 
    ofstream astridouts(file.c_str()); 


	    
	      
    // plot values along line p1-p2

    int i, j, k;

    int numpoints = 10;
    int numpoints2 = 10;
    Vec<2> point1, point2;
    point1(0) = 0.2; point2(0) = 0.8;

    const BilinearFormIntegrator & bfiu = (bfu) ? *bfu->GetIntegrator(0) : *gfu->GetFESpace().GetEvaluator();
    const BilinearFormIntegrator & bfis = (bfsigma) ? *bfsigma->GetIntegrator(0) : *gfu->GetFESpace().GetEvaluator();

	void * heapp = lh.GetPointer();

    bool applyd = 0;
    for (j = 0; j <= numpoints2; j++)	
      {
	point1(1) = 0.6 * j/ numpoints2 + 0.2;
	point2(1) = 0.6 * j/ numpoints2 + 0.2;
	for (i = 0; i <= numpoints; i++)
	  {
	    lh.CleanUp( heapp);
	    FlatVector<double> p(2, lh);
	    p = point1 + (double(i)/numpoints) * (point2-point1);
	    
	    FlatVector<double> pflux;
	    bool ok =
	      CalcPointFlux (ma, *gfu, p,
			     pflux, bfiu, applyd, lh);
	    
	    if (i==0 && j==0)
	      {
		astridout << "u" << number << " = [ ";
// 		for (k = 0; k < p.Size(); k++) astridout << "coord" << k+1 << " ";
// 		for (k = 0; k < pflux.Size(); k++) astridout << "flux" << k+1 << " ";
		astridout << endl;
	      }
	    
// 	    for (k = 0; k < p.Size(); k++)
// 	      astridout << p(k) << " ";
	    for (k = 0; ok && k < pflux.Size(); k++)
	      astridout << pflux(k) << ", ";
	    astridout << endl;

	      CalcPointFlux (ma, *gfu, p,
			     pflux, bfis, applyd, lh);
	      if (i==numpoints && j==numpoints2)
	      {
		astridout << " 0, 0];";
		astridout << endl;
	      }

	    if (i==0 && j==0)
	      {
		astridouts << "sigma" << number << " = [ ";
// 		for (k = 0; k < p.Size(); k++) astridouts << "coord" << k+1 << " ";
// 		for (k = 0; k < pflux.Size(); k++) astridouts << "flux" << k+1 << " ";
		astridouts << endl;
	      }
	    
// 	    for (k = 0; k < p.Size(); k++)
// 	      astridouts << p(k) << " ";
	    for (k = 0; ok && k < pflux.Size(); k++)
	      astridouts << pflux(k) << ", ";
	    astridouts << endl;

	      if (i==numpoints && j==numpoints2)
	      {
		astridouts << " 0, 0, 0];";
		astridouts << endl;
	      }

	  }
      }


    /*


    CompoundFESpace     compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HCurlHighOrderFESpace & vspace = dynamic_cast<const HCurlHighOrderFESpace & > ( *compound_fespace[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const FESpaceHybrid & pspace = dynamic_cast < const FESpaceHybrid & > (*compound_fespace[2]);

    MeshAccess  ma =  bfa -> GetFESpace(). GetMeshAccess();

    int nd_sigma = sigmaspace.GetNDof();
    int nd_u = vspace.GetNDof();
    int levels = 0;

    Vec<2> point0, point1, point2;
    ARRAY<int> vnums;

    for( int el=0; el < ma.GetNE(); el++)
      {
	void * heapp = lh.GetPointer();

	const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetFE(el, lh));
		
	const HCurlHighOrderFiniteElement<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<2> & > (fel[1]);

	
	const HDivSymFiniteElement<2> & fel_sigma = dynamic_cast< const HDivSymFiniteElement<2> & > (fel[0]);


	ARRAY<int> dofsu (fel_u.GetNDof());
	ARRAY<int> dofssigma (fel_sigma.GetNDof());

	ElementTransformation eltrans;
	ma.GetElementTransformation ( el, eltrans, lh);
		
		
	vspace.GetDofNrs (el, dofsu);
	sigmaspace.GetDofNrs (el, dofssigma);

	FlatMatrixFixHeight<2> u(fel_u.GetNDof(), lh);
	FlatMatrixFixHeight<3> sigma(fel_sigma.GetNDof(), lh);

	IntegrationPoint ip (1.0 / 3.0, 1.0 / 3.0);
	SpecificIntegrationPoint<2,2> sip( ip, eltrans, lh );
		
	double det = fabs (sip.GetJacobiDet());

	DiffOpIdEdge<2> :: GenerateMatrix (fel_u, sip, u, lh);
		
	DiffOpIdHDivSym<2> :: GenerateMatrix (fel_sigma, sip, sigma, lh);

	Vec<3> sigmavalue = 0, sigmasolution;
	Vec<2> uvalue = 0, usolution;

	for( int i=0; i<dofssigma.Size(); i++)
	  for(int j=0; j<3; j++)
	  {
	    sigmavalue(j) += vecu.FVDouble() (dofssigma[i]) * sigma(j,i);
	  }

	for( int i=0; i<dofsu.Size(); i++)
	  for( int j=0; j<2; j++)
	    uvalue (j) += vecu.FVDouble() ( nd_sigma + dofsu[i] ) * u(j,i);

	astridout << "sigma " << endl << sigmavalue << "u " << endl << uvalue << endl << endl;
	lh.CleanUp ( heapp ) ;

	ma.GetElVertices (el, vnums);
	ma.GetPoint(vnums[0], point0);
	ma.GetPoint(vnums[1], point1);
	ma.GetPoint(vnums[2], point2);
	
	Vec<2> x = 1.0 / 3.0 * ( point1 + point2 + point2);
	double mu = 0.3334222;
	double lambda = 4.9980e+03;

	sigmasolution = 0;
// 	sigmasolution(2) = 0.1 * x(0) * x(0);
	sigmasolution(0) = 0.01*((lambda + 2*mu) * (1 - 2*x(0)) *(x(1)-x(1)*x(1)) + lambda * (x(0)-x(0)*x(0)) *(1-2*x(1)));
	sigmasolution(1) = 0.01*((lambda + 2*mu) * (1 - 2*x(1)) *(x(0)-x(0)*x(0)) + lambda * (x(1)-x(1)*x(1)) *(1-2*x(0)));
	sigmasolution(2) = 0.02* (mu * 2 * ((x(0) - x(0)*x(0))*(1-2*x(1)) + (1-2*x(0))*(x(1)-x(1)*x(1)) ) );
	usolution = 0;
	usolution(1) = 0.01 * x(0) * (1-x(0)) * x(1) * (1-x(1));
	usolution(0) = 0.01 * x(0) * (1-x(0)) * x(1) * (1-x(1));
	astridout << "sigmasol " << endl << sigmasolution << endl << "usol " << endl << usolution << endl << endl;

      }
*/
  }



















  


  namespace
#ifdef MACOS
  bvp_cpp
#endif
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("bvptest", NumProcBVPTest::Create, NumProcBVPTest::PrintDoc);
    }
    
    
    Init init;
    
  }
  


}
