#include <solve.hpp>
#include <comp.hpp>

using namespace ngsolve;


class ContactMesh
{
protected: 
  Vec<2> a, b;
  double length;
  int ncel;
  int ncvert;
  ARRAY<Vec<2> > vertices;
  double * vert_param;
  int ** eltovert;
  BitArray ContElemToSurfElem;
  int ndof;

  const FESpace * fespace;

public:

  const bool SelemToCelem ( const int selem, const int celem) const
  {
    return ContElemToSurfElem.Test(selem * ncel + celem);
  }


  ContactMesh( const MeshAccess & ma, const FESpace & afespace, const int cbound1, const int cbound2, 
	       const double* aa, const double* bb, const int ancel)
  {
    fespace = &afespace;
    a(0) = 1.0*aa[0]; a(1) = 0.0;//1.0*aa[1];
    b(0) = 1.0*bb[0]; b(1) = 0.0;//1.0*bb[1];
    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    ncel = ancel;
    ncvert = ancel + 1;
    ndof = ncvert;
    Vec<2> normal, tang;
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    int ii = 0;
    vert_param = new double[ncvert];
    vertices.SetSize(ncvert);
    for(int i=0; i<ncvert; i++)
      {
	vertices[i] = a + (length*i/ncel) * tang ;
	vert_param[i] = length * i / ncel;
      }

    eltovert = new int* [ncel];
    for(int i=0; i<ncel; i++)
      {
	eltovert[i] = new int[2];
	eltovert[i][0] = ii;
	ii++;
	eltovert[i][1] = ii;
      }

    int nsel = ma.GetNSE();
    ContElemToSurfElem.SetSize(ncel * nsel);
    ContElemToSurfElem.Clear();
    ARRAY<int> vnums(2), dnums(4);
    double point0[2], point1[2];
    double param;
    for(int i=0; i<nsel; i++)
      {
	if(ma.GetSElIndex(i) == cbound1  || ma.GetSElIndex(i) == cbound2)
	  {
	    ma.GetSElVertices (i, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
 	    fespace->GetSDofNrs( i, dnums );
	    param = (point0[0] - a(0));//*tang(0) + ( point0[1] - a(1)) * tang(1);
	    int cel1  = int ( param * ncel /length) ;
	    param = (point1[0] - a(0));//*tang(0) + ( point0[1] - a(1)) * tang(1);
	    int cel2  = int ( param * ncel /length) ;
	    if (cel2 < cel1) { int help = cel2; cel2=cel1; cel1 = help;}
	    if(cel2 == ncel) cel2 = ncel-1;
	    for(int cel =cel1; cel <= cel2; cel ++)
	      {ContElemToSurfElem.Set(i * ncel + cel);}
	    /*
	    for(int j=0; j<ncel; j++)
	      {
		if( param >= vert_param[ eltovert[j][0] ] && param <= vert_param[ eltovert[j][1] ] )
		  //if( param >= length * j / ncel && param <= length * (j+1) / ncel)
		  ContElemToSurfElem.Set(i * ncel + j);	    
		//cout << param << "  " << vert_param[j] << "   " << vert_param[j+1] << "  " << SelemToCelem(i,j) << "  " << SelemToCelem(j,i) << endl;

	      }
	    param = (point1[0] - a(0));//*tang(0) + ( point1[1] - a(1)) * tang(1);
	    for(int j=0; j<ncel; j++)
	      {
		if( param >= vert_param[ eltovert[j][0] ] && param <= vert_param[ eltovert[j][1] ] )
		  //if(param >= length * j / ncel && param <= length * (j+1) / ncel && !ContElemToSurfElem.Test(i * ncel + j) )
		  ContElemToSurfElem.Set(i * ncel + j);
		//cout << param << "  " << vert_param[j] << "   " << vert_param[j+1] << "  " << SelemToCelem(i,j) << "  " << SelemToCelem(j,i) << endl;

	      }
	    */
	  }
      }


  }

  ~ContactMesh()
  { for(int i=0; i<ncel; i++) delete eltovert[i];
    delete eltovert, vertices; }

  void GetDomain(Vec<2> & aa, Vec<2> & bb) const
  {
    aa = a; bb = b;
  }

  const int GetNCE() const
  {
    return ncel;
  }

  const int GetNDof() const
  {
    return ndof;
  }

  const FESpace* GetFESpace() const
  {
    return fespace;
  }
  const double GetParam(const int numvert) const
  {
    return vert_param[numvert];
  }

  const double GetCElVertices(const int cel, int & vnum1, int & vnum2) const
  {
    vnum1 = eltovert[cel][0]; vnum2 = eltovert[cel ][1];
    vnum1 = cel; vnum2 = cel+1;
  }
};




class NumProcContact : public NumProc
{
protected:
  // bilinear-form for the Hellinger Reissner system of elasticity
  BilinearForm * bfa;
  // linear-form providing the right hand side, contains (0, f)
  LinearForm * lff;
  // solution vector, containes (sigma, u)
  GridFunction * gfu;

  //contact surfaces
  int cbound1, cbound2;

  Preconditioner * pre;

  double taun, tautau;
  int numuzawasteps;
  int contactelements;
  double coeff_f;

public:
  
    
  /*
    In the constructor, the solver class gets the flags from the pde - input file.
    the PDE class apde constains all bilinear-forms, etc...
  */
  NumProcContact (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {

    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", "a"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    cbound1 = int(flags.GetNumFlag ( "cbound1" , 1.0));
    cbound2 = int(flags.GetNumFlag ( "cbound2" , 2.0));

    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);

    taun = flags.GetNumFlag ( "tau", 1e-5);
    tautau = taun;

    taun = flags.GetNumFlag ( "taun", taun);
    tautau = flags.GetNumFlag ( "tautau", taun);

    numuzawasteps = int(flags.GetNumFlag( "usteps", 10));
    contactelements = int(flags.GetNumFlag( "cels", 10));

    coeff_f = flags.GetNumFlag( "friction", 1.0);
  }


  virtual ~NumProcContact() 
  { ; }


  // creates an solver object
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcContact (pde, flags);
  }



  void Assembleg ( const MeshAccess & ma, const ContactMesh & cmesh, 
		   const CompoundFESpace & compound_fespace, Vector<double> & vecg)
  {
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    vecg.SetSize(cmesh.GetNDof());
    vecg = 0.0;
    vecg(0) = 1;
    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double param_u1, param_u0, param_c0, param_c1;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;

    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    for(int sel =0; sel<nsel; sel++)
      {
	// boundary in direction +normal
	if(ma.GetSElIndex(sel) == cbound1 || ma.GetSElIndex(sel) == cbound2)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    double fac = 1;
	    if(point0[0] > point1[0])
	      {
		int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		fac = -1;
	      }
	    param_u0 = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
	    param_u1 = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);
	    for(int cvert=0; cvert < ncel+1; cvert++)
	      {
		double param_c = cmesh.GetParam(cvert);
		if(param_u0 < param_c && param_u1 >= param_c)
		  {
		    vecg[cvert] += (point0[1] + (param_c - param_u0) / (param_u1 - param_u0) * (point1[1] - point0[1])) *fac;
		  }
	      }

	  }
      }
  

  }

  void ApplyCTranspose( const MeshAccess & ma, const ContactMesh & cmesh, const CompoundFESpace & compound_fespace,
			const Vector<Vec<2> > & w, const Vector<double> & lambda, BaseVector & rhs,LocalHeap & lh)
  {
   
    //rhs.FVDouble() = 1.0;
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    rhs.FVDouble() = 0.0;

    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double x_u1, x_u0, x_c0, x_c1;
    Vec<2> x_s, x_c, i_s, i_c;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;
    double len;
    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);

     for(int cel =0; cel<ncel; cel++)
       {
	 cmesh.GetCElVertices(cel, cvnums[0], cvnums[1]);
	 
	 x_c(0) = cmesh.GetParam(cvnums[0] );
	 x_c(1) = cmesh.GetParam(cvnums[1]);

	 for(int sel=0; sel<nsel; sel++)
	   {
	     if(cmesh.SelemToCelem(sel,cel))
	       {
	        	     lh.CleanUp();

		 const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

		 const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
		 const HCurlHighOrderFiniteElement<1> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<1> & > (fel[1]);
		 ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

		 ma.GetSElVertices (sel, vnums);
		 ma.GetPoint(vnums[0], point0);
		 ma.GetPoint(vnums[1], point1);
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);
		 double fac = 1;
		 if(point0[0] > point1[0])
		   {
		     int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		     ma.GetPoint(vnums[0], point0);
		     ma.GetPoint(vnums[1], point1);
		     fac = -1;
		   }
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);

		 GetIntegrationBounds(x_c, x_s, i_c, i_s, len);

		 if (len > 0 )
		   {
		 //if(len > 0) cout << cmesh.SelemToCelem(sel,cel) << endl;
		 const IntegrationRule & ir1d =
		   GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4);

		 // teil avg(sigma nn) * w(0) + jump(sigma_nn) * w(1);
		 sigmaspace.GetSDofNrs (sel, dofs);
		 //cout << "sigmadofs " << dofs << endl;
		 Vector<double> sigma(fel_sigma.GetNDof());
		 for(int ir=0; ir<ir1d.GetNIP(); ir++)
		   {
		     double  point;
		     if(fac < 0) point  = i_s(0) + ir1d[ir](0) * (i_s(1) - i_s(0)) ;
		     else {point = i_s(1) + ir1d[ir](0) * (i_s(0) - i_s(1)); point = 1 - point;}
		     IntegrationPoint ip( point );
		     double ipc = i_c(0) + ir1d[ir](0) * (i_c(1) - i_c(0));
		     fel_sigma.CalcShape(ip, sigma);
		     sigma *= 1.0 / (sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1]));
;
      		     for(int i=0; i<fel_sigma.GetNDof(); i++)
		       {
			 rhs.FVDouble()(dofs[i]) += sigma(i) * (1 - ipc) * 0.5 *
			   (-w(cel)(1)*fac - w(cel)(0) ) * len * ir1d[ir].Weight();
			 rhs.FVDouble()(dofs[i]) += sigma(i) * (ipc) * 0.5 *
			   (-w(cel+1)(1)*fac - w(cel+1)(0) ) * len * ir1d[ir].Weight();
			 //sigma = -fac * sigma;
		       }
		   }
		 // teil jump(u_tau) * lambda
		 uspace.GetSDofNrs (sel, dofsu);
		 //cout << dofs << endl;
		 FlatMatrix<> u(1, fel_u.GetNDof(), lh);
     		 for(int ir=0;  ir<ir1d.GetNIP(); ir++)
		   {
		     double  point;
// 		     if(fac < 0) point  = i_s(0) + ir1d[ir](0) * (i_s(1) - i_s(0)) ;
// 		     else {point = i_s(1) + ir1d[ir](0) * (i_s(0) - i_s(1)); point = 1 - point;}
		     if(fac < 0) point  = i_s(0) + ir1d[ir](0) * (i_s(1) - i_s(0)) ;
		     else {point = i_s(1) + ir1d[ir](0) * (i_s(0) - i_s(1)); point = 1 - point;}

		     IntegrationPoint ip( point );
		     double ipc = i_c(0) + ir1d[ir](0) * (i_c(1) - i_c(0));
		     fel_u.CalcShape(ip, u);
		     u *= 1.0 / sqrt(sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1]));
		     for(int i=0; i<fel_u.GetNDof(); i++)
		       {
			 rhs.FVDouble()(dofsu[i] + ndof_sigma_total) += u(0,i) * (1-ipc) * (lambda(cel) )   *len * ir1d[ir].Weight() ;
			 rhs.FVDouble()(dofsu[i] + ndof_sigma_total) += u(0,i) * ipc * (lambda(cel) ) *len * ir1d[ir].Weight() ;
		       }
		   } 
		   }
	       }
	   }


       }
     lh.CleanUp();
  }


  void GetIntegrationBounds(const Vec<2> & x_c, const Vec<2> & x_s, Vec<2> & i_c, Vec<2> &i_s, double & len)
  {

    if(x_c(0) < x_s(0))
      {
	i_s(0) = 0;
	i_c(0) = (x_s(0) - x_c(0)) / (x_c(1) - x_c(0));
      }
    else
      {
	i_s(0) = (x_c(0) - x_s(0)) / (x_s(1) - x_s(0));
	i_c(0) = 0;
      }
    if(x_c(1) > x_s(1))
      {
	i_s(1) = 1;
	i_c(1) = (x_s(1) - x_c(0)) / (x_c(1) - x_c(0));
      }
    else
      {
	i_s(1) = (x_c(1) - x_s(0)) / (x_s(1) - x_s(0));
	i_c(1) = 1;
      }
    len = (x_c(1) - x_c(0)) * (i_c(1) - i_c(0));
  }



  void ApplyCn ( const MeshAccess & ma, const ContactMesh & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, Vector<Vec<2> > & updatew)
  {
    LocalHeap lh(100000);
    lh.CleanUp();
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    updatew.SetSize(cmesh.GetNDof());
    updatew = 0.0;

    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double x_u1, x_u0, x_c0, x_c1;
    Vec<2> x_s, x_c, i_s, i_c;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;
    double len;
    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

     for(int cel =0; cel<ncel; cel++)
       {
	 cmesh.GetCElVertices(cel, cvnums[0], cvnums[1]);
	 
	 x_c(0) = cmesh.GetParam(cvnums[0] );
	 x_c(1) = cmesh.GetParam(cvnums[1]);

	 for(int sel=0; sel<nsel; sel++)
	   {
	     if(cmesh.SelemToCelem(sel,cel))
	       {
	           lh.CleanUp();

		 const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

		 const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
		 ARRAY<int> dofs(fel_sigma.GetNDof());

		 ma.GetSElVertices (sel, vnums);
		 ma.GetPoint(vnums[0], point0);
		 ma.GetPoint(vnums[1], point1);
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);
		 double fac = 1;
		 if(point0[0] > point1[0])
		   {
		     int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		     ma.GetPoint(vnums[0], point0);
		     ma.GetPoint(vnums[1], point1);
		     fac = -1;
		   }
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);

		 GetIntegrationBounds(x_c, x_s, i_c, i_s, len);
		 if(len > 0)
		   {
		 const IntegrationRule & ir1d =
		   GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4);

		 sigmaspace.GetSDofNrs (sel, dofs);
		 Vector<double> sigma(fel_sigma.GetNDof());
		 for(int ir=0; ir<ir1d.GetNIP(); ir++)
		   {
		     double  point;
		     if(fac < 0) point  = i_s(0) + ir1d[ir](0) * (i_s(1) - i_s(0)) ;
		     else{point = i_s(1) + ir1d[ir](0) * (i_s(0) - i_s(1)); point = 1 - point;}
		     IntegrationPoint ip( point );
		     double ipc = i_c(0) + ir1d[ir](0) * (i_c(1) - i_c(0));
		     fel_sigma.CalcShape(ip, sigma);
		     for(int i=0; i<fel_sigma.GetNDof(); i++)
		       {			 
			 double sig =  vecu.FVDouble()(dofs[i])  * sigma(i); 
			 updatew(cel)(0) +=  (1-ipc) * sig * len * ir1d[ir].Weight();
			 updatew(cel)(1) +=  (1-ipc) * sig * len * ir1d[ir].Weight()*fac;
			 updatew(cel+1)(0) +=  ipc * sig * len * ir1d[ir].Weight();
			 updatew(cel+1)(1) +=  ipc * sig * len * ir1d[ir].Weight()*fac;
			 //sigma = -fac * sigma;

		       }
		   }
		   }

	       }
	   }


       }
  }


  void ApplyCtau (const MeshAccess & ma, const ContactMesh & cmesh, const CompoundFESpace & compound_fespace,
		  const BaseVector & vecu, Vector<double > & updatelambda)
  {
    LocalHeap lh(100000);
    lh.CleanUp();
    int ndof_sigma_total = compound_fespace[0]->GetNDof();

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    updatelambda.SetSize(cmesh.GetNDof());
    updatelambda = 0.0;

    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double x_u1, x_u0, x_c0, x_c1;
    Vec<2> x_s, x_c, i_s, i_c;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;
    double len;
    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);

     for(int cel =0; cel<ncel; cel++)
       {
	 cmesh.GetCElVertices(cel, cvnums[0], cvnums[1]);
	 
	 x_c(0) = cmesh.GetParam(cvnums[0] );
	 x_c(1) = cmesh.GetParam(cvnums[1]);

	 for(int sel=0; sel<nsel; sel++)
	   {
      	     if(cmesh.SelemToCelem(sel,cel))
	       {
	            lh.CleanUp();

		 const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		 const HCurlHighOrderFiniteElement<1> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<1> & > (fel[1]);
		 ARRAY<int> dofs(fel_u.GetNDof());

		 ma.GetSElVertices (sel, vnums);
		 ma.GetPoint(vnums[0], point0);
		 ma.GetPoint(vnums[1], point1);
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);

		 double fac = 1;
		 if(point0[0] > point1[0])
		   {
		     int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		     ma.GetPoint(vnums[0], point0);
		     ma.GetPoint(vnums[1], point1);
		     fac = -1;
		   }
		 x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
		 x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);

		 GetIntegrationBounds(x_c, x_s, i_c, i_s, len);
		 const IntegrationRule & ir1d =
		   GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4);

		 uspace.GetSDofNrs (sel, dofs);
		 FlatMatrix<> u(fel_u.GetNDof(), 1, lh);
		 for(int ir=0; ir<ir1d.GetNIP(); ir++)
		   {
		     double  point;
		     if(fac < 0) point  = i_s(0) + ir1d[ir](0) * (i_s(1) - i_s(0)) ;
		     else {point = i_s(1) + ir1d[ir](0) * (i_s(0) - i_s(1)); point = 1 - point;}
		     IntegrationPoint ip( point );
		     double ipc = i_c(0) + ir1d[ir](0) * (i_c(1) - i_c(0));
		     fel_u.CalcShape(ip, u);
		     u *= 1.0 / sqrt(sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1]));

		     for(int i=0; i<fel_u.GetNDof(); i++)
		       {
			 
			 double sig = u(i,0) * vecu.FVDouble()(dofs[i] + ndof_sigma_total);
			 updatelambda(cel) += (1-ipc) * sig * len * ir1d[ir].Weight() ;
			 updatelambda(cel+1) += (ipc) * sig * len * ir1d[ir].Weight() ;//* fac;
		       }
		   }
		 
	       }
	   }


       }
  }

  void ProjectDisp (const Vector<double > & utauy,
		    const Vector<double > & vecg, Vector<Vec<2> > & w)
  {
    
    int ndof_g = vecg.Size();

     for(int i=0; i< ndof_g; i++)
       {
	 if ( w(i)(0) + utauy(i) > vecg(i))// + vecurel(i) )
	   w(i)(0) = vecg(i) - utauy(i);// + vecurel(i);

       }

  }


  void ProjectStress ( const double coeff_f, const Vector<double> & sigmann, Vector<double > & lambda )
  {


    int ndof_c = lambda.Size();
    for(int i=0; i < ndof_c; i++)
      {
	if(lambda(i) > abs(coeff_f * sigmann(i)) ) {lambda(i) = abs(coeff_f * sigmann(i)); }//cout << i << endl;}
	if(lambda(i) < -abs(coeff_f * sigmann(i)) ) {lambda(i) = -abs(coeff_f * sigmann(i));}// cout << i << endl;}
//     lambda(i)(1) = 0;

      }


  }


  void CalcSigmann (const MeshAccess & ma, const ContactMesh & cmesh, const CompoundFESpace & compound_fespace,
		      const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh)
  {

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double x_u1, x_u0, x_c0, x_c1;
    Vec<2> x_s, x_c, i_s, i_c;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;

    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);


    sigmann = 0.0;

    for(int sel =0; sel<nsel; sel++)
      {
	// boundary in direction +normal
	if(ma.GetSElIndex(sel) == cbound1 || ma.GetSElIndex(sel) == cbound2)
	  {
	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    ARRAY<int> dofs(fel_sigma.GetNDof());

	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    double fac = 1;
	    if(point0[0] > point1[0])
	      {
		int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		fac = -1;
	      }
	    x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
	    x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);
	    sigmaspace.GetSDofNrs (sel, dofs);
	    FlatVector<> sigma(fel_sigma.GetNDof(), lh);
	    sigmaspace.GetSDofNrs (sel, dofs);

	    for(int cvert=0; cvert < ncel+1; cvert++)
	      {
		double param_c = cmesh.GetParam(cvert);
		if(x_s(0) < param_c && x_s(1) >= param_c)
		  {
		    double  point;
		    if(fac < 0) point  = (param_c - x_s(0)) / (x_s(1) - x_s(0));
		    //else{point = (param_c - x_s(1)) / (x_s(0) - x_s(1)); point = 1 - point;}
		    else break;
		    IntegrationPoint ip( point );
		    fel_sigma.CalcShape(ip, sigma);
		    sigma *= 1.0 / (sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1]));
		    for(int i=0; i<fel_sigma.GetNDof(); i++)
		      {
		        sigmann(cvert) += sigma(i) * vecu.FVDouble()(dofs[i]) ;
// 			vecg[cvert] += (point0[1] + (param_c - param_u0) / (param_u1 - param_u0) * (point1[1] - point0[1])) *fac;
		      }
		  }
	      }

	  }
      }
  


  }




 void ProjectDisp (const MeshAccess & ma, const ContactMesh & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, const Vector<double> & vecg, Vector<Vec<2> > & w, LocalHeap & lh)
  {

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    ARRAY<int> vnums(2), dnums(4);
    int cvnums[2];
    double point0[2], point1[2];

    double x_u1, x_u0, x_c0, x_c1;
    Vec<2> x_s, x_c, i_s, i_c;
    Vec<2> a, b;
    cmesh.GetDomain(a,b);
    Vec<2> normal, tang;
    double length;

    Vector<double> wy(ndof_g), utauy(ndof_g);

    wy = 0.0; utauy = 0.0;

    length = sqrt( sqr(a(0) - b(0)) + sqr(a(1) - b(1)) );
    normal(0) = (a(1) - b(1))/length;
    normal(1) = (b(0) - a(0))/length;

    tang(0) = normal(1);
    tang(1) = -normal(0);

    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);

    int sigma_total_dofs = compound_fespace[0]->GetNDof();
    utauy = 0.0;

    for(int sel =0; sel<nsel; sel++)
      {
	// boundary in direction +normal
	if(ma.GetSElIndex(sel) == cbound1 || ma.GetSElIndex(sel) == cbound2)
	  {
	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HCurlHighOrderFiniteElement<1> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<1> & > (fel[1]);
	    ARRAY<int> dofs(fel_u.GetNDof());

	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    double fac = 1;
	    if(point0[0] > point1[0])
	      {
		int help = vnums[1]; vnums[1] = vnums[0]; vnums[0] = help;
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		fac = -1;
	      }
	    x_s(0) = (point0[0] - a(0))*tang(0) + ( point0[1] - a(1)) * tang(1);
	    x_s(1) = (point1[0] - a(0))*tang(0) + ( point1[1] - a(1)) * tang(1);
	    double tauy_elem = point1[1] - point0[1];
	    double ny_elem = point0[0] - point1[0];
	    if (fac < 0) {tauy_elem = -tauy_elem; ny_elem = -ny_elem;}
	    uspace.GetSDofNrs (sel, dofs);
	    FlatMatrixFixWidth<1> u(fel_u.GetNDof(), lh);
	    uspace.GetSDofNrs (sel, dofs);

	    for(int cvert=0; cvert < ncel+1; cvert++)
	      {
		double param_c = cmesh.GetParam(cvert);
		if(x_s(0) < param_c && x_s(1) >= param_c)
		  {
		    double  point;
		    if(fac < 0) point  = (param_c - x_s(0)) / (x_s(1) - x_s(0));
		    else{point = (param_c - x_s(1)) / (x_s(0) - x_s(1)); point = 1 - point;}
		    //else break;
		    IntegrationPoint ip( point );
		    fel_u.CalcShape(ip, u);
		    // factor (tau_elem)_y / norm(tau_elem) / h
		    u *= tauy_elem / (sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1]));
		    for(int i=0; i<fel_u.GetNDof(); i++)
		      {
		        utauy(cvert) -= u(i,0) * vecu.FVDouble()(sigma_total_dofs + dofs[i])*fac;
// 			vecg[cvert] += (point0[1] + (param_c - param_u0) / (param_u1 - param_u0) * (point1[1] - point0[1])) *fac;
		      }
		    wy(cvert) -= (0.5 * w(cvert)(0)*fac - 0.5*  w(cvert)(1)) * ny_elem / sqrt(sqr(point0[0] - point1[0]) + sqr(point1[1] - point0[1])) ;
		  }
	      }

	  }
      }
  
    //    cout << L2Norm(wy) << "     " << L2Norm(vecg) << endl;
    
    for(int i=0; i< ndof_g; i++)
       {
	 //cout << wy(i) << "  " <<  w(i)(0) << endl;
	 if ( wy(i) + utauy(i) > vecg(i))// + vecurel(i) )
	   w(i)(0) = (vecg(i) - utauy(i))/wy(i) * w(i)(0);// + vecurel(i);
       }

    //cout << endl << endl;

  }







  virtual void Do(LocalHeap & lh)
  {
    cout << "solve contact problem" << endl;
      

    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();

    CompoundFESpace     compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HCurlHighOrderFESpace & vspace = dynamic_cast<const HCurlHighOrderFESpace & > ( *compound_fespace[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

    MeshAccess  ma =  bfa -> GetFESpace(). GetMeshAccess();

    int ndof, nd_u, nd_sigma;


    ndof = compound_fespace.GetNDof();
    nd_u = compound_fespace[1]->GetNDof();
    nd_sigma = compound_fespace[0]->GetNDof();

    ///
    int maxsteps = 3;
    ///
    double prec = 1e-5;
    ///
    bool print=false;
    ///
    enum SOLVER { CG, GMRES, QMR, NCG };
    ///
    SOLVER solver = GMRES;

    // ********* in short **************
    // make contact mesh

    // find gap function g

    // solve variational inequality by uzawa type method

    // given w_0, lambda_0

    // while ( error > eps or n < nmax or whatever)

    //   right hand side
    //   rhs = f - (apply_cn(w), apply_ctau(lambda) )

    //   find (sigma, u) by solving system
    //   mata * vecu = rhs

    //   update w, lambda in direction of steepest descent/ascent
    //   w = w - tau * (apply_cn(sigma))
    //   lambda = lambda + tau * apply_ctau(u)
   
    //   project lambda, w onto admissible sets

    //   w = max ( w, g )

    //   lambda = max(min( lambda, 1), -1);

    // end;

    // ************ details ******************

    // make contact mesh
    int ncel = contactelements;
    double aa[2] = {-1.0 , 0.0 };
    double bb[2] = { 1.0 , 0.0 };

    ContactMesh cmesh ( ma, compound_fespace,  cbound1, cbound2, aa, bb,  ncel);
    int nd_c = cmesh.GetNDof();

    // find gap function g
    Vector<double> vecg (nd_c);

    Assembleg ( ma, cmesh,compound_fespace, vecg);
   
    // init lambda_0, w_0
    Vector<double > veclambda(nd_c), lambdaold(nd_c); Vector<Vec<2> > vecw(nd_c), wold(nd_c);
    Vector<double > updatelambda(nd_c); Vector<Vec<2> > updatew(nd_c);
    Vector<double> vecsigmann(nd_c);

    veclambda = 0.0;
    vecsigmann = 0.0;   
     for(int i=0; i<nd_c; i++)
       {
	 vecw(i)(0) = 0;
 	vecw(i)(1) = 0.0;
       }

    // solve by Uzawa type method
    BaseVector & rhs = *vecf.CreateVector();
    const BaseMatrix * premat = NULL;
    if (pre)  premat = &(pre->GetMatrix()); 
    KrylovSpaceSolver * invmat;

	switch (solver)
	  {
	  case CG:
	    cout << "cg solve for real problem" << endl;
	    invmat = new CGSolver<double>(mata, *premat);
	    break;
	  case QMR:
	    invmat = new QMRSolver<double>(mata, *premat);
	    break;
	  case GMRES:
	    invmat = new GMRESSolver<double>(mata, *premat);
	    break;
	  }

    invmat->SetMaxSteps (maxsteps);
    invmat->SetPrecision (prec);
    invmat->SetPrintRates ();
    invmat->SetInitialize (0);
    invmat->SetStatusHandler(ma);

//     clock_t starttime, endtime, soltime;
//     starttime = clock();

    int uzawasteps = 0;
    double elambda = 1.0, ew = 1.0;
    double eps = 1e-9;
    double epsw=0, epslambda =0;

    while( uzawasteps < numuzawasteps && (elambda > epslambda || ew > epsw))
      {
	wold = vecw; 
	lambdaold = veclambda;

    //   right hand side
    //   rhs = f - (apply_cn(w), apply_ctau(lambda) )
	rhs.FVDouble() = 0.0;
	ApplyCTranspose( ma, cmesh, compound_fespace, vecw, veclambda, rhs, lh);
	cout << L2Norm(rhs.FVDouble()) << "  " << L2Norm(vecf.FVDouble()) << endl;
	rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();

    //   find (sigma, u) by solving system
    //   mata * vecu = rhs

	invmat->Mult (rhs, vecu);


    //   update w, lambda in direction of steepest descent/ascent
    //   w = w - tau * (apply_cn(sigma))
    //   lambda = lambda + tau * apply_ctau(u)
	ApplyCn (ma, cmesh, compound_fespace, vecu, updatew);
      	ApplyCtau (ma, cmesh, compound_fespace,  vecu, updatelambda);
	//	vecw -= tau*updatew;
	for(int i = 0; i < nd_c; i++)
	  {
	    vecw(i)(0) -= taun*updatew(i)(0);
	    vecw(i)(1) -= taun*updatew(i)(1);
	    veclambda(i) -= tautau*updatelambda(i);
	  }
	//	veclambda -= tau*updatelambda;
    //   project lambda, w onto admissible set


	ProjectDisp(ma,cmesh, compound_fespace, vecu, vecg, vecw, lh);
	//	cout << vecutauy << endl;
	//ProjectDisp(vecutauy, vecg, vecw);
 	// for(int i = 0; i < nd_c; i++)
//  	  if(vecw(i)(0) > vecg(i))
//  	  {
//  	    vecw(i)(0) = vecg(i);
//  	  }

	CalcSigmann( ma, cmesh, compound_fespace, vecu, vecsigmann, lh );

	ProjectStress ( coeff_f, vecsigmann, veclambda);

    //   lambda = max(min( lambda, 1), -1);
	//	if (uzawasteps > numuzawasteps - 2) cout << veclambda << endl;
	ew = 0; elambda = 0; epsw=0; epslambda=0;
	for(int i=0; i<nd_c; i++)
	  {
	    ew += sqr(vecw(i)(0) - wold(i)(0));
	    epsw += sqr(vecw(i)(0)) * eps;
	    elambda += sqr(veclambda(i) - lambdaold(i));
	    epslambda += sqr(veclambda(i)) * eps;
	  }

	if(uzawasteps % 100 == 0 ) Ng_Redraw();
	uzawasteps ++;
	cout << ew / epsw  << "         " << elambda / epslambda  << endl;
      }

    cout << "lambda, sigma_nn = " << endl;
    for(int i=0; i<veclambda.Size(); i++)
      cout << veclambda(i) << "   " << vecsigmann(i) << endl;
    cout << "numberofsteps " << uzawasteps << endl;
    delete invmat;
  }







  virtual string GetClassName () const
  {
    return " Solver for Contact Problems ";
  }

  virtual void PrintReport (ostream & ost)
  {
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl;
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*    ost << 
      "\n\nNumproc Parabolic:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};





// declare the numproc 'contact' 

namespace
{
  class Init
  { 
  public: 
    Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("contact", NumProcContact::Create, NumProcContact::PrintDoc);
  }
    
  Init init;
}
  
