#include <solve.hpp>
#include <comp.hpp>

using namespace ngsolve;

#include "contact3d.hpp"



// Vec<3> Cross ( Vec<3> veca, Vec<3> vecb)
// {
//   Vec<3> vecc;
//   vecc(0) = veca(1)*vecb(2) - veca(2)*vecb(1);
//   vecc(1) = veca(2)*vecb(0) - veca(0)*vecb(2);
//   vecc(2) = veca(0)*vecb(1) - veca(1)*vecb(0);
//   return vecc;
// }



double InnerProduct ( Vector<double> & a, Vector<double> & b)
{
  double t=0;
  for(int i=0; i<a.Size(); i++)
    t += a(i) * b(i);
  return t;
}

ContactMesh3d :: ContactMesh3d( const MeshAccess & ma, const FESpace & afespace, const int cbound1, const int cbound2, 
				const Vec<3> apoint_0, const Vec<3> apoint_a, const Vec<3> apoint_b, const int ancel_a, const int ancel_b)
  {
    fespace = &afespace;
    point_0 = apoint_0; point_a = apoint_a; point_b = apoint_b;
    ncel_a = ancel_a;
    ncel_b = ancel_b;
    ncel = ncel_a * ncel_b;
    ncvert = (ncel_a + 1)*(ncel_b + 1);
    ndof = ncvert;

    double norma = L2Norm(point_0 - point_a);
    double normb = L2Norm(point_0 - point_b);

    Vec<3> normal = Cross(point_0 - point_a, point_0 - point_b);
    double length = sqrt( sqr(normal(0)) + sqr(normal(1)) + sqr(normal(2)));
    normal = 1.0 / length * normal;

    int ii = 0;
    vertices.SetSize(ncvert);
    for(int i=0; i<ncel_a + 1; i++)
      for(int j=0; j<ncel_b + 1; j++)
      {
	vertices[i + j*(ncel_a+1)] =  point_0 + 1.0*i/ncel_a*(point_a - point_0) + 1.0*j/ncel_b*(point_b - point_0);
      }

    eltovert = new int* [ncel];
    for(int ib=0; ib<ncel_b; ib++)
      {
	for(int ia=0; ia<ncel_a; ia++)
	  {
	    int i = ia + ib*ncel_a;
	    eltovert[i] = new int[4];
	    eltovert[i][0] = ii;
	    ii++;
	    eltovert[i][1] = ii;
	    eltovert[i][2] = ii + ncel_a + 1;
	    eltovert[i][3] = ii + ncel_a;
	  }
	ii++; 
      }

    // Surface Elements matching contact nodes
    ARRAY<int> vnums(3);
    Vec<3> point0, point1, point2, lambda;
    double t;

    int nsel = ma.GetNSE();
    MatchingSEl.SetSize(ncvert);
    for(int cvert=0; cvert<ncvert; cvert++)
      {
	// Vec<3> cpoint = vertices[cvert];
	MatchingSEl[cvert][0] = -1;
	MatchingSEl[cvert][1] = -1;
	for(int sel=0; sel<nsel; sel++)
	  {
	    if(ma.GetSElIndex(sel) == cbound1)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		CalcProjection(vertices[cvert], normal, point0, point1, point2, lambda, t);
		if(lambda(0) >=0 && lambda(0) <=1 && lambda(1) >=0 && lambda(1) <=1 && lambda(2) >=0 && lambda(2) <=1)
		  {
		    MatchingSEl[cvert][0] = sel; 
		  }
	      }

	    if(ma.GetSElIndex(sel) == cbound2)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		CalcProjection(vertices[cvert], normal, point0, point1, point2, lambda, t);
		if(lambda(0) >=0 && lambda(0) <=1 && lambda(1) >=0 && lambda(1) <=1 && lambda(2) >=0 && lambda(2) <=1)
		  {
		    MatchingSEl[cvert][1] = sel; 
		  }
	      }
	  }

      }

  }



const double ContactMesh3d :: EvaluateContactFunction(const Vector<double> & vecf, const Vec<3> & x) const
  {
    int x_cel = 0, x_cel_a, x_cel_b;
    Vec<3> veca = point_a - point_0, vecx = x - point_0, vecb = point_b - point_0;
    double a = vecx(0) * veca(0) + vecx(1) * veca(1) + vecx(2) * veca(2);
    double b = vecx(0) * vecb(0) + vecx(1) * vecb(1) + vecx(2) * vecb(2);

    x_cel_a = int ( a / (sqr(veca(0)) + sqr(veca(1)) + sqr(veca(2))) * ncel_a ) ; if(x_cel_a == ncel_a) x_cel_a --;
    x_cel_b = int ( b / (sqr(vecb(0)) + sqr(vecb(1)) + sqr(vecb(2))) * ncel_b ) ; if(x_cel_b == ncel_b) x_cel_b --;
    x_cel = x_cel_b * ncel_a + x_cel_a;
    ARRAY<int> vnums(4);
    GetCElVertices(x_cel, vnums);
    Vec<3> x0 = GetPoint(vnums[0]), x1 = GetPoint(vnums[1]),x2 = GetPoint(vnums[2]), x3 = GetPoint(vnums[3]);
    Vec<2> pp;
    pp(0) = (x(0) - x0(0))/(x1(0) - x0(0));
    pp(1) = (x(1) - x0(1))/(x3(1) - x0(1));

    double value = vecf(vnums[0])*(1-pp(0))*(1-pp(1)) + 
      vecf(vnums[1])*(1-pp(1))*pp(0) + 
      vecf(vnums[2])*pp(0)*pp(1) + 
      vecf(vnums[3])*pp(1)*(1-pp(0));

   
    return value;

  }


void CalcProjection ( const Vec<3> & xx, const Vec<3> & nn, const Vec<3> & aa, const Vec<3> & bb, const Vec<3> & cc, Vec<3> & lambda, double & t)
{

  Vec<3> ne = Cross( aa - bb, aa - cc);
  double nenn = 0, na=0;
  for(int i=0; i<3; i++)
    {
      nenn += ne(i) * nn(i);
      na += ne(i) * ( aa(i) - xx(i));
    }
  if (nenn == 0)
    {
      t = -1; lambda = -1; return;
    }
  t = 1.0 * na / nenn;

  Vec<3> ss = xx + t * nn;

  lambda(2) = 1.0 * ( (bb(1) - aa(1))*(ss(0) - aa(0)) + (bb(0) - aa(0))*(aa(1) - ss(1)) ) / 
    ( (cc(0) - aa(0))*(bb(1) - aa(1)) - (cc(1) - aa(1))*(bb(0) - aa(0)) );

  lambda(1) = 1.0 * ( ss(1) - aa(1) - (cc(1) - aa(1)) * lambda(2)) / (bb(1) - aa(1));

  lambda(0) = 1.0 - lambda(1) - lambda(2);

}






  void NumProcContact :: Assembleg ( const MeshAccess & ma, const ContactMesh3d & cmesh, 
		   const CompoundFESpace & compound_fespace, Vector<double> & vecg)
  {
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    //    int ncela = cmesh.GetNCEa(), ncelb = cmesh.GetNCEb();

    vecg.SetSize(cmesh.GetNDof());
    vecg = 0.0;

    Vec<3> point_0, point_a, point_b;
    cmesh.GetDomain(point_0, point_a, point_b);
    Vec<3> normal = Cross(point_0 - point_a, point_0 - point_b);
    double length = sqrt( sqr(normal(0)) + sqr(normal(1)) + sqr(normal(2)));
    normal = 1.0 / length * normal;

    ARRAY<int> vnums(3);
    Vec<3> point0, point1, point2, lambda;
    double t;
    int fac =-1;
    for(int cvert = 0; cvert < ndof_g; cvert++)
      {
	Vec<3> xx = cmesh.GetPoint(cvert);
	fac = -1;
	for(int bound =0; bound<=1; bound++)
	  {
	    int sel = cmesh.GetMatchingSEl(cvert, bound);
	    if(sel >= 0)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		CalcProjection(xx, normal, point0, point1, point2, lambda, t);
		if(lambda(0) >=0 && lambda(0) <=1 && lambda(1) >=0 && lambda(1) <=1 && lambda(2) >=0 && lambda(2) <=1)
		  {vecg(cvert) += t*fac; }
	      }
	    else vecg(cvert) += 10.0;
	    fac = 1;
	  }
      }


  }

  void NumProcContact :: ApplyCTranspose( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
			const Vector<double > & w0, const Vector<double> & w1, 
			const Vector<double> & lambdax, const Vector<double> & lambday, const Vector<double> & lambdaz, BaseVector & rhs,
			LocalHeap & lh)
  {
   
    //rhs.FVDouble() = 1.0;
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    rhs.FVDouble() = 0.0;

    ARRAY<int> vnums(3), dnums(4);
    int cvnums[2];
    Vec<3> point0, point1, point2;


    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);

    for(int sel=0; sel<nsel; sel++)
      {
       	if(ma.GetSElIndex(sel) == cbound1  || ma.GetSElIndex(sel) == cbound2)
	  {
	    double fac = 1.0;
	    if (ma.GetSElIndex(sel) == cbound1) fac = -1.0;
	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    const HCurlHighOrderFiniteElement<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<2> & > (fel[1]);

	    ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);

	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    
	    const IntegrationRule & ir2d = 
	      GetIntegrationRules().SelectIntegrationRule(fel.ElementType(), 6);
	    
	    sigmaspace.GetSDofNrs (sel, dofs);
	    uspace.GetSDofNrs (sel, dofsu);

	    Vector<double> sigma(fel_sigma.GetNDof());
	    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);

	    for(int ir=0; ir<ir2d.GetNIP(); ir++)
	      {
		Vec<3> ipx;
		ipx = (1 - ir2d[ir](0) - ir2d[ir](1)) * point0   + ( ir2d[ir](0) )*point1+ ir2d[ir](1)* point2 ;

		SpecificIntegrationPoint<2,3> sip( ir2d[ir], eltrans, lh );

 		double det = fabs (sip.GetJacobiDet());
    		fel_sigma.CalcShape(sip.IP(), sigma);
		sigma *= 1.0 / sqr(det);

		DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);

		
		for(int i=0; i<fel_sigma.GetNDof(); i++)
		  {
		    rhs.FVDouble()(dofs[i]) += sigma(i)  * 0.5 * 
		     ( - cmesh.EvaluateContactFunction(w1, ipx)*fac  -  cmesh.EvaluateContactFunction(w0, ipx) )*
		      det * ir2d[ir].Weight();
		  }	
		for(int i=0; i < fel_u.GetNDof(); i++)
		  {
		    rhs.FVDouble()(ndof_sigma_total + dofsu[i]) += 
		      ( u(0,i) * cmesh.EvaluateContactFunction(lambdax, ipx) + 
			u(1,i) * cmesh.EvaluateContactFunction(lambday, ipx) + 
		       	u(2,i) * cmesh.EvaluateContactFunction(lambdaz, ipx) 
			) * det * ir2d[ir].Weight() *fac;

		  }	  
	      }

	  }
      }
    lh.CleanUp();
  }





  void NumProcContact :: ApplyCn ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, Vector<double> & updatew0, Vector<double> & updatew1, LocalHeap & lh )
  {
    //rhs.FVDouble() = 1.0;
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    updatew0 = 0.0; updatew1 = 0.0;
    ARRAY<int> vnums(3), dnums(4);
    int cvnums[2];
    Vec<3>  point0, point1, point2;
//     point0 = new double (3);
//     point1 = new double (3);
//     point2 = new double (3);

    Vector<double> testvec(updatew0.Size());
    testvec=0.0;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);


    for(int sel=0; sel<nsel; sel++)
      {
	if(ma.GetSElIndex(sel) == cbound1  || ma.GetSElIndex(sel) == cbound2)
	  {
	    double fac = 1.0;
	    if (ma.GetSElIndex(sel) == cbound1) fac = -1.0;

	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    const HCurlHighOrderFiniteElement<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<2> & > (fel[1]);

	    ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);

		 ma.GetSElVertices (sel, vnums);
		 ma.GetPoint(vnums[0], point0);
		 ma.GetPoint(vnums[1], point1);
		 ma.GetPoint(vnums[2], point2);

// 		 Vec<3> aa = point1 - point0, bb = point2 - point0;
// 		 Vec<3> cc = Cross(aa,bb);
// 		 double astriddet = 0.5*L2Norm(cc);

	    const IntegrationRule & ir2d =
	      GetIntegrationRules().SelectIntegrationRule (ET_TRIG, 6);

	    sigmaspace.GetSDofNrs (sel, dofs);

	    Vector<double> sigma(fel_sigma.GetNDof());
	    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);

	    for(int ir=0; ir<ir2d.GetNIP(); ir++)
	      {
		Vec<3> ipx;
		ipx = (1 - ir2d[ir](0) - ir2d[ir](1)) * point0   + ( ir2d[ir](0) )*point1+ ir2d[ir](1)* point2 ;
		SpecificIntegrationPoint<2,3> sip( ir2d[ir], eltrans, lh );
		
// 		Mat<2> jac = sip.GetJacobian();
// 		Mat<2> inv_jac = sip.GetJacobianInverse();
 		double det = fabs (sip.GetJacobiDet());

		fel_sigma.CalcShape(sip.IP(), sigma);
		sigma *= 1.0 / sqr(det);

 // 		fel_u.CalcShape(sip, u);
//  		u = u * invjac;

		for(int i=0; i<fel_sigma.GetNDof(); i++)
		  for(int cvert =0; cvert < (ncela+1)*(ncelb+1); cvert++)
		    {
		      testvec =0;
		      testvec(cvert) = 1.0;

		      double sig =  vecu.FVDouble()(dofs[i])  * sigma(i); 
		      updatew0(cvert) +=   cmesh.EvaluateContactFunction(testvec, ipx) * sig * det * ir2d[ir].Weight();
		      updatew1(cvert) +=   cmesh.EvaluateContactFunction(testvec, ipx) * sig * det * ir2d[ir].Weight()*fac;
	
		  }
	      }
	  }
      }
    lh.CleanUp();
  }


  void NumProcContact :: ApplyCtau ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		   const BaseVector & vecu, Vector<double> & updatelambdax, Vector<double> & updatelambday, Vector<double> & updatelambdaz, LocalHeap & lh )
  {
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ndof_g = cmesh.GetNDof();
    updatelambdax = 0.0; updatelambday = 0.0; updatelambdaz = 0.0;
    ARRAY<int> vnums(3), cvnums(4);

   
    Vec<3>  point0, point1, point2;

    Vector<double> testvec(updatelambday.Size());
    testvec=0.0;

    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);


    for(int sel=0; sel<nsel; sel++)
      {
	if(ma.GetSElIndex(sel) == cbound1  || ma.GetSElIndex(sel) == cbound2)
	  {
	    double fac = 1.0;
	    if (ma.GetSElIndex(sel) == cbound1) fac = -1.0;

	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    const HCurlHighOrderFiniteElement<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElement<2> & > (fel[1]);

	    ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);

	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    
	    // 		 Vec<3> aa = point1 - point0, bb = point2 - point0;
	    // 		 Vec<3> cc = Cross(aa,bb);
	    // 		 double astriddet = 0.5*L2Norm(cc);
	    
	    const IntegrationRule & ir2d =
	      GetIntegrationRules().SelectIntegrationRule (ET_TRIG, 6);
	    
	    uspace.GetSDofNrs (sel, dofs);

	    //	    Vector<double> sigma(fel_sigma.GetNDof());
	    // 		 FlatMatrix<> u(fel_u.GetNDof(), 1, lh);

	    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
	    for(int ir=0; ir<ir2d.GetNIP(); ir++)
	      {
		Vec<3> ipx;
		ipx = (1 - ir2d[ir](0) - ir2d[ir](1)) * point0   + ( ir2d[ir](0) )*point1+ ir2d[ir](1)* point2 ;

		// find contact element in which ipx is
		int cel;
		cmesh.GetContactElement(ipx, cel);
		cmesh.GetCElVertices(cel, cvnums);
		SpecificIntegrationPoint<2,3> sip( ir2d[ir], eltrans, lh );
		
		// 		Mat<2> jac = sip.GetJacobian();
//  		Mat<2> inv_jac = sip.GetJacobianInverse();
 		double det = fabs (sip.GetJacobiDet());

		
		// 		fel_sigma.CalcShape(sip.IP(), sigma);
		// 		sigma *= 1.0 / sqr(det);
		
//  		fel_u.CalcShape(sip.IP(), u);
//  		u = u * inv_jac;

		DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);
		for(int j=0; j<4; j++)
		  {
		    int cvert = cvnums[j];
		    for(int i=0; i<fel_sigma.GetNDof(); i++)
		      {
		   	  testvec =0;
			  testvec(cvert) = 1.0;
			  
			  double ux =  vecu.FVDouble()(ndof_sigma_total + dofs[i])  * u(0,i); 
			  double uy =  vecu.FVDouble()(ndof_sigma_total + dofs[i])  * u(1,i); 
			  double uz =  vecu.FVDouble()(ndof_sigma_total + dofs[i])  * u(2,i); 
			  updatelambdax(cvert) +=   cmesh.EvaluateContactFunction(testvec, ipx) * ux * det * ir2d[ir].Weight() *fac;
			  updatelambday(cvert) +=   cmesh.EvaluateContactFunction(testvec, ipx) * uy * det * ir2d[ir].Weight() *fac;
			  updatelambdaz(cvert) +=   cmesh.EvaluateContactFunction(testvec, ipx) * uz * det * ir2d[ir].Weight() *fac;
			  
		      }
		  }
	      }
	  }
      }
    lh.CleanUp();
  }



  void NumProcContact :: ProjectDisp (
		    const Vector<double > & vecg, Vector<double> & w)
  {
    
    int ndof_g = vecg.Size();

     for(int i=0; i< ndof_g; i++)
       {
	 if ( w(i) > vecg(i))
	   w(i) = vecg(i);

       }

  }


  void NumProcContact :: ProjectStress ( const double coeff_f, const Vector<double> & sigmann, 
		       Vector<double > & lambdax, Vector<double > & lambday, Vector<double > & lambdaz )
  {


    int ndof_c = lambdax.Size();

    for(int i=0; i < ndof_c; i++)
      {
	double normlambda = sqrt( sqr(lambdax(i)) + sqr(lambday(i)) + sqr(lambdaz(i)) );

	if(normlambda > abs(coeff_f * sigmann(i)) )
	  {
	    double t = abs(coeff_f * sigmann(i)) / normlambda;
	    lambdax(i) = t * lambdax(i);
	    lambday(i) = t * lambday(i);
	    lambdaz(i) = t * lambdaz(i);
	  }
      }


  }


void NumProcContact :: CalcSigmann (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
				    const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh)
{

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ncvert = cmesh.GetNDof();
    //    int ncela = cmesh.GetNCEa(), ncelb = cmesh.GetNCEb();

    Vec<3> point_0, point_a, point_b;
    cmesh.GetDomain(point_0, point_a, point_b);
    Vec<3> normal = Cross(point_0 - point_a, point_0 - point_b);
    double length = sqrt( sqr(normal(0)) + sqr(normal(1)) + sqr(normal(2)));
    normal = 1.0 / length * normal;

    ARRAY<int> vnums(3);
    Vec<3> point0, point1, point2, lambda;
    double t;
    int fac =-1;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

    sigmann = 0.0;
    
    for(int cvert=0; cvert<ncvert; cvert++)
      {
	Vec<3> xx = cmesh.GetPoint(cvert);
	int sel = cmesh.GetMatchingSEl(cvert, 0);
	if(sel >= 0)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    CalcProjection(xx, normal, point0, point1, point2, lambda, t);
	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    ARRAY<int> dofs(fel_sigma.GetNDof());

	    sigmaspace.GetSDofNrs (sel, dofs);

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
	    IntegrationPoint ip(lambda(0), lambda(1));

	    SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );
	    
	    double det = fabs (sip.GetJacobiDet());
	    Vector<double> sigma(fel_sigma.GetNDof());

	    fel_sigma.CalcShape(sip.IP(), sigma);
	    sigma *= 1.0 / sqr(det);
	    for(int i=0; i<fel_sigma.GetNDof(); i++)
	      sigmann(cvert) += sigma(i) * vecu.FVDouble()(dofs[i]);
	  }
	else 
	  sigmann(cvert) = 0;
      }
}




void NumProcContact :: ProjectDisp (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
				    const BaseVector & vecu, const Vector<double> & vecg, Vector<double > & w0, Vector<double> & w1, LocalHeap & lh)
{

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ncvert = cmesh.GetNDof();

    Vec<3> point_0, point_a, point_b;
    cmesh.GetDomain(point_0, point_a, point_b);
    Vec<3> normal = Cross(point_0 - point_a, point_0 - point_b);
    double length = sqrt( sqr(normal(0)) + sqr(normal(1)) + sqr(normal(2)));
    normal = 1.0 / length * normal;

    ARRAY<int> vnums(3);
    Vec<3> point0, point1, point2, lambda;
    double t;
    int fac =1;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpace & uspace = dynamic_cast<const HCurlHighOrderFESpace & > (*compound_fespace[1]);

    int sigma_total_dofs = sigmaspace.GetNDof();
    Vector<double> vecutau(ncvert);
    vecutau = 0.0;
    
    for(int cvert=0; cvert<ncvert; cvert++)
      {
	Vec<3> xx = cmesh.GetPoint(cvert);
	fac = 1;
	for(int bound=0; bound <= 1; bound++)
	  {
	    int sel = cmesh.GetMatchingSEl(cvert, bound);
	    if(sel >= 0)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		CalcProjection(xx, normal, point0, point1, point2, lambda, t);
		const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		const HCurlHighOrderFiniteElement<2> & fel_u = dynamic_cast<const HCurlHighOrderFiniteElement<2> & > (fel[1]);
		const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
		ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());
		
		sigmaspace.GetSDofNrs (sel, dofs);
		uspace.GetSDofNrs (sel, dofsu);
		
		FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		
		ElementTransformation eltrans;
		ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		IntegrationPoint ip(lambda(0), lambda(1));
		
		SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );
		
		DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);
		for(int i=0; i<fel_sigma.GetNDof(); i++)
		  {
		    vecutau(cvert) += u(2,i) * vecu.FVDouble()(sigma_total_dofs + dofsu[i]) * fac;
		  }
	      }
	    fac = -1;
	  }
	vecutau = -vecutau;
	if(vecutau(cvert) + w0(cvert) > vecg(cvert))
	  {
	    w0(cvert) = (vecg(cvert) - vecutau(cvert)) * abs(vecutau(cvert) + w0(cvert)) / w0(cvert);
	  }
	  
      }

}









  void NumProcContact :: Do(LocalHeap & lh)
  {
    cout << "solve contact problem" << endl;

    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();

    CompoundFESpace     compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HCurlHighOrderFESpace & vspace = dynamic_cast<const HCurlHighOrderFESpace & > ( *compound_fespace[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

    MeshAccess  ma =  bfa -> GetFESpace(). GetMeshAccess();

    int ndof, nd_u, nd_sigma;


    ndof = compound_fespace.GetNDof();
    nd_u = compound_fespace[1]->GetNDof();
    nd_sigma = compound_fespace[0]->GetNDof();

    ///
    int maxsteps = 3;
    ///
    double prec = 1e-5;
    ///
    bool print=false;
    ///
    enum SOLVER { CG, GMRES, QMR, NCG };
    ///
    SOLVER solver = CG;

    // ********* in short **************
    // make contact mesh

    // find gap function g

    // solve variational inequality by uzawa type method

    // given w_0, lambda_0

    // while ( error > eps or n < nmax or whatever)

    //   right hand side
    //   rhs = f - (apply_cn(w), apply_ctau(lambda) )

    //   find (sigma, u) by solving system
    //   mata * vecu = rhs

    //   update w, lambda in direction of steepest descent/ascent
    //   w = w - tau * (apply_cn(sigma))
    //   lambda = lambda + tau * apply_ctau(u)
   
    //   project lambda, w onto admissible sets

    //   w = max ( w, g )

    //   lambda = max(min( lambda, 1), -1);

    // end;

    // ************ details ******************

    // make contact mesh
    // int ncel = contactelements;
    Vec<3> xx =0.0, aa = 0.0, bb = 0.0;
    aa(0) = 2.0; bb(1) = 2.0;
    ContactMesh3d cmesh( ma, compound_fespace, cbound1, cbound2, 
			 xx, aa, bb, ncela, ncelb);
 
      int ncel = cmesh.GetNCE();
    int nd_c = cmesh.GetNDof();
    // find gap function g
    Vector<double> vecg (nd_c);
 
    Assembleg ( ma, cmesh,compound_fespace, vecg);

    // init lambda_0, w_0
    Vector<double > veclambdax(nd_c),  veclambdaz(nd_c),  veclambday(nd_c), lambdaold(nd_c); Vector<double > vecw0(nd_c), wold0(nd_c), vecw1(nd_c), wold1(nd_c);
    Vector<double > updatelambdax(nd_c), updatelambday(nd_c), updatelambdaz(nd_c); Vector<double > updatew0(nd_c), updatew1(nd_c);
    Vector<double> vecsigmann(nd_c);
 
    veclambdax = 0.0;
    veclambday = 0.0;
    veclambdaz = 0.0;
    vecsigmann = 0.0;
 
     for(int i=0; i<nd_c; i++)
       {
	 vecw0(i) = 0.0;
 	vecw1(i) = 0.0;
       }
    // solve by Uzawa type method
    BaseVector & rhs = *vecf.CreateVector();
    const BaseMatrix * premat = NULL;
    if (pre)  premat = &(pre->GetMatrix()); 
     KrylovSpaceSolver * invmat;

	switch (solver)
	  {
	  case CG:
	    cout << "cg solve for real problem" << endl;
	    invmat = new CGSolver<double>(mata, *premat);
	    break;
	  case QMR:
	    invmat = new QMRSolver<double>(mata, *premat);
	    break;
	  case GMRES:
	    invmat = new GMRESSolver<double>(mata, *premat);
	    break;
	  }

    invmat->SetMaxSteps (maxsteps);
    invmat->SetPrecision (prec);
    invmat->SetPrintRates ();
    invmat->SetInitialize (0);
    invmat->SetStatusHandler(ma);

//     clock_t starttime, endtime, soltime;
//     starttime = clock();

    int uzawasteps = 0;
    double elambda = 1.0, ew = 1.0;
    double eps = 1e-9;
    double epsw=0, epslambda =0;
   


    
    while( uzawasteps < numuzawasteps && (elambda > epslambda || ew > epsw))
      {
	wold0 = vecw0; wold1 = vecw1;
	lambdaold = veclambdax;

    //   right hand side
    //   rhs = f - (apply_cn(w), apply_ctau(lambda) )
	rhs.FVDouble() = 0.0;
       	ApplyCTranspose( ma, cmesh, compound_fespace, vecw0, vecw1, veclambdax, veclambday, veclambdaz, rhs, lh);
	rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();

    //   find (sigma, u) by solving system
    //   mata * vecu = rhs
	
	invmat->Mult (rhs, vecu);


    //   update w, lambda in direction of steepest descent/ascent
    //   w = w - tau * (apply_cn(sigma))
    //   lambda = lambda + tau * apply_ctau(u)
	

	ApplyCn (ma, cmesh, compound_fespace, vecu, updatew0, updatew1, lh);
       	ApplyCtau (ma, cmesh, compound_fespace, vecu, updatelambdax, updatelambday, updatelambdaz, lh);

	
	for(int i = 0; i < nd_c; i++)
	  {
	    vecw0(i) -= taun*updatew0(i);
	    vecw1(i) -= taun*updatew1(i);
	    veclambdax(i) -= tautau*updatelambdax(i);
	    veclambday(i) -= tautau*updatelambday(i);
	    veclambdaz(i) -= tautau*updatelambdaz(i);
	  }
    //   project lambda, w onto admissible set
	
	ProjectDisp (ma, cmesh, compound_fespace, vecu, vecg, vecw0, vecw1, lh);
	//ProjectDisp (vecg, vecw0);
	CalcSigmann( ma, cmesh, compound_fespace, vecu, vecsigmann, lh );

	ProjectStress ( coeff_f, vecsigmann, veclambdax, veclambday, veclambdaz);

	ew = 0; elambda = 0; epsw=0; epslambda=0;
	for(int i=0; i<nd_c; i++)
	  {
	    ew += sqr(vecw0(i) - wold0(i));
	    epsw += sqr(vecw0(i)) * eps;
// 	    elambda += sqr(veclambda(i) - lambdaold(i));
// 	    epslambda += sqr(veclambda(i)) * eps;
	  }

	//	if(uzawasteps % 100 == 0 ) 
	  Ng_Redraw();
	uzawasteps ++;
	//cout << ew / epsw  << "         " << elambda / epslambda  << endl;
	 epslambda=0; elambda=1;
	 cout << "step " << uzawasteps << ":   " << ew / epsw << endl;

      }
    

    cout << "lambda_n = " << endl;
    for(int i=0; i<veclambdax.Size(); i++)
      cout << veclambdax(i) << "   " << veclambday(i) <<  "  " << veclambdaz(i) << " sigma " << vecsigmann(i) <<  endl;
    cout << "numberofsteps " << uzawasteps << endl;
    for(int ia =0; ia < ncela+1; ia++){
      for(int ib =0; ib < ncelb+1; ib++)
	cout << vecsigmann(ia + ib*(ncela+1)) << "  ";
      cout << endl;}
    delete invmat;
  }












// declare the numproc 'contact' 

namespace
{
  class Init
  { 
  public: 
    Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("contact", NumProcContact::Create, NumProcContact::PrintDoc);
  }
    
  Init init;
}
  
