#ifndef CONTACT3D
#define CONTACT3D

void CalcProjection ( const Vec<3> & xx, const Vec<3> & nn, const Vec<3> & aa, const Vec<3> & bb, const Vec<3> & cc, Vec<3> & lambda, double & t);


class ContactMesh3d
{
protected: 
  Vec<3> point_0, point_a, point_b;
  int ncel_a, ncel_b, ncel;
  int ncvert;
  ARRAY<Vec<3> > vertices;
  int ** eltovert;
  int ndof;

  const FESpace * fespace;
  ARRAY<int[2]> MatchingSEl;

public:



  ContactMesh3d( const MeshAccess & ma, const FESpace & afespace, const int cbound1, const int cbound2, 
		 const Vec<3> apoint_0, const Vec<3> apoint_a, const Vec<3> apoint_b, const int ancel_a, const int ancel_b);

  ~ContactMesh3d()
  { for(int i=0; i<ncel; i++) {delete eltovert[i];}
    delete eltovert; }

  void GetDomain(Vec<3> & p0, Vec<3> & pa, Vec<3> & pb) const
   {
     p0 = point_0;
     pa = point_a;
     pb = point_b;
   }

  const int GetNCE() const
  {
    return ncel;
  }
  const int GetNCEa() const
  {
    return ncel_a;
  }
  const int GetNCEb() const
  {
    return ncel_b;
  }

  const int GetNDof() const
  {
    return ndof;
  }

  const FESpace* GetFESpace() const
  {
    return fespace;
  }

  const Vec<3> GetPoint(const int numvert) const
  {
    return vertices[numvert];
  }

  void GetCElVertices(const int cel, ARRAY<int> & vnums) const
  {
    for(int i=0; i<4; i++)
      vnums[i] = eltovert[cel][i];
  }

  const double EvaluateContactFunction(const Vector<double> & vecf, const Vec<3> & x) const;

  int  GetMatchingSEl ( const int & cvert, const int & bound) const
  {
    int sel = MatchingSEl[cvert][bound];
    return sel;
  }

  void GetContactElement(const Vec<3> & ipx, int & cel) const
  {
    int cela = int ( ncel_a * 1.0 * (ipx(0) - point_0(0)) / (point_a(0) - point_0(0)));
    int celb = int ( ncel_b * 1.0 * (ipx(1) - point_0(1)) / (point_b(1) - point_0(1)));
    if(cela == ncel_a) cela --;
    if(celb == ncel_b) celb --;

    cel = cela + celb*ncel_a;

  }

};





class NumProcContact : public NumProc
{
protected:
  // bilinear-form for the Hellinger Reissner system of elasticity
  BilinearForm * bfa;
  // linear-form providing the right hand side, contains (0, f)
  LinearForm * lff;
  // solution vector, containes (sigma, u)
  GridFunction * gfu;

  //contact surfaces
  int cbound1, cbound2;

  Preconditioner * pre;

  double taun, tautau;
  int numuzawasteps;
  int contactelements;
  double coeff_f;
  int ncela, ncelb;

public:
  
    
  NumProcContact (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {

    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", "a"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    cbound1 = int(flags.GetNumFlag ( "cbound1" , 1.0));
    cbound2 = int(flags.GetNumFlag ( "cbound2" , 2.0));

    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);

    taun = flags.GetNumFlag ( "tau", 1e-5);
    tautau = taun;

    taun = flags.GetNumFlag ( "taun", taun);
    tautau = flags.GetNumFlag ( "tautau", taun);

    numuzawasteps = int(flags.GetNumFlag( "usteps", 10));
    contactelements = int(flags.GetNumFlag( "cels", 10));

    coeff_f = flags.GetNumFlag( "friction", 1.0);

    ncela = int( flags.GetNumFlag( "ncela", 10));
    ncelb = int( flags.GetNumFlag( "ncelb", 10));
  }


  virtual ~NumProcContact() 
  { ; }


  // creates an solver object
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcContact (pde, flags);
  }



  void Assembleg ( const MeshAccess & ma, const ContactMesh3d & cmesh, 
		   const CompoundFESpace & compound_fespace, Vector<double> & vecg);


  void ApplyCTranspose( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
			const Vector<double > & w0, const Vector<double> & w1, 
			const Vector<double> & lambdax, const Vector<double> & lambday, const Vector<double> & lambdaz, BaseVector & rhs,
			LocalHeap & lh);




  void ApplyCn ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, Vector<double> & updatew0, Vector<double> & updatew1, LocalHeap & lh );



  void ApplyCtau ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		   const BaseVector & vecu, Vector<double> & updatelambdax, Vector<double> & updatelambday, Vector<double> & updatelambdaz, LocalHeap & lh );


  void ProjectDisp ( const Vector<double > & vecg, Vector<double> & w);


  void ProjectStress ( const double coeff_f, const Vector<double> & sigmann, 
		       Vector<double > & lambdax, Vector<double > & lambday, Vector<double > & lambdaz );


  void CalcSigmann (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh);

  virtual void Do(LocalHeap & lh);


  void ProjectDisp (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, const Vector<double> & vecg, Vector<double > & w0, Vector<double> & w1, LocalHeap & lh);


  virtual string GetClassName () const
  {
    return " Solver for Contact Problems ";
  }

  virtual void PrintReport (ostream & ost)
  {
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl;
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*    ost << 
      "\n\nNumproc Parabolic:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};









#endif
