/*********************************************************************/
/* File:   h1hofespace.cpp                                           */
/* Author: Start                                                     */
/* Date:   10. Feb. 2003                                             */
/*********************************************************************/

/**
   High Order Finite Element Space
*/

#include <comp.hpp>
#include <multigrid.hpp> 
#include <solve.hpp>
#include <parallelngs.hpp>
#include <feti.hpp>

extern ngstd::AutoPtr<ngsolve::PDE> pde;

using namespace ngmg;

namespace ngla
{
  extern MatrixGraph * MergeGraphs ( const MatrixGraph & graph1, const MatrixGraph & graph2 );
}

namespace ngcomp
{
  using namespace ngcomp;
  using namespace ngparallel;

  DDH1HighOrderFESpace ::  
  DDH1HighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FESpace (ama, flags)
  {
    name = "DDH1HighOrderFESpace(h1ho)";
    // define h1ho flags
    DefineDefineFlag("ddh1ho");
    DefineNumFlag("augmented");
    DefineDefineFlag("plate");
    DefineNumFlag("relorder");
    DefineNumFlag("orderinner");
    DefineNumFlag("orderface");
    DefineNumFlag("orderedge");
    DefineNumFlag("orderquad");
    DefineNumFlag("ordertrig");
    DefineNumFlag("variableorder"); 
    //  DefineNumListFlag("dom_order_min_x");
    //  DefineNumListFlag("dom_order_max_x");
    //  DefineNumListFlag("dom_order_min_y");
    //  DefineNumListFlag("dom_order_max_y");
    //  DefineNumListFlag("dom_order_min_z");
    //  DefineNumListFlag("dom_order_max_z");
    DefineNumFlag("smoothing");
    DefineDefineFlag("minext");
    DefineDefineFlag("optext");
    DefineDefineFlag("fast");
    DefineDefineFlag("print");
    DefineDefineFlag("noprint");
    if (parseflags) ParseFlags(flags);
    
    print = (flags.GetDefineFlag("print")); 
  
    // Variable order space: 
    //      in case of (var_order && order) or (relorder) 
    var_order = flags.GetDefineFlag("variableorder");  
    order =  int (flags.GetNumFlag ("order",1)); 
    
    if(flags.NumFlagDefined("relorder") && !flags.NumFlagDefined("order")) 
       var_order = 1; 
    
    rel_order=int(flags.GetNumFlag("relorder",order-1)); 

    minext = flags.GetDefineFlag ("minext");
    optext = flags.GetDefineFlag ("optext");
    fast_pfem = flags.GetDefineFlag ("fast");

    print = flags.GetDefineFlag("print");

    if(flags.NumFlagDefined("order") && flags.NumFlagDefined("relorder") && !flags.GetDefineFlag("force_relorder")) 
      {
	order =  int (flags.GetNumFlag ("order",1));
	var_order=0; 
      }  
    else 
      {
	if(var_order)
	  cerr << " WARNING: DDH1HoFeSpace: inconsistent flags: variableorder, order and relorder "
	       << "-> variable order space with rel_order, but order is ignored " << endl; 
	else 
	  { 
	    var_order = 0; 
	      cerr << " WARNING: DDH1HoFeSpace: inconsistent flags: order and rel_order "
		   << "-> uniform order space with order " << endl; 
	  }
      } 

    

   
    uniform_order_inner = int (flags.GetNumFlag ("orderinner", -1));
    uniform_order_face = int (flags.GetNumFlag ("orderface", -1));
    uniform_order_edge = int (flags.GetNumFlag ("orderedge", -1));
    uniform_order_quad = int (flags.GetNumFlag ("orderquad", -1));
    uniform_order_trig = int (flags.GetNumFlag ("ordertrig", -1));
    
    if (flags.NumFlagDefined("smoothing")) 
    throw Exception ("Flag 'smoothing' for fespace is obsolete \n Please use flag 'blocktype' in preconditioner instead");
          
    Flags loflags;
    loflags.SetFlag ("order", 1);
    loflags.SetFlag ("dim", dimension);
    if (iscomplex) loflags.SetFlag ("complex");
    low_order_space = 0; // new NodalFESpace (ma, loflags);

          
    static ConstantCoefficientFunction one(1);
    if (ma.GetDimension() == 2)
      {
	evaluator = new MassIntegrator<2> (&one);
	boundary_evaluator = 0;
      }
    else
      {
	evaluator = new MassIntegrator<3> (&one);
	boundary_evaluator = new RobinIntegrator<3> (&one);
      }

    if (dimension > 1)
      {
	evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
	boundary_evaluator = 
	  new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
      }

    prol = new LinearProlongation(*this);

#ifdef PARALLEL
    paralleldofs = 0;
#endif
  }

  DDH1HighOrderFESpace :: ~DDH1HighOrderFESpace ()
  {
    delete usedel;
    delete usedvert;
    delete usededge;
    delete usedface;
    ;
  }

  FESpace * DDH1HighOrderFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new DDH1HighOrderFESpace (ma, flags, true); // true: parse flags
  }
    
  void DDH1HighOrderFESpace :: Update()
  {
    const int dim = ma.GetDimension();
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = ma.GetNFaces();
    nel = ma.GetNE();
    
    order_edge.SetSize (ned);
    order_face.SetSize (nfa);
    order_inner.SetSize (nel);
    order_avertex.SetSize (nv); 
    fine_edge.SetSize(ned); 
    fine_face.SetSize(nfa); 

    fine_edge = 0;
    fine_face = 0; 
    int p = var_order ?  1 : order; 
    
    order_edge = p; 
    order_face = INT<2>(p,p);
    order_inner = INT<3>(p,p,p); 
    order_avertex = p; 
	
    usedel = new BitArray(nel);
    usedvert = new BitArray(nv);
    usededge = new BitArray(ned);
    usedface = new BitArray(nfa);

    usedel->Clear();
    usedvert->Clear();
    usededge->Clear();
    usedface->Clear();

    ARRAY<int> eledges, elfaces, vnums;
    
     for ( int el = 0; el < nel; el++ )
      {
        int elindex = ma.GetElIndex ( el );
	if ( ! DefinedOn ( elindex ) ) continue;

	ma.GetElPNums ( el, vnums );
	ma.GetElEdges ( el, eledges ) ;
	ma.GetElFaces ( el, elfaces );

	for ( int i = 0; i < vnums.Size(); i++ )
	  usedvert->Set(vnums[i]);
	for ( int i = 0; i < eledges.Size(); i++ )
	  usededge->Set(eledges[i]);
	for ( int i = 0; i < elfaces.Size(); i++ )
	  usedface->Set(elfaces[i]);
	usedel->Set(el);
      }

    for (int i = 0; i < nel; i++)
      {	
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const FACE * faces = ElementTopology::GetFaces (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	ma.GetElVertices (i, vnums);
	ma.GetElEdges (i, eledges);		
	

	if(dim==3) 
	  {
	    ma.GetElFaces(i, elfaces);
	    for (int j=0;j<elfaces.Size();j++) fine_face[elfaces[j]] = 1; 
	  }
	for (int j=0;j<eledges.Size();j++) fine_edge[eledges[j]] = 1; 
	
	if(!var_order) continue;  
	
	INT<3> el_orders = ma.GetElOrders(i); 
	for(int l=0;l<3;l++) el_orders[l] += rel_order; 

	for(int j=0;j<dim;j++)
	  order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
	for(int j=0;j<eledges.Size();j++)
	  {
	    for(int k=0;k<dim;k++)
	      if(points[edges[j][0]][k] != points[edges[j][1]][k])
		{ 
		  order_edge[eledges[j]] = max2(order_edge[eledges[j]],el_orders[k]);
		  k=dim; 
		}
	  }
	
	if(dim==3)
	  {
	    for(int j=0;j<elfaces.Size();j++)
	      {
		// trig_face
		if(faces[j][3]==-1) 
		  {
		    order_face[elfaces[j]][0] = int(max2(order_face[elfaces[j]][0], 
							 el_orders[0]));
		    order_face[elfaces[j]][1] = order_face[elfaces[j]][0]; 
		  }
		else //quad_face
		  {
		    int fmax = 0;
		    for(int k = 1; k < 4; k++) 
		      if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		    
		    INT<2> f((fmax+3)%4,(fmax+1)%4); 
		    if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		    
		    
		    for(int l=0;l<2;l++)
		      for(int k=0;k<3;k++)
			if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			  {
			    order_face[elfaces[j]][l] = int(max2(order_face[elfaces[j]][l], 
								 el_orders[k]));
			    break;
			  }
		  }
	      }
	  }
      }
	  
    /*
    if(uniform_order_inner > -1) 
      order_inner = INT<3>(uniform_order_inner,uniform_order_inner,uniform_order_inner);
    if(uniform_order_face > -1) 
      order_face = INT<2>(uniform_order_face,uniform_order_face); 
    if(uniform_order_edge > -1) 
      order_edge = uniform_order_edge; 
    */ 

 
 

    /*
    // skip dofs on Dirichlet boundary hack [JS]
    int nsel = ma.GetNE();
    for (i = 0; i < nsel; i++)
      {
	  ma.GetSElEdges (i, eledges);
	  int face = ma.GetSElFace (i);
	  
	  for (j = 0; j < eledges.Size(); j++) 
	    order_edge[eledges[j]] = 1;  
	  order_face[face] = 1;
      }
    */

    /* 
    if (ma.GetDimension() == 2 && uniform_order_trig != -1 && uniform_order_quad != -1)
      {
	for (int i = 0; i < nel; i++)
	  {
	    if (ma.GetElType(i) == ET_TRIG)
	      order_inner = INT<3> (uniform_order_trig, uniform_order_trig, uniform_order_trig);
	    else
	      order_inner = INT<3> (uniform_order_quad, uniform_order_quad, uniform_order_quad);
	  }
      }
    */ 
    
    if(uniform_order_inner > -1) 
      order_inner = INT<3>(uniform_order_inner,uniform_order_inner,uniform_order_inner);
    if(uniform_order_face > -1 && dim==3) 
      order_face = INT<2>(uniform_order_face,uniform_order_face); 
    if(uniform_order_edge > -1) 
      order_edge = uniform_order_edge; 
    

    for(int i=0;i<ned;i++) if(!fine_edge[i]) {order_edge[i] = 1;}
    if(dim == 3) for(int i=0;i<nfa;i++) if(!fine_face[i]) order_face[i] = INT<2> (1,1); 
    if(dim==2) 
      { 
	nfa = 0; 
	fine_face.SetSize(0); 
	order_face.SetSize(0);
      } 

    if(print) 
      {
	*testout << " H1HoFESpace order " << order << " , var_order " << var_order << " , relorder " << rel_order << endl;  
	(*testout) << "fine_edge (h1): " << fine_edge << endl;
	(*testout) << "fine_face (h1): " << fine_face << endl;
	
	
	(*testout) << "order_edge (h1): " << order_edge << endl;
	
	(*testout) << "order_face (h1): " << order_face <<  endl;
	(*testout) << "order_inner (h1): " << order_inner << endl;
      }
    
#ifdef SABINE 
    // fuer EE damit ich mit rel_order auch p-ref machen kann :) 
      order=order_inner[0][0]; 
#endif 



    UpdateDofTables ();

#ifdef PARALLEL
      UpdateParallelDofs();
#endif

  }


  void DDH1HighOrderFESpace :: UpdateDofTables ()
  {
   
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = ma.GetNFaces();
    nel = ma.GetNE();

    ndof = 0; 

    first_vert_dof.SetSize (nv+1);
    first_edge_dof.SetSize (ned+1);
    first_face_dof.SetSize (nfa+1);

    for ( int i = 0; i < nv; i++ )
      {
	first_vert_dof[i] = ndof;
	if ( UsedVert(i) ) ndof++;
      }
    first_vert_dof[nv] = ndof;

    for (int i = 0; i < ned; i++)
      {
	first_edge_dof[i] = ndof;
	if ( !UsedEdge(i) ) continue;
	if(order_edge[i]>1)
	  ndof += order_edge[i] - 1;
      }
    first_edge_dof[ned] = ndof;
         
    ARRAY<int> fapnums;
    for (int i = 0; i < nfa; i++)
      {
	first_face_dof[i] = ndof;
	if ( !UsedFace(i) ) continue;

	ma.GetFacePNums (i, fapnums);
	INT<2> p = order_face[i];
	switch(fapnums.Size())
	  {
	  case 3: 
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)/2;
	    break;
	  case 4:
	    if(p[0] > 1 && p[1]>1)
	      ndof += (p[0]-1)*(p[1]-1);
	    break; 
	  }
      }
    first_face_dof[nfa] = ndof;
 
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	if ( !UsedEl(i) ) continue;

	INT<3> p = order_inner[i];	
	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)/2;
	    break;
	  case ET_QUAD:
	    if(p[0]>1 && p[1]>1)
	      ndof += (p[0]-1)*(p[1]-1);
	    break;
	  case ET_TET:
	    if(p[0] > 3)
	      ndof += (p[0]-1)*(p[0]-2)*(p[0]-3)/6;
	    break;
	  case ET_PRISM:
	    if(p[0]>2 && p[2]>1)
	      ndof += (p[0]-1)*(p[0]-2)*(p[2]-1)/2;
	    break;
	  case ET_PYRAMID:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)*(2*p[0]-3)/6;
	    break;
	  case ET_HEX:
	    if(p[0]>1 && p[1] > 1 && p[2]>1) 
	      ndof += (p[0]-1)*(p[1]-1)*(p[2]-1);
	    break;
	  }
      } 
    first_element_dof[nel] = ndof;
   
    /*
    (*testout) << "augmented " << augmented << endl; 
   
    (*testout) << "h1 first edge = " << first_edge_dof << endl;
    (*testout) << "h1 first face = " << first_face_dof << endl;
    (*testout) << "h1 first inner = " << first_element_dof << endl;
    
    (*testout) << "order_edge H1 = " << order_edge << endl;
    (*testout) << "order_face H1 = " << order_face << endl;
    (*testout) << "order_inner H1 = " << order_inner << endl;
    (*testout) << " NDOF H1" << ndof << endl; 
    
    (*testout) << specialelements.Size() << endl;
    */

    prol->Update();
  }


  void DDH1HighOrderFESpace :: PrintReport (ostream & ost)
  {
    FESpace::PrintReport (ost);

//     (*testout) << "first edge = " << first_edge_dof << endl;
//     (*testout) << "first face = " << first_face_dof << endl;
//     (*testout) << "first inner = " << first_element_dof << endl;

  }


  const FiniteElement & DDH1HighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    if ( !UsedEl (elnr) ) return * (new FE_SegmDummy);

    try
      {
	H1HighOrderFiniteElement * hofe;

	    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

	    switch (ma.GetElType(elnr))
	      {
	      case ET_TET:
		{ 
		  hofe = new (lh.Alloc (sizeof(H1HighOrderTet<T_ORTHOPOL>)))  H1HighOrderTet<T_ORTHOPOL> (order);
		  break;
		}
	      case ET_PYRAMID:
		{
		  hofe = new (lh.Alloc (sizeof(H1HighOrderPyramid<T_ORTHOPOL>)))  H1HighOrderPyramid<T_ORTHOPOL> (order);
		  break;
		}
	      case ET_PRISM:
		{
		  hofe = new (lh.Alloc (sizeof(H1HighOrderPrism<T_ORTHOPOL>)))  H1HighOrderPrism<T_ORTHOPOL> (order);
		  break;
		}
	      case ET_HEX:
		{
		  hofe = new (lh.Alloc (sizeof(H1HighOrderHex<T_ORTHOPOL>)))  H1HighOrderHex<T_ORTHOPOL> (order);
		  break;
		}
	      case ET_TRIG:
		{ 
		  hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (order);
		  break;
		}
	      case ET_QUAD:
		{
		  hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  H1HighOrderQuad<T_ORTHOPOL> (order);
		  break;
		}
	      default:
		hofe = 0; 
	      }

	if (!hofe)
	  {
	    stringstream str;
	    str << "H1HighOrderFESpace " << GetClassName() 
		<< ", undefined eltype " 
		<< ElementTopology::GetElementName(ma.GetElType(elnr))
		<< ", order = " << order << endl;
	    throw Exception (str.str());
	  }


    
	ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
	ArrayMem<int, 12> ednums, order_ed;
	ArrayMem<int, 6> fanums;
	ArrayMem<INT<2>, 6> order_fa;


	ma.GetElVertices(elnr, vnums);
	ma.GetElEdges(elnr, ednums);

	order_ed.SetSize (ednums.Size());

	for (int j = 0; j < ednums.Size(); j++)
	  order_ed[j] = order_edge[ednums[j]];

	hofe -> SetVertexNumbers (vnums, lh);
	hofe -> SetOrderEdge (order_ed);
    
	if (ma.GetDimension() == 3)
	  {
	    ma.GetElFaces(elnr, fanums);
	    order_fa.SetSize (fanums.Size());
	    for (int j = 0; j < fanums.Size(); j++)
	      order_fa[j] = order_face[fanums[j]];
	
	    hofe -> SetOrderFace (order_fa);
	  }
	
	ArrayMem<int, 8> order_vert(vnums.Size());
	for (int j = 0; j < vnums.Size(); j++)
	  order_vert[j] = order_avertex[vnums[j]];
	hofe -> SetOrderVertex (order_vert);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> ComputeNDof();
	return *hofe;
      }
    catch (Exception & e)
      {
	e.Append ("in DDH1HoFESpace::GetElement");
	e.Append ("\n");
	throw;
      }
  }
 
 


  const FiniteElement & DDH1HighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    /*
      FiniteElement * fe = 0;

      switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
      fe = segm; break;
      case ET_TRIG:
      fe = trig; break;
      case ET_QUAD:
      fe = quad; break;
      default:
      fe = 0;
      }

      H1HighOrderFiniteElement * hofe =
      dynamic_cast<H1HighOrderFiniteElement*> (fe);
    */
    if ( ma.GetDimension() == 2 )
      {
	ARRAY<int> edge(1);
	ma.GetSElEdges ( elnr, edge );
	if ( !UsedEdge(edge[0]) ) return *(new FE_SegmDummy);
      }
    else
      {
	int face = ma.GetSElFace(elnr);
	if ( !UsedFace(face) ) return *(new FE_SegmDummy);
      }

    H1HighOrderFiniteElement * hofe;

	typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
	
	switch (ma.GetSElType(elnr))
	  {
	  case ET_TRIG:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (1);
	      break;
	    }
	  case ET_QUAD:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  H1HighOrderQuad<T_ORTHOPOL> (1);
	      break;
	    }
	  case ET_SEGM:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderSegm<T_ORTHOPOL>)))  H1HighOrderSegm<T_ORTHOPOL> (1);
	      break;
	    }
	  default:
	    hofe = 0; 
	  }
      
    if (!hofe)
      {
	stringstream str;
	str << "FESpace " << GetClassName() 
	    << ", undefined surface eltype " << ma.GetSElType(elnr) 
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    ArrayMem<int, 8> vnums, order_vert(8);
    ArrayMem<int, 4> ednums, order_ed;

    ma.GetSElVertices(elnr, vnums);
    ma.GetSElEdges(elnr, ednums);

    order_ed.SetSize (ednums.Size());

    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order_edge[ednums[j]];

    for (int j = 0; j < 8; j++)
      order_vert[j] = order;
    hofe -> SetOrderVertex (order_vert);
    

    hofe -> SetVertexNumbers (vnums, lh);

    if (ma.GetDimension() == 3)
      {
	hofe -> SetOrderEdge (order_ed);
	INT<2> p = order_face[ma.GetSElFace(elnr)];
	hofe -> SetOrderInner (INT<3> (p[0], p[1], 0));
      }
    else
      {
	hofe -> SetOrderInner (order_ed[0]);
      }
 
    hofe -> ComputeNDof(); 
    return *hofe;
  }
 






  int DDH1HighOrderFESpace :: GetNDof () const
  {
    return ndof;
  }



  void DDH1HighOrderFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0); 
    if ( !UsedEl (elnr) )
      return;

    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0);

    
    if(order < 1 && !var_order) 
      throw Exception(" H1HighOrderFESpace :: GetDofNrs() order < 1 "); 

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (first_vert_dof[vnums[i]]);


    //(*testout) << "after verts dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1];
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    //(*testout) << "after edges dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < fanums.Size(); i++)
      {
	first = first_face_dof[fanums[i]];
	next = first_face_dof[fanums[i]+1];
	  
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    //(*testout) << "after faces dnums.Size() " << dnums.Size() << endl;

    first = first_element_dof[elnr];
    next = first_element_dof[elnr+1];
 
    
    for (j = first; j < next; j++)
      dnums.Append (j);
     
    //(*testout) << "after inner dnums.Size() " << dnums.Size() << endl;

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
   }


  void  DDH1HighOrderFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if ( !UsedEl ( elnr ) )
      return;
      

    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first,next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0); 
    dnums.SetSize(0);

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (first_vert_dof[vnums[i]]);

    if(order < 1) 
      throw Exception(" DDH1HighOrderFESpace :: GetDofNrs() order < 1 "); 

    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1];
	
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    if (ma.GetDimension() == 3)
      {
	int first_face = 0;
	for (i = first_face; i < fanums.Size(); i++)
	  {
	    first = first_face_dof[fanums[i]];
	    next = first_face_dof[fanums[i]+1]; 
	    for (j = first; j < next; j++)
	      dnums.Append (j);
	  }
      }
    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
  }

  int DDH1HighOrderFESpace :: GetNDofLevel (int alevel) const
  {
    return ndof;
  }




  void DDH1HighOrderFESpace :: GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);

    if ( !UsedEl ( elnr ) ) 
      return;

    ArrayMem<int,12> vnums, ednums;

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    
    for (int i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    for (int i = 0; i < ednums.Size(); i++)
      if (order_edge[ednums[i]] > 1)
	dnums.Append (first_edge_dof[ednums[i]]);
  }





  
  void DDH1HighOrderFESpace :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);

    if ( !UsedVert ( vnr ) ) 
      return;

    dnums.Append (first_vert_dof[vnr]);
  }  
  
  void DDH1HighOrderFESpace :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if ( !UsedEdge ( ednr ) ) 
      return;
    int first = first_edge_dof[ednr];
    int neddofs = first_edge_dof[ednr+1] - first;
    for (int j = 0; j < neddofs; j++)
      dnums.Append (first+j);
  }

  void DDH1HighOrderFESpace :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if ( !UsedFace ( fanr ) ) 
      return;
    if (ma.GetDimension() < 3) return;
    
    int first = first_face_dof[fanr];
    int nfadofs = first_face_dof[fanr+1] - first;
    for (int j = 0; j < nfadofs; j++)
      dnums.Append (first+j);
  }


   void DDH1HighOrderFESpace :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
   {
     dnums.SetSize(0);
     if ( !UsedEl ( elnr ) ) 
       return;

     int first = first_element_dof[elnr];
     int neldofs = first_element_dof[elnr+1] - first;
     
     for (int j = 0; j < neldofs; j++)
       dnums.Append (first+j);
   }
  
  void DDH1HighOrderFESpace :: 
  GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if ( ma.GetDimension() == 2 )
      {
	ARRAY<int> edge(1);
	ma.GetSElEdges ( elnr, edge );
	if ( !UsedEdge(edge[0]) ) return;
      }
    else
      {
	int face = ma.GetSElFace(elnr);
	if ( !UsedFace(face) ) return;
      }


    static bool getall = false;

    ARRAY<int> vnums, ednums;
    int fanum;
    int i, j;

    ma.GetSElVertices (elnr, vnums);
    ma.GetSElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      fanum = ma.GetSElFace (elnr);
    
    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (first_vert_dof[vnums[i]]);

    ARRAY<int> edge_start;

    for (i = 0; i < ednums.Size(); i++)
      {
	edge_start.Append(dnums.Size());
	int first = first_edge_dof[ednums[i]];
	int neddofs = first_edge_dof[ednums[i]+1] - first;
	for (j = 0; j < neddofs; j++)
	  dnums.Append (first+j);
	// dnums.Append (-1);
      }
    edge_start.Append(dnums.Size());

    int face_start  = dnums.Size();

    if (ma.GetDimension() == 3)
      {
	int first = first_face_dof[fanum];
	int nfadofs = first_face_dof[fanum+1] - first;
	for (j = 0; j < nfadofs; j++)
	  //	  dnums.Append (-1);
	  dnums.Append (first+j);
      }

    if (!DefinedOnBoundary (ma.GetSElIndex (elnr)) && !getall)
      dnums = -1;



  }
  

  Table<int> * DDH1HighOrderFESpace :: 
  CreateSmoothingBlocks (const Flags & precflags ) const
  {
    ARRAY<int> cnt(1);
    cnt = 0;
    Table<int> * table = new Table<int>(cnt);
    /*
    (*testout) << "H1HO-table = " << table << endl;
    */
    return table;
  }


  ARRAY<int> * 
  DDH1HighOrderFESpace :: CreateDirectSolverClusters (int type) const
  {
    int i, j, k;

    int nv = ma.GetNV();
    int nd = GetNDof();
    int ne = ma.GetNE();
    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());
    clusters = 0;



    return &clusters;
  }

  /* add all entries mat(i,j) where i, j are surface-vertex dofs */
  MatrixGraph * DDH1HighOrderFESpace :: GetGraph (int level, bool symmetric)
  {
    MatrixGraph * graph = FESpace :: GetGraph ( level, symmetric );
    int i, j, k;
    int ndof = GetNDof();
    int ne = GetMeshAccess().GetNE();
    int nse = GetMeshAccess().GetNSE();

    ARRAY<int> linesize (ndof);
    ARRAY<int> dnums;

    ARRAY<int> dof_num_el (ndof);
    ARRAY<int> el_num_dof (ne);


    BitArray surfacedofs ( ndof );
    surfacedofs.Clear();

    for (i = 0; i < nse; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostSEl(i)) continue;
#endif

	GetSDofNrs (i, dnums);

 	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1 )
	    surfacedofs.Set(dnums[j]);
      }

    int ndof_lo = first_vert_dof.Last();
    for ( int dof = ndof_lo; dof < ndof; dof++ )
      surfacedofs.Clear(dof);

    int nsurfdofs = 0;
    for ( int dof = 0; dof < ndof_lo; dof++ )
      if ( surfacedofs.Test(dof)) nsurfdofs++;

    ARRAY<int> linesize2(ndof);

    int ii = 1;
    for ( int dof = 0; dof < ndof; dof++ )
      if ( surfacedofs.Test(dof))
	if ( symmetric )
	  linesize2[dof] = ii++;
	else
	  linesize2[dof] = nsurfdofs;
      else
	linesize2[dof] = 1;

    MatrixGraph * fullgraph = new MatrixGraph (linesize2);

    for ( i = 0; i < ndof; i++ )
      {
	int * data = fullgraph->GetRowIndicesPointer(i);

	if ( surfacedofs.Test(i) )
	  {
	    int max = symmetric ? i+1 : ndof;
	    for ( j = 0; j < max; j++ )
	      if ( surfacedofs.Test(j) )
		{
		  *data = j;
		  data++;
		}
	  }
	else
	  {
	    *data = i; data++;
	  }

      }


    MatrixGraph * mergedgraph = ngla :: MergeGraphs ( *graph, *fullgraph );

    delete graph;
    delete fullgraph;
    return mergedgraph;
  }




#ifdef PARALLEL
  void DDH1HighOrderFESpace :: UpdateParallelDofs_hoproc()
  {
     // ******************************
     // update exchange dofs 
     // ******************************
    *testout << "H1Ho::UpdateParallelDofs_hoproc" << endl;
    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks);
    nexdof = 0;

    MPI_Request * sendrequest = new MPI_Request[ntasks];
    MPI_Request * recvrequest = new MPI_Request[ntasks];
    MPI_Status status;

    // number of vertex exchange dofs
    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	if ( !parallelma->IsExchangeVert ( vert ) ) continue;
	
	ARRAY<int> dnums;
	GetVertexDofNrs ( vert, dnums );
	nexdof[id] += dnums.Size();
	
	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantPNum ( dest, vert ) >= 0 )
	    nexdof[dest] += dnums.Size(); 
      }

    // + number of edge exchange dofs
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	if ( !parallelma->IsExchangeEdge ( edge ) ) continue;
	
	ARRAY<int> dnums;
	GetEdgeDofNrs ( edge, dnums );
	nexdof[id] += dnums.Size() ;  
	
	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantEdgeNum ( dest, edge ) >= 0 )
	    nexdof[dest] += dnums.Size() ;  
      }

    // + number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	if ( !parallelma->IsExchangeFace ( face ) ) continue;
	
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	nexdof[id] += dnums.Size() ; 

	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    nexdof[dest] += dnums.Size() ; 
      }

    // + number of inner exchange dofs --> set eliminate internal

    nexdof[0] = LowOrderFESpace() . GetNDof();

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

    ARRAY<int> ** owndofs, ** distantdofs;
    owndofs = new ARRAY<int>* [ntasks];
    distantdofs = new ARRAY<int>* [ntasks];

    for ( int i = 0; i < ntasks; i++ )
      {
	owndofs[i] = new ARRAY<int>(1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int>(0);
      }
    // *****************
    // Parallel Vertex dofs
    // *****************

    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;
    int exdof = 0;

    // find local and distant dof numbers for vertex exchange dofs
    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	if ( !parallelma->IsExchangeVert ( vert ) ) continue;
	ARRAY<int> dnums;
	GetVertexDofNrs ( vert, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distvert = parallelma -> GetDistantPNum ( dest, vert );
	    if( distvert < 0 ) continue;

	    owndofs[dest]->Append ( distvert );
	    owndofs[dest]->Append( int (paralleldofs->IsGhostDof(dnums[0])) );
	    for ( int i=0; i<dnums.Size(); i++)
	      {
 		paralleldofs->SetExchangeDof ( dest, dnums[i] );
 		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int dest = 1; dest < ntasks; dest ++ )
      if ( dest != id )
	{
	  MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
	}
	  
    for ( int sender = 1; sender < ntasks; sender ++ )
      if ( sender != id )
      {
	MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender] );
      }


    int ii = 1;
    for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest + dest, &status );
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int vnum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetVertexDofNrs (vnum, dnums);
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }
    
    // *****************
    // Parallel Edge dofs
    // *****************

    for ( int dest = 0; dest < ntasks; dest++)
      {
	owndofs[dest]->SetSize(1);
	distantdofs[dest]->SetSize(0);
      }
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      if ( parallelma->IsExchangeEdge ( edge ) )
      {
	ARRAY<int> dnums;
	GetEdgeDofNrs ( edge, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distedge = parallelma -> GetDistantEdgeNum ( dest, edge );
	    if( distedge < 0 ) continue;

	    owndofs[dest]->Append ( distedge );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int dest = 1; dest < ntasks; dest ++ )
      if ( dest != id )
	{
	  MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
	}
	  
    for ( int sender = 1; sender < ntasks; sender ++ )
      if ( sender != id )
      {
	MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender]);
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait(recvrequest+dest, &status );

	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int ednum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetEdgeDofNrs (ednum, dnums);
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }

    // *****************
    // Parallel Face dofs
    // *****************

    for ( int dest = 0; dest < ntasks; dest++)
      {
	owndofs[dest]->SetSize(1);
	distantdofs[dest]->SetSize(0);
      }
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      if ( parallelma->IsExchangeFace ( face ) )
      {
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;

	    owndofs[dest]->Append ( distface );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );

	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int dest = 1; dest < ntasks; dest ++ )
      if ( dest != id )
	{
	  MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
	}
	  
    for ( int sender = 1; sender < ntasks; sender ++ )
      if ( sender != id )
      {
	MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender]);
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	ii = 1;
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest + dest, &status );
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }

     /*******************************

         update low order space

     *****************************/


    int ndof_lo = low_order_space->GetNDof();

    // all dofs are exchange dofs
    nexdof = ndof_lo;
 
    exdof = 0;
    cnt_nexdof = 0;


    // *****************
    // Parallel Vertex dofs
    // *****************

    owndofs[0]->SetSize(1);
    (*owndofs[0])[0] = ndof;
    distantdofs[0]->SetSize(0);
    
     // find local and distant dof numbers for vertex exchange dofs
    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	int dest = 0;
	
	int distvert = parallelma -> GetDistantPNum ( dest, vert );
	owndofs[0]->Append ( distvert );
	paralleldofs->SetExchangeDof ( dest, vert );
	owndofs[0]->Append ( vert );
      }   
    
    int dest = 0;
    MyMPI_ISend ( *owndofs[0], dest, sendrequest[dest]);
    MyMPI_IRecv ( *distantdofs[0], dest, recvrequest[dest] );
   
    MPI_Wait ( recvrequest + dest, &status );

    ii = 1;
    while ( ii < distantdofs[0]->Size() )
      {
	if ( dest == id ) continue;
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
	int vnum = (*distantdofs[0])[ii++];
	(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = vnum;
	(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[0])[ii];
	ii++; cnt_nexdof[dest]++;
      }

    for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;

  }


  void DDH1HighOrderFESpace :: UpdateParallelDofs_loproc()
  {
    *testout << "H1Ho::UpdateParallelDofs_loproc" << endl;
    // lo-proc should never use high order space
    return;
 
  }
#endif

  // register FESpaces
  namespace
#ifdef MACOS
  ddh1hofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("ddh1ho", DDH1HighOrderFESpace::Create);
    }
    
#ifdef MACOS
    static
#endif
    Init init;
  }
  
}
