#include <comp.hpp>
#include <multigrid.hpp>

#include <parallelngs.hpp>

#include <feti.hpp>

#ifdef PARALLEL
extern MPI_Group MPI_HIGHORDER_WORLD;
extern MPI_Comm MPI_HIGHORDER_COMM;
#endif

using namespace ngmg;

namespace ngcomp
{
  using namespace ngcomp;
#ifdef PARALLEL
  using namespace ngparallel;
#endif

  FETIL2SurfaceHighOrderFESpace ::  
  FETIL2SurfaceHighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : L2SurfaceHighOrderFESpace (ama, flags, parseflags)
  {
    name="FETIL2SurfaceHighOrderFESpace(l2surf)";
    // defined flags 
    DefineDefineFlag("fetil2surf");

    if(parseflags) ParseFlags(flags);
    
    
    if(flags.NumFlagDefined("relorder"))
      throw Exception("Variable order not implemented for FETIL2SurfaceHighOrderFESpace"); 
    
    delete segm;
    delete trig;
    delete quad;

    segm = new FETIHighOrderSegm (order);
    trig = new FETIHighOrderTrig (order);
    quad = new FETIHighOrderQuad (order);

    definedon.SetSize (ma.GetNBoundaries());
    definedon = 0;

    int bound = int ( flags.GetNumFlag ("definedon", 0) );
    if ( bound )
      definedon[bound-1] = 1;
    else
      definedon = 1;

   }

  FETIL2SurfaceHighOrderFESpace :: ~FETIL2SurfaceHighOrderFESpace ()
  {
    ;
  }

  FESpace * FETIL2SurfaceHighOrderFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FETIL2SurfaceHighOrderFESpace (ma, flags, true);
  }

  void FETIL2SurfaceHighOrderFESpace :: Update()
  {
    nel = ma.GetNSE();

    ndof = 0;
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	if ( !definedon [ ma.GetSElIndex(i) ] )
	  continue;
	switch (ma.GetSElType(i))
	  {
	  case ET_SEGM:
	    ndof += order+1;
	    break;
	  case ET_TRIG:
	    ndof += (order+1)*(order+2)/2;
	    break;
	  case ET_QUAD:
	    ndof += (order+1)*(order+1);
	    break;
	  default:
	    ;
	  }
      }
    first_element_dof[nel] = ndof;
  }

  const FiniteElement & FETIL2SurfaceHighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    if ( !definedon [ ma.GetSElIndex(elnr) ] )
      {
	fe = new FE_SegmDummy;
	return *fe;
      }

    switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
	fe = segm; break;
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0;
      }
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetSElVertices(elnr, vnums);

    if (!fe)
      {
	stringstream str;
	str << "FETIL2SurfaceHighOrderFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    dynamic_cast<FETIHighOrderFiniteElement*> (fe) -> SetVertexNumbers (vnums);

    return *fe;
  }
 
  const FiniteElement & FETIL2SurfaceHighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    return * new FE_SegmDummy;
    // throw Exception ("Volume elements not available for FETIL2SurfaceHighOrderFESpace");
  }
 
  int FETIL2SurfaceHighOrderFESpace :: GetNDof () const
  {
    return ndof;
  }

  void FETIL2SurfaceHighOrderFESpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if (!DefinedOn (ma.GetSElIndex (elnr)))
      return;

   int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetSElIndex (elnr)))
      dnums = -1;
  }
  
  void FETIL2SurfaceHighOrderFESpace :: 
  GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
  
  Table<int> * FETIL2SurfaceHighOrderFESpace :: 
  CreateSmoothingBlocks ( int type) const
  {
    int i, j, first;
    ARRAY<int> cnt(nel);
    cnt = 0;
    for (i = 0; i < nel; i++)
      cnt[i] = first_element_dof[i+1]-first_element_dof[i];
	
    Table<int> & table = *new Table<int> (cnt);
    
    for (i = 0; i < nel; i++)
      {
	first = first_element_dof[i];
	for (j = 0; j < cnt[i]; j++)
	  table[i][j] = first+j;
      }
    return &table;
  }


  void  FETIL2SurfaceHighOrderFESpace :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  FETIL2SurfaceHighOrderFESpace ::GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  FETIL2SurfaceHighOrderFESpace ::GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { GetDofNrs ( fanr, dnums ); return; }
  
  void  FETIL2SurfaceHighOrderFESpace ::GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs ( elnr, dnums ); return; }
    

  /* FETI lagrange fespace *****/



  FETILagrangeFESpace:: FETILagrangeFESpace
  (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FESpace (ama, flags)
  {
    name="FETILagrangeFESpace";
    // defined flags 
    DefineDefineFlag("fetilagrange");

    if(parseflags) ParseFlags(flags);
    
    segm = new H1HighOrderSegm<IntegratedLegendreMonomialExt> (order);
    trig = new H1HighOrderTrig<IntegratedLegendreMonomialExt> (order);
    quad = new H1HighOrderQuad<IntegratedLegendreMonomialExt> (order);

    ndomains = ma.GetNDomains();

    definedon.SetSize (ma.GetNBoundaries());
    definedon = 0;

    int bound = int ( flags.GetNumFlag ("definedon", 0) );
    if ( bound )
      definedon[bound-1] = 1;
    else
      definedon = 1;
  }


  FETILagrangeFESpace :: ~FETILagrangeFESpace ()
  {
    ;
  }



  ///


  FESpace * FETILagrangeFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FETILagrangeFESpace (ma, flags, true);
  }

  void FETILagrangeFESpace ::
  Update ()
  {
    ndof = 0;
    int dof = 0;

    int dim = ma.GetDimension();
    int nv = ma.GetNV();
    int ned = ma.GetNEdges();
    int nfa = (dim == 3 ) ?  ma.GetNFaces() : 0 ;
    int nsel = ma.GetNSE();

    first_vertex_dof.SetSize(nv+1 );
    first_edge_dof.SetSize (ned+1 );
    first_face_dof.SetSize (nfa+1 );

    BitArray issurfvertex, issurfedge, issurfface;
    issurfvertex.SetSize(nv);
    issurfedge.SetSize(ned);
    issurfface.SetSize(nfa);

    issurfvertex.Clear();
    issurfedge.Clear();
    issurfface.Clear();
    for ( int sel = 0; sel < nsel; sel++ )
      {
	if ( !DefinedOn ( ma.GetSElIndex(sel) ) ) continue;

	ARRAY<int> verts, edges; 
	int facet;
	ma.GetSElVertices (sel, verts );
	ma.GetSElEdges (sel, edges );

	facet = (dim == 3) ? ma.GetSElFace(sel) : edges[0];

	for ( int j = 0; j < verts.Size(); j++ )
	  issurfvertex.Set(verts[j]);
	for ( int j = 0; j < edges.Size(); j++ )
	  issurfedge.Set(edges[j]);
	if ( dim == 3 ) 
	  issurfface.Set(facet);

      }

//     for ( int sel = 0; sel < nsel; sel++ )
//       {
// 	if ( DefinedOn ( ma.GetSElIndex(sel) ) ) continue;

// 	ARRAY<int> verts; 
// 	ma.GetSElVertices (sel, verts );

// 	for ( int j = 0; j < verts.Size(); j++ )
// 	  issurfvertex.Clear(verts[j]);
//       }

    for ( int vert = 0; vert < nv; vert++ )
      {
	first_vertex_dof[vert] = ndof;
	if ( !issurfvertex.Test(vert) ) continue;

	ARRAY<int> elems, vert_domains;
	ma.GetVertexElements(vert, elems);

	vert_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !vert_domains.Contains(index) )
	      vert_domains.Append(index);

	  }

	BubbleSort(vert_domains);

	int n_vdom = vert_domains.Size();
	int n_lag = ((n_vdom-1) * n_vdom) / 2;

	ndof += n_lag;

	for ( int dom1 = 0; dom1 < n_vdom; dom1++ )
	  for ( int dom2 = dom1; dom2 < n_vdom; dom2++ )
	    {
	      domains.Append (INT<2> (vert_domains[dom1], vert_domains[dom2]) );
	      dof++;
	    }
	
      }
    first_vertex_dof[nv] = ndof;


    for ( int edge = 0; edge < ned; edge ++ )
      {
	first_edge_dof[edge] = ndof;
	if ( !issurfedge.Test(edge) ) continue;

	ARRAY<int> elems, edge_domains;

	ma.GetEdgeElements(edge, elems);

	edge_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !edge_domains.Contains(index) )
	      edge_domains.Append(index);

	  }

	BubbleSort(edge_domains);

	int n_vdom = edge_domains.Size();
	int n_lag = (order-1) * ((n_vdom-1) * n_vdom) / 2;

	ndof += n_lag;

	for ( int dom1 = 0; dom1 < n_vdom; dom1++ )
	  for ( int dom2 = dom1+1; dom2 < n_vdom; dom2++ )
	    for ( int p = 0; p < order-1; p++ )
	      {
		domains.Append (INT<2> (edge_domains[dom1], edge_domains[dom2]) );
		dof++;
	      }
	
      }

    first_edge_dof[ned] = ndof;


    for ( int face = 0; face < nfa; face++ )
      {
	first_face_dof[face] = ndof;
	if ( !issurfface.Test(face) ) continue;
	ARRAY<int> elems, face_domains(2);

	ma.GetFaceElements(face, elems);

	face_domains[0] = ma.GetElIndex(elems[0]);
	face_domains[1] = ma.GetElIndex(elems[1]);


	BubbleSort(face_domains);

	int n_lag = ((order-1)*(order-2)) / 2;

	ndof += n_lag;

	for ( int p = 0; p < n_lag; p++ )
	  {
	    domains.Append (INT<2> (face_domains[0],face_domains[1]) );
	    dof++;
	  }


      }
    first_face_dof[nfa] = ndof;
    ;
  }


  const FiniteElement & FETILagrangeFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    return * new FE_SegmDummy;
    // throw Exception ("Volume elements not available for FETIL2SurfaceHighOrderFESpace");
  }

  ///
  const FiniteElement & FETILagrangeFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    return * new FE_SegmDummy;
  }
  ///

  int FETILagrangeFESpace :: GetNDof () const
  {
    return ndof;
  }

  void FETILagrangeFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  { 
    dnums.SetSize(0);
    return;
  }
  ///
  void FETILagrangeFESpace :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  { 
    dnums.SetSize(0);
    return;
  }

  void FETILagrangeFESpace :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { 
    dnums.SetSize(0);
    return;
  }


  void FETILagrangeFESpace ::  GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    for ( int i = first_vertex_dof[vnr]; i < first_vertex_dof[vnr+1]; i++ )
      dnums.Append(i);
    return;
  }

  void FETILagrangeFESpace ::  GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    for ( int i = first_edge_dof[ednr]; i < first_edge_dof[ednr+1]; i++ )
      dnums.Append(i);
    return;
  }

  void FETILagrangeFESpace ::  GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    for ( int i = first_face_dof[fanr]; i < first_face_dof[fanr+1]; i++ )
      dnums.Append(i);
    return;
  }






  FESpace * L2FETICompoundFESpace :: Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(ma.GetNDomains()+1);
    
    cout << "Create FETI Compound FE Space L2" << endl;
    
    int order = int (flags.GetNumFlag ("order", 1.0));
    
    int lagrangeorder = int ( flags.GetNumFlag("lagrangeorder", order ) );

    for ( int i = 0; i < ma.GetNDomains(); i++ )
      {
	Flags h1flags;
	h1flags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
	h1flags.SetFlag("definedon", i+1 );
	if (flags.GetDefineFlag ("eliminate_internal")) h1flags.SetFlag ("eliminate_internal");

	spaces[i] = new DDH1HighOrderFESpace (ma, h1flags);
      }

    Flags lagrangeflags;
    lagrangeflags.SetFlag ( "order", lagrangeorder );
    lagrangeflags.SetFlag ( "definedon", flags.GetNumFlag( "lagrangebound", 0) ); 
    spaces[ma.GetNDomains()] = new FETIL2SurfaceHighOrderFESpace(ma, lagrangeflags);

    L2FETICompoundFESpace * fes = new L2FETICompoundFESpace (ma, spaces, flags);
#ifdef PARALLEL
    fes->ResetParallelDofs();
#endif
    return fes;
  }



  FESpace * FETICompoundFESpace :: Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(ma.GetNDomains()+1);
    
    cout << "Create FETI Compound FE Space" << endl;
    
    int order = int (flags.GetNumFlag ("order", 1.0));
    
    int lagrangeorder = int ( flags.GetNumFlag("lagrangeorder", order ) );

    for ( int i = 0; i < ma.GetNDomains(); i++ )
      {
	Flags h1flags;
	h1flags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
	h1flags.SetFlag("definedon", i+1 );
	if (flags.GetDefineFlag ("eliminate_internal")) h1flags.SetFlag ("eliminate_internal");
	if (flags.GetDefineFlag ("complex")) h1flags.SetFlag ( "complex" );

	spaces[i] = new DDH1HighOrderFESpace (ma, h1flags);
      }

    Flags lagrangeflags;
    lagrangeflags.SetFlag ( "order", lagrangeorder );
    lagrangeflags.SetFlag ( "definedon", flags.GetNumFlag( "lagrangebound", 0) ); 
    if (flags.GetDefineFlag ("complex")) lagrangeflags.SetFlag ( "complex" );

    spaces[ma.GetNDomains()] = new FETILagrangeFESpace(ma, lagrangeflags);

    FETICompoundFESpace * fes = new FETICompoundFESpace (ma, spaces, flags);
    
#ifdef PARALLEL
    fes->ResetParallelDofs();
#endif
    return fes;
  }





  template<int D>
  void FETIMassIntegrator<D> ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();
    
//     cout << "assemble elmat" << endl;
    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    int domain = 0;
    while ( domain < 10  )
      {
	if ( dynamic_cast<const H1HighOrderFiniteElement *> (&cfel[domain] ) )
	  break;
	domain++;

      }

    const H1HighOrderFiniteElement & fel_h1 = 
      dynamic_cast<const H1HighOrderFiniteElement&> 
      (cfel[domain]);
      
    int nd_u = fel_h1.GetNDof();
    
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(fel_h1.ElementType(), 2*fel_h1.Order() );
  
    for ( int l = 0; l < ir.GetNIP(); l++ )
      {
	  SpecificIntegrationPoint<D,D> sip(ir[l], eltrans, lh);
  
	  FlatVector<double> shape( nd_u, lh);
	  
	  fel_h1.CalcShape ( sip.IP(), shape );
	  
	  double u = 0;
	  for( int i=0; i<nd_u; i++)
	    for ( int j=0; j< nd_u; j++ )
	      elmat(i,j) += shape(i)*shape(j);
      }
  }



  template<int D> 
  void FETIMassIntegrator<D> ::
  CalcFlux (const FiniteElement & bfel,
	    const BaseSpecificIntegrationPoint & bsip,
	    const FlatVector<double> & elx, 
	    FlatVector<double> & flux,
	    bool applyd,
	    LocalHeap & lh) const

  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    int domain = 0;
    while ( domain < 1e10 )
      {
	if ( dynamic_cast<const H1HighOrderFiniteElement *> (&cfel[domain] ) )
	  break;
	domain++;
      }

    const H1HighOrderFiniteElement & fel_h1 = 
      dynamic_cast<const H1HighOrderFiniteElement&> 
      (cfel[domain]);
      
    int ndof = cfel.GetNDof();
    int nd_u = fel_h1.GetNDof();
      
    flux=0; 
        
    FlatVector<double> shape( nd_u, lh);
  
    fel_h1.CalcShape ( bsip.IP(), shape );
  
    double u = 0;
    for( int i=0; i<nd_u; i++)
      u += shape(i) * elx(i);
  
    flux (0) = u;
//     cout << "element " << bsip.GetTransformation().GetElementNr()+1 << ": flux = " << endl <<  flux << endl;
  }
      




template class FETIMassIntegrator<2>;
template class FETIMassIntegrator<3>;





  // register FESpaces
  namespace
#ifdef MACOS
  fetil2hofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("l2surf", FETIL2SurfaceHighOrderFESpace::Create);
      GetFESpaceClasses().AddFESpace ("l2fetifespace", L2FETICompoundFESpace::Create);
      GetFESpaceClasses().AddFESpace ("fetifespace", FETICompoundFESpace::Create);
      GetIntegrators().AddBFIntegrator ("fetimass", 2, 1,
					FETIMassIntegrator<2>::Create);
      GetIntegrators().AddBFIntegrator ("fetimass", 3, 1,
					FETIMassIntegrator<3>::Create);

    }
    
    Init init;
  }
}
