#include <fem.hpp>
#include <feti.hpp>

namespace ngfem
{
  using namespace ngfem;
  
  FETIHighOrderFiniteElement ::   
  FETIHighOrderFiniteElement (int dim, ELEMENT_TYPE aeltype)
    : NodalFiniteElement (dim, aeltype, -1, -1) 
  { 
    for (int i = 0; i < 8; i++)
      vnums[i] = i;
    
  }
  
  void FETIHighOrderFiniteElement :: 
  SetVertexNumbers (FlatArray<int> & avnums)
  {
    for (int i=0; i<avnums.Size(); i++)
      vnums[i] = avnums[i];
    for (int i=avnums.Size(); i<8; i++)
      vnums[i] = -1;
  }


  void FETIHighOrderFiniteElement::
  SetOrder (int o)
  {
    order_inner = INT<3>(o,o,o); 
    ComputeNDof();
  }

  void FETIHighOrderFiniteElement::
  SetOrderInner (INT<3> oi)
  {
    order_inner = oi;   
    ComputeNDof();
  }

  void FETIHighOrderFiniteElement:: 
  GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize(0);
    for (int i = 0; i < GetNDof(); i++)
      idofs.Append (i);
  }


  /* *********************** Segment  **********************/

  FETIHighOrderSegm :: FETIHighOrderSegm (int aorder)
    : FETIHighOrderFiniteElement(1, ET_SEGM)
  {
    order = aorder;
    order_inner = INT<3> (aorder, aorder, aorder);
    ComputeNDof();
  }


  void FETIHighOrderSegm :: ComputeNDof()
  {
    order = order_inner[0];
    ndof = order+1;
  }

  void FETIHighOrderSegm :: CalcShape (const IntegrationPoint & ip, 
				     FlatVector<> shape) const
  {
    double x = ip(0);
    // orient
    if ( vnums[0] > vnums[1])
      x = 1-x;
    
    LegendrePolynomial (order, 2*x-1, shape);


//     if ( vnums[0] > vnums[1])
//       shape *= -1;

  }
  
  void FETIHighOrderSegm :: CalcDShape (const IntegrationPoint & ip, 
				     FlatMatrix<> dshape) const
  {
    AutoDiff<1> x(ip(0),0);

    if ( vnums[0] > vnums[1])
      x = 1-x;
    
    ArrayMem<AutoDiff<1>, 20> pol(order_inner[0]+1); 
    LegendrePolynomial (order_inner[0], 2*x-1, pol);

    for(int i=0; i<order_inner[0]+1; i++) 
      dshape(i,0) = pol[i].DValue(0); 

//     if ( vnums[0] > vnums[1])
//       dshape *= -1;
  }

  /* *********************** Triangle  **********************/

  FETIHighOrderTrig :: FETIHighOrderTrig (int aorder)
    : FETIHighOrderFiniteElement(2, ET_TRIG)
  {
    order_inner = INT<3> (aorder,aorder,aorder); 
    ComputeNDof();
  }


  void FETIHighOrderTrig :: ComputeNDof()
  {
    ndof = (order_inner[0]+1) * (order_inner[0]+2) / 2;
    order = order_inner[0]; 
  }




  void FETIHighOrderTrig :: CalcShape (const IntegrationPoint & ip, 
				     FlatVector<> shape) const
  {
    double x = ip(0);
    double y = ip(1);
    
    double lami[3] = { x, y, 1-x-y };
    int fav[3] = { 0, 1, 2};
    int factor = 1;
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); factor *= -1; }
    if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); factor *= -1; }
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); factor *= -1; }	
    x = lami[fav[0]]; y=lami[fav[1]];
    
    
     
    int ii = 0;

    int n = order_inner[0];

    ArrayMem<double, 20> polx(n+1), poly(n+1);

   ScaledLegendrePolynomial (n, 2*x+y-1, 1-y, polx);
    
    for (int i = 0; i <= order_inner[0]; i++)
      {
       JacobiPolynomial (n, 2*y-1, 2*i+1, 0, poly);
        for (int j = 0; j <= order_inner[0]-i; j++)
          shape(ii++) = polx[i] * poly[j];
      }
    // shape *= factor;
  }

  void FETIHighOrderTrig :: CalcDShape (const IntegrationPoint & ip, 
				     FlatMatrix<> dshape) const
  {
    AutoDiff<2> x(ip(0),0), y(ip(1),1);
      
    AutoDiff<2> lami[3] = { x, y, 1-x-y };
    int fav[3] = { 0, 1, 2};
    int factor = 1;
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); factor *= -1; }
    if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); factor *= -1; }
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); factor *= -1; }	
    x = lami[fav[0]]; y=lami[fav[1]];
    
    int ii = 0;
    int n = order_inner[0];

    ArrayMem<AutoDiff<2>, 20> polx(n+1), poly(n+1);

    ScaledLegendrePolynomial (n, 2*x+y-1, 1-y, polx);
    
    for (int i = 0; i <= order_inner[0]; i++)
    {
      JacobiPolynomial (n, 2*y-1, 2*i+1, 0, poly);
      for (int j = 0; j <= order_inner[0]-i; j++, ii++)
        for(int l=0;l<2;l++) 
          dshape(ii,l) = polx[i].DValue(l) * poly[j].Value() + polx[i].Value()*poly[j].DValue(l);
    }
    // dshape *= factor;

  }





  /* *********************** Quadrilateral  **********************/

  FETIHighOrderQuad :: FETIHighOrderQuad (int aorder)
    : FETIHighOrderFiniteElement(2, ET_QUAD)
  {
    order_inner = INT<3>(aorder,aorder,aorder);
    ComputeNDof();
  }

  void FETIHighOrderQuad :: ComputeNDof()
  {
    ndof = (order_inner[0]+1) * (order_inner[1]+1); 
    order = max(order_inner[0],order_inner[1]); 
  }

  void FETIHighOrderQuad :: CalcShape (const IntegrationPoint & ip, 
				     FlatVector<> shape) const
  {
    double x = ip(0);
    double y = ip(1);
    
    // orient: copied from h1
    double lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
    double sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  
    int fmax = 0;
    int factor = 1; 
    for (int j = 1; j < 4; j++)
      if (vnums[j] > vnums[fmax]) fmax = j;
    int f1 = (fmax+3)%4; 
    int f2 = (fmax+1)%4; 
    if(vnums[f2] > vnums[f1]) { swap(f1,f2); factor = -1; } // fmax > f1 > f2; 
    x = sigma[fmax]-sigma[f1]; 
    y = sigma[fmax]-sigma[f2]; 
    
    int ii = 0;
    
    int n = order;
    ArrayMem<double, 20> polx(n+1), poly(n+1);

    LegendrePolynomial (n, x, polx);
    LegendrePolynomial (n, y, poly);
    
    for (int i = 0; i <= order_inner[0]; i++)
      for (int j = 0; j <= order_inner[1]; j++)
	shape(ii++) = polx[i] * poly[j];
    // shape *= factor;
  }
  
  void FETIHighOrderQuad :: CalcDShape(const IntegrationPoint & ip, 
				     FlatMatrix<> dshape) const  
  {
    AutoDiff<2> x(ip(0), 0);
    AutoDiff<2> y(ip(1), 1);

      // orient: copied from h1
    AutoDiff<2> lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
    AutoDiff<2> sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  
    int fmax = 0; 
    for (int j = 1; j < 4; j++)
      if (vnums[j] > vnums[fmax]) fmax = j;
    int f1 = (fmax+3)%4; 
    int f2 = (fmax+1)%4; 
    int factor = 1;

    if(vnums[f2] > vnums[f1]) { swap(f1,f2); factor = -1; }  // fmax > f1 > f2; 
    x = sigma[fmax]-sigma[f1]; 
    y = sigma[fmax]-sigma[f2]; 
     
    
    int ii=0; 
    int n = max(order_inner[0],order_inner[1]); 
    
    ArrayMem<AutoDiff<2>,20> polx(n+1), poly(n+1); 
    LegendrePolynomial (n, x, polx); 
    LegendrePolynomial (n, y, poly); 
    
    for (int i = 0; i <= order_inner[0]; i++)
      for (int j = 0; j <= order_inner[1]; j++)
    { 
      dshape(ii,0) = polx[i].DValue(0) * poly[j].Value() + polx[i].Value() * poly[j].DValue(0);
      dshape(ii++,1) = polx[i].DValue(1) * poly[j].Value() + polx[i].Value() * poly[j].DValue(1);
    } 
    // dshape *= factor;
  }



} // namespace
