/*********************************************************************/
/* File:   h1hofespace.cpp                                           */
/* Author: Start                                                     */
/* Date:   10. Feb. 2003                                             */
/*********************************************************************/

/**
   High Order Finite Element Space
*/

#include <comp.hpp>
#include <multigrid.hpp> 
#include <solve.hpp>
#include <parallelngs.hpp>
#ifdef ASTRID
#include <feti.hpp>
#endif

extern ngstd::AutoPtr<ngsolve::PDE> pde;

using namespace ngmg;

#ifdef CLEMENS
namespace ngfem
{
#include "h1hofeclemens.hpp"
}

#include ...

#endif



namespace ngcomp
{
  using namespace ngcomp;
  using namespace ngparallel;

#include "specialh1hofespace.hpp"



  H1HighOrderFESpaceClemens ::  
  H1HighOrderFESpaceClemens (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FESpace (ama, flags)
  {
    name = "H1HighOrderFESpaceClemens(h1ho)";
    // define h1ho flags
    DefineDefineFlag("h1ho");
    DefineNumFlag("augmented");
    DefineDefineFlag("plate");
    DefineNumFlag("relorder");
    DefineNumFlag("orderinner");
    DefineNumFlag("orderface");
    DefineNumFlag("orderedge");
    DefineNumFlag("orderquad");
    DefineNumFlag("ordertrig");
    DefineNumFlag("variableorder"); 
    DefineNumFlag("smoothing");
    DefineDefineFlag("minext");
    DefineDefineFlag("optext");
    DefineDefineFlag("fast");
    DefineDefineFlag("print");
    DefineDefineFlag("noprint");
    if (parseflags) ParseFlags(flags);
    
    augmented = int (flags.GetNumFlag ("augmented", 0));
    print = (flags.GetDefineFlag("print")); 
  
    // Variable order space: 
    //      in case of (var_order && order) or (relorder) 
    var_order = flags.GetDefineFlag("variableorder");  
    order =  int (flags.GetNumFlag ("order",1)); 
    
    if(flags.NumFlagDefined("relorder") && !flags.NumFlagDefined("order")) 
       var_order = 1; 
    
    rel_order=int(flags.GetNumFlag("relorder",order-1)); 

    print = flags.GetDefineFlag("print");

    if(flags.NumFlagDefined("order") && flags.NumFlagDefined("relorder") && !flags.GetDefineFlag("force_relorder")) 
      {
	order =  int (flags.GetNumFlag ("order",1));
	var_order=0; 
      }  
    else 
      {
	if(var_order)
	  cerr << " WARNING: H1HoFeSpace: inconsistent flags: variableorder, order and relorder "
	       << "-> variable order space with rel_order, but order is ignored " << endl; 
	else 
	  { 
	    var_order = 0; 
	      cerr << " WARNING: H1HoFeSpace: inconsistent flags: order and rel_order "
		   << "-> uniform order space with order " << endl; 
	  }
      } 

    

   
    uniform_order_inner = int (flags.GetNumFlag ("orderinner", -1));
    uniform_order_face = int (flags.GetNumFlag ("orderface", -1));
    uniform_order_edge = int (flags.GetNumFlag ("orderedge", -1));
    uniform_order_quad = int (flags.GetNumFlag ("orderquad", -1));
    uniform_order_trig = int (flags.GetNumFlag ("ordertrig", -1));
    
    if (flags.NumFlagDefined("smoothing")) 
    throw Exception ("Flag 'smoothing' for fespace is obsolete \n Please use flag 'blocktype' in preconditioner instead");
          
    Flags loflags;
    loflags.SetFlag ("order", 1);
    loflags.SetFlag ("dim", dimension);
    if (iscomplex) loflags.SetFlag ("complex");
    low_order_space = 0; // new NodalFESpace (ma, loflags);

          
    static ConstantCoefficientFunction one(1);
    if (ma.GetDimension() == 2)
      {
	evaluator = new MassIntegrator<2> (&one);
	boundary_evaluator = 0;
      }
    else
      {
	evaluator = new MassIntegrator<3> (&one);
	boundary_evaluator = new RobinIntegrator<3> (&one);
      }

    if (dimension > 1)
      {
	evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
	boundary_evaluator = 
	  new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
      }

    prol = new LinearProlongation(*this);


    meshinfofile = flags.GetStringFlag("meshinfo", "");
  }

  H1HighOrderFESpaceClemens :: ~H1HighOrderFESpaceClemens ()
  {
    ;
  }

  FESpace * H1HighOrderFESpaceClemens :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new H1HighOrderFESpaceClemens (ma, flags, true); // true: parse flags
  }
    
  void H1HighOrderFESpaceClemens :: Update()
  {
    if (low_order_space)
      low_order_space -> Update();


    const int dim = ma.GetDimension();
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = (dim == 2) ? 0 : ma.GetNFaces();
    // nfa = ma.GetNFaces();
    nel = ma.GetNE();
    
    order_edge.SetSize (ned);
    order_face.SetSize (nfa);
    order_inner.SetSize (nel);
    order_avertex.SetSize (nv); 

    int p = var_order ?  1 : order; 
    
    order_edge = p; 
    order_face = INT<2>(p,p);
    order_inner = INT<3>(p,p,p); 
    order_avertex = p; 
	
    SetExternMeshInfo();

    ARRAY<int> eledges, elfaces, vnums;
    

    /* ************

       set polynomial degrees ---> done extern 

       *************** */


//     for (int i = 0; i < nel; i++)
//       {	
// 	ELEMENT_TYPE eltype=ma.GetElType(i); 
// 	const FACE * faces = ElementTopology::GetFaces (eltype);
// 	const EDGE * edges = ElementTopology::GetEdges (eltype);
// 	const POINT3D * points = ElementTopology :: GetVertices (eltype);
// 	ma.GetElVertices (i, vnums);
// 	ma.GetElEdges (i, eledges);		
	

// 	if(dim==3) 
// 	  {
// 	    ma.GetElFaces(i, elfaces);
// 	    for (int j=0;j<elfaces.Size();j++) fine_face[elfaces[j]] = 1; 
// 	  }
// 	for (int j=0;j<eledges.Size();j++) fine_edge[eledges[j]] = 1; 
	
// 	if(!var_order) continue;  
	
// 	INT<3> el_orders = ma.GetElOrders(i); 
// 	for(int l=0;l<3;l++) el_orders[l] += rel_order; 

// 	for(int j=0;j<dim;j++)
// 	  order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
// 	for(int j=0;j<eledges.Size();j++)
// 	  {
// 	    for(int k=0;k<dim;k++)
// 	      if(points[edges[j][0]][k] != points[edges[j][1]][k])
// 		{ 
// 		  order_edge[eledges[j]] = max2(order_edge[eledges[j]],el_orders[k]);
// 		  k=dim; 
// 		}
// 	  }
	
// 	if(dim==3)
// 	  {
// 	    for(int j=0;j<elfaces.Size();j++)
// 	      {
// 		// trig_face
// 		if(faces[j][3]==-1) 
// 		  {
// 		    order_face[elfaces[j]][0] = int(max2(order_face[elfaces[j]][0], 
// 							 el_orders[0]));
// 		    order_face[elfaces[j]][1] = order_face[elfaces[j]][0]; 
// 		  }
// 		else //quad_face
// 		  {
// 		    int fmax = 0;
// 		    for(int k = 1; k < 4; k++) 
// 		      if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		    
// 		    INT<2> f((fmax+3)%4,(fmax+1)%4); 
// 		    if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		    
		    
// 		    for(int l=0;l<2;l++)
// 		      for(int k=0;k<3;k++)
// 			if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
// 			  {
// 			    order_face[elfaces[j]][l] = int(max2(order_face[elfaces[j]][l], 
// 								 el_orders[k]));
// 			    break;
// 			  }
// 		  }
// 	      }
// 	  }
//       }
	  
        

    if(dim==2) 
      { 
	nfa = 0; 
	order_face.SetSize(0);
      } 

    if(print) 
      {
	*testout << " H1HoFESpace order " << order << " , var_order " << var_order << " , relorder " << rel_order << endl;  
	
	*testout << "first_edge_dof  " << intern_edge << endl;
	(*testout) << "order_edge (h1): " << order_edge << endl;
	
	(*testout) << "order_face (h1): " << order_face <<  endl;
	(*testout) << "order_inner (h1): " << order_inner << endl;
      }
    

    UpdateDofTables ();


  }


  void H1HighOrderFESpaceClemens :: SetExternMeshInfo()
  {
    ifstream infile ( meshinfofile.c_str() );

    // infile edges --> v1, v2, p_edge
    // infile faces --> v1, v2, v3, (v4,) p_face
    
    extern_edge.SetSize(ned+1);
    extern_face.SetSize(nfa+1);
    intern_edge.SetSize(ned+1);
    intern_face.SetSize(nfa+1);

    if ( print)
      cout << "Set extern mesh info" << endl;
    for ( int el = 0; el < nel; el++ )
      {
	int nvert, p;
	ARRAY<int> extvnums, extednums, vnums, ednums, p_edge;
	int v1, v2;
	infile >> nvert;
	infile >> p;

	order_inner[el] = INT<3> (p, p, p);

	extvnums.SetSize(nvert);
	extednums.SetSize(nvert);
	p_edge.SetSize(nvert);
	for ( int i = 0; i < nvert; i++ )
	  infile >> extvnums[i];
	for ( int i = 0; i < nvert; i++ )
	  infile >> extednums[i];
	for ( int i = 0; i < nvert; i++ )
	  infile >> p_edge[i];

	ma.GetElPNums(el, vnums);
	ma.GetElEdges(el, ednums);

	for ( int edi = 0; edi < ednums.Size(); edi++ )
	  {
	    int ed = ednums[edi];
	    ma.GetEdgePNums (ed, v1, v2);

	    int i1 = extvnums.Pos(v1);
	    int i2 = extvnums.Pos(v2);

	    int locednum;
	    if ( i1+i2 == 1 ) locednum = 0;
	    else if ( i1 + i2 == 3 ) locednum = 1;
	    else if ( i1 + i2 == 2 ) locednum = 2;

	    extern_edge[ednums[edi]] = extednums[locednum]; 
	    intern_edge[extednums[locednum]] = ednums[edi];

	    order_edge[ed] = p_edge[locednum];

	  }

      }

    extern_edge[ned] = ned;
    extern_face[nfa] = nfa;
    intern_edge[ned] = ned;
    intern_face[nfa] = nfa;
    if ( print )
      cout << "ok..." << endl;
    ;
  }


  void H1HighOrderFESpaceClemens :: UpdateDofTables ()
  {
   
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = (ma.GetDimension() == 2) ? 0 : ma.GetNFaces();
    nel = ma.GetNE();

    ndof = nv;
    if (augmented == 1) ndof += nv;
    if (augmented == 2) ndof = order*nv;

    if ( print )
      cout << "first edge" << endl;
    first_edge_dof.SetSize (ned+1);
    for (int i = 0; i < ned; i++)
      {
	first_edge_dof[intern_edge[i]] = ndof;
	if(order_edge[intern_edge[i]]>1)
	  ndof += order_edge[intern_edge[i]] - 1;
      }
    first_edge_dof[ned] = ndof;
         
    if ( print )
      cout << "first face" << endl;
    first_face_dof.SetSize (nfa+1);
    ARRAY<int> fapnums;
    for (int i = 0; i < nfa; i++)
      {
	first_face_dof[intern_face[i]] = ndof;
	ma.GetFacePNums (intern_face[i], fapnums);
	INT<2> p = order_face[intern_face[i]];
	switch(fapnums.Size())
	  {
	  case 3: 
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)/2;
	    break;
	  case 4:
	    if(p[0] > 1 && p[1]>1)
	      ndof += (p[0]-1)*(p[1]-1);
	    break; 
	  }
      }
    first_face_dof[nfa] = ndof;
 
    if ( print )
      cout << "first element " << endl;
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	INT<3> p = order_inner[i];	
	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)/2;
	    break;
	  case ET_QUAD:
	    if(p[0]>1 && p[1]>1)
	      ndof += (p[0]-1)*(p[1]-1);
	    break;
	  case ET_TET:
	    if(p[0] > 3)
	      ndof += (p[0]-1)*(p[0]-2)*(p[0]-3)/6;
	    break;
	  case ET_PRISM:
	    if(p[0]>2 && p[2]>1)
	      ndof += (p[0]-1)*(p[0]-2)*(p[2]-1)/2;
	    break;
	  case ET_PYRAMID:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)*(2*p[0]-3)/6;
	    break;
	  case ET_HEX:
	    if(p[0]>1 && p[1] > 1 && p[2]>1) 
	      ndof += (p[0]-1)*(p[1]-1)*(p[2]-1);
	    break;
	  }
      } 
    first_element_dof[nel] = ndof;
   
    if ( print )
      cout << "juhu.." << endl;
    /*
    (*testout) << "augmented " << augmented << endl; 
   
    (*testout) << "h1 first edge = " << first_edge_dof << endl;
    (*testout) << "h1 first face = " << first_face_dof << endl;
    (*testout) << "h1 first inner = " << first_element_dof << endl;
    
    (*testout) << "order_edge H1 = " << order_edge << endl;
    (*testout) << "order_face H1 = " << order_face << endl;
    (*testout) << "order_inner H1 = " << order_inner << endl;
    (*testout) << " NDOF H1" << ndof << endl; 
    
    (*testout) << specialelements.Size() << endl;
    */




    // wir machen sicher kein multigrid in der visualization :-)
    
//     while (ma.GetNLevels() > ndlevel.Size())
//       ndlevel.Append (ndof);
//     ndlevel.Last() = ndof;
//     prol->Update();
  }


  void H1HighOrderFESpaceClemens :: PrintReport (ostream & ost)
  {
    FESpace::PrintReport (ost);

//     (*testout) << "first edge = " << first_edge_dof << endl;
//     (*testout) << "first face = " << first_face_dof << endl;
//     (*testout) << "first inner = " << first_element_dof << endl;

  }


  const FiniteElement & H1HighOrderFESpaceClemens :: GetFE (int elnr, LocalHeap & lh) const
  {
    try
      {
	H1HighOrderFiniteElement * hofe;

	typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

	switch (ma.GetElType(elnr))
	  {
	  case ET_TET:
	    { 
	      hofe = new (lh.Alloc (sizeof(H1HighOrderTet<T_ORTHOPOL>)))  H1HighOrderTet<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_PYRAMID:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderPyramid<T_ORTHOPOL>)))  H1HighOrderPyramid<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_PRISM:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderPrism<T_ORTHOPOL>)))  H1HighOrderPrism<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_HEX:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderHex<T_ORTHOPOL>)))  H1HighOrderHex<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_TRIG:
	    { 
#ifdef CLEMENS
 	      hofe = new (lh.Alloc (sizeof(H1HighOrderTrigClemens)))  H1HighOrderTrigClemens (order);
#else
	      hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (order);
#endif
	      break;
	    }
	  case ET_QUAD:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  H1HighOrderQuad<T_ORTHOPOL> (order);
	      break;
	    }
	  default:
	    hofe = 0; 
	  }


	if (!hofe)
	  {
	    stringstream str;
	    str << "H1HighOrderFESpaceClemens " << GetClassName() 
		<< ", undefined eltype " 
		<< ElementTopology::GetElementName(ma.GetElType(elnr))
		<< ", order = " << order << endl;
	    throw Exception (str.str());
	  }


    
	ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
	ArrayMem<int, 12> ednums, order_ed;
	ArrayMem<int, 6> fanums;
	ArrayMem<INT<2>, 6> order_fa;


	ma.GetElVertices(elnr, vnums);
	ma.GetElEdges(elnr, ednums);

	order_ed.SetSize (ednums.Size());

	for (int j = 0; j < ednums.Size(); j++)
	  order_ed[j] = order_edge[ednums[j]];

	hofe -> SetVertexNumbers (vnums, lh);
	hofe -> SetOrderEdge (order_ed);
    
	if (ma.GetDimension() == 3)
	  {
	    ma.GetElFaces(elnr, fanums);
	    order_fa.SetSize (fanums.Size());
	    for (int j = 0; j < fanums.Size(); j++)
	      order_fa[j] = order_face[fanums[j]];
	
	    hofe -> SetOrderFace (order_fa);
	  }
	
	ArrayMem<int, 8> order_vert(vnums.Size());
	for (int j = 0; j < vnums.Size(); j++)
	  order_vert[j] = order_avertex[vnums[j]];
	hofe -> SetOrderVertex (order_vert);
	hofe -> SetAugmented (augmented);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> ComputeNDof();
	return *hofe;
      }
    catch (Exception & e)
      {
	e.Append ("in H1HoFESpace::GetElement");
	e.Append ("\n");
	throw;
      }
  }
 
 


  const FiniteElement & H1HighOrderFESpaceClemens :: GetSFE (int elnr, LocalHeap & lh) const
  {
    /*
      FiniteElement * fe = 0;

      switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
      fe = segm; break;
      case ET_TRIG:
      fe = trig; break;
      case ET_QUAD:
      fe = quad; break;
      default:
      fe = 0;
      }

      H1HighOrderFiniteElement * hofe =
      dynamic_cast<H1HighOrderFiniteElement*> (fe);
    */

    H1HighOrderFiniteElement * hofe;


	typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
	
	switch (ma.GetSElType(elnr))
	  {
	  case ET_TRIG:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (1);
	      break;
	    }
	  case ET_QUAD:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  H1HighOrderQuad<T_ORTHOPOL> (1);
	      break;
	    }
	  case ET_SEGM:
	    {
	      hofe = new (lh.Alloc (sizeof(H1HighOrderSegm<T_ORTHOPOL>)))  H1HighOrderSegm<T_ORTHOPOL> (1);
	      break;
	    }
	  default:
	    hofe = 0; 
	  }
      
    if (!hofe)
      {
	stringstream str;
	str << "FESpace " << GetClassName() 
	    << ", undefined surface eltype " << ma.GetSElType(elnr) 
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    hofe -> SetAugmented (augmented);
    
    ArrayMem<int, 8> vnums, order_vert(8);
    ArrayMem<int, 4> ednums, order_ed;

    ma.GetSElVertices(elnr, vnums);
    ma.GetSElEdges(elnr, ednums);

    order_ed.SetSize (ednums.Size());

    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order_edge[ednums[j]];

    for (int j = 0; j < 8; j++)
      order_vert[j] = order;
    hofe -> SetOrderVertex (order_vert);
    

    hofe -> SetVertexNumbers (vnums, lh);

    if (ma.GetDimension() == 3)
      {
	hofe -> SetOrderEdge (order_ed);
	INT<2> p = order_face[ma.GetSElFace(elnr)];
	hofe -> SetOrderInner (INT<3> (p[0], p[1], 0));
      }
    else
      {
	hofe -> SetOrderInner (order_ed[0]);
      }
 
    hofe -> ComputeNDof(); 
    return *hofe;
  }
 






  int H1HighOrderFESpaceClemens :: GetNDof () const
  {
    return ndof;
  }



  int H1HighOrderFESpaceClemens :: GetNDofLevel (int alevel) const
  {
    return ndlevel[alevel];
  }



  void H1HighOrderFESpaceClemens :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if ( print )
      cout << "get dof nrs" << endl;
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0);
    dnums.SetSize(0); 

    
    if(order < 1 && !var_order) 
      throw Exception(" H1HighOrderFESpaceClemens :: GetDofNrs() order < 1 "); 

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);


    if (augmented == 1)
      for (i = 0; i < vnums.Size(); i++)
	dnums.Append (nv+vnums[i]);
    else if (augmented == 2)
      for (i = 0; i < vnums.Size(); i++)
	for (j = 0; j < order-1; j++)
	  dnums.Append (nv+(order-1)*vnums[i]+j);

    if ( print )
      (*testout) << "after verts dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[intern_edge[extern_edge[ednums[i]]+1]];
// 	*testout << "edge " << ednums[i] << " first edge dof.." << first << ", next " << next << endl;
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    if ( print )
      (*testout) << "after edges dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < fanums.Size(); i++)
      {
	int p_face = order_face[fanums[i]][0];
	first = first_face_dof[fanums[i]];
        next = first_face_dof[intern_face[extern_face[fanums[i]]+1]];
	  
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    if ( print )
      (*testout) << "after faces dnums.Size() " << dnums.Size() << endl;

    first = first_element_dof[elnr];
    next = first_element_dof[elnr+1];
 
    
    for (j = first; j < next; j++)
      dnums.Append (j);
     
    if ( print )
      (*testout) << "after inner dnums.Size() " << dnums.Size() << endl;

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
    if ( print )
      cout << "getdofnrs done" << endl;
   }


  void  H1HighOrderFESpaceClemens :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first,next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0); 
    dnums.SetSize(0);

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    if(order < 1) 
      throw Exception(" H1HighOrderFESpaceClemens :: GetDofNrs() order < 1 "); 

    if (augmented == 1)
      for (i = 0; i < vnums.Size(); i++)
	dnums.Append (nv+vnums[i]);
    else if (augmented == 2)
      for (i = 0; i < vnums.Size(); i++)
	for (j = 0; j < order-1; j++)
	  dnums.Append (nv+(order-1)*vnums[i]+j);

    for (i = 0; i < ednums.Size(); i++)
      {
	
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[intern_edge[extern_edge[ednums[i]]+1]];
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }

    if (ma.GetDimension() == 3)
      {
	int first_face = 0;
	for (i = first_face; i < fanums.Size(); i++)
	  {
	    first = first_face_dof[fanums[i]];
	    next = first_face_dof[intern_face[extern_face[fanums[i]]+1]]; 
	    for (j = first; j < next; j++)
	      dnums.Append (j);
	  }
      }
    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
  }





  void H1HighOrderFESpaceClemens :: GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ArrayMem<int,12> vnums, ednums;

    dnums.SetSize(0);

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    
    for (int i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    for (int i = 0; i < ednums.Size(); i++)
      if (order_edge[ednums[i]] > 1)
	dnums.Append (first_edge_dof[ednums[i]]);
  }





  
  void H1HighOrderFESpaceClemens :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    dnums.Append (vnr);

    if (augmented == 1)
      dnums.Append (nv+vnr);
    else if (augmented == 2)
      for (int j = 0; j < order-1; j++)
	dnums.Append (nv+(order-1)*vnr+j);
  }
  
  
  void H1HighOrderFESpaceClemens :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_edge_dof[ednr];
    int neddofs = first_edge_dof[intern_edge[extern_edge[ednr]+1]] - first;
    for (int j = 0; j < neddofs; j++)
      dnums.Append (first+j);
  }

  void H1HighOrderFESpaceClemens :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if (ma.GetDimension() < 3) return;
    
    int first = first_face_dof[fanr];
    int nfadofs = first_face_dof[intern_face[extern_face[fanr]+1]] - first;
    for (int j = 0; j < nfadofs; j++)
      dnums.Append (first+j);
  }


   void H1HighOrderFESpaceClemens :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
   {
     dnums.SetSize(0);

     int first = first_element_dof[elnr];
     int neldofs = first_element_dof[elnr+1] - first;
     
     for (int j = 0; j < neldofs; j++)
       dnums.Append (first+j);
   }
  
  void H1HighOrderFESpaceClemens :: 
  GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    static bool getall = false;

    ARRAY<int> vnums, ednums;
    int fanum;
    int i, j;

    ma.GetSElVertices (elnr, vnums);
    ma.GetSElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      fanum = ma.GetSElFace (elnr);
    
    dnums.SetSize(0);
    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    if (augmented == 1)
      for (i = 0; i < vnums.Size(); i++)
	dnums.Append (nv+vnums[i]);
    else if (augmented == 2)
      for (i = 0; i < vnums.Size(); i++)
	for (j = 0; j < order-1; j++)
	  dnums.Append (nv+(order-1)*vnums[i]+j);

    
    ARRAY<int> edge_start;

    for (i = 0; i < ednums.Size(); i++)
      {
	edge_start.Append(dnums.Size());
	int first = first_edge_dof[ednums[i]];
	int neddofs = first_edge_dof[intern_edge[extern_edge[ednums[i]]+1]] - first;
	for (j = 0; j < neddofs; j++)
	  dnums.Append (first+j);
	// dnums.Append (-1);
      }
    edge_start.Append(dnums.Size());

    int face_start  = dnums.Size();

    if (ma.GetDimension() == 3)
      {
	int first = first_face_dof[fanum];
	int nfadofs = first_face_dof[intern_edge[extern_edge[fanum]+1]] - first;
	for (j = 0; j < nfadofs; j++)
	  //	  dnums.Append (-1);
	  dnums.Append (first+j);
      }

    if (!DefinedOnBoundary (ma.GetSElIndex (elnr)) && !getall)
      dnums = -1;



  }
  

  
  

  // register FESpaces
  namespace
#ifdef MACOS
  specialh1hofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("h1hoclemens", H1HighOrderFESpaceClemens::Create);
    }
    
#ifdef MACOS
    static
#endif
    Init init;
  }
  
}
