#include <la.hpp>
#include <comp.hpp>
#include <fem.hpp>
#include <solve.hpp>
#include <feti.hpp>

namespace ngla
{
  using namespace ngla;
  using namespace ngcomp;
  using namespace ngsolve;

  template <class SCAL>
  FETIDP_Matrix<SCAL> :: FETIDP_Matrix () 
    : BaseMatrix(),
      height(0),
      width(0),
      invspp(0),
      invsdd(0),
      matadd(0),
      matapd(0),
      matapp(0),
      mats(0),
      primaldofs(0),
      dualdofs(0),
      lagrangedofs(0),
      matb(0),
      matbarb(0),
      inversetype(SPARSECHOLESKY)
  {
    ;
  }

  template <class SCAL>
  FETIDP_Matrix<SCAL> ::
  FETIDP_Matrix (const MatrixGraph & agraph_matapp, 
		 const MatrixGraph & agraph_matapd, 
		 const MatrixGraph & agraph_matadd,
		 bool stealgraph,
		 INVERSETYPE ainversetype)
    : BaseMatrix(),
      height(agraph_matapp.Size()),
      width (agraph_matapp.Size() ),
      invspp(0),
      invsdd(0),
      matadd(0),
      matapp(0),
      matb(0),
      matbarb(0),
      inversetype(ainversetype)
 {
   matapp = new SparseMatrix<SCAL> ( agraph_matapp, false );
   matadd = new SparseMatrix<SCAL> ( agraph_matadd, stealgraph );
   matapd = new SparseMatrix<SCAL> ( agraph_matapd, stealgraph );
   mats = new SparseMatrix<SCAL> ( agraph_matapp, stealgraph );

   matapd -> AsVector() = 0.0;
   mats -> AsVector() = 0.0;
   matapp -> AsVector() = 0.0;
   matadd -> AsVector() = 0.0;

    primaldofs = new BitArray(height);
    dualdofs = new BitArray(height);
    lagrangedofs = new BitArray(height);
    invsdd_dofs = new BitArray(height);

    primaldofs->Clear();
    dualdofs->Clear();
    lagrangedofs -> Clear();
    invsdd_dofs->Clear();

  }

  template <class SCAL>
  FETIDP_Matrix<SCAL> :: ~FETIDP_Matrix()
  {
    delete invsdd;
    delete invspp;
    delete matapd;
    delete matadd;
    delete matapp;
    delete mats;
    delete matb;
    if ( matbarb ) delete matbarb;
    delete primaldofs;
    delete dualdofs;
    delete lagrangedofs;
    delete invsdd_dofs;
  }

  template <class SCAL>
  ostream & FETIDP_Matrix<SCAL> :: Print (ostream & ost) const
  {
    ost << endl << " ****************************** " 
	<< endl << "      FETIDP Matrix  " 
	<< endl << " ****************************** " << endl;
    ost << endl << " ****************************** " 
	<< endl << "      Inverses   " 
	<< endl << " ****************************** " << endl;
    ost << endl << " ** invsdd ** " << endl;
    if ( invsdd ) invsdd->Print(ost);
    ost << endl << " ** invspp ** " << endl;
    if ( invspp ) invspp->Print(ost);
    ost << endl << " ****************************** " 
	<< endl << "      Matrix A " 
	<< endl << " ****************************** " << endl;
    ost << endl << " ** matapp ** " << endl;
    if ( matapd ) matapp -> Print(ost) ;
    ost << endl << " ** matapd ** " << endl;
    if ( matapd ) matapd -> Print(ost) ;
    ost << endl << " ** matadd ** " << endl;
    if ( matapd ) matadd -> Print(ost) ;
    ost << endl << " ****************************** " 
	<< endl << "      Matrix S " 
	<< endl << " ****************************** " << endl;
    if ( mats ) mats -> Print(ost) ;
    ost << endl << " ****************************** " 
	<< endl << "      Matrices B " 
	<< endl << " ****************************** " << endl;
    ost << endl << " ** matb ** " << endl;
    if ( matb ) matb -> Print(ost);
    ost << endl << " ** matbarb ** " << endl;
    if ( matbarb ) matbarb -> Print(ost);
    ost << endl << endl;
  }
 
  /// creates matrix of same type
  template <class SCAL>
  BaseMatrix * FETIDP_Matrix<SCAL> :: CreateMatrix () const
  {
    cout << "FETIDP_Matrix -- CreateMatrix not implemented" << endl;
    return 0;
  }

  /// creates a compativle vector, size = width
  template <class SCAL>
  BaseVector * FETIDP_Matrix<SCAL> :: CreateRowVector () const
  { return new VVector<SCAL> (width); }

  /// creates a compativle vector, size = height
  template <class SCAL>
  BaseVector * FETIDP_Matrix<SCAL> :: CreateColVector () const
  { return new VVector<SCAL> (height); }

  /// creates a fitting vector (for square matrices)
  template <class SCAL>
  BaseVector * FETIDP_Matrix<SCAL> :: CreateVector () const
  { return new VVector<SCAL> (height); }


  /// y = matrix * x. Multadd should be implemented, instead
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: Mult (const BaseVector & x, BaseVector & y) const
  {
    BaseVector & constx = const_cast<BaseVector&> (x);
    BaseVector & help = *y.CreateVector();

    FlatVector<SCAL> fvx = 
      dynamic_cast<T_BaseVector<SCAL> &> (constx).FV();
    FlatVector<SCAL> fvy = 
      dynamic_cast<T_BaseVector<SCAL> &> (y).FV();
    FlatVector<SCAL> fvh = 
      dynamic_cast<T_BaseVector<SCAL> &> (help).FV();

    y = Transpose(*matb) * x;
    help = (*invsdd) * y;

    y += (*matapd) * help ;

    help = (*invspp) * y;

    y += Transpose(*matapd) * help;

    help = (*invsdd) * y;

    y = (*matb) * help;

    delete &help;
  }
  /// y += s matrix * x
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: MultAdd (double s, const BaseVector & x, BaseVector & y) const
  { 
    BaseVector & constx = const_cast<BaseVector&> (x);
    BaseVector & help1 = *y.CreateVector();
    BaseVector & help2 = *y.CreateVector();

    FlatVector<SCAL> fvx = 
      dynamic_cast<T_BaseVector<SCAL> &> (constx).FV();
    FlatVector<SCAL> fvy = 
      dynamic_cast<T_BaseVector<SCAL> &> (y).FV();
    FlatVector<SCAL> fvh1 = 
      dynamic_cast<T_BaseVector<SCAL> &> (help1).FV();
    FlatVector<SCAL> fvh2 = 
      dynamic_cast<T_BaseVector<SCAL> &> (help2).FV();
//     FlatVector<SCAL> & fvx = dynamic_cast<FlatVector<SCAL> & > (constx);
//     FlatVector<SCAL> & fvy = dynamic_cast<FlatVector<SCAL> & > (y);
//     FlatVector<SCAL> & fvh1 = dynamic_cast<FlatVector<SCAL> & > (help1);
//     FlatVector<SCAL> & fvh2 = dynamic_cast<FlatVector<SCAL> & > (help2);

    help1 = Transpose(*matb) * x;
    help2 = (*invsdd) * help1;

    help1 += (*matapd) * help2;

    help2 = (*invspp) * help1;

    help1 += Transpose(*matapd) * help2;

    help2 = (*invsdd) * help1;

    (*matb).MultAdd(s, help2, y);

    for ( int i = 0; i < height; i++ )
      if ( !lagrangedofs -> Test(i) )
	fvy(i) = 0.0;

    delete &help1;
    delete &help2;

  }
   
  /// y += s matrix * x
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: MultAdd (Complex s, const BaseVector & x, BaseVector & y) const
  { 
    BaseVector & constx = const_cast<BaseVector&> (x);
    BaseVector & help1 = *y.CreateVector();
    BaseVector & help2 = *y.CreateVector();

    FlatVector<SCAL> fvx = 
      dynamic_cast<T_BaseVector<SCAL> &> (constx).FV();
    FlatVector<SCAL> fvy = 
      dynamic_cast<T_BaseVector<SCAL> &> (y).FV();
    FlatVector<SCAL> fvh1 = 
      dynamic_cast<T_BaseVector<SCAL> &> (help1).FV();
    FlatVector<SCAL> fvh2 = 
      dynamic_cast<T_BaseVector<SCAL> &> (help2).FV();

//     FlatVector<SCAL> & fvx = dynamic_cast<FlatVector<SCAL> & > (constx);
//     FlatVector<SCAL> & fvy = dynamic_cast<FlatVector<SCAL> & > (y);
//     FlatVector<SCAL> & fvh1 = dynamic_cast<FlatVector<SCAL> & > (help1);
//     FlatVector<SCAL> & fvh2 = dynamic_cast<FlatVector<SCAL> & > (help2);

    for ( int i = 0; i < height; i++ )
      if ( ! lagrangedofs->Test(i) ) 
	fvx(i) = 0.0;

    help1 = Transpose(*matb) * x;
    help2 = (*invsdd) * help1;

    help1 += (*matapd) * help2;

    help2 = (*invspp) * help1;

    help1 += Transpose(*matapd) * help2;

    help2 = (*invsdd) * help1;

    (*matb).MultAdd(s, help2, y);

    for ( int i = 0; i < height; i++ )
      if ( !lagrangedofs -> Test(i) )
	fvy(i) = 0.0;

    delete &help1;
    delete &help2;

  }
  
  /// y += s Trans(matrix) * x
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: MultTransAdd (double s, const BaseVector & x, BaseVector & y) const
  {
    MultAdd ( s, x, y );
  }

  /// y += s Trans(matrix) * x
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: MultTransAdd (Complex s, const BaseVector & x, BaseVector & y) const
  { 
    MultAdd ( s, x, y );
  }
  
  template <class SCAL>
  void FETIDP_Matrix<SCAL> :: CalcInverses ( bool eliminate_internal ) 
  {
    mats->SetInverseType(inversetype);
    matadd->SetInverseType(inversetype);

    if ( eliminate_internal )
      for ( int i = 0; i < matadd->Height(); i++ )
	{
	  if ( dualdofs->Test(i) ) invsdd_dofs->Set(i);
	} 
    else
      for ( int i = 0; i < matadd->Height(); i++ )
	{
	  if ( !primaldofs->Test(i) && !lagrangedofs->Test(i) ) invsdd_dofs->Set(i);
	} 

    if ( invsdd ) delete invsdd;
    invsdd = matadd->InverseMatrix();
    if ( invspp ) delete invspp;
    invspp = mats->InverseMatrix(primaldofs);
  }

  template class FETIDP_Matrix<double>;
  template class FETIDP_Matrix<Complex>;


  template <class SCAL>
  FETIDP_PrecMatrix<SCAL> :: FETIDP_PrecMatrix ( const FETIDP_Matrix<SCAL> & afetidpmat )
    : fetidpmat(afetidpmat),
      BaseMatrix()
  {
    ;
  }

  template <class SCAL>
  FETIDP_PrecMatrix<SCAL> :: ~FETIDP_PrecMatrix ()
  {
    ;
  }

  template <class SCAL>
  ostream & FETIDP_PrecMatrix<SCAL> :: Print (ostream & ost) const
  {
    ost << "FETIDP_Prec Print... " << endl << endl;
  }
 
  /// creates matrix of same type
  template <class SCAL>
  BaseMatrix * FETIDP_PrecMatrix<SCAL> :: CreateMatrix () const
  {
    FETIDP_PrecMatrix * newmat = new FETIDP_PrecMatrix ( fetidpmat );
    return newmat;
  }

  /// creates a compativle vector, size = width
  template <class SCAL>
  BaseVector * FETIDP_PrecMatrix<SCAL> :: CreateRowVector () const
  {
    fetidpmat . CreateRowVector();
  }

  /// creates a compativle vector, size = height
  template <class SCAL>
  BaseVector * FETIDP_PrecMatrix<SCAL> :: CreateColVector () const
  {
    fetidpmat . CreateColVector();
  }

  /// creates a fitting vector (for square matrices)
  template <class SCAL>
  BaseVector * FETIDP_PrecMatrix<SCAL> :: CreateVector () const
  {
    fetidpmat.CreateVector();
  }

  /// y = matrix * x. Multadd should be implemented, instead
  template <class SCAL>
  void FETIDP_PrecMatrix<SCAL> :: Mult (const BaseVector & x, BaseVector & y) const
  {
    BaseVector & help = *x.CreateVector();
    BaseVector & constx = const_cast<BaseVector&> (x);
    int nd = x.Size();
    const BaseMatrix & addmat = fetidpmat.GetAddMat();
    const BaseMatrix & barbmat = fetidpmat.GetBarBMat();

    FlatVector<SCAL> fx = 
      dynamic_cast<T_BaseVector<SCAL> &> (constx).FV();
    FlatVector<SCAL> fy = 
      dynamic_cast<T_BaseVector<SCAL> &> (y).FV();
    FlatVector<SCAL> fh = 
      dynamic_cast<T_BaseVector<SCAL> &> (help).FV();


    // y = B^T x
    y = Transpose(barbmat) * x;

    // help = A_dd B^T x
    help = addmat * y;

    // y = ( .. , .. , B A_dd B^T x )
    y = barbmat * help;

//     for ( int i = 0; i < nd; i++ )
//       if ( !fetidpmat.LagrangeDofs().Test(i) ) fy(i) = 0.0;

    delete &help;
  }

  /// y += s matrix * x
  template <class SCAL>
  void FETIDP_PrecMatrix<SCAL> :: MultAdd (double s, const BaseVector & x, BaseVector & y) const
  {
    BaseVector & help = *x.CreateVector();
    BaseVector & help2 = *x.CreateVector();
    BaseVector & constx = const_cast<BaseVector&> (x);
    int nd = x.Size();
    const BaseMatrix & addmat = fetidpmat.GetAddMat();
    const BaseMatrix & barbmat = fetidpmat.GetBarBMat();

    FlatVector<SCAL> fx = 
      dynamic_cast<T_BaseVector<SCAL> &> (constx).FV();
    FlatVector<SCAL> fy = 
      dynamic_cast<T_BaseVector<SCAL> &> (y).FV();
    FlatVector<SCAL> fh = 
      dynamic_cast<T_BaseVector<SCAL> &> (help).FV();
    FlatVector<SCAL> fh2 = 
      dynamic_cast<T_BaseVector<SCAL> &> (help2).FV();

    // y = B^T x
    help2 = Transpose(barbmat) * x;

    // help = A_dd B^T x
    help = addmat * help2;

    // y = ( .. , .. , B A_dd B^T x )
    barbmat.MultAdd ( s, help, y );

    delete &help;
    delete &help2;

  }
  /// y += s matrix * x
  template <class SCAL>
  void FETIDP_PrecMatrix<SCAL> :: MultAdd (Complex s, const BaseVector & x, BaseVector & y) const
  {
    cout << "fetidp precond MultAdd Complex not implemented, copy paste.. " << endl;
  }
  
  /// y += s Trans(matrix) * x
  template <class SCAL>
  void FETIDP_PrecMatrix<SCAL> :: MultTransAdd (double s, const BaseVector & x, BaseVector & y) const
  {
    cout << "fetidp precond MultTransAdd double not implemented, copy paste.. " << endl;
  }
  /// y += s Trans(matrix) * x
  template <class SCAL>
  void FETIDP_PrecMatrix<SCAL> :: MultTransAdd (Complex s, const BaseVector & x, BaseVector & y) const
  {
    cout << "fetidp precond MultTransAdd Complex not implemented, copy paste.. " << endl;
  }

  template class FETIDP_PrecMatrix<double>;
  template class FETIDP_PrecMatrix<Complex>;



  MatrixGraph *  MatrixGraph :: CreatePartialGraph ( BitArray & rows, BitArray & cols)
  {
    ARRAY<int> elsperrow(size);
    elsperrow = 0;

    for ( int i = 0; i < size; i++ )
      {
	if ( !rows.Test(i) ) continue;
	for ( int j_ind = firsti[i]; j_ind < firsti[i+1]; j_ind++ )
	  {
	    if ( cols.Test(colnr[j_ind]) )
	      elsperrow[i] ++;
	  }

      }

    MatrixGraph * newgraph = new MatrixGraph ( elsperrow );

    for ( int i = 0; i < size; i++ )
      {
	if ( !rows.Test(i) ) continue;
	for ( int j_ind = firsti[i]; j_ind < firsti[i+1]; j_ind++ )
	  {
	    if ( cols.Test(colnr[j_ind]) )
	      newgraph->CreatePosition( i, colnr[j_ind]);
	  }

      }
    
    return newgraph;
  }



}
