#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>
#include <feti.hpp>

extern ngsolve::PDE * pde;

namespace ngcomp
{

  using namespace ngfem;
  using namespace ngsolve;
  using namespace ngcomp;

  double GetMeanValueOnElement ( int elnr, CoefficientFunction * coef, 
				 const MeshAccess & ma, LocalHeap & lh)
  {
    void * heapp = lh.GetPointer();
    ElementTransformation eltrans;
    ma.GetElementTransformation ( elnr, eltrans, lh);
    IntegrationPoint * ip;
    switch ( ma.GetElType(elnr) )
      {
      case ET_TRIG:
	ip = new IntegrationPoint(1./3., 1./3.);
	break;
      case ET_QUAD:
	ip = new IntegrationPoint(0.5, 0.5);
	break;
      case ET_TET:
	ip = new IntegrationPoint (0.25, 0.25, 0.25);
	break;
      case ET_PRISM:
	ip = new IntegrationPoint (1./3., 1./3., 0.5);
	break;
      case ET_PYRAMID:
	ip = new IntegrationPoint(0.5, 0.5, 0.25);
	break;
      default:
	ip = new IntegrationPoint(0.5, 0.5, 0.5);
      }

    double value;
    if ( ma.GetDimension() == 2 )
      {
	SpecificIntegrationPoint<2,2> sip( *ip, eltrans, lh);
	value = coef->Evaluate(sip);
      }
    else
      {
	SpecificIntegrationPoint<3,3> sip( *ip, eltrans, lh);
	value = coef->Evaluate(sip);
      }
    lh.CleanUp(heapp);
    return value;

  }



  template <class SCAL>
  FETIDP_BilinearForm<SCAL> :: 
  FETIDP_BilinearForm (const FESpace & afespace, const string & aname,
		       const Flags & flags)
    : S_BilinearForm<SCAL> ( afespace, aname, flags )
  {
    inversetype = SPARSECHOLESKY;
    string it = flags.GetStringFlag("inverse", "" );
    if ( it == "pardiso" )
      {
	inversetype = PARDISO;
      }
    else if ( it == "pardisospd" )
      inversetype = PARDISOSPD;
    else if ( it == "superlu" )
      inversetype = SUPERLU;
    else if ( it == "sparsecholesky" )
      inversetype = SPARSECHOLESKY;

    string coeffname = flags.GetStringFlag ( "feticoeff" , "");

    if ( coeffname != "" )
      {
	deletefeticoeff = false;
	feticoeff = pde->GetCoefficientFunction ( coeffname );
      }
    else
      {
	deletefeticoeff = true;
	feticoeff = new ConstantCoefficientFunction(1);
      }

  }


  template <class SCAL>
  FETIDP_BilinearForm<SCAL> :: 
  ~FETIDP_BilinearForm ()
  {
    if ( deletefeticoeff ) delete feticoeff;
    ;
  }
  
  template <class SCAL> 
  void FETIDP_BilinearForm<SCAL> ::
  AllocateMatrix ()
  {
    const FETIDP_CompoundSpace& feticompoundspace = 
      dynamic_cast< const FETIDP_CompoundSpace& > (this->fespace);

//     MatrixGraph * graph_mats =
//       const_cast<FETIDP_CompoundSpace& >(feticompoundspace) . GetGraph_S (false);

    int matrixsize = feticompoundspace . GetNDof();
    int nd_primalspace = feticompoundspace[0]->GetNDof();

    MatrixGraph * graph_mata =
      const_cast<FETIDP_CompoundSpace& >(feticompoundspace) . GetGraph (this->ma.GetNLevels()-1, false );


    BitArray primaldofs(matrixsize), dualinnerdofs(matrixsize);
    primaldofs.Clear();
    dualinnerdofs.Clear();
    for ( int i = 0; i < nd_primalspace; i++ ) 
      {
	if ( feticompoundspace . PrimalSpace() . IsPrimalDof(i) )
	  primaldofs.Set(i);
	else if ( feticompoundspace . PrimalSpace() . IsDualDof(i) )
	  dualinnerdofs.Set(i);
	else if ( !this->eliminate_internal )
	  dualinnerdofs.Set(i);
      }

    MatrixGraph * graph_matapp = graph_mata->CreatePartialGraph (primaldofs, primaldofs);
    MatrixGraph * graph_matapd = graph_mata->CreatePartialGraph (primaldofs, dualinnerdofs);
    MatrixGraph * graph_matadd = graph_mata->CreatePartialGraph (dualinnerdofs, dualinnerdofs);

    this->mats.Append (new FETIDP_Matrix<SCAL> (*graph_matapp, *graph_matapd, 
						*graph_matadd, 1, inversetype));
    delete graph_matapp;
    delete graph_matapd;
    delete graph_matadd;
    delete graph_mata;

    FETIDP_Matrix<SCAL> & fetidpmat = dynamic_cast<FETIDP_Matrix<SCAL> & > (*this->mats.Last()) ;

    for ( int i = 0; i < nd_primalspace; i++ )
      {
	if ( feticompoundspace . PrimalSpace() . IsPrimalDof(i) )
	  fetidpmat.PrimalDofs().Set(i);
	else if ( feticompoundspace . PrimalSpace() . IsDualDof(i) )
	  fetidpmat.DualDofs().Set(i);
      }

    int offset = feticompoundspace.GetStorageStart(1);
    int last = feticompoundspace[1]->GetNDof() + offset;
    for ( int j = offset; j < last; j++ )
      fetidpmat.LagrangeDofs().Set(j);

    //   *testout << "matrix graphs..  " << endl  << *this->mats.Last() << endl;

  }


  template <class SCAL> 
  BaseVector *  FETIDP_BilinearForm<SCAL> ::
  CreateVector() const
  {
    return new VVector<SCAL> (this->fespace.GetNDof());
  }

  template <class SCAL>
  void FETIDP_BilinearForm<SCAL> :: DoAssemble (LocalHeap & lh)
  {
    static int mattimer = NgProfiler::CreateTimer ("Matrix assembling");
    NgProfiler::RegionTimer reg (mattimer);
  
    try
      {
	this->ma.PushStatus ("Assemble Matrix");
      
	const FETIDP_CompoundSpace& feticompoundspace = 
	  dynamic_cast< const FETIDP_CompoundSpace& > (this->fespace);
	FETIDP_Matrix<SCAL> & fetidp_mat = dynamic_cast<FETIDP_Matrix<SCAL> & > (this->GetMatrix());
	ARRAY<int> dnums, idofs, idofs1, odofs;
      
	ElementTransformation eltrans;
      
	int ndof = feticompoundspace[0]->GetNDof();
	BitArray useddof(ndof);
	useddof.Clear();
      
	int ne = this->ma.GetNE();
	int nse = this->ma.GetNSE();
      
	BaseMatrix & mat = this->GetMatrix();
	mat = 0.0;
      
	bool hasbound = 0;
	bool hasinner = 0;
      
	for (int j = 0; j < this->NumIntegrators(); j++)
	  {
	    if (this->parts[j] -> BoundaryForm())
	      hasbound = 1;
	    else
	      hasinner = 1;
	  }
      
	clock_t prevtime = clock();
	if (hasinner && !this->diagonal)
	  {
	    void * heapp = lh.GetPointer();
	    for (int i = 0; i < ne; i++)
	      {

		if(this->elmat_ev) *testout << " Assemble Element " << i << endl;  
		if (clock()-prevtime > 0.1 * CLOCKS_PER_SEC)
		  {
		    cout << "\rassemble element " << i << "/" << ne << flush;
		    this->ma.SetThreadPercentage ( 100.0*i / (ne+nse) );
		    prevtime = clock();
		  }

		if ( this->ma.IsGhostEl( i ) ) {continue;}
	      
		lh.CleanUp(heapp);
	      
		if (!this->fespace.DefinedOn (this->ma.GetElIndex (i))) continue;
	      
		const FiniteElement & fel = feticompoundspace[0]->GetFE (i, lh);
		const FiniteElement & fel2 = this->fespace.GetFE(i, lh);
		const CompoundFiniteElement & cfel = 
		  dynamic_cast<const CompoundFiniteElement & > (fel2);
		this->ma.GetElementTransformation (i, eltrans, lh);
		feticompoundspace[0]->GetDofNrs (i, dnums);
	      
		if(fel.GetNDof() != dnums.Size())
		  {
		    cout << "fel:GetNDof() = " << fel.GetNDof() << endl;
		    cout << "dnums.Size() = " << dnums.Size() << endl;
		  
		    (*testout) << "fel:GetNDof() = " << fel.GetNDof() << endl;
		    (*testout) << "dnums.Size() = " << dnums.Size() << endl;
		    (*testout) << "dnums = " << dnums << endl;
		    throw Exception ( "Inconsistent number of degrees of freedom " );
		  }
	      
		FlatMatrix<SCAL> sum_elmat(dnums.Size()*this->fespace.GetDimension(), lh);
		sum_elmat = 0;
		for (int j = 0; j < this->NumIntegrators(); j++)
		  {
		    BilinearFormIntegrator & bfi = *this->parts[j];
		  
		  
		    if (bfi.BoundaryForm()) continue;
		    if (!bfi.DefinedOn (this->ma.GetElIndex (i))) continue;
		      
		    FlatMatrix<SCAL> elmat;
		    try
		      {
			clock_t starttime;
			double time;
			starttime = clock();
		      
		      
			static int elementtimer = NgProfiler::CreateTimer ("Element matrix integration");
			NgProfiler::StartTimer (elementtimer);
		      
			if (!this->diagonal)
			  bfi.AssembleElementMatrix (fel, eltrans, elmat, lh);
			else
			  {
			    FlatVector<double> diag;
			    bfi.AssembleElementMatrixDiag (fel, eltrans, diag, lh);
			    elmat.AssignMemory (diag.Size(), diag.Size(), lh);
			    elmat = 0.0;
			    for (int k = 0; k < diag.Size(); k++)
			      elmat(k,k) = diag(k);
			  }
 
				
			  
			NgProfiler::StopTimer (elementtimer);
		      
			time = double(clock() - starttime) / CLOCKS_PER_SEC;
			// cout << "time = " << time << endl;
			  
			// 			    if(elmat.Height() != dnums.Size())
			// 			      (*testout) << "OJE" << endl;
			if (this->printelmat) //|| elmat.Height() != dnums.Size())  //  || fel.ElementType() == ET_TET)
			  {
			    testout->precision(8);
			    (*testout) << "elnum= " << i << endl;
			    (*testout) << "eltype " << fel.ElementType() << endl;
			    (*testout) << "integrator " << bfi.Name() << endl;
			    (*testout) << "dnums = " << endl << dnums << endl;
			    (*testout) << "elmat = " << endl << elmat << endl;
			  
			  }

			if (this->elmat_ev)
			  {
			    // (*testout) << "elind = " << eltrans.GetElementIndex() << endl;
		
#ifdef LAPACK
			    LapackEigenSystem(elmat, lh);
#else
			    Vector<> lami(elmat.Height());
			    Matrix<> evecs(elmat.Height());
			      
			    CalcEigenSystem (elmat, lami, evecs);
			    (*testout) << "lami = " << endl << lami << endl;
#endif
			    //< "evecs = " << endl << evecs << endl;
			  
			  } 
		      
		      }
		    catch (Exception & e)
		      {
			e.Append (string("in Assemble Element Matrix, bfi = ") + 
				  bfi.Name() + string("\n"));
			throw;
		      }
		    catch (exception & e)
		      {
			throw (Exception (string(e.what()) +
					  string("in Assemble Element Matrix, bfi = ") + 
					  bfi.Name() + string("\n")));
		      }
		  
		    sum_elmat += elmat;
		  }

		feticompoundspace[0]->TransformMat (i, false, sum_elmat, TRANSFORM_MAT_LEFT_RIGHT);

	      
		    
		if (this->elmat_ev)
		  {
		  
		    (*testout) << "sum matrix:" << endl;
		  
#ifdef LAPACK
		    LapackEigenSystem(sum_elmat, lh);
#else
		    Vector<> lami(sum_elmat.Height());
		    Matrix<> evecs(sum_elmat.Height());
		  
		    CalcEigenSystem (sum_elmat, lami, evecs);
		    (*testout) << "lami = " << endl << lami << endl ; // << "evecs = " << endl << evecs << endl;
#endif
		  
		    /*
		      Matrix<SCAL> hmat(sum_elmat.Height());
		      hmat = sum_elmat;
		      Vector<SCAL> lami2(sum_elmat.Height());
		      LapackEigenValuesSymmetric (hmat, lami2);
		      (*testout) << "lapack ev = " << lami2 << endl;
		    */
		  }
	      

		if (this->eliminate_internal)
		  {
		    static int statcondtimer = NgProfiler::CreateTimer ("static condensation");
		    NgProfiler::StartTimer (statcondtimer);
		  
		  
		    cfel[0].GetInternalDofs(idofs1);
		    if (idofs1.Size())
		      {
			void * heapp = lh.GetPointer();
		      
			int size = sum_elmat.Height();
			int dim = size / dnums.Size();
		      
			idofs.SetSize (0);
			for (int j = 0; j < idofs1.Size(); j++)
			  for (int jj = 0; jj < dim; jj++)
			    idofs.Append (dim*idofs1[j]+jj);
		      
			int sizei = idofs.Size();
			int sizeo = size - sizei;
		      
			odofs.SetSize (0);
			for (int j = 0; j < size; j++)
			  {
			    bool idof = 0;
			    for (int k = 0; k < idofs.Size(); k++)
			      if (idofs[k] == j) idof = 1;
			    if (!idof) odofs.Append (j);
			  }
		      
			if (this->printelmat)
			  {
			    (*testout) << "idofs = " << endl << idofs << endl;
			    (*testout) << "odofs = " << endl << odofs << endl;
			  }

			FlatMatrix<SCAL> a(sizeo, sizeo, lh);
			FlatMatrix<SCAL> b(sizeo, sizei, lh);
			FlatMatrix<SCAL> c(sizeo, sizei, lh);
			FlatMatrix<SCAL> idc(sizeo, sizei, lh);
			FlatMatrix<SCAL> d(sizei, sizei, lh);
		      
			for (int k = 0; k < sizeo; k++)
			  for (int l = 0; l < sizeo; l++)
			    a(k,l) = sum_elmat(odofs[k], odofs[l]);
		      
			for (int k = 0; k < odofs.Size(); k++)
			  for (int l = 0; l < idofs.Size(); l++)
			    {
			      b(k,l) = sum_elmat(odofs[k], idofs[l]);
			      c(k,l) = sum_elmat(idofs[l], odofs[k]);
			    }
		      
			for (int k = 0; k < idofs.Size(); k++)
			  for (int l = 0; l < idofs.Size(); l++)
			    d(k,l) = sum_elmat(idofs[k], idofs[l]);
		      
#ifdef LAPACK

			NgProfiler::AddFlops (statcondtimer, double(sizei)*sizei*sizei/3);  // LU fact
			NgProfiler::AddFlops (statcondtimer, double(sizei)*sizei*sizeo);  
			NgProfiler::AddFlops (statcondtimer, double(sizei)*sizeo*sizeo);  
		      
		      
			// new Versions, July 07
			LapackAInvBt (d, b);
			LapackMultAddABt (b, c, -1, a);
		      
#else
			FlatMatrix<SCAL> invd(sizei, sizei, lh);
			CalcInverse (d, invd);
			d = invd;
			idc = c * Trans (invd);
			a -= b * Trans (idc);
#endif
		      
			if (this->printelmat) //  || fel.ElementType() == ET_TET)
			  {
			    testout->precision(8);
			    (*testout) << "Schur elmat = " << endl << a << endl;
			  }
		      
			if (this->elmat_ev)
			  {
			    testout->precision(8);
			  
			    (*testout) << "EV of Schur complement:" << endl;
#ifdef LAPACK
			    LapackEigenSystem(a, lh);
#else
			    Vector<> lami(a.Height());
			    Matrix<> evecs(a.Height());
			  
			    CalcEigenSystem (a, lami, evecs);
			    (*testout) << "lami = " << endl << lami << endl; // << "evecs = " << endl << evecs << endl;
#endif
			  
			  
			    /*
			      Matrix<SCAL> hmat(a.Height());
			      Vector<SCAL> lami2(a.Height());
			      hmat = a;
			      LapackEigenValuesSymmetric (hmat, lami2);
			      (*testout) << "lapack ev = " << lami2 << endl;
			    */
			  }
		      
		      
			for (int k = 0; k < odofs.Size(); k++)
			  for (int l = 0; l < odofs.Size(); l++)
			    sum_elmat(odofs[k], odofs[l]) = a(k,l);
		      
		      
                      
			if (this->linearform)
			  {
			    FlatVector<SCAL> elvec (size, lh);
			    this->linearform -> GetVector().GetIndirect (dnums, elvec);
                          
			    FlatVector<SCAL> hfi1(sizei, lh);
			    FlatVector<SCAL> hfi2(sizei, lh);
			    FlatVector<SCAL> hfo(sizeo, lh);
                          
			    for (int k = 0; k < idofs.Size(); k++)
			      hfi1(k) = elvec(idofs[k]);
			  
#ifdef LAPACK
			    hfo = b * hfi1;
#else
			    hfi2 = d * hfi1;
			    hfo = b * hfi2;
#endif
			    for (int k = 0; k < odofs.Size(); k++)
			      elvec(odofs[k]) -= hfo(k);
			  
			    this->linearform->GetVector().SetIndirect (dnums, elvec);
                          
			    for (int k = 0; k < idofs1.Size(); k++)
			      dnums[idofs1[k]] = -1;
			  }
		      
		      
			lh.CleanUp (heapp);
		      }
		    NgProfiler::StopTimer (statcondtimer);
		  }

		AddElementMatrix (dnums, dnums, sum_elmat, 1, i, lh);
	      
		for (int j = 0; j < dnums.Size(); j++)
		  if (dnums[j] != -1)
		    useddof.Set (dnums[j]);
	      }
	    lh.CleanUp(heapp);
	    cout << "\rassemble element " << ne << "/" << ne << endl;
	  }
      
	//(*testout) << "useddof0 " << useddof << endl;
      
	  









	if (hasbound)
	  {
	  
	    for (int i = 0; i < nse; i++)
	      {
		if (i % 100 == 0)
		  cout << "\rassemble surface element " << i << "/" << nse << flush;
	      
	      
		if ( this->ma.IsGhostSEl ( i ) ) continue;
	      
		this->ma.SetThreadPercentage ( 100.0*(ne+i) / (ne+nse) );
	      
		lh.CleanUp();
	      
		if (!this->fespace.DefinedOnBoundary (this->ma.GetSElIndex (i))) continue;
	      
		const FiniteElement & fel = feticompoundspace[0]->GetSFE (i, lh);
		const FiniteElement & fel2 = this->fespace.GetSFE (i, lh);
		const CompoundFiniteElement & cfel =
		  dynamic_cast<const CompoundFiniteElement & > (fel2);   
	      
		this->ma.GetSurfaceElementTransformation (i, eltrans, lh);
		feticompoundspace[0]->GetSDofNrs (i, dnums);
	      
		for (int j = 0; j < this->NumIntegrators(); j++)
		  {
		    const BilinearFormIntegrator & bfi = *this->parts[j];
		  
		    if (!bfi.BoundaryForm()) continue;
		  
		    if (!bfi.DefinedOn (this->ma.GetSElIndex (i))) continue;		    
		  
		  
		    for (int k = 0; k < dnums.Size(); k++)
		      if (dnums[k] != -1)
			useddof.Set (dnums[k]);
		  
		    FlatMatrix<SCAL> elmat;
		    bfi.AssembleElementMatrix (fel, eltrans, elmat, lh);
		  
		    feticompoundspace[0]->TransformMat (i, true, elmat, TRANSFORM_MAT_LEFT_RIGHT);
		  
		    if (this->printelmat)
		      {
			testout->precision(8);
		      
			(*testout) << "surface-elnum= " << i << endl;
			(*testout) << "integrator " << bfi.Name() << endl;
			(*testout) << "dnums = " << endl << dnums << endl;
			(*testout) << "element-index = " << eltrans.GetElementIndex() << endl;
			(*testout) << "elmat = " << endl << elmat << endl;
		      }
		  
		  
		    if (this->elmat_ev)
		      {
			testout->precision(8);
		      
			(*testout) << "elind = " << eltrans.GetElementIndex() << endl;
#ifdef LAPACK
			LapackEigenSystem(elmat, lh);
#else
			Vector<> lami(elmat.Height());
			Matrix<> evecs(elmat.Height());
		      
			CalcEigenSystem (elmat, lami, evecs);
			(*testout) << "lami = " << endl << lami << endl;
#endif
			// << "evecs = " << endl << evecs << endl;
		      }

		    AddElementMatrix (dnums, dnums, elmat, 0, i, lh);
		  }
	      }
	    cout << "\rassemble surface element " << nse << "/" << nse << endl;	  
	  }
      
	this->ma.SetThreadPercentage ( 100.0 );
      
	lh.CleanUp();
      
	bool assembledspecialelements(false);
	for (int i = 0; i < this->fespace.specialelements.Size(); i++)
	  {
	    if(i%10 == 0) cout << "\rassemble special element " << i << "/" << this->fespace.specialelements.Size() << flush;
	  
	    const SpecialElement & el = *this->fespace.specialelements[i];
		
	    el.GetDofNrs (dnums);
		
	    for (int j = 0; j < dnums.Size(); j++)
	      if (dnums[j] != -1)
		useddof.Set (dnums[j]);
	      
	    FlatMatrix<SCAL> elmat;
	    el.Assemble (elmat, lh);
	      
	    AddElementMatrix (dnums, dnums, elmat, 0, i, lh);
	    assembledspecialelements = true;

	    lh.CleanUp();
	  }
	if(assembledspecialelements) cout << "\rassemble special element " 
					  << this->fespace.specialelements.Size() << "/" << this->fespace.specialelements.Size() << endl;

	  
	// add eps to avoid empty lines
	FlatMatrix<SCAL> elmat (this->fespace.GetDimension(), lh);
	elmat = 0;
	dnums.SetSize(1);

	void * p = lh.GetPointer();

	if (this->print)
	  (*testout) << "mat = " << endl << this->GetMatrix() << endl;

	int cntused = 0;
	for (int i = 0; i < useddof.Size(); i++)
	  if (useddof.Test(i))
	    cntused++;
	cout << "used " << cntused
	     << ", unused = " << useddof.Size()-cntused
	     << ", total = " << useddof.Size() << endl;

	
	/// ***************************************************** ///
	  /// LOOP OVER ELEMENTS TO GET SCHUR COMPLEMENT S
	  /// ***************************************************
	//(*testout) << "useddof " << useddof << endl;

	  BaseMatrix & sbasemat = fetidp_mat.GetSMat();
	  BaseMatrix & appbasemat = fetidp_mat.GetAppMat();
	  SparseMatrix<SCAL> & smat = dynamic_cast<SparseMatrix<SCAL> & > (sbasemat );
	  SparseMatrix<SCAL> & appmat = dynamic_cast<SparseMatrix<SCAL> & > (appbasemat );

	  smat.AsVector() = appmat.AsVector();

	  for ( int el = 0; el < ne; el++ )
	    ComputeSMatrix ( el, lh );



	this->ma.PopStatus ();

      }
    catch (Exception & e)
      {
	e.Append (string ("in Assemble BilinearForm '") + 
		  this->GetName() + string("'\n"));
	throw;
      }
    catch (exception & e)
      {
	throw (Exception (string(e.what()) +
			  string("\n in Assemble BilinearForm '" + this->GetName() + "'\n")));
      }
    
    SparseMatrix<SCAL> * bmat = dynamic_cast<SparseMatrix<SCAL>*> 
      (this->CreateBMat()) ;
    SparseMatrix<SCAL> * barbmat = dynamic_cast<SparseMatrix<SCAL>*> 
      (this->CreateBarBMat(lh)) ;
    dynamic_cast<FETIDP_Matrix<SCAL> & > (*this->mats.Last() ).SetBMat(bmat);
    dynamic_cast<FETIDP_Matrix<SCAL> & > (*this->mats.Last() ).SetBarBMat(barbmat);
    
    dynamic_cast<FETIDP_Matrix<SCAL> & > (*this->mats.Last() ).CalcInverses(this->eliminate_internal);
    ModifyRightHandSide();
  }
  
  template <class SCAL> 
  void FETIDP_BilinearForm<SCAL> ::
  AddElementMatrix (const ARRAY<int> & dnums1,
		    const ARRAY<int> & dnums2,
		    const FlatMatrix<SCAL> & elmat,
		    bool inner_element, int elnr,
		    LocalHeap & lh)
  {
    const FETIDP_CompoundSpace& feticompoundspace = 
      dynamic_cast< const FETIDP_CompoundSpace& > (this->fespace);

    FETIDP_Matrix<SCAL> & fetidp_mat = dynamic_cast<FETIDP_Matrix<SCAL> & > (this->GetMatrix());
    BaseMatrix & sbasemat = fetidp_mat.GetSMat();
    BaseMatrix & apdbasemat = fetidp_mat.GetApdMat();
    BaseMatrix & appbasemat = fetidp_mat.GetAppMat();
    BaseMatrix & addbasemat = fetidp_mat.GetAddMat();
    BaseMatrix & invbasesdd = fetidp_mat.GetInvSdd();
    SparseMatrix<SCAL> & smat = dynamic_cast<SparseMatrix<SCAL> & > (sbasemat );
    SparseMatrix<SCAL> & appmat = dynamic_cast<SparseMatrix<SCAL> & > (appbasemat );
    SparseMatrix<SCAL> & apdmat = dynamic_cast<SparseMatrix<SCAL> & > (apdbasemat );
    SparseMatrix<SCAL> & addmat = dynamic_cast<SparseMatrix<SCAL> & > (addbasemat );


    bool elim_int = this->eliminate_internal;

    // structure of element matrix
    //
    //         A_pp   A_pd  A_pi
    //   A  =  A_dp   A_dd  A_di
    //         A_ip   A_id  A_ii

    if ( inner_element )
      {
	void * heapp = lh.GetPointer();
	int np, nd, ni;

	int nd_primalspace;
	nd_primalspace = feticompoundspace[0]->GetNDof();

	ARRAY<int> primalpos, dualpos, innerpos;
	ARRAY<int> dualdofs;
	BitArray ispddof(nd_primalspace);
	primalpos.SetSize(0);
	dualpos.SetSize(0);
	innerpos.SetSize(0);
	ispddof.Clear();
	dualdofs.SetSize(0);


	for ( int i = 0; i < dnums1.Size(); i++ )
	  {
	    if ( dnums1[i] == -1 ) continue;

	    if ( feticompoundspace . PrimalSpace() . IsPrimalDof( dnums1[i] ) )
	      { primalpos.Append(i); ispddof.Set(i); }
	    else if ( feticompoundspace . PrimalSpace() . IsDualDof ( dnums1[i] ) )
	      { dualpos.Append(i); ispddof.Set(i); dualdofs.Append(dnums1[i]); }
	    else
	      { innerpos.Append(i); if ( !this->eliminate_internal ) ispddof.Set(i); dualdofs.Append(dnums1[i]);}
	  }

	np = primalpos.Size();
	nd = dualpos.Size();
	ni = innerpos.Size();


	// Assemble A matrix
	//
	//           A_pp   A_pr   0
	//   A  =    A_rp   A_rr   0
	//            0      0     0
	// 
	// S_pp = A_pp - A_rp inv(A_rr) A_pr
	// is done later

	int nr = (elim_int) ? nd : nd + ni;
	/*
	FlatMatrix<SCAL> a_pp(np, np, lh);
	FlatMatrix<SCAL> a_pr(np, nr, lh);
	FlatMatrix<SCAL> a_rp(np, nr, lh);
	FlatMatrix<SCAL> a_rr(nr, nr, lh);
	FlatMatrix<SCAL> inv_arr_arp (np, nr, lh);
	FlatMatrix<SCAL> a_pdpd (np+nd, np+nd, lh);
	*/

	for ( int k = 0; k < np; k++ )
	  for ( int j = 0; j < np; j++ )
	    {
// 	      a_pp(k,j) = elmat(primalpos[k], primalpos[j]);
// 	      a_pdpd (k,j) = elmat(primalpos[k], primalpos[j]);
	      appmat(dnums1[primalpos[k]], dnums1[primalpos[j]] ) += elmat(primalpos[k], primalpos[j]);
	    }

	for ( int k = 0; k < np; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
// 	      a_pr (k,j) = elmat(primalpos[k], dualpos[j]);
// 	      a_rp (k,j) = elmat(dualpos[j], primalpos[k]);
// 	      a_pdpd (k,j+np) = elmat(primalpos[k], dualpos[j]);
// 	      a_pdpd (j+np,k) = elmat(dualpos[j], primalpos[k]);
	      apdmat(dnums1[primalpos[k]], dnums1[dualpos[j]] ) += elmat(primalpos[k], dualpos[j]);
	    }


	for ( int k = 0; k < nd; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
// 	      a_rr (k,j) = elmat(dualpos[k], dualpos[j]);
// 	      a_pdpd (k+nd,j+nd) = elmat(dualpos[k], dualpos[j]);
	      addmat(dnums1[dualpos[k]], dnums1[dualpos[j]] ) += elmat(dualpos[k], dualpos[j]);
	    }


	if ( !elim_int )
	  {
	    for ( int k = 0; k < np; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
// 		  a_pr (k,j+nd) = elmat(primalpos[k], innerpos[j]);
// 		  a_rp (k,j+nd) = elmat(innerpos[j], primalpos[k]);
		  apdmat(dnums1[primalpos[k]], dnums1[innerpos[j]] ) +=  elmat(primalpos[k], innerpos[j]);
		}
	  
	    for ( int k = 0; k < ni; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
// 		  a_rr (k+nd,j+nd) = elmat(innerpos[k], innerpos[j]);
		  addmat(dnums1[innerpos[k]], dnums1[innerpos[j]] ) += elmat(innerpos[k], innerpos[j]);
		}
	    for ( int k = 0; k < nd; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
// 		  a_rr (k,j+nd) = elmat(dualpos[k], innerpos[j]);
// 		  a_rr (j+nd,k) = elmat(innerpos[j], dualpos[k]);
		  addmat(dnums1[dualpos[k]], dnums1[innerpos[j]] ) += elmat(dualpos[k], innerpos[j]);
		  addmat(dnums1[innerpos[j]], dnums1[dualpos[k]] ) += elmat(innerpos[j], dualpos[k]);
		}
	  }

	lh.CleanUp(heapp);
      }
    else  // surface elements
      {


	void * heapp = lh.GetPointer();
	int np, nd, ni;

	int ndof = feticompoundspace.GetNDof();
	int dim = this->ma.GetDimension();
	int nd_primalspace;
	nd_primalspace = feticompoundspace[0]->GetNDof();

	ARRAY<int> primalpos, dualpos, innerpos;
	ARRAY<int> dualdofs;
	primalpos.SetSize(0);
	dualpos.SetSize(0);
	innerpos.SetSize(0);
	dualdofs.SetSize(0);
	BitArray ispddof ( ndof );
	ispddof.Clear();

	for ( int i = 0; i < dnums1.Size(); i++ )
	  {
	    if ( dnums1[i] == -1 ) continue;

	    if ( feticompoundspace . PrimalSpace() . IsPrimalDof( dnums1[i] ) )
	      { primalpos.Append(i); ispddof.Set(i); }
	    else if ( feticompoundspace . PrimalSpace() . IsDualDof ( dnums1[i] ) )
	      { dualpos.Append(i); ispddof.Set(i); dualdofs.Append( dnums1[i]); }
	    else
	      innerpos.Append(i);
	  }

	np = primalpos.Size();
	nd = dualpos.Size();

// 	// dofs on surface are primal --> no schur complement


	
	// Assemble A matrix
	//
	//           A_pp   A_pr  
	//   A  =    A_rp   A_rr  
	// 
	// S_pp = A_pp - A_rp inv(A_rr) A_pr
	// is done later

	/*
	FlatMatrix<SCAL> a_pp(np, np, lh);
	FlatMatrix<SCAL> a_pr(np, nd, lh);
	FlatMatrix<SCAL> a_rp(np, nd, lh);
	FlatMatrix<SCAL> a_rr(nd, nd, lh);
	FlatMatrix<SCAL> inv_arr_arp (np, nd, lh);
	*/

	for ( int k = 0; k < np; k++ )
	  for ( int j = 0; j < np; j++ )
	    {
// 	      a_pp(k,j) = elmat(primalpos[k], primalpos[j]);
	      appmat(dnums1[primalpos[k]], dnums1[primalpos[j]] ) += elmat(primalpos[k], primalpos[j]);
	    }

	for ( int k = 0; k < np; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
// 	      a_pr (k,j) = elmat(primalpos[k], dualpos[j]);
// 	      a_rp (k,j) = elmat(dualpos[j], primalpos[k]);
	      apdmat(dnums1[primalpos[k]], dnums1[dualpos[j]] ) += elmat(primalpos[k], dualpos[j]);
	    }

	for ( int k = 0; k < nd; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
// 	      a_rr (k,j) = elmat(dualpos[k], dualpos[j]);
	      addmat(dnums1[dualpos[k]], dnums1[dualpos[j]]) += elmat(dualpos[k], dualpos[j]);
	    }


	lh.CleanUp(heapp);

      }

    ;
  }

  template <class SCAL>
  void FETIDP_BilinearForm<SCAL> :: ComputeSMatrix ( int elnr, LocalHeap & lh )

  {
    const FETIDP_CompoundSpace& feticompoundspace = 
      dynamic_cast< const FETIDP_CompoundSpace& > (this->fespace);

    FETIDP_Matrix<SCAL> & fetidpmat = dynamic_cast<FETIDP_Matrix<SCAL>& > (this->GetMatrix() );
  
    BaseMatrix & sbasemat = fetidpmat.GetSMat();
    BaseMatrix & apdbasemat = fetidpmat.GetApdMat();
    BaseMatrix & addbasemat = fetidpmat.GetAddMat();
    BaseMatrix & appbasemat = fetidpmat.GetAppMat();
    SparseMatrix<SCAL> & smat = dynamic_cast<SparseMatrix<SCAL> & > (sbasemat );
    SparseMatrix<SCAL> & appmat = dynamic_cast<SparseMatrix<SCAL> & > (appbasemat );
    SparseMatrix<SCAL> & apdmat = dynamic_cast<SparseMatrix<SCAL> & > (apdbasemat );
    SparseMatrix<SCAL> & addmat = dynamic_cast<SparseMatrix<SCAL> & > (addbasemat );

    ARRAY<int> dnums;
    bool elim_int = this->eliminate_internal;
  
    feticompoundspace[0]->GetDofNrs ( elnr, dnums );

    void * heapp = lh.GetPointer();
    int np, nd, ni;
    int ndof = feticompoundspace . GetNDof();

    int nd_primalspace;
    nd_primalspace = feticompoundspace[0]->GetNDof();
    
    ARRAY<int> primalpos, dualpos, innerpos;
    ARRAY<int> dualdofs;
    BitArray ispddof(ndof);
    primalpos.SetSize(0);
    dualpos.SetSize(0);
    innerpos.SetSize(0);
    ispddof.Clear();
    dualdofs.SetSize(0);
    

    for ( int i = 0; i < dnums.Size(); i++ )
      {
	if ( dnums[i] == -1 ) continue;
	
	if ( feticompoundspace . PrimalSpace() . IsPrimalDof( dnums[i] ) )
	  { 
	    primalpos.Append(i); 
	    ispddof.Set(i); 
	  }
	else if ( feticompoundspace . PrimalSpace() . IsDualDof ( dnums[i] ) )
	  { 
	    dualpos.Append(i); 
	    ispddof.Set(i); dualdofs.Append(dnums[i]); 
	  }
	else
	  { 
	    innerpos.Append(i); 
	    if ( !this->eliminate_internal ) ispddof.Set(i); dualdofs.Append(dnums[i]);
	  }
      }

    np = primalpos.Size();
    nd = dualpos.Size();
    ni = innerpos.Size();

    
    // Assemble Schur complement S_pp w.r.t primal dofs
    //
    //           A_pp   A_pr   0
    //   A  =    A_rp   A_rr   0
    //            0      0     0
    // 
    // S_pp = A_pp - A_rp inv(A_rr) A_pr
    //

    int nr = (elim_int) ? nd : nd + ni;
    FlatMatrix<SCAL> a_pp(np, np, lh);
    FlatMatrix<SCAL> a_pr(np, nr, lh);
    FlatMatrix<SCAL> a_rp(np, nr, lh);
    FlatMatrix<SCAL> a_rr(nr, nr, lh);
    FlatMatrix<SCAL> inv_arr_arp (np, nr, lh);
	
    for ( int k = 0; k < np; k++ )
      for ( int j = 0; j < np; j++ )
	{
	  a_pp(k,j) = 0.0; // appmat( dnums[primalpos[k]], dnums[primalpos[j]] );
	}

	for ( int k = 0; k < np; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
	      a_pr (k,j) = apdmat ( dnums[primalpos[k]], dnums[dualpos[j]]);
	      a_rp (k,j) = a_pr (k,j);
	    }


	for ( int k = 0; k < nd; k++ )
	  for ( int j = 0; j < nd; j++ )
	    {
	      a_rr (k,j) = addmat (dnums[dualpos[k]], dnums[dualpos[j]]);
	    }


	if ( !elim_int )
	  {
	    for ( int k = 0; k < np; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
		  a_pr (k,j+nd) = apdmat( dnums[primalpos[k]], dnums[innerpos[j]]);
		  a_rp (k,j+nd) = a_pr(k,j+nd);
		}
	  
	    for ( int k = 0; k < ni; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
		  a_rr (k+nd,j+nd) = addmat( dnums[innerpos[k]], dnums[innerpos[j]]);
		}
	    for ( int k = 0; k < nd; k++ )
	      for ( int j = 0; j < ni; j++ )
		{
		  a_rr (k,j+nd) = addmat(dnums[dualpos[k]], dnums[innerpos[j]]);
		  a_rr (j+nd,k) = a_rr(k, nd+j);
		}
	  }


	if ( nr > 0 && np > 0 )
	  {
	    LapackInverse (a_rr);
	    LapackMultABt (a_rp, a_rr, inv_arr_arp);
	    LapackMultAddABt (a_pr, inv_arr_arp, -1, a_pp);
	  }

	// add schur complement element matrix

// 	cout << "adding to smat " << endl;

	for (int k = 0; k < np; k++ )
	  for ( int j = 0; j < np; j++ )
	    {
	      smat(dnums[primalpos[k]], dnums[primalpos[j]] ) += a_pp(k,j);
	    }


      
	lh.CleanUp(heapp);
    ;
  }

  template <class SCAL>
  void FETIDP_BilinearForm<SCAL> :: ModifyRightHandSide ()
  {
    const FETIDP_Matrix<SCAL> & fetidpmat = dynamic_cast<const FETIDP_Matrix<SCAL>& > (*this->mats.Last() );
  
    const BaseMatrix & smat = fetidpmat.GetSMat();
    const BaseMatrix & apdmat = fetidpmat.GetApdMat();
    const BaseMatrix & addmat = fetidpmat.GetAddMat();
    const BaseMatrix & bmat = fetidpmat.GetBMat();
  
    const BaseMatrix & inv_spp = fetidpmat.GetInvSpp();
    const BaseMatrix & inv_sdd = fetidpmat.GetInvSdd();
  
    BaseVector & vecf = this->linearform -> GetVector();
    BaseVector & help = *vecf.CreateVector();
    BaseVector & help2 = *vecf.CreateVector();
  
    FlatVector<SCAL> ff = 
      dynamic_cast<T_BaseVector<SCAL> &> 
      (vecf).FV();
    FlatVector<SCAL> fh = 
      dynamic_cast<T_BaseVector<SCAL> &> (help).FV();
    FlatVector<SCAL> fh2 = 
      dynamic_cast<T_BaseVector<SCAL> &> 
      (help2).FV();
  
    int nd = ff.Size();
  
    BitArray isiddof (nd);
    isiddof.Clear();
    for ( int i = 0; i < nd; i++ )
      if ( fetidpmat.DualDofs().Test(i) )
	isiddof.Set(i);
      else if ( !this->eliminate_internal )
	if ( !fetidpmat.PrimalDofs().Test(i) && !fetidpmat.LagrangeDofs().Test(i) )
	  isiddof.Set(i);

    help = inv_sdd * vecf;

    help2 = apdmat * help;

    help2 -= vecf;

    bool fkannmanweglassen = true;
    bool h2kannmanweglassen = true;

    for ( int i = 0; i < nd; i++ )
      if (fetidpmat.LagrangeDofs().Test(i) )
	{
	  ff(i) = fh2(i);
	}

    help = inv_spp * help2;

    help2 = Transpose(apdmat) * help;

    help2 += vecf;

    help = inv_sdd * help2;

    help2 = bmat * help;
    vecf += help2;

    delete &help;
    delete &help2;
  }


  template <class SCAL> 
  void FETIDP_BilinearForm<SCAL> :: ComputePrimalDual ( BaseVector & vecu, LocalHeap & lh ) const
  {
    const FETIDP_Matrix<SCAL> & fetidpmat = dynamic_cast<const FETIDP_Matrix<SCAL>& > (*this->mats.Last() );

    const FETIDP_CompoundSpace & compoundspace = dynamic_cast<const FETIDP_CompoundSpace &> (this->fespace);

    const BaseMatrix & smat = fetidpmat.GetSMat();
    const BaseMatrix & apdmat = fetidpmat.GetApdMat();
    const BaseMatrix & addmat = fetidpmat.GetAddMat();
    const BaseMatrix & bmat = fetidpmat.GetBMat();
  
    const BaseMatrix & inv_spp = fetidpmat.GetInvSpp();
    const BaseMatrix & inv_sdd = fetidpmat.GetInvSdd();
  
    BaseVector & vecf = this->linearform -> GetVector();
    BaseVector & help = *vecf.CreateVector();
    BaseVector & help2 = *vecf.CreateVector();
  
    const MeshAccess & ma = this->ma;
    FlatVector<SCAL> ff = 
      dynamic_cast<T_BaseVector<SCAL> &> 
      (vecf).FV();
    FlatVector<SCAL> fu = 
      dynamic_cast<T_BaseVector<SCAL> &> 
      (vecu).FV();
    FlatVector<SCAL> fh = 
      dynamic_cast<T_BaseVector<SCAL> &> (help).FV();
    FlatVector<SCAL> fh2 = 
      dynamic_cast<T_BaseVector<SCAL> &> 
      (help2).FV();

    int nd = ff.Size();

    BitArray isiddof (nd);
    isiddof.Clear();
    for ( int i = 0; i < nd; i++ )
      if ( fetidpmat.DualDofs().Test(i) )
	isiddof.Set(i);
      else if ( !this->eliminate_internal )
	if ( !fetidpmat.PrimalDofs().Test(i) && !fetidpmat.LagrangeDofs().Test(i) )
	  isiddof.Set(i);

    // 1) u_pi = Spp^-1 ( f_p - A_pd A_dd^-1 fd )
    help2 = vecf - Transpose(bmat) * vecu;

    help = inv_sdd * (help2);

    help2 -= apdmat * help;

    help = inv_spp * help2;

    for ( int i = 0; i < nd; i++ )
      if ( fetidpmat.PrimalDofs().Test(i) )
// 	fh(i) = 0;
//       else
	{
	  fu(i) = fh(i);
	}

    help2 = vecf - Transpose(bmat) * vecu - Transpose(apdmat) * help;
    help = inv_sdd * help2;

    for ( int i = 0; i < nd; i++ )
      if ( isiddof.Test(i) )
	fu(i) = fh(i);
      else if ( fetidpmat.LagrangeDofs().Test(i) )
	ff(i) = 0.0;

    const FETIDP_LagrangeSpace & lagspace = 
      dynamic_cast<const FETIDP_LagrangeSpace &> (*compoundspace[1]);

    /*
    for ( int i = 0; i < compoundspace[1]->GetNDof(); i++ )
      {
	INT<2> dofs = lagspace.DualDofs(i);
	*testout << "dof " << i << ": dual dofs " << dofs << ", values " << fu(dofs[0]);
	if ( dofs[1] >= 0 ) *testout << ", " << fu(dofs[1]) << endl;
	else *testout << endl;
      }
    */

  }


  template <class SCAL>
  void FETIDP_BilinearForm<SCAL> :: ComputeInternal (BaseVector & u, LocalHeap & lh) const
  {
    dynamic_cast<const FETIDP_BilinearForm<SCAL> &> (*this) . ComputePrimalDual( u, lh );

    if (!this->eliminate_internal) return;
    if (!this->linearform)
      throw Exception ("ComputeInternal needs a linear-form");


    static int timer = NgProfiler::CreateTimer ("Compute Internal");
    NgProfiler::RegionTimer reg (timer);

    const MeshAccess & ma = this->ma;
    const FESpace & fespace = this->fespace;
    const FETIDP_CompoundSpace & feticompoundspace = 
      dynamic_cast<const FETIDP_CompoundSpace &>(fespace);

    try
      {
	ma.PushStatus ("Compute Internal");

	ARRAY<int> dnums, idofs;
	ElementTransformation eltrans;
      
	int ne = ma.GetNE();

	bool hasinner = 0;

	for (int j = 0; j < this->NumIntegrators(); j++)
	  {
	    if (!this->parts[j] -> BoundaryForm())
	      hasinner = 1;
	  }	  
	 
	clock_t prevtime = clock();
	if (hasinner)
	  {
	    for (int i = 0; i < ne; i++)
	      {

		// PARALLEL
		if ( ma.IsGhostEl( i ) ) {continue;}
		
		if (clock()-prevtime > 0.1 * CLOCKS_PER_SEC)
		  {
		    cout << "\rcompute internal element " << i << "/" << ne << flush;
		    ma.SetThreadPercentage ( 100.0*i / ne );
		    prevtime = clock();
		  }
	      
		lh.CleanUp();
	      
		if (!fespace.DefinedOn (ma.GetElIndex (i))) continue;
		  
		const FiniteElement & fel = feticompoundspace[0]->GetFE (i, lh);
	        const FiniteElement & fel2 = fespace.GetFE(i, lh);
		const CompoundFiniteElement & cfel = dynamic_cast<const CompoundFiniteElement &> (fel2);

		ma.GetElementTransformation (i, eltrans, lh);
	        feticompoundspace[0]->GetDofNrs (i, dnums);
	      
	      
		FlatMatrix<SCAL> sum_elmat(dnums.Size()*fespace.GetDimension(), lh);
		sum_elmat = 0;
		for (int j = 0; j < this->NumIntegrators(); j++)
		  {
		    const BilinearFormIntegrator & bfi = *this->parts[j];
		  
		    if (bfi.BoundaryForm()) continue;
		    if (!bfi.DefinedOn (ma.GetElIndex (i))) continue;
		  
		    FlatMatrix<SCAL> elmat;
		    bfi.AssembleElementMatrix (fel, eltrans, elmat, lh);
		  
		    sum_elmat += elmat;
		  }
	      
		feticompoundspace[0]->TransformMat (i, false, sum_elmat, TRANSFORM_MAT_LEFT_RIGHT);
	      
		fel.GetInternalDofs(idofs);

		// (*testout) << "compute internal element = " << i << endl;
		//		(*testout) << "mat = " << endl << sum_elmat << endl;

		if (idofs.Size())
		  {
		    int dim = sum_elmat.Height() / dnums.Size();
		  
		    FlatVector<SCAL> elf (sum_elmat.Height(), lh);
		    FlatVector<SCAL> elu (sum_elmat.Height(), lh);
		    FlatMatrix<SCAL> ai(dim*idofs.Size(), lh);
		    FlatMatrix<SCAL> inv_ai(dim*idofs.Size(), lh);
		    FlatVector<SCAL> resi(dim*idofs.Size(), lh);
		    FlatVector<SCAL> wi(dim*idofs.Size(), lh);
		  
		    this->linearform->GetVector().GetIndirect (dnums, elf);
		    u.GetIndirect (dnums, elu);
		  

		    // compute residuum
		    elf -= sum_elmat * elu;

		    for (int jj = 0; jj < idofs.Size(); jj++)
		      for (int j = 0; j < dim; j++)
			{
			  for (int kk = 0; kk < idofs.Size(); kk++)
			    for (int k = 0; k < dim; k++)
			      {
				ai(jj*dim+j, kk*dim+k) =
				  sum_elmat(idofs[jj]*dim+j, idofs[kk]*dim+k);
			      }
			  resi(jj*dim+j) = elf(idofs[jj]*dim+j);
			}
		

		    //		    *testout << "residuum = " << resi << endl;

		    // CholeskyFactors<SCAL> inv_ai(ai);
		    // inv_ai.Mult (resi, wi);
		    
#ifdef LAPACK
                    FlatMatrix<SCAL> mresi(1,resi.Size(), &resi(0));
                    LapackAInvBt (ai, mresi, 'T');
// 		    LapackInverse(ai);
//                  wi = ai * Trans(mresi);
		    wi = resi;
#else
		    CalcInverse (ai, inv_ai);
		    wi = inv_ai * resi;
#endif

		    
		    //		    *testout << "inv_ai = " << endl << inv_ai << endl;
		    
		    for (int jj = 0; jj < idofs.Size(); jj++)
		      for (int j = 0; j < dim; j++)
			elu(idofs[jj]*dim+j) += wi(jj*dim+j);
		  
		    //		    *testout << "elu = " << endl << elu << endl;

		    u.SetIndirect (dnums, elu);
		  }
	      }
	    cout << "\rcompute internal element " << ne << "/" << ne << endl;
	  }
	  
	ma.SetThreadPercentage ( 100.0 );
      
	ma.PopStatus ();
      }
    catch (Exception & e)
      {
	stringstream ost;
	ost << "in ComputeInternal" << endl;
	e.Append (ost.str());
	throw;
      }
    catch (exception & e)
      {
	throw (Exception (string(e.what()) +
			  string("\n in ComputeInternal\n")));
      }

  }






  template <class SCAL>
  BaseMatrix * FETIDP_BilinearForm<SCAL> :: CreateBMat () const
  {

    int nv = this->ma.GetNV();
    
    const FETIDP_Matrix<SCAL> & fetidpmat = 
      dynamic_cast<const FETIDP_Matrix<SCAL>& > (*this->mats.Last() );
    
    const FETIDP_CompoundSpace & compoundspace = 
      dynamic_cast<const FETIDP_CompoundSpace &> (this->fespace);
    
    const FETIDP_LagrangeSpace & lagrangespace = 
      dynamic_cast<const FETIDP_LagrangeSpace&> (*compoundspace[1]);
    int ndof = compoundspace.GetNDof();

    int base_lag = compoundspace.GetStorageStart(1);
    int ndof_lag = compoundspace[1]->GetNDof();
    ARRAY<int> elsperrow(ndof);
    for ( int i = 0; i < ndof; i++ )
      if ( !fetidpmat.LagrangeDofs().Test(i) )
	elsperrow[i] = 0;
      else if ( lagrangespace.DualDofs(i-base_lag)[1] >= 0 )
	elsperrow[i] = 2;
      else
	elsperrow[i] = 1;

    MatrixGraph * graph = new MatrixGraph(elsperrow);

    SparseMatrix<SCAL> * bmat = new SparseMatrix<SCAL> (*graph, 1);
    delete graph;

    bmat->AsVector() = 0.0;

    for ( int i = base_lag; i < ndof; i++ )
      {
	INT<2> dualdofs = lagrangespace.DualDofs(i-base_lag);
	bmat->CreatePosition (i, dualdofs[0]);
	if ( dualdofs[1] >= 0 )
	bmat->CreatePosition (i, dualdofs[1]);
      }
    for ( int i = base_lag; i < ndof; i++ )
      {
	INT<2> dualdofs = lagrangespace.DualDofs(i-base_lag);
	// 	*testout << "to row " << i << ":  dualdofs[" << i-base_lag <<"] = " << dualdofs << endl;
	(*bmat)(i, dualdofs[0]) = 1;
	if ( dualdofs[1] >= 0 )
	(*bmat)(i, dualdofs[1]) = -1;
      }
    return bmat;
  }


  template <class SCAL>
  BaseMatrix * FETIDP_BilinearForm<SCAL> :: CreateBarBMat (LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();

    int nv = this->ma.GetNV();
    int dim = this->ma.GetDimension();
    int nel = this->ma.GetNE();
    int ned = this->ma.GetNEdges();
    int nfa = ( dim == 2 ) ? 0 : this->ma.GetNFaces();
    int nsel = this->ma.GetNSE();
    
    const FETIDP_Matrix<SCAL> & fetidpmat = 
      dynamic_cast<const FETIDP_Matrix<SCAL>& > (*this->mats.Last() );
    
    const FETIDP_CompoundSpace & compoundspace = 
      dynamic_cast<const FETIDP_CompoundSpace &> (this->fespace);
    
    const FETIDP_LagrangeSpace & lagrangespace = 
      dynamic_cast<const FETIDP_LagrangeSpace&> (*compoundspace[1]);

    const FETIDP_PrimalSpace & primalspace =
      compoundspace . PrimalSpace();

    int ndof = compoundspace.GetNDof();

    int base_lag = compoundspace.GetStorageStart(1);
    int ndof_lag = compoundspace[1]->GetNDof();


    ARRAY<int> elsperrow(ndof);
    for ( int i = 0; i < ndof; i++ )
      if ( !fetidpmat.LagrangeDofs().Test(i) )
	elsperrow[i] = 0;
      else if ( lagrangespace.DualDofs(i-base_lag)[1] >= 0 )
	elsperrow[i] = 2;
      else
	elsperrow[i] = 1;

    MatrixGraph * graph = new MatrixGraph(elsperrow);

    SparseMatrix<SCAL> * bmat = new SparseMatrix<SCAL> (*graph, 1);
    delete graph;

    bmat->AsVector() = 0.0;

    ARRAY<double> dofcoeffs(ndof); 
    ARRAY<double> elcoeffs(nel);
    ARRAY<int> elnums;
    ARRAY<int> dofs, lagdofs;
    for ( int i = 0; i < nel; i++ )
      {
	compoundspace[0]->GetDofNrs(i, dofs);
	elcoeffs[i]  = GetMeanValueOnElement(i, feticoeff, this->ma, lh );
	for ( int ii = 0; ii < dofs.Size(); ii++ )
	  dofcoeffs[dofs[ii]] = elcoeffs[i];
      }
//     for ( int i = 0; i < nsel; i++ )
//       {
// 	primalspace. GetDualSDofNrs(i, dofs);
// 	for ( int ii = 0; ii < dofs.Size(); ii++ )
// 	  dofcoeffs[dofs[ii]] = 1.;
//       }

    for ( int vert = 0; vert < nv; vert++ )
      {
	lagrangespace.GetVertexDofNrs (vert, lagdofs);
	if ( lagdofs.Size() == 0 ) continue;

	this->ma.GetVertexElements(vert, elnums);
	double sum_coeffs = 0;
	for ( int i  = 0; i < elnums.Size(); i++ )
	  sum_coeffs += elcoeffs[elnums[i]];

	for ( int j = 0; j < lagdofs.Size(); j++ )
	  {
	    INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);
	    bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[0] );
	    if ( dualdofs[1] >= 0 )
	    bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[1] );
	  }
	for ( int j = 0; j < lagdofs.Size(); j++ )
	  {
	    INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);
	    if ( dualdofs[1]  >= 0 )
	      {
		(*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = dofcoeffs[dualdofs[1]] / sum_coeffs;;
		(*bmat)(lagdofs[j] + base_lag, dualdofs[1] ) = -dofcoeffs[dualdofs[0]] / sum_coeffs;;
	      }
	    else
	      (*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = 1. / sum_coeffs;
	  }

      }

    for ( int edge = 0; edge < ned; edge++ )
      {

	lagrangespace.GetEdgeDofNrs (edge, lagdofs);
	if ( lagdofs.Size() == 0 ) continue;

	this->ma.GetEdgeElements(edge, elnums);
	double sum_coeffs = 0;
	for ( int i  = 0; i < elnums.Size(); i++ )
	  sum_coeffs += elcoeffs[elnums[i]];

	for ( int j = 0; j < lagdofs.Size(); j++ )
	  {
	    INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);
	    bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[0] );
	    if ( dualdofs[1] >= 0 )
	    bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[1] );
	  }

	for ( int j = 0; j < lagdofs.Size(); j++ )
	  {
	    INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);

	    if ( dualdofs[1] >= 0 )
	      {
		(*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = dofcoeffs[dualdofs[1]] / sum_coeffs;;
		(*bmat)(lagdofs[j] + base_lag, dualdofs[1] ) = -dofcoeffs[dualdofs[0]] / sum_coeffs;;
	      }
	    else
	      {
		(*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = 1. / sum_coeffs;
	      }
	  }

      }

    if ( dim == 3 )
      for ( int face = 0; face < nfa; face++ )
	{
	  lagrangespace.GetFaceDofNrs (face, lagdofs);
	  if ( lagdofs.Size() == 0 ) continue;
	  
	  this->ma.GetFaceElements(face, elnums);
	  double sum_coeffs = 0;
	  for ( int i  = 0; i < elnums.Size(); i++ )
	    sum_coeffs += elcoeffs[elnums[i]];
	  
	  for ( int j = 0; j < lagdofs.Size(); j++ )
	    {
	      INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);

	      bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[0] );
	      if ( dualdofs[1] >= 0 )
	      bmat->CreatePosition ( lagdofs[j]+base_lag, dualdofs[1] );
	    }
	  for ( int j = 0; j < lagdofs.Size(); j++ )
	    {
	      INT<2> dualdofs = lagrangespace.DualDofs(lagdofs[j]);
	      if ( dualdofs[1] >= 0 )
		{
		  (*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = dofcoeffs[dualdofs[1]] / sum_coeffs;;
		  (*bmat)(lagdofs[j] + base_lag, dualdofs[1] ) = -dofcoeffs[dualdofs[0]] / sum_coeffs;;
		}
	      else
		{
		  (*bmat)(lagdofs[j] + base_lag, dualdofs[0] ) = 1./sum_coeffs;
		}
	    }
	  
	  
	}
    
    lh.CleanUp(heapp);

    return bmat;
  }


  template class FETIDP_BilinearForm<Complex>;
  template class FETIDP_BilinearForm<double>;


  template <class SCAL>
  void FETIDP_Preconditioner<SCAL> :: Update ()
  {
    const FETIDP_Matrix<SCAL> & fetidpmat = 
      dynamic_cast<const FETIDP_Matrix<SCAL> & > (bfa->GetMatrix() );
    precmat = new FETIDP_PrecMatrix<SCAL> ( fetidpmat );
    ;
  }

  template class FETIDP_Preconditioner<double>;
  template class FETIDP_Preconditioner<Complex>;

  // register FESpaces
  namespace
#ifdef MACOS
  fetidp_bilinearform_cpp
#endif
  {
    class Init
    { 
    public:  
      Init ();
    };        
    
    Init::Init()
    {
      //        GetIntegrators().AddBFIntegrator ( "fetidp_b_h1", 2, 1,
      // 					FETIDP_BIntegrator<2>::Create);
      //       GetIntegrators().AddBFIntegrator ("fetidp_b_h1", 3, 1,
      // 					FETIDP_BIntegrator<3>::Create);
      GetPreconditionerClasses().AddPreconditioner ("fetidp_prec", FETIDP_Preconditioner<double>::Create);
      GetPreconditionerClasses().AddPreconditioner ("fetidp_prec_complex", FETIDP_Preconditioner<Complex>::Create);
    }
#ifdef MACOS
    static
#endif
    Init init;
  }

}
