#include <fem.hpp>
#include <comp.hpp>
#include <feti.hpp>


namespace ngcomp
{
  using namespace ngcomp;
  using namespace ngfem;

  FETIDP_PrimalSpace :: FETIDP_PrimalSpace ( const Flags & flags )
  {
    npedgedofs = int( flags.GetNumFlag("primal_edge_dofs", 0) );
    npfacedofs = int( flags.GetNumFlag("primal_face_dofs", 0) );
    npvertdofs = int (flags.GetNumFlag("primal_vertex_dofs", 1) );
    ;
  }

  FETIDP_PrimalSpace :: ~FETIDP_PrimalSpace ()
  {
    delete dualdofs;
    delete primaldofs;
  }


  FETIDP_H1HOSpace :: 
  FETIDP_H1HOSpace (const MeshAccess & ama, 
		    const Flags & flags, bool parseflags)
    : H1HighOrderFESpace ( ama, flags, parseflags ), FETIDP_PrimalSpace (flags),
      firstsdualvertexdof(0),
      firstsdualfacedof(0),
      firstsdualedgedof(0)
  {
    first_vertex_dof.SetSize(ma.GetNV());
    low_order_space = 0;
    refinep = flags.GetDefineFlag("refinep")    ;
  }


  void FETIDP_H1HOSpace :: Update()
  {

    int maxorder = -1; 
    int minorder = 99; 

    if (low_order_space)
      low_order_space -> Update();

    const int dim = ma.GetDimension();
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = (dim == 2) ? 0 : ma.GetNFaces();
    nel = ma.GetNE();
    
    if ( refinep ) order++;

    order_edge.SetSize (ned);
    order_face.SetSize (nfa);
    order_inner.SetSize (nel);
    order_avertex.SetSize (nv); 
    fine_edge.SetSize(ned); 
    fine_face.SetSize(nfa); 

    fine_edge = 0;
    fine_face = 0; 
    int p = var_order ?  1 : order; 
    
    order_edge = p; 
    order_face = INT<2>(p,p);
    order_inner = INT<3>(p,p,p); 
    order_avertex = p; 
	
    ARRAY<int> eledges, elfaces, vnums;
    
    for (int i = 0; i < nel; i++)
      {	
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const FACE * faces = ElementTopology::GetFaces (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	ma.GetElVertices (i, vnums);
	ma.GetElEdges (i, eledges);		
	

	if(dim==3) 
	  {
	    ma.GetElFaces(i, elfaces);
	    for (int j=0;j<elfaces.Size();j++) fine_face[elfaces[j]] = 1; 
	  }
	for (int j=0;j<eledges.Size();j++) fine_edge[eledges[j]] = 1; 
	
	if(!var_order) continue;  
	
	INT<3> el_orders = ma.GetElOrders(i); 
	for(int l=0;l<3;l++) el_orders[l] += rel_order; 

	for(int l=0;l<3;l++) 
	  maxorder = int(max2(el_orders[l],maxorder)); 
	for(int l=0;l<3;l++) 
	  minorder = int(min2(el_orders[l],minorder)); 
	

	for(int j=0;j<dim;j++)
	  order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
	for(int j=0;j<eledges.Size();j++)
	  {
	    for(int k=0;k<dim;k++)
	      if(points[edges[j][0]][k] != points[edges[j][1]][k])
		{ 
		  order_edge[eledges[j]] = max2(order_edge[eledges[j]],el_orders[k]);
		  k=dim; 
		}
	  }
	
	if(dim==3)
	  {
	    for(int j=0;j<elfaces.Size();j++)
	      {
		// trig_face
		if(faces[j][3]==-1) 
		  {
		    order_face[elfaces[j]][0] = int(max2(order_face[elfaces[j]][0], 
							 el_orders[0]));
		    order_face[elfaces[j]][1] = order_face[elfaces[j]][0]; 
		  }
		else //quad_face
		  {
		    int fmax = 0;
		    for(int k = 1; k < 4; k++) 
		      if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		    
		    INT<2> f((fmax+3)%4,(fmax+1)%4); 
		    if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		    
		    
		    for(int l=0;l<2;l++)
		      for(int k=0;k<3;k++)
			if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			  {
			    order_face[elfaces[j]][l] = int(max2(order_face[elfaces[j]][l], 
								 el_orders[k]));
			    break;
			  }
		  }
	      }
	  }
      }
	  
    if(uniform_order_inner > -1) 
      order_inner = INT<3>(uniform_order_inner,uniform_order_inner,uniform_order_inner);
    if(uniform_order_face > -1 && dim==3) 
      order_face = INT<2>(uniform_order_face,uniform_order_face); 
    if(uniform_order_edge > -1) 
      order_edge = uniform_order_edge; 
    

    for(int i=0;i<ned;i++) if(!fine_edge[i]) {order_edge[i] = 1;}
    if(dim == 3) for(int i=0;i<nfa;i++) if(!fine_face[i]) order_face[i] = INT<2> (1,1); 
    if(dim==2) 
      { 
	nfa = 0; 
	fine_face.SetSize(0); 
	order_face.SetSize(0);
      } 
    
    UpdateDofTables ();

#ifdef PARALLEL
      UpdateParallelDofs();
#endif

  }

  ///
  void FETIDP_H1HOSpace :: UpdateDofTables()
  {
    int nel = ma.GetNE();
    int nv = ma.GetNV();
    int ned = ma.GetNEdges();
    int nfa = (ma.GetDimension() == 2 ) ? 0 : ma.GetNFaces();
    int dim = ma.GetDimension();

    ndof = 0;

    int ndof_counter = ndof;
    first_edge_dof.SetSize (ned+1);
    first_vertex_dof.SetSize(nv+1);

    for (int i = 0; i < nv; i++)
      {
	first_vertex_dof[i] = ndof;
	  ndof += npvertdofs;
      }
    first_vertex_dof[nv] = ndof;


    for (int i = 0; i < ned; i++)
      {
	first_edge_dof[i] = ndof;
	ndof += npedgedofs;
      }
    first_edge_dof[ned] = ndof;
         
    first_face_dof.SetSize (nfa+1);
    for (int i = 0; i < nfa; i++)
      {
	first_face_dof[i] = ndof;
	ndof += npfacedofs;
      }

    first_face_dof[nfa] = ndof;
 
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	INT<3> p = order_inner[i];	
	ARRAY<int> elfaces, eledges, elverts;

	ma.GetElPNums(i,elverts); 
	ma.GetElEdges(i,eledges);
	if ( dim == 3 ) 
	  ma.GetElFaces(i,elfaces);
	else
	  elfaces.SetSize(0);

	for ( int i_vert = 0; i_vert < elverts.Size(); i_vert++ )
	  {
	    ndof += GetNDualVertexDofs(elverts[i_vert]); // ndualvertdofs;
	  }
	for ( int i_ed = 0; i_ed < eledges.Size(); i_ed++ )
	  {
	    ndof += GetNDualEdgeDofs(eledges[i_ed]);
	  }

	for ( int i_fa = 0; i_fa < elfaces.Size(); i_fa++ )
	  {
	    ndof += GetNDualFaceDofs(elfaces[i_fa]);
	  }

	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)/2;
	    break;
	  case ET_QUAD:
	    if(p[0]>1 && p[1]>1)
	      ndof += (p[0]-1)*(p[1]-1);
	    break;
	  case ET_TET:
	    if(p[0] > 3)
	      ndof += (p[0]-1)*(p[0]-2)*(p[0]-3)/6;
	    break;
	  case ET_PRISM:
	    if(p[0]>2 && p[2]>1)
	      ndof += (p[0]-1)*(p[0]-2)*(p[2]-1)/2;
	    break;
	  case ET_PYRAMID:
	    if(p[0]>2)
	      ndof += (p[0]-1)*(p[0]-2)*(2*p[0]-3)/6;
	    break;
	  case ET_HEX:
	    if(p[0]>1 && p[1] > 1 && p[2]>1) 
	      ndof += (p[0]-1)*(p[1]-1)*(p[2]-1);
	    break;
	  }
      } 
    first_element_dof[nel] = ndof;

    primaldofs = new BitArray(ndof);
    dualdofs = new BitArray(ndof);

    primaldofs->Clear();
    dualdofs->Clear();

    int first_primal_dof = first_vertex_dof[0];
    int last_primal_dof = (ma.GetDimension() == 2 ) ? first_edge_dof[ned] : first_face_dof[nfa];
    for ( int i = first_primal_dof; i < last_primal_dof; i++ )
      primaldofs -> Set(i);

    for (int i = 0; i < nel; i++)
      {
	INT<3> p = order_inner[i];	
	ARRAY<int> elfaces, eledges, elverts;

	ma.GetElPNums(i,elverts); 
	ma.GetElEdges(i,eledges);
	if ( dim == 3 ) 
	  ma.GetElFaces(i,elfaces);
	else
	  elfaces.SetSize(0);

	int ii = first_element_dof[i]; 
	for ( int i_vert = 0; i_vert < elverts.Size(); i_vert++ )
	  {
	    for ( int j = 0; j < GetNDualVertexDofs(elverts[i_vert]); j++ )
	      dualdofs->Set(ii++);
	  }

	for ( int i_ed = 0; i_ed < eledges.Size(); i_ed++ )
	  {
	    for ( int j = 0; j < GetNDualEdgeDofs(eledges[i_ed]); j++ )
	      dualdofs->Set(ii++);
	  }

	for ( int i_fa = 0; i_fa < elfaces.Size(); i_fa++ )
	  {
	    for ( int j = 0; j < GetNDualFaceDofs(elfaces[i_fa]); j++ )
	      dualdofs -> Set(ii++);
	  }

      }



    // *********************************
    //
    // Update surface-element tables
    //
    // *********************************

    int nsel = ma.GetNSE();
    ARRAY<int> cnt_vert(nsel), cnt_edge(nsel);
    ARRAY<int> verts, edges;
    int face;
    ARRAY<int> elements;
	cnt_vert = 0;
	cnt_edge = 0;
    
    for ( int sel = 0; sel < ma.GetNSE(); sel++ )
      {
	ma.GetSElPNums(sel, verts);
	ma.GetSElEdges(sel, edges);
	for ( int i = 0; i < verts.Size(); i++ )
	  cnt_vert[sel]++; //  += GetNDualVertexDofs(verts[i]);
	for ( int i = 0; i < edges.Size(); i++ )
	  cnt_edge[sel]++; //  += GetNDualEdgeDofs(edges[i]);
      }

    if ( firstsdualvertexdof ) delete firstsdualvertexdof;
    if ( firstsdualedgedof ) delete firstsdualedgedof;
    if ( firstsdualfacedof ) delete firstsdualfacedof;

    firstsdualvertexdof = new Table<int> (cnt_vert);
    firstsdualedgedof = new Table<int> (cnt_edge);
    if ( ma.GetDimension() == 3 )
      firstsdualfacedof = new ARRAY<int> (nsel);

    int changedverts = 0;
    for ( int sel = 0; sel < nsel; sel++ )
      {
// 	*testout << "surface element " << sel << endl;
	ma.GetSElPNums(sel, verts);
	ma.GetSElEdges(sel, edges);
	if ( ma.GetDimension() == 2 )
	  {
	    ma.GetEdgeElements(edges[0], elements);
// 	    if ( elements.Size() > 1 ) continue;
// 	    if ( elements.Size() > 1 )
// 	      {
// 		(*firstsdualvertexdof)[sel] = -1;
// 		(*firstsdualedgedof)[sel] = -1;
// 		continue;
// 	      }

	    ARRAY<int> dofs, elverts, eledges;
	    int firstinnerdof = first_element_dof[elements[0]];
	    ma.GetElPNums(elements[0],elverts);
// 	    *testout << "element " << elements[0] << ": vnums " << elverts << endl;
	    for ( int i = 0; i < verts.Size(); i++ )
	      {
		if ( GetNDualVertexDofs(verts[i]) )
		  {
		    int locvert = elverts.Pos(verts[i]);
		    (*firstsdualvertexdof)[sel][i] = first_element_dof[elements[0]] + locvert ;
		    primaldofs->Set( first_element_dof[elements[0]] + locvert );
		    dualdofs->Clear( first_element_dof[elements[0]] + locvert );
		  }
		else
		    (*firstsdualvertexdof)[sel][i] = first_element_dof[elements[0]];
	      }
	    for ( int i = 0; i < elverts.Size(); i++ )
	      {
		firstinnerdof += GetNDualVertexDofs(elverts[i]);
	      }
// 	    *testout << "first dual vertex dof " << (*firstsdualvertexdof)[sel] << endl;

	    ma.GetElEdges(elements[0], eledges);
	    int locedge = eledges.Pos(edges[0]);
	    for ( int i = 0; i < locedge; i++ )
	      firstinnerdof += GetNDualEdgeDofs(eledges[i]);
	    (*firstsdualedgedof)[sel][0] = firstinnerdof;
	    for ( int i = 0; i < GetNDualEdgeDofs(eledges[locedge]); i++ )
	      {
		dualdofs->Clear(firstinnerdof);
		primaldofs->Set(firstinnerdof++);
	      }

// 	    *testout << "first dual edge dof " << (*firstsdualedgedof)[sel] << endl;
	  }
	else // dim == 3
	  {
	    face = ma.GetSElFace(sel);

	    ma.GetFaceElements(face, elements);
// 	    if ( elements.Size() > 1 ) continue;
// 	    if ( elements.Size() > 1 )
// 	      {
// 		(*firstsdualvertexdof)[sel] = -1;
// 		(*firstsdualedgedof)[sel] = -1;
// 		(*firstsdualfacedof)[sel] = -1;
// 		continue;
// 	      }

	    ARRAY<int> dofs, elverts, eledges, elfaces;
	    int firstinnerdof = first_element_dof[elements[0]];
	    ma.GetElPNums(elements[0],elverts);
	    for ( int i = 0; i < verts.Size(); i++ )
	      {
		if ( GetNDualVertexDofs(verts[i]) )
		  {
		    int locvert = elverts.Pos(verts[i]);
		    (*firstsdualvertexdof)[sel][i] = first_element_dof[elements[0]] + locvert ;
		    primaldofs->Set( first_element_dof[elements[0]] + locvert );
		    dualdofs->Clear( first_element_dof[elements[0]] + locvert );
		  }
		else
		    (*firstsdualvertexdof)[sel][i] = first_element_dof[elements[0]];

	      }
	    for ( int el = 0; el < elverts.Size(); el++ )
	      firstinnerdof += GetNDualVertexDofs(elverts[el]);

	    ma.GetElEdges(elements[0], eledges);
	    ARRAY<int> firstedge(eledges.Size());
	    for ( int i = 0; i < eledges.Size(); i++ )
	      {
		firstedge[i] = firstinnerdof;
		firstinnerdof += GetNDualEdgeDofs(eledges[i]);
	      }

	    for ( int i = 0; i < edges.Size(); i++ )
	      {
		ma.GetElEdges(elements[0], eledges);
		int locedge = eledges.Pos(edges[i]);
		(*firstsdualedgedof)[sel][i] = firstedge[locedge];
		for ( int j = 0; j < GetNDualEdgeDofs(edges[i]); j++ )
		  {
		    primaldofs->Set(firstedge[locedge]+j);
		    dualdofs->Clear(firstedge[locedge]+j);
		  }
	      }

	    ma.GetElFaces(elements[0], elfaces);
	    int locface = elfaces.Pos(face);
	    for ( int i = 0; i < locface; i++ )
	      firstinnerdof += GetNDualFaceDofs(elfaces[i]);
	    (*firstsdualfacedof)[sel] = firstinnerdof;
	    for ( int j = 0; j < GetNDualFaceDofs(elfaces[locface]); j++ )
	      {
		primaldofs->Set(firstinnerdof);
		dualdofs->Clear(firstinnerdof++);
	      }
	  }


      }

  }

  /// 
  const FiniteElement & FETIDP_H1HOSpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    H1HighOrderFiniteElement * hofe;
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
    
    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	{ 
	  hofe = new (lh.Alloc (sizeof(H1HighOrderTet<T_ORTHOPOL>)))  
	    H1HighOrderTet<T_ORTHOPOL> (order);
	  break;
	}
      case ET_PYRAMID:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderPyramid<T_ORTHOPOL>)))  
	    H1HighOrderPyramid<T_ORTHOPOL> (order);
	  break;
	}
      case ET_PRISM:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderPrism<T_ORTHOPOL>)))  
	    H1HighOrderPrism<T_ORTHOPOL> (order);
	  break;
	}
      case ET_HEX:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderHex<T_ORTHOPOL>)))  
	    H1HighOrderHex<T_ORTHOPOL> (order);
	  break;
	}
      case ET_TRIG:
	{ 
	  hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  
	    H1HighOrderTrig<T_ORTHOPOL> (order);
	  break;
	}
      case ET_QUAD:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  
	    H1HighOrderQuad<T_ORTHOPOL> (order);
	  break;
	}
      default:
	hofe = 0; 
      }
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ArrayMem<int, 12> ednums, order_ed;
    ArrayMem<int, 6> fanums;
    ArrayMem<INT<2>, 6> order_fa;


    ma.GetElVertices(elnr, vnums);
    ma.GetElEdges(elnr, ednums);
    
    order_ed.SetSize (ednums.Size());
    
    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order_edge[ednums[j]];
    
    
    hofe -> SetOrderEdge (order_ed);
    
    if (ma.GetDimension() == 3)
      {
	ma.GetElFaces(elnr, fanums);
	order_fa.SetSize (fanums.Size());
	for (int j = 0; j < fanums.Size(); j++)
	  order_fa[j] = order_face[fanums[j]];
	
	hofe -> SetOrderFace (order_fa);
      }
	
    ArrayMem<int, 8> order_vert(vnums.Size());
    order_vert = 1;

    hofe -> SetOrderVertex (order_vert);
    hofe -> SetAugmented (0);
    hofe -> SetOrderInner (order_inner[elnr]);

    hofe -> ComputeNDof();
    hofe -> SetVertexNumbers (vnums, lh);

    return *hofe;
    ;
  }
  ///
  const FiniteElement & FETIDP_H1HOSpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    H1HighOrderFiniteElement * hofe;
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
    
    switch (ma.GetSElType(elnr))
      {
      case ET_TRIG:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (1);
	  break;
	}
      case ET_QUAD:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderQuad<T_ORTHOPOL>)))  H1HighOrderQuad<T_ORTHOPOL> (1);
	  break;
	}
      case ET_SEGM:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderSegm<T_ORTHOPOL>)))  H1HighOrderSegm<T_ORTHOPOL> (1);
	  break;
	}
      default:
	hofe = 0; 
      }
    
    hofe -> SetAugmented (0);
    
    ArrayMem<int, 8> vnums, order_vert(8);
    ArrayMem<int, 4> ednums, order_ed;
    
    ma.GetSElVertices(elnr, vnums);
    ma.GetSElEdges(elnr, ednums);
    
    order_ed.SetSize (ednums.Size());
    
    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order_edge[ednums[j]];

    for (int j = 0; j < 8; j++)
      order_vert[j] = order;
    hofe -> SetOrderVertex (order_vert);
    

    

    if (ma.GetDimension() == 3)
      {
	hofe -> SetOrderEdge (order_ed);
	INT<2> p = order_face[ma.GetSElFace(elnr)];
	hofe -> SetOrderInner (INT<3> (p[0], p[1], 0));
      }
    else
      {
	hofe -> SetOrderInner (order_ed[0]);
      }
 
    hofe -> ComputeNDof(); 
    hofe -> SetVertexNumbers (vnums, lh);  
    return *hofe;
    ;
  }
  ///
  void FETIDP_H1HOSpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0);
    dnums.SetSize(0); 

    int first_inner_dof = first_element_dof[elnr];

    if ( npvertdofs )
    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);
    else
      for ( i = 0; i < vnums.Size(); i++ )
	dnums.Append(first_inner_dof++);

    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1];
	for (j = first; j < next; j++)
	  dnums.Append (j);

	for ( int j = 0; j < GetNDualEdgeDofs(ednums[i]) ; j++ )
	  dnums.Append( first_inner_dof++ );
      }

    for (i = 0; i < fanums.Size(); i++)
      {
	first = first_face_dof[fanums[i]];
	next = first_face_dof[fanums[i]+1];
	  
	for (j = first; j < next; j++)
	  dnums.Append (j);

	for ( int j = 0; j < GetNDualFaceDofs(fanums[i]) ; j++ )
	  dnums.Append( first_inner_dof++ );
      }


    first = first_inner_dof; // first_element_dof[elnr];
    next = first_element_dof[elnr+1];
 
    
    for (j = first; j < next; j++)
      dnums.Append (j);
     

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
    ;
  }


  void FETIDP_H1HOSpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> vnums, ednums;
    int fanum;
    int i, j;

    ma.GetSElVertices (elnr, vnums);
    ma.GetSElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      fanum = ma.GetSElFace (elnr);
    
    dnums.SetSize(0);

    if ( npvertdofs )
    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);
    else
      for ( i = 0; i < vnums.Size(); i++ )
// 	if ( (*firstsdualvertexdof)[elnr][i] >= 0 )
	  dnums.Append((*firstsdualvertexdof)[elnr][i]);

    for (i = 0; i < ednums.Size(); i++)
      {
	int first = first_edge_dof[ednums[i]];
	int neddofs = first_edge_dof[ednums[i]+1] - first;
	for (j = 0; j < neddofs; j++)
	  dnums.Append (first+j);
	first = (*firstsdualedgedof)[elnr][i];
// 	if ( first < 0 ) continue;
	neddofs = GetNDualEdgeDofs(ednums[i]);
	for (j = 0; j < neddofs; j++)
	  dnums.Append (first+j);
      }

    if (ma.GetDimension() == 3)
      {
	int first = first_face_dof[fanum];
	int nfadofs = first_face_dof[fanum+1] - first;
	for (j = 0; j < nfadofs; j++)
	  //	  dnums.Append (-1);
	  dnums.Append (first+j);
	first = (*firstsdualfacedof)[elnr];
// 	if ( first >= 0 )
// 	  {
	    nfadofs = GetNDualFaceDofs(fanum);
	    for (j = 0; j < nfadofs; j++)
	      dnums.Append (first+j);
// 	  }
      }

//     (*testout) << "juhu, element (" << elnr <<"): my dnums are" << endl << dnums << endl;
    if (!DefinedOnBoundary (ma.GetSElIndex (elnr)) )
      dnums = -1;

  }



  void FETIDP_H1HOSpace :: GetPrimalDualDofNrs ( int elnr, ARRAY<int> & dnums ) const
  {
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0);
    dnums.SetSize(0); 

    int first_inner_dof = first_element_dof[elnr];

    if ( npvertdofs )
    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);
    else
      for ( i = 0; i < vnums.Size(); i++ )
	dnums.Append(first_inner_dof++);

    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1];
	for (j = first; j < next; j++)
	  dnums.Append (j);

	for ( int j = 0; j < GetNDualEdgeDofs(ednums[i]); j++ )
	  dnums.Append( first_inner_dof++ );
      }

    //(*testout) << "after edges dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < fanums.Size(); i++)
      {
	first = first_face_dof[fanums[i]];
	next = first_face_dof[fanums[i]+1];
	  
	for (j = first; j < next; j++)
	  dnums.Append (j);

	for ( int j = 0; j < GetNDualFaceDofs(fanums[i]); j++ )
	  dnums.Append( first_inner_dof++ );
      }

  }

  void FETIDP_H1HOSpace :: GetDualDofNrs ( int elnr, ARRAY<int> & dnums ) const 
  {
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 

    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    if (ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else 
      fanums.SetSize(0);
    dnums.SetSize(0); 

    int dof = first_element_dof[elnr];

    if ( !npvertdofs )
      for ( i = 0; i < vnums.Size(); i++ )
	dnums.Append(dof++);

    for (i = 0; i < ednums.Size(); i++)
      {
	for ( int j = 0; j < GetNDualEdgeDofs(ednums[i]); j++ )
	  dnums.Append (dof++);
      }

    //(*testout) << "after edges dnums.Size() " << dnums.Size() << endl;

    for (i = 0; i < fanums.Size(); i++)
      {
	for (j = 0; j < GetNDualFaceDofs(fanums[i]); j++)
	  dnums.Append (dof++);
      }

  }

  void FETIDP_H1HOSpace :: GetDualVertexDofNrs ( int elnr, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    if ( npvertdofs ) return;

    int dof = first_element_dof[elnr];
    int nvert = ElementTopology :: GetNVertices ( ma.GetElType(elnr) );
    for ( int i = 0; i < nvert; i++ )
      dnums.Append (dof++);
  }

  void FETIDP_H1HOSpace :: GetDualVertexDofNrs ( int elnr, int locvnum, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    if ( npvertdofs ) return;

    int dof = first_element_dof[elnr];
    dnums.Append (dof+locvnum);
  }

  void FETIDP_H1HOSpace :: GetDualEdgeDofNrs ( int elnr, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    ARRAY<int> ednums;
    ma.GetElEdges (elnr, ednums);

    int dof = first_element_dof[elnr];

    if ( !npvertdofs )
      dof += ElementTopology :: GetNVertices ( ma.GetElType(elnr) );

    for (int i = 0; i < ednums.Size(); i++)
      {
	for (int j = 0; j < GetNDualEdgeDofs(ednums[i]); j++ )
	  dnums.Append (dof++);
      }
  }

  void FETIDP_H1HOSpace :: GetDualEdgeDofNrs ( int elnr, int locednum, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    ARRAY<int> ednums;
    ma.GetElEdges (elnr, ednums);

    int dof = first_element_dof[elnr];
    int p;

    if ( !npvertdofs )
      dof += ElementTopology :: GetNVertices ( ma.GetElType(elnr) );

    for (int i = 0; i < locednum; i++)
      {
	dof += GetNDualEdgeDofs(ednums[i]);
      }

    for (int j = 0; j < GetNDualEdgeDofs(ednums[locednum]); j++ )
      dnums.Append (dof++);

  }

  void FETIDP_H1HOSpace :: GetDualFaceDofNrs ( int elnr, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    if (ma.GetDimension() < 3) return;

    int i, j;
    ARRAY<int> ednums, fanums, fapnums;
    ma.GetElEdges (elnr, ednums);
    ma.GetElFaces(elnr, fanums);

    int dof = first_element_dof[elnr];

    if ( !npvertdofs )
      dof += ElementTopology :: GetNVertices ( ma.GetElType(elnr) );

    for (int i = 0; i < ednums.Size(); i++)
      {
	dof += GetNDualEdgeDofs(ednums[i]);
      }


    for (i = 0; i < fanums.Size(); i++)
      {
	for (j = 0; j < GetNDualFaceDofs(fanums[i]); j++)
	  dnums.Append (dof++);
      }
    
  }

  void FETIDP_H1HOSpace :: GetDualFaceDofNrs ( int elnr, int locfanum, ARRAY<int> & dnums ) const 
  {
    dnums.SetSize(0);
    if (ma.GetDimension() < 3) return;

    int i, j;
    ARRAY<int> ednums, fanums, fapnums;
    ma.GetElEdges (elnr, ednums);
    ma.GetElFaces(elnr, fanums);

    int dof = first_element_dof[elnr];

    if ( !npvertdofs )
      dof += ElementTopology :: GetNVertices ( ma.GetElType(elnr) );

    for (int i = 0; i < ednums.Size(); i++)
      {
	dof += GetNDualEdgeDofs(ednums[i]); // p-1-npedgedofs;
      }

    for (i = 0; i < locfanum; i++)
      {
	dof += GetNDualFaceDofs(fanums[i]);
      }

    for (j = 0; j < GetNDualFaceDofs(fanums[locfanum]); j++)
      dnums.Append (dof++);
  }
  


  FETIDP_LagrangeSpace :: 
  FETIDP_LagrangeSpace (const MeshAccess & ama, 
			const Flags & flags, bool parseflags)
    : FESpace(ma, flags, parseflags)
  {
    int nfa = (ma.GetDimension() == 2 ) ? 0 : ma.GetNFaces();
    first_vertex_dof.SetSize(ma.GetNV()+1);
    first_edge_dof.SetSize(ma.GetNEdges()+1);
    first_face_dof.SetSize(nfa+1);
  }

  Table<int> *  FETIDP_LagrangeSpace :: CreateEdgeElementsTable () const 
  {
    int ned = ma.GetNEdges();
    int nel = ma.GetNE();
    ARRAY<int> eledges;
    ARRAY<int> cnt(ned);
    cnt = 0;
    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElEdges(el, eledges);
	for ( int i = 0; i < eledges.Size(); i++ )
	  cnt[eledges[i]]++;
      }

    Table<int> * table = new Table<int> (cnt);
    cnt = 0;

    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElEdges(el, eledges);
	for ( int i = 0; i < eledges.Size(); i++ )
	  (*table)[eledges[i]][cnt[eledges[i]]++] = el;
      }
    return table;
  }


  Table<int> *  FETIDP_LagrangeSpace :: CreateFaceElementsTable () const 
  {
    int nfa = ma.GetNFaces();
    int nel = ma.GetNE();
    ARRAY<int> elfaces;
    ARRAY<int> cnt(nfa);
    cnt = 0;
    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElFaces(el, elfaces);
	for ( int i = 0; i < elfaces.Size(); i++ )
	  cnt[elfaces[i]]++;
      }

    Table<int> * table = new Table<int> (cnt);
    cnt = 0;

    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElFaces(el, elfaces);
	for ( int i = 0; i < elfaces.Size(); i++ )
	  (*table)[elfaces[i]][cnt[elfaces[i]]++] = el;
      }
    return table;
  }

  Table<int> *  FETIDP_LagrangeSpace :: CreateVertexElementsTable () const 
  {
    int nv = ma.GetNV();
    int nel = ma.GetNE();
    ARRAY<int> elverts;
    ARRAY<int> cnt(nv);
    cnt = 0;
    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElPNums(el, elverts);
	for ( int i = 0; i < elverts.Size(); i++ )
	  cnt[elverts[i]]++;
      }

    Table<int> * table = new Table<int> (cnt);
    cnt = 0;

    for ( int el = 0; el < nel; el++ )
      {
	ma.GetElPNums(el, elverts);
	for ( int i = 0; i < elverts.Size(); i++ )
	  (*table)[elverts[i]][cnt[elverts[i]]++] = el;
      }
    return table;
  }


  void  FETIDP_LagrangeSpace :: Update( const FETIDP_PrimalSpace & primalspace)
  {
    FETIDP_PrimalSpace & constprimalspace = const_cast<FETIDP_PrimalSpace &> (primalspace);
    ndof = 0;
    ARRAY<int> elnums;
    int nfa = (ma.GetDimension() == 2 ) ? 0 : ma.GetNFaces();
    first_vertex_dof.SetSize(ma.GetNV()+1);
    first_edge_dof.SetSize(ma.GetNEdges()+1);
    first_face_dof.SetSize(nfa+1);
    dualdofs.SetSize(0);

    Table<int> & vertelements = *CreateVertexElementsTable();
    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	first_vertex_dof[vert] = ndof;
	int nvertels = vertelements[vert].Size();
	if ( !primalspace.GetNDualVertexDofs(vert) ) continue;
	ndof += primalspace.GetNDualVertexDofs(vert) * (nvertels * (nvertels-1)) / 2 ;


	for ( int i = 0; i < nvertels; i++ )
	  for ( int j = i+1; j < nvertels; j++ )
	    {
	      ARRAY<int> dnumsi, dnumsj;
	      ARRAY<int> vnumsi, vnumsj;

	      ma.GetElPNums ( vertelements[vert][i], vnumsi );
	      int locvnumi = vnumsi.Pos(vert);
	      ma.GetElPNums ( vertelements[vert][j], vnumsj );
	      int locvnumj = vnumsj.Pos(vert);
	      primalspace.GetDualVertexDofNrs(vertelements[vert][i], dnumsi);
	      primalspace.GetDualVertexDofNrs(vertelements[vert][j], dnumsj);
	      dualdofs.Append(INT<2> ( dnumsi[locvnumi], dnumsj[locvnumj]) );
		  constprimalspace.SetDualDof(dnumsi[locvnumi], 1);
		  constprimalspace.SetDualDof(dnumsj[locvnumj], 1);
	    }
      }
    delete &vertelements;

    first_vertex_dof[ma.GetNV()] = ndof;

    Table<int> & edgeelements = *CreateEdgeElementsTable();
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	first_edge_dof[edge] = ndof;
	int nedgeels = edgeelements[edge].Size();

	// edges in coarse mesh
	if ( !nedgeels ) continue;
	if ( ! primalspace.GetNDualEdgeDofs(edge) ) continue;
	ndof += (primalspace.GetNDualEdgeDofs(edge)) * (nedgeels * (nedgeels-1)) / 2;

	for ( int i = 0; i < nedgeels; i++ )
	  for ( int j = i+1; j < nedgeels; j++ )
	    {
	      ARRAY<int> dnumsi, dnumsj;
	      ARRAY<int> ednumsi, ednumsj;

	      ma.GetElEdges ( edgeelements[edge][i], ednumsi );
	      int locednumi = ednumsi.Pos(edge);
	      ma.GetElEdges ( edgeelements[edge][j], ednumsj );
	      int locednumj = ednumsj.Pos(edge);
	      primalspace.GetDualEdgeDofNrs(edgeelements[edge][i], locednumi, dnumsi);
	      primalspace.GetDualEdgeDofNrs(edgeelements[edge][j], locednumj, dnumsj);
	      for ( int k = 0; k < dnumsi.Size(); k++ )
		{
		  dualdofs.Append(INT<2> ( dnumsi[k], dnumsj[k] ) );
		  constprimalspace.SetDualDof(dnumsi[k], 1);
		  constprimalspace.SetDualDof(dnumsj[k], 1);
		}
	    }


	// hack for surface elements in 2d
	if ( ma.GetDimension() == 2 && nedgeels == 1 )
	  {
// 	    ndof += primalspace.GetNDualEdgeDofs(edge);
	    ARRAY<int> dnumsi, ednumsi;
	    ma.GetElEdges ( edgeelements[edge][0], ednumsi );
	    int locednumi = ednumsi.Pos(edge);
	    primalspace.GetDualEdgeDofNrs(edgeelements[edge][0], locednumi, dnumsi);
	    for ( int k = 0; k < dnumsi.Size(); k++ )
	      constprimalspace.SetDualDof(dnumsi[k], false );
 	  }

      }
    first_edge_dof[ma.GetNEdges()] = ndof;
    delete &edgeelements;

    if ( ma.GetDimension() == 3 )
      {
	Table<int> & faceelements = *CreateFaceElementsTable();
	for ( int face = 0; face < nfa; face++ )
	  {
	    first_face_dof[face] = ndof;

// 	    // faces in coarse mesh..
	    if ( faceelements[face].Size() == 0 ) continue;
// 	    if ( faceelements[face].Size() <= 1 ) continue;

	    if ( faceelements[face].Size() > 1 ) 
	      ndof += primalspace.GetNDualFaceDofs(face);
	    
	    ARRAY<int> dnumsi, dnumsj;
	    ARRAY<int> fanumsi, fanumsj;
	    
	    ma.GetElFaces ( faceelements[face][0], fanumsi );
	    int locfanumi = fanumsi.Pos(face);
	    primalspace.GetDualFaceDofNrs(faceelements[face][0], locfanumi, dnumsi);
	    if ( faceelements[face].Size() == 2 )
	      {
		ma.GetElFaces ( faceelements[face][1], fanumsj );
		int locfanumj = fanumsj.Pos(face);
		primalspace.GetDualFaceDofNrs(faceelements[face][1], locfanumj, dnumsj);
		for ( int k = 0; k < dnumsi.Size(); k++ )
		  {
		    dualdofs.Append(INT<2> ( dnumsi[k], dnumsj[k] ) );
		    constprimalspace.SetDualDof(dnumsi[k], true);
		    constprimalspace.SetDualDof(dnumsj[k], true);
		  }
	      }
	    else
	      {
		for ( int k = 0; k < dnumsi.Size(); k++ )
		constprimalspace.SetDualDof(dnumsi[k], false);
	      }

	  }
	first_face_dof[nfa] = ndof;
	delete & faceelements;
      }

  }
  
  ///
  const FiniteElement &  FETIDP_LagrangeSpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    ARRAY<int> fanums, ednums, vnums, elnums;
    int dim = ma.GetDimension();
    ma.GetElPNums(elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    ma.GetElFaces(elnr, fanums);

    int ndof_el = 0;
    for ( int i = 0; i < vnums.Size(); i++ )
      ndof_el += first_vertex_dof[vnums[i]+1] - first_vertex_dof[vnums[i]];
    for ( int i = 0; i < ednums.Size(); i++ )
      ndof_el += first_edge_dof[ednums[i]+1] - first_edge_dof[ednums[i]];
    if ( dim == 3 )
      for ( int i = 0; i < fanums.Size(); i++ )
	ndof_el += first_face_dof[fanums[i]+1] - first_face_dof[fanums[i]];

    ELEMENT_TYPE eltype = ma.GetElType ( elnr );

    LagrangeFiniteElement * fel_l = (new LagrangeFiniteElement ( dim, eltype, ndof_el, 0) );

    return *fel_l;

  }
  ///
  const FiniteElement &  FETIDP_LagrangeSpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    ARRAY<int> ednums, vnums;
    int dim = ma.GetDimension();
    ma.GetSElPNums(elnr, vnums);
    ma.GetSElEdges (elnr, ednums);
    int fanum = ma.GetSElFace (elnr);
    int ndof_el = 0;
    for ( int i = 0; i < vnums.Size(); i++ )
      ndof_el += first_vertex_dof[vnums[i]+1] - first_vertex_dof[vnums[i]];
    for ( int i = 0; i < ednums.Size(); i++ )
      ndof_el += first_edge_dof[ednums[i]+1] - first_edge_dof[ednums[i]];
    if ( dim == 3 )
      ndof_el += first_face_dof[fanum+1] - first_face_dof[fanum];

    ELEMENT_TYPE eltype = ma.GetSElType ( elnr );

    return *(new LagrangeFiniteElement ( dim, eltype, ndof_el, 0) );
  }
  //
  void  FETIDP_LagrangeSpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> fanums, ednums, vnums;
    int dim = ma.GetDimension();
    ma.GetElPNums(elnr, vnums);
    ma.GetElEdges (elnr, ednums);
    ma.GetElFaces(elnr, fanums);
    dnums.SetSize(0);

    for ( int i = 0; i < vnums.Size(); i++ )
      {
	int first = first_vertex_dof[vnums[i]];
	int next = first_vertex_dof[vnums[i]+1];
	  for ( int j = first; j < next; j++ )
	    dnums.Append(j);
      }
//     *testout << "lagrange element " << elnr << endl;
//     *testout << "vertex dofs " ;
//     for ( int i = 0; i < dnums.Size(); i++ )
//       (*testout) << dnums[i] << " for dofs " << dualdofs[i][0] << ", " << dualdofs[i][1] << endl;
//        int edgestart = dnums.Size();

    for ( int i = 0; i < ednums.Size(); i++ )
      {
	int first = first_edge_dof[ednums[i]];
	int next = first_edge_dof[ednums[i]+1];
	for ( int j = first; j < next; j++ )
	  dnums.Append(j);
      }
//     *testout << "edge dofs " ;
//     for ( int i = edgestart; i < dnums.Size(); i++ )
//       (*testout) << dnums[i] << " for dofs " << dualdofs[i][0] << ", " << dualdofs[i][1] << endl;

    if ( dim == 3 )
    for ( int i = 0; i < fanums.Size(); i++ )
      {
	int first = first_face_dof[fanums[i]];
	int next = first_face_dof[fanums[i]+1];
	for ( int j = first; j < next; j++ )
	  dnums.Append(j);
      }
  }
  ///
  void  FETIDP_LagrangeSpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> ednums, vnums;
    int dim = ma.GetDimension();
    ma.GetSElPNums(elnr, vnums);
    ma.GetSElEdges (elnr, ednums);
    int fanum = ma.GetSElFace ( elnr );

    dnums.SetSize(0);

    for ( int i = 0; i < vnums.Size(); i++ )
      {
	int first = first_vertex_dof[vnums[i]];
	int next = first_vertex_dof[vnums[i]+1];
	  for ( int j = first; j < next; j++ )
	    dnums.Append(j);
      }
    for ( int i = 0; i < ednums.Size(); i++ )
      {
	int first = first_edge_dof[ednums[i]];
	int next = first_edge_dof[ednums[i]+1];
	for ( int j = first; j < next; j++ )
	  dnums.Append(j);
      }
    if ( dim == 3 )
      {
	int first = first_face_dof[fanum];
	int next = first_face_dof[fanum+1];
	for ( int j = first; j < next; j++ )
	  dnums.Append(j);
      }
    ;
  }
  
  /// 
  void  FETIDP_LagrangeSpace :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
  }

  void  FETIDP_LagrangeSpace :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_vertex_dof[vnr];
    int next = first_vertex_dof[vnr+1];
    for ( int j = first; j < next; j++ )
      dnums.Append(j);
  }
  void  FETIDP_LagrangeSpace :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_edge_dof[ednr];
    int next = first_edge_dof[ednr+1];
    for ( int j = first; j < next; j++ )
      dnums.Append(j);
  }
  void  FETIDP_LagrangeSpace :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if ( ma.GetDimension() == 2 ) return;

    int first = first_face_dof[fanr];
    int next = first_face_dof[fanr+1];
    for ( int j = first; j < next; j++ )
      dnums.Append(j);
  }
 
  int FETIDP_LagrangeSpace :: GetNDof () const
  {
    return ndof;
  }

  FETIDP_CompoundSpace :: FETIDP_CompoundSpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
    : CompoundFESpace ( ama, aspaces, flags )
  {
    ;
  }


    FESpace * FETIDP_CompoundSpace :: Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(2);
      
      Flags primalflags(flags);
      Flags dualflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      primalflags.SetFlag ("order", order);

      string type = flags.GetStringFlag ( "space", "");
      if ( flags.GetDefineFlag("refinep") )
	primalflags.SetFlag("refinep");

      if ( type == "h1ho" )
	spaces[0] = new FETIDP_H1HOSpace (ma, primalflags);

      const FETIDP_H1HOSpace * aconstprimalspace = 
	dynamic_cast<const FETIDP_H1HOSpace*> (spaces[0]);

      FETIDP_H1HOSpace * aprimalspace = 
	const_cast<FETIDP_H1HOSpace*> (aconstprimalspace);
      spaces[1] = new FETIDP_LagrangeSpace ( ma, dualflags );

      FETIDP_CompoundSpace * fes = new FETIDP_CompoundSpace (ma, spaces, flags);
      fes->SetPrimalSpace(aprimalspace);

      return fes;
    }


  void FETIDP_CompoundSpace :: Update ()
  {
    cout << "\nFETIDP CompoundSpace: order " << order << endl;
    cummulative_nd.SetSize (spaces.Size()+1);
    cummulative_nd[0] = 0;

    const_cast<FESpace*> ( spaces[0] ) -> Update();
    
    const FETIDP_H1HOSpace * h1hospace = dynamic_cast<const FETIDP_H1HOSpace*> (spaces[0]);
    
    const FETIDP_PrimalSpace * primalspace;
    if ( h1hospace ) primalspace = h1hospace;
    
    const FETIDP_LagrangeSpace * lagrangespace = dynamic_cast<const FETIDP_LagrangeSpace *>
      (spaces[1]);
    
    const_cast<FETIDP_LagrangeSpace*> ( lagrangespace ) -> Update ( *primalspace );

    for (int i = 0; i < spaces.Size(); i++)
      {
	cummulative_nd[i+1] = cummulative_nd[i] + spaces[i]->GetNDof();
      }

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (cummulative_nd.Last());

    prol -> Update();

    (*testout) << "Update compound fespace" << endl;
    (*testout) << "cummulative dofs start at " << cummulative_nd << endl;

#ifdef PARALLEL
    UpdateParallelDofs();
#endif


  } 


  MatrixGraph * FETIDP_CompoundSpace :: GetGraph_S (bool symmetric)
  {
    int i, j, k;
    int ndof = GetNDof();
    int ne = GetMeshAccess().GetNE();
    int nse = GetMeshAccess().GetNSE();

    ARRAY<int> linesize (ndof);
    ARRAY<int> dnums, dnums2;

    ARRAY<int> dof_num_el (ndof);
    ARRAY<int> el_num_dof (ne);

    PrintReport (*testout);

    // generate dof -> volume element table
    dof_num_el = 0;
    el_num_dof = 0;

    for (i = 0; i < ne; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostEl(i)) continue;
#endif
	if (!DefinedOn (ma.GetElIndex(i))) continue;
// 	spaces[0]->GetExternalDofNrs (i, dnums);

	dnums.SetSize(0);
	spaces[0]->GetDofNrs (i, dnums2);
	for ( int ii = 0; ii < dnums2.Size(); ii++ )
	  if ( primalspace->IsPrimalDof(dnums2[ii]) )
	    dnums.Append(dnums2[ii]);

	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof_num_el[dnums[j]]++;
	      el_num_dof[i]++;
	    }
      }


    Table<int> dof2el (dof_num_el);
    Table<int> el2dof (el_num_dof);

    dof_num_el = 0;
    el_num_dof = 0;
    for (i = 0; i < ne; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostEl(i)) continue;
#endif
	if (!DefinedOn (ma.GetElIndex(i))) continue;
// 	spaces[0]->GetExternalDofNrs (i, dnums);
	dnums.SetSize(0);
	spaces[0]->GetDofNrs (i, dnums2);
	for ( int ii = 0; ii < dnums2.Size(); ii++ )
	  if ( primalspace->IsPrimalDof(dnums2[ii]) )
	    dnums.Append(dnums2[ii]);
	    
	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof2el[dnums[j]][dof_num_el[dnums[j]]++] = i;
	      el2dof[i][el_num_dof[i]++] = dnums[j];
	    }
      }


    // generate dof -> surface element table
    dof_num_el = 0;
    for (i = 0; i < nse; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostSEl(i)) continue;
#endif

	spaces[0]->GetSDofNrs (i, dnums);

	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    dof_num_el[dnums[j]]++;
      }

    Table<int> dof2sel (dof_num_el);

    dof_num_el = 0;
    for (i = 0; i < nse; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostSEl(i)) continue;
#endif
	if (!DefinedOnBoundary (ma.GetSElIndex(i))) continue;
	spaces[0]->GetSDofNrs (i, dnums);
	    
	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof2sel[dnums[j]][dof_num_el[dnums[j]]] = i;
	      dof_num_el[dnums[j]]++;
	    }
      }

     

    ARRAY<int> elflags(ndof);
    elflags = -1;

    for (i = 0; i < ndof; i++)
      {
	linesize[i] = 1;
	elflags[i] = i;
	for (j = 0; j < dof2el[i].Size(); j++)
	  {
	    int elnr = dof2el[i][j];

	    FlatArray<int> dnums2 = el2dof[elnr];
	    for (k = 0; k < dnums2.Size(); k++)
	      {
		int dnumk = dnums2[k];
		if (!symmetric || dnumk <= i)
		  {
		    if (elflags[dnumk] != i)
		      {
			elflags[dnumk] = i;
			linesize[i]++;
		      }
		  }
	      }


	  }

	for (j = 0; j < dof2sel[i].Size(); j++)
	  {
	    int elnr = dof2sel[i][j];
	    spaces[0]->GetSDofNrs (elnr, dnums);
	    for (k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1 &&
		  (!symmetric || dnums[k] <= i))
		{
		  if (elflags[dnums[k]] != i)
		    {
		      elflags[dnums[k]] = i;
		      linesize[i]++;
		    }
		}
	  }


      }

    int cnt = 0;
    for (int l = 0; l < linesize.Size(); l++)
      cnt += linesize[l];

    MatrixGraph * graph = new MatrixGraph (linesize);

    //      graph->Print (cout);

    ARRAY<int> help(ndof);
    elflags = -1;
      
    for (i = 0; i < ndof; i++)
      {
	int * data = graph->GetRowIndicesPointer(i);
	*data = i;
	data++;
	elflags[i] = i;

	for (j = 0; j < dof2el[i].Size(); j++)
	  {
	    int elnr = dof2el[i][j];

	    FlatArray<int> dnums2 = el2dof[elnr];
	    for (k = 0; k < dnums2.Size(); k++)
	      {
		int dnumk = dnums2[k];
		if (!symmetric || dnumk <= i)
		  {
		    if (elflags[dnumk] != i)
		      {
			elflags[dnumk] = i;
			*data = dnumk;
			data++;
		      }
		  }
	      }
	  }
	  
	for (j = 0; j < dof2sel[i].Size(); j++)
	  {
	    int elnr = dof2sel[i][j];
	    spaces[0]->GetSDofNrs (elnr, dnums);
	    for (k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1 &&
		  (!symmetric || dnums[k] <= i))
		{
		  if (elflags[dnums[k]] != i)
		    {
		      elflags[dnums[k]] = i;
		      *data = dnums[k];
		      data++;
		    }
		}
	  }
	  


	MergeSort (linesize[i], graph->GetRowIndicesPointer(i), &help[0]);
      }

    graph -> FindSameNZE();
    return graph;
  }


  MatrixGraph * FETIDP_CompoundSpace :: GetGraph (int level, bool symmetric)
  {
    int i, j, k;
    int ndof = GetNDof();
    int ne = GetMeshAccess().GetNE();
    int nse = GetMeshAccess().GetNSE();

    ARRAY<int> linesize (ndof);
    ARRAY<int> dnums;

    ARRAY<int> dof_num_el (ndof);
    ARRAY<int> el_num_dof (ne);

    PrintReport (*testout);

    // generate dof -> volume element table
    dof_num_el = 0;
    el_num_dof = 0;

    for (i = 0; i < ne; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostEl(i)) continue;
#endif
	if (!DefinedOn (ma.GetElIndex(i))) continue;
// 	GetPDLDofNrs (i, dnums);
	spaces[0]->GetDofNrs (i, dnums);
	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof_num_el[dnums[j]]++;
	      el_num_dof[i]++;
	    }
      }


    Table<int> dof2el (dof_num_el);
    Table<int> el2dof (el_num_dof);

    dof_num_el = 0;
    el_num_dof = 0;
    for (i = 0; i < ne; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostEl(i)) continue;
#endif
	if (!DefinedOn (ma.GetElIndex(i))) continue;
// 	GetPDLDofNrs (i, dnums);
	spaces[0]->GetDofNrs (i, dnums);
	    
	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof2el[dnums[j]][dof_num_el[dnums[j]]++] = i;
	      el2dof[i][el_num_dof[i]++] = dnums[j];
	    }
      }


    // generate dof -> surface element table
    dof_num_el = 0;
    for (i = 0; i < nse; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostSEl(i)) continue;
#endif

	spaces[0]->GetSDofNrs (i, dnums);

	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    dof_num_el[dnums[j]]++;
      }

    Table<int> dof2sel (dof_num_el);

    dof_num_el = 0;
    for (i = 0; i < nse; i++)
      {
#ifdef PARALLEL
	if ( ma.IsGhostSEl(i)) continue;
#endif
	if (!DefinedOnBoundary (ma.GetSElIndex(i))) continue;
	spaces[0]->GetSDofNrs (i, dnums);
	    
	for (j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	    {
	      dof2sel[dnums[j]][dof_num_el[dnums[j]]] = i;
	      dof_num_el[dnums[j]]++;
	    }
      }

     

    ARRAY<int> elflags(ndof);
    elflags = -1;

    for (i = 0; i < ndof; i++)
      {
	linesize[i] = 1;
	elflags[i] = i;
	for (j = 0; j < dof2el[i].Size(); j++)
	  {
	    int elnr = dof2el[i][j];

	    FlatArray<int> dnums2 = el2dof[elnr];
	    for (k = 0; k < dnums2.Size(); k++)
	      {
		int dnumk = dnums2[k];
		if (!symmetric || dnumk <= i)
		  {
		    if (elflags[dnumk] != i)
		      {
			elflags[dnumk] = i;
			linesize[i]++;
		      }
		  }
	      }


	  }

	for (j = 0; j < dof2sel[i].Size(); j++)
	  {
	    int elnr = dof2sel[i][j];
	    spaces[0]->GetSDofNrs (elnr, dnums);
	    for (k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1 &&
		  (!symmetric || dnums[k] <= i))
		{
		  if (elflags[dnums[k]] != i)
		    {
		      elflags[dnums[k]] = i;
		      linesize[i]++;
		    }
		}
	  }


      }

    int cnt = 0;
    for (int l = 0; l < linesize.Size(); l++)
      cnt += linesize[l];

    MatrixGraph * graph = new MatrixGraph (linesize);

    //      graph->Print (cout);

    ARRAY<int> help(ndof);
    elflags = -1;
      
    for (i = 0; i < ndof; i++)
      {
	int * data = graph->GetRowIndicesPointer(i);
	*data = i;
	data++;
	elflags[i] = i;

	for (j = 0; j < dof2el[i].Size(); j++)
	  {
	    int elnr = dof2el[i][j];

	    FlatArray<int> dnums2 = el2dof[elnr];
	    for (k = 0; k < dnums2.Size(); k++)
	      {
		int dnumk = dnums2[k];
		if (!symmetric || dnumk <= i)
		  {
		    if (elflags[dnumk] != i)
		      {
			elflags[dnumk] = i;
			*data = dnumk;
			data++;
		      }
		  }
	      }
	  }
	  
	for (j = 0; j < dof2sel[i].Size(); j++)
	  {
	    int elnr = dof2sel[i][j];
	    spaces[0]->GetSDofNrs (elnr, dnums);
	    for (k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1 &&
		  (!symmetric || dnums[k] <= i))
		{
		  if (elflags[dnums[k]] != i)
		    {
		      elflags[dnums[k]] = i;
		      *data = dnums[k];
		      data++;
		    }
		}
	  }
	  


	MergeSort (linesize[i], graph->GetRowIndicesPointer(i), &help[0]);
      }

    graph -> FindSameNZE();
    return graph;
  }

  void FETIDP_CompoundSpace :: GetPDLDofNrs ( int elnr, ARRAY<int> & dnums ) const
  {
    dnums.SetSize(0);
    ARRAY<int> dnums2;

    PrimalSpace() . GetPrimalDualDofNrs(elnr, dnums);
    spaces[1] -> GetExternalDofNrs( elnr, dnums2);

    for ( int i = 0; i < dnums2.Size(); i++ )
      dnums.Append(dnums2[i]);

  }

  // register FESpaces
  namespace
#ifdef MACOS
  fetidpfespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("fetidp_compound", FETIDP_CompoundSpace::Create);
      GetFESpaceClasses().AddFESpace ("fetidp_h1ho", FETIDP_H1HOSpace::Create);
      GetFESpaceClasses().AddFESpace ("fetidp_lagrange", FETIDP_LagrangeSpace::Create);
    }
    
#ifdef MACOS
    static
#endif
    Init init;
  }







}
