 /**
   High Order Finite Element Space for H(Curl) 
*/
#include <comp.hpp>
#include <fem.hpp> 
#include <multigrid.hpp>
#include "hcurlhofespace_a.hpp"

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif

namespace ngcomp 
{
  using namespace ngcomp; 

#ifdef PARALLEL
  using namespace ngparallel;
#endif

  HCurlHighOrderFESpaceA ::  
  HCurlHighOrderFESpaceA (const MeshAccess & ama, const Flags & aflags,  bool parseflags)
    : FESpace (ama, aflags),
      flags(aflags)
  {
    name="HCurlHighOrderFESpace(hcurlho)";
    
    // define flags
    DefineDefineFlag("hcurlho");
    DefineNumFlag("face");
    DefineNumListFlag("gradientdomains");
    DefineNumListFlag("gradientboundaries");
    DefineDefineFlag("nograds");
    DefineDefineFlag("variableorder");
    DefineNumFlag("orderinner");
    DefineNumFlag("orderface");
    DefineNumFlag("orderedge");
    DefineNumFlag("relorder");
    DefineNumListFlag("dom_order");
    DefineNumListFlag("dom_order_min_x");
    DefineNumListFlag("dom_order_max_x");
    DefineNumListFlag("dom_order_min_y");
    DefineNumListFlag("dom_order_max_y");
    DefineNumListFlag("dom_order_min_z");
    DefineNumListFlag("dom_order_max_z");
    DefineDefineFlag("print"); 
    DefineNumFlag("augmented");
    DefineNumFlag("alternative");
    
    if(parseflags) ParseFlags(flags);

    int ndom = ma.GetNDomains();
    gradientdomains.SetSize (ndom);
    gradientdomains.Set();

    gradientboundaries.SetSize (ma.GetNBoundaries());
    gradientboundaries.Set();

    fn=int(flags.GetNumFlag("face",1));

    if (flags.NumListFlagDefined("gradientdomains"))
      {
	const ARRAY<double> & graddomains = flags.GetNumListFlag ("gradientdomains");
	for (int i = 0; i < gradientdomains.Size(); i++)
	  if (!graddomains[i])
	    gradientdomains.Clear(i);
	cout << endl << " ******* gradientdomains" << gradientdomains << endl;
      }
    if (flags.NumListFlagDefined("gradientboundaries"))
      {
	const ARRAY<double> & gradbounds = flags.GetNumListFlag ("gradientboundaries");
	for (int i = 0; i < gradientboundaries.Size(); i++)
	  if (!gradbounds[i])
	    gradientboundaries.Clear(i);
	cout << endl << " ******* gradientboundaries" << gradientboundaries << endl;
      }
    
    if(flags.GetDefineFlag("nograds"))
      {
	gradientdomains.Clear();  
	gradientboundaries.Clear();   
	nograds = 1; 
      } 
       
    low_order_space = new  NedelecFESpace (ma, 0, dimension, iscomplex);
    prol = 0; // new ngmg::EdgeProlongation (*static_cast<NedelecFESpace*> (low_order_space));
        
    ndof_level.SetSize(0);

    var_order = flags.GetDefineFlag("variableorder");
    
    dom_order_min.SetSize(ndom); 
    dom_order_max.SetSize(ndom); 
	
    uniform_order_inner = int (flags.GetNumFlag ("orderinner", -1));
    uniform_order_face = int (flags.GetNumFlag ("orderface", -1));
    uniform_order_edge = int (flags.GetNumFlag ("orderedge", -1));
	
    orderz = int ( flags.GetNumFlag ("orderz", -1));
    //
    dom_order_min = INT<3> (order,order,order); 
    dom_order_max = INT<3> (99,99,99); 
    
    rel_order = int (flags.GetNumFlag ("relorder", -1)); 
    //

    if(var_order)
      {
	dom_order_min = INT<3> (order,order,order); 
	dom_order_max = INT<3> (99,99,99); 

	rel_order = int (flags.GetNumFlag ("relorder", -1)); 

	if(flags.NumListFlagDefined("dom_order"))
	  {
	    const ARRAY<double> & dom_order = flags.GetNumListFlag ("dom_order");
	    for (int i=0;i<ndom;i++)
	      for (int l=0;l<3;l++)
		{
		  dom_order_min[i][l] = int(dom_order[i]);
		  dom_order_max[i][l] = int(dom_order[i]); 
		}
	  }
	
	if (flags.NumListFlagDefined("dom_order_min_x"))
	  {
	    const ARRAY<double> & dom_order_min_x = flags.GetNumListFlag ("dom_order_min_x");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][0] = int(dom_order_min_x[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_x"))
	  {
	    const ARRAY<double> & dom_order_max_x = flags.GetNumListFlag ("dom_order_max_x");
	    for (int i=0;i<ndom;i++)
	      {
		dom_order_max[i][0] = int(dom_order_max_x[i]); 
		if(dom_order_min[i][0] > dom_order_max[i][0])
		  dom_order_min[i][0] = dom_order_max[i][0];
	      }
	  }
	if (flags.NumListFlagDefined("dom_order_min_y"))
	  {
	    const ARRAY<double> & dom_order_min_y = flags.GetNumListFlag ("dom_order_min_y");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][1] = int(dom_order_min_y[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_y"))
	  {
	    const ARRAY<double> & dom_order_max_y = flags.GetNumListFlag ("dom_order_max_y");
	    for (int i=0;i<ndom;i++)
	      {
		dom_order_max[i][1] = int(dom_order_max_y[i]); 
		if(dom_order_min[i][1] > dom_order_max[i][1])
		  dom_order_min[i][1] = dom_order_max[i][1];
	      }
	  }
	if (flags.NumListFlagDefined("dom_order_min_z"))
	  {
	    const ARRAY<double> & dom_order_min_z = flags.GetNumListFlag ("dom_order_min_z");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][2] = int(dom_order_min_z[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_z"))
	  {
	    const ARRAY<double> &  dom_order_max_z = flags.GetNumListFlag ("dom_order_max_z");
	    for (int i=0;i<ndom;i++)
	      {
		dom_order_max[i][2] = int(dom_order_max_z[i]); 
		if(dom_order_min[i][2] > dom_order_max[i][2])
		  dom_order_min[i][2] = dom_order_max[i][2];
	      }
	  }

	
	
      }
    else
      {
	dom_order_min = INT<3> (order,order,order); 
	dom_order_max = INT<3> (order,order,order); 
      }
    
    /*
    (*testout) << "dom_order_min = " << dom_order_min << endl;
    (*testout) << "dom_order_max = " << dom_order_max << endl;
    */
    
    if (flags.NumFlagDefined("smoothing")) 
      throw Exception ("Flag 'smoothing' for fespace is obsolete \n Please use flag 'blocktype' in preconditioner instead");
    if (flags.NumFlagDefined("cluster")) 
      throw Exception ("Flag 'cluster' for fespace is obsolete \n Please use flag 'ds_cluster' in preconditioner instead");

    //smoother = int (flags .GetNumFlag("smoothing",7));
    //clustertype = int (flags .GetNumFlag("cluster",2));
    //  level_adapted_order = flags.GetDefineFlag("level_adpated_order"); 

    
    augmented = int (flags.GetNumFlag ("augmented", 0));

    alternative = int( flags.GetNumFlag ( "alternative", 0) );
    if ( alternative ) augmented = 0;
    cout << "alternative " << alternative << endl;

    /*    if (flags.GetDefineFlag ("optext"))
	  { 
	  trig = new HCurlHighOrderTrig<TrigExtensionMin> (order);
	  tet =  new HCurlHighOrderTet<TrigExtensionMin> (order);
	  prism =  new HCurlHighOrderPrism<TrigExtensionMin> (order);
	  }
	  else
    */

    
    /*
    trig = new HCurlHighOrderTrig<IntegratedLegendreMonomialExt> (order);
    tet =  new HCurlHighOrderTet<IntegratedLegendreMonomialExt> (order);
    prism =  new HCurlHighOrderPrism<IntegratedLegendreMonomialExt> (order);
    quad = new HCurlHighOrderQuad<IntegratedLegendreMonomialExt> (order);
    hex = new HCurlHighOrderHex<IntegratedLegendreMonomialExt> (order);
    segm = new HCurlHighOrderSegm<IntegratedLegendreMonomialExt> (order); 
    pyramid = new HCurlHighOrderPyr<IntegratedLegendreMonomialExt> (order);
    
      
    dynamic_cast<HCurlHighOrderFiniteElement<1>*> (segm) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<2>*> (trig) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (tet) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (prism) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<2>*> (quad) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (pyramid) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (hex) -> SetAugmented (augmented);
    */    


    // Evaluator for shape tester 
    static ConstantCoefficientFunction one(1);
    if (ma.GetDimension() == 2)
      {
	ARRAY<CoefficientFunction*> coeffs(1);
	coeffs[0] = &one;
	evaluator = GetIntegrators().CreateBFI("massedge", 2, coeffs);
      }
    else if(ma.GetDimension() == 3) 
      {
	ARRAY<CoefficientFunction*> coeffs(1); 
	coeffs[0] = &one;
	evaluator = GetIntegrators().CreateBFI("massedge",3,coeffs); 
	boundary_evaluator = GetIntegrators().CreateBFI("robinedge",3,coeffs); 
	
      }


#ifdef PARALLEL
    ResetParallelDofs();
#endif

  }
  
  HCurlHighOrderFESpaceA :: ~HCurlHighOrderFESpaceA ()
  {
    ;
  }
  
  FESpace * HCurlHighOrderFESpaceA :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    /*
      int order = int(flags.GetNumFlag ("order", 0));
      int dim = int(flags.GetNumFlag ("dim", 1));
      bool iscomplex = flags.GetDefineFlag ("complex");

      if (order <= 0)
      return new NedelecFESpace (ma,0, dim, iscomplex);      // Space with MG   
      else       
    */
    return new HCurlHighOrderFESpaceA (ma, flags);
  }
  
  void HCurlHighOrderFESpaceA :: Update()
  {
    if (order < 0) 
      throw Exception("HCurlHighOrderFESpaceA::Update() order < 0 !" ) ;

    low_order_space -> Update();

    nv = ma.GetNV();
    nel = ma.GetNE();
    int nse = ma.GetNSE();
    maxorder = -1; 
    minorder = 99; 

    for(int i=0; i<specialelements.Size(); i++)
      delete specialelements[i];
    specialelements.DeleteAll();

    switch(ma.GetDimension())
      {
      case 2: 
	ned = ma.GetNEdges();
	nfa = 0;  
	break; 
      case 3: 
	ned = ma.GetNEdges();
	nfa = ma.GetNFaces();
	break; 
      }

    order_edge.SetSize (ned);   
    order_face.SetSize (nfa); 
    order_inner.SetSize (nel);
    usegrad_edge.SetSize (ned);                                
    usegrad_face.SetSize (nfa); 
    usegrad_cell.SetSize (nel);
    fine_edge.SetSize(ned); 
    fine_face.SetSize(nfa); 
     
    order_edge = 0; 
    order_face = INT<2> (0,0);
    order_inner = INT<3> (0,0,0);
    fine_edge = 0; 
    fine_face = 0; 
     
    usegrad_edge = 0;                                
    usegrad_face = 0; 
    usegrad_cell = 0; 

    ARRAY<int> eledges, elfaces, vnums;
    int dim = ma.GetDimension();

    /* int minuniform = 0; 
       if(uniform_order_inner !=-1) minuniform = uniform_order_inner;
       if(uniform_order_face !=-1)
       if(minuniform > 0) minuniform = min2(uniform_order_face,minuniform);
       else minuniform = uniform_order_face; 
       if(uniform_order_edge !=-1)
       if(minuniform > 0) minuniform = min2(uniform_order_edge,minuniform);
       else minuniform = uniform_order_edge; */

    

    for(int i=0; i< nel; i++) 
      if(gradientdomains[ma.GetElIndex(i)]) 
	{
	  ma.GetElEdges(i,eledges);
	  
	  for(int j=0;j<eledges.Size();j++)
	    usegrad_edge[eledges[j]]=1;
	  
	  if(ma.GetDimension()==3)
	    {
	      ma.GetElFaces(i,elfaces);
	      
	      for(int j=0;j<elfaces.Size();j++)
		usegrad_face[elfaces[j]]=1; 
	    }
	  usegrad_cell[i] = 1; 
	} 

    //    cout << "hack JS: usegrad_cell = 0" << endl;
    // usegrad_cell = 0;


    for(int i=0; i<nse && gradientboundaries.Size(); i++)
      if(gradientboundaries[ma.GetSElIndex(i)])
	{
	  ma.GetSElEdges(i,eledges);
	  for(int j=0; j<eledges.Size();j++)
	    usegrad_edge[eledges[j]] = 1;
	
	  if(ma.GetDimension()==3)
	    usegrad_face[ma.GetSElFace(i)] = 1;
	}

    for (int i = 0; i < nel; i++)
      {
	int index = ma.GetElIndex(i);
	// *testout << " GetElement " << i << " with index " << index << endl; 
	
	
	if (!DefinedOn (index)) continue;
	// *testout << " DefinedOn(index) " << endl; 
	INT<3> el_orders = ma.GetElOrders(i);
	
	for(int j=0;j<3;j++)
	  {
	    el_orders[j] = int(max2(dom_order_min[index][j],min2(dom_order_max[index][j],el_orders[j]+rel_order)));

	    if(el_orders[j] > maxorder) maxorder=el_orders[j];
	    if(el_orders[j] < minorder) minorder=el_orders[j];
	  }
	
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const FACE * faces = ElementTopology::GetFaces (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	ma.GetElVertices (i, vnums);

	ma.GetElEdges (i, eledges);		
	for(int j=0;j<eledges.Size();j++) fine_edge[eledges[j]] = 1; 

	if (dim == 3)
	  {
	    ma.GetElFaces(i, elfaces);
	    for(int j=0;j<elfaces.Size();j++) fine_face[elfaces[j]] = 1; 
	  }

	if(uniform_order_edge==-1)
	  {
	    for(int j=0;j<eledges.Size();j++)
	      for(int k=0;k<dim;k++)
		if(points[edges[j][0]][k] != points[edges[j][1]][k])
		  { 
		    order_edge[eledges[j]] = el_orders[k];
		    break; 
		  }
	  }
	else  
	  for(int j=0;j<eledges.Size();j++) 
	    order_edge[eledges[j]] = uniform_order_edge;  

	if(dim==3)
	  { 
	    if(uniform_order_face==-1)
	      for(int j=0;j<elfaces.Size();j++)
		{
		  if(faces[j][3]==-1) 
		    {
		      
		      order_face[elfaces[j]][0] = int(max2(order_face[elfaces[j]][0], 
							   el_orders[0]));
		      order_face[elfaces[j]][1] = order_face[elfaces[j]][0]; 
		    }
		  else //quad_face
		    {
		      int fmax = 0;
		      for(int k = 1; k < 4; k++) 
			if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   

		      
		      INT<2> f((fmax+3)%4,(fmax+1)%4); 
		      if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		      
		      // fmax > f[0] > f[1]
		      // p[0] for direction fmax,f[0] 
		      // p[1] for direction fmax,f[1] 
		      for(int l=0;l<2;l++)
			for(int k=0;k<3;k++)
			  if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			    {
			      order_face[elfaces[j]][l] = int(max2(order_face[elfaces[j]][l], 
				 				   el_orders[k]));
			      break; 
			    } 
		    }  

		}
	    else
	      for(int j=0;j<elfaces.Size();j++) order_face[elfaces[j]] = INT<2> (uniform_order_face,uniform_order_face);
	    
	  }
	
	if(uniform_order_inner==-1)
	  for(int j=0;j<dim;j++)
	    order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
	else 
	  order_inner[i] = INT<3> (uniform_order_inner,uniform_order_inner,uniform_order_inner); 
	
	/*
	  int minorder = el_orders[0]; 
	  for(int j=1;j<2;j++)
	  if(minorder > el_orders[j]) minorder = el_orders[j];
       
	  if(uniform_order_inner != -1 && uniform_order_face != -1 && uniform_order_edge !=-1)	   
	  minorder=minuniform; 
	  else
	  minorder = min2(minuniform,minorder); 
	   
	  for(int j=0;j<nv; j++) 
	  if(order_avertex[vnums[j]]==0) 
	  order_avertex[j] = minorder+1;  // +1 compatible with  H1Space
	  else 
	  if(minorder < order_avertex[vnums[j]])
	  order_avertex[vnums[j]] = minorder+1; 	 */ 
	
      }
   
    /*
    // skip dofs on Dirichlet boundary hack [JS]
    int nsel = ma.GetNE();
    for (i = 0; i < nsel; i++)
    {
    ma.GetSElEdges (i, eledges);
    int face = ma.GetSElFace (i);
	  
    for (j = 0; j < eledges.Size(); j++)
    order_edge[eledges[j]] = 0;
    order_face[face] = 0;
    }
    */

    //  for(i=0;i<ned;i++) order_edge[i] *= usegrad_edge[i]; 
    
    // pyramids 
    for ( int i = 0; i < nel; i++)
      { 
	if (!DefinedOn (ma.GetElIndex(i))) continue;
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	if(eltype == ET_PYRAMID)
	  {
	    /*
	      const FACE * faces = ElementTopology::GetFaces (eltype);
	      const EDGE * edges = ElementTopology::GetEdges (eltype);

	      ma.GetElEdges (i, eledges);		
	      ma.GetElFaces(i, elfaces);

	      for(int j=0;j<8;j++)  
	      order_edge[eledges[j]] = order; 
	    
	      for(int j=0;j<4;j++) order_face[elfaces[j]] = INT<2> (order,order); 
	    */
	    //order_face[elfaces[4]] = INT<2> (0,0); 
	    
	    // order_inner[i] = INT<3>(0,0,0); 
	    order_inner[i] = INT<3>(1,1,1); 
	    ;
	  }
	
      }

    // *testout << " hcurl fine_edge " << fine_edge << endl; 
    // *testout << " hcurl fine_face " << fine_face << endl; 

   
    for (int i=0;i<ned;i++) if(!fine_edge[i]) order_edge[i] = 0; 
    for (int i=0;i<nfa;i++) if(!fine_face[i]) order_face[i] = INT<2> (0,0); 
 

    // set orders
    if ( orderz >= 0 &&  ma.GetDimension() == 2)
      {
	// set orders on quads
	
	for (int el = 0; el < ma.GetNE(); el++)
	  {
	    ma.GetElEdges(el, eledges);
	    if ( ma.GetElType(el) == ET_TRIG ) continue;
	    
	    for ( int i = 0; i < 4; i++ )
	      { 
		int v1, v2;
		Vec<2> point1, point2, normal;
		ma.GetEdgePNums(eledges[i], v1, v2);
		ma.GetPoint(v1, point1);
		ma.GetPoint(v2, point2);
		normal(0) = point1(1) - point2(1);
		normal(1) = point2(0) - point2(0);
		normal /= L2Norm(normal);
		if ( fabs(normal(1)) < 1e-1 )
		  {
		    // *testout << "HCURL " << eledges[i] << endl;
		    order_edge[eledges[i]] = orderz;
		  }
	      }
	  }
     
	for ( int i = 0; i < ma.GetNE(); i++ )
	  {
	    ArrayMem<int,12> vnums; 
	    ma.GetElVertices(i, vnums);
	    
	    if ( ma.GetElType(i) == ET_TRIG ) continue;
	    Vec<2> point1, point2, normal;
	    ma.GetPoint(vnums[0], point1);
	    ma.GetPoint(vnums[1], point2);
	    normal(0) = point1(1) - point2(1);
	    normal(1) = point2(0) - point2(0);
	    normal /= L2Norm(normal);
	    if ( abs( normal(1) ) >  1e-1 )
	      {
		order_inner[i][0] = orderz;
	      }
	    else
	      order_inner[i][1] = orderz;
	    
	  }
	
      }


    UpdateDofTables(); 


#ifdef PARALLEL
    UpdateParallelDofs();
#endif

  }
		
  void HCurlHighOrderFESpaceA :: UpdateDofTables()
  {
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = (ma.GetDimension()==3) ? ma.GetNFaces() : 0;
    nel = ma.GetNE();
    ARRAY<int> pnums; 
    int i; 
    
 
    nedfine = 0; 
    for(int i=0; i<ned; i++) 
      if(fine_edge[i] == 1) nedfine++; 
 
    //ndof = nedfine; 
    ndof = ned; // Nedelec (order = 0) !!   
    if(augmented == 1 ) ndof+=nv; 
       
    first_edge_dof.SetSize(ned+1); 
    for (i = 0; i < ned; i++)
      {
	first_edge_dof[i] = ndof;
	if(order_edge[i]>0)
	 ndof += usegrad_edge[i]*order_edge[i];
       
	// only in 2d: all edges alternative, ho dofs noncontinuous
	if( alternative && order_edge[i] > 0) 
	  {
	    ndof --;
	  }
      }
    first_edge_dof[ned] = ndof;
    // cout << " after edges ndof is " << ndof << endl; 
    
    first_face_dof.SetSize(nfa+1);
    face_ngrad.SetSize(nfa); 
    face_ngrad = 0; 
    for (i=0; i< nfa; i++) 
      { 
	first_face_dof[i] = ndof; 
	ma.GetFacePNums(i,pnums);  
	INT<2> p = order_face[i]; 
	switch(pnums.Size())
	  {
	  case 3: //Triangle   
	    if(p[0]>1)
	      {
		ndof += ((usegrad_face[i]+1)*p[0] + 2)*(p[0]-1)/2;
		

		face_ngrad[i] = usegrad_face[i] *p[0]*(p[0]-1)/2;
	      }
	    break; 
	  case 4: //Quad 
	    if(p[0]>=0 &&  p[1]>=0)
	      {
		ndof += (usegrad_face[i]+1)*p[0]*p[1] + p[0] + p[1]; 
		face_ngrad[i] = usegrad_face[i]*p[0]*p[1];; 
		break; 
	      }
	  }
      } 
    first_face_dof[nfa] = ndof;   
    // cout << " after faces ndof is " << ndof << endl; 
    
    cell_ngrad.SetSize(nel); 
    cell_ngrad = 0; 
    first_inner_dof.SetSize(nel+1);
    for (i=0; i< nel; i++)
      {
	first_inner_dof[i] = ndof;
	INT<3> p = order_inner[i];
	switch(ma.GetElType(i)) 
	  {
	  case ET_TRIG:
	    if(p[0]>1)
	      {
		ndof += ((usegrad_cell[i]+1)*p[0] + 2) * (p[0]-1) /2;
		cell_ngrad[i] = ((usegrad_cell[i])*p[0]) * (p[0]-1) /2;
	      }
	    if ( alternative ) ndof += 3; // 3 high order nonconforming dofs	    
	    break; 
	  case ET_QUAD: 
	    if( (p[0]>=0 &&  p[1]>=0) ) 
	      {
		ndof += (usegrad_cell[i]+1) * p[0] * p[1] + p[0] + p[1]; 
		cell_ngrad[i] = (usegrad_cell[i]) * p[0] * p[1];
	      }
	    if ( alternative ) ndof += 4; // 4 ho nonconforming dofs
	    break; 
	  case ET_TET: 
	    if(p[0]>2)
	      {
		ndof += ((usegrad_cell[i] + 2) *  p[0] + 3) * (p[0]-2) * (p[0]-1) / 6; 
		cell_ngrad[i] = ((usegrad_cell[i] ) *  p[0]) * (p[0]-2) * (p[0]-1) / 6; 
	      }
	    if ( alternative == 1 )  ndof += 6; // 6 ho usw.
	    break; 
	  case ET_PRISM:
	    if( (p[0]>1 &&  p[2]>=0)) 
	      {
		ndof += (usegrad_cell[i]+2)*p[2] * p[0]*(p[0]-1)/2
		  + (p[0]-1)*p[2] + p[0]*(p[0]-1)/2; 
		cell_ngrad[i] = (usegrad_cell[i])*p[2] * p[0]*(p[0]-1)/2;
	      } 
	    if ( alternative && p[0]>0) ndof += 9;
	    break; 
	  case ET_HEX:
	    if(p[0]>=0 && p[1]>=0 && p[2] >=0)
	      {
		ndof += (usegrad_cell[i] + 2)* p[0]*p[1]*p[2] +  p[0]*p[1] 
		  + p[0]*p[2] + p[1]*p[2];
		cell_ngrad[i] = (usegrad_cell[i])* p[0]*p[1]*p[2]; 
	      }
	    if ( alternative && p[0]>0 ) ndof += 12;
	    break; 
	  case ET_PYRAMID:
	    if (p[0] > 0)
	      {
		ndof += 2 * p[0] * (p[0]+1) * (p[0]+2) / 3;
		ndof += usegrad_cell[i] * p[0]*(p[0]-1)*(2*p[0]-1)/6; 
		// ndof += 4;
		// ndof += 0; //(usegrad_cell[i] + 2)* p[0]*p[1]*p[2] +  p[0]*p[1] + p[0]*p[2] + p[1]*p[2];
	      }
	    break; 
	  }
      }
    first_inner_dof[nel] = ndof;    


  (*testout) << "HCurlSpace: " << endl;
  (*testout) << "ndof " << ndof << endl; 
  (*testout) << "number of elements " << ma.GetNE() << endl;
  (*testout) << "number of facets " << ned << endl;
  (*testout) << "ndof " << ndof << " =  2 * " << ma.GetNEdges() << " = " << 2 * ned << endl;

  ndof_level.Append(ndof);

    // cout << " after elements ndof is " << ndof << endl; 

    /*
    (*testout) << " HCURL " << endl; 
    (*testout) << "augmented " << augmented << endl; 
    (*testout) << "ndof  = " << ndof << endl; 

   
    
    (*testout) << "hc first_edge   = " << first_edge_dof << endl;
    (*testout) << "hc first_face   = " << first_face_dof << endl;
    (*testout) << "hc first_inner  = " << first_inner_dof << endl;
    */ 
    /*
    (*testout) << " order = " << order << endl; 
    (*testout) << "order_edge HC = " << order_edge << endl;
    (*testout) << "order_face HC = " << order_face << endl;
    (*testout) << "order_inner HC = " << order_inner << endl;

    (*testout) << " NDOF HC " << ndof << endl; 
    */
    

  }
  

  void HCurlHighOrderFESpaceA :: GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    // wire basket dofs are the nedelec dofs --> dnums = element edgenums
    ARRAY<int> edges;
    ma.GetElEdges ( elnr, edges )    ;
    for ( int i = 0; i < edges.Size(); i++)
      {
	dnums.Append(edges[i]);
	// using full p1
	// dnums.Append(first_edge_dof[edges[i]]);
      }
  }

  void HCurlHighOrderFESpaceA :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
  }

  void HCurlHighOrderFESpaceA :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    //dnums[0] = ednr;
    int first = first_edge_dof[ednr];
    int next = first_edge_dof[ednr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);
  }

  void HCurlHighOrderFESpaceA :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_face_dof[fanr];
    int next = first_face_dof[fanr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);
  }

  void HCurlHighOrderFESpaceA :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_inner_dof[elnr];
    int next = first_inner_dof[elnr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);  
}




  const FiniteElement & HCurlHighOrderFESpaceA :: GetFE (int elnr, LocalHeap & lh) const
    {

    

    FiniteElement * fe = 0;
    
    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	fe = new (lh.Alloc (sizeof(HCurlHighOrderTetA<IntegratedLegendreMonomialExt>))) HCurlHighOrderTetA<IntegratedLegendreMonomialExt> (order, alternative);

	//fe = tet; 
	break;
      case ET_PYRAMID: 
	fe = new (lh.Alloc (sizeof(HCurlHighOrderPyrA<IntegratedLegendreMonomialExt>))) HCurlHighOrderPyrA<IntegratedLegendreMonomialExt> (order);
	//fe = pyramid; 
	break;
      case ET_PRISM:
	fe = new (lh.Alloc (sizeof(HCurlHighOrderPrismA<IntegratedLegendreMonomialExt>))) HCurlHighOrderPrismA<IntegratedLegendreMonomialExt> (order);
	//fe = prism; 
	break;
      case ET_TRIG:
	fe = new (lh.Alloc (sizeof(HCurlHighOrderTrigA<IntegratedLegendreMonomialExt>))) HCurlHighOrderTrigA<IntegratedLegendreMonomialExt> (order, alternative);

	//fe = trig; 
	break;
      case ET_QUAD:

	fe = new (lh.Alloc (sizeof(HCurlHighOrderQuadA<IntegratedLegendreMonomialExt>))) HCurlHighOrderQuadA<IntegratedLegendreMonomialExt> (order, alternative);

	//fe = quad; 
	break;
      case ET_HEX: 
	fe = new (lh.Alloc (sizeof(HCurlHighOrderHexA<IntegratedLegendreMonomialExt>))) HCurlHighOrderHexA<IntegratedLegendreMonomialExt> (order);
	//fe = hex; 
	break; 
      default:
	fe = 0;
      }

    if (!fe)
      {
	stringstream str;
	str << "HCurlHighOrderFESpaceA " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }


    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetElVertices(elnr, vnums);
    // cout << " GetFE : vnums " << elnr << " \t " << vnums << endl;  

    if(ma.GetDimension() == 2) 
      {	
	HCurlHighOrderFiniteElementA<2> * hofe = 
	  dynamic_cast<HCurlHighOrderFiniteElementA<2>*> (fe);
	
    	ArrayMem<int, 4> ednums, ord_edge, ug_edge;
	
	ma.GetElEdges(elnr, ednums);
	
	ord_edge.SetSize (ednums.Size());
	ug_edge.SetSize (ednums.Size()); 
	
	for (int j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]];
	    ug_edge[j]  = usegrad_edge[ednums[j]]; 
	  }

	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> SetUsegradEdge (ug_edge); 
	hofe -> SetUsegradCell (usegrad_cell[elnr]); 
	hofe -> SetAugmented(augmented);
	// hofe -> SetOrderVertex (order_vert);

	hofe -> ComputeNDof();
      }   
    else if (ma.GetDimension() == 3) 
      {
	HCurlHighOrderFiniteElementA<3> * hofe = 
	  dynamic_cast<HCurlHighOrderFiniteElementA<3>*> (fe);

	ArrayMem<int, 12> ednums, ord_edge, ug_edge;
	ArrayMem<int, 6> fanums,  ug_face; 
	ArrayMem<INT<2>, 6> ord_face; 
	
	ma.GetElEdges(elnr, ednums);
	ma.GetElFaces(elnr, fanums);
	
	ord_edge.SetSize (ednums.Size());
	ord_face.SetSize (fanums.Size());
	ug_edge.SetSize (ednums.Size()); 
	ug_face.SetSize (fanums.Size()); 
	
	for (int j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]];
	    ug_edge[j] = usegrad_edge[ednums[j]]; 
	  }
	for (int j = 0; j < fanums.Size(); j++)
	  {
	    ord_face[j] = order_face[fanums[j]];
	    ug_face[j] = usegrad_face[fanums[j]];
	  }

       	hofe -> SetAugmented(augmented); 
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderFace (ord_face);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> SetUsegradEdge (ug_edge); 

       	hofe -> SetUsegradFace (ug_face); 
       	hofe -> SetUsegradCell (usegrad_cell[elnr]); 

	//	hofe -> SetAugmented(augmented); 

	/*	ArrayMem<int, 8> order_vert;
		for (int j = 0; j < vnums.Size(); j++)
		order_vert[j] = order_avertex[vnums[j]];
		hofe -> SetOrderVertex (order_vert);
	*/
	
       
	hofe -> ComputeNDof();
      
      }

    return *fe;
  }
 
  const FiniteElement & HCurlHighOrderFESpaceA :: GetSFE (int selnr, LocalHeap & lh) const
  {
    int i, j;

    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(selnr))
      {
      case ET_SEGM: 
	fe = new (lh.Alloc (sizeof(HCurlHighOrderSegmA<IntegratedLegendreMonomialExt>))) HCurlHighOrderSegmA<IntegratedLegendreMonomialExt> (order, alternative);
	//fe = segm; 
	break; 
      case ET_TRIG:
	fe = new (lh.Alloc (sizeof(HCurlHighOrderTrigA<IntegratedLegendreMonomialExt>))) HCurlHighOrderTrigA<IntegratedLegendreMonomialExt> (order, alternative);
	//fe = trig; 
	break;
      case ET_QUAD:
	fe = new (lh.Alloc (sizeof(HCurlHighOrderQuadA<IntegratedLegendreMonomialExt>))) HCurlHighOrderQuadA<IntegratedLegendreMonomialExt> (order, alternative);
	//fe = quad; 
	break;
      default:
	fe = 0;
      }
    if (!fe)
      {
	stringstream str;
	str << "HCurlHighOrderFESpaceA " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(selnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    ArrayMem<int, 8> vnums;
    ArrayMem<int, 4> ednums, ord_edge, ug_edge; 
       
    ma.GetSElVertices(selnr, vnums);
    // cout << " GetSFE : vnums " << selnr << " \t " << vnums << endl;  


    if(ma.GetSElType(selnr) == ET_SEGM)
      {
	HCurlHighOrderFiniteElementA<1> * hofe =
	  dynamic_cast<HCurlHighOrderFiniteElementA<1>*> (fe);

	hofe -> SetVertexNumbers (vnums);
	hofe -> SetAugmented(augmented); 
	ma.GetSElEdges(selnr, ednums);
	hofe -> SetOrderInner (order_edge[ednums[0]]);
	hofe -> SetUsegradCell (usegrad_edge[ednums[0]]);
	hofe -> ComputeSNDof();
      } 
else
      {     
	HCurlHighOrderFiniteElementA<2> * hofe =
	  dynamic_cast<HCurlHighOrderFiniteElementA<2>*> (fe);
      
	ma.GetSElEdges(selnr, ednums);
	
	ord_edge.SetSize (ednums.Size());
	ug_edge.SetSize (ednums.Size()); 
	
	for (j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]]; 
	    ug_edge[j] = usegrad_edge[ednums[j]];
	  }
	
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderInner (INT<3> (order_face[ma.GetSElFace(selnr)][0],order_face[ma.GetSElFace(selnr)][1],0));
      
	hofe -> SetUsegradEdge(ug_edge); 
	hofe -> SetUsegradCell(usegrad_face[ma.GetSElFace(selnr)]); 
	hofe -> SetAugmented(augmented); 

	/*ArrayMem<int, 8> order_vert;
	  for (int j = 0; j < vnums.Size(); j++)
	  order_vert[j] = order_avertex[vnums[j]];
	  hofe -> SetOrderVertex (order_vert);*/


	hofe -> ComputeSNDof();

	
      }
    
    return *fe;
  }
 


  int HCurlHighOrderFESpaceA :: GetNDof () const
  {

    return ndof;
  }

  void HCurlHighOrderFESpaceA :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    // Save phase-functions the way 
    // (1*e1),.. (1*e_ne)  
    // (p_t)*tang_e1, ... p_t*tang_ne
    // p_n * el1, p_i * el1, ... , p_n * nel_n , p_i *el_nel 

    ARRAY<int> vnums, ednums, fnums;
    int i, j;
    int first,next; 

    ma.GetElEdges (elnr, ednums);
    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fnums); 
    else 
      fnums.SetSize(0); 
   
    dnums.SetSize(0);

 
    if(order < 0) 
      throw Exception(" HCurlHighOrderFESpaceA :: GetDofNrs() order < 0 "); 

    //Nedelec0
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);
//     (*testout) << "after nedelec0 dnums.Size() " << dnums.Size() << endl;
    
    // augmented 
    if(augmented==1) 
      {
	cout << " GetDofNrs" << endl; 
	ma.GetElVertices (elnr, vnums);
	cout << " augmented " << vnums << endl; 
	for(i=0;i< vnums.Size();i++) 
	  dnums.Append(ned+vnums[i]); 
      }

    //edges
    for (i = 0; i < ednums.Size(); i++) 
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	
	for (j = first; j <next; j++)
	  dnums.Append (j);
      }
//     (*testout) << "after edges dnums.Size() " << dnums.Size() << endl;
    
    // faces 
    for(i=0; i<fnums.Size(); i++)     
      { 
	first = first_face_dof[fnums[i]]; 
	next = first_face_dof[fnums[i]+1];
	for(j=first ; j<next; j++) 
	  dnums.Append(j); 
      } 
//     (*testout) << "after faces dnums.Size() " << dnums.Size() << endl;
    
    //inner 
    first = first_inner_dof[elnr];
    next = first_inner_dof[elnr+1]; 

    for(j=first; j<next; j++) 
      dnums.Append(j);
    //(*testout) << "after inner dnums.Size() " << dnums.Size() << endl;

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;

//      *testout << " GetElDNums elnr " << elnr << " \t " << dnums << endl; 

  }
 

  void HCurlHighOrderFESpaceA :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    ARRAY<int> vnums, ednums, fnums;
    int i, j;
    int first,next; 

    ma.GetElEdges (elnr, ednums);
    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fnums); 
    else 
      fnums.SetSize(0); 
    dnums.SetSize(0);

 
    if(order < 0) 
      throw Exception(" HCurlHighOrderFESpaceA :: GetDofNrs() order < 0 "); 

    //Nedelec0
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);
    
    //edges
    for (i = 0; i < ednums.Size(); i++) 
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	
	for (j = first; j <next; j++)
	  dnums.Append (j);
      }
    // faces 

    for(i=0; i<fnums.Size(); i++)     
      { 
	first = first_face_dof[fnums[i]]; 
	next = first_face_dof[fnums[i]+1];
	for(j=first ; j<next; j++) 
	  dnums.Append(j); 
      } 

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
  }
 


  void HCurlHighOrderFESpaceA :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> vnums, ednums;
    int fnum; 
    int i, j;
    int first,next; 

    ma.GetSElEdges (selnr, ednums);
    fnum = ma.GetSElFace (selnr); 

    dnums.SetSize(0);

    if(order <0) throw (" HCurlHighOrderFESpaceA :: GetSDofNrs() order < 0 "); 
    //Nedelec
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);

    if(augmented==1) 
      {
	cout << " GetSDofNrs " << endl; 
	ma.GetElVertices (selnr, vnums);
	cout << "vnums " << vnums << endl; 
	for(i=0;i< vnums.Size();i++) 
	  dnums.Append(ned+vnums[i]); 
      } 
    
    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	for (j = first; j <next; j++)
	  dnums.Append (j);
      }

    if(first_face_dof.Size()>1)		       
      {
	// inner = normal + bubbles 
	first = first_face_dof[fnum];
	next = first_face_dof[fnum+1]; 
	for(j=first; j<next; j++) 
	  dnums.Append(j);
      }
    
    if (!DefinedOnBoundary (ma.GetSElIndex (selnr)))
      dnums = -1;

    // *testout << " GetSDNums selnr " << selnr << " \t " << dnums << endl; 
  
  }


  void HCurlHighOrderFESpaceA ::
  SetGradientDomains (const BitArray & adoms)
  {
    gradientdomains = adoms;
  }

//   double InnerProduct (const Vec<3> & a, const Vec<3> & b)
//   {
//     return a(0)*b(0) + a(1)*b(1) + a(2)*b(2);
//   }

  void HCurlHighOrderFESpaceA ::
  SetGradientBoundaries (const BitArray & abnds)
  {
    gradientboundaries = abnds;
  }

  
#ifndef SZ_OLDBLOCKTYPES 
  Table<int> * HCurlHighOrderFESpaceA :: 
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, k, first,ii;
    int ncnt; 
    int ni = nel;
    if (eliminate_internal) ni = 0; 
    
    int SmoothingType = int(precflags.GetNumFlag("blocktype",2)); 
    bool excl_grads = precflags.GetDefineFlag("exclude_grads"); 
    
    // #define SZ_OLDBLOCKTYPES        
      
    ARRAY<int> vnums,elnums; 
    ARRAY<int> orient; 
    ARRAY<int> ednums, fanums;
    
    int augv = augmented; 

        
    if(nfa == 0 && SmoothingType == 1) 
      SmoothingType = 4; 
    if(nfa == 0 && (SmoothingType == 2 || SmoothingType ==3)) 
      SmoothingType = 5;


    cout << "SmoothingType " << SmoothingType << endl; 
    cout << " Use H(Curl)-Block smoothing " ;
    switch(SmoothingType) 
      {
      case 4: // former 11 
	cout << " AFW(loE) + E + F + I (noClusters)" << endl; 
	ncnt = nv  + ned + nfa + ni;
	break;  
      case 5: //	
	cout << " AFW(hoE) + E + F + I (noClusters)" << endl; 
	ncnt = nv  + ned + nfa + ni;
	break; 	
      case 1: // former 10 
	cout << " Line-Clustering and  AFW(loE)  " << endl; 
	ncnt = nv + ned + nfa + ni; 
	break;
      case 2: // former 26 
	cout << " Line-Clustering and AFW(hoE)  " << endl; 
	ncnt = nv + ned + nfa + ni;
	break;
      case 3: // former 23 (equal to HCurl-Old-SmoothingBlocks)  
	cout << " Line-Clustering: AFW(hoE) - (horizE)F - (trigF)I " << endl; 
	// horiz-quadface-stack(horiz-edge,top-bot-trig-Faces, inner)" << endl; 
	ncnt = nv + ned + nfa +ni; 
	break;
      case 6: // der hier is nur fuer testzwecke !!!  
	cout << "Jacobi (Diag)" << endl; 
	ncnt = nv  + ned + nfa + ni;
	break;  
      default: 
	throw Exception("HCurlHoFeSpace:: CreateSmoothingBlocks chosen Blocktype not valid \n Choose blocktype 1-6 "); 
	ncnt =0; 
	return 0; 
      } 

        
    ARRAY<int> cnt(ncnt); 
    cnt = 0;
    ii=0; 

    switch (SmoothingType)
      {
      case 4: // AFW(loE) + E + F + I (noCluster)
	{
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		cnt[pn1]++;
		cnt[pn2]++;
	      }
	  int nnv = nv; 
	  /* if(augmented==1 && !excl_grads)
	    { 
	      nnv += nv; 
	      for (i=0;i<nv;i++)
		cnt[nv+i] = 1; 
	    } */ 
	 
	  for(i=0;i<ned;i++)
	    cnt[nnv+i] =  (first_edge_dof[i+1] - first_edge_dof[i])*(1-excl_grads);
	  for (i = 0; i < nfa; i++)
	    cnt[nnv+ned+i] = first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[nnv+ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i] - excl_grads*cell_ngrad[i];
	  break;
	}
      case 5: // AFW(hoE) + E + F + I (noCluster)
	{
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		cnt[pn1]+= 1 + (first_edge_dof[i+1]-first_edge_dof[i])*(1-excl_grads);
		cnt[pn2]+= 1 + (first_edge_dof[i+1]-first_edge_dof[i])*(1-excl_grads);
	      }
	  int nnv = nv; 
	  /* 
	  if(augmented==1 && wgrads)
	    { 
	      nnv += nv; 
	      for (i=0;i<nv;i++)
		cnt[nv+i] = 1; 
	    } */ 
	  
	  for (i = 0; i < nfa; i++)
	    cnt[nnv+ned+i] = first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[nnv+ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i] - excl_grads*cell_ngrad[i];
	  break;
	}

      case 1: // Clustering with AFW(loE) 
	{
	  
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
	 
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
	  
		pn2 = ma.GetClusterRepVertex(pn2);
		cnt[pn1]++;
		if (pn1 != pn2)
		  cnt[pn2]++;
		// ev. noch lo-edge zu horiz.edges 
		cnt[ma.GetClusterRepEdge(i)] +=
		  (first_edge_dof[i+1] - first_edge_dof[i])*(1-excl_grads);
	      }
	 
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i];
	 
	  for (i = 0; i < ni; i++)
	    cnt[ma.GetClusterRepElement(i)] += 
	      first_inner_dof[i+1] - first_inner_dof[i] - excl_grads*cell_ngrad[i];
	  
	  break;
	}
      case 2: // Clustering with AFW(hoE)
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ (1-excl_grads)*(first_edge_dof[i+1] - first_edge_dof[i]); 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i]; 
	     	 
	  for(i=0;i<ni;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      // Stack: plane-face + inner 
	      cnt[ccl] += first_inner_dof[i+1] - first_inner_dof[i]- excl_grads*cell_ngrad[i]; 
	    }
	  break;
	}

      case 3: 
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ (1-excl_grads)*(first_edge_dof[i+1] - first_edge_dof[i]); 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i]; 
	  
	  for(i=0;i<ni;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      int ndi = first_inner_dof[i+1] - first_inner_dof[i] - excl_grads*cell_ngrad[i]; 
	      // Stack: plane-face + inner 
	      cnt[ccl] += ndi; 
	      
	      
	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      
	      ma.GetElFaces(i,fanums); 
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl) 
		    cnt[fcl] += ndi; 
		}  
	      // if nocluster -> FI + I 
	    }  
	  
	  for(i=0;i<nfa;i++) // Trig-Faces to horiz-edge stack 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		int nd = first_face_dof[i+1] - first_face_dof[i] - excl_grads*face_ngrad[i]; 
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    cnt[ecl] += nd; 
		  }
		  // if nocluster -> AFW(hoE) + EF + FI + I 
	      } 
	  
	  break;
	}
	 case 6: // Jacobi(Diag)
	{
	  int ii=0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      cnt[ii++]=1; 
	  			 
	  for(i=0;i<ned;i++)
	    if(fine_edge[i])
	      for(int k= first_edge_dof[i]; k< first_edge_dof[i+1]; k++) 
		cnt[ii++] = 1;
	  
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      for(int k= first_face_dof[i]; k< first_face_dof[i+1]; k++) 
		cnt[ii++] = 1; 
 
	  for (i = 0; i < ni; i++)
	    for(int k= first_inner_dof[i]; k< first_inner_dof[i+1]; k++) 
	      cnt[ii++] = 1; 
	  
	  break;
	}
      
      }
    
   
        
    Table<int> & table = *new Table<int> (cnt); 
    ii = 0; 
    cnt = 0;
    switch(SmoothingType) 
      {
      case 4: // AFW(loE) + E  + F + I (noLineClusters) 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  int nnv = nv; 
	  
	 
	  /* if(augmented==1)
	    { 
	      nnv += nv; 
	      for(i=0;i<nv;i++)
		table[i][cnt[i]++] = ned+i;
		} */ 
	  
	  if(!excl_grads)
	    for (i =0; i<ned ; i++)
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nnv+i][cnt[nnv+i]++] = j;
	  
	  for (i = 0; i < nfa; i++)
	    { 
	      int first = first_face_dof[i] + excl_grads*face_ngrad[i]; 
	      for (j = first; j < first_face_dof[i+1]; j++)
		table[nnv+ned+i][cnt[nnv+ned+i]++] = j;
	    }
	      
	  for (i = 0; i < ni; i++)
	    { 
	      int first = first_inner_dof[i] + excl_grads*cell_ngrad[i];
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nnv+ned+nfa+i][cnt[nnv+ned+nfa+i]++] = j;
	    }
	  break;
	}
      case 5: // AFW(hoE) + E  + F + I (noLineClusters) 
	{
	  cnt = 0;
	  
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		table[pn1][cnt[pn1]++]  = i; 
		table[pn2][cnt[pn2]++]  = i; 
		if(!excl_grads)
		  for(j=first_edge_dof[i];j<first_edge_dof[i+1];j++) 
		    {
		      table[pn1][cnt[pn1]++]  = j; 
		      table[pn2][cnt[pn2]++]  = j;
		    }
	      }
	  int nnv = nv; 
	  /* if(augmented==1)
	     { 
	     nnv += nv; 
	     for(i=0;i<nv;i++)
	     table[i][cnt[i]++] = ned+i;
	     } */ 
	  for (i = 0; i < nfa; i++)
	    { 
	      int first = first_face_dof[i] + excl_grads*face_ngrad[i]; 
	      for (j = first; j < first_face_dof[i+1]; j++)
		table[nnv+ned+i][cnt[nnv+ned+i]++] = j;
	    }
	  
	  for (i = 0; i < ni; i++)
	    { 
	      int first = first_inner_dof[i] + excl_grads*cell_ngrad[i];
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nnv+ned+nfa+i][cnt[nnv+ned+nfa+i]++] = j;
	    }
	  break;
	}
	
      case 1: // Clustering + AFW(loE) 
	{
	  cnt = 0; 

	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		table[pn1][cnt[pn1]++] = i;
		if (pn1 != pn2)
		  table[pn2][cnt[pn2]++] = i;

		int nr = ma.GetClusterRepEdge (i);
		if(!excl_grads)
		  for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		    table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		int first = first_face_dof[i] + excl_grads*face_ngrad[i];
		for (j = first; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      int first = first_inner_dof[i] + excl_grads*cell_ngrad[i];
	      for (j = first; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  break;
	}
	
      case 2: // Clustering with AFW(hoEdges) 
	{
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		if(!excl_grads)
		  for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		    table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    if(!excl_grads)
		      for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
			{		    
			  table[pn2][cnt[pn2]++] = j;
			  table[ecl][cnt[ecl]++] = j;
			}
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		int first = first_face_dof[i] + excl_grads*face_ngrad[i];
		//	if(fcl  >= nv + ned) // quad-faces (prism) 
		for(j=first; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 
	      }
	  
	  for(i=0;i<ni;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i); 
	      int first = first_inner_dof[i] + excl_grads*cell_ngrad[i];
	      for(j=first;j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	    }
	  break;
	}
      case 3: 
	{
	  // Stack-AFW(hoEdges) 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		if(!excl_grads)
		  for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		    table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    if(!excl_grads)
		      for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
			{		    
			  table[pn2][cnt[pn2]++] = j;
			  table[ecl][cnt[ecl]++] = j;
			}
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		int first = first_face_dof[i] + excl_grads*face_ngrad[i];
		
		for(j=first; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 
	      }
	  
	  for(i=0;i<ni;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i);
	      int first = first_inner_dof[i] + excl_grads*cell_ngrad[i]; 
	      for(j=first;j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	         	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      ma.GetElFaces(i,fanums); 
	      
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl)
		    for(k=first;k<first_inner_dof[i+1];k++) 
		      table[fcl][cnt[fcl]++] = k; 
		} 
	      
	    }  
	  
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		 if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    int first = first_face_dof[i] + excl_grads*face_ngrad[i];
		    for(k=first; k<first_face_dof[i+1]; k++)
		      table[ecl][cnt[ecl]++] = k; 
		  }
	      }
	  
	  break;
	}
      case 6: // Jacobi(Diag)
	{
	  cnt = 0;
	  int ii=0; 
	  
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
		table[ii++][0] = i;

	    for (i =0; i<ned ; i++)
	      if(fine_edge[i])
		for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		  table[ii++][0] = j;
	  
	  for (i = 0; i < nfa; i++)
	    { 
	      if(fine_face[i]) 
		{
		  int first = first_face_dof[i]; 
		  for (j = first; j < first_face_dof[i+1]; j++)
		    table[ii++][0] = j;
		}
	    }
	      
	  for (i = 0; i < ni; i++)
	    { 
	      int first = first_inner_dof[i];
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[ii++][0] = j;
	    }
	  break;
	}
	
      
      }
    //(*testout) << "H(Curl)-table = " << table << endl;	
	
    return & table;
#else 

  Table<int> * HCurlHighOrderFESpaceA :: 
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, k, first,ii;
    int ncnt; 
    int ni = nel;
    if (eliminate_internal) ni = 0; 
    
    int SmoothingType = int(precflags.GetNumFlag("blocktype",2)); 
    bool excl_grads = precflags.GetDefineFlag("exclude_grads"); 
    

    ARRAY<int> vnums,elnums; 
    ARRAY<int> orient; 
    ARRAY<int> ednums, fanums;
    
    if(augmented!=1 && SmoothingType == 9) 
      SmoothingType = 8; 
    if(augmented==1 && SmoothingType != 11) 
      SmoothingType = 9;
    
    int augv = augmented; 


    cout << "SmoothingType " << SmoothingType << endl; 
    cout << " Use H(Curl)-Block smoothing " ;
    switch(SmoothingType) 
      {
      case 8: // former 8 
	cout << " Stack-AFW(hoE) + Clustering E-F-I " << endl; 
	ncnt = nv + ned + nfa + ni;
	break;
      case 10: // former 10 
	cout << " Clustering: AFW + E + F + I  " << endl; 
	ncnt = nv + ned + nfa + ni; 
	break;
      case 11: // former 11 
	cout << " AFW + E + F + I " << endl; 
	ncnt = nv  + ned + nfa + ni;
	break; 
      case 19: // former 19 
	cout << " AFW + E-F-AFW + F + I " << endl; 
	ncnt = nv + ned + nfa + ni; 
	break;
      case 23: // former 23 (HCurl-Old)  
	cout << " Schaumamoi HCURL-OLD SmoothingBlocks 23 " << endl; 
	ncnt = nv + ned + nfa +ni; 
	break; 
      case 26: // former 26 
	cout << " Schaumamoi HCURL-OLD SmoothingBlocks 26 " << endl; 
	ncnt = nv + ned + nfa +ni; 
	break; 
	
      case 1: 
	cout << " 2d Block(N) + E + I " << endl;
	ncnt = 1 + ned + ni; 	    
	break; 
      case 2: 
	cout << " 2d N + E + I " << endl; 
	ncnt = 2*ned + ni; 	    
	break; 
      case 3: 
	cout << "2d  NE + I " << endl;
	ncnt = ned + ni; 	  
	break; 
      case 4: 
	cout << " 2d E + I (no Ned0)" << endl; 
	ncnt = ned + ni; 	    
	break; 
      case 5:
	cout << " 2d AFW +  E + I " << endl;  
	ncnt = nv + ned + ni; 
	break; 
      case 6:
	cout << " E + FI  " << endl; 
	ncnt = ned + nfa;
	break;
      case 7: 
	cout << " Clustering: N + E + F + I " << endl; 
	ncnt = nv + ned + nfa + ni;
	break;
      case 9: 
	cout << " AugmentedV + EI + FI  " << endl; 
	ncnt = nv + ned + nfa;
	break;   
      case 12: 
	cout << " AFW + F + I  without ho grads " << endl; 
	ncnt = nv + nfa + ni; 
	break; 
      case  13: 
	cout << " Clusters ned  + E + F + I without ho-grads " << endl; 
	ncnt = nv + ned + nfa + ni;
      case  14:
	cout << " Clustering: AFW + E + F + I  without ho-grads" << endl; 
	ncnt = nv + ned + nfa + ni; 
	break;
      case  15:
	cout << " AFW + I  without ho-grads " << endl;  
	ncnt = nv +  ni; 
	break; 
      case 16: 
	cout << " F + I without ho-grad " << endl; 
	ncnt = nfa + ni; 
	break; 
      case 17: // Anisotropic Face-AFW with ho-grads 
	cout << " AFW-Face Smoothing for around long anisotopic edges: AFW + E +  F-AFW + F + I " << endl; 
	ncnt = nv + 2*ned + nfa + ni; 
	break;
      case 18: // Anisotropic Face-AFW wo ho-grads 
	cout << " AFW-Face Smoothing for around long anisotopic edges : AFW +  F-AFW + F + I  without ho-grads" << endl; 
	ncnt = nv + ned + nfa + ni; 
	break; 
    
      case 20: // 
	cout << " AFW + F-AFW + F + I withoud grads " << endl; 
	ncnt = nv + ned + nfa + ni; 
	break;
      case 21 : // AFW + EF + I 
	cout << " AFW + EF + I  " << endl; 
	ncnt = nv + ned  + ni; 
	break; 
      case 22  : // AFW + EF + I 
	cout << " AFW + EF + I  without grads " << endl; 
	ncnt = nv + ned  + ni; 
	break;	  
      
      case 24: // HCURL-OLD 
	cout << " Schaumamoi HCURL-OLD SmoothingBlocks 24 " << endl; 
	ncnt = nv + ned + nfa +ni; 
	break; 
      case 25: // HCURL-OLD 
	cout << " Schaumamoi HCURL-OLD SmoothingBlocks 25 " << endl; 
	ncnt = nv + ned + nfa +ni; 
	break; 
    
      default: 
	cout << " Error in HCurlHOFESpace :: CreateSmoothingBlocks no SmoothingType " << endl; 
	ncnt =0; 
	return 0; 
      } 

    ARRAY<int> cnt(ncnt); 
    cnt = 0;
    ii=0; 

    // ARRAY<INT<2> > AnisotropicFaceEdges(nfa);
    ARRAY<INT<4> > AnisotropicFaceEdges(nfa);
    for(i=0;i<nfa;i++)
      for(int k=0;k<4;k++)
	AnisotropicFaceEdges[i][k] = -1; 
    // AnisotropicFaceEdges = INT<4> (-1,-1,-1,-1); 
    ARRAY<int> FaceAnisotropicDir(nfa); 
    FaceAnisotropicDir=-1; 


    
    if(SmoothingType == 17 || SmoothingType==18 || SmoothingType==19)
      {
	double ratio = 20; 
	for(i=0;i<nfa;i++)
	  if(fine_face[i])
	    {
	      ARRAY<int> vnums,enums;

	      ma.GetFacePNums(i,vnums);
	      if(vnums.Size() == 4) // quad_face
		{
		  *testout << " quad_face " << i << endl; 
		  *testout << " order_face[i] " << order_face[i] << endl; 
		  
		  Vec<3> p[4]; 
		  for(j=0;j<4;j++)
		    ma.GetPoint(vnums[j],p[j]); 		     
		    
		  ma.GetFaceEdges(i,enums);
		  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
		  		  
		  // f[0] > f[1] > f[2] ... f[3] diagonal of f[0]
		  int f[4]; 
		  f[0]=0; 
		  for (j = 1; j < 4; j++)
		    if (vnums[j] > vnums[f[0]])
		      f[0] = j;
		  f[1] = (f[0]+3)%4; 
		  f[2] = (f[0]+1)%4; 
		  if(vnums[f[1]] < vnums[f[2]]) 
		    swap(f[1],f[2]);  // f1 > f2; 
		  f[3] = 6 - f[0] - f[1] - f[2];		  

		  *testout << " Vnums " << vnums[0] << " \t " << vnums[1] << " \t " << vnums[2] << "\t" << vnums[3] << endl; 
		  *testout << " Vnums f[i]" << vnums[f[0]] << " \t " << vnums[f[1]] << "\t" 
			   << vnums[f[2]] << "\t" << vnums[f[3]] << endl;   
		  // xi 
		  double l1 = L2Norm(p[f[0]]-p[f[1]]);
		  double l2 = L2Norm(p[f[2]]-p[f[3]]);
		  double lxi = min(l1,l2); 
		  // eta 
		  l1 = L2Norm(p[f[0]]-p[f[2]]);
		  l2 = L2Norm(p[f[1]]-p[f[3]]);
		  double leta = min(l1,l2);
		
		  *testout << " f[0] " << f[0] <<  " f[1] " << f[1] << " f[2] " << f[2] << " f[3] " << f[3] << endl; 
		  *testout << " enums[0]" << enums[0] <<  " enums[1] " << enums[1] << " enums[2] " << enums[2] << " enums[3] " << enums[3] << endl; 
		  
		  *testout << " lxi " << lxi << "\t leta " << leta << endl; 
		  *testout << " lxi / leta " << lxi / leta  << endl; 
		  *testout << " leta / lxi " << leta / lxi  << endl; 
		  /* if(lxi > ratio*leta) // xi is the longer direction
		    {
		      // ordne edges
		      // GetPEdge(
		      AnisotropicFaceEdges[i] =  

		      FaceAnisotropicDir[i] = 0; 
		      if(f[0] +  f[1] == 3) // the longer edge (xi) is a vertical one 
			{
			  *testout << " xi anisotropic with edges "<< enums[2] << "\t" << enums[3] << endl; 
			  AnisotropicFaceEdges[i] = INT<2>
			    (enums[2],enums[3]);
			}
			  else 
			    { 
			      *testout << " xi anisotropic with edges " <<  enums[0] << "\t" << enums[1]  << endl; 
			    AnisotropicFaceEdges[i] = INT<2> (enums[0],enums[1]);
			    } 
		    }
		  else if (leta > ratio*lxi) // eta is the longer direction  
		    {
		      FaceAnisotropicDir[i] = 1; 
		      
		      if(f[0] +  f[2] == 3) // the longer edge (eta) is a vertical one
			{
			   *testout << " eta anisotropic with edges "<< enums[2] << "\t" << enums[3] << endl; 
			   AnisotropicFaceEdges[i] = INT<2> (enums[2],enums[3]);
			}
		      else
			{
			  *testout << " xi anisotropic with edges "<< enums[0] << "\t" << enums[1] << endl; 
			  AnisotropicFaceEdges[i] = INT<2> (enums[0],enums[1]);
			}
		    }
		  */
		  if(lxi > ratio*leta || leta > ratio*lxi) 
		    for(k=0;k<4;k++) 
		      AnisotropicFaceEdges[i][k] = enums[k];  
		}
	    }
	*testout << " AnisotropicEdges " << endl << AnisotropicFaceEdges << endl; 
	// 	*testout << " FaceAnisotropicDir " << endl << FaceAnisotropicDir << endl; 
      }
    
    

    switch (SmoothingType)
      {
      case 8: // former 8 : AFW(hoE) + F + I 
	{
	  cnt = 0;

	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2, cl;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		cl = ma.GetClusterRepEdge(i);
		cnt[pn1] += 1 + first_edge_dof[i+1] - first_edge_dof[i];
		if (pn1 != pn2)
		  {
		    cnt[pn2] += 1 + first_edge_dof[i+1] - first_edge_dof[i];
		    cnt[cl] += 1 + first_edge_dof[i+1] - first_edge_dof[i];
		  }
	      }

	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i];
	  
	  for (i = 0; i < ni; i++)
	    cnt[nv+ned+nfa+i] += 
	      first_inner_dof[i+1] - first_inner_dof[i];
	  break;
	}

      case 1: 
	for(i=0;i<ned;i++) if(fine_edge[i]) cnt[0]++; 
	ii=1; 
	for(i=0; i< ned;i++,ii++ ) 
	  if(fine_edge[i])
	    cnt[ii] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< ni;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 	
	break; 
      case 2: 
	for(i=0; i< ned; i++)  cnt[ii++] = fine_edge[i]; 
	for(i=0; i< ned;i++,ii++ )
	  if(fine_edge[i])
	    cnt[ii] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< ni;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 
	break; 
      case 3: // NE + I 
	for(i=0; i< ned; i++)  cnt[i] = fine_edge[i]; 
	for(i=0; i< ned;i++ ) 
	  if(fine_edge[i])
	    cnt[i] += first_edge_dof[i+1] - first_edge_dof[i];
	ii=ned;
	for(i=0; i< ni;i++)
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 
	break; 
      case 4: // E + I 
	for(i=0; i< ned;i++,ii++ )
	  if(fine_edge[i]) cnt[ii] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< ni;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 	
	break; 	
      case 5: // AFW + E + I 
	{
	  int v1, v2; 
	for(i=0; i< ned; i++) 
	  if(fine_edge[i])
	    {
	      ma.GetEdgePNums (i, v1, v2);
	      cnt[v1]++; 
	      cnt[v2]++;
	    }
	ii = nv; 
	for(i=0; i< ned;i++, ii++ )
	  if(fine_edge[i]) cnt[ii] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< ni;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i] ; 	
	}
	break; 
      
     
      case 6: // E + FI 
	{
	  for (i = 0; i < ned; i++)
	    cnt[i] = first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < ni; i++)
	    {	   
	      if(ma.GetDimension() == 3) ma.GetElFaces (i, fanums);
	      else fanums.SetSize(0); 
	      for (j = 0; j < fanums.Size(); j++)
		cnt[ned+fanums[j]] += first_inner_dof[i+1]-first_inner_dof[i];
	    }
	  break;
	}
      case 7:
	{
	  /*  
	      for (i = 0; i < ned; i++)
	      cnt[i] = first_edge_dof[i+1] - first_edge_dof[i];
	      for (i = 0; i < nfa; i++)
	      cnt[ned+i] = first_face_dof[i+1] - first_face_dof[i];
	      for (i = 0; i < nel; i++)
	      cnt[ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i];
	      break;
	  */

	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      cnt[ma.GetClusterRepEdge(i)] +=
		first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < ni; i++)
	    cnt[ma.GetClusterRepElement(i)] += 
	      first_inner_dof[i+1] - first_inner_dof[i];
	  
	  /*
	    for(i=0; i<nfa; i++)
	    {
	    ma.GetFaceEdges(i,ednums);
	    if(ednums.Size() != 4) continue;
	      
	    int nr_self = ma.GetClusterRepFace(i);

	    ARRAY<int> rep;
	    for(j=0; j<ednums.Size(); j++)
	    {
	    int nr = ma.GetClusterRepEdge(ednums[j]);
	    if(nr != nr_self && !rep.Contains(nr))
	    rep.Append(nr);
	    }
	    for(k=0; k<rep.Size(); k++)
	    cnt[rep[k]] += first_face_dof[i+1] - first_face_dof[i];
	    }
	  */
	  break;
	}
      
      case 9: // AugV + EI + FI 
	{
	  for (i = 0; i < nv; i++)
	    cnt[i] = 1; 
	  for (i = 0; i < ned; i++)
	    cnt[nv+i] = first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[nv+ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < ni; i++)
	    {
	      ma.GetElEdges (i,ednums,orient);
	      for (k = 0; k < ednums.Size(); k++)
		cnt[nv+ednums[k]] += first_inner_dof[i+1] - first_inner_dof[i];

	      ma.GetElFaces (i,fanums,orient);
	      for (k = 0; k < fanums.Size(); k++)
		cnt[nv+ned+fanums[k]] += first_inner_dof[i+1] - first_inner_dof[i];
	    }
	  break;
	}
      case 10: // Clustering : AFW + E + F + I 
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		cnt[pn1]++;
		if (pn1 != pn2)
		  cnt[pn2]++;
		cnt[ma.GetClusterRepEdge(i)] +=
		  first_edge_dof[i+1] - first_edge_dof[i];
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < ni; i++)
	    cnt[ma.GetClusterRepElement(i)] += 
	      first_inner_dof[i+1] - first_inner_dof[i];
	  break;
	}
      case 11: // AFW + E + F + I 
	{
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		cnt[pn1]++;
		cnt[pn2]++;
	      }
	  int nnv = nv; 
	  if(augmented==1)
	    { 
	      nnv += nv; 
	      for (i=0;i<nv;i++)
		cnt[nv+i] = 1; 
	    }
	  
	  for(i=0;i<ned;i++)
	    cnt[nnv+i] =  first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[nnv+ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < ni; i++)
	    cnt[nnv+ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i];
	  break;
	}
      case 12: // AFW  + F + I without grads  
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		cnt[pn1]++;
		cnt[pn2]++;
	      }
	  
	  for (i = 0; i < nfa; i++)
	    cnt[nv+i] = first_face_dof[i+1] - first_face_dof[i] - face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[nv+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	} 
      case 13: // Clustering w/o ho-grads
	{
	  cnt = 0; 
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i] - face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[ma.GetClusterRepElement(i)] += 
	      first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break; 
	}
      case 14: // Clustering : AFW + E + F + I 
	{
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		cnt[pn1]++;
		if (pn1 != pn2)
		  cnt[pn2]++;
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      cnt[ma.GetClusterRepFace(i)] += 
		first_face_dof[i+1] - first_face_dof[i] - face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[ma.GetClusterRepElement(i)] += 
	      first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	}
      case 15: // AFW + I  without ho-grads 
	{
	  int v1, v2; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
	      }
	  ii = nv; 
	  for(i=0; i< ni;i++ )
	    cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i] ; 	
	  break; 	
	}
      case 16: //  F + I without grads  
	{
	  cnt = 0; 
	  for (i = 0; i < nfa; i++)
	    cnt[i] = first_face_dof[i+1] - first_face_dof[i] - face_ngrad[i];
	  for (i = 0; i < ni; i++)
	    cnt[nfa+i] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	} 
      case 17: // Anisotropic Face-AFW with ho-grads 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
		cnt[nv +i] =  first_edge_dof[i+1]-first_edge_dof[i]; 
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
	
		/* if(FaceAnisotropicDir[i] == 0)  // xi - anisotropic dir
		  nfafw = order_face[i][1]; 
		else if(FaceAnisotropicDir[i]==1)
		nfafw = order_face[i][0]; */  
		
		int nfafw = 0; 
		if(AnisotropicFaceEdges[i][0]>=0) 
		  { 
		    nfafw = order_face[i][0]+order_face[i][1]; 
		    for(k=0;k<4;k++)
		      cnt[nv + AnisotropicFaceEdges[i][k]] += nfafw; 
		  }
		//   cnt[nv + ned + AnisotropicFaceEdges[i][1]] += nfafw; 
		cnt[nv + 2*ned + i] = first_face_dof[i+1]-first_face_dof[i] - nfafw;
	      }
	  ii=nv+2*ned+nfa;
	  for (i = 0; i < ni; i++)
	    cnt[ii+i] = first_inner_dof[i+1] - first_inner_dof[i];
	  
	  *testout << " cnt smoothing 17 " << cnt << endl; 
	  break;
	}
      case 18: // Anisotropic Face-AFW wo ho-grads 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
		int nfafw = 0; 
		/*	if(FaceAnisotropicDir[i] == 0)  // xi - anisotropic dir
		  nfafw = order_face[i][1]; 
		else if(FaceAnisotropicDir[i]==1)
		  nfafw = order_face[i][0]; 
		*/
		if(AnisotropicFaceEdges[i][0]>=0) 
		  {
		    nfafw = order_face[i][0]+order_face[i][1]; 
		    cnt[nv + ned + AnisotropicFaceEdges[i][0]] += nfafw; 
		    cnt[nv + ned + AnisotropicFaceEdges[i][1]] += nfafw; 
		    cnt[nv + ned + AnisotropicFaceEdges[i][2]] += nfafw; 
		    cnt[nv + ned + AnisotropicFaceEdges[i][3]] += nfafw; 
		  }

		cnt[nv + 2*ned + i] = first_face_dof[i+1]-first_face_dof[i] - nfafw - face_ngrad[i];
	      }
	  ii=nv+ned+nfa;
	  for (i = 0; i < ni; i++)
	    cnt[ii+i] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	}
      case 19: // QUAD-Face-AFW with ho-grads 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
		cnt[nv + i] = first_edge_dof[i+1] - first_edge_dof[i]; 
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
		cnt[nv+ned+i] = first_face_dof[i+1]-first_face_dof[i]; 
		ARRAY<int> vnums; 
		ma.GetFacePNums(i,vnums);
		if(vnums.Size() == 4) // quad_face
		  { 
		    ARRAY<int> f2ed; 
		    int nfafw = order_face[i][0] + order_face[i][1]; 
		    ma.GetFaceEdges (i, f2ed);
		    for (j = 0; j < f2ed.Size(); j++)
		      cnt[nv+f2ed[j]] += nfafw; 
		    cnt[nv+ned+i] -= nfafw; 
		  } 
		//cnt[nv + ned + i] = first_face_dof[i+1]-first_face_dof[i] - nfafw;  
	      }
	  
	  for (i = 0; i < ni; i++)
	    cnt[nv + ned + nfa + i] = first_inner_dof[i+1] - first_inner_dof[i]; // - cell_ngrad[i];
	  break;
	}
      case 20: // QUAD-Face-AFW wo ho-grads 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
		//cnt[nv + i] = first_edge_dof[i+1] - first_edge_dof[i]; 
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
		cnt[nv + ned + i] = first_face_dof[i+1]-first_face_dof[i] - face_ngrad[i];  
		ARRAY<int> vnums; 
		ma.GetFacePNums(i,vnums);
		if(vnums.Size() == 4) // quad_face
		  { 
		    ARRAY<int> f2ed; 
		    int nfafw = order_face[i][0] + order_face[i][1]; 
		    ma.GetFaceEdges (i, f2ed);
		    for (j = 0; j < f2ed.Size(); j++)
		      cnt[nv+f2ed[j]] += nfafw;
		    cnt[nv+ned+i] =- nfafw; 
		  } 
		
	      }
	  
	  for (i = 0; i < ni; i++)
	    cnt[nv + ned + nfa + i] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	}
      case 21: // AFW + EF + I 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
		cnt[nv + i] = first_edge_dof[i+1] - first_edge_dof[i]; 
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
		ARRAY<int> f2ed; 
		ma.GetFaceEdges (i, f2ed);
		for (j = 0; j < f2ed.Size(); j++)
		  cnt[nv+f2ed[j]] += first_face_dof[i+1] - first_face_dof[i];  
	      }
	  
	  for (i = 0; i < ni; i++)
	    cnt[nv + ned  + i] = first_inner_dof[i+1] - first_inner_dof[i]; 
	  break;
	}
      case 22: // Anisotropic Face-AFW wo ho-grads 
	{
	  int v1, v2; 
	  ii=nv; 
	  for(i=0; i< ned; i++) 
	    if(fine_edge[i])
	      {
		ma.GetEdgePNums (i, v1, v2);
		cnt[v1]++; 
		cnt[v2]++;
		//cnt[nv + i] = first_edge_dof[i+1] - first_edge_dof[i]; 
	      }
	  for(i=0;i<nfa;i++)
	    if(fine_face[i])
	      {
		ARRAY<int> f2ed; 
		ma.GetFaceEdges (i, f2ed);
		for (j = 0; j < f2ed.Size(); j++)
		  cnt[nv+f2ed[j]] += first_face_dof[i+1] - first_face_dof[i] - face_ngrad[i];  
	      }
	  
	  for (i = 0; i < ni; i++)
	    cnt[nv + ned  + i] = first_inner_dof[i+1] - first_inner_dof[i] - cell_ngrad[i];
	  break;
	}
      case 23: 
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ first_edge_dof[i+1] - first_edge_dof[i]; 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i]; 
	     	 
	  for(i=0;i<nel;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      // Stack: plane-face + inner 
	      cnt[ccl] += first_inner_dof[i+1] - first_inner_dof[i]; 
	      
	       
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      
	       ma.GetElFaces(i,fanums); 
	       for(j=0;j<fanums.Size();j++) 
		 { 
		   int fcl = ma.GetClusterRepFace(fanums[j]);
		   if(fcl != ccl) 
		     cnt[fcl] += first_inner_dof[i+1] - first_inner_dof[i];
		  }
	       
	    }
	  
	  for(i=0;i<nfa;i++) // Trig-Faces to horiz-edge stack 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		int nd = first_face_dof[i+1] - first_face_dof[i]; 
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    cnt[ecl] += nd; 
		  }
	      } 
	  
	  break;
	}
      case 24: 
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ first_edge_dof[i+1] - first_edge_dof[i]; 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i]; 
	     	 
	  for(i=0;i<nel;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      // Stack: plane-face + inner 
	      cnt[ccl] += first_inner_dof[i+1] - first_inner_dof[i]; 
	      
	      /* 
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      
	       ma.GetElFaces(i,fanums); 
	       for(j=0;j<fanums.Size();j++) 
		 { 
		   int fcl = ma.GetClusterRepFace(fanums[j]);
		   if(fcl != ccl) 
		     cnt[fcl] += first_inner_dof[i+1] - first_inner_dof[i];
		  }
	      */
	    }
	  
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		int nd = first_face_dof[i+1] - first_face_dof[i]; 
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    cnt[ecl] += nd; 
		  }
	      } 
	  
	  break;
	}
      case 25: 
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ first_edge_dof[i+1] - first_edge_dof[i]; 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i]; 
	     	 
	  for(i=0;i<nel;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      // Stack: plane-face + inner 
	      cnt[ccl] += first_inner_dof[i+1] - first_inner_dof[i]; 
	      
	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      
	       ma.GetElFaces(i,fanums); 
	       for(j=0;j<fanums.Size();j++) 
		 { 
		   int fcl = ma.GetClusterRepFace(fanums[j]);
		   if(fcl != ccl) 
		     cnt[fcl] += first_inner_dof[i+1] - first_inner_dof[i];
		  }
	      
	    }
	  /*
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		int nd = first_face_dof[i+1] - first_face_dof[i]; 
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    cnt[ecl] += nd; 
		  }
	      } 
	  */
	  break;
	}
        case 26: 
	{
	  // Stack-AFW(hoEdges) and Stack of horiz edges 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2); 
		int nde = 1+ first_edge_dof[i+1] - first_edge_dof[i]; 
		cnt[pn1] += nde; 
		if (pn1 != pn2)
		  { 
		    cnt[pn2]+= nde;
		    cnt[ma.GetClusterRepEdge(i)] +=nde;
		  }
	      }
	  
	  // Stack of horizontal edges: + quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      cnt[ma.GetClusterRepFace(i)] += first_face_dof[i+1] - first_face_dof[i]; 
	     	 
	  for(i=0;i<nel;i++) 
	    { 
	      int ccl = ma.GetClusterRepElement(i);
	      // Stack: plane-face + inner 
	      cnt[ccl] += first_inner_dof[i+1] - first_inner_dof[i]; 
	      
	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      /* 
	       ma.GetElFaces(i,fanums); 
	       for(j=0;j<fanums.Size();j++) 
		 { 
		   int fcl = ma.GetClusterRepFace(fanums[j]);
		   if(fcl != ccl) 
		     cnt[fcl] += first_inner_dof[i+1] - first_inner_dof[i];
		  }
	      */
	    }
	  
	  /*  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		int nd = first_face_dof[i+1] - first_face_dof[i]; 
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    cnt[ecl] += nd; 
		  }
	      } 
	  */ 
	  break;
	}
      }
  
    
    //  (*testout) << " nv " << nv << " ned " << ned << "nfa " << nfa << "ni " << ni  << " ndofs " << GetNDof() <<  endl;  
     
    //   (*testout) << "Hcurlho::createsmoothingblocks" << endl;
    //(*testout) << "cnt = " << cnt << endl;      
    
    
    Table<int> & table = *new Table<int> (cnt); 
    ii = 0; 
    cnt = 0;
    switch(SmoothingType) 
      {
      case 7: 
	for(i=0;i<ndof;i++) 
	  table[0][i] = i; 
	break; 
      case 1: // Block(N) + E + I 
	for(i=0; i<ned; i++) table[0][i] = i*fine_edge[i]; //Ned0
	ii++;
	for(i=0; i<ned; i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<ni; i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 	
	break; 
      case 2: // N + E + I 
	for(i=0; i<ned; i++) if(fine_edge[i]) table[i][0] = i; //Ned0
	ii = ned; 
	for(i=0; i<ned;i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<ni;i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 
	break; 
      case 3: // NE + I 
	for(i=0; i<ned; i++)   
	  {
	    int m = 0; 
	    if(fine_edge[i]) table[i][m++] = i;  
	    first = first_edge_dof[i]; 
	    for(j=m; j< cnt[i]; j++)
	      table[i][j] = j+first-1;  
	  } 
	ii = ned; 
	for(i=0; i<ni;i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 
	break; 
      case 4: // E + I  (no Neds) 
	for(i=0; i<ned; i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[i]; j++)
	      table[i][j] = j+first; 
	  } 
	for(i=0; i<ni; i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ned+i]; j++) 
	      table[ned+i][j] =  j+ first; 
	  } 	
	break;
      case 5: // AFW + E + I 
	{
	int v1,v2; 
	for(i=0; i< ned; i++) 
	  if(fine_edge[i])
	    {
	      ma.GetEdgePNums (i, v1, v2);
	      table[v1][--cnt[v1]] = i; 
	      table[v2][--cnt[v2]] = i; 
	    }
	ii = nv; 
	 for(i=0; i<ned; i++,ii++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	      }
	for(i=0; i<ni; i++, ii++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	  } 
	break;  
	}
      case 15: // AFW + E + I  without ho-grads 
	{
	  int v1,v2; 
	for(i=0; i< ned; i++) 
	  if(fine_edge[i])
	    {
	      ma.GetEdgePNums (i, v1, v2);
	      table[v1][--cnt[v1]] = i; 
	      table[v2][--cnt[v2]] = i; 
	    }
	ii = nv; 
	for(i=0; i<ni; i++, ii++) 
	  { 
	    first = first_inner_dof[i] + cell_ngrad[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j + first; 
	  } 
	break; 
	}
      case 6: // E + FI 
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[i][cnt[i]++] = j;

	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[ned+i][cnt[ned+i]++] = j;
	  for (i = 0; i < nel; i++)
	    {
	      if(ma.GetDimension() == 3)  ma.GetElFaces (i, fanums);
	      else fanums.SetSize(0); 
	      for (j = 0; j < fanums.Size(); j++)
		for (k = first_inner_dof[i]; k < first_inner_dof[i+1]; k++)
		  table[ned+fanums[j]][cnt[ned+fanums[j]]++] = k;
	    }
	  break;
	}
      case 7: // Clustering N + E + F + I 
	{
	  
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int nr = ma.GetClusterRepEdge (i);
		for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  /*
	    for(i=0; i<nfa; i++)
	    {	      
	    ma.GetFaceEdges(i,ednums);
	    if(ednums.Size() != 4) continue;
	      
	    int nr_self = ma.GetClusterRepFace(i);

	    ARRAY<int> rep;
	    for(j=0; j<ednums.Size(); j++)
	    {
	    int nr = ma.GetClusterRepEdge(ednums[j]);
	    if(nr != nr_self && !rep.Contains(nr))
	    rep.Append(nr);
	    }
	    for(k=0; k<rep.Size(); k++)
	    for(int l=first_face_dof[i]; l<first_face_dof[i+1]; l++)
	    table[rep[k]][cnt[rep[k]]++] = l;
	    }
	  */
	  
	  /*
	    cnt = 0;
	    for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	    table[i][cnt[i]++] = j;
	    for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	    table[ned+i][cnt[ned+i]++] = j;
	    for (i = 0; i < nel; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	    table[ned+nfa+i][cnt[ned+nfa+i]++] = j;
	  */ 
	  break;
	}
      case 8: // AFW(hoE)  + F + I 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2, cl;
		ma.GetEdgePNums (i,pn1,pn2);	      
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		cl = ma.GetClusterRepEdge(i);

		table[pn1][cnt[pn1]++] = i;
		for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		  table[pn1][cnt[pn1]++] = j;
		if (pn1 != pn2)
		  {
		    table[pn2][cnt[pn2]++] = i;
		    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		      table[pn2][cnt[pn2]++] = j;

		    table[cl][cnt[cl]++] = i;
		    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		      table[cl][cnt[cl]++] = j;
		  }
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 9: // AugV + EI + FI 
	{
	  cnt = 0;
	  for (i = 0; i < nv; i++)
	    table[i][cnt[i]++] = ned+i;
	  for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[nv+i][cnt[nv+i]++] = j;
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[nv+ned+i][cnt[nv+ned+i]++] = j;
	  for (i = 0; i < ni; i++)
	    {
	      ma.GetElEdges (i,ednums,orient);
	      for (k = 0; k < ednums.Size(); k++)
		{
		  int dof = nv+ednums[k];
		  for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		    table[dof][cnt[dof]++] = j;
		}
	      ma.GetElFaces (i,fanums,orient);
	      for (k = 0; k < fanums.Size(); k++)
		{
		  int dof = nv+ned+fanums[k];
		  for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		    table[dof][cnt[dof]++] = j;
		}
	    }
	  break;
	}
      case 10: // Clustering + AFW 
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		table[pn1][cnt[pn1]++] = i;
		if (pn1 != pn2)
		  table[pn2][cnt[pn2]++] = i;
		
		int nr = ma.GetClusterRepEdge (i);
		for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  break;
	}
      case 11: // AFW + E  + F + I 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  int nnv = nv; 
	  
	  if(augmented==1)
	    { 
	      nnv += nv; 
	      for(i=0;i<nv;i++)
		table[i][cnt[i]++] = ned+i;
	    }
	  
	  for (i =0; i<ned ; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[nnv+i][cnt[nnv+i]++] = j;
	  
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[nnv+ned+i][cnt[nnv+ned+i]++] = j;
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[nnv+ned+nfa+i][cnt[nnv+ned+nfa+i]++] = j;
	  break;
	}
      case 12: // AFW  + F + I no ho-grads 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i] + face_ngrad[i]; j < first_face_dof[i+1]; j++)
	      table[nv+i][cnt[nv+i]++] = j;
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i] + cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+nfa+i][cnt[nv+nfa+i]++] = j;
	  break;
	}

      case 13: // Clustering N + E + F + I  without ho-grads 
	{
	  cnt = 0; 
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		for (j = first_face_dof[i] + face_ngrad[i]; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      for (j = first_inner_dof[i] + cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  break;
	}
      case 14: // Clustering + AFW  without ho-grads 
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		table[pn1][cnt[pn1]++] = i;
		if (pn1 != pn2)
		  table[pn2][cnt[pn2]++] = i;
		
	      }
	  for (i = 0; i < nfa; i++)
	    if(fine_face[i])
	      {
		int nr = ma.GetClusterRepFace (i);
		for (j = first_face_dof[i] + face_ngrad[i]; j < first_face_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  for (i = 0; i < ni; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      for (j = first_inner_dof[i] + cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  break;
	}
	
      case 16: // F + I no ho-grads 
	{
	  cnt = 0;
	  	  
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i] + face_ngrad[i]; j < first_face_dof[i+1]; j++)
	      table[i][cnt[i]++] = j;
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i] + cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
	      table[nfa+i][cnt[nfa+i]++] = j;
	  break;
	}
      case 17: // Anisotropic Face-AFW with ho-grads 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  for (i =0; i<ned ; i++)
	    if(fine_edge[i])
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nv+i][cnt[nv+i]++] = j;
	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
	
		/* 
		   if(FaceAnisotropicDir[i] == 0)  // xi - anisotropic dir => eta const. 
		   {
		   first_fafw -= order_face[i][0] + order_face[i][1]; 
		   last_fafw -= order_face[i][1];
		   }
		   else if(FaceAnisotropicDir[i] == 1) // eta-anisotropic dir => xi const. 
		   first_fafw -=order_face[i][1]; 
		*/
		int nfafw =  0; 
		if (AnisotropicFaceEdges[i][0] >= 0) nfafw = (order_face[i][0] + order_face[i][1]);  
		int first_fafw = first_face_dof[i+1] - nfafw; 
		int last_fafw = first_face_dof[i+1];
		
		for(int k=first_face_dof[i];k<first_fafw;k++) 
		  table[nv+2*ned+i][cnt[nv+2*ned+i]++] = k; 
		
	
		  
		// int em = AnisotropicFaceEdges[i][m]; 
		// int e2 = AnisotropicFaceEdges[i][1]; 
		for(int k=first_fafw;k<last_fafw;k++) 
		  for(int m=0;m<4 && AnisotropicFaceEdges[i][m]>=0;m++)
		    table[nv+ned+AnisotropicFaceEdges[i][m]][cnt[nv+ned+AnisotropicFaceEdges[i][m]]++] = k; 
		
		/*for(int k=last_fafw;k<first_face_dof[i+1];k++) 
		  table[nv+2*ned+i][cnt[nv+2*ned+i]++] = k; 
		*/
	      }
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+2*ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 18: // Anisotropic Face-AFW wo ho-grads
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
		int nfafw =  0; 
		if (AnisotropicFaceEdges[i][0] >= 0) nfafw = (order_face[i][0] + order_face[i][1]);  
		int first_fafw = first_face_dof[i+1] - nfafw - face_ngrad[i]; 
		int last_fafw = first_face_dof[i+1] - face_ngrad[i];
		
		for(int k=first_face_dof[i];k<first_fafw;k++) 
		  table[nv+2*ned+i][cnt[nv+2*ned+i]++] = k; 
		for(int k=first_fafw;k<last_fafw;k++) 
		  for(int m=0;m<4;m++)
		    table[nv+ned+AnisotropicFaceEdges[i][m]][cnt[nv+ned+AnisotropicFaceEdges[i][m]]++] = k; 
	      }
		/*




		int first_fafw = first_face_dof[i+1] - face_ngrad[i]; 
		int last_fafw = first_face_dof[i+1] - face_ngrad[i];
		if(FaceAnisotropicDir[i] == 0)  // xi - anisotropic dir => eta const. 
		  {
		    first_fafw -= order_face[i][0] + order_face[i][1]; 
		    last_fafw -= order_face[i][1];
		  }
		else if(FaceAnisotropicDir[i] == 1) // eta-anisotropic dir => xi const. 
		  first_fafw -=order_face[i][1]; 
		
		for(int k=first_face_dof[i];k<first_fafw;k++) 
		  table[nv+2*ned+i][cnt[nv+2*ned+i]++] = k; 
		int e1 = AnisotropicFaceEdges[i][0]; 
		int e2 = AnisotropicFaceEdges[i][1]; 
		for(int k=first_fafw;k<last_fafw;k++) 
		  {
		    table[nv+ned+e1][cnt[nv+ned+e1]++] = k; 
		    table[nv+ned+e2][cnt[nv+ned+e2]++] = k; 
		  }
		for(int k=last_fafw;k<first_face_dof[i+1]-face_ngrad[i];k++) 
		  table[nv+2*ned+i][cnt[nv+2*ned+i]++] = k; 
		  }*/ 

	  
	  
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]-cell_ngrad[i]; j++)
	      table[nv+2*ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 19: // QUAD-FACE-AFW with grads  
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  for (i =0; i<ned ; i++)
	    if(fine_edge[i])
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nv+i][cnt[nv+i]++] = j;

  	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
		ARRAY<int> vnums; 
		ma.GetFacePNums(i,vnums);
		first = first_face_dof[i]; 
		int last = first_face_dof[i+1];
		if(vnums.Size() == 4) // quad_face
		{ 
		  ARRAY<int> f2ed; 
		  ma.GetFaceEdges (i, f2ed);
		  
		  int nfafw = order_face[i][0] + order_face[i][1]; 
		  
		  for (j = 0; j < f2ed.Size(); j++)
		    for(k=last-nfafw; k<last; k++) 
		      table[nv+f2ed[j]][cnt[nv+f2ed[j]]++] = k;
		  last -= nfafw; 
		}
		
		for(k=first;k<last;k++) 
		  table[nv+ned+i][cnt[nv + ned + i]++] = k; 
	      }
	  
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 20: 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
		ARRAY<int> vnums; 
		ma.GetFacePNums(i,vnums);
		int first = first_face_dof[i]+face_ngrad[i];
		int last = first_face_dof[i+1]; 
		
		if(vnums.Size() == 4) // quad_face
		{ 
		  ARRAY<int> f2ed; 
		  ma.GetFaceEdges (i, f2ed);
		  
		  int nfafw = order_face[i][0] + order_face[i][1]; 
		  int fafw_first = first_face_dof[i+1] - nfafw;
		 
		  for (j = 0; j < f2ed.Size(); j++)
		    for(k=fafw_first; k<last; k++) 
		      table[nv+f2ed[j]][cnt[nv+f2ed[j]]++] = k;
		  last -= nfafw;
		}
		
		for(k=first;k<last;k++) 
		  table[nv+ned+i][cnt[nv + ned + i]++] = k; 
	      }
	  
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]+cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 21: 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  for (i =0; i<ned ; i++)
	    if(fine_edge[i])
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nv+i][cnt[nv+i]++] = j;

  	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
		ARRAY<int> f2ed; 
		ma.GetFaceEdges (i, f2ed);
		
				
		for (j = 0; j < f2ed.Size(); j++)
		  for(k=first_face_dof[i];k<first_face_dof[i+1];k++) 
		    table[nv+f2ed[j]][cnt[nv+f2ed[j]]++] = k; 
	      }
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+ned+i][cnt[nv+ned+i]++] = j;
	  break;
	}
      case 22: 
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		table[pn1][cnt[pn1]++] = i;
		table[pn2][cnt[pn2]++] = i;
	      }
	  /*  for (i =0; i<ned ; i++)
	    if(fine_edge[i])
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[nv+i][cnt[nv+i]++] = j;*/ 

  	  for (i=0;i<nfa;i++) 
	    if(fine_face[i])
	      {
		ARRAY<int> f2ed; 
		ma.GetFaceEdges (i, f2ed);
		
				
		for (j = 0; j < f2ed.Size(); j++)
		  for(k=first_face_dof[i]+face_ngrad[i];k<first_face_dof[i+1];k++) 
		    table[nv+f2ed[j]][cnt[nv+f2ed[j]]++] = k; 
	      }
	  for (i = 0; i < ni; i++)
	    for (j = first_inner_dof[i]+cell_ngrad[i]; j < first_inner_dof[i+1]; j++)
	      table[nv+ned+i][cnt[nv+ned+i]++] = j;
	  break;
	}
      case 23: 
	{
	  cnt = 0; 
	  
	  // Stack-AFW(hoEdges) 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		  table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		      {		    
			table[pn2][cnt[pn2]++] = j;
			table[ecl][cnt[ecl]++] = j;
		      }
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		//	if(fcl  >= nv + ned) // quad-faces (prism) 
		for(j=first_face_dof[i]; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 
	      }
	  
	  for(i=0;i<nel;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i); 
	      for(j=first_inner_dof[i];j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	         	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      ma.GetElFaces(i,fanums); 
	      
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl)
		    for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		      table[fcl][cnt[fcl]++] = k; 
		} 
	     
	    }
	  
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		 if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    for(k=first_face_dof[i]; k<first_face_dof[i+1]; k++)
		      table[ecl][cnt[ecl]++] = k; 
		  }
	      }
	  
	  break;
	}
      case 24: 
	{
	  cnt = 0; 
	  
	  // Stack-AFW(hoEdges) 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		  table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		      {		    
			table[pn2][cnt[pn2]++] = j;
			table[ecl][cnt[ecl]++] = j;
		      }
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		//	if(fcl  >= nv + ned) // quad-faces (prism) 
		for(j=first_face_dof[i]; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 
	      }
	  
	  for(i=0;i<nel;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i); 
	      for(j=first_inner_dof[i];j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	         	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      /*  ma.GetElFaces(i,fanums); 
	      
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl)
		    for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		      table[fcl][cnt[fcl]++] = k; 
		} 
	      */
	    }
	  
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		 if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    for(k=first_face_dof[i]; k<first_face_dof[i+1]; k++)
		      table[ecl][cnt[ecl]++] = k; 
		  }
	      }
	  
	  break;
	}
      case 25: 
	{
	  cnt = 0; 
	  
	  // Stack-AFW(hoEdges) 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		  table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		      {		    
			table[pn2][cnt[pn2]++] = j;
			table[ecl][cnt[ecl]++] = j;
		      }
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		//	if(fcl  >= nv + ned) // quad-faces (prism) 
		for(j=first_face_dof[i]; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 


	      }
	  
	  for(i=0;i<nel;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i); 
	      for(j=first_inner_dof[i];j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	      ma.GetElFaces(i,fanums); 
	      
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl)
		    for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		      table[fcl][cnt[fcl]++] = k; 
		} 
	         	      
	    }
	  
	  /*  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		 if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    for(k=first_face_dof[i]; k<first_face_dof[i+1]; k++)
		      table[ecl][cnt[ecl]++] = k; 
		  }
		  }
	  */
	  
	  break;
	}
      case 26: 
	{
	  cnt = 0; 
	  
	  // Stack-AFW(hoEdges) 
	  for(i=0;i<ned;i++) 
	    if(fine_edge[i])
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);
		pn1 = ma.GetClusterRepVertex(pn1);
		pn2 = ma.GetClusterRepVertex(pn2);
		int ecl = ma.GetClusterRepEdge(i);
		table[pn1][cnt[pn1]++] = i;
		for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		  table[pn1][cnt[pn1]++] = j;
		if(pn1!=pn2)
		  {
		    table[pn2][cnt[pn2]++] = i; 
		    table[ecl][cnt[ecl]++] = i; 
		    // ho-edges
		    for(j=first_edge_dof[i]; j<first_edge_dof[i+1]; j++)
		      {		    
			table[pn2][cnt[pn2]++] = j;
			table[ecl][cnt[ecl]++] = j;
		      }
		  }
	      }
	  
	  // Stack of horizontal edges, quadfaces and prism faces 
	  //  (+ inner) 
	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		int fcl = ma.GetClusterRepFace(i); 
		//	if(fcl  >= nv + ned) // quad-faces (prism) 
		for(j=first_face_dof[i]; j< first_face_dof[i+1]; j++) 
		  table[fcl][cnt[fcl]++] = j; 
	      }
	  
	  for(i=0;i<nel;i++) 
	    { 
	      // Stack: plane-face + inner
	      int ccl = ma.GetClusterRepElement(i); 
	      for(j=first_inner_dof[i];j<first_inner_dof[i+1];j++) 
		table[ccl][cnt[ccl]++] = j; 
	         	      
	      // inner to quad-face (horizontal edge stack) 
	      // each face different stack 
	      
	      ma.GetElFaces(i,fanums); 
	      /* 
	      for(j=0;j<fanums.Size();j++) 
		{ 
		  int fcl = ma.GetClusterRepFace(fanums[j]);
		  if(fcl != ccl)
		    for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		      table[fcl][cnt[fcl]++] = k; 
		} 
	      */
	    }
	  
	  /*	  for(i=0;i<nfa;i++) 
	    if(fine_face[i]) 
	      { 
		ma.GetFacePNums(i,vnums); 
		 if(vnums.Size()==4) continue; 
		int fcl = ma.GetClusterRepFace(i); 
		ma.GetFaceEdges(i,ednums); 
		
		for(j=0;j<ednums.Size();j++)
		  {
		    int ecl = ma.GetClusterRepEdge(ednums[j]); 
		    if(ecl==fcl) continue; 
		    for(k=first_face_dof[i]; k<first_face_dof[i+1]; k++)
		      table[ecl][cnt[ecl]++] = k; 
		  }
	      }
	  */
	  break;
	}
      }
    //(*testout) << "H(Curl)-table = " << table << endl;	
	
     
    return & table;
   #endif

  }

  ARRAY<int> * 
  HCurlHighOrderFESpaceA :: CreateDirectSolverClusters (const Flags & precflags) const
    {

      int clustertype = int(precflags.GetNumFlag("ds_cluster",4)); 
      cout << " DirectSolverCluster Clustertype " << clustertype << endl; 
      if(clustertype==0)
	return(0);
	   
      int nv = ma.GetNV();
      int nd = GetNDof();
      int ne = ma.GetNE();
      int ned = ma.GetNEdges();

      ARRAY<int> ednums, fnums, pnums;

    int i, j, k;
    bool hasprism = false;

    for (i = 0; !hasprism && i < ne; i++)
      if (ma.GetElType(i) == ET_PRISM)
	hasprism = true;
    
    if (!hasprism && adddirectsolverdofs.Size() == 0) 
      return 0;
        
    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());

    if(hasprism)
      {
	// for long prisms, JS June 2005
	switch(clustertype)
	  {

	  case 0: 
	    clusters = 0; 
	    break; 
	  case 1:
	    clusters = 0;
	    for (i = 0; i < ne; i++)
	      {
		if (ma.GetElType(i) == ET_PRISM)
		  {
		    ma.GetElEdges (i, ednums);
		
		    for (j = 6; j < 9; j++)  //vertical Edges 
		      { 
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 1;      //verthoedge
			clusters[ednums[j]] = 1;
		      }

		ma.GetElFaces (i, fnums); 

		    for (j=2; j<5 ; j++)  // vertical faces
		      {
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			for(k=first; k<next; k++)
			  clusters[k] = 1;
		    
			// INT<2> p = order_face[fnums[j]]; 
			// for(k=next-p[0]-p[1]; k<next; k++)
			// clusters[k] = 1;
		      }

		    int first = first_inner_dof[i]; 
		    int next = first_inner_dof[i+1]; 

		for(k=first; k<next; k++)
		  clusters[k] = 1;
		  }
	      }
	    break; 
	  case 2: 
	    clusters = 0;
	
	    // All Vertical Edges in one Cluster for Hex and Prism + const_x*poly faces + const_y*polx faces (-> 2d Problems !) 
	
	
	    //lo 
	    for(i=0;i<ma.GetNEdges();i++)
	      clusters[i]=1; 

       
	    for (i = 0; i < ne; i++)
	      {
		// 	if (ma.GetElType(i) == ET_PYRAMID)
		// 	  {
		// 	    GetDofNrs(i,ednums);
		// 	    for(j=0; j<ednums.Size(); j++) clusters[ednums[j]] = 3;
	    
		// 	  }   
	    
		if (ma.GetElType(i) == ET_PRISM)
		  {
		    ma.GetElEdges (i, ednums);
		    for (j = 0; j < 6; j++)  //horizontal Edges 
		      { 
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 1; 
			clusters[ednums[j]]=1;
		      }
		
		
		    for (j = 6; j < 9; j++)  //vertical Edges 
		      { 
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 1;      //verthoedge
			clusters[ednums[j]]=1; //ned
		      }
		    ma.GetElFaces (i, fnums); 
		
		    for (j=2; j<5 ; j++)  // vertical faces
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			// TEST SZ eigentlich nurin eine richtung konst ausreichend
			for (k=first; k < next; k++) 
			  clusters[k]=1; 
		    
		    
			INT<2> p = order_face[fnums[j]]; 
		    
			for(k=next-p[0]-p[1]; k<next; k++)
			  clusters[k] = 1;
			for(k=0; k<4; k++)
			  clusters[k] = 1;
		    
		    
			// for (k=first; k < next; k++) 
			//   clusters[k]=3; 
		    
		      }
		    for (j=0; j<2 ; j++)  // horizontal faces
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			for (k=first; k < next; k++) 
			  clusters[k]=1; 
		      }
		  }
	    
		else if (ma.GetElType(i) == ET_HEX)
		  {
		    ma.GetElEdges (i, ednums);
		
		    for(j=0;j<8;j++) // horizontal edges
		      {	
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 0;
		
			clusters[ednums[j]]=0; 
		    
		      }
			
		    for (j = 8; j < 12; j++)  // vertical edges 
		      {	
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 3;
			clusters[ednums[j]]=0; 
		      }
		
		    ma.GetElFaces(i,fnums); // vertical faces 
		    for(j=2;j<6;j++) 
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			/*	for (k=first; k < next; k++) 
			  clusters[k]=0; 
		      
			  INT<2> p = order_face[fnums[j]];
			  for(k=2*(p[0]+1)*(p[0]+1);k<next;k++) //achtung
			  clusters[k]=3;  
			*/ 
			// TEST SZ  eigentlich in eine richtung konstante nur benoetigt
			for (k=first; k < next; k++) 
			  clusters[k]=3; 
		    
		      }
		    for(j=0;j<2;j++)  //horizontal faces 
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			for (k=first; k < next; k++) 
			  clusters[k]=0; 
		      } 
		  }
	    
		for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		  clusters[k]=0; 
	    
	      }
    
	    break; 
	  case 3: 
	    //lo 
	    for(i=0;i<ma.GetNEdges();i++)
	      clusters[i]=0; 
	
	    for (i = 0; i < ne; i++)
	      {
		ma.GetElPNums(i,pnums); 
		if (ma.GetElType(i) == ET_PRISM)
		  {
		    ma.GetElEdges (i, ednums);
		    for (j = 0; j < 6; j++)  //horizontal Edges 
		      { 
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 0;
			clusters[ednums[j]]=0; 
		      }
		    for (j = 6; j < 9; j++)  //vertical Edges 
		      { 
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 3;      //verthoedge
			clusters[ednums[j]]=0; //ned
		      }
		    ma.GetElFaces (i, fnums); 
		
		    for (j=2; j<5 ; j++)  // vertical faces
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			/* TEST SZ eigentlich nurin eine richtung konst ausreichend
			   for (k=first; k < next; k++) 
			   clusters[k]=0; 
		       
			   int p = order_face[fnums[j]][0]; // achtung !!! 
			   for(k=2*(p+1)*(p+1);k<next;k++)
			   clusters[k]=3;  
		       
			*/
		    
			for (k=first; k < next; k++) 
			  clusters[k]=3; 
		    
		      }
		    for (j=0; j<2 ; j++)  // horizontal faces
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			for (k=first; k < next; k++) 
			  clusters[k]=0; 
		      }
		  }
	    
		else if (ma.GetElType(i) == ET_HEX)
		  {
		    ma.GetElEdges (i, ednums);
		    for(j=0;j<8;j++) // horizontal edges
		      {	
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 0;
		    
			clusters[ednums[j]]=0; 
		    
		      }
		    for (j = 8; j < 12; j++)  // vertical edges 
		      {	
			int first = first_edge_dof[ednums[j]];
			int next = first_edge_dof[ednums[j]+1];
			for (k = first; k < next; k++)
			  clusters[k] = 3;
			clusters[ednums[j]]=0; 
		      }
		
		    ma.GetElFaces(i,fnums); // vertical faces 
		    for(j=2;j<6;j++) 
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			/*	for (k=first; k < next; k++) 
			  clusters[k]=0; 
		      
			  INT<2> p = order_face[fnums[j]];
			  for(k=2*(p[0]+1)*(p[0]+1);k<next;k++) //achtung
			  clusters[k]=3;  
			*/ 
			// TEST SZ  eigentlich in eine richtung konstante nur benoetigt
			for (k=first; k < next; k++) 
			  clusters[k]=3; 
		    
		      }
		    for(j=0;j<2;j++)  //horizontal faces 
		      {
		    
			int first = first_face_dof[fnums[j]]; 
			int next = first_face_dof[fnums[j]+1]; 
		    
			for (k=first; k < next; k++) 
			  clusters[k]=0; 
		      } 
		  }
	    
		for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
		  clusters[k]=0; 
	    
	      }
    
	    //  (*testout) << "direct clusters = " << endl << clusters << endl;
	
	    for(i=0; directsolverclustered.Size() > 0 && i<ne; i++)
	      {
		if(directsolverclustered[ma.GetElIndex(i)])
		  {
		    GetDofNrs(i,ednums);
		    for(k=0; k<ednums.Size(); k++)
		      {
			clusters[ednums[k]] = 4;
		      }
		  }
	      }
	    break; 

	  case 4:  // just like the old hcurl 
	    //lo 
	    clusters = 0; 
	    for(i=0;i<ned;i++) 
	      { 
		int pi1,pi2; 
		ma.GetEdgePNums(i,pi1,pi2); 
		pi1 = ma.GetClusterRepVertex (pi1); 
		pi2 = ma.GetClusterRepVertex (pi2);
		if(pi1 == pi2) 
		  // stack of horizontal edge-blocks (without low-order)
		  // -> decoupled 
		  for(j=first_edge_dof[i];j <first_edge_dof[i+1];j++) 
		    clusters[j] = 1; 
	      }

	    for(i=0;i<nfa;i++)
	      {
		// Attenzione das is zuviel !!! 

		ma.GetFacePNums(i,pnums); 
		if(pnums.Size() == 4) // quad face 
		  { 
		    INT<2> p = order_face[i];         
		    int first =first_face_dof[i];  
		    int next = first_face_dof[i+1]; 
		    // Ned_0*pol_z 
		    // int first = next - p[0] - p[1]; 
	
		    for(j= first ; j<next; j++)
		      clusters[j] = 1; 
		  } 
	  
	    
	    
	      }
	    break; 
	  case 5:  // just like the old hcurl horizontal only constant ... 
	    //lo 
	    clusters = 0; 
	    for(i=0;i<ned;i++) 
	      { 
	    
		int pi1,pi2; 
		ma.GetEdgePNums(i,pi1,pi2); 
		pi1 = ma.GetClusterRepVertex (pi1); 
		pi2 = ma.GetClusterRepVertex (pi2);
		if(pi1 == pi2) 
		  // stack of horizontal edge-blocks (without low-order)
		  // -> decoupled 
		  {
		    //clusters[i] = 1; 
		    for(j=first_edge_dof[i];j <first_edge_dof[i+1];j++) 
		      clusters[j] = 1; 
		  }
	      }

	    for(i=0;i<nfa;i++)
	      {
		// Attenzione das is zuviel !!! 

		ma.GetFacePNums(i,pnums); 
		if(pnums.Size() == 4) // quad face 
		  { 
		    INT<2> p = order_face[i];        
		    // int first =first_face_dof[i];  
		    int next = first_face_dof[i+1]; 
		    // Ned_0*pol_z 
		    int first = next - p[0] - p[1]; 
	
		    for(j= first ; j<next; j++)
		      clusters[j] = 1; 
		  } 
	  
	    
	    
	      }


	    //  (*testout) << "direct clusters = " << endl << clusters << endl;
	
	    for(i=0; directsolverclustered.Size() > 0 && i<ne; i++)
	      {
		if(directsolverclustered[ma.GetElIndex(i)])
		  {
		    ELEMENT_TYPE eltype = ma.GetElType(i); 
		    if(eltype != ET_PRISM) continue; 
		
		    GetDofNrs(i,ednums);
		    for(k=0; k<ednums.Size(); k++)
		      if(ednums[k]>=0) clusters[ednums[k]] = 2;
		  }
	      }
	    break; 
      
      
	  }

      }

    if(adddirectsolverdofs.Size())
      (*testout) << "addirectsolverdofs " << adddirectsolverdofs << endl;

    for(i=0; i< adddirectsolverdofs.Size(); i++)
      {
	clusters[adddirectsolverdofs[i]] = 5;
      }

    

    //(*testout) << "clusters = " << clusters << endl;
    return &clusters;
    
  }
  


    
  
  /*  BitArray * HCurlHighOrderFESpaceA :: 
      CreateIntermediatePlanes (int type) const
      {
      int i;
      ARRAY<int> vnums;
 
      //bool has_cluster = 0;
      for (i = 0; i < ned; i++)
      {
      int pi1, pi2;
      ma.GetEdgePNums (i, pi1, pi2);
	
      if (ma.GetClusterRepVertex (pi1) ==
      ma.GetClusterRepVertex (pi2))
      has_cluster = 1;
      }

      if (!has_cluster) return 0;
    
   
      BitArray & ba = *new BitArray (GetNDof());
      ba.Clear();

      for (i = 0; i < ned; i++)
      {
      int pi1, pi2;
      ma.GetEdgePNums (i, pi1, pi2);
	
      if (ma.GetClusterRepVertex (pi1) ==
      ma.GetClusterRepVertex (pi2))
      {
      ba.Set (i);
      for (int l = first_edge_dof[i];
      l < first_edge_dof[i+1]; l++)
      ba.Set (l);
      }
      }
    
    
      for (i = 0; i < nfa; i++)
      {
      ma.GetFacePNums (i, vnums);
      if (vnums.Size() == 4)
      {
      for (int l = first_face_dof[i];
      l < first_face_dof[i+1]; l++)
      ba.Set (l);
      }	    
      }
      return &ba;
      }*/


  SparseMatrix<double> * 
  HCurlHighOrderFESpaceA :: CreateGradient() const
  {
    *testout << " CreateGradient " << endl; 

    Flags flags2 (flags);
    flags2.SetFlag("order", order+1);
    flags2.SetFlag("relorder", rel_order+1); 
    flags2.SetFlag("orderinner",uniform_order_inner+1); 
    flags2.SetFlag("orderface",uniform_order_face+1); 
    flags2.SetFlag("orderedge",uniform_order_edge+1); 
    flags2.SetFlag ("orderquad", -1);
    flags2.SetFlag ("ordertrig", -1);  
    flags2.SetFlag("variableorder",var_order); 
    
    int ndom = ma.GetNDomains();
    ARRAY<double> dom_order_h;
    dom_order_h.SetSize(ndom); 
    
    int i; 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][0]+1; 
    flags2.SetFlag("dom_order_min_x", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][1]+1; 
    flags2.SetFlag("dom_order_min_y", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][2]+1; 
    flags2.SetFlag("dom_order_min_z", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][0]+1; 
    flags2.SetFlag("dom_order_max_x", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][1]+1; 
    flags2.SetFlag("dom_order_max_y", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][2]+1; 
    flags2.SetFlag("dom_order_max_z", dom_order_h); 
    
    H1HighOrderFESpace  fesh1(ma, flags2); 
    fesh1.Update();
     
    int ned = ma.GetNEdges(); 
    int nv  = ma.GetNV(); 
    int nel = ma.GetNE(); 
    int nfa = 0;
    if(ma.GetDimension()==3) nfa= ma.GetNFaces();  

    int dim     = fesh1.GetDimension();
    int dimcurl = GetDimension();

    ARRAY<int> dnums_h1l; 
    ARRAY<int> dnums_hcl;
    LocalHeap lh (1000000);
    
    // Matrix Graph for gradient matrix , start (end) position for assembling 
    ARRAY<int> cnts(ndof);
    cnts = 0; 

    //*testout << "grad fine edges " << fine_edge << endl ; 
    //*testout << "grad fine faces " << fine_edge << endl ; 
   
    for(int i=0; i<ned; i++)
      {
	if(fine_edge[i])
	  {
	    cnts[i] = 2;  // vertices-nedelec
	    int l = first_edge_dof[i]; 
	    for( int k = fesh1.GetFirstEdgeDof(i); k< fesh1.GetFirstEdgeDof(i+1); 
		 k++, l++)
	      cnts[l] = 1;
	  }
      }
    
    for(int i=0; i< nfa; i++) 
      {
	if(fine_face[i])
	  {
	    int l= first_face_dof[i]; 
	    for (int k = fesh1.GetFirstFaceDof(i); k<fesh1.GetFirstFaceDof(i+1); 
		 k++, l++) 
	      cnts[l] = 1;
	  }
      }
    
    for(int i=0; i<nel; i++)
      {
	int l= first_inner_dof[i]; 
	for (int k = fesh1.GetFirstElementDof(i); k<fesh1.GetFirstElementDof(i+1); 
	     k++, l++) 
	  cnts[l] = 1; 
      }
  
    //sparse matrix with above matrix graph cnts 
    SparseMatrix<double> & grad = *new SparseMatrix<double>(cnts);  
    
    // vertices - nedelec
    // ho-edges
    for (int i = 0; i < ned; i++) 
      {
	if(fine_edge[i])
	  { 
	    int p1,p2; 
	    ma.GetEdgePNums(i,p1,p2); 
	    grad.CreatePosition(i,p1); 
	    grad.CreatePosition(i,p2); 
	    if (p1 < p2) 
	      {	      
		grad(i,p1) = -1.;
		grad(i,p2) = 1.; 
	      }
	    else
	      {
		grad(i,p1) = 1.; 
		grad(i,p2) = -1.; 
	      }
	    
	    int l = first_edge_dof[i]; 
	    for( int k = fesh1.GetFirstEdgeDof(i); k< fesh1.GetFirstEdgeDof(i+1); 
		 k++, l++)
	      {
		grad.CreatePosition(l,k); 
		grad(l,k)=1.; 
	      }
	  }
      }
    
    for(int i=0; i< nfa; i++) 
      {
	if(fine_face[i])
	  {
	    int l= first_face_dof[i]; 
	    for (int k = fesh1.GetFirstFaceDof(i); k<fesh1.GetFirstFaceDof(i+1); 
		 k++, l++) 
	      {
		grad.CreatePosition(l,k); 
		grad(l,k)=1.;
	      }
	  }
      }
    
    for(int i=0; i<nel; i++)
      {
	int l= first_inner_dof[i]; 
	for (int k = fesh1.GetFirstElementDof(i); k<fesh1.GetFirstElementDof(i+1); 
	     k++, l++) 
	  {
	    grad.CreatePosition(l,k); 
	    grad(l,k)=1.;
	  }	
      }
    
    (*testout) << " Global grad " << grad << endl; 
    return &grad;
  }

#ifdef PARALLEL

void HCurlHighOrderFESpaceA :: UpdateParallelDofs_hoproc()
  {
     // ******************************
     // update exchange dofs 
     // ******************************
    *testout << "HCurlHOA::UpdateParallelDofs_hoproc" << endl;
    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks);
    nexdof = 0;

    MPI_Status status;
    MPI_Request * sendrequest = new MPI_Request [ntasks];
    MPI_Request * recvrequest = new MPI_Request [ntasks];

    // number of edge exchange dofs
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	if ( !parallelma->IsExchangeEdge ( edge ) ) continue;
	
	ARRAY<int> dnums;
	GetEdgeDofNrs ( edge, dnums );
	nexdof[id] += dnums.Size() ;  
	
	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantEdgeNum ( dest, edge ) >= 0 )
	    nexdof[dest] += dnums.Size() ;  
      }

    // + number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	if ( !parallelma->IsExchangeFace ( face ) ) continue;
	
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	nexdof[id] += dnums.Size() ; 

	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    nexdof[dest] += dnums.Size() ; 
      }

    // + number of inner exchange dofs --> set eliminate internal

    nexdof[0] = LowOrderFESpace() . GetNDof();

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

    ARRAY<int> ** owndofs, ** distantdofs;
    owndofs = new ARRAY<int>* [ntasks];
    distantdofs = new ARRAY<int>* [ntasks];

    for ( int i = 0; i < ntasks; i++ )
      {
	owndofs[i] = new ARRAY<int>(1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int>(0);
      }

    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;
    int exdof = 0;
    int ii;
    // *****************
    // Parallel Edge dofs
    // *****************


   for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      if ( parallelma->IsExchangeEdge ( edge ) )
      {
	ARRAY<int> dnums;
	GetEdgeDofNrs ( edge, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distedge = parallelma -> GetDistantEdgeNum ( dest, edge );
	    if( distedge < 0 ) continue;

	    owndofs[dest]->Append ( distedge );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int sender = 1; sender < ntasks; sender ++ )
      {
        if ( id == sender )
          for ( int dest = 1; dest < ntasks; dest ++ )
            if ( dest != id )
              {
		MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
              }
	  
        if ( id != sender )
          {
	    MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender]);
          }
 	  
	  
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest + dest, &status );
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
	// low order nedelec dofs first
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int ednum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetEdgeDofNrs (ednum, dnums);
	    (*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[0];
	    (*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
	    if ( dest < id && !isdistghost )
	      paralleldofs->ismasterdof.Clear ( dnums[0] ) ;
	    ii += dnums.Size();
	    cnt_nexdof[dest]++;
	  }
	// then the high order dofs, without nedelecs
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int ednum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetEdgeDofNrs (ednum, dnums);
	    ii++; 
	    for ( int i=1; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }

    // *****************
    // Parallel Face dofs
    // *****************

    for ( int dest = 0; dest < ntasks; dest++)
      {
	owndofs[dest]->SetSize(1);
	distantdofs[dest]->SetSize(0);
      }
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      if ( parallelma->IsExchangeFace ( face ) )
      {
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;

	    owndofs[dest]->Append ( distface );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );

	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int sender = 1; sender < ntasks; sender ++ )
      {
        if ( id == sender )
          for ( int dest = 1; dest < ntasks; dest ++ )
            if ( dest != id )
              {
		MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
              }
	  
        if ( id != sender )
          {
	    MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender] );
          }
 	  
	  
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest + dest, &status );
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }

     /*******************************

         update low order space

     *****************************/

    int ndof_lo = low_order_space->GetNDof();

    // all dofs are exchange dofs
    nexdof = ndof_lo;
 
    exdof = 0;
    cnt_nexdof = 0;


    // *****************
    // Parallel Edge dofs
    // *****************

    owndofs[0]->SetSize(1);
    (*owndofs[0])[0] = ndof;
    distantdofs[0]->SetSize(0);
    
     // find local and distant dof numbers for vertex exchange dofs
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	int dest = 0;
	
	int distedge = parallelma -> GetDistantEdgeNum ( dest, edge );
	owndofs[0]->Append ( distedge );
	paralleldofs->SetExchangeDof ( dest, edge );
	owndofs[0]->Append ( edge );
      }   
    
    int dest = 0;
    MyMPI_ISend ( *owndofs[0], dest, sendrequest[dest]);
    MyMPI_IRecv ( *distantdofs[0], dest, recvrequest[dest]);
   
    MPI_Wait ( recvrequest+dest, &status );

    paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
    ii = 1;
    while ( ii < distantdofs[0]->Size() )
      {
	int ednum = (*distantdofs[0])[ii++];
	(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = ednum;
	(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[0])[ii];
	ii++; cnt_nexdof[dest]++;
      }

    for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;

  }

  void HCurlHighOrderFESpaceA :: UpdateParallelDofs_loproc()
  {
    *testout << "HCurlHoA::UpdateParallelDofs_loproc" << endl;
    // lo-proc should never use high order space
    return;
 
  }
#endif // PARALLEL


  // register FESpaces
  namespace
#ifdef MACOS
  hcurlhofespace_a_cpp
#endif
  {
     
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hcurlho_a", HCurlHighOrderFESpaceA::Create);
    }
    
    Init init;

  }
  

}
