

#include <fem.hpp>


namespace ngfem
{  
  using namespace ngfem;

#include "hdivhofe_a.hpp" 

  
  // hv.DValue() = (grad u) x (grad v) 
  extern AutoDiff<3> GradCrossGrad (const AutoDiff<3> & u,
				    const AutoDiff<3> & v);

  extern Vec<3> GradCrossVec (const AutoDiff<3> & u,
                              const Vec<3> & v);
  
  
  // 2d (gradu) x (grad v)
  extern double  GradCrossGrad2 (const AutoDiff<2> & u, const AutoDiff<2> & v) ;
        

  //------------------------------------------------------------------------
  // HDivHighOrderTrig
  //------------------------------------------------------------------------

  template <class T_ORTHOPOL>
  HDivHighOrderTrigA<T_ORTHOPOL> :: HDivHighOrderTrigA (int aorder)
    : HDivHighOrderFiniteElement<2>(ET_TRIG)
  {

    order_inner = INT<3>(aorder,0,0);
    for (int i = 0; i < 3; i++)
      order_edge[i] = aorder;

    ComputeNDof();
  }

  template <class T_ORTHOPOL>
  void HDivHighOrderTrigA<T_ORTHOPOL> :: ComputeNDof()
  {
    ndof_edge =0;
    ndof = 3; // Thomas-Raviart 
    // if oder_edge < 1 --> Thomas Raviart
    int i;

    for (i = 0; i < 3; i++)
      {  
	ndof += order_edge[i];
	ndof_edge += order_edge[i];
      }
    
    if (order_inner[0] > 1)
      { 
        ndof += order_inner[0]*order_inner[0]-1;
	//ndof += order_inner_curl*(order_inner_curl-1)/2;
	//ndof += order_inner*(order_inner-1)/2 + order_inner-1;
      }

    order = 0; // max(order_edges_normal,order_inner);
    for (i = 0; i < 3; i++)
      {
	if (order_edge[i] > order)
	  order = order_edge[i];
      }

    if (order_inner[0] > order) 
      order = order_inner[0];

    order++;

  }

  template <class T_ORTHOPOL>
  void HDivHighOrderTrigA<T_ORTHOPOL> ::
  GetInternalDofs (ARRAY<int> & idofs) const
  {
    if (discontinuous)
    {
      idofs.SetSize(0);
      for (int i=0; i<ndof; i++)
        idofs.Append(i);
      return ;
    }
    else
    {
      idofs.SetSize (0);

      int base = 3;
      for (int i = 0; i < 3; i++)
      {
        base += order_edge[i];
      }


      if(order_inner[0] > 1)
      {
        int p = order_inner[0];
        int ni = p*p -1 ;
        for (int i = 0; i < ni; i++)
          idofs.Append (base+i);
      }
    //(*testout) << "idofs = " << idofs << endl;
    }
  }
  
  template <class T_ORTHOPOL>
  void HDivHighOrderTrigA<T_ORTHOPOL> :: GetFacetDofs(int fa, ARRAY<int> & dnums) const 
  {
    if (fa > 2 ) 
    {
      cout << " Warning HDIVHighOrderTrigSZ::GetFacetDofNrs() index out of range" << endl; 
      
      dnums.SetSize(0); 
      return; 
    } 
  
    dnums.SetSize(0); 
    dnums.Append(fa); 
    
    int ii = 3; // Thomas-Raviart 
    for (int i = 0; i < fa; i++)
      ii += order_edge[i];

    for(int i = 0; i < order_edge[fa]; i++)  
      dnums.Append(ii+i); 
  }                  

  

  template <class T_ORTHOPOL>
  void HDivHighOrderTrigA<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip,
						   FlatMatrixFixWidth<2> shape) const
  {
    double x = ip(0);
    double y = ip(1);

    if ( discontinuous )
      {
	shape = 0.0;
	int ii = 0;
	// no orientation necessary
	// use L2 shape functions, they are L2 orthogonal
	int n=order_inner[0];
	ArrayMem<double, 20> polx(n+1), poly(n+1);


	// Polynomials orthogonal w.r.t L2 inner product
	ScaledLegendrePolynomial (n, 2*x+y-1, 1-y, polx);
	
	for (int i = 0; i <= order_inner[0]; i++)
	  {
	    JacobiPolynomial (n, 2*y-1, 2*i+1, 0, poly);
	    for (int j = 0; j <= order_inner[0]-i; j++)
	      for ( int m = 0; m < 2; m++ )
		shape(ii++, m) = polx[i] * poly[j];
	  }
	return;
      }


    double lami[3];

    lami[0] = x;
    lami[1] = y;
    lami[2] = 1-x-y;



    Mat<3,2> dlami(0);
    dlami(0,0) = 1.;
    dlami(1,1) = 1.;
    dlami(2,0) = -1.;
    dlami(2,1) = -1.;

    // Curl(lami_i) 
    Mat<3,2> clami(0); 
    clami(0,1) = -1.;
    clami(1,0) = 1.;
    clami(2,1) = 1.;
    clami(2,0) = -1.;

    shape = 0.0;

    int p, p_curl;

    int i, j, k, l;
    int ii = 3;




    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);

    int is, ie, io;
    for (i = 0; i < 3; i++)
      {
	p = order_edge[i];
	is = edges[i][0];
	ie = edges[i][1];
	io = 3-is-ie;

	if (vnums[is] > vnums[ie])
	  swap (is, ie);

	//RT low order edge shape function
	for(j=0; j<2; j++)
	  shape(i,j) = lami[ie]*clami(is,j) - lami[is]*clami(ie,j);
	
	//Edge normal shapes
	if(p>0)
	  {
	    MatrixFixWidth<2> DExt(p+2);

	    T_ORTHOPOL::CalcTrigExtDeriv(p+1, lami[ie]-lami[is], lami[io], DExt);

	    for(j=0; j<= p-1;j++)
	      {
		shape(ii,0) = (clami(ie,0) - clami(is,0)) * DExt(j,0) + clami(io,0) * DExt(j,1);
		shape(ii++,1) = (clami(ie,1) - clami(is,1)) * DExt(j,0) + clami(io,1) * DExt(j,1);

	      }
	  }

      }

    if(order_inner[0] < 2) return;

    p = order_inner[0];
    p_curl = order_inner[0];

    //Inner shapes
    if(p>1)
      {

	ArrayMem<double,10> rec_pol(order-1), drec_polx(order-1),  drec_polt(order-1);
	ArrayMem<double,10> rec_pol2(order-1), drec_pol2x(order-1), drec_pol2t(order-1);
	ArrayMem<double,30> mem(3*order*sizeof(double));
	ArrayMem<double,30> mem2(3*order*sizeof(double));
	FlatMatrixFixWidth<3> curl_recpol(order, &mem[0]);
	FlatMatrixFixWidth<3> curl_recpol2(order, &mem2[0]);

	int fav[3];
	for(i=0;i<3;i++) fav[i] = i;

	//Sort vertices  first edge op minimal vertex
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);


	double ls = lami[fav[0]];
	double le = lami[fav[1]];
      double lo = lami[fav[2]];

	ScaledLegendrePolynomialandDiff(p_curl-2, le-ls, 1-lo,
					rec_pol, drec_polx, drec_polt);
	LegendrePolynomialandDiff(p_curl-2, 2*lo-1,rec_pol2, drec_pol2x);
	// \curl (ls * le * rec_pol), and \curl (lo rec_pol2)
	for (j = 0; j <= p_curl-2; j++)
	  {
	    for (l = 0; l < 2; l++)
	      {
		curl_recpol(j,l) = ls * le * (drec_polx[j] * (clami(fav[1],l) - clami(fav[0],l)) -
					      drec_polt[j] * (clami(fav[2],l))) +
		  (ls * clami(fav[1],l) + le * clami(fav[0],l)) * rec_pol[j];

		curl_recpol2(j,l) = lo * (drec_pol2x[j] * 2*clami(fav[2],l)) +
		  clami(fav[2],l) * rec_pol2[j];
	      }
	  }

	// curls:
	for (j = 0; j <= p_curl-2; j++)
	  for (k = 0; k <= p_curl-2-j; k++, ii++)
	    for (l = 0; l < 2; l++)
	      shape(ii, l) = ls*le*rec_pol[j]*curl_recpol2(k,l) + curl_recpol(j,l)*lo*rec_pol2[k];

	// rotations of curls
	for (j = 0; j <= p-2; j++)
	  for (k = 0; k <= p-2-j; k++, ii++)
	    for (l = 0; l < 2; l++)
	      shape(ii, l) = ls*le*rec_pol[j]*curl_recpol2(k,l) - curl_recpol(j,l) * lo * rec_pol2[k];

	// rec_pol2 * RT_0
	for (j = 0; j <= p-2; j++, ii++)
	  for (l = 0; l < 2; l++)
	    shape(ii,l) = lo*rec_pol2[j] * (ls*clami(fav[1],l)-le*clami(fav[0],l));


      }



  }

  template <class T_ORTHOPOL>
  void HDivHighOrderTrigA<T_ORTHOPOL> :: CalcDivShape (const IntegrationPoint & ip,
						      FlatVector<> shape) const
  {

    double lami[3];
    lami[0] = ip(0);
    lami[1] = ip(1);
    lami[2] = 1-ip(0)-ip(1);

    Mat<3,2> dlami(0.);
    dlami(0,0) = 1.;
    dlami(1,1) = 1.;
    dlami(2,0) = -1.;
    dlami(2,1) = -1.;



    Mat<3,2> clami(0.);
    clami(0,1) = -1.;
    clami(1,0) = 1.;
    clami(2,1) = 1.;
    clami(2,0) = -1.;

    shape = 0.0;

    int p, p_curl;

    int i, j, k, l;
    int ii = 3; 


    if ( discontinuous )
      {
	AutoDiff<3> x(ip(0),0), y(ip(1),1), z(ip(2),2);
	shape = 0.0;
	ii = 0;
	// no orientation necessary
	// use L2 shape functions, they are L2 orthogonal
	int n=order_inner[0];
	ArrayMem<AutoDiff<3>, 20> polx(n+1), poly(n+1), polz(n+1);


	// Polynomials orthogonal w.r.t L2 inner product
	ScaledLegendrePolynomial (n, 2*x+y-1, 1-y, polx);
	
	for (int i = 0; i <= order_inner[0]; i++)
	  {
	    JacobiPolynomial (n, 2*y-1, 2*i+1, 0, poly);
	    for (int j = 0; j <= order_inner[0]-i; j++)
	      for ( int m = 0; m < 2; m++ )
		shape(ii++) = polx[i].DValue(m) * poly[j].Value() 
		  + polx[i].Value() * poly[j].DValue(m);
	  }
      
	return;
      }

    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    for (i = 0; i < 3; i++)
      {
        p = order_edge[i];

	int is = edges[i][0]; 
	int ie = edges[i][1];
	
	if (vnums[is] > vnums[ie])
	  swap (is, ie);

	//RT low order edge shape function  
	//shape(i) = 2*(dlami(ie,0)*clami(is,0) +  dlami(ie,1)*clami(ie,1));
	for(j=0; j<= 1; j++)
	  shape(i) += dlami(ie,j)*clami(is,j) - dlami(is,j)*clami(ie,j);
	

	//Edge normal shapes (curls) 
	if(p>0)
	  {
	    for(j=0; j<= p-1;j++)
	      ii++; 
	  }
      }
    
    if(order_inner[0] < 2) return;
    
    p = order_inner[0];
    p_curl = order_inner[0];
        
    //Inner shapes (Face) 
    if(p>1) 
      {
	
	ArrayMem<double,10> rec_pol(order-1), drec_polx(order-1),  drec_polt(order-1);
	ArrayMem<double,10> rec_pol2(order-1), drec_pol2x(order-1), drec_pol2t(order-1);  
	ArrayMem<double,30> mem(3*order*sizeof(double));
	ArrayMem<double,30> mem2(3*order*sizeof(double));
	FlatMatrixFixWidth<2> grad_recpol(order, &mem[0]);
	FlatMatrixFixWidth<2> grad_recpol2(order, &mem2[0]);

	int fav[3];
	for(i=0;i<3;i++) fav[i] = i;

	//Sort vertices  first edge op minimal vertex
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);

	
	double ls = lami[fav[0]]; 
	double le = lami[fav[1]]; 
       	double lo = lami[fav[2]];  
	
	ScaledLegendrePolynomialandDiff(p_curl-2, le-ls, 1-lo,
					rec_pol, drec_polx, drec_polt);  	    
	LegendrePolynomialandDiff(p_curl-2, 2*lo-1,rec_pol2, drec_pol2x);
	// \curl (ls * le * rec_pol), and \curl (lo rec_pol2)
	for (j = 0; j <= p_curl-2; j++)
	  {
	    for (l = 0; l < 2; l++)
	      {
		grad_recpol(j,l) =
		  ls * le * (drec_polx[j] * (dlami(fav[1],l) - dlami(fav[0],l)) -
			     drec_polt[j] * (dlami(fav[2],l))) +
		  (ls * dlami(fav[1],l) + le * dlami(fav[0],l)) * rec_pol[j];
		grad_recpol2(j,l) =
		  lo * (drec_pol2x[j] * 2*dlami(fav[2],l)) + 
		  dlami(fav[2],l) * rec_pol2[j];
	      }
	  }



	// curls:
	for (j = 0; j <= p_curl-2; j++)
	  for (k = 0; k <= p_curl-2-j; k++)
	    ii++;
	     
	// rotations of curls  
	for (j = 0; j <= p-2; j++)
	  for (k = 0; k <= p-2-j; k++, ii++)
	    for (l = 0; l < 2; l++)
	      shape(ii) = 2 * (grad_recpol(j,0) * grad_recpol2(k,1)
			       - grad_recpol(j,1) * grad_recpol2(k,0));
	// shape (ii) = grad u . curl v = 2 (u_x v_y - u_y v_x) ; 
		
	// div(rec_pol * RT_0)  =
 	double divrt0 = 2*(dlami(fav[0],0)*clami(fav[1],0) +  dlami(fav[0],1)*clami(fav[1],1));
	Vec<2> rt0;
	for (l = 0; l<2; l++)
	  rt0(l) = lami[fav[0]]*clami(fav[1],l)-lami[fav[1]]*clami(fav[0],l);
	for (j = 0; j <= p-2; j++, ii++)
	  shape(ii) = divrt0 * lo*rec_pol2[j] + rt0(0)*grad_recpol2(j,0) +
	    rt0(1)*grad_recpol2(j,1); 
      }
  }

  template class HDivHighOrderTrigA<IntegratedLegendreMonomialExt>;
  template class HDivHighOrderTrigA<TrigExtensionMonomial>;
  //  template class HDivHighOrderTrigA<TrigExtensionOptimal>;
  template class HDivHighOrderTrigA<TrigExtensionMin>;


 
  //------------------------------------------------------------------------
  // HDivHighOrderTet
  //------------------------------------------------------------------------

  template <class T_ORTHOPOL>
  HDivHighOrderTetA<T_ORTHOPOL> :: HDivHighOrderTetA (int aorder)
    : HDivHighOrderFiniteElement<3>(ET_TET)
  {
    nv = 4;
    ned = 6;
    nf = 4;

    augmented =0;

    int i;
    order_inner = INT<3>(aorder,aorder,aorder);
    for (i = 0; i < nf; i++)
      {
	order_face[i] = INT<2>(aorder,aorder);
      }

    ComputeNDof();

  }

 

  template <class T_ORTHOPOL>
  void HDivHighOrderTetA<T_ORTHOPOL> :: ComputeNDof()
  {
    ndof_face = 0;
  

    // RT_0
    ndof = 4;

    //ndof = 0;
    int i;
    int p;

    for (i = 0; i < 4; i++)
      {
        if (order_face[i][0] > 0)
	  { 
	    p = order_face[i][0];

	    ndof_face += (p*p+3*p)/2;

	    ndof += (p*p+3*p)/2;

	  }


      }


    
    p = order_inner[0];
    int pc = order_inner[0]; // should be order_inner_curl!!!  
    if(pc > 1) 
      ndof += pc*(pc+1)*(pc-1)/3 + pc*(pc-1)/2;
    if(p > 1) 
      ndof += p*(p+1)*(p-1)/6 + p*(p-1)/2 + p-1;


    order = 0; // max(order_face_normal,order_inner);
    for (i = 0; i < 4; i++)
      {
	if (order_face[i][0] > order)
	  order = order_face[i][0];

      }
    if (order_inner[0] > order)
      order = order_inner[0];

    order++;


  }



  template <class T_ORTHOPOL>
  void HDivHighOrderTetA<T_ORTHOPOL> ::
  GetInternalDofs (ARRAY<int> & idofs) const
  {
    if (discontinuous)
    {
      idofs.SetSize(0);
      for (int i=0; i<ndof; i++)
        idofs.Append(i);
      return ;
    }
    else
    {
      idofs.SetSize (0);

      int base = 4;
      for (int i = 0; i < 4; i++)
      {
        int p = order_face[i][0];
        base += (p*p+3*p)/2;
      }


      if(order_inner[0] >= 2)
      {
        int p = order_inner[0];
	//int ni = 6*(p-1) + 4*(p-1)*(p-2) + (p-1)*(p-2)*(p-3)/2;
        int ni = p*(p+1)*(p-1)/2 + p*(p-1)+(p-1);
        for (int i = 0; i < ni; i++)
          idofs.Append (base+i);
      }
    //(*testout) << "idofs = " << idofs << endl;
    }
  }
  
  
  
  
  
  template <class T_ORTHOPOL>
  void HDivHighOrderTetA<T_ORTHOPOL> :: GetFacetDofs(int fa, ARRAY<int> & dnums) const 
  {
    if (fa >= 4 ) 
    {
      cout << " Warning HDIVHighOrderTet::GetFacetDofNrs() index out of range" << endl; 
      
      dnums.SetSize(0); 
      return; 
    } 
  
    dnums.SetSize(0); 
    dnums.Append(fa);  // lowest-order
    
    int ii = 4; // Thomas-Raviart 
     
    for (int i = 0; i < fa; i++)
    {     
      int p = order_face[i][0];
      ii+= (p*p+3*p)/2; 
    }
    
    int p = order_face[fa][0];
    int nf = (p*p+3*p)/2;
    
    for(int i = 0; i < nf; i++)  
      dnums.Append(ii+i); 

    *testout << " HDIVHO2 FacetDofs " << dnums << endl; 
  }                  

  template <class T_ORTHOPOL>
  void HDivHighOrderTetA<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip,
						  FlatMatrixFixWidth<3> shape) const
  {

    double x = ip(0);
    double y = ip(1);
    double z = ip(2);

    if ( discontinuous )
      {
	shape = 0.0;
	// no orientation necessary
	// use L2 shape functions, they are L2 orthogonal
	int n=order_inner[0];
	ArrayMem<double, 20> polx(n+1), poly(n+1), polz(n+1);


	// Polynomials orthogonal w.r.t L2 inner product
	ScaledLegendrePolynomial ( n, 2*x+y+z-1, 1-y-z, polx);
	
	int ii = 0;
	for (int i = 0; i <= n; i++)
	  {
	    ScaledJacobiPolynomial (n, 2*y+z-1, 1-z, 2*i+1, 0, poly);
	    for (int j = 0; j <= n-i; j++)
	      {
		JacobiPolynomial(n, 2*z-1, 2*i+2*j+2, 0, polz);
		for (int k = 0; k <= n-i-j; k++)
		  for ( int m = 0; m < 3; m++ )
		    shape(ii++, m) = polx[i] * poly[j] * polz[k];
	      }
	  }
      
	return;
      }

    int i, j, ii, k,l, m, n;

    int print=0;

    double lami[4];
    for (j=0;j<3;j++) lami[j] = ip(j);
    lami[3] = 1-ip(0)-ip(1)-ip(2);

    int index[3][3];

    int node[3];

    Mat<4,3> dlami;
    dlami = 0.0;
    dlami(0,0) = 1.;
    dlami(1,1) = 1.;
    dlami(2,2) = 1.;
    dlami(3,0) = -1.;
    dlami(3,1) = -1.;
    dlami(3,2) = -1.;

    shape = 0.0;

    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    const POINT3D * vertices = ElementTopology::GetVertices (ET_TET);


    Vec<3> tang, tang1, normal, vp;

    int is, ie, iop, iop1, iop2;

    int p=(order+1)*(order+2)*(order+3)/2;
    

    ArrayMem<double,10> rec_pol(p), rec_pol_1(p), rec_pol_2(p);

    //********************************FACE SHAPES********************************************************
	// RT_0
	// Typ 1
	// Typ 2
    
	ii = 4;
    for (i = 0; i < 4; i++)
      {
	p = order_face[i][0];
	
	if (p >= 0)
	  {
            int fav[3];
	    for(j=0; j<3; j++) fav[j]=faces[i][j];

	    //Sort vertices  first edge op minimal vertex
	    if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	    if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	    if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	    int fop = 6 - fav[0] - fav[1] - fav[2];

            is = fav[0]; ie = fav[1]; iop = fav[2];

	    AutoDiff<3> ls = lami[is];
	    AutoDiff<3> le = lami[ie];
	    AutoDiff<3> lo = lami[iop];
	    AutoDiff<3> lz = lami[fop];
           

	    for (j = 0; j < 3; j++)
	      {
		ls.DValue(j) = dlami(is,j);
		le.DValue(j) = dlami(ie,j);
		lo.DValue(j) = dlami(iop,j);
		lz.DValue(j) = dlami(fop,j);

	      }
	    AutoDiff<3> lsle=ls*le;

            Vec<3> vp, vp1, vp2, vp3, nedelec;
	    Vec<3> dlamis, dlamie, dlamio;

	    for (n = 0; n < 3; n++)
	      {
		dlamis(n) = dlami(is,n);
		dlamie(n) = dlami(ie,n);
		dlamio(n) = dlami(iop,n);
		nedelec(n) = ls.Value()*dlamie(n) - le.Value()*dlamis(n);
	      }
	    vp1 = Cross (dlamis, dlamie);
	    vp2 = Cross (dlamio, dlamis);
	    vp3 = Cross (dlamie, dlamio);



	    // RT_0 low order shapes
	    for (j = 0; j < 3; j++)
	      shape(i,j) = ls.Value()*vp3(j)+le.Value()*vp2(j)+lo.Value()*vp1(j);


            Vec<3> grad1, grad2;
	    ArrayMem<AutoDiff<3>, 10> ad_rec_pol1(order+1);
	    ArrayMem<AutoDiff<3>, 10> ad_rec_pol2(order+1);

         // *testout << " p " << p << endl; 
            
          
	    ScaledLegendrePolynomial(p-1, le-ls, ls+le, ad_rec_pol1);
	    ScaledLegendrePolynomial(p-1, lo-le-ls, lo+ls+le, ad_rec_pol2);

            for (k = 0; k <= p-1; k++)
	      { 
		ad_rec_pol1[k] *= lsle;
		ad_rec_pol2[k] *= lo;
	      }
           
            // Typ 1
            for (k = 0; k <= p-1; k++)
	      {
		for (l = 0; l <= p-1-k; l++, ii++)
		  {

		    for (j = 0; j < 3; j++)
		      {
			grad1(j) = ad_rec_pol1[k].DValue(j);
			grad2(j) = ad_rec_pol2[l].DValue(j);
		      }
		    vp = Cross(grad2,grad1);
		    for (m = 0; m < 3; m++)
		      shape(ii,m) = 2.*vp(m);
		  }
	      }

	    // Typ 2
	    for (k = 0; k <= p-1; k++, ii++)
	      {
		for (j = 0; j < 3; j++)
                  grad2(j) = ad_rec_pol2[k].DValue(j);
		vp = Cross(nedelec,grad2);
		for (m = 0; m < 3; m++)
	          shape(ii,m) = (-1.*vp(m) + 2 * vp1(m) * ad_rec_pol2[k].Value());
	      }

	  }
      }
    


   // if(order_inner[0] < 2) return;
    // **********************************INNER*********************************************

    
      
    // **************curl inner shapes****************************************************
    p = order_inner[0];
    Vec<3> grad1, grad2, grad3, vp1, vp2, vp3;
    ArrayMem<AutoDiff<3>, 10> ad_poli(order-1);
    ArrayMem<AutoDiff<3>, 10> ad_polj(order-1);
    ArrayMem<AutoDiff<3>, 10> ad_polk(order-1);
    AutoDiff<3> pol_prod;
    AutoDiff<3> x1 (ip(0),0);
    AutoDiff<3> y1 (ip(1),1);
    AutoDiff<3> z1 (ip(2),2);
    
    //AutoDiff<3> lz = lami[fop];
    T_INNERSHAPES::CalcSplitted(p+2, x1-(1-x1-y1-z1), y1, z1,ad_poli, ad_polj, ad_polk );
    //   ScaledLegendrePolynomial(p-1, le-ls, ls+le, ad_rec_pol1);
    //   ScaledLegendrePolynomial(p-1, lo-le-ls, lo+ls+le, ad_rec_pol2);
    for (i = 0; i <= p-2; i++)
      {
	
	for (j = 0; j <= p-2-i; j++)
	  {
	    
	    for (k = 0; k <= p-2-i-j; k++, ii+=3)
	      {
		
		for (l = 0; l < 3; l++)
		  {
		    grad1(l) = ad_poli[i].DValue(l);
		    grad2(l) = ad_polj[j].DValue(l);
		    grad3(l) = ad_polk[k].DValue(l);
		  }
		
		vp1 = Cross(grad2,grad1);
		vp2 = Cross(grad2,grad3);
		vp3 = Cross(grad3,grad1);

		
		for (m = 0; m < 3; m++)
		  {
		    shape(ii,m)   = 2.* (ad_polk[k].Value()*vp1(m) + ad_poli[i].Value()*vp2(m));
		    shape(ii+1,m) = 2.* (ad_polj[j].Value()*vp3(m) - ad_poli[i].Value()*vp2(m));
		    shape(ii+2,m) = ad_polk[k].Value()*vp1(m);
		  }
              
             
            
	      }
	  }
      }
    AutoDiff<3> le = 1-x1-y1-z1;
    
    Vec<3> ned=0.;;
    ned(0) = - x1.Value() - le.Value(); 
    ned(1) =  - x1.Value(); 
    ned(2)= - x1.Value();
   
    Vec<3> curlned=0.;
    curlned(1) = (-x1.DValue(0)*le.DValue(2));
    curlned(2) = (x1.DValue(0)*le.DValue(1));
    // Typ 4  curl(Nedelec*ad_polk*ad_polj)
    for (i = 0; i <= p-2; i++)
      {
	for (j = 0; j <= p-2-i; j++, ii+=2)
	  {
	    pol_prod = ad_polj[j]*ad_polk[i];
	    for (k = 0; k < 3; k++)
	      {
		grad1(k) = pol_prod.DValue(k);
		grad2(k) = ad_polj[i].DValue(k);
	      }
	    vp1 = Cross(grad1,ned);
	    vp2 = Cross(ned,grad2);
	    
         
          
	    for (m = 0; m < 3; m++)
	      {
		shape(ii,m) = 2.*(curlned(m)*pol_prod.Value()) + vp1(m);
		shape(ii+1,m) = vp2(m)*ad_polk[j].Value();
	      }
                       
	  }
      }
    for (i = 0; i <= p-2; i++, ii++)
      {
	for (j = 0; j < 3; j++)
	  grad1(j) = z1.DValue(j);
      vp1 = Cross(ned,grad1);   
	for (m = 0; m < 3; m++)
	  {
	    shape(ii,m) = vp1(m)* ad_polj[i].Value();
          
	    
	  }
      }
   
  // (*testout)<<"shape tet="<<shape<<endl;
     
  }

  template <class T_ORTHOPOL>
  void HDivHighOrderTetA<T_ORTHOPOL> :: CalcDivShape (const IntegrationPoint & ip,
						     FlatVector<> divshape) const
  {


    if ( discontinuous )
      {
	AutoDiff<3> x(ip(0),0), y(ip(1),1), z(ip(2),2);
	divshape = 0.0;
	// no orientation necessary
	// use L2 shape functions, they are L2 orthogonal
	int n=order_inner[0];
	ArrayMem<AutoDiff<3>, 20> polx(n+1), poly(n+1), polz(n+1);


	// Polynomials orthogonal w.r.t L2 inner product
	ScaledLegendrePolynomial ( n, 2*x+y+z-1, 1-y-z, polx);
	
	int ii = 0;
	for (int i = 0; i <= n; i++)
	  {
	    ScaledJacobiPolynomial (n, 2*y+z-1, 1-z, 2*i+1, 0, poly);
	    for (int j = 0; j <= n-i; j++)
	      {
		JacobiPolynomial(n, 2*z-1, 2*i+2*j+2, 0, polz);
		for (int k = 0; k <= n-i-j; k++)
		  for ( int m = 0; m < 3; m++ )
		    {
		      divshape(ii++) = polx[i].DValue(m) * poly[j].Value() * polz[k].Value() 
			+ polx[i].Value() * poly[j].DValue(m) * polz[k].Value()
			+ polx[i].Value() * poly[j].Value() * polz[k].DValue(m);
		    }
	      }
	  }
       
	return;
      }

    double x = ip(0);
    double y = ip(1);
    double z = ip(2);

    int i, j, ii, k, l, m, n;
    int is, ie, iop, iop1, iop2;




    double lami[4];
    for (j=0;j<3;j++) lami[j] = ip(j);
    lami[3] = 1-ip(0)-ip(1)-ip(2);

    Mat<4,3> dlami;
    dlami = 0.0;
    dlami(0,0) = 1.;
    dlami(1,1) = 1.;
    dlami(2,2) = 1.;
    dlami(3,0) = -1.;
    dlami(3,1) = -1.;
    dlami(3,2) = -1.;

    divshape = 0.0;

    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    const POINT3D * vertices = ElementTopology::GetVertices (ET_TET);

    Vec<3> tang, tang1, normal, vp;

    int p=(order+1)*(order+2)*(order+3)/2;




    ArrayMem<double,10> rec_pol(p), rec_pol_1(p), rec_pol_2(p);
    ArrayMem<double,10> drec_pol(p), drec_pol_1(p), drec_pol_2(p);
    ArrayMem<AutoDiff<3>, 100> ad_rec_pol(100);


    //*******************************DIV FACE-SHAPES*********************************************



	// RT_0
	// Typ 1
	// Typ 2

	ii = 4;
    for (i = 0; i < 4; i++)
      {
	p = order_face[i][0];

	if (p >= 0)
	  {
            int fav[3];
	    for(j=0; j<3; j++) fav[j]=faces[i][j];

	    //Sort vertices  first edge op minimal vertex
	    if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	    if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	    if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
	    int fop = 6 - fav[0] - fav[1] - fav[2];

            is = fav[0]; ie = fav[1]; iop = fav[2];

	    AutoDiff<3> ls = lami[is];
	    AutoDiff<3> le = lami[ie];
	    AutoDiff<3> lo = lami[iop];
	    AutoDiff<3> lz = lami[fop];


	    for (j = 0; j < 3; j++)
	      {
		ls.DValue(j) = dlami(is,j);
		le.DValue(j) = dlami(ie,j);
		lo.DValue(j) = dlami(iop,j);
		lz.DValue(j) = dlami(fop,j);

	      }
	    AutoDiff<3> lsle=ls*le;

            Vec<3> vp, vp1, vp2, vp3, nedelec;
	    Vec<3> dlamis, dlamie, dlamio;
	    for (n = 0; n < 3; n++)
	      {
		dlamis(n) = dlami(is,n);
		dlamie(n) = dlami(ie,n);
		dlamio(n) = dlami(iop,n);
		nedelec(n) = ls.Value()*dlamie(n) - le.Value()*dlamis(n);
	      }
	    vp1 = Cross (dlamis, dlamie);
	    vp2 = Cross (dlamio, dlamis);
	    vp3 = Cross (dlamie, dlamio);



	    // Divergenz RT_0 low order divshapes
	    for (j = 0; j < 3; j++)
	      divshape(i) += dlamis(j)*vp3(j) + dlamie(j)*vp2(j) + dlamio(j)*vp1(j);

	    // Divergenz Typ 1 = 0

	    // Divergenz Typ 2 = 0


	  }
      }
    ii += 4*(p*p+3*p)/2;


    //**********************************DIV INNER-SHAPES*********************************************
	if(order_inner[0] < 2) return;

    
      //***************DIV curl shapes*****************************************************
	  p = order_inner[0];
    Vec<3> grad1, grad2, grad3, vp1, vp2, vp3;
    ArrayMem<AutoDiff<3>, 100> ad_poli(100);
    ArrayMem<AutoDiff<3>, 100> ad_polj(100);
    ArrayMem<AutoDiff<3>, 100> ad_polk(100);
    AutoDiff<3> pol_prod;
    AutoDiff<3> x1 (ip(0),0);
    AutoDiff<3> y1 (ip(1),1);
    AutoDiff<3> z1 (ip(2),2);

    T_INNERSHAPES::CalcSplitted(p+2, x1-(1-x1-y1-z1), y1, z1,ad_poli, ad_polj, ad_polk );
    //   ScaledLegendrePolynomial(p-1, le-ls, ls+le, ad_rec_pol1);
    //   ScaledLegendrePolynomial(p-1, lo-le-ls, lo+ls+le, ad_rec_pol2);

    for (i = 0; i <= p-2; i++)
      {

	for (j = 0; j <= p-2-i; j++)
	  {

	    for (k = 0; k <= p-2-i-j; k++, ii+=3)
	      {

		for (l = 0; l < 3; l++)
		  {
		    grad1(l) = ad_poli[i].DValue(l);
		    grad2(l) = ad_polj[j].DValue(l);
		    grad3(l) = ad_polk[k].DValue(l);
		  }
		vp1 = Cross(grad2,grad1);
		//  Divergenz Typ 1 = 0
		//  Divergenz Typ 2 = 0
		//  Divergenz Typ 3 div((adpolj.D X adpoli.D)adpolk) = - adpolj.D*(adpolk.D X adpoli.D)

            
            
		divshape(ii+2) = InnerProduct(grad3,vp1);
           
        
	      }
	  }
      }

  

    AutoDiff<3> le = 1-x1-y1-z1;

    Vec<3> ned=0.;;
    ned(0) = - x1.Value() - le.Value(); ned(1) =  - x1.Value(); ned(2)= - x1.Value();
    Vec<3> curlned=0.;
    curlned(1) = 2; // (-x1.DValue(0)*le.DValue(2));
    curlned(2) = -2; // (x1.DValue(0)*le.DValue(1));
    // Divergenz Typ 4  div(curl(Nedelec*ad_polk*ad_polj)) = 0
    // Divergenz Typ 5 div((Nedelec X ad_polj)ad_polk) = ad_polj.D*ad_polk*curlned + ad_polk.D*(ned X ad_polj.D)
    for (i = 0; i <= p-2; i++)
      {
	for (j = 0; j <= p-2-i; j++, ii+=2)
	  {

	    for (k = 0; k < 3; k++)
	      {
		grad1(k) = ad_polj[i].DValue(k);
		grad2(k) = ad_polk[j].DValue(k);

	      }
	    vp1 = Cross(ned,grad1);
	    double h1 = InnerProduct(grad2,vp1);
	    double h2 = InnerProduct(grad1,curlned);

	    divshape(ii+1) = h2*ad_polk[j].Value() + h1;
         
          
	  }
      }
    // Divergenz Typ 6 div((Nedelec X grad(lam4))*ad_polj) = grad(lam4)*ad_polj*curlned + ad_polj.D*(ned X grad(lam4))
    for (i = 0; i <= p-2; i++, ii++)
      {
	for (j = 0; j < 3; j++)
	  {
	    grad1(j) = z1.DValue(j);
	    grad2(j) = ad_polj[i].DValue(j);
	  }
	vp1 = Cross(ned,grad1);

	double h1 = InnerProduct(grad2,vp1);
	double h2 = InnerProduct(grad1,curlned);

	divshape(ii) = ad_polj[i].Value()*h2 + h1;
     
     
      }

    //(*testout)<<"divshape analytisch tet="<<divshape<<endl;
	    
    //Vector<> divshape_num(divshape.Size());
    //HDivFiniteElement<3> :: CalcDivShape (ip, divshape_num);
	    
     //(*testout) << "divshape numerisch = " << endl << divshape_num << endl;

    // (*testout) << "difference = " << endl << divshape-divshape_num << endl;

  }

  template class HDivHighOrderTetA<IntegratedLegendreMonomialExt>;
  template class HDivHighOrderTetA<TrigExtensionMonomial>;
  //  template class HDivHighOrderTetA<TrigExtensionOptimal>;
  template class HDivHighOrderTetA<TrigExtensionMin>;


}
