/*********************************************************************/
/* File:   hdivsurfacehofespace.cpp                                  */
/*********************************************************************/

/**
   High Order Finite Element Space for H div on surface 
   Finite Elements (BDM)
*/


#include <comp.hpp>
#include <fem.hpp> 
#include <solve.hpp>

namespace ngcomp
{

using namespace ngsolve;
using namespace ngcomp;
using namespace ngfem;

#ifdef PARALLEL
using namespace ngparallel;
#endif

  HDivSurfaceFiniteElement :: HDivSurfaceFiniteElement (ELEMENT_TYPE etype)
    : FiniteElement (ElementTopology::GetSpaceDim (etype), etype, -1, -1) 
  {
    for (int i = 0; i < 8; i++)
      vnums[i] = i;
  }




  HDivSurfaceFiniteElement ::  ~HDivSurfaceFiniteElement() 
  { ; }


  void HDivSurfaceFiniteElement :: SetVertexNumbers (FlatArray<int> & avnums)
  {
    for (int i = 0; i < avnums.Size(); i++)
      vnums[i] = avnums[i];
  }

 
  void HDivSurfaceFiniteElement ::  SetOrderEdge (FlatArray<int> & oe)
  {
    for (int i = 0; i < oe.Size(); i++)
      order_edge[i] = oe[i];
    ComputeNDof();
  }


  void HDivSurfaceFiniteElement ::  SetOrderInner (int oi)
  {
    order_inner = oi;
    ComputeNDof();

  }



  HDivSurfaceTrig :: HDivSurfaceTrig (int aorder) : HDivSurfaceFiniteElement (ET_TRIG)
  {
    order = aorder; 
    // full P^k       (k+1)*(k+2) 
    // edges             3 *(k+1)
    // inner          (k-1)*(k+1)
    ninner = (sqr(order) - 1 );

    // 2 dofs per edge
    ndof = 3*(order+1);

    eltype = ET_TRIG;
    cout << "trig, order = " << order << endl;
    cout << "ndof = " << ndof << ", ninner = " << ninner << endl;

  }





  void HDivSurfaceTrig :: GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize(0);
    for (int i = ndof-ninner; i < ndof; i++)
      idofs.Append(i);
  }


  void HDivSurfaceTrig :: CalcShape (const IntegrationPoint & ip, 
				     FlatMatrixFixWidth<3> shape) const   
  {
    double x = ip(0), y=ip(1); 
    double lami[3] = {x,y,1-x-y};
    double ddlami[3] = {x,y,1-x-y};

    shape = 0.0; 
    
    int i, j, k, l;
    int ii = 0; 
    
//     shape(ii,0) = x;
//     shape(ii,1) = -x;
//     ii++;

//     shape(ii,1) = -1 + x + y;
//     ii++;

//     shape(ii,0) = -y;
//     shape(ii,1) = y;
//     ii++;

//     shape(ii,0) = -1 + x + y;
//     ii++;

//     shape(ii,1) = y;
//     ii++;

//     shape(ii,0) = x;
//     ii++;

    ArrayMem< Vec<2>, 10 > T(3), n(3);
    T[0](0) = 0.0; T[0](1) = 1.0;
    T[1](0) = -1.0; T[1](1) = 0.0;
    T[2](0) = 1.0; T[2](1) = -1.0;

    n[0](0) = -1.0; n[0](1) = 0.0;
    n[1](0) = 0.0; n[1](1) = -1.0;
    n[2](0) = 1.0; n[2](1) = 1.0;

    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    ArrayMem<double, 20> phiedge(order_inner + 1);

    //-------------------------
    // edge basis functions
    //-------------------------
    for (i = 0; i < 3; i++)
      {
	int es = edges[i][0]; // EDGE i
	int ee = edges[i][1]; // edge i+1
	
	int et = 3 - ee - es; // third edge i+2
	
      	if (vnums[es] > vnums[ee]) { swap (es, ee);}

	
	
	double ls = lami[es];
	double le = lami[ee];
	double lt = lami[et];
	


	shape(ii,0) = T[ee](0) * lami[es];
	shape(ii,1) = T[ee](1) * lami[es];
	ii++;
	shape(ii,0) = T[es](0) * lami[ee];
	shape(ii,1) = T[es](1) * lami[ee];
	ii++;

	IntegratedLegendreMonomialExt::CalcTrigExt (order_inner+2,le + ls, (1 + ls - le), phiedge);


	for(int l=2; l<=order; l++)
	  {
	    shape(ii,0) = phiedge[l] * n[et](0);
	    shape(ii,1) = phiedge[l] * n[et](1);
	    ii++;
	  }

  	
      }
    
    //-----------------------
    // inner basis functions
    //-----------------------

    for(i=0; i<3; i++)
      {
	int es = edges[i][0]; // EDGE i
	int ee = edges[i][1]; // edge i+1
	
	int et = 3 - ee - es; // third edge i+2
	
	if (vnums[es] > vnums[ee]) { swap (es, ee);}

	
	
	double ls = lami[es];
	double le = lami[ee];
	double lt = lami[et];


	IntegratedLegendreMonomialExt::CalcTrigExt (order_inner+2,le + ls, (1 + ls - le), phiedge);

	for(int l=2; l <= order; l++)
	  {
	    shape(ii,0) = phiedge[l] * T[et](0);
	    shape(ii,1) = phiedge[l] * T[et](1);
	    ii++;
	  }
      }

    if(order > 2)
      {
	ArrayMem<double, 20> pol(order-1);

	double bub = lami[0] * lami[1] * lami[2];
	double ls = lami[1], le = lami[0];
	IntegratedLegendreMonomialExt::CalcTrigExt (order_inner+2,le + ls, (1 + ls - le), phiedge);
	
	LegendrePolynomial (order-2, 2*lami[2] - 1, pol);
	
	for(int l=2;  l<=order_inner-1; l++)
	  for(int k=0; k+l<=order_inner-1; k++)
	    {
	      double p = phiedge[l]*pol[k]*lami[2];
	      shape(ii,0) = p;
	      ii++;
	      shape(ii,1) = p;
	      ii++;
	    }
      }

  }


  void HDivSurfaceTrig :: ComputeNDof()
  {
    ndof = (order_inner + 1) * (order_inner + 2);
    order = order_inner;
    order++;
  }




  HDivSurfaceQuad :: HDivSurfaceQuad (int aorder) : HDivSurfaceFiniteElement(ET_QUAD)
  {
    order = aorder;

    ninner = 0;

    ndof = 4*(order+1);
    cout << "quad, order = " << order << endl;
    cout << "ndof = " << ndof << ", ninner = " << ninner << endl;



  }


  void HDivSurfaceQuad :: GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize(0);
    for (int i = ndof-ninner; i < ndof; i++)
      idofs.Append(i);
  }


  void HDivSurfaceQuad :: CalcShape (const IntegrationPoint & ip, 
				     FlatMatrixFixWidth<3> shape) const   
  {
    double x = ip(0), y=ip(1);

    double lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  

    shape = 0.0; 
    
    int i, j, k, l;
    int ii = 0; 
    
    const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
    ArrayMem<double, 20> ha(order+1);
	
    for (i = 0; i < 4; i++)
      {
	;
      }
  }

  void HDivSurfaceQuad :: ComputeNDof()
  {
    ndof = 2 * (order_inner + 1) * (order_inner + 1);
    order = order_inner;
    order++;
  }




  HDivSurfaceHighOrderFESpace ::  HDivSurfaceHighOrderFESpace (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama, flags)
  {
    trig = new HDivSurfaceTrig (order);
    quad = new HDivSurfaceQuad (order);
  
    if (ma.GetDimension() == 2)
      {
	cout << "use H1 on 1D boundary" << endl;
      }
    else
      {
	boundary_evaluator = 
	  new RobinHDivSIntegrator (new ConstantCoefficientFunction(1));
      }
 

  }


  HDivSurfaceHighOrderFESpace :: ~HDivSurfaceHighOrderFESpace ()
  {
    ;
  }

  FESpace * HDivSurfaceHighOrderFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new HDivSurfaceHighOrderFESpace (ma, flags);
  }  
  

  void HDivSurfaceHighOrderFESpace :: Update()
  {
    cout << "HDivSurfaceFESpace" << endl;

    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nel = ma.GetNE();

    ndof = 0;
    if (ma.GetDimension() == 3)
      {


	order_edge.SetSize (ned);
	order_inner.SetSize (nel);


	order_edge = order;
	order_inner = order;

	/*
	(*testout) << "H(div) on surface: order_edge = " << order << endl << "           order_inner = " << order_inner << endl;
	*/

	ndof = 0;
	first_edge_dof.SetSize(ned+1); // last element = ndof;
	for (int i = 0; i < ned; i++)
	  {
	    first_edge_dof[i] = ndof;
	    ndof += order_edge[i] + 1;
	  }
	first_edge_dof[ned] = ndof;


	first_inner_dof.SetSize(nel+1);
	int inci = 0;
	for(int i=0; i< nel; i++)
	  {
	    switch(ma.GetElType(i))
	      {
	      case ET_TRIG:
		inci = order_inner[i]*order_inner[i]-1;
		break;
	      case ET_QUAD:
		inci = 2*(order_inner[i]*order_inner[i] - 1);
		break;
	      default:
		;
	      }
	    if (inci < 0) inci = 0;

	    first_inner_dof[i] = ndof;
	    ndof+=inci;
	  }
	first_inner_dof[nel] = ndof;

      }
    else 
      cout << "no HDiv on 1D Surface " << endl;
  }


  const FiniteElement & HDivSurfaceHighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
  
    switch (ma.GetSElType(elnr))
      {
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0;
      }
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetSElVertices(elnr, vnums);
   
    if (!fe)
      {
	stringstream str;
	str << "HDivSurfaceHighOrderFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }


    HDivSurfaceFiniteElement * hofe =
      dynamic_cast<HDivSurfaceFiniteElement* > (fe);
    ArrayMem<int, 12> ednums, order_ed;
    
    ma.GetElEdges(elnr, ednums);
    order_ed.SetSize (ednums.Size());
    
    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order_edge[ednums[j]];
    
    hofe -> SetVertexNumbers (vnums);
    hofe -> SetOrderEdge (order_ed);
    
    hofe -> SetOrderInner (order_inner[elnr]);
    hofe -> ComputeNDof();
    

    return *fe;
  }
 
  const FiniteElement & HDivSurfaceHighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    throw Exception ("Volume elements not available for HDivSurfaceHighOrderFESpace");
  }
 


  int HDivSurfaceHighOrderFESpace :: GetNDof () const
  {
    return ndof;
  }

  void HDivSurfaceHighOrderFESpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    ARRAY<int> ednums, fnums;
    int i, j;
    int first,next;

    // ma.GetElEdges (elnr, ednums);
    if(ma.GetDimension() == 3)
      {
	ma.GetElEdges (elnr, ednums); //neu
	fnums.SetSize(0);
	dnums.SetSize(0);

	//edges
	for ( i = 0; i < ednums.Size(); i++)
	  {
	    first = first_edge_dof[ednums[i]];
	    next = first_edge_dof[ednums[i]+1];
	  
	    for (j = first; j <next; j++)
	      dnums.Append (j);
	  }

	//inner
	first = first_inner_dof[elnr];
	next = first_inner_dof[elnr+1];
	for(j=first; j<next; j++)
	  dnums.Append(j);

      }


    first = first_inner_dof[elnr];
    int neldofs = first_inner_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetSElIndex (elnr)))
      dnums = -1;
  }


  void HDivSurfaceHighOrderFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }



  namespace
#ifdef MACOS
  hdivsurfacehofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hdivsurf", HDivSurfaceHighOrderFESpace::Create);

      GetIntegrators().AddBFIntegrator ("robinhdivs", 3, 1,
					RobinHDivSIntegrator::Create);

      GetIntegrators().AddLFIntegrator ("neumannhdiv", 3, 1,
					NeumannHDivSIntegrator::Create);

    }
    
    Init init;
  }


}
