#include <fem.hpp>    

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif

namespace ngfem
{

  using namespace ngfem;

#ifdef PARALLEL
  using namespace ngparallel;
#endif



  inline AutoDiff<3> Cross (const AutoDiff<3> & u,
			    const AutoDiff<3> & v)
  {
    AutoDiff<3> hv;
    hv.Value() = 0.0;
    hv.DValue(0) = u.DValue(1)*v.DValue(2)-u.DValue(2)*v.DValue(1);
    hv.DValue(1) = u.DValue(2)*v.DValue(0)-u.DValue(0)*v.DValue(2);
    hv.DValue(2) = u.DValue(0)*v.DValue(1)-u.DValue(1)*v.DValue(0);
    return hv;
  }



  // calculate sym [ (nabla a) x ( nabla^2 b ) x (nabla c) ]
  // where the first cross product is row-wise, the second is column-wise
  

  inline AutoDiffDiff<3>  CrossCross ( const AutoDiffDiff<3> & a, const AutoDiffDiff<3> & b,
			   const AutoDiffDiff<3> & c )
  {
    AutoDiffDiff<3> sigma;
    sigma.Value() = 0.0;
    sigma.DValue(0) = sigma.DValue(1) = sigma.DValue(2) = 0.0;

    const double & a0 = a.DValue(0), & a1 = a.DValue(1), & a2 = a.DValue(2);
    const double & c0 = c.DValue(0), & c1 = c.DValue(1), & c2 = c.DValue(2);
    const double & b00 = b.DDValue(0,0), & b01 = b.DDValue(0,1), & b02 = b.DDValue(0,2),
      & b11 = b.DDValue(1,1), & b12 = b.DDValue(1,2), & b22 = b.DDValue(2,2),
      & b10 = b.DDValue(1,0), & b20 = b.DDValue(2,0), & b21 = b.DDValue(2,1);

    // diagonal
    sigma.DDValue(0,0) = c2 * ( a1*b12 - a2*b11 ) - c1 * ( a1*b22 - a2*b21 );
    sigma.DDValue(1,1) = c0 * ( a2*b20 - a0*b22 ) - c2 * ( a2*b00 - a0*b02 );
    sigma.DDValue(2,2) = c1 * ( a0*b01 - a1*b00 ) - c0 * ( a0*b11 - a1*b10 );

    // off diagonal
    sigma.DDValue(0,1) = 0.5 * ( c0 * ( a1*b22 - a2*b21 ) - c1 * ( a2*b20 - a0*b22 ) 
				 - c2 * ( a0*b12 - a2*(b10+b01) + a1*b02 ) ) ;
    sigma.DDValue(0,2) = 0.5 * ( c2 * ( a0*b11 - a1*b10 ) - c0 * ( a1*b12 - a2*b11 )
				 - c1 * ( a2*b01 - a1*(b02+b20) + a0*b21 ) ) ;
    sigma.DDValue(1,2) = 0.5 * ( c1 * ( a2*b00 - a0*b02 ) - c2 * ( a0*b01 - a1*b00 ) 
				 - c0 * ( a1*b20 - a0*(b12+b21) + a2*b10 ) );

    // symmetric
    sigma.DDValue(1,0) = sigma.DDValue(0,1);
    sigma.DDValue(2,0) = sigma.DDValue(0,2);
    sigma.DDValue(2,1) = sigma.DDValue(1,2);

//     *testout << " sigma " << sigma << endl;
    return sigma;

  }


  // calculate sym [ (nabla a) x ( nabla^2 b ) x (nabla c) ]
  // where the first cross product is row-wise, the second is column-wise
  // first and last argument depend on x, y, second one on z

  inline AutoDiffDiff<3>  CrossCross_prismz ( const AutoDiffDiff<2> & a, const AutoDiffDiff<1> & b,
			   const AutoDiffDiff<2> & c )
  {
    AutoDiffDiff<3> sigma;
    sigma.Value() = 0.0;
    sigma.DValue(0) = sigma.DValue(1) = sigma.DValue(2) = 0.0;

    const double & a0 = a.DValue(0), & a1 = a.DValue(1);
    const double & c0 = c.DValue(0), & c1 = c.DValue(1);
    const double & b22 = b.DDValue(0,0);

    // diagonal
    sigma.DDValue(0,0) = - c1 * ( a1*b22 );
    sigma.DDValue(1,1) = - c0 * ( a0*b22 );
    sigma.DDValue(2,2) = 0.0;

    // off diagonal
    sigma.DDValue(0,1) = 0.5 * ( c0 * a1 + c1 * a0 ) * b22 ;
    sigma.DDValue(0,2) = 0.0;
    sigma.DDValue(1,2) = 0.0;
    // symmetric
    sigma.DDValue(1,0) = sigma.DDValue(0,1);
    sigma.DDValue(2,0) = sigma.DDValue(0,2);
    sigma.DDValue(2,1) = sigma.DDValue(1,2);

//     *testout << " sigma " << sigma << endl;
    return sigma;

  }

  // calculate sym [ (nabla a) x ( nabla^2 b ) x (nabla c) ]
  // where the first cross product is row-wise, the second is column-wise
  // first and second argument only depend on x,y, third only on z  

  inline AutoDiffDiff<3>  CrossCross_prismxy ( const AutoDiffDiff<2> & a, const AutoDiffDiff<2> & b,
			   const AutoDiffDiff<1> & c )
  {
    AutoDiffDiff<3> sigma;
    sigma.Value() = 0.0;
    sigma.DValue(0) = sigma.DValue(1) = sigma.DValue(2) = 0.0;

    const double & a0 = a.DValue(0), & a1 = a.DValue(1);
    const double & c2 = c.DValue(0);
    const double & b00 = b.DDValue(0,0), & b01 = b.DDValue(0,1), & b10 = b.DDValue(1,0),
      & b11 = b.DDValue(1,1);

    // diagonal
    sigma.DDValue(0,0) = 0.0;
    sigma.DDValue(1,1) = 0.0;
    sigma.DDValue(2,2) = 0.0;

    // off diagonal
    sigma.DDValue(0,1) = 0.0;
    sigma.DDValue(0,2) = 0.5 * ( c2 * ( a0*b11 - a1*b10 ) ) ;
    sigma.DDValue(1,2) = - 0.5 * ( c2 * ( a0*b01 - a1*b00 ) );

    // symmetric
    sigma.DDValue(1,0) = sigma.DDValue(0,1);
    sigma.DDValue(2,0) = sigma.DDValue(0,2);
    sigma.DDValue(2,1) = sigma.DDValue(1,2);

//     *testout << " sigma " << sigma << endl;
    return sigma;

  }




  // calculate sym ( curl curl ( nabla b nabla c^T ) )
  // first curl is row wise, second curl columnwise

  inline AutoDiffDiff<3> CurlCurl ( const AutoDiffDiff<3> & b, const AutoDiffDiff<3> & c  )
  {
    AutoDiffDiff<3> sigma;
    sigma.Value() = 0.0;
    sigma.DValue(0) = sigma.DValue(1) = sigma.DValue(2) = 0.0;

    const double & b00 = b.DDValue(0,0), & b01 = b.DDValue(0,1), & b02 = b.DDValue(2,0),
      & b10 = b.DDValue(1,0), & b11 = b.DDValue(1,1), & b12 = b.DDValue(1,2),
      & b20 = b.DDValue(2,0), & b21 = b.DDValue(2,1), & b22 = b.DDValue(2,2);

    const double & c00 = c.DDValue(0,0), & c01 = c.DDValue(0,1), & c02 = c.DDValue(2,0),
      & c10 = c.DDValue(1,0), & c11 = c.DDValue(1,1), & c12 = c.DDValue(1,2),
      & c20 = c.DDValue(2,0), & c21 = c.DDValue(2,1), & c22 = c.DDValue(2,2);

    // diagonal 
    sigma.DDValue(0,0) = 2*b21*c21 + 2*b12*c12 - b12*c21 - b21*c12 - b22*c11 - b11*c22;;
    sigma.DDValue(1,1) = 2*b02*c02 + 2*b20*c20 - b02*c20 - b20*c02 - b22*c00 - b00*c22;
    sigma.DDValue(2,2) = 2*b10*c10 + 2*b01*c01 - b10*c01 - b01*c10 - b11*c00 - b00*c11;

    // off diagonal
    sigma.DDValue(0,1) = - b02*c12 - b12*c02 - b21*c20 - b20*c21 +
      0.5 * ( b01*c22 + b02*c21 + b22*c10 + b20*c12 + b22*c01 + b21*c02 + b10*c22 + b12*c20 ); 
    sigma.DDValue(0,2) = - b01*c21 - b21*c01 - b10*c12 - b12*c10 +
      0.5 * ( b11*c20 + b10*c21 + b20*c11 + b01*c12 + b20*c11 + b21*c10 + b11*c02 + b12*c01 );
    sigma.DDValue(1,2) = - b20*c10 - b10*c20 - b02*c01 - b01*c02 +
      0.5 * ( b12*c00 + b10*c02 + b00*c21 + b01*c20 + b00*c12 + b02*c10 + b12*c00 + b20*c01 );

    // symmetric
    sigma.DDValue(1,0) = sigma.DDValue(0,1);
    sigma.DDValue(2,0) = sigma.DDValue(0,2);
    sigma.DDValue(2,1) = sigma.DDValue(1,2);

    return sigma;
  }


  // calculate sym ( curl curl ( nabla b nabla c^T ) )
  // first curl is row wise, second curl columnwise
  // first argument only depends on x, y, second only on z (used for prismatic elements)

  inline AutoDiffDiff<3> CurlCurl_prism ( const AutoDiffDiff<2> & b, const AutoDiffDiff<1> & c  )
  {
    AutoDiffDiff<3> sigma;
    sigma.Value() = 0.0;
    sigma.DValue(0) = sigma.DValue(1) = sigma.DValue(2) = 0.0;

    const double & b00 = b.DDValue(0,0), & b01 = b.DDValue(0,1), 
      & b10 = b.DDValue(1,0), & b11 = b.DDValue(1,1);

    const double & c22 = c.DDValue(2,2);

    // diagonal 
    sigma.DDValue(0,0) = - b11*c22;;
    sigma.DDValue(1,1) = - b00*c22;
    sigma.DDValue(2,2) = 0.0;

    // off diagonal
    sigma.DDValue(0,1) = 0.5 * ( b01*c22 + b10*c22 ); 
    sigma.DDValue(0,2) = 0.0;
    sigma.DDValue(1,2) = 0.0;

    // symmetric
    sigma.DDValue(1,0) = sigma.DDValue(0,1);
    sigma.DDValue(2,0) = sigma.DDValue(0,2);
    sigma.DDValue(2,1) = sigma.DDValue(1,2);


    return sigma;
  }




  // assigns the DDValue-matrix of sigma to the vector representation shape
  inline void AssignSymmetricTensor ( const AutoDiffDiff<3> & sigma, int & ii,  
				      FlatMatrixFixWidth<6> & shape )
  {
    shape ( ii, 0 ) = sigma.DDValue(0,0);
    shape ( ii, 1 ) = sigma.DDValue(1,1);
    shape ( ii, 2 ) = sigma.DDValue(2,2);
    shape ( ii, 3 ) = sigma.DDValue(0,1);
    shape ( ii, 4 ) = sigma.DDValue(0,2);
    shape ( ii, 5 ) = sigma.DDValue(1,2);
    return;
  }

template <int DIM>  
HDivSymFiniteElement<DIM> :: HDivSymFiniteElement (ELEMENT_TYPE etype)
  : FiniteElement (ElementTopology::GetSpaceDim (etype), etype, -1, -1) 
{
  for (int i = 0; i < 8; i++)
    vnums[i] = i;

}


template <int DIM>  
HDivSymFiniteElement<DIM> ::  ~HDivSymFiniteElement() 
{ ; }


template <int DIM>  
void HDivSymFiniteElement<DIM> :: 
SetVertexNumbers (FlatArray<int> & avnums)
{
  for (int i = 0; i < avnums.Size(); i++)
    vnums[i] = avnums[i];
}



template <int DIM>  
void  HDivSymFiniteElement<DIM> :: CalcDivShape (const IntegrationPoint & ip, 
						 FlatMatrixFixWidth<DIM> divshape) const 
{
  double eps = 1e-7;  

  ArrayMem<double, 200> 
    hm1(DIM_STRESS*ndof), hm2(DIM_STRESS*ndof), 
    hm3(DIM_STRESS*ndof), hm4(DIM_STRESS*ndof), hmi(DIM_STRESS*ndof);
  
  FlatMatrixFixWidth<DIM_STRESS> shape1(ndof, &hm1[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape2(ndof, &hm2[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape3(ndof, &hm3[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape4(ndof, &hm4[0]);
  FlatMatrixFixWidth<DIM_STRESS> dshapei(ndof, &hmi[0]);

  divshape = 0;

  for (int i = 0; i < DIM; i++)
    {
      IntegrationPoint ip1 = ip;
      IntegrationPoint ip2 = ip;
      ip1(i) -= eps;
      ip2(i) += eps;
      
      CalcShape (ip1, shape1);
      CalcShape (ip2, shape2);
      
      ip1(i) -= eps;
      ip2(i) += eps;
      CalcShape (ip1, shape3);
      CalcShape (ip2, shape4);
      
      dshapei = 
	2/(3*eps) * (shape2 - shape1) 
	- 1/(12*eps) * (shape4 - shape3);
      
      
      if (DIM == 2)
	{
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,0);
		  divshape(j,1) += dshapei(j,2);
		}
	      break;
	      
	    case 1:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,2);
		  divshape(j,1) += dshapei(j,1);
		}
	      break;
	    }
	}
      else
	{
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,0);   // sigmaxx
		  divshape(j,1) += dshapei(j,3);   // sigmaxy
		  divshape(j,2) += dshapei(j,4);   // sigmaxz
		}
	      break;
	    
	    case 1:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,3);
		  divshape(j,1) += dshapei(j,1);
		  divshape(j,2) += dshapei(j,5);
		}
	      break;

	    case 2:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,4);
		  divshape(j,1) += dshapei(j,5);
		  divshape(j,2) += dshapei(j,2);
		}
	      break;
	    }
	}
    }
  // (*testout) << "div-shape = " << divshape << endl;
}








template <int DIM>  
void  HDivSymFiniteElement<DIM> :: CalcDivDivShape (const IntegrationPoint & ip, 
						    FlatVector<> ddshape) const 
{
  double eps = 1e-4;  

  if (ElementTopology::GetSpaceDim (ElementType()) == 2)
    {
      ArrayMem<double, 200> hm1(3*ndof), hm2(3*ndof), 
	hm3(3*ndof), hm4(3*ndof), hmi(3*ndof);
    
      FlatMatrix<> shape1(ndof, 2, &hm1[0]);
      FlatMatrix<> shape2(ndof, 2, &hm2[0]);
      FlatMatrix<> shape3(ndof, 2, &hm3[0]);
      FlatMatrix<> shape4(ndof, 2, &hm4[0]);
      FlatMatrix<> dshapei(ndof, 2, &hmi[0]);

      ddshape = 0;
    
      for (int i = 0; i < 2; i++)
	{
	  IntegrationPoint ip1 = ip;
	  IntegrationPoint ip2 = ip;
	  ip1(i) -= eps;
	  ip2(i) += eps;

	  CalcDivShape (ip1, shape1);
	  CalcDivShape (ip2, shape2);

	  ip1(i) -= eps;
	  ip2(i) += eps;
	  CalcDivShape (ip1, shape3);
	  CalcDivShape (ip2, shape4);
	
	  dshapei = 
	    2/(3*eps) * (shape2 - shape1) 
	    - 1/(12*eps) * (shape4 - shape3);
	
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		ddshape(j) += dshapei(j,0);
	      break;
	    
	    case 1:
	      for (int j = 0; j < ndof; j++)
		ddshape(j) += dshapei(j,1);
	      break;
	    }
	}
    }
}




template class  HDivSymFiniteElement<2>;
template class  HDivSymFiniteElement<3>;
















// Trig with full basis (non-curved elements)

HDivSymTrig :: HDivSymTrig (int aorder, int aorder_facet, const int noncont) : HDivSymFiniteElement<2> (ET_TRIG)
{
  nfacets = 3;

  INT<2> of = INT<2>(aorder_facet, aorder_facet);
  INT<3> oi = INT<3>(aorder, aorder, aorder);
  order = max2(aorder, aorder_facet);
  for ( int i = 0; i < 6; i++ )
    order_facet[i] = of; 
  order_inner = oi;

  ninner = 3*order_inner[0]*(order_inner[0]+1)/2 ;

  ndof = (order_facet[0][0] + order_facet[0][1] + order_facet[0][2]) + 3  + ninner ;

//   cout << "trig, order_facet = " << order_facet << ", order_inner = " << order_inner << endl;
//   cout << "ndof = " << ndof << ", ninner = " << ninner << endl;

  noncontinuous = noncont;
  if(noncontinuous) ninner=ndof;

}

  void HDivSymTrig :: ComputeNDof() 
  {
    order = order_inner[0];
    for ( int i = 0; i < 2; i++)
      if ( order < order_facet[i][0] ) order = order_facet[i][0];

    ninner = 3*order_inner[0]*(order_inner[0]+1)/2 ;

    ndof = (order_facet[0][0] + order_facet[1][0] + order_facet[2][0]) + 3  + ninner ;

  }


void HDivSymTrig :: GetInternalDofs (ARRAY<int> & idofs) const
{

  idofs.SetSize(0);
  if(noncontinuous)
    {
      for (int i = 0; i < ndof; i++)
	idofs.Append(i);
    }
  else
    {
      for (int i = ndof-ninner; i < ndof; i++)
	idofs.Append(i);
    }
  
}

void HDivSymTrig :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  int offset = 0;
  fdofs.SetSize(0);

  for ( int i = 0; i < face; i++ )
    offset += order_facet[i][0] + 1;

  for ( int l = 0; l < order_facet[face][0] + 1; l++ )
    fdofs.Append(offset + l);
}



void HDivSymTrig :: CalcShape (const IntegrationPoint & ip, 
			       FlatMatrixFixWidth<3> shape) const   // n * 3 
{
  AutoDiffDiff<2> x(ip(0),0); 
  AutoDiffDiff<2> y(ip(1),1); 
//   AutoDiff<2> lami[3] = {x,y,1-x-y};
  AutoDiffDiff<2> ddlami[3] = {x,y,1-x-y};

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  int maxorder_facet =
    max2 ( order_facet[0][0], max2(order_facet[1][0], order_facet[2][0]) );
  const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
  ArrayMem<AutoDiffDiff<2>, 20> ha(maxorder_facet+1);
  
  for (i = 0; i < 3; i++)
    {
      int es = edges[i][0]; 
      int ee = edges[i][1];
	
      if (vnums[es] > vnums[ee])  swap (es, ee);

      AutoDiffDiff<2> ls = ddlami[es];
      AutoDiffDiff<2> le = ddlami[ee];

      IntegratedLegendreMonomialExt::CalcTrigExt (maxorder_facet+2, le-ls, 1-le-ls, ha);
      for (int l = 0; l <= order_facet[i][0]; l++)
	{
	  shape(ii,0) = ha[l].DDValue(1,1);
	  shape(ii,1) = ha[l].DDValue(0,0);
	  shape(ii,2) = -ha[l].DDValue(1,0);
	    
	  ii++;
	}
    }
  
  
  int es = 0; int ee = 1; int et = 2;
  AutoDiffDiff<2> ls = ddlami[es];
  AutoDiffDiff<2> le = ddlami[ee];
  AutoDiffDiff<2> lt = ddlami[et];
  ArrayMem<AutoDiffDiff<2>, 20> u(order_inner[0]+1);
  ArrayMem<AutoDiffDiff<2>, 10>  P_lam2(order_inner[0]-1), v(order_inner[0]+1);

  AutoDiffDiff<2> phie0 = le;
  AutoDiffDiff<2> phie1 = -ls;

  IntegratedLegendreMonomialExt::CalcTrigExt (order_inner[0]+2, le-ls, 1-le-ls, u);

  LegendrePolynomial (order_inner[0], 2*lt-1, v );
  for ( int i = 0; i <= order_inner[0]; i++ )
    v[i] *= lt;
//   if(order_inner > 2)
//     LegendrePolynomial (order_inner-2, 2*lt - 1, P_lam2);
//     v[0] = lt;
//   for(int i=0; i<=order_inner[0]-1; i++)
//     {
//       v[i+1] = v[i] * (1 - 2*lt);
//     }

//   for(int i=2; i<=order_inner; i++)
//     for(int j=1; j+i<=order_inner+1; j++)
  if ( order_inner[0]==0 )
    {
      i=0; j=0;
	shape(ii,0) = u[i].DDValue(1,1)*v[j].Value() + 2*u[i].DValue(1)*v[j].DValue(1) + u[i].Value()*v[j].DDValue(1,1);
	shape(ii,1) = u[i].DDValue(0,0)*v[j].Value() + 2*u[i].DValue(0)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(0,0);
	shape(ii,2) = -(u[i].DDValue(0,1)*v[j].Value() + u[i].DValue(0)*v[j].DValue(1) + 
			u[i].DValue(1)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(1,0));
	ii++;
	
	shape(ii,0) = u[i].DDValue(1,1)*v[j].Value() - 2*u[i].DValue(1)*v[j].DValue(1) + u[i].Value()*v[j].DDValue(1,1);
	shape(ii,1) = u[i].DDValue(0,0)*v[j].Value() - 2*u[i].DValue(0)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(0,0);
	shape(ii,2) = -(u[i].DDValue(0,1)*v[j].Value() - u[i].DValue(0)*v[j].DValue(1) - 
			u[i].DValue(1)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(1,0));
	ii++;
	return;
    }

  for(int i=0; i<=order_inner[0]-1; i++)
    for(int j=0; j+i<=order_inner[0]-1; j++)
      {
	shape(ii,0) = u[i].DDValue(1,1)*v[j].Value() + 2*u[i].DValue(1)*v[j].DValue(1) + u[i].Value()*v[j].DDValue(1,1);
	shape(ii,1) = u[i].DDValue(0,0)*v[j].Value() + 2*u[i].DValue(0)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(0,0);
	shape(ii,2) = -(u[i].DDValue(0,1)*v[j].Value() + u[i].DValue(0)*v[j].DValue(1) + 
			u[i].DValue(1)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(1,0));
	ii++;
	
	shape(ii,0) = u[i].DDValue(1,1)*v[j].Value() - 2*u[i].DValue(1)*v[j].DValue(1) + u[i].Value()*v[j].DDValue(1,1);
	shape(ii,1) = u[i].DDValue(0,0)*v[j].Value() - 2*u[i].DValue(0)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(0,0);
	shape(ii,2) = -(u[i].DDValue(0,1)*v[j].Value() - u[i].DValue(0)*v[j].DValue(1) - 
			u[i].DValue(1)*v[j].DValue(0) + u[i].Value()*v[j].DDValue(1,0));
	ii++;
	
	if( j > 0 )
	  {
	    shape(ii,0) = u[i].DDValue(1,1)*v[j].Value() - u[i].Value()*v[j].DDValue(1,1);
	    shape(ii,1) = u[i].DDValue(0,0)*v[j].Value() - u[i].Value()*v[j].DDValue(0,0);
	    shape(ii,2) = -(u[i].DDValue(0,1)*v[j].Value() - 
			    u[i].Value()*v[j].DDValue(1,0));
	ii++;
	  }
      }

  for( int i=0; i<order_inner[0]; i++)
    {
      shape(ii,0) = phie1.DValue(1)*v[i].Value() + phie1.Value()*v[i].DValue(1);
      shape(ii,1) = phie0.DValue(0)*v[i].Value() + phie0.Value()*v[i].DValue(0);
      shape(ii,2) = ( phie1.DValue(0)*v[i].Value() + phie1.Value()*v[i].DValue(0) +  
		      phie0.DValue(1)*v[i].Value() + phie0.Value()*v[i].DValue(1)) * 0.5;
      ii++;
    }  
   
//   cout << "order - inner " << order_inner[0] << order  << endl;
//   cout << "order - facet " << order_facet[0]  << order_facet[1] << order_facet[2]  << endl;
  return;
  /*
  double xs = ip(0), ys = ip(1);
  
  if(order_inner > 0)
    {
      Vec<3> lambdai;
      lambdai(0) = xs;
      lambdai(1) = ys;
      lambdai(2) = 1 - xs - ys;
      ArrayMem<double, 10>  P_lam2(order_inner-1);
      
      if(order_inner > 2)
	LegendrePolynomial (order_inner-2, 2*lambdai(2) - 1, P_lam2);



      shape(ii++,2) = lambdai(2);

      shape(ii, 0) = 2*lambdai(0);
      shape(ii++,2) = -lambdai(0);

      shape(ii,1) = 2*lambdai(1);
      shape(ii++,2) = -lambdai(1);

      ArrayMem<double, 20> phiedge(order_inner + 1);
      

      for( int i=0; i<3; i++)
	{

//  	  int es = edges[i][0]; 
//  	  int ee = edges[i][1];
	  
//  	  if (vnums[es] > vnums[ee])  swap (es, ee);
 	  int es = (i+1) % 3; int ee = (i+2) % 3;	  
  	  if (vnums[es] > vnums[ee])  swap (es, ee);

	  double lam1 = lambdai(es);
	  double lam2 = lambdai(ee);

	  //IntegratedLegendreMonomialExt::CalcTrigExt (order_facet+2, le-ls, 1-le-ls, ha);

	  IntegratedLegendreMonomialExt::CalcTrigExt (order_inner+2,lam2 + lam1, (1 + lam1 - lam2), phiedge);

	  for(int l=2; l<=order_inner; l++)
	    {
	      // i=0
	      if(i == 2)
		{
		  shape(ii,0) = 2*(phiedge[l]);
		  shape(ii,2) = -phiedge[l];
		  ii++;
		  shape(ii,1) = 2*(phiedge[l]);
		  shape(ii,2) = -phiedge[l];
		  ii++;
		}
	      else if (i == 0)
		{
		  shape(ii,1) = 2*(phiedge[l]);
		  shape(ii,2) = -phiedge[l];
		  ii++;
		  shape(ii,2) = phiedge[l];
		  ii++;
		}
	      else if (i == 1)
		{
		  shape(ii,2) = phiedge[l];
		  ii++;
		  shape(ii,0) = 2*(phiedge[l]);
		  shape(ii,2) = -phiedge[l];
		  ii++;
		}
	    }
// 	  double tl;
// 	  //	  ls[0] = -1.0;
// 	  phiedge[0](i) = -1.0;
// 	  tl = t;
// 	  //	  ls[1] = xx;
// 	  phiedge[1](i) = xx * tl;
// 	  for(int l=2; l<=order_inner; l++)
// 	    {
// 	      tl = tl*t;
// 	      ls[l] = (2*l-3)*xx*ls[l-1]/l - (l-3)*ls[l-2]/l;
// 	      phiedge[l](i) = ls[l] * tl;
// 	    }
	}
      
      

 
  /*  
  AutoDiffDiff<2> ddbub = ddlami[0] * ddlami[1] * ddlami[2];

  AutoDiffDiff<2> add_x[10];
  add_x[0] = 1.0;
  AutoDiffDiff<2> add_y[10];
  add_y[0] = 1.0;

  for (int i = 1; i <= order_inner+1; i++)
    {
      add_x[i] = add_x[i-1] * AutoDiffDiff<2> (1-2.0*x);
      add_y[i] = add_y[i-1] * AutoDiffDiff<2> (1-2.0*y);
    }

  */
  
    /*
  for (int k = 0; k < 3; k++)
  {
  AutoDiffDiff<2> bub = ddlami[k] * ddlami[(k+1)%3];
  double lami = ddlami[(k+2)%3].Value();
      
  for (int i = 0; i <= order_inner-1; i++)
  for (int j = 0; i+j <= order_inner-1; j++)
  {
  double fac = lami * add_x[i].Value() * add_y[j].Value();
  shape(ii,0) = bub.DDValue(1,1) * fac;
  shape(ii,1) = bub.DDValue(0,0) * fac;
  shape(ii,2) = -bub.DDValue(1,0) * fac;
  ii++;
  }
  }


  // one order higher for second derivatives
  for (int i = 0; i <= order_inner; i++)
  {
  AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner-i];
  shape(ii,0) = pot.DDValue(1,1);
  shape(ii,1) = pot.DDValue(0,0);
  shape(ii,2) = -pot.DDValue(1,0);
  ii++;
  }
  // and one more
  for (int i = 0; i <= order_inner+1; i++)
  {
  AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner+1-i];
  shape(ii,0) = pot.DDValue(1,1);
  shape(ii,1) = pot.DDValue(0,0);
  shape(ii,2) = -pot.DDValue(1,0);
  ii++;
  }
  
  

    for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j+i <= order_inner; j++)
    {
    // AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner-i];
    AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[j];
	
    shape(ii,0) = pot.DDValue(1,1);
    shape(ii,1) = pot.DDValue(0,0);
    shape(ii,2) = -pot.DDValue(1,0);
    ii++;
    }
*/
}












HDivSymQuad :: HDivSymQuad (int aorder, int aorder_facet, const int noncont) : HDivSymFiniteElement<2> (ET_QUAD)
{
  nfacets = 4;

  order = max2(aorder, aorder_facet);
  for ( int i = 0; i < 6; i++ )
    order_facet[i] = INT<2> (aorder_facet, aorder_facet);
  order_inner = INT<3> (aorder, aorder, aorder);
 
  ndof = 3 * (order_inner[0]+1) * (order_inner[1]+1);

  ninner = ndof - 2; // 2*(order_inner[0]+1) - 2*(order_inner[1]+1);

  ndof = ninner;
  for ( int ed = 0; ed < 4; ed++ )
    ndof += order_facet[ed][0] + 1;

  noncontinuous = noncont;



}

  void HDivSymQuad :: ComputeNDof ()
  {
    order = order_inner[0];
    for ( int i = 0; i < 4; i++ )
      if ( order < order_facet[i][0] ) order = order_facet[i][0];

    ndof = 3 * (order_inner[0]+1) * (order_inner[1]+1);
    
    ninner = ndof - 2; // 2*(order_inner[0]+1) - 2*(order_inner[1]+1);
    
    ndof = ninner;
    for ( int ed = 0; ed < 4; ed++ )
      ndof += order_facet[ed][0] + 1;

  }
void HDivSymQuad :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  if(noncontinuous)
    {
      for (int i = 0; i < ndof; i++)
	idofs.Append(i);
    }
  else
    {
      for (int i = ndof-ninner; i < ndof; i++)
	idofs.Append(i);
    }

}


void HDivSymQuad :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  int offset = 0;
  fdofs.SetSize(0);

  for ( int i = 0; i < face; i++ )
    offset += order_facet[i][0] + 1;

  for ( int l = 0; l < order_facet[face][0]+1; l++ )
    fdofs.Append(offset + l);
}


void HDivSymQuad :: CalcShape (const IntegrationPoint & ip, 
			       FlatMatrixFixWidth<3> shape) const   // n * 3 
{
  AutoDiffDiff<2> x(ip(0),0); 
  AutoDiffDiff<2> y(ip(1),1); 

  AutoDiffDiff<2> lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  AutoDiffDiff<2> sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    

  int maxorder_facet = 0;
  for ( int i = 0; i < 4; i++ )
    if ( maxorder_facet < order_facet[i][0] ) maxorder_facet = order_facet[i][0];
  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
  ArrayMem<AutoDiffDiff<2>, 20> ha(maxorder_facet+1);
	
  for (i = 0; i < 4; i++)
    {
      int es = edges[i][0], ee = edges[i][1];
      if (vnums[es] > vnums[ee])  swap (es, ee);

      AutoDiffDiff<2> xi = sigma[es]-sigma[ee];
      AutoDiffDiff<2> eta = lami[es]+lami[ee];

      IntegratedLegendreMonomialExt::Calc (order_facet[i][0]+2, -xi, ha);

      for (int l = 0; l <= order_facet[i][0]; l++)
	{
	  ha[l] *= eta;
	  shape(ii,0) = ha[l].DDValue(1,1);
	  shape(ii,1) = ha[l].DDValue(0,0);
	  shape(ii,2) = -ha[l].DDValue(1,0);
	    
	  ii++;
	}
    }


  /* 
  // inner shapes

  double polx[3] = { 1, 2*x.Value()-1, x.Value()*(1-x.Value()) };
  double poly[3] = { 1, 2*y.Value()-1, y.Value()*(1-y.Value()) };

  AutoDiffDiff<2> add_x[10];
  add_x[0] = x * (1-x);
  AutoDiffDiff<2> add_y[10];
  add_y[0] = y * (1-y);

  for (int i = 1; i <= order_inner; i++)
    {
      add_x[i] = add_x[i-1] * (1-2.0*x);
      add_y[i] = add_y[i-1] * (1-2.0*y);
    }
 
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j <= order_inner; j++)
      {
	AutoDiffDiff<2> prod = add_x[i] * add_y[j];

	shape(ii,0) = add_x[i].Value() * add_y[j].DDValue(1,1);
	shape(ii,1) = add_x[i].DDValue(0,0) * add_y[j].Value();
	shape(ii,2) = -add_x[i].DValue(0) * add_y[j].DValue(1);
	ii++;

	shape(ii,0) = -add_x[i].Value() * add_y[j].DDValue(1,1);
	shape(ii,1) = add_x[i].DDValue(0,0) * add_y[j].Value();
	shape(ii,2) = 0;
	ii++;
      }

  shape(ii,2) = 1;
  ii++;

  for (int i = 0; i <= order_inner; i++)
    {
      shape(ii,2) = -0.5 * add_y[i].DValue(1);  
      ii++;
      shape(ii,2) = -0.5 * add_x[i].DValue(0);  
      ii++;
    }
  
  */
  
  

      ArrayMem<double, 10> polx(order_inner[0]+2), poly(order_inner[1]+2);
      
      LegendrePolynomial (order_inner[0]+1, 2*ip(0)-1, polx);
      LegendrePolynomial (order_inner[1]+1, 2*ip(1)-1, poly);
      
      double fac;
      for (int i = 0; i <= order_inner[0]-1; i++)
	for (int j = 0; j <= order_inner[1]+1; j++)
	  {
	    fac = polx[i]*poly[j];
	      shape(ii++,0) = ip(0)*(1-ip(0))*fac;
	  }
      for (int i = 0; i <= order_inner[0]+1; i++)
	for (int j = 0; j <= order_inner[1]-1; j++)
	  {
	    fac = polx[i]*poly[j];
	      shape(ii++,1) = ip(1)*(1-ip(1))*fac;
	  }
      for (int i = 0; i <= order_inner[0]; i++)
	for (int j = 0; j <= order_inner[1]; j++)
	  {
	    fac = polx[i]*poly[j];
	      shape(ii++,2) = fac;
	  }
 



}



void HDivSymQuad :: CalcDivShape (const IntegrationPoint & ip, 
				  FlatMatrixFixWidth<2> divshape) const   // n * 3 
{

  AutoDiffDiff<2> x(ip(0),0); 
  AutoDiffDiff<2> y(ip(1),1); 

  AutoDiffDiff<2> lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  AutoDiffDiff<2> sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  

  divshape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  int maxorder_facet = 0;
  for ( int i = 0; i < 4; i++ )
    if ( maxorder_facet < order_facet[i][0] ) maxorder_facet = order_facet[i][0];
  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
  ArrayMem<AutoDiffDiff<2>, 20> ha(maxorder_facet+1);
	
  for (i = 0; i < 4; i++)
    {
      int es = edges[i][0], ee = edges[i][1];
      if (vnums[es] > vnums[ee])  swap (es, ee);

      ii += order_facet[i][0] +1;
    }


  AutoDiff<1> adx(ip(0),0); 
  AutoDiff<1> ady(ip(1),0); 

  if(order_inner[0] > 0 || order_inner[1] > 0)
    {
      ArrayMem<AutoDiff<1>, 10> polx(order_inner[0]+2), poly(order_inner[1]+2);
      ArrayMem<AutoDiff<1>, 10> polxbub(order_inner[0]+2), polybub(order_inner[1]+2);
      
      LegendrePolynomial (order_inner[0]+1, 2*adx-1, polx);
      LegendrePolynomial (order_inner[1]+1, 2*ady-1, poly);
      
      for (int i = 0; i <= order_inner[0]+1; i++)
	{
	  polxbub[i] = adx*(1-adx)*polx[i];
	}
      for (int i = 0; i <= order_inner[1]+1; i++)
	{
	  polybub[i] = ady*(1-ady)*poly[i];
	}
      
      for (int i = 0; i <= order_inner[0]-1; i++)
	for (int j = 0; j <= order_inner[1]+1; j++)
	  {
	    divshape(ii++,0) = polxbub[i].DValue(0) * poly[j].Value();
	  }
      for (int i = 0; i <= order_inner[0]+1; i++)
	for (int j = 0; j <= order_inner[1]-1; j++)
	  {
	    divshape(ii++,1) = polybub[j].DValue(0) * polx[i].Value();
	  }
      for (int i = 0; i <= order_inner[0]; i++)
	for (int j = 0; j <= order_inner[1]; j++)
	  {
	    divshape(ii  ,1) = polx[i].DValue(0) * poly[j].Value();
	    divshape(ii++,0) = polx[i].Value() * poly[j].DValue(0);
	      
	  }
    }
}











HDivSymTet :: HDivSymTet (int aorder, int aorder_facet, const int noncont) 
  : HDivSymFiniteElement<3> (ET_TET)
{
  nfacets = 4;

  order = max2 (aorder, aorder_facet);

  for ( int i = 0; i < 6; i++ )
    order_facet[i] = INT<2> (aorder_facet, aorder_facet);
  order_inner = INT<3> (aorder, aorder, aorder); 

  ninner = (order_inner[0]+1)*(order_inner[0]+2)*(order_inner[0]+3) - 
    2 * (order_inner[0]+1)*(order_inner[0]+2);
  
  ndof = ninner;
  for ( int fa = 0; fa < 4; fa++ )
    ndof += (order_facet[fa][0]+1)*(order_facet[fa][0]+2) / 2 ;
  noncontinuous = noncont;
  
  
}

  void HDivSymTet :: ComputeNDof ()
  {
    order = order_inner[0];
    for ( int i = 0; i < 4; i++ )
      if( order < order_facet[i][0] ) order = order_facet[i][0];

    ninner = 6 * (order_inner[0]+1)*(order_inner[0]+2)*(order_inner[0]+3)/6 - 
      4 * (order_inner[0]+1)*(order_inner[0]+2)/2;
    
    ndof = ninner;
    for ( int fa = 0; fa < 4; fa++ )
      ndof += (order_facet[fa][0]+1)*(order_facet[fa][0]+2) / 2 ;

  }

void HDivSymTet :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  if(noncontinuous)
    {
      for (int i = 0; i < ndof; i++)
	idofs.Append(i);
    }
  else
    {
      for (int i = ndof-ninner; i < ndof; i++)
	idofs.Append(i);
    }
  

}

void HDivSymTet :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  int offset = 0;
  fdofs.SetSize(0);

  for ( int i = 0; i < face; i++ )
    offset += int ( 0.5 * (order_facet[i][0] + 1) * (order_facet[i][0]+2) );

  int facendof = int ( 0.5 * (order_facet[face][0] + 1) * (order_facet[face][0]+2) );
  for ( int l = 0; l < facendof; l++ )
    fdofs.Append(offset + l);
}


void HDivSymTet :: CalcShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<6> shape) const   // n * 6
{
  AutoDiffDiff <3> sigma;

  AutoDiffDiff<3> x (ip(0) ,0);
  AutoDiffDiff<3> y (ip(1) ,1);
  AutoDiffDiff<3> z (ip(2) ,2); //(1-ip(0)-ip(1), 2); //
    
  AutoDiffDiff<3> lami[4] = {x, y, z, 1-x-y-z}; 



  double li[4] = { ip(0), ip(1), ip(2), 1-ip(0)-ip(1)-ip(2) };
    
  // int i, j, ii, k,l, m, n;
  shape = 0.0; 
  
  const FACE * faces = ElementTopology::GetFaces (ET_TET);
  
  int is, ie, iop, iop1, iop2;
    
  ArrayMem<AutoDiffDiff<3>, 20> uf(order+3), vf(order+3);

  // ********** BOUNDARY SHAPES ************ //
  
  int ii = 0;

  
  for (int fa = 0; fa < 4; fa++)
    {
      int fav[3] =  { faces[fa][0], faces[fa][1], faces[fa][2] };
      
      int p = order_facet[fa][0];
      //Sort vertices  first edge op minimal vertex
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      

      ScaledLegendrePolynomial (order_facet[fa][0]+1, lami[fav[1]]-lami[fav[2]], 1-lami[fav[0]], uf);
      LegendrePolynomial (order_facet[fa][0]+1, 2*lami[fav[0]]-1, vf);
      
      AutoDiffDiff<3> bub1 = 4 * lami[fav[1]] * lami[fav[2]];
      AutoDiffDiff<3> bub2 = lami[fav[0]]; 

      for (int i = 0; i <= p+1; i++)
	{
	  uf[i] *= bub1;
	  vf[i] *= bub2;
	} 
      
      for (int j = 1; j <= order_facet[fa][0]+1; j++)
	for (int k = 1; j+k <= order_facet[fa][0]+2; k++)
	  {
	    sigma = CurlCurl( uf[j-1], vf[k] );
	    AssignSymmetricTensor ( sigma, ii, shape );
//  	    *testout << sigma.DDValue(0,0) +sigma.DDValue(1,0) +sigma.DDValue(2,0) + 
//  	      sigma.DDValue(0,1) +sigma.DDValue(1,1) +sigma.DDValue(2,1) + 
//  	      sigma.DDValue(0,2) +sigma.DDValue(1,2) +sigma.DDValue(2,2)  << endl;
	    ii++;
	  }
    }
  int p;


  // ************ INNER SHAPES ****************** //



  int es = 0; int ee = 1; int et = 2; int ev=3;
  AutoDiffDiff<3> ls = lami[es];
  AutoDiffDiff<3> le = lami[ee];
  AutoDiffDiff<3> lt = lami[et];
  AutoDiffDiff<3> lv = lami[ev];

  ArrayMem<AutoDiffDiff<3>, 20> u(order_inner[0]+3), v(order_inner[0]+3), w(order_inner[0]+3);


  p = max2 ( 1, order_inner[0]) ;

  //  T_INNERSHAPES::CalcSplitted(p+4, ls-lv, le, lt, u, v, w );

  ScaledLegendrePolynomial (p, ls - le, ls+le, u);
  ScaledLegendrePolynomial (p, lt-ls-le, ls+le+lt, v);

  LegendrePolynomial (p, 2*lv-1, w);

  AutoDiffDiff<3> bub1 = 4 * le * ls;
  AutoDiffDiff<3> bub2 = lt; 
  AutoDiffDiff<3> bub3 = lv; 
  
  for (int i = 0; i <= p; i++)
    {
      u[i] *= bub1;
      v[i] *= bub2;
      w[i] *= bub3;
    } 
  
  
  for ( int i = 0; i <= order_inner[0]; i++ )
    for ( int j = 0; j <= order_inner[0]-i; j++ )
      for ( int k = 0; k <= order_inner[0]-i-j; k++ )
	{
	  AutoDiffDiff<3> vw = w[k]*v[j];
	  
	  sigma = CurlCurl(vw, u[i] ) ;
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	  
	}
  
  
  for ( int i = 0; i <= p; i++ )
    for ( int j = 1; j <= p-i; j++ )
      for ( int k = 0; k <= p-i-j; k++ )
	{
	  AutoDiffDiff <3> uv = u[i] * v[k];
	  AutoDiffDiff<3> uw = u[i]*w[k];
	  
	  if ( order_inner[0] > 0 )
	    {
	  sigma = CurlCurl ( uv, w[j] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	  sigma =  CurlCurl ( uw, v[j] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	    }

	  sigma = CrossCross ( u[i], w[j], v[k] ) - w[j].Value() * CurlCurl ( u[i], v[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
 
	  sigma = v[j].Value() * CurlCurl ( u[i], w[k] ) - 
	    CrossCross ( u[i], v[j], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
 	  ii++;
	}

  for ( int i = 0; i <= p; i++ )
    for ( int j = 1; j <= p-i; j++ )
      for ( int k = 1; k <= p-i-j; k++ )
	{
	  sigma =  CrossCross ( v[j], u[i], w[k] )  - u[i].Value() * CurlCurl ( v[j], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	}

  // shapes including the nedelec edge basis function
  // phi_e = nabla ls le - ls nabla le

  AutoDiffDiff<3> phi_e;

  phi_e.Value() = 0.0;
  for( int i = 0; i < 3; i++ )
    phi_e.DValue(i) = ls.DValue(i) * le.Value() - ls.Value() * le.DValue(i);

  for( int i = 0; i < 3; i++)
    for ( int j = 0; j < 3; j++ )
      phi_e.DDValue(i,j) = ls.DValue(i) * le.DValue(j) 	- ls.DValue(j) * le.DValue(i);

  
  for ( int j = 0; j <= p; j++ )
    for ( int k = 0; k <= p-j; k++ )
      {
 	sigma = CrossCross ( v[j], phi_e, w[k] );
 	AssignSymmetricTensor ( sigma, ii, shape );
 	ii++;
      }  
  
  for ( int j = 1; j <= p; j++ )
    for ( int k = 0; k <= p-j; k++ )
      {
 	sigma = CrossCross ( phi_e, v[j], w[k] );
 	AssignSymmetricTensor ( sigma, ii, shape );
 	ii++;
      }

}













// [P^2(x,y) x P^2(z)]^6  =   108
// sigma_nn = P^1   bot/top  ->  -2 * 3
//                  quad-faces -> -3 * 5
//                            == 87
// div sigma in [P^0]^3       -15

// quad-face ... 4    -> 12
// trig-face ... 3    -> 6
//  n-inner = 54

// inner,  sigma(x,y)-tensor  Trig-inner tensor P2(z)  -> 5 x 3 = 15
// inner,  sigma_xz, sigma_yz   voll P2                -> 2 x 18 = 36
// inner,  sigma_zz, P2 * z(1-z)                                    6
//                                                              = 57
// sigma_xx, sigma_yy, sigma_zz, sigma_xy, sigma_xz, sigma_yz


// first trig faces, then quads

  HDivSymPrism :: HDivSymPrism (int aorder, int aorder_facet, const int noncont) 
  : HDivSymFiniteElement<3> (ET_PRISM)
{
  nfacets = 5;

  order = max2(aorder, aorder_facet);

  for ( int i = 0; i < 6; i++ )
    order_facet[i] = aorder_facet;
  order_inner = aorder;

  ndof = int ( 3 * ((order_inner[0]+1)*(order_inner[0]+2)) *(order_inner[2]+1) ); 
  ninner = ndof - (order_inner[0]+1)*(order_inner[0]+2) -
    3*(order_inner[2]+1)*(order_inner[0]+1);

  ndof = ninner;
  for ( int i = 0; i < 2; i++)
    ndof += (order_facet[i][0]+1)*(order_facet[i][0]+2) / 2;
  for ( int i = 2; i < 5; i++)
    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1); 

  noncontinuous = noncont;

}

  void HDivSymPrism :: ComputeNDof ()
  {
  ndof = (order_inner[0]+1)*(order_inner[0]+2)*(order_inner[2]+1) * 3; 
  ninner = ndof - (order_inner[0]+1)*(order_inner[0]+2) -
    3*(order_inner[0]+1)*(order_inner[2]+1);

  ndof = ninner;
  for ( int i = 0; i < 2; i++)
    ndof += (order_facet[i][0]+1)*(order_facet[i][0]+2) / 2;
  for ( int i = 2; i < 5; i++)
    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1); 

  order = max2(order_inner[0], order_inner[2]);
  for ( int fa = 0; fa < 5; fa++ )
    if ( order < order_facet[fa][0] ) order = order_facet[fa][0];
  for ( int fa = 2; fa < 5; fa++ )
    if ( order < order_facet[fa][1] ) order = order_facet[fa][1];
  }

void HDivSymPrism :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  int offset = 0;
  fdofs.SetSize(0);

  int facendof;

  for ( int i = 0; i < face; i++ )
    {
      facendof = ( i < 2 ) ?  
	((order_facet[i][0]+1)*(order_facet[i][0]+2) / 2) :  
	((order_facet[i][0]+1)*(order_facet[i][1]+1));
      offset += facendof;
    }

  facendof = ( face < 2 ) ?  
	((order_facet[face][0]+1)*(order_facet[face][0]+2) / 2) :  
	((order_facet[face][0]+1)*(order_facet[face][1]+1));

  for ( int l = 0; l < facendof; l++ )
    fdofs.Append(offset + l);
}


void HDivSymPrism :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  if(noncontinuous)
    {
      for (int i = 0; i < ndof; i++)
	idofs.Append(i);
    }
  else
    {
      for (int i = ndof-ninner; i < ndof; i++)
	idofs.Append(i);
    }

}


void HDivSymPrism :: CalcShape (const IntegrationPoint & ip, 
				FlatMatrixFixWidth<6> shape) const   // n * 3 
{
  AutoDiff<2> x(ip(0),0); 
  AutoDiff<2> y(ip(1),1); 
  AutoDiff<1> z(ip(2),0);
  AutoDiff<2> lami[6] = {x,y,1-x-y, x, y, 1-x-y };
  
  // lami w.r.t trig surface, constant in z direction
  AutoDiffDiff<2> ddlami[6] = {x,y,1-x-y, x,y,1-x-y};
  // lami w.r.t z direction
  AutoDiffDiff<1> lamiz[6] = {1-z,1-z,1-z,z,z,z};
  
  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 

  //************* BOUNDARY SHAPES *************//
      
      // u,v -- scaled legendre/legendre polynomial for trig surface
      int maxorder_trigface = order_facet[0][0];
  
  ArrayMem<AutoDiffDiff<2>, 10 > uf(order+2), vf(order+3);
  // w   -- legendre polynomial for z direction
  ArrayMem<AutoDiffDiff<1>, 10 > wf(order+2);



  const FACE * faces = ElementTopology::GetFaces (ET_PRISM);

  // trig faces

  // curl curl (uv) (i,j) = 0  if (i,j) != (2,2)
  // curl curl (uv) (2,2) = 2*u10*v10 + 2*u01*v01 - u10*v01 - u01*v10 - u11*v00 - u00*v11;

  for (int fa = 0; fa < 2; fa++)
    {
      int fav[3] =  { faces[fa][0], faces[fa][1], faces[fa][2] };
      
      //Sort vertices  first edge op minimal vertex
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      
      AutoDiffDiff<2> ddbubxy = ddlami[fav[1]] * ddlami[fav[2]];

      ScaledLegendrePolynomial (order+1, ddlami[fav[1]]-ddlami[fav[2]], 1-ddlami[fav[0]], uf);
      LegendrePolynomial (order+2, 2*ddlami[fav[0]]-1, vf);
      
      for(int i = 0; i <= order_facet[fa][0]+1; i++)
	  uf[i] *= ddbubxy;
      for(int i = 0; i <= order_facet[fa][0]+2; i++)
	  vf[i] *= ddlami[fav[0]];
      

      // calculate curl curl ( nabla u nabla v^T )
      // all derivatives w.r.t z vanish
      // sigma (2,2) =  2*u10*v10 + 2*u01*v01 - u10*v01 - u01*v10 - u11*v00 - u00*v11;
      // sigma (i,j) = 0 else
       
      for (int j = 1; j <= order_facet[fa][0]+1; j++)
	for (int k = 1; j+k <= order_facet[fa][0]+2; k++)
	  {
	    // sigma = lamiz[fav[0]] * CurlCurl( uf[j-1], vf[k] );
	    shape(ii,2) = lamiz[fav[0]].Value() * (2 * uf[j-1].DDValue(1,0) * vf[k].DDValue(1,0) 
					   - uf[j-1].DDValue(1,1) * vf[k].DDValue(0,0) 
					   - uf[j-1].DDValue(0,0) * vf[k].DDValue(1,1) );
	    ii++;
	  }

    }
 
  // quad faces
  for ( int fa = 2; fa < 5; fa++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[fa][j]] > vnums[faces[fa][fmax]]) fmax = j;
	
      int fz, ftrig;

      fmax = faces[fa][fmax];
      fz = (fmax + 3) % 6;

      switch( fmax )
	{
	case 0: 
	  if ( fa == 2 ) ftrig = 2; else ftrig = 1;
	  break;
	case 1:
	  if ( fa == 3 ) ftrig = 0; else ftrig = 2;
	  break;
	case 2:
	  if ( fa == 4 ) ftrig = 1; else ftrig = 0;
	  break;
	case 3:
	  if ( fa == 2 ) ftrig = 5; else ftrig = 4;
	  break;
	case 4: 
	  if ( fa == 3 ) ftrig = 3; else ftrig = 5;
	  break;
	case 5:
	  if ( fa == 4 ) ftrig = 4; else ftrig = 3;
	  break;
	}

//       if ( fz > ftrig ) swap(fz, ftrig);


      AutoDiffDiff<2> bubxy = ddlami[fmax] * ddlami[ftrig];
      AutoDiffDiff<1> lz = lamiz[fmax] - lamiz[fz];
      AutoDiffDiff<1> bubz = lamiz[fmax] * lamiz[fz ];

      int orderz = order_facet[fa][1];

      LegendrePolynomial (orderz, lz , wf);
      for ( int i = 0; i <= orderz; i++)
	wf[i] *= bubz;

      ScaledLegendrePolynomial (order+1, ddlami[fmax]-ddlami[ftrig], ddlami[fmax]+ddlami[ftrig], uf);
      
      for(int i = 0; i <= order_facet[fa][0]; i++)
	{
	  uf[i] *= bubxy;
	}

      // shape function curl curl ( nabla u  nabla w^T )
      // contains only derivatives of w w.r.t z
      //                              v w.r.t x,y

      //     // diagonal 
      //     sigma.DDValue(0,0) = - u11*w22;;
      //     sigma.DDValue(1,1) = - u00*w22;
      //     sigma.DDValue(2,2) = 0.0;
      
      //     // off diagonal
      //     sigma.DDValue(0,1) = 0.5 * ( u01 + u10 ) *w22 ;
      //     sigma.DDValue(0,2) = 0.0;
      //     sigma.DDValue(1,2) = 0.0;
      
     for (int l = 0; l <= order_facet[fa][0]; l++)
       for (k = 0; k <= order_facet[fa][1]; k++)
	 {
	   shape(ii,0) = - uf[l].DDValue(1,1) * wf[k].DDValue(0,0);
	   shape(ii,1) = - uf[l].DDValue(0,0) * wf[k].DDValue(0,0);
	   shape(ii,3) = wf[k].DDValue(0,0) * uf[l].DDValue(0,1);
	   ii++;
	 }
    }

  // ************** INTERIOR SHAPES ***************//

  // inner shapes, sigma_xy - tensor,  3 x 5
  
  int orderz = order_inner[2];
  // u,v -- scaled legendre/legendre polynomial for trig surface
  ArrayMem<AutoDiffDiff<2>, 10 > u(order_inner[0]+2), v(order_inner[0]+3);
  // w   -- legendre polynomial for z direction
  ArrayMem<AutoDiffDiff<1>, 10 > w(orderz+3);

  AutoDiffDiff<2> ddbub = ddlami[0] * ddlami[1] * ddlami[2];
  AutoDiffDiff<2> ddbub1 = ddlami[0] * ddlami[1];
  AutoDiffDiff<2> ddbub2 = ddlami[2];
  AutoDiffDiff<1> ddbubz = lamiz[0]*lamiz[3];

  ScaledLegendrePolynomial (order_inner[0]+1, ddlami[0]-ddlami[1], ddlami[0]+ddlami[1], u);
  LegendrePolynomial (order_inner[0]+2, 1-2*ddlami[2], v);
  LegendrePolynomial (order_inner[2]+2, 1-2*lamiz[0], w);

  for ( int i = 0; i <= order_inner[0]+1; i++ )
    u[i] *= ddbub1;
  for ( int i = 0; i <= order_inner[0]+2; i++ )
    v[i] *= ddbub2;
  for ( int i = 0; i <= order_inner[2]+2; i++ )
    w[i] *= ddbubz;

  AutoDiffDiff<3> sigma = 0;

  for ( int k = 0; k <= order_inner[2]-2; k++ )
    for (int i = 0; i <= order_inner[0]; i++)
      for (int j = 1; j <= order_inner[0]-i+1; j++)
	{
	  shape(ii,2) = w[k].Value() * (2 * u[i].DDValue(1,0) * v[j].DDValue(1,0) 
					   - u[i].DDValue(1,1) * v[j].DDValue(0,0) 
					   - u[i].DDValue(0,0) * v[j].DDValue(1,1) );
	  ii++;
	}


  for ( int k = 0; k <= order_inner[2]; k++ )
    for (int i = 0; i <= order_inner[0]-1; i++)
      for (int j = 0; j <= order_inner[0]-1-i; j++)
	{
	  // v * curlcurl ( nabla u nabla w^T )
	  shape(ii,0) = v[j].Value() * (- u[i].DDValue(1,1) * w[k].DDValue(0,0));
	  shape(ii,1) = v[j].Value() * (- u[i].DDValue(0,0) * w[k].DDValue(0,0));
	  shape(ii,3) = v[j].Value() * (w[k].DDValue(0,0) * u[i].DDValue(0,1));
	  ii++;

	  // u * curlcurl ( nabla v nabla w^T )
	  if ( j == 0 ) continue;
	  shape(ii,0) = u[i].Value() * (- v[j].DDValue(1,1) * w[k].DDValue(0,0));
	  shape(ii,1) = u[i].Value() * (- v[j].DDValue(0,0) * w[k].DDValue(0,0));
	  shape(ii,3) = u[i].Value() * (w[k].DDValue(0,0) * v[j].DDValue(0,1));
	  ii++;
	}
  
  for ( int k = 0; k <= order_inner[2]-1; k++ )
    {
      for ( int i = 0; i <= order_inner[0]; i++)
	for ( int j = 0; j <= order_inner[0]-i; j++)
	  {
	    sigma = CrossCross_prismxy ( v[j], u[i], w[k] );
	    AssignSymmetricTensor ( sigma, ii, shape );
	    ii++;
	  }
      for ( int i = 0; i <= order_inner[0]; i++)
	for ( int j = 1; j <= order_inner[0]-i; j++)
	  {
	    sigma = CrossCross_prismxy ( u[i], v[j], w[k] );
	    AssignSymmetricTensor ( sigma, ii, shape );
	    ii++;
	  }
    }

  for ( int k = 0; k <= order_inner[2]; k++)
    for ( int i = 0; i <= order_inner[0]-1; i++ )
      for ( int j = 0; j <= order_inner[0]-1-i; j++)
	{
	  sigma = CrossCross_prismz ( u[i], w[k], v[j] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	}

  AutoDiffDiff<2> phi_e;

  orderz = order_inner[2];

  phi_e.Value() = 0.0;
  for( int i = 0; i < 2; i++ )
    phi_e.DValue(i) = ddlami[0].DValue(i) * ddlami[1].Value() - ddlami[0].Value() * ddlami[1].DValue(i);

  for( int i = 0; i < 2; i++)
    for ( int j = 0; j < 2; j++ )
      phi_e.DDValue(i,j) = ddlami[0].DValue(i) * ddlami[1].DValue(j) 	- ddlami[0].DValue(j) * ddlami[1].DValue(i);

  for ( int k = 0; k <= orderz-1; k++ )
    {
      for ( int j = 1; j <= order_inner[0]+1; j++)
	{
	  sigma = CrossCross_prismxy ( phi_e, v[j], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	}
    }

  
  for ( int k = 0; k <= orderz; k++)
    for ( int j = 0; j <= order_inner[0]; j++)
      {
	sigma = CrossCross_prismz ( phi_e, w[k], v[j] );
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;
      }

  // phi is nedelec edge function in z direction ... phi = (0,0,1)
  AutoDiffDiff<1> phi = 0;
  phi.DValue(0) = 1;
  for ( int i = 0; i <= order_inner[0]; i++)
    for ( int j = 0; j <= order_inner[0]-i; j++)
      {
	sigma = CrossCross_prismxy ( v[j], u[i], phi );
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;
      }

  for ( int i = 0; i <= order_inner[0]; i++)
    for ( int j = 1; j <= order_inner[0]-i; j++)
      {
	sigma = CrossCross_prismxy ( u[i], v[j], phi );
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;
      }
  
  return;
  for ( int k = 0; k <= orderz; k++)
    for ( int i = 0; i <= order_inner[0]-1; i++ )
      for ( int j = 0; j <= order_inner[0]-1-i; j++)
	{
	  sigma = CrossCross_prismz ( u[i], w[k], v[j] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;
	}

  return;

  /*
  for (int l = 0; l < 3; l++)
    {
      AutoDiffDiff<2> bub = ddlami[l] * ddlami[(l+1)%3];
      double lami = ddlami[(l+2)%3].Value();
      
      for (int i = 0; i <= order_inner-1; i++)
	for (int j = 0; i+j <= order_inner-1; j++)
	  {
	    double fac = lami * polx[i] * poly[j];

	    for (k = 0; k <= orderz; k++)
	      {
		shape(ii,0) = bub.DDValue(1,1) * fac;
		shape(ii,1) = bub.DDValue(0,0) * fac;
		shape(ii,3) = -bub.DDValue(1,0) * fac;
		fac *= 1-2*z.Value();
		ii++;
	      }
	  }
    }



  // sigma_xz, sigma_yz,  2 x (orderz+1) * (order_inner+1) * (order_inner+2) / 2
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      {
	double pot = polx[i] * poly[j];
	  
	for (int k = 0; k < orderz; k++)
	  {
	    shape(ii++,4) = pot;
	    shape(ii++,5) = pot;
	    pot *= 1-2*z.Value();
	  }
      }



  // sigma_zz,   6
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      {
	double pot = polx[i] * poly[j] * z.Value()*(1-z.Value());
	  
	for (int k = 0; k < orderz-1; k++)
	  {
	    shape(ii++,2) = pot;
	    pot *= 1-2*z.Value();
	  }
      }
  */
}












HDivSymHex :: HDivSymHex (int aorder, int aorder_facet, const int noncont) 
  : HDivSymFiniteElement<3> (ET_HEX)
{
  nfacets = 6;

  order = max2(aorder, aorder_facet);
  for ( int i = 0; i < 6; i++ )
    order_facet[i] = aorder_facet;
  order_inner = aorder;

  ndof = (order_inner[0]+1) * (order_inner[1]+1) * (order_inner[2]+1);
  ninner = ndof - 2*(order_inner[0]+1)*(order_inner[1]+1)
    - 2*(order_inner[2]+1)*(order_inner[1]+1)
    - 2*(order_inner[0]+1)*(order_inner[2]+1);

  ndof = ninner;
  for ( int i = 0; i < 6; i++ )
    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1);

  noncontinuous = noncont;
}

  void HDivSymHex :: ComputeNDof()
  {
  ndof = (order_inner[0]+1) * (order_inner[1]+1) * (order_inner[2]+1);
  ninner = ndof - 2*(order_inner[0]+1)*(order_inner[1]+1)
    - 2*(order_inner[2]+1)*(order_inner[1]+1)
    - 2*(order_inner[0]+1)*(order_inner[2]+1);

  ndof = ninner;
  for ( int i = 0; i < 6; i++ )
    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1);

  order = max2 ( order_inner[0], max2 (order_inner[1],order_inner[2]) );
  for ( int i = 0; i < 6; i++ )
    {
      if ( order < order_facet[i][0] ) order = order_facet[i][0];
      if ( order < order_facet[i][1] ) order = order_facet[i][1];
    }
  }

void HDivSymHex :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  if(noncontinuous)
    {
      for (int i = 0; i < ndof; i++)
	idofs.Append(i);
    }
  else
    {
      for (int i = ndof-ninner; i < ndof; i++)
	idofs.Append(i);
    }
}


void HDivSymHex :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  int offset = 0;
  fdofs.SetSize(0);

  for ( int i = 0; i < face; i++ )
    offset += (order_facet[i][0] + 1) * (order_facet[i][1]+1) ;

  int facendof = (order_facet[face][0] + 1) * (order_facet[face][1]+2) ;
  for ( int l = 0; l < facendof; l++ )
    fdofs.Append(offset + l);
}


void HDivSymHex :: CalcShape (const IntegrationPoint & ip, 
				FlatMatrixFixWidth<6> shape) const   // n * 3 
{
  AutoDiffDiff<3> x(ip(0),0); 
  AutoDiffDiff<3> y(ip(1),1); 
  AutoDiffDiff<3> z(ip(2),2);


  AutoDiffDiff<3> lami[8]={(1-x)*(1-y)*(1-z),x*(1-y)*(1-z),x*y*(1-z),(1-x)*y*(1-z),
			   (1-x)*(1-y)*z,x*(1-y)*z,x*y*z,(1-x)*y*z}; 
  shape = 0.0; 
    
  int ii = 0; 
  
  const FACE * faces = ElementTopology::GetFaces (ET_HEX);

  ArrayMem<AutoDiffDiff<3>, 20> u(order+3), v(order+3), w(order+3);

  AutoDiffDiff <3> sigma;


  //*********** BOUNDARY SHAPES ********//


    
  LegendrePolynomial (order_inner[0], 2*x-1, u);
  LegendrePolynomial (order_inner[1], 2*y-1, v);
  LegendrePolynomial (order_inner[2], 2*z-1, w);

  AutoDiffDiff<3> bub1 = x*(1-x);
  AutoDiffDiff<3> bub2 = y*(1-y); 
  AutoDiffDiff<3> bub3 = z*(1-z);

  for (int i = 0; i <= order_inner[0]; i++)
    {
      u[i] *= bub1;
    }  
  for (int i = 0; i <= order_inner[1]; i++)
    {
      v[i] *= bub2;
    }
  for (int i = 0; i <= order_inner[2]; i++)
    {
      w[i] *= bub3;
    } 

  // quad faces
  for (int i = 0; i < 6; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;

      int f1 = faces[i][(fmax+3)%4];
      int f2 = faces[i][(fmax+1)%4]; 
      fmax = faces[i][fmax]; 
      int frem = 6 - fmax - f1 - f2;
      
      if(vnums[f2]>vnums[f1]) swap(f1,f2);  // fmax > f1 > f2 

      // x(1-x) bubble
      LegendrePolynomial (order_facet[i][0], 2*(lami[fmax]+lami[f1]) - 1, u);
      // y(1-y) bubble
      LegendrePolynomial (order_facet[i][1], 2*(lami[fmax]+lami[f2]) - 1, v);

      AutoDiffDiff<3> zbub = 1 - (lami[faces[i][0]] + lami[faces[i][1]] + lami[faces[i][2]] + lami[faces[i][3]] );
      AutoDiffDiff<3> xbub = (lami[fmax]+lami[f1]) * ( 1-lami[fmax]-lami[f1] );
      AutoDiffDiff<3> ybub = (lami[fmax]+lami[f2]) * ( 1-lami[fmax]-lami[f2] );

      for (int j = 0; j <= order_facet[i][0]; j++)
	{
	  u[j] *= xbub;
	}
      for (int j = 0; j <= order_facet[i][1]; j++)
	{
	  v[j] *= ybub;
	} 
      
      for ( int j = 0; j <= order_facet[i][1]; j++ )
	for ( int k = 0; k <= order_facet[i][0]; k++ )
	  {
	    sigma = zbub * CurlCurl ( v[j], u[k] );
	    AssignSymmetricTensor ( sigma, ii, shape );
	    ii++;
	  }
    }

  // ********* INTERNAL SHAPES ****************** //

  int p = max2(order_inner[0], max2(order_inner[1], order_inner[2]) );

  LegendrePolynomial (p, 2*x-1, u);
  LegendrePolynomial (p, 2*y-1, v);
  LegendrePolynomial (p, 2*z-1, w);
  
  for (int i = 0; i <= p; i++)
    {
      u[i] *= bub1;
      v[i] *= bub2;
      w[i] *= bub3;
    } 

  for ( int i = 0; i <=order_inner[0]; i++ )
    for ( int j = 0; j <= order_inner[1]-1; j++ )
      for ( int k = 0; k <= order_inner[2]-1; k++ )
	{
	  sigma = CrossCross ( u[j], v[i], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	  sigma = CrossCross ( v[j], u[i], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	  sigma = CrossCross ( v[j], w[i], u[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	}

  for ( int i = 0; i <=p-2; i++ )
    for ( int j = 0; j <= p; j++ )
      for ( int k = 0; k <= p; k++ )
	{
	  sigma = u[i] * CurlCurl ( v[j], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	  sigma = v[i] * CurlCurl ( u[j], w[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	  sigma = w[i] * CurlCurl ( v[j], u[k] );
	  AssignSymmetricTensor ( sigma, ii, shape );
	  ii++;

	}

  // shapes using the unit vectors
  // e_i = nabla (x_i) (1-x_i) - x_i nabla (1-x_i)

  AutoDiffDiff<3> e_0, e_1, e_2;

  e_0 = 0.0;
  e_1 = 0.0;
  e_2 = 0.0;

  e_0.DValue(0) = 1.0;
  e_1.DValue(1) = 1.0;
  e_2.DValue(2) = 1.0;

 for ( int i = 0; i <=p; i++ )
    for ( int j = 0; j <= p-1; j++ )
      {
	sigma = CrossCross ( e_0, v[i], w[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_1, w[i], u[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_2, u[i], v[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_0, w[i], v[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_1, u[i], w[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_2, v[i], u[j]);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;
      }

  for ( int i = 0; i <= p; i++ )
    {
      sigma = CrossCross ( e_0, v[i], e_2);
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_1, w[i], e_0 );
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;

	sigma = CrossCross ( e_2, u[i], e_1 );
	AssignSymmetricTensor ( sigma, ii, shape );
	ii++;
  }
  
}






















void HDivSymNormalFiniteElement :: 
SetVertexNumbers (FlatArray<int> & avnums)
{
  for (int i = 0; i < avnums.Size(); i++)
    vnums[i] = avnums[i];
}




HDivSymNormalSegm :: HDivSymNormalSegm (int aorder)
{
  order = aorder;
  order_inner = aorder;
  eltype = ET_SEGM;
  ndof = aorder+1; 
}

  void HDivSymNormalSegm :: ComputeNDof()
  {
    ndof = order_inner[0]+1;
  }

void HDivSymNormalSegm :: CalcShape (const IntegrationPoint & ip, 
				  FlatVector<> shape) const   
{
  AutoDiff<1> x(ip(0),0); 
  AutoDiff<1> lami[3] = {x,1-x};
  AutoDiffDiff<1> ddlami[3] = {x,1-x};
    
  int es = 0, ee = 1;
  if (vnums[es] > vnums[ee])  swap (es, ee);

  AutoDiffDiff<1> xi = lami[ee]-lami[es];

  ArrayMem<AutoDiffDiff<1>, 20> ha(order+1);
  IntegratedLegendreMonomialExt::Calc (order_inner[0]+2, xi, ha);

  for (int l = 0; l <= order_inner[0]; l++)
    shape(l) = ha[l].DDValue(0,0);
}




HDivSymNormalTrig :: HDivSymNormalTrig (int aorder)
{
  order = aorder;
  order_inner = aorder;
  eltype = ET_TRIG;
  ndof = (aorder+1)*(aorder+2)/2;
}

  void HDivSymNormalTrig :: ComputeNDof()
  {
    ndof = (order_inner[0]+1)*(order_inner[0]+2) / 2;
  }

void HDivSymNormalTrig :: CalcShape (const IntegrationPoint & ip, 
				  FlatVector<> shape) const   
{
  shape = 0.0;

  double li[3] = { ip(0), ip(1), 1-ip(0)-ip(1) };

  ArrayMem<double,20> polxi(order_inner[0]+1), poleta(order_inner[0]+1);

  ScaledLegendrePolynomial (order_inner[0], li[0]-li[1], li[0]+li[1], polxi);
  LegendrePolynomial (order_inner[0], 2*li[2]-1, poleta);

  int ii = 0;
  for (int i = 0; i <= order_inner[0]; i++)
    for (int j = 0; j+i <= order_inner[0]; j++)
      shape(ii++) = polxi[i] * poleta[j];

  /*
  shape(0) = ip(0);
  shape(1) = ip(1);
  shape(2) = 1-ip(0)-ip(1);
  */
}





HDivSymNormalQuad :: HDivSymNormalQuad (int aorder)
{
  order = aorder;
  order_inner = aorder;
  eltype = ET_QUAD;
  ndof = sqr(order+1);
}

  void HDivSymNormalQuad :: ComputeNDof ()
  {
    ndof = (order_inner[0]+1)*(order_inner[1]+1);
  }
void HDivSymNormalQuad :: CalcShape (const IntegrationPoint & ip, 
				     FlatVector<> shape) const   
{
  ArrayMem<double,10> polx(order_inner[0]+1), poly(order_inner[1]+1);

  LegendrePolynomial (order_inner[0], 2*ip(0)-1, polx);
  LegendrePolynomial (order_inner[1], 2*ip(1)-1, poly);

  int ii = 0;
  for (int i = 0; i <= order_inner[0]; i++)
    for (int j = 0; j <= order_inner[1]; j++)
      shape(ii++) = polx[i] * poly[j];
}







}
