/*

tensor-valued, symmetric, normal-normal-continuous FESpace for  H(divdiv)

*/

#include <solve.hpp>
#ifdef PARALLEL
#include <parallelngs.hpp>
#endif

namespace ngcomp 
{
  using namespace ngcomp; 

#ifdef PARALLEL
  using namespace ngparallel;
#endif

using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace *************************** */




  HDivSymHighOrderFESpace :: HDivSymHighOrderFESpace (const MeshAccess & ama, const Flags & aflags,  bool parseflags)
    : FESpace (ama, aflags)
{
  name="HDivHighOrderFESpace(hdivho)";
  // allowed flags
  DefineNumFlag("order");
  DefineNumFlag("facetorder");
  DefineDefineFlag("hdivsym");
  DefineNumFlag("noncontinuous");
  
  if(parseflags) ParseFlags(aflags);


  uniform_order_facet = int( aflags.GetNumFlag ("facetorder", order));
  order_facet.SetSize(ma.GetNFacets());  
  order_facet = INT<2> (uniform_order_facet, uniform_order_facet);

  uniform_order_inner = int ( aflags.GetNumFlag("orderinner", order));
  order_inner.SetSize(ma.GetNE());
  order_inner = INT<3> (uniform_order_inner, uniform_order_inner, uniform_order_inner);

  orderz = int(aflags.GetNumFlag("orderz", -1.0));

  //noncontinuous = 0;   // for hybrid
  noncontinuous = int(aflags.GetNumFlag ("noncontinuous", 0));

  low_order_space = 0;

  complex_element = aflags.GetDefineFlag("sequence");

  int oi = order_inner[0][0];
  int of = order_facet[0][0];

  if (ma.GetDimension() == 2)
    {
      if (noncontinuous)
 	segm = new HDivSymNormalSegm (-1);
      else
	segm = new HDivSymNormalSegm (of);

      if ( complex_element )
	{
	  trig = new HDivSymTrig (oi, of, noncontinuous);
	  quad = new HDivSymQuad (oi, of, noncontinuous);
	}
      else
	{
	  trig = new HDivSymTrig (oi, of, noncontinuous);
	  quad = new HDivSymQuad (oi, of, noncontinuous);
	}      

    }
  else
    {
      if(noncontinuous)
	{
	  trig = new HDivSymNormalTrig(-1);
	  quad = new HDivSymNormalTrig(-1);
	}
      else
	{
	  trig = new HDivSymNormalTrig (of); //(1);
	  quad = new HDivSymNormalQuad (of);
	}

      if ( complex_element )
	{
	  prism = new HDivSymPrism (oi, of, noncontinuous);
	  tet = new HDivSymTet (oi, of, noncontinuous);
	  hex = new HDivSymHex (oi, of, noncontinuous);
	}
      else
	{
	  prism = new HDivSymPrism (oi, of, noncontinuous);
	  tet = new HDivSymTet (oi, of, noncontinuous);
	  hex = new HDivSymHex (oi, of, noncontinuous);
	}

    }

  ndof_level.SetSize(0);

  cout << "**********************+ HDivSym, order = " << order << endl;
  cout << "facetorder = " << uniform_order_facet << endl;
  if ( orderz >= 0 ) 
    cout << "anisotropic elements: orderz = " << orderz << endl;
  if ( noncontinuous ) cout << "noncontinuous space" << endl;

}

HDivSymHighOrderFESpace :: ~HDivSymHighOrderFESpace ()
{
  ;
}


void HDivSymHighOrderFESpace :: Update()
{

  nel = ma.GetNE();
  ndof = 0;


  INT<2> oldorderfacet = order_facet[0];
//   oldorderfacet[0]++;
//   oldorderfacet[1]++;
  order_facet.SetSize(ma.GetNFacets());  
  order_facet = oldorderfacet;

  INT<3> oldorderinner = order_inner[0];
//   oldorderinner[0]++;
//   oldorderinner[1]++;
//   oldorderinner[2]++;
  order_inner.SetSize(ma.GetNE());
  order_inner = oldorderinner;

//   order++;

  // set orders
  if ( orderz >= 0 &&  ma.GetDimension() == 2)
    {
      // set orders on quads
      nfa = ma.GetNEdges();
      ARRAY<int> elfacets;
      for (int el = 0; el < ma.GetNE(); el++)
	{
	  ma.GetElEdges(el, elfacets);
	  if ( ma.GetElType(el) == ET_TRIG ) continue;

	  for ( int i = 0; i < 4; i++ )
	    { 
	      int v1, v2;
	      Vec<2> point1, point2, normal;
	      ma.GetEdgePNums(elfacets[i], v1, v2);
	      ma.GetPoint(v1, point1);
	      ma.GetPoint(v2, point2);
	      normal(0) = point1(1) - point2(1);
	      normal(1) = point2(0) - point2(0);
	      normal /= L2Norm(normal);
	      if ( fabs(normal(1)) < 1e-1 )
		{
		  *testout << "HDIVSYM " << elfacets[i] << endl;
		    order_facet[elfacets[i]] = orderz;
		}
	    }
	}
      
      for ( int i = 0; i < ma.GetNE(); i++ )
	{
	  ArrayMem<int,12> vnums; 
	  ma.GetElVertices(i, vnums);

	  if ( ma.GetElType(i) == ET_TRIG ) continue;
	  Vec<2> point1, point2, normal;
	  ma.GetPoint(vnums[0], point1);
	  ma.GetPoint(vnums[1], point2);
	  normal(0) = point1(1) - point2(1);
	  normal(1) = point2(0) - point2(0);
	  normal /= L2Norm(normal);
	  if ( abs( normal(1) ) < 1e-1 )
	    {
	      order_inner[i][0] = orderz - 1;
	    }
	  else
	    order_inner[i][1] = orderz - 1;

	}

    }


  ndof = 0;

  ARRAY<int> elfacets;
  if (ma.GetDimension() == 2)
    {
      nfa = ma.GetNEdges();

      first_face_dof.SetSize(nfa+1);
      for (int i = 0; i < nfa; i++)
	{
	  first_face_dof[i] = ndof;
	  ndof += order_facet[i][0]+1 ; //  + 1;
	}
      first_face_dof[nfa] = ndof;    

      fine_face.SetSize(nfa); 
      fine_face = 0; 
      for ( int i = 0; i < nel; i++)
	{
	  ma.GetElEdges (i, elfacets);		
	  for(int j=0;j<elfacets.Size();j++) fine_face[elfacets[j]] = 1; 

	}  
    }
  else
    {
      nfa = ma.GetNFaces();
      first_face_dof.SetSize(nfa+1);
      ARRAY<int> pnums;

      cout << "nfa " << nfa << endl;
      for (int i = 0; i < nfa; i++)
	{
	  first_face_dof[i] = ndof;

	  ma.GetFacePNums(i, pnums);

	  if (pnums.Size() == 3)
	    ndof += (order_facet[i][0]+1)*(order_facet[i][0]+2)/2;
	  else
	    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
	}
      first_face_dof[nfa] = ndof;    

      fine_face.SetSize(nfa); 
      fine_face = 0; 
      
      for ( int i = 0; i < nel; i++)
	{
	  ma.GetElFaces (i, elfacets);		
	  for(int j=0;j<elfacets.Size();j++) fine_face[elfacets[j]] = 1; 
	   
	}  

    }

  // (*testout) << "first face dofs = " << endl << first_face_dof << endl;

  first_element_dof.SetSize(nel+1);
  for (int i = 0; i < nel; i++)
    {
      first_element_dof[i] = ndof;
      int oi0 = order_inner[i][0];
      int oiy;
      int oiz;
      switch (ma.GetElType(i))
	{
	case ET_TRIG:
	  ndof += 3*oi0*(oi0+1)/2 ;
	  break;
	case ET_QUAD:
	  oiz = order_inner[i][1];
	  ndof += 3 * (oi0+1) * (oiz+1) - 2;
	  // 2*(oi0+1) - 2*(oiz+1);
	  break;
	case ET_PRISM:
	  oiz = order_inner[i][2];
	  ndof += 3 * (oi0+1)*(oi0+2)*(oiz+1) - 
	    (oi0+1)*(oi0+2) -
	    3*(oiz+1)*(oi0+1);
	  break;
	case ET_TET:
	  ndof += (oi0+1)*(oi0+2)*(oi0+3) - 
	    2 * (oi0+1)*(oi0+2);
	  break;
	case ET_HEX:
	  oiy = order_inner[i][1]; 
	  oiz = order_inner[i][2];
	  ndof += (oi0+1) * (oiy+1) * (oiz+1) - 2*(oi0+1)*(oiy+1)
	    - 2*(oiz+1)*(oi0+1)
	    - 2*(oiy+1)*(oiz+1);
	  break;
	}
    }

  first_element_dof[nel] = ndof;    


  if (noncontinuous)
    {
      first_face_dof = 0;
      ndof = 0;
      
      ARRAY<int> facets, pnums;
      int of0, ofz;
      int oi0;
      int oiy;
      int oiz;

      for (int i = 0; i < nel; i++)
	{
	  first_element_dof[i] = ndof;
	  ma.GetElFacets(i, facets);

	  // add facet dofs
	  for ( int fa = 0; fa < facets.Size(); fa++ )
	    {
	      if ( ma.GetDimension() == 2 )
		{
		  ndof += order_facet[facets[fa]][0] +1;	  
		  //(*testout) << "order facet " << order_facet[facets[fa]][0] << endl;
		}
	      else 
		{
		  of0 = order_facet[facets[fa]][0];
		  ma.GetFacePNums(facets[fa], pnums);  
		  if ( pnums.Size() == 3 )
		    ndof += (of0+1)*(of0+2)/2;
		  else
		    {
		      ofz = order_facet[facets[fa]][1];
		      ndof += ( ofz+1 ) * (of0 + 1);
		    }
		}
	      
	    }
	  
	  //  *testout << "element " << i << " has " << ndof << " facet dofs" << endl;

	  oi0 = order_inner[i][0];
	  // add inner dofs
	  switch (ma.GetElType(i))
	    {
	    case ET_TRIG:
	      ndof += 3*oi0*(oi0+1)/2 ;
	      break;
	    case ET_QUAD:
	      oiz = order_inner[i][1];
	      ndof += 3 * (oi0+1) * (oiz+1) - 2;
	      // 2*(oi0+1) - 2*(oiz+1);
	      break;
	    case ET_PRISM:
	      oiz = order_inner[i][2];
	      ndof += 3 * (oi0+1)*(oi0+2)*(oiz+1) - 
		(oi0+1)*(oi0+2) -
		3*(oiz+1)*(oi0+1);
	      break;
	    case ET_TET:
	      ndof += (oi0+1)*(oi0+2)*(oi0+3) - 
		2 * (oi0+1)*(oi0+2);
	      break;
	    case ET_HEX:
	      oiy = order_inner[i][1]; 
	      oiz = order_inner[i][2];
	      ndof += (oi0+1) * (oiy+1) * (oiz+1) - 2*(oi0+1)*(oiy+1)
		- 2*(oiz+1)*(oi0+1)
		- 2*(oiy+1)*(oiz+1);
	      break;
	    }
	  //*testout << "element " << i << " has " << ndof << " dofs" << endl;

	}
      first_element_dof[nel] = ndof;    
    }
  
  (*testout) << "HDIVSYM ndof " << ndof << endl; 
  (*testout) << "number of elements " << ma.GetNE() << endl;
  (*testout) << "number of facets " << ma.GetNFacets() << endl;
  (*testout) << "ndof " << ndof << " =  9 * " << ma.GetNE() << " = " << 9 * ma.GetNE() << endl;
  /*
  (*testout) << "first face dofs = " << endl << first_face_dof << endl;
  (*testout) << "first element dofs = " << endl << first_element_dof << endl;
  (*testout) << "done " << endl;
  */

	  ndof_level.Append(ndof);

#ifdef PARALLEL
    UpdateParallelDofs();
#endif

}



const FiniteElement & HDivSymHighOrderFESpace :: GetFE (int elnr, LocalHeap & lh) const
{
  HDivSymFiniteElement<2> * fe2d = 0;
  HDivSymFiniteElement<3> * fe3d = 0;
 
  if (ma.GetElType(elnr) == ET_TRIG)
    fe2d = dynamic_cast<HDivSymFiniteElement<2>*> (trig);
  else if (ma.GetElType(elnr) == ET_QUAD)
    fe2d = dynamic_cast<HDivSymFiniteElement<2>*> (quad);
  else if (ma.GetElType(elnr) == ET_TET)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (tet);
  else if (ma.GetElType(elnr) == ET_PRISM)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (prism);
  else if (ma.GetElType(elnr) == ET_HEX)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (hex);
  


  ArrayMem<int,12> vnums; 
  ma.GetElVertices(elnr, vnums);
  
  if (fe2d)
    {
      ArrayMem<int, 12> ednums;
      ARRAY<INT<2> > order_ed;
      
      ma.GetElEdges(elnr, ednums);
      order_ed.SetSize (ednums.Size());
      for (int j = 0; j < ednums.Size(); j++)
	order_ed[j] = order_facet[ednums[j]];

      fe2d -> SetVertexNumbers (vnums);
      fe2d -> SetOrderFacet (order_ed);
      fe2d -> SetOrderInner (order_inner[elnr]);
      fe2d -> ComputeNDof ();
      return *fe2d;
    }
  if (fe3d)
    {
      ArrayMem<int, 6> fanums; 
      ARRAY<INT<2> > order_fa;
      ma.GetElFaces(elnr, fanums);
      order_fa.SetSize (fanums.Size());
      
      for (int j = 0; j < fanums.Size(); j++)
	order_fa[j] = order_facet[fanums[j]];
      
      fe3d -> SetVertexNumbers (vnums);
	
      fe3d -> SetOrderFacet (order_fa);
      fe3d -> SetOrderInner (order_inner[elnr]);
      
      fe3d -> ComputeNDof ();
      return *fe3d;
    }
  
}

const FiniteElement & HDivSymHighOrderFESpace :: GetSFE (int elnr, LocalHeap & lh) const
{

  HDivSymNormalFiniteElement * fe;

  if (ma.GetSElType(elnr) == ET_SEGM)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (segm);
  else if (ma.GetSElType(elnr) == ET_TRIG)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (trig);
  else if (ma.GetSElType(elnr) == ET_QUAD)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (quad);


  ArrayMem<int,12> vnums, ednums; 
  ma.GetSElVertices(elnr, vnums);
  fe -> SetVertexNumbers (vnums);

  if ( noncontinuous ) 
    {
      fe->SetOrderInner ( -1 );
      fe->ComputeNDof();
      return *fe;
    }

  if(ma.GetSElType(elnr) == ET_SEGM)
    {
      ma.GetSElEdges(elnr, ednums);
	
      fe -> SetOrderInner (order_facet[ednums[0]]);
      fe -> ComputeNDof();
    }
  else
    {
      INT<2> order_fa = order_facet[ma.GetSElFace(elnr)];
      fe -> SetOrderInner (order_fa);
      fe -> ComputeNDof();
    }

  return *fe;
}



void HDivSymHighOrderFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
{   
  dnums.SetSize(0);

  ArrayMem<int,12> vnums, ednums, fanums; 
  int i, j;
  int first, next; 

  if (ma.GetDimension() == 2)
    ma.GetElEdges (elnr, ednums);
  else
    ma.GetElFaces (elnr, ednums);

  for (i = 0; i < ednums.Size(); i++)
    {
      first = first_face_dof[ednums[i]];
      next = first_face_dof[ednums[i]+1];
      for (j = first; j < next; j++)
	dnums.Append (j);
    }

//   (*testout) << "after faces sigma_dnums.size " << dnums.Size() << endl;

  first = first_element_dof[elnr];
  next = first_element_dof[elnr+1];
  for (int j = first; j < next; j++)
    dnums.Append (j);

//   *testout << " GetElDNums sigma elnr " << elnr << " \t " << dnums << endl; 

}



  void HDivSymHighOrderFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }
    
    dnums.SetSize(0);
    
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 
    
    if (ma.GetDimension() == 2)
      ma.GetElEdges (elnr, ednums);
    else
      ma.GetElFaces (elnr, ednums);
    
    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_face_dof[ednums[i]];
	next = first_face_dof[ednums[i]+1];
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }


  }
 

void HDivSymHighOrderFESpace :: GetInternalDofs (int elnr, ARRAY<int> & dnums ) const
{
  dnums.SetSize(0);
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 
    
  first = first_element_dof[elnr];
  next = first_element_dof[elnr+1];
  for (int j = first; j < next; j++)
    dnums.Append (j);



}







void HDivSymHighOrderFESpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
{
  dnums.SetSize (0);

    
  ArrayMem<int,12> vnums, ednums, fanums; 
  int i, j;
  int first, next; 

  if (ma.GetDimension() == 2)
    ma.GetSElEdges (elnr, ednums);
  else
    {
      ednums.SetSize(1);
      ednums[0] = ma.GetSElFace (elnr);
    }

    
  for (i = 0; i < ednums.Size(); i++)
    {
      first = first_face_dof[ednums[i]];
      next = first_face_dof[ednums[i]+1];
      for (j = first; j < next; j++)
	dnums.Append (j);
    }
}




#ifdef PARALLEL

void HDivSymHighOrderFESpace :: UpdateParallelDofs_hoproc()
  {
    
     // ******************************
     // update exchange dofs 
     // ******************************
    *testout << "HDivSymHO::UpdateParallelDofs_hoproc" << endl;
    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks);
    nexdof = 0;


    // noncontinuous case: no exchange between processes
    if ( noncontinuous )
      {
	paralleldofs->SetNExDof(nexdof);
	
	paralleldofs->localexchangedof = new Table<int> (nexdof);
	paralleldofs->distantexchangedof = new Table<int> (nexdof);
	return;
      }

    MPI_Status status;
    MPI_Request * sendrequest = new MPI_Request[ntasks];
    MPI_Request * recvrequest = new MPI_Request[ntasks];

    // number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	if ( !parallelma->IsExchangeFace ( face ) ) continue;
	
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	nexdof[id] += dnums.Size() ; 

	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    nexdof[dest] += dnums.Size() ; 
      }

    // + number of inner exchange dofs --> set eliminate internal

    nexdof[0] = ma.GetNFaces();

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

    ARRAY<int> ** owndofs, ** distantdofs;
    owndofs = new ARRAY<int>* [ntasks];
    distantdofs = new ARRAY<int>* [ntasks];

    for ( int i = 0; i < ntasks; i++ )
      {
	owndofs[i] = new ARRAY<int>(1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int>(0);
      }

    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;
    int exdof = 0;
    int ii;

    // *****************
    // Parallel Face dofs
    // *****************


   for ( int face = 0; face < ma.GetNFaces(); face++ )
      if ( parallelma->IsExchangeFace ( face ) )
      {
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;

	    owndofs[dest]->Append ( distface );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int sender = 1; sender < ntasks; sender ++ )
      {
        if ( id == sender )
          for ( int dest = 1; dest < ntasks; dest ++ )
            if ( dest != id )
              {
		MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
              }
	  
        if ( id != sender )
          {
	    MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender]);
          }
 	  
	  
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait(recvrequest+dest, &status);
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;

	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    ii++; 
	    for ( int i=1; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }



     /*******************************

         update low order space

     *****************************/


     int ndof_lo = ma.GetNFaces();

     // all dofs are exchange dofs
     nexdof = ndof_lo;
     
     exdof = 0;
     cnt_nexdof = 0;
     

     // *****************
     // Parallel Face dofs
     // *****************
     
     owndofs[0]->SetSize(1);
     (*owndofs[0])[0] = ndof;
     distantdofs[0]->SetSize(0);
     
     // find local and distant dof numbers for face exchange dofs
     for ( int face = 0; face < ma.GetNFaces(); face++ )
       {
	 int dest = 0;
	 
	 ARRAY<int> dnums;
	 GetFaceDofNrs ( face, dnums );
	 int distface = parallelma -> GetDistantFaceNum ( dest, face );
	 owndofs[0]->Append ( distface );
	 paralleldofs->SetExchangeDof ( dest, dnums[0] );
	 owndofs[dest]->Append ( dnums[0] );
      }   
    
    int dest = 0;
    MyMPI_ISend ( *owndofs[0], dest, sendrequest[dest]);
    MyMPI_IRecv ( *distantdofs[0], dest, recvrequest[dest]);
   
    MPI_Wait ( recvrequest+dest, &status);


    // denk nach und machs gscheit...

    
    paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
    ii = 1;
    while ( ii < distantdofs[0]->Size() )
      {
	int fanum = (*distantdofs[0])[ii++];
	int dnum = (*distantdofs[0])[ii++];
	ARRAY<int> dnums;
	GetFaceDofNrs (fanum, dnums);
	(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[0];
	(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = dnum;
	cnt_nexdof[dest]++;
      }
    
    for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;

  }

  void HDivSymHighOrderFESpace :: UpdateParallelDofs_loproc()
  {
    *testout << "HDivHOFESpace::UpdateParallelDofs_loproc" << endl;

    const MeshAccess & ma = (*this). GetMeshAccess();

    int ndof = GetNDof();

    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks); 
    nexdof = 0;

    // noncontinuous case: no exchange between processes
    if ( noncontinuous )
      {
	paralleldofs->SetNExDof(nexdof);
	
	paralleldofs->localexchangedof = new Table<int> (nexdof);
	paralleldofs->distantexchangedof = new Table<int> (nexdof);
	return;
      }


    MPI_Status status;
    MPI_Request * sendrequest = new MPI_Request[ntasks];
    MPI_Request * recvrequest = new MPI_Request[ntasks];

    // number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	nexdof[id] ++;//= dnums.Size() ; 
	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    { 
	      nexdof[dest] ++; 
	    }
      }

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);



    ARRAY<int> ** owndofs,** distantdofs;
    owndofs = new ARRAY<int> * [ntasks];
    distantdofs = new ARRAY<int> * [ntasks];

    for ( int i = 0; i < ntasks; i++)
      {
	owndofs[i] = new ARRAY<int> (1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int> (0);
      }

    int exdof = 0;
    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;

    // *****************
    // Parallel Face dofs
    // *****************


    // find local and distant dof numbers for face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	(*(paralleldofs->localexchangedof))[id][exdof++] = face;

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;
	    ARRAY<int> dnums;
	    GetFaceDofNrs ( face, dnums );
	    owndofs[dest]->Append ( distface );
	    paralleldofs->SetExchangeDof ( dest, dnums[0] );
	    paralleldofs->SetExchangeDof ( dnums[0] );
	    owndofs[dest]->Append (dnums[0]);
  	  }
      }   


    for ( int dest = 1; dest < ntasks; dest ++ )
      {
	MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
	MyMPI_IRecv ( *distantdofs[dest], dest, recvrequest[dest]);
      }
    


    int ii = 1;
    for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest+dest, &status );
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int dnum = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    (*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[0];
	    (*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = dnum;
	    cnt_nexdof[dest]++;
	     
	  }
      }
     for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;
 
  }
#endif // PARALLEL



  // register FESpaces
  namespace
#ifdef MACOS
  hdivsymhofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hdivsym", HDivSymHighOrderFESpace::Create);
    }
    
#ifdef MACOS
    static
#endif
    Init init;
  }
  
}



