#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

#include <helmholtz_single_exp.hpp>



extern ngsolve::MeshAccess * ma;
extern ngsolve::PDE * pde;

namespace helmholtz_exp_cpp
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace helmholtz_exp_cpp;

  
  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */



  ////////////////////////////////////////////////////////////////////////////////


template <int D>
void HybridHelmholtzIntegrator<D> :: AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      int elnum = eltrans.GetElementNr();
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");
      clock_t start_add, end_add, time_add = 0;

      NgProfiler::RegionTimer reg (timer);

      const HybridHelmholtzFiniteElement & cfel = 
        dynamic_cast<const HybridHelmholtzFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const FacetVolumeFiniteElement & fel_hdivn = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[1]);

      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();

      int base_hdiv = 0;
      int base_hdivn = nd_hdiv;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      enum { BLOCK = 5} ;
      
      FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      
      FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh), scalbmat_sigma(nd_hdiv, lh)  ; 
      FlatVector<> div_sigma(nd_hdiv, lh), scaldiv_sigma(nd_hdiv, lh) ; // (nd_hdiv, lh);
      FlatVector<> div_sigma2(nd_hdiv, lh), scaldiv_sigma2(nd_hdiv, lh) ; // (nd_hdiv, lh);
      
      FlatMatrixFixHeight<3, Complex> bmat(nd, lh), bmat2(nd, lh);
      FlatMatrixFixHeight<3, Complex> dbmat(nd, lh);
      Mat<3, 3, Complex> dmat, dmat2;

      // int ik sigma tau dx
      FlatMatrix<> mat_iksigmasigma (nd_hdiv, lh);
      // int 1/ik div sigma div tau dx
      FlatMatrix<> mat_iikdivdiv_real (nd_hdiv, lh), mat_iikdivdiv_imag (nd_hdiv, lh);

      FlatMatrixFixHeight<D> diag_iksigmasigma (nd_hdiv, lh);

      Vec<D> kvec;
      double kappa = coef_kappa->EvaluateConst();
          
      bool const_kvec =
	dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec0 ) && 
	dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec1 ) && 
	dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec2 ); 

      const IntegrationRule & irDd =
        SelectIntegrationRule (eltype, 2*fel_hdiv.Order() );
      
      int with_exp = 0;
      if ( pde -> ConstantUsed ("with_exp") )
	with_exp = int(pde->GetConstant("with_exp") );

      void * heapp = lh.GetPointer();
      mat_iksigmasigma = 0.0;
      mat_iikdivdiv_real = 0.0;
      mat_iikdivdiv_imag = 0.0;
      diag_iksigmasigma = 0.0;

      bool curvedelement = true; // eltrans.IsCurvedElement();
      if ( curvedelement || with_exp )      
	{
      for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      IntegrationPoint ip0 (0., 1.);
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      kvec(0) = coef_kvec0->Evaluate(sip);
	      kvec(1) = coef_kvec1->Evaluate(sip);
	      if ( D == 3 )
		kvec(2) = coef_kvec2->Evaluate(sip);


	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double det = sip.GetJacobiDet();
	      double invdet = 1.0 / det;
	      Mat<D> jac = sip.GetJacobian();
	      Mat<D> invjac = sip.GetJacobianInverse();
	      Vec<D> point;
	      eltrans.CalcPoint(sip.IP(), point, lh);

	      bmat_sigma = invdet
		* jac * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
    
	      // HACK for sigma = s e^ikx  
	      // compute (ik.x)' . s
	      Vec<D> d_kx, d_kx_ref;
	      Mat<D> dk, dkref;

	      if ( with_exp )
		{
		  if ( const_kvec )
		    {
		      d_kx = kvec;
		    }
		  else
		    {
		      for ( int d = 0; d < D; d++ )
			{
			  IntegrationPoint ip1, ip2;
			  ip1 = sip.IP(); ip2 = sip.IP();
			  ip1(d) += 1e-6;
			  ip2(d) -= 1e-6;
			  SpecificIntegrationPoint<D,D> sip1 (ip1, eltrans, lh);
			  SpecificIntegrationPoint<D,D> sip2 (ip2, eltrans, lh);
			  dkref(d, 0) = 0.5e6 * 
			    (coef_kvec0->Evaluate(sip1) - coef_kvec0->Evaluate(sip2));
			  dkref(d, 1) = 0.5e6 * 
			    (coef_kvec1->Evaluate(sip1) - coef_kvec1->Evaluate(sip2));
			  if ( D == 3 )
			    dkref(d, 2) = 0.5e6 * 
			      (coef_kvec2->Evaluate(sip1) - coef_kvec2->Evaluate(sip2));
			  
			}
		      dk = Trans(invjac) * dkref;
		      d_kx =  dk * point + kvec;
		  
		    }
		  d_kx_ref = Trans(jac) * d_kx;
		}

// 	      *testout << "dk/dxi " << dkref << endl << "dx/dxi" << invjac << endl << "dxi/dx" << jac << endl;

// 	      Vec<D> kvec2 = (kappa / L2Norm(point) ) * point;
// 	      *testout << "integration point " << point << endl << "kvec " << kvec << endl
// 		       << "kvec_ref " << kvec2 << endl
// 		       << "d_k . x " << d_kx << endl << "with L2Norm " << L2Norm(d_kx) << endl;;
// 	      *testout << "identity * k = " << dk << endl;
// 	      *testout << " ****************************************** " << endl;


	      scalbmat_sigma = kappafactor * bmat_sigma;
 	      scaldiv_sigma = ikappafactor * div_sigma;

	      mat_iikdivdiv_imag += (scaldiv_sigma * Trans (div_sigma));
	      if ( with_exp )
		{
		  div_sigma2 = Trans(bmat_sigma) * d_kx;
		  scaldiv_sigma2 = ikappafactor * div_sigma2;

		  mat_iikdivdiv_imag += (scaldiv_sigma2 * Trans ( div_sigma2 ) );
		  mat_iikdivdiv_real += (scaldiv_sigma2 * Trans(div_sigma));
		  mat_iikdivdiv_real -= (scaldiv_sigma * Trans(div_sigma2));
		}
	      mat_iksigmasigma += Trans(scalbmat_sigma) * (bmat_sigma);
	      lh.CleanUp(heapp);
	    }

	  //	  *testout << "mat_iksigmasigma " << endl << mat_iksigmasigma << endl;
	  // int i k sigma . tau
      if ( curvedelement )
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      imag(elmat(base_hdiv+i, base_hdiv+j)) +=  mat_iksigmasigma(i,j);

// 	  *testout << "mat_iksigmasigma( " << elnum << ") " << endl << mat_iksigmasigma << endl;

	  // int 1/ik divsigma divtau == - int i/k divsigma divtau 
      if ( curvedelement )
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv_imag(i,j);
	      }
	  if ( with_exp )
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		real(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv_real(i,j);
	      }
	}

      else // FAST

	{
	  
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      kvec(0) = coef_kvec0->Evaluate(sip);
	      kvec(1) = coef_kvec1->Evaluate(sip);
	      if ( D == 3 )
		kvec(2) = coef_kvec2->Evaluate(sip);
// 	      kappa = L2Norm(kvec);

	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
	      
	      

	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	       // mass for l2^3 is diagonal
	      for ( int ii = 0; ii < nd_hdiv / D ; ii ++ )
		{
		  for ( int mi = 0; mi < D; mi++ )
		    for ( int mj = 0; mj < D; mj++ )
		      for ( int j = 0; j < D; j++)
			diag_iksigmasigma(mi, D*ii + mj) += scalbmat_sigma(j, D*ii+mi) * bmat_sigma(j, D*ii+mj);
		}
	       
	      mat_iikdivdiv_imag += (scaldiv_sigma * Trans (div_sigma));

	      lh.CleanUp(heapp);
	    }
	  
	  
	  // int i k sigma . tau
      if ( !curvedelement )
	  for (int i = 0; i < nd_hdiv / D; i++)
	    for ( int mi = 0; mi < D; mi++ )
	      for ( int mj = 0; mj < D; mj++ )
		imag(elmat(base_hdiv+i*D+mi, base_hdiv+i*D+mj)) +=  diag_iksigmasigma(mi,D*i+mj);
	  
	 	  
	  // - int i / k sigma . tau  --> diagonal
	  if ( ! curvedelement )
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv_imag(i,j);
	      }
	   
	  lh.CleanUp(heapp);
	  
  	}
      
      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     
      
      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;


      FlatMatrix<> mat_sigmansigman (int(nd_hdiv/D), lh);
      mat_sigmansigman = 0.0;
      mat_iksigmasigma = 0.0;

      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

      for (int k = 0; k < ned; k++)
        {
	  mat_sigmansigman = 0.0;
	  Vec<D> normal_ref;
	  for (int i=0; i<D; i++)
	    normal_ref(i) = normals[k][i];
        
          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*fel_hdiv.Order() );

          void * heapp = lh.GetPointer();

	  Mat<D> ninj;
	  ninj = 0.0;
          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
	      IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

	      kvec(0) = coef_kvec0->Evaluate(sip);
	      kvec(1) = coef_kvec1->Evaluate(sip);
	      if ( D == 3 )
		kvec(2) = coef_kvec2->Evaluate(sip);

              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();
	      double invdet = 1.0 / det;

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

	      Vec<D> point;
	      eltrans.CalcPoint ( sip.IP(), point, lh);
              
              Vec<D> nt = Trans(jac) * normal;
              nt /= det;

	      // calc the tensor product n n^T for fast evaluation of (sigma . n) (tau . n)
	      if ( l == 0 )
		for (int i = 0; i < D; i++ )
		  for ( int j = 0; j < D; j++ )
		    ninj(i,j) = nt(i) * nt(j);
              //Piola

	      // Vec<3> e1;
	      // e1 = 1.0;
	      bmat_sigma = Trans( fel_hdiv.GetShape(sip.IP(), lh) );
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
              mat_hdiv_n = Trans ( bmat_sigma ) * nt;
	      mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh);


	      // compute (ik.x)' . s
	      Vec<D> d_kx, d_kx_ref;
	      Mat<D> dk, dkref;
	      if ( const_kvec )
		{
		  d_kx = kvec;
		}
	      else
		{
		  for ( int d = 0; d < D; d++ )
		    {
		      IntegrationPoint ip1, ip2;
		      ip1 = sip.IP(); ip2 = sip.IP();
		      ip1(d) += 1e-6;
		      ip2(d) -= 1e-6;
		      SpecificIntegrationPoint<D,D> sip1 (ip1, eltrans, lh);
		      SpecificIntegrationPoint<D,D> sip2 (ip2, eltrans, lh);
		      dkref(d, 0) = 0.5e6 * 
			(coef_kvec0->Evaluate(sip1) - coef_kvec0->Evaluate(sip2));
		      dkref(d, 1) = 0.5e6 * 
			(coef_kvec1->Evaluate(sip1) - coef_kvec1->Evaluate(sip2));
		      if ( D == 3 )
			dkref(d, 2) = 0.5e6 * 
			  (coef_kvec2->Evaluate(sip1) - coef_kvec2->Evaluate(sip2));
		      
		    }
		  dk = Trans(inv_jac) * dkref;
		  d_kx =  dk * point + kvec;
		  
		}
	      
              d_kx_ref = Trans(jac) * d_kx;

	      bmat_sigma = invdet
		* jac * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
	      div_sigma2 = Trans(bmat_sigma) * d_kx;
// 	      div_sigma2 *= Complex(0., 1.);
// 	      div_sigma2 += div_sigma;
// 	      div_sigma2 = div_sigma;

              double factor = len * ir1d[l].Weight();

	      // penalty ... high on master, 0 on slave
	      double penalty = 1e7;
	      
	      double nx = 0.0;

	      for ( int i = 0; i < D; i++ )
		nx += kvec(i) * normal(i);

	      nx /= L2Norm(kvec);
	      nx /= L2Norm(normal);

	      if ( nx < 0 ) penalty = 0;
	      // isslave = 0 --> im smoothing block
// 	      if ( nx <= -0.0 ) const_cast<HybridHelmholtzFiniteElement&> (cfel) .SetSlave(k, 0); // [elnum][k] = 0;



 	      double fac = 1.0;
 	      if ( nx < 0. ) fac = -1.0;
//               bmat = 0.0;
//               bmat.Row(0).Range (base_hdiv, base_hdiv +nd_hdiv) = div_sigma;
//               bmat.Row(2).Range (base_hdiv, base_hdiv +nd_hdiv) = mat_hdiv_n;
//               bmat.Row(2).Range (base_hdivn, base_hdivn +nd_hdivn) = -mat_hdivn_n;
// 	      bmat.Row(2).Range (base_hdiv, base_hdiv +nd_hdiv) += 
// 		Complex(0, fac/kappa) * div_sigma;

// 	      bmat.Row(1).Range (base_hdiv , base_hdiv +nd_hdiv) = mat_hdiv_n;
// 	      bmat.Row(1).Range (base_hdivn , base_hdivn +nd_hdivn) = -mat_hdivn_n;


//               /*
// 		0     i/k
// 		i/k   p+1
//               */


//               dmat = 0.0;
// 	      dmat(0,0) = -fac/kappa/kappa;
// 	      dmat(1,0) = dmat(0,1) = Complex(0.0, 1.0/kappa);
// 	      dmat(2,2) = 1; 

	      dmat = 0.0; 
	      bmat = 0.0;
	      bmat2 = 0.0;
	      if ( nx > 0. ) // T_out
		{
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma; //  + 1.0 / kappa * div_sigma2;

		  bmat2 = bmat;

		  if ( with_exp )
		    {
		      bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv ) =
			-Complex(0, 1.0/kappa) * div_sigma + 1.0 / kappa * div_sigma2;
		      bmat2.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv ) =
			-Complex(0, 1.0/kappa) * div_sigma - 1.0 / kappa * div_sigma2;
		    }

		  /* JOACHIM */
// 		  // V1
//  		  dmat(0,0) = -1.0;
//  		  dmat(1,0) = dmat(0,1) = -1.0;
// // 		  // V2
// //  		  dmat(0,0) = 1.0;
// //  		  dmat(1,0) = dmat(0,1) = -1.0;

// // 		  dmat(2,2) = 1.0;
// // 		  dmat(0,2) = dmat(2,0) = -1.0;
// // 		  dmat(1,2) = dmat(2,1) = -1.0;
// // 		  dmat(1,1) = 1.0;

// 		  // stabilization joachim
// // 		  Complex pen = 1;
// // 		  dmat(1,1) += pen;
// // 		  dmat(0,0) += pen;
// // 		  dmat(2,2) += pen;
// // 		  dmat(0,1) += pen;
// // 		  dmat(1,0) += pen;
// // 		  dmat(2,0) -= pen;
// // 		  dmat(0,2) -= pen;
// // 		  dmat(1,2) -= pen;
// // 		  dmat(2,1) -= pen;

		  /* ASTRID */
		  dmat(1,0) = -1.0; 
		  dmat(0,1) = -1.0;
		  dmat(2,1) = 1.0;
		  dmat(1,2) = 1.0;
		  dmat(2,2) = -1.0;

		  // stabilization, (u-lambda)(v-mu) daempft
 		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
 		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;
		}

	      else // T_in
		{
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma; //  + 1.0 / kappa * div_sigma2;

		  bmat2 = bmat;

		  if ( with_exp )
		    {
		      bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv ) =
			-Complex(0, 1.0/kappa) * div_sigma + 1.0 / kappa * div_sigma2;
		      bmat2.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv ) =
			-Complex(0, 1.0/kappa) * div_sigma - 1.0 / kappa * div_sigma2;
		    }

		  /* JOACHIM */
// // 		  // V1
// // 		  dmat(2,2) = -1.0;
// // 		  dmat(0,2) = dmat(2,0) = -1.0;
// // 		  dmat(1,2) = dmat(2,1) = 1.0;
// // 		  dmat(1,1) = -1.0;

// // 		  // V2
// // 		  dmat(2,2) = 1.0;
// // 		  dmat(0,2) = dmat(2,0) = -1.0;
// // 		  dmat(1,2) = dmat(2,1) = -1.0;
// // 		  dmat(1,1) = 1.0;

//  		  dmat(0,0) = 1.0;
//  		  dmat(1,0) = dmat(0,1) = -1.0;
// // 		  // stabilization joachim
// // 		  Complex pen = 1;
// // 		  dmat(1,1) += pen;
// // 		  dmat(0,0) += pen;
// // 		  dmat(2,2) += pen;
// // 		  dmat(0,1) -= pen;
// // 		  dmat(1,0) -= pen;
// // 		  dmat(2,0) += pen;
// // 		  dmat(0,2) += pen;
// // 		  dmat(1,2) -= pen;
// // 		  dmat(2,1) -= pen;


		  /* ASTRID */
		  dmat(0,1) = -1.0;
		  dmat(1,1) = -1.0;
		  dmat(1,0) = -1.0;

		  // stabilization, (u-lambda)(v-mu) daempft
 		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
 		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;

		}

              dmat *= factor;

//  	      *testout << "dmat = " << endl << dmat << endl;
//  	      *testout << "bmat = " << endl << bmat << endl;

              dbmat = dmat * bmat;

              elmat += Trans (bmat2) * dbmat;

	      lh.CleanUp(heapp);
            } 
	  
	
        }

    }
  

  template class HybridHelmholtzIntegrator<2>;
  template class HybridHelmholtzIntegrator<3>;









  // register FESpaces
  namespace
#ifdef MACOS
  hh_exp_cpp
#endif
  {
   
  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
  
    GetIntegrators().AddBFIntegrator ("helmholtz_exp", 2, 3,
                                      helmholtz_exp_cpp::HybridHelmholtzIntegrator<2>::Create);

    GetIntegrators().AddBFIntegrator ("helmholtz_exp", 3, 4,
                                      helmholtz_exp_cpp::HybridHelmholtzIntegrator<3>::Create);
  }

  Init init;
  }

} // namespace

