
#ifndef FILE_HELMHOLTZ_EXP_HPP
#define FILE_HELMHOLTZ_EXP_HPP


// solving the helmholtz equation
// unknowns sigma, lambda
// matches single.pdf
// possibility to include e^ikx into basis function

namespace helmholtz_exp_cpp
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace helmholtz_exp_cpp;

template<class TM, 
	 class TV_ROW = typename mat_traits<TM>::TV_ROW, 
	 class TV_COL = typename mat_traits<TM>::TV_COL>
class BlockJacobiPrecond_ElByEl;


  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
    Table <int> * isslave;

  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags);

    virtual ~HybridHelmholtzFESpace () { if ( isslave ) delete isslave; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(2);
      
      Flags hdivflags(flags);
      Flags hdivnflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      hdivflags.SetFlag ("order", order);
      hdivflags.SetFlag ("discontinuous");
      hdivflags.SetFlag ("orderinner", order);
      hdivflags.SetFlag ("curlorder", order);
      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderinner", 1.0);

      spaces[0] = new HDivHighOrderFESpaceA (ma, hdivflags);
      spaces[1] = new FacetFESpace (ma, hdivnflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }


    // direct solver clusters:

    // 0 ...... no clusters
    // 1 ...... low order facets
    // >= 2 ... divide sphere into n layers

    virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const;
      
    // #define VER1
    Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
    {
      int blocktype = int(precflags.GetNumFlag("blocktype",1)); 
      Table<int> * it = CreateSmoothingBlocks ( blocktype );

    }

  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;

    Table<int> * CreateSmoothingBlocks (int type=1) const;

    virtual void GetWireBasketDofNrs (int vnr, ARRAY<int> & dnums) const;


    inline void SetSlave ( int elnum, int fanum, int issl = 1 )
    {
      (*isslave)[elnum][fanum] = issl;
    }

    inline int IsSlave ( int elnum, int fanum ) const
    {
      return (*isslave)[elnum][fanum];
    }
  };




  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
    CoefficientFunction *coef_kvec0;
    CoefficientFunction *coef_kvec1;
    CoefficientFunction *coef_kvec2;

  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa = coeffs[0];
      coef_kvec0  = coeffs[1];
      coef_kvec1  = coeffs[2];
      if ( D == 3 )
	coef_kvec2  = coeffs[3];
      else
	{
	  ARRAY<double> coef(1);
	  coef = 0.;
	  coef_kvec2 = new DomainConstantCoefficientFunction(coef);
	}
    }

    virtual ~HybridHelmholtzIntegrator () 
    { 
      if ( D == 2 )
	delete coef_kvec2; 
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const;

  };



class HybridHelmholtzFiniteElement : public CompoundFiniteElement
{
protected:
  ///
  const HybridHelmholtzFESpace * hhspace;
  int globelnum;

public:
  /// 
  HybridHelmholtzFiniteElement (ARRAY<const FiniteElement*> & afea);

  ~HybridHelmholtzFiniteElement () 
  { ; }

  /// select i-th component
  const FiniteElement & operator[] (int i) const { return *fea[i]; }
  /// 
  inline void SetSlave (int fanum, int issl = 1 )
  { 
    HybridHelmholtzFESpace & consthhspace = const_cast<HybridHelmholtzFESpace&> (*hhspace);
    consthhspace.SetSlave(globelnum, fanum, issl); 
  }

  inline int IsSlave ( int fanum ) const
  { return hhspace -> IsSlave ( globelnum, fanum); }

  inline int ElNum () const 
  { return globelnum; }

  inline void SetElNum ( int aglobelnum ) 
  { globelnum = aglobelnum; }

  inline void SetHelmholtzSpace ( const HybridHelmholtzFESpace * ahhspace )
  { hhspace = ahhspace;}

//   template <int D>
//   const ngfem::HDivFiniteElement<D> & HDivElement<D> () const
//   { return dynamic_cast<const HDivFiniteElement<D> & > (*fea[0]); }

//   const ngfem::FacetFiniteElement & FacetElement () const
//   { return dynamic_cast<const FacetFiniteElement & > (*fea[1]); }

};




  template <class SCAL>
  class HelmholtzPreconditioner : public Preconditioner
  {
    const ElementByElement_BilinearForm<SCAL> * bfa;
    BaseMatrix * pre;
  public:
    HelmholtzPreconditioner (const PDE * pde, const Flags & aflags, const string aname = "hhprecond")
      : Preconditioner (pde, aflags, aname)
    {
      cout << "create helmholtz precond" << endl;
      bfa = dynamic_cast<const ElementByElement_BilinearForm<SCAL>*>(pde->GetBilinearForm (aflags.GetStringFlag ("bilinearform", NULL)));

    }

    virtual ~HelmholtzPreconditioner() 
    { ; }

    static Preconditioner * Create (const PDE & pde, const Flags & flags)
    {
      return new HelmholtzPreconditioner<SCAL> (&pde, flags);
    }

    virtual void Update ();

    virtual const BaseMatrix & GetAMatrix() const
    {
      return bfa->GetMatrix();
    }

    virtual const BaseMatrix & GetMatrix() const
    {
      return *pre;
    }
  };


}




#endif
