#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>
#include <helmholtz_single_exp.hpp>




extern ngsolve::MeshAccess * ma;
extern ngsolve::PDE * pde;

namespace helmholtz_exp_cpp
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace helmholtz_exp_cpp;

  HybridHelmholtzFESpace :: 
  HybridHelmholtzFESpace (const MeshAccess & ama, 		   
			  const ARRAY<const FESpace*> & aspaces,
			  const Flags & flags)
    : CompoundFESpace (ama, aspaces, flags)
  { 
    LocalHeap lh (10000000);

    isslave = new Table<int> (ma.GetNE(), 4);
    for ( int i = 0; i < ma.GetNE(); i++ )
      (*isslave)[i] = 1;
    
    BitArray issurffacet(ma.GetNFacets());
    ARRAY<int> facets;
    issurffacet.Clear();
    for ( int sel = 0; sel < ma.GetNSE(); sel++ )
//       if ( ma.GetSElBCName(sel) == "sourcebc" || ma.GetSElBCName(sel) == "internal" ) 
	//ma.GetSElIndex(sel) == 0 || ma.GetSElIndex(sel) == 2 )
	{
	  ma.GetSElFacets(sel, facets);
	  issurffacet.Set(facets[0] );
	}
    
    
    for ( int el = 0; el < ma.GetNE(); el++ )
      {
	ma.GetElFacets(el, facets);
	for ( int i = 0; i < facets.Size(); i++ )
	  {
	    if ( issurffacet.Test(facets[i]) ) 
	      (*isslave)[el][i] = 0;
	  }
      }


    if ( flags.GetStringFlag("kx","") == "" )
      return;

    // find slave elements
    CoefficientFunction *coef_kvec0 = pde->GetCoefficientFunction(flags.GetStringFlag("kx",""));
    CoefficientFunction *coef_kvec1 = pde->GetCoefficientFunction(flags.GetStringFlag("ky",""));;
    CoefficientFunction *coef_kvec2;

    int dim = ma.GetDimension();
    if ( dim == 3 )
      coef_kvec2  = pde->GetCoefficientFunction(flags.GetStringFlag("kz",""));
    else
      {
	ARRAY<double> coef(ma.GetNDomains() );
	coef = 0.;
	coef_kvec2 = new DomainConstantCoefficientFunction(coef);
      }

    for ( int el = 0; el < ma.GetNE(); el++ )
      {
	ElementTransformation eltrans;
	ma.GetElementTransformation ( el, eltrans, lh);

	ELEMENT_TYPE eltype = ma.GetElType(el);
	const POINT3D * points = ElementTopology::GetVertices (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	int ned = (eltype == ET_TRIG) ? 3: 4;
	
	Facet2ElementTrafo transform(eltype); 
	const NORMAL * normals = ElementTopology::GetNormals(eltype);
	
	ELEMENT_TYPE etfacet = (dim == 2) ? ET_SEGM : ET_TRIG;
  
	bool const_kvec =
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec0 ) && 
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec1 ) && 
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec2 ); 

	const IntegrationRule & ir1d =
	  SelectIntegrationRule (etfacet, 0 );
	for (int k = 0; k < ned; k++)
	  {
	    if ( dim == 2 )
	      {
		Vec<2> kvec;
		Vec<2> normal_ref;
		for (int i=0; i<2; i++)
		  normal_ref(i) = normals[k][i];
 		IntegrationPoint ip = transform(k, ir1d[0]);
		SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
		kvec(0) = coef_kvec0->Evaluate(sip);
		kvec(1) = coef_kvec1->Evaluate(sip);

		Mat<2> inv_jac = sip.GetJacobianInverse();
		double det = sip.GetJacobiDet();
		
		Vec<2> normal = det * Trans (inv_jac) * normal_ref;
		double len = L2Norm(normal);
		normal /= len;

		double nx = 0.0;
		
		for ( int i = 0; i < 2; i++ )
		  nx += kvec(i) * normal(i);
		
		nx /= L2Norm(kvec);
		nx /= L2Norm(normal);
		
		// isslave = 0 --> im smoothing block
		if ( nx <= 0.20 ) 
		  {
		    (*isslave)[el][k] = 0; // [elnum][k] = 0;
		  }
	      }
	    else
	      {
		Vec<3> kvec;
		Vec<3> normal_ref;
		for (int i=0; i<3; i++)
		  normal_ref(i) = normals[k][i];
 		IntegrationPoint ip = transform(k, ir1d[0]);
		SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
		kvec(0) = coef_kvec0->Evaluate(sip);
		kvec(1) = coef_kvec1->Evaluate(sip);
		kvec(2) = coef_kvec2->Evaluate(sip);

		Mat<3> inv_jac = sip.GetJacobianInverse();
		double det = sip.GetJacobiDet();
		
		Vec<3> normal = det * Trans (inv_jac) * normal_ref;
		double len = L2Norm(normal);
		normal /= len;

		double nx = 0.0;
		
		for ( int i = 0; i < 3; i++ )
		  nx += kvec(i) * normal(i);
		
		nx /= L2Norm(kvec);
		nx /= L2Norm(normal);
		
		// isslave = 0 --> im smoothing block
		if ( nx <= 0.25 ) 
		  {
		    (*isslave)[el][k] = 0; // [elnum][k] = 0;
		  }
		;
	      }
	  }

      }


    if ( ma.GetDimension() == 2 ) delete coef_kvec2;
    
  }
  


  ARRAY<int> * HybridHelmholtzFESpace :: CreateDirectSolverClusters (const Flags & precflags) const
  {
    int clustertype = int(precflags.GetNumFlag("ds_cluster",0)); 
    
    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());
    clusters = 0;
    
    int nd = GetNDof();
    int nse = ma.GetNSE();
    int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
    int ned = ma.GetNEdges();
    int nel = ma.GetNE();
    
    int dim = ma.GetDimension();
    
    int order_sigmaE = (*this)[1]->GetOrder();
    int order_sigma = (*this)[0]->GetOrder();
      
    ARRAY<int> facets, dnums;
    
    const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace& > (*spaces[1]);
    
    if ( clustertype == 0 )
	{
	  clusters = 0;
	  return 0; // &clusters;
	}
    else if ( clustertype == 1 )
	{
	  int nf = (ma.GetDimension() == 2 ) ? 3 : 4;
	  int nt = nf;
	  nfacets = ma.GetNFacets();
	  for ( int el = 0; el < nel; el++ )
	    {
// 	      spaces[0]->GetDofNrs(el, dnums);
// 	      for (int i = 0; i < nt; i++ )
// 		clusters[dnums[i]] = 1;
	      for ( int i = 0; i < nfacets; i++ )
		{
		  ARRAY<int> facetdofs;
		  facetspace.GetFacetDofNrs(i, facetdofs);
		  for ( int  j = 0; j < 2; j++ )
		    clusters[GetStorageStart(1) + facetdofs[j] ] = 1;
		  for ( int j = 2; j < facetdofs.Size(); j++ )
		    clusters[facetdofs[j]] = facetdofs[j];
		}
	    }
// 	  for ( int i = spaces[0]->GetNDof() + nf; i < GetNDof(); i++)
// 	    clusters[i] = i;
	  return &clusters;
	}


      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < nse; sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 2 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
   		    int cluster = int( (1.0 * (nel-1) * clustertype) / nel) - int( (1.0 * nmarkedels * clustertype) / nel) +1 ;
//   		    int cluster = int( (1.0 * nmarkedels * clustertype) / nel) +1 ;
		    if ( ma.GetDimension() == 3 )
		      ma.SetElementPartition(el, cluster);
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      newmarkedfacets.Set(facets[j]);

		    // set cluster .. 1 if el < nel/2, 2 else

		    GetDofNrs (el, dnums);
		    for (int j = 0; j < dnums.Size(); j++)
		     if ( !clusters[dnums[j]] )
		      clusters[dnums[j]] = cluster;

		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa) ) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;

      return &clusters;
    } 


  Table<int> * HybridHelmholtzFESpace :: CreateSmoothingBlocks (int type) const

     {
      int nd = GetNDof();
      int nse = ma.GetNSE();
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
            
      int dim = ma.GetDimension();

      int order_sigmaE = (*this)[1]->GetOrder();
      int order_sigma = (*this)[0]->GetOrder();
      

      const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace &> (*spaces[1]);

      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel), orderedfacets(nfacets);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);
	  int ndof_per_edge, ndof_per_trig;

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      int ifacet = 0;

      ARRAY<int> facets;
      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < nse; sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 2 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	  orderedfacets[ifacet++] = facets[0];
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      {
			if (!markedfacets.Test(facets[j]) && !newmarkedfacets.Test(facets[j]))
			  orderedfacets[ifacet++] = facets[j];
			newmarkedfacets.Set(facets[j]);
		      }
		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa)) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;

      cout << "SmoothingType " << type << endl; 
      // edge by edge

      Table<int> * it;
      int base_hdivn = GetStorageStart(1);
      int ndof = GetNDof();
      ARRAY<int> cnt;
      ARRAY<int> dnums;
      ARRAY<int> block;
      
//       Smoothing types:
//      1 .. facet by facet (eliminate_internal)
// 	2 .. element by element  (eliminate_internal)
// 	3 .. nichts
// 	4 .. master-edges/element (noelim_int)
// 	5 .. ds_clusters (noelim_int)
// 	6 .. vertex-patches (noelim_int)
// 	62 .. vertex-patches (elim_int)
// 	7 .. facet-patches (elim_int)

      if ( dim == 2 ) 
	{
	  ndof_per_edge = order_sigmaE +1;
	  ndof_per_trig = int((order_sigma+1)*(order_sigma+2));
	}
      else
	{
	  ndof_per_edge = int(0.5*(order_sigmaE+1)*(order_sigmaE+2));
	  ndof_per_trig = ((order_sigma+1)*(order_sigma+2)*(order_sigma+3))/2;
	}


      

      switch (type)
	{
	case 1:
	case 12:
	  cout << "facet by facet, " ;
	  if ( type == 1 ) cout << "no" ;
	  cout << "eliminate_internal";

	  cnt.SetSize(nfacets); // ned/* +nel */);

	  for (int i = 0; i < nfacets; i++)
	      cnt[i] = ndof_per_edge;

	  it = new Table<int>(cnt);
	  // Table<int> * it = new Table<int>(ned,2*order+2);
	  cnt = 0;
	  for (int i = 0; i < nfacets; i++)
	    {
	      int ifacet = orderedfacets[i];
	      facetspace.GetFacetDofNrs(ifacet, dnums);
	      for ( int j = 0; j < dnums.Size(); j++)
		(*it)[i][cnt[i]++] = base_hdivn + dnums[j];
 
// 	      (*it)[i][0] = base_hdivn + ifacet;
// 	      for (int j = 0; j < ndof_per_edge-1; j++)
// 		(*it)[i][j+1] = base_hdivn + nfacets + (ifacet * (ndof_per_edge-1)) + j;
	    }


	  // *testout << "table = " << endl << *it << endl;
	  return it;
	  break;

	case 2:
	case 22:
	  // #ifdef VER2
	  // all external dofs of an element
	  cout << "element by element" << endl;
	  if ( type == 2 ) cout << "no" ;
	  cout << "eliminate_internal";
	  
	  cnt.SetSize(nel);
	  
	  for (int i = 0; i < nel; i++)
	    {
	      if ( type == 2 )
		GetDofNrs (orderedels[i], dnums);
	      else 
		GetExternalDofNrs(orderedels[i], dnums);
	      cnt[i] = dnums.Size();
	    }
	  
	  it = new Table<int>(cnt);    
	  for (int i = 0; i < nel; i++)
	    {
	      if ( type == 2 )
		GetDofNrs (orderedels[i], dnums);
	      else
		{
		  GetExternalDofNrs(orderedels[i], dnums);
// 		  for ( int j = 0; j < dnums.Size(); j++)
// 		    dnums[j] += spaces[0]->GetNDof();
		}
	      for (int j = 0; j < dnums.Size(); j++)
		(*it)[i][j] = dnums[j];
	    }
	  return it;
	  // #endif
	  break;

	case 3:
	  // no blocks
	  cnt.SetSize(1);
	  cnt[0] = 0;
	  it = new Table<int> (cnt);
	  return it;
	  break;

	case 5:
	  cnt.SetSize(nel);
	  block.SetSize(ndof);
	  cnt = 0;
	  for ( int eli = 0; eli < nel; eli++ )
	    {
	      int el = orderedels[eli];
	      GetDofNrs(el, dnums);
	      for ( int i = 0; i < dnums.Size(); i++ )
		block[dnums[i]] = (eli)/20;

	    }
	  for ( int dof = 0; dof < ndof; dof++ )
	    {
	      cnt[block[dof]]++;
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for ( int dof = 0; dof < ndof; dof++ )
	    (*it)[block[dof]][cnt[block[dof]]++] = dof;

	  return it;
	  break;
	case 4:
	case 42:
	  cout << "master edges of element" << endl;
	  if ( type == 2 ) cout << "no" ;
	  cout << "eliminate_internal";

	  cnt.SetSize(nel);
	 
	  for ( int i = 0; i < nel; i++ )
	    {
	      ma.GetElFacets(i, facets);
	      cnt[i] = facets.Size();
	    }
	  // Table<int> isslave (cnt);
	  cnt = 0;

	      for ( int eli = 0; eli < nel; eli++ )
		{
		  int el = orderedels[eli];
		  if ( type == 4 )
		    GetDofNrs(el, dnums); 
		  else
		    spaces[1]->GetDofNrs(el, dnums);

		  cnt[eli] += dnums.Size();
		  ma.GetElFacets(el, facets);
		  for ( int i = 0; i < facets.Size(); i++ )
		    {

		      if ( (*isslave)[el][i] )
			cnt[eli] -= ndof_per_edge;

		    }
		}

	  it = new Table<int> (cnt);
	  cnt = 0;

	      for ( int eli = 0; eli < nel; eli++)
		{
		  int el = orderedels[eli];
		  ma.GetElFacets(el, facets);
		  GetDofNrs(el, dnums); 
		  if ( type == 4 )
		    for ( int i = 0; i < dnums.Size() - facets.Size() * (ndof_per_edge) ; i++)
		      (*it)[eli][cnt[eli]++] = dnums[i];
		  for ( int i = 0; i < facets.Size(); i++ )
		    {
		      if ( !(*isslave)[el][i] )
			{
			  facetspace.GetFacetDofNrs(facets[i], dnums);
			  for ( int j = 0; j < dnums.Size(); j++)
			    (*it)[eli][cnt[eli]++] = base_hdivn + dnums[j];

// 			  (*it)[eli][cnt[eli]++] = facets[i] + base_hdivn;
// 			  for (int j = 0; j < ndof_per_edge-1; j++)
// 			    (*it)[eli][cnt[eli]++] = base_hdivn + nfacets + (facets[i] * (ndof_per_edge-1)) + j;

// 			  for (int j = 0; j < ndof_per_edge; j++)
// 			    (*it)[eli][cnt[eli]++] = base_hdivn +  (facets[i] * ndof_per_edge ) + j;
			}
		      
		    }
		}
	  // *testout << "it " << it << endl;
	  return it;
	  break;

	case 6:
	  cnt.SetSize(ma.GetNV() );
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		  cnt[verts[i]] += ndof_per_trig;
		  for ( int k = 0; k < facets.Size(); k++)
		    if ( !(*isslave)[el][k] )
		      cnt[verts[i]] += ndof_per_edge;
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		 spaces[0]->GetDofNrs(el, dnums); 
 		 for ( int j = 0; j < dnums.Size() ; j++)
 		   (*it)[verts[i]][cnt[verts[i]]++] = dnums[j];
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + dnums[j];
// 		       (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + nfacets + 
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  // *testout << "table.. " << endl << *it << endl;
	  return it;
	  break;

	case 62:
	  // mit eliminate internal
	  cnt.SetSize(ma.GetNV() );
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		  for ( int k = 0; k < facets.Size(); k++)
		    if ( !(*isslave)[el][k] )
		      cnt[verts[i]] += ndof_per_edge;
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + dnums[j];
// 		       (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + nfacets + 
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  //*testout << "table.. " << endl << *it << endl;
	  return it;
	  break;

	case 7:
	case 72:

	  if ( type == 7 )
	    cnt.SetSize(ma.GetNFacets() );
	  else
	    cnt.SetSize(ma.GetNFacets()  + ma.GetNE() );

	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  spaces[0]->GetDofNrs(el, dnums); 
		  if ( type == 7 )
		    cnt[facets[i]] += dnums.Size();
		  else
		    cnt[nfacets + el] += dnums.Size();
 
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       cnt[facets[i]] += dnums.Size();
		     }
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  spaces[0]->GetDofNrs(el, dnums); 
		  if ( type == 7 )
		    {
		      for ( int j = 0; j < dnums.Size() ; j++)
			(*it)[facets[i]][cnt[facets[i]]++] = dnums[j];
		    }
		  else
		    for ( int j = 0; j < dnums.Size() ; j++)
		      (*it)[nfacets+el][cnt[nfacets+el]++] = dnums[j];

		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn + dnums[j];

// 		       (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn  + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn  + nfacets +
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  // *testout << "table.. " << endl << *it << endl;
	  return it;
	  break;


	case 8:
	  cnt.SetSize(nel);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ma.GetElFacets(el, facets);

	      cnt[el] += ndof_per_trig;

		  for ( int k = 0; k < facets.Size(); k++)
		    {
		      cnt[el] += ndof_per_edge;
		      if ( !(*isslave)[el][k] )
			cnt[el] += ndof_per_trig;
		    }
	    }

	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> neighbours;
	      ma.GetElFacets(el, facets);
	      spaces[0]->GetDofNrs(el, dnums); 
	      for ( int j = 0; j < dnums.Size() ; j++)
		(*it)[el][cnt[el]++] = dnums[j];
	      for ( int k = 0; k < facets.Size(); k++)
		{
		  (*it)[el][cnt[el]++] = base_hdivn  + facets[k];
		  for (int j = 0; j < ndof_per_edge-1; j++)
		    (*it)[el][cnt[el]++] = base_hdivn  + 
		      (facets[k] * (ndof_per_edge-1) ) + j;
		  if ( !(*isslave)[el][k] )
		    {
		      if ( dim == 2 )
			ma.GetEdgeElements(facets[k], neighbours);
		      else
			ma.GetFaceElements(facets[k], neighbours);
		      for( int ii = 0; ii < neighbours.Size(); ii++ )
			if ( neighbours[ii] != el )
			  {
			    //*testout << "el " << el << "facet " << k << ": " << neighbours[ii] << endl;
			    spaces[0]->GetDofNrs(neighbours[ii], dnums);
			    for ( int j = 0; j < dnums.Size() ; j++)
			      (*it)[el][cnt[el]++] = dnums[j];
			  }
		    }
		}
	      
	    }
	  
	  break;

	case 9:
	  cnt.SetSize(spaces[1]->GetNDof() );
	  cnt = 1;
	  it = new Table<int> (cnt);
	  for ( int i = base_hdivn; i < ndof; i++ )
	    (*it)[i-base_hdivn][0] = i;
	  break;


	default:

	  cout << "Illegal Smoothing Type " << type << endl;
	  cout << "set flag -blocktype=<type> where <type> is from" << endl <<
	    "#       Smoothing types:\n #       1 .. facet by facet (noeliminate_internal) \n #       12 .. facet by facet (eliminate_internal)\n # 	2 .. element by element  (noeliminate_internal)\n # 	22 .. element by element  (eliminate_internal)\n # 	3 .. nichts\n # 	4 .. master-edges/element (noelim_int)\n # 	42 .. master-edges/element (elim_int)\n # 	5 .. ds_clusters (noelim_int)\n # 	6 .. vertex-patches (noelim_int)\n # 	62 .. vertex-patches (elim_int)\n # 	7 .. facet-patches (elim_int)" << endl;
	}
      return it;
    }

  void HybridHelmholtzFESpace :: GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ArrayMem<int,500> hdnums;
    ARRAY<int> facets, facetdofs;
    dnums.SetSize(0);

    if ( !eliminate_internal)
      {
	spaces[0] -> GetDofNrs(elnr, dnums);
      }

    // wirebasket dofs are facet-dofs of master facets
    ma.GetElFacets (elnr, facets);
    const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace & > (*spaces[1]);
    for (int i = 0; i < facets.Size(); i++)
      {
	facetspace.GetFacetDofNrs(facets[i], facetdofs);
	dnums.Append( GetStorageStart(1) + facetdofs[0]);
//     	if ( !(*isslave)[elnr][i] )
	for ( int j = 1; j < 3; j++ ) // facetdofs.Size(); j++ )
	  dnums.Append (GetStorageStart(1) + facetdofs[j]);
      }

  }

  const FiniteElement & HybridHelmholtzFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    ARRAY<const FiniteElement*> fea(2);
    for (int i = 0; i < fea.Size(); i++)
      fea[i] = &spaces[i]->GetFE(elnr, lh);
    
    void * mem = lh.Alloc (sizeof(HybridHelmholtzFiniteElement));

    HybridHelmholtzFiniteElement & fel = *new (mem) HybridHelmholtzFiniteElement (fea);
    fel.SetElNum ( elnr );
    fel.SetHelmholtzSpace ( this );

    return fel;
  }



  HybridHelmholtzFiniteElement ::  
  HybridHelmholtzFiniteElement (ARRAY<const FiniteElement*> & afea)
    : CompoundFiniteElement (afea)
  {
    hhspace = 0;
    globelnum = -1;
  }


  // register FESpaces
  namespace
#ifdef MACOS
  hhfem_cpp
#endif
  {
   
  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("helmholtz_exp_space", helmholtz_exp_cpp::HybridHelmholtzFESpace::Create);
  
  }

  Init init;
  }


}
