#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

#include <helmholtz_single_exp.hpp>

extern ngsolve::MeshAccess * ma;
extern ngsolve::PDE * pde;


namespace helmholtz_exp_cpp
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace helmholtz_exp_cpp;
  using namespace ngcomp;



  template <class SCAL>
  void HelmholtzPreconditioner<SCAL> :: Update()
  {
    LocalHeap lh(100000);
    pre = new ElementByElementMatrix<SCAL>( bfa -> GetFESpace() . GetNDof() );
    const FESpace & fespace = bfa -> GetFESpace();
    const HybridHelmholtzFESpace & hhspace = 
      dynamic_cast<const HybridHelmholtzFESpace &> (fespace);
    const MeshAccess & ma = bfa -> GetMeshAccess();
    const ElementByElementMatrix<SCAL> & amat = 
      dynamic_cast<const ElementByElementMatrix<SCAL> &> (bfa -> GetMatrix());
    ElementByElementMatrix<SCAL> & premat = 
      dynamic_cast<ElementByElementMatrix<SCAL> &> (*pre);

    int nel = ma.GetNE();
 
    for ( int el = 0; el < nel; el++ )
      {
	ARRAY<int> dofs, blockdofs, boundi, inneri;
	hhspace.GetExternalDofNrs(el, dofs);
 	hhspace.GetWireBasketDofNrs(el, blockdofs);


	int nd_bound = blockdofs.Size();
	int nd = dofs.Size();
	int nd_inner = nd - nd_bound;
	
	if ( nd_bound == 0 )
	  {
	    FlatMatrix<SCAL> blockmat(0, 0, lh);
	    premat.AddElementMatrix( blockdofs, blockdofs, blockmat );
	    continue;
	  }

	boundi.SetSize(nd_bound);
	inneri.SetSize(0);

	for ( int i = 0; i < nd_bound; i++ )
	  boundi[i] = dofs.Pos(blockdofs[i]);

	for ( int i = 0; i < nd; i++ )
	  if ( !boundi.Contains(i) )
	    inneri.Append(i);

	FlatMatrix<SCAL> a(nd_bound, nd_bound, lh);
	FlatMatrix<SCAL> b(nd_bound, nd_inner, lh);
	FlatMatrix<SCAL> c(nd_bound, nd_inner, lh);
	FlatMatrix<SCAL> idc(nd_bound, nd_inner, lh);
	FlatMatrix<SCAL> d(nd_inner, nd_inner, lh);
	const FlatMatrix<SCAL> elmat = amat.GetElementMatrix (el);
	FlatMatrix<SCAL> invmat (nd, nd, lh);

	for ( int i = 0; i < nd_bound; i++ )
	  for ( int j = 0; j < nd_bound; j++ )
	    a(i, j) = elmat(boundi[i], boundi[j]);

	for ( int i = 0; i < nd_bound; i++ )
	  for ( int j = 0; j < nd_inner; j++ )
	    {
	      b(i, j) = elmat(boundi[i], inneri[j]);
	      c(i, j) = elmat(inneri[j], boundi[i]);
	    }

	for ( int i = 0; i < nd_inner; i++ )
	  for ( int j = 0; j < nd_inner; j++ )
	    d(i, j) = elmat(inneri[i], inneri[j]);



// 	if ( nd_inner > 0 )
// 	  {
// 	    LapackInverse (d);
// 	    LapackMultABt (c, d, idc);
// 	    LapackMultAddABt (b, idc, -1, a);
// 	  }
	LapackInverse(a);

// 	*testout << "element " << el << ": ";
// 	*testout << "a " << endl << a << endl << "b " << endl << b << endl;
// 	*testout << "c" << endl << c << endl << "d" << endl << d << endl;

// 	CalcInverse ( elmat, invmat);
// 	dnums.Append (dn);
	
// 	elmats.SetSize (elmats.Size()+1);
// 	elmats.Last().AssignMemory (s, s, &invmat(0,0));
	premat.AddElementMatrix ( blockdofs, blockdofs, a);

	lh.CleanUp();
      }
    
    
  }

  template class HelmholtzPreconditioner<double>;
  template class HelmholtzPreconditioner<Complex>;


  // register FESpaces
  namespace
#ifdef MACOS
  ho_biform_cpp
#endif
  {
   
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetPreconditionerClasses().AddPreconditioner ("helmholtzprec", HelmholtzPreconditioner<double>::Create);
      GetPreconditionerClasses().AddPreconditioner ("helmholtz_complex", HelmholtzPreconditioner<Complex>::Create);
    }
    
    Init init;
  }



}
