#include "../include/solve.hpp"

namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;



  class SymmetricGaussSeidel : public BaseMatrix
  {
    BlockJacobiPrecond<Complex> * blockjacobi;

  public:
    SymmetricGaussSeidel ( BlockJacobiPrecond<Complex> * ablockjacobi )
    {
      blockjacobi = ablockjacobi;
    }

    ~SymmetricGaussSeidel ()
    { ; }

    virtual void Mult ( const BaseVector & x, BaseVector & y ) const
    {
      y.FVDouble () = 0.0;

      blockjacobi->GSSmooth ( y, x, 1);
      blockjacobi->GSSmoothBack ( y, x, 1);
    }

  };


  /* ********** mein liebes kleines helmholtz-1d programm  ************* */


  ///
  class NumProcHH1d : public NumProc
  {
  protected:
    int n;
    int p;
    double kappa;
    double tau;
    int maxsteps;
    double precision;
    int smoothingtype;
    ///
    enum SOLVER { CG, CGMULT, CGDIRECT, GMRES, GMRESMULT, QMR, NCG, SIMPLE, DIRECT };
    ///
    SOLVER solver;
    
  public:
    ///
    NumProcHH1d (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcHH1d();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcHH1d (pde, flags);
    }


    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "1d helmholtz";
    }

    virtual void PrintReport (ostream & ost)
    {
      ;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcHH1d :: NumProcHH1d (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    n = int(flags.GetNumFlag ("nel", 200));
    p = int(flags.GetNumFlag ("order", 1));
    kappa = flags.GetNumFlag ("kappa", 1.0);
    smoothingtype = int(flags.GetNumFlag ("smoother", 0));
    const char* sol = flags.GetStringFlag("solver","cg");
    if (strcmp(sol,"cg")==0) solver = CG;
    if (strcmp(sol,"cgmult")==0) solver = CGMULT;
    if (strcmp(sol,"cgdirect")==0) solver = CGDIRECT;
    if (strcmp(sol,"qmr")==0) solver = QMR;
    if (strcmp(sol,"gmres")==0) solver = GMRES;
    if (strcmp(sol,"gmresmult")==0) solver = GMRESMULT;
    if (strcmp(sol,"simple")==0) solver = SIMPLE;
    if (strcmp(sol,"direct")==0) solver = DIRECT;
    tau = flags.GetNumFlag("tau", 0.1);
    precision = flags.GetNumFlag("prec", 1e-8);
    maxsteps = int ( flags.GetNumFlag("maxsteps",100) );
  }

  NumProcHH1d :: ~NumProcHH1d()
  {
    ;
  }

  void NumProcHH1d :: PrintDoc (ostream & ost)
  {
    ;
  }


  void NumProcHH1d :: Do(LocalHeap & lh)
  {

  cout << "nel = " << n << endl;
  cout << "order = " << p << endl;
  cout << "kappa = " << kappa << endl;

  int nel = n;
  int nv = n+1;
  double a = -1.0, b = 1.0;
  ARRAY<double> points(nv);

  // mesh
  for ( int i = 0; i < nv; i++ )
    points[i] =  a + 1.0*(b-a)/ nel * i;


  Table<int> el2vert(nel, 2);
  for ( int el = 0; el < nel; el++ )
    {
      el2vert[el][0] = el;
      el2vert[el][1] = el+1;
    }

  // fespaces
  int nd, nd_hdiv, nd_hdivn, ndof;
  nd_hdiv = (p+1) * nel;
  nd_hdivn = nv;
  nd = nd_hdiv + nd_hdivn;
  ndof = nd;

  Table<int> el2dof(nel, p+1+2);
  Table<int> vert2dof(nv, 1);
  for ( int el = 0; el < nel; el++ )
    {
      for ( int j = 0; j < p+1; j++ )
	el2dof[el][j] = el*(p+1) + j;
      el2dof[el][p+1] = nd_hdiv + el;
      el2dof[el][p+2] = nd_hdiv + el+1;
    }
  for ( int vert = 0; vert < nv; vert++ )
    {
      vert2dof[vert][0] = nd_hdiv + vert;
    }

  int nd_el = p+1+2;
  int nd_hdiv_el = p+1;
  int nd_hdivn_el = 2;

  // generate matrix graph
  ARRAY<int> elsperrow(ndof);
  elsperrow = 0;
  for ( int el = 0; el < nel; el++ )
    {
      for ( int j = 0; j < el2dof[el].Size(); j++ )
	elsperrow[el2dof[el][j]] += el2dof[el].Size(); 
    }
  for ( int j = 1; j < nv-1; j++ )
    elsperrow[nd_hdiv + j] --;

  MatrixGraph * graph = new MatrixGraph( elsperrow );

  for ( int el = 0; el < nel; el++ )
    {
      for ( int j = 0; j < el2dof[el].Size(); j++ )
	for ( int k = 0; k < el2dof[el].Size(); k++ )
	{
	  graph->CreatePosition( el2dof[el][j], el2dof[el][k] );
	}
    }

  SparseMatrix<Complex> * amat = new SparseMatrix<Complex> (*graph, 0);
  BaseVector & vecf = *amat->CreateVector();
  BaseVector & vecu = *amat->CreateVector();

  delete graph;

  amat->AsVector().FVComplex() = 0.0;

  // assemble elmat

  const IntegrationRule & ir = 
    GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*p);

  FlatMatrix<Complex> elmat(nd_el, lh);
  FlatVector<> mat_hdiv(nd_hdiv_el, lh);
  FlatMatrix<> mat_divhdiv(nd_hdiv_el, 1, lh);

  L2HighOrderSegm fel_hdiv(p);

  for ( int el = 0; el < nel; el++ )
    {
      elmat = 0.0;

      double v1 = points[el2vert[el][0]];
      double v2 = points[el2vert[el][1]];

      double det = v2 - v1;

      for ( int l = 0; l < ir.GetNIP(); l++ )
	{
	  double factor = ir[l].Weight() * det;
	  const IntegrationPoint & ip = ir[l];

	  fel_hdiv.CalcShape(ip, mat_hdiv);
	  // piola: 1/h * h * shape = shape

	  fel_hdiv.CalcDShape(ip, mat_divhdiv);
	  mat_divhdiv *= 1.0 / det;

	  for ( int i = 0; i < nd_hdiv_el; i++ )
	    for ( int j = 0; j < nd_hdiv_el; j++ )
	      {
		// int i kappa sigma tau 
		imag(elmat (i,j) ) += mat_hdiv(i) * mat_hdiv(j) * factor * kappa;

		// - int i / kappa sigma' tau'
		imag(elmat (i,j) ) -= mat_divhdiv(i,0) * mat_divhdiv(j,0) * factor / kappa;

	      }
 	}


      // facet --> points
      // vertex 0
      double n = -1.0;
      IntegrationPoint ip1(0.0);
      double point = points[el2vert[el][0]];

      double penalty = 1e7;
      if ( n * point < 0 ) penalty = 1;

      fel_hdiv.CalcShape(ip1, mat_hdiv);
      fel_hdiv.CalcDShape(ip1, mat_divhdiv);
      mat_divhdiv /=  det;
      double mat_hdivn = n;
      int k = nd_hdiv_el;
      for ( int i = 0; i < nd_hdiv_el; i++ )
	{
	  for ( int j = 0; j < nd_hdiv_el; j++ )
	    {
	      imag( elmat(i,j) ) += 1.0 / kappa * ( n * mat_hdiv(i) ) * mat_divhdiv(j,0);
	      imag( elmat(j,i) ) += 1.0 / kappa * ( n * mat_hdiv(i) ) * mat_divhdiv(j,0);
	      elmat (i,j) -= penalty * ( mat_hdiv(i) ) * mat_hdiv(j);
	    }
	  imag( elmat(i,k) ) += -1.0 / kappa * ( mat_hdivn ) * mat_divhdiv(i,0);
	  imag( elmat(k,i) ) += -1.0 / kappa * ( mat_hdivn ) * mat_divhdiv(i,0);

	  elmat (i,k) += penalty * n * mat_hdiv(i) * mat_hdivn;
	  elmat (k,i) += penalty * n * mat_hdiv(i) * mat_hdivn;
	}
	 
      elmat (k,k) -= penalty * mat_hdivn * mat_hdivn;
 
      n = 1.0;
      IntegrationPoint ip2(1.0);
      point = points[el2vert[el][1]];

      penalty = 1e7;
      if ( n * point < 0 ) penalty = 1;

      fel_hdiv.CalcShape(ip2, mat_hdiv);
      fel_hdiv.CalcDShape(ip2, mat_divhdiv);
      mat_divhdiv /= det;
      mat_hdivn = n;
      k = nd_hdiv_el +1;
 
      for ( int i = 0; i < nd_hdiv_el; i++ )
	{
	  for ( int j = 0; j < nd_hdiv_el; j++ )
	    {
	      imag( elmat(i,j) ) += 1.0 / kappa * ( n * mat_hdiv(i) ) * mat_divhdiv(j,0);
	      imag( elmat(j,i) ) += 1.0 / kappa * ( n * mat_hdiv(i) ) * mat_divhdiv(j,0);
	      elmat (i,j) -= penalty * ( mat_hdiv(i) ) * mat_hdiv(j);
	    }
	  imag( elmat(i,k) ) += -1.0 / kappa * ( mat_hdivn ) * mat_divhdiv(i,0);
	  imag( elmat(k,i) ) += -1.0 / kappa * ( mat_hdivn ) * mat_divhdiv(i,0);

	  elmat (i,k) += penalty * n * mat_hdiv(i) * mat_hdivn;
	  elmat (k,i) += penalty * n * mat_hdiv(i) * mat_hdivn;
	}
	 
      elmat (k,k) -= penalty * mat_hdivn * mat_hdivn;
 
      for ( int i = 0; i < nd_el; i++ )
	for ( int j = 0; j < nd_el; j++ )
	  (*amat)(el2dof[el][i], el2dof[el][j]) += elmat(i,j);
    }


  // boundary terms: - int_gamma sigma_E tau_E
  (*amat)(vert2dof[0][0], vert2dof[0][0]) -= -1.0;
  (*amat)(vert2dof[nv-1][0], vert2dof[nv-1][0]) -= -1.0;

//   for ( int i = 0; i < ndof; i++ )
//     (*amat)(i,i) -= Complex(1e-5, 1e-5);

  *testout << *amat;
  vecf.FVComplex() = 0.0;
  FlatVector<Complex> elvec (nd_el, lh);
  for ( int el = 0; el < nel; el++ )
    {
      elvec = 0.0;

      double v1 = points[el2vert[el][0]];
      double v2 = points[el2vert[el][1]];

      if ( v1 < -0.1 || v2 > 0.1 ) continue;
      double det = v2 - v1;

      for ( int l = 0; l < ir.GetNIP(); l++ )
	{
	  double factor = ir[l].Weight() * det;
	  const IntegrationPoint & ip = ir[l];
	  fel_hdiv.CalcDShape(ip, mat_divhdiv);
	  mat_divhdiv /= det;

	  for ( int i = 0; i < nd_hdiv_el; i++ )
	    elvec(i) += mat_divhdiv(i,0) * factor; 
	}

      for ( int i = 0; i < nd_el; i++ )
	vecf.FVComplex()(el2dof[el][i]) += elvec(i);
    }

  amat->SetInverseType(SPARSECHOLESKY);


  // gauss-seidel
  ARRAY<int> cnt(nel), block(ndof);
  cnt = 0; block = 0;
  Table<int> * blocktable;

  int slavev1, slavev2;
  switch ( smoothingtype )
    {
    case 0:
      for ( int i = nel-1; i >= 0 ; i-- )
	for ( int j = 0; j < el2dof[i].Size(); j++ )
	  block[el2dof[i][j]] = i;
      
      for ( int dof = 0; dof < ndof; dof++ )
	cnt[block[dof]]++;
      
      blocktable = new Table<int>(cnt);
      cnt = 0;
      for ( int dof = 0; dof < ndof; dof++ )
	(*blocktable)[block[dof]][cnt[block[dof]]++] = dof;
      break;

    case 1:
      for ( int i = 0; i < nel; i++ )
	cnt[i] = el2dof[i].Size();
      
      blocktable = new Table<int>(cnt);
      for ( int i = 0; i < nel; i++ )
	for ( int j = 0; j < el2dof[i].Size(); j++ )
	  (*blocktable)[i][j] = el2dof[i][j];
      break;

    case 2:
      cnt.SetSize(ndof);
      cnt = 1;
      blocktable = new Table<int> (cnt);
      for ( int dof = 0; dof < ndof; dof++ )
	(*blocktable)[dof][0] = dof;

      break;
    case 3:
      cnt.SetSize(1);
      cnt[0]= ndof;
      blocktable = new Table<int> (cnt);
      for ( int dof = 0; dof < ndof; dof++ )
	(*blocktable)[0][dof] = dof;
      break;

    case 4:
      cnt.SetSize(nel);
      for ( int el = 0; el < nel; el++ )
	{
	  double v1 = points[el2vert[el][0]], v2 = points[el2vert[el][1]];
	  if ( v1 > 0 ) slavev1 = 1; else slavev1 = 0;
	  if ( v2 <= 0 ) slavev2 = 1; else slavev2 = 0;
	  if ( el == 0 || el == nel - 1 ) slavev1 = slavev2 = 1;
	  cnt[el] += p+1 + slavev1 + slavev2;
	}

      blocktable = new Table<int> (cnt);
      cnt = 0;
      for ( int el = 0; el < nel; el++ )
	{
	  double v1 = points[el2vert[el][0]], v2 = points[el2vert[el][1]];
	  if ( v1 > 0 ) slavev1 = 1; else slavev1 = 0;
	  if ( v2 <= 0 ) slavev2 = 1; else slavev2 = 0;
	  if ( el == 0 || el == nel - 1 ) slavev1 = slavev2 = 1;
	  for ( int i = 0; i < el2dof[el].Size()-2; i++)
	    (*blocktable)[el][cnt[el]++] = el2dof[el][i];
	  if ( slavev1 ) (*blocktable)[el][cnt[el]++] = el2dof[el][p+1];
	  if ( slavev2 ) (*blocktable)[el][cnt[el]++] = el2dof[el][p+2];
	}
      break;

    case 10:
      cnt.SetSize(nel);
      for ( int el = 0; el < nel; el++ )
	{
	  double v1 = points[el2vert[el][0]], v2 = points[el2vert[el][1]];
	  if ( v1 <= 0 ) slavev1 = 1; else slavev1 = 0;
	  if ( v2 >= 0 ) slavev2 = 1; else slavev2 = 0;
	  //if ( el == 0 || el == nel - 1 ) slavev1 = slavev2 = 1;
	  cnt[el] += p+1 + slavev1 + slavev2;
	}

      blocktable = new Table<int> (cnt);
      cnt = 0;
      for ( int el = 0; el < nel; el++ )
	{
	  double v1 = points[el2vert[el][0]], v2 = points[el2vert[el][1]];
	  if ( v1 <= 0 ) slavev1 = 1; else slavev1 = 0;
	  if ( v2 >= 0 ) slavev2 = 1; else slavev2 = 0;
	  //if ( el == 0 || el == nel - 1 ) slavev1 = slavev2 = 1;
	  for ( int i = 0; i < el2dof[el].Size()-2; i++)
	    (*blocktable)[el][cnt[el]++] = el2dof[el][i];
	  if ( slavev1 ) (*blocktable)[el][cnt[el]++] = el2dof[el][p+1];
	  if ( slavev2 ) (*blocktable)[el][cnt[el]++] = el2dof[el][p+2];
	}
      break;


}
//   cout << *amat << endl;

//   cout << *blocktable << endl;
  BlockJacobiPrecond<Complex> * blockprecond = 
    new BlockJacobiPrecond<Complex> (*amat, *blocktable);

  vecu.FVDouble () = 0.0;
  
  SymmetricGaussSeidel * premat = new SymmetricGaussSeidel ( blockprecond );
  KrylovSpaceSolver * invmat;
  BaseMatrix * invmat2;
  

  
  switch (solver)
    {
    case CG:
      cout << "cg solve for complex problem" << endl;
      invmat = new CGSolver<Complex>(*amat, *blockprecond);
      break;
    case CGMULT:
      cout << "cg solve for complex problem" << endl;
      invmat = new CGSolver<Complex>(*amat, *premat);
      break;
    case CGDIRECT:
      cout << "cg solve for complex problem" << endl;
      invmat2 = dynamic_cast<const BaseSparseMatrix&> (*amat) . InverseMatrix(); 
      invmat = new CGSolver<Complex>(*amat, *invmat2);
      break;
    case QMR:
      cout << "qmr solve for complex problem" << endl;
      invmat = new QMRSolver<Complex>(*amat, *premat);
      break;
    case GMRES:
      cout << "gmres solve for complex problem" << endl;
      invmat = new GMRESSolver<Complex>(*amat, *blockprecond);
      break;
    case GMRESMULT:
      cout << "gmres solve for complex problem" << endl;
      invmat = new GMRESSolver<Complex>(*amat, *premat);
      break;
    case SIMPLE:
      {
	cout << "simple solve for real problem" << endl;
	SimpleIterationSolver<Complex> * hinv = new SimpleIterationSolver<Complex>(*amat, *premat);
	hinv -> SetTau (tau);
	invmat = hinv;
	break;
      }
    case DIRECT:
      cout << "direct solve for real problem" << endl;
      invmat2 = dynamic_cast<const BaseSparseMatrix&> (*amat) . InverseMatrix(); 
      break;
    }
  
  
    if (solver != DIRECT)  
    {
        ma.PushStatus ("Iterative solver");
        invmat->SetMaxSteps (maxsteps);
        invmat->SetPrecision (precision);
        invmat->SetPrintRates ();
        invmat->SetInitialize (0);
      }
    else
      ma.PushStatus ("Direct solver");

    
    if (solver != DIRECT)
      invmat->Mult (vecf, vecu);
    else 
      invmat2->Mult (vecf, vecu);
    
    /*

    BaseVector & vecr = *amat->CreateVector();
    BaseVector & vecw = *amat->CreateVector();
    vecw = 0.0;
    for ( int step = 0; step < maxsteps; step++)
      {
	
	// r = f - A * u
	vecr = vecf - (*amat) * vecu;
	// w = C * r
 	blockprecond -> GSSmooth (vecw, vecr, 1);
 	blockprecond -> GSSmoothBack (vecw, vecr, 1);
	cout << step << "  " << L2Norm(vecr) << endl;
	// vecw = vecr;
	vecu += tau * vecw;

      }
    */
  //  cout << vecu << endl;



  // plot values
  ofstream divsigmaout ("divsigma.dat");
  ofstream sigmaout ("sigma.dat");
  ofstream sigmaeout ( "sigmae.dat");

  int np = 10;
  for ( int el = 0; el < nel; el++ )
    {
      double v1 = points[el2vert[el][0]];
      double v2 = points[el2vert[el][1]];

      double det = v2 - v1;
      for ( int l = 0; l < np; l++ )
	{
	  const IntegrationPoint & ip = 1.0 / np * l;
	  double point = v1 + ip(0)*(v2-v1);
	  fel_hdiv.CalcShape(ip, mat_hdiv);
	  fel_hdiv.CalcDShape(ip, mat_divhdiv);
// 	  cout << " **  shape " << endl <<  mat_hdiv << endl;
// 	  cout << " **  dshape " << endl <<  mat_divhdiv << endl;
	  mat_divhdiv /= det;

	  sigmaout << point << "  ";
	  divsigmaout << point << "  ";
	  double val = 0.0, valdiv = 0.0;
	  for ( int i = 0; i < nd_hdiv_el; i++ )
	    val += mat_hdiv(i) * real(vecu.FVComplex()(el2dof[el][i]));
	  for ( int i = 0; i < nd_hdiv_el; i++ )
	    valdiv -= mat_divhdiv(i,0) * imag(vecu.FVComplex()(el2dof[el][i]))  / kappa;
	  sigmaout << val << endl;
	  divsigmaout << valdiv  << endl;
	}

    }

  for ( int vert = 0; vert < nv; vert++)
    {
      sigmaeout << points[vert] << "  " << real(vecu.FVComplex()(vert2dof[vert][0])) << endl;
    }

  delete amat;
  // delete blocktable;
  delete blockprecond;
  
  delete premat;
  if (solver != DIRECT)  
    delete invmat;
  else
    delete invmat2;
  
  if ( solver == CGDIRECT )
    delete invmat2;

  *testout << vecu << endl;
  }
















  


  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("hh1d", NumProcHH1d::Create, NumProcHH1d::PrintDoc);
 
    }
    
    
    Init init;
    
  }
  


}
