#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

#include <parallelngs.hpp>

// helmholtz equation
// unknowns u, lambda, sigma^E

namespace hybrid_cpp
{

  using namespace ngfem;
  using namespace ngsolve;
  using namespace ngparallel;

  static const MeshAccess * myma;


  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      myma = &ama;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);
      
      Flags l2flags(flags);
      Flags hdivnflags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      l2flags.SetFlag ("order", max(0,order)+2 );
      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderinner", 1.0);
      facetflags.SetFlag("order", order+2);

      spaces[0] = new L2HighOrderFESpace (ma, l2flags);    
      //spaces[0] = new H1HighOrderFESpace (ma, l2flags);    
      spaces[1] = new HDivHighOrderFESpace (ma, hdivnflags);
      spaces[2] = new FacetFESpace (ma, facetflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



#define VER2

  Table<int> * CreateSmoothingBlocks (int type) const
  {
#ifdef VER2
    int ne = ma.GetNE();
    ARRAY<int> dnums;

    ARRAY<int> cnt(ne);
    
    for (int i = 0; i < ne; i++)
      {
        GetExternalDofNrs (i, dnums);
        cnt[i] = dnums.Size();
      }

    Table<int> & it = *new Table<int>(cnt);    
    for (int i = 0; i < ne; i++)
      {
        GetExternalDofNrs (i, dnums);
        for (int j = 0; j < dnums.Size(); j++)
          it[i][j] = dnums[j];
      }
    return &it;
#endif

#ifdef VER1

    das ist nonsense

    // edge by edge
    int nd = GetNDof();
    int ned = ma.GetNEdges();
    int nel = ma.GetNE();
    
    int order_u = (*this)[1]->GetOrder();
    int ndel = ((order_u+1)*(order_u+2))/2;

    ARRAY<int> cnt(ned/* +nel */);
    for (int i = 0; i < ned; i++)
      cnt[i] = 2*order+2;
    /*
    for (int i = ned; i < nel+ned; i++)
      cnt[i] = ndel;
    */
    Table<int> * it = new Table<int>(cnt);
    // Table<int> * it = new Table<int>(ned,2*order+2);
    int base_hdivn = GetStorageStart(2);
    int base_facet = GetStorageStart(3);
    int base_el = GetStorageStart(1);
    for (int i = 0; i < ned; i++)
      {
        (*it)[i][0] = base_hdivn+i;
        for (int j = 0; j < order; j++)
          (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
        (*it)[i][order+1] = base_facet + i;
        for (int j = 0; j < order; j++)
          (*it)[i][order+2+j] = base_facet + ned + i * order + j;
      }
    /*
    for (int i = 0; i < nel; i++)
      {
        (*it)[ned+i] = base_el+i;
        for (int j = 0; j < ndel-1; j++)
          (*it)[ned+i][j+1] = base_el+nel+i*(ndel-1)+j;
      }
    */
    *testout << "table = " << endl << *it << endl;
    return it;
#endif
  }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");


      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
         dynamic_cast<const CompoundFiniteElement&> (fel);

       const NodalFiniteElement & fel_l2 = 
         dynamic_cast<const NodalFiniteElement&> (cfel[0]);

//       const H1HighOrderFiniteElement & fel_l2 = 
//         dynamic_cast<const H1HighOrderFiniteElement&> (cfel[0]);
      const HDivFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
      
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();
      int base_l2 = 0;
      int base_hdivn = base_l2+nd_l2;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      FlatVector<> mat_l2(nd_l2, lh);
      FlatVector<> fmat_l2(nd_l2, lh);

      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      FlatVector<> mat_dudn(nd_l2, lh);
      
      FlatMatrix<> hmat_l2(nd_l2, lh);
      FlatMatrixFixHeight<D> bmat_grad(nd_l2, lh);
      FlatMatrixFixHeight<D> fbmat_grad(nd_l2, lh);

      FlatMatrixFixHeight<4> bmat(nd, lh);
      FlatMatrixFixHeight<4, Complex> dbmat(nd, lh);
      Mat<4,4,Complex> dmat;

      FlatMatrix<> mat_l2l2 (nd_l2, lh);
      FlatMatrix<> mat_gradgrad (nd_l2, lh);

      double kappa = coef_kappa->EvaluateConst();
          
      const IntegrationRule & ir2d =
        SelectIntegrationRule (eltype, 2*fel_l2.Order());

      
      void * heapp = lh.GetPointer();

      /*
      hmat_l2 = 0.0;
      mat_l2l2  = 0.0;
      mat_gradgrad = 0.0;

      for (int l=0; l<ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<2,2> sip(ir2d[l], eltrans, lh);
          
          double factor = sip.GetJacobiDet() * ir2d[l].Weight();
          
          bmat_grad = Trans (sip.GetJacobianInverse ()) * 
            Trans (fel_l2.GetDShape(sip.IP(),lh));

          fel_l2.CalcShape(sip.IP(), mat_l2);
          
          fmat_l2 = factor * mat_l2;
          mat_l2l2 += fmat_l2 * Trans (mat_l2);

          fbmat_grad = factor * bmat_grad,
            mat_gradgrad +=  Trans (fbmat_grad) * bmat_grad;

          lh.CleanUp(heapp);
        }
      
      mat_l2l2 -= 1.0/(kappa*kappa) * mat_gradgrad;
      for (int i = 0; i < nd_l2; i++)
        for (int j = 0; j < nd_l2; j++)
          elmat(base_l2+j, base_l2+i) += Complex(0, -kappa) * mat_l2l2(i,j);
      */

      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned;
      switch (eltype)
        {
        case ET_TRIG : ned = 3; break;
        case ET_QUAD : ned = 4; break;
        case ET_TET : ned = 4; break;
        }

      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      vu = 0.0;
      vv = 0.0;


      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);


      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

      for (int k = 0; k < ned; k++)
        {
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];
        
          const IntegrationRule & ir1d =
            GetIntegrationRules().SelectIntegrationRule (etfacet, 2 * fel_l2.Order());

          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

              Vec<D> nt = Trans(sip.GetJacobian()) * normal;
              nt /= sip.GetJacobiDet();

              //Piola

              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              fel_l2.CalcShape(sip.IP(), mat_l2);

              Vec<D> nt2 = sip.GetJacobianInverse() * normal;
              mat_dudn = fel_l2.GetDShape(sip.IP(), lh) * nt2;

              double factor = len * ir1d[l].Weight();

              bmat = 0.0;
              bmat.Row(0).Range (base_l2   , base_l2   +nd_l2   ) = mat_dudn;
              bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;

              /*
                1/ik   -1   1     1
                -1     -ik    0    ik
                -1     0   -ik   0
                1     ik    0    -ik
              */


              dmat = 0.0;
              Complex ik (0, kappa);

              dmat(1,0) = -1;
              dmat(3,0) = 1;

              dmat(0,1) = -1;
              dmat(0,3) = 1;

              dmat(2,2) = -ik;
              dmat(2,0) = -1;
              dmat(0,2) = 1;
              dmat(0,0) = 1.0 / ik;

	      
              dmat(1,1) = -ik;
              dmat(1,3) = ik;
              dmat(3,1) = ik;
              dmat(3,3) = -ik;
	      

              dmat *= factor;
              dbmat = dmat * bmat;

              for (int i = base_l2; i < base_l2 + nd_l2; i++)
                elmat.Row(i) += mat_l2(i) * dbmat.Row(1) + mat_dudn(i) * dbmat.Row(0);

              for (int i = 0; i < nd_facet; i++)
                elmat.Row(base_facet+i) += mat_facet(i) * dbmat.Row(3);

              for (int i = 0; i < nd_hdivn; i++)
                elmat.Row(base_hdivn+i) += mat_hdivn_n(i) * dbmat.Row(2);

              lh.CleanUp(heapp);
            }
        }
    }
  };



  template <int D>
  class HybridRobinIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<> & elmat,
                                        LocalHeap & lh) const
    {
      cerr << "Helmholtz - Robin real called" << endl;
    }




    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const HDivNormalFiniteElement<D-1> & hdivnfel = 
        dynamic_cast<const HDivNormalFiniteElement<D-1>&> (cfel[1]);
      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int nd_hdivn = hdivnfel.GetNDof();
      int nd_facet = ffel.GetNDof();
      int base_facet = nd-nd_facet;
      int base_hdivn = base_facet-nd_hdivn;
      
      FlatVector<> shape(nd_facet, lh), shapehdivn(nd_hdivn, lh);
      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ffel.ElementType(), 2*ffel.Order());
      
      FlatMatrixFixHeight<2> bmat(nd, lh);
      FlatMatrixFixHeight<2> dbmat(nd, lh);
      Mat<2> dmat(2);

      vu = 0.0;
      vv = 0.0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<D-1,D> sip(ir1d[l], eltrans, lh);
          double val = coef->Evaluate(sip);

          if (val == 0) continue;

          hdivnfel.CalcShape(ir1d[l], shapehdivn);
          shapehdivn *= 1.0 / sip.GetJacobiDet();

          ffel.CalcShape(ir1d[l], shape);
          /*
          vu.Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
          vu.Range (base_facet, nd) = shape;
          elmat +=  (sip.GetJacobiDet()*ir1d[l].Weight())*(vu*Trans(vu));
          */

          double factor = sip.GetJacobiDet()*ir1d[l].Weight();

          bmat = 0.0;
          bmat.Row(0).Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
          bmat.Row(1).Range (base_facet, nd) = shape;

          dmat = 0.0;
          dmat(1,1) =  1;

          dbmat = dmat * bmat;
          elmat += Complex (0, -val*factor) * Trans (bmat) * dbmat;
        }
    }
  };



  class HybridNeumannIntegrator2d : public LinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridNeumannIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : LinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridNeumannIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridNeumannIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementVector (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatVector<double> & elvec,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      //       cout << "** neumann: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      int nd = cfel.GetNDof();  
      
      elvec.AssignMemory(nd, lh);
      elvec = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      

      FlatVector<> vec(ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      vec=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          vec +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*shape;
          //         vec +=  ir1d[l].Weight()*coef->Evaluate(sip)*shape;
        }
      
      // copy to global matrix
      elvec.Range(ndr,nd) = vec;
    }


  };

  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz", HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 2, 1,
                                      HybridHelmholtzIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 3, 1,
                                      HybridHelmholtzIntegrator<3>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator<2>::Create);

    GetIntegrators().AddBFIntegrator ("hybridrobin", 3, 1,
                                      HybridRobinIntegrator<3>::Create);

    GetIntegrators().AddLFIntegrator ("hybridneumann", 2, 1,
                                      HybridNeumannIntegrator2d::Create);
  
  }

  Init init;

} // namespace

