#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


// Helmholtz equation, unknowns sigma on tets, lambda, on facets
// matches single.pdf

namespace hybrid_sigma_cpp
{

  using namespace ngfem;
  using namespace ngsolve;
  using namespace hybrid_sigma_cpp;

  Table<int> * isslave;
  MeshAccess * ma2;
  
  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {

  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      ;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(2);
      
      Flags hdivflags(flags);
      Flags hdivnflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      hdivflags.SetFlag ("order", order);
      hdivflags.SetFlag ("discontinuous");
      hdivflags.SetFlag ("orderinner", order);
      hdivflags.SetFlag ("curlorder", order);
      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderinner", 1.0);

      spaces[0] = new HDivHighOrderFESpaceA (ma, hdivflags);
      spaces[1] = new FacetFESpace (ma, hdivnflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      isslave = new Table<int> (ma.GetNE(), 4);
      for ( int i = 0; i < ma.GetNE(); i++ )
	(*isslave)[i] = 1;
      ma2 = const_cast<MeshAccess*>(&ma);
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }


    // direct solver clusters:

    // 0 ...... no clusters
    // 1 ...... low order facets
    // >= 2 ... divide sphere into n layers

    virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const
    {
      int clustertype = int(precflags.GetNumFlag("ds_cluster",0)); 

      ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());
      clusters = 0;

      int nd = GetNDof();
      int nse = ma.GetNSE();
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
            
      int dim = ma.GetDimension();

      int order_sigmaE = (*this)[1]->GetOrder();
      int order_sigma = (*this)[0]->GetOrder();
      
      ARRAY<int> facets, dnums;
      if ( clustertype == 0 )
	{
	  clusters = 0;
	  return 0; // &clusters;
	}
      else if ( clustertype == 1 )
	{
	  int nf = (ma.GetDimension() == 2 ) ? 3 : 4;
	  int nt = nf;
	  nfacets = ma.GetNFacets();
	  for ( int el = 0; el < nel; el++ )
	    {
// 	      spaces[0]->GetDofNrs(el, dnums);
// 	      for (int i = 0; i < nt; i++ )
// 		clusters[dnums[i]] = 1;
	      for ( int i = 0; i < nfacets; i++ )
		clusters[GetStorageStart(1) + i ] = 1;

	    }
// 	  for ( int i = spaces[0]->GetNDof() + nf; i < GetNDof(); i++)
// 	    clusters[i] = i;
	  return &clusters;
	}


      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < nse; sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 0 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
   		    int cluster = int( (1.0 * (nel-1) * clustertype) / nel) - int( (1.0 * nmarkedels * clustertype) / nel) +1 ;
//   		    int cluster = int( (1.0 * nmarkedels * clustertype) / nel) +1 ;
		    if ( ma.GetDimension() == 3 )
		      ma.SetElementPartition(el, cluster);
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      newmarkedfacets.Set(facets[j]);

		    // set cluster .. 1 if el < nel/2, 2 else

		    GetDofNrs (el, dnums);
		    for (int j = 0; j < dnums.Size(); j++)
		     if ( !clusters[dnums[j]] )
		      clusters[dnums[j]] = cluster;

		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa) ) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;

      return &clusters;
    } 


    // #define VER1
    Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
    {
      int blocktype = int(precflags.GetNumFlag("blocktype",1)); 
      Table<int> * it = CreateSmoothingBlocks ( blocktype );

    }


    Table<int> * CreateSmoothingBlocks (int type=1) const
    {
      int nd = GetNDof();
      int nse = ma.GetNSE();
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
            
      int dim = ma.GetDimension();

      int order_sigmaE = (*this)[1]->GetOrder();
      int order_sigma = (*this)[0]->GetOrder();
      

      const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace &> (*spaces[1]);

      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel), orderedfacets(nfacets);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);
	  int ndof_per_edge, ndof_per_trig;

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      int ifacet = 0;

      ARRAY<int> facets;
      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < nse; sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 0 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	  orderedfacets[ifacet++] = facets[0];
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      {
			if (!markedfacets.Test(facets[j]) && !newmarkedfacets.Test(facets[j]))
			  orderedfacets[ifacet++] = facets[j];
			newmarkedfacets.Set(facets[j]);
		      }
		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa)) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;

      cout << "SmoothingType " << type << endl; 
      // edge by edge

      Table<int> * it;
      int base_hdivn = GetStorageStart(1);
      int ndof = GetNDof();
      ARRAY<int> cnt;
      ARRAY<int> dnums;
      ARRAY<int> block;
      BitArray issurffacet(nfacets);
      
//       Smoothing types:
//      1 .. facet by facet (eliminate_internal)
// 	2 .. element by element  (eliminate_internal)
// 	3 .. nichts
// 	4 .. master-edges/element (noelim_int)
// 	5 .. ds_clusters (noelim_int)
// 	6 .. vertex-patches (noelim_int)
// 	62 .. vertex-patches (elim_int)
// 	7 .. facet-patches (elim_int)

      if ( dim == 2 ) 
	{
	  ndof_per_edge = order_sigmaE +1;
	  ndof_per_trig = int((order_sigma+1)*(order_sigma+2));
	}
      else
	{
	  ndof_per_edge = int(0.5*(order_sigmaE+1)*(order_sigmaE+2));
	  ndof_per_trig = ((order_sigma+1)*(order_sigma+2)*(order_sigma+3))/2;
	}

      issurffacet.Clear();
      for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	if ( ma.GetSElIndex(sel) == 0 || ma.GetSElIndex(sel) == 2 )
	  {
	    ma.GetSElFacets(sel, facets);
	    issurffacet.Set(facets[0] );
	  }

      
      for ( int el = 0; el < nel; el++ )
	{
	  ma.GetElFacets(el, facets);
	  for ( int i = 0; i < facets.Size(); i++ )
	    {
	      if ( issurffacet.Test(facets[i]) ) 
		(*isslave)[el][i] = 0;
	    }
	}
      

      switch (type)
	{
	case 1:
	case 12:
	  cout << "facet by facet, " ;
	  if ( type == 1 ) cout << "no" ;
	  cout << "eliminate_internal";

	  cnt.SetSize(nfacets); // ned/* +nel */);

	  for (int i = 0; i < nfacets; i++)
	      cnt[i] = ndof_per_edge;

	  it = new Table<int>(cnt);
	  // Table<int> * it = new Table<int>(ned,2*order+2);
	  cnt = 0;
	  for (int i = 0; i < nfacets; i++)
	    {
	      int ifacet = orderedfacets[i];
	      facetspace.GetFacetDofNrs(ifacet, dnums);
	      for ( int j = 0; j < dnums.Size(); j++)
		(*it)[i][cnt[i]++] = base_hdivn + dnums[j];
 
// 	      (*it)[i][0] = base_hdivn + ifacet;
// 	      for (int j = 0; j < ndof_per_edge-1; j++)
// 		(*it)[i][j+1] = base_hdivn + nfacets + (ifacet * (ndof_per_edge-1)) + j;
	    }


	  // *testout << "table = " << endl << *it << endl;
	  return it;
	  break;

	case 2:
	case 22:
	  // #ifdef VER2
	  // all external dofs of an element
	  cout << "element by element" << endl;
	  if ( type == 2 ) cout << "no" ;
	  cout << "eliminate_internal";
	  
	  cnt.SetSize(nel);
	  
	  for (int i = 0; i < nel; i++)
	    {
	      if ( type == 2 )
		GetDofNrs (orderedels[i], dnums);
	      else 
		GetExternalDofNrs(orderedels[i], dnums);
	      cnt[i] = dnums.Size();
	    }
	  
	  it = new Table<int>(cnt);    
	  for (int i = 0; i < nel; i++)
	    {
	      if ( type == 2 )
		GetDofNrs (orderedels[i], dnums);
	      else
		{
		  GetExternalDofNrs(orderedels[i], dnums);
// 		  for ( int j = 0; j < dnums.Size(); j++)
// 		    dnums[j] += spaces[0]->GetNDof();
		}
	      for (int j = 0; j < dnums.Size(); j++)
		(*it)[i][j] = dnums[j];
	    }
	  return it;
	  // #endif
	  break;

	case 3:
	  // no blocks
	  cnt.SetSize(1);
	  cnt[0] = 0;
	  it = new Table<int> (cnt);
	  return it;
	  break;

	case 5:
	  cnt.SetSize(nel);
	  block.SetSize(ndof);
	  cnt = 0;
	  for ( int eli = 0; eli < nel; eli++ )
	    {
	      int el = orderedels[eli];
	      GetDofNrs(el, dnums);
	      for ( int i = 0; i < dnums.Size(); i++ )
		block[dnums[i]] = (eli)/20;

	    }
	  for ( int dof = 0; dof < ndof; dof++ )
	    {
	      cnt[block[dof]]++;
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for ( int dof = 0; dof < ndof; dof++ )
	    (*it)[block[dof]][cnt[block[dof]]++] = dof;

	  return it;
	  break;
	case 4:
	case 42:
	  cout << "master edges of element" << endl;
	  if ( type == 2 ) cout << "no" ;
	  cout << "eliminate_internal";

	  cnt.SetSize(nel);
	 
	  for ( int i = 0; i < nel; i++ )
	    {
	      ma.GetElFacets(i, facets);
	      cnt[i] = facets.Size();
	    }
	  // Table<int> isslave (cnt);
	  cnt = 0;

	      for ( int eli = 0; eli < nel; eli++ )
		{
		  int el = orderedels[eli];
		  if ( type == 4 )
		    GetDofNrs(el, dnums); 
		  else
		    spaces[1]->GetDofNrs(el, dnums);

		  cnt[eli] += dnums.Size();
		  ma.GetElFacets(el, facets);
		  for ( int i = 0; i < facets.Size(); i++ )
		    {

		      if ( (*isslave)[el][i] )
			cnt[eli] -= ndof_per_edge;

		    }
		}

	  it = new Table<int> (cnt);
	  cnt = 0;

	      for ( int eli = 0; eli < nel; eli++)
		{
		  int el = orderedels[eli];
		  ma.GetElFacets(el, facets);
		  GetDofNrs(el, dnums); 
		  if ( type == 4 )
		    for ( int i = 0; i < dnums.Size() - facets.Size() * (ndof_per_edge) ; i++)
		      (*it)[eli][cnt[eli]++] = dnums[i];
		  for ( int i = 0; i < facets.Size(); i++ )
		    {
		      if ( !(*isslave)[el][i] )
			{
			  facetspace.GetFacetDofNrs(facets[i], dnums);
			  for ( int j = 0; j < dnums.Size(); j++)
			    (*it)[eli][cnt[eli]++] = base_hdivn + dnums[j];

// 			  (*it)[eli][cnt[eli]++] = facets[i] + base_hdivn;
// 			  for (int j = 0; j < ndof_per_edge-1; j++)
// 			    (*it)[eli][cnt[eli]++] = base_hdivn + nfacets + (facets[i] * (ndof_per_edge-1)) + j;

// 			  for (int j = 0; j < ndof_per_edge; j++)
// 			    (*it)[eli][cnt[eli]++] = base_hdivn +  (facets[i] * ndof_per_edge ) + j;
			}
		      
		    }
		}
	  // *testout << "it " << it << endl;
	  return it;
	  break;

	case 6:
	  cnt.SetSize(ma.GetNV() );
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		  cnt[verts[i]] += ndof_per_trig;
		  for ( int k = 0; k < facets.Size(); k++)
		    if ( !(*isslave)[el][k] )
		      cnt[verts[i]] += ndof_per_edge;
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		 spaces[0]->GetDofNrs(el, dnums); 
 		 for ( int j = 0; j < dnums.Size() ; j++)
 		   (*it)[verts[i]][cnt[verts[i]]++] = dnums[j];
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + dnums[j];
// 		       (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + nfacets + 
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  // *testout << "table.. " << endl << *it << endl;
	  return it;
	  break;

	case 62:
	  // mit eliminate internal
	  cnt.SetSize(ma.GetNV() );
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		  for ( int k = 0; k < facets.Size(); k++)
		    if ( !(*isslave)[el][k] )
		      cnt[verts[i]] += ndof_per_edge;
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < verts.Size(); i++ )
		{
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + dnums[j];
// 		       (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[verts[i]][cnt[verts[i]]++] = base_hdivn + nfacets + 
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  //*testout << "table.. " << endl << *it << endl;
	  return it;
	  break;

	case 7:
	case 72:

	  if ( type == 7 )
	    cnt.SetSize(ma.GetNFacets() );
	  else
	    cnt.SetSize(ma.GetNFacets()  + ma.GetNE() );

	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums(el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  spaces[0]->GetDofNrs(el, dnums); 
		  if ( type == 7 )
		    cnt[facets[i]] += dnums.Size();
		  else
		    cnt[nfacets + el] += dnums.Size();
 
		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       cnt[facets[i]] += dnums.Size();
		     }
		}
	    }
	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> verts;
	      ma.GetElPNums( el, verts);
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  spaces[0]->GetDofNrs(el, dnums); 
		  if ( type == 7 )
		    {
		      for ( int j = 0; j < dnums.Size() ; j++)
			(*it)[facets[i]][cnt[facets[i]]++] = dnums[j];
		    }
		  else
		    for ( int j = 0; j < dnums.Size() ; j++)
		      (*it)[nfacets+el][cnt[nfacets+el]++] = dnums[j];

		 for ( int k = 0; k < facets.Size(); k++)
		   if ( !(*isslave)[el][k] )
		     {
		       facetspace.GetFacetDofNrs(facets[k], dnums);
		       for ( int j = 0; j < dnums.Size(); j++)
			 (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn + dnums[j];

// 		       (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn  + facets[k];
// 		       for (int j = 0; j < ndof_per_edge-1; j++)
// 			 (*it)[facets[i]][cnt[facets[i]]++] = base_hdivn  + nfacets +
// 			   (facets[k] * (ndof_per_edge-1) ) + j;
		     }
		}
	    }
 	  // *testout << "table.. " << endl << *it << endl;
	  return it;
	  break;


	case 8:
	  cnt.SetSize(nel);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ma.GetElFacets(el, facets);

	      cnt[el] += ndof_per_trig;

		  for ( int k = 0; k < facets.Size(); k++)
		    {
		      cnt[el] += ndof_per_edge;
		      if ( !(*isslave)[el][k] )
			cnt[el] += ndof_per_trig;
		    }
	    }

	  it = new Table<int> (cnt);
	  cnt = 0;
	  for( int el=0; el<nel; el++)
	    {	
	      ARRAY<int> neighbours;
	      ma.GetElFacets(el, facets);
	      spaces[0]->GetDofNrs(el, dnums); 
	      for ( int j = 0; j < dnums.Size() ; j++)
		(*it)[el][cnt[el]++] = dnums[j];
	      for ( int k = 0; k < facets.Size(); k++)
		{
		  (*it)[el][cnt[el]++] = base_hdivn  + facets[k];
		  for (int j = 0; j < ndof_per_edge-1; j++)
		    (*it)[el][cnt[el]++] = base_hdivn  + 
		      (facets[k] * (ndof_per_edge-1) ) + j;
		  if ( !(*isslave)[el][k] )
		    {
		      if ( dim == 2 )
			ma.GetEdgeElements(facets[k], neighbours);
		      else
			ma.GetFaceElements(facets[k], neighbours);
		      for( int ii = 0; ii < neighbours.Size(); ii++ )
			if ( neighbours[ii] != el )
			  {
			    //*testout << "el " << el << "facet " << k << ": " << neighbours[ii] << endl;
			    spaces[0]->GetDofNrs(neighbours[ii], dnums);
			    for ( int j = 0; j < dnums.Size() ; j++)
			      (*it)[el][cnt[el]++] = dnums[j];
			  }
		    }
		}
	      
	    }
	  
	  break;

	case 9:
	  cnt.SetSize(spaces[1]->GetNDof() );
	  cnt = 1;
	  it = new Table<int> (cnt);
	  for ( int i = base_hdivn; i < ndof; i++ )
	    (*it)[i-base_hdivn][0] = i;
	  break;

	default:

	  cout << "Illegal Smoothing Type " << type << endl;
	  cout << "set flag -blocktype=<type> where <type> is from" << endl <<
	    "#       Smoothing types:\n #       1 .. facet by facet (noeliminate_internal) \n #       12 .. facet by facet (eliminate_internal)\n # 	2 .. element by element  (noeliminate_internal)\n # 	22 .. element by element  (eliminate_internal)\n # 	3 .. nichts\n # 	4 .. master-edges/element (noelim_int)\n # 	42 .. master-edges/element (elim_int)\n # 	5 .. ds_clusters (noelim_int)\n # 	6 .. vertex-patches (noelim_int)\n # 	62 .. vertex-patches (elim_int)\n # 	7 .. facet-patches (elim_int)" << endl;
	}
      return it;
    }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      int elnum = eltrans.GetElementNr();
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");
      clock_t start_calcshape, end_calcshape, time_cs = 0;
      clock_t start_mult, end_mult, time_mult = 0;
      clock_t start_add, end_add, time_add = 0;

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const FacetVolumeFiniteElement & fel_hdivn = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[1]);

      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();

      int base_hdiv = 0;
      int base_hdivn = nd_hdiv;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      enum { BLOCK = 5} ;
      
      FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      
      FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh), scalbmat_sigma(nd_hdiv, lh)  ; 
      FlatVector<> div_sigma(nd_hdiv, lh), scaldiv_sigma(nd_hdiv, lh) ; // (nd_hdiv, lh);
      
      FlatMatrixFixHeight<3, Complex> bmat(nd, lh);
      FlatMatrixFixHeight<3, Complex> dbmat(nd, lh);
      Mat<3, 3, Complex> dmat, dmat2;
      FlatMatrix<Complex> dbdmat1(nd, lh), dbdmat2(nd, lh);
      dbdmat1 = 0.0;
      dbdmat2 = 0.0;
      // int ik sigma tau dx
      FlatMatrix<> mat_iksigmasigma (nd_hdiv, lh);
      // int 1/ik div sigma div tau dx
      FlatMatrix<> mat_iikdivdiv (nd_hdiv, lh);

      FlatMatrixFixHeight<D> diag_iksigmasigma (nd_hdiv, lh);

      double kappa = coef_kappa->EvaluateConst();
          

      const IntegrationRule & irDd =
        SelectIntegrationRule (eltype, 2*fel_hdiv.Order() );
      
      
      void * heapp = lh.GetPointer();
      mat_iksigmasigma = 0.0;
      mat_iikdivdiv = 0.0;
      diag_iksigmasigma = 0.0;

      bool curvedelement =  eltrans.IsCurvedElement();
      if ( curvedelement )      
	{
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
      
          
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));
	      mat_iksigmasigma += Trans(scalbmat_sigma) * (bmat_sigma);
	      lh.CleanUp(heapp);
	    }

	  //	  *testout << "mat_iksigmasigma " << endl << mat_iksigmasigma << endl;
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      imag(elmat(base_hdiv+i, base_hdiv+j)) +=  mat_iksigmasigma(i,j);


	  // int 1/ik divsigma divtau == - int i/k divsigma divtau 
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	}

      else // FAST

	{
	  
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
	      
	      
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	       // mass for l2^3 is diagonal
	      for ( int ii = 0; ii < nd_hdiv / D ; ii ++ )
		{
		  for ( int mi = 0; mi < D; mi++ )
		    for ( int mj = 0; mj < D; mj++ )
		      for ( int j = 0; j < D; j++)
			diag_iksigmasigma(mi, D*ii + mj) += scalbmat_sigma(j, D*ii+mi) * bmat_sigma(j, D*ii+mj);
		}
	       
	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));

	      lh.CleanUp(heapp);
	    }
	  
	  
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv / D; i++)
	    for ( int mi = 0; mi < D; mi++ )
	      for ( int mj = 0; mj < D; mj++ )
		imag(elmat(base_hdiv+i*D+mi, base_hdiv+i*D+mj)) +=  diag_iksigmasigma(mi,D*i+mj);
	  
	 	  
	  // - int i / k sigma . tau  --> diagonal
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	   
	  lh.CleanUp(heapp);
	  
  	}
      
   //    *testout << "elmat inner " << endl << elmat << endl;

      /*
	cout << "timings: cs: " << double(time_cs)/CLOCKS_PER_SEC << 
	"\n      mult: " << double(time_mult)/CLOCKS_PER_SEC << 
	"\n     add: " << double(time_add)/CLOCKS_PER_SEC << endl;
      */
      
      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     
      
      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;


      FlatMatrix<> mat_sigmansigman (int(nd_hdiv/D), lh);
      mat_sigmansigman = 0.0;
      mat_iksigmasigma = 0.0;

      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;
      ARRAY<int> facetk(ned);
      for ( int i = 0; i < ned; i++ )
	facetk[i] = i;

    //   *testout  << "inner: as(elmat) = " << elmat-Trans(elmat) << endl;
      for (int k = 0; k < ned; k++)
        {
	  mat_sigmansigman = 0.0;
	  Vec<D> normal_ref;
	  for (int i=0; i<D; i++)
	    normal_ref(i) = normals[k][i];
        
	  /*
          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
        
          tau_ref  = pe-ps;
	  */

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*fel_hdiv.Order() );

          void * heapp = lh.GetPointer();

	  Mat<D> ninj;
	  ninj = 0.0;
          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
	      IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

            // Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
	      // IntegrationPoint ip( p2d(0), p2d(1) );
              // SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();
	      double invdet = 1.0 / det;

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

              
              Vec<D> nt = Trans(jac) * normal;
              nt /= det;

	      // calc the tensor product n n^T for fast evaluation of (sigma . n) (tau . n)
	      if ( l == 0 )
		for (int i = 0; i < D; i++ )
		  for ( int j = 0; j < D; j++ )
		    ninj(i,j) = nt(i) * nt(j);
              //Piola

	      // Vec<3> e1;
	      // e1 = 1.0;
	      bmat_sigma = Trans( fel_hdiv.GetShape(sip.IP(), lh) );
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
              mat_hdiv_n = Trans ( bmat_sigma ) * nt;
	      mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh);


              double factor = len * ir1d[l].Weight();

	      // penalty ... high on master, 0 on slave
	      double penalty = 1e7;
	      
	      double nx = 0.0;
	      Vec<D> point;
	      eltrans.CalcPoint ( ip, point, lh);

	      for ( int i = 0; i < D; i++ )
		nx += point(i) * normal(i);

	      nx /= L2Norm(point);
	      nx /= L2Norm(normal);

	      if ( nx < 0 ) penalty = 0;
	      // isslave = 0 --> im smoothing block
	      if ( nx < 0.0 ) (*isslave)[elnum][k] = 0;



 	      double fac = 1.0;
 	      if ( nx < 0.0 ) fac = -1.0;
//               bmat = 0.0;
//               bmat.Row(0).Range (base_hdiv, base_hdiv +nd_hdiv) = div_sigma;
//               bmat.Row(2).Range (base_hdiv, base_hdiv +nd_hdiv) = mat_hdiv_n;
//               bmat.Row(2).Range (base_hdivn, base_hdivn +nd_hdivn) = -mat_hdivn_n;
// 	      bmat.Row(2).Range (base_hdiv, base_hdiv +nd_hdiv) += 
// 		Complex(0, fac/kappa) * div_sigma;

// 	      bmat.Row(1).Range (base_hdiv , base_hdiv +nd_hdiv) = mat_hdiv_n;
// 	      bmat.Row(1).Range (base_hdivn , base_hdivn +nd_hdivn) = -mat_hdivn_n;


//               /*
// 		0     i/k
// 		i/k   p+1
//               */


//               dmat = 0.0;
// 	      dmat(0,0) = -fac/kappa/kappa;
// 	      dmat(1,0) = dmat(0,1) = Complex(0.0, 1.0/kappa);
// 	      dmat(2,2) = 1; 

	      dmat = 0.0; 
	      bmat = 0.0;
	      if ( nx > 0 ) // T_out
		{
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma;

		  /* JOACHIM */
// 		  // V1
 		  dmat(0,0) = -1.0;
 		  dmat(1,0) = dmat(0,1) = -1.0;
// 		  // V2
//  		  dmat(0,0) = 1.0;
//  		  dmat(1,0) = dmat(0,1) = -1.0;

// 		  dmat(2,2) = 1.0;
// 		  dmat(0,2) = dmat(2,0) = -1.0;
// 		  dmat(1,2) = dmat(2,1) = -1.0;
// 		  dmat(1,1) = 1.0;

		  // stabilization joachim
// 		  Complex pen = 1;
// 		  dmat(1,1) += pen;
// 		  dmat(0,0) += pen;
// 		  dmat(2,2) += pen;
// 		  dmat(0,1) += pen;
// 		  dmat(1,0) += pen;
// 		  dmat(2,0) -= pen;
// 		  dmat(0,2) -= pen;
// 		  dmat(1,2) -= pen;
// 		  dmat(2,1) -= pen;

// 		  /* ASTRID */
// 		  dmat(1,0) = -1.0; 
// 		  dmat(0,1) = -1.0;
// 		  dmat(2,1) = 1.0;
// 		  dmat(1,2) = 1.0;
// 		  dmat(2,2) = -1.0;

// 		  // stabilization, (u-lambda)(v-mu) daempft
//  		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
//  		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// // 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// // 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;
		}

	      else // T_in
		{
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma;

		  /* JOACHIM */
// 		  // V1
// 		  dmat(2,2) = -1.0;
// 		  dmat(0,2) = dmat(2,0) = -1.0;
// 		  dmat(1,2) = dmat(2,1) = 1.0;
// 		  dmat(1,1) = -1.0;
//
		  // V2
// 		  dmat(2,2) = 1.0;
// 		  dmat(0,2) = dmat(2,0) = -1.0;
// 		  dmat(1,2) = dmat(2,1) = -1.0;
// 		  dmat(1,1) = 1.0;

 		  dmat(0,0) = 1.0;
 		  dmat(1,0) = dmat(0,1) = -1.0;
		  // stabilization joachim
		  Complex pen = 1;
		  dmat(1,1) += pen;
		  dmat(0,0) += pen;
		  dmat(2,2) += pen;
		  dmat(0,1) -= pen;
		  dmat(1,0) -= pen;
		  dmat(2,0) += pen;
		  dmat(0,2) += pen;
		  dmat(1,2) -= pen;
		  dmat(2,1) -= pen;


// 		  /* ASTRID */
// 		  dmat(0,1) = -1.0;
// 		  dmat(1,1) = -1.0;
// 		  dmat(1,0) = -1.0;

// 		  // stabilization, (u-lambda)(v-mu) daempft
//  		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
//  		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// // 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// // 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;

		}

              dmat *= factor;

//  	      *testout << "dmat = " << endl << dmat << endl;
//  	      *testout << "bmat = " << endl << bmat << endl;

              dbmat = dmat * bmat;

              elmat += Trans (bmat) * dbmat;

	      lh.CleanUp(heapp);
            } 
	  
	
        }

    }
  };
  




  template <int D>
  class HybridHelmholtzIntegratorP : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegratorP (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegratorP () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegratorP (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      int elnum = eltrans.GetElementNr();
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");
      clock_t start_calcshape, end_calcshape, time_cs = 0;
      clock_t start_mult, end_mult, time_mult = 0;
      clock_t start_add, end_add, time_add = 0;

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const FacetVolumeFiniteElement & fel_hdivn = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[1]);

      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();

      int base_hdiv = 0;
      int base_hdivn = nd_hdiv;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      enum { BLOCK = 5} ;
      
      FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      
      FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh), scalbmat_sigma(nd_hdiv, lh)  ; 
      FlatVector<> div_sigma(nd_hdiv, lh), scaldiv_sigma(nd_hdiv, lh) ; // (nd_hdiv, lh);
      
      FlatMatrixFixHeight<3, Complex> bmat(nd, lh);
      FlatMatrixFixHeight<3, Complex> dbmat(nd, lh);
      Mat<3, 3, Complex> dmat, dmat2;
      FlatMatrix<Complex> dbdmat1(nd, lh), dbdmat2(nd, lh);
      dbdmat1 = 0.0;
      dbdmat2 = 0.0;
      // int ik sigma tau dx
      FlatMatrix<> mat_iksigmasigma (nd_hdiv, lh);
      // int 1/ik div sigma div tau dx
      FlatMatrix<> mat_iikdivdiv (nd_hdiv, lh);

      FlatMatrixFixHeight<D> diag_iksigmasigma (nd_hdiv, lh);

      double kappa = coef_kappa->EvaluateConst();
          

      const IntegrationRule & irDd =
        SelectIntegrationRule (eltype, 2*fel_hdiv.Order() );
      
      
      void * heapp = lh.GetPointer();
      mat_iksigmasigma = 0.0;
      mat_iikdivdiv = 0.0;
      diag_iksigmasigma = 0.0;

      bool curvedelement =  eltrans.IsCurvedElement();
      if ( curvedelement )      
	{
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
      
          
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));
	      mat_iksigmasigma += Trans(scalbmat_sigma) * (bmat_sigma);
	      lh.CleanUp(heapp);
	    }

	  //	  *testout << "mat_iksigmasigma " << endl << mat_iksigmasigma << endl;
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      imag(elmat(base_hdiv+i, base_hdiv+j)) +=  mat_iksigmasigma(i,j);


	  // int 1/ik divsigma divtau == - int i/k divsigma divtau 
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	}

      else // FAST

	{
	  
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
	      
	      
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	       // mass for l2^3 is diagonal
	      for ( int ii = 0; ii < nd_hdiv / D ; ii ++ )
		{
		  for ( int mi = 0; mi < D; mi++ )
		    for ( int mj = 0; mj < D; mj++ )
		      for ( int j = 0; j < D; j++)
			diag_iksigmasigma(mi, D*ii + mj) += scalbmat_sigma(j, D*ii+mi) * bmat_sigma(j, D*ii+mj);
		}
	       
	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));

	      lh.CleanUp(heapp);
	    }
	  
	  
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv / D; i++)
	    for ( int mi = 0; mi < D; mi++ )
	      for ( int mj = 0; mj < D; mj++ )
		imag(elmat(base_hdiv+i*D+mi, base_hdiv+i*D+mj)) +=  diag_iksigmasigma(mi,D*i+mj);
	  
	 	  
	  // - int i / k sigma . tau  --> diagonal
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	   
	  lh.CleanUp(heapp);
	  
  	}
      
   //    *testout << "elmat inner " << endl << elmat << endl;

      /*
	cout << "timings: cs: " << double(time_cs)/CLOCKS_PER_SEC << 
	"\n      mult: " << double(time_mult)/CLOCKS_PER_SEC << 
	"\n     add: " << double(time_add)/CLOCKS_PER_SEC << endl;
      */
      
      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     
      
      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;


      FlatMatrix<> mat_sigmansigman (int(nd_hdiv/D), lh);
      mat_sigmansigman = 0.0;
      mat_iksigmasigma = 0.0;

      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;
      ARRAY<int> facetk(ned);
      for ( int i = 0; i < ned; i++ )
	facetk[i] = i;

    //   *testout  << "inner: as(elmat) = " << elmat-Trans(elmat) << endl;
      for (int k = 0; k < ned; k++)
        {
	  mat_sigmansigman = 0.0;
	  Vec<D> normal_ref;
	  for (int i=0; i<D; i++)
	    normal_ref(i) = normals[k][i];
        
          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*fel_hdiv.Order() );

          void * heapp = lh.GetPointer();

	  Mat<D> ninj;
	  ninj = 0.0;
          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
	      IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

            // Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
	      // IntegrationPoint ip( p2d(0), p2d(1) );
              // SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();
	      double invdet = 1.0 / det;

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

              
              Vec<D> nt = Trans(jac) * normal;
              nt /= det;

	      // calc the tensor product n n^T for fast evaluation of (sigma . n) (tau . n)
	      if ( l == 0 )
		for (int i = 0; i < D; i++ )
		  for ( int j = 0; j < D; j++ )
		    ninj(i,j) = nt(i) * nt(j);
              //Piola

	      // Vec<3> e1;
	      // e1 = 1.0;
	      bmat_sigma = Trans( fel_hdiv.GetShape(sip.IP(), lh) );
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
              mat_hdiv_n = Trans ( bmat_sigma ) * nt;
	      mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh);


              double factor = len * ir1d[l].Weight();

	      // penalty ... high on master, 0 on slave
	      double penalty = 1e7;
	      
	      double nx = 0.0;
	      Vec<D> point;
	      eltrans.CalcPoint ( ip, point, lh);

	      for ( int i = 0; i < D; i++ )
		nx += point(i) * normal(i);

	      nx /= L2Norm(point);
	      nx /= L2Norm(normal);

	      if ( nx < 0 ) penalty = 0;
// 	      // isslave = 0 --> im smoothing block
// 	      if ( nx >= 0.0 ) (*isslave)[elnum][k] = 1;
// 	      else (*isslave)[elnum][k] = 0;



 	      double fac = 1.0;
 	      if ( nx < 0.0 ) fac = -1.0;

	      dmat = 0.0; 
	      bmat = 0.0;
	      if ( nx > 0 ) // T_out
		{
		 
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma;

		  /* JOACHIM */
// 		  // V1
//  		  dmat(0,0) = -1.0;
//  		  dmat(1,0) = dmat(0,1) = -1.0;
// 		  // V2
//  		  dmat(0,0) = 1.0;
//  		  dmat(1,0) = dmat(0,1) = -1.0;

		  dmat(2,2) = 1.0;
		  dmat(0,2) = dmat(2,0) = -1.0;
		  dmat(1,2) = dmat(2,1) = -1.0;
		  dmat(1,1) = 1.0;

		  // stabilization joachim
		  Complex pen = 1;
		  dmat(1,1) += pen;
		  dmat(0,0) += pen;
		  dmat(2,2) += pen;
		  dmat(0,1) += pen;
		  dmat(1,0) += pen;
		  dmat(2,0) -= pen;
		  dmat(0,2) -= pen;
		  dmat(1,2) -= pen;
		  dmat(2,1) -= pen;

// 		  /* ASTRID */
// 		  dmat(1,0) = -1.0; 
// 		  dmat(0,1) = -1.0;
// 		  dmat(2,1) = 1.0;
// 		  dmat(1,2) = 1.0;
// 		  dmat(2,2) = -1.0;

// 		  // stabilization, (u-lambda)(v-mu) daempft
//  		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
//  		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// // 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// // 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;

		  if ( (*isslave)[elnum][k] ) 
		    {
		      dmat (1,0) = 0;
		      dmat(0,1) = 0;
		      dmat(2,1) = 0;
		      dmat(1,2) = 0;
		      // dmat(1,1) = 0;
		    }
		}

	      else // T_in
		{
		  bmat.Row(0).Range(base_hdiv, base_hdiv + nd_hdiv) = mat_hdiv_n;
		  bmat.Row(1).Range(base_hdivn, base_hdivn + nd_hdivn) = mat_hdivn_n;
		  bmat.Row(2).Range(base_hdiv, base_hdiv + nd_hdiv) = 
		    -Complex(0, 1.0/kappa) * div_sigma;

		  /* JOACHIM */
// 		  // V1
// 		  dmat(2,2) = -1.0;
// 		  dmat(0,2) = dmat(2,0) = -1.0;
// 		  dmat(1,2) = dmat(2,1) = 1.0;
// 		  dmat(1,1) = -1.0;
// 		  // V2
// 		  dmat(2,2) = 1.0;
// 		  dmat(0,2) = dmat(2,0) = -1.0;
// 		  dmat(1,2) = dmat(2,1) = -1.0;
// 		  dmat(1,1) = 1.0;

 		  dmat(0,0) = 1.0;
 		  dmat(1,0) = dmat(0,1) = -1.0;
// 		  stabilization joachim
		  Complex pen = 1;
		  dmat(1,1) += pen;
		  dmat(0,0) += pen;
		  dmat(2,2) += pen;
		  dmat(0,1) -= pen;
		  dmat(1,0) -= pen;
		  dmat(2,0) += pen;
		  dmat(0,2) += pen;
		  dmat(1,2) -= pen;
		  dmat(2,1) -= pen;


// 		  /* ASTRID */
// 		  dmat(0,1) = -1.0;
// 		  dmat(1,1) = -1.0;
// 		  dmat(1,0) = -1.0;

// 		  // stabilization, (u-lambda)(v-mu) daempft
//  		  dmat(1,1) -= Complex(0,1.0); dmat(2,2) -= Complex(0,1.0);
//  		  dmat(1,2) += Complex(0,1.0); dmat(2,1) += Complex(0,1.0);
// // 		  dmat(1,1) += 1.0; dmat(2,2) += 1.0;
// // 		  dmat(1,2) -= 1.0; dmat(2,1) -= 1.0;

		}

              dmat *= factor;

//  	      *testout << "dmat = " << endl << dmat << endl;
//  	      *testout << "bmat = " << endl << bmat << endl;

              dbmat = dmat * bmat;

              elmat += Trans (bmat) * dbmat;

	      lh.CleanUp(heapp);
            } 
	  
	
        }

    }
  };
  










  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtzs", hybrid_sigma_cpp::HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtzs", 2, 1,
                                      hybrid_sigma_cpp::HybridHelmholtzIntegrator<2>::Create);

    GetIntegrators().AddBFIntegrator ("hybridhelmholtzs", 3, 1,
                                      hybrid_sigma_cpp::HybridHelmholtzIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("hybridhelmholtzp", 2, 1,
                                      hybrid_sigma_cpp::HybridHelmholtzIntegratorP<2>::Create);

    GetIntegrators().AddBFIntegrator ("hybridhelmholtzp", 3, 1,
                                      hybrid_sigma_cpp::HybridHelmholtzIntegratorP<3>::Create);

  }

  Init init;

} // namespace

