#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

// Helmholtz equation, unknowns sigma on tets, lambda, sigma_E on facets
// juhu
namespace hybrid_sigma2_cpp
{

  using namespace ngfem;
  using namespace ngsolve;

  Table<int> * isslave;

  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      ;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);
      
      Flags hdivflags(flags), l2flags(flags);
      Flags hdivnflags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      hdivflags.SetFlag ("order", order);
      hdivflags.SetFlag ("discontinuous");
      hdivflags.SetFlag ("orderinner", order);
      hdivflags.SetFlag ("curlorder", order);
      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderinner", 1.0);
      hdivnflags.SetFlag ("curlorder", 1.0);
      facetflags.SetFlag ("order", order);

      spaces[0] = new HDivHighOrderFESpaceA (ma, hdivflags);
      spaces[1] = new HDivHighOrderFESpace (ma, hdivnflags);
      spaces[2] = new FacetFESpace (ma, facetflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      isslave = new Table<int> (ma.GetNE(), 4);
      for ( int i = 0; i < ma.GetNE(); i++ )
	(*isslave)[i] = 1;

      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



    // #define VER1
    Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
    {
      int blocktype = int(precflags.GetNumFlag("blocktype",1)); 
      Table<int> * it = CreateSmoothingBlocks ( blocktype );
    }


    Table<int> * CreateSmoothingBlocks (int type=1) const
    {
      cout << "SmoothingType " << type << endl; 
      // #ifdef VER1
      // edge by edge
      
      int nd = GetNDof();
      
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
      
      int base_hdivn = GetStorageStart(1);
      int base_facet = GetStorageStart(2);
      int base_hdiv = 0;

      ARRAY<int> dnums;
      
      ARRAY<int> cnt; // ned/* +nel */);
      Table<int> *it;  

      BitArray issurffacet( nfacets );
  
      const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace &> (*spaces[2]);
      const HDivHighOrderFESpace & hdivnspace = dynamic_cast<const HDivHighOrderFESpace &> (*spaces[1]);

      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel), orderedfacets(nfacets);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      int ifacet = 0;

      ARRAY<int> facets;
      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 0 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	  orderedfacets[ifacet++] = facets[0];
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      {
			if (!markedfacets.Test(facets[j]) && !newmarkedfacets.Test(facets[j]))
			  orderedfacets[ifacet++] = facets[j];
			newmarkedfacets.Set(facets[j]);
		      }
		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa)) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;


      issurffacet.Clear();
      for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	if ( ma.GetSElIndex(sel) == 0 || ma.GetSElIndex(sel) == 2 )
	  {
	    ma.GetSElFacets(sel, facets);
	    issurffacet.Set(facets[0] );
	  }

      
      for ( int el = 0; el < nel; el++ )
	{
	  ma.GetElFacets(el, facets);
	  for ( int i = 0; i < facets.Size(); i++ )
	    {
	      if ( issurffacet.Test(facets[i]) ) 
		(*isslave)[el][i] = 0;
	    }
	}



      switch(type)
	{
	case 1:
	  cout << "facet by facet" << endl;
	  cnt.SetSize(nfacets); // ned/* +nel */);
// 	  for (int i = 0; i < ned; i++)
	  for (int i = 0; i < nfacets; i++)
	    if ( ma.GetDimension() == 2 )
	      cnt[i] = 2*order+2;
	    else
	      cnt[i] = (order+1)*(order+2);
	  /*
	    for (int i = ned; i < nel+ned; i++)
	    cnt[i] = ndel;
	  */
	  
	  it = new Table<int>(cnt);
	  // Table<int> * it = new Table<int>(ned,2*order+2);
// 	  for (int i = 0; i < ned; i++)
	  for (int i = 0; i < nfacets; i++)
	    {
	      if ( ma.GetDimension() == 2 )
		{
		  (*it)[i][0] = base_hdivn+i;
		  for (int j = 0; j < order; j++)
		    (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
		  (*it)[i][order+1] = base_facet + i;
		  for (int j = 0; j < order; j++)
		    (*it)[i][order+2+j] = base_facet + ned + i * order + j;
		}
	      else
		{
		  (*it)[i][0] = base_hdivn+i;
		  // ndof_hdiv -- ho dofs on one facet
		  int ndof_hdiv = int(0.5 * (order+1)*(order+2)) - 1;
		  for (int j = 0; j < ndof_hdiv; j++)
		    (*it)[i][1+j] = base_hdivn + nfacets + (i * ndof_hdiv ) + j;
		  (*it)[i][ndof_hdiv+1] = base_facet + i;
		  for (int j = 0; j < ndof_hdiv; j++)
		    (*it)[i][ndof_hdiv+2+j] = base_facet + nfacets + i * ndof_hdiv + j;
		}
	    }
	  /*
	    for (int i = 0; i < nel; i++)
	    {
	    (*it)[ned+i] = base_el+i;
	    for (int j = 0; j < ndel-1; j++)
	    (*it)[ned+i][j+1] = base_el+nel+i*(ndel-1)+j;
	    }
	  */
	  // *testout << "table = " << endl << *it << endl;
	  break;
	case 2:

	  // #ifdef VER2
	  // all external dofs of an element
	  cout << "element by element" << endl;
	  cnt.SetSize(nel);
	  
	  for (int i = 0; i < nel; i++)
	    {
	      int el = orderedels[i];
	      GetExternalDofNrs (el, dnums);
	      cnt[i] = dnums.Size();
	    }
	  
	  it = new Table<int> (cnt);

	  for (int i = 0; i < nel; i++)
	    {
	      int el = orderedels[i];
	      GetExternalDofNrs (el, dnums);
	      for (int j = 0; j < dnums.Size(); j++)
		(*it)[i][j] = dnums[j];
	    }
	  break;

	case 3: // only master
	case 31: // all hdivn, master facet
	case 32: // all facet, master hdivn
	  cout << "element by element, only outflow edges for sigma" << endl;

	  cnt.SetSize(nel);
	 
	  cnt = 0;

	  for ( int eli = 0; eli < nel; eli++ )
	    {
	      int el = orderedels[eli];
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  hdivnspace.GetFacetDofNrs( facets[i], dnums );
		  if ( !(*isslave)[el][i] || type == 31)
		    cnt[eli] += dnums.Size();
		  facetspace.GetFacetDofNrs( facets[i], dnums );
		  if ( !(*isslave)[el][i] || type == 32 )
		    cnt[eli] += dnums.Size();
		  
		}
	    }

	  *testout << "Table cnt " << endl <<  cnt << endl;
	  it = new Table<int> (cnt);
	  cnt = 0;

	  for ( int eli = 0; eli < nel; eli++)
	    {
	      int el = orderedels[eli];
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  if ( !(*isslave)[el][i] || type == 32 )
		    {
		      facetspace.GetFacetDofNrs(facets[i], dnums);
		      for ( int j = 0; j < dnums.Size(); j++)
			(*it)[eli][cnt[eli]++] = base_facet + dnums[j];
		    }
		  if ( !(*isslave)[el][i] || type == 31 )
		    {
		      hdivnspace.GetFacetDofNrs(facets[i], dnums);
		      for ( int j = 0; j < dnums.Size(); j++)
			(*it)[eli][cnt[eli]++] = base_hdivn + dnums[j];
		      
		    }
		  
		}
	    }
	  break;



	}
      *testout << "smoothing blocks " << endl << *it << endl;

      return it;
    }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      int elnum = eltrans.GetElementNr();

      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const HDivFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);

      ELEMENT_TYPE eltype = cfel.ElementType();
      
        
      int nd = cfel.GetNDof();          
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();

      int base_hdiv = 0;
      int base_hdivn = nd_hdiv;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);
      // FlatMatrixFixWidth<2> mat_hdivn(nd_hdivn, lh); 
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh), scalbmat_sigma(nd_hdiv, lh)  ; 
      FlatVector<> div_sigma(nd_hdiv, lh), scaldiv_sigma(nd_hdiv, lh) ; // (nd_hdiv, lh);

      FlatMatrixFixHeight<4> bmat(nd, lh);
      FlatMatrixFixHeight<4> dbmat(nd, lh);
      Mat<4> dmat;

      // int ik sigma tau dx
      FlatMatrix<> mat_iksigmasigma (nd_hdiv, lh);
      // int 1/ik div sigma div tau dx
      FlatMatrix<> mat_iikdivdiv (nd_hdiv, lh);

      FlatMatrixFixHeight<D> diag_iksigmasigma (nd_hdiv, lh);

      double kappa = coef_kappa->EvaluateConst();
          

      const IntegrationRule & irDd =
        SelectIntegrationRule (eltype, 2*fel_hdiv.Order() );
      
      
      void * heapp = lh.GetPointer();
      mat_iksigmasigma = 0.0;
      mat_iikdivdiv = 0.0;
      diag_iksigmasigma = 0.0;

      bool curvedelement =  eltrans.IsCurvedElement();
      if ( curvedelement )      
	{
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
      
          
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));
	      mat_iksigmasigma += Trans(scalbmat_sigma) * (bmat_sigma);
	      lh.CleanUp(heapp);
	    }

	  //	  *testout << "mat_iksigmasigma " << endl << mat_iksigmasigma << endl;
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      imag(elmat(base_hdiv+i, base_hdiv+j)) +=  mat_iksigmasigma(i,j);


	  // int 1/ik divsigma divtau == - int i/k divsigma divtau 
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	}

      else // FAST

	{
	  
	  for (int l=0; l<irDd.GetNIP(); l++)
	    {
	      const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
	      double factor = sip.GetJacobiDet() * irDd[l].Weight();
	      double kappafactor = kappa*factor;
	      double ikappafactor = factor / kappa;
	      double invdet = 1.0 / sip.GetJacobiDet();
	      bmat_sigma = invdet
		* sip.GetJacobian() * Trans(fel_hdiv.GetShape(sip.IP(), lh));   //  * Trans(sip.GetJacobian()));
          
	      div_sigma = invdet *
		fel_hdiv.GetDivShape(sip.IP(), lh);
	      
	      
	      scalbmat_sigma = kappafactor * bmat_sigma;
	      scaldiv_sigma = ikappafactor * div_sigma;

	       // mass for l2^3 is diagonal
	      for ( int ii = 0; ii < nd_hdiv / D ; ii ++ )
		{
		  for ( int mi = 0; mi < D; mi++ )
		    for ( int mj = 0; mj < D; mj++ )
		      for ( int j = 0; j < D; j++)
			diag_iksigmasigma(mi, D*ii + mj) += scalbmat_sigma(j, D*ii+mi) * bmat_sigma(j, D*ii+mj);
		}
	       
	      mat_iikdivdiv += (scaldiv_sigma * Trans (div_sigma));

	      lh.CleanUp(heapp);
	    }
	  
	  
	  // int i k sigma . tau
	  for (int i = 0; i < nd_hdiv / D; i++)
	    for ( int mi = 0; mi < D; mi++ )
	      for ( int mj = 0; mj < D; mj++ )
		imag(elmat(base_hdiv+i*D+mi, base_hdiv+i*D+mj)) +=  diag_iksigmasigma(mi,D*i+mj);
	  
	 	  
	  // - int i / k sigma . tau  --> diagonal
	  for (int i = 0; i < nd_hdiv; i++)
	    for (int j = 0; j < nd_hdiv; j++)
	      {
		imag(elmat(base_hdiv+i, base_hdiv+j)) -= mat_iikdivdiv(i,j);
	      }
	   
	  lh.CleanUp(heapp);
	  
  	}




      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;


      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      vu = 0.0;
      vv = 0.0;
      FlatMatrix<> mat_sigmansigman (int(nd_hdiv/D), lh);
      mat_sigmansigman = 0.0;
      mat_iksigmasigma = 0.0;

      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

      for (int k = 0; k < ned; k++)
        {
	  mat_sigmansigman = 0.0;

	  Vec<D> normal_ref;
	  for (int i=0; i<D; i++)
	    normal_ref(i) = normals[k][i];
        
          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*fel_hdiv.Order());

          void * heapp = lh.GetPointer();

	  Mat<D> ninj;
	  ninj = 0.0;

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
	      IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

              
	      Vec<D> nt = normal_ref; // 
	      nt /= len;

	      double nx = 0.0;
	      Vec<D> point;
	      eltrans.CalcPoint ( ip, point, lh);

	      for ( int i = 0; i < D; i++ )
		nx += point(i) * normal(i);

	      // isslave = 0 --> im smoothing block
	      if ( nx < 0.0 ) (*isslave)[elnum][k] = 0;

              //Piola
	      // calc the tensor product n n^T for fast evaluation of (sigma . n) (tau . n)
	      if ( l == 0 )
		for (int i = 0; i < D; i++ )
		  for ( int j = 0; j < D; j++ )
		    ninj(i,j) = nt(i) * nt(j);

	      bmat_sigma = Trans( fel_hdiv.GetShape(sip.IP(), lh) );
              mat_hdiv_n = Trans ( bmat_sigma ) * nt;
              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;


              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              // fel_l2.CalcShape(sip.IP(), mat_l2);
          
              double factor = len * ir1d[l].Weight();

              bmat = 0.0;
              bmat.Row(0).Range (base_hdiv , base_hdiv +nd_hdiv) = mat_hdiv_n;
              // bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;

	      // ##############################################

	      /*
		slow and simple ...
	      */

	      /*
	      dmat = 0.0;
	      
	      if ( nx > 0 || !discont )
		{
		  dmat(0,0) = 1;
		  dmat(0,2) = -1;
		  dmat(2,0) = -1;
		  dmat(2,2) = 1;
		}
	      
              dmat(0,3) = 1;
              dmat(3,0) = 1;
              
	      dmat *= factor;

              dbmat = dmat * bmat;

              elmat += Trans (bmat) * dbmat;
	      */


	      // #################################################

	      /*
		fast...
	      */

	      dbmat = 0.0;
	      for ( int j = 0; j < nd_hdiv; j++ )
		{
		  dbmat(0,j) += factor * bmat(0,j);
		  dbmat(2,j) = -dbmat(0,j);
		  dbmat(3,j) = dbmat(0,j);
		}


	      for ( int j = base_hdivn; j < base_facet; j++)
		{
		  dbmat(0,j) -= factor * bmat(2,j);
		  dbmat(2,j) = -dbmat(0,j);
		}

	      for ( int j = base_facet; j < nd; j++ )
		{
		  dbmat(0,j) += factor * bmat(3,j);
		}

              // dbmat = dmat * bmat;

              // elmat += factor * Trans (bmat) * dbmat;

	      // int sigma_n tau_n


	      if ( curvedelement )
		{ 
		  for ( int i = 0; i < base_hdivn; i++ )
		    {
		      for ( int j = 0; j < nd; j++ )
			elmat(i,j) += bmat(0,i) * dbmat(0,j);
		    }
		}
	      else
		{
		  for ( int i = 0; i < base_hdivn; i++ )
		    {
		      for ( int j = nd_hdiv; j < nd; j++ )
			elmat(i,j) += bmat(0,i) * dbmat(0,j);
		    }
		  
		  // calculate scalar inner product of l2-shapes
		  for ( int i = 0; i < base_hdivn / D ; i ++)
		    for ( int j = 0; j < base_hdivn / D ; j ++ )
		      {
			mat_sigmansigman(i, j) += factor * bmat_sigma(0,D*i) * bmat_sigma(0,D*j);
		      }
		}


	      for ( int i = base_hdivn; i < base_facet; i++ )
		{
		  for ( int j = 0; j < nd; j++ )
		    elmat(i,j) += bmat(2,i) * dbmat(2,j);
		}
	      for ( int i = base_facet; i < nd; i++ )
		{
		  for ( int j = 0; j < base_hdivn; j++ )
		    elmat(i,j) += bmat(3,i) * dbmat(3,j);
		}
	      //  lh.CleanUp(heapp);
            } 
	  
	  // add scalar inner product times tensor-product of normal vectors
	  if ( ! curvedelement )
  	  for ( int i = 0; i < base_hdivn / D; i++ )
  	    for ( int j = 0; j < base_hdivn / D; j++ )
  	      for ( int mi = 0; mi < D; mi++ )
  		for ( int mj = 0; mj < D; mj ++ )
  		  elmat(i*D+mi, j*D+mj) += ninj(mi,mj) * mat_sigmansigman(i,j);


        }
    }
  };




  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz2", hybrid_sigma2_cpp::HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz2", 2, 1,
                                      hybrid_sigma2_cpp::HybridHelmholtzIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz2", 3, 1,
                                      hybrid_sigma2_cpp::HybridHelmholtzIntegrator<3>::Create);
  
  
  }

  Init init;

} // namespace

