#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

// helmholtz equation,
// mixed formulation with u, sigma
// hybridized with 2 lagrange parameters lambda, sigma^E

namespace hybrid_cpp
{

  using namespace ngfem;
  using namespace ngsolve;

  bool discont;
  Table<int> * isslave;

  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      ;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(4);
      
      Flags hdivflags(flags), l2flags(flags);
      Flags hdivnflags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      hdivflags.SetFlag ("order", order);
      hdivflags.SetFlag ("discontinuous");
      hdivflags.SetFlag ("orderinner", order);
      hdivflags.SetFlag ("curlorder", order);
      l2flags.SetFlag ("order", max(0,order) );
      // l2flags.SetFlag ("order", 1.0);
      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderinner", 1.0);
      hdivnflags.SetFlag ("curlorder", 1.0);
      discont = flags.GetDefineFlag("discontinuous");
      if (discont) 
	hdivnflags.SetFlag("discontinuous");
      facetflags.SetFlag("order", order);

      spaces[0] = new HDivHighOrderFESpaceA (ma, hdivflags);
      spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
      spaces[2] = new HDivHighOrderFESpace (ma, hdivnflags);
      spaces[3] = new FacetFESpace (ma, facetflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      isslave = new Table<int> (ma.GetNE(), 4);
      for ( int i = 0; i < ma.GetNE(); i++ )
	(*isslave)[i] = 1;

      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



    // #define VER1
    Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
    {
      int blocktype = int(precflags.GetNumFlag("blocktype",1)); 
      Table<int> * it = CreateSmoothingBlocks ( blocktype );
    }


    Table<int> * CreateSmoothingBlocks (int type=1) const
    {
      cout << "SmoothingType " << type << endl; 
      // #ifdef VER1
      // edge by edge
      
      int nd = GetNDof();
      
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
      
      int order_u = (*this)[1]->GetOrder();
      int ndel = ((order_u+1)*(order_u+2))/2;
      
      int base_hdivn = GetStorageStart(2);
      int base_facet = GetStorageStart(3);
      int base_el = GetStorageStart(1);
      
      ARRAY<int> dnums;
      
      ARRAY<int> cnt; // ned/* +nel */);
      Table<int> *it;  

      BitArray issurffacet( nfacets );
  
      const FacetFESpace & facetspace = dynamic_cast<const FacetFESpace &> (*spaces[3]);
      const HDivHighOrderFESpace & hdivnspace = dynamic_cast<const HDivHighOrderFESpace &> (*spaces[2]);

      cout << "Order Elements .. " << endl;
      ARRAY<int> orderedels(nel), orderedfacets(nfacets);
      BitArray markedels(nel), markedfacets(nfacets), newmarkedfacets(nfacets);

      markedels.Clear();
      markedfacets.Clear();
      newmarkedfacets.Clear();

      int ifacet = 0;

      ARRAY<int> facets;
      // mark boundary facets --> facets where boundary index == 0
      for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	{
	  if ( ma.GetSElIndex(sel) != 0 ) continue;
	  ma.GetSElFacets(sel, facets);
	  markedfacets.Set(facets[0]);
	  orderedfacets[ifacet++] = facets[0];
	}

      int nmarkedels = 0;
      while (nmarkedels < nel )
	{
	  for ( int el = 0; el < nel; el++ )
	    {
	      if ( markedels.Test(el) ) continue;

	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++)
		if ( markedfacets.Test(facets[i]) )
		  {
		    markedels.Set(el);
		    orderedels[nmarkedels++] = el;
		    for ( int j = 0; j < facets.Size(); j++ )
		      {
			if (!markedfacets.Test(facets[j]) && !newmarkedfacets.Test(facets[j]))
			  orderedfacets[ifacet++] = facets[j];
			newmarkedfacets.Set(facets[j]);
		      }
		    break;
		  }
	    }
	  for ( int fa = 0; fa < nfacets; fa++ )
	    if ( newmarkedfacets.Test(fa)) markedfacets.Set(fa);

	}
      cout << " .. Order Elements done" << endl;


      issurffacet.Clear();
      for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	if ( ma.GetSElIndex(sel) == 0 || ma.GetSElIndex(sel) == 2 )
	  {
	    ma.GetSElFacets(sel, facets);
	    issurffacet.Set(facets[0] );
	  }

      
      for ( int el = 0; el < nel; el++ )
	{
	  ma.GetElFacets(el, facets);
	  for ( int i = 0; i < facets.Size(); i++ )
	    {
	      if ( issurffacet.Test(facets[i]) ) 
		(*isslave)[el][i] = 0;
	    }
	}



      switch(type)
	{
	case 1:
	  cout << "facet by facet" << endl;
	  cnt.SetSize(nfacets); // ned/* +nel */);
// 	  for (int i = 0; i < ned; i++)
	  for (int i = 0; i < nfacets; i++)
	    if ( ma.GetDimension() == 2 )
	      cnt[i] = 2*order+2;
	    else
	      cnt[i] = (order+1)*(order+2);
	  /*
	    for (int i = ned; i < nel+ned; i++)
	    cnt[i] = ndel;
	  */
	  
	  it = new Table<int>(cnt);
	  // Table<int> * it = new Table<int>(ned,2*order+2);
// 	  for (int i = 0; i < ned; i++)
	  for (int i = 0; i < nfacets; i++)
	    {
	      if ( ma.GetDimension() == 2 )
		{
		  (*it)[i][0] = base_hdivn+i;
		  for (int j = 0; j < order; j++)
		    (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
		  (*it)[i][order+1] = base_facet + i;
		  for (int j = 0; j < order; j++)
		    (*it)[i][order+2+j] = base_facet + ned + i * order + j;
		}
	      else
		{
		  (*it)[i][0] = base_hdivn+i;
		  // ndof_hdiv -- ho dofs on one facet
		  int ndof_hdiv = int(0.5 * (order+1)*(order+2)) - 1;
		  for (int j = 0; j < ndof_hdiv; j++)
		    (*it)[i][1+j] = base_hdivn + nfacets + (i * ndof_hdiv ) + j;
		  (*it)[i][ndof_hdiv+1] = base_facet + i;
		  for (int j = 0; j < ndof_hdiv; j++)
		    (*it)[i][ndof_hdiv+2+j] = base_facet + nfacets + i * ndof_hdiv + j;
		}
	    }
	  /*
	    for (int i = 0; i < nel; i++)
	    {
	    (*it)[ned+i] = base_el+i;
	    for (int j = 0; j < ndel-1; j++)
	    (*it)[ned+i][j+1] = base_el+nel+i*(ndel-1)+j;
	    }
	  */
	  // *testout << "table = " << endl << *it << endl;
	  break;
	case 2:

	  // #ifdef VER2
	  // all external dofs of an element
	  cout << "element by element" << endl;
	  cnt.SetSize(nel);
	  
	  for (int i = 0; i < nel; i++)
	    {
	      int el = orderedels[i];
	      GetExternalDofNrs (el, dnums);
	      cnt[i] = dnums.Size();
	    }
	  
	  it = new Table<int> (cnt);

	  for (int i = 0; i < nel; i++)
	    {
	      int el = orderedels[i];
	      GetExternalDofNrs (el, dnums);
	      for (int j = 0; j < dnums.Size(); j++)
		(*it)[i][j] = dnums[j];
	    }
	  break;

	case 3: // only master
	case 31: // all hdivn, master facet
	case 32: // all facet, master hdivn
	  cout << "element by element, only outflow edges for sigma" << endl;

	  cnt.SetSize(nel);
	 
	  cnt = 0;

	  for ( int eli = 0; eli < nel; eli++ )
	    {
	      int el = orderedels[eli];
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  hdivnspace.GetFacetDofNrs( facets[i], dnums );
		  if ( !(*isslave)[el][i] || type == 31)
		    cnt[eli] += dnums.Size();
		  facetspace.GetFacetDofNrs( facets[i], dnums );
		  if ( !(*isslave)[el][i] || type == 32 )
		    cnt[eli] += dnums.Size();
		  
		}
	    }

	  *testout << "Table cnt " << endl <<  cnt << endl;
	  it = new Table<int> (cnt);
	  cnt = 0;

	  for ( int eli = 0; eli < nel; eli++)
	    {
	      int el = orderedels[eli];
	      ma.GetElFacets(el, facets);
	      for ( int i = 0; i < facets.Size(); i++ )
		{
		  if ( !(*isslave)[el][i] || type == 32 )
		    {
		      facetspace.GetFacetDofNrs(facets[i], dnums);
		      for ( int j = 0; j < dnums.Size(); j++)
			(*it)[eli][cnt[eli]++] = base_facet + dnums[j];
		    }
		  if ( !(*isslave)[el][i] || type == 31 )
		    {
		      hdivnspace.GetFacetDofNrs(facets[i], dnums);
		      for ( int j = 0; j < dnums.Size(); j++)
			(*it)[eli][cnt[eli]++] = base_hdivn + dnums[j];
		      
		    }
		  
		}
	    }
	  break;



	}
      *testout << "smoothing blocks " << endl << *it << endl;

      return it;
    }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      int elnum = eltrans.GetElementNr();

      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const NodalFiniteElement & fel_l2 = 
        dynamic_cast<const NodalFiniteElement&> (cfel[1]);
      const HDivFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[2]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[3]);

      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();

      int base_hdiv = 0;
      int base_l2 = nd_hdiv;
      int base_hdivn = base_l2+nd_l2;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

        
      FlatVector<> mat_l2(nd_l2, lh), scalmat_l2(nd_l2, lh);
      FlatMatrixFixWidth<D> mat_hdiv(nd_hdiv, lh); 
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);
      // FlatMatrixFixWidth<2> mat_hdivn(nd_hdivn, lh); 
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrixFixHeight<D> bmat_sigma(nd_hdiv, lh);
      FlatVector<> div_sigma(nd_hdiv, lh);
      FlatMatrixFixHeight<D> scalbmat_sigma(nd_hdiv, lh);
      FlatVector<> scaldiv_sigma(nd_hdiv, lh);
      FlatMatrix<> hmat_l2(nd_l2, lh);
      
      FlatMatrixFixHeight<4> bmat(nd, lh);
      FlatMatrixFixHeight<4> dbmat(nd, lh);
      Mat<4> dmat;

      FlatMatrix<> mat_sigmasigma (nd_hdiv, lh);
      FlatMatrix<> mat_divdiv (nd_hdiv, lh);
      FlatMatrix<> mat_divsigma (nd_hdiv, nd_l2, lh);
      FlatMatrix<> mat_l2l2 (nd_l2, lh);

      double kappa = coef_kappa->EvaluateConst();
          
      const IntegrationRule & irDd =
        SelectIntegrationRule (eltype, 2*max2(fel_l2.Order(), fel_hdiv.Order()));
      
      
      void * heapp = lh.GetPointer();
      hmat_l2 = 0.0;
      mat_sigmasigma = 0.0;
      mat_divsigma = 0.0;
      mat_divdiv = 0.0;
      mat_l2l2  = 0.0;

      for (int l=0; l<irDd.GetNIP(); l++)
        {
	  const SpecificIntegrationPoint<D,D> sip(irDd[l], eltrans, lh);
          
          double factor = sip.GetJacobiDet() * irDd[l].Weight();
          
          bmat_sigma = (1.0/sip.GetJacobiDet()) 
            * Trans (fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian()));
          
          div_sigma = 1.0/sip.GetJacobiDet() *
            fel_hdiv.GetDivShape(sip.IP(), lh);
      
          
          fel_l2.CalcShape(sip.IP(), mat_l2);
          
	  scalmat_l2 = -kappa * factor * mat_l2;
	  scalbmat_sigma = -kappa * factor * bmat_sigma;
	  scaldiv_sigma = factor * div_sigma;

          mat_sigmasigma += Trans (scalbmat_sigma) * bmat_sigma ;
          mat_divsigma += (scaldiv_sigma * Trans (mat_l2));
          mat_divdiv += (scaldiv_sigma * Trans (div_sigma));
          mat_l2l2 += (scalmat_l2 * Trans (mat_l2));

          lh.CleanUp(heapp);
        }


      for (int i = 0; i < nd_hdiv; i++)
        for (int j = 0; j < nd_hdiv; j++)
          imag(elmat(base_hdiv+i, base_hdiv+j)) +=  mat_sigmasigma(i,j);

      /*
      for (int i = 0; i < nd_hdiv; i++)
        for (int j = 0; j < nd_hdiv; j++)
          elmat(base_hdiv+i, base_hdiv+j) += Complex(0, 1.0/kappa) * mat_divdiv(i,j);
      */
      for (int i = 0; i < nd_hdiv; i++)
        for (int j = 0; j < nd_l2; j++)
	  {
	    elmat(base_hdiv+i, base_l2+j) += mat_divsigma(i,j);
	    elmat(base_l2+j, base_hdiv+i) += mat_divsigma(i,j);
	  }
      /*
      for (int i = 0; i < nd_hdiv; i++)
        for (int j = 0; j < nd_l2; j++)
          elmat(base_l2+j, base_hdiv+i) -= mat_divsigma(i,j);
      */  
      for (int i = 0; i < nd_l2; i++)
        for (int j = 0; j < nd_l2; j++)
          imag(elmat(base_l2+j, base_l2+i)) -= mat_l2l2(i,j);


      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;


      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      vu = 0.0;
      vv = 0.0;

      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

      for (int k = 0; k < ned; k++)
        {
	  Vec<D> normal_ref;
	  for (int i=0; i<D; i++)
	    normal_ref(i) = normals[k][i];
        
	  /*
          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
        
          tau_ref  = pe-ps;
	  */

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*max (fel_hdiv.Order(),fel_l2.Order()));

          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
	      IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

              // Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
	      // IntegrationPoint ip( p2d(0), p2d(1) );
              // SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

	      /*
              tau = jac * tau_ref;
              double len = L2Norm(tau);
	    
              tau *= (1.0/len);
              normal(0) = tau(1);
              normal(1) = -tau(0);
	      
	      normal = det * Trans (inv_jac) * normal_ref;
	      len = L2Norm(normal);
	      normal /= len;
	      */

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;
              double len = L2Norm(normal);
              normal /= len;

              
	      Vec<D> nt = normal_ref; // 
	      nt /= len;

	      double nx = 0.0;
	      Vec<D> point;
	      eltrans.CalcPoint ( ip, point, lh);

	      for ( int i = 0; i < D; i++ )
		nx += point(i) * normal(i);

	      // isslave = 0 --> im smoothing block
	      if ( nx < 0.0 ) (*isslave)[elnum][k] = 0;

              //Piola

	      bmat_sigma = Trans( fel_hdiv.GetShape(sip.IP(), lh) );
              mat_hdiv_n = Trans ( bmat_sigma ) * nt;
              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;


              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              fel_l2.CalcShape(sip.IP(), mat_l2);
          
              double factor = len * ir1d[l].Weight();

              bmat = 0.0;
              bmat.Row(0).Range (base_hdiv , base_hdiv +nd_hdiv) = mat_hdiv_n;
              bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;


              /*
                1   0   1  -1
                0   1   0  -1
                1   0   1   0
                1  -1   0   1
              */


              dmat = 0.0;
	      
	      dmat(0,0) = -1;
	      dmat(0,2) = 1;
	      dmat(2,0) = 1;
	      dmat(2,2) = -1;



                dmat(1,1) = 1;
                dmat(1,3) = 1;
                dmat(3,1) = 1;
                dmat(3,3) = 1;
	      
              dmat(0,3) = 1;
              dmat(3,0) = 1;
              
              dmat *= factor;

// 	      dbmat = 0.0;
// 	      for ( int j = 0; j < nd_hdivn; j++ )
// 		{
// 		  dbmat(0,j) += bmat(0,j);
// 		  dbmat(2,j) += bmat(0,j);
// 		  dbmat(3,j) += bmat(0,j);
// 		}

// // 	      for ( int j = base_l2; j < base_hdivn; j++ )
// // 		{
// // 		  ;
// // 		}

// 	      for ( int j = base_hdivn; j < base_facet; j++)
// 		{
// 		  dbmat(0,j) += bmat(2,j);
// 		  dbmat(2,j) += bmat(2,j);
// 		}

// 	      for ( int j = base_facet; j < nd; j++ )
// 		{
// 		  dbmat(0,j) += bmat(3,j);
// 		}

              dbmat = dmat * bmat;

              elmat += Trans (bmat) * dbmat;

// 	      for ( int i = 0; i < base_l2; i++ )
// 		{
// 		  for ( int j = 0; j < nd; j++ )
// 		    elmat(i,j) += factor * bmat(0,i) * dbmat(0,j);
// 		}


// 	      for ( int i = base_hdivn; i < base_facet; i++ )
// 		{
// 		  for ( int j = 0; j < nd; j++ )
// 		    elmat(i,j) += factor * bmat(2,i) * dbmat(2,j);
// 		}
// 	      for ( int i = base_facet; i < nd; i++ )
// 		{
// 		  for ( int j = 0; j < base_l2; j++ )
// 		    elmat(i,j) += factor * bmat(3,i) * dbmat(3,j);
// 		}
              lh.CleanUp(heapp);
            }
        }
    }
  };



  template <int D>
  class HybridRobinIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<> & elmat,
                                        LocalHeap & lh) const
    {
      cerr << "Helmholtz - Robin real called" << endl;
    }




    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const HDivNormalFiniteElement<D-1> & hdivnfel = 
        dynamic_cast<const HDivNormalFiniteElement<D-1>&> (cfel[2]);
      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[3]);
    
      int nd = cfel.GetNDof();  
      
      ELEMENT_TYPE eltype = cfel.ElementType();

      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int nd_hdivn = hdivnfel.GetNDof();
      int nd_facet = ffel.GetNDof();
      int base_facet = nd-nd_facet;
      int base_hdivn = base_facet-nd_hdivn;
      
      FlatVector<> shape(nd_facet, lh), shapehdivn(nd_hdivn, lh);
      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (eltype, 2*ffel.Order());
      
      FlatMatrixFixHeight<2> bmat(nd, lh);
      FlatMatrixFixHeight<2> dbmat(nd, lh);
      Mat<2> dmat(2);

      vu = 0.0;
      vv = 0.0;

 
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
	  SpecificIntegrationPoint<D-1,D> sip(ir1d[l], eltrans, lh);
          double val = coef->Evaluate(sip);

          if (val == 0) continue;

          hdivnfel.CalcShape(ir1d[l], shapehdivn);
          shapehdivn *= 1.0 / sip.GetJacobiDet();

          ffel.CalcShape(ir1d[l], shape);
          /*
            vu.Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
            vu.Range (base_facet, nd) = shape;
            elmat +=  (sip.GetJacobiDet()*ir1d[l].Weight())*(vu*Trans(vu));
          */

          double factor = sip.GetJacobiDet()*ir1d[l].Weight();

          bmat = 0.0;
          bmat.Row(0).Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
          bmat.Row(1).Range (base_facet, nd) = shape;

          dmat = 0.0;
          /*
            dmat(0,0) = 1;
            dmat(0,1) = 1;
            dmat(1,0) = 1;
            dmat(1,1) = 1;
          */

          // dmat(0,0) =  1;
          // dmat(0,1) =  -1;
          // dmat(1,0) = -1;
          dmat(1,1) =  -1;


          dbmat = dmat * bmat;
          elmat += factor * Trans (bmat) * dbmat;
        }
    }
  };



  class HybridNeumannIntegrator2d : public LinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridNeumannIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : LinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridNeumannIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridNeumannIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementVector (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatVector<double> & elvec,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      //       cout << "** neumann: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      int nd = cfel.GetNDof();  
      
      elvec.AssignMemory(nd, lh);
      elvec = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      

      FlatVector<> vec(ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      vec=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          vec +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*shape;
          //         vec +=  ir1d[l].Weight()*coef->Evaluate(sip)*shape;
        }
      
      // copy to global matrix
      elvec.Range(ndr,nd) = vec;
    }


  };

  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz", HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 2, 1,
                                      HybridHelmholtzIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator<2>::Create);

    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 3, 1,
                                      HybridHelmholtzIntegrator<3>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 3, 1,
                                      HybridRobinIntegrator<3>::Create);

    GetIntegrators().AddLFIntegrator ("hybridneumann", 2, 1,
                                      HybridNeumannIntegrator2d::Create);
  
  }

  Init init;

} // namespace

