

#ifndef HYBRIDCONTACT
#define HYBRIDCONTACT


class NumProcContHybrid : public NumProc
{
protected:
  // bilinear-form for the Hellinger Reissner system of elasticity
  BilinearForm * bfa;
  // linear-form providing the right hand side, contains (0, f)
  LinearForm * lff;
  // solution vector, containes (sigma, u)
  GridFunction * gfu;

  //contact surfaces
  int cbound1, cbound2;
  int mbound;

  Preconditioner * pre;

  double taun, tautau, taup;
  int numuzawasteps;
  int contactelements;
  double coeff_f;
  int ncela, ncelb;
  double tmax;



public:
  
    
  NumProcContHybrid (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {

    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", "a"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    cbound1 = int(flags.GetNumFlag ( "cbound1" , 1.0));
    cbound2 = int(flags.GetNumFlag ( "cbound2" , 2.0));
    mbound = int (flags.GetNumFlag ( "masterbound", cbound1 ) );

    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);

    taun = flags.GetNumFlag ( "tau", 1e-5);
    tautau = taun;
    taup = flags.GetNumFlag ("taup", 1e-5);

    taun = flags.GetNumFlag ( "taun", taun);
    tautau = flags.GetNumFlag ( "tautau", taun);

    numuzawasteps = int(flags.GetNumFlag( "usteps", 10));
    contactelements = int(flags.GetNumFlag( "cels", 10));

    coeff_f = flags.GetNumFlag( "friction", 1.0);

    ncela = int( flags.GetNumFlag( "ncela", 10));
    ncelb = int( flags.GetNumFlag( "ncelb", 10));
    tmax = flags.GetNumFlag( "gap", 0.1);


  }


  virtual ~NumProcContHybrid() 
  { ; }


  // creates an solver object
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcContHybrid (pde, flags);
  }



  void Assembleg (  const ContactInfo & cinfo, 
		    Vector<double > & vecg);


  void ApplyCTranspose( const ContactInfo & cinfo,
			const HDivSurfaceHighOrderFESpace & lambdatfespace,
			const Vector<double > & veclambdan, const Vector<double> & veclambda, BaseVector & rhs,
			LocalHeap & lh);


  void  ApplyC(  const ContactInfo & cinfo, 
		const HDivSurfaceHighOrderFESpace & lambdatspace,
		const BaseVector & rhs, Vector<double> & veclambdan, Vector<double> & veclambda, LocalHeap & lh);


  bool ProjectStress ( const ContactInfo & cinfo,  const HDivSurfaceHighOrderFESpace & lambdtspace,
		       const double coeff_f, const Vector<double> & vecsigmann, Vector<double > & lambdan, 
		       Vector<double> & lambda, LocalHeap & lh);

  void CalcURel ( const ContactInfo & cinfo, 
		  const BaseVector & vecu, Vector<double > & vecurel, LocalHeap & lh);

  void FindPhiBeta(const Vector<double> & veclambdan, 
		   const Vector<double> & r, Vector<double> & phi, Vector<double> & beta);

  void FindPhiBetaTau(const ContactInfo & cinfo, const HDivSurfaceHighOrderFESpace & lambdatspace,
		      const double & coeff_f,
		      const Vector<double> & veclambdan, const Vector<double> & veclambda,
		      const Vector<double> & rtau, Vector<double> & phitau, Vector<double> & betatau, LocalHeap & lh);

  double TangL2Product( const MeshAccess & ma, const ContactInfo & cinfo,
			const Vector<Vec<3> > & veca, const Vector<Vec<3> > & vecb);
    
  double NormalL2Product( const MeshAccess & ma, const ContactInfo & cinfo,
			  const Vector<double> & veca, const Vector<double> & vecb);

  /*
  void ApplyCTranspose( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
			const Vector<double > & w0, const Vector<double> & w1, 
			const Vector<double> & lambdax, const Vector<double> & lambday, const Vector<double> & lambdaz, BaseVector & rhs,
			LocalHeap & lh);




  void ApplyCn ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, Vector<double> & updatew0, Vector<double> & updatew1, LocalHeap & lh );



  void ApplyCtau ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		   const BaseVector & vecu, Vector<double> & updatelambdax, Vector<double> & updatelambday, Vector<double> & updatelambdaz, LocalHeap & lh );


  void ProjectDisp ( const Vector<double > & vecg, Vector<double> & w);


  void ProjectStress ( const double coeff_f, const Vector<double> & sigmann, 
		       Vector<double > & lambdax, Vector<double > & lambday, Vector<double > & lambdaz );


  void CalcSigmann (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh);
  */
  virtual void Do(LocalHeap & lh);

  /*
  void ProjectDisp (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, const Vector<double> & vecg, Vector<double > & w0, Vector<double> & w1, LocalHeap & lh);

  */
  virtual string GetClassName () const
  {
    return " Solver for Contact Problems ";
  }

  virtual void PrintReport (ostream & ost)
  {
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl;
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*    ost << 
      "\n\nNumproc Parabolic:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};









#endif
