/*

Mixed elasticity -  hybrid finite element for Lagrange parameters on edges/faces

*/



#include <solve.hpp>
#include <comp.hpp>
#include <fem.hpp>

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif


namespace ngfem
{

using namespace ngfem;
using namespace ngsolve;
using namespace ngcomp;

#ifdef PARALLEL
using namespace ngparallel;
#endif

HybridFiniteElement :: HybridFiniteElement (ELEMENT_TYPE etype)
  : FiniteElement (ElementTopology::GetSpaceDim (etype), etype, -1, -1)
{
  nfacets = ElementTopology :: GetNFacets ( etype );
  for (int i = 0; i < 8; i++)
    vnums[i] = i;
}




HybridFiniteElement ::  ~HybridFiniteElement() 
{ ; }



HybridNormalFiniteElement :: HybridNormalFiniteElement (ELEMENT_TYPE etype)
  : FiniteElement (ElementTopology::GetSpaceDim (etype), etype, -1, -1)
{
    for (int i = 0; i < 4; i++)
      vnums[i] = i;
}





HybridNormalFiniteElement ::  ~HybridNormalFiniteElement() 
{ ; }






  HybridNormalSegm :: HybridNormalSegm(int aorder) : HybridNormalFiniteElement (ET_SEGM)
  {
    order = aorder;
    order_inner = INT<2> (aorder, aorder);
    ndof = aorder+1; 
  }

void HybridNormalSegm :: ComputeNDof ()
{
  ndof = order_inner[0]+1;
}


void HybridNormalSegm :: CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const
  {
    if ( order_inner[0] < 0 ) return;

    double x = ip(0); 
    double lami[2] = {x,1-x};
    
    
    int es = 0, ee = 1;
    double fac=1.0;
    if (vnums[es] > vnums[ee]) { swap (es, ee); fac=-1.0;}
    

    double ls = lami[es];
    double le = lami[ee];
    
    ArrayMem<double, 20> polx(order_inner[0]+1);
    LegendrePolynomial (order_inner[0]+1, ls - le, polx);

    int ii = 0;
    for(int l=0; l<=order_inner[0]; l++)
      {
	shape(ii) = polx[l] *fac;
	ii++;
      }


  }

HybridNormalTrig :: HybridNormalTrig(int aorder) : HybridNormalFiniteElement (ET_TRIG)
  {
    order = aorder;
    order_inner = INT<2> (aorder, aorder);
    ndof = (aorder+1)*(aorder+2)/2; 
  }

void HybridNormalTrig :: ComputeNDof ()
{
  ndof = (order_inner[0] + 1)*(order_inner[0]+2) / 2;
}

void HybridNormalTrig :: CalcShape (const IntegrationPoint & ip,
			  FlatVector<> shape) const
{
  if ( order_inner[0] < 0 ) return;

    double x = ip(0);
    double y = ip(1);
    double lami[3] = {x, y, 1-x-y};
    int ii = 0;

    int n = order_inner[0];

    int fac = 1;
    ARRAY<int> fav(3);

    fav[0] = 0;
    fav[1] = 1;
    fav[2] = 2;
    
    //Sort vertices  first edge op minimal vertex
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
    if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); fac *= -1;}
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
    
    ArrayMem<double, 20> polx(n+1), poly(n+1);
    
    ScaledLegendrePolynomial (order_inner[0], 2*lami[fav[0]]+lami[fav[1]]-1, 1-lami[fav[1]], polx);
    
    for (int i = 0; i <= order_inner[0]; i++)
      {
	JacobiPolynomial (n, 2*lami[fav[1]]-1, 2*i+1, 0, poly);
        for (int j = 0; j <= order_inner[0]-i; j++)
          shape(ii++) = polx[i] * poly[j] * fac;
      }
    
 
      /*
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
    if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); fac *= -1;}
    if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
    
  
    ArrayMem<double, 20> polx(n+1), poly(n+1);

    ScaledLegendrePolynomial (order, lami[fav[0]]-lami[fav[1]], lami[fav[0]] + lami[fav[1]], polx);
    JacobiPolynomial (order, 2*lami[fav[2]]-1, 1, 0, poly);
      
      for (int i2 = 0; i2 <= order; i2++)
	for (int j = 0; j <= order-i2; j++)
	  {
	    shape(ii) = polx[i2] * poly[j] * fac;
	    ii++;
	  }

      */
      /*
    ScaledLegendrePolynomial (order, lami[fav[0]]-lami[fav[1]], lami[fav[0]] + lami[fav[1]], polx);
    //    LegendrePolynomial (order, 2*lami[fav[2]]-1, poly);
    JacobiPolynomial (order, 2*lami[fav[2]]-1, 1, 0, poly);

    shape(ii++) = lami[fav[0]] * fac;
    shape(ii++) = lami[fav[1]] * fac;
    shape(ii++) = lami[fav[2]] * fac;

      
    for (int i2 = 0; i2 <= order; i2++)
      for (int j = max2 (2 - i2, 0); j <= order-i2; j++)
	{
	  shape(ii) = polx[i2] * poly[j] * fac;
	  ii++;
	}
      */
}







HybridNormalQuad :: HybridNormalQuad(int aorder) : HybridNormalFiniteElement (ET_QUAD)
  {
    order = aorder;
    eltype = ET_QUAD;
    ndof = sqr(aorder+1); 
    order_inner = INT<2> (aorder, aorder);
  }


void HybridNormalQuad :: ComputeNDof ()
{
  ndof = (order_inner[0] + 1)*(order_inner[1]+1);
}

void HybridNormalQuad :: CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const
  {
  if ( order_inner[0] < 0 ) return;

    double x = ip(0);
    double y = ip(1);
    int ii = 0;

    //int n = order_inner[0];

    ArrayMem<double, 20> polx(order_inner[0]+1), poly(order_inner[1]+1);

    LegendrePolynomial (order_inner[0], 2*x-1, polx);
    LegendrePolynomial (order_inner[1], 2*y-1, poly);
    
    for (int i = 0; i <= order_inner[0]; i++)
      for (int j = 0; j <= order_inner[1]; j++)
	shape(ii++) = polx[i] * poly[j];
  }

  HybridTrig :: HybridTrig (int aorder) : HybridFiniteElement (ET_TRIG)

{
  nfacets = 3;
  order = aorder;
  for ( int i = 0; i < 6; i++)
    order_facet[i] = INT<2> (aorder, aorder);

  ndof = order_facet[0][0] + order_facet[1][0] + order_facet[2][0] + 3;

  int dof = 0;
  for(int i=0; i<3; i++)
    {
      firstfacedof[i] = dof;
      dof += order_facet[i][0] + 1;
    }
  firstfacedof[3] = dof;
  ndof = dof;
  //  if (order_facet > order_inner)
  //    throw Exception ("HDivSymTrig, order_face > order_inner)");
}


void HybridTrig :: ComputeNDof ()
{
  int dof = 0;
  for(int i=0; i<3; i++)
    {
      firstfacedof[i] = dof;
      dof += order_facet[i][0] + 1;
    }
  firstfacedof[3] = dof;
  ndof = dof;

  order = max2(order_facet[0][0], max2(order_facet[1][0], order_facet[2][0]));
}



void HybridTrig :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
}


void HybridTrig :: CalcShape (const IntegrationPoint & ip, 
			       FlatVector<> shape) const   // n * 3 
{
  double x = ip(0), y=ip(1); 
  double lami[3] = {x,y,1-x-y};
  double ddlami[3] = {x,y,1-x-y};

  shape = 0.0; 
    
  int i, j, k;
  int ii = 0; 
    
 
  
  const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
  ArrayMem<double, 20> ha(order+1);
  
  for (i = 0; i < 3; i++)
    {
      int es = edges[i][0]; 
      int ee = edges[i][1];
	
      int fac=1;

      if (vnums[es] > vnums[ee]) { swap (es, ee); fac = -1;}

      
     
      double ls = lami[es];
      double le = lami[ee];

      ArrayMem<double, 20> polx(order_facet[i][0]+1);
      LegendrePolynomial (order_facet[i][0], ls - le, polx);

       for(int l=0; l<=order_facet[i][0]; l++)
	{
	  shape(ii) = polx[l] *fac;
	  ii++;
	}

	
    }
}


void HybridTrig :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  fdofs.SetSize(firstfacedof[face+1] - firstfacedof[face]);

  for(int i=firstfacedof[face]; i<firstfacedof[face+1]; i++)
    fdofs[i - firstfacedof[face] ] = i;
 
}



  HybridTet :: HybridTet (int aorder) : HybridFiniteElement (ET_TET)
{
  nfacets = 4;
  order = aorder; //aorder+3;//aorder;

  for ( int i = 0; i < 6; i++ )
    order_facet[i] = INT<2> (aorder, aorder);

  int dof = 0;
  int nfa = 4;
  for(int i=0; i<nfa; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][0]+2) / 2;
    }
  firstfacedof[nfa] = dof;
  ndof = firstfacedof[nfa];

  //  if (order_facet > order_inner)
  //    throw Exception ("HDivSymTrig, order_face > order_inner)");
}


void HybridTet :: ComputeNDof()
{
  int dof = 0;
  int nfa = 4;
  for(int i=0; i<nfa; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][0]+2) / 2;
    }
  firstfacedof[nfa] = dof;
  ndof = firstfacedof[nfa];

  order = order_facet[0][0];
  for ( int i = 1; i < 4; i++ )
    if ( order < order_facet[i][0] ) order = order_facet[i][0];

}



void HybridTet :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
}



void HybridTet :: CalcShape (const IntegrationPoint & ip, 
			       FlatVector<> shape) const   // n * 3 
{
  double x = ip(0), y=ip(1), z = ip(2); 
  double lami[4] = {x,y,z,1-x-y-z};
  double ddlami[4] = {x,y,z,1-x-y-z};

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  
  
  const FACE * faces = ElementTopology::GetFaces (ET_TET);
  int nf = ElementTopology::GetNFaces (ET_TET);
  ArrayMem<double, 20> ha(order+1);
  
  for (i = 0; i < 4; i++)
    {
      int fav[3] =  { faces[i][0], faces[i][1], faces[i][2] };
      int point = 6 - fav[0] - fav[1] - fav[2];
      int fac=1;
      
      int n = order_facet[i][0];

      //Sort vertices  first edge op minimal vertex
      if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
      if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); fac *= -1;}
      if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
     
      ArrayMem<double, 20> polx(n+1), poly(n+1);
      
      ScaledLegendrePolynomial (n, 2 * lami[fav[0]]+lami[fav[1]]-1, 1-lami[fav[1]], polx);
      
//       shape(ii++) = lami[fav[0]] * fac;
//       shape(ii++) = lami[fav[1]] * fac;
//       shape(ii++) = lami[fav[2]] * fac;

      
      for (int k = 0; k <= n; k++)
	{
	  JacobiPolynomial (n, 2*lami[fav[1]]-1, 2*k+1, 0, poly);
	  for (int j = 0; j <= n-k; j++)
	    shape(ii++) = polx[k] * poly[j] * fac;
	}
      
    }
}

void HybridTet :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  fdofs.SetSize(firstfacedof[face+1] - firstfacedof[face]);
  for(int i=firstfacedof[face]; i<firstfacedof[face+1]; i++)
    fdofs[i - firstfacedof[face] ] = i;
}


/****************************HEX**************************/

  HybridHex :: HybridHex (int aorder) : HybridFiniteElement (ET_HEX)
{
  nfacets = 6;
  order = aorder; //aorder+3;//aorder;

  for ( int i = 0; i < 6; i++)
    order_facet[i] = INT<2> (aorder, aorder);

  eltype = ET_HEX;
  cout << "Hybrid space: " << endl;
  cout << "hex, order = " << order << endl;
  cout << "ndof = " << ndof << endl;

  int dof = 0;
  int nfa = 6;
  for(int i=0; i<nfa; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
    }
  firstfacedof[nfa] = dof;
  ndof = dof;

  //  if (order_facet > order_inner)
  //    throw Exception ("HDivSymTrig, order_face > order_inner)");
}


void HybridHex :: ComputeNDof ()
{
  int dof = 0;
  int nfa = 6;
  for(int i=0; i<nfa; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
    }
  firstfacedof[nfa] = dof;
  ndof = dof;

  order = order_facet[0][0];
  for ( int i = 0; i < 6; i++ )
    if ( order < max2(order_facet[i][0], order_facet[i][1] )) 
      order = max2(order_facet[i][0], order_facet[i][1] );
}



void HybridHex :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
}



void HybridHex :: CalcShape (const IntegrationPoint & ip, 
			       FlatVector<> shape) const   // n * 3 
{
  double x = ip(0), y=ip(1), z = ip(2); 
  double lami[8]={(1-x)*(1-y)*(1-z),x*(1-y)*(1-z),x*y*(1-z),(1-x)*y*(1-z),
		(1-x)*(1-y)*z,x*(1-y)*z,x*y*z,(1-x)*y*z}; 

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  
  
  const FACE * faces = ElementTopology::GetFaces (ET_HEX);
  int nf = ElementTopology::GetNFaces (ET_HEX);
  ArrayMem<double, 20> ha(order+1);
  
  for (i = 0; i < 6; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;

      int f1 = faces[i][(fmax+3)%4];
      int f2 = faces[i][(fmax+1)%4]; 
      fmax = faces[i][fmax]; 
      

      int fac = 1;
      if(vnums[f2]>vnums[f1]) { swap(f1,f2); fac = -1;} // fmax > f1 > f2 
       
      ArrayMem<double, 20> polx(order+1), poly(order+1);
      double x1 = lami[f1] - lami[fmax];
      double y1 = lami[f2] - lami[fmax];
      LegendrePolynomial (order_facet[i][0], 2*x1-1.0, polx);
      LegendrePolynomial (order_facet[i][1], 2*y1-1.0, poly);
      
      for (int k = 0; k <= order_facet[i][0]; k++)
	for (int j = 0; j <= order_facet[i][1]; j++)
	  shape(ii++) = polx[k] * poly[j] * fac;
    }

     

}

void HybridHex :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  fdofs.SetSize(firstfacedof[face+1] - firstfacedof[face]);

  for(int i=firstfacedof[face]; i<firstfacedof[face+1]; i++)
    fdofs[i - firstfacedof[face] ] = i;
 
}

/**************************** PRISM **************************/

  HybridPrism :: HybridPrism (int aorder) : HybridFiniteElement (ET_PRISM)
{
  nfacets = 5;
  order = aorder; //aorder+3;//aorder;
  for ( int i = 0; i < 6; i++)
    order_facet[i] = INT<2> (aorder, aorder);

  // 2 trigs + 3 quads

  eltype = ET_PRISM;
  cout << "Hybrid space: " << endl;
  cout << "prism, order = " << order << endl;
  cout << "ndof = " << ndof << endl;

  int dof = 0;
  int nfa = 5;
  int ndofperface = sqr(order_facet[0][0]+1);

  // trig faces
  for(int i=0; i<2; i++)
    {
      firstfacedof[i] = dof;
      dof += int ( (order_facet[i][0]+1) * (order_facet[i][0]+2) / 2 ) ;
    }
  // quad faces
  for(int i=2; i<5; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
    }
  firstfacedof[nfa] = dof;
  ndof = dof;

  //  if (order_facet > order_inner)
  //    throw Exception ("HDivSymTrig, order_face > order_inner)");
}

void HybridPrism :: ComputeNDof()
{
  int dof = 0;
  int nfa = 5;
  int ndofperface = sqr(order+1);

  // trig faces
  for(int i=0; i<2; i++)
    {
      firstfacedof[i] = dof;
      dof += int ( ((order_facet[i][0]+1) * (order_facet[i][0]+2)) / 2 ) ;
    }
  // quad faces
  for(int i=2; i<5; i++)
    {
      firstfacedof[i] = dof;
      dof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
    }
  firstfacedof[nfa] = dof;
  ndof = dof;

  order = order_facet[0][0];
  for ( int i = 0; i < 5; i++ )
    if ( order < max2(order_facet[i][0], order_facet[i][1] )) 
      order = max2(order_facet[i][0], order_facet[i][1] );
}




void HybridPrism :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
}



void HybridPrism :: CalcShape (const IntegrationPoint & ip, 
			       FlatVector<> shape) const   // n * 3 
{
  double x = ip(0), y=ip(1), z = ip(2); 
  double lami[6]={ x*(1-z), y*(1- z), (1-x-y)*(1-z), x*z, y*z, (1-x-y)*z };
		

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  
  
  const FACE * faces = ElementTopology::GetFaces (ET_PRISM);
  int nf = ElementTopology::GetNFaces (ET_PRISM);
  ArrayMem<double, 20> ha(order+1);
  

  // trig faces
  for (i = 0; i < 2; i++)
    {
      int fav[3] =  { faces[i][0], faces[i][1], faces[i][2] };
      int fac=1;
      
      
      //Sort vertices  first edge op minimal vertex
      if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
      if(vnums[fav[1]] > vnums[fav[2]]) {swap(fav[1],fav[2]); fac *= -1;}
      if(vnums[fav[0]] > vnums[fav[1]]) {swap(fav[0],fav[1]); fac *= -1;}
     
      int n = order_facet[i][0];

      ArrayMem<double, 20> polx(n+1), poly(n+1);
      
      ScaledLegendrePolynomial (n, lami[fav[1]]-lami[fav[2]], 1-lami[fav[0]], polx);
      
//       shape(ii++) = lami[fav[0]] * fac;
//       shape(ii++) = lami[fav[1]] * fac;
//       shape(ii++) = lami[fav[2]] * fac;

      
      for (int k = 0; k <= n; k++)
	{
	  JacobiPolynomial (n, 2*lami[fav[0]]-1, 2*k+1, 0, poly);
	  for (int j = 0; j <= n-k; j++)
	    shape(ii++) = polx[k] * poly[j] * fac;
	}
      
      
    }

  // quad faces
  for( i = 2; i<5; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;

      int f1 = faces[i][(fmax+3)%4];
      int f2 = faces[i][(fmax+1)%4]; 
      fmax = faces[i][fmax]; 
      

      int fac = 1;
      if(vnums[f2]>vnums[f1]) { swap(f1,f2); fac = -1;} // fmax > f1 > f2 
       
      ArrayMem<double, 20> polx(order+1), poly(order+1);
      double x1 = lami[f1] - lami[fmax];
      double y1 = lami[f2] - lami[fmax];
      LegendrePolynomial (order_facet[i][0], 2*x1-1.0, polx);
      LegendrePolynomial (order_facet[i][1], 2*y1-1.0, poly);
      
      for (int k = 0; k <= order_facet[i][0]; k++)
	for (int j = 0; j <= order_facet[i][1]; j++)
	  shape(ii++) = polx[k] * poly[j] * fac;
    }

     

}

void HybridPrism :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  fdofs.SetSize(firstfacedof[face+1] - firstfacedof[face]);

  for(int i=firstfacedof[face]; i<firstfacedof[face+1]; i++)
    fdofs[i - firstfacedof[face] ] = i;
 
}




/***************************Quad********************************/



  HybridQuad :: HybridQuad (int aorder) : HybridFiniteElement(ET_QUAD)
{
  nfacets = 4;
  order = aorder;

  for (int i = 0; i < 6; i++ )
    order_facet[i] = INT<2> (aorder, aorder);

  cout << "quad, order = " << order << endl;
  cout << "ndof = " << ndof << endl;

  int dof = 0;

  for(int i=0; i<nfacets; i++)
    {
      firstfacedof[i] = dof;
      dof += order_facet[i][0] + 1;
    }
  firstfacedof[nfacets] = dof;
  ndof = dof;

}

void HybridQuad :: ComputeNDof ()
{
  int dof = 0;
  for(int i=0; i<nfacets; i++)
    {
      firstfacedof[i] = dof;
      dof += order_facet[i][0] + 1;
    }
  firstfacedof[nfacets] = dof;
  ndof = dof;

  order = order_facet[0][0];
  for ( int i = 1; i < 4; i++ )
    if ( order < order_facet[i][0] ) order = order_facet[i][0];
}

void HybridQuad :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
}


void HybridQuad :: CalcShape (const IntegrationPoint & ip, 
			       FlatVector<> shape) const   // n * 3 
{
  double x = ip(0), y=ip(1);

  double lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
  ArrayMem<double, 20> ha(order+1);
 
  
  
  for (i = 0; i < 4; i++)
    {
      int es = edges[i][0]; 
      int ee = edges[i][1];
	
      int fac=1;

      if (vnums[es] > vnums[ee]) { swap (es, ee); fac = -1;}

      
     
      double ls = lami[es];
      double le = lami[ee];

     
      ArrayMem<double, 20> polx(order_facet[i][0]+1);
      LegendrePolynomial (order_facet[i][0], ls - le, polx);

       for(int l=0; l<=order_facet[i][0]; l++)
	{
	  if ( (ls+le) > 0.99 )
	  shape(ii) = polx[l] *fac;
	  ii++;
	}

	
    }
}


void HybridQuad :: GetFDofNrs(const int & face, ARRAY<int> & fdofs) const
{
  fdofs.SetSize(firstfacedof[face+1] - firstfacedof[face]);

  for(int i=firstfacedof[face]; i<firstfacedof[face+1]; i++)
    fdofs[i - firstfacedof[face] ] = i;
 
}







template <int D>
class HybridRobinIntegrator 
  : public T_BDBIntegrator<DiffOpBoundIdHybrid<D>, DiagDMat<1>, FiniteElement>
{
public:
  HybridRobinIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpBoundIdHybrid<D>, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HybridRobinIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "Hybrid-Robin"; }
  virtual bool BoundaryForm () const { return 1; }
};



template <int D>
class HybridNeumannIntegrator 
  : public T_BIntegrator<DiffOpBoundIdHybrid<D>, DVec<1>, FiniteElement>
{
public:
  HybridNeumannIntegrator (CoefficientFunction * coef)
    :  T_BIntegrator<DiffOpBoundIdHybrid<D>, DVec<1>, FiniteElement>
  (DVec<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HybridNeumannIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "Hybrid-Neumann"; }
  virtual bool BoundaryForm () const { return 1; }
};



namespace  
{
  using namespace ngfem;

  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetIntegrators().AddBFIntegrator ("robinhybrid", 2, 1,
				      HybridRobinIntegrator<2>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhybrid", 2, 1,
				      HybridNeumannIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("robinhybrid", 3, 1,
				      HybridRobinIntegrator<3>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhybrid", 3, 1,
				      HybridNeumannIntegrator<3>::Create);

  }
  
  Init init;

}


}



