/*

Mixed elasticity -  hybrid fespace for Lagrange parameters on edges/faces

*/



#include <solve.hpp>
#include <comp.hpp>
#include <fem.hpp>

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif


namespace ngcomp
{

using namespace ngfem;
using namespace ngsolve;
using namespace ngcomp;

#ifdef PARALLEL
using namespace ngparallel;
#endif


/**************************
  FESpace for lagrange parameters
  *****************************/




FESpaceHybrid :: FESpaceHybrid (const MeshAccess & ama, const Flags & aflags, bool parseflags)
  : FESpace (ama, aflags)
{
  name="FESpaceHybrid(hybrid)";
  // allowed flags
  DefineNumFlag("order");
  DefineNumFlag("facetorder");
  DefineDefineFlag("hybridspace");
  
  if(parseflags) ParseFlags(aflags);

  cout << "**********************+ Hybrid space, order = " << order << endl;
  cout << "flags = " << aflags << endl;

  INT<2> orderfacet (order, order);
  order_facet.SetSize(ma.GetNFacets());
  order_facet = orderfacet;

  orderz = int(aflags.GetNumFlag("orderz", -1));

   if (ma.GetDimension() == 2)
     {
       segm = new HybridNormalSegm (order);
       trig = new HybridTrig (order);
       quad = new HybridQuad (order);
     }
   else
     {
       trig = new HybridNormalTrig (order); // order_facet);
       quad = new HybridNormalQuad (order);
       prism = new HybridPrism (order);
       tet = new HybridTet (order);
       // hex = new HybridHex (order, order_facet);
     }
   prol = 0;

   ndof_level.SetSize(0);

#ifdef PARALLEL
   ResetParallelDofs();
#endif

}


FESpaceHybrid :: ~FESpaceHybrid ()
{
  ;
}

void FESpaceHybrid :: PrintReport (ostream & ost)
{
  ost << "FESpaceHybrid, order " << order << endl;
  int uniformfacetorder = order_facet[0][0];
  for( int i = 0; i < ma.GetNFacets(); i++ )
    if ( uniformfacetorder != order_facet[i][0] || uniformfacetorder != order_facet[i][1] )
      uniformfacetorder = -1; 
  if ( uniformfacetorder >= 0 )
    ost << "facetorder " << uniformfacetorder << endl;
  ost << "ndof = " << ndof << endl;

}


void FESpaceHybrid :: Update()
{

  nel = ma.GetNE();
  ndof = 0;
  ARRAY<int> elfacets, verts;

  INT<2> oldorderfacet = order_facet[0];
//   oldorderfacet[0]++;
//   oldorderfacet[1]++;

  order_facet.SetSize(ma.GetNFacets());  
  order_facet = oldorderfacet;

//   order++;

  // set orders
  if ( orderz >= 0 &&  ma.GetDimension() == 2)
    {
      // set orders on quads
      nfa = ma.GetNEdges();

      for (int el = 0; el < ma.GetNE(); el++)
	{
	  ma.GetElEdges(el, elfacets);
	  if ( ma.GetElType(el) == ET_TRIG ) continue;

	  for ( int i = 0; i < 4; i++ )
	    { 
	      int v1, v2;
	      Vec<2> point1, point2, normal;
	      ma.GetEdgePNums(elfacets[i], v1, v2);
	      ma.GetPoint(v1, point1);
	      ma.GetPoint(v2, point2);
	      normal(0) = point1(1) - point2(1);
	      normal(1) = point2(0) - point2(0);
	      normal /= L2Norm(normal);
	      if ( fabs(normal(1)) < 1e-1 )
		{
		  order_facet[elfacets[i]] = 1;
		  // *testout << "HYBRID order 1: facet " << elfacets[i] << endl;
		}
	    }
	}
    }

  ndof = 0;

  if ( ma.GetDimension() == 2)
    {

      nfa = ma.GetNEdges();
      first_face_dof.SetSize(nfa+1);
      first_face_dof = 0;

      fine_face.SetSize(nfa); 
      fine_face = 0; 

      for ( int i = 0; i < nel; i++)
	{
	  ma.GetElEdges (i, elfacets);		
	  for(int j=0;j<elfacets.Size();j++) fine_face[elfacets[j]] = 1; 

	}  
      for (int i = 0; i < nfa; i++)
	{
	  first_face_dof[i] = ndof;
	  if ( fine_face[i] )
	    ndof += order_facet[i][0] + 1 ;
	}
      first_face_dof[nfa] = ndof;  
    }
   else
     {
       nfa = ma.GetNFaces();
       first_face_dof.SetSize(nfa+1);
       ARRAY<int> pnums;

       fine_face.SetSize(nfa); 
       fine_face = 0; 
       
       for ( int i = 0; i < nel; i++)
	 {
	   ma.GetElFaces (i, elfacets);		
	   for(int j=0;j<elfacets.Size();j++) fine_face[elfacets[j]] = 1; 
	   
	 }  
  
       for (int i = 0; i < nfa; i++)
 	{
 	  first_face_dof[i] = ndof;

	  if ( !fine_face[i]) continue;
 	  ma.GetFacePNums(i, pnums);
 	  if (pnums.Size() == 3)
 	    ndof += (order_facet[i][0]+1)*(order_facet[i][0]+2)/2;
 	  else
 	    ndof += (order_facet[i][0]+1)*(order_facet[i][1]+1);
 	}
       first_face_dof[nfa] = ndof;  
     }

  first_element_dof.SetSize(nel+1);
  for (int i = 0; i < nel; i++)
    {
      first_element_dof[i] = ndof; // no element dof

    }

  first_element_dof[nel] = ndof;    

  
  (*testout) << "hybrid space : " << endl;
  (*testout) << "ndof " << ndof << endl; 
  (*testout) << "number of elements " << ma.GetNE() << endl;
  (*testout) << "number of facets " << ma.GetNFacets() << endl;
  (*testout) << "ndof " << ndof << " =  2 * " << ma.GetNFacets() << " = " << 2 * ma.GetNFacets() << endl;

  ndof_level.Append(ndof);

#ifdef PARALLEL
    UpdateParallelDofs();
#endif

}






 const FiniteElement & FESpaceHybrid:: GetFE (int elnr, LocalHeap & lh) const
 {
   HybridFiniteElement * fe2d = 0;
   HybridFiniteElement * fe3d = 0;
   if (ma.GetElType(elnr) == ET_TRIG)
     fe2d = dynamic_cast<HybridFiniteElement*> (trig);
   else if (ma.GetElType(elnr) == ET_QUAD)
     fe2d = dynamic_cast<HybridFiniteElement*> (quad);
    else if (ma.GetElType(elnr) == ET_TET)
      fe3d = dynamic_cast<HybridFiniteElement*> (tet);
    else if (ma.GetElType(elnr) == ET_PRISM)
      fe3d = dynamic_cast<HybridFiniteElement*> (prism);
    else if (ma.GetElType(elnr) == ET_HEX)
      fe3d = dynamic_cast<HybridFiniteElement*> (hex);
//     else if (ma.GetElType(elnr) == ET_PYR)
//       fe3d = dynamic_cast<HybridFiniteElement*> (pyr);
   else cout << "not implemented " << endl;

   ArrayMem<int,12> vnums; 
   ma.GetElVertices(elnr, vnums);
  
  if (fe2d)
    {
      ArrayMem<int, 12> ednums;
      ArrayMem<INT<2>, 12> order_ed;
      
      ma.GetElEdges(elnr, ednums);
      order_ed.SetSize (ednums.Size());
      for (int j = 0; j < ednums.Size(); j++)
	order_ed[j] = order_facet[ednums[j]];

      fe2d -> SetVertexNumbers (vnums);
      fe2d -> SetOrderFacet (order_ed);
      fe2d -> ComputeNDof ();
      return *fe2d;
    }
  if (fe3d)
    {
      ArrayMem<int, 6> fanums; 
      ArrayMem<INT<2>, 6> order_fa;
      ma.GetElFaces(elnr, fanums);
      order_fa.SetSize (fanums.Size());
      
      for (int j = 0; j < fanums.Size(); j++)
	order_fa[j] = order_facet[fanums[j]];
      
      fe3d -> SetVertexNumbers (vnums);
	
      fe3d -> SetOrderFacet (order_fa);
      
      fe3d -> ComputeNDof ();
      return *fe3d;
    }
  
 }

const FiniteElement & FESpaceHybrid :: GetSFE (int elnr, LocalHeap & lh) const
{
  		
  HybridNormalFiniteElement * fe;

  if (ma.GetSElType(elnr) == ET_SEGM)
    fe = dynamic_cast<HybridNormalFiniteElement*> (segm);
   else if (ma.GetSElType(elnr) == ET_TRIG)
     fe = dynamic_cast<HybridNormalFiniteElement*> (trig);
   else if (ma.GetSElType(elnr) == ET_QUAD)
     fe = dynamic_cast<HybridNormalFiniteElement*> (quad);


   else cout << "not implemented " << endl;

  ArrayMem<int,12> vnums, ednums; 
  ma.GetSElVertices(elnr, vnums);
  fe -> SetVertexNumbers (vnums);

  if(ma.GetSElType(elnr) == ET_SEGM)
    {
      ma.GetSElEdges(elnr, ednums);
	
      fe -> SetOrderInner (order_facet[ednums[0]]);
      fe -> ComputeNDof();
    }
  else
    {
      INT<2> order_fa = order_facet[ma.GetSElFace(elnr)];
      fe -> SetOrderInner (order_fa);
      fe -> ComputeNDof();
    }

  return *fe;    
}



 void FESpaceHybrid :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
 {   
   dnums.SetSize(0);
   ArrayMem<int,12> vnums, ednums, fanums; 
   int i, j;
   int first, next; 

   if (ma.GetDimension() == 2)
     ma.GetElEdges (elnr, ednums);
   else
     ma.GetElFaces (elnr, ednums);

   int nelfacets;
   switch ( ma.GetElType(elnr) )
     {
     case ET_TRIG :
       nelfacets = 3;
       break;
     case ET_QUAD:
     case ET_TET:
       nelfacets = 4;
       break;
     case ET_PRISM:
       nelfacets = 5;
       break;
     case ET_HEX:
       nelfacets = 6;
       break;
     }
   for (i = 0; i < nelfacets; i++)
     {
       first = first_face_dof[ednums[i]];
       next = first_face_dof[ednums[i]+1];
       for (j = first; j < next; j++)
 	dnums.Append (j);
     }

   /*   
	(*testout) << "after faces Hybrid_dnums.size " << dnums.Size() << endl;
	*testout << " GetElDNums hybrid elnr " << elnr << " \t " << dnums << endl; 
	*/

//    first = first_element_dof[elnr];
//    next = first_element_dof[elnr+1];
//    for (int j = first; j < next; j++)
//      dnums.Append (j);


 }



   void FESpaceHybrid :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
   {
     // all dofs are external

       {
 	GetDofNrs (elnr, dnums);
 	return;
       }


   }
 









void FESpaceHybrid :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
{
  dnums.SetSize (0);

  ArrayMem<int,12> vnums, ednums, fanums; 
  int i, j;
  int first, next; 
  if (ma.GetDimension() == 2)
    ma.GetSElEdges (elnr, ednums);
  else
    {
      ednums.SetSize(1);
      ednums[0] = ma.GetSElFace (elnr);
    }

  first = first_face_dof[ednums[0]];
  next = first_face_dof[ednums[0]+1];
  for (j = first; j < next; j++)
    dnums.Append (j);
}


void FESpaceHybrid :: GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
{
  dnums.SetSize(0);

  ARRAY<int> faces;
  ma.GetElFaces ( elnr, faces );
  for ( int i = 0; i < faces.Size(); i++ )
    {
      for ( int j = first_face_dof[faces[i]]; j < first_face_dof[faces[i]]+1; j++)
	dnums.Append(j);
    }
}

void FESpaceHybrid :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
{ 
  dnums.SetSize(0); 
  for ( int i = first_face_dof[fanr]; i < first_face_dof[fanr+1]; i++)
    dnums.Append(i);

}


#ifdef PARALLEL

void FESpaceHybrid :: UpdateParallelDofs_hoproc()
  {
    
     // ******************************
     // update exchange dofs 
     // ******************************
    *testout << "Hybrid::UpdateParallelDofs_hoproc" << endl;
    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks);
    nexdof = 0;

    MPI_Status status;
    MPI_Request * sendrequest = new MPI_Request[ntasks];
    MPI_Request * recvrequest = new MPI_Request[ntasks];

    // number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	if ( !parallelma->IsExchangeFace ( face ) ) continue;
	
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	nexdof[id] += dnums.Size() ; 

	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    nexdof[dest] += dnums.Size() ; 
      }

    // + number of inner exchange dofs --> set eliminate internal

    nexdof[0] = ma.GetNFaces();

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);

    ARRAY<int> ** owndofs, ** distantdofs;
    owndofs = new ARRAY<int>* [ntasks];
    distantdofs = new ARRAY<int>* [ntasks];

    for ( int i = 0; i < ntasks; i++ )
      {
	owndofs[i] = new ARRAY<int>(1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int>(0);
      }

    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;
    int exdof = 0;
    int ii;

    // *****************
    // Parallel Face dofs
    // *****************


   for ( int face = 0; face < ma.GetNFaces(); face++ )
      if ( parallelma->IsExchangeFace ( face ) )
      {
	ARRAY<int> dnums;
	GetFaceDofNrs ( face, dnums );
	if ( dnums.Size() == 0 ) continue;

 	for ( int i=0; i<dnums.Size(); i++ )
 	  (*(paralleldofs->localexchangedof))[id][exdof++] = dnums[i];

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;

	    owndofs[dest]->Append ( distface );
	    owndofs[dest]->Append (int(paralleldofs->IsGhostDof(dnums[0])) );
	    for ( int i=0; i<dnums.Size(); i++)
	      {
		paralleldofs->SetExchangeDof ( dest, dnums[i] );
		paralleldofs->SetExchangeDof ( dnums[i] );
		owndofs[dest]->Append ( dnums[i] );
	      }
  	  }
      }   


    for ( int sender = 1; sender < ntasks; sender ++ )
      {
        if ( id == sender )
          for ( int dest = 1; dest < ntasks; dest ++ )
            if ( dest != id )
              {
		MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
              }
	  
        if ( id != sender )
          {
	    MyMPI_IRecv ( *distantdofs[sender], sender, recvrequest[sender]);
          }
 	  
	  
      }

     for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait(recvrequest+dest, &status);
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;

	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int isdistghost = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    ii++; 
	    for ( int i=1; i<dnums.Size(); i++)
	      {
		(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[i];
		(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
		if ( dest < id && !isdistghost )
		  paralleldofs->ismasterdof.Clear ( dnums[i] ) ;
		ii++; cnt_nexdof[dest]++;
	      }
	  }
      }



     /*******************************

         update low order space

     *****************************/


     int ndof_lo = ma.GetNFaces();

     // all dofs are exchange dofs
     nexdof = ndof_lo;
     
     exdof = 0;
     cnt_nexdof = 0;
     

     // *****************
     // Parallel Face dofs
     // *****************
     
     owndofs[0]->SetSize(1);
     (*owndofs[0])[0] = ndof;
     distantdofs[0]->SetSize(0);
     
     // find local and distant dof numbers for face exchange dofs
     for ( int face = 0; face < ma.GetNFaces(); face++ )
       {
	 int dest = 0;
	 
	 ARRAY<int> dnums;
	 GetFaceDofNrs ( face, dnums );
	 int distface = parallelma -> GetDistantFaceNum ( dest, face );
	 owndofs[0]->Append ( distface );
	 paralleldofs->SetExchangeDof ( dest, dnums[0] );
	 owndofs[dest]->Append ( dnums[0] );
      }   
    
    int dest = 0;
    MyMPI_ISend ( *owndofs[0], dest, sendrequest[dest]);
    MyMPI_IRecv ( *distantdofs[0], dest, recvrequest[dest]);
   
    MPI_Wait ( recvrequest+dest, &status);


    // denk nach und machs gscheit...

    
    paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
    ii = 1;
    while ( ii < distantdofs[0]->Size() )
      {
	int fanum = (*distantdofs[0])[ii++];
	int dnum = (*distantdofs[0])[ii++];
	ARRAY<int> dnums;
	GetFaceDofNrs (fanum, dnums);
	(*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[0];
	(*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = dnum;
	cnt_nexdof[dest]++;
      }
    
    for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;

  }

  void FESpaceHybrid :: UpdateParallelDofs_loproc()
  {
    *testout << "Hybrid::UpdateParallelDofs_loproc" << endl;

    const MeshAccess & ma = (*this). GetMeshAccess();

    int ndof = GetNDof();

    // Find number of exchange dofs
    ARRAY<int> nexdof(ntasks); 
    nexdof = 0;

    MPI_Status status;
    MPI_Request * sendrequest = new MPI_Request[ntasks];
    MPI_Request * recvrequest = new MPI_Request[ntasks];

    // number of face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	nexdof[id] ++;//= dnums.Size() ; 
	for ( int dest = 1; dest < ntasks; dest ++ )
	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
	    { 
	      nexdof[dest] ++; 
	    }
      }

    paralleldofs->SetNExDof(nexdof);

    paralleldofs->localexchangedof = new Table<int> (nexdof);
    paralleldofs->distantexchangedof = new Table<int> (nexdof);



    ARRAY<int> ** owndofs,** distantdofs;
    owndofs = new ARRAY<int> * [ntasks];
    distantdofs = new ARRAY<int> * [ntasks];

    for ( int i = 0; i < ntasks; i++)
      {
	owndofs[i] = new ARRAY<int> (1);
	(*owndofs[i])[0] = ndof;
	distantdofs[i] = new ARRAY<int> (0);
      }

    int exdof = 0;
    ARRAY<int> cnt_nexdof(ntasks);
    cnt_nexdof = 0;

    // *****************
    // Parallel Face dofs
    // *****************


    // find local and distant dof numbers for face exchange dofs
    for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	(*(paralleldofs->localexchangedof))[id][exdof++] = face;

	for ( int dest = 1; dest < ntasks; dest++ )
	  {
	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
	    if( distface < 0 ) continue;
	    ARRAY<int> dnums;
	    GetFaceDofNrs ( face, dnums );
	    owndofs[dest]->Append ( distface );
	    paralleldofs->SetExchangeDof ( dest, dnums[0] );
	    paralleldofs->SetExchangeDof ( dnums[0] );
	    owndofs[dest]->Append (dnums[0]);
  	  }
      }   


    for ( int dest = 1; dest < ntasks; dest ++ )
      {
	MyMPI_ISend ( *owndofs[dest], dest, sendrequest[dest]);
	MyMPI_IRecv ( *distantdofs[dest], dest, recvrequest[dest]);
      }
    


    int ii = 1;
    for( int dest=1; dest<ntasks; dest++) 
      {
	if ( dest == id ) continue;
	MPI_Wait ( recvrequest+dest, &status );
	paralleldofs -> SetDistNDof( dest, (*distantdofs[dest])[0]) ;
	ii = 1;
	while ( ii < distantdofs[dest]->Size() )
	  {
	    int fanum = (*distantdofs[dest])[ii++];
	    int dnum = (*distantdofs[dest])[ii++];
	    ARRAY<int> dnums;
	    GetFaceDofNrs (fanum, dnums);
	    (*(paralleldofs->localexchangedof))[dest][ cnt_nexdof[dest] ] = dnums[0];
	    (*(paralleldofs->distantexchangedof))[dest][ cnt_nexdof[dest] ] = dnum;
	    cnt_nexdof[dest]++;
	     
	  }
      }
     for ( int i = 0; i < ntasks; i++ )
       delete distantdofs[i], owndofs[i];

     delete [] owndofs, distantdofs;
     delete [] sendrequest, recvrequest;
 
  }
#endif // PARALLEL



namespace  
{
  
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridspace", FESpaceHybrid::Create);
  }
  
  Init init;

}



}





