#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


#include <maxwell.hpp>

// maxwell equation, mixed formulation
//   
//   curl E   -  i omega mu H = 0
//   curl H   +  i omega eps E = j
//
// boundary conditions
//    omega mu H_tau = abs(k) E_tau    on GammaR

// hybridized system for E, H (discont. HCurl), 
//                       E_F, H_F (Lagrange parameters for E_tau, H_tau
//
//  ( E, curl h )_T  - (E_F, h_tau )_F  - (i omega mu H, h ) = 0
//  ( i omega eps E, e)  + ( curl H, e )_T  = ( j, e)
//  ( H_tau, e_F )_F  - (e_F , abs(k)/omega mu E_F )_GammaR   = 0
//  ( E_tau - E_F, e_tau - e_F )_F = 0
//  ( H_tau - H_F, h_tau - h_F )_F = 0
//



// compound fespace
// [ hcurl(discont), hcurl(discont), vectorfacet, vectorfacet ]
// for [ E, H, E_F, H_F ]

// flags kx, ky, kz to define the k-vector, otherwise 
// all elements are slave elements

// compound FiniteElement:
// finite element for MaxwellFESpace
// has pointer to fespace (and therefore to *isslave)

extern ngsolve::MeshAccess * ma;
extern ngsolve::PDE * pde;


namespace hybrid_maxwell
{
  using namespace ngfem;
  using namespace ngsolve;
  using namespace hybrid_maxwell;



  HybridMaxwellFESpace :: HybridMaxwellFESpace ( const MeshAccess & ama, 
						 const ARRAY<const FESpace*> & afespaces,
						 const Flags & flags)
    : CompoundFESpace (ama, afespaces, flags )
  {
    // build table isslave,
    // isslave[el][i] is true, if facet i of element el is a slave element
    // this means: dofs for this facet are in element-smoothing-block
    LocalHeap lh (10000000);

    isslave = new Table<int> (ma.GetNE(), 6);
    for ( int i = 0; i < ma.GetNE(); i++ )
      (*isslave)[i] = 1;
    
    // BitArray containing facets on surface
    BitArray issurffacet(ma.GetNFacets());
    ARRAY<int> facets;
    issurffacet.Clear();
    for ( int sel = 0; sel < ma.GetNSE(); sel++ )
      {
	ma.GetSElFacets(sel, facets);
	issurffacet.Set(facets[0] );
      }
    
    
    for ( int el = 0; el < ma.GetNE(); el++ )
      {
	ma.GetElFacets(el, facets);
	for ( int i = 0; i < facets.Size(); i++ )
	  {
	    if ( issurffacet.Test(facets[i]) ) 
	      (*isslave)[el][i] = 0;
	  }
      }


    if ( flags.GetStringFlag("kx","") == "" )
      {
	with_kvec = 0;
	for ( int i = 0; i < ma.GetNE(); i++ )
	  (*isslave)[i] = 0;
	return;
      }
    else
      with_kvec = true;


    // find slave elements
    CoefficientFunction *coef_kvec0 = pde->GetCoefficientFunction(flags.GetStringFlag("kx",""));
    CoefficientFunction *coef_kvec1 = pde->GetCoefficientFunction(flags.GetStringFlag("ky",""));;
    CoefficientFunction *coef_kvec2;

    int dim = ma.GetDimension();
    if ( dim == 3 )
      coef_kvec2  = pde->GetCoefficientFunction(flags.GetStringFlag("kz",""));
    else
      {
	ARRAY<double> coef(ma.GetNDomains() );
	coef = 0.;
	coef_kvec2 = new DomainConstantCoefficientFunction(coef);
      }

    for ( int el = 0; el < ma.GetNE(); el++ )
      {
	ElementTransformation eltrans;
	ma.GetElementTransformation ( el, eltrans, lh);

	ELEMENT_TYPE eltype = ma.GetElType(el);
	const POINT3D * points = ElementTopology::GetVertices (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	int nfa = ElementTopology :: GetNFacets(eltype);
	
	Facet2ElementTrafo transform(eltype); 
	const NORMAL * normals = ElementTopology::GetNormals(eltype);
	
	bool const_kvec =
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec0 ) && 
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec1 ) && 
	  dynamic_cast<const DomainConstantCoefficientFunction *> ( coef_kvec2 ); 

	for (int k = 0; k < nfa; k++)
	  {
	    if ( dim == 2 )
	      {
		ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);
		const IntegrationRule & ir1d =
		  SelectIntegrationRule (etfacet, 0 );
  
		Vec<2> kvec;
		Vec<2> normal_ref;
		for (int i=0; i<2; i++)
		  normal_ref(i) = normals[k][i];
 		IntegrationPoint ip = transform(k, ir1d[0]);
		SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
		kvec(0) = coef_kvec0->Evaluate(sip);
		kvec(1) = coef_kvec1->Evaluate(sip);

		Mat<2> inv_jac = sip.GetJacobianInverse();
		double det = sip.GetJacobiDet();
		
		Vec<2> normal = det * Trans (inv_jac) * normal_ref;
		double len = L2Norm(normal);
		normal /= len;

		double nx = 0.0;
		
		for ( int i = 0; i < 2; i++ )
		  nx += kvec(i) * normal(i);
		
		nx /= L2Norm(kvec);
		nx /= L2Norm(normal);
		
		// isslave = 0 --> im smoothing block
		if ( nx <= 0.0 ) 
		  {
		    (*isslave)[el][k] = 0; // [elnum][k] = 0;
		  }
	      }
	    else
	      {
		ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);
		const IntegrationRule & ir1d =
		  SelectIntegrationRule (etfacet, 0 );

		Vec<3> kvec;
		Vec<3> normal_ref;
		for (int i=0; i<3; i++)
		  normal_ref(i) = normals[k][i];
 		IntegrationPoint ip = transform(k, ir1d[0]);
		SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
		kvec(0) = coef_kvec0->Evaluate(sip);
		kvec(1) = coef_kvec1->Evaluate(sip);
		kvec(2) = coef_kvec2->Evaluate(sip);

		Mat<3> inv_jac = sip.GetJacobianInverse();
		double det = sip.GetJacobiDet();
		
		Vec<3> normal = det * Trans (inv_jac) * normal_ref;
		double len = L2Norm(normal);
		normal /= len;

		double nx = 0.0;
		
		for ( int i = 0; i < 3; i++ )
		  nx += kvec(i) * normal(i);
		
		nx /= L2Norm(kvec);
		nx /= L2Norm(normal);
		
		// isslave = 0 --> im smoothing block
		if ( nx <= 0.0 ) 
		  {
		    (*isslave)[el][k] = 0; // [elnum][k] = 0;
		  }
		;
	      }
	  }

      }


    if ( ma.GetDimension() == 2 ) delete coef_kvec2;
  }



  const FiniteElement & HybridMaxwellFESpace :: GetFE ( int elnr, LocalHeap & lh) const
  {
    ARRAY<const FiniteElement*> fea(4);
    for ( int i = 0; i < fea.Size(); i++ )
      fea[i] = &spaces[i]->GetFE(elnr, lh);

    void * mem = lh.Alloc (sizeof(HybridMaxwellFiniteElement));

    HybridMaxwellFiniteElement & fel = *new (mem) HybridMaxwellFiniteElement (fea);
    fel.SetElNum ( elnr );
    fel.SetMaxwellSpace ( this );

    return fel;

  }


  Table<int> * HybridMaxwellFESpace :: CreateSmoothingBlocks (int type) const
  {
      cout << "SmoothingType " << type << endl; 
      // #ifdef VER1
      // edge by edge
      
      int nd = GetNDof();
      
      int nfacets = (ma.GetDimension() == 2 ) ? ma.GetNEdges() : ma.GetNFaces();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();
      
      int order_u = (*this)[1]->GetOrder();
      int ndel = ((order_u+1)*(order_u+2))/2;
      
      int base_hdivn = GetStorageStart(2);
      int base_facet = GetStorageStart(3);
      int base_el = GetStorageStart(1);
      
      ARRAY<int> dnums;
      
      ARRAY<int> cnt; // ned/* +nel */);
      Table<int> *it;  

      switch ( type )
	{
	case 1:
	  cout << "element by element" << endl;
	  cnt.SetSize(nel);
	  
	  for (int i = 0; i < nel; i++)
	    {
	      GetExternalDofNrs (i, dnums);
	      cnt[i] = dnums.Size();
	    }
	  
	  it = new Table<int> (cnt);

	  for (int i = 0; i < nel; i++)
	    {
	      GetExternalDofNrs (i, dnums);
	      for (int j = 0; j < dnums.Size(); j++)
		(*it)[i][j] = dnums[j];
	    }
	  break;
	case 2:
	  cout << "facet by facet" << endl;
	  cnt.SetSize(nfacets);

	  for ( int i = 0; i < nfacets; i++ )
	    {
	      ;
	    }
	}

      return it;

  }

  FESpace * HybridMaxwellFESpace :: Create (const MeshAccess & ma, const Flags & flags )
  {
    ARRAY<const FESpace*> spaces(4);

    Flags Hflags(flags), Eflags(flags);
    Flags H_Fflags(flags), E_Fflags(flags);

    int order = int (flags.GetNumFlag ("order", 1));
  
    Hflags.SetFlag("order", order);
    Eflags.SetFlag("order", order);
    Hflags.SetFlag("orderinner", order);
    Eflags.SetFlag("orderinner", order);
    Hflags.SetFlag("orderedge", order);
    Eflags.SetFlag("orderedge", order);
    Hflags.SetFlag("orderface", order);
    Eflags.SetFlag("orderface", order);
    H_Fflags.SetFlag("order", order);
    E_Fflags.SetFlag("order", order);

    Hflags.SetFlag("discontinuous");
    Eflags.SetFlag("discontinuous");

    spaces[0] = new HCurlHighOrderFESpace (ma, Eflags);
    spaces[1] = new HCurlHighOrderFESpace (ma, Hflags);
    spaces[2] = new VectorFacetFESpace (ma, E_Fflags);
    spaces[3] = new VectorFacetFESpace (ma, H_Fflags);

    HybridMaxwellFESpace * fes = new HybridMaxwellFESpace (ma, spaces, flags);

    return fes;

  }




  string HybridMaxwellFESpace :: GetClassName() const
  { 
    return "HybridMaxwellFESpace"; 
  }


  HybridMaxwellFiniteElement ::  HybridMaxwellFiniteElement (ARRAY<const FiniteElement* > & afea )
    : CompoundFiniteElement (afea),
      maxwellspace(0),
      globelnum(-1)
  {
    ;
  }





  // register FESpaces
  namespace
#ifdef MACOS
  hybrid_maxwell_init_cpp
#endif
  {
   
    class Init
    { 
    public: 
      Init ();
    };
   
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hybridmaxwell", HybridMaxwellFESpace::Create);
  
    }

    Init init;
  }



}
