#ifndef FILE_HYBRID_MAXWELL__
#define FILE_HYBRID_MAXWELL__


class HybridMaxwellFESpace : public CompoundFESpace
{
protected:

  bool with_kvec;
  Table<int> * isslave;

public:
  HybridMaxwellFESpace ( const MeshAccess & ama, 
			 const ARRAY<const FESpace*> & afespaces,
			 const Flags & flags);

  virtual ~HybridMaxwellFESpace () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags );

  virtual string GetClassName() const;

  virtual const FiniteElement & GetFE ( int elnr, LocalHeap & lh ) const;

  Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
  {
    int blocktype = int(precflags.GetNumFlag("blocktype",1)); 
    Table<int> * it = CreateSmoothingBlocks ( blocktype );
  }


  Table<int> * CreateSmoothingBlocks (int type=1) const;

    inline void SetSlave ( int elnum, int fanum, int issl = 1 )
    {
      (*isslave)[elnum][fanum] = issl;
    }

    inline int IsSlave ( int elnum, int fanum ) const
    {
      return (*isslave)[elnum][fanum];
    }

  inline bool WithKVec () { return with_kvec; }
};


// A FiniteElement for HybridMaxwell, has pointer to FESpace
// and IsSlave-member

class HybridMaxwellFiniteElement : public CompoundFiniteElement
{

protected:
  const HybridMaxwellFESpace * maxwellspace;
  int globelnum;


public:
  HybridMaxwellFiniteElement (ARRAY<const FiniteElement* > & afea );

  ~HybridMaxwellFiniteElement ()
  { ; }

  /// select i-th component
  const FiniteElement & operator[] (int i) const { return *fea[i]; }
  /// 
  inline void SetSlave (int fanum, int issl = 1 )
  { 
    HybridMaxwellFESpace & constmaxwellspace = const_cast<HybridMaxwellFESpace&> (*maxwellspace);
    constmaxwellspace.SetSlave(globelnum, fanum, issl); 
  }

  inline int IsSlave ( int fanum ) const
  { return maxwellspace -> IsSlave ( globelnum, fanum); }

  inline int ElNum () const 
  { return globelnum; }

  inline void SetElNum ( int aglobelnum ) 
  { globelnum = aglobelnum; }

  inline void SetMaxwellSpace ( const HybridMaxwellFESpace * amaxwellspace )
  { maxwellspace = amaxwellspace;}



};

// Integrator for the mixed curl-curl problem, 
// only domain integrals

class HybridMaxwellIntegrator : public BilinearFormIntegrator
{
protected:
  CoefficientFunction * coeff_omega;
  CoefficientFunction * coeff_mu_real;
  CoefficientFunction * coeff_eps_real;
  CoefficientFunction * coeff_mu_imag;
  CoefficientFunction * coeff_eps_imag;

public:

  enum { D = 3 };

  HybridMaxwellIntegrator ( ARRAY<CoefficientFunction*> & coeffs );

  ~HybridMaxwellIntegrator () { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs );


  virtual bool BoundaryForm () const { return 0; }

  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<double> & elmat,
				      LocalHeap & lh) const
  {
    throw Exception ("HybridMaxwellIntegrator::AssembleElementMatrix for double called, equation is complex");
  }
  
  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<Complex> & elmat,
				      LocalHeap & lh) const;
  


};


// integrator for mixed hybrid maxwell problem
// facet contributions

class HybridMaxwellFacetIntegrator : public BilinearFormIntegrator
{
protected:

public:

  enum { D = 3 };

  HybridMaxwellFacetIntegrator ();

  ~HybridMaxwellFacetIntegrator () { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs );


  virtual bool BoundaryForm () const { return 0; }

  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<double> & elmat,
				      LocalHeap & lh) const
  {
    throw Exception ("HybridMaxwellFacetIntegrator::AssembleElementMatrix for double called, equation is complex");
  }
  
  virtual void AssembleElementMatrix (const FiniteElement & fel,
				      const ElementTransformation & eltrans, 
				      FlatMatrix<Complex> & elmat,
				      LocalHeap & lh) const;
  


};


void CrossNormal ( FlatMatrixFixHeight<3> shape, Vec<3> & normal );

#endif
