/*

Mixed elasticity

*/


#include <solve.hpp>
#include <fem.hpp>

using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace for mixed mechanics *************************** */
/* Compound FESpace:
   H(divdiv)
   H(curl)
*/

namespace ngcomp
{

  using namespace ngcomp;

class FESpaceMixedMech : public CompoundFESpace
{
public:
  FESpaceMixedMech (const MeshAccess & ama, 		   
		    const ARRAY<const FESpace*> & aspaces,
		    const Flags & aflags)
    : CompoundFESpace (ama, aspaces, aflags)
  {
    ;
  }

  virtual ~FESpaceMixedMech () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(2);

    cout << "Create EAS FE Space" << endl;
    
    int order = int (flags.GetNumFlag ("order", 1.0));

    int orderz = int(flags.GetNumFlag ("orderz", -1));

    Flags sigmaflags;
    sigmaflags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
    sigmaflags.SetFlag ("facetorder", flags.GetNumFlag ("facetorder", order));
    sigmaflags.SetFlag ("noncontinuous", 0.0);
    if (flags.GetDefineFlag ("eliminate_internal")) sigmaflags.SetFlag ("eliminate_internal");
    int sigmaorderz = int ( flags.GetNumFlag("sigmaorderz", orderz));
    sigmaflags.SetFlag("orderz", sigmaorderz);

    spaces[0] = new HDivSymHighOrderFESpace (ma, sigmaflags);
    
    Flags uflags;
    int curlorder = int (flags.GetNumFlag ("curlorder", order));
    uflags.SetFlag ("order", curlorder);
    uflags.SetFlag ("orderinner", flags.GetNumFlag ("curlorderinner", curlorder));
    uflags.SetFlag ("orderface", flags.GetNumFlag ("orderface", curlorder));
    uflags.SetFlag ("orderedge", flags.GetNumFlag ("orderedge", curlorder));
    uflags.SetFlag ("orderz", flags.GetNumFlag("curlorderz", orderz));


    ARRAY<double> maxz(ma.GetNDomains());
    maxz = 1;
    uflags.SetFlag ("dom_order_min_z", maxz);
    uflags.SetFlag ("dom_order_max_z", maxz);
    uflags.SetFlag ("variableorder");

    if (flags.GetDefineFlag ("eliminate_internal")) uflags.SetFlag ("eliminate_internal");
    spaces[1] = new HCurlHighOrderFESpaceA (ma, uflags);
    
    FESpaceMixedMech * fes = new FESpaceMixedMech (ma, spaces, flags);
    return fes;
  }
    
  virtual string GetClassName () const { return "MixedMech-FESpace"; }
};





















/* ******************** Integrator ************************** */









template <int D>
class HDivSymMassIntegrator 
  : public T_BDBIntegrator<DiffOpIdHDivSym<D>, DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT>, FiniteElement>
{
public:
  HDivSymMassIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpIdHDivSym<D>, DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT>, FiniteElement>
  (DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymMassIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Mass"; }
};


template <int D> class HDivSymSourceIntegrator { ; };




template <>
class HDivSymSourceIntegrator<2> 
  : public T_BIntegrator<DiffOpIdHDivSym<2>, DVec<3>, FiniteElement>
{
public:
  HDivSymSourceIntegrator (CoefficientFunction * coeff1,
			   CoefficientFunction * coeff2,
			   CoefficientFunction * coeff3)
    :  T_BIntegrator<DiffOpIdHDivSym<2>, DVec<3>, FiniteElement>
  (DVec<3> (coeff1, coeff2, coeff3))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymSourceIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }

  virtual string Name () const { return "HDivSym-Source"; }
};



template <>
class HDivSymSourceIntegrator<3> 
  : public T_BIntegrator<DiffOpIdHDivSym<3>, DVec<6>, FiniteElement>
{
public:
  HDivSymSourceIntegrator (CoefficientFunction * coeff1,
			   CoefficientFunction * coeff2,
			   CoefficientFunction * coeff3,
			   CoefficientFunction * coeff4,
			   CoefficientFunction * coeff5,
			   CoefficientFunction * coeff6)
    :  T_BIntegrator<DiffOpIdHDivSym<3>, DVec<6>, FiniteElement>
  (DVec<6> (coeff1, coeff2, coeff3, coeff4, coeff5, coeff6)) 
  { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymSourceIntegrator (coeffs[0], coeffs[1], coeffs[2],
					coeffs[3], coeffs[4], coeffs[5]);
  }

  virtual string Name () const { return "HDivSym-Source"; }
};




template <int D>
class HDivSymRobinIntegrator 
  : public T_BDBIntegrator<DiffOpBoundIdHDivSym<D>, DiagDMat<1>, FiniteElement>
{
public:
  HDivSymRobinIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpBoundIdHDivSym<D>, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymRobinIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Robin"; }
  virtual bool BoundaryForm () const { return 1; }
};



template <int D>
class HDivSymNeumannIntegrator 
  : public T_BIntegrator<DiffOpBoundIdHDivSym<D>, DVec<1>, FiniteElement>
{
public:
  HDivSymNeumannIntegrator (CoefficientFunction * coef)
    :  T_BIntegrator<DiffOpBoundIdHDivSym<D>, DVec<1>, FiniteElement>
  (DVec<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymNeumannIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Neumann"; }
  virtual bool BoundaryForm () const { return 1; }
};







template <int D>
class HDivSymDivDivIntegrator
  : public T_BDBIntegrator<DiffOpDivHDivSym<D>, DiagDMat<D>, FiniteElement>
{
public:
  HDivSymDivDivIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpDivHDivSym<D>, DiagDMat<D>, FiniteElement>
  (DiagDMat<D> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivDivIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-DivDiv"; }
};





class HDivSymDivIntegrator 
  : public T_BIntegrator<DiffOpDivHDivSym<2>, DVec<2>, FiniteElement>
{
public:
  HDivSymDivIntegrator (CoefficientFunction * coeff1,
			CoefficientFunction * coeff2)
    :  T_BIntegrator<DiffOpDivHDivSym<2>, DVec<2>, FiniteElement>
  (DVec<2> (coeff1, coeff2))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivIntegrator (coeffs[0], coeffs[1]);
  }

  virtual string Name () const { return "HDivSym-Div"; }
};




class HDivSymDivIntegrator3d 
  : public T_BIntegrator<DiffOpDivHDivSym<3>, DVec<3>, FiniteElement>
{
public:
  HDivSymDivIntegrator3d (CoefficientFunction * coeff1,
			CoefficientFunction * coeff2,
			CoefficientFunction * coeff3)
    :  T_BIntegrator<DiffOpDivHDivSym<3>, DVec<3>, FiniteElement>
  (DVec<3> (coeff1, coeff2, coeff3))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivIntegrator3d (coeffs[0], coeffs[1], coeffs[2]);
  }

  virtual string Name () const { return "HDivSym-Div"; }
};









template <int D>
class HDivSymDivDivDivDivIntegrator
  : public T_BDBIntegrator<DiffOpDivDivHDivSym<D>, DiagDMat<1>, FiniteElement>
{
public:
  HDivSymDivDivDivDivIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpDivDivHDivSym<D>, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivDivDivDivIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-DivDivDivDiv"; }
};









template <int D>
class DiffOpMixedMech { };


template <>
class DiffOpMixedMech<2> : public DiffOp<DiffOpMixedMech<2> >
{

  // (stress, div stress, u)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 7 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<2> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;

    Mat<2> jac = sip.GetJacobian();
    Mat<2> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());

    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    FlatMatrix<> div_sigma_bmat(2, nd_sigma, lh);
    DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

    for (int i = 0; i < nd_sigma; i++)
      {
	mat(0,i) = sigma_bmat(0,i);
	mat(1,i) = sigma_bmat(1,i);
	mat(2,i) = sigma_bmat(2,i);
	mat(3,i) = div_sigma_bmat(0,i);
	mat(4,i) = div_sigma_bmat(1,i);
      }

    FlatMatrix<> u_ref(nd_u, 2, lh);
    FlatMatrixFixWidth<2> u(nd_u, lh);

    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(5,i+nd_sigma) = u(i,0);
	mat(6,i+nd_sigma) = u(i,1);
      }
  }
};





template <>
class DiffOpMixedMech<3> : public DiffOp<DiffOpMixedMech<3> >
{

  // (stress, div stress, u)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 12 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<3> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
    const HCurlFiniteElementD<3> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;


    Mat<3> jac = sip.GetJacobian();
    Mat<3> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());



    FlatMatrix<> sigma_bmat(6, nd_sigma, lh);
    DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    FlatMatrix<> div_sigma_bmat(3, nd_sigma, lh);
    DiffOpDivHDivSym<3>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

    for (int i = 0; i < nd_sigma; i++)
      {
	for (int j = 0; j < 6; j++)
	  mat(j,i) = sigma_bmat(j,i);
	for (int j = 0; j < 3; j++)
	  mat(6+j,i) = div_sigma_bmat(j,i);
      }


    FlatMatrix<> u_ref(nd_u, 3, lh);
    FlatMatrixFixWidth<3> u(nd_u, lh);

    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(9,i+nd_sigma) = u(i,0);
	mat(10,i+nd_sigma) = u(i,1);
	mat(11,i+nd_sigma) = u(i,2);
      }
  }
};






















template <int D> class DMatMixedMech { };

template <>
class DMatMixedMech<2> : public DMatOp<DMatMixedMech<2> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
public:

  enum { DIM_DMAT = 7 };

  DMatMixedMech (CoefficientFunction * ae,
		 CoefficientFunction * anu) : e(ae), nu(anu) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);


    mat(0,0) = mat(1,1) = 1/ei * 1;
    mat(0,1) = mat(1,0) = -1/ei * nui;
    mat(2,2) = 2/ei * (1+nui);

    mat(3,5) = mat(5,3) = 1;
    mat(4,6) = mat(6,4) = 1;
    
    //    mat(5,5) = -1e-6;
    //    mat(6,6) = -1e-6;
  }  
};


template <>
class DMatMixedMech<3> : public DMatOp<DMatMixedMech<3> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
public:

  enum { DIM_DMAT = 12 };

  DMatMixedMech (CoefficientFunction * ae,
		 CoefficientFunction * anu) : e(ae), nu(anu) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);


    mat(0,0) = mat(1,1) = mat(2,2) = 1/ei;
    mat(0,1) = mat(1,0) = 
      mat(0,2) = mat(2,0) = 
      mat(2,1) = mat(1,2) = -nui / ei;

    mat(3,3) = 2 / ei * (1+nui);
    mat(4,4) = 2 / ei * (1+nui);
    mat(5,5) = 2 / ei * (1+nui);

    
    mat(6,9) = mat(9,6) = 1;
    mat(7,10) = mat(10,7) = 1;
    mat(8,11) = mat(11,8) = 1;
  }  
};








template <int D>
class MixedMechIntegrator 
  : public T_BDBIntegrator<DiffOpMixedMech<D>, DMatMixedMech<D>, FiniteElement>
{
public:
  ///
  MixedMechIntegrator (CoefficientFunction * ae,
		       CoefficientFunction * anu)
    :  T_BDBIntegrator<DiffOpMixedMech<D>, DMatMixedMech<D>, FiniteElement>
  (DMatMixedMech<D> (ae, anu))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechIntegrator (coeffs[0], coeffs[1]);
  }
  
  ///
  virtual string Name () const { return "MixedMech"; }
};




















template <int D>
class DiffOpMixedMechStab : public DiffOp<DiffOpMixedMechStab<D> >
{
public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = D*(D+1) };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<D> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<D>&> (cfel[0]);
    const HCurlFiniteElementD<D> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<D>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;

    Mat<D> jac = sip.GetJacobian();
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());

    FlatMatrix<> sigma_bmat(DIM_DMAT/2, nd_sigma, lh);
    DiffOpIdHDivSym<D>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < DIM_DMAT/2; j++)
	mat(j,i) = sigma_bmat(j,i);

    FlatMatrixFixWidth<D> u_ref(nd_u, lh);
    FlatMatrixFixWidth<D> ur(nd_u, lh);
    FlatMatrixFixWidth<D> ul(nd_u, lh);
    FlatMatrixFixWidth<D> dudx(nd_u, lh);
    FlatMatrixFixWidth<D> dudy(nd_u, lh);
    FlatMatrixFixWidth<D> dudz(nd_u, lh);
	
    for (int k = 0; k < D; k++)
      {
	IntegrationPoint ipr(sip.IP());
	IntegrationPoint ipl(sip.IP());
	ipr(k) += 1e-6;
	ipl(k) -= 1e-6;

	SpecificIntegrationPoint<D,D> sipr (ipr,sip.GetTransformation(), lh);
	SpecificIntegrationPoint<D,D> sipl (ipl,sip.GetTransformation(), lh);

	fel_u.CalcShape (ipr, u_ref);
	ur = u_ref * sipr.GetJacobianInverse();

	fel_u.CalcShape (ipl, u_ref);
	ul = u_ref * sipl.GetJacobianInverse();

	ur -= ul;
	ur /= (2e-6);

	switch (k)
	  {
	  case 0: dudx = ur; break;
	  case 1: dudy = ur; break;
	  case 2: dudz = ur; break;
	  }
      }


    if (D == 2)
      for (int i = 0; i < nd_u; i++)
	{
	  Mat<2,2> gradur, gradu;
	  gradur(0,0) = dudx(i,0);
	  gradur(0,1) = dudy(i,0);
	  gradur(1,0) = dudx(i,1);
	  gradur(1,1) = dudy(i,1);
	  
	  gradu = gradur * inv_jac;
	  mat(3,nd_sigma+i) = gradu(0,0);
	  mat(4,nd_sigma+i) = gradu(1,1);
	  mat(5,nd_sigma+i) = gradu(0,1)+gradu(1,0);
	}
    else
      for (int i = 0; i < nd_u; i++)
	{
	  Mat<3,3> gradur, gradu;
	  for (int j = 0; j < 3; j++)
	    {
	      gradur(j,0) = dudx(i,j);
	      gradur(j,1) = dudy(i,j);
	      gradur(j,2) = dudz(i,j);
	    }
	  gradu = gradur * inv_jac;

	  mat(6,nd_sigma+i) = gradu(0,0);
	  mat(7,nd_sigma+i) = gradu(1,1);
	  mat(8,nd_sigma+i) = gradu(2,2);
	  mat(9,nd_sigma+i) = gradu(0,1)+gradu(1,0);
	  mat(10,nd_sigma+i) = gradu(0,2)+gradu(2,0);
	  mat(11,nd_sigma+i) = gradu(1,2)+gradu(2,1);
	}

    // (*testout) << "bmat = " << endl << mat << endl;

    /*

    FlatMatrix<> u_refr(nd_u, 2, lh);
    FlatMatrix<> u_refl(nd_u, 2, lh);
    FlatMatrix<> dudx(nd_u, 2, lh);
    FlatMatrix<> dudy(nd_u, 2, lh);
	
    for (int k = 0; k < 2; k++)
      {
	IntegrationPoint ipr(sip.IP());
	IntegrationPoint ipl(sip.IP());
	ipr(k) += 1e-6;
	ipl(k) -= 1e-6;

	fel_u.CalcShape (ipr, u_refr);
	fel_u.CalcShape (ipl, u_refl);

	u_refr -= u_refl;
	u_refr /= (2e-6);

	if (k == 0)
	  dudx = u_refr;
	else
	  dudy = u_refr;
      }

    for (int i = 0; i < nd_u; i++)
      {
	Mat<2,2> gradur, gradu;
	gradur(0,0) = dudx(i,0);
	gradur(0,1) = dudy(i,0);
	gradur(1,0) = dudx(i,1);
	gradur(1,1) = dudy(i,1);
	
	gradu = Trans(inv_jac) * gradur * inv_jac;
	mat(3,nd_sigma+i) = gradu(0,0);
	mat(4,nd_sigma+i) = gradu(1,1);
	mat(5,nd_sigma+i) = gradu(0,1)+gradu(1,0);
      }
    */


    /*
    FlatMatrix<> u(nd_u, 2, lh);
    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(4,i+nd_sigma) = u(i,0);
	mat(5,i+nd_sigma) = u(i,0);
	mat(6,i+nd_sigma) = u(i,1);
      }
    */
  }
};


class MixedMechBFIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coef_e;
  CoefficientFunction * coef_nu;
  CoefficientFunction *coef_reg;
  
public:
  MixedMechBFIntegrator (CoefficientFunction * acoef_e, 
			   CoefficientFunction * acoef_nu, 
			   CoefficientFunction * acoef_reg) 
    : BilinearFormIntegrator()
  { 
    coef_e  = acoef_e;
    coef_nu = acoef_nu;
    coef_reg = acoef_reg;
  }

  virtual ~MixedMechBFIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();
      const int D = 2;

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // (\int sigma_nt q_t \ds)^2

    
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<2> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
 
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(fel_u.ElementType(), integration_order);
    
    void * heapp = lh.GetPointer();
    
    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    FlatMatrix<> div_sigma_bmat(2, nd_sigma, lh);

    FlatMatrix<> u_ref(nd_u, 2, lh);
    FlatMatrixFixWidth<2> u(nd_u, lh);

    Mat<7> mat;
    mat=0;
    for ( int i = 0; i<ir.GetNIP(); i++)
      {
	SpecificIntegrationPoint<D,D> sip(ir[i], eltrans, lh);

	Mat<2> jac = sip.GetJacobian();
	Mat<2> inv_jac = sip.GetJacobianInverse();
	double det = fabs (sip.GetJacobiDet());
	
	// sigma
	DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	// div(sigma)
	DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);
	// u
	fel_u.CalcShape (sip.IP(), u_ref);
	u = u_ref * inv_jac;

	double weight = ir[i].Weight() * det;
	mat = 0;

	double ei = coef_e->Evaluate (sip);
	double nui = coef_nu->Evaluate (sip);
	double regi = coef_reg->Evaluate (sip) * weight;

	mat(0,0) = mat(1,1) = weight/ei ;
	mat(0,1) = mat(1,0) = -weight/ei * nui;
	mat(2,2) = 2*weight/ei * (1+nui);
	
	mat(3,5) = mat(5,3) = weight;
	mat(4,6) = mat(6,4) = weight;
	
	for(int m=0; m<nd_sigma; m++)
	  for(int j=0; j<nd_sigma; j++)
	    for(int l=0; l<3; l++)
	      for(int k=0; k<3; k++)
		{
		  elmat(m,j) += sigma_bmat(l,m) * mat(l,k) * sigma_bmat(k,j);
		}
	for(int m=0; m<nd_sigma; m++)
	  for(int j=0; j<nd_u; j++)
	    for(int l=3; l<=4; l++)
	      for(int k=5; k<=6; k++)
		{
		  elmat(m,j+nd_sigma) += div_sigma_bmat(l-3,m) * mat(l,k) * u(j,k-5);
		  elmat(j+nd_sigma,m) += div_sigma_bmat(l-3,m) * mat(l,k) * u(j,k-5);
		}
	for(int m=0; m<nd_u; m++)
	  for(int j=0; j<nd_u; j++)
	    {
	      elmat(m+nd_sigma,j+nd_sigma) += regi*u(m,0)*u(j,0);

	      elmat(m+nd_sigma,j+nd_sigma) += regi*u(m,1)*u(j,1);
	  }
	
	//    mat(5,5) = -1e-6;
	//    mat(6,6) = -1e-6;
      }
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechBFIntegrator (coeffs[0], coeffs[1], coeffs[2]); 
  }

};





template <int D>
class DMatMixedMechStab { };

template <>
class DMatMixedMechStab<2> : public DMatOp<DMatMixedMechStab<2> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
  CoefficientFunction * alpha;
public:

  enum { DIM_DMAT = 6 };

  DMatMixedMechStab (CoefficientFunction * ae,
		     CoefficientFunction * anu,
		     CoefficientFunction * aalpha) : e(ae), nu(anu), alpha(aalpha) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);
    double alphai = alpha->Evaluate (sip);
    
    mat(0,0) = mat(1,1) = 1/ei * 1;
    mat(0,1) = mat(1,0) = -1/ei * nui;
    mat(2,2) = 2/ei * (1+nui);

    double fac = ei / (1-nui*nui);
    mat(3,3) = mat(4,4) = fac;
    mat(3,4) = mat(4,3) = fac*nui;
    mat(5,5) = fac*(1-nui)/2;

    for (int i = 0; i < 3; i++)
      {
	mat(i,3+i) = -1;
	mat(3+i,i) = -1;
      }

    mat *= -alphai;
  }  
};


template <>
class DMatMixedMechStab<3> : public DMatOp<DMatMixedMechStab<3> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
  CoefficientFunction * alpha;
public:

  enum { DIM_DMAT = 12 };

  DMatMixedMechStab (CoefficientFunction * ae,
		     CoefficientFunction * anu,
		     CoefficientFunction * aalpha) : e(ae), nu(anu), alpha(aalpha) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);
    double alphai = alpha->Evaluate (sip);
    
    mat(0,0) = mat(1,1) = mat(2,2) = 1/ei;
    mat(0,1) = mat(1,0) =
      mat(0,2) = mat(2,0) =
      mat(1,2) = mat(2,1) = -nui/ei;

    mat(3,3) = mat(4,4) = mat(5,5) = 2/ei * (1+nui);


    mat(6,6) = mat(7,7) = mat(8,8) = ei * (1-nui) / ( (1+nui)*(1-2*nui) );

    for (int i = 6; i < 9; i++)
      for (int j = 6; j < i; j++)
	mat(i,j) = mat(j,i) = ei * nui  / ( (1+nui)*(1-2*nui) );

    for (int i = 9; i < 12; i++)
      mat(i,i) = ei/ (2*(1+nui));

    for (int i = 0; i < 6; i++)
      {
	mat(i,6+i) = -1;
	mat(6+i,i) = -1;
      }

    mat *= -alphai;
  }  
};














template <int D>
class MixedMechStabIntegrator 
  : public T_BDBIntegrator<DiffOpMixedMechStab<D>, DMatMixedMechStab<D>, FiniteElement>
{
public:
  ///
  MixedMechStabIntegrator (CoefficientFunction * ae,
			   CoefficientFunction * anu,
			   CoefficientFunction * aalpha)
    :  T_BDBIntegrator<DiffOpMixedMechStab<D>, DMatMixedMechStab<D>, FiniteElement>
  (DMatMixedMechStab<D> (ae, anu, aalpha))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechStabIntegrator<D> (coeffs[0], coeffs[1], coeffs[2]);
  }
  
  ///
  virtual string Name () const { return "MixedMechStab"; }
};

































class MixedMechJumpsIntegrator : public BilinearFormIntegrator
{
public:
  MixedMechJumpsIntegrator () 
    : BilinearFormIntegrator()
  { ; }

  virtual ~MixedMechJumpsIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  /*
  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // \int sigma_nt u_t \ds

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();
    int nd = bfel.GetNDof();

    // FlatMatrix<> sigma_ref(nd_sigma, 3, lh);
    FlatMatrix<> u_ref(nd_u, 2, lh);

    // FlatMatrix<> sigma(nd_sigma, 3, lh);
    FlatMatrix<> u(nd_u, 2, lh);

    FlatMatrix<> mat_u(1, nd, lh);
    FlatMatrix<> mat_sigma(1, nd, lh);

    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    
    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_sigma.Order());


    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);


	    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	    for (int i = 0; i < nd_sigma; i++)	    
	      mat_sigma(0,i) = 
		normal(0) * tau(0) * sigma_bmat(0,i) +
		normal(1) * tau(1) * sigma_bmat(1,i) +
		(normal(0) * tau(1) + normal(1) * tau(0)) * sigma_bmat(2,i);

	    
	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;
	    
	    for (int i = 0; i < nd_u; i++)
	      mat_u(0,i+nd_sigma) = u(i,0)*tau(0) + u(i,1)*tau(1);
	    
	    mat_u *= -len * ir1d[l].Weight();
	    elmat += Trans(mat_u) * mat_sigma;
	    elmat += Trans(mat_sigma) * mat_u;
	  }
      }
  }
*/




  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // \int sigma_nt u_t \ds

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<2> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();
    int nd = bfel.GetNDof();

    FlatMatrix<> u_ref(nd_u, 2, lh);
    FlatMatrix<> u(nd_u, 2, lh);

    FlatVector<> mat_u(nd_u, lh);
    FlatVector<> mat_sigma(nd_sigma, lh);

    FlatMatrix<> mat_u_sigma(nd_u, nd_sigma, lh);
    
    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    
    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_sigma.Order());

    mat_u_sigma = 0.0;
    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	    for (int i = 0; i < nd_sigma; i++)	    
	      mat_sigma(i) = 
		normal(0) * tau(0) * sigma_bmat(0,i) +
		normal(1) * tau(1) * sigma_bmat(1,i) +
		(normal(0) * tau(1) + normal(1) * tau(0)) * sigma_bmat(2,i);

	    
	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;
	    
	    for (int i = 0; i < nd_u; i++)
	      mat_u(i) = u(i,0)*tau(0) + u(i,1)*tau(1);
	    
	    mat_u *= -len * ir1d[l].Weight();
	    mat_u_sigma += mat_u * Trans (mat_sigma);
	    lh.CleanUp (heapp);
	  }
      }

    for (int i = 0; i < nd_u; i++)
      for (int j = 0; j < nd_sigma; j++)
	{
	  elmat(nd_sigma+i, j) += mat_u_sigma(i,j);
	  elmat(j, nd_sigma+i) += mat_u_sigma(i,j);
	}
  }













  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechJumpsIntegrator (); 
  }

};













class MixedMechJumpsIntegrator3d : public BilinearFormIntegrator
{
public:
  MixedMechJumpsIntegrator3d () 
    : BilinearFormIntegrator()
  { ; }

  virtual ~MixedMechJumpsIntegrator3d () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    // \int sigma_nt u_t \ds

    int ndof = bfel.GetNDof();

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<3> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
    const HCurlFiniteElementD<3> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    FlatMatrixFixWidth<3> u_ref(nd_u, lh);
    FlatMatrixFixWidth<3> u(nd_u, lh);

    FlatMatrixFixHeight<3> mat_u(nd_u, lh);
    FlatMatrixFixHeight<3> mat_sigma(nd_sigma, lh);
    FlatMatrix<> hmat(nd_sigma, nd_u, lh);

    FlatMatrixFixHeight<6> sigma_bmat(nd_sigma, lh);

    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const FACE * faces = ElementTopology::GetFaces (fel_sigma.ElementType());
    int nf = ElementTopology::GetNFaces (fel_sigma.ElementType());
    hmat = 0;
    
    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );

	const IntegrationRule & ir2d =
	  GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, 8);

	Vec<3> pstart, vx, vy;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }

	Vec<3> taux, tauy, normal;

	for (int l = 0; l < ir2d.GetNIP(); l++)
	  {
	    Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
	    Mat<3> jac = sip.GetJacobian();
	    Mat<3> inv_jac = sip.GetJacobianInverse();
	    // double det = fabs (sip.GetJacobiDet());

	    taux = jac * vx;
	    tauy = jac * vy;

	    normal(0) = taux(1) * tauy(2) - taux(2) * tauy(1);
	    normal(1) = taux(2) * tauy(0) - taux(0) * tauy(2);
	    normal(2) = taux(0) * tauy(1) - taux(1) * tauy(0);

	    
	    DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	    // sigma n
	    for (int i = 0; i < nd_sigma; i++)	    
	      {
		mat_sigma(0,i) = normal(0) * sigma_bmat(0,i) + normal(1) * sigma_bmat(3,i) + normal(2) * sigma_bmat(4,i);
		mat_sigma(1,i) = normal(0) * sigma_bmat(3,i) + normal(1) * sigma_bmat(1,i) + normal(2) * sigma_bmat(5,i);
		mat_sigma(2,i) = normal(0) * sigma_bmat(4,i) + normal(1) * sigma_bmat(5,i) + normal(2) * sigma_bmat(2,i);
	      }

	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;
	    
	    double inv_area2 = 1.0 / L2Norm2(normal);
	    for (int i = 0; i < nd_u; i++)
	      {
		// n * u
		double un = 0;
		for (int j = 0; j < 3; j++)
		  un += normal(j) * u(i,j);
		un *= inv_area2;

		for (int j = 0; j < 3; j++)
		  mat_u(j,i) = u(i,j) - normal(j) * un;
	      }

	    mat_u *= ir2d[l].Weight();
	    hmat += Trans(mat_sigma) * mat_u;
	  }
      }

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;
    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_u; j++)
	elmat(j+nd_sigma,i) = elmat(i,j+nd_sigma) = hmat(i,j);
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechJumpsIntegrator3d (); 
  }

};

































/*

class MixedMechStabIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coef;
public:
  MixedMechStabIntegrator (CoefficientFunction * acoef) 
    : BilinearFormIntegrator()
  { 
    coef  = acoef;
  }

  virtual ~MixedMechStabIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // (\int sigma_nt q_t \ds)^2

    
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();


    FlatMatrix<> sigma_ref(nd_sigma, 3, lh);
    FlatMatrix<> sigma(nd_sigma, 3, lh);

    FlatVector<> vec(nd_sigma+nd_u, lh);

    vec = 0;

    const POINT3D * points = ElementTopology::GetVertices (ET_TRIG);
    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    
    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 7);

    double cval = 1;
    double len = 1;

    for (int k = 0; k < 3; k++)
      {
	vec = 0;
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    // IntegrationPoint ip( 0.5 * (ps(0)+pe(0)), 0.5 * (ps(1)+pe(1)));
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);

	    cval = coef -> Evaluate (sip);

	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);
	    
	    fel_sigma.CalcShape (ip, sigma_ref);
	    for (int i = 0; i < nd_sigma; i++)
	      {
		Mat<2> sig_ref, sig;
		sig_ref(0,0) = sigma_ref(i, 0);
		sig_ref(1,1) = sigma_ref(i, 1);
		sig_ref(0,1) = sig_ref(1,0) = sigma_ref(i, 2);
		
		sig = (jac) * sig_ref * Trans(jac);
		sig *= (1.0 / sqr (det));
		sig_n = sig * normal;
		vec(i) += ir1d[l].Weight() * (2*ir1d[l](0)-1) 
		  * InnerProduct (tau, sig_n) ;
	      }
	  }
	elmat += (cval) * vec * Trans(vec);
      }
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechStabIntegrator (coeffs[0]); 
  }

};



*/


























  

class NumProcMixedTest : public NumProc
{
public: 
  ///
  NumProcMixedTest (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedTest (pde, flags);
  }

  virtual void Do(LocalHeap & lh)
  {
    FlatMatrix<> elmat1, elmat2;
    
    MixedMechIntegrator<2> mmi(new ConstantCoefficientFunction(1),
			       new ConstantCoefficientFunction(0.2));

    MixedMechJumpsIntegrator mmj;

    
    const FiniteElement & fel = pde.GetFESpace("v")->GetFE (0, lh);

    ElementTransformation eltrans;
    ma.GetElementTransformation (0, eltrans, lh);


    mmi.AssembleElementMatrix (fel, eltrans, elmat1, lh);
    mmj.AssembleElementMatrix (fel, eltrans, elmat2, lh);
    
    (*testout) << "mmi.mat = " << endl << elmat1 << endl;
    (*testout) << "mmj.mat = " << endl << elmat2 << endl;
    (*testout) << "diff = " << endl << (elmat1-elmat2) << endl;
    (*testout) << "sum  = " << endl << (elmat1+elmat2) << endl;
  }
};
  



}

namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hdivsymnn", HDivSymHighOrderFESpace::Create);
    GetFESpaceClasses().AddFESpace ("mixedmech", FESpaceMixedMech::Create);


    GetIntegrators().AddBFIntegrator ("mixedmech", 2, 2,
				      MixedMechIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("mixedmech", 3, 2,
				      MixedMechIntegrator<3>::Create);
    GetIntegrators().AddBFIntegrator ("mixedmechbf", 2, 3,
				      MixedMechBFIntegrator::Create);

    GetIntegrators().AddBFIntegrator ("mixedmechjumps", 2, 0,
				      MixedMechJumpsIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedmechjumps", 3, 0,
				      MixedMechJumpsIntegrator3d::Create);


    GetIntegrators().AddBFIntegrator ("mixedmechstab", 2, 3,
				      MixedMechStabIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("mixedmechstab", 3, 3,
				      MixedMechStabIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("masshdivsym", 2, 1,
				      HDivSymMassIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("masshdivsym", 3, 1,
				      HDivSymMassIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("robinhdivsym", 2, 1,
				      HDivSymRobinIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("robinhdivsym", 3, 1,
				      HDivSymRobinIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("divdivhdivsym", 2, 1,
				      HDivSymDivDivIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("divdivhdivsym", 3, 1,
				      HDivSymDivDivIntegrator<3>::Create);


    GetIntegrators().AddBFIntegrator ("divdivdivdivhdivsym", 2, 1,
				      HDivSymDivDivDivDivIntegrator<2>::Create);


    GetIntegrators().AddLFIntegrator ("sourcehdivsym", 2, 3,
				      HDivSymSourceIntegrator<2>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhdivsym", 2, 1,
				      HDivSymNeumannIntegrator<2>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhdivsym", 3, 1,
				      HDivSymNeumannIntegrator<3>::Create);
    GetIntegrators().AddLFIntegrator ("sourcehdivsym", 3, 6,
				      HDivSymSourceIntegrator<3>::Create);
    GetIntegrators().AddLFIntegrator ("divhdivsym", 2, 2,
				      HDivSymDivIntegrator::Create);
    GetIntegrators().AddLFIntegrator ("divhdivsym", 3, 3,
				      HDivSymDivIntegrator3d::Create);


    GetNumProcs().AddNumProc ("mixedtest", NumProcMixedTest::Create);
  }
  
  Init init;
}
