/*

Mixed elasticity -  hybrid fespace

*/



#include <solve.hpp>
#include <comp.hpp>
#include <fem.hpp>

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif

namespace ngcomp
{

  using namespace ngfem;
  using namespace ngsolve;
  using namespace ngcomp;

#ifdef PARALLEL
  using namespace ngparallel;
#endif







  /********************************************************

  Compound FESpace for mixed mechanics - hybridized version

  ******************************************************/

  class FESpaceMixedMechHybrid : public CompoundFESpace
  {
  public:
    FESpaceMixedMechHybrid (const MeshAccess & ama, 		   
			    const ARRAY<const FESpace*> & aspaces,
			    const Flags & aflags)
      : CompoundFESpace (ama, aspaces, aflags)
    {
      ;
    }

    virtual ~FESpaceMixedMechHybrid () { ; }

    static FESpace *   Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);

      cout << "Create EAS FE Space" << endl;
    
      int order = int (flags.GetNumFlag ("order", 1.0));

      int orderz = int(flags.GetNumFlag ("orderz", -1));

      int facetorder = int ( flags.GetNumFlag("facetorder", max2(order-1,1) ) );
      facetorder = max2 ( facetorder, int(flags.GetNumFlag("lagrangeorder", facetorder) ) );

      int sigmaorderz = int ( flags.GetNumFlag("sigmaorderz", orderz) );
      sigmaorderz = max2 ( sigmaorderz, int(flags.GetNumFlag("lagrangeorderz", sigmaorderz) ) );

      int curlorderz = int( flags.GetNumFlag("curlorderz", orderz) );

      Flags sigmaflags;
      sigmaflags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
      sigmaflags.SetFlag("orderz", flags.GetNumFlag ("orderz", 1.0));
      sigmaflags.SetFlag ("facetorder", flags.GetNumFlag ("facetorder", facetorder ));
      sigmaflags.SetFlag ("noncontinuous", 1.0);
      if (flags.GetDefineFlag ("eliminate_internal")) sigmaflags.SetFlag ("eliminate_internal");
      sigmaflags.SetFlag("orderz", sigmaorderz);
      if ( flags.GetDefineFlag("sequence") ) sigmaflags.SetFlag("sequence");

      spaces[0] = new HDivSymHighOrderFESpace (ma, sigmaflags);
    
      Flags uflags;
      int curlorder = int (flags.GetNumFlag ("curlorder", order));
      uflags.SetFlag ("order", curlorder);
      uflags.SetFlag ("orderz", curlorderz);
      uflags.SetFlag ("orderinner", flags.GetNumFlag ("curlorderinner", curlorder));
      uflags.SetFlag ("orderface", flags.GetNumFlag ("curlorderface", curlorder));
      uflags.SetFlag ("orderedge", flags.GetNumFlag ("curlorderedge", curlorder));
      uflags.SetFlag ("nograds", flags.GetDefineFlag ( "nograds") );
      uflags.SetFlag ("alternative", flags.GetNumFlag ( "alternative", 0) );

      ARRAY<double> maxz(ma.GetNDomains());
      maxz = 1;
      uflags.SetFlag ("dom_order_min_z", maxz);
      uflags.SetFlag ("dom_order_max_z", maxz);
      uflags.SetFlag ("variableorder");

      if (flags.GetDefineFlag ("eliminate_internal")) uflags.SetFlag ("eliminate_internal");
      spaces[1] = new HCurlHighOrderFESpaceA (ma, uflags);
    
      Flags pflags;
      pflags.SetFlag("order", facetorder);
      pflags.SetFlag("orderz", flags.GetNumFlag ("lagrangeorderz", sigmaorderz));
      spaces[2] = new FESpaceHybrid (ma, pflags);

      FESpaceMixedMechHybrid * fes = new FESpaceMixedMechHybrid (ma, spaces, flags);

#ifdef PARALLEL
      fes->ResetParallelDofs();
#endif
      return fes;
    }
    
    Table<int> *  CreateSmoothingBlocks (const Flags & precflags) const;

    
    virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const;


    virtual string GetClassName () const 
    { 
      return "Hybrid-MixedMech-FESpace"; 
    }



  };





  ///***************
  // 
  //  Smoothing blocks:
  //      0 ..... jacobi
  //      1 ..... div-free p, AFW - edges - faces - diag
  //      2 ..... div-free p, loAFW - edges - faces - diag
  //      3 ..... div-free p, diag
  //      4 ..... lo div-free p, AFW - edges - faces - diag
  //      5 ..... lo div-free p, loAFW - edges - faces - diag
  //      6 ..... lo div-free p, diag
  //      7 ..... edge-face-blocks
  //
  //*****************



  Table<int> * FESpaceMixedMechHybrid :: CreateSmoothingBlocks (const Flags & precflags) const
  {
  
    if ( ! eliminate_internal ) 
      {
	cerr << "Set flag \"-eliminate_internal\" for the MixedMechHybridFESpace!" << endl;
	return 0;
      }
    const HCurlHighOrderFESpaceA & uspace = 
      dynamic_cast<const HCurlHighOrderFESpaceA & > (*spaces[1]);
    const HDivSymHighOrderFESpace & sigmaspace =dynamic_cast<const HDivSymHighOrderFESpace & > (*spaces[0]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*spaces[2]);
  
    bool highnlo = true;
    if ( precflags.GetDefineFlag("p0") )
      highnlo = 0;

    int i, j, k, first,ii;
    int ncnt; 
    int ni = ma.GetNE();//nel;
  
    int dim = ma.GetDimension();
  
  
    int SmoothingType = int(precflags.GetNumFlag("blocktype",0)); 
  
  
    ARRAY<int> vnums,elnums; 
    ARRAY<int> orient; 
    ARRAY<int> ednums, fanums;
    ARRAY<int> edges(3);
  
    int porder = spaces[2]->GetOrder();
    int nd_sigma = spaces[0]->GetNDof(), nd_v = spaces[1]->GetNDof(), nd_p = spaces[2]->GetNDof();
    int ndof = nd_sigma + nd_v + nd_p;
    int nd_u = nd_v;
  
    int ned = ma.GetNEdges();
    int nfa = ma.GetNFaces();
    int nnv = ma.GetNV();
    int nel = ma.GetNE();
  
    BitArray usedpdof(nd_p);
    usedpdof.Clear();

    // compute used number of degrees of freedom
    int nd_sigma_used = nd_sigma, nd_u_used = nd_u, nd_p_used = nd_p;
    if(!ni) 
      {
	nd_sigma_used = 0;
	nd_u_used = nd_u;
	for(i=0; i<nel; i++)
	  {
	    nd_u_used -= uspace.GetFirstCellDof(i+1) - uspace.GetFirstCellDof(i);
	  }
      }

    int ndof_used = nd_sigma_used + nd_u_used + nd_p_used;
    cout << "SmoothingType " << SmoothingType << endl; 
    switch(SmoothingType) 
      {
      case 0: 
	cout << " simple jacobi" << endl; 
	ncnt = ndof; //_used;
	break;  
      case 1:
      case 4:
	if ( SmoothingType == 1 )
	  cout << "div-free(hybrid) - AFW - E(Hcurl) - F(Hcurl) - I" << endl;
	else
	  cout << "lo div-free(hybrid) - AFW - E(Hcurl) - F(Hcurl) - I" << endl;

	if( dim == 2 )
	  {
	    // 1) Hcurl :   nnv + ned + nfa
	    // 2) Hybrid :  vertex blocks -- nnv
	    ncnt = nnv + nnv + ned ;
	    if ( SmoothingType == 4 )
	      ncnt += ned;
	  }
	else
	  {
	    // 1) Hcurl :   nnv + ned + nfa
	    // 2) Hybrid :  edge blocks -- ned
	    ncnt = nnv + nfa + ned + ned;
	    if ( SmoothingType == 4 ) 
	      ncnt += nfa;
	  }
	break;
      case 2:
      case 5:
	if ( SmoothingType == 2 )
	  cout << "div-free(hybrid) - loAFW - E(Hcurl) - F(Hcurl) - I" << endl;
	else
	  cout << "lo div-free(hybrid) - loAFW - E(Hcurl) - F(Hcurl) - I" << endl;
	if( dim == 2 )
	  {
	    // 1) Hcurl :   nnv + ned
	    // 2) Hybrid :  vertex blocks -- nnv
	    ncnt = nnv + nnv + ned ;
	    if ( SmoothingType == 5 )
	      ncnt += ned;

	  }
	else
	  {
	    // 1) Hcurl :   nnv + ned + nfa
	    // 2) Hybrid :  edge blocks -- ned
	    ncnt = nnv + nfa + 2*ned;
	    if ( SmoothingType == 5 )
	      ncnt += nfa;
	  }
	break;
      case 3:
      case 6:
	if ( SmoothingType == 3 )
	  cout << "div-free(hybrid) - diag" << endl;
	else
	  cout << "lo div-free(hybrid) - diag" << endl;
	if( dim == 2 )
	  {
	    // 1) Hcurl :   ndof_u
	    // 2) Hybrid :  vertex blocks -- nnv
	    ncnt = nnv + nd_u_used;
	    if ( SmoothingType == 6 )
	      ncnt += ned;

	  }
	else
	  {
	    // 1) Hcurl :   ndof_u
	    // 2) Hybrid :  edge blocks -- ned
	    ncnt = nd_u_used + ned;
	    if ( SmoothingType == 6 )
	      ncnt += nfa;
	  }
	break;

      case 7:
	if ( dim == 2 )
	  ncnt = ned + ned;
	else
	  ncnt = ned + nfa + nfa;
	break;

      case 8:
	if ( dim == 2 )
	  ncnt = ned;
	else
	  ncnt = ned + nfa;
	break;

      }


    ARRAY<int> cnt(ncnt); 
    cnt = 0;
    ii=0; 

    int offset = 0;



    int nlo;
    if ( highnlo )
      {
	nlo = (dim == 2) ? 2 : 3;
      }
    else
      nlo = 1;

    switch (SmoothingType)
      {
	//      0 ..... jacobi
      case 0:   // diagonal
	for(i=0; i<ndof; i++ ) //_used; i++)
	  cnt[i] = 1;
	break;

	//      1 ..... div-free p, AFW - edges - faces - diag
      case 1:   
	if( dim == 2 )
	  {
	    // AFW vertex patches
	    for( i=0; i<ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums ( i, pn1, pn2);
		  cnt[offset + pn1] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		  cnt[offset + pn2] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		
		}

	    offset += nnv;
	    // div-free vertex patches
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		  cnt[offset + pn2] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		}

	  }

	else
	  {
	    // AFW blocks
	    for (i = 0; i < ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);
		  cnt[offset + pn1] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		  cnt[offset + pn2] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		}
	    offset += nnv;
	    
	    /*
	    
	    for( i=0; i<ned; i++)
	    {
	    cnt[offset + i] =   uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
	    }
	    
	    offset += ned;
	    */

	    // face blocks
	    for( i=0; i<nfa; i++)
	      if( uspace.GetFineFace (i) )
		cnt[offset + i] = uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i);

	    offset += nfa;
	    
	    // div-free
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += 
		      pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) ;
		}
	  }
	break;

      case 2:
	//      2 ..... div-free p, loAFW - edges - faces - diag
	if( dim == 2 )
	  {
	    // loAFW vertex patches
	    for( i=0; i<ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums ( i, pn1, pn2);
		  cnt[offset + pn1] ++;
		  cnt[offset + pn2] ++;
		}

	    offset += nnv;

	    // edge blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 
		    1 + uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}

	    offset += ned;

	    // div-free vertex patches
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		  cnt[offset + pn2] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		}

	  }

	else
	  {
	    // loAFW blocks
	    for (i = 0; i < ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);
		  cnt[offset + pn1] ++;
		  cnt[offset + pn2] ++;
		}
	    offset += nnv;
	    
	    // edge blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 1 + uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}
	    
	    offset += ned;

	    // face blocks
	    for( i=0; i<nfa; i++)
	      if( uspace.GetFineFace (i) )
		cnt[offset + i] = uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i);

	    offset += nfa;
	    
	    // div-free
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += 
		      pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) ;
		}
	  }

	break;

	//      3 ..... div-free p, diag
      case 3:
	if( dim == 2 )
	  {
	    // diagonal prec for H(curl)
	    for( i=0; i<nd_u_used; i++)
	      cnt[i] ++;

	    offset += nd_u_used;

	    // div-free vertex patches
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		  cnt[offset + pn2] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i);
		}

	  }

	else
	  {
	    // diagonal prec for H(curl)
	    for( i=0; i<nd_u_used; i++)
	      cnt[i] ++;

	    offset += nd_u_used;
	    
	    // div-free
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += 
		      pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) ;
		}
	  }

	break;

      case 4:  
	//      4 ..... lo div-free p, AFW - edges - faces - diag
	if( dim == 2 )
	  {
	    // AFW vertex patches
	    for( i=0; i<ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums ( i, pn1, pn2);
		  cnt[offset + pn1] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		  cnt[offset + pn2] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		
		}

	    offset += nnv;
	    // div-free vertex patches
	    // 2 low-order edge dofs
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += nlo;
		  cnt[offset + pn2] += nlo;

		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - nlo;

	  }

	else
	  {
	    // AFW blocks
	    for (i = 0; i < ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);
		  cnt[offset + pn1] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		  cnt[offset + pn2] += uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) + 1;
		}
	    offset += nnv;
	    
	    // face blocks
	    for( i=0; i<nfa; i++)
	      if( uspace.GetFineFace (i) )
		cnt[offset + i] = uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i);

	    offset += nfa;
	    
	    // div-free
	    // 3 low order face dofs
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += nlo; 
		}

	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - 
		  pspace.GetFirstFaceDof(i) - nlo;

	  }
	break;

      case 5:
	//      5 ..... lo div-free p, loAFW - edges - faces - diag
	if( dim == 2 )
	  {
	    // loAFW vertex patches
	    for( i=0; i<ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums ( i, pn1, pn2);
		  cnt[offset + pn1] ++;
		  cnt[offset + pn2] ++;
		}

	    offset += nnv;
	    for( i=0; i<ned; i++)
	      {
		cnt[offset + i] = 
		  1 + uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
	      }
	    offset += ned;
	    // div-free vertex patches
	    // 2 low order edge dofs
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += nlo;
		  cnt[offset + pn2] += nlo;
		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - nlo;

	  }

	else
	  {
	    // loAFW blocks
	    for (i = 0; i < ned; i++)
	      if(uspace.GetFineEdge(i))
		{
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);
		  cnt[offset + pn1] ++;
		  cnt[offset + pn2] ++;
		}
	    offset += nnv;
	    
	    // edge blocks
	    for( i=0; i<ned; i++)
	      {
		cnt[offset + i] = 1 + uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
	      }
	    
	    offset += ned;

	    // face blocks
	    for( i=0; i<nfa; i++)
	      if( uspace.GetFineFace (i) )
		cnt[offset + i] = uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i);

	    offset += nfa;
	    
	    // div-free
	    // 3 low order dofs
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += nlo; 
		}

	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - 
		  pspace.GetFirstFaceDof(i) - nlo;

	  }

	break;

      case 6:   
	//      6 ..... lo div-free p, diag
	if( dim == 2 )
	  {
	    // diagonal prec for H(curl)
	    for( i=0; i<nd_u_used; i++)
	      cnt[i] ++;

	    offset += nd_u_used;

	    // div-free vertex patches
	    // 2 low order edge dofs
	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{
		  int pn1, pn2;
		  ma.GetEdgePNums( i, pn1, pn2);
		  cnt[offset + pn1] += nlo;
		  cnt[offset + pn2] += nlo;
		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - nlo;

	  }

	else
	  {
	    // diagonal prec for H(curl)
	    for( i=0; i<nd_u_used; i++)
	      cnt[i] ++;

	    offset += nd_u_used;
	    
	    // div-free
	    // 3 low order edge dofs
	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    cnt[offset +  edges[ed]] += nlo; 
		}

	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		cnt[offset + i] += pspace.GetFirstFaceDof(i+1) - 
		  pspace.GetFirstFaceDof(i) - nlo;
	  }

	break;

      case 7:
	if( dim == 2 )
	  {
	    // edge blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}

	    offset += ned;
	    int nlo = 2;
	    for( i=0; i<ned; i++)
	      if ( pspace.GetFineFace(i) )
		{
		  cnt[offset + i] = 
		    pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - nlo; 
		}

	  }
	else
	  {
	    // loAFW blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}

	    offset += ned;

	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i) ; 
		}

	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      if ( pspace.GetFineFace(i) )
		{
		  cnt[offset + i] = 
		    pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - 3; 
		}


	  }
	break;

      case 8:
	if( dim == 2 )
	  {
	    // edge blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}

	    int nlo = 2;
	    for( i=0; i<ned; i++)
	      if ( pspace.GetFineFace(i) )
		{
		  cnt[offset + i] += 
		    pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - nlo; 
		}

	  }
	else
	  {
	    // loAFW blocks
	    for( i=0; i<ned; i++)
	      if ( uspace.GetFineEdge(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstEdgeDof(i+1) - uspace.GetFirstEdgeDof(i) ; 
		}

	    offset += ned;

	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace(i) )
		{
		  cnt[offset + i] = 
		    uspace.GetFirstFaceDof(i+1) - uspace.GetFirstFaceDof(i) ; 
		}

	    for( i=0; i<nfa; i++)
	      if ( pspace.GetFineFace(i) )
		{
		  cnt[offset + i] += 
		    pspace.GetFirstFaceDof(i+1) - pspace.GetFirstFaceDof(i) - 3; 
		}
	  }
	break;

      }


    Table<int> & table = *new Table<int> (cnt); 
  
    ii = 0; 
    cnt = 0;
  
    offset =0;
    BitArray used_pdofs(nd_p);
  
  
    switch(SmoothingType) 
      {
	//      0 ..... jacobi
      case 0:
	for(i=0; i<ndof; i++)
	  table[i][cnt[i]] = i;
	// 	for(i=0; i<nd_u_used; i++)
	// 	  table[i][cnt[i]] = i + nd_sigma;
	// 	for(i=0; i<nd_p_used; i++)
	// 	  table[nd_u_used + i][cnt[nd_u_used + i]] = i + nd_sigma + nd_u;
	break;

	//      1 ..... div-free p, AFW - edges - faces - diag
      case 1:

	if ( dim == 2 )
	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		    
		    first = uspace.GetFirstEdgeDof(i);
		    int last = uspace.GetFirstEdgeDof(i+1);
		    for( int l=first; l<last; l++)
		      {
			table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma;
			table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma;
		      }
		  }
	      }

	    offset += nnv;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  int last = pspace.GetFirstFaceDof(i+1);
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}

	    (*testout) << table << endl;
	  }

	else // 3d

	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;

		    first = uspace.GetFirstEdgeDof(i);
		    int last = uspace.GetFirstEdgeDof(i+1);
		    for( int l=first; l<last; l++)
		      {
			table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma;
			table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma;
		      }
		  }

	      }


	    offset += nnv;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}
	    
	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges(i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      int last = pspace.GetFirstFaceDof(i+1);
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}
	    
	  }
	break;

	//      2 ..... div-free p, loAFW - edges - faces - diag
      case 2:
	if ( dim == 2 )
	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		  }
	      }

	    offset += nnv;

	    for ( i = 0; i < ned; i++)
	      {
		table[offset + i][cnt[offset+i]++] = i + nd_sigma;
		int first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		for ( int l=first; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma;
	      }
	    offset += ned;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  int last = pspace.GetFirstFaceDof(i+1);
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}

	    (*testout) << table << endl;
	  }

	else // 3d

	  {
	    offset = 0;
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		  }

	      }

	    offset += nnv;

	    for( i=0; i<ned; i++)
	      {
		first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() + i;
		for(j=first; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() +  j ; 
		
	      }
	    offset += ned;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}
	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      int last = pspace.GetFirstFaceDof(i+1);
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}
	  }
	break;

	//      3 ..... div-free p, diag
      case 3:
	offset = 0; 
	if ( dim == 2 )
	  {
	    for (i = 0; i < nd_u_used; i++)
	      {
		table[offset + i][cnt[offset + i]++] = i + nd_sigma;
	      }
	    
	    offset += nd_u_used;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  int last = pspace.GetFirstFaceDof(i+1);
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}
	  }

	else // 3d

	  {
	    for (i = 0; i < nd_u_used; i++)
	      {
		table[offset + i][cnt[offset + i]++] = i + nd_sigma;
	      }
	    
	    offset += nd_u_used;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      int last = pspace.GetFirstFaceDof(i+1);
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}
	    
	  }
	break;

	//      4 ..... lo div-free p, AFW - edges - faces - diag
      case 4:
	offset = 0;
	if ( dim == 2 )
	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		    
		    first = uspace.GetFirstEdgeDof(i);
		    int last = uspace.GetFirstEdgeDof(i+1);
		    for( int l=first; l<last; l++)
		      {
			table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma;
			table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma;
		      }
		  }
	      }

	    offset += nnv;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  // 2 low-order edge dofs
		  int last = first + nlo;
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}
	  }

	else // 3d

	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;

		    first = uspace.GetFirstEdgeDof(i);
		    int last = uspace.GetFirstEdgeDof(i+1);
		    for( int l=first; l<last; l++)
		      {
			table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma;
			table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma;
		      }
		  }

	      }


	    offset += nnv;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}
	    
	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      // 3 low-order face dofs
		      int last = first+3;
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}
	    
	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}
	  }
	break;

	//      5 ..... lo div-free p, loAFW - edges - faces - diag
      case 5:
	if ( dim == 2 )
	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		  }
	      }

	    offset += nnv;

	    for ( i = 0; i < ned; i++)
	      {
		table[offset + i][cnt[offset+i]++] = i + nd_sigma;
		int first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		for ( int l=first; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma;
	      }
	    offset += ned;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  int last = first+2;//pspace.GetFirstFaceDof(i+1);
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}

	  }

	else // 3d

	  {
	    for (i = 0; i < ned; i++)
	      {
		int pn1, pn2;
		ma.GetEdgePNums (i,pn1,pn2);	      
		if( uspace.GetFineEdge(i) )
		  {
		    table[offset + pn1][cnt[offset + pn1]++] = i + nd_sigma;
		    table[offset + pn2][cnt[offset + pn2]++] = i + nd_sigma;
		  }

	      }


	    offset += nnv;

	    for( i=0; i<ned; i++)
	      {
		first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() + i;
		for(j=first; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() +  j ; 
		
	      }

	    offset += ned;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}
	    
	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      int last = first + 3;
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}

	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}
	    
	  }
	break;

	//      6 ..... lo div-free p, diag
      case 6:
	offset = 0; 
	if ( dim == 2 )
	  {
	    for (i = 0; i < nd_u_used; i++)
	      {
		table[offset + i][cnt[offset + i]++] = i + nd_sigma;
	      }
	    
	    offset += nd_u_used;

	    for( i=0; i<ned; i++)
	      if( pspace.GetFineFace (i) )
		{		
		  int pn1, pn2;
		  ma.GetEdgePNums (i,pn1,pn2);	      
		  first = pspace.GetFirstFaceDof(i);
		  int last = first + 2;
		  for( int l=first; l<last; l++)
		    {
		      table[offset + pn1][cnt[offset + pn1]++] = l + nd_sigma + nd_u;
		      table[offset + pn2][cnt[offset + pn2]++] = l + nd_sigma + nd_u;
		    }
		}

	    offset += nnv;
	    for ( i = 0; i < ned; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}
	  }

	else // 3d

	  {
	    for (i = 0; i < nd_u_used; i++)
	      {
		table[offset + i][cnt[offset + i]++] = i + nd_sigma;
	      }
	    
	    offset += nd_u_used;

	    for( i=0; i<nfa; i++)
	      if( pspace.GetFineFace (i) )
		{
		  // div-free functions per edge: 
		
		  ma.GetFaceEdges (i, edges);
		  for(int ed=0; ed<edges.Size(); ed++)
		    {
		      first = pspace.GetFirstFaceDof(i);
		      int last = first + 3;
		      for( j=first; j<last; j++)
			{
			  table[offset + edges[ed] ][cnt[offset + edges[ed] ]] = j + nd_sigma + nd_u;
			  cnt[offset + edges[ed] ]++;
			}
		    }
		
		}

	    offset += ned;
	    for ( i = 0; i < nfa; i++ )
	      if ( pspace.GetFineFace(i) )
		{
		  first = pspace.GetFirstFaceDof(i) + nlo;
		  int last = pspace.GetFirstFaceDof(i+1);
		  for ( int l = first; l < last; l++ )
		    table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
		}
	    
	  }
	break;

      case 7:
	if ( dim == 2 )
	  {
	    for ( i = 0; i < ned; i++)
	      {
		if ( !uspace.GetFineEdge(i) ) continue;
		// table[offset + i][cnt[offset+i]++] = i + nd_sigma;
		int first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		for ( int l=first; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma;
	      }
	    offset += ned;

	    for ( i = 0; i < ned; i++)
	      {
		if ( !pspace.GetFineFace(i) ) continue;
		int first = pspace.GetFirstFaceDof(i);
		int last = pspace.GetFirstFaceDof(i+1);
		for ( int l=first+2; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
	      }
	  }
	else // 3d

	  {
	    offset = 0;

	    for( i=0; i<ned; i++)
	      {
		if ( !uspace.GetFineEdge(i) ) continue;

		first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		// table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() + i;
		for(j=first; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() +  j ; 
		
	      }
	    offset += ned;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}
	    offset += nfa;

	    for( i=0; i<nfa; i++)
	      {
		if ( !pspace.GetFineFace(i) ) continue;
		first = pspace.GetFirstFaceDof(i);
		int last = pspace.GetFirstFaceDof(i+1);
		for(j=first+3; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = nd_sigma + nd_u +  j ; 
		
	      }

	  }
	break;
 
      case 8:
	if ( dim == 2 )
	  {
	    for ( i = 0; i < ned; i++)
	      {
		// table[offset + i][cnt[offset+i]++] = i + nd_sigma;
		if ( !uspace.GetFineEdge(i) ) continue;
		int first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		for ( int l=first; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma;
	      }

	    for ( i = 0; i < ned; i++)
	      {
		if ( !pspace.GetFineFace(i) ) continue;
		int first = pspace.GetFirstFaceDof(i);
		int last = pspace.GetFirstFaceDof(i+1);
		for ( int l=first+2; l < last; l++ )
		  table[offset + i][cnt[offset+i]++] = l + nd_sigma + nd_u;
	      }
	  }
	else // 3d

	  {
	    offset = 0;

	    for( i=0; i<ned; i++)
	      {
		if ( !uspace.GetFineEdge(i) ) continue;

		first = uspace.GetFirstEdgeDof(i);
		int last = uspace.GetFirstEdgeDof(i+1);
		// table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() + i;
		for(j=first; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = sigmaspace.GetNDof() +  j ; 
		
	      }
	    offset += ned;
	
	    for( i=0; i<nfa; i++)
	      if ( uspace.GetFineFace (i) )
		{
		  first = uspace.GetFirstFaceDof(i);
		  int last = uspace.GetFirstFaceDof(i+1);
		  for(j=first; j<last; j++)
		    table[offset + i][cnt[offset+i]++] = sigmaspace.GetNDof() + j;
		}

	    for( i=0; i<nfa; i++)
	      {
		if ( !pspace.GetFineFace(i) ) continue;
		first = pspace.GetFirstFaceDof(i);
		int last = pspace.GetFirstFaceDof(i+1);
		for(j=first+3; j<last; j++)
		  table[offset + i][cnt[offset + i] ++] = nd_sigma + nd_u +  j ; 
		
	      }

	  }
	break;
      }

    return & table;

  }







  // ******************** //
  // Direct Solver Clusters
  // 0) H(curl) edges + hybrid faces
  // 1) all
  // 2) none
  // 3) low-order dofs
  // 4) H(curl) edges + low-order hybrid faces
  // 5) H(curl) edges (order-1) + low-order hybrid faces


  ARRAY<int> * FESpaceMixedMechHybrid :: CreateDirectSolverClusters (const Flags & precflags) const
  {
    int clustertype = int(precflags.GetNumFlag("ds_cluster",-1)); 
    if ( clustertype == -1 ) return 0;

    cout << " DirectSolverCluster Clustertype " << clustertype << endl; 
  
    if ( ! eliminate_internal ) 
      {
	cerr << "Set flag \"-eliminate_internal\" for the MixedMechHybridFESpace!" << endl;
	return 0;
      }

    int nv = ma.GetNV();
    int nd = GetNDof();
    int ne = ma.GetNE();
    int ned = ma.GetNEdges();
  
    int dim = ma.GetDimension();

    int ni = ne;
    if (precflags.GetDefineFlag("eliminate_internal") ) ni = 0; 
    
    int nlo;
    bool highnlo = !precflags.GetDefineFlag("p0");

    ARRAY<int> vnums,elnums; 
    ARRAY<int> orient; 
  
    ARRAY<int> edges(3);
        
    int porder = spaces[2]->GetOrder();

    int nd_sigma = spaces[0]->GetNDof(), nd_v = spaces[1]->GetNDof(), nd_p = spaces[2]->GetNDof();
    int ndof = nd_sigma + nd_v + nd_p;
    int nd_u = nd_v;

  
    int nfa = ma.GetNFaces();
    int nnv = ma.GetNV();
    int nel = ma.GetNE();
    int nfa_p = (dim == 2 ) ? ned : nfa;

    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*spaces[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*spaces[0]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*spaces[2]);
    ARRAY<int> ednums, fnums, pnums;
  
    int i, j, k;
        
    ARRAY<int> & clusters = *new ARRAY<int> (nd_u + nd_sigma + nd_p);
    clusters = 0;


    switch (clustertype)
      {

	// 0) H(curl) edges + hybrid faces
      case 0:

	for( i=0; i<ned; i++ )
	  if( uspace.GetFineEdge (i) )
	    {
	      clusters[i + nd_sigma] = 1;
	      int first = uspace.GetFirstEdgeDof(i) + nd_sigma;
	      int next = uspace.GetFirstEdgeDof(i+1) + nd_sigma;
	      for (k = first; k < next; k++)
		{
		  clusters[k] = 1;      
		}
	    
	    }


	for ( i=0; i<nfa_p; i++)
	  {
	   
	    int first = pspace.GetFirstFaceDof(i) + nd_u + nd_sigma;
	    int next = pspace.GetFirstFaceDof(i+1) + nd_u + nd_sigma;
	    for (k = first; k < next; k++)
	      {
		clusters[k] = 1;      
	      }
	  
	  }


	break;

	// 1) all
      case 1:
	clusters = 1;

	break;

	// 2) none
      case 2:
	clusters = 0;

	break;

	// 3) low-order dofs
      case 3:

	for( i=0; i<ned; i++ )
	  {
	    if ( !uspace.GetFineEdge(i) ) continue;
	    clusters[i + nd_sigma] = 1;

	  }

	// nur low order sigma dofs
	for ( i=0; i<nfa_p; i++)
	  {
	   
	    if ( !pspace.GetFineFace(i) ) continue;
	    int first = pspace.GetFirstFaceDof(i) + nd_u + nd_sigma;
	    int nlo = 2;
	    if ( dim == 3 )
	      {
		ma.GetFaceEdges(i, ednums);
		nlo = ednums.Size();
	      }

	    if ( !highnlo ) nlo = 1;

	    for (k = first; k < first + nlo; k++)
	      {
		clusters[k] = 1;      
	      }
	  
	  }

	break;

	// 4) H(curl) edges + low-order hybrid faces
      case 4:
	clusters = 0;

	for( i=0; i<ned; i++ )
	  if( uspace.GetFineEdge (i) )
	    {
	      clusters[i + nd_sigma] = 1;
	      int first = uspace.GetFirstEdgeDof(i) + nd_sigma;
	      int next = uspace.GetFirstEdgeDof(i+1) + nd_sigma;
	      for (k = first; k < next; k++)
		{
		  clusters[k] = 1;      
		}
	    
	    }


	for ( i=0; i<nfa_p; i++)
	  {
	   
	    int first = pspace.GetFirstFaceDof(i) + nd_u + nd_sigma;
	    // 2d: 2 low order edge dofs
	    int next = first + 2;
	    // 3d: 3 low order face dofs
	    if ( dim == 3 )
	      next ++;

	    for (k = first; k < next; k++)
	      {
		clusters[k] = 1;      
	      }
	  
	  }


	break;
	// 5) H(curl) edges (order-1) + low-order hybrid faces
      case 5:
	clusters = 0;

	for( i=0; i<ned; i++ )
	  if( uspace.GetFineEdge (i) )
	    {
	      clusters[i + nd_sigma] = 1;
	      int first = uspace.GetFirstEdgeDof(i) + nd_sigma;
	      int next = uspace.GetFirstEdgeDof(i+1) + nd_sigma - 1;
	      for (k = first; k < next; k++)
		{
		  clusters[k] = 1;      
		}
	    
	    }


	for ( i=0; i<nfa_p; i++)
	  {
	   
	    int first = pspace.GetFirstFaceDof(i) + nd_u + nd_sigma;
	    // 2d: 2 low order edge dofs
	    int next = first + 2;
	    // 3d: 3 low order face dofs
	    if ( dim == 3 )
	      next ++;

	    for (k = first; k < next; k++)
	      {
		clusters[k] = 1;      
	      }
	  
	  }


	break;
      }

    //       cout << clusters << endl;

    return &clusters;

  }










  /* ******************** Integrator ************************** */



  class MixedMechHIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction * coef_e;
    CoefficientFunction * coef_nu;
    CoefficientFunction *coef_reg;
    CoefficientFunction * coef_div;
  
    enum { DIM = 1 };
    enum { DIM_SPACE = 2 };
    enum { DIM_ELEMENT = 2 };
    enum { DIM_DMAT = 8 };
    enum { DIFFORDER = 0 };

  public:
    MixedMechHIntegrator (CoefficientFunction * acoef_e, 
			  CoefficientFunction * acoef_nu, 
			  CoefficientFunction * acoef_reg,
			  CoefficientFunction * acoef_div)
      : BilinearFormIntegrator()
    { 
      coef_e  = acoef_e;
      coef_nu = acoef_nu;
      coef_reg = acoef_reg;
      coef_div = acoef_div;
    }

    virtual ~MixedMechHIntegrator () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }

    virtual int DimFlux () const { return DIM_DMAT; }

    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();
      const int D = 2;

      elmat.AssignMemory (ndof, ndof, lh);
      elmat = 0;

      // (\int sigma_nt q_t \ds)^2

    
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const HDivSymFiniteElement<2> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
      const HCurlFiniteElementD<2> & fel_u = 
	dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement &> (cfel[2]);

      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();
      int nd_p = fel_p.GetNDof();

      // sigma is highest order
      int intorder = fel_sigma.Order() * 2;

      const IntegrationRule & ir = 
	GetIntegrationRules().SelectIntegrationRule(fel_u.ElementType(), intorder);
    
      const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
      const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
      int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;
   
    
      FlatMatrixFixHeight<3> sigma_bmat( nd_sigma, lh);
      FlatMatrixFixHeight<2> div_sigma_bmat( nd_sigma, lh);
      FlatVector<> divdiv_sigma_bmat( nd_sigma, lh);

      FlatMatrix<> u_ref(nd_u, 2, lh);
      FlatMatrixFixWidth<2> u(nd_u, lh);

 
      FlatVector<> mat_u(nd_u, lh);
      FlatVector<> mat_sigmantau(nd_sigma, lh);

      FlatMatrix<> mat_u_sigma(nd_u, nd_sigma, lh);
    
    
      mat_u = 0;
      mat_sigmantau = 0;


      Mat<7> mat;
      mat=0;
      double h;
      for ( int i = 0; i<ir.GetNIP(); i++)
	{
	  SpecificIntegrationPoint<D,D> sip(ir[i], eltrans, lh);

	  Mat<2> jac = sip.GetJacobian();
	  Mat<2> inv_jac = sip.GetJacobianInverse();
	  double det =  fabs(sip.GetJacobiDet());

	  Vec<2> ps, pe, tau_ref, sig_n, tau, normal, tau2, tau3, tau4, p3, p4;
	  int es = 0; 
	  int ee = 1;
	
	  ps(0) = points[es][0];
	  ps(1) = points[es][1];
	  pe(0) = points[ee][0];
	  pe(1) = points[ee][1];

	  tau_ref = pe-ps;
	  tau = jac * tau_ref;


	  if ( i == 0 )
	    {
	      p3(0) = points[2][0];
	      p3(1) = points[2][1];
	      if( ned == 4)
		{
		  p4(0) = points[3][0];
		  p4(1) = points[3][1];
		}
	    

	      tau2 = jac * (p3 - pe);
	      if ( ned == 4)
		{
		  tau3 = jac * (p4 - p3);
		  tau4 = jac * (ps - p4);
		}
	      else
		tau3 = jac * (ps - p3);
	    }

	  // sigma
	  DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	  // div(sigma)
	  DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);
	  //	*testout << "mat " << endl << mat << endl;
	  DiffOpDivDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, divdiv_sigma_bmat, lh);


	  // u
	  fel_u.CalcShape (sip.IP(), u_ref);
	  u = u_ref * inv_jac;

	  double weight = ir[i].Weight() * det;
	  mat = 0;

	  double ei = coef_e->Evaluate (sip);
	  double nui = coef_nu->Evaluate (sip);
	  double regi = coef_reg->Evaluate (sip)*weight;
	  double coeff_divdiv = coef_div -> Evaluate (sip) * weight;

	  if ( i == 0 ) 
	    {
	      h = min2(L2Norm(tau), L2Norm(tau2));
	    
	      h = min2 ( h, L2Norm(tau3));
	      if ( ned == 4)
		h = min2 ( h, L2Norm (tau4) );
	    }

	  mat(0,0) = mat(1,1) = (1+nui)*(1-nui) * weight/ei ;
	  mat(0,1) = mat(1,0) = - (1+nui) * weight * nui/ ei;
	  mat(2,2) = 2.0*weight * (1+nui) / ei;
	
	  mat(3,5) = mat(5,3) = weight;
	  mat(4,6) = mat(6,4) = weight;
	
	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_sigma; j++)
	      for(int l=0; l<3; l++)
		for(int k=0; k<3; k++)
		  {
		    elmat(m,j) += sigma_bmat(l,m) * mat(l,k) * sigma_bmat(k,j);
		  }


	  // old stabilization ######################################
	  
	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_sigma; j++)
	      {
		elmat(m,j) += 
		  (div_sigma_bmat(0,m)* div_sigma_bmat(0,j) 
		   + div_sigma_bmat(1,m) *  div_sigma_bmat(1,j))
		  * coeff_divdiv * h * h;
	      }
	  
	  // new stabilization ######################################
	  /*
	    for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_sigma; j++)
	    {
	    elmat(m,j) += 
	    (divdiv_sigma_bmat(m)* divdiv_sigma_bmat(j))
	    * coeff_divdiv * h * h * h * h;
	    }
	  */
	  
	  // end stabilization ######################################

	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_u; j++)
	      for(int l=3; l<=4; l++)
		for(int k=5; k<=6; k++)
		  {
		    elmat(m,j+nd_sigma) += div_sigma_bmat(l-3,m) * mat(l,k) * u(j,k-5);
		    elmat(j+nd_sigma,m) += div_sigma_bmat(l-3,m) * mat(l,k) * u(j,k-5);
		  }

	  for(int m=0; m<nd_u; m++)
	    for(int j=0; j<nd_u; j++)
	      {
		elmat(m+nd_sigma,j+nd_sigma) += regi*u(m,0)*u(j,0);

		elmat(m+nd_sigma,j+nd_sigma) += regi*u(m,1)*u(j,1);
	      }
	
	  //    mat(5,5) = -1e-6;
	  //    mat(6,6) = -1e-6;
	}
   
      // hybridization terms
      FlatVector<> p(nd_p, lh);
    
      p = 0;


      int nfdof = fel_sigma.GetNDof();//ned*(fel_sigma.Order()+1);

      FlatMatrix<> mat_p_sigma(nd_p, nfdof, lh);
      FlatVector<> mat_sigmann(nd_sigma, lh);
      mat_sigmann = 0.0;
      mat_p_sigma = 0.0;
      mat_u_sigma = 0.0;

      intorder = 2 * fel_p.Order();

      const IntegrationRule & ir1d =
	GetIntegrationRules().SelectIntegrationRule (ET_SEGM, intorder);

      mat_p_sigma = 0.0;
      for (int k = 0; k < ned; k++)
	{
	  Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	  int es = edges[k][0]; 
	  int ee = edges[k][1];
	
	  //     int et = 3 - ee - es; // third edge

	  //if (vnums[es] > vnums[ee])  swap (es, ee);

	  ps(0) = points[es][0];
	  ps(1) = points[es][1];
	  pe(0) = points[ee][0];
	  pe(1) = points[ee][1];

	  tau_ref = pe-ps;
	  void * heapp = lh.GetPointer();


	  ARRAY<int> fdofs(0);
	  fel_p.GetFDofNrs(k, fdofs);

	  for (int l = 0; l < ir1d.GetNIP(); l++)
	    {
	      Vec<2> p2d = ps + ir1d[l](0)* (pe-ps);
	      IntegrationPoint ip( p2d(0), p2d(1) );
	    
	      SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	      Mat<2> jac = sip.GetJacobian();
	      Mat<2> inv_jac = sip.GetJacobianInverse();
	      double det = fabs (sip.GetJacobiDet());

	      tau = jac * tau_ref;
	      double len = L2Norm(tau);
	    
	      tau *= (1.0/len);
	      normal(0) = tau(1);
	      normal(1) = -tau(0);

	      DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	      DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

	      //for (int i = 0; i < nd_sigma; i++)
	      for(int i=0; i<nfdof; i++)	    
		mat_sigmann(i) = 
		  normal(0) * normal(0) * sigma_bmat(0,i) +
		  normal(1) * normal(1) * sigma_bmat(1,i) +
		  (normal(0) * normal(1) + normal(1) * normal(0)) * sigma_bmat(2,i);

	      for (int i = 0; i < nd_sigma; i++)	    
		mat_sigmantau(i) = 
		  normal(0) * tau(0) * sigma_bmat(0,i) +
		  normal(1) * tau(1) * sigma_bmat(1,i) +
		  (normal(0) * tau(1) + normal(1) * tau(0)) * sigma_bmat(2,i);

	      // scaling for lagrangeparameters
	      fel_p.CalcShape ( ip, p);
 	      p *= sqr(len);

	      double weight = ir1d[l].Weight();
	  
	      //mat_p_sigma += p * Trans (mat_sigma);
	      for(int i = 0; i < fdofs.Size(); i++)
		for(int j = 0; j < nfdof; j++)
		  mat_p_sigma(fdofs[i],j) += p(fdofs[i]) * mat_sigmann(j) * weight;
	      // 	    cout << mat_sigma << endl << "p    " << p << endl;;
	      // 	    cout << "mult " << mat_p_sigma << endl;
	   
	      for (int i=0; i<fdofs.Size(); i++)
		for(int j=0; j<fdofs.Size(); j++)
		  {
		    double regi = -coef_reg->Evaluate(sip) ;
		    elmat(fdofs[i]+nd_u + nd_sigma , fdofs[j]+nd_u+nd_sigma) += p(fdofs[i])*p(fdofs[j])*regi * len * weight;
		  }
		    
	      fel_u.CalcShape (ip, u_ref);
	      u = u_ref * inv_jac;
	    
	      for (int i = 0; i < nd_u; i++)
		mat_u(i) = u(i,0)*tau(0) + u(i,1)*tau(1);
	    
	      mat_u *= -len * ir1d[l].Weight();
	      mat_u_sigma += mat_u * Trans (mat_sigmantau);

	  // new stabilization ######################################

	      /*
	      double coeff_divdiv = coef_div -> Evaluate (sip) * weight * len;
	      for(int m=0; m<nd_sigma; m++)
		for(int j=0; j<nd_sigma; j++)
		  {
		    elmat(m,j) += 
		      (div_sigma_bmat(0,m) * normal(0) + div_sigma_bmat(1,m) * normal(1) ) * 
		      (div_sigma_bmat(0,j) * normal(0) + div_sigma_bmat(1,j) * normal(1) ) 
		      * coeff_divdiv * h * h * h;
		  }
	      */
	  // end stabilization ######################################
	    }
	}

      for (int i = 0; i < nd_p; i++)
	for (int j = 0; j < nfdof; j++)
	  {
	    elmat(nd_sigma+nd_u+i, j) += mat_p_sigma(i,j);
	    elmat(j, nd_sigma+nd_u+i) += mat_p_sigma(i,j);
	  
	  }

      for (int i = 0; i < nd_u; i++)
	for (int j = 0; j < nd_sigma; j++)
	  {
	    elmat(nd_sigma+i, j) += mat_u_sigma(i,j);
	    elmat(j, nd_sigma+i) += mat_u_sigma(i,j);
	  }

    
    }


    virtual void
    CalcFlux (const FiniteElement & bfel,
	      const ElementTransformation & eltrans,
	      const IntegrationPoint & ip,
	      const FlatVector<double> & elx, 
	      FlatVector<double> & flux,
	      bool applyd,
	      LocalHeap & lh) const

    {
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const HDivSymFiniteElement<2> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
      const HCurlFiniteElementD<2> & fel_u = 
	dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement &> (cfel[2]);

      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();
      int nd_p = fel_p.GetNDof();
      int ndof = cfel.GetNDof();

      const int D = 2;
    
      flux=0; 
	  
      SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
 
      FlatMatrixFixHeight<3> sigma_bmat( nd_sigma, lh);
      FlatMatrixFixHeight<2> div_sigma_bmat( nd_sigma, lh);

      FlatMatrix<> u_ref(nd_u, 2, lh);
      FlatMatrixFixWidth<2> u(nd_u, lh);
      FlatVector<> p(nd_p, lh);

      DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
      // div(sigma)
      DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);
    
      Mat<D> inv_jac = sip.GetJacobianInverse();
      // u
      fel_u.CalcShape (sip.IP(), u_ref);
      u = u_ref * inv_jac;
    
      fel_p.CalcShape ( sip.IP(), p);

      FlatMatrixFixHeight<3> sigma ( 1, lh );
      FlatMatrixFixHeight<2> div_sigma (1, lh), disp(1, lh);

      for( int i=0; i<nd_sigma; i++)
	{
	  for( int j=0; j<3; j++)
	    sigma(j,0) = sigma_bmat(j,i) * elx(i);
	  for( int j=0; j<2; j++)
	    div_sigma(j,0) = div_sigma_bmat(j,i) * elx(i);
	}
      for( int i=0; i<nd_u; i++)
	for( int j=0; j<2; j++)
	  disp(j,0) = u(j,i) * elx(i);


      double pp = 0;
      for( int i=0; i<nd_p; i++)
	pp += p(i) * elx(nd_sigma + nd_u + i);

      if (D == 2)
	{

	  flux (0) = sigma(0,0);
	  flux (1) = sigma(1,0);
	  flux (2) = sigma(2,0);
	  flux (6) = div_sigma(0,0);
	  flux (7) = div_sigma(1,0);
	  flux (3) = disp(0,0);
	  flux (4) = disp(1,0);
	  flux (5) = pp;
	}
      
    }




    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHIntegrator (coeffs[0], coeffs[1], coeffs[2], coeffs[3]); 
    }

  };



  /*******************
complete hybrid integrator 3d
  *****************/


  class MixedMechHIntegrator3d : public BilinearFormIntegrator
  {
    CoefficientFunction * coef_e;
    CoefficientFunction * coef_nu;
    CoefficientFunction *coef_reg;
    CoefficientFunction * coef_div;
  public:
    MixedMechHIntegrator3d (CoefficientFunction * acoef_e, 
			    CoefficientFunction * acoef_nu, 
			    CoefficientFunction * acoef_reg,
			    CoefficientFunction * acoef_div) 
      : BilinearFormIntegrator()
    { 
      coef_e  = acoef_e;
      coef_nu = acoef_nu;
      coef_reg = acoef_reg;
      coef_div = acoef_div;
      
   
    }

    virtual ~MixedMechHIntegrator3d () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }


    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {


      int ndof = bfel.GetNDof();
      const int D = 3;
      elmat.AssignMemory (ndof, ndof, lh);

      elmat = 0;

      // (\int sigma_nt q_t \ds)^2


    
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);

   
      const HDivSymFiniteElement<3> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
      const HCurlFiniteElementD<3> & fel_u = 
	dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement &> (cfel[2]);
   
      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();
      int nd_p = fel_p.GetNDof();
      int nfdof = fel_p.GetNDof() ;


      int intorder = 2 * fel_sigma.Order() + 6;

      const IntegrationRule & ir = 
	GetIntegrationRules().SelectIntegrationRule(fel_u.ElementType(), intorder);

      void * heapp = lh.GetPointer();
    
      FlatMatrixFixHeight<3> div_sigma_bmat(nd_sigma, lh);
      FlatVector<> divdiv_sigma_bmat(nd_sigma, lh);

      FlatMatrixFixHeight<3> testintegral(nd_u, lh);
      testintegral = 0;
 
      FlatVector<> mat_sigmann(nd_sigma, lh);
      FlatVector<> mat_un ( nd_u, lh);
      FlatMatrix<> mat_u_sigma(nd_u, nd_sigma, lh);
    
      FlatMatrixFixWidth<3> u_ref(nd_u, lh);
      FlatMatrixFixWidth<3> u(nd_u, lh);

      FlatMatrixFixHeight<3> mat_u(nd_u, lh);
      FlatMatrixFixHeight<3> mat_u_nc(nd_u, lh);
      FlatMatrixFixHeight<3> mat_sigma(nd_sigma, lh);
      FlatMatrix<> hmat(nd_sigma, nd_u, lh);
      FlatMatrix<> hmat_nc(nd_sigma, nd_u, lh);

      FlatMatrixFixHeight<6> sigma_bmat(nd_sigma, lh);
      FlatMatrix<> mat_p_sigma(nd_sigma, nd_p, lh);

      FlatVector<> p(nd_p, lh);


 

      const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
      const FACE * faces = ElementTopology::GetFaces (fel_sigma.ElementType());
      int nf = ElementTopology::GetNFaces (fel_sigma.ElementType());
   
      Vec<3> ps, pe, tau_ref, tau;
      int es = 0; 
      int ee = 1;
    
      ps(0) = points[es][0];
      ps(1) = points[es][1];
      ps(2) = points[es][2];
      pe(0) = points[ee][0];
      pe(1) = points[ee][1];
      pe(2) = points[ee][2];
 
      mat_u = 0;

      double regi;
      Mat<12> mat;
      mat=0;


    
      for ( int lip = 0; lip<ir.GetNIP(); lip++)
	{
	  SpecificIntegrationPoint<D,D> sip(ir[lip], eltrans, lh);

	  Mat<3> jac = sip.GetJacobian();
	  Mat<3> inv_jac = sip.GetJacobianInverse();
	  double det = fabs (sip.GetJacobiDet());

	  tau_ref = pe-ps;
	  tau = jac * tau_ref;

	  double h = L2Norm(tau);

	  // sigma
	  DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	  // 	  if ( lip == 0 ) *testout << "****** sigma bmat " << endl << sigma_bmat << endl << endl ;
	  // div(sigma)
	  DiffOpDivHDivSym<3>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);


	  DiffOpDivDivHDivSym<3>::GenerateMatrix (fel_sigma, sip, divdiv_sigma_bmat, lh);
	  // u
	  fel_u.CalcShape (sip.IP(), u_ref);
	  u = u_ref * inv_jac;

	  double weight = ir[lip].Weight() * det;
	  mat = 0;

	  double ei = coef_e->Evaluate (sip);
	  double nui = coef_nu->Evaluate (sip);
	  regi = coef_reg->Evaluate (sip) ;
	  double coeff_div = coef_div->Evaluate (sip) ;

	  mat(0,0) = mat(1,1) = mat(2,2) = 1.0/ei;
	  mat(0,1) = mat(1,0) = 
	    mat(0,2) = mat(2,0) = 
	    mat(2,1) = mat(1,2) = -nui / ei;
	
	  mat(3,3) = 2.0 / ei * (1+nui);
	  mat(4,4) = 2.0 / ei * (1+nui);
	  mat(5,5) = 2.0 / ei * (1+nui);
	
	
	  mat(6,9) = mat(9,6) = 1;
	  mat(7,10) = mat(10,7) = 1;
	  mat(8,11) = mat(11,8) = 1;


	  // 	for (int i = 6; i < 9; i++)
	  // 	  for (int j = 6; j < i; j++)
	  // 	    mat(i,j) = mat(j,i) = ei * nui  / ( (1+nui)*(1-2*nui) );
	
	  // 	for (int i = 9; i < 12; i++)
	  // 	  mat(i,i) = ei/ (2*(1+nui));
	
	  // 	for (int i = 0; i < 6; i++)
	  // 	  {
	  // 	    mat(i,6+i) = -1;
	  // 	    mat(6+i,i) = -1;
	  // 	  }
	
	  mat *= weight;


	  // A sigma : tau	
	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_sigma; j++)
	      {
		for(int l=0; l<3; l++)
		  for(int k=0; k<3; k++)
		    {
		      elmat(m,j) += sigma_bmat(l,m) * mat(l,k) * sigma_bmat(k,j);
		    }

		for ( int l = 3; l < 6; l++ )
		  elmat(m,j) += sigma_bmat(l,m) * mat(l,l) * sigma_bmat(l,j);
	      }


	  // old stabilisation ###############################

	  // h^2  div sigma . div tau 
	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_sigma; j++)
	      for(int l=0; l<3; l++)
		{
		  elmat(m,j) += div_sigma_bmat(l,m) * coeff_div  * div_sigma_bmat(l,j) * weight * h * h;
		}

//           // new stabilization ############################
// 	  for(int m=0; m<nd_sigma; m++)
// 	    for(int j=0; j<nd_sigma; j++)
// 	      {
// 		elmat(m,j) += 
// 		  (divdiv_sigma_bmat(m)* divdiv_sigma_bmat(j))
// 		  * coeff_div * weight * h * h * h * h;
// 	      }

          // end stabilization ############################

	  // div sigma . u
// 	  (6, 9), (7, 10), (8, 11)
	  for(int m=0; m<nd_sigma; m++)
	    for(int j=0; j<nd_u; j++)
	      for(int l=6; l<=8; l++)
// 		for(int k=9; k<=11; k++)
		  {
		    int k = l+3;
		    elmat(m,j+nd_sigma) += div_sigma_bmat(l-6,m) * mat(l,k) * u(j,k-9);
		    elmat(j+nd_sigma,m) += div_sigma_bmat(l-6,m) * mat(l,k) * u(j,k-9);
		  }

		
	  // regularization:         reg u . u
	  for(int m=0; m<nd_u; m++)
	    for(int j=0; j<nd_u; j++)
	      {
		elmat(m+nd_sigma,j+nd_sigma) -= regi*u(m,0)*u(j,0)*weight;

		elmat(m+nd_sigma,j+nd_sigma) -= regi*u(m,1)*u(j,1)*weight;
	      }

      
	
	
	  //    mat(5,5) = -1e-6;
	  //    mat(6,6) = -1e-6;


	
// 	  // int_omega (Ainv * eps(u)) : eps(v)
// 	  const int D = 3;
// 	  FlatMatrixFixHeight<6> epsmat(nd_u, lh), aepsmat(nd_u, lh);
	
// 	  DiffOpEpsHCurlA<3>::GenerateMatrix (fel_u, sip, epsmat, lh);



	  // nonconforming 
	  /*
	
	  for (int i = 0; i < nd_u; i++)
 	  {
	    
	  double c, dd, e;
	  c = ei / (1.0 + nui);
	  dd = ei * nui / ( (1.0+nui) * ( 1.0+2*nui) );
	  e = ei * 2.0 / (1.0 + nui);
	  aepsmat(0,i) = c* epsmat(0,i)
	  + dd * (epsmat(0,i)+epsmat(1,i)+epsmat(2,i));
	  aepsmat(1,i) = c * epsmat(1,i) 
	  + dd * (epsmat(0,i)+epsmat(1,i)+epsmat(2,i));
	  aepsmat(2,i) = c * epsmat(2,i) 
	  + dd * (epsmat(0,i)+epsmat(1,i)+epsmat(2,i));
	  aepsmat(3,i) = epsmat(3,i) * e * 2;
	  aepsmat(4,i) = epsmat(4,i) * e * 2;
	  aepsmat(5,i) = epsmat(5,i) * e * 2;
	    
 	  }
	
	  //       // 2nd Piola-Kirchhoff (Braess, p251)
	  // 	       stress = (e/(1+nu)) * strain 
	  // 	 			+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();
	
	  for(int m=6; m<nd_u; m++)
 	  for(int j=6; j<nd_u; j++)
	  for( int d=0; d<6; d++)
	  {
	  elmat(m+nd_sigma,j+nd_sigma) += aepsmat(d,m) * epsmat(d,j) * weight;
	  }
	  */
	
	}
    
    
    
    
    
    
    
    



      // hybridization terms, jump terms


      mat_u = 0;
      mat_u_nc = 0;
      mat_sigma = 0;
      p = 0;
      mat_p_sigma = 0;

      hmat = 0;
      hmat_nc = 0;
   
      int order_p = fel_p.Order();

      //*testout << " *** new el" << endl;
      for (int k = 0; k < nf; k++)
	{
	  bool trig = ( faces[k][3] == -1 );
	  
	  intorder = 2 * fel_sigma.Order() + 6;

	  const IntegrationRule & ir2d =
	    GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, intorder);
    
	  Vec<3> pstart, vx, vy;
	  Vec<3> s, s2;
    
	  int vv = 6 - faces[k][0] - faces[k][1] - faces[k][2];


	  for (int j = 0; j < 3; j++)
	    {
	      pstart(j) = points[faces[k][0]][j];
	      vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	      vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];

	      s(j) = points[vv][j];
	    }

	  Vec<3> taux, tauy, normal;


	  ARRAY<int> fdofs(0), fdofs_sigma(0);
	  fel_p.GetFDofNrs(k, fdofs);
	  fel_sigma.GetFDofNrs(k, fdofs_sigma);
	  int nd_sigma_face = fdofs_sigma.Size();

	  for (int l = 0; l < ir2d.GetNIP(); l++)
	    {
	      Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	      IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	      SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
	      Mat<3> jac = sip.GetJacobian();
	      Mat<3> inv_jac = sip.GetJacobianInverse();

	      taux = jac * vx;
	      tauy = jac * vy;

	      normal(0) = taux(1) * tauy(2) - taux(2) * tauy(1);
	      normal(1) = taux(2) * tauy(0) - taux(0) * tauy(2);
	      normal(2) = taux(0) * tauy(1) - taux(1) * tauy(0);

	      //  	    normal *= -1.0;


	      double inv_area2 = 1.0 / L2Norm2(normal);
	      double area = 1.0 / sqrt(inv_area2);
	      double h = sip.GetJacobiDet() / area;
	      double fac = 1.0; // trig ? 0.5 : 1.0;

	      DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	      DiffOpDivHDivSym<3>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

	      fel_p.CalcShape(ip,p);

	      // sigma n
	      for (int i = 0; i < nd_sigma; i++)	    
		{
		  mat_sigma(0,i) = normal(0) * sigma_bmat(0,i) + normal(1) * sigma_bmat(3,i) + normal(2) * sigma_bmat(4,i);
		  mat_sigma(1,i) = normal(0) * sigma_bmat(3,i) + normal(1) * sigma_bmat(1,i) + normal(2) * sigma_bmat(5,i);
		  mat_sigma(2,i) = normal(0) * sigma_bmat(4,i) + normal(1) * sigma_bmat(5,i) + normal(2) * sigma_bmat(2,i);
// 		}

// 	      // sigma_nn
// 	      for ( int ii = 0; ii < nd_sigma_face; ii++ )
// 		{
// 		  int i = fdofs_sigma[ii];
		  mat_sigmann(i) = normal(0) * mat_sigma(0,i)  
		    + normal(1) * mat_sigma(1,i)  
		    + normal(2) * mat_sigma(2,i);
		}



	      // 	    for ( int i = 0; i < nd_sigma; i++)
	      // 	      for( int j = 0; j < 3; j++ )
	      // 	      {
	      // 		mat_sigma(j,i) -= inv_area2 * mat_sigmann(i) * normal(j);
	      // 	      }

	      fel_u.CalcShape (ip, u_ref);
	      u = u_ref * inv_jac;
	    


	      // calc normal and tangential component u_n = mat_un, u_tau = mat_u
	      for (int i = 0; i < nd_u; i++)
		{
		  // n * u
		  double un = 0;
		  for (int j = 0; j < 3; j++)
		    un += normal(j) * u(i,j);


		  un *= inv_area2;

		  mat_un(i) =  un * area ;

// 		  for (int j = 0; j < 3; j++)
// 		    if ( i < 6 ) 
// 		      mat_u_nc(j,i) = u(i,j) - normal(j) * un;
// 		    else 
// 		      mat_u_nc(j,i) = u(i,j);

		  for (int j = 0; j < 3; j++)
		    mat_u(j,i) = u(i,j) - normal(j) * un;
	
		}


	      // sigma_ntau . u_tau
	      mat_u *= ir2d[l].Weight() * fac;
// 	      mat_u_nc *= ir2d[l].Weight() * fac;
	      hmat += Trans(mat_sigma) * mat_u;
// 	      hmat_nc += Trans(mat_sigma) * mat_u_nc;

 
// 	      // p_n . sigma_nn
// 	      for(int i=0; i<nd_sigma_face; i++)
// 		for(int j=0; j<fdofs.Size(); j++)
// 		  {
// 		    mat_p_sigma(fdofs_sigma[i],fdofs[j]) += p(fdofs[j]) * mat_sigmann(fdofs_sigma[i]) * ir2d[l].Weight() ;
// 		  }

	      // p_n . sigma_nn
	      for(int i=0; i<nd_sigma; i++)
		for(int j=0; j<fdofs.Size(); j++)
		  {
		    mat_p_sigma(i,fdofs[j]) += fac * p(fdofs[j]) * mat_sigmann(i) * ir2d[l].Weight() ;
		  }


	      // regularization:       reg * p^2
	      for(int i=0; i<fdofs.Size(); i++)
		for(int j=0; j<fdofs.Size(); j++)
		  elmat(fdofs[i]+nd_sigma + nd_u, fdofs[j]+nd_sigma + nd_u) -= fac * p(fdofs[i])*p(fdofs[j]) * ir2d[l].Weight() * area * regi;



	      
//               // new stabilization ############################
// 	      double coeff_divdiv = coef_div -> Evaluate (sip) * ir2d[l].Weight();
// 	      for(int m=0; m<nd_sigma; m++)
// 		for(int j=0; j<nd_sigma; j++)
// 		  {
// 		    elmat(m,j) += 
// 		      (div_sigma_bmat(0,m) * normal(0) + div_sigma_bmat(1,m) * normal(1) + div_sigma_bmat(2,m) * normal(2) ) * 
// 		      (div_sigma_bmat(0,j) * normal(0) + div_sigma_bmat(1,j) * normal(1) + div_sigma_bmat(2,j) * normal(2)) 
// 		      * coeff_divdiv * h * h * h;
// 		  }
             // end stabilization ############################



	      



	      // 	     	    // sigma_nn * v_n nonconforming
	    
	      //   	    for( int i = 0; i < nd_sigma; i++)
	      //   	      for(int j = 6; j<nd_u; j++)
	      //   		{
	      //     		  elmat(j+nd_sigma, i) -= mat_un(j) * mat_sigmann(i) * ir2d[l].Weight() / area;
	      // //    		  elmat(i, j+nd_sigma) -= mat_un(j) * mat_sigmann(i) * ir2d[l].Weight() / area;
	      //  		}
	    


	    

	      /*
		double weightdeth = ir2d[l].Weight() * sqrt(area);
		// B^T D B for Projection
		double mat_P = 0;
	    
		for( int  k = 0; k < nfdof; k++)
		//	      for ( int ij = 0; ij < nfdof; ij++)
		{
		mat_P += mat_sigmann(k) * mat_sigmann(k);
		}

		if( fel_sigma.GetOrderFacet() == 0  )
		{
		mat_P = 1;
		}

	
		// // 	    // - lambda_n u_n
		for( int i=0; i<fdofs.Size(); i++)
		for(int j=0; j<nd_u; j++)
		{
		elmat(fdofs[i]+nd_sigma+nd_u, nd_sigma+j) += p(fdofs[i]) * mat_un(j) * mat_P * weightdeth;
		elmat(j + nd_sigma, fdofs[i]+nd_sigma+nd_u) += p(fdofs[i]) * mat_un(j) * mat_P * weightdeth;
		}
	    
		// // 	    //  v_n  u_n
		for( int i=0; i<nd_u; i++)
		for(int j=0; j<nd_u; j++)
		{
		elmat(i+nd_sigma, nd_sigma+j) -=  mat_un(i) * mat_un(j)  * mat_P * weightdeth;
		}
	    
		// // 	    //  lambda_n mu_n
		for( int i=0; i<fdofs.Size(); i++)
		for(int j=0; j<fdofs.Size(); j++)
		{
		elmat(fdofs[i]+nd_sigma+nd_u, nd_sigma+nd_u + fdofs[j]) -= 
		p(fdofs[i]) * p(fdofs[j]) * mat_P * weightdeth;
		}

	      */

	    }
	

	
// 	  // add p.sigma_nn to elmat
// 	  for(int i=0; i<nd_sigma_face; i++)
// 	    for(int j=0; j<fdofs.Size(); j++)
// 	      {
// 		elmat(fdofs_sigma[i], nd_sigma + nd_u + fdofs[j]) += 
// 		  mat_p_sigma(fdofs_sigma[i],fdofs[j]);
// 		elmat(nd_sigma + nd_u + fdofs[j], fdofs_sigma[i]) += 
// 		  mat_p_sigma(fdofs_sigma[i],fdofs[j]);
// 	      }

	  // add p.sigma_nn to elmat
	  for(int i=0; i<nd_sigma; i++)
	    for(int j=0; j<fdofs.Size(); j++)
	      {
		elmat(i, nd_sigma + nd_u + fdofs[j]) += mat_p_sigma(i,fdofs[j]);
		elmat(nd_sigma + nd_u + fdofs[j], i) += mat_p_sigma(i,fdofs[j]);
	      }
       
	}
    

   
      // add u_tau . sigmantau to elmat
      for (int i = 0; i < nd_sigma; i++)
	for (int j = 0; j < nd_u; j++)
	  {
	    //  	  elmat(j+nd_sigma,i) += hmat_nc(i,j); 
	    //  	  elmat(i,j+nd_sigma) += hmat_nc(i,j);
	    elmat(j+nd_sigma,i) += hmat(i,j); 
	    elmat(i,j+nd_sigma) += hmat(i,j);
	  }



      elmat *= -1;
    }
    
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHIntegrator3d (coeffs[0], coeffs[1], coeffs[2], coeffs[3]); 
    }

  };











  /****************************
Hybrid Integrator
  *****************************/


  class MixedMechHybridIntegrator : public BilinearFormIntegrator
  {
  public:
    MixedMechHybridIntegrator () 
      : BilinearFormIntegrator()
    { ; }

    virtual ~MixedMechHybridIntegrator () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }




    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();

      elmat.AssignMemory (ndof, ndof, lh);
      elmat = 0;

      // \int sigma_nt u_t \ds

      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const HDivSymFiniteElement<2> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
      const HCurlFiniteElementD<2> & fel_u = 
	dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement&> (cfel[2]);

      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();
      int nd_p = fel_p.GetNDof();
      int nd = bfel.GetNDof();

      ARRAY<INT<2> > order_f;
      fel_sigma.GetOrderFacet(order_f);
      FlatVector<> p(nd_p, lh);

      FlatVector<> mat_sigma(nd_sigma, lh);

    
      FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    
      p = 0;
      mat_sigma = 0;

      const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
      const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
      int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;

      int nfdof = 0;
      for ( int i = 0; i < order_f.Size(); i++ )
	nfdof += order_f[i][0]+1;

      FlatMatrix<> mat_p_sigma(nd_p, nfdof, lh);


      const IntegrationRule & ir1d =
	GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_p.Order());

      mat_p_sigma = 0.0;
      for (int k = 0; k < ned; k++)
	{
	  Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	  int es = edges[k][0]; 
	  int ee = edges[k][1];
	
	  //     int et = 3 - ee - es; // third edge

	  //if (vnums[es] > vnums[ee])  swap (es, ee);

	  ps(0) = points[es][0];
	  ps(1) = points[es][1];
	  pe(0) = points[ee][0];
	  pe(1) = points[ee][1];

	  tau_ref = pe-ps;
	  void * heapp = lh.GetPointer();

	  for (int l = 0; l < ir1d.GetNIP(); l++)
	    {
	      Vec<2> p2d = ps + ir1d[l](0)* (pe-ps);
	      IntegrationPoint ip( p2d(0), p2d(1) );
	    
	      SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	      Mat<2> jac = sip.GetJacobian();
	      Mat<2> inv_jac = sip.GetJacobianInverse();
	      double det = fabs (sip.GetJacobiDet());

	      tau = jac * tau_ref;
	      double len = L2Norm(tau);
	    
	      tau *= (1.0/len);
	      normal(0) = tau(1);
	      normal(1) = -tau(0);

	      DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	      //for (int i = 0; i < nd_sigma; i++)
	      for(int i=0; i<nfdof; i++)	    
		mat_sigma(i) = 
		  normal(0) * normal(0) * sigma_bmat(0,i) +
		  normal(1) * normal(1) * sigma_bmat(1,i) +
		  (normal(0) * normal(1) + normal(1) * normal(0)) * sigma_bmat(2,i);

	    
	      fel_p.CalcShape (ip,p);
	    
	      p *= len* ir1d[l].Weight();
	      //mat_p_sigma += p * Trans (mat_sigma);
	      for(int i = 0; i < nd_p; i++)
		for(int j = 0; j < nfdof; j++)
		  mat_p_sigma(i,j) += p(i) * mat_sigma(j);
	      // 	    cout << mat_sigma << endl << "p    " << p << endl;;
	      // 	    cout << "mult " << mat_p_sigma << endl;
	      lh.CleanUp (heapp);
	    }
	}

      for (int i = 0; i < nd_p; i++)
	for (int j = 0; j < nfdof; j++)
	  {
	    elmat(nd_sigma+nd_u+i, j) += mat_p_sigma(i,j);
	    elmat(j, nd_sigma+nd_u+i) += mat_p_sigma(i,j);
	  
	  }
    }













    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHybridIntegrator (); 
    }

  };







  /**************************
hybrid mass integrator
  ************************/


  class MixedMechHybridMassIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;

  public:

    MixedMechHybridMassIntegrator (CoefficientFunction * acoef) 
      : BilinearFormIntegrator()
    { coef = acoef; }

    virtual ~MixedMechHybridMassIntegrator () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }




    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();

      elmat.AssignMemory (ndof, ndof, lh);
      elmat = 0;

      // \int sigma_nt u_t \ds

      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement&> (bfel);

      int nd_p = fel_p.GetNDof();
      int nd = bfel.GetNDof();

      FlatVector<> p(nd_p, lh);
      FlatMatrix<> mat_p(nd_p,nd_p,lh);
    
      p = 0;
      mat_p=0;

      const POINT3D * points = ElementTopology::GetVertices (fel_p.ElementType());
      const EDGE * edges = ElementTopology::GetEdges (fel_p.ElementType());
      int ned = (fel_p.ElementType() == ET_TRIG) ? 3: 4;

      const IntegrationRule & ir1d =
	GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_p.Order());

      for (int k = 0; k < ned; k++)
	{
	  Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	  ps(0) = points[edges[k][0]][0];
	  ps(1) = points[edges[k][0]][1];
	  pe(0) = points[edges[k][1]][0];
	  pe(1) = points[edges[k][1]][1];

	  tau_ref = pe-ps;

	  void * heapp = lh.GetPointer();

	  for (int l = 0; l < ir1d.GetNIP(); l++)
	    {
	      Vec<2> p2d = ps + ir1d[l](0)* (pe-ps);
	      IntegrationPoint ip( p2d(0), p2d(1) );
	    
	      SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	      double ei = coef -> Evaluate(sip);
	    
	      fel_p.CalcShape (ip,p);
	    
	      p *=  ir1d[l].Weight();
	      mat_p += ei*p * Trans (p);
	      lh.CleanUp (heapp);
	    }
	}
      //double eps = 1e-7;
      for (int i = 0; i < nd_p; i++)
	for (int j = 0; j < nd_p; j++)
	  {
	    elmat(i, j) += mat_p(i,j);
	    elmat(j, i) += mat_p(i,j);	  
	  }
    }













    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHybridMassIntegrator (coeffs[0]); 
    }

  };





  class MixedMechHybridMassIntegrator3d : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;

  public:

    MixedMechHybridMassIntegrator3d (CoefficientFunction * acoef) 
      : BilinearFormIntegrator()
    { coef = acoef; }

    virtual ~MixedMechHybridMassIntegrator3d () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }




    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();

      elmat.AssignMemory (ndof, ndof, lh);
      elmat = 0;

      // \int sigma_nt u_t \ds

      const HybridFiniteElement & fel_p =
	dynamic_cast<const HybridFiniteElement&> (bfel);

      int nd_p = fel_p.GetNDof();
      int nd = bfel.GetNDof();

      FlatVector<> p(nd_p, lh);
      FlatMatrix<> mat_p(nd_p,nd_p,lh);
    
      p = 0;
      mat_p=0;

      const POINT3D * points = ElementTopology::GetVertices (ET_TET);
      const FACE * faces = ElementTopology::GetFaces (ET_TET);
      int nf = ElementTopology::GetNFaces (ET_TET);
    
      for (int k = 0; k < nf; k++)
	{
	  bool trig = ( faces[k][3] == -1 );

	  const IntegrationRule & ir2d =
	    GetIntegrationRules().SelectIntegrationRule ( ET_TRIG, 2*fel_p.Order());

	  Vec<3> pstart, vx, vy;

	  for (int j = 0; j < 3; j++)
	    {
	      pstart(j) = points[faces[k][0]][j];
	      vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	      vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	    }

	  Vec<3> taux, tauy, normal;

	  for (int l = 0; l < ir2d.GetNIP(); l++)
	    {
	      Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	      IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	      SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);

	      double ei = coef -> Evaluate(sip);
	    
	      fel_p.CalcShape (ip,p);




	    
	      p *=  ir2d[l].Weight();
	      mat_p += ei*p * Trans (p);
	    }
	}
      //double eps = 1e-7;
      for (int i = 0; i < nd_p; i++)
	for (int j = 0; j < nd_p; j++)
	  {
	    elmat(i, j) += mat_p(i,j);
	    elmat(j, i) += mat_p(i,j);
	  
	  }
    }













    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHybridMassIntegrator3d (coeffs[0]); 
    }

  };




































  class MixedMechHybridSourceIntegrator2d : public LinearFormIntegrator
  {
    CoefficientFunction * coef1;
    CoefficientFunction * coef2;
    CoefficientFunction * coef_div;

  public:

    MixedMechHybridSourceIntegrator2d (CoefficientFunction * acoef1, CoefficientFunction * acoef2,
				       CoefficientFunction * acoef_div) 
      : LinearFormIntegrator()
    { 
      coef1 = acoef1; 
      coef2 = acoef2;
      coef_div = acoef_div;
    }

    virtual ~MixedMechHybridSourceIntegrator2d () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }




    virtual void
    AssembleElementVector (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatVector<double> & elvec,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();

      const int D = 2;

      elvec.AssignMemory (ndof,  lh);
      elvec = 0;


    
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const HDivSymFiniteElement<2> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
      const HCurlFiniteElementD<2> & fel_u = 
	dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);

      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();

      int intorder = 2 * fel_sigma.Order();
      const IntegrationRule & ir = 
	GetIntegrationRules().SelectIntegrationRule(fel_u.ElementType(), intorder);
    
      const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
      const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
      int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;
   
    
      FlatMatrixFixHeight<2> div_sigma_bmat(nd_sigma, lh);

      FlatMatrix<> u_ref(nd_u, 2, lh);
      FlatMatrixFixWidth<2> u(nd_u, lh);

      double h;
 
      for ( int i = 0; i<ir.GetNIP(); i++)
	{
	  SpecificIntegrationPoint<D,D> sip(ir[i], eltrans, lh);

	  Mat<2> jac = sip.GetJacobian();
	  Mat<2> inv_jac = sip.GetJacobianInverse();
	  double det = fabs (sip.GetJacobiDet());
	

	  Vec<2> ps, pe, tau_ref, sig_n, tau, normal, tau2, tau3, tau4, p3, p4;
	  int es = 0; 
	  int ee = 1;
	
	  ps(0) = points[es][0];
	  ps(1) = points[es][1];
	  pe(0) = points[ee][0];
	  pe(1) = points[ee][1];

	  tau_ref = pe-ps;
	  tau = jac * tau_ref;

	  if ( i == 0 )
	    {
	      p3(0) = points[2][0];
	      p3(1) = points[2][1];
	      if( ned == 4)
		{
		  p4(0) = points[3][0];
		  p4(1) = points[3][1];
		}
	    
	    
	      tau2 = jac * (p3 - pe);
	      if ( ned == 4)
		{
		  tau3 = jac * (p4 - p3);
		  tau4 = jac * (ps - p4);
		}
	      else
		tau3 = jac * (ps - p3);
	    }
	  // div(sigma)
	  DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);



	  // u
	  fel_u.CalcShape (sip.IP(), u_ref);
	  u = u_ref * inv_jac;

	  double weight = ir[i].Weight() * det;

	  double f1 = coef1->Evaluate (sip);
	  double f2 = coef2->Evaluate (sip);
	  double coeff_divdiv = coef_div -> Evaluate (sip);


	  if ( i == 0)
	    {
	      h = min2(L2Norm(tau), L2Norm(tau2));
	    
	      h = min2 ( h, L2Norm(tau3));
	      if ( ned == 4)
		h = min2 ( h, L2Norm (tau4) );
	    }
	  for(int m=0; m<nd_sigma; m++)
	    {
	      elvec(m) += (div_sigma_bmat(0,m) * f1 + div_sigma_bmat(1,m) * f2 ) * coeff_divdiv    * weight * h * h;
	    }

	  for(int m=0; m<nd_u; m++)
	    {
	      elvec(m + nd_sigma) += (u(m,0) * f1 + u(m,1) * f2 )  * weight;
	    }


	}


    }
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHybridSourceIntegrator2d (coeffs[0], coeffs[1], coeffs[2]); 
    }

  };



  class MixedMechHybridSourceIntegrator3d : public LinearFormIntegrator
  {
    CoefficientFunction * coef1;
    CoefficientFunction * coef2;
    CoefficientFunction * coef3;
    CoefficientFunction * coef_div;

  public:

    MixedMechHybridSourceIntegrator3d (CoefficientFunction * acoef1, CoefficientFunction * acoef2,
				       CoefficientFunction * acoef3,
				       CoefficientFunction * acoef_div) 
      : LinearFormIntegrator()
    { 
      coef1 = acoef1; 
      coef2 = acoef2;
      coef3 = acoef3;
      coef_div = acoef_div;
    }

    virtual ~MixedMechHybridSourceIntegrator3d () { ; }

    virtual bool BoundaryForm () const 
    { return 0; }




    virtual void
    AssembleElementVector (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatVector<double> & elvec,
			   LocalHeap & lh) const
    {
      int ndof = bfel.GetNDof();

      const int D = 3;

      elvec.AssignMemory (ndof,  lh);
      elvec = 0;


    
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const HDivSymFiniteElement<D> & fel_sigma = 
	dynamic_cast<const HDivSymFiniteElement<D>&> (cfel[0]);
      const HCurlHighOrderFiniteElementA<D> & fel_u = 
	dynamic_cast<const HCurlHighOrderFiniteElementA<D>&> (cfel[1]);

      int nd_sigma = fel_sigma.GetNDof();
      int nd_u = fel_u.GetNDof();

      int intorder = 2 * fel_sigma.Order() + 2;
      const IntegrationRule & ir = 
	GetIntegrationRules().SelectIntegrationRule(fel_u.ElementType(), intorder);
    
      const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
      const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
      int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;
   
    
      FlatMatrixFixHeight<D> div_sigma_bmat(nd_sigma, lh);

      FlatMatrix<> u_ref(nd_u, 3, lh);
      FlatMatrixFixWidth<D> u(nd_u, lh);

 
      FlatVector<> mat_sigmantau(nd_sigma, lh);

      for ( int i = 0; i<ir.GetNIP(); i++)
	{
	  SpecificIntegrationPoint<D,D> sip(ir[i], eltrans, lh);

	  Mat<D> jac = sip.GetJacobian();
	  Mat<D> inv_jac = sip.GetJacobianInverse();
	  double det = fabs (sip.GetJacobiDet());
	

	  Vec<D> ps, pe, tau_ref, sig_n, tau, normal;
	  int es = 0; 
	  int ee = 1;
	
	  ps(0) = points[es][0];
	  ps(1) = points[es][1];
	  ps(2) = points[es][2];
	  pe(0) = points[ee][0];
	  pe(1) = points[ee][1];
	  pe(2) = points[ee][2];

	  tau_ref = pe-ps;
	  tau = jac * tau_ref;



	  // div(sigma)
	  DiffOpDivHDivSym<D>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);



	  // u
	  fel_u.CalcShape (sip.IP(), u_ref);
	  u = u_ref * inv_jac;

	  double weight = ir[i].Weight() * det;

	  double f1 = coef1->Evaluate (sip);
	  double f2 = coef2->Evaluate (sip);
	  double f3 = coef3->Evaluate (sip);
	  double coeff_divdiv = coef_div -> Evaluate (sip);

	  double h = L2Norm(tau);
	
	  for(int m=0; m<nd_sigma; m++)
	    {
	      elvec(m) += (div_sigma_bmat(0,m) * f1 + div_sigma_bmat(1,m) * f2 + div_sigma_bmat(2,m) * f3 ) * coeff_divdiv    * weight * h * h;
	    }


	  int nd_u_c = nd_u;
	  if ( fel_u.GetAlternative() )
	    nd_u_c = 6;
	  for(int m=0; m<nd_u_c; m++)
	    {
	      elvec(m + nd_sigma) += ( u(m,0) * f1 + u(m,1) * f2 + u(m,2) * f3 )  * weight;
	    }

	
	}


    
    }




    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MixedMechHybridSourceIntegrator3d (coeffs[0], coeffs[1], coeffs[2], coeffs[3]); 
    }

  };















  namespace  
  {
  
    class Init
    { 
    public: 
      Init ();
    };
  
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("mixedmechhybrid", FESpaceMixedMechHybrid::Create);

      GetIntegrators().AddBFIntegrator ("mixedmechh", 2, 4,
					MixedMechHIntegrator::Create);
      GetIntegrators().AddBFIntegrator ("mixedmechh", 3, 4,
					MixedMechHIntegrator3d::Create);

      GetIntegrators().AddBFIntegrator ("mixedmechhybridjumps", 2, 0,
					MixedMechHybridIntegrator::Create);


      GetIntegrators().AddBFIntegrator ("mixedmechhybridmass", 2, 1,
					MixedMechHybridMassIntegrator::Create);
      GetIntegrators().AddBFIntegrator ("mixedmechhybridmass", 3, 1,
					MixedMechHybridMassIntegrator3d::Create);



      GetIntegrators().AddLFIntegrator ("mixedmechsource", 2, 3,
					MixedMechHybridSourceIntegrator2d ::Create);
      GetIntegrators().AddLFIntegrator ("mixedmechsource", 3, 4,
					MixedMechHybridSourceIntegrator3d ::Create);

    }
  
    Init init;

  }





}






