/* expfit
 */

#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>
#include "marinipietrafespace.hpp"
#include "marinipietrafe.hpp"

namespace expfit_cpp
{

  using namespace ngfem;
  using namespace ngsolve;


  
  /* *********************************************************** */
  /* ************* HybridExpFittingIntegrator ********************* */
  /* *********************************************************** */
  template<int D>
  class HybridExpFittingIntegrator : public BilinearFormIntegrator
  {
    protected:
      CoefficientFunction *coef_eps, *coef_b[D];
    public:
      HybridExpFittingIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
      { 
        coef_eps  = coeffs[0];
        for (int i=0; i<D; i++)
          coef_b[i]  = coeffs[i+1];
      }

      virtual ~HybridExpFittingIntegrator () { ; }

      static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
      {
        return new HybridExpFittingIntegrator<D> (coeffs);
      }
    
      virtual bool BoundaryForm () const 
      { return 0; }


      virtual void AssembleElementMatrix (const FiniteElement & fel,
                                          const ElementTransformation & eltrans, 
                                          FlatMatrix<double> & elmat,
                                          LocalHeap & lh) const
      {
        const CompoundFiniteElement & cfel = 
            dynamic_cast<const CompoundFiniteElement&> (fel);
    
        const HDivFiniteElement<D> & fel_hdiv = 
            dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
        const L2HighOrderFiniteElement & fel_l2 = 
            dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
        const FacetVolumeFiniteElement & fel_facet = 
            dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
        ELEMENT_TYPE eltype = cfel.ElementType();
      
        int nd = cfel.GetNDof();
        int nd_l2 = fel_l2.GetNDof();
        int nd_hdiv = fel_hdiv.GetNDof();
        int nd_facet = fel_facet.GetNDof();
      
        elmat.AssignMemory(nd, nd, lh);
        elmat = 0.0;

        
        
        // calculate the R_K's and S_K's
        int nfa = ElementTopology :: GetNFacets(eltype);
        double R_K=0, S_K[nfa], Smax, vol=0, length=0;
        double RK, SK[nfa];
        int facorder=10;

        int nv = ElementTopology::GetNVertices(eltype);
        const POINT3D* points = ElementTopology::GetVertices(eltype);
        
        double bmax=-1e99;
        for (int i=0; i<nv; i++)
        {
          IntegrationPoint ip(points[i], 0);
          SpecificIntegrationPoint<D,D> sip(ip, eltrans, lh);
          Vec<D> b, x;
          for (int i=0; i<D; i++)
          {
            b(i) = -coef_b[i]->Evaluate(sip);
            x(i) = ip(i);
          }
          Vec<1> bx;
          bx = Trans(b)*sip.GetJacobian()*x;
          bmax = max(bx(0),bmax);
        }
        
//         bmax = 0;
        
        IntegrationPoint ip(points[0], 0);
        SpecificIntegrationPoint<D,D> sip(ip, eltrans, lh);
        
        Mat<D,D> jac=sip.GetJacobian();
        Vec<D> b;
        for (int i=0; i<D; i++)
          b(i)=-coef_b[i]->Evaluate(sip);
        
        Vec<D> jtb=Trans(jac)*b;
        if (D==2)
        {
          double a=jtb(0), b=jtb(1), thresh=1e-10; 
          if (abs(a)<thresh) a=thresh;
          if (abs(b)<thresh) b=-thresh;
          
          RK = 2*( (exp(a)-1)/a - (exp(b)-1)/b )/(a-b);
          SK[0] = (exp(a)-1)/a;
          SK[1] = (exp(b)-1)/b;
          SK[2] = (exp(a)-exp(b))/(a-b);
          
//           cerr << "** RK=" << RK << ", S_K=[" << SK[0] << "," << SK[1] << "," << SK[2] << endl;
        }
        else // 3D
        {
          cerr << "** not implemented for 3D" << endl;
          exit(0);
        }
        
        
        
        
        bmax=0;
//         cerr << "** bmax = " << bmax << endl; 
        
        const IntegrationRule & irv =
            GetIntegrationRules().SelectIntegrationRule (eltype, facorder);
        R_K=0;
        for (int l=0; l<irv.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<D,D> sip(irv[l], eltrans, lh);
          double det = sip.GetJacobiDet();
          Vec<D> b, x; 
          for (int i=0; i<D; i++)
          {
            b(i) = -coef_b[i]->Evaluate(sip);
            x(i) = irv[l](i);
          }
          Vec<1> bx;
          bx=Trans(b)*sip.GetJacobian()*x; 
          R_K += exp(bx(0)-bmax) * det * sip.IP().Weight(); 
          vol += det * sip.IP().Weight();
        }
        R_K /= vol;
        
        
        
        
       for (int k = 0; k < nfa; k++)
        {
          S_K[k]=0;
          length = 0;
          ELEMENT_TYPE facettype = ElementTopology :: GetFacetType(eltype, k);
          const IntegrationRule & irf =
              GetIntegrationRules().SelectIntegrationRule (facettype, facorder);
          
//        normal on reference element with |normal| = length of tangential (in 3D: area of face)
          const NORMAL * normals = ElementTopology::GetNormals(eltype);
            
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];
      
//        tranformation of facet integration points (1D) to element points (2D):
          Facet2ElementTrafo transform(eltype); 
          for (int l = 0; l < irf.GetNIP(); l++)
          {
            const IntegrationPoint  & facet_ip = irf[l];
            IntegrationPoint volume_ip = transform(k, facet_ip);
          
            SpecificIntegrationPoint<D,D> sip (volume_ip, eltrans, lh);
            
//          piolamapping of normal
            Vec<D> normal = sip.GetJacobiDet() * Trans( sip.GetJacobianInverse()) * normal_ref;
	    double ds = irf[l].Weight() * L2Norm(normal);
            
            Vec<D> b, x;
            for (int i=0; i<D; i++)
            {
              b(i) = -coef_b[i]->Evaluate(sip);
              x(i) = volume_ip(i);
            }
            Vec<1> bx;
            bx = Trans(b)*sip.GetJacobian()*x;
            S_K[k] += exp(bx(0)-bmax) * ds;
            
            length += ds;
            
          }
          S_K[k] /= length;
        } // loop over facets

        
/*        swap(RK,R_K);
        for (int i=0; i<nfa; i++) swap(SK[i],S_K[i]);*/
        
        
        for (int i=1; i<nfa; i++) Smax=max(Smax,SK[i]);
        
        
//         cerr << "*** RK=" << RK << ", R_K=" << R_K << endl;
//         for (int i=0; i<nfa; i++)
//           cerr << "*** SK[" << i << "]=" << SK[i] << ", S_K[" << i << "]=" << S_K[i] << endl;
        
//         Smax = RK;
        
        // ****************************************************************************
        // the volume terms 1/eps (tau,sigma) + (div(tau),u) + (v, div(sigma))
        // ****************************************************************************
        FlatVector<> mat_l2(nd_l2, lh);
        FlatMatrix<> mat_hdiv(nd_hdiv, D, lh); 
        FlatVector<> mat_div_hdiv(nd_hdiv, lh);
        FlatVector<> mat_facet(nd_facet, lh);
      
        FlatMatrix<> mat_tau_sigma(nd_hdiv,nd_hdiv,lh) ;
        FlatMatrix<> mat_tau_u(nd_hdiv,nd_l2,lh); 
        FlatMatrix<> mat_v_sigma(nd_l2,nd_hdiv,lh); 
          
        const IntegrationRule & volume_ir =
            GetIntegrationRules().SelectIntegrationRule (eltype, 2*fel_hdiv.Order());
        
        void * heapp = lh.GetPointer();
        for (int l=0; l<volume_ir.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<D,D> sip(volume_ir[l], eltrans, lh);
        
          double det = sip.GetJacobiDet();
          fel_l2.CalcShape (sip.IP(), mat_l2) ;
        
          // piola transformation
          mat_hdiv =  (1.0/det) * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());
          mat_div_hdiv = (1.0/det) * fel_hdiv.GetDivShape(sip.IP(), lh);

          double dx = det * volume_ir[l].Weight();
          double eps = coef_eps->Evaluate(sip);
        
          //  1/eps * tau * sigma * dx
          mat_tau_sigma = (RK*dx/eps) * mat_hdiv * Trans( mat_hdiv );
        
          //  div(tau) * u * dx
          mat_tau_u = dx * mat_div_hdiv * Trans(mat_l2);
          mat_v_sigma = Trans(mat_tau_u);
          mat_tau_u *= Smax;
        
          // copy into the element matrix
          for (int i=0; i<nd_hdiv; i++)
          {
            for (int j=0; j<nd_hdiv; j++)
            {
              elmat(i,j) +=  mat_tau_sigma(i,j);
            }
            double sum=0;
            for (int j=0; j<nd_l2; j++)
            {
              elmat(i,nd_hdiv+j) += mat_tau_u(i,j);       //  (div(tau), u)
              elmat(nd_hdiv+j,i) += mat_v_sigma(j,i);     //  (v, div(sigma)
            }
          }
          lh.CleanUp(heapp);
        }
        
        // **********************************************************************************
        // the facet contribution: <tau.n, lambda>, <mu, sigma.n>
        // **********************************************************************************
//         int nfa = ElementTopology :: GetNFacets(eltype);
        
        FlatMatrix<> mat_tau_lambda(nd_hdiv, nd_facet, lh);
        FlatMatrix<> mat_mu_sigma(nd_facet, nd_hdiv, lh);
        FlatVector<> mat_hdiv_n(nd_hdiv, lh);
        
        mat_tau_lambda = 0;
        mat_mu_sigma = 0;
        
        for (int k = 0; k < nfa; k++)
        {
          ELEMENT_TYPE facettype = ElementTopology :: GetFacetType(eltype, k);
          const IntegrationRule & facet_ir =
              GetIntegrationRules().SelectIntegrationRule (facettype, 2*fel_hdiv.Order());
      
          // tranformation of facet integration points (1D) to element points (2D):
          Facet2ElementTrafo transform(eltype); 

          // normal on reference element with |normal| = length of tangential (in 3D: area of face)
          const NORMAL * normals = ElementTopology::GetNormals(eltype);
            
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];
            
          void * heapp = lh.GetPointer();
          

          for (int l = 0; l < facet_ir.GetNIP(); l++)
          {
            const IntegrationPoint  & facet_ip = facet_ir[l];
            IntegrationPoint volume_ip = transform(k, facet_ip);
          
            SpecificIntegrationPoint<D,D> sip (volume_ip, eltrans, lh);
	    
            // piola mapping of the normal: normal = det * F^(-T) normal_ref
            // this mapping has the property that 
            //    length of mapped normal = length of mapped edge (2D)
            //    length of mapped normal = area of mapped surface (3D)
            // area = |normal| and  normal0 = normal/|normal|
            //
            // sigma * normal_0 = 1/det sigma_ref F^T * det F^(-T) normal_ref / |normal|
            //                  = sigma_ref * normal_ref / area
            // the term 1/area cancels out below where ds = area * weight 
            mat_hdiv_n = fel_hdiv.GetShape(sip.IP(), lh) * normal_ref;
            
//             if (l==0) cerr << "** sigma.n(" << k << ")=" << Trans(mat_hdiv_n);
         
            // use (d-1) dimensional integration point to evaluate lambda
            fel_facet.CalcFacetShape(k, facet_ir[l], mat_facet); 

            // ds = area * weight: here area cancels out with the 1/area in sigma.n
            double ds = facet_ir[l].Weight();

            mat_tau_lambda += (SK[k] * ds * mat_hdiv_n) * Trans(mat_facet);
            mat_mu_sigma   += Trans((ds * mat_hdiv_n) * Trans(mat_facet));
            lh.CleanUp(heapp);
            
          }
        } // loop over facets
        
        // copy to element matrix
        for (int i=0; i<nd_hdiv; i++)
        {
          for (int j=0; j<nd_facet; j++)
          {
            elmat(i,nd_hdiv+nd_l2+j) -= mat_tau_lambda(i,j); // <tau.n, l>
            elmat(j+nd_hdiv+nd_l2,i) -= mat_mu_sigma(j,i); // <mu, sigma.n>
          }
        }
      }
  };


  template class HybridExpFittingIntegrator<2>;
  template class HybridExpFittingIntegrator<3>;


  
  
  
  
  
  
  
  
  
  
  class NumProcMpOrient : public NumProc
  {
    protected:
      
      FESpace * fes;
      CoefficientFunction * coef_b[3];

    public:
    
      NumProcMpOrient (PDE & apde, const Flags & flags)
      : NumProc (apde)
      {

        fes = pde.GetFESpace (flags.GetStringFlag ("fespace", "v"));
        coef_b[0] = pde.GetCoefficientFunction (flags.GetStringFlag ("b1", "b1"));
        coef_b[1] = pde.GetCoefficientFunction (flags.GetStringFlag ("b2", "b2"));
        if (ma.GetDimension()==3)
          coef_b[2] = pde.GetCoefficientFunction (flags.GetStringFlag ("b3", "b3"));
      }

      virtual ~NumProcMpOrient() 
      { ; }


      static NumProc * Create (PDE & pde, const Flags & flags)
      {
        return new NumProcMpOrient (pde, flags);
      }

      virtual void Do(LocalHeap & lh)
      {
        cout << "mporient: reorient the Marini-Pietra elements" << endl;
      
        int ne=ma.GetNE();
        const int D=2;
        
        ElementTransformation eltrans;
        for (int i=0; i<ne; i++)
        {
          const FiniteElement & fel = fes->GetFE(i, lh);
          ma.GetElementTransformation (i, eltrans, lh);
          
          IntegrationPoint ip(0, 0, 0);
          SpecificIntegrationPoint<D,D> sip(ip, eltrans, lh);
        
          Mat<D,D> jac=sip.GetJacobian();
          Vec<D> b;
          for (int j=0; j<D; j++)
            b(j)=-coef_b[j]->Evaluate(sip);
        

          Vec<D> jtb=Trans(jac)*b;
          double SK[3];
          int imax=2;
          
          if (D==2)
          {

            double a=jtb(0), b=jtb(1), thresh=1e-10; 
            if (abs(a)<thresh) a=thresh;
            if (abs(b)<thresh) b=-thresh;
          
            SK[0] = (exp(a)-1)/a;
            SK[1] = (exp(b)-1)/b;
            SK[2] = (exp(a)-exp(b))/(a-b);
            if (SK[1]>SK[2]) imax=1;
            if (SK[0]>SK[2] && SK[0]>SK[1]) imax=0;
            

            const CompoundFESpace & cfes = 
                dynamic_cast<const CompoundFESpace &>(*fes);
            
            const MariniPietraFESpace & mpfes = 
                dynamic_cast<const MariniPietraFESpace & >(*cfes[0] );
            
//             cerr << "*** setting orientation of element " << i << " to " << imax << endl;
            mpfes.SetOrientation(i, imax);
            
          }
          else // 3D
          {
            cerr << "** not implemented for 3D" << endl;
            exit(0);
          }
        
          
          
          

        }
  
  
      }






      virtual string GetClassName () const
      {
        return "MpOrient: ReOrient MArini Pietra Elements";
      }

      virtual void PrintReport (ostream & ost)
      {
        ost << GetClassName() << endl;
      }

      ///
      static void PrintDoc (ostream & ost)
      {
        ost << 
            "\n\nNumproc MpOrient:"   << endl;
      }
  };



  
  class Init
  { 
    public: 
      Init ();
  };
   
  Init::Init()
  {
    GetIntegrators().AddBFIntegrator ("expfitlaplace", 2, 3,
    HybridExpFittingIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("expfitlaplace", 3, 4,
    HybridExpFittingIntegrator<3>::Create);
  
    GetNumProcs().AddNumProc ("mporient", NumProcMpOrient::Create, NumProcMpOrient::PrintDoc);
  }

  Init init;

} // namespace

