#include "marinipietrafe.hpp"

namespace ngfem
{ 
  // Base Volume Element 
  template<int D>
  MariniPietraFiniteElement<D>::MariniPietraFiniteElement (ELEMENT_TYPE aeltype)
  : HDivFiniteElement<D> (aeltype, -1, -1) 
  { 
    fmax=D;
  }

  
  
  
  /****************************************************************
   * Triangle
   ****************************************************************/
  MariniPietraTrig :: MariniPietraTrig ()  
  : MariniPietraFiniteElement<2>(ET_TRIG) 
  {
    order=2;  // use 2nd order integration rule!
    ndof = 3;
  }
  
  void MariniPietraTrig :: 
  GetInternalDofs (ARRAY<int> & idofs) const
  {
      idofs.SetSize(0);
      for (int i=0; i<ndof; i++)  // everything external!
        idofs.Append(i);
  }


  void MariniPietraTrig :: 
  CalcShape (const IntegrationPoint & ip,  FlatMatrixFixWidth<2> shape) const
  {
    double x=ip(0), y=ip(1);
    shape = 0.0;
    shape(0,0) = 1;              shape(0,1) = 0; 
    shape(1,0) = 0;              shape(1,1) = 1;
    if (fmax==2)
    {
      shape(2,0) = 4*x*(x+y)-3*x;  shape(2,1) = 4*y*(y+x)-3*y; 
    }
    else if (fmax==1)
    {
      shape(2,0) = (1-x)*(1-4*x);  shape(2,1) = -y*(1-4*x);
    }
    else // if (fmax==0)
    {
      shape(2,0) = -x*(1-4*y);  shape(2,1) = (1-y)*(1-4*y);
    }
      
  }

  void MariniPietraTrig :: 
  CalcDivShape (const IntegrationPoint & ip,  FlatVector<> shape) const
  {
    double x=ip(0), y=ip(1);
    
    shape = 0.0;
    if (fmax==2)
      shape(2) = 12.*(x+y) - 6;
    else if (fmax==1)
      shape(2) = 12.*x - 6;
    else
      shape(2) = 12.*y - 6;
  }
  

  
  
  
  /****************************************************************
   * Tetrahedron
   ****************************************************************/

  MariniPietraTet :: MariniPietraTet ()  
  : MariniPietraFiniteElement<3>(ET_TET) 
  {
    order = 2; // use 2nd order integration rule
    ndof  = 4;
  }
  
  void MariniPietraTet :: GetInternalDofs (ARRAY<int> & idofs) const
  {
      idofs.SetSize(0);
      for (int i=0; i<ndof; i++) idofs.Append(i);
  }


  void MariniPietraTet :: 
      CalcShape (const IntegrationPoint & ip,  FlatMatrixFixWidth<3> shape) const
  {
    double x=ip(0), y=ip(1), z=ip(2);
    

    shape = 0.0;
    shape(0,0) = 1.; shape(0,1) = 0.; shape(0,2) = 0.;  
    shape(1,0) = 0.; shape(1,1) = 1.; shape(1,2) = 0.; 
    shape(2,0) = 0.; shape(2,1) = 0.; shape(2,2) = 1.; 
    
    if (fmax==0) { 
      shape(3,0)=(1-x)*(1-5*x); shape(3,1)=-y*(1-5*x); shape(3,2)=-z*(1-5*x); 
    }
    else if (fmax==1) { 
      shape(3,0)=-x*(1-5*y); shape(3,1)=(1-y)*(1-5*y); shape(3,2)=-z*(1-5*y); 
    }
    else if (fmax==2) { 
      shape(3,0)=-x*(1-5*z); shape(3,1)=-y*(1-5*z); shape(3,2)=(1-z)*(1-5*z); 
    }
    else  { // if fmax==3 
      shape(3,0)=-x*(1-1.2*(x+y+z)); shape(3,1)=y*(1-1.2*(x+y+z)); shape(3,2)=z*(1-1.2*(x+y+z)); 
    }
  }

  void MariniPietraTet :: 
  CalcDivShape (const IntegrationPoint & ip,  FlatVector<> shape) const
  {
    double x=ip(0), y=ip(1), z=ip(2);
    shape = 0.0;
    if (fmax < 3) 
      shape(3) = 20*ip(fmax) - 8.;
    else // fmax==3
      shape(3) = 20.*(x+y+z) - 12.; 
  }
  
  
  template class  MariniPietraNormalFiniteElement<2>;
  template class  MariniPietraNormalFiniteElement<3>;
  
  template class  MariniPietraFiniteElement<2>;
  template class  MariniPietraFiniteElement<3>;
}






