/**
   Marini Pietra FE Space: H(Div) discontinuous
*/

#include <comp.hpp>
#include <fem.hpp>  

#include "marinipietrafespace.hpp"
#include "marinipietrafe.hpp"

namespace ngcomp
{
  using namespace ngfem;
  
  MariniPietraFESpace ::  
  MariniPietraFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags)
  : FESpace (ama, flags)
  {
    name="MariniPietraFESpace(mphdiv)";
    DefineDefineFlag("mphdiv");

    if(parseflags) ParseFlags(flags);

    order = 0;
    low_order_space = 0; // space is already lowest order

    // Evaluator for shape tester 
    if (ma.GetDimension() == 2)
    {
      ARRAY<CoefficientFunction*> coeffs(1);
      coeffs[0] = new ConstantCoefficientFunction(1);
      evaluator = GetIntegrators().CreateBFI("masshdiv", 2, coeffs);
    }
    else
    {
      ARRAY<CoefficientFunction*> coeffs(1);
      coeffs[0] = new ConstantCoefficientFunction(1);
      evaluator = GetIntegrators().CreateBFI("masshdiv", 3, coeffs);
    }
    if (dimension > 1)
    {
      evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
    }
  }
  
  MariniPietraFESpace:: ~MariniPietraFESpace ()
  {
    ;
  }

  FESpace * MariniPietraFESpace ::
  Create (const MeshAccess & ma, const Flags & flags)
  {
     return new MariniPietraFESpace (ma, flags, true);
  }

  void MariniPietraFESpace :: Update()
  {
    nel = ma.GetNE();

    maxedge.SetSize(nel);
    maxedge = ma.GetDimension();

    first_element_dof.SetSize(nel+1);
    ndof = 0;
    for (int i=0; i<nel; i++)
    {
      first_element_dof[i] = ndof;
      switch(ma.GetElType(i)) 
      {
        case ET_TRIG:
          ndof += 3;
          break;
        case ET_TET:
          ndof += 4;
          break;
        default:
          cerr << "*** error in MariniPietraFESpace::Update(): unknown eltype" << endl;
          exit(0);
      }
      first_element_dof[nel] = ndof;
    }
    
    // standard floskelfor multigrid, would not be needed here...
    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;
  }

  const FiniteElement & MariniPietraFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe;

    switch (ma.GetElType(elnr))
    {
      case ET_TRIG:
        fe = new (lh.Alloc (sizeof(MariniPietraTrig)))  MariniPietraTrig();
        break;
      case ET_TET:
        fe = new (lh.Alloc (sizeof(MariniPietraTet)))  MariniPietraTet();
        break;
      default:
        fe = 0; 
    }
  
    if (!fe)
    {
      stringstream str;
      str << "MariniPietraFESpace " << GetClassName() 
          << ", undefined eltype " 
          << ElementTopology::GetElementName(ma.GetElType(elnr)) << endl;
      throw Exception (str.str());
    }

    if (ma.GetDimension()==2)
    {
      MariniPietraFiniteElement<2> * mpfe = dynamic_cast<MariniPietraFiniteElement<2> * >(fe);
      mpfe->SetOrientation(maxedge[elnr]); 
    }
    else // dim==3
    {
      MariniPietraFiniteElement<3> * mpfe = dynamic_cast<MariniPietraFiniteElement<3> * >(fe);
      mpfe->SetOrientation(maxedge[elnr]); 
    }
    return *fe;
  }

  const FiniteElement & MariniPietraFESpace :: GetSFE (int selnr, LocalHeap & lh) const
  {
    FiniteElement * fe;
    if (ma.GetDimension()==2)
      fe = new (lh.Alloc (sizeof(MariniPietraNormalFiniteElement<2>)))  
          MariniPietraNormalFiniteElement<2>(ma.GetSElType(selnr));
    else
      fe = new (lh.Alloc (sizeof(MariniPietraNormalFiniteElement<3>)))  
          MariniPietraNormalFiniteElement<3>(ma.GetSElType(selnr));
    return *fe;
  }
  
  int MariniPietraFESpace :: GetNDof () const
  {
    return ndof;
  }

  int MariniPietraFESpace :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }

  void MariniPietraFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    for (int i=first_element_dof[elnr]; i<first_element_dof[elnr+1]; i++)
      dnums.Append(i); 
  }

  void MariniPietraFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0); 
  }

  void MariniPietraFESpace :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
  }
  
  void MariniPietraFESpace :: SetOrientation(int elnr, int fmax) const 
  {
    maxedge[elnr]=fmax;
  }
} // namespace
  
namespace marini_pietra_fespace 
{
  using namespace ngcomp;
  class Init
  {
    public:
      Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("mphdiv", MariniPietraFESpace::Create);
  }
  
  Init init;
}




