/*
 * Braess-Schberl error estimator - Laplace version
 * =================================================
 * 
 * outline:
 * 
 *   (1) calculate fe-solution u_h
 *   (2) loop over all elements and
 *        i)   calc f^T = 
 *       ii)   calc f^E and add to global vector. 
 *   (3) loop over boundary edges and use 
 *        i)   du_n = g on neumann boundary (TODO) 
 *       ii)   choose in order to minimize L_2 norm of sigma_{w_n} (stage 1)        
 *   (4) define f - f_h = sum_{w_v} f_{w_v}  ??     -        
 *   (5) loop over vertices == patches
 *        i)   find local patch w_v
 *       ii)   f	T_{w_v} = 1/3 f^{T}_{w_v},  f^E_{w_v} = -1/2  f^E_{w_v}
 *      iii)   solve div sigma_{w_v} = f_{w_v} by solving linear (singular) systems
 *       iv)   assemble sigma^D = sum_{w_v} sigma_{w_v}
 *   --------------------------------------------------- 
 *   (6) check norm{grad(u) - grad(u_h)}_0 \le \norm{sigma^D}_0 (should be so by construction)
 *   (7) construct sigma :=  IntegrationPoint (double p1 = 0, double p2 = 0, double p3 = 0, double aw = 0); //!
 grad(u_h) + sigma^D (continuous Raviar Thomas function)  
 */

// TODOS:
//  - boundary condition
//  - calculate L2 norm   

#define DEBUG
#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;

namespace{
   inline void rotate(const Vec<2>& x, Vec<2>& y) { y(0) = -x(1); y(1) = x(0); };
   inline double spr(const Vec<2>& x, const Vec<2>& y) { return x(0)*y(0)+x(1)*y(1); };
}

/* ***************** Braess-Schberl error estimator ************************ */

class NumProcBrSchEE : public NumProc
{
   protected:
   public:
      ///
      NumProcBrSchEE (PDE & apde, const Flags & flags) : NumProc (apde)
      { 
      }
      ///
      virtual ~NumProcBrSchEE()
      { ; }
  
      static NumProc * Create (PDE & pde, const Flags & flags)
      {
         return new NumProcBrSchEE (pde, flags);
      }


      // The main routine -------------------------------------------------------------------
      virtual void Do(LocalHeap & lh)
      {
         // do not know what for?? 
         // int level = ma.GetNLevels() - 1;
    
         cout << "**************************************************************" << endl;
         cout << "called Braess Schberl error estimator" << endl;
         
         time_t start=clock();
         
         const int ned = ma.GetNEdges();
         const int ne = ma.GetNE();
         const int np = ma.GetNP();
         
         cout << "general information: ne=" << ne << ", ned=" << ned << ", np=" << np << endl;
         
         void * hp = lh.GetPointer();

         
             
         // storage for global vectors
         FlatVector<> fhE(ned,lh), fT(ne,lh), sigmaD(3*ne,lh);
               
             
         // this should probably be updated. 
         // u, a, ... should be given as arguments to numproc BrSchEE  
         S_GridFunction<double> & gfu = 
               *dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction("u"));
         BilinearForm & bfa = *pde.GetBilinearForm ("a");
         const FESpace & fes = gfu.GetFESpace();
         const BilinearFormIntegrator & bli = *bfa.GetIntegrator(0);
         
         // elementwise constant gridfunction containing the error contributions
         S_GridFunction<double> & gfe =
               *dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction("e"));
 
         // the rhs of the pde
         CoefficientFunction * coef_f = pde.GetCoefficientFunction ("coef_f");
               
         // those are often needed
         ARRAY<int> vnums, ednums, dnums, pnums, elnums;
         ARRAY<int> elsonvertex;

         // integration rules
         const IntegrationPoint ipT(1./3.,1./3.,0,0.5);  // over triangle

         
         
         
         
                           
         // initialize global vectors
         fhE=0;  fT=0, sigmaD=0;
         
//          NgProfiler::StartTimer(tim1);
         // loop over elements  (2)
         // and build fhE, and fT 
         for (int i = 0; i < ne; i++)
         {
//             NgProfiler::StartTimer(tim1_1);

            const NodalFiniteElement & fel = 
                  dynamic_cast<const NodalFiniteElement&>(fes.GetFE(i, lh));

            ElementTransformation eltrans;
            ma.GetElementTransformation (i, eltrans, lh);
            
            // part one (2i): int_el f dx
            SpecificIntegrationPoint<2,2> sipT(ipT,eltrans,lh);
            
              
            // fTi = int_T f dx = |T| f(xi) = |det(J)| 0.5 f(xi)
            fT(i) = Evaluate(*coef_f, sipT) * 0.5 * sipT.GetJacobiDet();  
            
            
            // part two (2ii): int_ed f_h ds = int_ed (uh)_n ds
            // this is done for each edge on each element. 
            // globally this amounts to int_ed [uh_n]_(jump) ds   for each edge!!
            int nd = fel.GetNDof();    
                    
            FlatVector<> elu(nd, lh);  // local dofs for u
            
            fes.GetDofNrs (i, dnums);
            gfu.GetElementVector (dnums, elu);
            ma.GetElPNums(i, pnums);
            ma.GetElEdges(i, ednums);
            
//             NgProfiler::StopTimer(tim1_1);
            // loop over the three edges
            for (int j=0; j<3; j++)
            {
//                NgProfiler::StartTimer(tim1_2);

               int v1, v2, v3;  
               ma.GetEdgePNums(ednums[j],v1,v2);
               v3=0;
               if (v1 != pnums[0] && v2 != pnums[0])
                  v3=pnums[0];
               else if (v1 != pnums[1] && v2 != pnums[1])
                  v3=pnums[1];
               else
                  v3=pnums[2];
               
               // calculate tangential and normal vector
               Vec<2> p1, p2, p3;
               ma.GetPoint(v1, p1);
               ma.GetPoint(v2, p2);
               ma.GetPoint(v3, p3);
               
               Vec<2> n, tau = p2-p1, rho = p3-p1;
               double h = L2Norm(tau);
               
               n(0) = -tau(1); 
               n(1) = tau(0);
               n /= h;
               // flip n if not pointing outwards
               if ( rho(0)*n(0) + rho(1)*n(1) > 0 )
                  n = -n; 
            
               
               const IntegrationRule & ir = 
                     GetIntegrationRules().SelectIntegrationRule(ET_SEGM, 1); // over edge: 
//                NgProfiler::StopTimer(tim1_2);

//                NgProfiler::StartTimer(tim1_3);

               double fhE_loc=0;
               for (int ipi=0; ipi<ir.GetNIP(); ipi++)
               {
                  const IntegrationPoint& ip=ir[ipi];
                  Vec<2> pref = p1 + ip(0)*tau;
                  IntegrationPoint ipel(pref(0), pref(1),0,0);
                  
                  // gradient in natural coordinates
                  FlatVector<> gradu(2, lh);                  
                  bli.CalcFlux(fel, eltrans, ipel, elu, gradu, 0, lh);
                 
                  fhE_loc += (gradu(0) * n(0) + gradu(1)*n(1))*h*ip.Weight();
                  
               }  // loop over integration points 
     
                      
               // add to global jump terms
               ma.GetEdgeElements( ednums[j], elnums);
               if (elnums.Size()>1) // TODO: boundary edges not treated yet!!
                  fhE(ednums[j]) += fhE_loc;
               
//                NgProfiler::StopTimer(tim1_3);

            }// loop over edges on element
            
         } // loop over elements
//          NgProfiler::StopTimer(tim1);
         
                           
         
         /* now build local contributions to sigma!!
          *   (5) loop over vertices => patches
          *        i)   find local patch w_v
          *       ii)   f_T_{w_v} = 1/3 f^{T}_{w_v},  f^E_{w_v} = -1/2  f^E_{w_v}
          *      iii)   solve div sigma_{w_v} = f_{w_v} by solving linear (singular) systems
          *       iv)   assemble sigma^D = sum_{w_v} sigma_{w_v}
         */
         
          // (5) loop over points => patches
         for (int i=0; i<np; i++)
         {
            // find local patch
            ma.GetVertexElements( i, elsonvertex);
            int ne_loc = elsonvertex.Size();  
            
//             NgProfiler::StartTimer(tim5);
// -----------------------------------------------------------------               
            // build local connectivities: edge-list, edge->point mapping, edge->element mapping
            // this is probably not needed but simplifies the notation below
            ARRAY<int> edges(2*ne_loc+1);   // edge-list
            edges=-1; 
            
            int ned_loc=0;                      // length of edge-list
            for (int j=0; j<ne_loc; j++)
            {
               ARRAY<int> ednums;
               int pn1, pn2;
               int elj = elsonvertex[j];    // current element
               
               ma.GetElEdges(elj, ednums);
               for (int k=0; k<ned_loc; k++)    // look if edge already in list
               {
                  for (int l=0; l<3; l++)   
                  {
                     if(ednums[l]==edges[k])  // mark as already listed
                        ednums[l]=-1;
                  }
               }
               
               for (int l=0; l<3; l++)
               {
                  if (ednums[l]>=0)
                  {
                     ma.GetEdgePNums( ednums[l], pn1, pn2);
                     if (pn1 == i || pn2 == i) // interesting edge should be added
                     {
                        edges[ned_loc]=ednums[l];
                        ned_loc++;
                     }
                  }
               }
             }
             
             // now build connectivity: loc_edge_number->loc_element_number
             ARRAY<int> elv1(ned_loc), elv2(ned_loc);
             elv1=-1; elv2=-1;
             for (int j=0; j<ned_loc; j++)
             {
                ARRAY<int> elnums;
                ma.GetEdgeElements(edges[j], elnums);
                
                // assign global element number
                elv1[j] = elnums[0];
                if (elnums.Size() == 2)
                {
                   elv2[j] = elnums[1];
                }
                
                // now transform global to local element numbers
                for (int k=0; k<ne_loc; k++)
                {
                   int elk=elsonvertex[k];
                   if ( elv1[j] == elk)
                   {
                      elv1[j] = k;
                   }
                   if ( elv2[j] == elk)
                   {
                      elv2[j] = k;
                   }
                }

             } // end of building local connectivity
// -----------------------------------------------------------------------------                           
             
//              NgProfiler::StopTimer(tim5);
             
             
//              NgProfiler::StartTimer(tim2);
            
             //TODO: boundary elements not treated at the moment
             if (ne_loc != ned_loc)  // boundary elements
             {
                continue;
             }
             
             
             FlatMatrix<> M( ned_loc+ne_loc, 2*ned_loc, lh); M=0;               
             FlatVector<> f( ned_loc+ne_loc, lh); f=0;
             // NOW loop over edges and assemble
             for (int j=0; j<ned_loc; j++)
             {
                // int_E [sigma] ds
                M(j,j)         = 1; // tau
                M(j,j+ne_loc)  = 1; // tau
                f(j)           = fhE(edges[j])*0.5;
                
                // int_T sigma dx = int_Es dsigma/dn ds
                // jth edge has  2  adjacent elements
                M(ne_loc+elv1[j],j)        = 1; // tau
                M(ne_loc+elv2[j],j+ne_loc) = 1; // tau
                
                // die eintraege kommen jeweils 2x vor !!
                f(ne_loc+elv1[j]) += fT(elsonvertex[elv1[j]])/6.0;
                f(ne_loc+elv2[j]) += fT(elsonvertex[elv2[j]])/6.0; 
             }         
//              NgProfiler::StopTimer(tim2);

                      
             // Now build system and solve 
            
//              NgProfiler::StartTimer(tim3);
             FlatMatrix<> A( 2*ned_loc, 2*ned_loc, lh);
             FlatVector<> b( 2*ned_loc, lh);
             A = Trans(M) * M ;
            
            // eliminate curls by regularization: A + alpha I
             double alpha=1e-6;
             for (int i=0; i<2*ned_loc; i++)
                A(i,i) = A(i,i)+alpha;
            
             b = Trans(M) * f;
            
             FlatMatrix<> invA( 2*ned_loc, 2*ned_loc, lh);
             CalcInverse(A, invA);
            
             FlatVector<> lsd( 2*ned_loc, lh); 
             lsd = invA * b;

                       
            
            // add to global vector
             for (int j=0; j<ned_loc; j++)
             {
                ARRAY<int> ednums;
                int el, ed;
                ed = edges[j];
               
               // edge 1
                el = elsonvertex[elv1[j]]; // global element number
                ma.GetElEdges(el, ednums);
                for (int k=0; k<3; k++)
                {
                   if (ed==ednums[k])
                  {
                     sigmaD(k*ne+el) += lsd(j);
                     break;
                  }
                }
               
               // edge 2
                el = elsonvertex[elv2[j]]; // global element number
                ma.GetElEdges(el, ednums);
                for (int k=0; k<3; k++)
                {
                   if (ed==ednums[k])
                   {
                      sigmaD(k*ne+el) += lsd(j+ned_loc);
                      break;
                   }
                }
               
               
             } // add to global vector
            
//              NgProfiler::StopTimer(tim3);

          
         } // end loop over patches
         
 


//          NgProfiler::StartTimer(tim4);
         // TODO: calculate |simga_D|_0^2 !! 
         ConstantCoefficientFunction one(1.0); 
         const MassHDivIntegrator<2> mhdiv(&one);
         const FE_RTTrig0 fel;
         
         FlatMatrix<> elmat(3,3,lh);
         FlatVector<> elsigma(3,lh), melsigma(3,lh);
         double l2sigma;
         
         double sum=0;
         const FESpace &fese = gfe.GetFESpace();
         for (int i=0; i<ne; i++)
         {
            
            // calculate L2 norm of sigma
            ElementTransformation eltrans;
            ma.GetElementTransformation (i, eltrans, lh);
            
         
            mhdiv.AssembleElementMatrix(fel,eltrans,elmat,lh);
            elsigma(0) = sigmaD(i);
            elsigma(1) = sigmaD(i+ne);   
            elsigma(2) = sigmaD(i+2*ne);   
            
            l2sigma = (Trans(elsigma) * elmat * elsigma)(0,0);
            sum += l2sigma;
            
            
            const NodalFiniteElement & fel = 
                  dynamic_cast<const NodalFiniteElement&>(fese.GetFE(i, lh));
       
            ARRAY<int> dnums;
            int nd = fel.GetNDof();
       
            FlatVector<> ele(nd,lh);
            fese.GetDofNrs (i, dnums);
            ele = l2sigma;
            gfe.SetElementVector(dnums, ele);
         }
    
         std::cout << "  total estimated error: " << sqrt(sum) << std::endl;

         time_t stop=clock();
         cout << "  ellapsed time: " << ((double) stop - start)/CLOCKS_PER_SEC << endl;
         
         cout << "**************************************************************" << endl;
         
//          NgProfiler::StopTimer(tim4);
    
            
        
      } // end Do


 
};





namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("eest", NumProcBrSchEE::Create);
  }
  
  Init init;
}
