#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


namespace boltzmann_cpp
{

  using namespace ngfem;
  using namespace ngsolve;


  
  class HermiteExpansionFESpace;
  
  HermiteExpansionFESpace * fes;


  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  class HermiteExpansionFESpace : public CompoundFESpace
  {
    int hermite_order;
    int dimv;

  public:
    HermiteExpansionFESpace (const MeshAccess & ama, 		   
                             const ARRAY<const FESpace*> & aspaces,
                             const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      hermite_order = int (flags.GetNumFlag ("hermiteorder", 2));
      dimv = (hermite_order+1) * (hermite_order+2) / 2;

      evaluator = 
        new CompoundBilinearFormIntegrator (*spaces[0]->GetEvaluator(), 0);
    }

    virtual ~HermiteExpansionFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(2);
      
      Flags l2flags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));
      
      l2flags.SetFlag ("order", order);
      facetflags.SetFlag("order", order);

      spaces[0] = new L2HighOrderFESpace (ma, l2flags);    
      spaces[1] = new FacetFESpace (ma, facetflags);

      fes = new HermiteExpansionFESpace (ma, spaces, flags);
      return fes;
    }
    
    virtual string GetClassName () const { return "HermiteExpansionFESpace"; }

    int GetDimV () { return dimv; }
    int GetHermiteOrder () { return hermite_order; }


    virtual void Update(LocalHeap & lh)
    {
      CompoundFESpace :: Update (lh);
    }

  };




  
  
  
  class NumProcBoltzmann : public NumProc
  {
  protected:
    double dt, te;
    int dimv, hermite_order;
    ARRAY<double> hx, hw;   // hermite integration-rule
    Matrix<> mat_hermite, mat_v1, mat_v2, mat_dv1, mat_dv2;
    Matrix<> hermite_xi;
    Vector<> hermite_fac;
    MatrixFixWidth<2> hermite_v;

    ARRAY<int[6]> neighbour;
    ARRAY<Vec<2> > elementnormals;
    ARRAY<FlatMatrix<> > mats_vnplus;  // 3 per element
    ARRAY<FlatMatrix<> > mats_vnminus; // 3 per element
    ARRAY<FlatMatrix<> > mats_absvn;   // one per edge
    ARRAY<Vector<>*> shapes_ipedge; // trig-shapes on the edges;

    Matrix<> mat_collision;
    double knudsen;
    bool collision;

    int num_udofs;
    int num_lamdofs;

    const BaseMatrix * invm_facet;
    const BaseMatrix * masse1;
    const BaseMatrix * masse2;
  public:
    
    NumProcBoltzmann (PDE & apde, const Flags & flags)
      : NumProc (apde)
    { 
      dt = flags.GetNumFlag ("dt", 1e-3);
      te = flags.GetNumFlag ("te", 1.0);
      knudsen = flags.GetNumFlag ("kn", 1.0);
      cout << "Knudsen number = " << knudsen << endl;

      collision = flags.GetDefineFlag ("collision");

      if (collision)
        cout << "apply collision" << endl;
      else
        cout << "no collision operator" << endl;

      dimv = fes->GetDimV();
      hermite_order = fes->GetHermiteOrder();

      ComputeHermiteRule (2 * hermite_order+3, hx, hw);
      for (int i = 0; i < hx.Size(); i++)
        hw[i] *= exp(sqr(hx[i]));

      mat_hermite.SetSize(dimv);
      mat_v1.SetSize(dimv);
      mat_v2.SetSize(dimv);
      mat_dv1.SetSize(dimv);
      mat_dv2.SetSize(dimv);

      if (collision)
        mat_collision.SetSize (dimv, dimv*dimv);

      Vector<AutoDiff<1> > herm1(hermite_order+1), herm2(hermite_order+1);
      Vector<> herm(dimv);
      MatrixFixWidth<2> dherm(dimv);

      mat_hermite = 0.0;
      mat_v1 = 0.0;
      mat_v2 = 0.0;
      mat_dv1 = 0.0;
      mat_dv2 = 0.0;
      hermite_fac.SetSize (sqr(hx.Size()));
      hermite_xi.SetSize (sqr(hx.Size()), dimv);
      hermite_v.SetSize (sqr(hx.Size()));


      {
        int n = 50;
        ARRAY<double> x, w;
        
        ComputeHermiteRule (n, x, w);
        
        *testout << "hermiterule, x = " << endl << x << ", w = " << endl << w << endl;
    
        Matrix<> mass(6);
        Vector<> hermite(6);
        
        mass = 0.0;
        for (int i = 0; i < n; i++)
          {
            HermitePolynomial (5, x[i], hermite);
            mass += w[i] * hermite * Trans(hermite);
          }
        
        *testout << "hermite-mass = " << endl << mass << endl;
      }




      for (int l1 = 0, ll = 0; l1 < hx.Size(); l1++)
        for (int l2 = 0; l2 < hx.Size(); l2++, ll++)
          {
            Vec<2> v (hx[l1], hx[l2]);
            
            hermite_v.Row(ll) = v;
            hermite_fac(ll) = exp(-v(0)*v(0)-v(1)*v(1)) * hw[l1] * hw[l2];

            AutoDiff<1> ad1(hx[l1], 0);
            AutoDiff<1> ad2(hx[l2], 0);

            HermitePolynomial (hermite_order, ad1, herm1);
            HermitePolynomial (hermite_order, ad2, herm2);
            for (int k1 = 0, kk=0; k1 <= hermite_order; k1++)
              for (int k2 = 0; k2 <= k1; k2++, kk++)
                {
                  herm(kk) = herm1(k1-k2).Value() * herm2(k2).Value();
                  dherm(kk,0) = herm1(k1-k2).DValue(0) * herm2(k2).Value();
                  dherm(kk,1) = herm1(k1-k2).Value() * herm2(k2).DValue(0);
                }

            hermite_xi.Row(ll) = herm;
            
            double facv = exp(-v(0)*v(0)-v(1)*v(1)) * hw[l1] * hw[l2];

            mat_hermite += facv * herm * Trans(herm);
            mat_v1 += (facv * v(0)) * herm * Trans(herm);
            mat_v2 += (facv * v(1)) * herm * Trans(herm);
            mat_dv1 += -facv * dherm.Col(0) * Trans(herm);
            mat_dv2 += -facv * dherm.Col(1) * Trans(herm);
          }

      for (int i = 0; i < dimv; i++)
        for (int j = 0; j < dimv; j++)
          {
            if (sqr (mat_hermite(i,j)) < 1e-12 * mat_hermite(i,i) * mat_hermite(j,j) ) mat_hermite(i,j) = 0.0;
            if (sqr (mat_v1(i,j)) < 1e-12 * mat_hermite(i,i) * mat_hermite(j,j) ) mat_v1(i,j) = 0.0;
            if (sqr (mat_v2(i,j)) < 1e-12 * mat_hermite(i,i) * mat_hermite(j,j) ) mat_v2(i,j) = 0.0;

            if (sqr (mat_dv1(i,j)) < 1e-12 * mat_hermite(i,i) * mat_hermite(j,j) ) mat_dv1(i,j) = 0.0;
            if (sqr (mat_dv2(i,j)) < 1e-12 * mat_hermite(i,i) * mat_hermite(j,j) ) mat_dv2(i,j) = 0.0;
          }
      
      *testout << "mat_hermite = " << endl << mat_hermite << endl;
      *testout << "mat_v1 = " << endl << mat_v1 << endl;
      *testout << "mat_dv1 = " << endl << mat_dv1 << endl;
    }

    virtual ~NumProcBoltzmann()
    { ; }

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcBoltzmann (pde, flags);
    }

    virtual string GetClassName () const
    {
      return "Boltzmann";
    }

    virtual void PrintReport (ostream & ost)
    {
      ;
    }

    virtual void Do(LocalHeap & lh)
    {
      cout << "**************************" << endl;
      cout << "***  Boltzmann solver  ***" << endl;
      cout << "**************************" << endl << endl;

      GridFunction & gf = *pde.GetGridFunction("f");
      GridFunction & gfhv = *pde.GetGridFunction("hv");
      GridFunction & gfhv2 = *pde.GetGridFunction("hv2");

      const BilinearForm & bfm = *pde.GetBilinearForm("m");
      const BilinearForm & bfm_facet = *pde.GetBilinearForm("mfacet");

      const LinearForm & lfinitial = *pde.GetLinearForm("initial");


      num_udofs = fes->GetStorageEnd(0);
      num_lamdofs = fes->GetStorageEnd(1) - fes->GetStorageStart(1);

      cout << "elemente dofs = " << num_udofs << endl;
      cout << "edge dofs     = " << num_lamdofs << endl;
      cout << "dimv          = " << dimv << endl << endl;

      if (collision)
        ComputeCollision ();




      const L2HighOrderFESpace & fes_l2 = *dynamic_cast<const L2HighOrderFESpace*> ((*fes)[0]);
      const FacetFESpace & fes_facet = *dynamic_cast<const FacetFESpace*> ((*fes)[1]);
    
      BitArray udofs(fes->GetNDof());
      udofs.Clear();
      for (int i = 0; i < num_udofs; i++)
        udofs.Set(i);

      BitArray lamdofs(fes->GetNDof());
      lamdofs.Clear();
      for (int i = num_udofs; i < num_udofs+num_lamdofs; i++)
        lamdofs.Set(i);

      
      const BaseSparseMatrix & matm = 
        dynamic_cast<const BaseSparseMatrix&> (bfm.GetMatrix());

      ARRAY<int> dnums;

      ARRAY<int> cntud(ma.GetNE());
      for (int i = 0; i < cntud.Size(); i++)
        cntud[i] = fes_l2.GetNElemDofs(i);
      Table<int> element_dofs(cntud);
      for (int i = 0; i < cntud.Size(); i++)
        {
          fes_l2.GetInnerDofNrs(i, dnums);
          for (int j = 0; j < dnums.Size(); j++)
            element_dofs[i][j] = dnums[j];
        }



      ARRAY<int> cntfd(ma.GetNFacets());
      for (int i = 0; i < cntfd.Size(); i++)
        cntfd[i] = fes_facet.GetNFacetDofs(i);
      Table<int> facet_dofs(cntfd);
      for (int i = 0; i < cntfd.Size(); i++)
        {
          fes_facet.GetFacetDofNrs(i, dnums);
          for (int j = 0; j < dnums.Size(); j++)
            facet_dofs[i][j] = dnums[j] + num_udofs;
        }

      const BaseSparseMatrix & matm_facet = 
        dynamic_cast<const BaseSparseMatrix&> (bfm_facet.GetMatrix());

      const BaseMatrix & invm = *matm.CreateBlockJacobiPrecond (element_dofs);
      invm_facet = matm_facet.CreateBlockJacobiPrecond (facet_dofs);

      gf.GetVector(0) = invm * lfinitial.GetVector();
      for (int j = 1; j < gf.GetMultiDim(); j++)
        gf.GetVector(j) = 0.0;



      int ne = ma.GetNE();
      neighbour.SetSize(ne);
      ARRAY<int> eledges, edgeels;
      for (int i = 0; i < ne; i++)
        {
          ma.GetElEdges (i, eledges);
          for (int j = 0; j < 3; j++)
            {
              ma.GetEdgeElements (eledges[j], edgeels);
              if (edgeels.Size() == 2)
                neighbour[i][j] = edgeels[0]+edgeels[1]-i;
              else
                neighbour[i][j] = -1;
            }
        }


      elementnormals.SetSize(0);

      ElementTransformation eltrans;
      for (int elnr = 0; elnr < ma.GetNE(); elnr++)
        {
          HeapReset hr(lh);
		      
          const CompoundFiniteElement & fel = 
            dynamic_cast<const CompoundFiniteElement&> (fes->GetFE (elnr, lh));

          ma.GetElementTransformation (elnr, eltrans, lh);

          const NodalFiniteElement & fel_l2 = 
            dynamic_cast<const NodalFiniteElement&> (fel[0]);


          ELEMENT_TYPE eltype = fel.ElementType();

          int ned = ElementTopology::GetNFacets(eltype);
          Facet2ElementTrafo transform(eltype); 
          const NORMAL * normals = ElementTopology::GetNormals(eltype);

          for (int k = 0; k < ned; k++)
            {
              ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);
              
              Vec<2> normal_ref;
              for (int i=0; i<2; i++)
                normal_ref(i) = normals[k][i];
              
              const IntegrationRule & ir1d =
                SelectIntegrationRule (etfacet, 2*fel.Order());
              
              for (int l = 0; l < ir1d.GetNIP(); l++)
                {
                  IntegrationPoint ip = transform(k, ir1d[l]);
                  SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
                  
                  Mat<D> jac = sip.GetJacobian();
                  Mat<D> inv_jac = sip.GetJacobianInverse();
                  double det = sip.GetJacobiDet();

                  Vec<D> normal = det * Trans (inv_jac) * normal_ref;       
                  elementnormals.Append (normal);
                  
                  shapes_ipedge.Append (new Vector<> (fel_l2.GetNDof()));
                  fel_l2.CalcShape (ip, *shapes_ipedge.Last());
                }
            }
        }


      mats_vnplus.SetSize (ma.GetNE()*3);
      mats_vnminus.SetSize (ma.GetNE()*3);
      mats_absvn.SetSize (ma.GetNFacets());

      for (int elnr = 0, cntnormal=0; elnr < ma.GetNE(); elnr++)
        {
          HeapReset hr(lh);
		      
          const CompoundFiniteElement & fel = 
            dynamic_cast<const CompoundFiniteElement&> (fes -> GetFE (elnr, lh));

          ELEMENT_TYPE eltype = fel.ElementType();
          ARRAY<int> fanums;

          ma.GetElFacets (elnr, fanums);

          FlatMatrix<> mat_hermite (dimv, lh);

          int ned = ElementTopology::GetNFacets(eltype);
          for (int k = 0; k < ned; k++)
            {
              ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

              const IntegrationRule & ir1d =
                SelectIntegrationRule (etfacet, 2*fel.Order());

              Vec<D> normal = elementnormals[cntnormal]; cntnormal += ir1d.GetNIP();

              mat_hermite = 0.0;
              for (int ll = 0; ll < hermite_fac.Size(); ll++)
                {
                  double vn = InnerProduct (hermite_v.Row(ll), normal);
                  if (vn < 0) continue;
                  mat_hermite += (vn * hermite_fac(ll))  * hermite_xi.Row(ll) * Trans (hermite_xi.Row(ll));
                }
              
              mats_vnplus[elnr*3+k].AssignMemory (dimv, dimv, new double[dimv*dimv]);
              mats_vnplus[elnr*3+k] = mat_hermite;

              mat_hermite = 0.0;
              for (int ll = 0; ll < hermite_fac.Size(); ll++)
                {
                  double vn = InnerProduct (hermite_v.Row(ll), normal);
                  if (vn > 0) continue;
                  mat_hermite += (vn * hermite_fac(ll))  * hermite_xi.Row(ll) * Trans (hermite_xi.Row(ll));
                }
              
              mats_vnminus[elnr*3+k].AssignMemory (dimv, dimv, new double[dimv*dimv]);
              mats_vnminus[elnr*3+k] = mat_hermite;

              mats_absvn[fanums[k]].AssignMemory (dimv, dimv, new double[dimv*dimv]);

              mat_hermite = mats_vnplus[elnr*3+k] - mats_vnminus[elnr*3+k];
              mat_hermite *= 1.0 / L2Norm (normal);
              CalcInverse (mat_hermite, mats_absvn[fanums[k]]);
            }
        }






      clock_t starttime, time;
      starttime = clock();

      // simple Euler
      int cnt = 0;
      for (double t = 0; t <= te; t += dt, cnt++)
        {
          cout << "t = " << t  
               << "   per step = " << double(clock()-starttime) / (double (CLOCKS_PER_SEC) * cnt)
               << endl;

          // calc facet dofs:
          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j) = 0.0;
          ApplyFacetFacet (gf, gfhv, lh);
          ApplyBoundaryFacet (gfhv, lh);
          SolveFacet (gfhv, gf, bfm_facet, lh);
          
          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j) = 0.0;

          ApplyElementElement (gf, gfhv, lh);
          ApplyFacetElement (gf, gfhv, lh);

          for (int j = 0; j < dimv; j++)
            gf.GetVector(j) -= (dt/mat_hermite(j,j)) * (invm * gfhv.GetVector(j));

          Ng_Redraw();
        }

      /*
      // improved Euler
      int cnt = 0;
      for (double t = 0; t <= te; t += dt, cnt++)
      {
      cout << "t = " << t  
      << "   per step = " << double(clock()-starttime) / (double (CLOCKS_PER_SEC) * cnt)
      << endl;

      // calc facet dofs:
      for (int j = 0; j < dimv; j++)
      gfhv.GetVector(j) = 0.0;
      ApplyFacetFacet (gf, gfhv, lh);
      ApplyBoundaryFacet (gfhv, lh);
      SolveFacet (gfhv, gf, bfm_facet, lh);

      for (int j = 0; j < dimv; j++)
      gfhv.GetVector(j) = 0.0;

      ApplyElementElement (gf, gfhv, lh);
      ApplyFacetElement (gf, gfhv, lh);

      for (int j = 0; j < dimv; j++)
      gfhv2.GetVector(j) = gf.GetVector(j) - dt * (invm * gfhv.GetVector(j));


      // calc facet dofs:
      for (int j = 0; j < dimv; j++)
      gfhv.GetVector(j) = 0.0;
      ApplyFacetFacet (gfhv2, gfhv, lh);
      ApplyBoundaryFacet (gfhv, lh);
      SolveFacet (gfhv, gfhv2, bfm_facet, lh);

      for (int j = 0; j < dimv; j++)
      gfhv.GetVector(j) = 0.0;

      ApplyElementElement (gfhv2, gfhv, lh);
      ApplyFacetElement (gfhv2, gfhv, lh);

      for (int j = 0; j < dimv; j++)
      gf.GetVector(j) -= dt * (invm * gfhv.GetVector(j));

      Ng_Redraw();
      }
      */


    }



    void ComputeCollision ()
    {
      cout << "compute collision " << flush;
      const IntegrationRule & ir1d =
        SelectIntegrationRule (ET_SEGM, 6*dimv);

      mat_collision = 0.0;

      ARRAY<double> hx, hw;
      ComputeHermiteRule (2 * hermite_order+3, hx, hw);
      for (int i = 0; i < hx.Size(); i++)
        hw[i] *= exp(sqr(hx[i]));


      /*
        double sum = 0;
        for (int i_theta = 0; i_theta < ir1d.GetNIP(); i_theta++)
        for (int i_phi = 0; i_phi < ir1d.GetNIP(); i_phi++)
        {
        double theta = M_PI * ir1d[i_theta](0);
        double phi = 2*M_PI * ir1d[i_phi](0);

        Vec<3> eprime (cos(theta), 
        sin(theta) * cos(phi), 
        sin(theta) * sin(phi));
            
        double weight = 2*sqr(M_PI) * sin(theta) * ir1d[i_theta].Weight() * ir1d[i_phi].Weight();



        for (int j_v = 


        sum += weight;
        }
        cout << "sum = " << sum << " =?= " << 4*M_PI << endl;
      */



      Vector<> vec_v(dimv), vec_w(dimv);
      Vector<> vec_vp(dimv), vec_wp(dimv), vec_test(dimv);
      Vector<> hmat(dimv*dimv);

      /*
        Matrix<> mat_collision1(dimv, dimv*dimv);
        mat_collision1 = 0.0;
        for (int i_phi = 0; i_phi < ir1d.GetNIP(); i_phi++)
        {
        cout << "+" << flush;
        double phi = 2*M_PI * ir1d[i_phi](0);
          
        Vec<2> eprime (cos(phi), sin(phi));
            
        double weight = 2*M_PI * ir1d[i_phi].Weight();

        for (int j_v1 = 0; j_v1 < hx.Size(); j_v1++)
        for (int j_v2 = 0; j_v2 < hx.Size(); j_v2++)
        for (int j_w1 = 0; j_w1 < hx.Size(); j_w1++)
        for (int j_w2 = 0; j_w2 < hx.Size(); j_w2++)
        {
        // cout << "." << flush;
        Vec<2> v(hx[j_v1], hx[j_v2]);
        Vec<2> w(hx[j_w1], hx[j_w2]);

        Vec<2> vp, wp;
        vp = 0.5 * (v+w) + 0.5 * L2Norm(v-w) * eprime;
        wp = 0.5 * (v+w) - 0.5 * L2Norm(v-w) * eprime;

        HermiteShape(v, vec_v);
        HermiteShape(w, vec_w);
        HermiteShape(vp, vec_vp);
        HermiteShape(wp, vec_wp);
        vec_test = vec_v;
        vec_v *= exp (-L2Norm2(v));
        vec_w *= exp (-L2Norm2(w));
        vec_vp *= exp (-L2Norm2(vp));
        vec_wp *= exp (-L2Norm2(wp));
        // int fvp fwp - fv fw
        double fac = collision * hw[j_v1] * hw[j_v2] * hw[j_w1] * hw[j_w2] * weight;

        for (int k1 = 0; k1 < dimv; k1++)
        for (int k2 = 0; k2 < dimv; k2++)
        mat_collision1.Col(k1+dimv*k2) += 
        (fac * (vec_vp(k1) * vec_wp(k2) - vec_v(k1) * vec_w(k2))) * vec_test;
        }
        }
      */


      for (int j_v1 = 0; j_v1 < hx.Size(); j_v1++)
        for (int j_v2 = 0; j_v2 < hx.Size(); j_v2++)
          {
            cout << "+" << flush;

            Vec<2> v(hx[j_v1], hx[j_v2]);
            HermiteShape(v, vec_v);
            vec_test = vec_v;
            vec_v *= exp (-L2Norm2(v));
            
            hmat = 0.0;

                
            for (int j_w1 = 0; j_w1 < hx.Size(); j_w1++)
              for (int j_w2 = 0; j_w2 < hx.Size(); j_w2++)
                {
                  Vec<2> w(hx[j_w1], hx[j_w2]);
                  HermiteShape(w, vec_w);
                  vec_w *= exp (-L2Norm2(w));
                  
                  for (int i_phi = 0; i_phi < ir1d.GetNIP(); i_phi++)
                    {
                      double phi = 2*M_PI * ir1d[i_phi](0);
                      Vec<2> eprime (cos(phi), sin(phi));
                      double weight = 2*M_PI * ir1d[i_phi].Weight();
                      
                      Vec<2> vp, wp;
                      vp = 0.5 * (v+w) + 0.5 * L2Norm(v-w) * eprime;
                      wp = 0.5 * (v+w) - 0.5 * L2Norm(v-w) * eprime;
                      
                      HermiteShape(vp, vec_vp);
                      HermiteShape(wp, vec_wp);
                      vec_vp *= exp (-L2Norm2(vp));
                      vec_wp *= exp (-L2Norm2(wp));
                      // int fvp fwp - fv fw

                      double costheta = InnerProduct (v, eprime) / sqrt (L2Norm2(v) * L2Norm2(eprime));
                      double collision = 1.0; //  1.0 / sqrt (1-costheta*costheta);
                      double fac = collision * hw[j_v1] * hw[j_v2] * hw[j_w1] * hw[j_w2] * weight;
                      
                      for (int k1 = 0; k1 < dimv; k1++)
                        for (int k2 = 0; k2 < dimv; k2++)
                          hmat(k1+dimv*k2) += 
                            fac * (vec_vp(k1) * vec_wp(k2) - vec_v(k1) * vec_w(k2));
                    }
                }
            mat_collision += vec_test * Trans (hmat);
          }

      /*
        double sum = 0;
        for (int i = 0; i < dimv*dimv; i++)
        sum +=  L2Norm2 (mat_collision.Col(i)-mat_collision1.Col(i));
        cout << "diff = " << sqrt(sum) << endl;
      */

      cout << "done " << endl;

      for (int k = 0; k < dimv; k++)
        {
          FlatMatrix<> collision_row(dimv, dimv, &mat_collision(k,0));
          Matrix<> hmat(dimv, dimv);
          *testout << "row " << k << " before: " << endl;
          *testout << collision_row << endl;

          for (int i = 0; i < dimv; i++)
            for (int j = 0; j < i; j++)
              {
                double mean = 0.5 * (collision_row(i,j) + collision_row(j,i));
                collision_row(i,j) = mean;
                collision_row(j,i) = mean;
              }

          *testout << "row " << k << " after: " << endl;
          *testout << collision_row << endl;
        }

      for (int k = 0; k < dimv*dimv*dimv; k++)
        if (fabs (mat_collision(k)) < 1e-10) 
          mat_collision(k) = 0;

      for (int k = 0; k < dimv; k++)
        {
          *testout << "collision, row " << k << " = " << endl;
          FlatMatrix<> collision_row(dimv, dimv, &mat_collision(k,0));
          *testout << collision_row << endl;
          *testout << "symmetric part = " << endl;
          *testout << collision_row+Trans(collision_row) << endl;
        }
    }

    void HermiteShape (const Vec<2> & v, FlatVector<> shape)
    {
      VectorMem<10> herm1(hermite_order+1);
      VectorMem<10> herm2(hermite_order+1);

      HermitePolynomial (hermite_order, v(0), herm1);
      HermitePolynomial (hermite_order, v(1), herm2);
      for (int k1 = 0, kk=0; k1 <= hermite_order; k1++)
        for (int k2 = 0; k2 <= k1; k2++, kk++)
          shape(kk) = herm1(k1-k2) * herm2(k2);
    }



    enum { D = 2 };

    
    void ApplyFacetFacet (GridFunction & gf,
                          GridFunction & gfhv,
                          LocalHeap & lh)
    {
      static int timer = NgProfiler::CreateTimer ("boltzmann::applyfacetfacet");
      static int timer1 = NgProfiler::CreateTimer ("boltzmann::applyfacetfacet1");
      static int timer2 = NgProfiler::CreateTimer ("boltzmann::applyfacetfacet2");
      static int timer3 = NgProfiler::CreateTimer ("boltzmann::applyfacetfacet3");
      NgProfiler::RegionTimer reg (timer);

      /*
        \int_V \int_E   vn  { f/f_in } \mu  dx dv
      */

      ARRAY<int> dnums;
      ElementTransformation eltrans;



      for (int elnr = 0, cntnormal=0; elnr < ma.GetNE(); elnr++)
        {
          HeapReset hr(lh);
		      
          const CompoundFiniteElement & fel = 
            dynamic_cast<const CompoundFiniteElement&> (fes -> GetFE (elnr, lh));

          ma.GetElementTransformation (elnr, eltrans, lh);
          fes->GetDofNrs (elnr, dnums);

          const NodalFiniteElement & fel_l2 = 
            dynamic_cast<const NodalFiniteElement&> (fel[0]);
          const FacetVolumeFiniteElement & fel_facet = 
            dynamic_cast<const FacetVolumeFiniteElement&> (fel[1]);
          
          ELEMENT_TYPE eltype = fel_l2.ElementType();

          int nd_facet = fel_facet.GetNDof();
          int nd_l2 = fel_l2.GetNDof();
          int nd = nd_facet+nd_l2;

          FlatMatrix<> elx(dimv, nd_l2, lh);
          FlatMatrix<> ely(dimv, nd_facet, lh);

          for (int j = 0; j < dimv; j++)
            gf.GetVector(j).GetIndirect(dnums.Range(0, nd_l2), elx.Row(j));

          int nipv = sqr (hx.Size());

          FlatVector<> vec_l2(nd_l2, lh);
          FlatVector<> vec_facet(nd_facet, lh);

          int ned = ElementTopology::GetNFacets(eltype);
          Facet2ElementTrafo transform(eltype); 
          const NORMAL * normals = ElementTopology::GetNormals(eltype);

          for (int k = 0; k < ned; k++)
            {
              ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

              Vec<2> normal_ref;
              for (int i=0; i<2; i++)
                normal_ref(i) = normals[k][i];

              const IntegrationRule & ir1d =
                SelectIntegrationRule (etfacet, 2*fel.Order());

              int first = fel_facet.GetFirstFacetDof(k);
              int next = fel_facet.GetFirstFacetDof(k+1);

              FlatMatrix<> mat_x (next-first, nd_l2, lh);
              FlatMatrix<> elz(dimv, next-first, lh);

              NgProfiler::StartTimer (timer2);

              mat_x = 0.0;
              for (int l = 0; l < ir1d.GetNIP(); l++)
                {
                  IntegrationPoint ip = transform(k, ir1d[l]);
                  
                  fel_l2.CalcShape (ip, vec_l2);
                  fel_facet.CalcFacetShape(k, ir1d[l], vec_facet);
                  mat_x += ir1d[l].Weight() * (vec_facet.Range(first,next) * Trans (vec_l2));
                }

              NgProfiler::StopTimer (timer2);
              NgProfiler::StartTimer (timer3);


              elz = elx * Trans (mat_x);

              SliceMatrix<> elys = ely.HRange (first, next);
              elys = mats_vnplus[elnr*3+k] * elz;

              NgProfiler::StopTimer (timer3);
            }
            
          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j).AddIndirect(dnums.Range(nd_l2, nd), ely.Row(j));
        }
    }






    
    void ApplyBoundaryFacet (GridFunction & gfhv,
                             LocalHeap & lh)
    {
      static int timer = NgProfiler::CreateTimer ("boltzmann::applyboundaryfacet");
      NgProfiler::RegionTimer reg (timer);

      /*
        \int_V \int_E   vn  { f/f_in } \mu  dx dv
      */

      ARRAY<int> dnums;
      ElementTransformation eltrans;

      for (int elnr = 0, cntnormal=0; elnr < ma.GetNE(); elnr++)
        {
          HeapReset hr(lh);
		      
          const CompoundFiniteElement & fel = 
            dynamic_cast<const CompoundFiniteElement&> (fes -> GetFE (elnr, lh));

          ma.GetElementTransformation (elnr, eltrans, lh);
          fes->GetDofNrs (elnr, dnums);

          const NodalFiniteElement & fel_l2 = 
            dynamic_cast<const NodalFiniteElement&> (fel[0]);
          const FacetVolumeFiniteElement & fel_facet = 
            dynamic_cast<const FacetVolumeFiniteElement&> (fel[1]);
          
          ELEMENT_TYPE eltype = fel_l2.ElementType();

          int nd_facet = fel_facet.GetNDof();
          int nd_l2 = fel_l2.GetNDof();
          int nd = nd_facet+nd_l2;

          FlatMatrix<> ely(dimv, nd_facet, lh);
          ely = 0.0;


          int nipv = sqr (hx.Size());

          FlatVector<> vec_l2(nd_l2, lh);
          FlatVector<> vec_facet(nd_facet, lh);
          FlatVector<> hvx (dimv, lh), hvy(dimv, lh);
          FlatVector<> hvx2 (nipv, lh), hvy2(nipv, lh);
          FlatVector<> hvn(nipv, lh);

          int ned = ElementTopology::GetNFacets(eltype);
          Facet2ElementTrafo transform(eltype); 
          const NORMAL * normals = ElementTopology::GetNormals(eltype);

          for (int k = 0; k < ned; k++)
            {
              ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

              const IntegrationRule & ir1d =
                SelectIntegrationRule (etfacet, 2*fel.Order());

              if (neighbour[elnr][k] != -1)
                {
                  cntnormal += ir1d.GetNIP();
                  continue;
                }

              int first = fel_facet.GetFirstFacetDof(k);
              int next = fel_facet.GetFirstFacetDof(k+1);

              for (int l = 0; l < ir1d.GetNIP(); l++)
                {
                  IntegrationPoint ip = transform(k, ir1d[l]);
                  Vec<D> normal = elementnormals[cntnormal];  cntnormal++; 
              
                  fel_l2.CalcShape (ip, vec_l2);
                  fel_facet.CalcFacetShape(k, ir1d[l], vec_facet);

                  double facx = ir1d[l].Weight();

                  hvn = hermite_v * normal;

                  for (int ll = 0; ll < nipv; ll++)
                    {
                      double vn = hvn(ll);
                      double fx = 0;
                      if (vn < 0) fx = -1;     // boundary value

                      hvy2(ll) = facx * hermite_fac(ll) * vn * fx;
                    }

                  hvy = Trans (hermite_xi) * hvy2;

                  SliceMatrix<> elys = ely.HRange (first, next);
                  elys += hvy * Trans (vec_facet.Range(first,next));
                }
            }
            
          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j).AddIndirect(dnums.Range(nd_l2, nd), ely.Row(j));
        }
    }










    virtual void ApplyElementElement (GridFunction & gf,
                                      GridFunction & gfhv,
                                      LocalHeap & lh)
    {
      static int timer = NgProfiler::CreateTimer ("boltzmann::applyelementelement");
      NgProfiler::RegionTimer reg (timer);

      /*
        \int_V \int_T  v nabla f g  dx dv    
      */

      const BaseMatrix & dudx1 = pde.GetBilinearForm("dudx1") -> GetMatrix();
      const BaseMatrix & dudx2 = pde.GetBilinearForm("dudx2") -> GetMatrix();


      BaseVector & hv = *gf.GetVector().CreateVector();

      for (int i = 0; i < dimv; i++)
        {
          hv = dudx1 * gf.GetVector(i);
          for (int j = 0; j < dimv; j++)
            if (fabs (mat_v1(j,i)) > 1e-14)
              gfhv.GetVector(j) += mat_v1(j,i) * hv;
        }

      for (int i = 0; i < dimv; i++)
        {
          hv = dudx2 * gf.GetVector(i);
          for (int j = 0; j < dimv; j++)
            if (fabs (mat_v2(j,i)) > 1e-14)
              gfhv.GetVector(j) += mat_v2(j,i) * hv;
        }


      const BaseMatrix & masse1 = pde.GetBilinearForm("masse1") -> GetMatrix();
      const BaseMatrix & masse2 = pde.GetBilinearForm("masse2") -> GetMatrix();

      hv = 0.0;
      for (int i = 0; i < dimv; i++)
        {
          hv = masse1 * gf.GetVector(i);
          for (int j = 0; j < dimv; j++)
            if (fabs (mat_dv1(j,i)) > 1e-14)
              gfhv.GetVector(j) += mat_dv1(j,i) * hv;
        }

      for (int i = 0; i < dimv; i++)
        {
          hv = masse2 * gf.GetVector(i);
          for (int j = 0; j < dimv; j++)
            if (fabs (mat_dv2(j,i)) > 1e-14)
              gfhv.GetVector(j) += mat_dv2(j,i) * hv;
        }

      // collision


      if (collision)
        {
          static int timer2 = NgProfiler::CreateTimer ("boltzmann::applyelementelement - collision");
          NgProfiler::RegionTimer reg2 (timer2);

          ARRAY<int> dnums;
          ElementTransformation eltrans;
          for (int elnr = 0; elnr < ma.GetNE(); elnr++)
            {
              HeapReset hr(lh);
		      
              const CompoundFiniteElement & fel = 
                dynamic_cast<const CompoundFiniteElement&> (fes -> GetFE (elnr, lh));

              ma.GetElementTransformation (elnr, eltrans, lh);
              fes->GetDofNrs (elnr, dnums);

              const NodalFiniteElement & fel_l2 = 
                dynamic_cast<const NodalFiniteElement&> (fel[0]);
          
              ELEMENT_TYPE eltype = fel.ElementType();

              int nd_l2 = fel_l2.GetNDof();

              FlatMatrix<> elx(dimv, nd_l2, lh);
              FlatMatrix<> ely(dimv, nd_l2, lh);

              ely = 0.0;

              for (int j = 0; j < dimv; j++)
                gf.GetVector(j).GetIndirect(dnums.Range(0, nd_l2), elx.Row(j));
          
              // *testout << "elx = " << endl << elx << endl;

              const IntegrationRule & ir =
                SelectIntegrationRule (eltype, 2*fel_l2.Order());
          
              FlatVector<> shape(nd_l2, lh);
              FlatVector<> func(dimv, lh), hv(dimv, lh), funcy(dimv, lh);

              for (int k = 0; k < ir.GetNIP(); k++)
                {
                  SpecificIntegrationPoint<2,2> sip(ir[k], eltrans, lh);
                  fel_l2.CalcShape (ir[k], shape);
              
                  func = elx * shape;

                  // *testout << "func = " << endl << func << endl;

                  for (int l = 0; l < dimv; l++)
                    {
                      FlatMatrix<> collision_row(dimv, dimv, &mat_collision(l,0));
                      hv = collision_row * func;
                      funcy(l) = InnerProduct (func, hv);
                    }

                  NgProfiler::AddFlops (timer2, dimv*dimv*dimv);

                  funcy *= -fabs (sip.GetJacobiDet()) * ir[k].Weight() / knudsen;

                  // *testout << "funcy = " << endl << funcy << endl;
                  ely += funcy * Trans (shape);
                }

              // *testout << "ely = " << endl << ely << endl;
              for (int j = 0; j < dimv; j++)
                gfhv.GetVector(j).AddIndirect(dnums.Range(0, nd_l2), ely.Row(j));
            }
        }
      delete &hv;
    }


    void ApplyFacetElement (GridFunction & gf,
                            GridFunction & gfhv,
                            LocalHeap & lh)
    {
      static int timer = NgProfiler::CreateTimer ("boltzmann::applyfacetelement");
      static int timer1 = NgProfiler::CreateTimer ("boltzmann::applyfacetelement1");
      static int timer2 = NgProfiler::CreateTimer ("boltzmann::applyfacetelement2");

      NgProfiler::RegionTimer reg (timer);

      /*
        \int_V \int_{gamma_in}  vn ( lam - f ) g  dx dv
      */


      ARRAY<int> dnums;
      ElementTransformation eltrans;


      for (int elnr = 0, cntnormal = 0; elnr < ma.GetNE(); elnr++)
        {
          HeapReset hr(lh);
		      
          const CompoundFiniteElement & fel = 
            dynamic_cast<const CompoundFiniteElement&> (fes -> GetFE (elnr, lh));

          ma.GetElementTransformation (elnr, eltrans, lh);
          fes->GetDofNrs (elnr, dnums);

          const NodalFiniteElement & fel_l2 = 
            dynamic_cast<const NodalFiniteElement&> (fel[0]);
          const FacetVolumeFiniteElement & fel_facet = 
            dynamic_cast<const FacetVolumeFiniteElement&> (fel[1]);
          
          ELEMENT_TYPE eltype = fel.ElementType();

          int nd_facet = fel_facet.GetNDof();
          int nd_l2 = fel_l2.GetNDof();
          int nd = fel.GetNDof();          

          // FlatMatrix<> elmat(nd_l2, nd, lh);
          // elmat = 0.0;

          FlatMatrix<> elx(dimv, nd, lh);
          FlatMatrix<> ely(dimv, nd_l2, lh);
          FlatMatrix<> elz(dimv, nd_l2, lh);
          ely = 0.0;

          for (int j = 0; j < dimv; j++)
            gf.GetVector(j).GetIndirect(dnums, elx.Row(j));

          FlatVector<> vec_l2(nd_l2, lh);
          FlatVector<> vec_facet(nd_facet, lh);
          FlatVector<> vec(nd, lh);

          FlatVector<> hv(dimv, lh), hvx (dimv, lh), hvy (dimv, lh);
          FlatVector<> hvx2 (sqr(hx.Size()), lh), hvy2 (sqr(hx.Size()), lh);
          // FlatMatrix<> mat_x (nd_facet, nd, lh);
          FlatMatrix<> mat_x (nd_l2, nd, lh);

          int ned = ElementTopology::GetNFacets(eltype);
          Facet2ElementTrafo transform(eltype); 
          const NORMAL * normals = ElementTopology::GetNormals(eltype);

          for (int k = 0; k < ned; k++)
            {
              ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

              Vec<2> normal_ref;
              for (int i=0; i<2; i++)
                normal_ref(i) = normals[k][i];

              
              const IntegrationRule & ir1d =
                SelectIntegrationRule (etfacet, 2*fel.Order());


              NgProfiler::StartTimer (timer1);

              /*
                mat_x = 0.0;
                for (int l = 0; l < ir1d.GetNIP(); l++)
                {
                IntegrationPoint ip = transform(k, ir1d[l]);
                  
                fel_l2.CalcShape (ip, vec_l2);
                fel_facet.CalcFacetShape(k, ir1d[l], vec_facet);
                vec.Range (0, nd_l2) = -vec_l2;
                vec.Range (nd_l2, nd) = vec_facet;
                mat_x += ir1d[l].Weight() * (vec_l2 * Trans (vec));
                }
                elz = elx * Trans (mat_x);
              */

              elz = 0;
              for (int l = 0; l < ir1d.GetNIP(); l++)
                {
                  IntegrationPoint ip = transform(k, ir1d[l]);

                  fel_l2.CalcShape (ip, vec_l2);
                  // vec_l2 = *shapes_ipedge[cntnormal]; cntnormal++;
                  fel_facet.CalcFacetShape(k, ir1d[l], vec_facet);
                  vec.Range (0, nd_l2) = -vec_l2;
                  vec.Range (nd_l2, nd) = vec_facet;
                  
                  hv = ir1d[l].Weight() * (elx * vec);
                  elz += hv * Trans (vec_l2);
                }

              NgProfiler::StopTimer (timer1);
              NgProfiler::AddFlops (timer1, nd * nd);
              NgProfiler::AddFlops (timer1, dimv * nd_l2);


              NgProfiler::StartTimer (timer2);
              
              ely += mats_vnminus[3*elnr+k] * elz;

              NgProfiler::AddFlops (timer2, dimv * dimv * nd_l2);
              NgProfiler::StopTimer (timer2);
            }
            
          // *testout << "elmat = " << endl << elmat << endl;

          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j).AddIndirect(dnums.Range(0, nd_l2), ely.Row(j));
        }
    }






    void SolveFacet (GridFunction & gfhv, 
                     GridFunction & gf, 
                     const BilinearForm & bfm_facet, 
                     LocalHeap & lh)
    {
      static int timer = NgProfiler::CreateTimer ("boltzmann::solvefacet");
      NgProfiler::RegionTimer reg (timer);

      int nfacets = ma.GetNFacets();
      const FacetFESpace & fes_facet = dynamic_cast<const FacetFESpace&> (*(*fes)[1]);

      ARRAY<int> fdnums;
      int first_facet_dof = fes->GetStorageStart(1);

      for (int i = 0; i < nfacets; i++)
        {
          HeapReset hr(lh);

          fes_facet.GetFacetDofNrs (i, fdnums);
          for (int j = 0; j < fdnums.Size(); j++)
            fdnums[j] += first_facet_dof;
          
          FlatMatrix<> elx(dimv, fdnums.Size(), lh);
          FlatMatrix<> ely(dimv, fdnums.Size(), lh);

          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j).GetIndirect(fdnums, elx.Row(j));

          ely = mats_absvn[i] * elx;

          for (int j = 0; j < dimv; j++)
            gfhv.GetVector(j).SetIndirect(fdnums, ely.Row(j));
        }

      for (int j = 0; j < dimv; j++)
        {
          gf.GetVector(j).FVDouble().Range (num_udofs, num_udofs+num_lamdofs) = 0.0;
          gf.GetVector(j) += (*invm_facet) * gfhv.GetVector(j);
        }
    }
  };





  /// Integrator for du/dxi v
  template <int comp>
  class dudxi_v_Integrator : public BilinearFormIntegrator
  {
  public:
    dudxi_v_Integrator () : BilinearFormIntegrator()
    { 
      ;
    }

    virtual ~dudxi_v_Integrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new dudxi_v_Integrator ();
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & bfel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      const MeshAccess & ma = fes->GetMeshAccess();

      const NodalFiniteElement & fel = 
        dynamic_cast<const NodalFiniteElement&> (bfel);
    
      ELEMENT_TYPE eltype = fel.ElementType();
      int nd = fel.GetNDof();          

      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      HeapReset hr(lh);

      const IntegrationRule & ir =
        SelectIntegrationRule (eltype, 2*fel.Order());

      FlatVector<> vec_l2(nd, lh);
      FlatMatrixFixWidth<2> mat_dl2(nd, lh);
      
      for (int i = 0; i < ir.GetNIP(); i++)
        {
          HeapReset hr(lh);
          
          IntegrationPoint ip = ir[i];
          SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
                  
          Mat<2> jac = sip.GetJacobian();
          Mat<2> inv_jac = sip.GetJacobianInverse();
          double det = sip.GetJacobiDet();
          
          fel.CalcShape (ip, vec_l2);
          mat_dl2 = fel.GetDShape (ip, lh) * inv_jac;

          double facx = fabs(det) * ir[i].Weight();
          
          elmat += facx * vec_l2 * Trans (mat_dl2.Col(comp));
        }
    }
  };






  
  template <int D>
  class MassFacetIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_mass;
  public:
    MassFacetIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_mass  = coeffs[0];
    }

    virtual ~MassFacetIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MassFacetIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & bfel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      const MeshAccess & ma = fes->GetMeshAccess();
      ARRAY<int> facets, facetels;
      ma.GetElEdges (eltrans.GetElementNr(), facets);

      const FacetVolumeFiniteElement & fel = 
        dynamic_cast<const FacetVolumeFiniteElement&> (bfel);
    
      ELEMENT_TYPE eltype = fel.ElementType();
      int nd = fel.GetNDof();          

      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      FlatVector<> vec_facet(nd, lh);

      int ned = ElementTopology::GetNFacets(eltype);
      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      for (int k = 0; k < ned; k++)
        {
          ma.GetEdgeElements (facets[k], facetels);
          if (facetels.Size() == 2 && min(facetels[0], facetels[1]) != eltrans.GetElementNr())
            continue;

          ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);
          
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*fel.Order());
          
          void * heapp = lh.GetPointer();
          
          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              Vec<D> normal = det * Trans (inv_jac) * normal_ref;       
              double len = L2Norm (normal);
              normal /= L2Norm (normal);

              fel.CalcFacetShape(k, ir1d[l], vec_facet);

              elmat += (len * ir1d[l].Weight()) * vec_facet * Trans(vec_facet);
            }
        }
    }
  };







  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    /*

    int nmax = 100;
    ARRAY<double> x, w;
    ComputeHermiteRule (nmax, x, w);

    ofstream hermout2 ("hermite2.out");

    Matrix<> solution (nmax, 1000);

    for (int n = 1; n < nmax; n++)
      {
        Matrix<> a(n, n), ainv(n,n);
        Vector<> f(n), u(n);
        Vector<> herm(n);

        a = 0.0;
        f = 0.0;

        for (int i = 0; i < x.Size(); i++)
          {
            HermitePolynomial (n-1, x[i], herm);
            a += w[i] * herm * Trans(herm);
            f += (exp(sqr(x[i])-sqr (4*x[i])) * w[i]) * herm;
          }
        
        // CalcInverse (a, ainv);
        // u = ainv * f;
        for (int i = 0; i < n; i++)
          u(i) = f(i) / a(i,i);

        hermout2 << "a = " << endl << a << endl;
        hermout2 << "f = " << endl << f << endl;
        hermout2 << "u = " << endl << u << endl;

        for (int i = 0; i < 1000; i++)
          {
            double x = -10 + 0.02 * i;
            HermitePolynomial (n-1, x, herm);
            solution(n, i) = exp (-x*x) * InnerProduct (herm, u);
          }
      }

       
    ofstream hermout ("hermite.out");
    for (int i = 0; i  < 1000; i++)
      {
        double x = -10 + 0.02 * i;
        hermout << x;
        for (int j = 1; j < 100; j++)
          hermout << " " << solution(j, i);
        hermout << endl;
      }
    */

    /*

    ofstream hermout ("hermite.out");

    int n = 100;
    ARRAY<double> x, w;
    
    ComputeHermiteRule (n, x, w);
    
    cout << "hermiterule, x = " << endl << x << ", w = " << endl << w << endl;
    
    Matrix<> mass(6);
    Vector<> hermite(100), sum(100);
    
    mass = 0.0;
    for (int i = 0; i < n; i++)
      {
        HermitePolynomial (5, x[i], hermite);
        mass += (fabs (x[i]) * w[i]) * hermite * Trans(hermite);
      }
    
    cout << "mass = " << endl << mass << endl;
    hermout << "hermite-|v| - mass = " << endl << mass << endl;
    

    int N = 10000;
    double a = -sqrt(N), b = sqrt(N);

    sum = 0;
    for (int j = 0; j < N; j++)
      {
        double h = (b-a)/N;
        double x = a + j * h;
        HermitePolynomial (99, x, hermite);
        sum += (fabs(x) * h * exp(-x*x)) * hermite;
      }

    hermout << "int |v| H_i" << sum << endl;
    */

      /*
      double integral = 0;
      cout << " a = " << a << endl;
      for (int j = 0; j < N; j++)
      {
      integral += (b-a)/N * exp(-sqr(a+ (b-a) * double(j)/N));
      }
      cout << "integral exp(-x*x) = " << integral << endl;
    
      ofstream out("hermite2.out");
      for (double x = -10; x <= 10; x+= 0.01)
      {
      HermitePolynomial (5, x, hermite);
      out << x;
      for (int j = 0; j < hermite.Size(); j++)
      out << " " << hermite(j) / sqrt(mass(j,j));
      out << endl;
      }
*/





    GetFESpaceClasses().AddFESpace ("hermiteexpansion", HermiteExpansionFESpace::Create);

    GetNumProcs().AddNumProc ("boltzmann", NumProcBoltzmann::Create);
    
    GetIntegrators().AddBFIntegrator ("massfacet", 2, 1,
                                      MassFacetIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("dudx1", 2, 0,
                                      dudxi_v_Integrator<0>::Create);
    GetIntegrators().AddBFIntegrator ("dudx2", 2, 0,
                                      dudxi_v_Integrator<1>::Create);
  }

  Init init;


}
