/*

Reissner-Mindlin finite element
by Chapelle and Stenberg

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace **************************** */

class FESpaceChapelleStenberg : public CompoundFESpace
{
public:
  FESpaceChapelleStenberg (const MeshAccess & ama, 		   
			   const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  {
    ;
  }

  virtual ~FESpaceChapelleStenberg () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    int order = int (flags.GetNumFlag ("order", 1));

    ARRAY<const FESpace*> spaces(5);

    Flags flags2;
    flags2.SetFlag ("order", order+1);
    spaces[0] = new H1HighOrderFESpace (ma, flags2);

    flags2.DeleteFlags();
    flags2.SetFlag ("order", order);
    flags2.SetFlag ("orderinner", order+2);
    spaces[1] = new H1HighOrderFESpace (ma, flags2);
    spaces[2] = new H1HighOrderFESpace (ma, flags2);


    flags2.DeleteFlags();
    flags2.SetFlag ("order", order-1);
    if (order <= 1)
      {
	spaces[3] = new ElementFESpace (ma, 0, 1, 0);
	spaces[4] = new ElementFESpace (ma, 0, 1, 0);
      }
    else
      {
	spaces[3] = new L2HighOrderFESpace (ma, flags2);
	spaces[4] = new L2HighOrderFESpace (ma, flags2);
      }

    FESpaceChapelleStenberg * fes = new FESpaceChapelleStenberg (ma, spaces);
    return fes;
  }
    
  virtual string GetClassName () const { return "ChapelleStenbergFESpace"; }
};






/* ********************+ Integrator ************************** */

class DiffOpRM : public DiffOp<DiffOpRM>
{
  // 7 components:
  // eps(beta), nabla w - beta, eta

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 7 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_w = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_beta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const NodalFiniteElement & fel_eta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[3]);
    
    int nd_w = fel_w.GetNDof();
    int nd_beta = fel_beta.GetNDof();
    int nd_eta = fel_eta.GetNDof();

    FlatMatrix<> gradw(2, nd_w, lh);
    gradw = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_w.GetDShape(sip.IP(), lh));
    
    FlatMatrix<> gradbeta(2, nd_beta, lh);
    gradbeta = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_beta.GetDShape(sip.IP(), lh));
    
    FlatVector<> vecbeta(nd_beta, lh);
    vecbeta = fel_beta.GetShape (sip.IP(), lh);

    FlatVector<> veceta(nd_eta, lh);
    veceta = fel_eta.GetShape (sip.IP(), lh);

    mat = 0;

    for (int j = 0; j < nd_w; j++)
      {
	mat(3,j) = gradw(0,j);
	mat(4,j) = gradw(1,j);
      }

    for (int j = 0; j < nd_beta; j++)
      {
	mat(0,nd_w+j)         = gradbeta(0,j);
	mat(1,nd_w+nd_beta+j) = gradbeta(1,j);
	mat(2,nd_w+j)         = gradbeta(1,j);
	mat(2,nd_w+nd_beta+j) = gradbeta(0,j);
	
	mat(3,nd_w+j)         = -vecbeta(j);
	mat(4,nd_w+nd_beta+j) = -vecbeta(j);
      }

    for (int j = 0; j < nd_eta; j++)
      {
	mat(5,nd_w+2*nd_beta+j)        = veceta(j);
	mat(6,nd_w+2*nd_beta+nd_eta+j) = veceta(j);
      }
  }
};


class RMDMat : public DMatOp<RMDMat>
{
  CoefficientFunction * coefe;
  CoefficientFunction * coefnu;
  CoefficientFunction * coeft;
public:

  enum { DIM_DMAT = 7 };

  RMDMat (CoefficientFunction * ae,
	      CoefficientFunction * anu,
	      CoefficientFunction * at)
    : coefe(ae), coefnu(anu), coeft(at) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double nu = Evaluate (*coefnu, sip);
    double e = Evaluate (*coefe, sip);
    double t = Evaluate (*coeft, sip);
    double h = sqrt (fabs (sip.GetJacobiDet()));


    mat(0,0) = mat(1,1) = 1;
    mat(0,1) = mat(1,0) = nu;
    mat(2,2) = (1-nu) / 2;

    mat *= (e / (1 - nu * nu));

    // double stab = 0.5 / (0.1*h+t);
    double stab = 1.0 / (h+t);
    double tnew = 1 / sqrt (1/(t*t) - stab*stab);

    // stab term
    mat(3,3) = stab*stab;
    mat(4,4) = stab*stab;
    
    /*
    mat(5,3) = mat(3,5) = 1-t*t*stab*stab;
    mat(6,4) = mat(4,6) = 1-t*t*stab*stab;

    mat(5,5) = -t*t+stab*stab*t*t*t*t;
    mat(6,6) = -t*t+stab*stab*t*t*t*t;
    */
    mat(5,3) = mat(3,5) = 1;
    mat(6,4) = mat(4,6) = 1;

    mat(5,5) = -tnew*tnew;   // -t*t+stab*stab*t*t*t*t;
    mat(6,6) = -tnew*tnew;   // -t*t+stab*stab*t*t*t*t;

  }
};


class RMIntegrator 
  : public T_BDBIntegrator<DiffOpRM, RMDMat, FiniteElement>
{
public:
  ///
  RMIntegrator (CoefficientFunction * ae,
		CoefficientFunction * anu,
		CoefficientFunction * at)
    :  T_BDBIntegrator<DiffOpRM, RMDMat, FiniteElement>
  (RMDMat (ae, anu, at))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new RMIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }
  
  ///
  virtual string Name () const { return "Reissner Mindlin"; }

};






class DiffOpIdRM : public DiffOp<DiffOpIdRM>
{
  // 5 components:
  // eps(beta), w, beta, eta

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 5 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_w = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_beta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const NodalFiniteElement & fel_eta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[3]);
    
    int nd_w = fel_w.GetNDof();
    int nd_beta = fel_beta.GetNDof();
    int nd_eta = fel_eta.GetNDof();

    
    FlatVector<> vecw(nd_w, lh);
    vecw = fel_w.GetShape (sip.IP(), lh);

    FlatVector<> vecbeta(nd_beta, lh);
    vecbeta = fel_beta.GetShape (sip.IP(), lh);

    FlatVector<> veceta(nd_eta, lh);
    veceta = fel_eta.GetShape (sip.IP(), lh);

    int i, j;
    mat = 0;

    for (j = 0; j < nd_w; j++)
      {
	mat(0,j) = vecw(j);
      }

    for (j = 0; j < nd_beta; j++)
      {
	mat(1,nd_w+j)         = vecbeta(j);
	mat(2,nd_w+nd_beta+j) = vecbeta(j);
      }

    for (j = 0; j < nd_eta; j++)
      {
	mat(3,nd_w+2*nd_beta+j)        = veceta(j);
	mat(4,nd_w+2*nd_beta+nd_eta+j) = veceta(j);
      }
  }
};


class RMIdDMat : public DMatOp<RMIdDMat>
{
public:

  enum { DIM_DMAT = 5 };

  RMIdDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;
    for (int i = 0; i < 5; i++)
      mat(i,i) = 1;
  }
};


class RMIdIntegrator 
  : public T_BDBIntegrator<DiffOpIdRM, RMIdDMat, FiniteElement>
{
public:
  ///
  RMIdIntegrator ()
    :  T_BDBIntegrator<DiffOpIdRM, RMIdDMat, FiniteElement>
  (RMIdDMat ())
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new RMIdIntegrator ();
  }
  
  ///
  virtual string Name () const { return "Reissner Mindlin Id"; }

};








/* ******************** Error Estimator ************** */


///
class NumProcCSEE : public NumProc
{
protected:
  double t;
public:
  ///
  NumProcCSEE (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    t = flags.GetNumFlag ("t",0.001); 
  }
  ///
  virtual ~NumProcCSEE()
  { ; }
  
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcCSEE (pde, flags);
  }
  
  ///
  virtual void Do(LocalHeap & lh)
  {
    int level = ma.GetNLevels() - 1;
    
    cout << "called Chapelle Stenberg error estimator" << endl;
    
    int ned = ma.GetNEdges();
    int ne = ma.GetNE();
    
    S_GridFunction<double> & gfu = 
      *dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction("u"));
    BilinearForm & bfa = *pde.GetBilinearForm ("a");
    const FESpace & fes = gfu.GetFESpace();
    const BilinearFormIntegrator & bli = *bfa.GetIntegrator(0);

    CoefficientFunction * coef_f = pde. GetCoefficientFunction ("coef_f");


    BaseVector & verr = pde.GetGridFunction("err")->GetVector();
    BaseVector & verr1 = pde.GetGridFunction("err1")->GetVector();
    BaseVector & verr2 = pde.GetGridFunction("err2")->GetVector();
    BaseVector & verr3 = pde.GetGridFunction("err3")->GetVector();

    BaseVector & verr_el1 = pde.GetGridFunction("err_el1")->GetVector();
    BaseVector & verr_el2 = pde.GetGridFunction("err_el2")->GetVector();
    BaseVector & verr_el3 = pde.GetGridFunction("err_el3")->GetVector();
    BaseVector & verr_el4 = pde.GetGridFunction("err_el4")->GetVector();

    verr = 0;
    verr1 = 0;
    verr2 = 0;
    verr3 = 0;
    verr_el1 = 0;
    verr_el2 = 0;    
    verr_el3 = 0;
    verr_el4 = 0;

    ARRAY<int> vnums, ednums, dnums;
    ARRAY<int> elsonvertex;

    void * hp = lh.GetPointer();
    for (int i = 0; i < ne; i++)
      {
	const FiniteElement & fel = fes.GetFE(i, lh);
	const NodalFiniteElement & felw = 
	  dynamic_cast<const NodalFiniteElement&> 
	  (dynamic_cast<const CompoundFiniteElement&> (fel)[0]);
	const NodalFiniteElement & felbeta = 
	  dynamic_cast<const NodalFiniteElement&> 
	  (dynamic_cast<const CompoundFiniteElement&> (fel)[1]);
	const NodalFiniteElement & felgamma = 
	  dynamic_cast<const NodalFiniteElement&> 
	  (dynamic_cast<const CompoundFiniteElement&> (fel)[3]);

	ElementTransformation eltrans;
	ma.GetElementTransformation (i, eltrans, lh);

	int nd = fel.GetNDof();
	int ndw = felw.GetNDof();
	int ndbeta = felbeta.GetNDof();
	int ndgamma = felgamma.GetNDof();

	FlatMatrix<> dshapew(ndw, 2, lh);
	FlatMatrix<> ddshapew(ndw, 4, lh);

	FlatVector<> shapebeta(ndbeta, lh);
	FlatMatrix<> dshapebeta(ndbeta, 2, lh);
	FlatMatrix<> ddshapebeta(ndbeta, 4, lh);

	FlatVector<> shapegamma(ndgamma, lh);
	FlatMatrix<> dshapegamma(ndgamma, 2, lh);


	FlatVector<> elvec(nd, lh);
	FlatVector<> elw(ndw, lh), elbeta1(ndbeta,lh), elbeta2(ndbeta, lh);
	FlatVector<> elgamma1(ndgamma, lh), elgamma2(ndgamma, lh);
	
	fes.GetDofNrs (i, dnums);
	gfu.GetElementVector (dnums, elvec);
	for (int j = 0; j < ndw; j++)
	  elw(j) = elvec(j);
	for (int j = 0; j < ndbeta; j++)
	  {
	    elbeta1(j) = elvec(ndw+j);
	    elbeta2(j) = elvec(ndw+ndbeta+j);
	  }
	for (int j = 0; j < ndgamma; j++)
	  {
	    elgamma1(j) = elvec(ndw+2*ndbeta+j);
	    elgamma2(j) = elvec(ndw+2*ndbeta+ndgamma+j);
	  }

	const IntegrationRule & ir = 
	  GetIntegrationRules().SelectIntegrationRule (fel.ElementType(), felw.Order()+2);

	for (int k = 0; k < ir.GetNIP(); k++)
	  {
	    const IntegrationPoint & ip = ir[k];
	    SpecificIntegrationPoint<2,2> sip(ip, eltrans, lh);
	    
	    double h = sqrt (fabs (sip.GetJacobiDet()));
	    double stab = 1.0 / (h+t);
	    double alpha = stab;
	    double tnew = 1 / sqrt (1/(t*t) - stab*stab);

	    double f = Evaluate (*coef_f, sip);


	    felw.CalcDShape (ip, dshapew);
	    felw.CalcDDShape (ip, ddshapew);

	    felbeta.CalcShape (ip, shapebeta);
	    felbeta.CalcDShape (ip, dshapebeta);
	    felbeta.CalcDDShape (ip, ddshapebeta);

	    felgamma.CalcShape (ip, shapegamma);
	    felgamma.CalcDShape (ip, dshapegamma);


	    Vec<2> gradwref = Trans (dshapew) * elw;
	    Vec<2> gradw = Trans (sip.GetJacobianInverse ()) * gradwref;

	    Vec<4> hdd = Trans (ddshapew) * elw;
	    Mat<2,2> ddwref;
	    ddwref(0,0) = hdd(0); ddwref(0,1) = hdd(1);
	    ddwref(1,0) = hdd(2); ddwref(1,1) = hdd(3);
	    Mat<2,2> ddw = Trans (sip.GetJacobianInverse ()) * ddwref * sip.GetJacobianInverse();




	    Vec<2> beta;
	    beta(0) = InnerProduct (shapebeta, elbeta1);
	    beta(1) = InnerProduct (shapebeta, elbeta2);

	    Vec<2> db1ref = Trans (dshapebeta) * elbeta1;
	    Vec<2> dbeta1 = Trans (sip.GetJacobianInverse ()) * db1ref;
	    Vec<2> db2ref = Trans (dshapebeta) * elbeta2;
	    Vec<2> dbeta2 = Trans (sip.GetJacobianInverse ()) * db2ref;


	    hdd = Trans (ddshapebeta) * elbeta1;
	    Mat<2,2> ddb1ref;
	    ddb1ref(0,0) = hdd(0); ddb1ref(0,1) = hdd(1);
	    ddb1ref(1,0) = hdd(2); ddb1ref(1,1) = hdd(3);
	    Mat<2,2> ddbeta1 = Trans (sip.GetJacobianInverse ()) * ddb1ref * sip.GetJacobianInverse();
	    
	    hdd = Trans (ddshapebeta) * elbeta2;
	    Mat<2,2> ddb2ref;
	    ddb2ref(0,0) = hdd(0); ddb2ref(0,1) = hdd(1);
	    ddb2ref(1,0) = hdd(2); ddb2ref(1,1) = hdd(3);
	    Mat<2,2> ddbeta2 = Trans (sip.GetJacobianInverse ()) * ddb2ref * sip.GetJacobianInverse();
	    
	    Vec<2> gamma;
	    gamma(0) = InnerProduct (shapegamma, elgamma1);
	    gamma(1) = InnerProduct (shapegamma, elgamma2);

	    Vec<2> dg1ref = Trans (dshapegamma) * elgamma1;
	    Vec<2> dgamma1 = Trans (sip.GetJacobianInverse ()) * dg1ref;
	    Vec<2> dg2ref = Trans (dshapegamma) * elgamma2;
	    Vec<2> dgamma2 = Trans (sip.GetJacobianInverse ()) * dg2ref;


	    Mat<3,2> deps, dsigma;
	    for (int j = 0; j < 2; j++)
	      {
		deps(0,j) = ddbeta1(0,j);
		deps(1,j) = ddbeta2(1,j);
		deps(2,j) = ddbeta1(1,j)+ddbeta2(0,j);
	      }

	    Mat<3,3> dmat;
	    dmat = 0.0;
	    double e = 1;
	    double nu = 0.2;
	    dmat(0,0) = dmat(1,1) = 1;
	    dmat(0,1) = dmat(1,0) = nu;
	    dmat(2,2) = (1-nu) / 2;
	    dmat *= (e / (1 - nu * nu));
	    
	    dsigma = dmat * deps;
	    Vec<2> divsigma;
	    divsigma(0) = dsigma(0,0) + dsigma(2,1);
	    divsigma(1) = dsigma(2,0) + dsigma(1,1);


	    /*
	    (*testout) << "divsigma = " << divsigma 
		       << " =?= " << (stab*stab*(beta-gradw)-gamma) << endl;
	    */

	    Vec<2> diff = stab*stab*(beta-gradw)-gamma - divsigma;
	    verr_el1.FVDouble()(i) += fabs (sip.GetJacobiDet()) * ip.Weight() *
	      h*h* L2Norm2(diff);

	    verr_el2.FVDouble()(i) += fabs (sip.GetJacobiDet()) * ip.Weight() *
	      h*h/stab/stab *
	      sqr (f - stab*stab*(dbeta1(0)+dbeta2(1)-ddw(0,0)-ddw(1,1)) + dgamma1(0)+dgamma2(1));

	    double curlbeta = dbeta1(1)-dbeta2(0);
	    verr_el3.FVDouble()(i) += fabs (sip.GetJacobiDet()) * ip.Weight() *
	      1.0 / sqr (1 + tnew/h) * 
	       sqr (dbeta1(1)-dbeta2(0) + tnew*tnew * (dgamma1(1)-dgamma2(0)));

	    Vec<2> res = beta - gradw + tnew*tnew * gamma;
	    verr_el4.FVDouble()(i) += fabs (sip.GetJacobiDet()) * ip.Weight() *
	      1.0 / sqr (1.0/stab + tnew) * L2Norm2 (res);

	    /*
	    (*testout) << "stab^2 = " <<  stab*stab
		       << ", f = " << f 
		       << ", stab .. = " << stab*stab*(dbeta1(0)+dbeta2(1)-ddw(0,0)-ddw(1,1))
		       << ", gradw = " << gradw << ", beta = " << beta
		       << ", div beta = " << dbeta1(0)+dbeta2(1)
		       << ", laplace w = " << ddw(0,0)+ddw(1,1)
		       << ", divgamma = " << dgamma1(0)+dgamma2(1) << endl;
	    */

	    /*

	    (*testout) << "elw = " << elw << endl;
	    (*testout) << "gradwref = " << gradwref << endl;
	    (*testout) << "gradw = " << gradw << endl;
	    (*testout) << "beta = " << beta << endl;
	    (*testout) << "gradw - beta = " << gradw-beta << endl;
	    (*testout) << "t^2 gamma = " << t*t*gamma << endl;
	    
	    (*testout) << "ddw = " << ddw << endl;
	    (*testout) << "dbeta1,2 = " << dbeta1 << "  " << dbeta2 << endl;
	    (*testout) << "div (beta-nabla w) = " 
		       << dbeta1(0)+dbeta2(1)-ddw(0,0)-ddw(1,1) << endl;
	    (*testout) << "div (gamma) = " 
		       << dgamma1(0)+dgamma2(1) << endl;
	    

	    (*testout) << "ddbeta1 = " << ddbeta1 << endl;
	    (*testout) << "ddbeta2 = " << ddbeta2 << endl;
	    */	    
	  }

	lh.CleanUp (hp);
      }


    for (int i = 0; i < ned; i++)
      {
	int vi1, vi2;
	ma.GetEdgePNums (i, vi1, vi2);

	int elnr1 = -1, elnr2 = -1;
	ma.GetVertexElements(vi1, elsonvertex);

	// for (int j = 0; j < ne; j++)
	for (int jj = 0; jj < elsonvertex.Size(); jj++)
	  {
	    int j = elsonvertex[jj];
	    ma.GetElEdges (j, ednums);
	    for (int k = 0; k < ednums.Size(); k++)
	      if (ednums[k] == i)
		{
		  elnr2 = elnr1; 
		  elnr1 = j;
		}
	  }
	// cout << "els = " << elnr1 << ", " << elnr2 << endl;

	//	(*testout) << "elnr1 = " << elnr1 << ", elnr2 = " << elnr2 << endl;
	if (elnr1 != -1)
	  {
	    // inner edge

	    Vec<2> p1, p2;
	    ma.GetPoint (vi1, p1);
	    ma.GetPoint (vi2, p2);

	    Vec<2> tau = p2 - p1;
	    Vec<2> n;
	    n(0) = tau(1); n(1) = -tau(0);
	    double h = L2Norm (tau);
	    n /= h;
	    tau /= h;


	    double alpha = 1.0 / (h+t);
	    double beta = 1 / sqrt (1/(t*t)-alpha*alpha);   // approx t for t << h

	    const IntegrationRule & ir = 
	      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 10);

	    double err1 = 0;
	    double err2 = 0;
	    double err3 = 0;
	    double err4 = 0;

	    for (int ipi = 0; ipi < ir.GetNIP(); ipi++)
	      {
		const IntegrationPoint & ip = ir[ipi];

		FlatVector<> diffstrain(7, lh), diffstress(7, lh);
		diffstrain = 0;
		diffstress = 0;
	   
		int nels = (elnr2 != -1) ? 2 : 1;
		for (int j = 0; j < nels; j++)
		  {
		    int elnr = (j == 0) ? elnr1 : elnr2;
		    double fac = (j == 0) ? 1 : -1;

		    const FiniteElement & fel = fes.GetFE(elnr, lh);
		    fes.GetDofNrs (elnr, dnums);
		    FlatVector<> elu(dnums.Size(), lh);
		    FlatVector<> strain(7, lh);
		    FlatVector<> stress(7, lh);
		    gfu.GetElementVector (dnums, elu);

		    ElementTransformation eltrans;
		    ma.GetElementTransformation (elnr, eltrans, lh);

		    
		    ma.GetElVertices (elnr, vnums);
		    const POINT3D *pelp1, *pelp2;
		    Vec<2> elp1, elp2;
		    for (int k = 0; k < vnums.Size(); k++)
		      {
			if (vnums[k] == vi1)
			  pelp1 = ElementTopology::GetVertices(ET_TRIG)+k;
			if (vnums[k] == vi2)
			  pelp2 = ElementTopology::GetVertices(ET_TRIG)+k;
		      }
		    for (int k = 0; k < 2; k++)
		      {
			elp1(k) = (*pelp1)[k];
			elp2(k) = (*pelp2)[k];
		      }

		    Vec<2> pref = elp1 + ip(0) * (elp2 - elp1);
		    IntegrationPoint ipel(pref(0), pref(1), 0, 0);
		    bli.CalcFlux (fel, eltrans, ipel, elu, strain, 0, lh);
		    bli.CalcFlux (fel, eltrans, ipel, elu, stress, 1, lh);
		    
		    // (*testout) << "elu = " << elu << endl;
		    // (*testout) << "strain = " << strain << endl;
		    // (*testout) << "stress = " << stress << endl;
		    
		    diffstrain += fac * strain;
		    diffstress += fac * stress;
		  }


		Vec<2> diffr;
		diffr(0) = diffstress(5);
		diffr(1) = diffstress(6);
		double diffrtau = InnerProduct(tau, diffr);

		double diffstabn = n(0)*diffstress(3) + n(1)*diffstress(4);

		if (nels == 2)
		  err1 += ip.Weight() * h * h *
		    (diffstress(0) * diffstress(0) + 
		     diffstress(1) * diffstress(1) + 
		     2 * diffstress(2) * diffstress(2));
		
		err2 += ip.Weight() * h * 
		  h / (beta * (h+beta)) * diffrtau * diffrtau;
		
		if (nels == 2)
		  err3 += ip.Weight() * h * 
		    h/sqr(alpha) * diffstabn * diffstabn;
		
		lh.CleanUp();
	      }
	    
	    verr.FVDouble()(elnr1) += err1+err2+err3;
	    verr1.FVDouble()(elnr1) += err1;
	    verr2.FVDouble()(elnr1) += err2;
	    verr3.FVDouble()(elnr1) += err3;

	    if (elnr2 != -1)
	      {
		verr.FVDouble()(elnr2) += err1+err2+err3;
		verr1.FVDouble()(elnr2) += err1;
		verr2.FVDouble()(elnr2) += err2;
		verr3.FVDouble()(elnr2) += err3;
	      }
	  }
      }

    verr += verr_el1;
    verr += verr_el2;
    verr += verr_el3;
    verr += verr_el4;

    double sum;


    static ofstream outfile ("rmerror.out");
    outfile << setw(7) << fes.GetNDof();

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr_el1.FVDouble()(i);
    cout << "element error1 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr_el2.FVDouble()(i);
    cout << "element error2 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr_el3.FVDouble()(i);
    cout << "element elerror3 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr_el4.FVDouble()(i);
    cout << "element elerror4 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);


    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr1.FVDouble()(i);
    cout << "edge error1 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr2.FVDouble()(i);
    cout << "edge error2 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr3.FVDouble()(i);
    cout << "edge error3 = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);


    sum = 0;
    for (int i = 0; i < ne; i++)
      sum += verr.FVDouble()(i);
    cout << "Total error = " << sqrt (sum) << endl;
    outfile << setw(12) << sqrt(sum);

    outfile << endl;

  }
};








namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("cserrorestimator", NumProcCSEE::Create);
    GetFESpaceClasses().AddFESpace ("reissnermindlin", FESpaceChapelleStenberg::Create);

    GetIntegrators().AddBFIntegrator ("reissnermindlin", 2, 3,
				      RMIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("reissnermindlinid", 2, 0,
				      RMIdIntegrator::Create);
  }
  
  Init init;
}
