#include <solve.hpp>


namespace ngsolve
{
  using namespace ngsolve;

  // contact element: vertex - cell
  template <int DIM>
  class ContactElement : public SpecialElement
  {
    ///
    const MeshAccess & ma;
    /// the distance function (logical dist to boundary)
    const FlatVector<double> & dist;
    /// displacement function 
    const FlatVector<Vec<DIM> > & u;
    /// point index, vertex
    int pi;
    /// point indices, cell
    int el_points[DIM+1];
    ///
    double penalty;
  public:
    ContactElement (const MeshAccess & ama,
		    const FlatVector<double> & adist,
		    const FlatVector<Vec<DIM> > & au,
		    int api, int * ael_points,
		    double apenalty)
      : ma(ama), dist(adist), u(au), pi(api), penalty(apenalty)
    {
      for (int i = 0; i <= DIM; i++)
	el_points[i] = ael_points[i];
    }

    virtual void GetDofNrs (ARRAY<int> & dnums) const
    {
      dnums.SetSize (DIM+2);
      dnums[0] = pi;
      for (int i = 0; i <= DIM; i++)
	dnums[i+1] = el_points[i];
    }

    virtual double Energy (const FlatVector<double> & elx, 
			   LocalHeap & lh) const
    {
      int i, j, k;

      FlatVector<Vec<DIM> > selx (DIM+2, 
				  static_cast<Vec<DIM>*> 
				  (const_cast<void*> (elx.Data())));

      Vec<DIM> p, p0;
      Mat<DIM, DIM+1> pmat;
      Mat<DIM> mat;
      Vec<DIM> rhs, sol;

      ma.GetPoint (pi, p);
      p += selx(0);

      ma.GetPoint (el_points[0], p0);
      p0 += selx(1);

      rhs = p - p0;
      for (i = 0; i < DIM; i++)
	{
	  Vec<DIM> pi;
	  ma.GetPoint (el_points[i+1], pi);
	  pi += selx(i+2);
	  pi -= p0;

	  for (k = 0; k < DIM; k++)
	    mat(k, i) = pi(k);
	}
      
      sol = Inv (mat) * rhs;

      double energy = 0;
      energy = dist(el_points[0]);
      for (i = 0; i < DIM; i++)
	energy += sol(i) * (dist(el_points[i+1]) - dist(el_points[0]));

      if (energy == 0)
	{
	  energy = 1;
	  for (j = 0; j < DIM; j++)
	    energy -= sol(j);
	  for (j = 0; j < DIM; j++)
	    if (sol(j) < energy)
	      energy = sol(j);
	}

      if (energy < 0) energy = 0;

      return penalty * sqr (energy);

      //return /*1e8* */ 1e-6 * sqr (energy); //pendel.pde
      //return 0.0000001 * sqr (energy);
    }




    void Assemble (FlatMatrix<double> & elmat,
		   LocalHeap & lh) const
    {
      int i, j;
      int nd = DIM+2;

      elmat.AssignMemory (DIM*nd, DIM*nd, lh);
      ARRAY<int> dnums;
      
      GetDofNrs (dnums);

      FlatVector<double> elx (DIM*nd, lh);
      FlatVector<Vec<DIM> > selx (nd, static_cast<Vec<DIM>*> (elx.Data()));
      for (j = 0; j < nd; j++)
	selx(j) = u(dnums[j]);
    
    
      FlatVector<double> hx1(elx.Size(), lh);
      FlatVector<double> hx2(elx.Size(), lh);
      FlatVector<double> hx3(elx.Size(), lh);
      FlatVector<double> hx4(elx.Size(), lh);

      double dx = 1e-12 + 1e-6 * L2Norm (elx);

      for (i = 0; i < elx.Size(); i++)
	for (j = 0; j < elx.Size(); j++)
	  {
	    hx1 = elx;
	    hx2 = elx;
	    hx3 = elx;
	    hx4 = elx;
	    
	    hx1(i) += dx;
	    hx2(i) -= dx;
	    hx3(i) += dx;
	    hx4(i) -= dx;
	    hx1(j) += dx;
	    hx2(j) += dx;
	    hx3(j) -= dx;
	    hx4(j) -= dx;
	    
	    elmat(i,j) = 
	      (Energy (hx1, lh) + Energy(hx4, lh) - Energy(hx2,lh) - Energy(hx3,lh))
	      / (4 * dx * dx);
	  }
      //(*testout) << "contact, elmat = " << endl << elmat << endl;
    }
  };

  void ComputeDistanceFunction (const MeshAccess & ma,
				FlatVector<double> & dist)
  {
    int i, j, k;
    dist = 1e12;
    ARRAY<int> pnums;

    int nse = ma.GetNSE();
    int ne = ma.GetNE();
    int np = ma.GetNP();

    ARRAY<int> nodeused(np);
    nodeused = 0;

    for (i = 0; i < nse; i++)
      {
	ma.GetSElPNums (i, pnums);
	for (j = 0; j < pnums.Size(); j++)
	  dist(pnums[j]) = 0;
      }

    bool change;
    do
      {
	change = 0;
	for (i = 0; i < ne; i++)
	  {
	    ma.GetElPNums (i, pnums);
	    
	    double minval = dist(pnums[0]);
	    for (j = 1; j < pnums.Size(); j++)
	      if (dist(pnums[j]) < minval)
		minval = dist(pnums[j]);
	    
	    minval++;
	    for (j = 1; j < pnums.Size(); j++)
	      if (minval < dist(pnums[j]))
		{
		  dist(pnums[j]) = minval;
		  change = 1;
		}
	  }
      }
    while (change);

    /*
    for (i = 0; i < ne; i++)
      {
        int ind = 1+ma.GetElIndex(i);
	    
	double fact = 1;
	if (ind == 3) fact = 20;

	for (j = 0; j < pnums.Size(); j++)
	{
      	  if (!nodeused[pnums[j]])
	  {
            nodeused[pnums[j]] = 1;
	    dist(pnums[j]) *= fact;
	  }
	}
      }
    */
  }

  
  template <int DIM>
  void FindOverlap (const MeshAccess & ma,
		    FESpace & fes,
		    FlatVector<Vec<DIM> > & u,
		    FlatVector<double> & dist)
  {
    int i, j, k, l;

    int np = ma.GetNV();
    int ne = ma.GetNE();
    int nse = ma.GetNSE();

    Vector<Vec<DIM> > points(np);
    BitArray bnd_pts(np);
    ARRAY<int> pnums;


    // remove old contact elements
    for (i = 0; i < fes.specialelements.Size(); i++)
      delete fes.specialelements[i];
    fes.specialelements.SetSize (0);


    // find vertices on boundary
    bnd_pts.Clear();
    for (i = 0; i < nse; i++)
      {
	ma.GetSElPNums (i, pnums);
	for (j = 0; j < pnums.Size(); j++)
	  if (pnums[j] < np)
	    bnd_pts.Set (pnums[j]);
      }

    for (i = 0; i < np; i++)
      {
	Vec<DIM> p;
	ma.GetPoint (i, p);
	points(i) = p + u(i);
      }


    // compute bounding boxes
    ARRAY<Vec<2*DIM> > bboxes(ne);
    for (i = 0; i < ne; i++)
      {
	ma.GetElPNums (i, pnums);
	Vec<DIM> p, pmin, pmax;
	pmin = 0;
	pmax = 0;
	for (j = 0; j < DIM+1; j++)
	  {
	    p = points (pnums[j]);
	    for (k = 0; k < DIM; k++)
	      {
		if (j == 0 || p(k) < pmin(k))
		  pmin(k) = p(k);
		if (j == 0 || p(k) > pmax(k))
		  pmax(k) = p(k);
	      }
	  }
	for (j = 0; j < DIM; j++)
	  {
	    bboxes[i](2*j) = pmin(j);
	    bboxes[i](2*j+1) = pmax(j);
	  }
      }

    for (i = 0; i < np; i++)
      {
	Vec<DIM> p;

	if (!bnd_pts.Test(i)) continue;
	p = points(i);

	for (j = 0; j < ne; j++)
	  {
	    bool outside = 0;
	    for (k = 0; k < DIM; k++)
	      if (bboxes[j](2*k) > p(k) || bboxes[j](2*k+1) < p(k))
		outside = 1;
	    if (outside) continue;


	    ma.GetElPNums (j, pnums);

	    bool haspoint = false;
	    for (k = 0; k < pnums.Size(); k++)
	      if (pnums[k] == i) haspoint = true;
	    if (haspoint) continue;


	    Mat<DIM> mat;
	    Vec<DIM> pk, p0, rhs, sol;

	    p0 = points(pnums[0]);
	    for (k = 0; k < DIM; k++)
	      {
		pk = points (pnums[k+1]) - p0;
		for (l = 0; l < DIM; l++)
		  mat(l,k) = pk(l);
	      }
	    
	    rhs = p - p0;
	    sol = Inv (mat) * rhs;

	    int index = 1+ma.GetElIndex(j);
	    double penalty = 200;
	    if (index == 3)
	      penalty = 200;
	    
	    if (index == 5)
	      penalty = 1e5;

	    penalty = 1;
	    if (DIM == 2)
	      {
		if (sol(0) >= 0 && sol(1) >= 0 && sol(0)+sol(1) <= 1)
		  {
		    fes.specialelements.Append 
		      (new ContactElement<DIM>(ma, dist, u, i, &pnums[0], penalty)); 
		    
		    //		cout << "point " << p << " in element !!!" << endl;
		  }
	      }
	    else
	      if (sol(0) >= 0 && sol(1) >= 0 && sol(2) >= 0 &&
		  sol(0)+sol(1)+sol(2) <= 1)
		{
		  fes.specialelements.Append 
		    (new ContactElement<DIM>(ma, dist, u, i, &pnums[0], penalty)); 
		  
		  //		cout << "point " << p << " in element !!!" << endl;
		}
	  }
      }
  }
  
  template 
  void FindOverlap<2> (const MeshAccess & ma,
		       FESpace & fes,
		       FlatVector<Vec<2> > & u,
		       FlatVector<double> & dist);
		    
  template 
  void FindOverlap<3> (const MeshAccess & ma,
		       FESpace & fes,
		       FlatVector<Vec<3> > & u,
		       FlatVector<double> & dist);
		    
}
