#include <solve.hpp>

using namespace ngsolve;



/*
  Solver for density function theory
*/



class NumProcDFT : public NumProc
{
protected:
  // bilinear-form for the Hamiltonian  (H-matrix)
  BilinearForm * bf_h;
  // bilinear-form for the overlap matrix (S-matrix)
  BilinearForm * bf_s;

  // bilinear-form for the potential problem
  BilinearForm * bf_pot;
  // mass matrix for the density function rho
  BilinearForm * bf_mass;

  // linear-form for the back-ground charge density
  LinearForm * lf_rhobg;

  // solution vectors
  GridFunction * gf_psi;
  GridFunction * gf_rho;
  GridFunction * gf_pot;


  int dftits, evpits;

  // help vectors for the evp
  ARRAY<BaseVector*> vecui;

  int pointnr;
  double rho_proton;

public:

  NumProcDFT (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bf_h = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformh", "h"));
    bf_s = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforms", "s"));

    bf_pot = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformpot", "bfpot"));
    bf_mass = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformmass", "bfmass"));

    lf_rhobg = pde.GetLinearForm (flags.GetStringFlag ("linearform", "rhobg"), 1);
    gf_psi = pde.GetGridFunction (flags.GetStringFlag ("gridfunctionpsi", "psi"));
    gf_rho = pde.GetGridFunction (flags.GetStringFlag ("gridfunctionrho", "rho"));
    gf_pot = pde.GetGridFunction (flags.GetStringFlag ("gridfunctionpot", "pot"));
    
    dftits = int (flags.GetNumFlag ("dftits", 20));
    evpits = int (flags.GetNumFlag ("evpits", 20));
    pointnr = int (flags.GetNumFlag ("pointnr", 0));
    rho_proton = flags.GetNumFlag ("rhoproton", -1);
  }

  virtual ~NumProcDFT() 
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcDFT (pde, flags);
  }



  virtual void Do(LocalHeap & lh)
  {
    cout << "solve DFT" << endl;
    
    int num = gf_psi->GetMultiDim();

    for (int i = 0; i < num; i++)
      gf_psi->GetVector(i).SetRandom();

    vecui.SetSize (2*num);
    for (int i = 0; i < 2*num; i++)
      vecui[i] = gf_psi->GetVector().CreateVector();

    Vector<> lami(2*num);

    // charge density functional
    BaseVector & rho_fct = *gf_psi->GetVector().CreateVector();

    BaseMatrix * inv_pot = dynamic_cast<const BaseSparseMatrix&> (bf_pot->GetMatrix()).InverseMatrix();
    BaseMatrix * inv_mass = dynamic_cast<const BaseSparseMatrix&> (bf_mass->GetMatrix()).InverseMatrix();


    GridFunctionCoefficientFunction coeffunc(*gf_pot);
    if (ma.GetDimension() == 2)
      bf_h -> AddIntegrator (new MassIntegrator<2> (&coeffunc));
    else
      bf_h -> AddIntegrator (new MassIntegrator<3> (&coeffunc));


    cout.precision(10);

    double lam_old1 = 1e8;

    for (int dftit = 1; dftit <= dftits; dftit++)
      {
	cout << endl << "DFT iteration " << dftit << endl;

	bf_h -> ReAssemble(lh);

	const BaseMatrix & math = bf_h->GetMatrix();
	const BaseMatrix & mats = bf_s->GetMatrix();
	
	BaseMatrix * preh = dynamic_cast<const BaseSparseMatrix&> (math).InverseMatrix();


	SolveEVP(math, mats, *preh, lami);

	delete preh;

	// compute new potential
	if (ma.GetDimension() == 2)
	  CalcChargeDensity<2> (rho_fct, lh);
	else
	  CalcChargeDensity<3> (rho_fct, lh);

	gf_rho->GetVector() = (*inv_mass) * rho_fct;
	
	if (lf_rhobg)
	  rho_fct += lf_rhobg->GetVector();
	if (pointnr > 0)
	  rho_fct.FVDouble()(pointnr-1) += rho_proton;

	gf_pot->GetVector() = (*inv_pot) * rho_fct;

	if (fabs (lami(num-1)-lam_old1) < 1e-8 * fabs(lami(num-1))) break;
	lam_old1 = lami(num-1);
      }
  }


  void SolveEVP (const BaseMatrix & math, 
		 const BaseMatrix & mats,
		 const BaseMatrix & preh,
		 Vector<> & lami)
  {
    int num = gf_psi->GetMultiDim();

    BaseVector & d = *gf_psi->GetVector().CreateVector();
    BaseVector & w = *gf_psi->GetVector().CreateVector();
    BaseVector & hu = *gf_psi->GetVector().CreateVector();
    BaseVector & su = *gf_psi->GetVector().CreateVector();

    int nv = ma.GetNV();

    Matrix<double> small_h(2*num), small_s(2*num);
    Matrix<double> small_h2(num), small_s2(num);
    Matrix<double> cg_h(nv+num), cg_s(nv+num), cg_h2(nv+num), cg_s2(nv+num);


    Vector<> lami2(nv+num);

    const SparseMatrix<double> & smath = dynamic_cast<const SparseMatrix<double> &> (math);
    const SparseMatrix<double> & smats = dynamic_cast<const SparseMatrix<double> &> (mats);

    // Matrix<double> coarseh(nv), coarsehinv(nv);

    cg_h2 = 0;
    cg_s2 = 0;
    for (int i = 0; i < nv; i++)
      for (int j = 0; j < nv; j++)
	{
	  cg_h2(num+i,num+j) = smath(i,j);
	  cg_s2(num+i,num+j) = smats(i,j);
	}
    



    double lam_old = 1e10;

    for (int it = 0; it < evpits; it++)
      {
	cout << "it = " << it << " " << flush;

	for (int ei = 0; ei < num; ei++)
	  *vecui[ei] = gf_psi->GetVector(ei);
	
	for (int ei = 0; ei < num; ei++)
	  {
	    hu = math * *vecui[ei];
	    su = mats * *vecui[ei];
		
	    double lam = 
	      InnerProduct (*vecui[ei], hu) /
	      InnerProduct (*vecui[ei], su);
	    
	    d = hu - lam * su;
	    w = preh * d;
	    double norm = sqrt (InnerProduct (w, d=mats*w));
	    *vecui[num+ei] = (1.0/norm) * w;
	    
	    for (int ej  = 0; ej <= ei; ej++)
	      {
		small_h(ei, ej) = small_h(ej,ei) = InnerProduct (hu, *vecui[ej]);
		small_s(ei, ej) = small_s(ej,ei) = InnerProduct (su, *vecui[ej]);
	      }
	  }

	for (int ei = num; ei < 2*num; ei++)
	  {
	    hu = math * *vecui[ei];
	    su = mats * *vecui[ei];
	    
	    for (int ej  = 0; ej <= ei; ej++)
	      {
		small_h(ei, ej) = small_h(ej,ei) = InnerProduct (hu, *vecui[ej]);
		small_s(ei, ej) = small_s(ej,ei) = InnerProduct (su, *vecui[ej]);
	      }
	  }
	
	LapackGHEPEPairs (2*num, &small_h(0,0), &small_s(0,0), &lami(0));
	// cout << "lami = " << lami << endl;	    

	for (int ei = 0; ei < num; ei++)
	  {
	    hu = 0;
	    for (int i = 0; i < 2*num; i++)
	      hu += small_h(ei,i) * *vecui[i];
	    
	    gf_psi->GetVector(ei) = hu;
	  }




	
	for (int ei = 0; ei < num; ei++)
	  *vecui[ei] = gf_psi->GetVector(ei);

	/*
	for (int i = 0; i < nv; i++)
	  for (int j = 0; j < nv; j++)
	    coarseh(i,j) = smath(i,j);
	CalcInverse (coarseh, coarsehinv);
	Vector<> dc(nv), wc(nv);

	for (int ei = 0; ei < num; ei++)
	  {
	    hu = math * *vecui[ei];
	    dc = hu.FVDouble().Range(0, nv-1);
	    wc = coarsehinv * dc;
	    (*vecui[ei]).FVDouble().Range(0,nv-1) -= wc;
	  }
	*/


	cg_h = cg_h2;
	cg_s = cg_s2;

	for (int ei = 0; ei < num; ei++)
	  {
	    hu = math * *vecui[ei];
	    su = mats * *vecui[ei];
		
	    for (int ej  = 0; ej <= ei; ej++)
	      {
		cg_h(ei, ej) = cg_h(ej,ei) = InnerProduct (hu, *vecui[ej]);
		cg_s(ei, ej) = cg_s(ej,ei) = InnerProduct (su, *vecui[ej]);
	      }

	    for (int ej = 0; ej < nv; ej++)
	      {
		cg_h(num+ej,ei) = cg_h(ei,num+ej) = hu.FVDouble()(ej);
		cg_s(num+ej,ei) = cg_s(ei,num+ej) = su.FVDouble()(ej);
	      }
	  }


	/*
	for (int ei = 0; ei < num; ei++)
	  for (int ej = 0; ej < num; ej++)
	    {
	      small_h2(ei,ej) = cg_h(ei,ej);
	      small_s2(ei,ej) = cg_s(ei,ej);
	    }

	(*testout) << "cg_h = " << endl << cg_h << endl;

	LapackGHEPEPairs (num, &small_h2(0,0), &small_s2(0,0), &lami(0));
	cout << "lam, complement = " << lami.Range(0,num-1) << endl;
	*/


	LapackGHEPEPairs (num+nv, &cg_h(0,0), &cg_s(0,0), &lami2(0));
	// cout << "lami2 = " << lami2 << endl;

	for (int ei = 0; ei < num; ei++)
	  {
	    hu = 0;
	    for (int i = 0; i < num; i++)
	      hu += cg_h(ei,i) * *vecui[i];
	    for (int i = 0; i < nv; i++)
	      hu.FVDouble()(i) += cg_h(ei,num+i);

	    gf_psi->GetVector(ei) = hu;
	  }

	// cout << " lam_max = " << lami(num-1);
	// cout << endl;

	cout << " diff = " << lami(num-1)-lam_old << endl;
	if (fabs (lami(num-1)-lam_old) < 1e-6 * fabs(lami(num-1)) && it >= 4) break;
	lam_old = lami(num-1);
      }
    
    for (int ei = 0; ei < num; ei++)
      cout << "lam(" << ei << ") = " << lami(ei) << endl;

    /*    
    for (int ei = 0; ei < num; ei++)
      gf_psi->GetVector(ei) = *vecui[ei];
    */

    delete &d;
    delete &w;
    delete &hu;
    delete &su;
  }


  
  // calculates the functionals
  template<int DIM>
  void CalcChargeDensity (BaseVector & rho, LocalHeap & lh)
  {
    // compute new density
    
    /*
      // for vertex dofs only 
    FlatVector<double> fvrho = gf_rho->GetVector().FVDouble();
    int num = gf_psi->GetMultiDim();
    
    fvrho = 0;
    for (int ei = 0; ei < num; ei++)
      {
	FlatVector<double> fvpsi = gf_psi->GetVector(ei).FVDouble();
	
	for (int j = 0; j < fvrho.Size(); j++)
	  fvrho(j) += sqr (fvpsi(j));
      }
    
    rho = bf_mass->GetMatrix() * gf_rho->GetVector();
    */



    const FESpace & fes = gf_pot->GetFESpace();
    const MeshAccess & ma = fes.GetMeshAccess();

    int num = gf_psi->GetMultiDim();

    ElementTransformation eltrans;
    ARRAY<int> dnums;

    rho = 0;

    void * heapp = lh.GetPointer();
    for (int i = 0; i < ma.GetNE(); i++)
      {
	const NodalFiniteElement & fel = 
	  dynamic_cast<const NodalFiniteElement&> (fes.GetFE(i, lh));

	const IntegrationRule & ir = 
	  GetIntegrationRules().SelectIntegrationRule(fel.ElementType(), 2*fel.Order());

	fes.GetDofNrs (i, dnums);
	ma.GetElementTransformation (i, eltrans, lh);

	FlatMatrix<> elvecs(num, fel.GetNDof(), lh);
	for (int j = 0; j < num; j++)
	  {
	    FlatVector<double> row = elvecs.Row(j);
	    const BaseVector & psii = gf_psi->GetVector(j);
	    psii.GetIndirect (dnums, row);
	  }


	FlatMatrix<> shapes (ir.GetNIP(), fel.GetNDof(), lh);
	for (int j = 0; j < ir.GetNIP(); j++)
	  fel.CalcShape (ir[j], shapes.Row(j));
	
	
	FlatMatrix<> ipvalues (ir.GetNIP(), num, lh);
	ipvalues = shapes * Trans(elvecs);

	FlatVector<> sum(fel.GetNDof(), lh);
	sum = 0;
	for (int j = 0; j < ir.GetNIP(); j++)
	  {
	    double s = 0;
	    for (int k = 0; k < num; k++)
	      s += sqr (ipvalues(j,k));
	    
	    SpecificIntegrationPoint<DIM,DIM> sip(ir[j], eltrans, lh);
	    double fac = s * ir[j].Weight() * fabs (sip.GetJacobiDet());

	    sum += fac * shapes.Row(j);
	  }

	rho.AddIndirect (dnums, sum);

	lh.CleanUp (heapp);
      }
  }
  





  virtual string GetClassName () const
  {
    return "DFT";
  }

  virtual void PrintReport (ostream & ost)
  {
    /*
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Bilinear-form M = " << bfm->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl
	<< "dt              = " << dt << endl
	<< "tend            = " << tend << endl;
    */
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*
    ost << 
      "\n\nNumproc Parabolic:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};





// declare the numproc 'parabolic' 

namespace
{
  class Init
  { 
  public: 
    Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("dft", NumProcDFT::Create, 0);
  }
    
  Init init;
}
  
