/*

EAS solver

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace **************************** */


class EASFiniteElement : public FiniteElement
{
public:
  virtual void CalcShapes (const IntegrationPoint & ip, 
			   FlatMatrix<> shape) const = 0;
};


class EASQuad : public EASFiniteElement
{
public:
  EASQuad (int aorder)
  {
    order = aorder;
    ndof = 4*order;
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
    for (int i = 0; i < ndof; i++)
      idofs.Append (i);
  }

  virtual void CalcShapes (const IntegrationPoint & ip, 
			   FlatMatrix<> shape) const   // n * 3 
  {
    double x = ip(0);
    double y = ip(1);

    ArrayMem<double, 10> polx(order+1), poly(order+1);

    LegendrePolynomial (order, 2*x-1, polx);
    LegendrePolynomial (order, 2*y-1, poly);

    shape = 0.0;
    for (int i = 0; i < order; i++)
      {
	shape(i,2) = polx[order] * poly[i];
	shape(order+i,2) = polx[i] * poly[order];

	shape(2*order+i,0) = polx[order] * poly[i];
	shape(3*order+i,1) = polx[i] * poly[order];
      }
    /*
    shape(0, 0) = 0;    // eps11
    shape(0, 1) = 0;    // eps22
    shape(0, 2) = x-0.5; 

    shape(1, 0) = 0;    // eps11
    shape(1, 1) = 0;    // eps22
    shape(1, 2) = y-0.5; 
    */
  }
};


// just to define it
class EASTrig : public EASFiniteElement
{
public:
  EASTrig (int aorder)
  {
    order = aorder;
    ndof = 0;
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
  }

  virtual void CalcShapes (const IntegrationPoint & ip, 
			   FlatMatrix<> shape) const   // n * 3
  {
    ;
  }
};



#ifdef ORIG

class EASPrism : public EASFiniteElement
{
  int orderz;
public:
  EASPrism (int aorder, int aorderz)
  {
    order = aorder;
    orderz = aorderz;

    cout << "prism: order = " << order << ", orderz = " << orderz << endl;

    eltype = ET_PRISM;
    dimspace = 3;

    switch (order)
      {
      case 1:
	ndof = 6;
	break;
      default:
	ndof = 3 * 3 * (orderz+1) + 3 * (order) * (order+1) / 2;
	break;
      }
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
    for (int i = 0; i < ndof; i++)
      idofs.Append (i);
  }

  virtual void CalcShapes (const IntegrationPoint & ip, 
			   FlatMatrix<> shape) const   // n * 6
  {
    double x = ip(0);
    double y = ip(1);
    double z = ip(2);

    shape = 0.0;
    
    if (order == 1)
      {
	// strain of  u_z = x y, u_z = x(1-x-y), u_z = y(1-x-y)
	
	shape(0, 4) = y;
	shape(0, 5) = x;
	
	shape(1, 4) = 1-2*x-y;
	shape(1, 5) = -x;
	
	shape(2, 4) = -y;
	shape(2, 5) = 1-x-2*y;
	
	shape(3, 4) = z-0.5;
	shape(4, 5) = z-0.5;
	shape(5, 2) = z-0.5;
      }
    else
      {
	AutoDiff<3> adx(x, 0), ady(y, 1), adz(z,2);
	
	int ii = 0;

	ArrayMem< AutoDiff<3>, 10 > pol(order+10);
	ArrayMem< AutoDiff<3>, 10 > polx(order+10);
	ArrayMem< AutoDiff<3>, 10 > poly(order+10);
	ArrayMem< AutoDiff<3>, 10 > polz(order+10);

	LegendrePolynomial (order, 2*adz-1, polz);

	AutoDiff<3> lami[] = { adx, ady, 1-adx-ady };
	for (int i = 0; i < 3; i++)
	  {
	    AutoDiff<3> l1 = lami[i];
	    AutoDiff<3> l2 = lami[(i+1)%3];

	    ScaledIntegratedLegendrePolynomial(order+9, l1-l2, l1+l2, pol);

	    /*
	    for (int k = order+1; k <= order+2; k++)
	      for (int j = 0; j <= orderz; j++)
		{
		  AutoDiff<3> val = pol[k] * polz[j];

		  // phi_x
		  shape(ii, 0) = val.DValue(0);
		  shape(ii, 3) = val.DValue(1);
		  shape(ii, 4) = val.DValue(2);
		  ii++;
		  
		  // phi_y
		  shape(ii, 3) = val.DValue(0);
		  shape(ii, 1) = val.DValue(1);
		  shape(ii, 5) = val.DValue(2);
		  ii++;

		  // phi_z
		  shape(ii, 4) = val.DValue(0);
		  shape(ii, 5) = val.DValue(1);
		  shape(ii, 2) = val.DValue(2);
		  ii++;
		}
	    */

	    for (int j = 0; j <= orderz; j++)
	      {
		AutoDiff<3> val = pol[order+2] * polz[j];
		
		// phi_x
		shape(ii, 0) = val.DValue(0);
		shape(ii, 3) = val.DValue(1);
		shape(ii, 4) = val.DValue(2);
		ii++;
		
		// phi_y
		shape(ii, 3) = val.DValue(0);
		shape(ii, 1) = val.DValue(1);
		shape(ii, 5) = val.DValue(2);
		ii++;
		

		val = pol[order+1] * polz[j];
		// phi_z
		shape(ii, 4) = val.DValue(0);
		shape(ii, 5) = val.DValue(1);
		shape(ii, 2) = val.DValue(2);
		ii++;
	      }
	  }
	

	ScaledLegendrePolynomial (order+2, lami[2]-lami[0], lami[2]+lami[0], polx);
	LegendrePolynomial (order+2, 2*lami[1]-1, poly);
	IntegratedLegendrePolynomial (order+1, 2*adz-1, polz);

	for (int j = 0; j <= order-1; j++)
	  for (int k = 0; k <= order-1-j; k++)
	    {
	      AutoDiff<3> val = polx[j] * poly[k] * polz[orderz+1];

	      // phi_x
	      shape(ii, 0) = val.DValue(0);
	      shape(ii, 3) = val.DValue(1);
	      shape(ii, 4) = val.DValue(2);
	      ii++;
	      
	      // phi_y
	      shape(ii, 3) = val.DValue(0);
	      shape(ii, 1) = val.DValue(1);
	      shape(ii, 5) = val.DValue(2);
	      ii++;
	      
	      // phi_z
	      shape(ii, 4) = val.DValue(0);
	      shape(ii, 5) = val.DValue(1);
	      shape(ii, 2) = val.DValue(2);
	      ii++;
	    }
	
	if (ii != ndof) 
	  {
	    cout << "EAS prism: ndof is wrong" << endl;
	    cout << "ndof = " << ndof << "; ii = " << ii << endl;
	    exit(1);
	  }
      }
  }
};

#endif


#define VER2
#ifdef VER2

class EASPrism : public EASFiniteElement
{
  int orderz;
public:
  EASPrism (int aorder, int aorderz)
  {
    order = aorder;
    orderz = aorderz;


    eltype = ET_PRISM;
    dimspace = 3;

    switch (order)
      {
      case 1:
	ndof = 6;
	break;
      default:
	ndof = 3 + 2 * (3*(order+1)+3);
	break;
      }
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
    for (int i = 0; i < ndof; i++)
      idofs.Append (i);
  }

  virtual void CalcShapes (const IntegrationPoint & ip, 
			   FlatMatrix<> shape) const   // n * 6
  {
    double x = ip(0);
    double y = ip(1);
    double z = ip(2);

    shape = 0.0;
    
    if (order == 1)
      {
	// strain of  u_z = x y, u_z = x(1-x-y), u_z = y(1-x-y)
	
	shape(0, 4) = y;
	shape(0, 5) = x;
	
	shape(1, 4) = 1-2*x-y;
	shape(1, 5) = -x;
	
	shape(2, 4) = -y;
	shape(2, 5) = 1-x-2*y;
	
	shape(3, 4) = z-0.5;
	shape(4, 5) = z-0.5;
	shape(5, 2) = z-0.5;
      }
    else
      {
	AutoDiff<3> adx(x, 0), ady(y, 1), adz(z,2);
	
	int ii = 0;

	ArrayMem< double, 10 > pol(order+10);
	ArrayMem< double, 10 > poly(order+10);
	ArrayMem< AutoDiff<3>, 10 > polz(order+10);

	ArrayMem< AutoDiff<3>, 10 > adpol(order+10);

	double li[] = { x, y, 1-x-y };
	AutoDiff<3> lami[] = { adx, ady, 1-adx-ady };


	// nabla w^p+1
	for (int i = 0; i < 3; i++)
	  {
	    AutoDiff<3> l1 = lami[i];
	    AutoDiff<3> l2 = lami[(i+1)%3];

	    ScaledIntegratedLegendrePolynomial(order+1, l1-l2, l1+l2, adpol);

	    shape (ii, 4) = adpol[order+1].DValue(0);
	    shape (ii, 5) = adpol[order+1].DValue(1);
	    ii++;
	  }

	ScaledLegendrePolynomial(order+2, li[0]-li[2], li[0]+li[2], pol);

	for (int i = 0; i <= order+2; i++)
	  {
	    JacobiPolynomial (order+2, 2*li[1]-1, 2*i+1, 0, poly);

	    for (int j = max2 (order-i, 0); j <= order+2-i; j++)
	      {
		shape(ii,4) = pol[i] * poly[j];
		ii++;
		shape(ii,5) = pol[i] * poly[j];
		ii++;
	      }
	  }

	if (ii != ndof) 
	  {
	    cout << "EAS prism: ndof is wrong" << endl;
	    cout << "ndof = " << ndof << "; ii = " << ii << endl;
	    exit(1);
	  }
      }
  }
};


#endif








class DummySegm : public FiniteElement
{
public:
  DummySegm ()
  {
    eltype = ET_SEGM;
    ndof = 0; 
  }
};








class FESpaceEnhancedStrain : public FESpace
{
private:
  
  // Number of Elements
  int nel;

  ARRAY<int> first_element_dof;
  ARRAY<int> ndlevel;
  int ndof;
  int orderz;
public:

  FESpaceEnhancedStrain (const MeshAccess & ama, const Flags & aflags)
    : FESpace (ama, aflags)
  {
    orderz = int (aflags.GetNumFlag ("orderz", 2));

    if (ma.GetDimension() == 2)
      {
	segm = new DummySegm ();
	quad = new EASQuad (order);
	trig = new EASTrig (order);
      }
    else
      {
	trig = new DummySegm ();
	quad = new DummySegm ();
	prism = new EASPrism (order, orderz);
      }
  }

  virtual ~FESpaceEnhancedStrain ()
  {
    ;
  }

  virtual string GetClassName () const
  {
    return "FESpaceEnhancedStrain";
  }


  ///
  virtual void Update()
  {
    nel = ma.GetNE();
    ndof = 0;
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;

	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    break;
	  case ET_QUAD:
	    ndof += 4*order;
	    break;
	  case ET_PRISM:
	    ndof += prism->GetNDof();
	    break;
	  }
      }
    first_element_dof[nel] = ndof;    

    if (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
  }

  ///
  virtual int GetNDof () const
  {
    return ndof;
  }

  virtual int GetNDofLevel (int level) const
  {
    return ndlevel[level]; 
  }



  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {   
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);
  }



  virtual void GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs (elnr, dnums); }


  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
};


class FESpaceEAS : public CompoundFESpace
{
  int orderz;

public:
  FESpaceEAS (const MeshAccess & ama, 		   
	      const ARRAY<const FESpace*> & aspaces,
	      const Flags & aflags)
    : CompoundFESpace (ama, aspaces)
  {
    order = int (aflags.GetNumFlag ("order", 2));
    orderz = int (aflags.GetNumFlag ("orderz", 2));
    if (orderz > order) orderz = order;
  }

  virtual ~FESpaceEAS () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(ma.GetDimension()+1);

    cout << "Create EAS FE Space" << endl;
    int order = int (flags.GetNumFlag ("order", 2));
    int orderz = int (flags.GetNumFlag ("orderz", 2));
    
    Flags uflags;
    uflags.SetFlag ("order", order);
    if (flags.GetDefineFlag ("plate")) uflags.SetFlag ("plate");

    for (int i = 0; i < ma.GetDimension(); i++)
      spaces[i] = new H1HighOrderFESpace (ma, uflags);
    
    Flags epsflags;
    epsflags.SetFlag ("order", order);
    epsflags.SetFlag ("orderz", orderz);
    spaces[ma.GetDimension()] = new FESpaceEnhancedStrain (ma, epsflags);

    FESpaceEAS * fes = new FESpaceEAS (ma, spaces, flags);
    return fes;
  }
    
  virtual string GetClassName () const { return "EASFESpace"; }


  virtual void Update()
  {
    cummulative_nd.SetSize (spaces.Size()+1);
    cummulative_nd[0] = 0;

    int dim = ma.GetDimension();
    for (int ii = 0; ii < dim; ii++)
      {
	H1HighOrderFESpace * h1ho = dynamic_cast<H1HighOrderFESpace*>
	  (const_cast<FESpace*> (spaces[ii]));

	h1ho->Update();

	if (ma.GetDimension() == 3)
	  {
	    for (int i = 0; i < ma.GetNEdges(); i++)
	      {
		int pi1, pi2;
		ma.GetEdgePNums (i, pi1, pi2);
		if (ma.GetClusterRepVertex (pi1) == 
		    ma.GetClusterRepVertex (pi2))
		  {
		    h1ho->SetEdgeOrder (i, orderz);
		  }
		else
		  {
		    if (ii < 2)
		      h1ho->SetEdgeOrder (i, order);
		    else
		      h1ho->SetEdgeOrder (i, order);
		  }
	      }


	    ARRAY<int> pnums;
	    for (int i = 0; i < ma.GetNFaces(); i++)
	      {
		ma.GetFacePNums (i, pnums);
		if (pnums.Size() == 3)
		  {
		    if (i < 2)
		      h1ho->SetFaceOrder (i, order+2);
		    else
		      h1ho->SetFaceOrder (i, order+2);
		  }
		else
		  {
		    int fmax = 0; 
		    for (int j = 1; j < 4; j++)
		      if (pnums[j] > pnums[fmax]) fmax = j;
		    
		    int f1 = (fmax+3)%4; 
		    int f2 = (fmax+1)%4; 
		    if(pnums[f2] > pnums[f1]) swap(f1,f2);  // fmax > f1 > f2;
		    
		    int orderx = (ii < 2) ? order : order+1;


		    if (ma.GetClusterRepVertex (pnums[fmax]) == 
			ma.GetClusterRepVertex (pnums[f1]))
		      {
			h1ho->SetFaceOrder (i, orderz, orderx);
		      }
		    else if (ma.GetClusterRepVertex (pnums[fmax]) == 
			     ma.GetClusterRepVertex (pnums[f2]))
		      {
			h1ho->SetFaceOrder (i, orderx, orderz);
		      }
		    else
		      {
			cerr << "quad face has no cluster" << endl;
			h1ho->SetFaceOrder (i, order, order);
		      }
		  }
	      }

	    for (int i = 0; i < ma.GetNE(); i++)
	      h1ho->SetElementOrder (i, order, order, orderz);
	  }
	h1ho->UpdateDofTables();
	cout << "space has " << spaces[ii]->GetNDof() << " unknowns" << endl;

	cummulative_nd[ii+1] = cummulative_nd[ii] + spaces[ii]->GetNDof();
      }

    const_cast<FESpace*> (spaces[dim])->Update();
    cout << "eas : " << spaces[dim]->GetNDof() << endl;

    cummulative_nd[dim+1] = cummulative_nd[dim] + spaces[dim]->GetNDof();

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (cummulative_nd.Last());
  }

  Table<int> * 
  CreateSmoothingBlocks ( int type) const
  {
    int ne = ma.GetNE();
    ARRAY<int> cnt(ne);
    ARRAY<int> dnums;

    for (int i = 0; i < ne; i++)
      {
	GetDofNrs (i, dnums);
	cnt[i] = dnums.Size();
      }

    Table<int> & table = *new Table<int> (cnt); 
	
    for (int i = 0; i < ne; i++)
      {
	GetDofNrs (i, dnums);
	for (int j = 0; j < dnums.Size(); j++)
	  table[i][j] = dnums[j];
      }

    return &table;
  }
};




/* ********************+ Integrator ************************** */

template <int D> class DiffOpEAS;

template <>
class DiffOpEAS<2> : public DiffOp<DiffOpEAS<2> >
{

  // (compatible strains, enhance strains)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 3 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_u = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const EASFiniteElement & fel_eps = 
      dynamic_cast<const EASFiniteElement&> (cfel[2]);
    
    int nd_u = fel_u.GetNDof();
    int nd_eps = fel_eps.GetNDof();

    FlatMatrix<> gradu(2, nd_u, lh);
    gradu = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_u.GetDShape(sip.IP(), lh));

    mat = 0;
    for (int i = 0; i < nd_u; i++)
      {
	mat(0, i  ) = gradu(0, i);
	mat(2, i  ) = gradu(1, i);
	mat(1, nd_u+i) = gradu(1, i);
	mat(2, nd_u+i) = gradu(0, i);
      }

    FlatMatrix<> eastrain (nd_eps, 3, lh);
    fel_eps.CalcShapes (sip.IP(), eastrain);
    
    for (int i = 0; i < nd_eps; i++)
      {
	Mat<2> eps, epstrans;
	eps(0,0) = eastrain(i, 0);
	eps(1,1) = eastrain(i, 1);
	eps(0,1) = eps(1,0) = eastrain(i, 2) / 2;

	double det = sip.GetJacobiDet();
	epstrans = Trans (sip.GetJacobianInverse ()) * eps * sip.GetJacobianInverse ();

	mat(0, 2*nd_u+i) = epstrans(0,0);
	mat(1, 2*nd_u+i) = epstrans(1,1);
	mat(2, 2*nd_u+i) = 2*epstrans(0,1);
      }

    /*
    (*testout) << "ip = " << sip.IP() << endl
	       << "bmat = " << endl << mat << endl;
    */
  }
};




template <>
class DiffOpEAS<3> : public DiffOp<DiffOpEAS<3> >
{

  // (compatible strains, enhance strains)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 12 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_uxy = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_uz = 
      dynamic_cast<const NodalFiniteElement&> (cfel[2]);
    const EASFiniteElement & fel_eps = 
      dynamic_cast<const EASFiniteElement&> (cfel[3]);
    
    int nd_uxy = fel_uxy.GetNDof();
    int nd_uz = fel_uz.GetNDof();
    int nd_eps = fel_eps.GetNDof();

    FlatMatrix<> graduxy(3, nd_uxy, lh);
    graduxy = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_uxy.GetDShape(sip.IP(), lh));
    FlatMatrix<> graduz(3, nd_uz, lh);
    graduz = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_uz.GetDShape(sip.IP(), lh));

    mat = 0;
    for (int i = 0; i < nd_uxy; i++)
      {
	mat(0, i  ) = graduxy(0, i);
	mat(3, i  ) = graduxy(1, i);
	mat(4, i  ) = graduxy(2, i);

	mat(3, nd_uxy+i) = graduxy(0, i);
	mat(1, nd_uxy+i) = graduxy(1, i);
	mat(5, nd_uxy+i) = graduxy(2, i);
      }
    for (int i = 0; i < nd_uz; i++)
      {
	mat(4, 2*nd_uxy+i) = graduz(0, i);
	mat(5, 2*nd_uxy+i) = graduz(1, i);
	mat(2, 2*nd_uxy+i) = graduz(2, i);
      }

    FlatMatrix<> eastrain (nd_eps, 6, lh);
    fel_eps.CalcShapes (sip.IP(), eastrain);
    
    int nd_u = 2*nd_uxy+nd_uz;

    for (int i = 0; i < nd_eps; i++)
      {
	Mat<3> eps, epstrans;
	eps(0,0) = eastrain(i, 0);
	eps(1,1) = eastrain(i, 1);
	eps(2,2) = eastrain(i, 2);
	eps(0,1) = eps(1,0) = eastrain(i, 3) / 2;
	eps(0,2) = eps(2,0) = eastrain(i, 4) / 2;
	eps(1,2) = eps(2,1) = eastrain(i, 5) / 2;

	// (*testout) << "EAS-eps = " << endl << eps << endl;
	double det = sip.GetJacobiDet();
	epstrans = Trans (sip.GetJacobianInverse ()) * eps * sip.GetJacobianInverse ();
	// (*testout) << "epstrans = " << epstrans << endl;


	int base = (i < 3) ? 0 : 6;
	base = 6;
	mat(base+0, nd_u+i) = epstrans(0,0);
	mat(base+1, nd_u+i) = epstrans(1,1);
	mat(base+2, nd_u+i) = epstrans(2,2);
	mat(base+3, nd_u+i) = 2*epstrans(0,1);
	mat(base+4, nd_u+i) = 2*epstrans(0,2);
	mat(base+5, nd_u+i) = 2*epstrans(1,2);
      }

    Mat<3> jac = sip.GetJacobian();
    double h = sqrt (sqr (jac(0,0)) + sqr (jac(0,1)));
    double t = fabs (jac(2,2));

    /*
    double fac = t / (0.2*h+t);
    for (int i = 0; i < mat.Width(); i++)
      {
	mat(4,i) *= fac;
	mat(5,i) *= fac;
      }
    */

    /*
    (*testout) << "ip = " << sip.IP() << endl
	       << "bmat = " << endl << mat << endl;
    */
  }
};






/// 2D plane strain, and 3D
template <int DIM>
class ElasticityDMatEAS : public DMatOp<ElasticityDMatEAS<DIM> >
{
public:
  CoefficientFunction * coefe;
  CoefficientFunction * coefnu;
public:
  enum { DIM_DMAT = 2 * (DIM * (DIM+1)) / 2 };  

  ElasticityDMatEAS (CoefficientFunction * acoefe,
		  CoefficientFunction * acoefnu) 
    : coefe(acoefe), coefnu(acoefnu) { ; }

  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;
    double nu = Evaluate (*coefnu, sip);
    double e = Evaluate (*coefe, sip);

    for (int i = 0; i < DIM; i++)
      {
	mat(i,i) = 1-nu;
	for (int j = 0; j < i; j++)
	  mat(i,j) = mat(j,i) = nu;
      }
    for (int i = DIM; i < (DIM*(DIM+1)/2); i++)
      mat(i,i) = 0.5 * (1-2*nu);

    mat *= (e / ((1 + nu) * (1 - 2 * nu)));



    Mat<3> jac = sip.GetJacobian();
    double h = sqrt (sqr (jac(0,0)) + sqr (jac(0,1)));
    double t = fabs (jac(2,2));

    for (int i = 0; i < 6; i++)
      for (int j = 0; j < 6; j++)
	{
	  mat(i+6, j) = mat(i,j);
	  mat(i, j+6) = mat(i,j);
	  mat(i+6, j+6) = (1 + 10 * t*t/(h*h)) * mat(i,j);
	}
  }  
};


template <int D>
class EASIntegrator 
  : public T_BDBIntegrator<DiffOpEAS<D>, ElasticityDMatEAS<D>, FiniteElement>
{
public:
  ///
  EASIntegrator (CoefficientFunction * acoefe, 
		 CoefficientFunction * acoefnu)
    :  T_BDBIntegrator<DiffOpEAS<D>,  ElasticityDMatEAS<D>, FiniteElement>
  (ElasticityDMatEAS<D> (acoefe, acoefnu))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new EASIntegrator (coeffs[0], coeffs[1]);
  }
  
  ///
  virtual string Name () const { return "EAS"; }
};







template <int D>
class DiffOpIdU : public DiffOp<DiffOpIdU<D> >
{
  // 2 components:
  // u1 u2

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = D };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    const NodalFiniteElement & fel_ux = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_uz = 
      dynamic_cast<const NodalFiniteElement&> (cfel[2]);
    
    int nd_ux = fel_ux.GetNDof();
    int nd_uz = fel_uz.GetNDof();

    FlatVector<> vecux(nd_ux, lh);
    vecux = fel_ux.GetShape (sip.IP(), lh);
    FlatVector<> vecuz(nd_uz, lh);
    vecuz = fel_uz.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_ux; j++)
      {
	mat(0,j) = vecux(j);
	mat(1,nd_ux+j) = vecux(j);
      }
    
    for (int j = 0; j < nd_uz; j++)
      mat(2,2*nd_ux+j) = vecuz(j);
  }
};


template <int D>
class EASUIntegrator 
  : public T_BDBIntegrator<DiffOpIdU<D>, DiagDMat<D>, FiniteElement>
{
public:
  ///
  EASUIntegrator ()
    :  T_BDBIntegrator<DiffOpIdU<D>, DiagDMat<D>, FiniteElement>
  (DiagDMat<D> (new ConstantCoefficientFunction(1)))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new EASUIntegrator ();
  }
  
  ///
  virtual string Name () const { return "EAS IdU"; }
};









template <int D>
class DiffOpIdUSurf : public DiffOp<DiffOpIdUSurf<D> >
{
  // 2 components:
  // u1 u2

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D-1 };
  enum { DIM_DMAT = D };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    const NodalFiniteElement & fel_ux = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_uz = 
      dynamic_cast<const NodalFiniteElement&> (cfel[2]);
    
    int nd_ux = fel_ux.GetNDof();
    int nd_uz = fel_uz.GetNDof();

    FlatVector<> vecux(nd_ux, lh);
    vecux = fel_ux.GetShape (sip.IP(), lh);
    FlatVector<> vecuz(nd_uz, lh);
    vecuz = fel_uz.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_ux; j++)
      {
	mat(0,j) = vecux(j);
	mat(1,nd_ux+j) = vecux(j);
      }
    
    for (int j = 0; j < nd_uz; j++)
      mat(2,2*nd_ux+j) = vecuz(j);
  }
};



template <int D>
class EASUSurfIntegrator 
  : public T_BDBIntegrator<DiffOpIdUSurf<D>, DiagDMat<D>, FiniteElement>
{
public:
  ///
  EASUSurfIntegrator ()
    :  T_BDBIntegrator<DiffOpIdUSurf<D>, DiagDMat<D>, FiniteElement>
  (DiagDMat<D> (new ConstantCoefficientFunction(1)))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new EASUSurfIntegrator ();
  }
  
  bool BoundaryForm() const { return 1; }
  ///
  virtual string Name () const { return "EAS IdU Surf"; }
};







namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("eas", FESpaceEAS::Create);

    GetIntegrators().AddBFIntegrator ("eas", 2, 2,
				      EASIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("easu", 2, 0,
				      EASUIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("eas", 3, 2,
				      EASIntegrator<3>::Create);
    GetIntegrators().AddBFIntegrator ("easu", 3, 0,
				      EASUIntegrator<3>::Create);
    GetIntegrators().AddBFIntegrator ("easusurf", 3, 0,
				      EASUSurfIntegrator<3>::Create);
  }
  
  Init init;
}
