#ifdef LAPACK
#include "../include/solve.hpp"

/*
namespace ngfem
{
  extern Complex alpha;
  extern bool apply_deriv_alpha;
}
*/

namespace ngsolve
{
  using namespace ngsolve;


  /* ***************************** Numproc EVP-Lapack *************************** */

  ///
  class NumProcEVPWaveguide : public NumProc
  {
  protected:
    BilinearForm * bfa;
    BilinearForm * bfm;

    GridFunction * gfu;
    int num;

    double prec, shift, shift_imag;
    bool square_root;
    bool print;

    enum SOLVER { DENSE, ARNOLDI };
    ///
    SOLVER solver;

  public:
    NumProcEVPWaveguide (PDE & apde, const Flags & flags);
    virtual ~NumProcEVPWaveguide();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcEVPWaveguide (pde, flags);
    }

    virtual void Do();

    virtual string GetClassName () const
    {
      return " Eigenvalue Problem (Lapack)";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl;
    }
  };


  NumProcEVPWaveguide :: NumProcEVPWaveguide (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    num = int(flags.GetNumFlag ("num", 500));
    shift = flags.GetNumFlag ("shift",0.); 
    shift_imag = flags.GetNumFlag ("shifti",0.); 
    square_root = flags.GetDefineFlag ("sqrt");
    solver = ARNOLDI;
    if (flags.GetDefineFlag("dense")) solver = DENSE;
  }

  NumProcEVPWaveguide :: ~NumProcEVPWaveguide()
  {
    ;
  }


  void NumProcEVPWaveguide :: Do()
  {
    cout << "solve evp with Lapack" << endl;

    int dim = bfa->GetFESpace().GetDimension();
    int size = bfa->GetMatrix().Height();
    bool iscomplex = bfa->GetFESpace().IsComplex();

    
    if (solver == DENSE)
      {
	int n = dim*size;

	if (!iscomplex)
	  {
	    (*testout) << "mata = " << endl << bfa->GetMatrix() << endl;
	    (*testout) << "matm = " << endl << bfm->GetMatrix() << endl;

	    VVector<> hv(n), hva(n), hvm(n);
	    Matrix<> mata(n), matm(n);
	    
	    for (int i = 0; i < n; i++)
	      {
		hv = 0.0;
		hv(i) = 1.0;
		hva = bfa->GetMatrix() * hv;
		hvm = bfm->GetMatrix() * hv;
		
		for (int j = 0; j < n; j++)
		  {
		    mata(j,i) = hva(j);
		    matm(j,i) = hvm(j);
		  }
	      }
	    cout << "has matrices" << endl;
	    
	    Vector<Complex> lami(n);
	    Matrix<Complex> evecs(n);

	    // LapackGHEPEPairs (n, &mata(0,0), &matm(0,0), &lami(0));
	    LaEigNSSolve (n, &mata(0,0), &matm(0,0), &lami(0), 0, 0, 0, 0);
	    cout.precision(12);
	    cout << "lami = " << lami << endl;
	    (*testout) << "lami = " << lami << endl;
	    
	    // LaEigNSSolve (n, &mata(0,0), &matm(0,0), &lami(0), 1, &evecs(0,0), 0, 'N');
	    
	    cout << "eigensystem done" << endl;
	  }
	else
	  {
	    cout << "dense complex solver" << endl;
	    VVector<Complex> hv(n), hva(n), hvm(n);
	    Matrix<Complex> mata(n), matm(n);
	    
	    for (int i = 0; i < n; i++)
	      {
		hv = 0.0;
		hv(i) = 1.0;
		hva = bfa->GetMatrix() * hv;
		hvm = bfm->GetMatrix() * hv;
		
		for (int j = 0; j < n; j++)
		  {
		    mata(j,i) = hva(j);
		    matm(j,i) = hvm(j);
		  }
	      }
	    cout << "has matrices" << endl;
	    
	    Vector<Complex> lami(n);
	    Matrix<Complex> evecs(n);
	    
	    LaEigNSSolve (n, &mata(0,0), &matm(0,0), &lami(0), 1, &evecs(0,0), 0, 'N');

	    for (int i = 0; i < min2 (gfu->GetMultiDim(), n); i++)
	      {
		cout << "vector is " << typeid(gfu->GetVector(i)).name() << endl;
		FlatVector<Complex> vu = gfu->GetVector(i).FVComplex();
		for (int j = 0; j < n; j++)
		  vu(j) = evecs(i,j);
	      }


	    cout.precision(12);
	    cout << "lami = " << lami << endl;

	    ofstream ev ("eigenvalues.out");
	    for (int i = 0; i < n; i++)
	      {
		Complex lam =  (lami(i));
		ev << i << " " << lam.real() << " " << lam.imag() << endl;
	      }
	    cout << "eigensystem done" << endl;
	  }
      }
    
    else if (solver == ARNOLDI)

      {
	if (!iscomplex)
	  {
	    int n = dim*size;
	    VVector<double> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);
	    
	    // Arnoldi:
	    int m = min2 (num, n);
	    
	    Matrix<> matV (m,n), matVt(n,m);
	    Matrix<> matH(m), matI(m);

	    const BaseSparseMatrix&  mata1 =
	      dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix());
	
	    BaseSparseMatrix&  mata =
	      dynamic_cast<BaseSparseMatrix&> (*mata1.CreateMatrix());
	    
	    const BaseSparseMatrix& matm = 
	      dynamic_cast<const BaseSparseMatrix&> (bfm->GetMatrix()); 
	
	    mata.AsVector() = mata1.AsVector() - shift*matm.AsVector();  
	    
	    BaseMatrix * inva = mata.InverseMatrix();	    
	    
	    // BaseMatrix * inva = dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix()).InverseMatrix();
	    
	    hv.SetRandom();
	    matV = 0.0;
	    matH = 0.0;
	    matI = 0.0;
	    for (int i = 0; i < m; i++)
	      matI(i,i) = 1.0;

	    hv /= L2Norm (hv);

	    for (int i = 0; i < m; i++)
	      {
		cout << "i = " << i << endl;
		for (int j = 0; j < n; j++)
		  matV(i,j) = hv(j);
		
		hva = bfm->GetMatrix() * hv;
		hvm = *inva * hva;
		
		for (int j = 0; j <= i; j++)
		  {
		    double sum = 0.0;
		    for (int k = 0; k < n; k++)
		      sum += hvm(k) * matV(j,k);
		    matH(j,i) = sum;
		    for (int k = 0; k < n; k++)
		      hvm(k) -= sum * matV(j,k);
		  }
		
		hv = hvm;
		double len = L2Norm (hv);
		if (i<m-1)
		  matH(i+1,i) = len; 
		
		//  cout << ", len = " << len << endl;
		hv /= len;
	      }
	    
	    
	    // (*testout) << "matH = " << endl << matH << endl;
	    //	(*testout) << "V^T V = " << endl << matV * Trans (matV) << endl;
	    cout << "has Hessenberg" << endl;
	    
	    Vector<Complex> lami(m);
	    Matrix<Complex> evecs(m);
	    Matrix<Complex> levecs(n,m);
	    Matrix<Complex> matHt(m);

	    matHt = Trans (matH);
	    
	    evecs = Complex (0.0);
	    lami = Complex (0.0);
	    
	    LapackHessenbergEP (matH.Height(), &matHt(0,0), &lami(0), &evecs(0,0));
	    
	    cout << "hessenberg is back" << endl;
	    
	    // levecs = Trans(matV) * Trans(evecs);
	    matVt = Trans (matV);
	    levecs = Trans(evecs * Trans (matVt));
	    
	    for (int i = 0; i < m; i++)
	      lami(i) =  1.0 / lami(i) + shift;

	    if (square_root)
	      for (int i = 0; i < m; i++)
		lami(i) =  sqrt (lami(i));


	    ofstream out ("eigenvalues.dat");
	    for (int i = 0; i < m; i++)
	      out << i << "  " << lami[i].real() << "  " << lami[i].imag() << endl;
	    cout << "wrote eigenvalues to file 'eigenvalues.dat'" << endl;

	    cout << "eigensystem done" << endl;

	    (*testout) << "lami = " << endl << lami << endl;
	    cout << "lami = " << endl << lami << endl;
	    
	    for (int i = 0; i < min2 (gfu->GetMultiDim(), m); i++)
	      {
		FlatVector<Complex> vu = gfu->GetVector(i).FVComplex();
		for (int j = 0; j < n; j++)
		  vu(j) = levecs(j, i);
	      }
	  }
	
	else
	  
	  {
	    int n = dim*size;
	    // VVector<Complex> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);



	    // Arnoldi:
	    int m = min2 (num, n);

	    Matrix<Complex> matV (m,n), matVt(n,m);
	    Matrix<Complex> matH(m), matHt(m), matI(m);
	
	    const BaseSparseMatrix&  mata1 =
	      dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix());
	
	    BaseSparseMatrix&  mata =
	      dynamic_cast<BaseSparseMatrix&> (*mata1.CreateMatrix());
	
	    const BaseSparseMatrix& matm = 
	      dynamic_cast<const BaseSparseMatrix&> (bfm->GetMatrix()); 


            BaseVector & hv = *mata.CreateVector();
            BaseVector & hva = *mata.CreateVector();
            BaseVector & hvm = *mata.CreateVector();
            BaseVector & hwa = *mata.CreateVector();
            BaseVector & hwm = *mata.CreateVector();
            BaseVector & w = *mata.CreateVector();




	
	    mata.AsVector() = mata1.AsVector() - Complex (shift, shift_imag)*matm.AsVector();  
	    BaseMatrix * inva = mata.InverseMatrix();	    
	
	    hv.SetRandom();
	    matV = Complex(0.0);
	    matH = Complex(0.0);
	    matI = Complex(0.0);
	    for (int i = 0; i < m; i++)
	      matI(i,i) = 1.0;

	    Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
	    hv /= len;


	    for (int i = 0; i < m; i++)
	      {
		cout << "i = " << i << endl;
		for (int j = 0; j < n; j++)
		  matV(i,j) = hv(j);
	    
		hva = bfm->GetMatrix() * hv;
		hvm = *inva * hva;

		for (int j = 0; j <= i; j++)
		  {
		    Complex sum = 0.0;
		    for (int k = 0; k < n; k++)
		      sum += hvm(k) * matV(j,k);
		    matH(j,i) = sum;
		    for (int k = 0; k < n; k++)
		      hvm(k) -= sum * matV(j,k);
		  }
	    
		hv = hvm;
		Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
		if (i<m-1)
		  matH(i+1,i) = len; 

		hv /= len;
	      }

	
	    //	(*testout) << "matH = " << endl << matH << endl;
	    //	(*testout) << "V^T V = " << endl << matV * Trans (matV) << endl;
	    cout << "has Hessenberg" << endl;

	    Vector<Complex> lami(m);
	    Matrix<Complex> evecs(m);
	    Matrix<Complex> levecs(n,m);
	    matHt = Trans (matH);

	    evecs = Complex (0.0);
	    lami = Complex (0.0);

	    LapackHessenbergEP (m, &matHt(0,0), &lami(0), &evecs(0,0));

	    for (int i = 0; i < m; i++)
	      lami(i) =  1.0 / lami(i) + Complex(shift, shift_imag);

	    if (square_root)
	      for (int i = 0; i < m; i++)
		lami(i) =  sqrt (lami(i));

	    for (int i = 0; i < m; i++)
	      lami(i) = lami(i) * Complex (0,-1);

	    ofstream out ("eigenvalues.dat");
	    for (int i = 0; i < m; i++)
	      if (lami[i].real() >= shift_imag-20 && lami[i].real() < shift_imag+20)
		out << i << "  " << lami[i].real() << "  " << lami[i].imag() << endl;
	    cout << "wrote eigenvalues to file 'eigenvalues.dat'" << endl;
	    cout << "***** cut spectrum to small range !!!!" << endl;


	    cout << "eigensystem done" << endl;

	    // levecs = Trans(matV) * Trans(evecs);
	    matVt = Trans (matV);
	    levecs = Trans(evecs * Trans (matVt));
	
	    for (int i = 0; i < m; i++)
	      cout << "lam(" << i << ") = " 
		   << lami(i).real() << "+i" << lami(i).imag() << endl;

	    for (int i = 0; i < min2 (gfu->GetMultiDim(), m); i++)
	      {
		FlatVector<Complex> vu = gfu->GetVector(i).FVComplex();
		for (int j = 0; j < n; j++)
		  vu(j) = levecs(j, i);
	      }



	    
	    apply_deriv_alpha = 0;

	    // check autodiff 
	    hv = 0;
	    alpha += 1e-6;
	    bfa -> AddMatrix (Complex(1,0), gfu->GetVector(0), hv);
	    alpha -= 2e-6;
	    bfa -> AddMatrix (Complex(-1,0), gfu->GetVector(0), hv);
	    hv *= 0.5e6;
	    (*testout) << "num diff = " << hv << endl;

	    apply_deriv_alpha = 1;
	    hv = 0;
	    bfa -> AddMatrix (Complex(1,0), gfu->GetVector(0), hv);
	    (*testout) << "auto diff = " << hv << endl;	    
	  }
    
      }
  }
  


  


  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("evp_waveguide", NumProcEVPWaveguide::Create, NumProcEVPWaveguide::PrintDoc);
    }
    
    
    Init init;
    
  }
  


}
#endif
