/*

Nonlinear magnetic field solver (2D)

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;



class FastLaplaceTetHOIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coef;
public:
#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;



class FastLaplaceTetHOIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coef;
  FastLaplaceTetHOIntegrator (CoefficientFunction * acoef)
    : coef(acoef)
  {
    ;
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new FastLaplaceTetHOIntegrator (coeffs[0]);
  }

  virtual bool BoundaryForm () const { return 0; }


  void CalcShapes1D (int n, double s, double * values) const
  {
    IntegratedLegendrePolynomial (n, -1+2*s, values);
    values[0] = 1-s;
    values[1] = s;
  }

  void CalcDShapes1D (int n, double s, double * values) const
  {
    double * hp = values+1;
    LegendrePolynomial (n-1, -1+2*s, hp);
    for (int i = 2; i <= n; i++)
      values[i] *= 2;
    values[0] = -1;
    values[1] = 1;
  }





  ///
  virtual void
  AssembleElementMatrix (const FiniteElement & bfel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & locheap) const
  {
    int order = 10;
    int ndof = (order+1)*(order+2)*(order+3)/6;
    int ndoftrig = (order+1)*(order+2)/2;

    const IntegrationRule & ir =  
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*order);

    elmat.AssignMemory (ndof, ndof, locheap);
    elmat = 0;
    FlatMatrix<double> hmat(ndoftrig, ndoftrig, locheap);
    FlatVector<double> shapex(ndof, locheap);
    FlatVector<double> shapey(ndof, locheap);
    FlatVector<double> shapez(ndof, locheap);
    FlatVector<double> dshapex(ndof, locheap);
    FlatVector<double> dshapey(ndof, locheap);
    FlatVector<double> dshapez(ndof, locheap);

    for (int s = 0; s < 3; s++)
      for (int t = 0; t < 3; t++)
	for (int ipx = 0; ipx < ir.GetNIP(); ipx++)
	  {
	    // comput y-z integrals:
	    hmat = 0;
	    double x = ir.GetIP(ipx)(0);
	    
	    for (int ipy = 0; ipy < ir.GetNIP(); ipy++)
	      for (int ipz = 0; ipz < ir.GetNIP(); ipz++)
		{
		  double y = ir.GetIP(ipy)(0);
		  double z = ir.GetIP(ipz)(0);
		  IntegrationPoint ip(x, y, t, 0);
		  IntegrationPoint iptet(x, y, t, 0);
		  Mat<3> f;
		  f = 0.0;
		  f(0,0) = f(1,1) = f(2,2) = 1;
		  Mat<3> a = f * Trans(f);
		  
		  CalcShapes1D (order, y, &shapey(0));
		  CalcDShapes1D (order, y, &dshapey(0));
		  CalcShapes1D (order, z, &shapez(0));
		  CalcDShapes1D (order, z, &dshapez(0));

		  int jj = 0;
		  for (int j = 0; j < order; j++)
		    for (int k = 0; k < order-j; k++)
		      {
			double fac = a(s,t) * shapey(j) * shapez(k);
			int jj2 = 0;
			for (int j2 = 0; j2 < order; j2++)
			  for (int k2 = 0; k2 < order-j2; k2++)
			    {
			      hmat(jj,jj2) += fac * shapez(j2) * shapez(k2);
			      jj2++;
			    }
			jj++;
		      }
		}
	    
	    CalcShapes1D (order, x, &shapex(0));
	    CalcDShapes1D (order, x, &dshapex(0));

	    int jj = 0;
	    for (int j = 0; j < order; j++)
	      for (int k = 0; k < (order-j+1)*(order-j+2)/2; k++)
		{
		  int jj2 = 0;
		  for (int j2 = 0; j2 < order; j2++)
		    for (int k2 = 0; k2 < (order-j2+1)*(order-j2+2)/2; k2++)
		      {
			elmat(jj,jj2) += 
			  shapex(j) * shapex(j2) * hmat(k,k2);
			jj2++;
		      }
		  jj++;
		}
	  }
  }



  ///
  virtual string Name () const { return "FastLaplace"; }
};



namespace {
  class Init
    { 
    public: 
      Init ();
  };        
  
  Init::Init()
  {
    GetIntegrators().AddBFIntegrator ("fastlaplace", 3, 1,
				      FastLaplaceTetHOIntegrator::Create);

    /*
    FastLaplaceTetHOIntegrator flap (new ConstantCoefficientFunction (1));
    FiniteElement tet;
    ElementTransformation eltrans;
    FlatMatrix<double> elmat;
    LocalHeap lh(100000000);
    
    cout << "do assembling" << endl;
    for (int i = 0; i < 1000; i++)
      {
	cout << "el " << i << endl;
	flap.AssembleElementMatrix (tet, eltrans, elmat, lh);
	lh.CleanUp();
      }
    cout << "done" << endl;
    */
  }

  Init init;

}
