#include <solve.hpp>

#include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;







  class DiffOpIdHCurlH1 : public DiffOp<DiffOpIdHCurlH1>
  {
  public:
    enum { DIM = 1 };
    enum { DIM_SPACE = 3 };
    enum { DIM_ELEMENT = 3 };
    enum { DIM_DMAT = 3 };
    enum { DIFFORDER = 0 };

    template <typename FEL, typename SIP, typename MAT>
    static void GenerateMatrix (const FEL & bfel, const SIP & sip,
				MAT & mat, LocalHeap & lh)
    {
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);

      const HCurlFiniteElementD<3> & felvec = 
	dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[0]);
      const NodalFiniteElement & felscal = 
	dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
      int ndvec = felvec.GetNDof();
      int ndscal = felscal.GetNDof();
      FlatMatrix<> matvec(3, ndvec, lh);
      matvec = Trans (sip.GetJacobianInverse ()) * 
	Trans (felvec.GetShape(sip.IP(), lh));

      FlatVector<> vecscal(ndscal, lh);
      vecscal = felscal.GetShape (sip.IP(), lh);

      int i, j;
      mat = 0;
      for (i = 0; i < 3; i++)
	for (j = 0; j < ndvec; j++)
	  mat(i,j) = matvec(i,j);
      for (j = 0; j < ndscal; j++)
	mat(1, ndvec+j) = vecscal(j);
    }

  };

  class MassHCurlH1Integrator 
    : public T_BDBIntegrator<DiffOpIdHCurlH1, DiagDMat<3>, FiniteElement>
  {
  public:
    ///
    MassHCurlH1Integrator (CoefficientFunction * coeff)
      :  T_BDBIntegrator<DiffOpIdHCurlH1, DiagDMat<3>, FiniteElement>
    (DiagDMat<3> (coeff))
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MassHCurlH1Integrator (coeffs[0]);
    }
  
    ///
    virtual string Name () const { return "MassHCurlH1"; }
  };









  class DiffOpIdBoundaryHCurlH1 : public DiffOp<DiffOpIdBoundaryHCurlH1>
  {
  public:
    enum { DIM = 1 };
    enum { DIM_SPACE = 3 };
    enum { DIM_ELEMENT = 2 };
    enum { DIM_DMAT = 3 };
    enum { DIFFORDER = 0 };

    template <typename FEL, typename SIP, typename MAT>
    static void GenerateMatrix (const FEL & bfel, const SIP & sip,
				MAT & mat, LocalHeap & lh)
    {
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);

      const HCurlFiniteElementD<2> & felvec = 
	dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[0]);
      const NodalFiniteElement & felscal = 
	dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
      int ndvec = felvec.GetNDof();
      int ndscal = felscal.GetNDof();
      FlatMatrix<> matvec(3, ndvec, lh);

      matvec = Trans (sip.GetJacobianInverse ()) *
	Trans (felvec.GetShape(sip.IP(),lh));

      FlatVector<> vecscal(ndscal, lh);
      vecscal = felscal.GetShape (sip.IP(), lh);

      int i, j;
      mat = 0;
      for (i = 0; i < 3; i++)
	for (j = 0; j < ndvec; j++)
	  mat(i,j) = matvec(i,j);

      Mat<3> proj = sip.GetJacobian() * sip.GetJacobianInverse();
    
      for (i = 0; i < 3; i++)
	for (j = 0; j < ndscal; j++)
	  mat(i, ndvec+j) = -proj(i,1) * vecscal(j);
    }

  };




  class RobinHCurlH1Integrator 
    : public T_BDBIntegrator<DiffOpIdBoundaryHCurlH1, DiagDMat<3>, FiniteElement>
  {
  public:
    ///
    RobinHCurlH1Integrator (CoefficientFunction * coeff)
      : T_BDBIntegrator<DiffOpIdBoundaryHCurlH1, DiagDMat<3>, FiniteElement>
    (DiagDMat<3> (coeff))
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new RobinHCurlH1Integrator (coeffs[0]);
    }
  
    ///
    virtual string Name () const { return "RobinHCurlH1"; }
  };






  /// Gradient operator of dimension D
  class DiffOpCurlHCurlH1 : public DiffOp<DiffOpCurlHCurlH1>
  {
  public:
    enum { DIM = 1 };
    enum { DIM_SPACE = 3 };
    enum { DIM_ELEMENT = 3 };
    enum { DIM_DMAT = 3 };
    enum { DIFFORDER = 1 };

    ///
    template <typename FEL, typename SIP, typename MAT>
    static void GenerateMatrix (const FEL & bfel, const SIP & sip,
				MAT & mat, LocalHeap & lh)
    {
      const CompoundFiniteElement & cfel = 
	dynamic_cast<const CompoundFiniteElement&> (bfel);

      const HCurlFiniteElementD<3> & felvec = 
	dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[0]);
      const NodalFiniteElement & felscal = 
	dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
      int ndvec = felvec.GetNDof();
      int ndscal = felscal.GetNDof();

      FlatMatrix<> grad(3, ndscal, lh);
      grad = Trans (sip.GetJacobianInverse ()) * 
	Trans (felscal.GetDShape(sip.IP(),lh));

      mat = 0;
      for (int i = 0; i < ndscal; i++)
	{
	  mat(0, ndvec+i) = -grad(2,i);
	  mat(2, ndvec+i) = grad(0,i);
	}
    }
  };


  class CurlHCurlH1Integrator 
    : public T_BDBIntegrator<DiffOpCurlHCurlH1, DiagDMat<3>, FiniteElement>
  {
  public:
    ///
    CurlHCurlH1Integrator (CoefficientFunction * coeff)
      :  T_BDBIntegrator<DiffOpCurlHCurlH1, DiagDMat<3>, FiniteElement>
    (DiagDMat<3> (coeff))
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new CurlHCurlH1Integrator (coeffs[0]);
    }
  
    ///
    virtual string Name () const { return "CurlHCurlH1"; }
  };








  class RobinHCurlH1Integrator2 
    : public BilinearFormIntegrator
  {
    CoefficientFunction * coeff;

  public:
    ///
    RobinHCurlH1Integrator2 (CoefficientFunction * acoeff)
      : coeff(acoeff)
    { ; }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new RobinHCurlH1Integrator2 (coeffs[0]);
    }
  
    ///
    virtual string Name () const { return "RobinHCurlH1"; }

    virtual bool BoundaryForm () const { return 1; }



    virtual void
    AssembleElementMatrix (const FiniteElement & bfel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      try
	{
	  const CompoundFiniteElement & cfel = 
	    dynamic_cast<const CompoundFiniteElement&> (bfel);
	
	  const HCurlFiniteElementD<2> & felvec = 
	    dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[0]);
	  const NodalFiniteElement & felscal = 
	    dynamic_cast<const NodalFiniteElement&> (cfel[1]);
	
	  int ndvec = felvec.GetNDof();
	  int ndscal = felscal.GetNDof();
	  int ndof = ndvec + ndscal;
	
	  elmat.AssignMemory (ndof, ndof, lh);
	  elmat = 0;
	
	  FlatMatrix<> matvec(3, ndvec, lh);
	  FlatMatrix<> matscal(3, ndscal, lh);
	  FlatMatrix<> matscal1(2, ndscal, lh);
	  FlatMatrix<> matscal2(3, ndscal, lh);
	  FlatVector<> vecscal(ndscal, lh);

	  FlatMatrix<> shape(2, 2*ndscal, lh);
	  FlatMatrix<> tau(2*ndscal, ndscal, lh);

	  tau = 0;
	  const IntegrationRule & nir = felscal.NodalIntegrationRule();
	  /*
	    const IntegrationRule & nir =
	    GetIntegrationRules().SelectNodalIntegrationRule (felvec.ElementType(), 2);
	  */
	  for (int ipnr = 0; ipnr < nir.GetNIP(); ipnr++)
	    {
	      void * heapp = lh.GetPointer();
	    
	      SpecificIntegrationPoint<2,3> sip(nir, ipnr, eltrans, lh);
	      tau(2*ipnr  , ipnr) = sip.GetJacobian()(1,0);
	      tau(2*ipnr+1, ipnr) = sip.GetJacobian()(1,1);
	    }

	  FlatMatrixFixHeight<3, double> bmat (ndof, lh);
	
	  int order = max2 (felvec.Order(), felscal.Order());

	  const IntegrationRule & ir =
	    GetIntegrationRules().SelectIntegrationRule (felvec.ElementType(), 2*order+2);

	  for (int ipnr = 0; ipnr < ir.GetNIP(); ipnr++)
	    {
	      void * heapp = lh.GetPointer();
	    
	      SpecificIntegrationPoint<2,3> sip(ir, ipnr, eltrans, lh);
	    
	      //	    DIFFOP::GenerateMatrix (fel, sip, bmat, locheap);

	      matvec = Trans (sip.GetJacobianInverse ()) *
		Trans (felvec.GetShape(sip.IP(),lh));
	      vecscal = felscal.GetShape (sip.IP(), lh);


	      int i, j;
	      bmat = 0;
	      for (i = 0; i < 3; i++)
		for (j = 0; j < ndvec; j++)
		  bmat(i,j) = matvec(i,j);


	      Mat<3> proj = sip.GetJacobian() * sip.GetJacobianInverse();
	      for (i = 0; i < 3; i++)
		for (j = 0; j < ndscal; j++)
		  bmat(i, ndvec+j) = -proj(i,1) * vecscal(j);

	      /*
		matscal = 0;
		for (j = 0; j < ndscal; j++)
		matscal(1, j) = vecscal(j);

		matscal1 = Trans (sip.GetJacobian()) * matscal;
		matscal2 = Trans (sip.GetJacobianInverse()) * matscal1;
	      */

	      shape = 0;
	      for (i = 0; i < 2; i++)
		for (j = 0; j < ndscal; j++)
		  shape(i,2*j+i) = vecscal(j);

	      matscal1 = shape * tau;
	      matscal2 = Trans (sip.GetJacobianInverse()) * matscal1;


	    
	      for (i = 0; i < 3; i++)
		for (j = 0; j < ndscal; j++)
		  bmat(i,ndvec+j) = -matscal2(i,j);

	      double fac = 
		fabs (sip.GetJacobiDet()) * sip.IP().Weight() * 
		Evaluate (*coeff, sip);

	      elmat += fac * (Trans (bmat) * bmat);
	    
	      lh.CleanUp (heapp);
	    } 
	}
    
      catch (Exception & e)
	{
	  stringstream ost;
	  ost << "in AssembleElementMatrix, type = " 
	      << typeid(*this).name() << endl;
	  e.Append (ost.str());
	  throw;
	}
      catch (exception & e)
	{
	  throw (Exception (string(e.what()) +
			    string("\n in AssembleElementMatrix\n")));
	}
    }
  

    virtual void 
    ApplyElementMatrix (const FiniteElement & fel, 
			const ElementTransformation & eltrans, 
			const FlatVector<Complex> & elx, 
			FlatVector<Complex> & ely,
			LocalHeap & locheap) const
    {
      FlatMatrix<double> mat;
      AssembleElementMatrix (fel, eltrans, mat, locheap);
      ely = mat * elx;
    }

  };






  


  




  /// A combination of fe-spaces
  class HCurlH1FESpace : public CompoundFESpace
  {
  public:
    HCurlH1FESpace (const MeshAccess & ama,
		    const ARRAY<const FESpace*> & aspaces)
      : CompoundFESpace (ama, aspaces) { ; }
    ///
    virtual ~HCurlH1FESpace () { ; }


    static FESpace * Create (const MeshAccess & ma, const Flags & flags);


    
    virtual string GetClassName () const
    {
      return "HCurlH1FESpace";
    }
    
    virtual Table<int> * CreateSmoothingBlocks (int type = 0) const;
    virtual BitArray * CreateIntermediatePlanes (int type = 0) const;
  };







Table<int> * HCurlH1FESpace :: CreateSmoothingBlocks (int type) const
{
  cout << "HCurlH1FESpace::CreateSmoothingBlocks" << endl;

  int nedge = ma.GetNEdges();
  int nv = ma.GetNV();
  int nfa = ma.GetNFaces();
  int nd0 = spaces[0]->GetNDof();
  int nd1 = spaces[1]->GetNDof();

  int order_hcurl = spaces[0]->GetOrder();
  int order_h1 = spaces[1]->GetOrder();
  cout << "hcurl, order = " << order_hcurl << ", ndof = " << nd0 << endl;
  cout << "h1,    order = " << order_h1 << ", ndof = " << nd1 << endl;

  int nbl = nv;
  if (order_hcurl >= 3)
    nbl += nfa;
  if (nd1 > nbl) nbl = nd1;

  ARRAY<int> cnts(nbl);
  Table<int> *node2edge;
  int j, k, i;

  for (k = 1; k <= 2; k++)
    {
      if (k == 2)
	node2edge = new Table<int>(cnts);

      cnts = 0;

      for (j = 0; j < nedge; j++)
	{
	  
	  int ecl = ma.GetClusterRepEdge(j);
	  if (ecl < 0) continue;
	    
	  int ep1, ep2;
	  ma.GetEdgePNums (j, ep1, ep2);
	    
	  // for anisotropic connections:
	  //	  int cep1 = ma.GetClusterRepVertex(ep1);
	  //	  int cep2 = ma.GetClusterRepVertex(ep2);
	  int cep1 = ep1;
	  int cep2 = ep2;
	  
	  if (k == 2)
	    {
	      (*node2edge)[cep1][cnts[cep1]++] = j;
	      if (cep1 != cep2)
		(*node2edge)[cep2][cnts[cep2]++] = j;

	      for (i = 1; i < order_hcurl; i++)
		{
		  (*node2edge)[cep1][cnts[cep1]++] = j+i*nedge;		  
		  if (cep1 != cep2)
		    (*node2edge)[cep2][cnts[cep2]++] = j+i*nedge;
		}
	    }
	  else
	    {
	      cnts[cep1] += order_hcurl;
	      if (cep1 != cep2)
		cnts[cep2] += order_hcurl;
	    }
	}
      if (order_hcurl >= 3)
	for (j = 0; j < nfa; j++)
	  {
	    if (k == 1)
	      {
		cnts[nv+j] += 3;
	      }
	    else
	      {
		for (i = 0; i < 3; i++)
		  (*node2edge)[nv+j][cnts[nv+j]++] = 3*nedge+3*j+i;
	      }
	  }


      for (j = 0; j < nd1; j++)
	{
	  if (k == 2)
	    (*node2edge)[j][cnts[j]++] = nd0+j;
	  else
	    cnts[j]++;
	}
    }

  (*testout) << "HCurlH1, blocks = " << endl 
	     << (*node2edge) << endl;
  cout << "blocks finished" << endl;

  //  (*testout) << "cnts = " << cnts << endl;
  //  (*testout) << "Smoothingblocks: " << endl << (*node2edge) << endl;

  /*
  int maxsize = 0;

  for (int i = 0; i < node2edge->Size(); i++)
    {
      (*testout) << "row " << i << ", s = " << (*node2edge)[i].Size() << endl;
      if ((*node2edge)[i].Size() > maxsize)
	maxsize = (*node2edge)[i].Size();
    }
  cout << "max size = " << maxsize << endl;
  */

  return node2edge;
}



BitArray * HCurlH1FESpace :: 
CreateIntermediatePlanes (int type) const
{
  cout << "HCurlH1FESpace::CreateIntermediatePlanes" << endl;
  BitArray & ba = *new BitArray (GetNDof());
  ba.Clear();

  int nd0 = spaces[0]->GetNDof();
  int nd1 = spaces[1]->GetNDof();

  ARRAY<int> dnums;
  int nv = ma.GetNV();
  int np = ma.GetNP();
  int nse = ma.GetNSE();
  int ne = ma.GetNE();
  int ned = ma.GetNEdges();
  int i, j;

  // all H1 volume dofs
  for (i = 0; i < ne; i++)
    {
      //      int ind = ma.GetElIndex (i);
      // if (ind != 0) continue;
      spaces[1]->GetDofNrs (i, dnums);
      for (j = 0; j < dnums.Size(); j++)
	if (dnums[j] != -1)
	  ba.Set (nd0+dnums[j]);
    }

  // all dofs on the H1 boundary
  for (i = 0; i < nse; i++)
    {
      int ind = ma.GetSElIndex (i);
      if (!spaces[1]->DefinedOnBoundary(ind)) continue;

      GetSDofNrs (i, dnums);
      for (j = 0; j < dnums.Size(); j++)
	if (dnums[j] != -1)
	  ba.Set (dnums[j]);
    }

  // all edges on H1/H(curl) interface

  BitArray interfacepts(np);
  interfacepts.Clear();
  for (i = 0; i < nse; i++)
    {
      int ind = ma.GetSElIndex (i);
      if (!spaces[1]->DefinedOnBoundary(ind)) continue;
      ma.GetSElPNums (i, dnums);
      for (j = 0; j < dnums.Size(); j++)
	interfacepts.Set (dnums[j]);
    }
  
  for (i = 0; i < ned; i++)
    {
      int pi1, pi2;
      ma.GetEdgePNums (i, pi1, pi2);
      if (interfacepts[pi1] || interfacepts[pi2])
	ba.Set (i);
    }
  
  //  (*testout) << "ba = " << endl << ba << endl;

  return &ba;
}



FESpace * HCurlH1FESpace :: Create (const MeshAccess & ma, const Flags & flags)
{
  /*
  int order = int(flags.GetNumFlag ("order", -1));
  int zorder = int(flags.GetNumFlag ("zorder", order));
  int dim = int(flags.GetNumFlag ("dim", 1));
  bool iscomplex = flags.GetDefineFlag ("complex");
  
  if (order >= 2)
    return new NedelecFESpace2 (ma, order, zorder, dim, iscomplex);
  else
    return new NedelecFESpace (ma, 1, dim, iscomplex);

      const ARRAY<char*> & spacenames = flags.GetStringListFlag ("spaces");
      ARRAY<const FESpace*> spaces (spacenames.Size());
      for (int i = 0; i < spaces.Size(); i++)
	spaces[i] = GetFESpace (spacenames[i]);
      space = new HCurlH1FESpace (GetMeshAccess(), spaces);
  */
  return 0;
}





  namespace InitHCH1 {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("masshcurlh1", 3, 1,
      					MassHCurlH1Integrator::Create);
      GetIntegrators().AddBFIntegrator ("robinhcurlh1", 3, 1,
      					RobinHCurlH1Integrator2::Create);
      GetIntegrators().AddBFIntegrator ("curlhcurlh1", 3, 1,
      					CurlHCurlH1Integrator::Create);

      GetFESpaceClasses().AddFESpace ("hcurlh1", HCurlH1FESpace::Create);
    }
    
    Init init;
  }



}
