/*
  Arnold Winther elasticity elements
*/

#include <comp.hpp>

namespace ngcomp
{
  using namespace ngcomp;


  /* ******************** HDivSym Base Finite Element ***************** */


  ///
  class HDivSymFiniteElement : public FiniteElement
  {
  
  protected:
    class IPData
    {
    public:
      Matrix<> shape;
      Matrix<> divshape;
      IPData (int ndof)
	: shape(ndof, 3), divshape(ndof, 2) { ; }
    };

  public:
    HDivSymFiniteElement (int adimspace = 0, ELEMENT_TYPE aeltype = ET_TRIG,
			  int andof = 0, int aorder = 0)
      : FiniteElement (adimspace, aeltype, andof, aorder) { ; }

    virtual ~HDivSymFiniteElement () { ; }
  
  
  
    template <class TP>
    FlatMatrix<> GetShape (TP & p, LocalHeap & lh) const
    {
      IntegrationPoint ip;
      for (int i = 0; i < 2; i++)
	ip(i) = p(i);
    
      FlatMatrix<> shape(ndof, 3, lh);
      CalcShape (ip, shape);
      return shape;
    }

    /*
    const FlatMatrix<> GetShape (const IntegrationPoint & ip, LocalHeap & lh) const
    {
      return GetShapeV (ip);
    }

    const FlatMatrix<> GetShape (const IntegrationPoint & ip) const
    {
      return GetShapeV (ip);
    }
    */

    virtual const FlatMatrix<> GetShapeV (const IntegrationPoint & ip) const
    {
      return GetIPData() [ip.IPNr()] -> shape;
    }


    template <class TP>
    FlatMatrix<> GetDivShape (TP & p, LocalHeap & lh) const
    {
      IntegrationPoint ip;
      for (int i = 0; i < 2; i++)
	ip(i) = p(i);

      FlatMatrix<> divshape(ndof, 2, lh);
      CalcDivShape (ip, divshape);
      return divshape;
    }
    /*
    const FlatMatrix<> GetDivShape (const IntegrationPoint & ip, 
				    LocalHeap & lh) const
    {
      return FlatMatrix<> (GetDivShapeV (ip));
    }

    const FlatMatrix<> GetDivShape (const IntegrationPoint & ip) const
    {
      return FlatMatrix<> (GetDivShapeV (ip));
    }
    */

    virtual const FlatMatrix<> GetDivShapeV (const IntegrationPoint & ip) const
    {
      return FlatMatrix<> (GetIPData() [ip.IPNr()] -> divshape);
    }


    virtual void CalcShape (const IntegrationPoint & ip, 
			    FlatMatrix<> & shape) const = 0;
  
    virtual void CalcDivShape (const IntegrationPoint & ip, 
			       FlatMatrix<> & divshape) const;

    virtual const ARRAY<IPData*> & GetIPData () const = 0;

  protected:
    void CalcIPData (ELEMENT_TYPE et,
		     ARRAY<IPData*> & ipdata);

    virtual void CalcShape1 (const IntegrationPoint & ip, 
			     FlatMatrix<> & shape) const { ; } 
  };




  
  void HDivSymFiniteElement ::
  CalcDivShape (const IntegrationPoint & ip, 
		FlatMatrix<> & divshape) const
  {
    double eps = 1e-5;
    Matrix<> shape1(ndof,3);
    Matrix<> shape2(ndof,3);
    Matrix<> shape3(ndof,3);
    Matrix<> shape4(ndof,3);
    
    Matrix<> dshapei(ndof,3);
    divshape = 0;

    for (int i = 0; i < 2; i++)
      {
	IntegrationPoint ip1 = ip;
	IntegrationPoint ip2 = ip;
	ip1(i) -= eps;
	ip2(i) += eps;

	CalcShape (ip1, shape1);
	CalcShape (ip2, shape2);

	ip1(i) -= eps;
	ip2(i) += eps;
	CalcShape (ip1, shape3);
	CalcShape (ip2, shape4);
	
	dshapei = 
	  2/(3*eps) * (shape2 - shape1) 
	  - 1/(12*eps) * (shape4 - shape3);

	for (int j = 0; j < ndof; j++)
	  {
	    divshape(j,0) += dshapei(j,i);
	    divshape(j,1) += dshapei(j,i+1);
	  }
      }
  }


  void HDivSymFiniteElement :: 
  CalcIPData (ELEMENT_TYPE et,
	      ARRAY<IPData*> & ipdata)
  {
    const ARRAY<IntegrationPoint*> & ipts = 
      GetIntegrationRules().GetIntegrationPoints (et);
    
    (*testout) << "Calc HDivSym-IP Data for element type  " << et 
	       << ", ndof = " << GetNDof() << ": " << ipts.Size() << endl;
    
    ipdata.SetSize (0);
    for (int i = 0; i < ipts.Size(); i++)
      {
	IPData * ipd = new IPData(ndof);
	CalcShape (*ipts[i], ipd->shape);
	CalcDivShape (*ipts[i], ipd->divshape);
	ipdata.Append (ipd);
      }
  }


  
  
  
  /* ***************** Non-conforming Arnold-Winther element ******************* */
  
  ///
  class FE_AWTrigNC : public HDivSymFiniteElement
  {
  protected:
    ///
    static ARRAY<IPData*> ipdata;
    ///
    static Matrix<> trans;
    
  public:
    ///
    FE_AWTrigNC();
    ///
    virtual ~FE_AWTrigNC();

    ///
    virtual void CalcShape (const IntegrationPoint & ip, 
			    FlatMatrix<> & shape) const;
    
    ///
    virtual void CalcShape1 (const IntegrationPoint & ip, 
			     FlatMatrix<> & shape) const;

    ///
    void Orthogonalize();
    
    ///
    virtual const ARRAY<IPData*> & GetIPData () const 
    { return ipdata; }
  };
  






  ARRAY<HDivSymFiniteElement::IPData*> FE_AWTrigNC::ipdata;
  Matrix<> FE_AWTrigNC::trans(18,15);

  FE_AWTrigNC :: FE_AWTrigNC()
    : HDivSymFiniteElement (2, ET_TRIG, 15, 2)
  {
    if (!ipdata.Size())
      {
	Orthogonalize();
	CalcIPData(ET_TRIG, ipdata);
      }
  }
  
  FE_AWTrigNC :: ~FE_AWTrigNC()
  {
    ;
  }

  void FE_AWTrigNC :: CalcShape (const IntegrationPoint & ip, 
				 FlatMatrix<> & shape) const
  {
    double sh1[3*18];
    FlatMatrix<> shape1 (18, 3, sh1);
    
    CalcShape1 (ip, shape1);
    shape = Trans (trans) * shape1;
  }
  
  void FE_AWTrigNC :: CalcShape1 (const IntegrationPoint & ip, 
				  FlatMatrix<> & shape) const
  {
    double x = ip(0);
    double y = ip(1);
    shape = 0;

    for (int i = 0; i < 3; i++)
      {
	int base = 6 * i;
	shape(base  ,i) = 1;
	shape(base+1,i) = x;
	shape(base+2,i) = y;
	shape(base+3,i) = x*x;
	shape(base+4,i) = x*y;
	shape(base+5,i) = y*y;
      }
  }
  

  void FE_AWTrigNC :: Orthogonalize()
  {
    cout << "compute AW trig 1" << endl;
    
    int nd = 18;
    int i, j;
    Matrix<double> fiphij(nd);
    Matrix<double> trans1(nd);
  
    Matrix<> shape(nd, 3);
    Matrix<> momentsnn(3,nd);
    Matrix<> momentsnt(3,nd);
    Matrix<> momentsnx(3,nd);
    Matrix<> momentsny(3,nd);
    FE_Segm2L2 segm2;
    

    // compute edge momemtns
    for (i = 0; i < 3; i++) 
      {  // 3 edges

	Matrix<> shapenn(nd, 1);
	Matrix<> shapent(nd, 1);
	Matrix<> shapenx(nd, 1);
	Matrix<> shapeny(nd, 1);
	Vector<> testshape(3);
	
	momentsnn = 0;
	momentsnt = 0;
	momentsnx = 0;
	momentsny = 0;

	const IntegrationRule & edgerule = 
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4);

	const POINT3D * points = ElementTopology::GetVertices (ElementType());
	const EDGE & edge = ElementTopology::GetEdges (ElementType()) [i];
       
	Vec<2> p1, p2, tv, nv;
	Vec<3> nvnv, nvtv, nvxv, nvyv;

	for (j = 0; j < 2; j++)
	  {
	    p1(j) = points[edge[0]][j];
	    p2(j) = points[edge[1]][j];
	  }

	tv = p2 - p1;
	nv(0) = -tv(1); 
	nv(1) = tv(0);

	nvnv(0) = nv(0) * nv(0);
	nvnv(1) = 2 * nv(0) * nv(1);
	nvnv(2) = nv(1) * nv(1);

	nvtv(0) = nv(0) * tv(0);
	nvtv(1) = nv(0) * tv(1) + nv(1) * tv(0);
	nvtv(2) = nv(1) * tv(1);

	nvxv(0) = nv(0);
	nvxv(1) = nv(1);
	nvxv(2) = 0;

	nvyv(0) = 0;
	nvyv(1) = nv(0);
	nvyv(2) = nv(1);

	for (j = 0; j < edgerule.GetNIP(); j++)
	  {
	    const IntegrationPoint & ip = edgerule.GetIP(j);

	    Vec<2> p2d = p1 + ip(0) * tv;
	    
	    IntegrationPoint ip2d(&p2d(0), 0);
	    
	    segm2.CalcShape (ip, testshape);
	    
	    CalcShape1 (ip2d, shape);

	    shapenn = shape * nvnv;
	    shapent = shape * nvtv;
	    shapenx = shape * nvxv;
	    shapeny = shape * nvyv;

	    double fac = ip.Weight(); //  / L2Norm (nv);
	    momentsnn += fac * (testshape * Trans (shapenn));
	    momentsnt += fac * (testshape * Trans (shapent));
	    momentsnx += fac * (testshape * Trans (shapenx));
	    momentsny += fac * (testshape * Trans (shapeny));
	  }


	for (j = 0; j < nd; j++)
	  {
	    fiphij(4*i  ,j) = momentsnx(0, j);
	    fiphij(4*i+1,j) = momentsny(0, j);
	    fiphij(4*i+2,j) = momentsnx(1, j);
	    fiphij(4*i+3,j) = momentsny(1, j);

	    fiphij(15+i,j)  = momentsnn(2, j);
	  }
      }
   
    // 3 times 0th moement
    const IntegrationRule & trigrule = 
      GetIntegrationRules().SelectIntegrationRule (ET_TRIG, 4);
    Matrix<> trigmoments(3,nd);
    trigmoments = 0;
    for (j = 0; j < trigrule.GetNIP(); j++)
      {
	const IntegrationPoint & ip = trigrule.GetIP(j);
	CalcShape1 (ip, shape);
	trigmoments += ip.Weight() * Trans (shape);
      }
    for (j = 0; j < nd; j++)
      {
	fiphij(12,j) = trigmoments(0, j);
	fiphij(13,j) = trigmoments(1, j);
	fiphij(14,j) = trigmoments(2, j);
      }

    CalcInverse (fiphij, trans1);
    for (i = 0; i < 18; i++)
      for (j = 0; j < 15; j++)
	trans(i,j) = trans1(i,j);

    (*testout) << "AWTrigNC" << endl
	       << "fiphij = " << endl << fiphij << endl
	       << "Trans(fiphij) = " << endl << Trans(fiphij) << endl
	       << "trans = " << endl << trans1 << endl;
  }




  /* ************************ Finite Element Space ****************** */







  HDivSymFESpace :: HDivSymFESpace (const MeshAccess & ama,
				    int aorder, int adim, bool acomplex)
    
    : FESpace (ama, aorder, adim, acomplex)
  {
    trig    = new FE_AWTrigNC;
    //    segm    = new FE_NedelecSegm1;
  }

  HDivSymFESpace :: ~HDivSymFESpace ()
  {
    ;
  }


  void HDivSymFESpace :: 
  GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (15);

    int enr[12], eoa[12];
    ARRAY<int> enums(12, enr), eorient(12, eoa);
    GetMeshAccess().GetElEdges (elnr, enums, eorient);

    int i, j;

    for (i = 0; i < 3; i++)
      for (j = 0; j < 4; j++)
	dnums[4*i+j] = 4 * enr[i] + j;
    for (j = 0; j < 3; j++)
      dnums[12+j] = 4 * n_edges + 3 * elnr + j;
  }

  void HDivSymFESpace :: 
  GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
  }



  template <class MAT>
  void HDivSymFESpace::TransformMat (int elnr, bool boundary,
				     MAT & mat, TRANSFORM_TYPE tt) const
  {
    int nd;
    int ena[12], eoa[12];
    ARRAY<int> enums(12, ena), eorient(12, eoa);
    
    if (boundary)
      {
	GetMeshAccess().GetSElEdges (elnr, enums, eorient);
	nd = GetSFE (elnr).GetNDof();
      }
    else
      {
	GetMeshAccess().GetElEdges (elnr, enums, eorient);
	nd = GetFE (elnr).GetNDof();
      }
    
    int i, j, k, l;
    ARRAY<double> fac(nd);
    fac = 1.0;
    for (i = 0; i < 3; i++)
      {
	fac[4*i+0] = eorient[i];
	fac[4*i+1] = eorient[i];
      }

    if (tt & TRANSFORM_MAT_LEFT)
      for (k = 0; k < dimension; k++)
	for (i = 0; i < nd; i++)
	  for (j = 0; j < mat.Width(); j++)
	    mat(k+i*dimension, j) *= fac[i];
    
    if (tt & TRANSFORM_MAT_RIGHT)
      for (l = 0; l < dimension; l++)
	for (i = 0; i < mat.Height(); i++)
	  for (j = 0; j < nd; j++)
	    mat(i, l+j*dimension) *= fac[j];
  }


  
  template <class VEC>
  void HDivSymFESpace::TransformVec (int elnr, bool boundary,
				     VEC & vec, TRANSFORM_TYPE tt) const
  {
    int nd;
    int ena[12], eoa[12];
    ARRAY<int> enums(12, ena), eorient(12, eoa);

    if (boundary)
      {
	GetMeshAccess().GetSElEdges (elnr, enums, eorient);
	nd = GetSFE (elnr).GetNDof();
      }
    else
      {
	GetMeshAccess().GetElEdges (elnr, enums, eorient);
	nd = GetFE (elnr).GetNDof();
      }

    ARRAY<double> fac(nd);
    fac = 1.0;
    for (int i = 0; i < 3; i++)
      {
	fac[4*i+0] = eorient[i];
	fac[4*i+1] = eorient[i];
      }
  
    if ((tt & TRANSFORM_RHS) || (tt & TRANSFORM_SOL))
      {
	for (int k = 0; k < dimension; k++)
	  for (int i = 0; i < nd; i++)
	    vec(k+i*dimension) *= fac[i];
      }
  }



  /* **************************** Integrator ********************** */



  
  class DiffOpIdHDivSym : public DiffOp<DiffOpIdHDivSym>
  {
  public:
    enum { DIM = 1 };
    enum { DIM_SPACE = 2 };
    enum { DIM_ELEMENT = 2 };
    enum { DIM_DMAT = 3 };
    enum { DIFFORDER = 0 };

    template <typename FEL, typename SIP, typename MAT>
    static void GenerateMatrix (const FEL & fel, const SIP & sip,
				MAT & mat, LocalHeap & lh)
    {
      FlatMatrix<> shape = fel.GetShape (sip.IP(), lh);
      int nd = fel.GetNDof();
      int i, j;
      for (i = 0; i < nd; i++)
	{
	  Mat<2> stress, trans, hmat;

	  stress(0,0) = shape(i,0);
	  stress(0,1) = stress(1,0) = shape(i,1);
	  stress(1,1) = shape(i,2);

	  trans = sip.GetJacobian() * stress * Trans(sip.GetJacobian());
	  trans /= sip.GetJacobiDet(); //  * sip.GetJacobiDet();

	  mat(0,i) = trans(0,0);
	  mat(1,i) = trans(0,1);
	  mat(2,i) = trans(1,1);
	}

      // transform normal and tangential dofs
      for (i = 0; i < 6; i++)
	for (j = 0; j < 3; j++)
	  {
	    Vec<2> dof, ndof;
	    dof(0) = mat(j,2*i);
	    dof(1) = mat(j,2*i+1);
	    //	    ndof = Trans(sip.GetJacobianInverse()) * dof;
	    ndof = Trans(sip.GetJacobianInverse()) * dof;

	    mat(j,2*i)   = ndof(0);
	    mat(j,2*i+1) = ndof(1);
	  }
    }

    /*
    template <typename FEL, typename SIP, class TVX, class TVY>
    static void Apply (const FEL & fel, const SIP & sip,
		       const TVX & x, TVY & y,
		       LocalHeap & lh) 
    {
      typedef typename TVX::TSCAL TSCAL;
      
      Vec<3,TSCAL> stressv = Trans (fel.GetShape(sip.IP(), lh)) * x;
      Mat<2,2,TSCAL> stress, trans;
      stress(0,0) = stressv(0);
      stress(0,1) = stressv(1);
      stress(1,0) = stressv(1);
      stress(1,1) = stressv(2);

      (*testout) << "Apply, ip = " << sip.IP() << endl
		 << "stress = " << stress << endl;

      trans = sip.GetJacobian() * stress * Trans(sip.GetJacobian());
      trans /= sip.GetJacobiDet() * sip.GetJacobiDet();
      
      (*testout) << "trans = " << trans << endl;
      
      y(0) = trans(0,0);
      y(1) = trans(0,1);
      y(2) = trans(1,1);
    }
    */
  };

  class HDivSymDMat : public DMatOp<HDivSymDMat>
  {
    CoefficientFunction * coef;
  public:
    enum { DIM_DMAT = 3 };
    
    HDivSymDMat (CoefficientFunction * acoef) : coef(acoef) { ; }
    
    template <typename FEL, typename SIP, typename MAT>
    void GenerateMatrix (const FEL & fel, const SIP & sip,
			 MAT & mat, LocalHeap & lh) const
    {
      mat = 0;
      mat(0,0) = mat(2,2) = Evaluate (*coef, sip);
      mat(1,1) = 2 * mat(0,0);
    }  
  };






  
  class DiffOpDivHDivSym : public DiffOp<DiffOpDivHDivSym>
  {
  public:
    enum { DIM = 1 };
    enum { DIM_SPACE = 2 };
    enum { DIM_ELEMENT = 2 };
    enum { DIM_DMAT = 2 };
    enum { DIFFORDER = 1 };

    template <typename FEL, typename SIP, typename MAT>
    static void GenerateMatrix (const FEL & fel, const SIP & sip,
				MAT & mat, LocalHeap & lh)
    {
      int nd = fel.GetNDof();
      int i, j;
      mat = sip.GetJacobian() * Trans (fel.GetDivShape(sip.IP(), lh));
      mat /= sip.GetJacobiDet();

      // transform normal and tangential dofs
      for (i = 0; i < 6; i++)
	for (j = 0; j < 3; j++)
	  {
	    Vec<2> dof, ndof;
	    dof(0) = mat(j,2*i);
	    dof(1) = mat(j,2*i+1);

	    ndof = Trans(sip.GetJacobianInverse()) * dof;

	    mat(j,2*i)   = ndof(0);
	    mat(j,2*i+1) = ndof(1);
	  }
    }
  };












  
  /// 
  template <typename FEL = HDivSymFiniteElement>
  class MassHDivSymIntegrator 
    : public T_BDBIntegrator<DiffOpIdHDivSym, HDivSymDMat, FEL>
  {
public:
    ///
    MassHDivSymIntegrator (CoefficientFunction * coeff)
      : T_BDBIntegrator<DiffOpIdHDivSym, HDivSymDMat, FEL> 
    (HDivSymDMat (coeff))
    {
      ;
    }
    
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MassHDivSymIntegrator (coeffs[0]);
    }

    ///
    virtual string Name () const { return "MassHDivSym"; }
  };
  



 
  /// 
  template <typename FEL = HDivSymFiniteElement>
  class DivHDivSymIntegrator 
    : public T_BDBIntegrator<DiffOpDivHDivSym, DiagDMat<2>, FEL>
  {
public:
    ///
    DivHDivSymIntegrator (CoefficientFunction * coeff)
      : T_BDBIntegrator<DiffOpDivHDivSym, DiagDMat<2>, FEL> 
    (DiagDMat<2> (coeff))
    {
      ;
    }
    
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new DivHDivSymIntegrator (coeffs[0]);
    }

    ///
    virtual string Name () const { return "DivHDivSym"; }
  };
  




  
  /// 
  template <typename FEL = HDivSymFiniteElement>
  class DivSourceHDivSymIntegrator 
    : public T_BIntegrator<DiffOpDivHDivSym, DVec<2>, FEL>
  {
public:
    ///
    DivSourceHDivSymIntegrator (CoefficientFunction * coeff1,
			     CoefficientFunction * coeff2)
      : T_BIntegrator<DiffOpDivHDivSym, DVec<2>, FEL> 
    (DVec<2> (coeff1, coeff2))
    {
      ;
    }
    
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new DivSourceHDivSymIntegrator (coeffs[0], coeffs[1]);
    }

    ///
    virtual string Name () const { return "DivSourceHDivSym"; }
  };
  



  /// 
  template <typename FEL = HDivSymFiniteElement>
  class SourceHDivSymIntegrator 
    : public T_BIntegrator<DiffOpIdHDivSym, DVec<3>, FEL>
  {
public:
    ///
    SourceHDivSymIntegrator (CoefficientFunction * coeff1,
			     CoefficientFunction * coeff2,
			     CoefficientFunction * coeff3)
      : T_BIntegrator<DiffOpIdHDivSym, DVec<3>, FEL> 
    (DVec<3> (coeff1, coeff2, coeff3))
    {
      ;
    }
    
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new SourceHDivSymIntegrator (coeffs[0], coeffs[1], coeffs[2]);
    }

    ///
    virtual string Name () const { return "SourceHDivSym"; }
  };
  




  /* ******************** Initialization *********************** */

  namespace InitAW {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("masshdivsym", 2, 1,
					MassHDivSymIntegrator<>::Create);
      GetIntegrators().AddBFIntegrator ("divhdivsym", 2, 1,
					DivHDivSymIntegrator<>::Create);
      GetIntegrators().AddLFIntegrator ("sourcehdivsym", 2, 3,
					SourceHDivSymIntegrator<>::Create);
      GetIntegrators().AddLFIntegrator ("divsourcehdivsym", 2, 2,
					DivSourceHDivSymIntegrator<>::Create);
    }

    Init init;
  }

}
