/*

Helmholtz discretization based on wave factorization

*/


#include <solve.hpp>

extern AutoPtr<ngsolve::PDE> pde;


/*
namespace ngfem
{
#include "../fem/pml.hpp"
}
*/



using namespace ngfem;
using namespace ngsolve;




/* ********************+ Integrator ************************** */

class DiffOpHelmholtzConvection : public DiffOp<DiffOpHelmholtzConvection>
{
  // 3 components:
  // u  u,x u,y 

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 3 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    typedef typename MAT::TSCAL TSCAL;
    const NodalFiniteElement & fel = 
      dynamic_cast<const NodalFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatMatrix<double> grad_ref(nd, 2, lh);
    FlatMatrix<TSCAL> grad(2, nd, lh);

    /*
    grad = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel.GetDShape(sip.IP(), lh));
    */

    fel.CalcDShape (sip.IP(), grad_ref);
    grad = Trans (sip.GetJacobianInverse ()) * Trans (grad_ref);

    FlatVector<TSCAL> vec(nd, lh);
    vec = fel.GetShape (sip.IP(), lh);

    for (int j = 0; j < nd; j++)
      {
	mat(0,j) = vec(j);
	mat(1,j) = grad(0,j);
	mat(2,j) = grad(1,j);
      }
  }
};


double omega = 1;
class DiffOpHelmholtzId : public DiffOp<DiffOpHelmholtzId>
{
  // 1 components:
  // u e^(i x)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const NodalFiniteElement & fel = 
      dynamic_cast<const NodalFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatVector<> vec(nd, lh);
    vec = fel.GetShape (sip.IP(), lh);

    double x = sip(0);
    double y = sip(1);
    Complex fac = exp (Complex(0,1)*omega*sqrt (x*x+y*y));
    for (int j = 0; j < nd; j++)
      mat(0,j) = ngbla::ReduceComplex<typename MAT::TSCAL> (fac) * vec(j);
  }



  template <typename FEL, typename SIP, class TVX, class TVY>
  static void Apply (const FEL & bfel, const SIP & sip,
		     const TVX & x, TVY & y,
                     LocalHeap & lh)
  {
    const NodalFiniteElement & fel = 
      dynamic_cast<const NodalFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatVector<> vec(nd, lh);
    vec = fel.GetShape (sip.IP(), lh);

    double hx = sip(0);
    double hy = sip(1);
    Complex fac = exp (Complex(0,1)*omega*sqrt (hx*hx+hy*hy));

    y = 0.0;
    for (int j = 0; j < nd; j++)
      y(0) += vec(j) * x(j);

    y(0) *=  ReduceComplex<typename TVY::TSCAL> (fac);
  }


  ///
  template <typename FEL, typename SIP, class TVX, class TVY>
  static void ApplyTrans (const FEL & fel, const SIP & sip,
			  const TVX & x, TVY & y,
			  LocalHeap & lh) 
  {
    cerr << "HelmholtzConvection :: ApplyTrans not impelemnted" << endl;
  }


};


class HelmholtzConvectionDMat : public DMatOp<HelmholtzConvectionDMat>
{
  CoefficientFunction * coef;
public:

  enum { DIM_DMAT = 3 };
  enum { SYMMETRIC = 0 };

  HelmholtzConvectionDMat (CoefficientFunction * acoef) : coef(acoef) { ; }

  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    double x = sip(0);
    double y = sip(1);
    
    mat = 0.0;

    double nx = -x / sqrt(x*x+y*y);
    double ny = -y / sqrt(x*x+y*y);

    mat(0,1) = nx;
    mat(0,2) = ny;
    mat(1,0) = -nx;
    mat(2,0) = -ny;
    
    omega = coef->Evaluate (sip);
    mat *= omega;
  }  
};


class HelmholtzConvectionIntegrator 
  : public T_BDBIntegrator<DiffOpHelmholtzConvection, HelmholtzConvectionDMat, FiniteElement>
{
public:
  ///
  HelmholtzConvectionIntegrator (CoefficientFunction * coeff)
    :  T_BDBIntegrator<DiffOpHelmholtzConvection, HelmholtzConvectionDMat, FiniteElement>
  (HelmholtzConvectionDMat (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HelmholtzConvectionIntegrator (coeffs[0]);
  }
  
  ///
  virtual string Name () const { return "HelmholtzConvection"; }
};






class PML_HelmholtzConvectionIntegrator 
  : public PML_BDBIntegrator<DiffOpHelmholtzConvection, HelmholtzConvectionDMat, FiniteElement>
{
public:
  ///
  PML_HelmholtzConvectionIntegrator (CoefficientFunction * coeff)
    :  PML_BDBIntegrator<DiffOpHelmholtzConvection, HelmholtzConvectionDMat, FiniteElement>
  (HelmholtzConvectionDMat (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new PML_HelmholtzConvectionIntegrator (coeffs[0]);
  }
  
  ///
  virtual string Name () const { return "PML_HelmholtzConvection"; }
};








class HelmholtzIdIntegrator 
  : public T_BDBIntegrator<DiffOpHelmholtzId, DiagDMat<1>, FiniteElement>
{
public:
  ///
  HelmholtzIdIntegrator (CoefficientFunction * coeff)
    :  T_BDBIntegrator<DiffOpHelmholtzId, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HelmholtzIdIntegrator (coeffs[0]);
  }
  
  ///
  virtual string Name () const { return "HelmholtzId"; }
};








int n_inf = 7;

class Infinite1_Integrator : public BilinearFormIntegrator
{
  double r;
  Matrix<double> mat_m, mat_k;

  LaplaceBoundaryIntegrator<2> lap_int;
  RobinIntegrator<2> mass_int;

public:
    
  Infinite1_Integrator ()
    : mat_m(n_inf), mat_k(n_inf),
      lap_int(new ConstantCoefficientFunction (1)),
      mass_int(new ConstantCoefficientFunction (1))
  {
    for (int i = 0; i < n_inf; i++)
      for (int j = 0; j < n_inf; j++)
	{
	  double di = i+0.5;
	  double dj = j+1.5;
	  mat_k(i,j) = 1.0 / (di+dj);
	  mat_m(i,j) = di*dj / (di+dj);
	}
    (*testout) << "infinite1, matk = " << endl << mat_k << endl;
    (*testout) << "infinite1, matm = " << endl << mat_m << endl;
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new Infinite1_Integrator ();
  }
    
  virtual string Name () const { return "Helmholtz infinite 2"; }    
    
    
  virtual bool BoundaryForm () const { return 1; }

  virtual void
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & locheap) const
  { 
    FlatMatrix<double> lap_mat, mass_mat;
      
    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);

    (*testout) << "lap_mat = " << endl << lap_mat << endl;
      
    (*testout) << "mass_mat = " << endl << mass_mat << endl;
      
    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;
      
    elmat.AssignMemory (n*m, n*m, locheap);
      
    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }
};







class Infinite2_Integrator : public BilinearFormIntegrator
{
  double r;
  Matrix<double> mat_m, mat_k;

  LaplaceBoundaryIntegrator<2> lap_int;
  RobinIntegrator<2> mass_int;

public:
    
  Infinite2_Integrator ()
    : mat_m(n_inf), mat_k(n_inf),
      lap_int(new ConstantCoefficientFunction (1)),
      mass_int(new ConstantCoefficientFunction (1))
  {
    mat_k = 0.0;
      
    for (int i = 0; i < n_inf; i++)
      for (int j = 0; j < n_inf; j++)
	{
	  double di = i+0.5;
	  double dj = j+1.5;
	  mat_m(i,j) = (dj-di) / (di+dj-1);
	}

    (*testout) << "infinite2, matk = " << endl << mat_k << endl;
    (*testout) << "infinite2, matm = " << endl << mat_m << endl;

  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new Infinite2_Integrator ();
  }
    
  virtual string Name () const { return "Helmholtz infinite 2"; }    
    
    
  virtual bool BoundaryForm () const { return 1; }

  virtual void
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & locheap) const
  { 
    FlatMatrix<double> lap_mat, mass_mat;
      
    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);
      
    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;
      
    elmat.AssignMemory (n*m, n*m, locheap);
      
    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }
};




























///
class NumProcHelmholtzTrans : public NumProc
{
protected:
  int num;
public:
  ///
  NumProcHelmholtzTrans (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    num = int (flags.GetNumFlag ("num", 300)); 
  }


  ///
  virtual ~NumProcHelmholtzTrans()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcHelmholtzTrans (pde, flags);
  }



  ///
  virtual void Do()
  {
    cout << "solve helmholtz with wave factorization" << endl;

    if (!pde.GetBilinearForm("a")->GetFESpace().IsComplex())
      {
	SparseMatrix<double> & mata =
	  dynamic_cast<SparseMatrix<double> & >
	  (pde.GetBilinearForm ("a")->GetMatrix());


	SparseMatrix<double> & matb =
	  dynamic_cast<SparseMatrix<double> & >
	  (pde.GetBilinearForm ("b")->GetMatrix());


	GridFunction & gfu = *pde.GetGridFunction ("u");


	(*testout) << "mata = " << endl << mata << endl;
	(*testout) << "matb = " << endl << matb << endl;


	int n = mata.Height();


	// Arnoldi solver


	VVector<double> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);


	PardisoInverse<double> inva(mata);
	// SparseCholesky<double> inva(mata);


	// Arnoldi:
	int m = min2 (num, n);
    
	Matrix<double> matV (m,n); 
	Matrix<Complex> matH(m), matHt(m);


	hv.SetRandom();
	matV = 0.0;
	matH = Complex(0.0);
    
	double len = sqrt (S_InnerProduct<double> (hv, hv));
	hv /= len;

	for (int i = 0; i < m; i++)
	  {
	    cout << "i = " << i << endl;
	    for (int j = 0; j < n; j++)
	      matV(i,j) = hv(j);
	
	    hva = matb * hv;
	    hvm = inva * hva;
	
	    for (int j = 0; j <= i; j++)
	      {
		double sum = 0.0;
		for (int k = 0; k < n; k++)
		  sum += hvm(k) * matV(j,k);
		matH(j,i) = sum;
		for (int k = 0; k < n; k++)
		  hvm(k) -= sum * matV(j,k);
	      }
	
	    hv = hvm;
	    double len = sqrt (S_InnerProduct<double> (hv, hv));
	    if (i<m-1)
	      matH(i+1,i) = len; 
	
	    //  cout << ", len = " << len << endl;
	    hv /= len;
	  }
    

	(*testout) << "hessenberg = " << endl << matH << endl;
	cout << "has Hessenberg" << endl;
    
	Vector<Complex> lami(m);
	Matrix<Complex> evecs(m);
	// Matrix<Complex> levecs(n,m);
	matHt = Trans (matH);
    
	evecs = Complex (0.0);
	lami = Complex (0.0);
    
    
	LapackHessenbergEP (matH.Height(), &matHt(0,0), &lami(0), &evecs(0,0));
    
	cout << "hessenberg is back" << endl;

	for (int i = 0; i < m; i++)
	  lami(i) =  1.0 / lami(i);

	ofstream evout("eigenvalues.out");
	for (int i = 0; i < m; i++)
	  {
	    evout << i << " " << lami(i).real() << " " << lami(i).imag();
	    Complex inv = 1.0 / lami(i);
	    evout << " " << inv.real() << " " << inv.imag() << endl;
	  }



	Vector<Complex> ev(m);
	Vector<Complex> lev(n);

	int cnt = 0;
	for (int i = 0; i < m; i++)
	  {
	    cout << "compute eigenvector " << i << endl;

	    for (int j = 0; j < m; j++)
	      ev(j) = evecs(i,j);

	    lev = Trans (matV) * ev;

	    FlatVector<Complex> vecu = gfu.GetVector(cnt).FVComplex();
	    vecu = lev;
	
	    cnt++;
	    if (cnt >= gfu.GetMultiDim()) break;
	  }
      }
    else
      {
	SparseMatrix<Complex> & mata =
	  dynamic_cast<SparseMatrix<Complex> & >
	  (pde.GetBilinearForm ("a")->GetMatrix());


	SparseMatrix<Complex> & matb =
	  dynamic_cast<SparseMatrix<Complex> & >
	  (pde.GetBilinearForm ("b")->GetMatrix());


	GridFunction & gfu = *pde.GetGridFunction ("u");


	(*testout) << "mata = " << endl << mata << endl;
	(*testout) << "matb = " << endl << matb << endl;


	int n = mata.Height();


	// Arnoldi solver


	VVector<Complex> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);


	// PardisoInverse<Complex> inva(mata);
	SparseCholesky<Complex> inva(mata);


	// Arnoldi:
	int m = min2 (num, n);
    
	Matrix<Complex> matV (m,n); 
	Matrix<Complex> matH(m), matHt(m);


	hv.SetRandom();
	matV = Complex(0.0);
	matH = Complex(0.0);
    
	Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
	hv /= len;

	for (int i = 0; i < m; i++)
	  {
	    cout << "i = " << i << endl;
	    for (int j = 0; j < n; j++)
	      matV(i,j) = hv(j);
	
	    hva = matb * hv;
	    hvm = inva * hva;
	
	    for (int j = 0; j <= i; j++)
	      {
		Complex sum = 0.0;
		for (int k = 0; k < n; k++)
		  sum += hvm(k) * matV(j,k);
		matH(j,i) = sum;
		for (int k = 0; k < n; k++)
		  hvm(k) -= sum * matV(j,k);
	      }
	
	    hv = hvm;
	    Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
	    if (i<m-1)
	      matH(i+1,i) = len; 
	
	    //  cout << ", len = " << len << endl;
	    hv /= len;
	  }
    

	(*testout) << "hessenberg = " << endl << matH << endl;
	cout << "has Hessenberg" << endl;
    
	Vector<Complex> lami(m);
	Matrix<Complex> evecs(m);
	// Matrix<Complex> levecs(n,m);
	matHt = Trans (matH);
    
	evecs = Complex (0.0);
	lami = Complex (0.0);
    
	LapackHessenbergEP (matH.Height(), &matHt(0,0), &lami(0), &evecs(0,0));
    
	cout << "hessenberg is back" << endl;

	//	for (int i = 0; i < m; i++)
	//	  lami(i) =  1.0 / lami(i);

	ofstream evout("eigenvalues.out");
	for (int i = 0; i < m; i++)
	  {
	    evout << i << " " << lami(i).real() << " " << lami(i).imag();
	    Complex inv = 1.0 / lami(i);
	    evout << " " << inv.real() << " " << inv.imag();
	    inv = sqrt (1.0 / lami(i));
	    evout << " " << inv.real() << " " << inv.imag() << endl;
	  }


	Vector<Complex> ev(m);
	Vector<Complex> lev(n);

	int cnt = 0;
	for (int i = 0; i < m; i++)
	  {
	    cout << "compute eigenvector " << i << endl;

	    for (int j = 0; j < m; j++)
	      ev(j) = evecs(i,j);

	    lev = Trans (matV) * ev;

	    FlatVector<Complex> vecu = gfu.GetVector(cnt).FVComplex();
	    vecu = lev;
	
	    cnt++;
	    if (cnt >= gfu.GetMultiDim()) break;
	  }
      }

    cout << "eigensystem done" << endl;
    
  }


};







///
class NumProcHelmholtzTransBVP : public NumProc
{
protected:
  double taur, taui;
  int maxsteps;
public:
  ///
  NumProcHelmholtzTransBVP (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    taur = (flags.GetNumFlag ("taur", 0.1)); 
    taui = (flags.GetNumFlag ("taui", 0.1)); 
    maxsteps = int(flags.GetNumFlag ("maxsteps", 1000)); 
  }


  ///
  virtual ~NumProcHelmholtzTransBVP ()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcHelmholtzTransBVP (pde, flags);
  }



  ///
  virtual void Do()
  {
    cout << "solve helmholtz with wave factorization" << endl;


    SparseMatrix<Complex> & mat =
      dynamic_cast<SparseMatrix<Complex> & >
      (pde.GetBilinearForm ("a")->GetMatrix());

    const BaseMatrix & pre =
      pde.GetPreconditioner ("c")->GetMatrix();


    BaseVector & vecf = pde.GetLinearForm ("f")->GetVector();
    BaseVector & vecfone = pde.GetLinearForm ("fone")->GetVector();
    BaseVector & vecu = pde.GetGridFunction ("u")->GetVector();

    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();
    BaseVector & wc = *vecu.CreateVector();
    BaseVector & uone = *vecu.CreateVector();
    BaseVector & aw = *vecu.CreateVector();
    BaseVector & aone = *vecu.CreateVector();

    int nv = ma.GetNV();
    uone = 0.0;
    for (int i = 0; i < nv; i++)
      uone.FVComplex()(i) = 1.0;

    aone = mat * uone;

    Mat<2,2, Complex> ms, invms;
    Vec<2, Complex> fs, us;

    FlatVector<Complex> fw = w.FVComplex();
    FlatVector<Complex> fwc = wc.FVComplex();

    /*
      for (int i = 0; i < 10; i++)
      {
      d = vecf - mat * vecu;
      cout << "i = " << i << ", err = " << L2Norm (d) << endl;

      w = pre * d;
      aw = mat * w;

	
      for (int j = 0; j < fw.Size(); j++)
      fwc(j) = conj (fw(j));

      fs(0) = S_InnerProduct<Complex> (d, wc);
      fs(1) = S_InnerProduct<Complex> (d, uone);

      ms(0,0) = S_InnerProduct<Complex> (aw, wc);
      ms(0,1) = S_InnerProduct<Complex> (aone, wc);

      ms(1,0) = S_InnerProduct<Complex> (aw, uone);
      ms(1,1) = S_InnerProduct<Complex> (aone, uone);

      CalcInverse (ms, invms);
      us = invms * fs;

      vecu += us(0) * w;
      vecu += us(1) * uone;
      }
    */


    /*
      Complex scale = 1.0 / S_InnerProduct<Complex> (vecfone, uone);

      for (int i = 0; i < 10000; i++)
      {
      d = vecf - mat * vecu;
      d -= (scale * S_InnerProduct<Complex> (d, uone)) * vecfone;

      cout << "i = " << i << ", err = " << L2Norm (d) << endl;
	
      w = pre * d;
      vecu += Complex (0.001, 0) * w;

      cout << "inner1 =                                    " << S_InnerProduct<Complex> (vecu, vecfone) << endl;
      vecu -= (scale * S_InnerProduct<Complex> (vecu, vecfone)) * uone;
      cout << "inner =                                     " << S_InnerProduct<Complex> (vecu, vecfone) << endl;
      }
    */


    /*
      for (int i = 0; i < maxsteps; i++)
      {
      d = vecf - mat * vecu;
      cout << "i = " << i << ", err = " << L2Norm (d) << endl;
	
      w = pre * d;
      vecu += Complex (taur, taui) * w;
      }
    */


    // PardisoInverse<Complex> inv(mat);
    // vecu = inv * vecf;
  }
};


































namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetIntegrators().AddBFIntegrator ("helmholtzconvection", 2, 1,
				      HelmholtzConvectionIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("PML_helmholtzconvection", 2, 1,
				      PML_HelmholtzConvectionIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("helmholtzid", 2, 1,
				      HelmholtzIdIntegrator::Create);


    GetIntegrators().AddBFIntegrator ("infinite1", 2, 0,
				      Infinite1_Integrator::Create);
    GetIntegrators().AddBFIntegrator ("infinite2", 2, 0,
				      Infinite2_Integrator::Create);

    GetNumProcs().AddNumProc ("helmholtz_trans", NumProcHelmholtzTrans::Create);
    GetNumProcs().AddNumProc ("helmholtz_trans_bvp", NumProcHelmholtzTransBVP::Create);
  }
  
  Init init;
}


