/*

unit-cell: domain = 1 / 2
(0,1)^2

global problem:
domain = 4 ... homogenized
domain = 3 ... standard
*/




#include <solve.hpp>

using namespace ngsolve;






class PeriodicVertex : public SpecialElement
{
  const MeshAccess & ma;
  int vnr1, vnr2;

public:
  PeriodicVertex (const MeshAccess & ama,
		  int avnr1, int avnr2)
    : ma(ama), vnr1(avnr1), vnr2(avnr2)
  { ; }
  
  void GetDofNrs (ARRAY<int> & dnums) const
  {
    dnums.SetSize (2);
    dnums[0] = vnr1;
    dnums[1] = vnr2;
  }

  virtual void Assemble (FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    elmat.AssignMemory (2, 2, lh);
    elmat = 0;
    elmat(0,0) = elmat(1,1) = 1e8;
    elmat(0,1) = elmat(1,0) = -1e8;
  }
};



class PeriodicEdge : public SpecialElement
{
  const MeshAccess & ma;
  int order;
  int ednr1, ednr2;

public:
  PeriodicEdge (const MeshAccess & ama,
			  int aorder,
			  int aednr1, int aednr2)
    : ma(ama), order(aorder), ednr1(aednr1), ednr2(aednr2)
  { ; }
  
  void GetDofNrs (ARRAY<int> & dnums) const
  {
    dnums.SetSize (2*(order-1));

    int nv = ma.GetNV();
    for (int i = 0; i < (order-1); i++)
      {
	dnums[2*i]   = nv+ednr1*(order-1)+i;
	dnums[2*i+1] = nv+ednr2*(order-1)+i;
      }
  }

  virtual void Assemble (FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    elmat.AssignMemory (2*(order-1), 2*(order-1), lh);
    elmat = 0;
    for (int i = 0; i < order-1; i++)
      {
	elmat(2*i,2*i) = elmat(2*i+1,2*i+1) = 1e8;
	elmat(2*i,2*i+1) = elmat(2*i+1,2*i) = -1e8;
      }
  }
};









class NumProcHomogenization : public NumProc
{
protected:
  BilinearForm * bfa_unit;
  BilinearForm * bfa_global;

  BilinearForm * bfm_unit;

  // linear-form providing the right hand side
  LinearForm * lff;
  // solution vector
  GridFunction * gfu;

  
  GridFunction * gfevecs;


public:
    
  NumProcHomogenization (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa_unit = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform_auc", "auc"));
    bfm_unit = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform_muc", "muc"));

    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    gfevecs = pde.GetGridFunction (flags.GetStringFlag ("eigenvectors", "evecs"));
  }

  virtual ~NumProcHomogenization() 
  { ; }


  // creates an solver object
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcHomogenization (pde, flags);
  }

  // solve at one level
  virtual void Do(LocalHeap & lh)
  {
    cout << "solve homogenization" << endl;


    // add periodic constraints:

    ARRAY<INT<2> > vertex_pairs;
    ma.GetPeriodicVertices (vertex_pairs);
    cout << "vertex pairs = " << endl << vertex_pairs << endl;

    ARRAY<INT<2> > edge_pairs;
    ma.GetPeriodicEdges (edge_pairs);
    cout << "edge pairs = " << endl << edge_pairs << endl;


    FESpace * space = pde.GetFESpace ("v");
    for (int i = 0; i < vertex_pairs.Size(); i++)
      space -> specialelements.Append (new PeriodicVertex (ma, vertex_pairs[i][0], vertex_pairs[i][1]));

    for (int i = 0; i < edge_pairs.Size(); i++)
      space -> specialelements.Append (new PeriodicEdge (ma, space->GetOrder(), edge_pairs[i][0], edge_pairs[i][1]));

    bfa_unit->ReAssemble (lh, 1);




      
    // solve evp
    
    const BaseMatrix & mata = bfa_unit->GetMatrix();
    const BaseMatrix & matm = bfm_unit->GetMatrix();

    int num = gfevecs->GetMultiDim();

    for (int i = 0; i < num; i++)
      gfevecs->GetVector(i).SetRandom();

    ARRAY<BaseVector*> vecui(2*num);
    for (int i = 0; i < 2*num; i++)
      vecui[i] = gfevecs->GetVector().CreateVector();

    Vector<> lami(2*num);


    BaseVector & d = *gfevecs->GetVector().CreateVector();
    BaseVector & w = *gfevecs->GetVector().CreateVector();
    BaseVector & au = *gfevecs->GetVector().CreateVector();
    BaseVector & mu = *gfevecs->GetVector().CreateVector();

    Matrix<double> small_a(2*num), small_m(2*num);

    SparseCholesky<double> inva (dynamic_cast<const SparseMatrix<double>&> (mata));

    double lam_old = 1e10;

    int evpits = 50;
    for (int it = 0; it < evpits; it++)
      {
	cout << "it = " << it << " " << flush;

	for (int ei = 0; ei < num; ei++)
	  *vecui[ei] = gfevecs->GetVector(ei);
	
	for (int ei = 0; ei < num; ei++)
	  {
	    au = mata * *vecui[ei];
	    mu = matm * *vecui[ei];
		
	    double lam = 
	      InnerProduct (*vecui[ei], au) /
	      InnerProduct (*vecui[ei], mu);
	    
	    d = au - lam * mu;

	    w = inva * d;
	    d = mata * w;

	    double norm = sqrt (InnerProduct (w, d=matm*w));
	    *vecui[num+ei] = (1.0/norm) * w;
	    
	    for (int ej  = 0; ej <= ei; ej++)
	      {
		small_a(ei, ej) = small_a(ej,ei) = InnerProduct (au, *vecui[ej]);
		small_m(ei, ej) = small_m(ej,ei) = InnerProduct (mu, *vecui[ej]);
	      }
	  }

	for (int ei = num; ei < 2*num; ei++)
	  {
	    au = mata * *vecui[ei];
	    mu = matm * *vecui[ei];
	    
	    for (int ej  = 0; ej <= ei; ej++)
	      {
		small_a(ei, ej) = small_a(ej,ei) = InnerProduct (au, *vecui[ej]);
		small_m(ei, ej) = small_m(ej,ei) = InnerProduct (mu, *vecui[ej]);
	      }
	  }
	

	LapackGHEPEPairs (2*num, &small_a(0,0), &small_m(0,0), &lami(0));

	for (int ei = 0; ei < num; ei++)
	  {
	    au = 0;
	    for (int i = 0; i < 2*num; i++)
	      au += small_a(ei,i) * *vecui[i];
	    
	    gfevecs->GetVector(ei) = au;
	  }


	cout << "err = " << lami(num-1)-lam_old << endl;
	if (fabs (lami(num-1)-lam_old) < 1e-8 * fabs(lami(num-1))) break;
	lam_old = lami(num-1);
      }
    
    for (int ei = 0; ei < num; ei++)
      cout << "lam(" << ei << ") = " << lami(ei) << endl;

    delete &d;
    delete &w;
    delete &au;
    delete &mu;


    
    
    testout -> precision(10);
    
    FESpace * fesglob = pde.GetFESpace ("vglob");
    int ndglob = fesglob -> GetNDof();

    FESpace * fesuc = pde.GetFESpace ("v");
    int nduc = fesuc -> GetNDof();

    ARRAY<int> dnumsglob, dnumsuc;

    cout << "ndglob = " << ndglob << endl;


    ARRAY<int> compress(num*ndglob);
    compress = 0;
    for (int elnr = 0; elnr < ma.GetNE(); elnr++)
      {
	if (ma.GetElIndex(elnr) <= 0) continue;
	fesglob -> GetDofNrs (elnr, dnumsglob);
	for (int i = 0; i < num; i++)
	  for (int j = 0; j < dnumsglob.Size(); j++)
	    compress[dnumsglob[j]+i*ndglob] = 1;
      }

    int useddofs = 0;
    for (int i = 0; i < compress.Size(); i++)
      if (compress[i])
	{
	  compress[i] = useddofs; 
	  useddofs++;
	}
      else
	compress[i] = -1;


    cout << "useddofs = " << useddofs << endl;

    Matrix<double> mat (useddofs);
    Vector<double> f(useddofs), u(useddofs);

    mat = 0;
    f = 0;

    ElementTransformation eltrans;

    void * heapp2 = lh.GetPointer();
    for (int elnr = 0; elnr < ma.GetNE(); elnr++)
      {
	if (ma.GetElIndex(elnr) <= 0) continue;
	cout << "." << flush;
	const NodalFiniteElement & felglob = 
	  dynamic_cast<const NodalFiniteElement&> (fesglob->GetFE(elnr, lh));
	fesglob -> GetDofNrs (elnr, dnumsglob);


	FlatMatrix<> elmat(num*felglob.GetNDof(), lh);
	elmat = 0;

	FlatVector<> elvec(num*felglob.GetNDof(), lh);
	elvec = 0;


	void * heapp1 = lh.GetPointer();
	for (int eluc = 0; eluc < ma.GetNE(); eluc++)
	  {
	    if (ma.GetElIndex(eluc) > 0) continue;

	    const NodalFiniteElement & feluc = 
	      dynamic_cast<const NodalFiniteElement&> (fesuc->GetFE(eluc, lh));
	    fesuc -> GetDofNrs (eluc, dnumsuc);

	    ma.GetElementTransformation (eluc, eltrans, lh);
	    

	    const IntegrationRule & ir =
	      GetIntegrationRules().SelectIntegrationRule (feluc.ElementType(), 2*feluc.Order()+2*felglob.Order());
	    

	    FlatMatrix<> elvecsuc(num, dnumsuc.Size(), lh);
	    for (int i = 0; i < num; i++)
	      gfevecs->GetVector(i).GetIndirect (dnumsuc, elvecsuc.Row(i));


	    FlatVector<> ui(num, lh);
	    FlatMatrix<> dui(num, 2, lh);

	    FlatMatrix<> bmat(2, num*felglob.GetNDof(), lh);
	    bmat = 0;

	    FlatVector<> bvec(num*felglob.GetNDof(), lh);
	    bvec = 0;


	    void * heapp = lh.GetPointer();
	    for (int ip = 0; ip < ir.GetNIP(); ip++)
	      {
		SpecificIntegrationPoint<2,2> sip(ir[ip], eltrans, lh);

		FlatVector<> shapeglob(felglob.GetNDof(), lh);
		FlatVector<> shapeuc (feluc.GetNDof(), lh);

		FlatMatrix<> dshapeglob (felglob.GetNDof(), 2, lh);
		FlatMatrix<> hdshapeuc  (feluc.GetNDof(), 2, lh);
		FlatMatrix<> dshapeuc  (feluc.GetNDof(), 2, lh);

		Vec<2> p = sip.GetPoint();
		IntegrationPoint ipglob (p(0), p(1));

		felglob.CalcShape (ipglob, shapeglob);
		felglob.CalcDShape (ipglob, dshapeglob);

		feluc.CalcShape (ir[ip], shapeuc);
		feluc.CalcDShape (ir[ip], hdshapeuc);
		dshapeuc = hdshapeuc * sip.GetJacobianInverse();

		ui = elvecsuc * shapeuc;
		dui = elvecsuc * dshapeuc;
		
		
		for (int i = 0; i < num; i++)
		  for (int j = 0; j < shapeglob.Size(); j++)
		    for (int k = 0; k < 2; k++)
		      bmat(k, i*shapeglob.Size()+j) = ui(i)*dshapeglob(j,k) + dui(i,k)*shapeglob(j);

		for (int i = 0; i < num; i++)
		  for (int j = 0; j < shapeglob.Size(); j++)
		    bvec(i*shapeglob.Size()+j) = ui(i)*shapeglob(j);
		
		elmat += (fabs (sip.GetJacobiDet())*ir[ip].Weight()) *  (Trans (bmat) * bmat);
		elvec += (fabs (sip.GetJacobiDet())*ir[ip].Weight()) * bvec;

		lh.CleanUp(heapp);
	      }

	    lh.CleanUp(heapp1);
	  }
	(*testout) << "elmat = " << endl << elmat << endl;
	(*testout) << "elvec = " << endl << elvec << endl;



	
	for (int i = 0; i < num; i++)
	  for (int j = 0; j < dnumsglob.Size(); j++)
	    {
	      f(compress[dnumsglob[j] + i*ndglob]) += elvec(j+i*dnumsglob.Size());

	      for (int k = 0; k < num; k++)
		for (int l = 0; l < dnumsglob.Size(); l++)
		  mat(compress[dnumsglob[j]+i*ndglob], compress[dnumsglob[l]+k*ndglob]) +=
		    elmat(j+i*dnumsglob.Size(), l+k*dnumsglob.Size());
	    }


	lh.CleanUp(heapp2);
      }


    cout << "implement bc" << endl;

    for (int elnr = 0; elnr < ma.GetNSE(); elnr++)
      {
	if (ma.GetSElIndex(elnr) != 4) continue;

	const NodalFiniteElement & fel = 
	  dynamic_cast<const NodalFiniteElement&> (fesglob->GetSFE(elnr, lh));
	fesglob -> GetSDofNrs (elnr, dnumsglob);

	for (int i = 0; i < num; i++)
	  for (int j = 0; j < dnumsglob.Size(); j++)
	    mat(compress[dnumsglob[j]+i*ndglob], compress[dnumsglob[j]+i*ndglob]) += 1e15;
	
	lh.CleanUp(heapp2);
      }


    (*testout) << "f = " << f << endl;

    Matrix<> invmat(mat.Height());

    cout << "calc inverse" << endl;

    LapackInverse (mat);

    u = mat * f;
    cout.precision(10);
    cout << "u*f = " << InnerProduct (u,f) << endl;
  }




  virtual string GetClassName () const
  {
    return "Homogenization Solver (Demo)";
  }

  virtual void PrintReport (ostream & ost)
  {
    /*
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Bilinear-form M = " << bfm->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl
	<< "dt              = " << dt << endl
	<< "tend            = " << tend << endl;
    */
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*
    ost << 
      "\n\nNumproc Homogenization:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};





// declare the numproc 'homogenization' 

namespace
{
  class Init
  { 
  public: 
    Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("homogenization", NumProcHomogenization::Create, NumProcHomogenization::PrintDoc);
  }
    
  Init init;
}
  
