/*

Mixed and hybrid finite elements

*/

#include <solve.hpp>


namespace  ngsolve
{


  class EquilibriumFESpace;





///
class EquilibriumFESpace : public FESpace
{
public:
  ///
  int nfa;
  /// dofs per face
  int n_face_dofs;
  ///
  bool peers;
  ///
  ARRAY<int> ndlevel;
public:
  ///
  EquilibriumFESpace (const MeshAccess & ama,
		      int aorder, int adim, bool acomplex);

  ///
  ~EquilibriumFESpace ();
  ///
  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  ///
  virtual string GetClassName() const
    { return "Equilibrium"; }

  ///
  virtual void Update();

  ///
  virtual int GetNDof () const;
  ///
  virtual int GetNDofLevel (int level) const;

  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;

  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;

  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  ///
  virtual Table<int> * CreateSmoothingBlocks (int type = 0) const;
  ///
  virtual ARRAY<int> * CreateDirectSolverClusters (int type = 0) const;
};







  class BaseEquilibriumProlongation : public Prolongation
{
protected:
  ///
  const MeshAccess & ma;
  ///
  const EquilibriumFESpace & space;
  ///
  ARRAY<int> parentfaces;
  ARRAY<ngstd::INT<2> > parents;
public:
  BaseEquilibriumProlongation (const EquilibriumFESpace & aspace)
    : ma(aspace.GetMeshAccess()), space(aspace) { ; }
  ///
  virtual void Update ();
  virtual BitArray * GetInnerDofs () const;
};


template <class TV>
class EquilibriumProlongation : public BaseEquilibriumProlongation
{
public:
  ///
  EquilibriumProlongation(const EquilibriumFESpace & aspace)
    : BaseEquilibriumProlongation (aspace) { ; }
  ///
  virtual ~EquilibriumProlongation() { ; }
  
  ///
  virtual void ProlongateInline (int finelevel, BaseVector & v) const
  {
    (*testout) << "prol, fine level = " << finelevel << endl;

    int nc = space.GetNDofLevel (finelevel-1);
    int nf = space.GetNDofLevel (finelevel);

    (*testout) << "nc = " << nc << " nf = " << nf << endl;
    
    //FlatVector<TV> & fv = 
    //      dynamic_cast<VFlatVector<TV> &> (v).FV();
    FlatVector<TV> fv = 
      dynamic_cast<T_BaseVector<TV> &> (v).FV();

    int i;
    for (i = nc; i < nf; i++)
      {
	if (parents[i][0] != -1)
	  {
	    fv(i) = 0.5 * (fv(parents[i][0]) + 
			   fv(parents[i][1]));
	  }
	else
	  fv(i) = 0;
      }
    for (i = 0; i < nc; i++)
      fv(i) = 0;
    (*testout) << "prol complete" << endl;
  }

  ///
  virtual void RestrictInline (int finelevel, BaseVector & v) const
  {
    (*testout) << "prol, fine level = " << finelevel << endl;

    int nc = space.GetNDofLevel (finelevel-1);
    int nf = space.GetNDofLevel (finelevel);

    (*testout) << "nc = " << nc << " nf = " << nf << endl;
    
    //    FlatVector<TV> & fv = 
    //      dynamic_cast<VFlatVector<TV> &> (v).FV();
    FlatVector<TV> fv = 
      dynamic_cast<T_BaseVector<TV> &> (v).FV();

    int i;
    for (i = 0; i < nc; i++)
      fv(i) = 0;
    for (i = nf-1; i >= nc; i--)
      {
	if (parents[i][0] != -1)
	  {
	    fv(parents[i][0]) += 0.5 * fv(i);
	    fv(parents[i][1]) += 0.5 * fv(i);
	  }
	fv(i) = 0;
      }
  }
};










void BaseEquilibriumProlongation :: Update ()
{
  cout << "Update EquilibriumProlongation" << endl;
  int i, j, k;

  int nfa = ma.GetNFaces();
  int oldnfa = parentfaces.Size();

  HashTable<INT<3>, int> faces(nfa+1);
  ARRAY<int> face, pface;
  cout << "nfa = " << nfa << " oldnfa = " << oldnfa << endl;
  for (i = 0; i < nfa; i++)
    {
      ma.GetFacePNums (i, face);
      INT<3> f;
      for (j = 0; j < 3; j++)
	f[j] = face[j];
      f.Sort();
      faces.Set (f, i);
    }

  parentfaces.SetSize (nfa);
  for (i = oldnfa; i < nfa; i++)
    {
      parentfaces[i] = -1;

      ma.GetFacePNums (i, face);
      INT<3> f;
      for (j = 0; j < 3; j++)
	f[j] = face[j];
      f.Sort();

      int pa[2];
      do
	{
	  ma.GetParentNodes (f[2], pa);

	  int pa1 = pa[0];
	  int pa2 = pa[1];

	  if (pa1 == -1 || pa2 == -1)
	    break;
	  if (pa1 == f[0] || pa1 == f[1])
	    f[2] = pa2;
	  else if (pa2 == f[0] || pa2 == f[1])
	    f[2] = pa1;
	  else
	    break;

	  f.Sort();
	  if (faces.Used(f))
	    {
	      parentfaces[i] = faces.Get(f);
	      break;
	    }
	}
      while (1);
    }
  
  (*testout) << "parentfaces = " << endl << parentfaces << endl;

  ARRAY<int> nchilds(nfa);
  nchilds = 0;
  for (i = 0; i < nfa; i++)
    if (parentfaces[i] != -1)
      nchilds[parentfaces[i]]++;
  (*testout) << "nchilds = " << endl << nchilds << endl;

  
  parents.SetSize (4*nfa);
  for (i = 0; i < parents.Size(); i++)
    parents[i][0] = parents[i][1] = -1;

  for (i = 0; i < nfa; i++)
    {
      if (parentfaces[i] != -1)
	{
	  int pf = parentfaces[i];
	  ma.GetFacePNums (pf, pface);
	  ma.GetFacePNums (i, face);
	  for (j = 0; j < 3; j++)
	    {
	      bool found = 0;
	      for (k = 0; k < 3; k++)
		if (face[j] == pface[k])
		  {
		    parents[4*i+j][0] =
		      parents[4*i+j][1] = 4*pf+k;
		    found = 1;
		  }
	      if (!found)
		{
		  int pa[2];
		  ma.GetParentNodes (face[j], pa);
		  for (k = 0; k < 3; k++)
		    {
		      if (pa[0] == pface[k])
			parents[4*i+j][0] = 4*pf+k;
		      if (pa[1] == pface[k])
			parents[4*i+j][1] = 4*pf+k;
		    }
		}
	    }
	}
    }
  (*testout) << "parents = " << parents << endl;
}

BitArray * BaseEquilibriumProlongation :: GetInnerDofs () const
{
  BitArray * id = new BitArray (parents.Size());
  id->Clear();
  for (int i = 0; i < parents.Size(); i++)
    if (parents[i][0] == -1)
      id->Set(i);
  return id;
}









  EquilibriumFESpace :: 
  EquilibriumFESpace (const MeshAccess & ama,
		      int aorder, int adim, bool acomplex)
    : FESpace (ama, aorder, adim, acomplex)
  {
    peers = 0;
    if (ma.GetDimension() == 2)
      {
	/*
	trig = new CompositeFiniteElement<HDivFiniteElement<2>,NodalFiniteElement,NodalFiniteElement>
	  (new FE_BDMTrig1, new FE_Trig2, new FE_Trig1);
	segm = new FE_Segm1;
	*/

	trig = new CompositeFiniteElement<HDivHighOrderFiniteElement<2>,L2HighOrderFiniteElement>
	  (new HDivHighOrderTrig (aorder), new L2HighOrderTrig (aorder-1));
	segm = new H1HighOrderSegm (aorder);

	n_face_dofs = order+1;
      }
    else
      {
	tet = new CompositeFiniteElement<HDivFiniteElement<3>,NodalFiniteElement,NodalFiniteElement>
	  (new FE_BDMTet1, new FE_Tet2, new FE_Tet1);

	trig = new FE_Trig1;
	n_face_dofs = 4;
      }

    CreateVecObject1(prol, EquilibriumProlongation,
		     dimension, iscomplex, *this);

    Update();
  }

  EquilibriumFESpace :: ~EquilibriumFESpace ()
  {
    ;
  }


  FESpace * EquilibriumFESpace :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    int order = int(flags.GetNumFlag ("order", 0));
    int dim = int(flags.GetNumFlag ("dim", 1));
    bool iscomplex = flags.GetDefineFlag ("complex");

    return new EquilibriumFESpace (ma, order, dim, iscomplex);
  }
  



  void EquilibriumFESpace :: Update()
  {
    if (ma.GetDimension() == 2)
      nfa = GetMeshAccess().GetNEdges();
    else
      nfa = GetMeshAccess().GetNFaces();

    while (ndlevel.Size() < ma.GetNLevels())
      if (!peers)
	ndlevel.Append (nfa*n_face_dofs);
      else
	ndlevel.Append (nfa+ma.GetNP());

    cout << "nfa = " << nfa << endl;
    cout << "ndlevel = " << ndlevel << endl;
  }

  int EquilibriumFESpace :: GetNDof () const
  {
    return ndlevel.Last();
  }

  int EquilibriumFESpace :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }




  const FiniteElement & EquilibriumFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    int j;

    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	fe = tet; break;
      case ET_PYRAMID:
	fe = pyramid; break;
      case ET_PRISM:
	fe = prism; break;
      case ET_HEX:
	fe = hex; break;
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0; 
      }

    //    H1HighOrderFiniteElement * hofe =
    //      dynamic_cast<HDivHighOrderFiniteElement<*> (fe);

    const HDivHighOrderFiniteElement<2> * hofe1 =
      &dynamic_cast<CompositeFiniteElement<HDivHighOrderFiniteElement<2>,L2HighOrderFiniteElement>*> (fe) -> GetFE0();

    HDivHighOrderFiniteElement<2> * hofe =
      const_cast<HDivHighOrderFiniteElement<2>*> (hofe1);


    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ArrayMem<int, 12> ednums, order_ed;
    ArrayMem<int, 6> fanums, order_fa;

    ma.GetElVertices(elnr, vnums);
    ma.GetElEdges(elnr, ednums);
    ma.GetElFaces(elnr, fanums);

    order_ed.SetSize (ednums.Size());
    order_fa.SetSize (fanums.Size());

    for (j = 0; j < ednums.Size(); j++)
      order_ed[j] = order;
    for (j = 0; j < fanums.Size(); j++)
      order_fa[j] = order;

    if (!fe)
      {
	stringstream str;
	str << "EquilibriumFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    hofe -> SetVertexNumbers (vnums);
    hofe -> SetOrderEdgeNormal (order_ed);
    hofe -> SetOrderFace (order_fa);
    hofe -> SetOrderInner (order);

    return *fe;
  }
 



  void EquilibriumFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    int i, j;
    if (!peers)
      {
	if (ma.GetDimension() == 2)
	  {
	    ARRAY<int> ednums, edorient;
	    ma.GetElEdges (elnr, ednums, edorient);
	    dnums.SetSize(n_face_dofs * ednums.Size());

	    for (i = 0; i < ednums.Size(); i++)
	      for (j = 0; j < n_face_dofs; j++)
		dnums[n_face_dofs*i+j] = n_face_dofs*ednums[i]+j;
	  }
	else
	  {
	    ARRAY<int> fanums, faorient;
	    ma.GetElFaces (elnr, fanums, faorient);

	    switch (ma.GetElType (elnr))
	      {
	      case ET_TET:
		dnums.SetSize(4*3);
		for (i = 0; i < 4; i++)
		  {
		    int facedir = faorient[i];
		  
		    static const int reorder[8][3] =
		      { { 1, 2, 3 }, { 2, 1, 3 }, { 1, 3, 2 }, { 2, 3, 1 },
			{ 2, 1, 3 }, { 1, 2, 3 }, { 3, 1, 2 }, { 3, 2, 1 } };
		  
		    dnums[3*i+reorder[facedir][0]-1] = n_face_dofs*fanums[i];
		    dnums[3*i+reorder[facedir][1]-1] = n_face_dofs*fanums[i]+1;
		    dnums[3*i+reorder[facedir][2]-1] = n_face_dofs*fanums[i]+2;
		  }
		break;
		
#ifdef NONE
	      case ET_PRISM:
		dnums.SetSize(18);
		for (i = 1; i <= 2; i++)
		  {
		    int facedir = faorient.Get(i);
		  
		    static const int reorder[8][3] =
		      { { 1, 2, 3 }, { 2, 1, 3 }, { 1, 3, 2 }, { 2, 3, 1 },
			{ 2, 1, 3 }, { 1, 2, 3 }, { 3, 1, 2 }, { 3, 2, 1 } };
		  
		    dnums.Elem(3*i-3+reorder[facedir][0]) = 4*fanums.Get(i) - 3;
		    dnums.Elem(3*i-3+reorder[facedir][1]) = 4*fanums.Get(i) - 2;
		    dnums.Elem(3*i-3+reorder[facedir][2]) = 4*fanums.Get(i) - 1;
		  }
		for (i = 1; i <= 3; i++)
		  {
		    int facedir = faorient.Get(i+2);
		  
		    static const int reorder[8][8] =
		      {
			{ 1, 2, 3, 4 }, { 4, 3, 2, 1 }, 
			{ 2, 1, 4, 3 }, { 3, 4, 1, 2 },
			{ 1, 3, 2, 4 }, { 4, 2, 3, 1 }, 
			{ 2, 4, 1, 3 }, { 3, 1, 4, 2 },
		      };
		  
		    dnums.Elem(4*i+2+reorder[facedir][0]) = 4*fanums.Get(i+2) - 3;
		    dnums.Elem(4*i+2+reorder[facedir][1]) = 4*fanums.Get(i+2) - 2;
		    dnums.Elem(4*i+2+reorder[facedir][2]) = 4*fanums.Get(i+2) - 1;
		    dnums.Elem(4*i+2+reorder[facedir][3]) = 4*fanums.Get(i+2)    ;
		  }
		break;
#endif
	      }
	  }
      }
    else
      {
#ifdef NONE
	ARRAY<int> ednums, edorient, pnums;
	ma.GetElEdges (elnr, ednums, edorient);
	ma.GetElPNums (elnr, pnums);

	dnums.SetSize(6);
	for (i = 1; i <= 3; i++)
	  {
	    dnums.Elem(i) = ednums.Get(i);
	    dnums.Elem(i+3) = nfa+pnums.Get(i);
	  }
#endif
      }
  }


  void EquilibriumFESpace :: 
  GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    int i;
    if (!peers)
      {
	if (ma.GetDimension() == 2)
	  {
	    ARRAY<int> ednums, edorient;
	    ma.GetSElEdges (selnr, ednums, edorient);
	    dnums.SetSize(2);
	    dnums[0] = 2*ednums[0];
	    dnums[1] = 2*ednums[0]+1;
	  }
	else
	  {
	    int fanum, facedir;
	    ma.GetSElFace (selnr, fanum, facedir);

	    switch (ma.GetSElType(selnr))
	      {
	      case ET_TRIG:
		{
		  dnums.SetSize(3);
		
		  static const int reorder[8][3] =
		    { { 1, 2, 3 }, { 2, 1, 3 }, { 1, 3, 2 }, { 2, 3, 1 },
		      { 2, 1, 3 }, { 1, 2, 3 }, { 3, 1, 2 }, { 3, 2, 1 } };
		
		  dnums[reorder[facedir][0]-1] = n_face_dofs*fanum;
		  dnums[reorder[facedir][1]-1] = n_face_dofs*fanum+1;
		  dnums[reorder[facedir][2]-1] = n_face_dofs*fanum+2;
		  break;
		}
#ifdef NONE
	      case ET_QUAD:
		{
		  dnums.SetSize(4);

		  static const int reorder[8][8] =
		    {
		      { 1, 2, 3, 4 }, { 4, 3, 2, 1 }, 
		      { 2, 1, 4, 3 }, { 3, 4, 1, 2 },
		      { 1, 3, 2, 4 }, { 4, 2, 3, 1 }, 
		      { 2, 4, 1, 3 }, { 3, 1, 4, 2 },
		    };
		
		  dnums.Elem(reorder[facedir][0]) = 4*fanum - 3;
		  dnums.Elem(reorder[facedir][1]) = 4*fanum - 2;
		  dnums.Elem(reorder[facedir][2]) = 4*fanum - 1;
		  dnums.Elem(reorder[facedir][3]) = 4*fanum    ;
		  break;
		}
#endif
	      }
	  }
      }
    else
      {
#ifdef NONE
	if (ma.GetDimension() == 2)
	  {
	    ARRAY<int> ednums, edorient, pnums;
	    ma.GetSElEdges (selnr, ednums, edorient);
	    ma.GetSElPNums(selnr, pnums);

	    dnums.SetSize(3);
	    dnums.Elem(1) = ednums.Get(1);
	    dnums.Elem(2) = nfa+pnums.Get(1);
	    dnums.Elem(3) = nfa+pnums.Get(2);
	  }
#endif
      }
    //    (*testout) << "sel " << selnr << ", dnums = " << dnums << endl;
  }


  Table<int> * EquilibriumFESpace :: 
  CreateSmoothingBlocks (int type) const
  {
    int i, j, k;
    Table<int> * it;

    if (ma.GetDimension() == 2)
      {
	it = new Table<int> (nfa, n_face_dofs);
      
	for (i = 0; i < nfa; i++)
	  for (j = 0; j < n_face_dofs; j++)
	    (*it)[i][j] = i*n_face_dofs+j;
      }
    else
      {
	/*
	int nv = ma.GetNV();
	ARRAY<int> cnts(nv);
	ARRAY<int> pnums;
	for (k = 1; k <= 2; k++)
	  {
	    if (k == 2)
	      it = new Table<int>(cnts);
	    cnts = 0;

	    for (i = 0; i < nfa; i++)
	      {
		ma.GetFacePNums (i, pnums);
		
		if (k == 2)
		  {
		    for (j = 0; j < pnums.Size(); j++)
		      {
			(*it)[pnums[j]][cnts[pnums[j]]] = n_face_dofs * i + j;
		      }
		  }
		for (j = 0; j < pnums.Size(); j++)
		  cnts[pnums[j]]++;
	      }
	  }

	(*testout) << "Table = " << (*it) << endl;
	*/
	
	// edge blocks:
	
	HashTable<INT<2>, int> edges(ma.GetNEdges());
	ARRAY<int> pnums(3), orient(4);
	for (i = 0; i < ma.GetNEdges(); i++)
	  {
	    INT<2> edge;
	    ma.GetEdgePNums (i, edge[0], edge[1]);
	    edge.Sort();
	    edges.Set (edge, i);
	  }

	int minface = 1000000;
	for (i = 0; i < ma.GetNE(); i++)
	  {
	    ma.GetElFaces (i, pnums, orient);
	    for (j = 0; j < pnums.Size(); j++)
	      if (pnums[j] < minface)
		minface = pnums[j];
	  }
	cout << "minface = " << minface << endl;
	
	ARRAY<int> cnts(ma.GetNEdges());
	for (k = 1; k <= 2; k++)
	  {
	    if (k == 2)
	      it = new Table<int>(cnts);
	    cnts = 0;

	    for (i = minface; i < nfa; i++)
	      {
		ma.GetFacePNums (i, pnums);
		
		for (j = 0; j < 3; j++)
		  {
		    INT<2> edge;
		    edge[0] = pnums[j];
		    edge[1] = pnums[(j+1)%3];
		    edge.Sort();
		    
		    int edgenr = edges.Get(edge);

		    if (k == 2)
		      {
			(*it)[edgenr][cnts[edgenr]] = n_face_dofs * i + j;
			(*it)[edgenr][cnts[edgenr]+1] = n_face_dofs * i + (j+1)%3;
			//			(*it)[edgenr][cnts[edgenr]+2] = n_face_dofs * i + (j+2)%3;
		      }
		    cnts[edgenr]+=2;
		  }
	      }
	  }
	
	(*testout) << "Table = " << (*it) << endl;



  

	/*
	  int nfa = ma.GetNFaces();
	  int nv = ma.GetNP();
	  it = new IntTable(nv);
	  ARRAY<int> pnums;

	  for (i = 1; i <= nfa; i++)
	  {
	  ma.GetFacePNums (i, pnums);
	  for (j = 1; j <= 3; j++)
	  //	    for (k = 1; k <= 3; k++)
	  it->AddUnique (pnums.Get(j), 3*(i-1)+j);
	  }

	int nfa = ma.GetNFaces();
	it = new IntTable(nfa);
      
	for (i = 1; i <= nfa; i++)
	  {
	    it->AddUnique (i, 4*i-3);
	    it->AddUnique (i, 4*i-2);
	    it->AddUnique (i, 4*i-1);
	    it->AddUnique (i, 4*i);
	  }
	*/
      }
    return it;
  }




  ARRAY<int> * 
  EquilibriumFESpace :: CreateDirectSolverClusters (int type) const
  {
    int nd = GetNDof();
    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());
    clusters = 0;

    for (int i = 0; i < nfa; i++)
      clusters[n_face_dofs*i] = 1;

    return &clusters;
  }






  /* ************************ Numproc ***************************** */







  template <class SCAL>
  void CalcEquilibriumFlux (const MeshAccess & ma, 
			    const S_GridFunction<SCAL> & u,
			    S_GridFunction<SCAL> & flux,
			    S_GridFunction<SCAL> & disp,
			    const EquilibriumIntegrator & bli,
			    int domain)
  {
    //    cout << "calc flux, new" << endl;
    const FESpace & fes = u.GetFESpace();
    const FESpace & fesflux = flux.GetFESpace();
    const FESpace & fesdisp = disp.GetFESpace();


    int ne      = ma.GetNE();
    int dim     = fes.GetDimension();
    int dimflux = fesflux.GetDimension();
    int dimdisp = fesdisp.GetDimension();

    int i, j;

    ElementTransformation eltrans;

    ARRAY<int> dnums, dnumsflux, dnumsdisp, dnumsrot;

    LocalHeap lh (1000000);
    
    for (i = 0; i < ne; i++)
      {
	lh.CleanUp();

	int eldom = ma.GetElIndex(i);
	
	if ((domain != -1) && (domain != eldom))
	  continue;

	const FiniteElement & fel = fes.GetFE (i, lh);

	const NodalFiniteElement & felflux = 
	  dynamic_cast<const NodalFiniteElement&> (fesflux.GetFE (i, lh));
	const NodalFiniteElement & feldisp = 
	  dynamic_cast<const NodalFiniteElement&> (fesdisp.GetFE (i, lh));


	ma.GetElementTransformation (i, eltrans, lh);
	fes.GetDofNrs (i, dnums);
	fesflux.GetDofNrs (i, dnumsflux);
	fesdisp.GetDofNrs (i, dnumsdisp);

	FlatVector<SCAL> elu(dnums.Size() * dim, lh);

	FlatVector<SCAL> elflux(dnumsflux.Size() * dimflux, lh);
	FlatVector<SCAL> eldisp(dnumsdisp.Size() * dimdisp, lh);
	FlatVector<SCAL> elfluxi(dnumsflux.Size() * dimflux, lh);
	FlatVector<SCAL> eldispi(dnumsdisp.Size() * dimdisp, lh);
	FlatVector<SCAL> fluxi(dimflux, lh);
	FlatVector<SCAL> dispi(dimdisp, lh);
	FlatVector<SCAL> elsigma, uint;
	FlatMatrix<> elmat;

	u.GetElementVector (dnums, elu);
	fes.TransformVec (i, 0, elu, TRANSFORM_SOL);

	(*testout) << "elu = " << endl << elu << endl;

	bli.ComputeInternalVariables (fel, eltrans, elu, elsigma, uint, lh);

	(*testout) << "elsigma = " << endl << elsigma << endl;
	(*testout) << "uint = " << endl << uint << endl;


	const BilinearFormIntegrator & fluxbli = *fesflux.GetEvaluator();
	const BilinearFormIntegrator & dispbli = *fesdisp.GetEvaluator();


	const IntegrationRule & ir = 
	  GetIntegrationRules().SelectIntegrationRule(fel.ElementType(), 2*felflux.Order());

	elflux = 0;
	eldisp = 0;
	for (j = 0; j < ir.GetNIP(); j++)
	  {
	    void * heapp = lh.GetPointer();

	    bli.ComputePointValues (fel, eltrans, ir.GetIP(j), elsigma, uint, fluxi, dispi, lh);

	    // bli.CalcFlux (fel, eltrans, ir.GetIP(j), elu, fluxi, applyd, lh);
	    fluxbli.ApplyBTrans (felflux, eltrans, ir.GetIP(j), fluxi, elfluxi, lh);
	    dispbli.ApplyBTrans (felflux, eltrans, ir.GetIP(j), dispi, eldispi, lh);

	    elflux += elfluxi;
	    eldisp += eldispi;

	    lh.CleanUp (heapp);
	  }

	fluxbli.AssembleElementMatrix (felflux, eltrans, elmat, lh);
	CholeskyFactors<SCAL> invelmat(elmat);

	invelmat.Mult (elflux, elfluxi);
	fesflux.TransformVec (i, 0, elfluxi, TRANSFORM_SOL);
	flux.SetElementVector (dnumsflux, elfluxi);

	dispbli.AssembleElementMatrix (feldisp, eltrans, elmat, lh);
	(*testout) << "elmat disp = " << endl << elmat << endl;
	CholeskyFactors<SCAL> invelmatdisp(elmat);

	(*testout) << "eldisprhs = " << endl << eldisp << endl;
	invelmatdisp.Mult (eldisp, eldispi);
	(*testout) << "eldisp = " << endl << eldispi << endl;

	fesdisp.TransformVec (i, 0, eldispi, TRANSFORM_SOL);
	disp.SetElementVector (dnumsdisp, eldispi);
      }
    

    /*


	const IntegrationRule & ir = felflux.NodalIntegrationRule();
	for (j = 0; j < ir.GetNIP(); j++)
	  {
	    //    bli.CalcFlux (fel, eltrans, ir.GetIP(j), elu, fluxi, applyd, lh);
	    bli.ComputePointValues (fel, eltrans, ir.GetIP(j), elsigma, uint, fluxi, dispi, lh);
	    (*testout) << "sigmai = " << fluxi << ", ui = " << dispi << endl;
	    elflux.Range (j*dimflux, (j+1)*dimflux-1) = fluxi;
	    eldisp.Range (j*dimdisp, ( j+1)*dimdisp-1) = dispi;
	  }

	flux.SetElementVector (dnumsflux, elflux);
	disp.SetElementVector (dnumsdisp, eldisp);
    */
  }


  template void CalcEquilibriumFlux<double> (const MeshAccess & ma, 
					     const S_GridFunction<double> & u,
					     S_GridFunction<double> & flux,
					     S_GridFunction<double> & disp,
					     const EquilibriumIntegrator & bli,
					     int domain);









  ///
  class NumProcCalcEquilibriumFlux : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    ///
    GridFunction * gfuface;
    ///
    GridFunction * gfusurf;
    ///
    GridFunction * gfflux;
    ///
    GridFunction * gfdisp;
  public:
    ///
    NumProcCalcEquilibriumFlux (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcCalcEquilibriumFlux();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcCalcEquilibriumFlux (pde, flags);
    }

    ///
    virtual void Do();
    ///
    virtual void PrintReport (ostream & ost);
  };




  NumProcCalcEquilibriumFlux :: NumProcCalcEquilibriumFlux (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    gfuface = pde.GetGridFunction (flags.GetStringFlag ("solution", ""));
    gfusurf = pde.GetGridFunction (flags.GetStringFlag ("usurf", ""),1);
    gfflux = pde.GetGridFunction (flags.GetStringFlag ("flux", ""));
    gfdisp = pde.GetGridFunction (flags.GetStringFlag ("disp", ""));
  }

  NumProcCalcEquilibriumFlux :: ~NumProcCalcEquilibriumFlux()
  {
    ;
  }

  void NumProcCalcEquilibriumFlux :: Do()
  {
    cout << "Num-proc equilibrium calc flux" << endl;

    CalcEquilibriumFlux (pde.GetMeshAccess(),
			 dynamic_cast<const S_GridFunction<double>&> (*gfuface), 
			 dynamic_cast<S_GridFunction<double>&> (*gfflux), 
			 dynamic_cast<S_GridFunction<double>&> (*gfdisp), 
			 dynamic_cast<const EquilibriumIntegrator&> (*bfa->GetIntegrator(0)),
			 -1);

    if (gfusurf)
      {
	const FlatVector<Vec<3> > & vu = 
	  dynamic_cast<const FlatVector<Vec<3> > &> (gfuface->GetVector());
	FlatVector<Vec<3> > & vusurf = 
	  dynamic_cast<FlatVector<Vec<3> > &> (gfusurf->GetVector());
	vusurf = 0;

	ARRAY<int> facedofs(3), surfdofs(3);
	int nse = pde.GetMeshAccess().GetNSE();
	for (int i = 0; i < nse; i++)
	  {
	    gfuface->GetFESpace().GetSDofNrs (i, facedofs);
	    gfusurf->GetFESpace().GetSDofNrs (i, surfdofs);

	    for (int j = 0; j < 3; j++)
	      {
		vusurf(surfdofs[j]) = vu(facedofs[j]);
	      }
	  }
      }
  }

  void NumProcCalcEquilibriumFlux :: PrintReport (ostream & ost)
  {
    ost << "NumProcCalcFlux:" << endl;
    ost << "Bilinear-form = " << endl;
  }
















  
  // register FESpaces
  namespace {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hybrid", EquilibriumFESpace::Create);
      GetNumProcs().AddNumProc ("calcequilibriumflux", NumProcCalcEquilibriumFlux::Create);
   }
    
    Init init;
  }

}
