#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


namespace hybrid_cpp
{

  using namespace ngfem;
  using namespace ngsolve;


  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridFESpace : public CompoundFESpace
  {
  public:
    HybridFESpace (const MeshAccess & ama, 		   
                   const ARRAY<const FESpace*> & aspaces)
      : CompoundFESpace (ama, aspaces)
    { }

    virtual ~HybridFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      /*      cout << "*** creating hybrid fespace" << endl;

      cout << "  * general info:" << endl;
      cout << "    np = " << ma.GetNP() << endl;     
      cout << "    ne = " << ma.GetNE() << endl;     
      cout << "    ned = " << ma.GetNEdges() << endl;     
      cout << "    nfa = " << ma.GetNFaces() << endl;     
      cout << "    ndom = " << ma.GetNDomains() << endl;     
      cout << "    nbound = " << ma.GetNBoundaries() << endl;     */
     
      ARRAY<const FESpace*> spaces(3);

      Flags hdivflags(flags), l2flags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));
      int order_u = int (flags.GetNumFlag ("order_u", order));
      
      l2flags.SetFlag ("order", max(order_u,0));
      hdivflags.SetFlag ("order", order);
      facetflags.SetFlag("order", order);
      hdivflags.SetFlag ("discontinuous");

      spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
      spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
      spaces[2] = new FacetFESpace (ma, facetflags);

      HybridFESpace * fes = new HybridFESpace (ma, spaces);
      
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridFESpace"; }
  };



  ////////////////////////////////////////////////////////////////////////////////

  class HybridLaplaceIntegrator2d : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_eps;
  public:
    HybridLaplaceIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_eps  = coeffs[0];
    }

    virtual ~HybridLaplaceIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridLaplaceIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridlaplace");
      static int timer2 = NgProfiler::CreateTimer ("hybridlaplace boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<2> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
      const L2HighOrderFiniteElement & fel_l2 = 
        dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_facet = fel_facet.GetNDof();
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

        
      FlatVector<> mat_l2(nd_l2, lh);
      FlatMatrix<> mat_hdiv(nd_hdiv, 2, lh); 
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrix<> bmat_grad(2, nd_l2, lh);
      FlatMatrixFixHeight<2> bmat_stab(nd_l2+nd_hdiv, lh);
      FlatMatrix<> mat_stab(nd_l2+nd_hdiv, lh);

          
      const IntegrationRule & ir2d =
        GetIntegrationRules().SelectIntegrationRule (eltype, fel_hdiv.Order()+fel_l2.Order());
      
      
      void * heapp = lh.GetPointer();
      for (int l=0; l<ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<2,2> sip(ir2d[l], eltrans, lh);
        
          double det = sip.GetJacobiDet();
        
          // piola
          mat_hdiv =  (1.0/det) 
            * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());

          double factor = det * ir2d[l].Weight() / coef_eps->Evaluate(sip);

          bmat_grad = Trans (sip.GetJacobianInverse()) * 
            Trans (fel_l2.GetDShape(sip.IP(), lh));

          for (int i = 0; i < nd_hdiv; i++)
            for (int j = 0; j < 2; j++)
              bmat_stab(j,i) = mat_hdiv(i, j);

          for (int i = 0; i < nd_l2; i++)
            for (int j = 0; j < 2; j++)
              bmat_stab(j,i+nd_hdiv) = bmat_grad(j, i);

          mat_stab = Trans (bmat_stab) * bmat_stab;

          double alpha = 0.5;

          for (int i = 0; i < nd_hdiv; i++)
            for (int j = 0; j < nd_hdiv; j++)
              elmat(i,j) += (1-alpha) * factor * mat_stab(i,j);

          for (int i=0; i<nd_hdiv; i++)
            for (int j=0; j<nd_l2; j++)
              {
                elmat(i,nd_hdiv+j) += (1-alpha) * factor * mat_stab(i,nd_hdiv+j);
                elmat(nd_hdiv+j,i) += (1-alpha) * factor * mat_stab(nd_hdiv+j,i);
              }
          
          for (int i=0; i<nd_l2; i++)
            for (int j=0; j<nd_l2; j++)
              elmat(nd_hdiv+i,nd_hdiv+j) -= alpha * factor * mat_stab(nd_hdiv+i,nd_hdiv+j);

          lh.CleanUp(heapp);
        }

      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;

      FlatMatrix<> mat_q_l(nd_hdiv, nd_facet, lh);
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);

      for (int k = 0; k < ned; k++)
        {
          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
        
          tau_ref  = pe-ps;
        
          const IntegrationRule & ir1d =
            GetIntegrationRules().SelectIntegrationRule (ET_SEGM, fel_hdiv.Order()+fel_l2.Order());
      

          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
              IntegrationPoint ip( p2d(0), p2d(1) );
          
              SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<2> jac = sip.GetJacobian();
              Mat<2> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              tau = jac * tau_ref;
              double len = L2Norm(tau);
	    
              tau *= (1.0/len);
              normal(0) = tau(1);
              normal(1) = -tau(0);
          
              //Piola
              mat_hdiv = (1.0/sip.GetJacobiDet()) 
                * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());
          
              mat_hdiv_n = mat_hdiv * normal;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
          
              double factor = len * ir1d[l].Weight();
              double eps = coef_eps->Evaluate(sip);
          
              mat_q_l = (factor * mat_hdiv_n) * Trans(mat_facet);

              fel_l2.CalcShape(sip.IP(), mat_l2);
              mat_l2 *= factor;

              for (int i=0; i<nd_hdiv; i++)
                for (int j=0; j<nd_facet; j++)
                  {
                    elmat(i,nd_hdiv+nd_l2+j) += mat_q_l(i,j);  // <q.n, l>
                    elmat(j+nd_hdiv+nd_l2,i) += mat_q_l(i,j);  // <p.n, mu>
                  }

              for (int i=0; i<nd_hdiv; i++)
                for (int j=0; j<nd_l2; j++)
                  {
                    elmat(i,nd_hdiv+j) -= mat_hdiv_n(i) * mat_l2(j);  // <q.n, l>
                    elmat(j+nd_hdiv,i) -= mat_hdiv_n(i) * mat_l2(j);  // <q.n, l>
                  }

              lh.CleanUp(heapp);
            }
        }
    }
  };



  class HybridConvectionNewIntegrator2d : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_b[2], *coef_fac;
  public:
    HybridConvectionNewIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_b[0] = coeffs[0];
      coef_b[1] = coeffs[1];
      coef_fac  = coeffs[2];
    }

    virtual ~HybridConvectionNewIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridConvectionNewIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridconvection tot");
      static int timer2 = NgProfiler::CreateTimer ("hybridconvection boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<2> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
      const L2HighOrderFiniteElement & fel_l2 = 
        dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_facet = fel_facet.GetNDof();
      int nd_u = nd_l2 + nd_facet;

      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      // u, grad u  
      FlatVector<> mat_l2(nd_l2, lh);
      FlatVector<> mat_dl2db(nd_l2, lh);

      // lambda
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrixFixWidth<2> bmat(nd_u, lh), dbmat(nd_u, lh);
      Mat<2,2> dmat;
      FlatMatrix<> elmatu(nd_u, lh);
      FlatMatrix<> elmatl2(nd_l2, lh);

      
      // volume contribution: -(u, b.grad(v))
      const IntegrationRule & ir2d =
        GetIntegrationRules().SelectIntegrationRule (eltype, 2*fel_l2.Order());
      
      void * heapp = lh.GetPointer();

      elmatl2 = 0.0;
      for (int l=0; l<ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<2,2> sip(ir2d[l], eltrans, lh);
          double factor = fabs (sip.GetJacobiDet()) * ir2d[l].Weight();
        
          fel_l2.CalcShape(sip.IP(), mat_l2);

          Vec<2> b, transb;
          b(0) = coef_b[0]->Evaluate(sip);
          b(1) = coef_b[1]->Evaluate(sip);

          transb = factor * (sip.GetJacobianInverse () * b); 

          mat_dl2db = fel_l2.GetDShape(sip.IP(), lh) * transb;

          elmatl2 += mat_dl2db * Trans (mat_l2);

          lh.CleanUp(heapp);
        }

      for (int i = 0; i < nd_l2; i++)
        for (int j = 0; j < nd_l2; j++)
          elmat(nd_hdiv+i, nd_hdiv+j) += 0.5 * (elmatl2(i,j) - elmatl2(j,i));

      

      
      // The facet contribution:
      NgProfiler::RegionTimer reg2 (timer2);

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;

      elmatu = 0.0;

      for (int k = 0; k < ned; k++)
        {
          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
        
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
        
          tau_ref  = pe-ps;
        
          const IntegrationRule & ir1d =
            GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*max(fel_l2.Order(),fel_facet.Order()));
      

          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
              IntegrationPoint ip( p2d(0), p2d(1) );
          
              SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<2> jac = sip.GetJacobian();
              double det = sip.GetJacobiDet();

              tau = jac * tau_ref;
              double len = L2Norm(tau);
	    
              tau *= (1.0/len);
              normal(0) = tau(1);
              normal(1) = -tau(0);
          
              double bn = normal(0) * coef_b[0]->Evaluate(sip) 
                + normal(1) * coef_b[1]->Evaluate(sip);
          
              // standard: 
              fel_l2.CalcShape(sip.IP(), mat_l2);
              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);

              bmat = 0.0;
              for (int i = 0; i < nd_l2; i++)
                bmat(i, 0) = mat_l2(i);

              for (int i = 0; i < nd_facet; i++)
                bmat(nd_l2+i, 1) = mat_facet(i);

              double stab = 0.5 * fabs(bn); 
              if (stab * len < 1) stab = 0; 

              dmat(0,0) = stab;
              dmat(0,1) = -stab - 0.5 * bn;
              dmat(1,0) = -stab + 0.5 * bn;
              dmat(1,1) = stab;

              double fac = ir1d[l].Weight() * len;
              dmat *= fac;

              dbmat = bmat * Trans (dmat);
              elmatu -= dbmat * Trans (bmat);
          
              lh.CleanUp(heapp);
            }
        }

      for (int i = 0; i < nd_u; i++)
        for (int j = 0; j < nd_u; j++)
          elmat(i+nd_hdiv, j+nd_hdiv) += elmatu(i,j);
    }
  };









  class HybridRobinIntegrator2d : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      //       cout << "** robin: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      
      FlatMatrix<> mat(ndf, ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      mat=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          mat +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*(shape*Trans(shape));
          //         mat +=  (ir1d[l].Weight()*coef->Evaluate(sip))*(shape*Trans(shape));
        }
      
      // copy to global matrix
      for (int i=0; i<ndf; i++)
        for (int j=0; j<ndf; j++)
          elmat(ndr+i,ndr+j) = mat(i,j);
      
    }


  };


  class HybridOutflowIntegrator2d : public BilinearFormIntegrator
  {
    CoefficientFunction *coef_b[2];
  public:
    HybridOutflowIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_b[0] = coeffs[0];
      coef_b[1] = coeffs[1];
    }

    virtual ~HybridOutflowIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridOutflowIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      
      FlatMatrix<> mat(ndf, ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      mat=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
        
          Mat<2,1> jac = sip.GetJacobian();
          Vec<2> tau, normal;
        
          tau(0) = jac(0,0); tau(1) = jac(0,1);
          double len = L2Norm(tau);
          tau *= (1.0/len);
        
          normal(0) = tau(1);
          normal(1) = -tau(0);
        
          double bn = coef_b[0]->Evaluate(sip)*normal(0) + coef_b[1]->Evaluate(sip)*normal(1);
         
        
          ffel.CalcShape(ir1d[l], shape);
          mat -=  (bn*len*ir1d[l].Weight())*(shape*Trans(shape));
          //         mat +=  (ir1d[l].Weight()*coef->Evaluate(sip))*(shape*Trans(shape));
        }
      
      // copy to global matrix
      for (int i=0; i<ndf; i++)
        for (int j=0; j<ndf; j++)
          elmat(ndr+i,ndr+j) = mat(i,j);
      
    }


  };


  class HybridNeumannIntegrator2d : public LinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridNeumannIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : LinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridNeumannIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridNeumannIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementVector (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatVector<double> & elvec,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      //       cout << "** neumann: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      int nd = cfel.GetNDof();  
      
      elvec.AssignMemory(nd, lh);
      elvec = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      

      FlatVector<> vec(ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      vec=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          vec +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*shape;
          //         vec +=  ir1d[l].Weight()*coef->Evaluate(sip)*shape;
        }
      
      // copy to global matrix
      elvec.Range(ndr,nd) = vec;
    }


  };

  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybrid", HybridFESpace::Create);
  
    // hybridlaplace eps
    GetIntegrators().AddBFIntegrator ("hybridlaplace", 2, 1,
                                      HybridLaplaceIntegrator2d::Create);
  
    // hybridconvectionnew b0 b1 fac: die neue version
    GetIntegrators().AddBFIntegrator ("hybridconvectionnew", 2, 3,
                                      HybridConvectionNewIntegrator2d::Create);
  
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator2d::Create);
    
    GetIntegrators().AddBFIntegrator ("hybridoutflow", 2, 2,
                                      HybridOutflowIntegrator2d::Create);
    
    GetIntegrators().AddLFIntegrator ("hybridneumann", 2, 1,
                                      HybridNeumannIntegrator2d::Create);
  
  }

  Init init;

} // namespace

