#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


namespace hybrid_cpp
{

  using namespace ngfem;
  using namespace ngsolve;


  static const MeshAccess * myma;


  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */

  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      myma = &ama;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);
      
      Flags hdivflags(flags), l2flags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));
      int order_u = int (flags.GetNumFlag ("order_u", order+1));
      
      l2flags.SetFlag ("order", max(order_u,0));
      hdivflags.SetFlag ("order", order);
      hdivflags.SetFlag ("orderinner", 1.0);
      facetflags.SetFlag("order", order);
      // hdivflags.SetFlag ("discontinuous");

      spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
      spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
      spaces[2] = new FacetFESpace (ma, facetflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



  Table<int> * CreateSmoothingBlocks (int type) const
  {
    int nd = GetNDof();
    int ned = ma.GetNEdges();
    int nel = ma.GetNE();
    
    int order_u = (*this)[1]->GetOrder();
    int ndel = ((order_u+1)*(order_u+2))/2;

    ARRAY<int> cnt(ned/* +nel */);
    for (int i = 0; i < ned; i++)
      cnt[i] = 2*order+2;
    /*
    for (int i = ned; i < nel+ned; i++)
      cnt[i] = ndel;
    */
    Table<int> * it = new Table<int>(cnt);
    // Table<int> * it = new Table<int>(ned,2*order+2);
    int base_facet = GetStorageStart(2);
    int base_el = GetStorageStart(1);
    for (int i = 0; i < ned; i++)
      {
        (*it)[i][0] = i;
        for (int j = 0; j < order; j++)
          (*it)[i][1+j] = ned + (i * order) + j;
        (*it)[i][order+1] = base_facet + i;
        for (int j = 0; j < order; j++)
          (*it)[i][order+2+j] = base_facet + ned + i * order + j;
      }
    /*
    for (int i = 0; i < nel; i++)
      {
        (*it)[ned+i] = base_el+i;
        for (int j = 0; j < ndel-1; j++)
          (*it)[ned+i][j+1] = base_el+nel+i*(ndel-1)+j;
      }
    */
    *testout << "table = " << endl << *it << endl;
    return it;
  }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<D> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
      const L2HighOrderFiniteElement & fel_l2 = 
        dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivb = nd_hdiv; // 3 * (fel_hdiv.Order());
      int nd_facet = fel_facet.GetNDof();
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
        
      FlatVector<> mat_l2(nd_l2, lh);
      FlatMatrix<> mat_hdiv(nd_hdiv, 2, lh); 
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrixFixHeight<D> bmat_grad(nd_l2, lh);
      FlatMatrixFixHeight<D> bmat_stab(nd_l2+nd_hdiv, lh);
      FlatMatrix<> mat_stab(nd_l2+nd_hdiv, lh);
      FlatMatrix<> mat_gradgrad(nd_l2, lh);
      FlatMatrix<> hmat_l2(nd_l2, lh);
      FlatMatrixFixWidth<3> bmat(nd_l2, lh);

      FlatMatrix<Complex> elmat2(nd, lh);
      elmat2 = 0.0;


      double kappa = coef_kappa->EvaluateConst();
          
      const IntegrationRule & ir2d =
        GetIntegrationRules().SelectIntegrationRule (eltype, 2*fel_l2.Order());
      
      
      void * heapp = lh.GetPointer();
      hmat_l2 = 0.0;
      for (int l=0; l<ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<2,2> sip(ir2d[l], eltrans, lh);
          double factor = sip.GetJacobiDet() * ir2d[l].Weight();

          bmat_grad = Trans (sip.GetJacobianInverse()) * 
            Trans (fel_l2.GetDShape(sip.IP(), lh));

          fel_l2.CalcShape(sip.IP(), mat_l2);

          hmat_l2 -= factor * Trans (bmat_grad) * bmat_grad;
          hmat_l2 += (factor*kappa*kappa) * mat_l2 * Trans(mat_l2);

          lh.CleanUp(heapp);
        }

      for (int i=0; i<nd_l2; i++)
        for (int j=0; j<nd_l2; j++)
          elmat(nd_hdiv+i,nd_hdiv+j) += hmat_l2(i,j);



      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;

      FlatMatrix<> mat_q_l(nd_hdiv, nd_facet, lh);
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);

      FlatMatrix<> mat_bound(nd_hdivb, nd_l2+nd_facet, lh);
      mat_bound = 0.0;
      FlatMatrix<> mat_sigman(nd_hdivb, lh);
      mat_sigman = 0.0;

      ARRAY<int> vnums;
      myma->GetElVertices (eltrans.GetElementNr(), vnums);

      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      vu = 0.0;
      vv = 0.0;


      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);
      ELEMENT_TYPE etfacet = (D == 2) ? ET_SEGM : ET_TRIG;

      for (int k = 0; k < ned; k++)
        {
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];
        
          const IntegrationRule & ir1d =
            GetIntegrationRules().SelectIntegrationRule (etfacet, 2 * max (fel_hdiv.Order(), fel_l2.Order()));

          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
          tau_ref  = pe-ps;
        
          // const IntegrationRule & ir1d =
          // GetIntegrationRules().SelectIntegrationRule (ET_SEGM, fel_hdiv.Order()+fel_l2.Order());
      
          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);

              // Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
              // IntegrationPoint ip( p2d(0), p2d(1) );
              // SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<2> jac = sip.GetJacobian();
              Mat<2> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              tau = jac * tau_ref;
              double len = L2Norm(tau);
	    
              tau *= (1.0/len);
              normal(0) = tau(1);
              normal(1) = -tau(0);
          
              //Piola
              mat_hdiv = (1.0/sip.GetJacobiDet()) 
                * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());
          
              mat_hdiv_n = mat_hdiv * normal;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
          
              double factor = len * ir1d[l].Weight();
          
              mat_q_l = (factor * mat_hdiv_n) * Trans(mat_facet);

              fel_l2.CalcShape(sip.IP(), mat_l2);

              bmat_grad = Trans (sip.GetJacobianInverse()) * 
                Trans (fel_l2.GetDShape(sip.IP(), lh));

              for (int i = 0; i < nd_l2; i++)
                {
                  for (int j = 0; j < nd_hdiv; j++)
                    elmat(nd_hdiv+i, j) -= factor * mat_l2(i) * mat_hdiv_n(j);
                  for (int j = 0; j < nd_l2; j++)
                    elmat(nd_hdiv+i, nd_hdiv+j) += Complex(0, -factor*kappa) * mat_l2(i) * mat_l2(j);
                  for (int j = 0; j < nd_facet; j++)
                    elmat(nd_hdiv+i, nd-nd_facet+j) += Complex(0, factor*kappa) * mat_l2(i) * mat_facet(j);
                }
              

              vu.Range(0, nd_hdiv) = mat_hdiv_n;
              for (int i = 0; i < nd_l2; i++)
                vu(nd_hdiv+i) = 
                  normal(0)*bmat_grad(0,i)+normal(1)*bmat_grad(1,i)-Complex(0,kappa)*mat_l2(i);
              vu.Range(nd_hdiv+nd_l2, nd) = Complex(0,kappa) * mat_facet;

              vv.Range(0, nd_hdiv) = -mat_hdiv_n;
              vv.Range(nd_hdiv+nd_l2, nd) = Complex(0,kappa) * mat_facet;

              for (int i = 0; i < nd; i++)
                for (int j = 0;j < nd; j++)
                  elmat(i,j) += (factor/kappa) * vu(j)*vv(i);

              lh.CleanUp(heapp);
            }
        }

      /*
      for (int i = 0; i < nd_hdiv; i++)
        elmat.Row(i) = elmat2.Row(i);
      for (int i = nd_hdiv+nd_l2; i < nd; i++)
        elmat.Row(i) = elmat2.Row(i);
      */











      /*
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const HDivFiniteElement<2> & fel_hdiv = 
        dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
      const L2HighOrderFiniteElement & fel_l2 = 
        dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdiv = fel_hdiv.GetNDof();
      int nd_hdivb = nd_hdiv; // 3 * (fel_hdiv.Order());
      int nd_facet = fel_facet.GetNDof();
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

        
      FlatVector<> mat_l2(nd_l2, lh);
      FlatMatrix<> mat_hdiv(nd_hdiv, 2, lh); 
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrixFixHeight<2> bmat_grad(nd_l2, lh);
      FlatMatrixFixHeight<2> bmat_stab(nd_l2+nd_hdiv, lh);
      FlatMatrix<> mat_stab(nd_l2+nd_hdiv, lh);
      FlatMatrix<> mat_gradgrad(nd_l2, lh);
      FlatMatrix<> hmat_l2(nd_l2, lh);
      FlatMatrixFixWidth<3> bmat(nd_l2, lh);

      FlatMatrix<Complex> elmat2(nd, lh);
      elmat2 = 0.0;


      double kappa = coef_kappa->EvaluateConst();
          
      const IntegrationRule & ir2d =
        GetIntegrationRules().SelectIntegrationRule (eltype, 2*fel_l2.Order());
      
      
      void * heapp = lh.GetPointer();
      hmat_l2 = 0.0;
      for (int l=0; l<ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<2,2> sip(ir2d[l], eltrans, lh);
          double factor = sip.GetJacobiDet() * ir2d[l].Weight();

          bmat_grad = Trans (sip.GetJacobianInverse()) * 
            Trans (fel_l2.GetDShape(sip.IP(), lh));

          fel_l2.CalcShape(sip.IP(), mat_l2);

          hmat_l2 -= factor * Trans (bmat_grad) * bmat_grad;
          hmat_l2 += (factor*kappa*kappa) * mat_l2 * Trans(mat_l2);

          lh.CleanUp(heapp);
        }

      for (int i=0; i<nd_l2; i++)
        for (int j=0; j<nd_l2; j++)
          elmat(nd_hdiv+i,nd_hdiv+j) += hmat_l2(i,j);



      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      const POINT3D * points = ElementTopology::GetVertices (eltype);
      const EDGE * edges = ElementTopology::GetEdges (eltype);
      int ned = (eltype == ET_TRIG) ? 3: 4;

      FlatMatrix<> mat_q_l(nd_hdiv, nd_facet, lh);
      FlatVector<> mat_hdiv_n(nd_hdiv, lh);

      FlatMatrix<> mat_bound(nd_hdivb, nd_l2+nd_facet, lh);
      mat_bound = 0.0;
      FlatMatrix<> mat_sigman(nd_hdivb, lh);
      mat_sigman = 0.0;

      ARRAY<int> vnums;
      myma->GetElVertices (eltrans.GetElementNr(), vnums);

      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      vu = 0.0;
      vv = 0.0;

      for (int k = 0; k < ned; k++)
        {
          Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
          ps(0) = points[edges[k][0]][0];
          ps(1) = points[edges[k][0]][1];
          pe(0) = points[edges[k][1]][0];
          pe(1) = points[edges[k][1]][1];
        
          tau_ref  = pe-ps;
        
          const IntegrationRule & ir1d =
            GetIntegrationRules().SelectIntegrationRule (ET_SEGM, fel_hdiv.Order()+fel_l2.Order());
      
          int sign = (vnums[edges[k][1]] > vnums[edges[k][0]]) ? 1 : -1;

          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              Vec<2> p2d = ps + (1-ir1d[l](0)) * (pe-ps);
              IntegrationPoint ip( p2d(0), p2d(1) );
          
              SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
              Mat<2> jac = sip.GetJacobian();
              Mat<2> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();

              tau = jac * tau_ref;
              double len = L2Norm(tau);
	    
              tau *= (1.0/len);
              normal(0) = tau(1);
              normal(1) = -tau(0);
          
              //Piola
              mat_hdiv = (1.0/sip.GetJacobiDet()) 
                * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());
          
              mat_hdiv_n = mat_hdiv * normal;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
          
              double factor = len * ir1d[l].Weight();
          
              mat_q_l = (factor * mat_hdiv_n) * Trans(mat_facet);

              fel_l2.CalcShape(sip.IP(), mat_l2);

              bmat_grad = Trans (sip.GetJacobianInverse()) * 
                Trans (fel_l2.GetDShape(sip.IP(), lh));

              for (int i=0; i<nd_hdivb; i++)
                for (int j=0; j<nd_facet; j++)
                  mat_bound(i,nd_l2+j) += mat_q_l(i,j);  // <q.n, l>

              for (int i=0; i<nd_hdivb; i++)
                for (int j=0; j<nd_l2; j++)
                  mat_bound(i,j) -= factor * mat_hdiv_n(i) * mat_l2(j);  // <q.n, l>

              for (int i = 0; i < nd_hdivb; i++)
                for (int j = 0; j < nd_hdivb; j++)
                  mat_sigman(i,j) += factor * mat_hdiv_n(i) * mat_hdiv_n(j);


              // ( (du/dn-iku + (ik gamma-sigma_n) , ik mu - tau_n)

              vu.Range(0, nd_hdiv) = mat_hdiv_n;
              for (int i = 0; i < nd_l2; i++)
                vu(nd_hdiv+i) = 
                  normal(0)*bmat_grad(0,i)+normal(1)*bmat_grad(1,i)-Complex(0,kappa)*mat_l2(i);
              vu.Range(nd_hdiv+nd_l2, nd) = Complex(0,kappa) * mat_facet;

              vv.Range(0, nd_hdiv) = -mat_hdiv_n;
              vv.Range(nd_hdiv+nd_l2, nd) = Complex(0,kappa) * mat_facet;

              for (int i = 0; i < nd; i++)
                for (int j = 0;j < nd; j++)
                  elmat2(i,j) += (factor/kappa) * vu(j)*vv(i);


              lh.CleanUp(heapp);
            }
        }

      for (int i = 0; i < nd_hdivb; i++)
        for (int j=0; j < nd_l2+nd_facet; j++)
          {
            elmat(i,nd_hdiv+j) += mat_bound(i,j);  // <q.n, l>
            elmat(j+nd_hdiv,i) += mat_bound(i,j);  // <p.n, mu>
          }

      FlatMatrix<> inv_sigman(nd_hdivb, lh);
      FlatMatrix<> hmat(nd_hdivb, nd_l2+nd_facet, lh);
      FlatMatrix<> stab(nd_l2+nd_facet, lh);

      CalcInverse (mat_sigman, inv_sigman);
      hmat = inv_sigman * mat_bound;
      stab = Trans (mat_bound) * hmat;

      for (int i = 0; i < nd_l2+nd_facet; i++)
        for (int j = 0; j < nd_l2+nd_facet; j++)
          elmat(nd_hdiv+i, nd_hdiv+j) -= Complex (0, kappa) * stab(i,j);
      
      for (int i = 0; i < nd_hdiv; i++)
        elmat.Row(i) = elmat2.Row(i);
      for (int i = nd_hdiv+nd_l2; i < nd; i++)
        elmat.Row(i) = elmat2.Row(i);
      */
    }
  };




  class HybridRobinIntegrator2d : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<> & elmat,
                                        LocalHeap & lh) const
    {
      cerr << "Helmholtz - Robin real called" << endl;
    }




    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const HDivNormalFiniteElement<1> & hdivfel = 
        dynamic_cast<const HDivNormalFiniteElement<1>&> (cfel[0]);
      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      //       cout << "** robin: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int ndhdiv = hdivfel.GetNDof();
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      
      FlatMatrix<> mat(ndf, ndf, lh);
      FlatVector<> shape(ndf, lh), shapehdiv(ndhdiv, lh);
      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      mat=0;
      vu = 0.0;
      vv = 0.0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          double val = coef->Evaluate(sip);

          if (val == 0) continue;


          hdivfel.CalcShape(ir1d[l], shapehdiv);
          shapehdiv *= 1.0 / sip.GetJacobiDet();

          ffel.CalcShape(ir1d[l], shape);

          vu.Range (0, ndhdiv) = shapehdiv;
          vu.Range (nd-ndf, nd) = Complex (0, val) * shape;

          
          elmat +=  (sip.GetJacobiDet()*ir1d[l].Weight())*(vu*Trans(vu));
        }
      

      /*
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      //       cout << "** robin: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      
      FlatMatrix<> mat(ndf, ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      mat=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          mat +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*(shape*Trans(shape));
          //         mat +=  (ir1d[l].Weight()*coef->Evaluate(sip))*(shape*Trans(shape));
        }
      
      // copy to global matrix
      for (int i=0; i<ndf; i++)
        for (int j=0; j<ndf; j++)
          elmat(ndr+i,ndr+j) = mat(i,j);
      */
      
    }


  };



  class HybridNeumannIntegrator2d : public LinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridNeumannIntegrator2d (ARRAY<CoefficientFunction*> & coeffs) 
      : LinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridNeumannIntegrator2d () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridNeumannIntegrator2d (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementVector (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatVector<double> & elvec,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      //       cout << "** neumann: ffel-vertices=" << ffel.GetVertexNumber(0) << "," << ffel.GetVertexNumber(1)  << endl;
      
      int nd = cfel.GetNDof();  
      
      elvec.AssignMemory(nd, lh);
      elvec = 0.0;
      
      
      int ndf = ffel.GetNDof();
      int ndr = nd-ndf;
      

      FlatVector<> vec(ndf, lh);
      FlatVector<> shape(ndf, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*ffel.Order());
      
      vec=0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<1,2> sip(ir1d[l], eltrans, lh);
          ffel.CalcShape(ir1d[l], shape);
          vec +=  (sip.GetJacobiDet()*ir1d[l].Weight()*coef->Evaluate(sip))*shape;
          //         vec +=  ir1d[l].Weight()*coef->Evaluate(sip)*shape;
        }
      
      // copy to global matrix
      elvec.Range(ndr,nd) = vec;
    }


  };

  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz", HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 2, 1,
                                      HybridHelmholtzIntegrator2d::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator2d::Create);

    GetIntegrators().AddLFIntegrator ("hybridneumann", 2, 1,
                                      HybridNeumannIntegrator2d::Create);
  
  }

  Init init;

} // namespace

