#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


namespace hybrid_helmholtz_cpp
{


  using namespace ngfem;
  using namespace ngsolve;


  /* *********************************************************** */
  /* ************************* FESpace ************************* */
  /* *********************************************************** */


  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    { 
      ;
    }

    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);
      
      Flags l2flags(flags);
      Flags hdivnflags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));

      l2flags.SetFlag ("order", order+1);

      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderfacet", order);
      hdivnflags.SetFlag ("orderinner", 1.0);
      hdivnflags.SetFlag ("curlorder", 1.0);

      facetflags.SetFlag("order", order);

      spaces[0] = new L2HighOrderFESpace (ma, l2flags);    
      spaces[1] = new HDivHighOrderFESpace (ma, hdivnflags);
      spaces[2] = new FacetFESpace (ma, facetflags);

      HybridHelmholtzFESpace * fes = new HybridHelmholtzFESpace (ma, spaces, flags);
      
      return fes;
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



#define VER2

    Table<int> * CreateSmoothingBlocks (int type) const
    {
      int nd = GetNDof();
      int ned = ma.GetNEdges();

      ARRAY<int> ednums;
      BitArray bndedges(ned);
      bndedges.Clear();
      /*
      for (int i = 0; i < ma.GetNSE(); i++)
        {
          if (ma.GetSElIndex(i) >= 4) continue;
          ma.GetSElEdges (i, ednums);
          bndedges.Set (ednums[0]);
        }
      */

#ifdef VER1
      // edge by edge
      ARRAY<int> cnt(ned);      
      cnt = 2*order+2;
      for (int i = 0; i < ned; i++)
        if (bndedges.Test(i))
          cnt[i] = 0;

      Table<int> * it = new Table<int>(cnt);

      int base_hdivn = GetStorageStart(2);
      int base_facet = GetStorageStart(3);
      int base_el = GetStorageStart(1);
      // for (int i = 0; i < 1; i++)
      for (int i = 0; i < ned; i++)
        {
          if (bndedges.Test(i)) continue;
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }
      return it;
#endif

#ifdef VER1a
      // edge by edge and direction by directon
      ARRAY<int> cnt(2*ned);
      cnt = order+1;
      for (int i = 0; i < ned; i++)
        if (bndedges.Test(i))
          cnt[2*i] = cnt[2*i+1] = 0;

      *testout << "cnt = " << endl << cnt << endl;

      Table<int> * it = new Table<int>(cnt);

      int base_hdivn = GetStorageStart(2);
      int base_facet = GetStorageStart(3);
      int base_el = GetStorageStart(1);
      for (int i = 0; i < ned; i++)
        {
          if (bndedges.Test(i)) continue;

          (*it)[2*i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[2*i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[2*i+1][0] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[2*i+1][1+j] = base_facet + ned + i * order + j;
        }

      *testout << "table = " << endl << *it << endl;
      return it;
#endif


#ifdef VER2
      // all external dofs of an element
      int ne = ma.GetNE();
      ARRAY<int> dnums;

      ARRAY<int> cnt(ne);
    
      for (int i = 0; i < ne; i++)
        {
          GetExternalDofNrs (i, dnums);
          cnt[i] = dnums.Size();
        }

      Table<int> & it = *new Table<int>(cnt);    
      for (int i = 0; i < ne; i++)
        {
          GetExternalDofNrs (i, dnums);
          for (int j = 0; j < dnums.Size(); j++)
            it[i][j] = dnums[j];
        }

      *testout << "table = " << endl << it << endl;
      return &it;
#endif


    }



  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 0; }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);


      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const L2HighOrderFiniteElement & fel_l2 = 
        dynamic_cast<const L2HighOrderFiniteElement&> (cfel[0]);
      const HDivHighOrderFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivHighOrderFiniteElement<D>&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
  
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();

      int base_l2 = 0;
      int base_hdivn = base_l2+nd_l2;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

        
      FlatVector<> mat_l2(nd_l2, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<> mat_dudn(nd_l2, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrix<> hmat_l2(nd_l2, lh);

      FlatMatrixFixHeight<4, Complex> bmat(nd, lh);
      FlatMatrixFixHeight<4, Complex> dbmat(nd, lh);
      Mat<4> dmat;

      FlatMatrix<> mat_l2l2 (nd_l2, lh);
      FlatMatrix<> mat_gradgrad (nd_l2, lh);

      FlatMatrix<> dshaper(nd_l2, D, lh);
      FlatMatrixFixWidth<D> dshape(nd_l2, lh);

      double kappa = coef_kappa->EvaluateConst();
          
      const IntegrationRule & ir2d =
        SelectIntegrationRule (eltype, 2*fel_l2.Order()+2);
      
      void * heapp = lh.GetPointer();
      hmat_l2 = 0.0;
      mat_l2l2  = 0.0;
      mat_gradgrad = 0.0;

      for (int l = 0; l < ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<D,D> sip(ir2d[l], eltrans, lh);
          
          double factor = sip.GetJacobiDet() * ir2d[l].Weight();

          fel_l2.CalcShape(sip.IP(), mat_l2);
          dshape = fel_l2.GetDShape (sip.IP(), lh) * sip.GetJacobianInverse();

          mat_gradgrad += factor * dshape * Trans (dshape);
          mat_l2l2 += factor *  (mat_l2 * Trans (mat_l2));

          lh.CleanUp(heapp);
        }

      double alpha = 0;

      for (int i = 0; i < nd_l2; i++)
        for (int j = 0; j < nd_l2; j++)
          elmat(base_l2+j, base_l2+i) -= Complex(0, -kappa) * mat_l2l2(i,j) + Complex (0, 1/kappa) * mat_gradgrad(i,j);


      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     


      int ned = ElementTopology::GetNFacets(eltype);
      
      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      for (int k = 0; k < ned; k++)
        {
          ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 2*max (0,fel_l2.Order())+2);
      
          void * heapp = lh.GetPointer();

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();


              Vec<D> normal = det * Trans (inv_jac) * normal_ref;       
              double len = L2Norm (normal);
              normal /= L2Norm (normal);

              Vec<D> nt = Trans(sip.GetJacobian()) * normal;
              nt /= det;

              //Piola

              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              fel_l2.CalcShape(sip.IP(), mat_l2);
              mat_dudn = fel_l2.GetDShape (sip.IP(), lh) * sip.GetJacobianInverse() * normal;
              
              bmat = 0.0;
              bmat.Row(0).Range (base_l2   , base_l2   +nd_l2 )   = Complex (0,1/kappa) * mat_dudn;
              bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;


              dmat = 0.0;

              dmat(2,1) = -1;
              dmat(1,2) = -1;
              dmat(2,3) = 1;
              dmat(3,2) = 1;


              dmat(1,1) = -1;
              dmat(1,3) = 1;
              dmat(3,1) = 1;
              dmat(3,3) = -1;

              dmat(2,2) += 1;
              dmat(2,0) += 1;
              dmat(0,2) += 1;
              dmat(0,0) += 1;

              dmat *= len * ir1d[l].Weight();
              dbmat = dmat * bmat;
              elmat += Trans (bmat) * dbmat;

              lh.CleanUp(heapp);
            }
        }
    }
  };








  template <int D>
  class HybridRobinIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<> & elmat,
                                        LocalHeap & lh) const
    {
      cerr << "Helmholtz - Robin real called" << endl;
    }




    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const HDivHighOrderNormalFiniteElement<D-1> & hdivnfel = 
        dynamic_cast<const HDivHighOrderNormalFiniteElement<D-1>&> (cfel[1]);
      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int nd_hdivn = hdivnfel.GetNDof();
      int nd_facet = ffel.GetNDof();
      int base_facet = nd-nd_facet;
      int base_hdivn = base_facet-nd_hdivn;
      
      FlatVector<> shape(nd_facet, lh), shapehdivn(nd_hdivn, lh);
      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ffel.ElementType(), 2*ffel.Order()+2);
      
      FlatMatrixFixHeight<2> bmat(nd, lh);
      FlatMatrixFixHeight<2> dbmat(nd, lh);
      Mat<2> dmat(2);

      vu = 0.0;
      vv = 0.0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<D-1,D> sip(ir1d[l], eltrans, lh);
          double val = coef->Evaluate(sip);

          if (val == 0) continue;

          hdivnfel.CalcShape(ir1d[l], shapehdivn);
          shapehdivn *= 1.0 / sip.GetJacobiDet();

          ffel.CalcShape(ir1d[l], shape);

          double factor = sip.GetJacobiDet()*ir1d[l].Weight();

          bmat = 0.0;
          bmat.Row(0).Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
          bmat.Row(1).Range (base_facet, nd) = shape;

          dmat = 0.0;

          dmat(1,1) =  -1;

          dbmat = dmat * bmat;
          elmat += factor * Trans (bmat) * dbmat;
        }
    }
  };


  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz", HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 2, 1,
                                      HybridHelmholtzIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 3, 1,
                                      HybridHelmholtzIntegrator<3>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("hybridrobin", 3, 1,
                                      HybridRobinIntegrator<3>::Create);
  }

  Init init;

} // namespace

