#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


extern ngsolve::PDE * pde;



namespace hybrid_helmholtz_cpp
{


  using namespace ngfem;
  using namespace ngsolve;

  static Complex I (0,1);




  class PlaneWaveElement : public FiniteElement
  {
    Mat<4,2> points;
    double kappa;
    int edge_order;

  public:
    enum { DOF_FACTOR = 3 };

    PlaneWaveElement (int order, double akappa)
      : FiniteElement (2, ET_TRIG, 3*order, order) 
    {
      edge_order = order;
      kappa = akappa; 
    }

    void SetPoints (const Mat<4,2> & apoints) 
    {
      points = apoints;
    }

    virtual void GetInternalDofs (ARRAY<int> & idofs) const
    {
      idofs.SetSize(0);
      for (int i = 0; i < ndof; i++) idofs.Append (i);
    }

    void CalcShapeDShape (const SpecificIntegrationPoint<2,2> & sip,
                          FlatVector<Complex> & shape,
                          FlatMatrix<Complex> & dshape) const
    {
      int jj = 0;

      const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
      for (int i = 0; i < 3; i++)
        {
          Vec<2> ps = points.Row(edges[i][0]);
          Vec<2> pe = points.Row(edges[i][1]);

          Vec<2> tau = pe-ps;
          double len = L2Norm (tau);
          tau /= len;
          Vec<2> normal (tau(1), -tau(0));
 
          int minorder = -edge_order/2;
          int maxorder = minorder+edge_order;

          for (int j = minorder; j < maxorder; j++)
            {
              // edge extension factor
              double fac = (edge_order < 4) ? 0.5 : 1.0 - 2.0/edge_order;

              double kt = fac * (2 * M_PI * j / len);
              double disc = kappa*kappa-kt*kt;
              Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));

              Vec<2, Complex> kvec = kt * tau - kn * normal;

              shape(jj) = exp ( I * InnerProduct (kvec, sip.GetPoint()-ps));
              dshape.Row(jj) = shape(jj) * I * kvec;
              jj++;
            }
        }
    }
  };








  // plane waves along one line
  class PlaneWaveElement2 : public FiniteElement
  {
    Mat<4,2> points;
    double kappa;
    int edge_order;

  public:
    enum { DOF_FACTOR = 2 };

    PlaneWaveElement2 (int order, double akappa)
      : FiniteElement (2, ET_TRIG, 2*order, order) 
    {
      edge_order = order;
      kappa = akappa; 
    }

    void SetPoints (const Mat<4,2> & apoints) 
    {
      points = apoints;
    }

    virtual void GetInternalDofs (ARRAY<int> & idofs) const
    {
      idofs.SetSize(0);
      for (int i = 0; i < ndof; i++) idofs.Append (i);
    }

    void CalcShapeDShape (const SpecificIntegrationPoint<2,2> & sip,
                          FlatVector<Complex> & shape,
                          FlatMatrix<Complex> & dshape) const
    {
      int jj = 0;

      const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
      for (int i = 0; i < 1; i++)
        {
          Vec<2> ps = points.Row(edges[i][0]);
          Vec<2> pe = points.Row(edges[i][1]);

          Vec<2> tau = pe-ps;
          double len = L2Norm (tau);
          tau /= len;
          Vec<2> normal (tau(1), -tau(0));
 
          int minorder = -edge_order/2;
          int maxorder = minorder+edge_order;

          for (int j = minorder; j < maxorder; j++)
            {
              double kt = 0.5 * (2 * M_PI * j / len);
              double disc = kappa*kappa-kt*kt;
              Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));

              Vec<2, Complex> kvec = kt * tau - kn * normal;
              shape(jj) = exp ( I * InnerProduct (kvec, sip.GetPoint()-ps));
              dshape.Row(jj) = shape(jj) * I * kvec;
              jj++;


              kvec = kt * tau + kn * normal;
              shape(jj) = exp ( I * InnerProduct (kvec, sip.GetPoint()-ps));
              dshape.Row(jj) = shape(jj) * I * kvec;
              jj++;
            }
        }
    }
  };







  // plane waves along two lines
  class PlaneWaveElement3 : public FiniteElement
  {
    Mat<4,2> points;
    double kappa;
    int edge_order;

  public:
    enum { DOF_FACTOR = 4 };

    PlaneWaveElement3 (int order, double akappa)
      : FiniteElement (2, ET_TRIG, 4*order, order) 
    {
      edge_order = order;
      kappa = akappa; 
    }

    void SetPoints (const Mat<4,2> & apoints) 
    {
      points = apoints;
    }

    virtual void GetInternalDofs (ARRAY<int> & idofs) const
    {
      idofs.SetSize(0);
      for (int i = 0; i < ndof; i++) idofs.Append (i);
    }

    void CalcShapeDShape (const SpecificIntegrationPoint<2,2> & sip,
                          FlatVector<Complex> & shape,
                          FlatMatrix<Complex> & dshape) const
    {
      int jj = 0;

      const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
      for (int i = 0; i < 2; i++)
        {
          Vec<2> ps = points.Row(edges[i][0]);
          Vec<2> pe = points.Row(edges[i][1]);

          Vec<2> tau = pe-ps;
          double len = L2Norm (tau);
          tau /= len;
          Vec<2> normal (tau(1), -tau(0));
 
          int minorder = -edge_order/2;
          int maxorder = minorder+edge_order;

          for (int j = minorder; j < maxorder; j++)
            {
              double kt = 0.5 * (2 * M_PI * j / len);
              double disc = kappa*kappa-kt*kt;
              Complex kn = (disc >= 0) ? sqrt(disc) : Complex (0, sqrt(-disc));

              Vec<2, Complex> kvec = kt * tau - kn * normal;
              shape(jj) = exp ( I * InnerProduct (kvec, sip.GetPoint()-ps));
              dshape.Row(jj) = shape(jj) * I * kvec;
              jj++;


              kvec = kt * tau + kn * normal;
              shape(jj) = exp ( I * InnerProduct (kvec, sip.GetPoint()-ps));
              dshape.Row(jj) = shape(jj) * I * kvec;
              jj++;
            }
        }
    }
  };








  typedef PlaneWaveElement WaveElement;








  class WaveFESpace : public FESpace
  {
    int order;
    int nel, ndof;
    double kappa;
    ARRAY<int> first_element_dof;
    
  public:
    WaveFESpace (const MeshAccess & ama, int aorder, const Flags & aflags)
      : FESpace (ama, aflags)
    {
      order = aorder;
      if ( pde -> ConstantUsed ("k") )
	kappa = int(pde->GetConstant("k") );
      else
        throw Exception ("constant k is needed");
    }

    virtual void Update()
    {
      nel = ma.GetNE();
      first_element_dof.SetSize(nel+1);

      ndof = 0;
      for (int i = 0; i < nel; i++)
        {
          first_element_dof[i] = ndof;
          ndof += WaveElement::DOF_FACTOR*order;
        }
      first_element_dof[nel] = ndof;
    }
    
    virtual int GetNDof () const
    {
      return ndof;
    }
    
    virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const
    {
      dnums.SetSize(0);
      for (int i = first_element_dof[elnr]; i < first_element_dof[elnr+1]; i++)
        dnums.Append(i);
    }

    virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
    {
      if (!eliminate_internal) 
        {
          GetDofNrs (elnr, dnums);
          return;
        }
      dnums.SetSize(0);
    }

    virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const
    {
      dnums.SetSize(0);
    }

    virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const
    {
      WaveElement * fe = new (lh.Alloc (sizeof(WaveElement)))  WaveElement (order, kappa);

      ArrayMem<int, 10> vnums;
      ma.GetElVertices (elnr, vnums);

      Mat<4,2> pmat;
      Vec<3> point;

      for (int i = 0; i < vnums.Size(); i++)
        {
          ma.GetPoint (vnums[i], point);
          pmat(i,0) = point(0);
          pmat(i,1) = point(1);
        }
      fe -> SetPoints (pmat);
      
      return *fe;
    }

    virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const
    {
      return *new (lh.Alloc (sizeof(FE_SegmDummy))) FE_SegmDummy;
    }
  };




  // product space for (u, p, l)
  class HybridHelmholtzFESpace : public CompoundFESpace
  {
    int neldof;
  public:
    HybridHelmholtzFESpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags)
      : CompoundFESpace (ama, aspaces, flags)
    {       
      neldof = 3 * int (flags.GetNumFlag ("orderl2", 1));
    }


    
    virtual ~HybridHelmholtzFESpace () { ; }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      ARRAY<const FESpace*> spaces(3);
      
      Flags l2flags(flags);
      Flags hdivnflags(flags), facetflags(flags);

      int order = int (flags.GetNumFlag ("order", 1));
      int orderl2 = int (flags.GetNumFlag ("orderl2", 1));
      l2flags.SetFlag ("order", orderl2);

      hdivnflags.SetFlag ("order", order);
      hdivnflags.SetFlag ("orderfacet", order);
      hdivnflags.SetFlag ("orderinner", 1.0);
      hdivnflags.SetFlag ("curlorder", 1.0);

      facetflags.SetFlag("order", order);

      /*
      double kappa = flags.GetNumFlag ("kappa", -1);
      if (kappa < 0) 
        {
          cerr << "kappa > 0 is required" << endl;
          exit(1);
        }
      */
      spaces[0] = new WaveFESpace (ma, orderl2, flags);    
      spaces[1] = new HDivHighOrderFESpace (ma, hdivnflags);
      spaces[2] = new FacetFESpace (ma, facetflags);

      return new HybridHelmholtzFESpace (ma, spaces, flags);
    }
    
    virtual string GetClassName () const { return "HybridHelmholtzFESpace"; }



#define VER3


    Table<int> * CreateSmoothingBlocks (int type) const
    {
      int nd = GetNDof();
      int ned = ma.GetNEdges();
      int nel = ma.GetNE();


      int base_hdivn = GetStorageStart(1);
      int base_facet = GetStorageStart(2);
      int base_el = GetStorageStart(0);



      ARRAY<int> ednums;
      BitArray bndedges(ned);
      bndedges.Clear();
      /*
      for (int i = 0; i < ma.GetNSE(); i++)
        {
          if (ma.GetSElIndex(i) >= 4) continue;
          ma.GetSElEdges (i, ednums);
          bndedges.Set (ednums[0]);
        }
      */

#ifdef VER1
      // edge by edge
      ARRAY<int> cnt(ned);      
      cnt = 2*order+2;
      /*
      for (int i = 0; i < ned; i++)
        if (bndedges.Test(i))
          cnt[i] = 0;
      */
      Table<int> * it = new Table<int>(cnt);

      // for (int i = 0; i < 1; i++)
      for (int i = 0; i < ned; i++)
        {
          // if (bndedges.Test(i)) continue;
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }
      return it;
#endif

#ifdef VER1a
      // edge by edge and direction by directon
      ARRAY<int> cnt(2*ned);
      cnt = order+1;
      for (int i = 0; i < ned; i++)
        if (bndedges.Test(i))
          cnt[2*i] = cnt[2*i+1] = 0;

      *testout << "cnt = " << endl << cnt << endl;

      Table<int> * it = new Table<int>(cnt);

      int base_hdivn = GetStorageStart(2);
      int base_facet = GetStorageStart(3);
      int base_el = GetStorageStart(1);
      for (int i = 0; i < ned; i++)
        {
          if (bndedges.Test(i)) continue;

          (*it)[2*i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[2*i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[2*i+1][0] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[2*i+1][1+j] = base_facet + ned + i * order + j;
        }

      *testout << "table = " << endl << *it << endl;
      return it;
#endif


#ifdef VER2
      // all external dofs of an element
      int ne = ma.GetNE();
      ARRAY<int> dnums;

      ARRAY<int> cnt(ne);
    
      for (int i = 0; i < ne; i++)
        {
          GetExternalDofNrs (i, dnums);
          cnt[i] = dnums.Size();
        }

      Table<int> & it = *new Table<int>(cnt);    
      for (int i = 0; i < ne; i++)
        {
          GetExternalDofNrs (i, dnums);
          for (int j = 0; j < dnums.Size(); j++)
            it[i][j] = dnums[j];
        }

      *testout << "table = " << endl << it << endl;
      return &it;
#endif







#ifdef VER3
      // edge by edge and element by element
      
      ARRAY<int> cnt(ned+nel);      
      cnt.Range (0, ned) = 2*order+2;
      cnt.Range (ned, ned+nel) = neldof;

      Table<int> * it = new Table<int>(cnt);

      for (int i = 0; i < ned; i++)
        {
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }

      for (int i = 0; i < nel; i++)
        for (int j = 0; j < neldof; j++) 
          (*it)[i+ned][j] = base_el+i*neldof+j;

      // *testout << "smblocks = " << *it << endl;

      return it;
#endif






#ifdef VER3a
      // edge by edge and element-edge by element-edge
      
      ARRAY<int> cnt(ned+3*nel);      
      cnt.Range (0, ned) = 2*order+2;
      cnt.Range (ned, ned+3*nel) = neldof/3;

      Table<int> * it = new Table<int>(cnt);

      for (int i = 0; i < ned; i++)
        {
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }

      for (int i = 0; i < nel; i++)
        for (int k = 0; k < 3; k++)
          for (int j = 0; j < neldof/3; j++) 
            (*it)[3*i+k+ned][j] = base_el+i*neldof+k*neldof/3+j;

      // *testout << "smblocks = " << *it << endl;

      return it;
#endif





#ifdef VER3b
      // edge by edge and diagonal elements
      
      ARRAY<int> cnt(ned+nel*neldof);      
      cnt.Range (0, ned) = 2*order+2;
      cnt.Range (ned, ned+nel*neldof) = 1;

      Table<int> * it = new Table<int>(cnt);

      for (int i = 0; i < ned; i++)
        {
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }

      for (int i = 0; i < nel*neldof; i++)
        (*it)[i+ned][0] = base_el+i;

      // *testout << "smblocks = " << *it << endl;

      return it;
#endif







#ifdef VER4
      // edge+element terms for each edge
      
      ARRAY<int> cnt(ned);      
      cnt = 2*order+2;
      for (int i = 0; i < nel; i++)
        {
          ma.GetElEdges (i, ednums);
          for (int j = 0; j < ednums.Size(); j++)
            cnt[ednums[j]] += neldof/3;
        }

      Table<int> * it = new Table<int>(cnt);
      for (int i = 0; i < ned; i++)
        {
          (*it)[i][0] = base_hdivn+i;
          for (int j = 0; j < order; j++)
            (*it)[i][1+j] = base_hdivn+ned + (i * order) + j;
          (*it)[i][order+1] = base_facet + i;
          for (int j = 0; j < order; j++)
            (*it)[i][order+2+j] = base_facet + ned + i * order + j;
        }

      cnt = 2*order+2;

      for (int i = 0; i < nel; i++)
        {
          ma.GetElEdges (i, ednums);
          for (int j = 0; j < ednums.Size(); j++)
            for (int k = 0; k < neldof/3; k++)
              (*it)[ednums[j]][cnt[ednums[j]]++] = i*neldof + j * neldof/3 + k;
        }

      *testout << "smblocks = " << *it << endl;

      return it;
#endif






    }
  };




 

  ////////////////////////////////////////////////////////////////////////////////

  template <int D>
  class HybridHelmholtzIntegrator : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzIntegrator (coeffs);
    }

    
    virtual bool BoundaryForm () const { return 0; }
    virtual int DimFlux () const { cout << "call dimflux" << endl; return 1; }
    virtual int DimElement () const { return D; }
    virtual int DimSpace () const { return D; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const WaveElement & fel_l2 = 
        dynamic_cast<const WaveElement&> (cfel[0]);
      const HDivHighOrderFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivHighOrderFiniteElement<D>&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
  
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();

      int base_l2 = 0;
      int base_hdivn = base_l2+nd_l2;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      HeapReset hr(lh); // void * heapp = lh.GetPointer();
        
      FlatVector<Complex> mat_l2(nd_l2, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<Complex> mat_dudn(nd_l2, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrix<Complex> hmat_l2(nd_l2, lh);

      FlatMatrixFixHeight<4, Complex> bmat(nd, lh);
      FlatMatrixFixHeight<4, Complex> dbmat(nd, lh);
      Mat<4,4,Complex> dmat, dmatstab;


      FlatMatrix<Complex> mat_l2l2 (nd_l2, lh);
      FlatMatrix<Complex> mat_gradgrad (nd_l2, lh);

      FlatMatrix<Complex> dshape(nd_l2, D, lh);

      double kappa = coef_kappa->EvaluateConst();

          
      /*
      const IntegrationRule & ir2d =
        SelectIntegrationRule (eltype, 40); // 2*fel_l2.Order()+6);

      hmat_l2 = 0.0;
      mat_l2l2  = 0.0;
      mat_gradgrad = 0.0;

      for (int l = 0; l < ir2d.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<D,D> sip(ir2d[l], eltrans, lh);

          CalcShapeDShape (fel_l2, sip, mat_l2, dshape, lh);
          
          double factor = sip.GetJacobiDet() * ir2d[l].Weight();

          mat_gradgrad += factor * dshape * Trans (dshape);
          mat_l2l2 += factor *  (mat_l2 * Trans (mat_l2));

          lh.CleanUp(heapp);
        }

      double alpha = 0;

      for (int i = 0; i < nd_l2; i++)
        for (int j = 0; j < nd_l2; j++)
          elmat(base_l2+j, base_l2+i) -= Complex(0, -kappa) * mat_l2l2(i,j) + Complex (0, 1/kappa) * mat_gradgrad(i,j);
      */



      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     

      ARRAY<int> facetdofs, hfd;


      int ned = ElementTopology::GetNFacets(eltype);
      
      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      for (int k = 0; k < ned; k++)
        {
          ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 3 * fel_facet.Order()+50);


          facetdofs.SetSize(0);
          for (int i = 0; i < nd_l2; i++)
            facetdofs.Append(base_l2+i);
          
          fel_hdivn.GetFacetDofs(k, hfd);
          for (int i = 0; i < hfd.Size(); i++)
            facetdofs.Append (base_hdivn+hfd[i]);
          for (int i = fel_facet.GetFirstFacetDof(k); i < fel_facet.GetFirstFacetDof(k+1); i++)
            facetdofs.Append (base_facet+i);

          // *testout << "facetdofs = " << endl << facetdofs << endl;

          int ndf = facetdofs.Size();
          FlatMatrixFixHeight<4, Complex> hbmat(ndf, lh), thishbmat(ndf, lh);
          FlatMatrixFixHeight<4, Complex> hdbmat(ndf, lh);
          FlatMatrix<Complex> helmat(ndf, lh);
          helmat = 0;

          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              HeapReset hr2(lh);

              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();


              Vec<D> normal = det * Trans (inv_jac) * normal_ref;       
              double len = L2Norm (normal);
              normal /= L2Norm (normal);

              Vec<D> nt = Trans(sip.GetJacobian()) * normal;
              nt /= det;

              //Piola

              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              fel_l2.CalcShapeDShape (sip, mat_l2, dshape);
              mat_dudn = dshape * normal;

              bmat = 0.0;
              bmat.Row(0).Range (base_l2   , base_l2   +nd_l2 )   = 1/kappa * mat_dudn;
              bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;

              for (int i = 0; i < ndf; i++)
                hbmat.Col(i) = bmat.Col(facetdofs[i]);

              int first_eled = nd_l2/3 * k;
              int last_eled = nd_l2/3 * (k+1);

              thishbmat = 0;
              for (int i = first_eled; i < last_eled; i++)
                thishbmat.Col(i) = bmat.Col(i);
              for (int i = nd_l2; i < ndf; i++)
                thishbmat.Col(i) = bmat.Col(facetdofs[i]);

              // thishbmat = hbmat;
              dmat = 0.0;

              /*
              dmat(0,1) = -0.5;
              dmat(1,0) = -0.5;

              dmat(2,1) = -1;
              dmat(1,2) = -1;
              dmat(2,3) = 1;
              dmat(3,2) = 1;

              dmat(1,1) = -1;
              dmat(1,3) = 1;
              dmat(3,1) = 1;
              dmat(3,3) = -1;

              dmat(2,2) += 1;
              dmat(2,0) += 1;
              dmat(0,2) += 1;
              dmat(0,0) += 1;
              */


              dmat(0,1) = -0.5;
              dmat(1,0) = -0.5;
              dmat(2,1) = -1;
              dmat(1,2) = -1;
              dmat(2,3) = 1;
              dmat(3,2) = 1;



              dmatstab = 0;

              double stab1[4][4] =   // the original one
                { {  1,  0,  1,  0 },
                  {  0, -1,  0,  1 },
                  {  1,  0,  1,  0 },
                  {  0,  1,  0, -1} };


              double stab2[4][4] = // incoming/outgoing fluxes
                { { -1, -1, -1, 1 },
                  {  1,  1,  1, -1 },
                  { -1, 1, -1, -1 },
                  { -1, 1, -1, -1 } };


              double fac[4] = { 1, 1, 1, 1 };

              for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                  dmat(i,j) += stab2[i][j] * fac[i];

              for (int j = 0; j < 4; j++)
                {
                  dmat(0,j) *= Complex(0,1);
                  dmat(j,0) *= Complex(0,1);
                }

              for (int j = 0; j < 4; j++)
                {
                  dmatstab(0,j) *= Complex(0,1);
                  dmatstab(j,0) *= Complex(0,1);
                }

              dmat *= len * ir1d[l].Weight();
              dmatstab *= len * ir1d[l].Weight();

              hdbmat = dmat * hbmat;
              helmat += Conj (Trans (hbmat)) * hdbmat;

              hdbmat = dmatstab * hbmat;
              helmat += Conj (Trans (thishbmat)) * hdbmat;
            }

          for (int i = 0; i < ndf; i++)
            for (int j = 0; j < ndf; j++)
              elmat(facetdofs[i], facetdofs[j]) += helmat(i,j);
        }



      // test matrix

      if (eltrans.GetElementNr() == 0)
        {

          FlatMatrix<Complex> elmatl2 (nd_l2, lh);
          elmatl2 = elmat.VRange (0, nd_l2).HRange (0, nd_l2);

          Matrix<Complex> pre(nd_l2), inv(nd_l2), prod(nd_l2);

          pre = Complex(0);
          for (int i = 0; i < nd_l2; i++)
            pre(i,i) = elmatl2(i,i);
          CalcInverse (pre, inv);
          prod = inv * elmatl2;

          *testout << "elmat = " << endl << elmat << endl;
          *testout << "elmatl2 = " << endl << elmatl2 << endl;

          Vector<Complex> lami(nd_l2);
          Matrix<Complex> evecs(nd_l2);

          LapackEigenValues(prod, lami, evecs);
          *testout << "lami = " << endl << lami << endl;
          *testout << "evecs = " << endl << evecs << endl;

          ofstream lamout ("lam.out");
          for (int i = 0; i < lami.Size(); i++)
            lamout << i << "  " << lami(i).real() << "   " << lami(i).imag() << "   " << abs(lami(i)-1.0) << endl;
        }
    }


    virtual void
    CalcFlux (const FiniteElement & bfel,
              const BaseSpecificIntegrationPoint & sip,
              const FlatVector<Complex> & elx, 
              FlatVector<Complex> & flux,
              bool applyd,
              LocalHeap & lh) const
    {
      flux.AssignMemory (1, lh);


      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (bfel);
    
      const WaveElement & fel_l2 = 
        dynamic_cast<const WaveElement&> (cfel[0]);
      
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();

      int base_l2 = 0;
        
      FlatVector<Complex> mat_l2(nd_l2, lh);
      FlatMatrix<Complex> dshape(nd_l2, D, lh);

      double kappa = coef_kappa->EvaluateConst();
      
      fel_l2.CalcShapeDShape(static_cast<const SpecificIntegrationPoint<D,D>&> (sip), 
                             mat_l2, dshape);

      flux(0) = InnerProduct (elx.Range(0, nd_l2), mat_l2);
    }
  };


















  template <int D>
  class HybridHelmholtzPrecond : public BilinearFormIntegrator
  {
  protected:
    CoefficientFunction *coef_kappa;
  public:
    HybridHelmholtzPrecond (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef_kappa  = coeffs[0];
    }

    virtual ~HybridHelmholtzPrecond () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridHelmholtzPrecond (coeffs);
    }

    
    virtual bool BoundaryForm () const { return 0; }
    virtual int DimFlux () const { cout << "call dimflux" << endl; return 1; }
    virtual int DimElement () const { return D; }
    virtual int DimSpace () const { return D; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<double> & elmat,
                                        LocalHeap & lh) const
    {
      throw Exception ("no real matrix for HybridHelmholtz");
    }

    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      static int timer = NgProfiler::CreateTimer ("hybridhelmholtz");
      static int timer2 = NgProfiler::CreateTimer ("hybridhelmholtz boundary");

      NgProfiler::RegionTimer reg (timer);

      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);
    
      const WaveElement & fel_l2 = 
        dynamic_cast<const WaveElement&> (cfel[0]);
      const HDivHighOrderFiniteElement<D> & fel_hdivn = 
        dynamic_cast<const HDivHighOrderFiniteElement<D>&> (cfel[1]);
      const FacetVolumeFiniteElement & fel_facet = 
        dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
  
      ELEMENT_TYPE eltype = cfel.ElementType();
      
      int nd = cfel.GetNDof();          
      int nd_l2 = fel_l2.GetNDof();
      int nd_hdivn = fel_hdivn.GetNDof();
      int nd_facet = fel_facet.GetNDof();

      int base_l2 = 0;
      int base_hdivn = base_l2+nd_l2;
      int base_facet = base_hdivn+nd_hdivn;
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;

      HeapReset hr(lh); // void * heapp = lh.GetPointer();
        
      FlatVector<Complex> mat_l2(nd_l2, lh);
      FlatVector<> mat_hdivn_n(nd_hdivn, lh);
      FlatVector<Complex> mat_dudn(nd_l2, lh);
      FlatVector<> mat_facet(nd_facet, lh);
      
      FlatMatrix<Complex> hmat_l2(nd_l2, lh);

      FlatMatrixFixHeight<4, Complex> bmat(nd, lh);
      FlatMatrixFixHeight<4, Complex> dbmat(nd, lh);
      Mat<4,4,Complex> dmat;

      FlatMatrix<Complex> mat_l2l2 (nd_l2, lh);
      FlatMatrix<Complex> mat_gradgrad (nd_l2, lh);

      FlatMatrix<Complex> dshape(nd_l2, D, lh);

      double kappa = coef_kappa->EvaluateConst();

      // The facet contribution
      NgProfiler::RegionTimer reg2 (timer2);     


      int ned = ElementTopology::GetNFacets(eltype);
      
      Facet2ElementTrafo transform(eltype); 
      const NORMAL * normals = ElementTopology::GetNormals(eltype);

      for (int k = 0; k < ned; k++)
        {
          ELEMENT_TYPE etfacet = ElementTopology::GetFacetType (eltype, k);

          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];

          const IntegrationRule & ir1d =
            SelectIntegrationRule (etfacet, 3 * fel_facet.Order()+50);


          for (int l = 0; l < ir1d.GetNIP(); l++)
            {
              HeapReset hr2(lh);

              IntegrationPoint ip = transform(k, ir1d[l]);
              SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
	    
              Mat<D> jac = sip.GetJacobian();
              Mat<D> inv_jac = sip.GetJacobianInverse();
              double det = sip.GetJacobiDet();


              Vec<D> normal = det * Trans (inv_jac) * normal_ref;       
              double len = L2Norm (normal);
              normal /= L2Norm (normal);

              Vec<D> nt = Trans(sip.GetJacobian()) * normal;
              nt /= det;

              //Piola

              mat_hdivn_n = fel_hdivn.GetShape(sip.IP(), lh) * nt;

              fel_facet.CalcFacetShape(k, ir1d[l], mat_facet);
              fel_l2.CalcShapeDShape (sip, mat_l2, dshape);
              mat_dudn = dshape * normal;

              bmat = 0.0;
              bmat.Row(0).Range (base_l2   , base_l2   +nd_l2 )   = 1/kappa * mat_dudn;
              bmat.Row(1).Range (base_l2   , base_l2   +nd_l2   ) = mat_l2;
              bmat.Row(2).Range (base_hdivn, base_hdivn+nd_hdivn) = mat_hdivn_n;
              bmat.Row(3).Range (base_facet, base_facet+nd_facet) = mat_facet;

              
              int first_eled = nd_l2/3 * k;
              int last_eled = nd_l2/3 * (k+1);

              for (int m = 0; m < first_eled; m++)
                bmat.Col(base_l2+m) = 0.0;
              for (int m = last_eled; m < nd_l2; m++)
                bmat.Col(base_l2+m) = 0.0;
              
              dmat = 0.0;


              // dmat(0,1) = -0.5;
              // dmat(1,0) = -1;
              // dmat(1,2) = -1;

              // dmat(2,1) = -1;
              // dmat(2,3) = 1;

              dmat(3,2) = 1;
              // dmat(3,0) = 1;

             double stab[4][4] = 
                { { -1, -1, -1, 1 },
                  {  1,  1,  1, -1 },
                  { -1, 1, -1, -1 },
                  { -1, 1, -1, -1 } };

              double fac[4] = { 100, 100, 1, 1 };

              for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                  dmat(i,j) += stab[i][j] * fac[i];

              for (int j = 0; j < 4; j++)
                {
                  dmat(0,j) *= Complex(0,1);
                  dmat(j,0) *= Complex(0,1);
                }

              dmat *= len * ir1d[l].Weight();
              dbmat = dmat * bmat;
              elmat += Conj (Trans (bmat)) * dbmat;
            }
        }

      // lh.CleanUp(heapp);
    }
  };











  template <int D>
  class HybridRobinIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction *coef;
  public:
    HybridRobinIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
    { 
      coef = coeffs[0];
    }

    virtual ~HybridRobinIntegrator () { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new HybridRobinIntegrator (coeffs);
    }
    
    virtual bool BoundaryForm () const 
    { return 1; }


    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<> & elmat,
                                        LocalHeap & lh) const
    {
      cerr << "Helmholtz - Robin real called" << endl;
    }




    virtual void AssembleElementMatrix (const FiniteElement & fel,
                                        const ElementTransformation & eltrans, 
                                        FlatMatrix<Complex> & elmat,
                                        LocalHeap & lh) const
    {
      const CompoundFiniteElement & cfel = 
        dynamic_cast<const CompoundFiniteElement&> (fel);

      const HDivHighOrderNormalFiniteElement<D-1> & hdivnfel = 
        dynamic_cast<const HDivHighOrderNormalFiniteElement<D-1>&> (cfel[1]);
      const FacetFacetFiniteElement & ffel = 
        dynamic_cast<const FacetFacetFiniteElement&> (cfel[2]);
    
      int nd = cfel.GetNDof();  
      
      elmat.AssignMemory(nd, nd, lh);
      elmat = 0.0;
      
      int nd_hdivn = hdivnfel.GetNDof();
      int nd_facet = ffel.GetNDof();
      int base_facet = nd-nd_facet;
      int base_hdivn = base_facet-nd_hdivn;
      
      FlatVector<> shape(nd_facet, lh), shapehdivn(nd_hdivn, lh);
      FlatVector<Complex> vu(nd, lh), vv(nd, lh);
      
      const IntegrationRule & ir1d =
        GetIntegrationRules().SelectIntegrationRule (ffel.ElementType(), 2*ffel.Order()+10);
      
      FlatMatrixFixHeight<2> bmat(nd, lh);
      FlatMatrixFixHeight<2> dbmat(nd, lh);
      Mat<2> dmat(2);

      vu = 0.0;
      vv = 0.0;
      for (int l=0; l<ir1d.GetNIP(); l++)
        {
          SpecificIntegrationPoint<D-1,D> sip(ir1d[l], eltrans, lh);
          double val = coef->Evaluate(sip);

          if (val == 0) continue;

          hdivnfel.CalcShape(ir1d[l], shapehdivn);
          shapehdivn *= 1.0 / sip.GetJacobiDet();

          ffel.CalcShape(ir1d[l], shape);

          double factor = sip.GetJacobiDet()*ir1d[l].Weight();

          bmat = 0.0;
          bmat.Row(0).Range (base_hdivn, base_hdivn+nd_hdivn) = shapehdivn;
          bmat.Row(1).Range (base_facet, nd) = shape;

          dmat = 0.0;

          dmat(1,1) =  -1;

          dbmat = dmat * bmat;
          elmat += factor * Trans (bmat) * dbmat;
        }
    }
  };


  class Init
  { 
  public: 
    Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybridhelmholtz", HybridHelmholtzFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 2, 1,
                                      HybridHelmholtzIntegrator<2>::Create);

    GetIntegrators().AddBFIntegrator ("hybridhelmholtzprecond", 2, 1,
                                      HybridHelmholtzPrecond<2>::Create);
  
    /*
    GetIntegrators().AddBFIntegrator ("hybridhelmholtz", 3, 1,
                                      HybridHelmholtzIntegrator<3>::Create);
    */

    GetIntegrators().AddBFIntegrator ("hybridrobin", 2, 1,
                                      HybridRobinIntegrator<2>::Create);
    /*
    GetIntegrators().AddBFIntegrator ("hybridrobin", 3, 1,
                                      HybridRobinIntegrator<3>::Create);
    */
  }

  Init init;
  
} // namespace

