/*

  Solver for computing the transfer function of hydraulic pipes

 */



#include <solve.hpp>

#include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;


  class Hydraulic : public NumProc
  {
  protected:
    BilinearForm *bfa, *bfm;
    LinearForm * lff;    
    LinearForm * lfb;
    GridFunction * gfu;

    double omega_min, omega_max, omega_step;
  public:

    Hydraulic (PDE & apde, const Flags & flags);

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new Hydraulic (pde, flags);
    }

    virtual void Do(LocalHeap & lh);

    virtual string GetClassName () const
    {
      return "Hydraulic solver";
    }
  };
  

  
  Hydraulic ::   
  Hydraulic (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    cout << "Create Hydraulic" << endl;
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", "a"));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", "m"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearformf", "f"));
    lfb = pde.GetLinearForm (flags.GetStringFlag ("linearformb", "b"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    omega_min = flags.GetNumFlag ("omega_min", 1);
    omega_max = flags.GetNumFlag ("omega_max", 10);
    omega_step = flags.GetNumFlag ("omega_step", 1);
  }


  void Hydraulic :: Do(LocalHeap & lh)
  {
    cout << "Hydraulic solver called" << endl;
    

    BaseVector & vecu = gfu->GetVector();

    BaseVector & vecf = lff->GetVector();
    BaseVector & vecb = lfb->GetVector();

    const SparseMatrixSymmetric<Complex> & mata =
      dynamic_cast<const SparseMatrixSymmetric<Complex>&> (bfa->GetMatrix());
    const BaseSparseMatrix & matm =
      dynamic_cast<const BaseSparseMatrix&> (bfm->GetMatrix());

    
    SparseMatrixSymmetric<Complex> & hmat =
      dynamic_cast<SparseMatrixSymmetric<Complex>&> (*mata.CreateMatrix());


    ofstream tout ("transfer.out");

    hmat.AsVector() = mata.AsVector() - matm.AsVector();
    SparseCholesky<Complex> inv_hmat (hmat);


    for (double omega = omega_min; omega <= omega_max; omega += omega_step)
      {
	cout << "omega = " << omega << endl;

	hmat.AsVector() = mata.AsVector() - omega*omega * matm.AsVector();
	inv_hmat.FactorNew (hmat);

	vecu = inv_hmat * vecf;

	Complex scal = S_InnerProduct<Complex> (vecu, vecb);
	tout << omega << "  " << scal.real() << "  " << scal.imag() << "  " << abs(scal) << "   " << arg(scal) << endl;
      }
    
    delete &hmat;
  }


  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("hydraulic", Hydraulic::Create);
    }

    Init init;
  }

}
