#include <solve.hpp>


namespace ngcomp
{
  
  
  class InfiniteFESpace : public FESpace
  {
    ARRAY<int> first_infinite_dof;
    int inf_per_node;
  public:
    InfiniteFESpace (const MeshAccess & ama, const Flags & flags);

    virtual ~InfiniteFESpace () { ; }

    virtual string GetClassName () const
    {
      return "InfiniteFESpace";
    }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      return new InfiniteFESpace (ma, flags); 
    }

    
    ///
    virtual void Update();
    ///
    virtual int GetNDof () const;
    ///
    virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
    ///
    virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;

    virtual Table<int> * CreateSmoothingBlocks (int type = 0) const;

    virtual ARRAY<int> * CreateDirectSolverClusters (int type = 0) const;
  };



    

  InfiniteFESpace :: InfiniteFESpace (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama, flags)
  {
    inf_per_node = 6;
    // inf_per_node = 6*6+3;

    tet     = new FE_Tet1;
    prism   = new FE_Prism1;
    pyramid = new FE_Pyramid1;
    hex     = new FE_Hex1;
    trig    = new FE_Trig1;
    quad    = new FE_Quad1;
    segm    = new FE_Segm1;

    if (ma.GetDimension() == 2)
      {
	evaluator = 
	  new MassIntegrator<2> (new ConstantCoefficientFunction(1));
	boundary_evaluator = 0;
      }
    else
      {
	evaluator = 
	  new MassIntegrator<3> (new ConstantCoefficientFunction(1));
	boundary_evaluator = 
	  new RobinIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
  }


  int InfiniteFESpace :: GetNDof () const
  {
    return first_infinite_dof[ma.GetNV()];
  }


  void InfiniteFESpace :: Update()
  {
    int i, j, k;
    int nse = ma.GetNSE(); 
    int nv = ma.GetNV();

    ARRAY<int> is_infinite(nv);
    first_infinite_dof.SetSize(nv+1);
    ARRAY<int> vnums;

    is_infinite = 0;
    for (i = 0; i < nse; i++)
      {
	if (ma.GetSElIndex(i) == 1)
	  {
	    ma.GetSElVertices (i, vnums);
	    for (j = 0; j < vnums.Size(); j++)
	      is_infinite[vnums[j]] = 1;
	  }
      }

    int cnt = nv;
    for (i = 0; i < nv; i++)
      {
	first_infinite_dof[i] = cnt;
	if (is_infinite[i]) cnt += inf_per_node;
      }
    first_infinite_dof[i] = cnt;
    (*testout) << "first_inf = " << endl << first_infinite_dof << endl;
  }

 
  void InfiniteFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ma.GetElVertices (elnr, dnums);
  }


  void InfiniteFESpace :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> vnums;
    ma.GetSElVertices (selnr, vnums);
    
    int i, j;
    dnums.SetSize (0);

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    int n_inf = first_infinite_dof[vnums[0]+1]-first_infinite_dof[vnums[0]];
    for (i = 0; i < n_inf; i++)
      for (j = 0; j < vnums.Size(); j++)
	dnums.Append (first_infinite_dof[vnums[j]]+i);
  }
  

  Table<int> * InfiniteFESpace :: CreateSmoothingBlocks (int type) const
  {

    {
      // vertices with their inf-dofs
    int i, j;
    int nv = ma.GetNV();
    ARRAY<int> cnt(nv);
    
    for (i = 0; i < nv; i++)
      cnt[i] = 1 + first_infinite_dof[i+1]-first_infinite_dof[i];

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < nv; i++)
      {
	tab[i][0] = i;
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	for (j = 0; j < ninf; j++)
	  tab[i][j+1] = first_infinite_dof[i]+j;
      }
    (*testout) << "table = " << tab << endl;
    return &tab;
    }


    /*
    // only boundary vertices with their inf-dofs
    int i, j;
    int nv = ma.GetNV();
    ARRAY<int> cnt(nv);
    
    for (i = 0; i < nv; i++)
      {
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	if (ninf)
	  cnt[i] = ninf+1;
	else
	  cnt[i] = 0;
      }

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < nv; i++)
      {
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	if (ninf)
	  {
	    tab[i][0] = i;
	    for (j = 0; j < ninf; j++)
	      tab[i][j+1] = first_infinite_dof[i]+j;
	  }
      }
    (*testout) << "table = " << tab << endl;
    return &tab;
    */


    /*
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(nv+1);
    
    for (i = 0; i < nv; i++)
      cnt[i] = 1 + first_infinite_dof[i+1]-first_infinite_dof[i];
    cnt[nv] = nbound+ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < nv; i++)
      {
	tab[i][0] = i;
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	for (j = 0; j < ninf; j++)
	  tab[i][j+1] = first_infinite_dof[i]+j;
      }
    for (i = 0; i < ninf; i++)
      tab[nv][i] = nv+i;
    int ct = ninf;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] != first_infinite_dof[i+1])
	{
	  tab[nv][ct] = i;
	  ct++;
	}
    
    (*testout) << "table = " << tab << endl;
    return &tab;
    */

    /*
    // only infinite dofs
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;

    ARRAY<int> cnt(1);
    
    cnt[0] = ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < ninf; i++)
      tab[0][i] = nv+i;
    
    (*testout) << "table = " << tab << endl;
    return &tab;
    */


    // only infinite dofs in seperate layers
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(inf_per_node);
    cnt = nbound;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < inf_per_node; i++)
      for (j = 0; j < nbound; j++)
	{
	  tab[i][j] = nv+i+j*inf_per_node;
	}

    (*testout) << "table = " << tab << endl;
    return &tab;



    /*
    // boundary finite and infinite in one block
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(1);
    
    cnt[0] = nbound + ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < ninf; i++)
      tab[0][i] = nv+i;
    int ct = ninf;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] != first_infinite_dof[i+1])
	{
	  tab[0][ct] = i;
	  ct++;
	}
    */

    
    (*testout) << "table = " << tab << endl;
    return &tab;

  }

  
  ARRAY<int> * InfiniteFESpace :: CreateDirectSolverClusters (int type) const
  {
    int i, j;
    int nv = ma.GetNV();

    ARRAY<int> & cl = *new ARRAY<int> (GetNDof());
    cl = 0;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] == first_infinite_dof[i+1])
	cl[i] = 1;

    return & cl;
  }




  
  















  class HOInfiniteFESpace : public H1HighOrderFESpace
  {
    ARRAY<int> first_infinite_dof;
    int inf_per_node;
    int inf_bound;
  public:
    HOInfiniteFESpace (const MeshAccess & ama, const Flags & flags);

    virtual ~HOInfiniteFESpace () { ; }

    virtual string GetClassName () const
    {
      return "HOInfiniteFESpace";
    }

    static FESpace * Create (const MeshAccess & ma, const Flags & flags)
    {
      cout << "Create HO InfiniteFESpace" << endl;
      return new HOInfiniteFESpace (ma, flags); 
    }
    
    virtual void Update();

    virtual int GetNDof () const;

    virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;

    virtual Table<int> * CreateSmoothingBlocks (int type = 0) const;

    virtual ARRAY<int> * CreateDirectSolverClusters (int type = 0) const;
  };



    

  HOInfiniteFESpace :: HOInfiniteFESpace (const MeshAccess & ama, const Flags & flags)
    : H1HighOrderFESpace (ama, flags)
  {
    inf_per_node = int(flags.GetNumFlag ("inf_per_node", 0));
    inf_bound = int(flags.GetNumFlag ("inf_bound", 1));

    if (ma.GetDimension() == 2)
      {
	evaluator = 
	  new MassIntegrator<2> (new ConstantCoefficientFunction(1));
	boundary_evaluator = 0;
      }
    else
      {
	evaluator = 
	  new MassIntegrator<3> (new ConstantCoefficientFunction(1));
	boundary_evaluator = 
	  new RobinIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);

    low_order_space = 0;
  }


  int HOInfiniteFESpace :: GetNDof () const
  {
    return first_infinite_dof[H1HighOrderFESpace::GetNDof()];
  }


  void HOInfiniteFESpace :: Update()
  {
    cout << "update HOInfiniteFESpace" << endl;
    H1HighOrderFESpace::Update();
    cout << "2" << endl;

    int i, j, k;
    int nse = ma.GetNSE(); 
    int nv = H1HighOrderFESpace::GetNDof();

    ARRAY<int> is_infinite(nv);
    first_infinite_dof.SetSize(nv+1);
    ARRAY<int> sdnums;

    is_infinite = 0;
    for (i = 0; i < nse; i++)
      {
	(*testout) << "surf-element " << i << " has ind " << ma.GetSElIndex(i) << endl;
	if (ma.GetSElIndex(i) == inf_bound-1)
	  {
	    H1HighOrderFESpace::GetSDofNrs (i, sdnums);
	    for (j = 0; j < sdnums.Size(); j++)
	      is_infinite[sdnums[j]] = 1;
	  }
      }

    (*testout) << "is_inf" << endl << is_infinite << endl;

    int cnt = nv;
    for (i = 0; i < nv; i++)
      {
	first_infinite_dof[i] = cnt;
	if (is_infinite[i]) cnt += inf_per_node;
      }
    first_infinite_dof[i] = cnt;
    (*testout) << "first_inf = " << endl << first_infinite_dof << endl;
  }

 
  void HOInfiniteFESpace :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    ARRAY<int> vnums;
    H1HighOrderFESpace::GetSDofNrs (selnr, vnums);

    int i, j, n_inf;
    dnums.SetSize (0);

    if (ma.GetSElIndex(selnr) == inf_bound-1)
      n_inf = inf_per_node;
    else
      n_inf = 0;
    // int n_inf = first_infinite_dof[vnums[0]+1]-first_infinite_dof[vnums[0]];

    for (i = 0; i < vnums.Size(); i++)
      dnums.Append (vnums[i]);

    for (i = 0; i < n_inf; i++)
      for (j = 0; j < vnums.Size(); j++)
	dnums.Append (first_infinite_dof[vnums[j]]+i);
  }
  



  Table<int> * HOInfiniteFESpace :: CreateSmoothingBlocks (int type) const
  {

    {
      // FEM-dofs together with their inf-dofs
      int i, j;
      int nv = H1HighOrderFESpace::GetNDof();
      ARRAY<int> cnt(nv);
    
      for (i = 0; i < nv; i++)
	cnt[i] = 1 + first_infinite_dof[i+1]-first_infinite_dof[i];
      
      Table<int> & tab = *new Table<int> (cnt);

      for (i = 0; i < nv; i++)
	{
	  tab[i][0] = i;
	  int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	  for (j = 0; j < ninf; j++)
	    tab[i][j+1] = first_infinite_dof[i]+j;
	}
      (*testout) << "table = " << tab << endl;
      return &tab;
    }


    /*
    // only boundary vertices with their inf-dofs
    int i, j;
    int nv = ma.GetNV();
    ARRAY<int> cnt(nv);
    
    for (i = 0; i < nv; i++)
      {
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	if (ninf)
	  cnt[i] = ninf+1;
	else
	  cnt[i] = 0;
      }

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < nv; i++)
      {
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	if (ninf)
	  {
	    tab[i][0] = i;
	    for (j = 0; j < ninf; j++)
	      tab[i][j+1] = first_infinite_dof[i]+j;
	  }
      }
    (*testout) << "table = " << tab << endl;
    return &tab;
    */


    /*
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(nv+1);
    
    for (i = 0; i < nv; i++)
      cnt[i] = 1 + first_infinite_dof[i+1]-first_infinite_dof[i];
    cnt[nv] = nbound+ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < nv; i++)
      {
	tab[i][0] = i;
	int ninf = first_infinite_dof[i+1]-first_infinite_dof[i];
	for (j = 0; j < ninf; j++)
	  tab[i][j+1] = first_infinite_dof[i]+j;
      }
    for (i = 0; i < ninf; i++)
      tab[nv][i] = nv+i;
    int ct = ninf;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] != first_infinite_dof[i+1])
	{
	  tab[nv][ct] = i;
	  ct++;
	}
    
    (*testout) << "table = " << tab << endl;
    return &tab;
    */

    /*
    // only infinite dofs
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;

    ARRAY<int> cnt(1);
    
    cnt[0] = ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < ninf; i++)
      tab[0][i] = nv+i;
    
    (*testout) << "table = " << tab << endl;
    return &tab;
    */


    // only infinite dofs in seperate layers
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(inf_per_node);
    cnt = nbound;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < inf_per_node; i++)
      for (j = 0; j < nbound; j++)
	{
	  tab[i][j] = nv+i+j*inf_per_node;
	}

    (*testout) << "table = " << tab << endl;
    return &tab;



    /*
    // boundary finite and infinite in one block
    int i, j;
    int nv = ma.GetNV();
    int ninf = first_infinite_dof[nv] - nv;
    int nbound = ninf / inf_per_node;

    ARRAY<int> cnt(1);
    
    cnt[0] = nbound + ninf;

    Table<int> & tab = *new Table<int> (cnt);

    for (i = 0; i < ninf; i++)
      tab[0][i] = nv+i;
    int ct = ninf;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] != first_infinite_dof[i+1])
	{
	  tab[0][ct] = i;
	  ct++;
	}
    */

    
    (*testout) << "table = " << tab << endl;
    return &tab;

  }

  
  ARRAY<int> * HOInfiniteFESpace :: CreateDirectSolverClusters (int type) const
  {
    int i, j;
    int nv = ma.GetNV();

    ARRAY<int> & cl = *new ARRAY<int> (GetNDof());
    cl = 0;
    for (i = 0; i < nv; i++)
      if (first_infinite_dof[i] == first_infinite_dof[i+1])
	cl[i] = 1;

    return & cl;
  }












  class AbsorbingPreconditioner : public Preconditioner
  {
  public:
    AbsorbingPreconditioner (const PDE & apde, const Flags & aflags)
      : Preconditioner (&apde, aflags)
    {
      ;
    }

    virtual ~AbsorbingPreconditioner() 
    { ; }

    static Preconditioner * Create (const PDE & pde, const Flags & flags)
    {
      return new AbsorbingPreconditioner (pde, flags);
    }

    virtual void Update ()
    { ; }
    virtual const BaseMatrix & GetMatrix() const
    { ; }
  };




    
  // register FESpaces
  namespace {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("infinitefe", HOInfiniteFESpace::Create);
      GetPreconditionerClasses().AddPreconditioner ("absorbing", AbsorbingPreconditioner::Create);
    }
    
    Init init;
  }
  
}
