/*

Drift Diffusion model in mixed form

*/


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


// unit length is 10^{-6} m

double dt = 0.1;
// double taurec = 0.01;

double kb = 1.38e-23;  // boltzmann constant [ J / K ]
double q  = 1.602e-19; // unit charge [ As ]
double temp = 300;       // lattice temperature [ K ]
double ut = kb * temp / q;
double eps0 = 8.854e-12; // dielectric constant in vacuum [ As / Vm ]
CoefficientFunction * cf_mu = 0;
CoefficientFunction * cf_eps = 0;
CoefficientFunction * cf_ni = 0;   // ni * q
CoefficientFunction * cf_n0 = 0;   
CoefficientFunction * cf_p0 = 0;   
CoefficientFunction * cf_rho = 0;  // doping * q
CoefficientFunction * cf_taurec = 0;   // ni * q

double nini = 9.29e31*q*q * 1e6;



// the product space for (D, phi, jn, n, jp, p, En, Ep)
class FESpaceMixedDD : public CompoundFESpace
{
public:
  FESpaceMixedDD (const MeshAccess & ama, 		   
		  const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  { 
    ;
  }

  virtual ~FESpaceMixedDD () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(8);

    Flags hdivflags, l2flags;

    int order = int (flags.GetNumFlag ("order", 1));

    l2flags.SetFlag ("order", max (order-1, 0));
    hdivflags.SetFlag ("order", order);

    if (order < 0)
      {
    	spaces[0] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[1] = new ElementFESpace (ma, 0, 0, 0);    
    	spaces[2] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[3] = new ElementFESpace (ma, 0, 0, 0);    
    	spaces[4] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[5] = new ElementFESpace (ma, 0, 0, 0);   
    	spaces[6] = new RaviartThomasFESpace (ma, 1, 0);
    	spaces[7] = new RaviartThomasFESpace (ma, 1, 0); 
      }
    else
      {
	spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
	spaces[2] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[3] = new L2HighOrderFESpace (ma, l2flags);
	spaces[4] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[5] = new L2HighOrderFESpace (ma, l2flags);
	spaces[6] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[7] = new HDivHighOrderFESpace (ma, hdivflags);
      }
    
    FESpaceMixedDD * fes = new FESpaceMixedDD (ma, spaces);
    return fes;
  }
    
  virtual string GetClassName () const { return "MixedDD - FESpace"; }
};


/* ************************************************************ */
/* ************************ Integrator ************************ */
/* ************************************************************ */

class DiffOpDD : public DiffOp<DiffOpDD>
{
  // 18 components:
  // Dx, Dy, div D, phi, jnx, jny, div jn, n, jpx, jpy, div jp, p, En, div En, Ep, div Ep

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 18 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();

    FlatVector<> shape_scal(nd_scal, lh);
    shape_scal = fel_scal.GetShape (sip.IP(), lh);

    FlatMatrix<> shape_vec(2, nd_vec, lh);
    shape_vec = (1.0/sip.GetJacobiDet()) * 
      (sip.GetJacobian() * Trans (fel_vec.GetShape(sip.IP(), lh)));

    FlatVector<> shape_div(nd_vec, lh);
    shape_div = 1.0/sip.GetJacobiDet() * fel_vec.GetDivShape(sip.IP(), lh);

    mat = 0;
    for (int i = 0; i < 3; i++)
      {
	int base = i * (nd_vec + nd_scal);
	for (int j = 0; j < nd_vec; j++)
	  {
	    mat(4*i  ,base+j) = shape_vec(0, j);
	    mat(4*i+1,base+j) = shape_vec(1, j);
	    mat(4*i+2,base+j) = shape_div(j);
	  }
	for (int j = 0; j < nd_scal; j++)
	  mat(4*i+3,base+nd_vec+j) = shape_scal(j);
      }
    for (int i = 0; i < 2; i++)
      {
	int base = 3 * (nd_vec + nd_scal) + i * nd_vec;
	for (int j = 0; j < nd_vec; j++)
	  {
	    mat(3*i+12,base+j) = shape_vec(0, j);
	    mat(3*i+13,base+j) = shape_vec(1, j);
	    mat(3*i+14,base+j) = shape_div(j);
	  }
      }
    // (*testout) << "bmat = " << endl << mat << endl;
  }

  template <typename FEL, typename SIP, class TVX, class TVY>
  static void Apply (const FEL & bfel, const SIP & sip,
		     const TVX & x, TVY & y,
		     LocalHeap & lh) 
  {
    const CompoundFiniteElement & cfel = 
      static_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_scal = 
      static_cast<const NodalFiniteElement&> (cfel[1]);
    const HDivFiniteElement<2> & fel_vec = 
      static_cast<const HDivFiniteElement<2>&> (cfel[0]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();
    
    FlatVector<> shape_scal(nd_scal, lh);
    fel_scal.CalcShape (sip.IP(), shape_scal);
    
    FlatMatrixFixWidth<2> shape_vec(nd_vec, lh);
    fel_vec.CalcShape (sip.IP(), shape_vec);
    
    FlatVector<> shape_div(nd_vec, lh);
    shape_div = fel_vec.GetDivShape(sip.IP(), lh);

    typedef typename TVX::TSCAL TSCAL;
    

    for (int i = 0; i < 3; i++)
      {
	Vec<2,TSCAL> h1, h2;
	Vec<1,TSCAL> hdiv, hs;

	int base = i * (nd_vec + nd_scal);

	h1 = Trans (shape_vec) * x.Range (base, base+nd_vec);
	h2 =  (1.0/sip.GetJacobiDet()) * (sip.GetJacobian() * h1);
	hdiv =  (1.0/sip.GetJacobiDet()) * (Trans (shape_div) * x.Range (base, base+nd_vec));
	hs = Trans (shape_scal) * x.Range (base+nd_vec, base+nd_vec+nd_scal);

	y(4*i  ) = h2(0);
	y(4*i+1) = h2(1);
	y(4*i+2) = hdiv(0);
	y(4*i+3) = hs(0);
      }

    for (int i = 0; i < 2; i++)
      {
	Vec<2,TSCAL> h1, h2;
	Vec<1,TSCAL> hdiv;

	int base = 3 * (nd_vec + nd_scal) + i*nd_vec;

	h1 = Trans (shape_vec) * x.Range (base, base+nd_vec);
	h2 =  (1.0/sip.GetJacobiDet()) * (sip.GetJacobian() * h1);
	hdiv =  (1.0/sip.GetJacobiDet()) * (Trans (shape_div) * x.Range (base, base+nd_vec));

	y(3*i+12) = h2(0);
	y(3*i+13) = h2(1);
	y(3*i+14) = hdiv(0);
      }
  }
};


class DDDMat : public DMatOp<DDDMat>
{
public:

  enum { DIM_DMAT = 18 };
  enum { SYMMETRIC = 0 };

  DDDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    Vec<18> vlin;
    vlin = 0.0;
    GenerateLinearizedMatrix (fel, sip, vlin, mat, lh);
  }  


  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    double coef_b = cf_mu -> Evaluate (sip);
    double coef_d = kb/q*temp * cf_mu -> Evaluate (sip);
    double coef_eps = cf_eps -> Evaluate (sip) * eps0;
    double coef_ni = cf_ni -> Evaluate (sip);
    double coef_n0 = cf_n0 -> Evaluate (sip);
    double coef_p0 = cf_p0 -> Evaluate (sip);
    double taurec = 1.0 / cf_taurec -> Evaluate (sip);

    mat = 0;
    mat(0,0) = mat(1,1) = 1.0 / coef_eps;
    mat(2,3) = mat(3,2) = -1;

    mat(3,7) = -1;        // div D = n - p;
    mat(3,11) = 1;        

    mat(4,4) = mat(5,5) = 1.0 / coef_d;
    mat(6,7) = mat(7,6) = -1;
    mat(7,7) = -1.0/dt;  

    mat(8,8) = mat(9,9) = 1.0 / coef_d;
    mat(10,11) = mat(11,10) = -1;
    mat(11,11) = -1.0/dt;


    // jn_drift / d = b/d n E
    
    double brel = -coef_b / coef_d / coef_eps;
    double brel2 = -coef_b / coef_d;
    mat(4,7) -= brel * vlin(0);
    mat(5,7) -= brel * vlin(1);
    mat(4,0) -= brel * vlin(7);
    mat(5,1) -= brel * vlin(7);

    mat(4,12) += brel2 * vlin(7);
    mat(5,13) += brel2 * vlin(7);


    // jp_drift / d = b/d p E
    mat(8,11) += brel * vlin(0);
    mat(9,11) += brel * vlin(1);
    mat(8,0) += brel * vlin(11);
    mat(9,1) += brel * vlin(11);

    mat(8,15) += brel2 * vlin(11);
    mat(9,16) += brel2 * vlin(11);


    // recombination term
    if (ReduceToReal (vlin(7)) > 0)
      {
	mat(7,11) -= taurec*vlin(7);
	mat(11,11) -= taurec*vlin(7);
      }
    
    if (ReduceToReal (vlin(11)) > 0)
      {
	mat(7,7) -= taurec*vlin(11);
	mat(11,7) -= taurec*vlin(11);
      }

    /*
    mat.Row(4) /= coef_n0;
    mat.Row(5) /= coef_n0;
    mat.Row(6) /= coef_n0;
    mat.Row(7) /= coef_n0;
    mat.Row(8) /= coef_p0;
    mat.Row(9) /= coef_p0;
    mat.Row(10) /= coef_p0;
    mat.Row(11) /= coef_p0;
    */

    mat(12,12) = 1;
    mat(13,13) = 1;
    mat(15,15) = 1;
    mat(16,16) = 1;
  }



  template <typename FEL, typename SIP, class VECX, class VECY>
  void Apply (const FEL & fel, const SIP & sip, 
	      const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    double coef_b = cf_mu -> Evaluate (sip);
    double coef_d = kb/q*temp * cf_mu -> Evaluate (sip);
    double coef_eps = cf_eps -> Evaluate (sip) * eps0;
    double coef_nini = sqr (cf_ni -> Evaluate (sip));
    double n0 = cf_n0 -> Evaluate (sip);
    double p0 = cf_p0 -> Evaluate (sip);
    double coef_rho = cf_rho -> Evaluate (sip);
    double taurec = 1.0 / cf_taurec -> Evaluate (sip);

    y(0) = x(0) / coef_eps;
    y(1) = x(1) / coef_eps;
    y(2) = -x(3);
    y(3) = -x(2) + coef_rho - x(7) + x(11);      // dif D + rho - n + p
    
    y(4) = x(4) / coef_d;
    y(5) = x(5) / coef_d;
    y(6) = -x(7);
    y(7) = -x(6) - 1.0/dt * x(7);

    y(8) = x(8) / coef_d;
    y(9) = x(9) / coef_d;
    y(10) = -x(11);
    y(11) = -x(10) - 1.0/dt * x(11);

    // jn = b n E
    double brel = -coef_b / coef_d / coef_eps;
    double brel2 = -coef_b / coef_d;
    y(4) -= brel * x(0) * x(7);
    y(5) -= brel * x(1) * x(7);
    y(4) += brel2 * x(12) * x(7);
    y(5) += brel2 * x(13) * x(7);


    // jp = b p E
    y(8) += brel * x(0) * x(11);
    y(9) += brel * x(1) * x(11);
    y(8) += brel2 * x(15) * x(11);
    y(9) += brel2 * x(16) * x(11);


    // recombination
    double prod = ReduceToReal (x(7)*x(11));
    if (prod<0) prod = 0;
    y(7)  -= taurec * (prod-coef_nini);   // recombination
    y(11) -= taurec * (prod-coef_nini);
    // cout << "prod = " << prod << ", nini = " << nini << endl;


    /*
    y(4) /= n0;
    y(5) /= n0;    
    y(6) /= n0;
    y(7) /= n0;
    y(8) /= p0;
    y(9) /= p0;
    y(10) /= p0;
    y(11) /= p0;
    */

    y(12) = x(12);
    y(13) = x(13);
    y(14) = ut/2 * log (n0*p0);
    y(15) = x(15);
    y(16) = x(16);
    y(17) = ut/2 * log (p0*n0);
  }

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }

};



class MixedDDIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpDD, DDDMat, FiniteElement>
{
public:
  MixedDDIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpDD, DDDMat, FiniteElement>
  (DDDMat ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedDDIntegrator ();
  }

  virtual string Name () const { return "DD"; }
};






class DDNormDMat : public DMatOp<DDNormDMat>
{
public:

  enum { DIM_DMAT = 18 };

  DDNormDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    double coef_b = cf_mu -> Evaluate (sip);
    double coef_d = kb/q*temp * cf_mu -> Evaluate (sip);
    double coef_eps = cf_eps -> Evaluate (sip) * eps0;
    // double coef_ni = cf_ni -> Evaluate (sip);
    double coef_n0 = cf_n0 -> Evaluate (sip);
    double coef_p0 = cf_p0 -> Evaluate (sip);
    double coef_nini = sqr (cf_ni -> Evaluate (sip));
    double taurec = 1.0 / cf_taurec -> Evaluate (sip);

    mat = 0;
    mat(0,0) = mat(1,1) = 1.0 / coef_eps;
    // mat(2,2) = 1;
    mat(3,3) = 1e-10;

    mat(4,4) = mat(5,5) = 1e6;   // 1.0 / coef_d;
    //     mat(6,6) = 1;
    mat(7,7) = 1e2 + taurec * sqr (coef_nini);  // 1.0/dt;  

    mat(8,8) = mat(9,9) = 1e6;  // 1.0 / coef_d;
    // mat(9,9) = 1;
    mat(11,11) = 1e2 + taurec * sqr (coef_nini);  // 1.0/dt;  

    /*
    mat(4,4) /= coef_n0;
    mat(5,5) /= coef_n0;
    mat(6,6) /= coef_n0;
    mat(7,7) /= coef_n0;
    mat(8,8) /= coef_p0;
    mat(9,9) /= coef_p0;
    mat(10,10) /= coef_p0;
    mat(11,11) /= coef_p0;
    */

    mat(12,12) = 1;
    mat(13,13) = 1;
    mat(15,15) = 1;
    mat(16,16) = 1;
  }

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }
};



class MixedDDNormIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpDD, DDNormDMat, FiniteElement>
{
public:
  MixedDDNormIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpDD, DDNormDMat, FiniteElement>
  (DDNormDMat ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedDDNormIntegrator ();
  }

  virtual string Name () const { return "DD-Norm"; }
};










class DDrhsDMat : public DMatOp<DDrhsDMat>
{
public:

  enum { DIM_DMAT = 18 };
  enum { SYMMETRIC = 0 };

  DDrhsDMat () { ; }


  template <typename FEL, typename SIP, class TVX, class TVY>
  void Apply (const FEL & fel, const SIP & sip,
	      const TVX & x, TVY & y,
	      LocalHeap & lh) const
  {
    y = 0.0;
    y(7) = -1.0 * x(7);
    y(11) = -1.0 * x(11);

    double coef_ni = cf_ni -> Evaluate (sip);
    double coef_n0 = cf_n0 -> Evaluate (sip);
    double coef_p0 = cf_p0 -> Evaluate (sip);

    /*
    y(7) /= coef_n0;
    y(11) /= coef_p0;
    */
  }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;
    mat(7,7) = -1.0;
    mat(11,11) = -1.0;

    double coef_ni = cf_ni -> Evaluate (sip);
    double coef_n0 = cf_n0 -> Evaluate (sip);
    double coef_p0 = cf_p0 -> Evaluate (sip);
    /*
    mat(7,7) /= coef_n0;
    mat(11,11) /= coef_p0;
    */
  }  

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }
};

class MixedDDrhsIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpDD, DDrhsDMat, FiniteElement>
{
public:
  ///
  MixedDDrhsIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpDD, DDrhsDMat, FiniteElement>
  (DDrhsDMat ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedDDrhsIntegrator ();
  }
  
  ///
  virtual string Name () const { return "DDrhs"; }
};












class DDJumpsIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction *fac;
public:
  DDJumpsIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
    : BilinearFormIntegrator()
  { 
    fac = coeffs[0];
  }

  virtual ~DDJumpsIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & fel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & locheap) const
  {
    elmat = 0.0;
  }

  virtual void
  AssembleLinearizedElementMatrix (const FiniteElement & bfel, 
				   const ElementTransformation & eltrans, 
				   FlatVector<double> & elveclin,
				   FlatMatrix<double> & elmat,
				   LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0.0;

    /*
    FlatMatrix<> diff (ndof, lh);
    FlatVector<> hx(ndof, lh), hy1(ndof, lh), hy2(ndof, lh);
    (*testout) << "elveclin = " << elveclin << endl;
    for (int i = 0; i < ndof; i++)
      {
	double eps = (i < 3) ? 1e-14 : 1e-5;
	hx = elveclin;
	hx(i) += eps;
	ApplyElementMatrix (bfel, eltrans, hx, hy1, 0, lh);

	hx = elveclin;
	hx(i) -= eps;
	ApplyElementMatrix (bfel, eltrans, hx, hy2, 0, lh);
	elmat.Col(i) = 0.5/eps * (hy1 - hy2);

      }
    (*testout) << "elmat,num = " << endl << elmat << endl;
    elmat = 0.0;
    */


    // j - (bE) n - D nabla n

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_vec = fel_vec.GetNDof();
    int nd_scal = fel_scal.GetNDof();
    int nd = bfel.GetNDof();


    FlatMatrix<> grad(nd_scal, 2, lh);
    FlatMatrix<> grad_ref(nd_scal, 2, lh);
    FlatVector<> shape_scal(nd_scal, lh);

    FlatMatrix<> mat_vec(nd_vec, 2, lh);
    FlatMatrix<> mat_vec_ref(nd_vec, 2, lh);
    FlatVector<> mat_vec_n(nd_vec, lh);

    const POINT3D * points = ElementTopology::GetVertices (fel_vec.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_vec.ElementType());
    int ned = (fel_vec.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel_vec.Order());

    FlatVector<>  bvec(nd, lh), bvec2(nd, lh);

    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);

	    double coef_b = cf_mu -> Evaluate (sip);
	    double coef_d = kb/q*temp * cf_mu -> Evaluate (sip);
	    double coef_eps = cf_eps -> Evaluate (sip) * eps0;
	    // double coef_ni = cf_ni -> Evaluate (sip);
	    double coef_n0 = cf_n0 -> Evaluate (sip);
	    double coef_p0 = cf_p0 -> Evaluate (sip);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    fel_scal.CalcShape (ip, shape_scal);
	    fel_scal.CalcDShape (ip, grad_ref);
	    grad = grad_ref * inv_jac;

	    fel_vec.CalcShape (ip, mat_vec_ref);
	    mat_vec = mat_vec_ref * Trans (jac);
	    mat_vec /= det;
	    
	    mat_vec_n = mat_vec * normal;

	    Vec<1> E, jn, jp;
	    Vec<1> n, p;

	    E = -(1.0/coef_eps) * Trans (mat_vec_n) * elveclin.Range (0, nd_vec);
	    E += Trans (mat_vec_n) * elveclin.Range(3*nd_vec+3*nd_scal, 4*nd_vec+3*nd_scal);

	    if (E(0) > 0)
	      {
		bvec = 0.0;
		bvec2 = 0.0;
		
		bvec.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = E(0) * mat_vec_n;
		bvec.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal) = (-coef_b*E(0)*E(0)) * shape_scal;

		jn = Trans (mat_vec_n) * elveclin.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal);
		n = Trans (shape_scal) * elveclin.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal);
		bvec.Range (0, nd_vec) = (-jn(0)+coef_b*2*n(0)*E(0)) * mat_vec_n;

		bvec2.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = mat_vec_n;

		double factor = ir1d[l].Weight() * coef_b/coef_d  * pow (len, 3) * fac -> Evaluate (sip);
		bvec *= factor;
		
		elmat += bvec2 * Trans (bvec);
	      }

	    E = (1.0/coef_eps) * Trans (mat_vec_n) * elveclin.Range (0, nd_vec);
	    E += Trans (mat_vec_n) * elveclin.Range(4*nd_vec+3*nd_scal, 5*nd_vec+3*nd_scal);

	    if (E(0) > 0)
	      {
		bvec = 0.0;
		bvec2 = 0.0;
		
		bvec.Range (2*nd_vec+2*nd_scal, 3*nd_vec+2*nd_scal) = (E(0)) * mat_vec_n;
		bvec.Range (3*nd_vec+2*nd_scal, 3*nd_vec+3*nd_scal) = (-coef_b*E(0)*E(0)) * shape_scal;

		jp = Trans (mat_vec_n) * elveclin.Range (2*nd_vec+2*nd_scal, 3*nd_vec+2*nd_scal);
		p = Trans (shape_scal) * elveclin.Range (3*nd_vec+2*nd_scal, 3*nd_vec+3*nd_scal);
		bvec.Range (0, nd_vec) = (jp(0)-coef_b*2*p(0)*E(0)) * mat_vec_n;

		bvec2.Range (2*nd_vec+2*nd_scal, 3*nd_vec+2*nd_scal) = mat_vec_n;

		double factor = ir1d[l].Weight() * coef_b/coef_d * pow (len, 3) * fac -> Evaluate (sip);
		bvec *= factor;
		
		elmat += bvec2 * Trans (bvec);
	      }
	  
	    lh.CleanUp (heapp);
	  }
      }
    // (*testout) << "jumps, elmat exact = " << endl << elmat << endl;
  }







  virtual void 
  ApplyElementMatrix (const FiniteElement & bfel, 
		      const ElementTransformation & eltrans, 
		      const FlatVector<double> & elx, 
		      FlatVector<double> & ely,
		      void * precomputed,
		      LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();
    ely = 0.0;

    // j - (bE) n - D nabla n

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_vec = fel_vec.GetNDof();
    int nd_scal = fel_scal.GetNDof();
    int nd = bfel.GetNDof();

    FlatMatrix<> grad(nd_scal, 2, lh);
    FlatMatrix<> grad_ref(nd_scal, 2, lh);
    FlatVector<> shape_scal(nd_scal, lh);

    FlatMatrix<> mat_vec(nd_vec, 2, lh);
    FlatMatrix<> mat_vec_ref(nd_vec, 2, lh);
    FlatVector<> mat_vec_n(nd_vec, lh);

    const POINT3D * points = ElementTopology::GetVertices (fel_vec.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_vec.ElementType());
    int ned = (fel_vec.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel_vec.Order());

    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);

	    double coef_b = cf_mu -> Evaluate (sip);
	    double coef_d = kb/q*temp * cf_mu -> Evaluate (sip);
	    double coef_eps = cf_eps -> Evaluate (sip) * eps0;
	    double coef_n0 = cf_n0 -> Evaluate (sip);
	    double coef_p0 = cf_p0 -> Evaluate (sip);


	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    fel_scal.CalcShape (ip, shape_scal);
	    fel_scal.CalcDShape (ip, grad_ref);
	    grad = grad_ref * inv_jac;

	    fel_vec.CalcShape (ip, mat_vec_ref);
	    mat_vec = mat_vec_ref * Trans (jac);
	    mat_vec /= det;
	    
	    mat_vec_n = mat_vec * normal;

	    Vec<1> E, jn, jp;
	    Vec<1> n, p;

	    E = -(1.0/coef_eps) * Trans (mat_vec_n) * elx.Range (0, nd_vec);
	    E += Trans (mat_vec_n) * elx.Range(3*nd_vec+3*nd_scal, 4*nd_vec+3*nd_scal);
	    if (E(0) > 0)
	      {
		jn = Trans (mat_vec_n) * elx.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal);
		n = Trans (shape_scal) * elx.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal);
		
		double resn = E(0) * (jn(0) - coef_b * E(0)*n(0));
		double factor = ir1d[l].Weight() * coef_b/coef_d * pow (len,3) * fac -> Evaluate (sip);

		ely.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) += (factor*resn) * mat_vec_n;
	      }

	    E = (1.0/coef_eps) * Trans (mat_vec_n) * elx.Range (0, nd_vec);
	    E += Trans (mat_vec_n) * elx.Range(4*nd_vec+3*nd_scal, 5*nd_vec+3*nd_scal);
	    if (E(0) > 0)
	      {
		jp = Trans (mat_vec_n) * elx.Range (2*nd_vec+2*nd_scal, 3*nd_vec+2*nd_scal);
		p = Trans (shape_scal) * elx.Range (3*nd_vec+2*nd_scal, 3*nd_vec+3*nd_scal);
		
		double resp = (E(0)) * (jp(0) - coef_b * E(0)*p(0));
		double factor = ir1d[l].Weight() * coef_b/coef_d * pow(len,3) * fac -> Evaluate (sip);
		
		ely.Range (2*nd_vec+2*nd_scal, 3*nd_vec+2*nd_scal) += (resp*factor) * mat_vec_n;
	      }

	    lh.CleanUp (heapp);
	  }
      }
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new DDJumpsIntegrator (coeffs); 
  }
};




























/* ************************************************************ */
/* ***********************   DDSolver  ************************ */
/* ************************************************************ */

class NumProcMixedDD : public NumProc
{
protected:
  double te;
  bool reassemble;
  int maxnewton;
  double dt0;
public:

  NumProcMixedDD (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    dt0 = flags.GetNumFlag ("dt", 0.1);
    // taurec = flags.GetNumFlag ("taurec", 0.001);
    te = flags.GetNumFlag ("te", 1);


    cf_mu = pde.GetCoefficientFunction ("mu");
    cf_eps = pde.GetCoefficientFunction ("eps");
    cf_ni = pde.GetCoefficientFunction ("ni");
    cf_n0 = pde.GetCoefficientFunction ("n0");
    cf_p0 = pde.GetCoefficientFunction ("p0");
    cf_rho = pde.GetCoefficientFunction ("rho");
    cf_taurec = pde.GetCoefficientFunction ("taurec");
    /*
    coef_b = flags.GetNumFlag ("mu", 1);   // mobility [ m^2 / Vs ]
    coef_d = kb / q * 300 * flags.GetNumFlag ("mu", 1); 
    coef_eps = flags.GetNumFlag ("eps", 1) * 8.854e-12;

    cout << "coef_d = " << coef_d << endl;
    cout << "coef_b = " << coef_b << endl;
    cout << "coef_eps = " << coef_eps << endl;
    */

    maxnewton = int (flags.GetNumFlag ("maxnewton", 5));
    reassemble = 1;
  }

  virtual ~NumProcMixedDD()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedDD (pde, flags);
  }




  virtual void Do(LocalHeap & lh)
  {
    cout << "mixed drift diffusion solver" << endl;

    dt = dt0;

    const BaseMatrix & mata = pde.GetBilinearForm("a")->GetMatrix();
    const BaseMatrix & matarhs = pde.GetBilinearForm("arhs")->GetMatrix();

    BilinearForm & bfa = *pde.GetBilinearForm ("a");
    BilinearFormApplication applya(&bfa);

    GridFunction * gfu = pde.GetGridFunction("u");
    BaseVector & vecu = gfu->GetVector();

    GridFunction * gfdudt = pde.GetGridFunction("w");
    BaseVector & vecdudt = gfdudt->GetVector();

    BilinearForm & bfnorm = *pde.GetBilinearForm ("norm");
    const BaseSparseMatrix & matnorm =
      dynamic_cast<const BaseSparseMatrix&> (bfnorm.GetMatrix());

     BaseMatrix & invnorm = *matnorm.InverseMatrix();
     // BaseMatrix & invnorm = *matnorm.CreateJacobiPrecond();

    const BaseVector & vecf = pde.GetLinearForm("f")->GetVector();




    

    const BaseMatrix & matid = pde.GetBilinearForm("id")->GetMatrix();
    const BaseVector & vecinitial = pde.GetLinearForm("initial")->GetVector();
    BaseMatrix * invid = dynamic_cast<const BaseSparseMatrix&> (matid) . InverseMatrix();
    vecu = (*invid) * vecinitial;


    BaseVector & d = *vecu.CreateVector();
    BaseVector & d2 = *vecu.CreateVector();
    BaseVector & d3 = *vecu.CreateVector();
    BaseVector & uold = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();
    BaseVector & w2 = *vecu.CreateVector();

    BaseMatrix * invmat = 0;

    if (!reassemble)
      invmat = dynamic_cast<const BaseSparseMatrix&> (mata) . InverseMatrix();

    int ndvec = ma.GetNEdges();
    int ndscal = ma.GetNE();


    ofstream newtonout ("newton.out");
    for (double t = 0; t <= te; )
      {
	cout << "                                t = " << t << ", dt = " << dt << endl;

	uold = vecu;

	int it = 0;
	for (it = 0; it < maxnewton; it++)
	  {
	    if (reassemble)
	      bfa.AssembleLinearization (vecu, lh);
	    BaseMatrix & mata2 = pde.GetBilinearForm("a")->GetMatrix();
	    if (reassemble)
	      {
		delete invmat;
		invmat = dynamic_cast<const BaseSparseMatrix&> (mata2) . InverseMatrix();
	      }


	    d = (1.0/dt) * (matarhs * uold) + vecf;
	    d -= applya * vecu;

	    w = (*invmat) * d;
	    vecu += w;

	    w = invnorm * d;
	    double err = sqrt (InnerProduct (w, d));

	    cout << "                                it = " << it << ", err = " << err << endl;
	    
	    TempVector d1 = d.Range (0, ndvec);
	    TempVector d2 = d.Range (ndvec, ndvec+ndscal);
	    TempVector d3 = d.Range (ndvec+ndscal, 2*ndvec+ndscal);
	    TempVector d4 = d.Range (2*ndvec+ndscal, 2*ndvec+2*ndscal);
	    TempVector d5 = d.Range (2*ndvec+2*ndscal, 3*ndvec+2*ndscal);
	    TempVector d6 = d.Range (3*ndvec+2*ndscal, 3*ndvec+3*ndscal);

	    TempVector w1 = w.Range (0, ndvec);
	    TempVector w2 = w.Range (ndvec, ndvec+ndscal);
	    TempVector w3 = w.Range (ndvec+ndscal, 2*ndvec+ndscal);
	    TempVector w4 = w.Range (2*ndvec+ndscal, 2*ndvec+2*ndscal);
	    TempVector w5 = w.Range (2*ndvec+2*ndscal, 3*ndvec+2*ndscal);
	    TempVector w6 = w.Range (3*ndvec+2*ndscal, 3*ndvec+3*ndscal);

	    newtonout << "t = " << t << " dt = " << dt 
		      << ", newton = " << it << ", err = " << err
		      << ", err1 = " << InnerProduct (*d1, *w1)
		      << ", err2 = " << InnerProduct (*d2, *w2)
		      << ", err3 = " << InnerProduct (*d3, *w3)
		      << ", err4 = " << InnerProduct (*d4, *w4)
		      << ", err5 = " << InnerProduct (*d5, *w5)
		      << ", err6 = " << InnerProduct (*d6, *w6)
		      << endl;

	    if (err > 1e20)
	      {
		it = 100;
		break;
	      }
	    if (err < 1e-5) break;
	  }

	if (it >= maxnewton)
	  {
	    vecu = uold;
	    dt *= 0.3;
	    if (dt < 1e-50)
	      {
		cerr << "Time step too small" << endl;
		break;
	      }
	  }
	else
	  {
	    vecdudt = (1.0/dt) * (vecu - uold);
	    t += dt;
	    dt *= 1.5;
	  }

	cout << "redraw " << flush;
	Ng_Redraw ();	  
	cout << " done" << endl;
      }

    delete invmat;



    /*
    ofstream newtonout ("newton.out");
    for (double t = 0; t <= te; )
      {
	cout << "t = " << t << ", dt = " << dt << endl;

	uold = vecu;

	int it = 0;
	//	for (it = 0; it < maxnewton; it++)
	  {

	    if (reassemble)
	      {
		bfa.AssembleLinearization (vecu, lh);
		BaseMatrix & mata2 = pde.GetBilinearForm("a")->GetMatrix();
		delete invmat;
		invmat = dynamic_cast<const BaseSparseMatrix&> (mata2) . InverseMatrix();
	      }


	    d = (1.0/dt) * (matarhs * uold) + vecf;
	    d -= applya * vecu;

	    cout << "                                it = " << it << ", err = " << L2Norm (d) << endl;
	    
	    w = (*invmat) * d;
	    vecu += w;

	    if (L2Norm (d) > 1e20)
	      {
		it = 100;
		break;
	      }
	    if (L2Norm (d) < 1e-6) break;
	  }

	    vecdudt = (1.0/dt) * (vecu - uold);
	    t += dt;
	    // dt *= 1.2;

	Ng_Redraw ();	  
      }

    delete invmat;
*/







    
    
    
    cout << "ZZ error-estimator" << endl;
    
    ConstantCoefficientFunction cf_one(1);
    MassHDivIntegrator<2> masshdiv (&cf_one);
    MassIntegrator<2> mass (&cf_one);

    CompoundBilinearFormIntegrator massn(mass, 3);
    CompoundBilinearFormIntegrator massp(mass, 5);

    Flags fesflags;
    fesflags.SetFlag ("order", 1);

    FESpace * fesflux = 
      new H1HighOrderFESpace (ma, fesflags);

    fesflux -> Update();

    Flags flags;
    GridFunction * flux = CreateGridFunction (fesflux, "smoothn", flags);
    flux->Update();

    GridFunction * gferr = pde.GetGridFunction ("err");

    FlatVector<double> err = 
      dynamic_cast<T_BaseVector<double>&> (gferr->GetVector()).FV();

    err = 0;
  
    int ndom = ma.GetNDomains();
  
    CalcFluxProject (ma, 
		     dynamic_cast<const S_GridFunction<double>&> (*gfu), 
		     dynamic_cast<S_GridFunction<double>&> (*flux), 
		     massn,
		     1, -1, lh);
	  
    CalcError (ma, 
	       dynamic_cast<const S_GridFunction<double>&> (*gfu), 
	       dynamic_cast<const S_GridFunction<double>&> (*flux), 
	       massn,
	       err, -1, lh);
    
    CalcFluxProject (ma, 
		     dynamic_cast<const S_GridFunction<double>&> (*gfu), 
		     dynamic_cast<S_GridFunction<double>&> (*flux), 
		     massp,
		     1, -1, lh);
	  
    CalcError (ma, 
	       dynamic_cast<const S_GridFunction<double>&> (*gfu), 
	       dynamic_cast<const S_GridFunction<double>&> (*flux), 
	       massp,
	       err, -1, lh);

  }


  ///
  virtual string GetClassName () const
  {
    return "DD";
  }

  virtual void PrintReport (ostream & ost)
  { ; }
};







namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("mixedDD", NumProcMixedDD::Create);
    GetFESpaceClasses().AddFESpace ("mixedDD", FESpaceMixedDD::Create);
  
    GetIntegrators().AddBFIntegrator ("mixedDD", 2, 0,
				      MixedDDIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedDD_norm", 2, 0,
				      MixedDDNormIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedDDrhs", 2, 0,
				      MixedDDrhsIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedDDjumps", 2, 1,
				      DDJumpsIntegrator::Create);
  }
  Init init;
}
