/*

diffusion reaction equation in mixed form:


a p =  nabla u + b u 
div p = f

*/


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


/* *********************************************************** */
/* ************************* FESpace ************************* */
/* *********************************************************** */

class FESpaceMixed : public CompoundFESpace
{
public:
  FESpaceMixed (const MeshAccess & ama, 		   
		const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  { ; }

  virtual ~FESpaceMixed () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(2);

    Flags hdivflags, l2flags;

    int order = int (flags.GetNumFlag ("order", 1));

    l2flags.SetFlag ("order", max (order-1, 0));
    hdivflags.SetFlag ("order", order);
    
    if (order == 1)
      {
	spaces[0] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[1] = new ElementFESpace (ma, 0, 0, 0);    
      }
    else
      {
	spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
      }

    FESpaceMixed * fes = new FESpaceMixed (ma, spaces);
    return fes;
  }
    
  virtual string GetClassName () const { return "Mixed - FESpace"; }
};


/* ************************************************************ */
/* ************************ Integrator ************************ */
/* ************************************************************ */

class DiffOpMixed : public DiffOp<DiffOpMixed>
{
  // 6 components:
  // px, py, div p, u, dudx, dudy

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 6 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();

    FlatVector<> shape_scal(nd_scal, lh);
    shape_scal = fel_scal.GetShape (sip.IP(), lh);

    FlatMatrix<> grad(2, nd_scal, lh);
    grad = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_scal.GetDShape(sip.IP(), lh));


    FlatMatrix<> shape_vec(2, nd_vec, lh);
    shape_vec = (1.0/sip.GetJacobiDet()) * 
      (sip.GetJacobian() * Trans (fel_vec.GetShape(sip.IP(), lh)));

    FlatVector<> shape_div(nd_vec, lh);
    shape_div = 1.0/sip.GetJacobiDet() * fel_vec.GetDivShape(sip.IP(), lh);


    mat = 0;
    for (int j = 0; j < nd_vec; j++)
      {
	mat(0,j) = shape_vec(0, j);
	mat(1,j) = shape_vec(1, j);
	mat(2,j) = shape_div(j);

	// mat(4,j) = -shape_vec(0, j);
	// mat(5,j) = -shape_vec(1, j);
      }

    for (int j = 0; j < nd_scal; j++)
      {
	mat(3,nd_vec+j) = shape_scal(j);
	mat(4,nd_vec+j) = grad(0,j); //  - bx * shape_scal(j);
	mat(5,nd_vec+j) = grad(1,j); //  - by * shape_scal(j);
      }
  }
};



class DMatMixed : public DMatOp<DMatMixed>
{
  CoefficientFunction *bx, *by, *stab;
public:

  enum { DIM_DMAT = 6 };
  enum { SYMMETRIC = 0 };

  DMatMixed (ARRAY<CoefficientFunction*> & coeffs) 
  { 
    bx = coeffs[0];
    by = coeffs[1];
    stab = coeffs[2];
  }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    double cbx = bx -> Evaluate (sip);
    double cby = by -> Evaluate (sip);
    double cstab = stab -> Evaluate (sip);
    mat = 0;
    mat(0,0) = mat(1,1) = 1;
    mat(2,3) = mat(3,2) = 1;
    mat(0,3) = cbx;
    mat(1,3) = cby;
    mat(3,3) = -1e-5;

    /*
    // stabilization
    // - | p - nabla u - b u |^2
    for (int j = 0; j < 2; j++)
      {
	Vec<6> v;
	v = 0.0;
	v(j) = 1;
	v(3) = (j == 0) ? cbx : cby;
	v(4+j) = -1;
	
	mat -= cstab * (v * Trans (v));
      }
    */
  }  
};


typedef T_BDBIntegrator<DiffOpMixed, DMatMixed> MixedIntegrator;

















class MixedJumpsIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction *bx, *by, *fac;
public:
  MixedJumpsIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
    : BilinearFormIntegrator()
  { 
    bx = coeffs[0];
    by = coeffs[1];
    fac = coeffs[2];
  }

  virtual ~MixedJumpsIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }



  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;
    
    // (p - nabla u + b u).n

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_p = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_u = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_p = fel_p.GetNDof();
    int nd_u = fel_u.GetNDof();
    int nd = bfel.GetNDof();


    FlatMatrix<> gradu(nd_u, 2, lh);
    FlatMatrix<> gradu_ref(nd_u, 2, lh);
    FlatVector<> vecu(nd_u, lh);

    FlatMatrix<> matp(nd_p, 2, lh);
    FlatMatrix<> matp_ref(nd_p, 2, lh);

    const POINT3D * points = ElementTopology::GetVertices (fel_p.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_p.ElementType());
    int ned = (fel_p.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel_p.Order());

    FlatVector<>  bvec(nd, lh), bvec2(nd, lh);

    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    fel_u.CalcShape (ip, vecu);
	    fel_u.CalcDShape (ip, gradu_ref);
	    gradu = gradu_ref * inv_jac;

	    fel_p.CalcShape (ip, matp_ref);
	    matp = matp_ref * Trans (jac);
	    matp /= det;

	    Vec<2> b;
	    b(0) = bx -> Evaluate (sip);
	    b(1) = by -> Evaluate (sip);
	    double factor = ir1d[l].Weight() * L2Norm(b) * pow (len, 3) * fac -> Evaluate (sip);

	    bvec = 0.0;
	    if (InnerProduct (b, normal) > 0)
	      {
		for (int i = 0; i < nd_p; i++)
		  for (int j = 0; j < 2; j++)
		    bvec(i) += normal(j) * matp(i,j);

		for (int i = 0; i < nd_u; i++)
		  for (int j = 0; j < 2; j++)
		    bvec(nd_p+i) += normal(j) * (b(j)*vecu(i)-gradu(i,j));

		bvec2 = bvec;
		// bvec2.Range(nd_p, nd) = -bvec.Range(nd_p, nd);
		bvec2.Range(nd_p, nd) = 0.0;

		/*
		(*testout) << "ip = " << ip << endl;
		(*testout) << "bvec = " << bvec << endl;
		(*testout) << "factor = " << factor << endl;
		*/
		elmat += factor * (bvec2 * Trans (bvec));
	      }
	  
	    lh.CleanUp (heapp);
	  }
      }
    // (*testout) << "elmat = " << endl << elmat << endl;
  }


  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedJumpsIntegrator (coeffs); 
  }

};













namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("mixed", FESpaceMixed::Create);
    GetIntegrators().AddBFIntegrator ("mixed", 2, 3, MixedIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedjumps", 2, 3, MixedJumpsIntegrator::Create);
  }
  Init init;
}
