/*

Helmholtz discretization based on wave factorization

*/


#include <solve.hpp>

extern ngsolve::PDE * pde;


/*
  namespace ngfem
  {
  #include "../fem/pml.hpp"
  }
*/



using namespace ngfem;
using namespace ngsolve;




class FE_Quad1_Aniso : public NodalFiniteElement
{
  static IPDataArray ipdata;

public:
  FE_Quad1_Aniso()
    : NodalFiniteElement (2, ET_QUAD, 2, 1)
  {
    if (!ipdata.Size())
      CalcIPData(ET_QUAD, ipdata);
  }
  virtual ~FE_Quad1_Aniso()
  { ; }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const
  {
    double x = ip.Point()[0];
    double y = ip.Point()[1];
    shape(0) = x;
    shape(1) = 1-x;
  }

  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const
  {
    double x = ip.Point()[0];
    double y = ip.Point()[1];

    dshape(0,0) = 1;
    dshape(1,0) = -1;
  }
}; 

NodalFiniteElement::IPDataArray FE_Quad1_Aniso::ipdata;






class H1HighOrder_Quad_Aniso : public H1HighOrderFiniteElement
{
  typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

public:
  H1HighOrder_Quad_Aniso (int aorder)
    : H1HighOrderFiniteElement(2, ET_QUAD)
  {
    order_inner = INT<3> (aorder,aorder,aorder);
    for (int i = 0; i < 4; i++)
      order_edge[i] = aorder;
    ComputeNDof();
  }    

  virtual void ComputeNDof()
  {
    ndof = order_inner[0] * (order_inner[0]+1);
    order = order_inner[0]+1;
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
  }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const
  {
    double x = ip(0);
    double y = ip(1);

    int p = order_inner[0];
    ArrayMem<double,20> polx(p+1), poly(p);
    LegendrePolynomial (p-1, y, poly);
    IntegratedLegendrePolynomial (p, x, polx);

    int ii = 0;
    for (int i = 0; i < p; i++)
      shape(ii++) = poly[i] * x;

    for (int i = 0; i < p; i++)
      shape(ii++) = poly[i] * (1-x);

    for (int i = 0; i < p; i++)
      for (int j = 2; j <= p; j++)
	shape(ii++) = polx[j] * poly[i];
  }
};










class DummySegm : public FiniteElement
{
public:
  DummySegm ()
  {
    eltype = ET_SEGM;
    ndof = 0; 
  }
};







class DualResonanceFESpace : public FESpace
{
  ARRAY<int> vert_dofs;
  ARRAY<int> edge_dofs;
  ARRAY<int> cell_dofs;
  int nd;

public:

  DualResonanceFESpace (const MeshAccess & ama, const Flags & aflags)
    : FESpace (ama, aflags)
  {
    trig    = 0; // new FE_Trig1;
    quad    = new FE_Quad1_Aniso;
    segm    = new DummySegm;
  }

  virtual ~DualResonanceFESpace ()
  {
    ;
  }

  virtual string GetClassName () const
  {
    return "DualResonanceFESpace";
  }

  virtual void Update()
  {
    int nv = ma.GetNV();
    int ne = ma.GetNE();
    int ned = ma.GetNEdges();

    ArrayMem<int,4> vnums, enums;

    vert_dofs.SetSize(nv+1);
    edge_dofs.SetSize(ned+1);
    cell_dofs.SetSize(ne+1);

    vert_dofs = -1;
    edge_dofs = -1;
    cell_dofs = -1;

    for (int i = 0; i < ne; i++)
      {
	ma.GetElVertices (i, vnums);
	ma.GetElEdges (i, enums);

	if (vnums.Size() == 3)
	  {
	    for (int j = 0; j < 3; j++)
	      vert_dofs[vnums[j]] = 1;
	    for (int j = 0; j < 3; j++)
	      edge_dofs[enums[j]] = order-1;
	    cell_dofs[i] = (order-1)*(order-2)/2;
	  }
	else
	  {
	    edge_dofs[enums[3]] = order;
	    edge_dofs[enums[2]] = order;
	    cell_dofs[i] = order*(order-1);
	  }
      }

    nd = 0;
    for (int i = 0; i < nv; i++)
      if (vert_dofs[i] != -1)
	{
	  vert_dofs[i] = nd;
	  nd++;
	}
    vert_dofs[nv] = nd;
    for (int i = 0; i < ned; i++)
      {
	int ndi = (edge_dofs[i] != -1) ? edge_dofs[i] : 0;
	edge_dofs[i] = nd;
	nd+=ndi;
      }
    edge_dofs[ned] = nd;
    for (int i = 0; i < ne; i++)
      {
	int ndi = (cell_dofs[i] != -1) ? cell_dofs[i] : 0;
	cell_dofs[i] = nd;
	nd+=ndi;
      }
    cell_dofs[ne] = nd;

    (*testout) << "edge_dofs = " << endl << edge_dofs << endl;
    (*testout) << "cell_dofs = " << endl << cell_dofs << endl;
    cout << "update dual, nd = " << nd << endl;
  }  

  virtual int GetNDof () const
  {
    return nd;
  }

  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const
  {
    H1HighOrderFiniteElement * hofe;

    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
    
    switch (ma.GetElType(elnr))
      {
      case ET_TRIG:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrderTrig<T_ORTHOPOL>)))  H1HighOrderTrig<T_ORTHOPOL> (order);
	  break;
	}
      case ET_QUAD:
	{
	  hofe = new (lh.Alloc (sizeof(H1HighOrder_Quad_Aniso)))  H1HighOrder_Quad_Aniso (order);
	  hofe -> SetOrderInner (order);
	  return *hofe;
	}
      default:
	hofe = 0; 
      }

    if (!hofe)
      {
	stringstream str;
	str << "H1HighOrderFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }
    
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ArrayMem<int, 12> ednums, order_ed;
    ArrayMem<INT<2>, 6> order_fa;
    
    ma.GetElVertices(elnr, vnums);
    ma.GetElEdges(elnr, ednums);
    
    order_ed.SetSize (ednums.Size());
    for (int j = 0; j < ednums.Size(); j++)
      order_ed[j] = order;

    hofe -> SetVertexNumbers (vnums, lh);
    hofe -> SetOrderEdge (order_ed);
    
    ArrayMem<int, 8> order_vert;
    for (int j = 0; j < vnums.Size(); j++)
      order_vert[j] = order;
    hofe -> SetOrderVertex (order_vert);
    hofe -> SetOrderInner (order);
    
    hofe -> ComputeNDof();
    return *hofe;
  }


  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    ArrayMem<int,4> vnums, enums;
    ma.GetElVertices (elnr, vnums);
    ma.GetElEdges (elnr, enums);
    if (vnums.Size() == 3)
      {
	dnums.SetSize(0);
	for (int j = 0; j < 3; j++)
	  dnums.Append (vert_dofs[vnums[j]]);
	for (int j = 0; j < 3; j++)
	  for (int k = edge_dofs[enums[j]]; k < edge_dofs[enums[j]+1]; k++)
	    dnums.Append (k);
	for (int k = cell_dofs[elnr]; k < cell_dofs[elnr+1]; k++)
	  dnums.Append (k);
      }
    else
      {
	dnums.SetSize(0);
	//dnums[0] = edge_dofs[enums[3]];
	//dnums[1] = edge_dofs[enums[2]];

	for (int k = edge_dofs[enums[3]]; k < edge_dofs[enums[3]+1]; k++)
	  dnums.Append (k);
	for (int k = edge_dofs[enums[2]]; k < edge_dofs[enums[2]+1]; k++)
	  dnums.Append (k);
	for (int k = cell_dofs[elnr]; k < cell_dofs[elnr+1]; k++)
	  dnums.Append (k);
      }

    (*testout) << "elnr" << elnr << endl;
    (*testout) << "enums" << enums << endl;
    (*testout) << "dnums = " << dnums << endl;
  }

  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
};






class FESpaceMixedResonance : public CompoundFESpace
{
public:
  FESpaceMixedResonance (const MeshAccess & ama, 		   
			 const ARRAY<const FESpace*> & aspaces,
			 const Flags & aflags)
    : CompoundFESpace (ama, aspaces, aflags)
  {
    ;
  }
  
  virtual ~FESpaceMixedResonance () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(2);

    cout << "Create Mixed Res Space" << endl;
    
    int order = int (flags.GetNumFlag ("order", 1.0));
    bool is_complex = flags.GetDefineFlag ("complex");

    Flags uflags;
    uflags.SetFlag ("order", order);
    if (is_complex)
      uflags.SetFlag ("complex");
    spaces[0] = new H1HighOrderFESpace (ma, uflags);
    
    Flags pflags;
    pflags.SetFlag ("order", order);
    if (is_complex)
      pflags.SetFlag ("complex");
    
    if (flags.GetDefineFlag ("reduced"))
      spaces[1] = new DualResonanceFESpace (ma, uflags);
    else
      spaces[1] = new H1HighOrderFESpace (ma, uflags);
    /*
    if (order > 1)
      spaces[1] = new L2HighOrderFESpace (ma, pflags);
    else
      spaces[1] = new ElementFESpace (ma, 0, 1, is_complex);
    */
    FESpaceMixedResonance * fes = new FESpaceMixedResonance (ma, spaces, flags);
    return fes;
  }
    
  virtual string GetClassName () const { return "MixedResonance-FESpace"; }
};









/* ********************+ Integrator ************************** */

class DiffOpMixedResonance : public DiffOp<DiffOpMixedResonance>
{
  // 2 components:
  // u, v

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 2 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    typedef typename MAT::TSCAL TSCAL;
    
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_u = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_p = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_u = fel_u.GetNDof();
    int nd_p = fel_p.GetNDof();
    int nd = nd_u + nd_p;
    
    FlatVector<> vu(nd_u, lh);
    FlatVector<> vp(nd_p, lh);

    fel_u.CalcShape (sip.IP(), vu);
    fel_p.CalcShape (sip.IP(), vp);
    
    mat = TSCAL (0.0);
    for (int j = 0; j < nd_u; j++)
      mat(0,j) = vu(j);
    for (int j = 0; j < nd_p; j++)
      mat(1,j+nd_u) = vp(j);
  }
};


class MixedResonanceDMat : public DMatOp<MixedResonanceDMat>
{
public:

  enum { DIM_DMAT = 2 };
  //   enum { SYMMETRIC = 0 };

  MixedResonanceDMat () { ; }

  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat(0,0) = 0;
    mat(1,0) = 1;
    mat(0,1) = 1;
    mat(1,1) = 0;
  }  
};


class MixedResonanceIntegrator 
  : public T_BDBIntegrator<DiffOpMixedResonance, MixedResonanceDMat, FiniteElement>
{
public:
  ///
  MixedResonanceIntegrator ()
    :  T_BDBIntegrator<DiffOpMixedResonance, MixedResonanceDMat, FiniteElement>
  (MixedResonanceDMat ())
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedResonanceIntegrator ();
  }
  
  ///
  virtual string Name () const { return "MixedResonance"; }
};




class PML_MixedResonanceIntegrator 
  : public PML_BDBIntegrator<DiffOpMixedResonance, MixedResonanceDMat, FiniteElement>
{
public:
  ///
  PML_MixedResonanceIntegrator ()
    :  PML_BDBIntegrator<DiffOpMixedResonance, MixedResonanceDMat, FiniteElement>
  (MixedResonanceDMat ())
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new PML_MixedResonanceIntegrator ();
  }
  
  ///
  virtual string Name () const { return "PML_MixedResonance"; }
};










///
class NumProcMixedResonance : public NumProc
{
protected:
  int num;
public:
  ///
  NumProcMixedResonance (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    num = int (flags.GetNumFlag ("num", 300)); 
  }


  ///
  virtual ~NumProcMixedResonance()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedResonance (pde, flags);
  }



  ///
  virtual void Do()
  {
    cout << "solve helmholtz with wave factorization" << endl;

    if (!pde.GetBilinearForm("a")->GetFESpace().IsComplex())
      {
	SparseMatrix<double> & mata =
	  dynamic_cast<SparseMatrix<double> & >
	  (pde.GetBilinearForm ("a")->GetMatrix());


	SparseMatrix<double> & matb =
	  dynamic_cast<SparseMatrix<double> & >
	  (pde.GetBilinearForm ("b")->GetMatrix());


	GridFunction & gfu = *pde.GetGridFunction ("u");


	// (*testout) << "mata = " << endl << mata << endl;
	// (*testout) << "matb = " << endl << matb << endl;


	int n = mata.Height();


	// Arnoldi solver


	VVector<double> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);


	PardisoInverse<double> inva(mata, 0, 0, 1);
	// SparseCholesky<double> inva(mata);


	// Arnoldi:
	int m = min2 (num, n);
    
	Matrix<double> matV (m,n); 
	Matrix<Complex> matH(m), matHt(m);


	//	for (int i = 0; i < n; i++)
	// hv.FVDouble()(i) = sin (i);

	hv.SetRandom();


	matV = 0.0;
	matH = Complex(0.0);
    
	double len = sqrt (S_InnerProduct<double> (hv, hv));
	hv /= len;

	for (int i = 0; i < m; i++)
	  {
	    cout << "i = " << i << endl;
	    for (int j = 0; j < n; j++)
	      matV(i,j) = hv(j);
	
	    hva = matb * hv;
	    hvm = inva * hva;

	    for (int j = 0; j <= i; j++)
	      {
		double sum = 0.0;
		for (int k = 0; k < n; k++)
		  sum += hvm(k) * matV(j,k);
		matH(j,i) = sum;
		for (int k = 0; k < n; k++)
		  hvm(k) -= sum * matV(j,k);
	      }
	
	    hv = hvm;
	    double len = sqrt (S_InnerProduct<double> (hv, hv));
	    if (i<m-1)
	      matH(i+1,i) = len; 
	
	    //  cout << ", len = " << len << endl;
	    hv /= len;
	  }
    

	(*testout) << "hessenberg = " << endl << matH << endl;
	cout << "has Hessenberg" << endl;
    
	Vector<Complex> lami(m);
	Matrix<Complex> evecs(m);
	// Matrix<Complex> levecs(n,m);
	matHt = Trans (matH);
    
	evecs = Complex (0.0);
	lami = Complex (0.0);
    
    
	LapackHessenbergEP (matH.Height(), &matHt(0,0), &lami(0), &evecs(0,0));
    
	cout << "hessenberg is back" << endl;

	for (int i = 0; i < m; i++)
	  lami(i) =  1.0 / lami(i);

	ofstream evout("eigenvalues.out");
	for (int i = 0; i < m; i++)
	  {
	    evout << i << " " << lami(i).real() << " " << lami(i).imag();
	    Complex inv = 1.0 / lami(i);
	    evout << " " << inv.real() << " " << inv.imag() << endl;
	  }



	Vector<Complex> ev(m);
	Vector<Complex> lev(n);

	int cnt = 0;
	for (int i = 0; i < m; i++)
	  {
	    cout << "compute eigenvector " << i << endl;

	    for (int j = 0; j < m; j++)
	      ev(j) = evecs(i,j);

	    lev = Trans (matV) * ev;

	    FlatVector<Complex> vecu = gfu.GetVector(cnt).FVComplex();
	    vecu = lev;
	
	    cnt++;
	    if (cnt >= gfu.GetMultiDim()) break;
	  }
      }
    else
      {
	SparseMatrix<Complex> & mata =
	  dynamic_cast<SparseMatrix<Complex> & >
	  (pde.GetBilinearForm ("a")->GetMatrix());


	SparseMatrix<Complex> & matb =
	  dynamic_cast<SparseMatrix<Complex> & >
	  (pde.GetBilinearForm ("b")->GetMatrix());


	GridFunction & gfu = *pde.GetGridFunction ("u");


	// (*testout) << "mata = " << endl << mata << endl;
	// (*testout) << "matb = " << endl << matb << endl;


	int n = mata.Height();


	// Arnoldi solver


	VVector<Complex> hv(n), hva(n), hvm(n), hwa(n), hwm(n), w(n);
	VVector<double> hvr(n);

	// PardisoInverse<Complex> inva(mata, 0, 0, 1);
	SparseCholesky<Complex> inva(mata);


	// Arnoldi:
	int m = min2 (num, n);
    
	Matrix<Complex> matV (m,n); 
	Matrix<Complex> matH(m), matHt(m);


	hv.SetRandom();

	// hvr.SetRandom();
	//	hv.FVComplex() = hvr.FVDouble();

	matV = Complex(0.0);
	matH = Complex(0.0);
    
	Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
	hv /= len;


	for (int i = 0; i < m; i++)
	  {
	    cout << "i = " << i << endl;
	    for (int j = 0; j < n; j++)
	      matV(i,j) = hv(j);
	
	    hva = matb * hv;
	    hvm = inva * hva;
	
	    // (*testout) << "hvm (" << i << ") = " << endl << hvm << endl;

	    for (int j = 0; j <= i; j++)
	      {
		Complex sum = 0.0;
		for (int k = 0; k < n; k++)
		  sum += hvm(k) * matV(j,k);
		matH(j,i) = sum;
		for (int k = 0; k < n; k++)
		  hvm(k) -= sum * matV(j,k);
	      }
	
	    hv = hvm;
	    Complex len = sqrt (S_InnerProduct<Complex> (hv, hv));
	    if (i<m-1)
	      matH(i+1,i) = len; 
	
	    //  cout << ", len = " << len << endl;
	    hv /= len;
	  }
    

	(*testout) << "hessenberg = " << endl << matH << endl;
	cout << "has Hessenberg" << endl;
    
	Vector<Complex> lami(m);
	Matrix<Complex> evecs(m);
	// Matrix<Complex> levecs(n,m);
	matHt = Trans (matH);
    
	evecs = Complex (0.0);
	lami = Complex (0.0);
    
	LapackHessenbergEP (matH.Height(), &matHt(0,0), &lami(0), &evecs(0,0));
    
	cout << "hessenberg is back" << endl;

	for (int i = 0; i < m; i++)
	  lami(i) =  1.0 / lami(i);  // shift

	ofstream evout("eigenvalues.out");
	for (int i = 0; i < m; i++)
	  {
	    evout << i << " " << lami(i).real() << " " << lami(i).imag();
	    Complex inv = 1.0 / lami(i);
	    evout << " " << inv.real() << " " << inv.imag();
	    inv = sqrt (1.0 / lami(i));
	    evout << " " << inv.real() << " " << inv.imag() << endl;
	  }


	Vector<Complex> ev(m);
	Vector<Complex> lev(n);

	int cnt = 0;
	for (int i = 0; i < m; i++)
	  {
	    cout << "compute eigenvector " << i << endl;

	    for (int j = 0; j < m; j++)
	      ev(j) = evecs(i,j);

	    lev = Trans (matV) * ev;

	    FlatVector<Complex> vecu = gfu.GetVector(cnt).FVComplex();
	    vecu = lev;
	
	    cnt++;
	    if (cnt >= gfu.GetMultiDim()) break;
	  }
      }

    cout << "eigensystem done" << endl;
    
  }


};







///
class NumProcMixedResonanceBVP : public NumProc
{
protected:
  double taur, taui;
  int maxsteps;
public:
  ///
  NumProcMixedResonanceBVP (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    taur = (flags.GetNumFlag ("taur", 0.1)); 
    taui = (flags.GetNumFlag ("taui", 0.1)); 
    maxsteps = int(flags.GetNumFlag ("maxsteps", 1000)); 
  }


  ///
  virtual ~NumProcMixedResonanceBVP ()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedResonanceBVP (pde, flags);
  }



  ///
  virtual void Do()
  {
    cout << "solve helmholtz with wave factorization" << endl;


    SparseMatrix<Complex> & mat =
      dynamic_cast<SparseMatrix<Complex> & >
      (pde.GetBilinearForm ("a")->GetMatrix());

    const BaseMatrix & pre =
      pde.GetPreconditioner ("c")->GetMatrix();


    BaseVector & vecf = pde.GetLinearForm ("f")->GetVector();
    BaseVector & vecfone = pde.GetLinearForm ("fone")->GetVector();
    BaseVector & vecu = pde.GetGridFunction ("u")->GetVector();

    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();
    BaseVector & wc = *vecu.CreateVector();
    BaseVector & uone = *vecu.CreateVector();
    BaseVector & aw = *vecu.CreateVector();
    BaseVector & aone = *vecu.CreateVector();

    int nv = ma.GetNV();
    uone = 0.0;
    for (int i = 0; i < nv; i++)
      uone.FVComplex()(i) = 1.0;

    aone = mat * uone;

    Mat<2,2, Complex> ms, invms;
    Vec<2, Complex> fs, us;

    FlatVector<Complex> fw = w.FVComplex();
    FlatVector<Complex> fwc = wc.FVComplex();

    /*
      for (int i = 0; i < 10; i++)
      {
      d = vecf - mat * vecu;
      cout << "i = " << i << ", err = " << L2Norm (d) << endl;

      w = pre * d;
      aw = mat * w;

	
      for (int j = 0; j < fw.Size(); j++)
      fwc(j) = conj (fw(j));

      fs(0) = S_InnerProduct<Complex> (d, wc);
      fs(1) = S_InnerProduct<Complex> (d, uone);

      ms(0,0) = S_InnerProduct<Complex> (aw, wc);
      ms(0,1) = S_InnerProduct<Complex> (aone, wc);

      ms(1,0) = S_InnerProduct<Complex> (aw, uone);
      ms(1,1) = S_InnerProduct<Complex> (aone, uone);

      CalcInverse (ms, invms);
      us = invms * fs;

      vecu += us(0) * w;
      vecu += us(1) * uone;
      }
    */


    /*
      Complex scale = 1.0 / S_InnerProduct<Complex> (vecfone, uone);

      for (int i = 0; i < 10000; i++)
      {
      d = vecf - mat * vecu;
      d -= (scale * S_InnerProduct<Complex> (d, uone)) * vecfone;

      cout << "i = " << i << ", err = " << L2Norm (d) << endl;
	
      w = pre * d;
      vecu += Complex (0.001, 0) * w;

      cout << "inner1 =                                    " << S_InnerProduct<Complex> (vecu, vecfone) << endl;
      vecu -= (scale * S_InnerProduct<Complex> (vecu, vecfone)) * uone;
      cout << "inner =                                     " << S_InnerProduct<Complex> (vecu, vecfone) << endl;
      }
    */


    /*
      for (int i = 0; i < maxsteps; i++)
      {
      d = vecf - mat * vecu;
      cout << "i = " << i << ", err = " << L2Norm (d) << endl;
	
      w = pre * d;
      vecu += Complex (taur, taui) * w;
      }
    */


    // PardisoInverse<Complex> inv(mat);
    // vecu = inv * vecf;
  }
};


































  namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("mixedresonance", FESpaceMixedResonance::Create);
    GetIntegrators().AddBFIntegrator ("mixedresonance", 2, 0,
				      MixedResonanceIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("PML_mixedresonance", 2, 0,
				      PML_MixedResonanceIntegrator::Create);

    GetNumProcs().AddNumProc ("mixedresonance", NumProcMixedResonance::Create);
  }
  
  Init init;
}


