/*

EHD model in mixed form
Joachim Schoeberl
May 2006
*/


#include <solve.hpp>

using namespace ngsolve;


static double dt = 0.1;             // time step
static double coef_b = 1;           // coefficient for convection
static double coef_elec = 100;      // penalty coefficient for boundary condition

/* *********************************************************** */
/* ************************* FESpace ************************* */
/* *********************************************************** */

// the product space for (D, phi, j, n)
class FESpaceMixedEHD : public CompoundFESpace
{
public:
  FESpaceMixedEHD (const MeshAccess & ama, 		   
	      const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  { ; }

  virtual ~FESpaceMixedEHD () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(4);

    Flags hdivflags, l2flags;

    int order = int (flags.GetNumFlag ("order", 1));

    l2flags.SetFlag ("order", order-1);
    hdivflags.SetFlag ("order", order);

    if (order < 0)
      {
    	spaces[0] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[1] = new ElementFESpace (ma, 0, 0, 0);    
    	spaces[2] = new RaviartThomasFESpace (ma, 1, 0);
	spaces[3] = new ElementFESpace (ma, 0, 0, 0);    
      }
    else
      {
	spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[2] = new HDivHighOrderFESpace (ma, hdivflags);
	spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
	spaces[3] = new L2HighOrderFESpace (ma, l2flags);
      }
    
    FESpaceMixedEHD * fes = new FESpaceMixedEHD (ma, spaces);
    return fes;
  }
    
  virtual string GetClassName () const { return "MixedEHD - FESpace"; }
};


/* ************************************************************ */
/* ************************ Integrator ************************ */
/* ************************************************************ */

class DiffOpEHD : public DiffOp<DiffOpEHD>
{
  // 8 components:
  // Dx, Dy, div D, phi, jx, jy, div j, n

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 8 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();

    FlatVector<> shape_scal(nd_scal, lh);
    shape_scal = fel_scal.GetShape (sip.IP(), lh);

    FlatMatrix<> shape_vec(2, nd_vec, lh);
    shape_vec = (1.0/sip.GetJacobiDet()) * 
      (sip.GetJacobian() * Trans (fel_vec.GetShape(sip.IP(), lh)));

    FlatVector<> shape_div(nd_vec, lh);
    shape_div = 1.0/sip.GetJacobiDet() * fel_vec.GetDivShape(sip.IP(), lh);

    mat = 0;
    int base_j = nd_vec + nd_scal;
    for (int j = 0; j < nd_vec; j++)
      {
	mat(0,j) = shape_vec(0, j);
	mat(1,j) = shape_vec(1, j);
	mat(2,j) = shape_div(j);

	mat(4,base_j+j) = shape_vec(0, j);
	mat(5,base_j+j) = shape_vec(1, j);
	mat(6,base_j+j) = shape_div(j);
      }

    for (int j = 0; j < nd_scal; j++)
      {
	mat(3,nd_vec+j) = shape_scal(j);
	mat(7,base_j+nd_vec+j) = shape_scal(j);
      }
  }


  template <typename FEL, typename SIP, class TVX, class TVY>
  static void Apply (const FEL & bfel, const SIP & sip,
		     const TVX & x, TVY & y,
		     LocalHeap & lh) 
  {
    const CompoundFiniteElement & cfel = 
      static_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_scal = 
      static_cast<const NodalFiniteElement&> (cfel[1]);
    const HDivFiniteElement<2> & fel_vec = 
      static_cast<const HDivFiniteElement<2>&> (cfel[0]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();

    FlatVector<> shape_scal(nd_scal, lh);
    fel_scal.CalcShape (sip.IP(), shape_scal);

    FlatMatrixFixWidth<2> shape_vec(nd_vec, lh);
    fel_vec.CalcShape (sip.IP(), shape_vec);

    FlatVector<> shape_div(nd_vec, lh);
    shape_div = fel_vec.GetDivShape(sip.IP(), lh);

    typedef typename TVX::TSCAL TSCAL;

    Vec<2, TSCAL> h2, E, j;
    Vec<1, TSCAL> divE, divj, phi, n;

    h2 = Trans (shape_vec) * x.Range (0, nd_vec);
    E =  (1.0/sip.GetJacobiDet()) * (sip.GetJacobian() * h2);

    divE =  (1.0/sip.GetJacobiDet()) * (Trans (shape_div) * x.Range (0, nd_vec));
    phi = Trans (shape_scal) * x.Range (nd_vec, nd_vec+nd_scal);

    int basej = nd_vec+nd_scal;

    h2 = Trans (shape_vec) * x.Range (basej, basej+nd_vec);
    j =  (1.0/sip.GetJacobiDet()) * (sip.GetJacobian() * h2);

    divj =  (1.0/sip.GetJacobiDet()) * (Trans (shape_div) * x.Range (basej, basej+nd_vec));
    n = Trans (shape_scal) * x.Range (basej+nd_vec, basej+nd_vec+nd_scal);

    y(0) = E(0);
    y(1) = E(1);
    y(2) = divE(0);
    y(3) = phi(0);
    y(4) = j(0);
    y(5) = j(1);
    y(6) = divj(0);
    y(7) = n(0);
  }

};



class DiffOpEHDbound : public DiffOp<DiffOpEHDbound>
{
  // 2 components:
  // Dn, jn

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 1 };
  enum { DIM_DMAT = 2 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);

    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const HDivNormalFiniteElement<1> & fel_vec = 
      dynamic_cast<const HDivNormalFiniteElement<1>&> (cfel[0]);
    
    int nd_scal = fel_scal.GetNDof();
    int nd_vec = fel_vec.GetNDof();

    FlatVector<> shape(nd_vec, lh);
    shape = 1.0/sip.GetJacobiDet() * fel_vec.GetShape(sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_vec; j++)
      {
	mat(0,j) = shape(j);
	mat(1,nd_vec+nd_scal+j) = shape(j);
      }
  }
};









class EHDDMat : public DMatOp<EHDDMat>
{
public:

  enum { DIM_DMAT = 8 };
  enum { SYMMETRIC = 0 };

  EHDDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    Vec<8> vlin;
    vlin = 0.0;
    GenerateLinearizedMatrix (fel, sip, vlin, mat, lh);
  }  

  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    mat = 0;
    mat(0,0) = mat(1,1) = 1;  // 1/eps    1/eps D
    mat(2,3) = mat(3,2) = 1;  //  div D v
    mat(3,3) = -1e-5;         // regularization
    
    mat(3,7) = -1;            // div D = n;
    
    mat(4,4) = mat(5,5) = 1;  // j = 
    mat(6,7) = mat(7,6) = -1;
    mat(7,7) = -1.0/dt;       // div j - 1/dt dn = 0;

    // j = b n E
    mat(4,7) = -coef_b * vlin(0);
    mat(5,7) = -coef_b * vlin(1);
  }


  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }
};



class MixedEHDIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpEHD, EHDDMat, FiniteElement>
{
public:
  ///
  MixedEHDIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpEHD, EHDDMat, FiniteElement>
  (EHDDMat ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedEHDIntegrator ();
  }
  
  ///
  virtual string Name () const { return "EHD"; }
};






class EHDDMatbound : public DMatOp<EHDDMatbound>
{
public:

  enum { DIM_DMAT = 2 };
  enum { SYMMETRIC = 0 };

  EHDDMatbound () { ; }

  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;
  }  


  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    mat = 0.0;

    int nr = sip.GetTransformation().GetElementIndex();
    if ( nr == 2 )
      {
	if (ReduceToReal(vlin(0)) > 0.3)
	  mat(1,0) = -coef_elec * 1e8;
	mat(1,1) = 1e8;
      }
  }

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }
};


class EHDrhsDMatbound : public DMatOp<EHDrhsDMatbound>
{
public:

  enum { DIM_DMAT = 2 };
  enum { SYMMETRIC = 0 };

  EHDrhsDMatbound () { ; }


  template <typename FEL, typename SIP, class TVX, class TVY>
  void Apply (const FEL & fel, const SIP & sip,
	      const TVX & x, TVY & y,
	      LocalHeap & lh) const
  {
    int nr = sip.GetTransformation().GetElementIndex();

    y = 0.0;
    if ( nr == 2 && ReduceToReal(x(0)) > 0.3)
      y(1) = coef_elec * 1e8 * ( /* x(0) */ - 0.3);
  }

  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;
  }  

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    cout << "apply lin should not be used" << endl;
  }
};



class MixedEHDrhsboundIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpEHDbound, EHDrhsDMatbound, FiniteElement>
{
public:
  MixedEHDrhsboundIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpEHDbound, EHDrhsDMatbound, FiniteElement>
  (EHDrhsDMatbound ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedEHDrhsboundIntegrator ();
  }
  
  ///
  virtual string Name () const { return "EHDrhsbound"; }
};



class MixedEHDboundIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpEHDbound, EHDDMatbound, FiniteElement>
{
public:
  MixedEHDboundIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpEHDbound, EHDDMatbound, FiniteElement>
  (EHDDMatbound ())
  {
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedEHDboundIntegrator ();
  }
  
  ///
  virtual string Name () const { return "EHDrhsbound"; }
};














class EHDJumpsIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction *fac;
public:
  EHDJumpsIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
    : BilinearFormIntegrator()
  { 
    fac = coeffs[0];
  }

  virtual ~EHDJumpsIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & fel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & locheap) const
  {
    elmat = 0.0;
  }


  virtual void
  AssembleLinearizedElementMatrix (const FiniteElement & bfel, 
				   const ElementTransformation & eltrans, 
				   FlatVector<double> & elveclin,
				   FlatMatrix<double> & elmat,
				   LocalHeap & lh) const
  {
    static int maintimer = NgProfiler::CreateTimer ("Jump-stabilization, Assemblelinearized");
    NgProfiler::RegionTimer reg(maintimer);



    int ndof = bfel.GetNDof();
    elmat = 0.0;

    // j - (bE) n - D nabla n

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_vec = fel_vec.GetNDof();
    int nd_scal = fel_scal.GetNDof();
    int nd = bfel.GetNDof();


    FlatMatrix<> grad(nd_scal, 2, lh);
    FlatMatrix<> grad_ref(nd_scal, 2, lh);
    FlatVector<> shape_scal(nd_scal, lh);

    FlatMatrix<> mat_vec(nd_vec, 2, lh);
    FlatMatrix<> mat_vec_ref(nd_vec, 2, lh);
    FlatVector<> mat_vec_n(nd_vec, lh);

    const POINT3D * points = ElementTopology::GetVertices (fel_vec.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_vec.ElementType());
    int ned = (fel_vec.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel_vec.Order());

    FlatVector<>  bvec(nd, lh), bvec2(nd, lh);

    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    fel_scal.CalcShape (ip, shape_scal);
	    fel_scal.CalcDShape (ip, grad_ref);
	    grad = grad_ref * inv_jac;

	    fel_vec.CalcShape (ip, mat_vec_ref);
	    mat_vec = mat_vec_ref * Trans (jac);
	    mat_vec /= det;
	    
	    mat_vec_n = mat_vec * normal;

	    Vec<1> En, jn;
	    Vec<1> n;

	    En = Trans (mat_vec_n) * elveclin.Range (0, nd_vec);

	    // jn = Trans (mat_vec_n) * elx.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal);
	    // n = Trans (shape_scal) * elx.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal);

	    //	    Vec<2> b;
	    //	    b(0) = bx -> Evaluate (sip);
	    //	    b(1) = by -> Evaluate (sip);

	    if (En(0) > 0)
	      {
		bvec = 0.0;
		bvec2 = 0.0;
		
		bvec.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = mat_vec_n;
		bvec.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal) = -(coef_b*En(0)) * shape_scal  - (grad * normal);
		
		bvec2.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = mat_vec_n;

		// double resn = jn(0) - coef_b * En(0)*n(0); // - dn/dn

		/*
		(*testout) << "jn = " << jn << ", en = " << En(0) << ", n = " << n(0) 
			   << " bEn = " << coef_b * En(0)*n(0) << endl;
		*/
		double factor = ir1d[l].Weight() * coef_b * L2Norm(En) * pow (len, 3) * fac -> Evaluate (sip);
		bvec *= factor;
		
		elmat += bvec2 * Trans (bvec);
		// ely.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) += resn * mat_vec;
	      }
	  
	    lh.CleanUp (heapp);
	  }
      }
    // (*testout) << "elmat = " << endl << elmat << endl;
  }










  virtual void 
  ApplyElementMatrix (const FiniteElement & bfel, 
		      const ElementTransformation & eltrans, 
		      const FlatVector<double> & elx, 
		      FlatVector<double> & ely,
		      void * precomputed,
		      LocalHeap & lh) const
  {
    cout << "apply is used - surprise ! " << endl;
    int ndof = bfel.GetNDof();
    ely = 0.0;

    // j - (bE) n - D nabla n

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivFiniteElement<2> & fel_vec = 
      dynamic_cast<const HDivFiniteElement<2>&> (cfel[0]);
    const NodalFiniteElement & fel_scal = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_vec = fel_vec.GetNDof();
    int nd_scal = fel_scal.GetNDof();
    int nd = bfel.GetNDof();


    FlatMatrix<> grad(nd_scal, 2, lh);
    FlatMatrix<> grad_ref(nd_scal, 2, lh);
    FlatVector<> shape_scal(nd_scal, lh);

    FlatMatrix<> mat_vec(nd_vec, 2, lh);
    FlatMatrix<> mat_vec_ref(nd_vec, 2, lh);
    FlatVector<> mat_vec_n(nd_vec, lh);

    const POINT3D * points = ElementTopology::GetVertices (fel_vec.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_vec.ElementType());
    int ned = (fel_vec.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel_vec.Order());

    //    FlatVector<>  bvec(nd, lh), bvec2(nd, lh);

    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = sip.GetJacobiDet();

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    fel_scal.CalcShape (ip, shape_scal);
	    fel_scal.CalcDShape (ip, grad_ref);
	    grad = grad_ref * inv_jac;

	    fel_vec.CalcShape (ip, mat_vec_ref);
	    mat_vec = mat_vec_ref * Trans (jac);
	    mat_vec /= det;
	    
	    mat_vec_n = mat_vec * normal;

	    Vec<1> En, jn;
	    Vec<1> n;
	    Vec<2> gradn;

	    En = Trans (mat_vec_n) * elx.Range (0, nd_vec);
	    jn = Trans (mat_vec_n) * elx.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal);
	    n = Trans (shape_scal) * elx.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal);
	    gradn = Trans (grad) * elx.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal);

	    //	    Vec<2> b;
	    //	    b(0) = bx -> Evaluate (sip);
	    //	    b(1) = by -> Evaluate (sip);

	    if (En(0) < 0)
	      {
		(*testout) << "jn(0) = " << jn(0) << " conv = " << coef_b * En(0)*n(0) << " diff = " << InnerProduct (gradn, normal) << endl;

		double resn = jn(0) - coef_b * En(0)*n(0) - InnerProduct (gradn, normal); // - dn/dn

		/*
		(*testout) << "jn = " << jn << ", en = " << En(0) << ", n = " << n(0) 
			   << " bEn = " << coef_b * En(0)*n(0) << endl;
		*/
		double factor = ir1d[l].Weight() * coef_b * L2Norm(En) * pow (len, 3) * fac -> Evaluate (sip);
		resn *= factor;
		
		ely.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) += resn * mat_vec;
	      }
	  
	    lh.CleanUp (heapp);
	  }
      }
    // (*testout) << "elmat = " << endl << elmat << endl;
  }


  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new EHDJumpsIntegrator (coeffs); 
  }

};




























/* ************************************************************ */
/* ************************ TestSolver ************************ */
/* ************************************************************ */

class NumProcMixedEHD : public NumProc
{
protected:
  double te;
  bool reassemble;
  double dt0;
public:

  NumProcMixedEHD (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    dt0 = flags.GetNumFlag ("dt", 0.1);
    te = flags.GetNumFlag ("te", 1);
    coef_b = flags.GetNumFlag ("b", 1);
    coef_elec = flags.GetNumFlag ("elec", 100);
    reassemble = 1;
  }

  virtual ~NumProcMixedEHD()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedEHD (pde, flags);
  }




  virtual void Do(LocalHeap & lh)
  {
    cout << "mixed EHD solver" << endl;

    const BaseMatrix & mata = pde.GetBilinearForm("a")->GetMatrix();
    const BaseMatrix & matnorm = pde.GetBilinearForm("norm")->GetMatrix();
    const BaseMatrix & matarhs = pde.GetBilinearForm("arhs")->GetMatrix();
    const BaseMatrix & matmassn = pde.GetBilinearForm("massn")->GetMatrix();

    BaseVector & vecu = pde.GetGridFunction("u")->GetVector();
    const BaseVector & vecf = pde.GetLinearForm("f")->GetVector();


    BaseVector & d = *vecu.CreateVector();
    BaseVector & res = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();


    BaseMatrix * inv_norm =
      dynamic_cast<const BaseSparseMatrix&> (matnorm) . InverseMatrix();



    /*
    BitArray potdofs(d.Size());
    BitArray chargedofs(d.Size());
    potdofs.Clear();
    for (int i = 0; i < potdofs.Size()/2; i++)
      potdofs.Set(i);
    chargedofs = potdofs;
    chargedofs.Invert();


    BaseMatrix * invmat_pot = dynamic_cast<const BaseSparseMatrix&> (mata) . InverseMatrix(&potdofs);
    BaseMatrix * invmat_charge = 0;

    if (!reassemble)
      invmat_charge = dynamic_cast<const BaseSparseMatrix&> (mata) . InverseMatrix(&chargedofs);

    // implicite Euler method
    if (ma.GetNLevels() == 1)
      vecu = 0;
    
    dt = dt0;
    for (double t = 0; t <= te; t += dt)
      {
	cout << "t = " << t << endl;

	// update potential:

	// d = matarhs * vecu - (1.0/dt) * (matmassn * vecu) + vecf;
	d = matarhs * vecu;
	d -= (1.0/dt) * (matmassn * vecu);
	d += vecf;
	d -= mata * vecu;

	w = (*invmat_pot) * d;
	vecu += w;


	// update charges
	
	if (reassemble)
	  {
	    pde.GetBilinearForm("a")->AssembleLinearization (vecu, lh);
	    BaseMatrix & mata2 = pde.GetBilinearForm("a")->GetMatrix();
	    delete invmat_charge;
	    invmat_charge = dynamic_cast<const BaseSparseMatrix&> (mata2) . InverseMatrix(&chargedofs);
	  }

	cout << "t = " << t << endl;
	d = matarhs * vecu - (1.0/dt) * (matmassn * vecu) + vecf;
	d -= mata * vecu;

	if (L2Norm (d) < 1e-3)
	  dt *= 2;

	cout << "err = " << L2Norm (d) << ", dt = " << dt << endl;

	w = (*invmat_charge) * d;

	vecu += w;

	// pde.GetBilinearForm("a")->ApplyMatrix (vecu, w);

	Ng_Redraw ();	  
      }

    delete invmat_pot;
    delete invmat_charge;
*/



    BaseMatrix * invmat = 0;
    if (!reassemble)
      invmat = dynamic_cast<const BaseSparseMatrix&> (mata) . InverseMatrix();

    // implicite Euler method
    //    if (ma.GetNLevels() == 1)
      vecu = 0;
    
    dt = dt0;
    for (double t = 0; t <= te; t += dt)
      {
	if (reassemble)
	  {
	    pde.GetBilinearForm("a")->AssembleLinearization (vecu, lh);
	    BaseMatrix & mata2 = pde.GetBilinearForm("a")->GetMatrix();
	    delete invmat;
	    invmat = dynamic_cast<const BaseSparseMatrix&> (mata2) . InverseMatrix();
	  }

	cout << "t = " << t << endl;
	// d = matarhs * vecu - (1.0/dt) * (matmassn * vecu) + vecf;
	d = matarhs * vecu;
	d -= (1.0/dt) * (matmassn * vecu);
	d += vecf;


	d -= mata * vecu;
	w = (*invmat) * d;
	vecu += w;

	// vecu = (*invmat) * d;


	w = (*inv_norm) * d;
	double err = sqrt (InnerProduct (w, d));
	cout << "err = " << err << ", dt = " << dt << endl;

	/*
	if (err < 1)
	  dt *= 2;
	if (err > 10 && dt > dt0)
	  dt /= 2;
	*/
	
	// pde.GetBilinearForm("a")->ApplyMatrix (vecu, w);

	Ng_Redraw ();	  

	// char ch;
	// cin >> ch;
      }

    delete invmat;
  }


  ///
  virtual string GetClassName () const
  {
    return "EHD";
  }

  virtual void PrintReport (ostream & ost)
  { ; }
};








class NumProcMixedEHDPost : public NumProc
{
protected:

public:

  NumProcMixedEHDPost (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    ;
  }

  virtual ~NumProcMixedEHDPost()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedEHDPost (pde, flags);
  }


  virtual void Do(LocalHeap & lh)
  {
    cout << "EHD Postprocessing" << endl;


    const FESpace & fes = *pde.GetFESpace("v");
    const FESpace & fesnpost = *pde.GetFESpace("vpost");

    const S_GridFunction<double> & gfu = 
      dynamic_cast<const S_GridFunction<double>&> (*pde.GetGridFunction ("u"));
     S_GridFunction<double> & gfnpost = 
       dynamic_cast<S_GridFunction<double>&> (*pde.GetGridFunction ("npost"));

    
    int ne      = ma.GetNE();

    ElementTransformation eltrans;

    ARRAY<int> dnums;
    ARRAY<int> dnumsnpost;

    gfnpost.GetVector() = 0.0;

    void * heapp1 = lh.GetPointer();
    for (int i = 0; i < ne; i++)
      {
	ma.SetThreadPercentage ( 100.0*i / ne );

	const FiniteElement & fel = fes.GetFE (i, lh);
	const NodalFiniteElement & felnpost = dynamic_cast<const NodalFiniteElement&> (fesnpost.GetFE (i, lh));

	ma.GetElementTransformation (i, eltrans, lh);
	fes.GetDofNrs (i, dnums);
	fesnpost.GetDofNrs (i, dnumsnpost);

	FlatVector<> elu(dnums.Size(), lh);
	FlatVector<> elnpost(dnumsnpost.Size(), lh);
	FlatVector<> elnposti(dnumsnpost.Size(), lh);
	FlatVector<> nposti(1, lh);

	FlatVector<> shapenpost(dnumsnpost.Size(), lh);
	FlatMatrix<> dshapenpostref(dnumsnpost.Size(), 2, lh);
	FlatMatrix<> dshapenpost(dnumsnpost.Size(), 2, lh);
	FlatMatrix<> elmat(dnumsnpost.Size(), lh);

	FlatVector<> elnpost2(dnumsnpost.Size(), lh);
	elnpost2 = 0.0;
	FlatMatrix<> elmat2(dnumsnpost.Size(), lh);
	elmat2 = 0.0;

	elmat = 0.0;
	gfu.GetElementVector (dnums, elu);
	fes.TransformVec (i, 0, elu, TRANSFORM_SOL);

	const IntegrationRule & ir = 
	  GetIntegrationRules().SelectIntegrationRule(fel.ElementType(), 2*felnpost.Order());
	elnpost = 0;
	Vec<8> vec_all;

	for (int j = 0; j < ir.GetNIP(); j++)
	  {
	    void * heapp2 = lh.GetPointer();

	    SpecificIntegrationPoint<2,2> sip (ir.GetIP(j), eltrans, lh);
	    double fac = sip.IP().Weight() * fabs (sip.GetJacobiDet());

	    DiffOpEHD::Apply (fel, sip, elu, vec_all, lh);
	    Vec<2> vj, E;
	    vj(0) = vec_all(4);
	    vj(1) = vec_all(5);
	    E(0) = vec_all(0);
	    E(1) = vec_all(1);

	    felnpost.CalcShape (ir.GetIP(j), shapenpost);
	    elnpost += (coef_b * fac * InnerProduct (vj, E)) * shapenpost;

	    elmat += (fac * coef_b * coef_b * InnerProduct (E, E)) * (shapenpost * Trans (shapenpost));
	    lh.CleanUp (heapp2);
	  }


	const POINT3D * points = ElementTopology::GetVertices (felnpost.ElementType());
	const EDGE * edges = ElementTopology::GetEdges (felnpost.ElementType());
	int ned = (felnpost.ElementType() == ET_TRIG) ? 3: 4;
	
	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*felnpost.Order());

	//	FlatVector<>  bvec(nd, lh), bvec2(nd, lh);

	for (int k = 0; k < ned; k++)
	  {
	    Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	    ps(0) = points[edges[k][0]][0];
	    ps(1) = points[edges[k][0]][1];
	    pe(0) = points[edges[k][1]][0];
	    pe(1) = points[edges[k][1]][1];
	    
	    tau_ref = pe-ps;
	
	    void * heapp = lh.GetPointer();
	    
	    for (int l = 0; l < ir1d.GetNIP(); l++)
	      {
		Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
		IntegrationPoint ip( p2d(0), p2d(1) );
		
		SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);

		DiffOpEHD::Apply (fel, sip, elu, vec_all, lh);
		Vec<2> vj, E;
		vj(0) = vec_all(4);
		vj(1) = vec_all(5);
		E(0) = vec_all(0);
		E(1) = vec_all(1);


		Mat<2> jac = sip.GetJacobian();
		Mat<2> inv_jac = sip.GetJacobianInverse();
		double det = sip.GetJacobiDet();

		tau = jac * tau_ref;
		double len = L2Norm(tau);
		
		tau *= (1.0/len);
		normal(0) = tau(1);
		normal(1) = -tau(0);
		

		
		// felnpost.CalcShape (ip, shapenpost);
		// felnpost.CalcDShape (ip, grad_ref);
		// grad = grad_ref * inv_jac;

		/*
	    fel_vec.CalcShape (ip, mat_vec_ref);
	    mat_vec = mat_vec_ref * Trans (jac);
	    mat_vec /= det;
	    
	    mat_vec_n = mat_vec * normal;

	    Vec<1> En, jn;
	    Vec<1> n;

	    En = Trans (mat_vec_n) * elu.Range (0, nd_vec);
		*/
		double En = InnerProduct (normal, E);
		double jn = InnerProduct (normal, vj);
		
		if (En < 0)
		  {
		    double factor = ir1d[l].Weight() * coef_b * fabs (En) * pow (len, 3) * 10; // fac -> Evaluate (sip);
		    felnpost.CalcShape (ip, shapenpost);
		    
		    elnpost2 += (factor * coef_b * jn*En) * shapenpost;
		    elmat2 += (factor * coef_b * coef_b * En * En) * (shapenpost * Trans (shapenpost));
		    // lh.CleanUp (heapp2);

		    /*
		    bvec = 0.0;
		    bvec2 = 0.0;
		    
		    bvec.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = mat_vec_n;
		    bvec.Range (2*nd_vec+nd_scal, 2*nd_vec+2*nd_scal) = -(coef_b*En(0)) * shape_scal  - (grad * normal);
		    
		    bvec2.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) = mat_vec_n;
		    
		    // double resn = jn(0) - coef_b * En(0)*n(0); // - dn/dn
		    

		    bvec *= factor;
		    
		    elmat += bvec2 * Trans (bvec);
		    // ely.Range (nd_vec+nd_scal, 2*nd_vec+nd_scal) += resn * mat_vec;
		    */
		  }
		
		lh.CleanUp (heapp);
	      }
	  }


	(*testout) << "elmat = " << endl << elmat << endl;
	(*testout) << "elnpost = " << endl << elnpost << endl;
	(*testout) << "elmat2 = " << endl << elmat2 << endl;
	(*testout) << "elnpost2 = " << endl << elnpost2 << endl;

	FlatCholeskyFactors<double> invelmat(elmat, lh);
	invelmat.Mult (elnpost, elnposti);
	
	gfnpost.GetElementVector (dnumsnpost, elnpost);
	elnposti += elnpost;
	gfnpost.SetElementVector (dnumsnpost, elnposti);
	
	


	(*testout) << "sol = " << endl << elnposti << endl;
	elmat += elmat2;
	elnpost += elnpost2;
	FlatCholeskyFactors<double> invelmat2(elmat, lh);
	invelmat2.Mult (elnpost, elnposti);


	(*testout) << "sol2 = " << endl << elnposti << endl;




	lh.CleanUp(heapp1);
      }
  




    // error estimator

    S_GridFunction<double> & gferr = 
      dynamic_cast<S_GridFunction<double>&> (*pde.GetGridFunction ("err"));
    ARRAY<int> elsonvertex, ednums, vnums;

    gferr.GetVector() = 0.0;
    int ned = ma.GetNEdges();
    for (int i = 0; i < ned; i++)
      {
	int vi1, vi2;
	ma.GetEdgePNums (i, vi1, vi2);

	int elnr1 = -1, elnr2 = -1;
	ma.GetVertexElements(vi1, elsonvertex);

	// for (int j = 0; j < ne; j++)
	for (int jj = 0; jj < elsonvertex.Size(); jj++)
	  {
	    int j = elsonvertex[jj];
	    ma.GetElEdges (j, ednums);
	    for (int k = 0; k < ednums.Size(); k++)
	      if (ednums[k] == i)
		{
		  elnr2 = elnr1; 
		  elnr1 = j;
		}
	  }

	if (elnr1 != -1)
	  {
	    Vec<2> p1, p2;
	    ma.GetPoint (vi1, p1);
	    ma.GetPoint (vi2, p2);

	    Vec<2> tau = p2 - p1;
	    Vec<2> n;
	    n(0) = tau(1); n(1) = -tau(0);
	    double h = L2Norm (tau);
	    n /= h;
	    tau /= h;

	    const IntegrationRule & ir = 
	      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 10);

	    double err1 = 0;

	    for (int ipi = 0; ipi < ir.GetNIP(); ipi++)
	      {
		const IntegrationPoint & ip = ir[ipi];
		
		double diffu = 0.0;

		int nels = (elnr2 != -1) ? 2 : 1;
		for (int j = 0; j < nels; j++)
		  {
		    int elnr = (j == 0) ? elnr1 : elnr2;
		    double fac = (j == 0) ? 1 : -1;

		    const NodalFiniteElement & fel = 
		      dynamic_cast<const NodalFiniteElement&> (fesnpost.GetFE(elnr, lh));
		    fesnpost.GetDofNrs (elnr, dnums);
		    FlatVector<> elu(dnums.Size(), lh);
		    FlatVector<> shapeu(dnums.Size(), lh);
		    gfnpost.GetElementVector (dnums, elu);

		    ElementTransformation eltrans;
		    ma.GetElementTransformation (elnr, eltrans, lh);

		    ma.GetElVertices (elnr, vnums);
		    const POINT3D *pelp1, *pelp2;
		    Vec<2> elp1, elp2;
		    for (int k = 0; k < vnums.Size(); k++)
		      {
			if (vnums[k] == vi1)
			  pelp1 = ElementTopology::GetVertices(ET_TRIG)+k;
			if (vnums[k] == vi2)
			  pelp2 = ElementTopology::GetVertices(ET_TRIG)+k;
		      }
		    for (int k = 0; k < 2; k++)
		      {
			elp1(k) = (*pelp1)[k];
			elp2(k) = (*pelp2)[k];
		      }

		    Vec<2> pref = elp1 + ip(0) * (elp2 - elp1);
		    IntegrationPoint ipel(pref(0), pref(1), 0, 0);
		    
		    fel.CalcShape (ipel, shapeu);
		    
		    diffu += fac * InnerProduct (shapeu, elu);
		  }


		if (nels == 2)
		  err1 += ip.Weight() * /* h * */ sqr (diffu);
		
		lh.CleanUp();
	      }
	    
	    if (elnr2 != -1)
	      {
		gferr.GetVector().FVDouble()(elnr1) += err1;
		gferr.GetVector().FVDouble()(elnr2) += err1;
	      }
	  }
      }
  }
};








namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("mixedEHD", NumProcMixedEHD::Create);
    GetNumProcs().AddNumProc ("mixedEHDpost", NumProcMixedEHDPost::Create);

    GetFESpaceClasses().AddFESpace ("mixedEHD", FESpaceMixedEHD::Create);
  
    GetIntegrators().AddBFIntegrator ("mixedEHD", 2, 0,
  				      MixedEHDIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedEHDbound", 2, 0,
  				      MixedEHDboundIntegrator::Create);

    GetIntegrators().AddBFIntegrator ("mixedEHDjumps", 2, 1,
  				      EHDJumpsIntegrator::Create);

    GetIntegrators().AddBFIntegrator ("mixedEHDrhsbound", 2, 0,
  				      MixedEHDrhsboundIntegrator::Create);
  }
  Init init;
}
