/*

Mixed elasticity

*/


#include <solve.hpp>

using namespace ngfem;
using namespace ngsolve;

#include "mixedmech.hpp"

/* ********************* FESpace **************************** */


  inline AutoDiff<3> Cross (const AutoDiff<3> & u,
			    const AutoDiff<3> & v)
  {
    AutoDiff<3> hv;
    hv.Value() = 0.0;
    hv.DValue(0) = u.DValue(1)*v.DValue(2)-u.DValue(2)*v.DValue(1);
    hv.DValue(1) = u.DValue(2)*v.DValue(0)-u.DValue(0)*v.DValue(2);
    hv.DValue(2) = u.DValue(0)*v.DValue(1)-u.DValue(1)*v.DValue(0);
    return hv;
  }


template <int DIM>  
HDivSymFiniteElement<DIM> :: HDivSymFiniteElement (ELEMENT_TYPE etype)
  : FiniteElement (ElementTopology::GetSpaceDim (etype), etype, -1, -1) 
{
  for (int i = 0; i < 8; i++)
    vnums[i] = i;
}


template <int DIM>  
HDivSymFiniteElement<DIM> ::  ~HDivSymFiniteElement() 
{ ; }


template <int DIM>  
void HDivSymFiniteElement<DIM> :: 
SetVertexNumbers (FlatArray<int> & avnums)
{
  for (int i = 0; i < avnums.Size(); i++)
    vnums[i] = avnums[i];
}



template <int DIM>  
void  HDivSymFiniteElement<DIM> :: CalcDivShape (const IntegrationPoint & ip, 
						 FlatMatrixFixWidth<DIM> divshape) const 
{
  double eps = 1e-4;  

  ArrayMem<double, 200> 
    hm1(DIM_STRESS*ndof), hm2(DIM_STRESS*ndof), 
    hm3(DIM_STRESS*ndof), hm4(DIM_STRESS*ndof), hmi(DIM_STRESS*ndof);
  
  FlatMatrixFixWidth<DIM_STRESS> shape1(ndof, &hm1[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape2(ndof, &hm2[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape3(ndof, &hm3[0]);
  FlatMatrixFixWidth<DIM_STRESS> shape4(ndof, &hm4[0]);
  FlatMatrixFixWidth<DIM_STRESS> dshapei(ndof, &hmi[0]);

  divshape = 0;

  for (int i = 0; i < DIM; i++)
    {
      IntegrationPoint ip1 = ip;
      IntegrationPoint ip2 = ip;
      ip1(i) -= eps;
      ip2(i) += eps;
      
      CalcShape (ip1, shape1);
      CalcShape (ip2, shape2);
      
      ip1(i) -= eps;
      ip2(i) += eps;
      CalcShape (ip1, shape3);
      CalcShape (ip2, shape4);
      
      dshapei = 
	2/(3*eps) * (shape2 - shape1) 
	- 1/(12*eps) * (shape4 - shape3);
      
      
      if (DIM == 2)
	{
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,0);
		  divshape(j,1) += dshapei(j,2);
		}
	      break;
	      
	    case 1:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,2);
		  divshape(j,1) += dshapei(j,1);
		}
	      break;
	    }
	}
      else
	{
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,0);   // sigmaxx
		  divshape(j,1) += dshapei(j,3);   // sigmaxy
		  divshape(j,2) += dshapei(j,4);   // sigmaxz
		}
	      break;
	    
	    case 1:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,3);
		  divshape(j,1) += dshapei(j,1);
		  divshape(j,2) += dshapei(j,5);
		}
	      break;

	    case 2:
	      for (int j = 0; j < ndof; j++)
		{
		  divshape(j,0) += dshapei(j,4);
		  divshape(j,1) += dshapei(j,5);
		  divshape(j,2) += dshapei(j,2);
		}
	      break;
	    }
	}
    }
  // (*testout) << "div-shape = " << divshape << endl;
}








template <int DIM>  
void  HDivSymFiniteElement<DIM> :: CalcDivDivShape (const IntegrationPoint & ip, 
						    FlatVector<> ddshape) const 
{
  double eps = 1e-4;  

  if (ElementTopology::GetSpaceDim (ElementType()) == 2)
    {
      ArrayMem<double, 200> hm1(3*ndof), hm2(3*ndof), 
	hm3(3*ndof), hm4(3*ndof), hmi(3*ndof);
    
      FlatMatrix<> shape1(ndof, 2, &hm1[0]);
      FlatMatrix<> shape2(ndof, 2, &hm2[0]);
      FlatMatrix<> shape3(ndof, 2, &hm3[0]);
      FlatMatrix<> shape4(ndof, 2, &hm4[0]);
      FlatMatrix<> dshapei(ndof, 2, &hmi[0]);

      ddshape = 0;
    
      for (int i = 0; i < 2; i++)
	{
	  IntegrationPoint ip1 = ip;
	  IntegrationPoint ip2 = ip;
	  ip1(i) -= eps;
	  ip2(i) += eps;

	  CalcDivShape (ip1, shape1);
	  CalcDivShape (ip2, shape2);

	  ip1(i) -= eps;
	  ip2(i) += eps;
	  CalcDivShape (ip1, shape3);
	  CalcDivShape (ip2, shape4);
	
	  dshapei = 
	    2/(3*eps) * (shape2 - shape1) 
	    - 1/(12*eps) * (shape4 - shape3);
	
	  switch (i)
	    {
	    case 0:
	      for (int j = 0; j < ndof; j++)
		ddshape(j) += dshapei(j,0);
	      break;
	    
	    case 1:
	      for (int j = 0; j < ndof; j++)
		ddshape(j) += dshapei(j,1);
	      break;
	    }
	}
    }
}




template class  HDivSymFiniteElement<2>;
template class  HDivSymFiniteElement<3>;
















// Trig with full basis (non-curved elements)

#ifdef OLD
HDivSymTrig :: HDivSymTrig (int aorder, int aorder_facet) : HDivSymFiniteElement<2> (ET_TRIG)
{
  order = aorder+3;
  order_facet = aorder_facet;
  order_inner = aorder;
  
  // ninner = 3*(order_inner+1)*(order_inner)/2 + order_inner+1;
  ninner = 3*(order_inner+1)*(order_inner)/2 + 2*order_inner+3;

  // just div-free
  // ninner = (order_inner+1)*(order_inner+2)/2;

  ndof = 3*(order_facet+1) + ninner;

  cout << "trig, order_facet = " << order_facet << ", order_inner = " << order_inner << endl;
  cout << "ndof = " << ndof << ", ninner = " << ninner << endl;

  //  if (order_facet > order_inner)
  //    throw Exception ("HDivSymTrig, order_face > order_inner)");
}


void HDivSymTrig :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymTrig :: CalcShape (const IntegrationPoint & ip, 
			       FlatMatrixFixWidth<3> shape) const   // n * 3 
{
  AutoDiff<2> x(ip(0),0); 
  AutoDiff<2> y(ip(1),1); 
  AutoDiff<2> lami[3] = {x,y,1-x-y};
  AutoDiffDiff<2> ddlami[3] = {x,y,1-x-y};

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
  ArrayMem<AutoDiffDiff<2>, 20> ha(order_facet+1);

  for (i = 0; i < 3; i++)
    {
      int es = edges[i][0]; 
      int ee = edges[i][1];
	
      if (vnums[es] > vnums[ee])  swap (es, ee);

      AutoDiffDiff<2> ls = lami[es];
      AutoDiffDiff<2> le = lami[ee];

      IntegratedLegendreMonomialExt::CalcTrigExt (order_facet+2, le-ls, 1-le-ls, ha);
      for (int l = 0; l <= order_facet; l++)
	{
	  shape(ii,0) = ha[l].DDValue(1,1);
	  shape(ii,1) = ha[l].DDValue(0,0);
	  shape(ii,2) = -ha[l].DDValue(1,0);
	    
	  ii++;
	}
    }


  // inner shapes

  AutoDiffDiff<2> ddbub = ddlami[0] * ddlami[1] * ddlami[2];

  AutoDiffDiff<2> add_x[10];
  add_x[0] = 1.0;
  AutoDiffDiff<2> add_y[10];
  add_y[0] = 1.0;

  for (int i = 1; i <= order_inner+1; i++)
    {
      add_x[i] = add_x[i-1] * AutoDiffDiff<2> (1-2.0*x);
      add_y[i] = add_y[i-1] * AutoDiffDiff<2> (1-2.0*y);
    }


  for (int k = 0; k < 3; k++)
    {
      AutoDiffDiff<2> bub = ddlami[k] * ddlami[(k+1)%3];
      double lami = ddlami[(k+2)%3].Value();
      
      for (int i = 0; i <= order_inner-1; i++)
	for (int j = 0; i+j <= order_inner-1; j++)
	  {
	    double fac = lami * add_x[i].Value() * add_y[j].Value();
	    shape(ii,0) = bub.DDValue(1,1) * fac;
	    shape(ii,1) = bub.DDValue(0,0) * fac;
	    shape(ii,2) = -bub.DDValue(1,0) * fac;
	    ii++;
	  }
    }


  // one order higher for second derivatives
  for (int i = 0; i <= order_inner; i++)
    {
      AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner-i];
      shape(ii,0) = pot.DDValue(1,1);
      shape(ii,1) = pot.DDValue(0,0);
      shape(ii,2) = -pot.DDValue(1,0);
      ii++;
    }
  // and one more
  for (int i = 0; i <= order_inner+1; i++)
    {
      AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner+1-i];
      shape(ii,0) = pot.DDValue(1,1);
      shape(ii,1) = pot.DDValue(0,0);
      shape(ii,2) = -pot.DDValue(1,0);
      ii++;
    }

  /*
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j+i <= order_inner; j++)
      {
	// AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[order_inner-i];
	AutoDiffDiff<2> pot = ddbub * add_x[i] * add_y[j];
	
	shape(ii,0) = pot.DDValue(1,1);
	shape(ii,1) = pot.DDValue(0,0);
	shape(ii,2) = -pot.DDValue(1,0);
	ii++;
      }
  */
}

#endif


// version Feb 10, Astrid

HDivSymTrig :: HDivSymTrig (int aorder, int aorder_facet) : HDivSymFiniteElement<2> (ET_TRIG)
{
  order = 2;
  order_facet = 1;
  order_inner = 2;
  

  ninner = 5;
  ndof = 11;
}


void HDivSymTrig :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymTrig :: CalcShape (const IntegrationPoint & ip, 
			       FlatMatrixFixWidth<3> shape) const   // n * 3 
{
  AutoDiff<2> x(ip(0),0); 
  AutoDiff<2> y(ip(1),1); 
  AutoDiff<2> lami[3] = {x,y,1-x-y};
  AutoDiffDiff<2> ddlami[3] = {x,y,1-x-y};

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
  ArrayMem<AutoDiffDiff<2>, 20> ha(order_facet+1);

  for (i = 0; i < 3; i++)
    {
      int es = edges[i][0]; 
      int ee = edges[i][1];
	
      if (vnums[es] > vnums[ee])  swap (es, ee);

      AutoDiffDiff<2> ls = lami[es];
      AutoDiffDiff<2> le = lami[ee];

      IntegratedLegendreMonomialExt::CalcTrigExt (order_facet+2, le-ls, 1-le-ls, ha);
      for (int l = 0; l <= order_facet; l++)
	{
	  shape(ii,0) = ha[l].DDValue(1,1);
	  shape(ii,1) = ha[l].DDValue(0,0);
	  shape(ii,2) = -ha[l].DDValue(1,0);
	    
	  ii++;
	}
    }


  // inner shapes, P1
  double xs = ip(0), ys = ip(1);
  /*
  shape(6,0) = 2 * xs;
  shape(6,1) = 2 * ys;
  shape(6,2) = 1-2*xs-2*ys;
  */
  shape(7,0) = 0;
  shape(7,1) = 2*ys;
  shape(7,2) = -ys;

  shape(6,0) = 2*xs;
  shape(6,1) = 0;
  shape(6,2) = -xs;


  for (int j = 0; j < 3; j++)
    {
      AutoDiffDiff<2> phi = ddlami[0] * ddlami[1] * ddlami[2] * ddlami[j];

      shape(8+j,0) = phi.DDValue(1,1);
      shape(8+j,1) = phi.DDValue(0,0);
      shape(8+j,2) = -phi.DDValue(1,0);
    }
}









HDivSymQuad :: HDivSymQuad (int aorder, int aorder_facet) : HDivSymFiniteElement<2> (ET_QUAD)
{
  order = aorder+3;
  order_facet = aorder_facet;
  order_inner = aorder;

  // ninner = 2 * sqr(order_inner+1) + 2*(order_inner+1) + 1;
  ninner = 3 * sqr(order_inner+1);
  ndof = 4*(order_facet+1) + ninner;
}


void HDivSymQuad :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymQuad :: CalcShape (const IntegrationPoint & ip, 
			       FlatMatrixFixWidth<3> shape) const   // n * 3 
{
  AutoDiffDiff<2> x(ip(0),0); 
  AutoDiffDiff<2> y(ip(1),1); 

  AutoDiffDiff<2> lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  AutoDiffDiff<2> sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
  ArrayMem<AutoDiffDiff<2>, 20> ha(order_facet+1);
	
  for (i = 0; i < 4; i++)
    {
      int es = edges[i][0], ee = edges[i][1];
      if (vnums[es] > vnums[ee])  swap (es, ee);

      AutoDiffDiff<2> xi = sigma[es]-sigma[ee];
      AutoDiffDiff<2> eta = lami[es]+lami[ee];

      IntegratedLegendreMonomialExt::Calc (order_facet+2, -xi, ha);

      for (int l = 0; l <= order_facet; l++)
	{
	  ha[l] *= eta;
	  shape(ii,0) = ha[l].DDValue(1,1);
	  shape(ii,1) = ha[l].DDValue(0,0);
	  shape(ii,2) = -ha[l].DDValue(1,0);
	    
	  ii++;
	}
    }

  /*
  // inner shapes

  double polx[3] = { 1, 2*x.Value()-1, x.Value()*(1-x.Value()) };
  double poly[3] = { 1, 2*y.Value()-1, y.Value()*(1-y.Value()) };

  AutoDiffDiff<2> add_x[10];
  add_x[0] = x * (1-x);
  AutoDiffDiff<2> add_y[10];
  add_y[0] = y * (1-y);

  for (int i = 1; i <= order_inner; i++)
    {
      add_x[i] = add_x[i-1] * (1-2.0*x);
      add_y[i] = add_y[i-1] * (1-2.0*y);
    }

  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j <= order_inner; j++)
      {
	AutoDiffDiff<2> prod = add_x[i] * add_y[j];

	shape(ii,0) = add_x[i].Value() * add_y[j].DDValue(1,1);
	shape(ii,1) = add_x[i].DDValue(0,0) * add_y[j].Value();
	shape(ii,2) = -add_x[i].DValue(0) * add_y[j].DValue(1);
	ii++;

	shape(ii,0) = -add_x[i].Value() * add_y[j].DDValue(1,1);
	shape(ii,1) = add_x[i].DDValue(0,0) * add_y[j].Value();
	shape(ii,2) = 0;
	ii++;
      }

  shape(ii,2) = 1;
  ii++;

  for (int i = 0; i <= order_inner; i++)
    {
      shape(ii,2) = -0.5 * add_y[i].DValue(1);  
      ii++;
      shape(ii,2) = -0.5 * add_x[i].DValue(0);  
      ii++;
    }
  */
  
  ArrayMem<double, 10> polx(order_inner+1), poly(order_inner+1);
  LegendrePolynomial (order_inner, 2*ip(0)-1, polx);
  LegendrePolynomial (order_inner, 2*ip(1)-1, poly);

  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j <= order_inner; j++)
      {
	shape(ii++,0) = ip(0)*(1-ip(0))*polx[i]*poly[j];
	shape(ii++,1) = ip(1)*(1-ip(1))*polx[i]*poly[j];
	shape(ii++,2) = polx[i]*poly[j];
      }
}





void HDivSymQuad :: CalcDivShape (const IntegrationPoint & ip, 
				  FlatMatrixFixWidth<2> divshape) const   // n * 3 
{
  //  Matrix<> hdivshape (ndof, 2);
  // HDivSymFiniteElement::CalcDivShape (ip, hdivshape);

  AutoDiffDiff<2> x(ip(0),0); 
  AutoDiffDiff<2> y(ip(1),1); 

  AutoDiffDiff<2> lami[4] = {(1-x)*(1-y),x*(1-y),x*y,(1-x)*y};  
  AutoDiffDiff<2> sigma[4] = {(1-x)+(1-y),x+(1-y),x+y,(1-x)+y};  

  divshape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const EDGE * edges = ElementTopology::GetEdges (ET_QUAD);
  ArrayMem<AutoDiffDiff<2>, 20> ha(order_facet+1);
	
  for (i = 0; i < 4; i++)
    {
      int es = edges[i][0], ee = edges[i][1];
      if (vnums[es] > vnums[ee])  swap (es, ee);

      ii += order_facet+1;
    }


  AutoDiff<1> adx(ip(0),0); 
  AutoDiff<1> ady(ip(1),0); 

  
  ArrayMem<AutoDiff<1>, 10> polx(order_inner+1), poly(order_inner+1);
  ArrayMem<AutoDiff<1>, 10> polxbub(order_inner+1), polybub(order_inner+1);

  LegendrePolynomial (order_inner, 2*adx-1, polx);
  LegendrePolynomial (order_inner, 2*ady-1, poly);

  for (int i = 0; i <= order_inner; i++)
    {
      polxbub[i] = adx*(1-adx)*polx[i];
      polybub[i] = ady*(1-ady)*poly[i];
    }

  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; j <= order_inner; j++)
      {
	divshape(ii++,0) = polxbub[i].DValue(0) * poly[j].Value();
	divshape(ii++,1) = polybub[j].DValue(0) * polx[i].Value();
	divshape(ii  ,1) = polx[i].DValue(0) * poly[j].Value();
	divshape(ii++,0) = polx[i].Value() * poly[j].DValue(0);
	/*
	shape(ii++,0) = ip(0)*(1-ip(0))*polx[i]*poly[j];
	shape(ii++,1) = ip(1)*(1-ip(1))*polx[i]*poly[j];
	shape(ii++,2) = polx[i]*poly[j];
	*/
      }

  //   (*testout) << "num-div = " << endl << hdivshape << endl;
  // (*testout) << "eaxt-div = " << endl << divshape << endl;
}












HDivSymTet :: HDivSymTet (int aorder, int aorder_facet) 
  : HDivSymFiniteElement<3> (ET_TET)
{
  order = max (aorder, aorder_facet)+1;
  order_facet = aorder_facet;
  order_inner = aorder;

  ninner = 6 * (order_inner+1)*(order_inner+2)*(order_inner+3)/6 - 4 * (order_inner+1)*(order_inner+2)/2;
  ndof = 4 * (order_facet+1)*(order_facet+2) / 2 + ninner;

  ninner = ndof;  // hybrid version !!!
}


void HDivSymTet :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymTet :: CalcShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<6> shape) const   // n * 6
{
  AutoDiff<3> x (ip(0),0);
  AutoDiff<3> y (ip(1),1);
  AutoDiff<3> z (ip(2),2);
    
  AutoDiff<3> lami[4] = {x, y, z, 1-x-y-z}; 
  double li[4] = { ip(0), ip(1), ip(2), 1-ip(0)-ip(1)-ip(2) };
    
  // int i, j, ii, k,l, m, n;
  shape = 0.0; 
  
  const FACE * faces = ElementTopology::GetFaces (ET_TET);
  
  int is, ie, iop, iop1, iop2;
    
  ArrayMem<double,20> polxi(order+1), poleta(order+1), polzeta(order+1);
     
  int ii = 0;
  for (int i = 0; i < 4; i++)
    {
      int fav[3] =  { faces[i][0], faces[i][1], faces[i][2] };
      
      //Sort vertices  first edge op minimal vertex
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
      if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
      
      AutoDiff<3> cr0 = Cross(lami[fav[1]],lami[fav[2]]); 
      AutoDiff<3> cr1 = Cross(lami[fav[2]],lami[fav[0]]);

      Mat<3> sigma;
      for (int j = 0; j < 3; j++)
	for (int k = 0; k < 3; k++)
	  sigma(j,k) = cr0.DValue(j) * cr1.DValue(k) + cr0.DValue(k) * cr1.DValue(j);

      ScaledLegendrePolynomial (order_facet, li[fav[1]]-li[fav[2]], 1-li[fav[0]], polxi);
      LegendrePolynomial (order_facet, 2*li[fav[0]]-1, poleta);

      for (int j = 0; j <= order_facet; j++)
	for (int k = 0; j+k <= order_facet; k++)
	  {
	    double val = polxi[j]*poleta[k];
	    shape(ii,0) = val*sigma(0,0);
	    shape(ii,1) = val*sigma(1,1);
	    shape(ii,2) = val*sigma(2,2);
	    shape(ii,3) = val*sigma(0,1);
	    shape(ii,4) = val*sigma(0,2);
	    shape(ii,5) = val*sigma(1,2);
	    ii++;
	  }
    }


  ScaledLegendrePolynomial (order_inner, li[0]-li[1], li[0]+li[1], polxi);
  ScaledLegendrePolynomial (order_inner, li[2]-li[0]-li[1], li[0]+li[1]+li[2], poleta);
  LegendrePolynomial (order_inner, 2*li[3]-1, polzeta);

  AutoDiff<3> cr0 = Cross (lami[0], lami[1]);
  AutoDiff<3> cr1 = Cross (lami[2], lami[3]);

  Mat<3> sigma;
  for (int j = 0; j < 3; j++)
    for (int k = 0; k < 3; k++)
      sigma(j,k) = cr0.DValue(j) * cr1.DValue(k) + cr0.DValue(k) * cr1.DValue(j);

  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      for (int k = 0; i+j+k <= order_inner; k++)
	{
	  double val = polxi[i]*poleta[j]*polzeta[k];
	  shape(ii,0) = val*sigma(0,0);
	  shape(ii,1) = val*sigma(1,1);
	  shape(ii,2) = val*sigma(2,2);
	  shape(ii,3) = val*sigma(0,1);
	  shape(ii,4) = val*sigma(0,2);
	  shape(ii,5) = val*sigma(1,2);
	  ii++;
	}

  cr0 = Cross (lami[0], lami[2]);
  cr1 = Cross (lami[1], lami[3]);

  for (int j = 0; j < 3; j++)
    for (int k = 0; k < 3; k++)
      sigma(j,k) = cr0.DValue(j) * cr1.DValue(k) + cr0.DValue(k) * cr1.DValue(j);

  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      for (int k = 0; i+j+k <= order_inner; k++)
	{
	  double val = polxi[i]*poleta[j]*polzeta[k];
	  shape(ii,0) = val*sigma(0,0);
	  shape(ii,1) = val*sigma(1,1);
	  shape(ii,2) = val*sigma(2,2);
	  shape(ii,3) = val*sigma(0,1);
	  shape(ii,4) = val*sigma(0,2);
	  shape(ii,5) = val*sigma(1,2);
	  ii++;
	}



  for (int i = 0; i < 4; i++)
    {
      int fav[3] =  { faces[i][0], faces[i][1], faces[i][2] };
      
      AutoDiff<3> cr0 = Cross(lami[fav[1]],lami[fav[2]]); 
      AutoDiff<3> cr1 = Cross(lami[fav[2]],lami[fav[0]]);

      Mat<3> sigma;
      for (int j = 0; j < 3; j++)
	for (int k = 0; k < 3; k++)
	  sigma(j,k) = cr0.DValue(j) * cr1.DValue(k) + cr0.DValue(k) * cr1.DValue(j);

      for (int j = 0; j <= order_inner-1; j++)
	for (int k = 0; j+k <= order_inner-1; k++)
	  for (int l = 0; j+k+l <= order_inner-1; l++)
	    {
	      double val = polxi[j]*poleta[k]*polzeta[l] * (1-li[fav[0]]-li[fav[1]]-li[fav[2]]);
	      shape(ii,0) = val*sigma(0,0);
	      shape(ii,1) = val*sigma(1,1);
	      shape(ii,2) = val*sigma(2,2);
	      shape(ii,3) = val*sigma(0,1);
	      shape(ii,4) = val*sigma(0,2);
	      shape(ii,5) = val*sigma(1,2);
	      ii++;
	    }
    }





  // (*testout) << "sigma = " << endl << shape << endl;
}













// [P^2(x,y) x P^2(z)]^6  =   108
// sigma_nn = P^1   bot/top  ->  -2 * 3
//                  quad-faces -> -3 * 5
//                            == 87
// div sigma in [P^0]^3       -15

// quad-face ... 4    -> 12
// trig-face ... 3    -> 6
//  n-inner = 54

// inner,  sigma(x,y)-tensor  Trig-inner tensor P2(z)  -> 5 x 3 = 15
// inner,  sigma_xz, sigma_yz   voll P2                -> 2 x 18 = 36
// inner,  sigma_zz, P2 * z(1-z)                                    6
//                                                              = 57
// sigma_xx, sigma_yy, sigma_zz, sigma_xy, sigma_xz, sigma_yz



HDivSymPrism :: HDivSymPrism (int aorder, int aorder_facet) 
  : HDivSymFiniteElement<3> (ET_PRISM)
{
  order = aorder+3;
  order_facet = aorder_facet;
  order_inner = aorder;
  orderz = 1;

  ninner = 57;   // 54 - 57(order_inner+1)*(order_inner+2)/2 + 2;  
  // ninner = 15 + 2*18 + 6;
  ninner = (orderz+1) * 3 * (order_inner+1) * order_inner / 2 + 
    2*(orderz)*(order_inner+1)*(order_inner+2)/2  +  
    (orderz-1) * (order_inner+1)*(order_inner+2)/2;
  ndof = 3 * sqr(order_facet+1) + 2 * 3  + ninner;
}


void HDivSymPrism :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize (0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymPrism :: CalcShape (const IntegrationPoint & ip, 
				FlatMatrixFixWidth<6> shape) const   // n * 3 
{
  AutoDiff<2> x(ip(0),0); 
  AutoDiff<2> y(ip(1),1); 
  AutoDiff<1> z(ip(2),0);
  AutoDiff<2> lami[6] = {x,y,1-x-y, x, y, 1-x-y };
  AutoDiffDiff<2> ddlami[6] = {x,y,1-x-y, x,y,1-x-y};
    
  AutoDiff<1> lamiz[6] = {1-z,1-z,1-z,z,z,z};

  shape = 0.0; 
    
  int i, j, k, l;
  int ii = 0; 
    
  const FACE * faces = ElementTopology::GetFaces (ET_PRISM);

  // trig faces
  for (i = 0; i < 2; i++)
    for (int l = 0; l < 3; l++, ii++)
      shape(ii,2) = lami[faces[i][l]].Value() * lamiz[faces[i][l]].Value();

	
  /*
  // quad faces
  for (i = 2; i < 5; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;
	
      int fz =    faces[i][3-fmax]; 
      int ftrig = faces[i][fmax^1]; 
      fmax = faces[i][fmax];

      AutoDiffDiff<2> ls = lami[fmax];
      AutoDiffDiff<2> le = lami[ftrig];
      double lz = lamiz[fmax].Value() - lamiz[fz].Value();


      AutoDiffDiff<2> pot = ls * le;

      for (int l = 0; l <= order_facet; l++)
	{
	  double facz = 1;

	  for (k = 0; k <= order_facet; k++)
	    {
	      shape(ii,0) = facz * pot.DDValue(1,1);
	      shape(ii,1) = facz * pot.DDValue(0,0);
	      shape(ii,3) = -facz * pot.DDValue(1,0);
	    
	      facz *= 1-2*lz;
	      ii++;
	    }
	  pot *= (ls - le);
	}
    }
  */


  // quad faces
  for (i = 2; i < 5; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;
	
      int fz =    faces[i][3-fmax]; 
      int ftrig = faces[i][fmax^1]; 
      fmax = faces[i][fmax];

      AutoDiffDiff<2> ls = lami[fmax];
      AutoDiffDiff<2> le = lami[ftrig];
      AutoDiffDiff<1> lz = lamiz[fmax] - lamiz[fz];

      AutoDiffDiff<2> pot = ls * le;

      for (int l = 0; l <= order_facet; l++)
	{
	  AutoDiffDiff<1> facz = lz * (1-lz);

	  for (k = 0; k <= order_facet; k++)
	    {
	      shape(ii,0) = facz.DDValue(0,0) * pot.DDValue(1,1);
	      shape(ii,1) = facz.DDValue(0,0) * pot.DDValue(0,0);
	      shape(ii,3) = -facz.DDValue(0,0) * pot.DDValue(1,0);
	    
	      facz *= 1-2*lz;
	      ii++;
	    }
	  pot *= (ls - le);
	}
    }






  // inner shapes, sigma_xy - tensor,  3 x 5
  

  AutoDiffDiff<2> ddbub = ddlami[0] * ddlami[1] * ddlami[2];
  ArrayMem<double, 10>  polx(order_inner+1), poly(order_inner+1);

  ScaledLegendrePolynomial (order_inner, 1.0 - 2*x.Value() - y.Value(), 1-y.Value(), polx);
  LegendrePolynomial (order_inner, 1.0 - 2*y.Value(), poly);


  for (int l = 0; l < 3; l++)
    {
      AutoDiffDiff<2> bub = ddlami[l] * ddlami[(l+1)%3];
      double lami = ddlami[(l+2)%3].Value();
      
      for (int i = 0; i <= order_inner-1; i++)
	for (int j = 0; i+j <= order_inner-1; j++)
	  {
	    double fac = lami * polx[i] * poly[j];

	    for (k = 0; k <= orderz; k++)
	      {
		shape(ii,0) = bub.DDValue(1,1) * fac;
		shape(ii,1) = bub.DDValue(0,0) * fac;
		shape(ii,3) = -bub.DDValue(1,0) * fac;
		fac *= 1-2*z.Value();
		ii++;
	      }
	  }
    }



  // sigma_xz, sigma_yz,  2 x (orderz+1) * (order_inner+1) * (order_inner+2) / 2
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      {
	double pot = polx[i] * poly[j];
	  
	for (int k = 0; k < orderz; k++)
	  {
	    shape(ii++,4) = pot;
	    shape(ii++,5) = pot;
	    pot *= 1-2*z.Value();
	  }
      }



  // sigma_zz,   6
  for (int i = 0; i <= order_inner; i++)
    for (int j = 0; i+j <= order_inner; j++)
      {
	double pot = polx[i] * poly[j] * z.Value()*(1-z.Value());
	  
	for (int k = 0; k < orderz-1; k++)
	  {
	    shape(ii++,2) = pot;
	    pot *= 1-2*z.Value();
	  }
      }
}











HDivSymHex :: HDivSymHex (int aorder, int aorder_facet) 
  : HDivSymFiniteElement<3> (ET_HEX)
{
  order = aorder+3;
  order_facet = aorder_facet;
  order_inner = aorder;

  ninner = 0;
  ndof = 6 * sqr(order_facet+1) + ninner;
}


void HDivSymHex :: GetInternalDofs (ARRAY<int> & idofs) const
{
  idofs.SetSize(0);
  for (int i = ndof-ninner; i < ndof; i++)
    idofs.Append(i);
}


void HDivSymHex :: CalcShape (const IntegrationPoint & ip, 
				FlatMatrixFixWidth<6> shape) const   // n * 3 
{
  AutoDiffDiff<3> x(ip(0),0); 
  AutoDiffDiff<3> y(ip(1),1); 
  AutoDiffDiff<3> z(ip(2),2);

  AutoDiff<1> x1(ip(0),0); 
  AutoDiff<1> y1(ip(1),0); 
  AutoDiff<1> z1(ip(2),0);

  AutoDiff<3> x2(ip(0),0); 
  AutoDiff<3> y2(ip(1),1); 
  AutoDiff<3> z2(ip(2),2);

  AutoDiffDiff<3> lami[8]={(1-x)*(1-y)*(1-z),x*(1-y)*(1-z),x*y*(1-z),(1-x)*y*(1-z),
			   (1-x)*(1-y)*z,x*(1-y)*z,x*y*z,(1-x)*y*z}; 
  AutoDiff<1> sigma[8]={(1-x1)+(1-y1)+(1-z1),x1+(1-y1)+(1-z1),x1+y1+(1-z1),(1-x1)+y1+(1-z1),
			(1-x1)+(1-y1)+z1,x1+(1-y1)+z1,x1+y1+z1,(1-x1)+y1+z1}; 
  AutoDiff<3> sigma2[8]={(1-x2)+(1-y2)+(1-z2),x2+(1-y2)+(1-z2),x2+y2+(1-z2),(1-x2)+y2+(1-z2),
			 (1-x2)+(1-y2)+z2,x2+(1-y2)+z2,x2+y2+z2,(1-x2)+y2+z2}; 

  shape = 0.0; 
    
  int ii = 0; 
    
  const FACE * faces = ElementTopology::GetFaces (ET_HEX);

  ArrayMem<AutoDiffDiff<1>,10> polx(order_facet+1);
  ArrayMem<AutoDiffDiff<1>,10> poly(order_facet+1);

  // quad faces
  for (int i = 0; i < 6; i++)
    {
      int fmax = 0;
      for(int j = 1; j < 4; j++) 
	if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;

      int f1 = faces[i][(fmax+3)%4];
      int f2 = faces[i][(fmax+1)%4]; 
      fmax = faces[i][fmax]; 
      
      if(vnums[f2]>vnums[f1]) swap(f1,f2);  // fmax > f1 > f2 

      AutoDiffDiff<1> xi  = sigma[fmax] - sigma [f1];
      AutoDiffDiff<1> eta = sigma[fmax] - sigma [f2];

      AutoDiff<3> nv = Cross (sigma2[fmax]-sigma2[f1], sigma2[fmax]-sigma2[f1]);

      polx[0] = 1 - xi * xi;
      poly[0] = 1 - eta * eta;
      for (int k = 1; k <= order_facet; k++)
	{
	  polx[k] = polx[k-1] * xi;
	  poly[k] = poly[k-1] * eta;
	}

      AutoDiffDiff<3> lam_f = 0;
      for (int j = 0; j < 4; j++)
	lam_f += lami[faces[i][j]];

      for (int k = 0; k <= order_facet; k++) 
	{
	  double pxl = polx[k].DDValue(0,0) * lam_f.Value();
	  for (int j = 0; j <= order_facet; j++) 
	    {
	      double val = pxl * poly[j].DDValue(0);

	      shape(ii,0) += nv.DValue(0) * val;
	      shape(ii,1) += nv.DValue(1) * val;
	      shape(ii,2) += nv.DValue(2) * val;
	      ii++;
	    }
	}
    }
}






















void HDivSymNormalFiniteElement :: 
SetVertexNumbers (FlatArray<int> & avnums)
{
  for (int i = 0; i < avnums.Size(); i++)
    vnums[i] = avnums[i];
}




HDivSymNormalSegm :: HDivSymNormalSegm (int aorder)
{
  order = aorder;
  eltype = ET_SEGM;
  ndof = aorder+1; 
}

void HDivSymNormalSegm :: CalcShape (const IntegrationPoint & ip, 
				  FlatVector<> shape) const   
{
  AutoDiff<1> x(ip(0),0); 
  AutoDiff<1> lami[3] = {x,1-x};
  AutoDiffDiff<1> ddlami[3] = {x,1-x};
    
  int es = 0, ee = 1;
  if (vnums[es] > vnums[ee])  swap (es, ee);

  AutoDiffDiff<1> xi = lami[ee]-lami[es];

  ArrayMem<AutoDiffDiff<1>, 20> ha(order+1);
  IntegratedLegendreMonomialExt::Calc (order+2, xi, ha);

  for (int l = 0; l <= order; l++)
    shape(l) = ha[l].DDValue(0,0);
}




HDivSymNormalTrig :: HDivSymNormalTrig (int aorder)
{
  order = aorder;
  eltype = ET_TRIG;
  ndof = (aorder+1)*(aorder+2)/2;
}

void HDivSymNormalTrig :: CalcShape (const IntegrationPoint & ip, 
				  FlatVector<> shape) const   
{
  shape = 0.0;

  double li[3] = { ip(0), ip(1), 1-ip(0)-ip(1) };

  int fav[3] =  { 0, 1, 2 };
  
  //Sort vertices  first edge op minimal vertex
  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);
  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]);


  ArrayMem<double,20> polxi(order+1), poleta(order+1);

  ScaledLegendrePolynomial (order, li[fav[1]]-li[fav[2]], li[fav[1]]+li[fav[2]], polxi);
  LegendrePolynomial (order, 2*li[fav[0]]-1, poleta);
  // ScaledLegendrePolynomial (order, li[0]-li[1], li[0]+li[1], polxi);
  // LegendrePolynomial (order, 2*li[2]-1, poleta);

  int ii = 0;
  for (int i = 0; i <= order; i++)
    for (int j = 0; j+i <= order; j++)
      shape(ii++) = 2*polxi[i] * poleta[j];

  /*
  shape(0) = ip(0);
  shape(1) = ip(1);
  shape(2) = 1-ip(0)-ip(1);
  */
}





HDivSymNormalQuad :: HDivSymNormalQuad (int aorder)
{
  order = aorder;
  eltype = ET_QUAD;
  ndof = sqr(order+1);
}

void HDivSymNormalQuad :: CalcShape (const IntegrationPoint & ip, 
				     FlatVector<> shape) const   
{
  ArrayMem<double,10> polx(order+1), poly(order+1);

  LegendrePolynomial (order, 2*ip(0)-1, polx);
  LegendrePolynomial (order, 2*ip(1)-1, poly);

  int ii = 0;
  for (int i = 0; i <= order; i++)
    for (int j = 0; j <= order; j++)
      shape(ii++) = polx[i] * poly[j];
}







