#include <solve.hpp>

#include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;




  // from contact module ...
  extern void ComputeDistanceFunction (const MeshAccess & ma,
				       FlatVector<double> & dist);

  template <int DIM>
  void FindOverlap (const MeshAccess & ma,
		    FESpace & fes,
		    FlatVector<Vec<DIM> > & u,
		    FlatVector<double> & dist);




  template <int D, class SPEC>
  class Material
  {
  public:
    enum { DIM_DMAT = D * (D+1) / 2 }; 

    SPEC & Spec() { return static_cast<SPEC&> (*this); }
    const SPEC & Spec() const { return static_cast<const SPEC&> (*this); }

    
    void ComputeStress (const SpecificIntegrationPoint<D,D> & sip,
			const Mat<D> & strain, Mat<D> & stress) const
    {
      Mat<D> strain1, strain2;
      Mat<D> stress2;
      double eps = 1e-8;
      for (int i = 0; i < D; i++)
	for (int j = 0; j < D; j++)
	  {
	    strain1 = strain;
	    strain2 = strain;
	    strain1(i,j) += eps;
	    strain2(i,j) -= eps;
	    stress (i,j) = 
	      (Spec().Energy (sip, strain1) - Spec().Energy (sip, strain2)) 
	      / (2*eps);
	  }

      /*eps = 1e-9;
      for (int i = 0; i < D; i++)
	for (int j = 0; j < D; j++)
	  {
	    strain1 = strain;
	    strain2 = strain;
	    strain1(i,j) += eps;
	    strain2(i,j) -= eps;
	    stress2 (i,j) = 
	      (Spec().Energy (sip, strain1) - Spec().Energy (sip, strain2)) 
	      / (2*eps);
	  }
      testout->precision(16);
      (*testout) << "stress1=" << stress << ", stress2=" << stress2 << endl;
      testout->precision(6);
      */
    }


    void ComputeDMatrix (const SpecificIntegrationPoint<D,D> & sip,
			 const Mat<D> & strain, Mat<DIM_DMAT> & dmat) const
    {
      Mat<D> strain1, strain2, strain3, strain4;
      Mat<D*D> hesse;
      int i, j;

      double eps = 1e-5;
      for (int i1 = 0; i1 < D; i1++)
	for (int j1 = 0; j1 < D; j1++)
	  for (int i2 = 0; i2 < D; i2++)
	    for (int j2 = 0; j2 < D; j2++)
	      {
		strain1 = 
		strain2 = 
		strain3 = 
		strain4 = strain;
		
		strain1(i1,j1) += eps;
		strain2(i1,j1) += eps;
		strain3(i1,j1) -= eps;
		strain4(i1,j1) -= eps;

		strain1(i2,j2) += eps;
		strain2(i2,j2) -= eps;
		strain3(i2,j2) += eps;
		strain4(i2,j2) -= eps;

		hesse(D*i1+j1, D*i2+j2) = 
		  (Spec().Energy (sip, strain1) + Spec().Energy (sip, strain4) -
		   Spec().Energy (sip, strain2) - Spec().Energy (sip, strain3))
		  / (4*eps*eps);
	      }

      Mat<D*D, DIM_DMAT> trans;
      trans = 0;
      for (i = 0; i < D; i++)
	trans(i*(D+1), i) = 1;
      int ei = D;
      for (i = 0; i < D; i++)
	for (j = 0; j < i; j++)
	  {
	    trans(i*D+j, ei) = 0.5;
	    trans(j*D+i, ei) = 0.5;
	    ei++;
	  }
      dmat = Trans (trans) * hesse * trans;
    }
  };


  template <int D>
  class LinearMaterial : public Material<D,LinearMaterial<D> >
  {
  public:
    enum { DIM_DMAT = D * (D+1) / 2 }; 

    CoefficientFunction * coefe;
    CoefficientFunction * coefnu;

    LinearMaterial (CoefficientFunction * acoefe,
		    CoefficientFunction * acoefnu)
      : coefe(acoefe), coefnu(acoefnu) 
    {
      ;
    }

		    
    double Energy (const SpecificIntegrationPoint<D,D> & sip,
		   const Mat<D> & strain) const
    {
      double nu = Evaluate (*coefnu, sip);
      double e = Evaluate (*coefe, sip);

      // 2nd Piola-Kirchhoff (Braess, p251)
      Mat<D,D,double> piola2 = (e/(1+nu)) * strain 
	+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();
      
      double energy = 0;
      for (int j = 0; j < D; j++)
	for (int k = 0; k < D; k++)
	  energy += 0.5 * piola2(j,k) * strain(j,k);      
      return energy;
    }

    void ComputeStress (const SpecificIntegrationPoint<D,D> & sip,
			const Mat<D> & strain, Mat<D> & stress) const
    {
      double nu = Evaluate (*coefnu, sip);
      double e = Evaluate (*coefe, sip);

      // 2nd Piola-Kirchhoff (Braess, p251)
      stress = (e/(1+nu)) * strain 
			+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();
    }

    void ComputeDMatrix (const SpecificIntegrationPoint<D,D> & sip,
			 const Mat<D> & strain, Mat<DIM_DMAT> & dmat) const
    {
      //      Material<D,LinearMaterial<D> >::ComputeDMatrix (sip, strain, dmat);
      dmat = 0;
      
      double nu = Evaluate (*coefnu, sip);
      double e = Evaluate (*coefe, sip);

      int j, k;
      for (j = 0; j < D; j++)
	{
	  dmat(j,j) = 1-nu;
	  for (k = 0; k < j; k++)
	    dmat(j,k) = dmat(k,j) = nu;
	}
      for (j = D; j < DIM_DMAT; j++)
	dmat(j,j) = 0.5 * (1-2*nu);
      dmat *= (e / ((1 + nu) * (1 - 2 * nu)));
    }
  };







  template <int D>
  class NeoHookTypeMaterial : public Material<D,NeoHookTypeMaterial<D> >
  {
  public:
    CoefficientFunction * coefe;
    CoefficientFunction * coefnu;

    NeoHookTypeMaterial (CoefficientFunction * acoefe,
			 CoefficientFunction * acoefnu)
      : coefe(acoefe), coefnu(acoefnu) { ; }
    
    double Energy (const SpecificIntegrationPoint<D,D> & sip,
		   const Mat<D> & strain) const
    {
      double nu = Evaluate (*coefnu, sip);
      double e = Evaluate (*coefe, sip);
      double g = e/(2*(1+nu)); //shear modulus
      double km;
      if (1-2*nu > 0) 
        {km = e/(3*(1-2*nu));}//compression module
      else 
	{km = 1e24; (*testout) << "WARNING: compression module is infinite or negative!!!" << endl;}

      /*
      //joachim
      Mat<D> s1 = strain + Id<D>();
      double lam3 = Det (s1); 
      if (lam3 = 0) return 1e24;
      
      double lam2 = Trace (Trans (strain) * strain);
      return e * lam2 + e * (lam3 + 1/lam3 - 2); //joachim inkompressible material?
      */

      Mat<D> C = 2.*strain + Id<D>();
	        
      double I1 = Trace (C);
      double I2 = 0.5*(I1*I1-Trace (Trans (C) * C));
      double I3 = Det(C);
      if (I3 <= 0) return 1e24;
     
      double c1q = 0.75;
      double c2q = 1-c1q;
      double c1 = g/2.*c1q;
      double c2 = g/2.*c2q;
      double j1 = I1*pow(I3,-1./(double)D);
      double j2 = I2*pow(I3,-2./(double)D);
      double j3 = sqrt(I3);

      //(*testout) << "strain=" << strain << ", C=" << C << ", j1=" << j1 << ", I3=" << I3 << endl;
      
      //      (*testout) << "lam1 = " << lam1 << endl;
      //      (*testout) << "lam2 = " << lam2 << endl;
      
      return c1*(j1-D)+c2*(j2-D)+0.5*km*(j3-1)*(j3-1); //Gummi, beinahe inkompr., Bathe S.701
      //return c1*(j1-(double)D); //Neo-hooke
      
    }
  };







  /* ****************  NonlinearElasticityIntegrator *********+ */


  template <int D>
  class NonlinearElasticityIntegrator : public BilinearFormIntegrator
  {
    typedef LinearMaterial<D> MATERIAL;
    //typedef NeoHookTypeMaterial<D> MATERIAL;
    enum { DIM_DMAT = D * (D+1) / 2 }; 

    double sigmay;

    GridFunction * gfu;
    GridFunction * gfplast;

    MATERIAL material;
    ElasticityIntegrator<D> * linelast;
  public:
    ///
    NonlinearElasticityIntegrator (CoefficientFunction * coefe,
				   CoefficientFunction * coefnu)
      : material (coefe, coefnu) 
    { 
      gfu = 0;
      gfplast = 0;
      linelast = new ElasticityIntegrator<D> (coefe, coefnu);
      sigmay = 0.05;
    }
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new NonlinearElasticityIntegrator (coeffs[0], coeffs[1]);
    }

    virtual bool BoundaryForm () const { return 0; }

    void SetGridFunctionU (GridFunction * agfu)
    {
      gfu = agfu;
    }

    void SetGridFunctionPlast (GridFunction * agfplast)
    {
      gfplast = agfplast;
    }
	        
    virtual void
    AssembleElementMatrix (const FiniteElement & bfel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      int i, j, k, l, m, n;
      if (!gfu)
	{
	  linelast -> AssembleElementMatrix (bfel, eltrans, elmat, lh);
	  return;
	}


      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int nd = fel.GetNDof ();

      elmat.AssignMemory (D*nd, D*nd, lh);
      elmat = 0;

      const FESpace & fes = gfu->GetFESpace();
      int elnr = eltrans.GetElementNr();
      const FiniteElement & el = fes.GetFE(elnr);

      ARRAY<int> dnums (nd); // , (int*)locheap.Alloc (nd*sizeof(int)));
      fes.GetDofNrs (elnr, dnums);

      FlatVector<double> elu (D*nd, lh);
      
      gfu->GetVector().GetIndirect (dnums, elu);
      fes.TransformVec (elnr, false, elu, TRANSFORM_SOL);

      Mat<D> elplast;
      elplast = Id<D>();
      FlatVector<double> velplast(D*D, lh);
      ARRAY<int> dnumsplast(1);
      dnumsplast[0] = elnr;
      if (gfplast)
	{
	  gfplast->GetVector().GetIndirect (dnumsplast, velplast);
	  for (i = 0; i < D; i++)
	    for (j = 0; j < D; j++)
	      elplast(i, j) = velplast(D*i+j);
	}

      FlatMatrix<double> melu(nd, D, 
			      const_cast<double*>
			      (reinterpret_cast<const double*>
			       (elu.Data())));

      int order = 2 * (fel.Order()-1);
      const IntegrationRule & ir =
	GetIntegrationRules().SelectIntegrationRule (fel.ElementType(), order);

      for (i = 0; i < ir.GetNIP(); i++)
	{
	  const IntegrationPoint & ip = ir.GetIP(i);
	  SpecificIntegrationPoint<D,D> sip (ir, i, eltrans, lh);
	  double fac = fabs (sip.GetJacobiDet()) * ip.Weight();


	  // gradient in reference coordinates:
	  Mat<D> graduref = Trans (melu) * fel.GetDShape(ip, lh);
	  
	  // gradient in real coordinates:
	  Mat<D> gradu = graduref * sip.GetJacobianInverse();
	  
	  // deformation gradient:

	  Mat<D> matf_tot = gradu + Id<D>();
	  Mat<D> matf = matf_tot * elplast;

	  //	  Mat<D> matf = gradu + Id<D>();

	  Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());


	  FlatMatrix<> grad1 (D, nd, lh);
	  FlatMatrix<> grad(D,nd,lh);
	  grad1 = Trans (sip.GetJacobianInverse ()) * Trans (fel.GetDShape(sip.IP(), lh));

	  FlatMatrix<double> bmat(DIM_DMAT, D*nd, lh);
	  bmat = 0;
	  Mat<D> gradv;

	  /*
	  for (j = 0; j < D; j++)
	    for (l = 0; l < nd; l++)
	      {
		gradv = 0;
		for (k = 0; k < D; k++)
		  gradv(j,k) = grad1(k,l);
		grad = Trans (matf) * gradv;

		if (D == 2)
		{
		bmat(0, D*l+j) = grad(0,0);
		bmat(1, D*l+j) = grad(1,1);		
		bmat(2, D*l+j) = grad(0,1)+grad(1,0);
	        }
		else if (D == 3)
		{
		  bmat(0, D*l+j) = grad(0,0);
		  bmat(1, D*l+j) = grad(1,1);		
		  bmat(2, D*l+j) = grad(2,2);		
		  bmat(3, D*l+j) = grad(0,1)+grad(1,0);
		  bmat(4, D*l+j) = grad(0,2)+grad(2,0);
		  bmat(5, D*l+j) = grad(1,2)+grad(2,1);
		}
	      }
	  */

	  if (D == 2)
	    {
	      for (j = 0; j < D; j++)
		for (l = 0; l < nd; l++)
		  {
		    gradv = 0;
		    for (k = 0; k < D; k++)
		      gradv(j,k) = grad1(k,l);
		    grad = Trans (matf) * gradv;
		    
		    bmat(0, D*l+j) = grad(0,0);
		    bmat(1, D*l+j) = grad(1,1);		
		    bmat(2, D*l+j) = grad(0,1)+grad(1,0);
		  }
	    }
	  else if (D == 3)
	    {
	      for (j = 0; j < D; j++)
		for (l = 0; l < nd; l++)
		  {
		    gradv = 0;
		    for (k = 0; k < D; k++)
		      gradv(j,k) = grad1(k,l);
		    grad = Trans (matf) * gradv;

		    bmat(0, D*l+j) = grad(0,0);
		    bmat(1, D*l+j) = grad(1,1);		
		    bmat(2, D*l+j) = grad(2,2);		
		    bmat(3, D*l+j) = grad(0,1)+grad(1,0);
		    bmat(4, D*l+j) = grad(0,2)+grad(2,0);
		    bmat(5, D*l+j) = grad(1,2)+grad(2,1);
		  }
	    }

	  /*
	  grad = matf * grad1;

	  FlatMatrix<double> bmat(DIM_DMAT, D*nd, lh);
	  bmat = 0;
	  for (j = 0; j < D; j++)
	    for (k = 0; k < nd; k++)
	      bmat(j, D*k+j) = grad(j,k);

	  int ei = D;
	  for (j = 0; j < D; j++)
	    for (k = 0; k < j; k++)
	      {
		for (l = 0; l < nd; l++)
		  {
		    bmat(ei, D*l+j) = grad(k,l);
		    bmat(ei, D*l+k) = grad(j,l);
		  }
		ei++;
		}
	  */


	  Mat<DIM_DMAT> dmat;
	  material.ComputeDMatrix (sip, strain, dmat);


	  FlatMatrix<double> dbmat(DIM_DMAT, D*nd, lh);
	  dbmat = dmat * bmat;
	  
	  elmat += fac * (Trans (bmat) * dbmat);
	  
	  
	  Mat<D> piola2;
	  material.ComputeStress (sip, strain, piola2);

	  /*
	  if (eltrans.GetElementNr() == 0)
	    {
	      (*testout) << "matf = " << endl << matf << endl;
	      (*testout) << "piola2 = " << endl << piola2 << endl;
	      (*testout) << "elmat, bef = " << endl << elmat << endl;
	    }
	  */

	  for (j = 0; j < nd; j++)
	    for (k = 0; k < nd; k++)
	      for (l = 0; l < D; l++)
		for (m = 0; m < D; m++)
		  {
		    double val = (fac * piola2(l,m) * grad1(l,j) * grad1(m,k));
		    for (n = 0; n < D; n++)
		      {
			elmat(j*D+n, k*D+n) += val;
		      }
		  }

	  /*
	  if (eltrans.GetElementNr() == 0)
	    {
	      (*testout) << "elmat, after = " << endl << elmat << endl;
	    }
	  */
	}
      /*

      if (eltrans.GetElementNr() == 0)
	(*testout) << "elmat1 = " << endl << elmat << endl;
      

      // finite differences:
      FlatMatrix<double> elmat2 (D*nd, D*nd, lh);
      FlatVector<double> hx1 (D*nd, lh);
      FlatVector<double> hx2 (D*nd, lh);
      FlatVector<double> hy1 (D*nd, lh);
      FlatVector<double> hy2 (D*nd, lh);
      FlatVector<double> hy (D*nd, lh);

      double dx = 1e-10 + 1e-5 * L2Norm (elu);
      for (int i = 0; i < D*nd; i++)
	{
	  hx1 = elu;
	  hx2 = elu;
	  hx1(i) += dx;
	  hx2(i) -= dx;

	  ApplyElementMatrix (fel, eltrans, hx1, hy1, lh);
	  ApplyElementMatrix (fel, eltrans, hx2, hy2, lh);

	  hy = (0.5 / dx ) * (hy1 - hy2);
	  for (int j = 0 ; j < D*nd; j++)
	    elmat2(i,j) = hy(j);
	}

      if (eltrans.GetElementNr() == 0)
	(*testout) << "elmat, version2 = " << endl << elmat2 << endl;

      elmat -= elmat2;
      (*testout) << "diff = " << Trace (Trans (elmat) * elmat) << endl;

      elmat = elmat2;
*/
    }



    virtual void 
    ApplyElementMatrix (const FiniteElement & bfel, 
			const ElementTransformation & eltrans, 
			const FlatVector<double> & elx, 
			FlatVector<double> & ely,
			LocalHeap & lh) const
    {
      int i, j;

      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int ndof = fel.GetNDof ();
    
      ely = 0;
      
      FlatMatrix<double> melx(ndof, D, 
			      const_cast<double*>
			      (static_cast<const double*>
			       (elx.Data())));

      FlatMatrix<double> mely(ndof, D, 
			      static_cast<double*>
			       (ely.Data()));

      int order = 2 * (fel.Order()-1);
      const IntegrationRule & ir =
	GetIntegrationRules().SelectIntegrationRule (fel.ElementType(), order);


      Mat<D> elplast = Id<D>();   // inverse of plastic deformation
      FlatVector<double> velplast(D*D, lh);
      ARRAY<int> dnumsplast(1);
      dnumsplast[0] = eltrans.GetElementNr();
      if (gfplast)
	{
cout << "plast" << endl;
	  gfplast->GetVector().GetIndirect (dnumsplast, velplast);
	  for (i = 0; i < D; i++)
	    for (j = 0; j < D; j++)
	      elplast(i, j) = velplast(D*i+j);
	}

      for (i = 0; i < ir.GetNIP(); i++)
	{
	  const IntegrationPoint & ip = ir.GetIP(i);
	  
	  SpecificIntegrationPoint<D,D> sip (ir, i, eltrans, lh);

	  // gradient in reference coordinates:
	  Mat<D,D,double> gradref = Trans (melx) * fel.GetDShape(ip, lh);

	  // gradient in real coordinates:
	  Mat<D> grad = gradref * sip.GetJacobianInverse();
	  
	  // deformation gradient:
	  
	  Mat<D> matf_tot = grad + Id<D>();
	  Mat<D> matf = matf_tot * elplast;
      	  	
	  //Mat<D> matf = grad + Id<D>();

	  // Green-Lagrange strain tensor
	  Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());

	  Mat<D> piola2;
	  material.ComputeStress (sip, strain, piola2);

	  Mat<D> ppiola2 = elplast * piola2;
	  piola2 = ppiola2 * Trans (elplast);

	  Mat<D> piola1 = matf * piola2;

	  double fac = fabs (sip.GetJacobiDet()) * ip.Weight();

	  // test with nabla v:
	  Mat<D,D,double> trans = sip.GetJacobianInverse() * Trans (piola1);
	  mely += fac * (fel.GetDShape(sip.IP(),lh) * trans);
	}
      //(*testout) << "elx=" << elx << endl;
      //(*testout) << "ely=" << ely << endl;
    }
    






    virtual double 
    Energy (const FiniteElement & bfel, 
	    const ElementTransformation & eltrans, 
	    const FlatVector<double> & elx, 
	    LocalHeap & lh) const
    {
      int i, j;

      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int ndof = fel.GetNDof ();
    
      
      FlatMatrix<double> melx(ndof, D, 
			      const_cast<double*>
			      (static_cast<const double*>
			       (elx.Data())));
      
      int order = 2 * (fel.Order()-1);
      const IntegrationRule & ir =
	GetIntegrationRules().SelectIntegrationRule (fel.ElementType(), order);
      /*
      int order = IntegrationOrder (fel);
      
      const IntegrationRule & ir = (Lumping() && fel.Order() == 1 && 0) ?
	GetIntegrationRules().SelectLumpingIntegrationRule (fel.ElementType()) :
	GetIntegrationRules().SelectIntegrationRule (fel.ElementType(), order);
      */
      double energy = 0;



      Mat<D> elplast = Id<D>();
      FlatVector<double> velplast(D*D, lh);
      ARRAY<int> dnumsplast(1);
      dnumsplast[0] = eltrans.GetElementNr();
      if (gfplast)
	{
	  gfplast->GetVector().GetIndirect (dnumsplast, velplast);
	  for (i = 0; i < D; i++)
	    for (j = 0; j < D; j++)
	      elplast(i, j) = velplast(D*i+j);
	}


      for (int i = 0; i < ir.GetNIP(); i++)
	{
	  const IntegrationPoint & ip = ir.GetIP(i);
	  
	  SpecificIntegrationPoint<D,D> sip (ir, i, eltrans, lh);

	  // gradient in reference coordinates:
	  Mat<D> gradref = Trans (melx) * fel.GetDShape(ip, lh);

	  // gradient in real coordinates:
	  Mat<D> grad = gradref * sip.GetJacobianInverse();
	  
	  /*
	  // deformation gradient:
	  Mat<D> matf_tot = grad + Id<D>();
	  Mat<D> matf = matf_tot * elplast;
	  */
	  Mat<D> matf = grad + Id<D>();

	  // Green-Lagrange strain tensor
	  Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());

	  double fac = fabs (sip.GetJacobiDet()) * ip.Weight();

	  energy += fac * material.Energy (sip, strain);
	}     
      return energy;
    }



    virtual void
    CalcFlux (const FiniteElement & bfel,
	      const ElementTransformation & eltrans,
	      const IntegrationPoint & ip,
	      const FlatVector<double> & elx, 
	      FlatVector<double> & flux,
	      bool applyd,
	      LocalHeap & lh) const
    {
      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int ndof = fel.GetNDof ();
      
      FlatMatrix<double> melx(ndof, D, 
			      const_cast<double*>
			      (static_cast<const double*>
			       (elx.Data())));

	  
      SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
      
      // gradient in reference coordinates:
      Mat<D> gradref = Trans (melx) * fel.GetDShape(ip, lh);
      
      // gradient in real coordinates:
      Mat<D> grad = gradref * sip.GetJacobianInverse();
      
      // deformation gradient:
      Mat<D> matf = grad + Id<D>();
      
      // Green-Lagrange strain tensor
      Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());
      
      double nu = Evaluate (*material.coefnu, sip);
      double e = Evaluate (*material.coefe, sip);
      
      // 2nd Piola-Kirchhoff (Braess, p251)
      Mat<D> piola2 = (e/(1+nu)) * strain 
	+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();

      Mat<D> piola1 = matf * piola2;
      Mat<D> stress = piola1 * Trans (matf);

      if (D == 2)
      {

        flux (0) = stress(0,0);
        flux (1) = stress(1,1);
        flux (2) = stress(0,1);
      }
      else if (D == 3)
      {
        flux (0) = stress(0,0);
        flux (1) = stress(1,1);
        flux (2) = stress(2,2);
        flux (3) = stress(0,1);
        flux (4) = stress(0,2);
        flux (5) = stress(1,2);
      }
      
    }




    virtual void
    CalcDissipation (const FiniteElement & bfel,
		     const ElementTransformation & eltrans,
		     const IntegrationPoint & ip,
		     const FlatVector<double> & elx, 
		     Mat<D> & plastprime,
		     LocalHeap & lh) const
    {
      int i, j;

      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int ndof = fel.GetNDof ();
      

      Mat<D> elplast = Id<D>();
      FlatVector<double> velplast(D*D, lh);
      ARRAY<int> dnumsplast(1);
      dnumsplast[0] = eltrans.GetElementNr();
      if (gfplast)
	{
	  gfplast->GetVector().GetIndirect (dnumsplast, velplast);
	  for (i = 0; i < D; i++)
	    for (j = 0; j < D; j++)
	      elplast(i, j) = velplast(D*i+j);
	}


      FlatMatrix<double> melx(ndof, D, 
			      const_cast<double*>
			      (static_cast<const double*>
			       (elx.Data())));

	  
      SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
      
      // gradient in reference coordinates:
      Mat<D> gradref = Trans (melx) * fel.GetDShape(ip, lh);
      
      // gradient in real coordinates:
      Mat<D> grad = gradref * sip.GetJacobianInverse();
      
      // deformation gradient:
      Mat<D> matf_tot = grad + Id<D>();
      Mat<D> matf = matf_tot * elplast;

      // Green-Lagrange strain tensor
      Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());
      
      double nu = Evaluate (*material.coefnu, sip);
      double e = Evaluate (*material.coefe, sip);
      
      // 2nd Piola-Kirchhoff (Braess, p251)
      Mat<D> piola2 = (e/(1+nu)) * strain 
	+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();


      Mat<D> dev = piola2 - (Trace (piola2) / D) * Id<D> ();
      Mat<D> dd = Trans (dev) * dev;
      double normdev = sqrt (Trace (dd));

      double fac = 0;
      if (normdev > sigmay)
	fac = 1000 * (normdev - sigmay);
      
      dev *= fac;
      Mat<D> pprime = Trans (Inv (elplast) * dev);
      
      for (i = 0; i < D; i++)
	for (j = 0; j < D; j++)
	  plastprime(i*D+j) = pprime(i,j);
    }





    ///
    virtual string Name () const { return "NonlinearElasticity"; }
  };










  ///
  template <int DIM>
  class NonlinearElasticity : public NumProc
  {
  protected:
    ///
    BilinearForm *bfa, *bfm;
    ///
    LinearForm * lff;
    ///
    GridFunction * gfu;
    GridFunction * gfplast;
    GridFunction * gfstress;
    GridFunction * gfstrain;
    ///
    Preconditioner * pre;
    ///
    int loadsteps;
    int newtonsteps; 
    double damping;
    enum SOLVER { STATIC_SOLVER, QUASI_STATIC_SOLVER, DYNAMIC_SOLVER };
    SOLVER solver;
    bool contact;
    bool plasticity;
    double tend, dt;
  public:
    ///
    NonlinearElasticity (PDE & apde, const Flags & flags);
    ///
    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      if (pde.GetMeshAccess().GetDimension() == 2)
	return new NonlinearElasticity<2> (pde, flags);
      else
	return new NonlinearElasticity<3> (pde, flags);
    }
    ///
    virtual void Do();

    virtual string GetClassName () const
    {
      if (DIM == 2)
	return "NonlinearElasticity 2D";
      else
	return "NonlinearElasticity 3D";
    }

    void ComputeU(const FlatVector<double> & point, 
	      	  FlatVector<double> & up,
 		  const S_GridFunction<double>& bu,
      	 	  LocalHeap & lh)
    {
    
      int elnr;
      double lami[3]; //lokale koordinaten
      ARRAY<int> dnums;
      ElementTransformation eltrans;

      elnr = Ng_FindElementOfPoint (const_cast<double*>(&point(0)), lami);
    
      if (!elnr) return;
      elnr--;

      const S_GridFunction<double> & u = 
        dynamic_cast<const S_GridFunction<double>&> (bu);

      const FESpace & fes = u.GetFESpace();
      const FiniteElement & fel = fes.GetFE (elnr);
      ma.GetElementTransformation (elnr, eltrans, lh);
      fes.GetDofNrs (elnr, dnums);

      FlatVector<double> elu(dnums.Size() * fes.GetDimension(),lh);

      u.GetElementVector (dnums, elu);
    //fes.TransformVec (elnr, 0, elu, TRANSFORM_SOL);

	
      if (fel.ElementType() == ET_TRIG)
        {
	  double hlam1 = lami[0];
  	  double hlam2 = lami[1];
	  lami[0] = 1-hlam1-hlam2;
	  lami[1] = hlam1;
      	}

      if (fel.ElementType() == ET_TET)
        {
	  double hlam1 = lami[0];
	  double hlam2 = lami[1];
	  double hlam3 = lami[2];
	  lami[0] = 1-hlam1-hlam2-hlam3;
	  lami[1] = hlam1;
      	  lami[2] = hlam2;
        }

      IntegrationPoint ip(lami, 1);
      FlatVector<double> shape(dnums.Size(),lh);
    
      const NodalFiniteElement & nfel = 
  	dynamic_cast<const NodalFiniteElement &>(fel);
    
     nfel.CalcShape(ip,shape);
     //FlatVector<double> sum(D);
     up = 0;

     int i,j;        
     for (i=0; i<=DIM; i++)
      {
        for (j=0; j<=dnums.Size(); j++)
         up(i)+=shape(j)*elu(DIM*j+i);
      }
      
    }    
    
  };
  

  template <int DIM>
  NonlinearElasticity<DIM> ::   
  NonlinearElasticity (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    cout << "Create NonlinearElasticity" << endl;
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("mass", ""),1);
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    gfstress = pde.GetGridFunction ("stress");
    gfstrain = pde.GetGridFunction ("strain");
    gfplast = pde.GetGridFunction ("plast", 1);
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    loadsteps = int (flags.GetNumFlag ("loadsteps", 10));
    newtonsteps = int (flags.GetNumFlag ("newtonsteps", 10));
    damping = flags.GetNumFlag ("damping", 0.5);

    contact = flags.GetDefineFlag ("contact");
    plasticity = flags.GetDefineFlag ("plasticity");

    solver = STATIC_SOLVER;
    if (flags.GetDefineFlag ("dynamic"))
      solver = DYNAMIC_SOLVER;
    if (flags.GetDefineFlag ("quasistatic"))
      solver = QUASI_STATIC_SOLVER;
    
    dt = flags.GetNumFlag ("dt", 0.1);
    tend = flags.GetNumFlag ("tend", 1);
    cout << "dt = " << dt << endl;
  }

  template <int DIM>
  void NonlinearElasticity<DIM> :: Do()
  {
    cout << "solve nonlin elast" << endl;
    
    int i, j, k;

    BilinearFormApplication app(bfa);

    BaseVector & u = gfu->GetVector();
    BaseVector & ugood = pde.GetGridFunction("ugood")->GetVector();

    const BaseVector & f = lff->GetVector();

    BaseVector & d = *u.CreateVector();
    BaseVector & w = *u.CreateVector();
    BaseVector & au = *u.CreateVector();
    BaseVector & uold = *u.CreateVector();


    int ne = ma.GetNE();
    
    // for contact
    Vector<double> dist(ma.GetNP());
    ComputeDistanceFunction (ma, dist);
    //cout << dist << endl;
    
    cout << "has dist" << endl;

    Ng_SolutionData soldata;
    Ng_InitSolutionData (&soldata);
    
    soldata.name = "dist";
    
    Vector<double> * ddist = new Vector<double>(ma.GetNP());

    int id;
    for (id = 0; id < ma.GetNP(); id++)
      {
	(*ddist)(id) = dist(id);
	double d = dist(id);
	//(*testout) << "id=" << id << " " << d << endl;
      }

    soldata.data = static_cast<double*> (ddist->Data());
    //soldata.data = ddist->Data();
    soldata.components = 1;
    
    soldata.dist = soldata.components;
    soldata.soltype = NG_SOLUTION_NODAL;
    soldata.order = 1;

    Ng_SetSolutionData (&soldata);    


    cout << "soldata set" << endl;

    NonlinearElasticityIntegrator<DIM> & nonlin =
      dynamic_cast<NonlinearElasticityIntegrator<DIM>&> (*bfa->GetIntegrator(0));
    nonlin.SetGridFunctionU (gfu);

    cout << "has nonlin" << endl;

    if (plasticity)
      {
	nonlin.SetGridFunctionPlast (gfplast);

	FlatVector<Vec<DIM*DIM> > vplast =
	  dynamic_cast<VFlatVector<Vec<DIM*DIM> >& > (gfplast->GetVector()).FV();
	Vec<DIM*DIM> id;
	id = 0;
	for (i = 0; i < DIM; i++)
	  id(i*(DIM+1)) = 1;
	for (i = 0; i < vplast.Size(); i++)
	  vplast(i) = id;
      }

    LocalHeap lh (10000000);

    cout << "has lh" << endl;

    switch (solver)
      {
      case STATIC_SOLVER:
	{
	  cout << "static solver" << endl;
	  double t, tau, err;
	  
	  for (i = 1; i <= loadsteps; i++)
	    {
	      t = double(i) / loadsteps;
	      cout << endl << endl << endl << "t = " << t << endl;
	      int cnt = 0;
	      do
		{
		  //Newton-loop		  
		  bfa->ReAssemble(); // neue Newtonmatrix
		  pre->Update();     // neu faktorisieren, bzw. multigrid update

		  CGSolver<double> inva(bfa->GetMatrix(), pre->GetMatrix());
		  inva.SetMaxSteps (1000);
		  
		  (*testout) << "Newton-iteration " << cnt << endl;
		  
		  au = app * u;
		  d = t * f - au;   // berechne Defekt
		  w = inva * d;
		  
		  cout << "|d| = " << L2Norm (d)
		       << ", |u| = " << L2Norm (u) 
		       << ", |au| = " << L2Norm (au) 
		       << ", |f| = " << L2Norm (f) 
		       << ", |w| = " << L2Norm (w) 
		       << " (d,w) = " << InnerProduct (d, w) 
		       << endl;

		  //(*testout) << " a = " << inva << endl;
		  		  
		  cout << "CG-steps = " << inva.GetSteps() << endl;
		  uold = u;

		  double energyold = bfa->Energy (uold) - t * InnerProduct (f, uold);
		  double energy;
		  tau = 1;
		  cout << "energyold = " << energyold << endl;
		  //	    cout << "dWi/dw = " << -InnerProduct (w, d) << endl;
		  do
		    {
		      u = uold + tau * w;
		      
		      if (contact)
			FindOverlap (ma, const_cast<FESpace&> (gfu->GetFESpace()),
				     dynamic_cast<VFlatVector<Vec<DIM> >&> (u).FV(),
				     dist);
		      
		      energy = bfa->Energy (u) - t * InnerProduct (f, u);
		      cout << "tau = " << tau << "energy = " << energy << endl;
		      //		cout << "                         numerisch dWi/dw = " << (energy-energyold)/tau << endl;
		      tau *= 0.5;
		    }
		  while ( energy > energyold * (1-1e-7) && tau > 1e-6);
		  
		  err = InnerProduct (w, d);
		  
		  cout << "Newton-it " << cnt << ", err = " << err << endl;

		  CalcFlux (ma, 
      	              dynamic_cast<const S_GridFunction<double>&> (*gfu),
                      dynamic_cast<S_GridFunction<double>&> (*gfstress),
                      *bfa->GetIntegrator(0), 1, 0, -1);
		}
	      while (fabs (err) > 1e-8 && cnt++ < newtonsteps);

              pde.GetGridFunction ("ugood")->GetVector() = u;

	      /*
      	      FlatVector<double> point(DIM,lh);
  	      FlatVector<double> up(DIM,lh);

	      ofstream def("deflec");
      	      double x = 0;

	      while(x<=200)
	      {
		point(0)=x;
		point(1)=10;
		point(2)=20;
      	      	ComputeU(point, up, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
		//cout << "x=" << x << ", u=" << up(1) << endl;
		def << x << " " << up(1) << endl;
      	      	x+=1;
	      }
	      */
            
	    }
 	    	  
	    break;
	}

      case QUASI_STATIC_SOLVER:
	{
	  double t, told, err, tau;
	  
	  BaseVector & plast = gfplast->GetVector();
	  BaseVector & pold = *plast.CreateVector();
	  tau = dt;
	  u = 0;

	  t = 0;
	  while (t <= tend)
	    {
	      int cnt = 0;

	      told = t;

	      pold = plast;
	      ugood = u;

	      do
		{

		  if (tau > dt) tau = dt;

		  t = told + tau;
		  u = ugood;
		  cout << endl << endl << " t = " << t << ", tau = " << tau << endl;

		  cnt = 0;
		  do
		    { 
		      // Newton-loop
		      
		      cnt++;
		      
		      if (contact)
			FindOverlap (ma, const_cast<FESpace&> (gfu->GetFESpace()),
				     dynamic_cast<VFlatVector<Vec<DIM> >&> (u).FV(),
				     dist);
		  
		      bfa->ReAssemble(); 
		      
		      //		      d = (t * f) - app * u;
		      d = app * u - (t*f);
		      d*= -1;

		      err = L2Norm(d);
		      cout << "Newton it " << cnt << ", err = " << err << endl;
		      (*testout) << "Newton it " << cnt << ", err = " << err << endl;

		      if (err < 1e-9) break;

		      pre->Update();
		      CGSolver<double> inva(bfa->GetMatrix(), pre->GetMatrix());
		      inva.SetMaxSteps (1000);
		      
		      w = inva * d;
		      u += w;
		    }
		  while (cnt < newtonsteps);
		  cout << "newton done, cnt =" << cnt << endl;
		  (*testout) << "t = " << t << " tau = " << tau << " newton = " << cnt << " err = " << err << endl;
		  if (cnt >= newtonsteps)
		    tau *= 0.5;
		  else
		    tau *= 1.2;
		}
	      while (cnt >= newtonsteps);
	      
	      // update plastic deformation by explicit visco-plastic law:

	      CalcFlux (ma, 
			dynamic_cast<const S_GridFunction<double>&> (*gfu),
			dynamic_cast<S_GridFunction<double>&> (*gfstress),
			*bfa->GetIntegrator(0), 1, 0, -1);


	      
	      ARRAY<int> dnums, dnumsplast(1);
	      const FESpace & fes = gfu->GetFESpace();
	      for (i = 0; i < ne; i++)
		{
		  lh.CleanUp();
		  NonlinearElasticityIntegrator<DIM> & nonlin =
		    dynamic_cast<NonlinearElasticityIntegrator<DIM>&> (*bfa->GetIntegrator(0));

		  const FiniteElement & fel = fes.GetFE(i);
		  
		  ElementTransformation eltrans;
		  ma.GetElementTransformation (i, eltrans, lh);

		  const IntegrationRule & rule = GetIntegrationRules().SelectIntegrationRule (ET_TRIG, 0);

		  fes.GetDofNrs (i, dnums);
		  FlatVector<double> elu(dnums.Size() * 2, lh);

		  gfu->GetVector().GetIndirect (dnums, elu);
		  fes.TransformVec (i, 0, elu, TRANSFORM_SOL);

		  Mat<DIM> plastprime;
		  nonlin.CalcDissipation (fel, eltrans, rule.GetIP(0), elu,
					  plastprime, lh);


		  plastprime *= -tau;
		  FlatVector<double> vplastprime(4, lh);
		  for (j = 0; j < 2; j++)
		    for (k = 0; k < 2; k++)
		      vplastprime(2*j+k) = plastprime(j,k);
		  dnumsplast[0] = i;
		  gfplast->GetVector().AddIndirect (dnumsplast, vplastprime);
		}


	      // compute equivalent stress
	      /*	      
	      for (i = 0; i < 
      const NodalFiniteElement & fel = 
	dynamic_cast<const NodalFiniteElement&> (bfel);
      int ndof = fel.GetNDof ();
      
      FlatMatrix<double> melx(ndof, D, 
			      const_cast<double*>
			      (static_cast<const double*>
			       (elx.Data())));

	  
      SpecificIntegrationPoint<D,D> sip (ip, eltrans, lh);
      
      // gradient in reference coordinates:
      Mat<D> gradref = Trans (melx) * fel.GetDShape(ip, lh);
      
      // gradient in real coordinates:
      Mat<D> grad = gradref * sip.GetJacobianInverse();
      
      // deformation gradient:
      Mat<D> matf = grad + Id<D>();
      
      // Green-Lagrange strain tensor
      Mat<D> strain = 0.5 * (Trans (matf) * matf - Id<D>());
      
      double nu = Evaluate (*material.coefnu, sip);
      double e = Evaluate (*material.coefe, sip);
      
      // 2nd Piola-Kirchhoff (Braess, p251)
      Mat<D> piola2 = (e/(1+nu)) * strain 
	+ (e*nu/((1+nu)*(1-2*nu)) * Trace(strain)) * Id<D>();
	      */



	    }	      
	  break;
	}
	
      case DYNAMIC_SOLVER:
	{
	  cout << "dynamic solver" << endl;

      	  // dynamic solver
	  double t, told, err, tau;
	  
	  BaseVector & uold = *u.CreateVector();
	  BaseVector & v = *u.CreateVector();
	  BaseVector & vold = *u.CreateVector();
	  BaseVector & force = *u.CreateVector();
	  BaseVector & forceold = *u.CreateVector();

	  BaseVector & hv = *u.CreateVector();

	  const SparseMatrixSymmetric<Mat<DIM> > & matm =
	    dynamic_cast<const SparseMatrixSymmetric<Mat<DIM> > &> (bfm->GetMatrix());


	  u = 0;
	  v = 0;
	  force = app * u - f;
	  t = 0;

          ofstream def("kurbel.out");
          ofstream def2("kurbel.dat");

	  tau = dt;
	  while (t <= tend)
	    {

	      cout << endl << endl << endl << "t = " << t << endl;
	      (*testout) << "t = " << t << endl;
	      
	      told = t;
	      uold = u;
	      vold = v;
	      forceold = force;
	      
	      int cnt;
	      tau *= 5;
	      if (tau > 4*dt) tau = 4*dt;
	      
	      do
		{ // try time step tau
		  tau /= 4;
		  t = told + tau;
		  u = uold + tau * v;
		  cnt = 0;
		  do
		    {
		      // Newton-loop

		      if (contact)
			FindOverlap (ma, const_cast<FESpace&> (gfu->GetFESpace()),
				     dynamic_cast<VFlatVector<Vec<DIM> >&> (u).FV(),
				     dist);

		      bfa->ReAssemble(); // neue Newtonmatrix

		      SparseMatrixSymmetric<Mat<DIM> > & mata =
			dynamic_cast<SparseMatrixSymmetric<Mat<DIM> > &> (bfa->GetMatrix());

		      double factf = 1;
		      //if (t > 0.1) {factf = 0; cout << "hacking" << endl;}
		      force = app * u - factf*f;

		      hv = u - uold - tau * vold;
		      d = matm * hv + (tau*tau/4) * (force + forceold);

		      err = L2Norm (d);
		      cout << "t = " << t << " tau = " << tau << endl;
		      cout << "Newton-it " << cnt << ", def = " << err << endl;
		      if (err < 1e-10) break;

		      mata.AddMerge (4/tau/tau, matm);

		      pre->Update();     // neu faktorisieren, bzw. multigrid update
		      CGSolver<double> inva(bfa->GetMatrix(), pre->GetMatrix());
		      inva.SetMaxSteps (1000);

		      w = inva * d;
		      w *= 4/(tau*tau);

		      u -= w;
		      cout << "CG-steps: " << inva.GetSteps() << endl;
		    }
		  while (err > 1e-5 && cnt++ < newtonsteps);
		}
	      while (cnt >= newtonsteps);


	      v = (2/tau) * u - (2/tau) * uold - vold;

	      pde.GetGridFunction ("ugood")->GetVector() = u;

	      CalcFlux (ma,
			dynamic_cast<const S_GridFunction<double>&> (*gfu),
			dynamic_cast<S_GridFunction<double>&> (*gfstress),
			*bfa->GetIntegrator(0), 1, 0, -1);


	      //2d-pendel:
	      FlatVector<double> point1(DIM,lh);
	      FlatVector<double> point2(DIM,lh);
	      FlatVector<double> pointm(DIM,lh);
  	      FlatVector<double> up1(DIM,lh);
  	      FlatVector<double> up2(DIM,lh);
  	      FlatVector<double> upm(DIM,lh);
	      FlatVector<double> vm(DIM,lh);
	      FlatVector<double> v1(DIM,lh);
	      FlatVector<double> v2(DIM,lh);

	      point1(0)=25; point1(1)=25;
	      point2(0)=1025; point2(1)=25;
	      pointm(0)=525; pointm(1)=25;
      	      ComputeU(point1, up1, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
      	      ComputeU(point2, up2, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
      	      ComputeU(pointm, upm, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);

	      vm(0) = 0.5*(up2(0)+1000);
	      vm(1) = 0.5*(up2(1));
	      v1(0) = upm(0)+500;
	      v1(1) = upm(1);
	      v2(0) = v1(0)-vm(0);
	      v2(1) = v1(1)-vm(1);
	      double l = sqrt(vm(0)*vm(0)+vm(1)*vm(1));
	      if (l == 0) {l=1;}
	      vm(0) /= l;vm(1) /= l;
	      double vmx = vm(0);
	      vm(0) = vm(1);
	      vm(1) =-vmx;
    	      double wm = vm(0)*v2(0) + vm(1)*v2(1);

	      def << t
	          << " " << wm
		  << " " << atan2(up2(1),up2(0)+1000)
		  << endl;


	      /*kurbelwelle:
	      FlatVector<double> point(DIM,lh);
  	      FlatVector<double> up(DIM,lh);

	      point(0)=15;
	      point(1)=0;
	      point(2)=20;
      	      ComputeU(point, up, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
		//cout << "x=" << x << ", u=" << up(1) << endl;

	      def << t << " " << (sqrt(up(1)*up(1)+(up(2)+20)*(up(2)+20))-20)
		  << " " << atan2(up(1),up(2)+20) << endl;


	      def2 << t << " " << up(0)  << " " << up(1)  << " " << up(2);

	      point(0)=-14.99; point(1)=0; point(2)=0;
      	      ComputeU(point, up, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
	      def2 << " " << up(0)  << " " << up(1)  << " " << up(2);

	      point(0)=44.99; point(1)=0; point(2)=0;
      	      ComputeU(point, up, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
	      def2 << " " << up(0)  << " " << up(1)  << " " << up(2);

	      point(0)=15.; point(1)=0; point(2)=80; //Kolben
      	      ComputeU(point, up, dynamic_cast<const S_GridFunction<double>&> (*gfu),lh);
	      def2 << " " << up(0)  << " " << up(1)  << " " << up(2);

	      def2 << endl;
	*/

	    }
	  break;
	}

      }
  }
  




  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("nonlinelast", NonlinearElasticity<2>::Create);
      GetIntegrators().AddBFIntegrator ("nonlinearelasticity", 2, 2,
					NonlinearElasticityIntegrator<2>::Create);
      GetIntegrators().AddBFIntegrator ("nonlinearelasticity", 3, 2,
					NonlinearElasticityIntegrator<3>::Create);
    }

    Init init;
  }

}
