/*

Nonlinear magnetic field solver (2D)

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;




class NonLinMagDMat : public DMatOp<NonLinMagDMat>
{
  CoefficientFunction * mu1;
  CoefficientFunction * mu2;
  CoefficientFunction * bsat;

public:

  enum { DIM_DMAT = 2 };

  NonLinMagDMat (CoefficientFunction * amu1,
		 CoefficientFunction * amu2,
		 CoefficientFunction * absat)
    : mu1(amu1), mu2(amu2), bsat(absat) { ; }


  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    double mu1i = Evaluate (*mu1, sip);
    double mu2i = Evaluate (*mu2, sip);
    double bsati = Evaluate (*bsat, sip);

    double babs = L2Norm (vlin);

    if (babs < bsati)
      {
	mat = 0;
	mat(0,0) = mat(1,1) = 1/mu1i;
      }

    else
      
      {
	int i, j;
	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);
	double nu = habs / babs;
	double nuprime = (1/mu2i * babs - habs) / (babs*babs);
	
	for (i = 0; i < 2; i++)
	  for (j = 0; j < 2; j++)
	    mat(i,j) = nuprime / babs * vlin(i) * vlin(j);
	for (i = 0; i < 2; i++)
	  mat(i,i) += nu;
      }
  }




  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    double mu1i = Evaluate (*mu1, sip);
    mat = 0;
    for (int i = 0; i < 2; i++)
      mat(i,i) = 1.0/mu1i;
  }


  template <typename FEL, typename SIP, class VECX, class VECY>
  void Apply (const FEL & fel, const SIP & sip,
	      const VECX & x, VECY & y, LocalHeap & lh) const
  {
    typedef typename VECX::TSCAL TSCAL;

    double mu1i = Evaluate (*mu1, sip);
    double mu2i = Evaluate (*mu2, sip);
    double bsati = Evaluate (*bsat, sip);

    double babs = L2Norm (x);

    if (babs < bsati)
      {
	y = 1/mu1i * x;
      }

    else
      
      {
	// cout << " nonlin, b = " << vlin << ", babs = " << babs << endl;
	// cout << "mu1 = " << mu1i << ", mu2 = " << mu2i << endl;

	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);
	double nu = habs / babs;

	y = nu * x;
      }
  }



  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    typedef typename VECX::TSCAL TSCAL;
    Mat<2,2, TSCAL> mat;

    double mu1i = Evaluate (*mu1, sip);
    double mu2i = Evaluate (*mu2, sip);
    double bsati = Evaluate (*bsat, sip);

    double babs = L2Norm (vlin);

    if (babs < bsati)
      {
	mat = 0;
	mat(0,0) = mat(1,1) = 1/mu1i;
      }

    else
      
      {
	// cout << " nonlin, b = " << vlin << ", babs = " << babs << endl;
	// cout << "mu1 = " << mu1i << ", mu2 = " << mu2i << endl;

	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);
	double nu = habs / babs;
	double nuprime = (1/mu2i * babs - habs) / (babs*babs);
	int i, j;
	// cout << "nu1 = " << 1.0/mu1i << ", nu2 = " << 1.0/mu2i << ", nuprime = " << nuprime << endl;
	// cout << "nuprime*babs = " << nuprime * babs << ", nu = " << nu << endl;
	for (i = 0; i < 2; i++)
	  for (j = 0; j < 2; j++)
	    mat(i,j) = nuprime / babs * vlin(i) * vlin(j);
	// cout << " mat1 = " << mat << endl;
	for (i = 0; i < 2; i++)
	  mat(i,i) += nu;
	// cout << " mat = " << mat << endl;
      }

    y = mat * x;
  }
};




/// Gradient operator of dimension D
class DiffOpCurl2d : public DiffOp<DiffOpCurl2d >
{
public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 2 };
  enum { DIFFORDER = 1 };

  ///
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & fel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    typedef typename MAT::TSCAL TSCAL;
    
    mat = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel.GetDShape(sip.IP(),lh));
    for (int i = 0; i < mat.Width(); i++)
      {
	TSCAL hv;
	hv = mat(0,i);
	mat(0,i) = mat(1,i);
	mat(1,i) = -hv;
      }
  }

  ///
  template <typename FEL, typename SIP, class TVX, class TVY>
  static void Apply (const FEL & fel, const SIP & sip,
		     const TVX & x, TVY & y,
		     LocalHeap & lh) 
  {
    typedef typename TVX::TSCAL TSCAL;

    Vec<2,TSCAL> hv = Trans (fel.GetDShape(sip.IP(), lh)) * x;
    y = Trans (sip.GetJacobianInverse()) * hv;

    TSCAL hval;
    hval = y(0);
    y(0) = y(1);
    y(1) = -hval;
  }

  ///
  template <typename FEL, typename SIP, class TVX, class TVY>
  static void ApplyTrans (const FEL & fel, const SIP & sip,
			  const TVX & x, TVY & y,
			  LocalHeap & lh) 
  {
    typedef typename TVX::TSCAL TSCAL;
    Vec<2,TSCAL> hx;
    hx(0) = -x(1);
    hx(1) = x(0);
    Vec<2,TSCAL> hv = sip.GetJacobianInverse() * hx;
    y = fel.GetDShape(sip.IP(),lh) * hv;
  }
};

  

class NonLinMagIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpCurl2d, NonLinMagDMat, NodalFiniteElement>
{
public:
  ///
  NonLinMagIntegrator (CoefficientFunction * coeff1,
		       CoefficientFunction * coeff2,
		       CoefficientFunction * coeff3)
    :  T_NonlinearBDBIntegrator<DiffOpCurl2d, NonLinMagDMat, NodalFiniteElement>
  (NonLinMagDMat (coeff1, coeff2, coeff3))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new NonLinMagIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }
  
  ///
  virtual string Name () const { return "NonLinMag"; }
};





///
class NumProcNonLinMag : public NumProc
{
protected:

public:
  ///
  NumProcNonLinMag (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { ; }
  ///
  virtual ~NumProcNonLinMag()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcNonLinMag (pde, flags);
  }

  ///
  virtual void Do(LocalHeap & lh)
  {
    int level = ma.GetNLevels() - 1;
    cout << "nonlinmag solver called" << endl;
    
    BilinearForm & bfa = const_cast<BilinearForm&> (*pde.GetBilinearForm("a"));
    const BilinearForm & bfm = *pde.GetBilinearForm("m");
    const LinearForm & lff = *pde.GetLinearForm("f");
    GridFunction & gfu = *pde.GetGridFunction("u");
    const Preconditioner & pre = *pde.GetPreconditioner("c");
    // double tau = pde.GetConstantTable()["tau"];
    // double te = pde.GetConstantTable()["te"];
    cout << "tau = " << tau << endl;
    cout << "te = " << te << endl;

    const BaseMatrix & c = pre.GetMatrix();
    const BaseMatrix & mata = bfa.GetMatrix();
    const BaseMatrix & matm = bfm.GetMatrix();
    BaseVector & vecu = gfu.GetVector();
    const BaseVector & vecf = lff.GetVector();

    BaseVector & uold = *vecu.CreateVector();
    BaseVector & deltaa = *vecu.CreateVector();
    BaseVector & uk = *vecu.CreateVector();
    BaseVector & d = *vecu.CreateVector();
    BaseVector & rhs = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();

    BilinearFormApplication applya(&bfa);
    /*
    LinearizedBilinearFormApplication applyalin(&bfa, &uk);
    CGSolver<double> inva (applyalin, c);
    inva.SetMaxSteps (500);
    */

    // instationary solver:

    ofstream lossout ("loss.out");
    double t;
    for (t = 0; t < te; t += tau)
      {
	cout << "***** t = " << t << endl;
	rhs = sin(100*M_PI*t) * vecf + matm * vecu;
	uk = vecu;
	
	double err, errold;
	d = rhs - applya * uk;
	err = L2Norm(d);
	for (int i = 1; i <= 30; i++)
	  {
	    cout << "newton it " << i << endl;
	    cout << "err = " << err << endl;
	    
	    bfa.AssembleLinearization (uk, lh);
	    BaseMatrix & inva = 
	      *dynamic_cast<BaseSparseMatrix&> (bfa.GetMatrix()).InverseMatrix();
	    /*
	    CGSolver<double> inva (bfa.GetMatrix(), c);
	    inva.SetMaxSteps (500);
	    */

	    errold = err;
	    w = inva * d;
	    // cout << "cgsteps. " << inva.GetSteps() << endl;
	    delete &inva;
	    uold = uk;
	    do
	      {
		uk = uold + w;
		d = rhs - applya * uk;
		w *= 0.5;
		err = L2Norm(d);
		cout << "linsearch, err = " << err << endl;
	      }
	    while (err > errold);
	    
	    if (err < 1e-8) break;
	  }
	
	deltaa = uk - vecu;
	d = matm * deltaa;
	double loss = 1/tau * InnerProduct (d, deltaa);
	lossout << t << " " << loss << endl;

	vecu = uk;
      }
    

    /*
      // stationary solver:
      
    double err, errold;
    d = vecf - applya * vecu;
    err = L2Norm(d);
    for (int i = 1; i <= 30; i++)
      {
	cout << "newton it " << i << endl;
	cout << "err = " << err << endl;
	
	errold = err;
	w = inva * d;
	cout << "cgsteps. " << inva.GetSteps() << endl;
	uold = vecu;
	do
	  {
	    vecu = uold + w;
	    d = vecf - applya * vecu;
	    w *= 0.5;
	    err = L2Norm(d);
	    cout << "linsearch, err = " << err << endl;
	  }
	while (err > errold);

	if (err < 1e-8) break;
      }
    */
  }
};








namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("nonlinmag", NumProcNonLinMag::Create);

    GetIntegrators().AddBFIntegrator ("nonlinmag", 2, 3,
				      NonLinMagIntegrator::Create);
  }
  
  Init init;
}
