/*
  Mindlin Reissner Plate (solver)

*/

/*********************************************************************/
/* File:   mr_plate.cpp                                              */
/* Author: Almedin Becirovic / Joachim Schberl                      */
/* Date:   10. Feb 2005                                              */
/*********************************************************************/

#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;

int equ_order = 1;

class EASFiniteElement : public FiniteElement
{

public:
  EASFiniteElement (ELEMENT_TYPE aeltype = ET_TRIG, int andof = 0, int aorder = 0)
    : FiniteElement (2, aeltype, andof, aorder) { ; } 

  virtual ~EASFiniteElement () 
  { ; }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<2> shape) const = 0;
};


class MREASQuad : public EASFiniteElement
{
public:
  MREASQuad (int aorder)
    : EASFiniteElement (ET_QUAD, 2 * (4*aorder) + 4*equ_order, aorder)
    // p=2 -> 3^2-1^2 = 8
    // p=3 -> 4^2-2^2 = 12
  {
    ;
  }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
    for (int i = 0; i < ndof; i++)
      idofs.Append (i);
  }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<2> shape) const  
  {
    double x = ip(0);
    double y = ip(1);

    int p = order;
    int ii = 0;
    
    ArrayMem<double, 10> pol_x(p+1), pol_y(p+1);
    ArrayMem<AutoDiff<2>, 10> adpol_x(p+1);
    LegendrePolynomial (p, 2*x-1, pol_x);
    LegendrePolynomial (p, 2*y-1, pol_y);
    
    for (int i = 0; i <= p; i++)
      for (int j = 0; j <= p; j++)
	if ( (i>p-2) || (j>p-2) )
	  {
	    shape(ii,0) = pol_x[i]*pol_y[j];
	    shape(ii,1) = 0;
	    ii++;
	    shape(ii,0) = 0;
	    shape(ii,1) = pol_x[i]*pol_y[j];
	    ii++;
	  }

    if (equ_order)
      {
	AutoDiff<2> adx(x, 0);
	AutoDiff<2> ady(y, 1);
	AutoDiff<2> n, tau;
	for (int i = 0; i < 4; i++)
	  {
	    switch (i)
	      {
	      case 0: tau = adx; n = ady; break;
	      case 1: tau = ady; n = 1.0-adx; break;
	      case 2: tau = adx; n = 1.0-ady; break;
	      case 3: tau = ady; n = adx; break;
	      }
	    
	    IntegratedLegendrePolynomial (p, 2*tau-1, adpol_x);
	    AutoDiff<2> val = adpol_x[p] * (1.0-n);
	    shape(ii,0) = val.DValue(0);
	    shape(ii,1) = val.DValue(1);
	    ii++;
	  }
      }
  }
};


class MREASTrig : public EASFiniteElement
{
public:
  MREASTrig (int aorder)
    : EASFiniteElement (ET_TRIG, 2 * (3*aorder) + 3*equ_order, aorder)
  { ; }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);
    for (int i = 0; i < ndof; i++)
      idofs.Append (i);
  }
  

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<2> shape) const   // n  
  {
    double x = ip(0);
    double y = ip(1);

    int p = order;
    int ii = 0;
    
    ArrayMem<double, 10> pol_1(p+1), pol_2(p+1);
    ScaledLegendrePolynomial(p, 2*x+y-1, 1-y, pol_1);

    shape = 0;
    for (int i = 0; i <= p; i++)
      {
	JacobiPolynomial (p, 2*y-1,2*i+1, 0, pol_2);
        shape(ii,0) = shape(ii+1,1) = pol_1[i]*pol_2[p-i];
	ii += 2;
      }
    for (int i = 0; i <= p-1; i++)
      {
	JacobiPolynomial (p, 2*y-1,2*i+1, 0, pol_2);
        shape(ii,0) = shape(ii+1,1) = pol_1[i]*pol_2[p-i-1];
	ii += 2;
      }

    for (int i = 0; i <= p-2; i++)
      {
	JacobiPolynomial (p, 2*y-1,2*i+1, 0, pol_2);
        shape(ii,0) = shape(ii+1,1) = pol_1[i]*pol_2[p-i-2];
	ii += 2;
      }

    if (equ_order)
      {
	AutoDiff<2> lami[3];
	lami[0] = AutoDiff<2> (x, 0);
	lami[1] = AutoDiff<2> (y, 1);
	lami[2] = 1.0 - lami[0] - lami[1];
	ArrayMem<AutoDiff<2>, 10> adpol_x(p+1);
	for (int i = 0; i < 3; i++)
	  {
	    AutoDiff<2> l1 = lami[i];
	    AutoDiff<2> l2 = lami[(i+1)%3];
	    ScaledIntegratedLegendrePolynomial (p-1, l1-l2, l1+l2, adpol_x);
	    shape(ii,0) = adpol_x[p-1].DValue(0);
	    shape(ii,1) = adpol_x[p-1].DValue(1);
	    ii++;
	  }
      }
  }
};



class FESpaceMREnhancedStrain : public FESpace
{
private:
  
  int nel;
  ARRAY<int> first_element_dof;
  int ndof; 

  
public:

  FESpaceMREnhancedStrain (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama, flags)
  {
    segm = new FE_Segm0;
    quad = new MREASQuad (order+1);
    trig = new MREASTrig (order+2);
  }

  virtual ~FESpaceMREnhancedStrain ()
  {
    ;
  }

  virtual string GetClassName () const
  {
    return "FESpaceEnhancedStrain for Mindlin-Reissner-Plate";
  }

  ///
  virtual void Update()
  {
    nel = ma.GetNE();
    ndof = 0;
    first_element_dof.SetSize(nel+1);
   
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
        switch (ma.GetElType(i))
	  {
	    
	  case ET_TRIG:
	    ndof += trig->GetNDof();
	    break;
	  case ET_QUAD:
	    ndof += quad->GetNDof();
	    break;
	  }
      }
    first_element_dof[nel] = ndof;  
  }

  ///
  virtual int GetNDof () const
  {
    return ndof;
  }

  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {  
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);
  }

  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }


  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize (0); }
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize (0); }

};



/* ********************* FESpace **************************** */

class FESpaceMRPlate : public CompoundFESpace
{
  enum TYPE { HOMRFE};
  TYPE type;
public:
  FESpaceMRPlate (const MeshAccess & ama,
		  const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  {
    ;
  }

  virtual ~FESpaceMRPlate () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(4);

    TYPE type = HOMRFE;
    if (flags.GetDefineFlag("homrplate")) type = HOMRFE;

    int order = int (flags.GetNumFlag ("order", 1));
    if (type == HOMRFE)
      {
	cout << " Create Mindlin-Reissner Plate HO-FE Space" << endl;
           
	Flags wflags;
	wflags.SetFlag ("order", order+1-equ_order);
	wflags.SetFlag ("orderinner", order+1 );
	spaces[0] = new H1HighOrderFESpace (ma, wflags);

	Flags betaflags;
	betaflags.SetFlag ("order", order);
	betaflags.SetFlag ("orderquad", order+1);
	betaflags.SetFlag ("ordertrig", order+2 );
	spaces[1] = new H1HighOrderFESpace (ma, betaflags);
	spaces[2] = new H1HighOrderFESpace (ma,  betaflags);
	   
	Flags etaflags;
	etaflags.SetFlag ("order", order);
	spaces[3] = new FESpaceMREnhancedStrain (ma, etaflags);
      }

    FESpaceMRPlate * fes = new FESpaceMRPlate (ma, spaces);
    fes->type = type;
    return fes;
  }

  virtual string GetClassName () const { return "Mindlin-Reissner-Plate FESpace"; }

  virtual Table<int> * CreateSmoothingBlocks (int stype = 0) const
  {
    return 0;
  }

  virtual void GetWireBasketDofNrs (int enr, ARRAY<int> & dnums) const
  {
    ARRAY<int> vnums, vdnums, enums, ednums;
    dnums.SetSize (0);
    ma.GetElVertices (enr, vnums);
    for (int i = 0; i < vnums.Size(); i++)
      {      
	GetVertexDofNrs (vnums[i], vdnums);
	for (int j = 0; j < vdnums.Size(); j++)
	  dnums.Append (vdnums[j]);
      }

    /*
    ma.GetElEdges (enr, enums);
    for (int i = 0; i < enums.Size(); i++)
      {
	for (int k = 0; k < 3; k++)
	  {
	    spaces[k] -> GetEdgeDofNrs (enums[i], ednums);
	    dnums.Append (cummulative_nd[k]+ednums[0]);
	  }
      }
    (*testout) << "wirebasket : " << endl << dnums << endl;
    */
  }


  virtual void VTransformMR (int elnr, bool boundary,
			     FlatMatrix<double> & mat, TRANSFORM_TYPE tt) const 
  {;}

  virtual void VTransformMC (int elnr, bool boundary,
			     FlatMatrix<Complex> & mat, TRANSFORM_TYPE tt) const
  { ; }

  virtual void VTransformVR (int elnr, bool boundary,
			     FlatVector<double> & vec, TRANSFORM_TYPE tt) const 
  {;}

  virtual void VTransformVC (int elnr, bool boundary,
			     FlatVector<Complex> & vec, TRANSFORM_TYPE tt) const 
  {;}


};



/* ********************+ Integrator ************************** */

template <int D>
class DiffOpMRPlate : public DiffOp<DiffOpMRPlate<D> >
{
};

template <>
class DiffOpMRPlate<2> : public DiffOp<DiffOpMRPlate<2> >
{
  // 7 components:
  // eps1,1 ; eps2,2 ; eps1,2 ; dw/dx-beta1; dw/dy-beta2; dw/dx-beta1-eta1; dw/dy-beta2-eta2

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 7 };
  enum { DIFFORDER = 0 };

  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = dynamic_cast<const CompoundFiniteElement&> (bfel);

    
    const NodalFiniteElement & fel_w = dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_beta = dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const EASFiniteElement & fel_eta = dynamic_cast<const EASFiniteElement&> (cfel[3]);
    
    
    int nd_w = fel_w.GetNDof ();
    int nd_beta = fel_beta.GetNDof();
    int nd_eta = fel_eta.GetNDof();
    

    FlatMatrix<> grad_beta (2, nd_beta, lh);
    grad_beta = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_beta.GetDShape(sip.IP(), lh));

    FlatMatrix<> gradw (2, nd_w, lh);
    gradw = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_w.GetDShape(sip.IP(), lh));

    FlatVector<> beta(nd_beta, lh);
    beta = fel_beta.GetShape (sip.IP(), lh);

    mat = 0;
    for (int i = 0; i < nd_beta; i++)
      {
	mat(0, nd_w+i  ) = grad_beta(0, i);
	mat(1, nd_w+nd_beta+i) = grad_beta(1, i);
	mat(2, nd_w+i  ) = grad_beta(1, i);
	mat(2, nd_w+nd_beta+i) = grad_beta(0, i);
        mat(3, nd_w+i) = -beta(i);
	mat(4, nd_w+nd_beta+i) = -beta(i);
	//        mat(5, nd_w+i) = -beta(i);
	//	mat(6, nd_w+nd_beta+i) = -beta(i);
      }

    for (int i = 0; i < nd_w; i++)
      {
	mat(3, i) = gradw(0, i);
	mat(4, i) = gradw(1, i);
	//        mat(5, i) = gradw(0, i);
	//	mat(6, i) = gradw(1, i);
      }


	FlatMatrixFixWidth<2> eta1 (nd_eta, lh);
	FlatMatrixFixWidth<2> eta (nd_eta, lh);
	fel_eta.CalcShape(sip.IP(), eta1);
	eta = eta1 * sip.GetJacobianInverse();

	for (int i = 0; i < nd_eta; i++)
	  {
	    mat(3,nd_w+2*nd_beta+i) = -eta(i,0);
	    mat(4,nd_w+2*nd_beta+i) = -eta(i,1);
	  }

	int nvert = (fel_w.ElementType() == ET_TRIG) ? 3 : 4;
	if (!equ_order) nvert = 0;
	// nvert = 0;
	for (int i = 0; i < nd_eta-nvert; i++)
	  {
	    mat(5,nd_w+2*nd_beta+i) = -eta(i,0);
	    mat(6,nd_w+2*nd_beta+i) = -eta(i,1);
	  }

    /* (*testout)<<"nd_w = "<<nd_w<<endl;
       (*testout)<<"nd_beta = "<<nd_beta<<endl;
       (*testout)<<"nd_eta = "<<nd_eta<<endl;
       (*testout)<<"mat = "<<mat<<endl;*/
  }
};






class MRPlateDMat : public DMatOp<MRPlateDMat>
{
  CoefficientFunction * coefe;
  CoefficientFunction * coefnu;
  CoefficientFunction * coeft;
public:

  enum { DIM_DMAT = 7 };

  MRPlateDMat (CoefficientFunction * aE, CoefficientFunction * anu, CoefficientFunction * at) : coefe(aE),coefnu(anu),coeft(at) { ; }

  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    // (C*eps(beta), eps(beta))
    double E = Evaluate (*coefe, sip);
    double nu = Evaluate (*coefnu, sip);
    double t = Evaluate (*coeft, sip);
    
    mat(0,0) = mat(1,1) = 1-nu;
    mat(0,1) = mat(1,0) = nu;
    mat(2,2) = 0.5 * (1-2*nu);

    mat *= (E / ((1 + nu) * (1 - 2 * nu)));

    double h = 0;
    h = sqrt(sip.GetJacobiDet());
    // (*testout) << "h1 = " << h << endl;


    h = 0.0;
    Mat<2,2> J;
    J = sip.GetJacobian();
    for (int i = 0; i < 2; i++)
      for (int j = 0; j < 2; j++)
	h += sqr (J(i,j));
    h = sqrt(h);

    h /= (5*fel.Order());
    // (*testout) << "h2 = " << h << endl;
    // (*testout) << "J = " << endl << J << endl;
   
    // E*t^-2(grad(w)-beta-eta, grad(w)-beta-eta)
    //    mat(3,3) = mat(4,4) = E/((t*t)+(h*h));
    //    mat(5,5) = mat(6,6) = E/(t*t)-E/((t*t)+(h*h));

    mat(3,3) = mat(4,4) = E/(t*t);
    mat(5,5) = mat(6,6) = E/(h*h);
  }
};




//*********************************************************************************
//                      Mindlin-Reissner-Plate Integrators                              *
//*********************************************************************************

class MRPlateIntegrator
  : public T_BDBIntegrator<DiffOpMRPlate<2>, MRPlateDMat, FiniteElement>
{
public:
  ///
  MRPlateIntegrator (CoefficientFunction * coefE, CoefficientFunction * coefnu, CoefficientFunction * coeft)
    :  T_BDBIntegrator<DiffOpMRPlate<2>, MRPlateDMat, FiniteElement>
  (MRPlateDMat (coefE,coefnu,coeft))
  { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MRPlateIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }

  ///
  virtual string Name () const { return "Mindlin-Reissner-Plate"; }
};




//********************************************************************************************************
//                 Mindlin-Reissner-Plate for viewing of displacement W and rotation beta1 and beta2     *
//********************************************************************************************************

class DiffOpIdWBE : public DiffOp<DiffOpIdWBE>
{
  // 5 components:
  // w, beta1, beta2, eta1, eta2,

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 5 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = dynamic_cast<const CompoundFiniteElement&> (bfel);
    const NodalFiniteElement & fel_w = dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_beta = dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    const EASFiniteElement & fel_eta = dynamic_cast<const EASFiniteElement&> (cfel[3]);
   
    int nd_w = fel_w.GetNDof();
    int nd_beta = fel_beta.GetNDof();
    int nd_eta = fel_eta.GetNDof();
   
    FlatVector<> vecw(nd_w, lh);
    fel_w.CalcShape (sip.IP(), vecw);

    FlatVector<> vecbeta(nd_beta, lh);
    fel_beta.CalcShape (sip.IP(), vecbeta);

    FlatMatrixFixWidth<2> eta (nd_eta, lh);
    fel_eta.CalcShape (sip.IP(), eta);

    mat = 0;

    for (int j = 0; j < nd_w; j++)
      {
	mat(0,j) = vecw(j);
      }
    for (int j = 0; j < nd_beta; j++)
      {
	mat(1,nd_w+j) = vecbeta(j);
        mat(2,nd_w+nd_beta+j) = vecbeta(j);
      }
    for (int j = 0; j < nd_beta; j++)
      {
	mat(3,nd_w+2*nd_beta+j) = eta(j,0);
        mat(4,nd_w+2*nd_beta+j) = eta(j,1);
      }

  }
};


class MRPlateWBEIntegrator 
  : public T_BDBIntegrator<DiffOpIdWBE, DiagDMat<5>, FiniteElement>
{
public:
  MRPlateWBEIntegrator ()
    :  T_BDBIntegrator<DiffOpIdWBE, DiagDMat<5>, FiniteElement>
  (DiagDMat<5> (new ConstantCoefficientFunction(1)))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MRPlateWBEIntegrator ();
  }

  virtual string Name () const { return "Mindlin-Reissner-Plate Id-W-Beta-Eta"; }
};




















  //  ******************************* Error computation **************************************


namespace ngcomp
{

  using namespace ngcomp;

 
  void CalcMREASError(const MeshAccess & ma,
		      const S_GridFunction<double> &  u1,
		      const S_GridFunction<double> &  u2,
		      const BilinearFormIntegrator & bfi_eas,
		      const BilinearFormIntegrator & bfi_id, 
		      FlatVector<double> & err, 
		      int domain)
  {
    ostream * test;
    test = new ofstream ("file_p2.out");
    const FESpace & fes1 = u1.GetFESpace();
    const FESpace & fes2 = u2.GetFESpace();
    
    //   (*test)<<" elu1 = "<< u1.GetVector()<<endl<<endl;
    bool bound = bfi_eas.BoundaryForm();

    int ne = ma.GetNE();
    int dim1 = fes1.GetDimension();
    int dim2 = fes2.GetDimension();
    int dimflux = bfi_eas.DimFlux();
    int dim = bfi_id.DimFlux();
    


    
    ElementTransformation eltrans;

    ARRAY<int> dnums1;
    ARRAY<int> dnums2;
  
    double errw = 0;
    double errbeta1 = 0; double errbeta2 = 0;
    double errshear2 = 0; double errshear1 =0;
    double errshear3 = 0; double errshear4 =0;
    double errshear5 = 0; double errshear6 =0;


    double err_shear = 0;
    double err_bending = 0;


    LocalHeap lh (10000000);

    for (int i = 0; i < ne; i++)
      {
	lh.CleanUp();

	int eldom = ma.GetElIndex(i);
        
	if ((domain != -1) && (domain != eldom))
	  continue;

	const FiniteElement & fel1 = fes1.GetFE (i, lh);
	const FiniteElement & fel2 = fes2.GetFE (i, lh);



	ma.GetElementTransformation (i, eltrans, lh);
	fes1.GetDofNrs (i, dnums1);
	fes2.GetDofNrs (i, dnums2);
	   
	FlatVector<double> elu1(dnums1.Size() * dim1, lh);
	FlatVector<double> elu2(dnums2.Size() * dim2, lh);
        FlatVector<double> flux1(dimflux, lh);
	FlatVector<double> flux2(dimflux, lh);
        FlatVector<double> u_1(dim, lh);
	FlatVector<double> u_2(dim, lh);


	u1.GetElementVector (dnums1, elu1);
	fes1.TransformVec (i, bound, elu1, TRANSFORM_SOL);
	//	(*test)<<" elu1 = "<< elu1<<endl<<endl;

	u2.GetElementVector (dnums2, elu2);
	fes2.TransformVec (i, bound, elu2, TRANSFORM_SOL);
	// (*test)<<" elu2 = "<< elu2<<endl<<endl;

	double elerr = 0;
        
	const IntegrationRule & ir =
	  GetIntegrationRules().SelectIntegrationRule(fel1.ElementType(),
						      2*fel1.Order()+4);
	double vol = ma.ElementVolume(i);

	for (int j = 0; j < ir.GetNIP(); j++)
	  {
	    void * heapp = lh.GetPointer();

	    bfi_eas.CalcFlux (fel1, eltrans, ir.GetIP(j), elu1, flux1, 1, lh);
	    bfi_eas.CalcFlux (fel2, eltrans, ir.GetIP(j), elu2, flux2, 1, lh);
            bfi_id.CalcFlux(fel1, eltrans, ir.GetIP(j), elu1, u_1, 0, lh);
            bfi_id.CalcFlux(fel2, eltrans, ir.GetIP(j), elu2, u_2, 0, lh);
	    // (*test)<<" IP = "<< ir.GetIP(j)<<endl;
	    // (*test)<<" elu1 = "<< elu2<<endl<<endl;
            u_1 -= u_2;
	    flux1 -= flux2;

            
	    elerr += ir.GetIP(j).Weight() * vol *sqr( flux1(6));

            errw  += ir.GetIP(j).Weight() * vol *sqr( u_1(0));
            errbeta1 += ir.GetIP(j).Weight() * vol *sqr( u_1(1));
            errbeta2 += ir.GetIP(j).Weight() * vol *sqr( u_1(2));
            // errshear1 += ir.GetIP(j).Weight() * vol *sqr( flux1(3)+flux1(5));
            // errshear2 += ir.GetIP(j).Weight() * vol *sqr( flux1(4)+flux1(6));
            errshear3 += ir.GetIP(j).Weight() * vol *sqr( flux1(3));
            errshear4 += ir.GetIP(j).Weight() * vol *sqr( flux1(4));
            errshear5 += ir.GetIP(j).Weight() * vol *sqr( flux1(5));
            errshear6 += ir.GetIP(j).Weight() * vol *sqr( flux1(6));

            err_bending += ir.GetIP(j).Weight() * vol * 
	      (sqr(flux1(0)) + sqr (flux1(1)) + sqr(flux1(2)));
            err_shear += ir.GetIP(j).Weight() * vol * 
	      (sqr(flux1(3)) + sqr (flux1(4)));

	    lh.CleanUp (heapp);
	  }

	err(i) += elerr;
	
      }

    cout<<errw<<" "<<errbeta1  /* << " "<<errshear1 */ <<" "<<errshear3<<" "<<errshear5<<endl;
    
    // cout << "err_shear = " << sqrt(err_shear) << ", err_
    cout << "N       err_shear       err_bending" << endl;
    cout << u1.GetFESpace().GetNDof() << "      " << sqrt (err_shear) << "       " << sqrt (err_bending) << endl;

    // (*testout)<<"Error Size="<<err.Size()<<endl;
    //	(*testout)<<err<<endl;
  }
}// end of ngcomp














namespace ngsolve
{
  using namespace ngsolve;
  using namespace ngbla;
  using namespace ngla;



  class NumProcMREASErrorEstimator : public NumProc
  {
  private:
    ///
    BilinearForm * bfa_eas;
    ///
    BilinearForm * bfa_id;
    ///
    GridFunction * gfu1;
    ///
    //  BilinearForm * bfa_eas2;
    ///
    // BilinearForm * bfa_id2;
    ///
    GridFunction * gfu2;
    ///
    GridFunction * gferr;

    string filename;

  public:
    NumProcMREASErrorEstimator (PDE & apde, const Flags & flags);
    virtual ~NumProcMREASErrorEstimator();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcMREASErrorEstimator (pde, flags);
    }

    virtual void Do();
    virtual void PrintReport (ostream & ost);

    virtual string GetClassName () const
    {
      return "MR-EAS Error Estimator";
    }
  };

  NumProcMREASErrorEstimator :: NumProcMREASErrorEstimator (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
                       
    bfa_eas = pde.GetBilinearForm (flags.GetStringFlag ("easbilinearform", ""));

                                                        
    bfa_id = pde.GetBilinearForm (flags.GetStringFlag ("idbilinearform", ""));

    gfu1 = pde.GetGridFunction (flags.GetStringFlag ("solution", ""));
    // bfa_eas2 = pde.GetBilinearForm (flags.GetStringFlag ("eas_p", ""));
    // bfa_id2 = pde.GetBilinearForm (flags.GetStringFlag ("id_p", ""));
    gfu2 = pde.GetGridFunction (flags.GetStringFlag ("solution_p", ""));
    gferr = pde.GetGridFunction (flags.GetStringFlag ("error", ""));
    filename = flags.GetStringFlag ("filename","");

  }

  NumProcMREASErrorEstimator :: ~NumProcMREASErrorEstimator()
  {
    ;
  }

  void NumProcMREASErrorEstimator :: Do()
  {
    cout << "MR-EAS ErrorEstimator" << endl;

    if (bfa_eas->NumIntegrators() == 0)
      throw Exception ("MR-EAS ErrorEstimator: Bilinearform eas needs an integrator");
    if (bfa_id->NumIntegrators() == 0)
      throw Exception ("MR-EAS ErrorEstimator: Bilinearform id needs an integrator");
    // if (bfa_eas2->NumIntegrators() == 0)
    // throw Exception ("EAS ErrorEstimator: Bilinearform eas2 needs an integrator");
    // if (bfa_id2->NumIntegrators() == 0)
    //  throw Exception ("EAS ErrorEstimator: Bilinearform id2 needs an integrator");

    BilinearFormIntegrator * bfi_eas = bfa_eas->GetIntegrator(0);
    BilinearFormIntegrator * bfi_id = bfa_id->GetIntegrator(0);
    // BilinearFormIntegrator * bfi_eas2 = bfa_eas2->GetIntegrator(0);
    //  BilinearFormIntegrator * bfi_id2 = bfa_id2->GetIntegrator(0);

  

    cout<<"LEVEL = "<<ma.GetNLevels()<<endl;

   



    FlatVector<double> err =
      dynamic_cast<T_BaseVector<double>&> (gferr->GetVector()).FV();

    err = 0;

    int i, j, k;
    int ndom = ma.GetNDomains();

    for (k = 0; k < ndom; k++){
      CalcMREASError(ma,
		     dynamic_cast<const S_GridFunction<double>&>  (*gfu1),
		     dynamic_cast<const S_GridFunction<double>&>  (*gfu2),
		     *bfi_eas, *bfi_id, err, k);
    }
    double sum = 0;
    for (i = 0; i < err.Size(); i++)
      sum += err(i);
    cout<< "estimated error="<<sqrt(sum) << endl;
    static ofstream errout ("error.out");
    errout << ma.GetNLevels()
	   << " " << bfa_eas->GetFESpace().GetNDof()
	   << " " << sqrt(double (bfa_eas->GetFESpace().GetNDof()))
	   << " " << sqrt(sum) << endl;
    static ofstream ofile (filename.c_str());
    ofile << "  " << bfa_eas->GetFESpace().GetOrder()
	  << "  " << sqrt(sum)<<endl;



  }
  void NumProcMREASErrorEstimator :: PrintReport (ostream & ost)
  {
    ost << "NumProcMREASErrorEstimator:" << endl;
    ost << "Bilinear-form = " << endl;
  }

  namespace {
    class Init
    {
    public:
      Init();
    };
    Init :: Init()
    {

      GetNumProcs().AddNumProc ("mreasee", NumProcMREASErrorEstimator::Create);
	   
    }

    Init init;
  }
} //namespace ngsolve
















namespace  {
  class Init
  {
  public:
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("mrplate", FESpaceMRPlate::Create);

    GetIntegrators().AddBFIntegrator ("mrplate", 2, 3,
				      MRPlateIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mrplatewbe", 2, 0,
				      MRPlateWBEIntegrator::Create);
  }

  Init init;
}
