/*

Drift Diffusion model

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


double eps = 11.8 * 8.85e-12;	
double ni = 1e19;
double e_ = 1.6e-19;
double uT = 1.38e-23*300./e_;     // kb * T / e_
double mu = 0.1;




/* ********************* FESpace **************************** */

// the product space for (phi, zeta_p, zeta_n)
class FESpaceSemiconductor : public CompoundFESpace
{
public:
  FESpaceSemiconductor (const MeshAccess & ama, 		   
			const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  { ; }

  virtual ~FESpaceSemiconductor () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(3);

    spaces[0] = new H1HighOrderFESpace (ma, flags);
    spaces[1] = new H1HighOrderFESpace (ma, flags);
    spaces[2] = new H1HighOrderFESpace (ma, flags);

    FESpaceSemiconductor * fes = new FESpaceSemiconductor (ma, spaces);
    return fes;
  }
    
  virtual string GetClassName () const { return "Semiconductor - FESpace"; }
};



/* ********************+ Integrator ************************** */

class DiffOpSemiconductor : public DiffOp<DiffOpSemiconductor>
{
  // 9 components:
  // phi, phi_x, phi_y, zetap, zetap_x, zetap_y, zetan, zetan_x, zetan_y

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 9 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_phi = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_zeta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_phi = fel_phi.GetNDof();
    int nd_zeta = fel_zeta.GetNDof();

    FlatMatrix<> grad_phi(2, nd_phi, lh);
    grad_phi = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_phi.GetDShape(sip.IP(), lh));

    FlatMatrix<> grad_zeta(2, nd_phi, lh);
    grad_zeta = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_zeta.GetDShape(sip.IP(), lh));
    
    FlatVector<> vec_phi(nd_phi, lh);
    vec_phi = fel_phi.GetShape (sip.IP(), lh);
    FlatVector<> vec_zeta(nd_zeta, lh);
    vec_zeta = fel_zeta.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_phi; j++)
      {
	mat(0,j) = vec_phi(j);
	mat(1,j) = grad_phi(0,j);
	mat(2,j) = grad_phi(1,j);
      }
    for (int j = 0; j < nd_zeta; j++)
      {
	mat(3,nd_phi+j) = vec_zeta(j);
	mat(4,nd_phi+j) = grad_zeta(0,j);
	mat(5,nd_phi+j) = grad_zeta(1,j);
	mat(6,nd_phi+nd_zeta+j) = vec_zeta(j);
	mat(7,nd_phi+nd_zeta+j) = grad_zeta(0,j);
	mat(8,nd_phi+nd_zeta+j) = grad_zeta(1,j);
      }
  }
};


class SemiconductorDMat : public DMatOp<SemiconductorDMat>
{
  CoefficientFunction * nu;

public:

  enum { DIM_DMAT = 9 };

  SemiconductorDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    Vec<9> vlin;
    vlin = 0.0;
    GenerateLinearizedMatrix (fel, sip, vlin, mat, lh);
  }  

  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    mat = 0.0;

    mat(0,0) = e_* ni / uT * ( exp ( (vlin(0) - vlin(6)) / uT) + exp ( (vlin(3) - vlin(0))/uT) );

    mat(1,1) = eps;
    mat(2,2) = eps;
    
    mat(4,4) = mat(5,5) = exp ( (vlin(3)-vlin(0)) / uT);   
    mat(7,7) = mat(8,8) = exp ( (vlin(0)-vlin(6)) / uT);   
  }


  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    ;
  }


  template <typename FEL, typename SIP, class TVX, class TVY>
  void Apply (const FEL & fel, const SIP & sip,
	      const TVX & x, TVY & y,
	      LocalHeap & lh) const
  {
    y(0) = e_* ni * ( exp ( (x(0) - x(6)) / uT) - exp ( (x(3) - x(0))/uT) );
    y(1) = eps * x(1);
    y(2) = eps * x(2);

    y(3) = 0.0;
    y(4) = exp ( (x(3) - x(0)) / uT) * x(4);
    y(5) = exp ( (x(3) - x(0)) / uT) * x(5);

    y(6) = 0.0;
    y(7) = exp ( (x(0) - x(6)) / uT) * x(7);
    y(8) = exp ( (x(0) - x(6)) / uT) * x(8);
  }



  template <typename FEL, typename SIP, class TVX>
  double Energy (const FEL & fel, const SIP & sip,
		 const TVX & x, LocalHeap & lh) const
  {
    // (*testout) << "x = " << x << "  ";
    double energy = 0.5 * eps * (x(1)*x(1) + x(2)*x(2));
    
    // (*testout) << "energy1 = " << energy << " ";
    energy += e_ * ni * uT * ( exp ( (x(0) - x(6)) / uT) + exp ( (x(3) - x(0))/uT));
    // (*testout) << ", E2 = " << energy << endl;
    return energy;
  }
  

};




  

class SemiconductorIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpSemiconductor, SemiconductorDMat, FiniteElement>
{
public:
  ///
  SemiconductorIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpSemiconductor, SemiconductorDMat, FiniteElement>
  (SemiconductorDMat ())
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new SemiconductorIntegrator ();
  }
  
  ///
  virtual string Name () const { return "Semiconductor"; }
};
















class DiffOpEvalSemiconductor : public DiffOp<DiffOpEvalSemiconductor>
{
  // 5 components:
  // phi, zeta_p, zeta_n, n, p

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 5 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_phi = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_zeta = 
      dynamic_cast<const NodalFiniteElement&> (cfel[1]);
    
    int nd_phi = fel_phi.GetNDof();
    int nd_zeta = fel_zeta.GetNDof();

    FlatVector<> vec_phi(nd_phi, lh);
    vec_phi = fel_phi.GetShape (sip.IP(), lh);
    FlatVector<> vec_zeta(nd_zeta, lh);
    vec_zeta = fel_zeta.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_phi; j++)
      mat(0,j) = vec_phi(j);

    for (int j = 0; j < nd_zeta; j++)
      {
	mat(1,nd_phi+j) = vec_zeta(j);
	mat(2,nd_phi+nd_zeta+j) = vec_zeta(j);
      }
  }
};


class EvalSemiconductorDMat : public DMatOp<EvalSemiconductorDMat>
{
  CoefficientFunction * nu;

public:

  enum { DIM_DMAT = 5 };

  EvalSemiconductorDMat () { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  { ; }

  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  { ; }

  template <typename FEL, typename SIP, class TVX, class TVY>
  void Apply (const FEL & fel, const SIP & sip,
	      const TVX & x, TVY & y,
	      LocalHeap & lh) const
  {
    y(0) = x(0);
    y(1) = x(1);
    y(2) = x(2);
    y(3) = ni * exp ( (x(1) - x(0)) / uT);
    y(4) = ni * exp ( (x(0) - x(2)) / uT);
  }
};


class EvalSemiconductorIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpEvalSemiconductor, EvalSemiconductorDMat, FiniteElement>
{
public:
  ///
  EvalSemiconductorIntegrator ()
    :  T_NonlinearBDBIntegrator<DiffOpEvalSemiconductor, EvalSemiconductorDMat, FiniteElement>
  (EvalSemiconductorDMat ())
  { 
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new EvalSemiconductorIntegrator ();
  }
  
  ///
  virtual string Name () const { return "Eval-Semiconductor"; }
};












/* ******************** TestSolver ************** */


class NumProcSemiconductor : public NumProc
{
protected:
  double tau2;
  double prec1, prec2;
  int nit, n1, n2;
public:

  NumProcSemiconductor (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    prec1 = flags.GetNumFlag ("prec1", 1e-16);
    prec2 = flags.GetNumFlag ("prec2", 1e-16);
    tau2 = flags.GetNumFlag ("tau2", 1e-3); 
    nit = int (flags.GetNumFlag ("n", 10));   
    n1 = int (flags.GetNumFlag ("n1", 100));
    n2 = int (flags.GetNumFlag ("n2", 100));
  }

  virtual ~NumProcSemiconductor()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcSemiconductor (pde, flags);
  }


  virtual void Do(LocalHeap & lh)
  {

    cout << "**************************" << endl;
    cout << "** Semiconductor solver **" << endl;
    cout << "**************************" << endl;


    BilinearForm & bfa = const_cast<BilinearForm&> (*pde.GetBilinearForm("a"));
    const LinearForm & lff = *pde.GetLinearForm("f");
    GridFunction & gfu = *pde.GetGridFunction("u");

    BilinearForm & bfa2 = const_cast<BilinearForm&> (*pde.GetBilinearForm("a2"));
    const LinearForm & lff2 = *pde.GetLinearForm("f2");



    const SparseMatrix<double> & mata = dynamic_cast<const SparseMatrix<double>&> (bfa.GetMatrix());
    BaseVector & u = gfu.GetVector();
    const BaseVector & vecf = lff.GetVector();
    const BaseVector & vecf2 = lff2.GetVector();




    BaseVector & uold = *u.CreateVector();
    BaseVector & hu = *u.CreateVector();
    BaseVector & d = *u.CreateVector();
    BaseVector & w = *u.CreateVector();




    int n = u.Size();

    BitArray phi_dofs(n);
    BitArray zeta_dofs(n);

    phi_dofs.Clear();
    for (int i = 0; i < n/3; i++)
      phi_dofs.Set(i);

    zeta_dofs = phi_dofs;
    zeta_dofs.Invert();



    BilinearFormApplication applya(&bfa);

    u = 0.0;

    SparseCholesky<double> invphi (mata, new BitArray (phi_dofs), 0, 1);
    SparseCholesky<double> invzeta (mata, new BitArray (zeta_dofs), 0, 1);

    
    for (int i = 1; i <= nit; i++)
      {
	for (int j = 1; j <= n1; j++)
	  {
	    bfa.AssembleLinearization (u, lh);
	    invphi.FactorNew (mata);
	    
	    d = vecf - applya * u;
	    w = invphi * d;
	    
	    double err = InnerProduct (w, d);
	    cout << "                                            loop " << i << ", 1, j = " << j << ", |err| = " << err << endl;
	    
	    if (err < prec1) break;
	    
	    uold = u;
	    double funcold = bfa2.Energy (u) - InnerProduct (vecf2, u);
	    double func;
	      
	    cout << "old energy = " << funcold << endl;
	    cout << "df/dw = " << err << endl;
	    
	    
	    if (i == 1 && j == 1)
	      u = w;
	    else
	      {
		double tau = 1;
		do
		  {
		    u = uold + tau * w;
		    func = bfa2.Energy (u) - InnerProduct (vecf2, u);
		    cout << "tau = " << tau << ", energy = " << func << "df/dw(tau) = " << (func-funcold)/tau << endl;
		    tau /= 2;
		    }
		while (func > funcold && tau > 1e-6);
	      }
	  }

	double err;
	for (int j = 1; j <= n2; j++)
	  {
	    bfa.AssembleLinearization (u, lh);

	    invzeta.FactorNew (mata);
	    w = invzeta * vecf;
	    
	    // keep phi-values;

	    FlatVector<> fvw = w.FVDouble();
	    FlatVector<> fvu = u.FVDouble();

	    for (int k = 0; k < n/3; k++)
	      fvw(k) = fvu(k);


	    double tau2a = tau2; // (i <= 2) ? tau2 : 1;

	    // if (i >= 3 || j >= 5) tau2 = 1;

	    uold = u;
	    u *= (1-tau2a);
	    u += tau2a * w;

	    uold -= u;

	    err = L2Norm (uold);

	    cout << "tau2 = " << tau2 << endl;
	    cout << "                                                      loop " << i << ", 2, j = " << j << ", |err| = " << err << endl;

	    if (err < prec2) break;
	  }

	if (err*tau2 < 1e-5 && tau2 < 0.05) tau2 *= 2;
	if (err > 0.1) tau2 /= 2;

	Ng_Redraw();
      }
  }

  ///
  virtual string GetClassName () const
  {
    return "Semiconductor";
  }

  virtual void PrintReport (ostream & ost)
  {
    ;
  }
};







namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("semiconductor", NumProcSemiconductor::Create);
    GetFESpaceClasses().AddFESpace ("semiconductor", FESpaceSemiconductor::Create);

    GetIntegrators().AddBFIntegrator ("semiconductor", 2, 0,
				      SemiconductorIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("evalsemiconductor", 2, 0,
				      EvalSemiconductorIntegrator::Create);
  }
  
  Init init;
}
