/*

Stokes solver

 */


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace **************************** */

class FESpaceStokes : public CompoundFESpace
{
  enum TYPE { P2P0 , MINI, TAYLORHOOD };
  TYPE type;
public:
  FESpaceStokes (const MeshAccess & ama, 		   
		 const ARRAY<const FESpace*> & aspaces)
    : CompoundFESpace (ama, aspaces)
  {
    ;
  }

  virtual ~FESpaceStokes () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(3);

    TYPE type = P2P0;
    if (flags.GetDefineFlag("taylorhood")) type = TAYLORHOOD;
    if (flags.GetDefineFlag("mini")) type = MINI;

    if (type == TAYLORHOOD)
      {
	cout << "Create Taylor-Hood FE Space" << endl;
	// Taylor-Hood
	Flags uflags;
	uflags.SetFlag ("order", 2);
	spaces[0] = new H1HighOrderFESpace (ma, uflags);
	spaces[1] = new H1HighOrderFESpace (ma, uflags);

	uflags.SetFlag ("order", 1);
	spaces[2] = new H1HighOrderFESpace (ma, uflags);
	// spaces[0] = new NodalFESpace (ma, 2, 1, 0, 1);
	// spaces[1] = new NodalFESpace (ma, 2, 1, 0, 1);
	// spaces[2] = new NodalFESpace (ma, 1, 1, 0);
      }
    else if (type == MINI)
      {
	cout << "Create Mini-element FE Space" << endl;
	Flags uflags;
	uflags.SetFlag ("order", 1);
	uflags.SetFlag ("orderinner", 3);
	spaces[0] = new H1HighOrderFESpace (ma, uflags);
	spaces[1] = new H1HighOrderFESpace (ma, uflags);

	Flags pflags;
	pflags.SetFlag ("order", 1);
	spaces[2] = new H1HighOrderFESpace (ma, pflags);
      }
    else
      {
	cout << "Create P2-P0 FE Space" << endl;
	// P2-P0
	Flags uflags;
	uflags.SetFlag ("order", 2);
	spaces[0] = new H1HighOrderFESpace (ma, uflags);
	spaces[1] = new H1HighOrderFESpace (ma, uflags);
	// spaces[0] = new NodalFESpace (ma, 2, 1, 0);
	// spaces[1] = new NodalFESpace (ma, 2, 1, 0);
	spaces[2] = new ElementFESpace (ma, 0, 1, 0);
      }

    FESpaceStokes * fes = new FESpaceStokes (ma, spaces);
    fes->type = type;
    return fes;
  }
    
  virtual string GetClassName () const { return "StokesFESpace"; }
  
  virtual Table<int> * CreateSmoothingBlocks (int stype = 0) const
  {
    if (type == P2P0)
      {
	// blocks associated with vertices
	// velocity V+E,  pressure I
	
	int i, j, k;
	int nv = ma.GetNV();
	int ne = ma.GetNE();

	Table<int> * table;
	ARRAY<int> dnums;

	BitArray has_el(nv);
	has_el.Clear();

	/*
	static const int blocks[3][13] = 
	  { 
	    { 1, 0, 0, 1, 0, 1, 
	      1, 0, 0, 1, 0, 1,
	      1 },
	    { 0, 1, 0, 0, 1, 1,
	      0, 1, 0, 0, 1, 1,
	      1 },
	    { 0, 0, 1, 1, 1, 0,
	      0, 0, 1, 1, 1, 0,
	      1 }
	  };
	*/
	static const int blocks[3][13] = 
	  { 
	    { 1, 0, 0, 0, 1, 1, 
	      1, 0, 0, 0, 1, 1,
	      1 },
	    { 0, 1, 0, 1, 0, 1,
	      0, 1, 0, 1, 0, 1,
	      1 },
	    { 0, 0, 1, 1, 1, 0,
	      0, 0, 1, 1, 1, 0,
	      1 }
	  };

	// inefficient, but simple:
	// first, generate a dynamic table
	// then, copy to the compressed one:

	IntTable temptab(nv);

	for (i = 0; i < ne; i++)
	  {
	    GetDofNrs (i, dnums);
	    (*testout) << "el " << i << ", dnums = " << dnums << endl;
	    for (j = 0; j < 3; j++)
	      {
		for (k = 0; k < 13; k++)
		  if (blocks[j][k])
		    {
		      temptab.AddUnique (dnums[j], dnums[k]);
		      (*testout) << "add " << dnums[j] << " to " << dnums[k] << endl;
		    }
	      }
	  }

	ARRAY<int> cnt(nv);
	for (i = 0; i < nv; i++)
	  cnt[i] = temptab[i].Size();
	table = new Table<int> (cnt);
	for (i = 0; i < nv; i++)
	  for (j = 0; j < temptab[i].Size(); j++)
	    (*table)[i][j] = temptab[i][j];

	(*testout) << "Smoothingblocks = " << endl << (*table) << endl;
	return table;
      }

    else if (type == MINI)
      {
	// blocks associated with vertices
	// vertex u, element u, all vertex p
	
	int i, j;
	int nv = ma.GetNV();
	int ne = ma.GetNE();

	Table<int> * table;
	ARRAY<int> dnums;

	IntTable temptab(nv);

	for (i = 0; i < ne; i++)
	  {
	    GetDofNrs (i, dnums);
	    (*testout) << "dnums = " << endl << dnums << endl;
	    for (j = 0; j < 3; j++)
	      {
		int vi = dnums[j];
		temptab.AddUnique (vi, dnums[j]);   // vertex u1
		temptab.AddUnique (vi, dnums[j+4]); // vertex u2
		temptab.AddUnique (vi, dnums[3]);   // element u1
		temptab.AddUnique (vi, dnums[7]);   // element u2
		temptab.AddUnique (vi, dnums[8]);   // vertex p
		temptab.AddUnique (vi, dnums[9]);   // vertex p
		temptab.AddUnique (vi, dnums[10]);  // vertex p
	      }
	  }

	ARRAY<int> cnt(nv);
	for (i = 0; i < nv; i++)
	  cnt[i] = temptab[i].Size();
	table = new Table<int> (cnt);
	for (i = 0; i < nv; i++)
	  for (j = 0; j < temptab[i].Size(); j++)
	    (*table)[i][j] = temptab[i][j];

	(*testout) << "Smoothingblocks = " << endl << (*table) << endl;
	return table;
      }


    else if (type == TAYLORHOOD)


      {
	(*testout) << "create smoothing blocks for taylor hood" << endl;
	// blocks associated with vertices
	// velocity: all nodes of connected elements
	// pressure: all nodes of connected elements+opposite over edge

	
	int i, j, k;
	int nv = ma.GetNV();
	int ne = ma.GetNE();
	int ned = ma.GetNEdges();

	ARRAY<INT<2> > edge2el(ned);
	ARRAY<int> eledges(4), orient(4);
	ARRAY<int> dnums, dnums2;


	INT<2> hi2;
	hi2[0] = hi2[1] = -1;
	edge2el = hi2;

	for (i = 0; i < ne; i++)
	  {
	    ma.GetElEdges(i, eledges, orient);
	    for (j = 0; j < eledges.Size(); j++)
	      {
		int enr = eledges[j];
		edge2el[enr][1] = edge2el[enr][0];
		edge2el[enr][0] = i;
	      }
	  }
	(*testout) << "edge2el = " << edge2el << endl;

	IntTable temptab(nv);

	for (i = 0; i < ne; i++)
	  {
	    GetDofNrs (i, dnums);
	    for (j = 0; j < 3; j++)
	      {
		for (k = 0; k < 12; k++)
		  {
		    temptab.AddUnique (dnums[j], dnums[k]);
		    (*testout) << "add to " << dnums[j] << ": " << dnums[k] << endl;
		  }
	      }
	  }
	for (i = 0; i < ned; i++)
	  for (j = 0; j < 2; j++)
	    for (k = 0; k < 2; k++)
	      {
		int elnr1 = edge2el[i][j];
		int elnr2 = edge2el[i][k];
		if (elnr1 != -1 && elnr2 != -1)
		  {
		    GetDofNrs (elnr1, dnums);
		    GetDofNrs (elnr2, dnums2);
		    
		    for (int jj = 0; jj < 3; jj++)
		      for (int kk = 12; kk < 15; kk++)
			temptab.AddUnique (dnums[jj], dnums[kk]);
		  }
	      }
	
	
	ARRAY<int> cnt(nv);
	for (i = 0; i < nv; i++)
	  cnt[i] = temptab[i].Size();
	Table<int> * table = new Table<int> (cnt);
	for (i = 0; i < nv; i++)
	  for (j = 0; j < temptab[i].Size(); j++)
	    (*table)[i][j] = temptab[i][j];

	(*testout) << "Smoothingblocks = " << endl << (*table) << endl;
	return table;
      }

    return 0;
  }
};



/* ********************+ Integrator ************************** */

class DiffOpStokes : public DiffOp<DiffOpStokes>
{
  // 5 components:
  // u1,1 u1,2 u2,1 u2,2 p

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 5 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const NodalFiniteElement & fel_u = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const NodalFiniteElement & fel_p = 
      dynamic_cast<const NodalFiniteElement&> (cfel[2]);
    
    int nd_u = fel_u.GetNDof();
    int nd_p = fel_p.GetNDof();

    FlatMatrix<> gradu(2, nd_u, lh);
    gradu = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_u.GetDShape(sip.IP(), lh));
    
    FlatVector<> vecp(nd_p, lh);
    vecp = fel_p.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_u; j++)
      {
	mat(0,j) = gradu(0,j);
	mat(1,j) = gradu(1,j);
	mat(2,nd_u+j) = gradu(0,j);
	mat(3,nd_u+j) = gradu(1,j);
      }
    for (int j = 0; j < nd_p; j++)
      mat(4, 2*nd_u+j) = vecp(j);
  }
};


class StokesDMat : public DMatOp<StokesDMat>
{
  CoefficientFunction * nu;
public:

  enum { DIM_DMAT = 5 };

  StokesDMat (CoefficientFunction * anu) : nu(anu) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    // (nu \nabla u, \nabla v)
    double val = Evaluate (*nu, sip);
    for (int i = 0; i < 4; i++)
      mat(i, i) = val;

    // (u1,1+u2,2, p)
    mat(4,0) = mat(0,4) = 1;
    mat(4,3) = mat(3,4) = 1;
  }  
};


class StokesIntegrator 
  : public T_BDBIntegrator<DiffOpStokes, StokesDMat, FiniteElement>
{
public:
  ///
  StokesIntegrator (CoefficientFunction * coeff)
    :  T_BDBIntegrator<DiffOpStokes, StokesDMat, FiniteElement>
  (StokesDMat (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new StokesIntegrator (coeffs[0]);
  }
  
  ///
  virtual string Name () const { return "Stokes"; }
};







class DiffOpIdU : public DiffOp<DiffOpIdU>
{
  // 2 components:
  // u1 u2

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 2 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    const NodalFiniteElement & fel_u = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    
    int nd_u = fel_u.GetNDof();

    FlatVector<> vecu(nd_u, lh);
    vecu = fel_u.GetShape (sip.IP(), lh);

    mat = 0;
    for (int j = 0; j < nd_u; j++)
      {
	mat(0,j) = vecu(j);
	mat(1,nd_u+j) = vecu(j);
      }
  }
};



class StokesUIntegrator 
  : public T_BDBIntegrator<DiffOpIdU, DiagDMat<2>, FiniteElement>
{
public:
  ///
  StokesUIntegrator ()
    :  T_BDBIntegrator<DiffOpIdU, DiagDMat<2>, FiniteElement>
  (DiagDMat<2> (new ConstantCoefficientFunction(1)))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new StokesUIntegrator ();
  }
  
  ///
  virtual string Name () const { return "Stokes IdU"; }
};






/* ******************** TestSolver ************** */


///
class NumProcTestStokes : public NumProc
{
protected:

public:
  ///
  NumProcTestStokes (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { ; }
  ///
  virtual ~NumProcTestStokes()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcTestStokes (pde, flags);
  }

  ///
  virtual void Do()
  {
    int level = ma.GetNLevels() - 1;

    cout << "test stokes, l = " << level << endl;
    if (level <= 1) return;

    const BaseSparseMatrix & mat = 
      dynamic_cast<const BaseSparseMatrix&> 
      (pde.GetBilinearForm("a")->GetMatrix());

    const BaseMatrix & norm = 
      pde.GetBilinearForm("norm")->GetMatrix();

    const MultigridPreconditioner & pre = 
      dynamic_cast<const MultigridPreconditioner &>
      (pde.GetPreconditioner("c")->GetMatrix());


    const BaseVector & constraint = pde.GetLinearForm ("constraint")->GetVector();

    BaseVector & hv1 = *mat.CreateVector();
    BaseVector & hv2 = *mat.CreateVector();
    BaseVector & hv3 = *mat.CreateVector();

    
    hv1.SetRandom();

    for (int i = 1; i < 100; i++)
      {
	double scal1 = InnerProduct (constraint, hv1);
	double scal2 = InnerProduct (constraint, constraint);
	hv1 -= (scal1 / scal2) * constraint;

	hv2 = norm * hv1;
	double norm = sqrt (InnerProduct (hv1, hv2));
	cout << "norm = " << norm << endl;
	hv1 /= norm;
	
	hv2 = mat * hv1;
	hv3 = pre * hv2;
	hv1 -= hv3;
      }
  }

  ///
  virtual string GetClassName () const
  {
    return "TestStokes";
  }

  virtual void PrintReport (ostream & ost)
  {
    ;
  }
};







namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("teststokes", NumProcTestStokes::Create);
    GetFESpaceClasses().AddFESpace ("stokes", FESpaceStokes::Create);

    GetIntegrators().AddBFIntegrator ("stokes", 2, 1,
				      StokesIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("stokesu", 2, 0,
				      StokesUIntegrator::Create);
  }
  
  Init init;
}
