/** 
   Solver for Helmholtz

   based on Cessenat and Despres, SINUM 35, 1998, pp 255-299
*/


#include <solve.hpp>

using namespace ngfem;
using namespace ngsolve;


namespace ngsolve
{



///
class NPWaves : public NumProc
{
  int p;
  double omega;
  double relp;

  ARRAY<int[3]> nbels;
  SparseMatrix<Complex> * mata;
  SparseMatrix<Complex> * matd;
  BaseMatrix * diag;

  ///
  VVector<Complex> *vecu, *vecf;
  ///
  ARRAY< ARRAY<Vec<2> >* > vl;
  ///
  ARRAY<int> eldof1;

  /// elements of finest computed level current 
  ARRAY<Vec<2> > pi;

  ARRAY<int[3]> els;

public:
  NPWaves (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    p = int (flags.GetNumFlag("order",11));
    omega = flags.GetNumFlag ("omega",1);
    relp = flags.GetNumFlag ("relp",11);
  }

  ~NPWaves ()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NPWaves (pde, flags);
  }

  ///
  virtual void Do()
  {
    GenerateTables();
    AssembleMatrix();
    Solve();
  }


  ///
  void GenerateTables ();
  ///
  void AssembleMatrix ();
  ///
  void Solve();
  ///
  Complex Evaluate (int elnr, const Vec<2> & p) const;
  /*
  ///
  void ErrorEstimator();
  ///
  void Analyze ();
  ///
  void Interpolate ();
  ///
  int ElementOfPoint (const SysVector2d & p);
  */
};










class VisualizeWave : public netgen::SolutionData
{
  const MeshAccess & ma;
  const NPWaves & npw;

public:
  VisualizeWave (const MeshAccess & ama,
		 const NPWaves & anpw)
    : SolutionData ("u", -1, 1), ma(ama), npw(anpw)
  { 
    components = 2;
  }

  virtual ~VisualizeWave ()
  { ; }
  
  virtual bool GetSurfValue (int elnr,
			     double lam1, double lam2, 
			     double * values) 
  { 
    values[0] = elnr;
    values[1] = elnr;
    Vec<2> p0, p1, p2, p;
    ArrayMem<int,10> pnums;
    ma.GetElPNums (elnr, pnums);
    ma.GetPoint (pnums[0], p0);
    ma.GetPoint (pnums[1], p1);
    ma.GetPoint (pnums[2], p2);
    p = p2 + lam1 * (p0-p2) + lam2 * (p1-p2);

    Complex val = npw.Evaluate (elnr, p);

    /*
    Complex val = 0.0;
    double fac = 0;
    if (pnums[0] == 18) fac = lam1;
    if (pnums[1] == 18) fac = lam2;
    if (pnums[2] == 18) fac = 1-lam1-lam2;
    val = fac * cos (p(0)*100 + p(1)*30);
    */
    /*
    if (elnr == 21)
      val = cos (p(0)*100 + p(1)*30);
    */
    values[0] = val.real();
    values[1] = val.imag();
    return 1;
  }
};









#ifdef ABC
  Waves :: Waves (MeshAccess & m, double aomega, int ap)
    : ma (m), vl(1)
  {
    cout << "Wave solver called" << endl;
    int i, j;

    p = ap;
    omega = aomega;

    vecu = NULL;
    
    GenerateTables();
    
    for (i = 1; i <= 0; i++)
      {
	cout << "adaptive iteration " << i << endl;
	AssembleMatrix();
	Solve();
	if (i%10 == 0)
	  {
	    Solve();
	    Solve();
	    Solve();
	    Solve();
	    Solve();
	  }
	ErrorEstimator ();
	delete &mldata.Last()->mata->GetGraph();
	delete mldata.Last()->mata;
	delete mldata.Last()->matd;
      }

    AssembleMatrix();
    Solve();

    for (i = 1; i <= 4; i++)
      {
	for (j = 1; j <= ma.GetNE(); j++)
	  Ng_SetRefinementFlag (j, 1);
	Ng_Refine ();
      }
  }

  Waves :: ~Waves ()
  {
    ;
  }


  void Waves :: Analyze ()
  {
    int i, j, k, ii, jj;
    
    SparseSystemMatrix<SysMatrix2d,SysVector2d> * mata = 
      mldata.Last()->mata;

    //    (*testout) << "matrix = " << endl << (*mata) << endl;


    EigenSystem eigen (*mata);
    eigen.Calc();
    for (i = 1; i <= eigen.NumEigenValues(); i++)
      cout << "lam (" << i << ") = " << eigen.EigenValue(i) << endl;

    DenseMatrix d1(2*p), f(2*p), dhat(2*p), hmat(2*p), dhatinv(2*p);
    

    for (i = 0; i < p; i++)
      for (j = 0; j < p; j++)
	{
	  d1.Elem(2*i+1,2*j+1) = mata->Elem(i+1, j+1, 1, 1);
	  d1.Elem(2*i+1,2*j+2) = mata->Elem(i+1, j+1, 1, 2);
	  d1.Elem(2*i+2,2*j+1) = mata->Elem(i+1, j+1, 2, 1);
	  d1.Elem(2*i+2,2*j+2) = mata->Elem(i+1, j+1, 2, 2);
	}

    
    int p1, p2;

    for (p1 = 1; p1 <= p; p1++)
      {

	if (p % p1 != 0)
	  continue;

	p2 = p/p1;

	cout << "p1 = " << p1 << " p2 = " << p2 << endl;

	f.SetScalar(0);
	for (k = 0; k < p1; k++)
	  for (ii = 0; ii < p2; ii++)
	    for (jj = 0; jj < p2; jj++)
	      {
		i = k+ii*p1;
		j = k+jj*p1;
		
		double phi = 2 * M_PI * ii * jj / p2;
		double co = cos (phi);
		double si = sin (phi);
		
		f.Elem(2*i+1,2*j+1) = co;
		f.Elem(2*i+1,2*j+2) = si;
		f.Elem(2*i+2,2*j+1) = -si;
		f.Elem(2*i+2,2*j+2) = co;
	      }
	
	//	(*testout) << "d = " << d1 << endl;
	
	// dhat = f^T d1 f
	CalcAtB (f, d1, hmat);
	Mult (hmat, f, dhat);
	
	(*testout) << "dhat = " << dhat << endl;
	
	dhatinv.SetScalar(0);
	for (i = 1; i <= 2*p; i++)
	  dhatinv.Elem(i,i) = 1 / dhat.Get(i, i);
	
	EigenSystem eigen2 (dhat, dhatinv);
	eigen2.Calc();
	/*
	for (i = 1; i <= eigen2.NumEigenValues(); i++)
	  cout << "lam (" << i << ") = " << eigen2.EigenValue(i) << endl;
	*/
	cout << "cond = " << eigen2.EigenValue(eigen2.NumEigenValues()) / eigen2.EigenValue(1) << endl;
	/*
	cout << "diag dhat = ";
	for (i = 1; i <= dhat.Height(); i++)
	  cout << dhat.Get(i,i) << " ";
	cout << endl;
	*/
      }
    
  }
  
#endif




  void NPWaves :: GenerateTables ()
  {
    cout << "Find neighbour elements" << endl;

    int ne = ma.GetNE();
    int nse = ma.GetNSE();
    int nedge = ma.GetNEdges();

    int i, j, k;
    ARRAY<int> pnums, eledges;

    HashTable<INT<2>, int> edges(5*ne+1);
    HashTable<INT<2>, int> boundedges(2*nse+1);
    ARRAY<int[2]> edge2element(nedge);

    

    for (i = 0; i < nedge; i++)
      {
	int pi1, pi2;
	ma.GetEdgePNums (i, pi1, pi2);
	INT<2> edge(pi1, pi2);
	edge.Sort();
	edges.Set (edge, i);
      }

    for (i = 0; i < nse; i++)
      {
	ma.GetSElPNums (i, pnums);
	INT<2> edge(pnums[0], pnums[1]);
	edge.Sort();
	boundedges.Set (edge, i);
      }


    for (i = 0; i < nedge; i++)
      {
	edge2element[i][0] = -1;
	edge2element[i][1] = -1;
      }


    for (i = 0; i < ne; i++)
      {
	ma.GetElPNums (i, pnums);
	for (j = 0; j < 3; j++)
	  {
	    INT<2> edge(pnums[j], pnums[(j+1)%3]);
	    edge.Sort();

	    int edgenr = edges.Get(edge);

	    edge2element[edgenr][1] = edge2element[edgenr][0];
	    edge2element[edgenr][0] = i;
	  }
      }

    nbels.SetSize (ne);
    for (i = 0; i < ne; i++)
      {
	ma.GetElPNums (i, pnums);
	for (j = 0; j < 3; j++)
	  {
	    INT<2> edge(pnums[j], pnums[(j+1)%3]);
	    edge.Sort();

	    int edgenr = edges.Get(edge);

	    nbels[i][j] = 
	      edge2element[edgenr][0] +
	      edge2element[edgenr][1] - i;
	  }
      }
	    
    

    
    for (i = 0; i < ne; i++)
      {
	ma.GetElPNums (i, pnums);
	for (j = 0; j < 3; j++)
	  {
	    INT<2> edge(pnums[j], pnums[(j+1)%3]);
	    edge.Sort();

	    if (boundedges.Used (edge))
	      {
		int els = boundedges.Get(edge);
		nbels[i][j] = -1-els;
	      }
	  }	    
      }

    (*testout) << "nbels = " << endl;
    for (i = 0; i < ne; i++)
      {
	(*testout) << "el " << i << ":";
	for (j = 0; j < 3; j++)
	  (*testout) << " " << nbels[i][j];
	(*testout) << endl;
      }


    /*
    // copy mesh points and elements
    int np = ma.GetNP();
    double pa[3];
    pi.SetSize(np);
    for (i = 1; i <= np; i++)
      {
	ma.GetPoint (i, pa);
	pi.Elem(i) = SysVector2d (pa[0], pa[1]);
      }
    els.SetSize (ne);
    for (i = 1; i <= ne; i++)
      {
	ma.GetElPNums (i, pnums);
	for (j = 1; j <= 3; j++)
	  els.Elem(i).I(j) = pnums.Get(j);
      }
    */


    // ray vectors
    vl.SetSize (ne);

    for (j = 0; j < ne; j++)
      {
	int pj = p;
	
	double h = sqrt (ma.ElementVolume(j));

	int pjm = int (relp * omega * h);
	if (pjm < 5) pjm = 5;
	
	if (pj > pjm) pj = pjm;
	if (pj % 2 == 0) pj++;


	cout << "el " << j << ", h = " << h;
	cout << ", pj = " << pj << ", pjm = " << pjm << endl;

	vl[j] = new ARRAY<Vec<2> > (pj);
	ARRAY<Vec<2> > & vlj = *vl[j];
	for (i = 0; i < pj; i++)
	  {
	    vlj[i](0) = cos (2 * M_PI * i / pj + 0.1);
	    vlj[i](1) = sin (2 * M_PI * i / pj + 0.1);
	  }
      }

    eldof1.SetSize(ne+1);
    int cnt = 0;
    for (i = 0; i < ne; i++)
      {
	eldof1[i] = cnt;
	cnt += vl[i]->Size();
      }
    eldof1[ne] = cnt;
  }



  void NPWaves :: AssembleMatrix ()
  {
    int i, j, jj, k, l, m, n;
    int ne = ma.GetNE();

    ARRAY<int> pnums;

    // generate matrix graph

    int dim = eldof1[ne];
    ARRAY<int> nze(dim);

    for (i = 0; i < ne; i++)
      {
	int nnb = vl[i]->Size();
	for (j = 0; j < 3; j++)
	  if (nbels[i][j] >= 0)
	    nnb += vl[nbels[i][j]]->Size();
	
	for (j = eldof1[i]; j < eldof1[i+1]; j++)
	  nze[j] = nnb;
      }

    MatrixGraph * graph = new MatrixGraph (nze);

    for (i = 0; i < ne; i++)
      {
	int pi = vl[i]->Size();
	int eldof1i = eldof1[i];

	for (k = 0; k < pi; k++)
	  for (l = 0; l < pi; l++)
	    graph->CreatePosition ( eldof1i+k, eldof1i+l);

	for (jj = 0; jj < 3; jj++)
	  {
	    j = nbels[i][jj];

	    if (j >= 0)
	      {
		int pj = vl[j]->Size();
		int eldof1j = eldof1[j];
		for (k = 0; k < pi; k++)
		  for (l = 0; l < pj; l++)
		    graph->CreatePosition ( eldof1i+k, eldof1j+l);
	      }
	  }
      }

    cout << "has graph" << endl;
    // (*testout) << "graph = " << (*graph) << endl;

    mata = new SparseMatrix<Complex> (*graph);
    matd = new SparseMatrix<Complex> (*graph);

    vecu = new VVector<Complex> (dim);
    vecf = new VVector<Complex> (dim);

    (*vecu) = 0.0;
    (*vecf) = 0.0;

    for (k = 0; k < ne; k++)
      {
	ma.GetElPNums (k, pnums);
	for (n = 0; n < 3; n++)
	  {
	    Vec<2> p1, p2;
	    ma.GetPoint (pnums[n], p1);
	    ma.GetPoint (pnums[(n+1)%3], p2);
	    
	    Vec<2> v = p2-p1;
	    double len = L2Norm(v);
	    double nu1 = (v(1)) / len;
	    double nu2 = (-v(0)) / len;

	    Vec<2> nu(nu1, nu2);
	    
	    // assemble D matrix
	    int pk = vl[k]->Size();
	    for (l = 0; l < pk; l++)
	      for (m = 0; m < pk; m++)
		{
		  Vec<2> nukl = (*vl[k])[l];
		  Vec<2> nukm = (*vl[k])[m];

		  double zexp = omega * (InnerProduct (nukm,p1) - InnerProduct(nukl,p1));
		  double hexp = omega * (InnerProduct (nukm,v) - InnerProduct(nukl,v)) / 2;

		  double fac = sqr (omega) * len * 
		    ( 1 - InnerProduct (nu, nukm)) * (1 - InnerProduct (nu, nukl));

		  if (fabs (hexp) > 1e-8)
		    fac *= sin (hexp) / hexp;

		  Complex z = fac * exp (Complex (0,1) * (zexp+hexp));

		  int mati = eldof1[k] + l;
		  int matj = eldof1[k] + m;

		  (*mata)(mati, matj) += z;
		  (*matd)(mati, matj) += z;

		  if (nbels[k][n] < 0)
		    { // boundary

		      int bc = ma.GetSElIndex (-nbels[k][n]-1);
		      (*testout) << "bc = " << bc << endl;
		      /*
			// 30 fingers:
		      const double ta[] =
		      { -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
			-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
			-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
			-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
			-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
			-1, 1, -1, 1, -1, 1, -1, 1, -1,
			  1, 0, 0, 1 };
		      const double ga[] =
		      { 1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1, 0, 
			1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1, 0, 
			1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1,
			0, 0, 0, 0, 0 };
		      */
		      const double ta[] =
		      { -1, -1, 1, 0 };
		      const double ga[] =
		      { -1, 1, 0, 0 };

		      double t = 0;
		      //		      if (nu2 < -0.5) t = -1;
		      t = ta[bc]; 

		      // boundary cond. 
		      double fac = t * sqr (omega) * len * 
			(1 - InnerProduct(nu, nukm)) * (1 + InnerProduct (nu, nukl));
		      
		      if (fabs (hexp) > 1e-8)
			fac *= sin (hexp) / hexp;
		      
		      (*mata)(mati, matj) -= fac * exp (Complex (0,1) * (zexp+hexp));

		      if (m == 1)
			{
			  Vec<2> nu0(0,-1);

			  double zexp = omega * (InnerProduct (nu0,p1) - InnerProduct(nukl,p1));
			  double hexp = omega * (InnerProduct (nu0,v) - InnerProduct(nukl,v)) / 2;

			  double fac = ga[bc] * sqr (omega) * len * 
			    ((1+t)*(InnerProduct(nu0, nu)) + (1-t)) * 
			    (1 + InnerProduct(nu, nukl));
		      
			  if (fabs (hexp) > 1e-8)
			    fac *= sin (hexp) / hexp;
			  
			  (*vecf)(mati) += fac * exp (Complex(0,1) * (zexp+hexp));
			}
		    }
		}


	    // assemble C matrix
	    j = nbels[k][n];
	    if (j >= 0)
	      {
		int pj = vl[j]->Size();
		for (l = 0; l < pk; l++)
		  for (m = 0; m < pj; m++)
		    {
		      Vec<2> nujm = (*vl[j])[m];
		      Vec<2> nukl = (*vl[k])[l];
		      
		      double zexp = omega * (InnerProduct(nujm,p1) - InnerProduct(nukl,p1));
		      double hexp = omega * (InnerProduct(nujm,v) - InnerProduct(nukl,v))/2;
		      
		      double fac = sqr (omega) * len * ( 1 + InnerProduct(nu,nujm)) * 
			(1 + InnerProduct (nu, nukl));
		      if (fabs (hexp) > 1e-8)
			fac *= sin (hexp) / hexp;
		      
		      int mati = eldof1[k] + l;
		      int matj = eldof1[j] + m;

		      (*mata) (mati, matj) -= fac * exp (Complex(0,1)*(zexp+hexp));
		    }
	      }
	  }
      }
    
    cout << "matrix done" << endl;
    //    (*testout) << "mat = " << *mata << endl;

    ARRAY<int> cnt(ne);
    for (i = 0; i < ne; i++)
      cnt[i] = eldof1[i+1]-eldof1[i];

    Table<int> * blocks = new Table<int>(cnt);
    for (i = 0; i < ne; i++)
      for (j = 0; j < vl[i]->Size(); j++)
	(*blocks)[i][j] = eldof1[i]+j;
    
    (*testout) << "blocks = " << blocks << endl;

    diag = matd -> CreateBlockJacobiPrecond (*blocks);

    cout << "has blockjac" << endl;

    /*
      // point source
    SysVector2d p0 (0.55, 0.45);
    int elnr =  ElementOfPoint (p0);
    cout << "el0 = " << elnr << endl;
    
    for (l = 1; l <= p; l++)
      {
	double exp = -omega * (vl.Get(l) * p0);
	vecf -> Elem( (elnr-1)*p+l , 1 ) = 
	  2 * omega * sin (exp);
	vecf -> Elem( (elnr-1)*p+l , 2 ) = 
	  -2 * omega * cos (exp);
      }
    */

    //    (*testout) << "vecf = " << (*vecf) << endl;
    //    vecu -> SetScalar (0);
  }


  void NPWaves :: Solve ()
  {
    int i, j;

    /*
    //    BiCGStabSolver ainv (*mldata.Last()->mata, *mldata.Last()->diag);
    //    NormalCGSolver ainv (*mldata.Last()->mata, *mldata.Last()->diag);

    ainv.SetMaxSteps (500);
    ainv.SetPrintRates();
    cout << "normf = " << vecf->L2Norm() << endl;
    vecu -> SetScalar (0);

    ainv.Mult (*vecf, *vecu);
    cout << "steps = " << ainv.GetSteps() << endl;
    */

      // simple iteration
    int n = 0;
    double err0 = -1, err = 1;

    BaseVector & d = *vecf->CreateVector();
    BaseVector & w = *vecf->CreateVector();


    //    vecu -> SetScalar (0);
    while (err > 1e-10 * err0 && n < 2000)
      {
	d = (*vecf) - (*mata) * (*vecu);
	w = (*diag) * d;


	//	d = )mata -> Residuum (*vecu, *vecf, d);
	//	mldata.Last()->diag -> Mult (d, w);
	err = L2Norm (d);
	// cout << "IP = " << S_InnerProduct<Complex> (d, w) << endl;
	if (n == 0) err0 = err;
	(*vecu) += 0.8 * w;
	n++;
	if (n % 10 == 0)
	  cout << "n = " << n << ", err = " << err << endl;
      }



    netgen::SolutionData * vis =
      new VisualizeWave (ma, *this);
    
    Ng_SolutionData soldata;
    Ng_InitSolutionData (&soldata);
	
    soldata.name = const_cast<char*> ("wave");
    soldata.data = 0;
    soldata.components = 2;
    soldata.iscomplex = 1;
    soldata.draw_surface = 1;
    soldata.draw_volume  = 0;
    soldata.dist = soldata.components;
    soldata.soltype = NG_SOLUTION_VIRTUAL_FUNCTION;
    soldata.solclass = vis;
    Ng_SetSolutionData (&soldata);    

    /*
    // multigrid
    cout << "call mg" << endl;

    int n = 0;
    double err0 = -1, err = 1;

    SystemVector<SysVector2d> d(vecu->Size());
    SystemVector<SysVector2d> w(vecu->Size());

    int ne = mldata.Last() -> ne;

    int cl = mldata.Size() - 1;

    int nec = mldata.Get(cl) -> ne;

    ARRAY<int> coarseof(ne);
    for (i = 1; i <= ne; i++)
      {
	int c = i;
	while (c > nec)
	  c = Ng_GetParentElement(c);
	coarseof.Elem(i) = c;
      }

    SystemVector<SysVector2d> dc(p * nec);
    SystemVector<SysVector2d> wc(p * nec);

    NormalCGSolver acinv (*mldata.Get(cl)->mata, 
			  *mldata.Get(cl)->diag);
    acinv.SetMaxSteps (5000);
    acinv.SetPrintRates(0);


    vecu -> SetScalar (0);
    while (err > 1e-8 * err0 && n < 1000)
      {
	for (i = 1; i <= 3; i++)
	  {
	    mldata.Last()->mata -> Residuum (*vecu, *vecf, d);
	    mldata.Last()->diag -> Mult (d, w);
	    vecu -> Add (0.8, w);
	  }

	err = d.L2Norm();
	if (n == 0) err0 = err;

	mldata.Last()->mata -> Residuum (*vecu, *vecf, d);

	(*testout) << "d = " << d << endl;

	dc.SetScalar (0);
	for (i = 1; i <= ne; i++)
	  for (j = 1; j <= p; j++)
	    dc.Elem((coarseof.Get(i)-1) * p + j) += d.Elem( (i-1)*p + j );

	//	mldata.Elem(1)->diag -> Mult (dc, wc);
	acinv.Mult (dc, wc);

	for (i = 1; i <= ne; i++)
	  for (j = 1; j <= p; j++)
	    w.Elem( (i-1)*p + j) = wc.Elem( (coarseof.Get(i)-1)*p+j );

	vecu -> Add (1, w);

	n++;
	cout << "n = " << n << ", err = " << err << endl;
      }

    
    d.Set (1, *vecf);

    (*testout) << "d = " << d << endl;
    dc.SetScalar (0);
    for (i = 1; i <= ne; i++)
      for (j = 1; j <= p; j++)
	dc.Elem((coarseof.Get(i)-1) * p + j) += d.Elem( (i-1)*p + j );
    (*testout) << "dc = " << dc << endl;


    acinv.Mult (dc, wc);
    cout << "steps = " << acinv.GetSteps() << endl;
    
    for (i = 1; i <= ne; i++)
      for (j = 1; j <= p; j++)
	w.Elem( (i-1)*p + j) = wc.Elem( (coarseof.Get(i)-1)*p+j );

    vecu -> Set (1, w);
    */

    /*
    int ne = mldata.Last() -> ne;
    for (i = 1; i <= ne; i++)
      {
	(*testout) << "El " << i << endl;
	for (j = 1; j <= vl.Get(i)->Size(); j++)
	  {
	    (*testout) << vecu->Get(eldof1.Get(i)+j-1) << " ";
	  }
	(*testout) << endl;
      }
    */
  }


Complex NPWaves :: Evaluate (int elnr, const Vec<2> & p) const
{
  const ARRAY<Vec<2> > & vi = *vl[elnr];
  int first = eldof1[elnr];
  
  Complex val = 0.;
  //  (*testout) << "val, elnr = " << elnr << ", p = " << p << endl;
  for (int i = 0; i < vi.Size(); i++)
    {
      val += (*vecu)(first+i) * exp (omega * Complex (0,1) * InnerProduct (vi[i], p));
    }
  //  (*testout) << "val = " << val << endl;
  return val;
}


#ifdef ABC
  void NPWaves :: ErrorEstimator ()
{
  int i, j, k, l, n, m;

  int ne = ma.GetNE();

  ARRAY<int> pnums;
  ARRAY<INT_3> & nbels = *mldata.Last()->nbels;

  ARRAY< SystemVector<SysVector2d>* > vltest;

  vltest.SetSize (ne);

  int ptest = 299;
  
  for (j = 1; j <= ne; j++)
    {
      int pj = ptest;
      
      vltest.Elem(j) = new SystemVector<SysVector2d> (pj);
      SystemVector<SysVector2d> & vlj = *vltest.Elem(j);
      for (i = 0; i < pj; i++)
	{
	  vlj.Elem(i+1, 1) = cos (2 * M_PI * i / pj + 0.1);
	  vlj.Elem(i+1, 2) = sin (2 * M_PI * i / pj + 0.1);
	}
    }
  
  
  SystemVector<SysVector2d> res(ne*ptest);
  res.SetScalar (0);
  

  for (k = 1; k <= ne; k++)
    {
      ma.GetElPNums (k, pnums);
      for (n = 1; n <= 3; n++)
	{
	  double p1[3], p2[3];
	  ma.GetPoint (pnums.Get(n), p1);
	  ma.GetPoint (pnums.Get(n%3+1), p2);
	  
	  double len = sqrt ( sqr (p1[0]-p2[0]) + sqr (p1[1]-p2[1]));
	  double nu1 = (p2[1]-p1[1]) / len;
	  double nu2 = (p1[0]-p2[0]) / len;
	  
	  SysVector2d nu(nu1, nu2);
	  SysVector2d vp1(p1[0], p1[1]);
	  SysVector2d vp2(p2[0], p2[1]);
	  SysVector2d vp12(vp2);
	  vp12.Add (-1, vp1);

	    
	  // assemble D matrix
	  int pk = vl.Get(k)->Size();
	  for (l = 1; l <= ptest; l++)
	    for (m = 1; m <= pk; m++)
	      {
		SysVector2d nukl = vltest.Get(k)->Get(l);
		SysVector2d nukm = vl.Get(k)->Get(m);

		double zexp = omega * (nukm*vp1 - nukl*vp1);
		double hexp = omega * (nukm*vp12 - nukl*vp12) / 2;

		double fac = sqr (omega) * len * 
		  ( 1 - nu * nukm) * (1 - nu * nukl);
		
		if (fabs (hexp) > 1e-8)
		  fac *= sin (hexp) / hexp;
		
		double zre = fac * cos (zexp+hexp);
		double zim = fac * sin (zexp+hexp);
		
		int mati = (k-1) * ptest + l;
		int matj = eldof1.Get(k) + m-1;
		
		res.Elem(mati, 1) += 
		  zre * vecu->Get (matj, 1) - zim * vecu->Get (matj, 2);
		res.Elem(mati, 2) += 
		  zim * vecu->Get (matj, 1) + zre * vecu->Get (matj, 2);


		if (nbels.Get(k).I(n) < 0)
		  { // boundary
		    
		    int bc = ma.GetSElIndex (-nbels.Get(k).I(n));

		    const double ta[] =
		    { -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
		      -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
		      -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
		      -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
		      -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, 
		      -1, 1, -1, 1, -1, 1, -1, 1, -1,
		      1, 0, 0, 1 };
		    const double ga[] =
		      { 1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1, 0, 
			1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1, 0, 
			1, 0, -1, 0, 1, 0, -1, 0, 1, 0, 
			-1, 0, 1, 0, -1, 0, 1, 0, -1,
			0, 0, 0, 0, 0 };
		    /*
		      const double ta[] =
		      { -1, 1, 0, 1 };
		      const double ga[] =
		      { 1, 1, 0, 1 };
		    */

		    double t = 0;
		    //		      if (nu2 < -0.5) t = -1;
		    t = ta[bc-1]; 
		    
		    // boundary cond. 
		    double fac = t * sqr (omega) * len * 
		      (1 - nu * nukm) * (1 + nu * nukl);
		    
		    if (fabs (hexp) > 1e-8)
		      fac *= sin (hexp) / hexp;
		    
		    double zre = fac * cos (zexp+hexp);
		    double zim = fac * sin (zexp+hexp);

		    res.Elem(mati, 1) -= 
		      zre * vecu->Get (matj, 1) - zim * vecu->Get (matj, 2);
		    res.Elem(mati, 2) -= 
		      zim * vecu->Get (matj, 1) + zre * vecu->Get (matj, 2);

		    if (m == 1)
		      {
			SysVector2d nu0(0, -1);
			
			double zexp = omega * (nu0*vp1 - nukl*vp1);
			double hexp = omega * (nu0*vp12 - nukl*vp12) / 2;
			
			double fac = ga[bc-1] * sqr (omega) * len * 
			  ((1+t)*(nu0 * nu) + (1-t)) * 
			  (1 + nu * nukl);
			
			if (fabs (hexp) > 1e-8)
			  fac *= sin (hexp) / hexp;
			
			double zre = fac * cos (zexp+hexp);
			double zim = fac * sin (zexp+hexp);

			res.Elem (mati, 1) -= zre;
			res.Elem (mati, 2) -= zim;

		      }
		  }
	      }
	  

	    // assemble C matrix
	    j = nbels.Get(k).I(n);
	    if (j > 0)
	      {
		int pj = vl.Get(j)->Size();
		for (l = 1; l <= ptest; l++)
		  for (m = 1; m <= pj; m++)
		    {
		      SysVector2d nujm = vl.Get(j)->Get(m);
		      SysVector2d nukl = vltest.Get(k)->Get(l);
		      
		      double zexp = omega * (nujm*vp1 - nukl*vp1);
		      double hexp = omega * (nujm*vp12 - nukl*vp12)/2;
		      
		      double fac = sqr (omega) * len * ( 1 + nu * nujm) * 
			(1 + nu * nukl);
		      if (fabs (hexp) > 1e-8)
			fac *= sin (hexp) / hexp;
		      
		      double zre = fac * cos (zexp+hexp);
		      double zim = fac * sin (zexp+hexp);

		      int mati = (k-1) * ptest + l;
		      int matj = eldof1.Get(j) + m-1;

		      res.Elem(mati, 1) -= 
			zre * vecu->Get (matj, 1) - zim * vecu->Get (matj, 2);
		      res.Elem(mati, 2) -= 
			zim * vecu->Get (matj, 1) + zre * vecu->Get (matj, 2);

		      /*		      
		      mata->Elem (mati, matj, 1, 1) -= zre;
		      mata->Elem (mati, matj, 1, 2) -= -zim;
		      mata->Elem (mati, matj, 2, 1) -= zim;
		      mata->Elem (mati, matj, 2, 2) -= zre;
		      */
		    }
	      }
	  }
      }



  
  // prolongate u
  ARRAY<int> eldof1old(ne);
  ARRAY<int> elpiold(ne);
  SystemVector<SysVector2d> *
    vecuold = new SystemVector<SysVector2d> (vecu->Size());
  vecuold->Set (1, *vecu);
  vecu->SetScalar (0);

  for (i = 1; i <= ne; i++)
    {
      eldof1old.Elem(i) = eldof1.Get(i);
      elpiold.Elem(i) = vl.Elem(i)->Size();
    }


  
  (*testout) << "residuum" << endl;
  /*
  for (i = 1; i <= ne; i++)
    {
      (*testout) << "el " << i << endl;
      for (j = 1; j <= ptest; j++)
	{
	  (*testout) << res.Elem((i-1)*ptest+j);
	}
    }
  */
  double maxerrel = 0;
  int refels = 0;
  for (i = 1; i <= ne; i++)
    {
      (*testout) << "el " << i << endl;
      double maxerr = 0.001;
      int maxind = 0;
      for (j = 1; j <= ptest; j++)
	{
	  double err =
	    sqr (res.Elem((i-1)*ptest+j).Get(1)) +
	    sqr (res.Elem((i-1)*ptest+j).Get(2));

	  int hasit = 0;
	  SysVector2d diff;
	  for (k = 1; k <= vl.Get(i)->Size(); k++)
	    {
	      diff = vltest.Get(i)->Get(j);
	      diff.Add (-1, vl.Get(i)->Get(k));
	      if (sqr (diff.Get(1)) + sqr (diff.Get(2)) < 1e-4)
		{
		  hasit = 1;
		}
	    }
	  /*
	  if (err > maxerr && hasit)
	    cout << "has vector !!" << endl;
	  */
	  if (err > maxerr && !hasit)
	    {
	      maxerr = err;
	      maxind = j;
	    }
	}

      if (maxerr > maxerrel)
	maxerrel = maxerr;

      if (maxind)
	{
	  (*testout) << "maxerr = " << maxerr << " dir = " << maxind << endl;
	  vl.Elem(i)->Append (vltest.Get(i)->Get(maxind));
	  refels++;
	}
    }

  eldof1.SetSize(ne);
  int cnt = 1;
  for (i = 1; i <= ne; i++)
    {
      eldof1.Elem(i) = cnt;
      cnt += vl.Get(i)->Size();
    }

  vecu = new SystemVector<SysVector2d> (cnt);


  for (i = ne; i >= 1; i--)
    {
      int ed1old = eldof1old.Get(i);
      int ed1    = eldof1.Get(i);
      for (j = elpiold.Get(i)-1; j >= 0; j--)
	{
	  vecu->VElem(ed1+j, 1) = vecuold->VElem(ed1old+j,1);
	  vecu->VElem(ed1+j, 2) = vecuold->VElem(ed1old+j,2);
	}
    }

  delete vecuold;

  cout << "Error estimator: " << maxerrel << endl;
  cout << "refine els: " << refels << endl;
}



  void NPWaves :: Interpolate ()
  {
    int i, j;
    int np = ma.GetNP();
    int ne = ma.GetNE();

    cout  << "Interpolate wave" << endl;

    SystemVector<SysVector2d> * vecufine =
      new SystemVector<SysVector2d> (np);    
    
    Vector & order = *new Vector (ne);


    for (i = 1; i <= np; i++)
      {
	double point[3];
	ma.GetPoint (i, point);
	SysVector2d pv(point[0], point[1]);
	
	int el = ElementOfPoint (pv);
	if (el == 0)
	  {
	    cout << "el = 0 !!!" << endl;
	    el = 1;
	  }
	double vr = 0, vi = 0;
	for (j = 1; j <= vl.Get(el)->Size(); j++)
	  {
	    // u = sum uj e^(vl * x)
	    SysVector2d nuij = vl.Get(el)->Get(j);
	    double co = cos (omega * (nuij * pv));
	    double si = sin (omega * (nuij * pv));

	    double vur = vecu->Get( eldof1.Get(el)+j-1, 1);
	    double vui = vecu->Get( eldof1.Get(el)+j-1, 2);

	    vr += co * vur - si * vui;
	    vi += co * vui + si * vur;
	  }

	vecufine->Elem(i, 1) = vr;
	vecufine->Elem(i, 2) = vi;
      }


    ARRAY<int> pnums;
    double point[3], cp[3];
    for (i = 1; i <= ne; i++)
      {
	ma.GetElPNums (i, pnums);
	cp[0] = cp[1] = cp[2] = 0;
	for (j = 1; j <= pnums.Size(); j++)
	  {
	    ma.GetPoint (pnums.Get(j), point);
	    cp[0] += 1.0/3.0 * point[0];
	    cp[1] += 1.0/3.0 * point[1];
	  }

	SysVector2d pv(point[0], point[1]);
	
	int el = ElementOfPoint (pv);
	if (el)
	  {
	    order.Elem(i) = vl.Get(el)->Size();
	  }
      }
    

    
    Ng_SolutionData soldata;
    Ng_InitSolutionData (&soldata);
    
    soldata.name = "solution";
    
    soldata.data = &vecufine->Elem(1,1);    
    soldata.components = 2;
    soldata.dist = 2;
    soldata.soltype = NG_SOLUTION_NODAL;
    Ng_SetSolutionData (&soldata);


    soldata.name = "order";
    
    soldata.data = &order.Elem(1);    
    soldata.components = 1;
    soldata.dist = 1;
    soldata.soltype = NG_SOLUTION_SURFACE_ELEMENT;
    Ng_SetSolutionData (&soldata);

  }


  int NPWaves :: ElementOfPoint (const SysVector2d & p)
  {
    int i, j;
    
    for (i = 1; i <= els.Size(); i++)
      {
	SysVector2d p1 = pi.Get(els.Get(i).I1());
	SysVector2d p2 = pi.Get(els.Get(i).I2());
	SysVector2d p3 = pi.Get(els.Get(i).I3());

	p2.Add (-1, p1);
	p3.Add (-1, p1);

	SysMatrix2d m;
	m.Elem(1,1) = p2.Get(1);
	m.Elem(2,1) = p2.Get(2);

	m.Elem(1,2) = p3.Get(1);
	m.Elem(2,2) = p3.Get(2);

	SysVector2d rs(p), lam;
	rs.Add (-1, p1);
	
	m.Solve (rs, lam);
	if (lam.Get(1) > -1e-6 && lam.Get(2) > -1e-6 &&
	    lam.Get(1) + lam.Get(2) < 1+1e-6)
	  return i;
      }
    return 0;
  }

}

#endif

  
  


  

  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("waves", NPWaves::Create, NPWaves::PrintDoc);
    }
    
    Init init;
  }
}





