
#include "../include/solve.hpp"
#include <comp.hpp>
#include <fem.hpp>
namespace ngsolve
{
  using namespace ngsolve; 
  #include "fourier.hpp"

inline Complex Conj (Complex c)
{
  return Complex (c.real(), -c.imag());
}


 
//Functions of FourierFESpace********************************************************************************************



  void FourierFESpace :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if(ma.GetSElIndex(elnr)==boundary)
    {
      dnums.SetSize(ndof);
      for(int i=0; i<ndof;i++)
        dnums[i] = i; 
    }
    else dnums.SetSize(0);

  }

 FourierFESpace :: FourierFESpace (const MeshAccess & ama, const Flags & aflags) : FESpace (ama, aflags)
  {
    boundary = int (aflags.GetNumFlag ("bc", 1))-1;
    above = aflags.GetDefineFlag("above");

    int i = int (aflags.GetNumFlag ("fourierorder", 0));
    ndof = 2*i+1;
    if(above)
      ndof += 1;

    segm     = new FE_SegmDummy;

    prol = new LinearProlongation(*this);
  }


//Functions of Periodic2DSpace*******************************************************************************

Periodic2DSpace :: Periodic2DSpace (const MeshAccess & ama, const Flags & flags)
    : H1HighOrderFESpace (ama, flags)
{ ;  }


  Periodic2DSpace :: ~Periodic2DSpace ()
  {
    ;
  }


  void Periodic2DSpace :: Update ()
{
   H1HighOrderFESpace::Update();

   dofmap.SetSize (GetNDof());
   for (int i = 0; i < dofmap.Size(); i++)
      dofmap[i] = i;

    ARRAY<INT<2> > pbcPairs;


     // periodic Vertices 
    // pbcPairs.SetSize(ma.GetNPairsPeriodicVertices());
    ma.GetPeriodicVertices(0, pbcPairs);

    for (int i = 0; i < pbcPairs.Size(); i++)
      dofmap[pbcPairs[i][1]] = pbcPairs[i][0];
    
    // periodic Edges 
    ARRAY<INT<2> > perEdges; 

    int npe = ma.GetNPairsPeriodicEdges(0);
    perEdges.SetSize(npe); 
    ma.GetPeriodicEdges(0,perEdges); 
	
    for(int i=0; i<npe; i++)
    {
      int dof_e00 = GetFirstEdgeDof(perEdges[i][0]); 
      int dof_e01 = GetFirstEdgeDof(perEdges[i][0]+1); 
      int dof_e10 = GetFirstEdgeDof(perEdges[i][1]); 
      int dof_e11 = GetFirstEdgeDof(perEdges[i][1]+1); 

      int diff = dof_e01-dof_e00;

      if(diff != dof_e11-dof_e10)
         throw Exception ("Dofs of periodic edges don't match");

      for(int j=0;j<diff;j++)
	 dofmap[dof_e10+j] = dof_e00+j;
    }
}


  void Periodic2DSpace::GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    H1HighOrderFESpace::GetDofNrs(elnr,dnums);
    for(int i=0; i<dnums.Size(); i++)
      dnums[i] = dofmap[dnums[i]];
  }

  void Periodic2DSpace::GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    H1HighOrderFESpace::GetSDofNrs(selnr,dnums);
    for(int i=0; i<dnums.Size(); i++)
      dnums[i] = dofmap[dnums[i]];
  }

 void Periodic2DSpace::VTransformMC (int elnr, bool boundary,
			FlatMatrix<Complex> & mat, TRANSFORM_TYPE type) const
{
  if (!boundary)
    {
      ARRAY<int> dnums;
      H1HighOrderFESpace::GetDofNrs (elnr, dnums);
      for (int i = 0; i < dnums.Size(); i++)
        if (dofmap[dnums[i]] != dnums[i])   // dof on right boundary
          {
            if (type & TRANSFORM_MAT_LEFT)
              for (int j = 0; j < mat.Width(); j++)
                 mat(i,j) *= Conj(gamma);
            if (type & TRANSFORM_MAT_RIGHT)
              for (int j = 0; j < mat.Height(); j++)
                 mat(j,i) *= gamma;
          }
   }
}

 void Periodic2DSpace::VTransformVC (int elnr, bool boundary,
			     FlatVector<Complex> & vec, TRANSFORM_TYPE type) const
{  
  if (!boundary)
    {
      ARRAY<int> dnums;
      H1HighOrderFESpace::GetDofNrs (elnr, dnums);

      if (type & TRANSFORM_SOL)
        for (int i = 0; i < dnums.Size(); i++)
          if (dofmap[dnums[i]] != dnums[i])   // dof on right boundary
            vec(i) *= gamma;
    }
}


//****************************Special Elements**********************************


//FourrierElement******************************************************************

FourierElement::FourierElement (ARRAY<Vec<2,Complex> >& ak,
                    ARRAY<Complex[2]>& aS,
		    int anormal,int aelnr,
		    CoefficientFunction *acoeff,
		    ARRAY<int>& adnums,
		    ElementTransformation aeltrans,
		    const FESpaceDefract& aspace)
      : k(ak), normal(anormal),elnr(aelnr),coeff(acoeff),eltrans(aeltrans),space(aspace),S(aS)
    {
       kSize = k.Size();
       ndof_k = (normal == 1) ?  (k.Size()) : (k.Size()/2);
       ndof = adnums.Size();
       ndof_f = ndof-ndof_k;
       dnums=adnums;
       alpha = 5.0;
    }

void FourierElement::Assemble (FlatMatrix<Complex> & elmat,LocalHeap & lh) const
    {

      static int timer = NgProfiler::CreateTimer ("AssembleFourierElement");
      NgProfiler::RegionTimer reg (timer);

      const Periodic2DSpace* h = dynamic_cast<const Periodic2DSpace*> (space[0]);
 
      elmat.AssignMemory (ndof,ndof, lh);

      const H1HighOrderSegm<> & seg = dynamic_cast<const H1HighOrderSegm<> & > (h->GetSFE(elnr,lh));
	
      Vector<> shapesr(ndof_f);
      Vector<Complex> shapes(ndof_f);

      ARRAY<int> dnums_orig(ndof_f);
      h -> H1HighOrderFESpace::GetSDofNrs (elnr, dnums_orig);

      const IntegrationRule & ir = GetIntegrationRules().SelectIntegrationRule(seg.ElementType(),20);

      elmat = 0;

      for ( int i = 0; i<ir.GetNIP(); i++)
      {
	SpecificIntegrationPoint<1,2> sip(ir[i], eltrans, lh);
        double det = fabs (sip.GetJacobiDet());
	double coefficient = coeff->Evaluate(sip);
        double weight = ir[i].Weight();

        Vec<2> p;
        eltrans.CalcPoint (ir[i],p,lh);

        seg.CalcShape(ir[i],shapesr);
        shapes = shapesr;
        for (int j = 0; j < ndof_f; j++)
          if (h->DofMap(dnums_orig[j]) != dnums_orig[j])
	    shapes(j) *= h->GetGamma();

        if(normal == 1)
          DoIntegrationAbove(elmat,shapes,p,det,coefficient,weight);
        else
          DoIntegrationBelow(elmat,shapes,p,det,coefficient,weight);

        for( int j=0; j<ndof_f; j++)
          for( int k=0; k<ndof_f; k++)
   	    elmat(j,k) += alpha * shapes[k]*Conj(shapes[j])*weight;
      }
    }


void FourierElement::DoIntegrationAbove (FlatMatrix<Complex> & elmat,Vector<Complex>& shapes, Vec<2> p,double det, double coef, double weight) const
    {
        Complex I = Complex(0,1);
        ARRAY<Complex> FourierShape(kSize),DFourierShape(kSize);
	for( int j=0; j < ndof_k; j++)
	{
	  FourierShape[j] = exp(I*k[j](0)*p(0));

	    // dv/dn = (i k v) . n
	    // outer normal n = (0,1) for upper boundary, n = (0,-1) for lower boundary
	  DFourierShape[j] = FourierShape[j]*double(normal)*k[j](1)*I;

	    // 1st line:  
	    // - \int c du_w/dn v_f
	    // second integrator 
	    // - 1/h \int u_w v_f
	    // 3rd line: 
	    // hermite
	}
	for( int j=0; j < ndof_k; j++)
	{
          for(int k=0; k < ndof_f; k++)
	  {
	    elmat(k,j+ndof_f) -= DFourierShape[j]*Conj(shapes[k])*(weight*coef*det);
   	    elmat(k,j+ndof_f) -= alpha * FourierShape[j] * (Conj(shapes[k])*weight);
            elmat(j+ndof_f,k) = Conj (elmat(k,j+ndof_f));
	  }

	    // int ( u_w dv_w / dn )
	    // 1/h int ( u_w v_w )


	  elmat(j+ndof_f,j+ndof_f) += FourierShape[j]*Conj(DFourierShape[j])*(coef*det*weight);
	  elmat(j+ndof_f,j+ndof_f) += alpha * FourierShape[j]*Conj(FourierShape[j])*weight;
	  if(j==(ndof_k/2-1))
	  {
	    elmat(ndof_k-1+ndof_f,j+ndof_f) += FourierShape[j]*Conj(DFourierShape[ndof_k-1])*(coef*det*weight);
	    elmat(ndof_k-1+ndof_f,j+ndof_f) += alpha * FourierShape[j]*Conj(FourierShape[ndof_k-1])*weight;
	    elmat(j+ndof_f,ndof_k-1+ndof_f) += FourierShape[ndof_k-1]*Conj(DFourierShape[j])*(coef*det*weight);
	    elmat(j+ndof_f,ndof_k-1+ndof_f) += alpha * FourierShape[ndof_k-1]*Conj(FourierShape[j])*weight;
          }
	}

    }


void FourierElement::DoIntegrationBelow (FlatMatrix<Complex> & elmat,Vector<Complex>& shapes, Vec<2> p,double det, double coef, double weight) const
    {
        Complex I = Complex(0,1);
        ARRAY<Complex> FourierShape(kSize),DFourierShape(kSize);
	for( int j=0; j < kSize; j++)
	{
	  FourierShape[j] = exp(I*k[j](0)*p(0));

	    // dv/dn = (i k v) . n
	    // outer normal n = (0,1) for upper boundary, n = (0,-1) for lower boundary
	  DFourierShape[j] = FourierShape[j]*double(normal)*k[j](1)*I;
	 }

	    // 1st line:  
	    // - \int c du_w/dn v_f
	    // second integrator 
	    // - 1/h \int u_w v_f
	    // 3rd line: 
	    // hermite
	 for( int j=0; j < kSize; j++)
	 {
	   int jcopy = j%ndof_k+ndof_f;

           Complex factor = Complex(1.0,0.0);

	   // int ( u_w dv_w / dn )
	   // 1/h int ( u_w v_w )
           if(j<(ndof_k))
           {
              factor *= S[j][0];
            
              // (up up) Term
	      elmat(jcopy,jcopy) += FourierShape[j]*factor*Conj(DFourierShape[j]*factor)*(coef*det*weight);
	      elmat(jcopy,jcopy) += alpha * FourierShape[j]*factor*Conj(FourierShape[j]*factor)*weight;
              // (down down) Term
	      elmat(jcopy,jcopy) += FourierShape[j+ndof_k]*Conj(DFourierShape[j+ndof_k])*(coef*det*weight);
	      elmat(jcopy,jcopy) += alpha * FourierShape[j+ndof_k]*Conj(FourierShape[j+ndof_k])*weight;
	      // (up down) Term
              elmat(jcopy,jcopy) += FourierShape[j+ndof_k]*Conj(DFourierShape[j]*factor)*(coef*det*weight);
	      elmat(jcopy,jcopy) += alpha * FourierShape[j+ndof_k]*Conj(FourierShape[j]*factor)*weight;
	      // (down up) Term
              elmat(jcopy,jcopy) += FourierShape[j]*factor*Conj(DFourierShape[j+ndof_k])*(coef*det*weight);
	      elmat(jcopy,jcopy) += alpha * FourierShape[j]*factor*Conj(FourierShape[j+ndof_k])*weight;

           }

	   for(int k=0; k < ndof_f; k++)
	   {
	      elmat(k,jcopy) -= DFourierShape[j]*factor*Conj(shapes[k])*(weight*coef*det);
   	      elmat(k,jcopy) -= alpha * FourierShape[j]*factor * (Conj(shapes[k])*weight);
              elmat(jcopy,k) = Conj(elmat(k,jcopy));
	   }

	 }
}


//Functions of IncomingWave************************************************************************************


void IncomingWave :: GetDofNrs (ARRAY<int> & dnums) const
{
  dnums.SetSize(1);
  dnums[0] = pos;
}


void IncomingWave :: Assemble (FlatMatrix<Complex> & elmat,LocalHeap & lh) const
{
  elmat.AssignMemory (1,1, lh);
  elmat(0,0) = Complex(penalty,0);
}









//Functions of NumprocBVPFourier ***************************************************** 





NumProcBVPFourier :: NumProcBVPFourier (PDE & apde, const Flags & flags) : NumProc (apde)
{
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    lambda = double(flags.GetNumFlag ("lambda", 1));
    alpha = double(flags.GetNumFlag ("alpha", 0));
    penalty = double(flags.GetNumFlag ("penalty", 1e10));
    pol = int(flags.GetNumFlag ("polarisation", 1));

    if (flags.GetDefineFlag ("layer")) layer = 1;
    else layer = 0;

    string ofile;
    ofile = flags.GetStringFlag("outfilereflect","./out/reflect.out");
    out1 = new ofstream (ofile.c_str());
    ofile = flags.GetStringFlag("outfiletransmit","./out/transmit.out");
    out2 = new ofstream (ofile.c_str());
    ofile = flags.GetStringFlag("infilelayer","./out/layer.out");
    layerfile = new ifstream (ofile.c_str());

    print = flags.GetDefineFlag ("print");
}


  void NumProcBVPFourier :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc BVPFourier:\n" \
      "------------\n" \
      "Solves the linear system resulting from a boundary value problem\n\n" \
      "Required flags:\n" 
      "-bilinearform=<bfname>\n" 
      "    bilinear-form providing the matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "\nOptional flags:\n"\
      "-print\n"
      "    write matrices and vectors into logfile\n"
	<< endl;
  }


void NumProcBVPFourier :: Do(LocalHeap & lh)
  {
    static int solvetimer = NgProfiler::CreateTimer ("Equation solving");
    NgProfiler::RegionTimer reg (solvetimer);

    const MeshAccess & ma = pde.GetMeshAccess();

    if (!lff->IsAssembled()) lff->Assemble(lh);

    cout << "solve bvpfourier" << endl;

    const FESpaceDefract& def = dynamic_cast<const FESpaceDefract&> (bfa->GetFESpace()); 

    const Periodic2DSpace* h1 = dynamic_cast<const Periodic2DSpace*> (def[0]);
    const FourierFESpace*  f1 = dynamic_cast<const FourierFESpace*> (def[1]);
    const FourierFESpace*  f2 = dynamic_cast<const FourierFESpace*> (def[2]);

    int nf1 = f1->GetNDof();
    int nf2 = f2->GetNDof();
    int nh1 = h1->GetNDof();
    int nse = ma.GetNSE();
    int dim = ma.GetDimension();
    ARRAY<INT<2> > pbcPairs; 
    ma.GetPeriodicVertices(0, pbcPairs);

   
    Vec<2> point1,point2;
    ma.GetPoint(pbcPairs[0][0],point1);
    ma.GetPoint(pbcPairs[0][1],point2);
    double widthx = fabs(point1(0)-point2(0));

    ARRAY<int> pnums,dnums;

    double n1 = pde.GetConstant("IndexAboveReal");
    double n2 = pde.GetConstant("IndexAboveImag");
    Complex n_above = Complex(n1,n2);
    n1 = pde.GetConstant("IndexBelowReal");
    n2 = pde.GetConstant("IndexBelowImag"); 
    Complex n_under = Complex(n1,n2);
	
    CoefficientFunction * coef = pde.GetCoefficientFunction("LaplaceBoundary");
    Complex n_layer;


    ARRAY<Vec<2,Complex> > k_above(nf1), k_under(2*nf2), k_layer(2*nf2);

    Complex k_above_abs = 2.0*M_PI/lambda*n_above;
    Complex k_under_abs = 2.0*M_PI/lambda*n_under;

    Complex gamma = (exp(Complex(0,widthx*sin(alpha))*k_above_abs)); 
    h1->SetGamma(gamma);

    for(int i = 0; i< (nf1-1); i++)
    {
	Complex c = (i-(nf1-2.0)/2.0)*2.0*M_PI/widthx;
	k_above[i](0) = k_above_abs*sin(alpha) + c;
        k_above[i](1) = sqrt(sqr(k_above_abs)-sqr(k_above[i](0)));
    }
    k_above[nf1-1](1) = -k_above[nf1/2-1](1);
    k_above[nf1-1](0) = k_above[nf1/2-1](0); 


    for(int i = 0; i< nf2; i++)
    {
	Complex c = (i-(nf2-1.0)/2.0)*2.0*M_PI/widthx;
	k_under[i](0) = k_layer[i](0) = k_above_abs*sin(alpha) + c;
        k_under[i](1) = sqrt(sqr(k_under_abs)-sqr(k_under[i](0)));
	k_under[i + nf2](1) = -k_under[i](1);
	k_under[i + nf2](0) = k_layer[i + nf2](0) = k_under[i](0);
    }

    ARRAY<Complex[2]> S(nf2);


    CalcSMatrix((nf2),S,k_layer,n_layer); 


    FourierElement* fe;
    IncomingWave* ic;

    ElementTransformation eltrans;

    for(int i=0; i<nse; i++)
    {
      if(ma.GetSElIndex(i)==f1->GetBoundary())
      {
        ma.GetSurfaceElementTransformation (i,eltrans,lh);
	def.GetSDofNrs(i,dnums);
        fe = new FourierElement (k_above,S,1,i,coef,dnums,eltrans,def);
	def.AppendSpecialElement(fe);

      }
      if(ma.GetSElIndex(i)==f2->GetBoundary())
      {
        ma.GetSurfaceElementTransformation (i,eltrans,lh);
	def.GetSDofNrs(i,dnums);
        fe = new FourierElement (k_under,S,(-1),i,coef,dnums,eltrans,def);
	def.AppendSpecialElement(fe);
      }
    }



    ic = new IncomingWave(nh1+nf1-1,penalty);
    def.AppendSpecialElement(ic);	  
    


    cout << "set nonassemble " << endl;
    bfa->SetNonAssemble(false);

    // reassemble matrix, with special elements included
     bfa->ReAssemble(lh, 1);




    const BaseMatrix & mat = bfa->GetMatrix();
    BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();

    FlatVector<Complex> fv = vecf.FVComplex();
 //   
    fv[nh1+nf1-1]=Complex(penalty,0);


    if (print)
      {
	(*testout) << "MatrixHeight = " << endl << mat.VHeight() << endl;
	(*testout) << "MatrixWidth = " << endl << mat.VWidth() << endl;
	(*testout) << "Matrix = " << endl << mat << endl;
	(*testout) << "RHS-Vector = " << endl << vecf << endl;
      }

    BaseMatrix * invmat; 

    cout << "direct solve for complex problem" << endl;
    invmat = dynamic_cast<const BaseSparseMatrix&> (mat) . InverseMatrix(); 

    ma.PushStatus ("Direct solver");

      
    clock_t starttime, endtime, soltime;
    starttime = clock();

    invmat->Mult (vecf, vecu);

/*
	// post iteration
	for (int k = 0; k < 10; k++)
	{
	  BaseVector & d = *vecu.CreateVector();
	  BaseVector & w = *vecu.CreateVector();
	  d = vecf - mat * vecu;
	  invmat2->Mult (d, w);
	  vecu += w;
	}
*/
    

    ma.PopStatus ();

    endtime = clock();
    cout << "Solution time = " << double(endtime - starttime)/CLOCKS_PER_SEC << endl;
    
    delete invmat;

    bfa -> ComputeInternal (vecu, lh);

    if (print)
      (*testout) << "Solution = " << endl << vecu << endl;

    if(!layer)
      Output(nh1,vecu,k_above,k_under,n_above, n_under);
    else
    {
      for( int i=0; i<nf2; i++)
        vecu.FVComplex()[i+nf1+nh1] *= S[i][1];

      if (print)
        (*testout) << "modified Solution = " << endl << vecu << endl;
      Output(nh1,vecu,k_above,k_under,n_above, n_layer);
    }

  }




void NumProcBVPFourier::Output(int h, BaseVector& v,ARRAY<Vec<2,Complex> >& ka,ARRAY<Vec<2,Complex> >& ku,Complex na, Complex nu)
{
  int f1 = ka.Size()-1;
  int f2 = ku.Size()/2;

  double energy=0;
  (*out1).precision(15);
  (*out2).precision(15);
  for(int i=0; i<f1; i++)
  {
    (*out1) <<endl<<"order= "<<(i-(f1-1)/2);
  //  (*out1) <<"   kx= "<<ka[i](0)<<"   ky= "<<ka[i](0);

    double angle;
    if(real(ka[i](1))==0)  angle = -1e-10;
    else  angle = atan(fabs(real(ka[i](0))/real(ka[i](1))));
    (*out1)<< "   angle="<<angle/M_PI*180;

    Complex amplitude = v.FVComplex()[h+i];
    (*out1)<<"   amplitude= "<<amplitude;

    Complex intensity = ka[i](1)/ka[(f1-1)/2](1)*amplitude*Conj(amplitude);
    (*out1)<<"   intensity= "<<fabs(real(intensity));
    energy += fabs(real(intensity));
  }
    cout<<endl<<endl <<"reflected energy: "<<energy; 
  for(int i=0; i<f2; i++)
  {  
    (*out2) <<endl<<"order= "<<(i-(f2-1)/2);
  //  (*out2) <<"   kx= "<<ku[i](0)<<"   ky= "<<ku[i](1);

    double angle;
    if(real(ku[i+f2](1))==0)  angle = -1e-10;
    else  angle = atan(fabs(real(ku[i+f2](0))/real(ku[i+f2](1))));
    (*out2) << "   angle="<<angle/M_PI*180;

    Complex amplitude = v.FVComplex()[h+f1+1+i];
    (*out2)<<"   amplitude= "<<amplitude;

    Complex intensity;
    intensity = ku[i+f2](1)/ka[(f1-1)/2](1)*amplitude*Conj(amplitude);
    if(pol==-1)
      intensity *= na*na/nu/nu;
    (*out2)<<"   intensity= "<<fabs(real(intensity));
    energy += fabs(real(intensity));
  }
  (*out1)<<endl;
  (*out2)<<endl;
  cout.precision(16);
  cout<<endl<<endl<<"Total Energy: " << energy<<endl<<endl;
}



void NumProcBVPFourier::CalcSMatrix(int dim, ARRAY<Complex[2]>& S, ARRAY<Vec<2,Complex> >& kl,Complex& n_layer)
{
  int nr_n, nr_thickness;
  ARRAY<double> thickness;
  ARRAY<Complex> n;
  double n_real,n_imag;

  if(layer)
  {
    (*layerfile) >> nr_n;
    n.SetSize(nr_n);
    for(int i=0; i<nr_n; i++)
    {
	*layerfile>>n_real;
        *layerfile>>n_imag;
	n[i] = Complex(n_real,n_imag);
    }
    *layerfile>>nr_thickness;
    thickness.SetSize(nr_thickness);
    for(int i=0; i<nr_thickness; i++)	*layerfile>>thickness[i];
    n_layer = n[nr_n-1];
  }



  for(int i=0; i<dim; i++)
  {
    S[i][1] = Complex(1,0);
    S[i][0] = Complex(0,0);
    if(layer)
    {
      ARRAY<Complex> Cos(nr_n),k(nr_n);
      for(int j=nr_n-1;j>=0; j--)
      {
        k[j] = n[j]*2.0*M_PI/lambda;
        Complex k_y;
        k_y=sqrt(k[j]*k[j]-kl[i](0)*kl[i](0));
        if(j == nr_n-1)
        {
          kl[i](1) = k_y;
          kl[i+dim](1) = -k_y;
        }
        Cos[j] = k_y/k[j];
      }

      Complex T[2][2];
      for(int j=nr_thickness-1; j>=0; j--)
      {

	//Calculation of actual T-Matrix
	//  (o+)  =   (T[0][0]  T[0][1])  (u+)  = ( T22 T21 ) (u+)
	//  (o-)  =   (T[1][0]  T[1][1])  (u-)  = ( T12 T11 ) (u-)
        if(pol == 1)
        {
           T[0][0]=T[1][1] = (n[j+1]*Cos[j+1]+n[j]*Cos[j])/(2.0*n[j]*Cos[j]);
           T[0][1]=T[1][0] = (-n[j+1]*Cos[j+1]+n[j]*Cos[j])/(2.0*n[j]*Cos[j]);
        }
        if(pol == -1)
        {
           T[0][0]=T[1][1] = (n[j+1]*Cos[j]+n[j]*Cos[j+1])/(2.0*n[j]*Cos[j]);
           T[0][1]=T[1][0] = (n[j+1]*Cos[j]-n[j]*Cos[j+1])/(2.0*n[j]*Cos[j]);
        }

	//Multiplying it onto S-Matrix
	//  (o+)  =   (S[0][0]  S[0][1])  (u+)  = ( S11 S12 ) (u+)
	//  (u-)  =   (S[1][0]  S[1][1])  (o-)  = ( S21 S22 ) (o-)

        // Z = (T11 + T12*S12')^(-1) 
	Complex Z =  Complex(1.0,0)/(T[1][1]+T[1][0]*S[i][0]);

        //S22 = S22' * Z
        S[i][1] *= Z;

        //S12 = (T21+T22*S12')*Z
        S[i][0] = (T[0][1]+T[0][0]*S[i][0])*Z;

        //Multiplying Propagationmatrix onto S-Matrix;

        Complex P22 = exp(Complex(0,1)*k[j]*thickness[j]*Cos[j]);
        Complex P11 = exp(-Complex(0,1)*k[j]*thickness[j]*Cos[j]);

        S[i][1] = S[i][1]/P11;
        S[i][0] = P22*S[i][0]/P11;

      }
    }
  }
}

    namespace
      fespace_cpp
  {
     
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("fourier", FourierFESpace::Create);
      GetFESpaceClasses().AddFESpace ("hofe", Periodic2DSpace::Create);
      GetFESpaceClasses().AddFESpace ("defract", FESpaceDefract::Create);
      GetNumProcs().AddNumProc ("bvpfourier", NumProcBVPFourier::Create, NumProcBVPFourier::PrintDoc);
     }
    
    Init init;
  }
}













