
#include "../include/solve.hpp"
#include <comp.hpp>
#include <fem.hpp>



namespace ngsolve
{
  using namespace ngsolve;


  inline Complex Conj (Complex c)
  {
    return Complex (c.real(), -c.imag());
  }



class Periodic2D : public H1HighOrderFESpace
{
  private:
    ARRAY<int> dofmapx;
    ARRAY<int> dofmapy;
    mutable Complex gammax ;
    mutable Complex gammay ;
    int idx;
    int idy;

  public:

    Periodic2D (const MeshAccess & ama, const Flags & flags);
  ///
    virtual ~Periodic2D ();

  void SetGammax(Complex agammax)const
  {
    gammax=agammax;
  }

  void SetGammay(Complex agammay)const
  {
    gammay=agammay;
  }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new Periodic2D (ma, flags);
  }

  virtual string GetClassName () const
  {
    return "Periodic2D";
  }

  double GetPeriodicWidthX();

  double GetPeriodicWidthY();

  virtual void Update ();
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;

  virtual void VTransformMC (int elnr, bool boundary,
			     FlatMatrix<Complex> & mat, TRANSFORM_TYPE type) const;
  virtual void VTransformVC (int elnr, bool boundary,
			     FlatVector<Complex> & vec, TRANSFORM_TYPE type) const;



  int DofMapx (int i) const { return dofmapx[i]; }
  int DofMapy (int i) const { return dofmapy[i]; }
  Complex GetGammax () const { return gammax; }
  Complex GetGammay () const { return gammay; }
};

double Periodic2D::GetPeriodicWidthX()
{
    ARRAY<INT<2> > pbcPairs(ma.GetNPairsPeriodicVertices(idx)); 
    ma.GetPeriodicVertices(idx, pbcPairs);

   
    Vec<2> point1,point2;
    ma.GetPoint(pbcPairs[0][0],point1);
    ma.GetPoint(pbcPairs[0][1],point2);
    return  fabs(point1(0)-point2(0));
}

double Periodic2D::GetPeriodicWidthY()
{
    ARRAY<INT<2> > pbcPairs(ma.GetNPairsPeriodicVertices(idy)); 
    ma.GetPeriodicVertices(idy, pbcPairs);

   
    Vec<2> point1,point2;
    ma.GetPoint(pbcPairs[0][0],point1);
    ma.GetPoint(pbcPairs[0][1],point2);
    return  fabs(point1(0)-point2(0));
}


Periodic2D :: Periodic2D (const MeshAccess & ama, const Flags & flags)
    : H1HighOrderFESpace (ama, flags)
{ 
   idx = int(flags.GetNumFlag ("idx", 2)); 
   idy = int(flags.GetNumFlag ("idy", 4));
}


  Periodic2D :: ~Periodic2D ()
  {
    ;
  }


  void Periodic2D :: Update ()
{
   H1HighOrderFESpace::Update();

   dofmapx.SetSize (GetNDof());
   for (int i = 0; i < dofmapx.Size(); i++)
      dofmapx[i] = i;

   dofmapy.SetSize (GetNDof());
   for (int i = 0; i < dofmapy.Size(); i++)
      dofmapy[i] = i;

    ARRAY<INT<2> > pbcPairs;
    ARRAY<INT<2> > perEdges; 

     // periodic Vertices 
    ma.GetPeriodicVertices(idx, pbcPairs);

    for (int i = 0; i < pbcPairs.Size(); i++)
      dofmapx[pbcPairs[i][1]] = pbcPairs[i][0];
    
    // periodic Edges 

    ma.GetPeriodicEdges(idx,perEdges); 
    int npe = perEdges.Size();
	
    for(int i=0; i<npe; i++)
    {
      int dof_e00 = GetFirstEdgeDof(perEdges[i][0]); 
      int dof_e01 = GetFirstEdgeDof(perEdges[i][0]+1); 
      int dof_e10 = GetFirstEdgeDof(perEdges[i][1]); 
      int dof_e11 = GetFirstEdgeDof(perEdges[i][1]+1); 

      int diff = dof_e01-dof_e00;

      if(diff != dof_e11-dof_e10)
         throw Exception ("Dofs of periodic edges don't match");

      for(int j=0;j<diff;j++)
	 dofmapx[dof_e10+j] = dof_e00+j;
      }

    ma.GetPeriodicVertices(idy, pbcPairs);

    for (int i = 0; i < pbcPairs.Size(); i++)
      dofmapy[pbcPairs[i][1]] = pbcPairs[i][0];
    
    // periodic Edges 

    ma.GetPeriodicEdges(idy,perEdges); 
    npe = perEdges.Size();	
    for(int i=0; i<npe; i++)
    {
      int dof_e00 = GetFirstEdgeDof(perEdges[i][0]);
      int dof_e01 = GetFirstEdgeDof(perEdges[i][0]+1); 
      int dof_e10 = GetFirstEdgeDof(perEdges[i][1]); 
      int dof_e11 = GetFirstEdgeDof(perEdges[i][1]+1);

      int diff = dof_e01-dof_e00;

      if(diff != dof_e11-dof_e10)
         throw Exception ("Dofs of periodic edges don't match");

      for(int j=0;j<diff;j++)
	 dofmapy[dof_e10+j] = dof_e00+j;
    }
}


  void Periodic2D::GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    H1HighOrderFESpace::GetDofNrs(elnr,dnums);
    for(int i=0; i<dnums.Size(); i++)
      dnums[i] = dofmapy[dofmapx[dnums[i]]];
  }

  void Periodic2D::GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    H1HighOrderFESpace::GetSDofNrs(selnr,dnums);
    for(int i=0; i<dnums.Size(); i++)
      dnums[i] = dofmapy[dofmapx[dnums[i]]];;
  }

 void Periodic2D::VTransformMC (int elnr, bool boundary,
			FlatMatrix<Complex> & mat, TRANSFORM_TYPE type) const
{
  if (!boundary)
    {
      ARRAY<int> dnums;
      H1HighOrderFESpace::GetDofNrs (elnr, dnums);
	for (int i = 0; i < dnums.Size(); i++)
	  {
	    if (dofmapx[dnums[i]] != dnums[i])   // dof on right boundary
	      {
		if (type & TRANSFORM_MAT_LEFT)
		  for (int j = 0; j < mat.Width(); j++)
		    mat(i,j) *= Conj(gammax);
		if (type & TRANSFORM_MAT_RIGHT)
		  for (int j = 0; j < mat.Height(); j++)
		    mat(j,i) *= gammax;
	      }
	    if (dofmapy[dnums[i]] != dnums[i])   // dof on right boundary
	      {
		if (type & TRANSFORM_MAT_LEFT)
		  for (int j = 0; j < mat.Width(); j++)
		    mat(i,j) *= Conj(gammay);
		if (type & TRANSFORM_MAT_RIGHT)
		  for (int j = 0; j < mat.Height(); j++)
		    mat(j,i) *= gammay;
	      }
	  }
   }
}

 void Periodic2D::VTransformVC (int elnr, bool boundary,
			     FlatVector<Complex> & vec, TRANSFORM_TYPE type) const
{  
  if (!boundary)
    {
      ARRAY<int> dnums;
      H1HighOrderFESpace::GetDofNrs (elnr, dnums);

      if (type & TRANSFORM_SOL)
      for (int i = 0; i < dnums.Size(); i++)
	{
	  if (dofmapx[dnums[i]] != dnums[i])   // dof on right boundary
	    vec(i) *= gammax;
	  if (dofmapy[dnums[i]] != dnums[i])   // dof on right boundary
	    vec(i) *= gammay;
	}
    }
}




  class NumProcEVPPeriodic : public NumProc
  {
  protected:

    BilinearForm * bfa;
    BilinearForm * bfm;
    GridFunction * gfu;

    Preconditioner * pre;
    /// 
    string filename;

    int stepx;

    int maxsteps;
    /// 
    int nev; 
    ///
    double prec;
    ///
    bool print;

  public:
    NumProcEVPPeriodic (PDE & apde, const Flags & flags);

    virtual ~NumProcEVPPeriodic();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcEVPPeriodic (pde, flags);
    }

    virtual void Do(LocalHeap & lh);

    virtual string GetClassName () const
    {
      return "Periodic EVP Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Bilinear-form = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcEVPPeriodic :: NumProcEVPPeriodic (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));

    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    
    maxsteps = int(flags.GetNumFlag ("maxsteps", 50));
    stepx = int(flags.GetNumFlag ("stepx", 1));
    
    nev =  gfu->GetMultiDim(); 
    filename = flags.GetStringFlag ("filename","lobpcg.out");
    prec = double(flags.GetNumFlag ("prec", 1.e-5));

  }

  NumProcEVPPeriodic :: ~NumProcEVPPeriodic()
  {
    ;
  }

  void NumProcEVPPeriodic :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc BVPFourierEM:\n" \
      "------------\n" \
      "Solves the linear system resulting from a boundary value problem\n\n" \
      "Required flags:\n" 
      "-bilinearform=<bfname>\n" 
      "    bilinear-form providing the matrix\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "\nOptional flags:\n"\
      "-preconditioner=<prename>\n"
      "-maxsteps=n\n"
      "-prec=eps\n"
      "-print\n"
      "    write matrices and vectors into logfile\n"
	<< endl;
  }


  void NumProcEVPPeriodic :: Do(LocalHeap & lh)
  {
    static int solvetimer = NgProfiler::CreateTimer ("Equation solving");
    NgProfiler::RegionTimer reg (solvetimer);

    const MeshAccess & ma = pde.GetMeshAccess();

    cout << "solve evp periodic" << endl;

    Periodic2D& per = dynamic_cast<Periodic2D&> (const_cast<FESpace&> (bfa->GetFESpace())); 

	cout << " periodic space ok " << endl; 

    int nper = per.GetNDof();
    int nse = ma.GetNSE();

    double px = per.GetPeriodicWidthX();
    double py = per.GetPeriodicWidthY();

    if (3*nev >= nper)
       throw Exception("Multidim too big vs ndofs");
    int N = 3*nev; // but in the first step use std. preconditioned gradient type method  	
cout<<endl<<"test " << endl; 
    ARRAY<BaseVector*> w(N);	
    // Create vector w with initial vector X and precond. Residuum and p=0;  
    // w = [X,C^[-1](AX- Lami MX),P]
    for(int m=0; m<N;m++)
      {
	w[m]=gfu->GetVector(0).CreateVector();
      }	
cout<<endl<<"test2 " << endl; 

    
    BaseVector & vw = *(w[0]->CreateVector());
    BaseVector & vd = *(w[0]->CreateVector());
    BaseVector & vv = *(w[0]->CreateVector());
    BaseVector & vmu = *(w[0]->CreateVector());
    BaseVector & vau = *(w[0]->CreateVector());
    
    cout.precision(12);
    (*testout).precision(12); 
    

    Vector<double> Lami(nev); 
    Vector<double> LamiOld(nev); 
     
    // lopcg=2: LOBPCG1  x_n+1 = alpha x_n + tau w_n + gamma p_n mit p_{n+1}^j = x_n+1^j - sum(tau^k x_n^k) 

    ARRAY<BaseVector*> wh(nev);
     
 cout << "vor Init " << endl;

  
   // InitEigenVector
    for(int m=0;m<nev; m++) 
      {
	wh[m] = w[0]->CreateVector();
	double norm; 
	

	FlatVector<Complex> fv = w[m]->FVComplex();
	FlatVector<Complex> fv2 = wh[m]->FVComplex();
	    
	time_t start=clock();
	int sstart = start; 
	srand(m*m*m*m*m*start); 
	for (int i = 0; i < fv.Size(); i++)
	   {
		
		double r1 = double(rand())/double(RAND_MAX);
		double r2 = double(rand())/double(RAND_MAX);
		fv(i) = Complex(r1,r2); 
            }
	    
	  
	    norm = 1/sqrt(real(InnerProduct(fv,Conj(fv))));
	    fv *= norm; 
      }

     cout << "nach Init " << endl;
    
    for( double ax = M_PI/2. ; ax <= M_PI; ax += M_PI/stepx)
    {
      per.SetGammax(exp(Complex(0,1)*ax));
      per.SetGammay(1.0);


      cout << "set nonassemble " << endl;
      bfa->SetNonAssemble(false);

      bfa->ReAssemble(lh, 1);

      //pre->Assemble(lh, 1); 	


    cout << "solve EVP via LOBPCG" << endl;
    *testout << "solve EVP via LOBPCG" << endl;
    
    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseMatrix & matm = bfm->GetMatrix();
    BaseVector & vu = gfu->GetVector();
    //const BaseMatrix & matpre = pre->GetMatrix();
    BaseMatrix * invmat; 

    cout << "direct solve for complex problem" << endl;
    invmat = dynamic_cast<const BaseSparseMatrix&> (mata) . InverseMatrix(); 

    if(print)
      (*testout)<<endl<<"invmat = "<<endl<< *invmat <<endl;

    
    
    
   



    // P^0 = 0  // can be skipped 
    for (int m=2*nev;m<N;m++) 
      *w[m]=0.;  
        
    // since first iterate is std. prec. gradient method 
    N = 2*nev; 
    // Improvement to implement (vg Knyazev Matlab Code) 
    // vmu[nev,nev,nev] und vau[nev,nev,nev] speichern 
    // nach preconditioning nur w-Teil Aw und Mw neu berechnen 
    // in X^T AX keine neue Anwendung von AX,MX 
    // ARRAY<BaseVector*> AV(N), MV(N); 
    
	cout << " complex " << endl;
	for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
	  {
	    cout << endl << "iteration i = " << ii << endl;
	    for(int m=0;m<nev;m++)
	      {
		//		*MX[m] = matm * (*w[m]); 	  		
		// *AX[m] = mata * (*w[m]); 

		vmu = matm * (*w[m]); 	  		
		vau = mata * (*w[m]); 
		cout << " vmu vau ok " << endl; 
		cout << " for vector " << m << endl; 
		//FlatVector<Complex> fvmu = MX[m]->FVComplex(); 
		//FlatVector<Complex> fvau = AX[m]->FVComplex(); 
		FlatVector<Complex> fvmu = vmu.FVComplex();
		FlatVector<Complex> fvau = vau.FVComplex(); 
		FlatVector<Complex> v = w[m]->FVComplex(); 
 		
	cout << " test2 " << endl; 

		if(ii == 1) 
		  LamiOld(m) = real(InnerProduct(fvau,Conj(v))/real(InnerProduct(fvmu,Conj(v)))); 
		
	cout << " calculated first lami " << endl; 
		
		// vd = (*AX[m])-LamiOld(m)*(*MX[m]); 
		vd = (vau)-LamiOld(m)*(vmu); 
		vmu = (*invmat)*vd; 
		
		double norm2 = 1/sqrt(real(InnerProduct(fvmu,Conj(fvmu)))); 	
		fvmu *= norm2; 
		*w[nev+m] = vmu;
		//*w[nev+m] = (*invmat)*vd; 
	      }
	    
	    Vector<double> Lami2(N);     
	    Matrix<Complex> asmall(2*N,N); 
	    Matrix<Complex> msmall(2*N,N); 
	  
	    
    
	    // [X,W,P]^T A [X,W,P] y = lami [X,W,P]^t M [X,W,P] y 
	    /*	    
	       for(int j=0;j<nev;j++)
	      {
		FlatVector<Complex> fvmu = MX[j]->FVComplex(); 
		FlatVector<Complex> fvau = AX[j]->FVComplex();
		
		
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    msmall(j,k) = InnerProduct(fvmu,Conj(fv));  	  		
		  } 
		
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    asmall(j,k) = InnerProduct(fvau,Conj(fv));  	  		
		  } 
		  } 
	    */ 
		cout << " Vor Create small Ritz-system " << endl; 
	    for(int j=0;j<N;j++) 
	      {
		vmu = matm * (*w[j]);
		cout << " t1 " << endl;  
		FlatVector<Complex> fvmu = vmu.FVComplex(); 
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
			cout << " IP " << InnerProduct(fvmu,Conj(fv)) << endl; 	
		    msmall(j,k) = InnerProduct(fvmu,Conj(fv));  	  		
		  } 
		
		vau = mata * (*w[j]); 
		FlatVector<Complex> fvau = vau.FVComplex(); 
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    asmall(j,k) = InnerProduct(fvau,Conj(fv));  	  		
		  } 
		cout << " j " << j << " N " << N << endl; 
	      } 
	    
	  cout << " Lapack  .... " << endl; 
	    int info = LapackGHEPEPairs(N, &asmall(0,0), &msmall(0,0), &Lami2(0)); 
	    cout << " .... is back " << endl; 
	  
	    
	    if(info>0)
	      {
		cout << " returned info " << info << endl; 
		throw Exception ("Error in zhegv_ LapackGHEPPairs() ");
	      }	
	    
	for(int m=0;m<nev;m++) 
	  Lami(m) = Lami2(m); 
	
	// use only the first nev eigenvectors and write to gfu
	for(int m=0;m<nev;m++)
	  {
	    *wh[m]=0.; 
	    for(int k=0;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }
	
	// new p^{j+1}_k = [0,w_j, p_j] v[k]  
	for(int m=0;m<nev;m++)
	  {
	    *w[m] = *wh[m]; 
	    *wh[m] = 0.;  
	    for(int k=nev;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }

	for(int m=0;m<nev;m++)
	  *w[m+2*nev] = *wh[m]; 
	
	// after first iteration use LOBPCG  
	N = 3*nev;
	double err=0.;
	for(int l=0;l<nev/2;l++)
	  err += (LamiOld(l)-Lami(l))*(LamiOld(l)-Lami(l));
	
	*testout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  *testout << " Lami(" << m << ") = " << Lami(m) << endl; 

	cout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  cout << " Lami(" << m << ") = " << Lami(m) << endl; 
		
	cout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	*testout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	if(sqrt(err)/double(nev/2)<prec) break; 

	LamiOld = Lami;

       
      

      } 
    
      for(int m=0;m<nev;m++)
        {
	  BaseVector & vu = gfu->GetVector(m);
	  vu = *w[m];
        }


}





  }







  namespace periodic
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("periodic", Periodic2D::Create);
      GetNumProcs().AddNumProc ("evpperiodic", NumProcEVPPeriodic::Create, NumProcEVPPeriodic::PrintDoc);
    }
    
    Init init;
  }

}




















