
/*
 * Braess-Schberl error estimator - Laplace version
 * =================================================
 * 
*/

// #define DEBUG
#include <solve.hpp>
#include <comp.hpp> 

#include "hdivhofespacesz.hpp"



using namespace ngfem;
using namespace ngsolve;
using namespace ngcomp; 

namespace{
  inline void rotate(const Vec<2>& x, Vec<2>& y){ y(0) = -x(1); y(1) = x(0); };
  inline double spr(const Vec<2>& x, const Vec<2>& y) { return x(0)*y(0)+x(1)*y(1); };
}

/* ***************** Braess-Schberl error estimator ************************ */
 
class NumProcBrSchEE : public NumProc
{
protected:
  
  BilinearForm * bfa;
  LinearForm * lff;
  S_GridFunction<double> * gfu;
  S_GridFunction<double> * gferr;
  
  CoefficientFunction * coef_f;


  string filename,filename2;
  ofstream fileout,fileout2;
  double refenergy;
public:

  NumProcBrSchEE (PDE & apde, const Flags & flags) : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction (flags.GetStringFlag ("solution", "")));
    gferr =  dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction (flags.GetStringFlag ("error", "")));

    coef_f = pde.GetCoefficientFunction (flags.GetStringFlag ("coeff", ""));

    filename = flags.GetStringFlag ("filename", "");
    filename2 = flags.GetStringFlag ("filename2", "");
    refenergy = flags.GetNumFlag ("energy", 0);
    if (filename.length())
      {
        fileout.open (filename.c_str());
        fileout.precision(16);
      }

    if (filename2.length())
      {
        fileout2.open (filename2.c_str());
        fileout2.precision(16);
      }
  }


  virtual ~NumProcBrSchEE()
  { ; }

  
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcBrSchEE (pde, flags);
  }


  virtual void Do(LocalHeap & lh)
  {
    cout << "**************************************************************" << endl;
    cout << "called Braess Schoeberl error estimator" << endl;
         

    static int timer = NgProfiler::CreateTimer ("Equilibration");
    static int timer1 = NgProfiler::CreateTimer ("Equilibration 1");
    static int timer2 = NgProfiler::CreateTimer ("Equilibration 2");
    static int timer3 = NgProfiler::CreateTimer ("Equilibration 3");
    static int timer4 = NgProfiler::CreateTimer ("Equilibration 4");
    static int timer5 = NgProfiler::CreateTimer ("Equilibration 5");

    testout->precision(12); 
    NgProfiler::StartTimer (timer);
    NgProfiler::StartTimer (timer1);

    time_t start=clock();

    const FESpace & fes = gfu->GetFESpace();
    const BilinearFormIntegrator & bli = *bfa->GetIntegrator(0);

    // int p = fes.GetOrder()-1;

    int psig = fes.GetOrder()-1;

    const H1HighOrderFESpace * h1fes =
      dynamic_cast<const H1HighOrderFESpace*> (&gfu->GetFESpace());
    
    psig = h1fes->GetOrder()-1; 
    cout << " ******* H1 GET ORDER  : " << fes.GetOrder() << endl; 
    cout << " ******* H1HO GET ORDER  : " << h1fes->GetOrder() << endl; 
    int pf = psig-1;
    int pe = psig;
         
    if (psig == 0)
      {  // RT 0
        pf = 0;
        pe = 0;
      }

    const int ned = ma.GetNEdges();
    const int ne = ma.GetNE();
    const int np = ma.GetNP();

    int ndt = (pf+2)*(pf+1)/2;
    int nde = pe+1;
    int ndsig = max(3, (psig+2)*(psig+1));
         
    cout << "general information: ne=" << ne << ", ned=" << ned << ", np=" << np << ", orderf = " << pf << endl;
                 
    ARRAY<int> vnums, ednums, dnums, pnums, elnums;
    Vector<> sigma_glob(ne * ndsig);
    sigma_glob = 0.0;

    BitArray dirichletedges (ned);
    dirichletedges.Clear();
    
    for (int i = 0; i < ma.GetNSE(); i++)
      if (ma.GetSElIndex (i) == 0)
        {
          ma.GetSElEdges(i, ednums);
          for (int j = 0; j < ednums.Size(); j++)
            dirichletedges.Set(ednums[j]);
        }
    *testout << "dirichletedges: " << endl << dirichletedges << endl;

    cout << "************ NEWFES order fessig hdiv " << psig << endl;  
    Flags flags;
    // flags.SetFlag("discontinuous"); 
    flags.SetFlag ("order", psig);
    HDivHighOrderFESpaceSZ fessig (ma, flags);
    fessig.Update();

    ARRAY<int> elsonvertex, edgesonvertex;
    ElementTransformation eltrans;
    double errt = 0, erre = 0, err = 0;

    void * hp = lh.GetPointer();
             
    for (int i = 0; i < np; i++)
      {
        ma.SetThreadPercentage ( 100.0 * i / np );

        NgProfiler::StartTimer (timer2);

        (*testout) << "vertex " << i << ":" << endl;

        ma.GetVertexElements( i, elsonvertex);

        // find edges on vertex
        edgesonvertex.SetSize (0);
        for (int j = 0; j < elsonvertex.Size(); j++)
          {
            ma.GetElEdges (elsonvertex[j], ednums);
            for (int k = 0; k < ednums.Size(); k++)
              {
                int pn1, pn2;
                ma.GetEdgePNums (ednums[k], pn1, pn2);
                if ((pn1 == i || pn2 == i) && 
                    !dirichletedges.Test(ednums[k]) &&
                    !edgesonvertex.Contains (ednums[k]))
                  edgesonvertex.Append (ednums[k]);
              }
          }

        *testout << "has elements " << elsonvertex << endl;
        *testout << "has edges " << edgesonvertex << endl;
        
        int nloc = ndt * elsonvertex.Size() + nde * edgesonvertex.Size();
        int nsig = ndsig * elsonvertex.Size();

        FlatVector<> res(nloc, lh);
        res = 0;

        FlatMatrix<> a(nsig, lh), aloc(ndsig, lh);
        FlatMatrix<> b(nloc, nsig, lh), bloct(ndt, ndsig, lh), bloce(nde, ndsig, lh);
        FlatMatrix<> ainv(nsig, lh), aib(nsig, nloc, lh), schur(nloc, lh), invschur(nloc, lh);
        FlatMatrix<> ai(ndsig, lh), aiinv(ndsig, lh);
        FlatMatrix<> bi(nloc, ndsig, lh), aibi(ndsig, nloc, lh);

        a = 0.0;
        b = 0.0;

        for (int j = 0; j < elsonvertex.Size(); j++)
          {
            ma.GetElementTransformation (elsonvertex[j], eltrans, lh);

            const NodalFiniteElement & fel = 
              dynamic_cast<const NodalFiniteElement&> (fes.GetFE(elsonvertex[j], lh));
            fes.GetDofNrs (elsonvertex[j], dnums);

            const HDivHighOrderFiniteElementSZ <2> & felsig =
              dynamic_cast<const HDivHighOrderFiniteElementSZ<2> &> (fessig.GetFE(elsonvertex[j], lh));
              
            aloc = 0.0;  
            bloct = 0.0;

            L2HighOrderTrig trig(pf);

            FlatVector<> elu (dnums.Size(), lh);
            gfu -> GetElementVector (dnums, elu);

            FlatMatrix<> ddshape(dnums.Size(), 4, lh);
            Vec<4> dduv;
            const IntegrationRule & ir = SelectIntegrationRule (ET_TRIG, 2*(2*pf+2));
            
            FlatMatrix<> shapesigref(ndsig, 2, lh);
            FlatMatrix<> shapesig(ndsig, 2, lh);
            FlatVector<> divsigref(ndsig, lh);
            FlatVector<> divsig(ndsig, lh);

            int vi = 0;
            ma.GetElVertices (elsonvertex[j], vnums);
            for (int k = 0; k < vnums.Size(); k++)
              if (vnums[k] == i) vi = k;

            FlatVector<> ri(ndt, lh), shape(ndt, lh);
            ri = 0;

            for (int k = 0; k < ir.GetNIP(); k++)
              {
                SpecificIntegrationPoint<2,2> sip(ir[k], eltrans, lh);

                fel.CalcDDShape (ir[k], ddshape);
                dduv = Trans (ddshape) * elu;
                
                Mat<2> ddu, transddu;
                ddu(0,0) = dduv(0);
                ddu(0,1) = dduv(1);
                ddu(1,0) = dduv(2);
                ddu(1,1) = dduv(3);

                transddu = Trans (sip.GetJacobianInverse ()) * ddu * sip.GetJacobianInverse();
                double lapu = transddu(0,0) + transddu(1,1);

                trig.CalcShape (ir[k], shape);
                double fi = coef_f -> Evaluate (sip);

                double jac = fabs (sip.GetJacobiDet());
                double lami[3] = { ir[k](0), ir[k](1), 1-ir[k](0)-ir[k](1) };

                ri += (jac * (fi+lapu) * ir[k].Weight() * lami[vi]) * shape;

                felsig.CalcShape (ir[k], shapesigref);
                shapesig = (1.0/sip.GetJacobiDet()) * shapesigref * Trans (sip.GetJacobian());
                felsig.CalcDivShape (ir[k], divsigref);
		
	

                divsig = (1.0/sip.GetJacobiDet()) * divsigref;
                
		*testout << "ir[" << k <<"]: " << ir[k] << endl;
		*testout << " divsigref " << divsigref << endl; 
		*testout << " shapesigref " << shapesigref << endl; 
		*testout << "shape_qt " << shape << endl; 
		*testout << "jac " << jac << endl; 
		*testout << " Trans(divsig) " << Trans(divsig) << endl; 
		

                bloct += (jac * ir[k].Weight()) * shape * Trans (divsig);
		*testout << " new bt_elmat " << bloct << endl; 
                aloc += (jac * ir[k].Weight()) * shapesig * Trans (shapesig);
              }



            res.Range (j*ndt, (j+1)*ndt) += ri;

            for (int k = 0; k < ndsig; k++)
              for (int l = 0; l < ndt; l++)
                b(j*ndt+l, j*ndsig+k) = bloct(l, k);
            for (int k = 0; k < ndsig; k++)
              for (int l = 0; l < ndsig; l++)
                a(j*ndsig+l, j*ndsig+k) = aloc(l, k);


	    *testout << " elsonvertex[j] " << elsonvertex[j] << endl; 
	    *testout << " aelmat "  << aloc ; 
	    *testout << " btelmat " << bloct ; 
	    

          }

        // zero boundary conditions
        for (int j = 0; j < elsonvertex.Size(); j++)
          {
            ma.GetElEdges (elsonvertex[j], ednums);
            for (int k = 0; k < ednums.Size(); k++)
              {
                int pn1, pn2;
                ma.GetEdgePNums (ednums[k], pn1, pn2);
                if (pn1 != i && pn2 != i)
                  {
		    *testout << "Local Patch Dirich-edge:  ednums[k] " << ednums[k] << endl; 
		    *testout << " position " <<  j*ndsig+k << endl; 
                    a(j*ndsig+k, j*ndsig+k) += 1e10;
                    int base = j*ndsig + 3 + k*(nde-1);
                    for (int l = 0; l < nde-1; l++)
		      {
			a(base+l, base+l) += 1e10;
			*testout << " and to position " << base+l << endl; 
		      } 
                  }

                /*
                int hi = j*ndsig+3+(k+1)*(nde-1)-1;
                a(hi, hi) += 1e10;
                */
              }
          }

        for (int j = 0; j < edgesonvertex.Size(); j++)
          {
            *testout << "edge " << j << endl;
            L2HighOrderSegm segm(pe);

            int pi1, pi2;
            ma.GetEdgePNums (edgesonvertex[j], pi1, pi2);
            *testout << "edge points = " << pi1 << ", " << pi2 << endl;
            int i2 = pi1+pi2-i;

            Vec<2> pt1, pt2;
            ma.GetPoint (i, pt1);
            ma.GetPoint (i2, pt2);

            const IntegrationRule & ir = SelectIntegrationRule (ET_SEGM, 2*(2*pe+2));

            for (int k = 0; k < elsonvertex.Size(); k++)
              {
                const HDivFiniteElement<2> & felsig =
                  dynamic_cast<const HDivFiniteElement<2> &> (fessig.GetFE(elsonvertex[k], lh));

                ma.GetElVertices (elsonvertex[k], vnums);
                int ind1 = -1, ind2 = -1;
                for (int l = 0; l < vnums.Size(); l++)
                  {
                    if (vnums[l] == i) ind1 = l;
                    if (vnums[l] == i2) ind2 = l;
                  }
                double sign = (ind2 == (ind1+1)%3) ? 1 : -1;


                if (ind1 >= 0 && ind2 >= 0)
                  {
                    const POINT3D * pts = ElementTopology::GetVertices (ET_TRIG);
                    Vec<2> p1, p2, nv;
                    p1(0) = pts[ind1][0];
                    p1(1) = pts[ind1][1];
                    p2(0) = pts[ind2][0];
                    p2(1) = pts[ind2][1];

                    nv(0) = pt1(1)-pt2(1);
                    nv(1) = pt2(0)-pt1(0);
                    double h = L2Norm (nv);
                    nv /= h;
                    

                    ma.GetElementTransformation (elsonvertex[k], eltrans, lh);
                    const FiniteElement & fel = fes.GetFE(elsonvertex[k], lh);
                    fes.GetDofNrs (elsonvertex[k], dnums);

                    FlatVector<> shape(nde, lh);
                    FlatVector<> elu (dnums.Size(), lh), ri(nde, lh);
                    FlatVector<> flux (2, lh);
                    gfu -> GetElementVector (dnums, elu);

                    FlatMatrix<> shapesigref(ndsig, 2, lh);
                    FlatMatrix<> shapesig(ndsig, 2, lh);
                    FlatVector<> shapesign(ndsig, lh);
                    
                    ri = 0;
                    bloce = 0;

                    for (int l = 0; l < ir.GetNIP(); l++)                      
                      {

                        Vec<2> p2d = p1 + ir[l](0) * (p2-p1);
                        IntegrationPoint ip2d (p2d(0), p2d(1));
                        SpecificIntegrationPoint<2,2> sip(ip2d, eltrans, lh);

                        bli . CalcFlux (fel, eltrans, ip2d, elu, flux, true, lh);
                        double dudn = InnerProduct (flux, nv);
                        // *testout << "dudn = " << dudn << endl;

                        IntegrationPoint ip1d (ir[l](0));
                        segm.CalcShape(ip1d, shape);

                        ri += (dudn*h*ir[l].Weight() * (1-ir[l](0))) * shape;
                        
                        felsig.CalcShape (ip2d, shapesigref);
                        shapesig = (1.0/sip.GetJacobiDet()) * shapesigref * Trans (sip.GetJacobian());
                        shapesign = shapesig * nv;

                        bloce += (h*ir[l].Weight()) * shape * Trans (shapesign);
                      }

                    res.Range (ndt*elsonvertex.Size()+nde*j,
                               ndt*elsonvertex.Size()+nde*(j+1)) += sign * ri;


		    *testout << " edge  " << j << ",  elsonvertex[k] " << elsonvertex[k] << endl; 
		    *testout << " sign*ri " << sign*ri ; 
		    *testout << " sign*bloce " << sign*bloce; 

                    for (int m = 0; m < ndsig; m++)
                      for (int l = 0; l < nde; l++)
                        b(ndt*elsonvertex.Size()+j*nde+l, k*ndsig+m) += sign * bloce(l, m);
                  }
              }
          }

        NgProfiler::StopTimer (timer2);
        NgProfiler::StartTimer (timer3);

        
        double errpatch = 0;
        
	bool printres = 1; 
	if(printres)	
	  {
	    *testout << " for vertex " << i << " patch matrices " << endl;  
	    *testout << "has elements " << elsonvertex << endl;
	    *testout << "has facets " << edgesonvertex << endl; 
	    *testout << "b_all  sz = " << endl << b << endl;
	    *testout << "a_all  sz = " << endl << a << endl;
	    *testout << "res_all sz =" << endl << res << endl;  
	  }
	        
        // cout << 'c' << flush;

        double errti = L2Norm2 (res.Range (0, ndt*elsonvertex.Size()));
        double errei = L2Norm2 (res.Range (ndt*elsonvertex.Size(),res.Size()));
        errt += errti;
        erre += errei;

        /*
        *testout << "errti = " << errti << ", errei = " << errei << endl;
        *testout << "res = " << endl << res << endl;
        double sum = 0;
        for (int j = 0; j < res.Size(); j++)
          sum += res(j); 
        *testout << "sum = " << sum << endl;
        */

	fileout2 << " a " << endl << a  << endl;  
	fileout2 << " b " << endl << b  << endl;  
	fileout2 << " res " << endl << res  << endl;  

        CalcInverse (a, ainv);
        aib = ainv * Trans(b);

	fileout2 << " ainv " << endl << ainv << endl;  

	fileout2 << " aib " << endl << aib << endl;  
        
	/*
        ainv = 0.0;
        for (int j = 0; j < elsonvertex.Size(); j++)
          {
            for (int k = 0; k < ndsig; k++)
              for (int l = 0; l < ndsig; l++)
                ai(k,l) = a(j*ndsig+k,j*ndsig+l);

            CalcInverse (ai, aiinv);

            for (int k = 0; k < ndsig; k++)
              for (int l = 0; l < ndsig; l++)
                ainv(j*ndsig+k,j*ndsig+l) = aiinv(k,l);


            for (int k = 0; k < nloc; k++)
              for (int l = 0; l < ndsig; l++)
                bi(k,l) = b(k,j*ndsig+l);
            aibi = aiinv * Trans (bi);
            for (int k = 0; k < nloc; k++)
              for (int l = 0; l < ndsig; l++)
                aib(j*ndsig+l, k) = aibi(l,k);
          }
	*/

        NgProfiler::StopTimer (timer3);
        NgProfiler::StartTimer (timer5);
        // schur = b * aib;
        LapackMultAB (b, aib, schur);

	fileout2 << " b " << endl << b << endl;  
	fileout2 << " schur " << endl << schur << endl;  
 
        NgProfiler::StopTimer (timer5);
        NgProfiler::StartTimer (timer4);

        for (int m = 0; m < schur.Height(); m++)
          schur(m,m) += 1e-10;

        // CalcInverse (schur, invschur);
        invschur = schur;
        LapackInverse (invschur);

	fileout2 << " invschur " << endl << invschur << endl;  
        
        NgProfiler::StopTimer (timer4);

        FlatVector<> w(nloc, lh), sig(nsig, lh), hv(nsig, lh);
        w = invschur * res;
        sig = aib * w;

	fileout2 << " res " << res << endl; 
	fileout2 << " w " << w << endl; 

	fileout2 << " sig " << sig << endl; 
	fileout2 << " i " << i << endl; 
       
      
        for (int j = 0; j < elsonvertex.Size(); j++)
          sigma_glob.Range(ndsig*elsonvertex[j], ndsig*(elsonvertex[j]+1)) +=
            sig.Range(ndsig*j, ndsig*(j+1));

	fileout << " sigma_glob " << sigma_glob << endl; 

        lh.CleanUp (hp);
      }

    *testout << " sigma_glob " << sigma_glob << endl; 

    NgProfiler::StopTimer (timer1);

    dnums.SetSize(1);
    Vector<> elerr(1);
    err = 0;
    for (int i = 0; i < ne; i++)
      {
        ma.GetElementTransformation (i, eltrans, lh);

        const HDivFiniteElement<2> & felsig =
          dynamic_cast<const HDivFiniteElement<2> &> (fessig.GetFE(i, lh));
        
        FlatVector<> elsigma (ndsig, lh);
        elsigma = sigma_glob.Range (i*ndsig, (i+1)*ndsig);

        const IntegrationRule & ir = SelectIntegrationRule (ET_TRIG, 2*(2*psig+2));
            
        FlatMatrix<> shapesigref(ndsig, 2, lh);
        FlatMatrix<> shapesig(ndsig, 2, lh);

        double erri = 0;
        for (int k = 0; k < ir.GetNIP(); k++)
          {
            SpecificIntegrationPoint<2,2> sip(ir[k], eltrans, lh);

            felsig.CalcShape (ir[k], shapesigref);
            shapesig = (1.0/sip.GetJacobiDet()) * shapesigref * Trans (sip.GetJacobian());

            Vec<2> sigmai = Trans (shapesig) * elsigma;
            double jac = fabs (sip.GetJacobiDet());

            erri += jac * ir[k].Weight() * L2Norm2 (sigmai);
          }        

        dnums[0] = i;
        elerr(0) = erri;
        gferr -> SetElementVector(dnums, elerr);

        err += erri;
        lh.CleanUp();
      }

    cout << "********  error is less than " << sqrt (err) << endl;
    *testout << "********  error is less than " << sqrt (err) << endl;

    BaseVector & vu = gfu->GetVector();
    BaseVector & hv = *vu.CreateVector();

    hv = bfa -> GetMatrix() * vu;
    double energy = InnerProduct (vu, hv);
    
    delete &hv;
    
    if (filename.length())
      {
        fileout << fes.GetNDof() 
                << " " << sqrt(err) 
                << " " << sqrt(fabs (energy-refenergy))
                << " " << energy << endl
                << flush;
      }


    NgProfiler::StopTimer (timer);

  }
};





namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("equilibrationee", NumProcBrSchEE::Create);
  }
  
  Init init;
}
