/*
 * Braess-Schöberl error estimator - Laplace version
 * =================================================
 * 
*/ 

// 
bool Equ_EE_Local_Solver = 1;

#undef NEW_SPACES 

// #define DEBUG
#include <solve.hpp>
#include <comp.hpp> 
#include <fem.hpp> 
#include "../hdivhofespace_ee.hpp"
#include "../hdivhofespacesz.hpp"
#include "../hdivhofenew.hpp" 
#include "../l2hofespace_ee.hpp" 
#include "../facetfespace_ee.hpp" 
#include "../hybrid.hpp"              

using namespace ngfem;
using namespace ngsolve;
using namespace ngcomp;  
using namespace laplacehybrid; 

namespace numproc_ee_sz {

/* namespace{
  inline void rotate(const Vec<2>& x, Vec<2>& y) { y(0) = -x(1); y(1) = x(0); };
  inline double spr(const Vec<2>& x, const Vec<2>& y) { return x(0)*y(0)+x(1)*y(1); };
  }*/

/* ***************** Braess-Schöberl error estimator ************************ */

template <int DIM>
class NumProcBrSchEESZ : public NumProc
{
protected: 
  
  BilinearForm * bfa;
  LinearForm * lff;
  S_GridFunction<double> * gfu;
  S_GridFunction<double> * gferr; 
  S_GridFunction<double> * gfsigma;
  
  CoefficientFunction * coef_f;
  CoefficientFunction * coef_eps;

  string filename, filename2;
  ofstream fileout, fileout2;
  double refenergy;

  bool laplace; 
  bool quads; 
  bool printflag; 
  int rel2; 

public:

  NumProcBrSchEESZ (PDE & apde, const Flags & flags) : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));    
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction (flags.GetStringFlag ("solution", "")));
    gferr =  dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction (flags.GetStringFlag ("error", "")));
    gfsigma = 0; 
    if(flags.StringFlagDefined("sigma_delta")) 
      gfsigma =  dynamic_cast<S_GridFunction<double>*> (pde.GetGridFunction (flags.GetStringFlag ("sigma_delta", "")));

    printflag = flags.GetDefineFlag("print"); 

    coef_f = pde.GetCoefficientFunction (flags.GetStringFlag ("coeff", ""));

    filename = flags.GetStringFlag ("filename", "equ_ee_sz.out");
    filename2 = flags.GetStringFlag ("filename2", "matsz.out");
   
    refenergy = flags.GetNumFlag ("energy", 0);
    if (filename.length())
      {
        fileout.open (filename.c_str());
        fileout.precision(16);
      } 
    if (filename2.length())
      {
        fileout2.open (filename2.c_str());
        fileout2.precision(16);
      } 
 
    rel2 = int(flags.GetNumFlag("rel2",0.)); 
    
    laplace = flags.GetDefineFlag("laplace");
    *testout << " defined laplace " << laplace << endl;
    quads = flags.GetDefineFlag("quads"); 
    *testout << " quad mesh defined " << quads << endl; 
  }

  virtual ~NumProcBrSchEESZ()
  { ; }
  
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcBrSchEESZ<DIM> (pde, flags);
  }

  virtual void Do(LocalHeap & lh)
  {
    bool ee_robin_penalty = 0; 
    bool print = printflag; 
    bool print2 = printflag; 

    cout << "**************************************************************" << endl;
    cout << "called Braess Schoeberl error estimator SZ2d " << endl;
         
    static int timer = NgProfiler::CreateTimer ("Equilibration");
    static int timer1 = NgProfiler::CreateTimer ("Equilibration 1");
    static int timer2 = NgProfiler::CreateTimer ("Equilibration 2");
    static int timer3 = NgProfiler::CreateTimer ("Equilibration 3");
    static int timer4 = NgProfiler::CreateTimer ("Equilibration 4");
    static int timer5 = NgProfiler::CreateTimer ("Equilibration 5");
    static int timer6 = NgProfiler::CreateTimer ("Equilibration Schur local patches");
    static int timer7 = NgProfiler::CreateTimer ("Equilibration Solve");

    NgProfiler::StartTimer (timer);
    NgProfiler::StartTimer (timer1);

    time_t start=clock();

    const int ned = ma.GetNEdges();
    const int nfa = ma.GetNFaces(); 
    const int dim = ma.GetDimension(); 
    const int nfacet = (dim == 2) ? ned : nfa; 
    const int nel = ma.GetNE();
    const int np = ma.GetNP(); 
  
    cout << "general information: nel=" << nel << ", ned=" << ned << ", np=" << np << endl; 
          
    ARRAY<int> vnums, ednums, dnums, pnums, elnums, fanums;
  
    // Attention: dirichedges are SElIndex == 0 !! 
    BitArray dirichletfacets (nfacet);
    dirichletfacets.Clear(); 
    fanums.SetSize(0); 
     
    for (int i = 0; i < ma.GetNSE(); i++)
      {
	if (ma.GetSElIndex (i) == 0)
	  {
	    
	    if(DIM==2) 
	      ma.GetSElEdges(i, fanums);
	    else 
	      fanums.Append(ma.GetSElFace(i)); 
	    
	    for (int j = 0; j < fanums.Size(); j++)
	      dirichletfacets.Set(fanums[j]);
	  }
      }
    *testout << "dirichletfacets: " << endl << dirichletfacets << endl;
    
    /************** Update FE-SPACES relative to mesh-order *****************************/ 
    // h1-fespace (order p) 
    const FESpace & fes = gfu->GetFESpace();
    const BilinearFormIntegrator & bli = *bfa->GetIntegrator(0);
    ARRAY<CoefficientFunction*> acoef; 
    acoef.SetSize(0); 
    acoef.Append(coef_eps);
    HybridLaplaceIntegrator<DIM> * bli_hybrid = new HybridLaplaceIntegrator<DIM>(acoef);  
    
    const H1HighOrderFESpace * h1fes =
      dynamic_cast<const H1HighOrderFESpace*> (&gfu->GetFESpace());
    
    
    testout->precision(12); 

    //  // rel2=1 yields the same space for sigma as used in ZZ for flux  
    // for quads: 
    //       u \in Q^k -> sig \in Q^(k+1,k)xQ^(k,k+1), q_t \in Q^k, q_f \in P^k 
    // for trigs: 
    //       u \in P^k -> sig \in P^(k-1), q_t \in P^(k-2), q_f \in P^(k-1) 
    double curlinc = 0; 
    Flags hdiv_flags, facet_flags, l2_flags; 

    printflag = 1; 
    if(printflag) 
      {
	hdiv_flags.SetFlag("print"); 
	l2_flags.SetFlag("print"); 
	facet_flags.SetFlag("print"); 
      } 

    if(h1fes->VarOrder()) 
      { 
	/*
	const int relorder_h1 = h1fes->GetRelOrder(); 
	hdiv_flags.SetFlag("relorder",max(double(relorder_h1-1+rel2),-1.)); 
	hdiv_flags.SetFlag("relcurlorder",max(double(relorder_h1+curlinc-1+rel2),-1.)); 
	facet_flags.SetFlag("relorder",max(double(relorder_h1-1+curlinc+rel2),-1.));  
	l2_flags.SetFlag("relorder",max(double(relorder_h1-2+rel2),-1.)); 
	*/	
	const int relo_h1 = h1fes->GetRelOrder(); 
	#ifdef NEW_SPACES 
	double trig_relo_hdiv = double(relo_h1-1+rel2); 
	double trig_relo_l2 =  double(relo_h1-2+rel2); 
	double trig_relo_facet = double(relo_h1-1+rel2); 

	// ATTENTION: for elements with h1_order = 1 this can be improved ...  
	double quad_relo_hdiv = double(relo_h1+rel2); 
	double quad_relo_l2 = double(relo_h1+rel2); 
	double quad_relo_facet = double(relo_h1+rel2); 

 
	cout << " trig_relo_hdiv  " << trig_relo_hdiv << endl; 
	cout << " trig_relo_l2    " << trig_relo_l2 << endl; 
	cout << " trig_relo_facet " << trig_relo_facet << endl; 

	cout << " quad_relo_hdiv  " << quad_relo_hdiv << endl; 
	cout << " quad_relo_l2    " << quad_relo_l2 << endl; 
	cout << " quad_relo_facet " << quad_relo_facet << endl; 

	//  default for trigs 
	hdiv_flags.SetFlag("relorder", trig_relo_hdiv); 
	//hdiv_flags.SetFlag("relcurlorder",); 
	facet_flags.SetFlag("relorder",trig_relo_facet); // curlinc?? 
	l2_flags.SetFlag("relorder",trig_relo_l2);
	// sonst kein var_order 

	hdiv_flags.SetFlag("trig_relorder",trig_relo_hdiv); 
	hdiv_flags.SetFlag("quad_relorder",quad_relo_hdiv); 
	facet_flags.SetFlag("trig_relorder",trig_relo_facet); 
	facet_flags.SetFlag("quad_relorder",quad_relo_facet); 
	l2_flags.SetFlag("trig_relorder",trig_relo_l2); 
	l2_flags.SetFlag("quad_relorder",quad_relo_l2); 
		
	#else 
	// the following works for pure trig or pure quad-meshes 	
	//
	if(!quads)
	  {
	    hdiv_flags.SetFlag("relorder", double(relo_h1-1+rel2)); 
	    hdiv_flags.SetFlag("relcurlorder",double(relo_h1+curlinc-1+rel2)); 
	    facet_flags.SetFlag("relorder",double(relo_h1-1+curlinc+rel2)); // curlinc?? 
	    l2_flags.SetFlag("relorder",double(relo_h1-2+rel2)); 
	  }
	else
	  {
	    // for quads 
	    hdiv_flags.SetFlag("relorder", double(relo_h1+rel2)); 
	    hdiv_flags.SetFlag("relcurlorder",double(relo_h1+curlinc+rel2)); 
	    facet_flags.SetFlag("relorder",double(relo_h1+curlinc+rel2));  
	    l2_flags.SetFlag("relorder",double(relo_h1+rel2)); 
	    }
	#endif 
      } 
    else
      { 
	int p_h1 = h1fes->GetOrder(); 
	hdiv_flags.SetFlag("order",max(double(p_h1-1),0.)); 
	hdiv_flags.SetFlag("curl_orer",max(double(p_h1-1+curlinc),0.)); 
	facet_flags.SetFlag("order",max(double(p_h1-1+curlinc),0.)); 
	l2_flags.SetFlag("order",max(double(p_h1-2),0.)); 
      } 


    #ifdef NEW_SPACES 
    //discontinuous HDiv-Space (order p-1)  
    *testout << " *********** Update discont hdiv-space  for equilibrated EE" << endl; 
    hdiv_flags.SetFlag("discontinuous"); 
    HDivHighOrderFESpaceEE fessig (ma, hdiv_flags);
    fessig.Update(); 
        
    Vector<> sigma_glob(fessig.GetNDof());
    sigma_glob = 0.0;    
    *testout << " *********** Update l2-space for equilibrated EE" << endl; 
    // L2-space (order p-2) 
    L2HighOrderFESpaceEE fesl2 (ma, l2_flags); 
    fesl2.Update();
    *testout << " *********** Update facet-fespace for equilibrated EE" << endl; 
    // Facet-space (order p-1) 
    FacetFESpaceEE fes_facet(ma, facet_flags);
    fes_facet.Update(); 
#else 
   //discontinuous HDiv-Space (order p-1)  
    *testout << " *********** Update discont hdiv-space  for equilibrated EE" << endl; 
    hdiv_flags.SetFlag("discontinuous"); 
    HDivHighOrderFESpace fessig (ma, hdiv_flags);
    fessig.Update(); 
        
    Vector<> sigma_glob(fessig.GetNDof());
    sigma_glob = 0.0;    
    *testout << " *********** Update l2-space for equilibrated EE" << endl; 
    // L2-space (order p-2) 
    L2HighOrderFESpace fesl2 (ma, l2_flags); 
    fesl2.Update();
    *testout << " *********** Update facet-fespace for equilibrated EE" << endl; 
    // Facet-space (order p-1) 
    FacetFESpace fes_facet(ma, facet_flags);
    fes_facet.Update(); 

    


#endif 

    ARRAY<int> elsonvertex, facetsonvertex, allfacetsonvertex, patch_bd;
    ElementTransformation eltrans;
    double errt = 0, erre = 0, err = 0;

    void * hp = lh.GetPointer();
    
    // run over all vertex-patches om_V_i 
    // Assemble and solve local patch problems 
    for (int i = 0; i < np; i++)
      {
	elsonvertex.SetSize(0);
	facetsonvertex.SetSize(0);
	allfacetsonvertex.SetSize(0); 
	patch_bd.SetSize(0); 

        ma.SetThreadPercentage ( 100.0 * i / np );

        NgProfiler::StartTimer (timer2);

        (*testout) << "vertex " << i << " : " << endl;

        ma.GetVertexElements( i, elsonvertex);

        // find facets on vertex 
	allfacetsonvertex.SetSize(0); 
        facetsonvertex.SetSize (0);

	ARRAY<int> dnums_dirich_fa; 
	dnums_dirich_fa.SetSize(0); 

	for (int j = 0; j < elsonvertex.Size(); j++)
	  {
	    if(DIM==2) 
	      ma.GetElEdges (elsonvertex[j], fanums);
	    else 
	      ma.GetElFaces (elsonvertex[j], fanums); 	
	    //    ma.GetElFacets(elsonvertex[j],fanums); 
	    for (int k = 0; k < fanums.Size();k++)
	      {
		//	ARRAY<int> pnums; 
		// 	ma.GetFacetPNums(fanums[k],pnums); 
		
		ARRAY<int> pnums(2); 
		if(dim == 2) 
		  ma.GetEdgePNums (fanums[k], pnums[0], pnums[1]); 
		else 
		  ma.GetFacePNums (fanums[k], pnums); 
		
		
		// facetsonvertex: facets on vertex excl. dirichedges 
		if(!dirichletfacets.Test(fanums[k]) && !facetsonvertex.Contains(fanums[k])) 
		  for(int l=0; l<pnums.Size(); l++ ) 
		    if(pnums[l] == i) 
		      facetsonvertex.Append(fanums[k]);
		
		// all facets on vertex incl. dirch-edges !! 
		if(!allfacetsonvertex.Contains(fanums[k])) 
		  for(int l=0; l<pnums.Size(); l++ ) 
		    if(pnums[l] == i) 
		      allfacetsonvertex.Append(fanums[k]); 

		// facets on boundary of vertex-patch  and not dirich-edge -> dirch edges of local patch
            bool v_on_face=0; 
            for(int l=0;l<pnums.Size();l++) if(pnums[l]==i) { v_on_face=1; break;}  
            
            if(!patch_bd.Contains(fanums[k]) && !v_on_face &&!dirichletfacets.Test(fanums[k]))
               patch_bd.Append(fanums[k]); 
            
	      }
	  }  
	*testout << "has elements " << elsonvertex << endl;
	*testout << "has facets " << facetsonvertex << endl;
	*testout << "has all facets " << allfacetsonvertex << endl;
	 		
	// Compute dofs of local vertex-patch problems  
	// for sigma, lambda_T, and lambda_E 
	ARRAY<int> first_sig_dof(elsonvertex.Size()+1); 
	ARRAY<int> first_qt_dof(elsonvertex.Size()+1); 
	ARRAY<int> first_qf_dof(facetsonvertex.Size()+1); 
	ARRAY<int> first_qff_dof(facetsonvertex.Size()+1); 
	ARRAY<int> lam_lo_patchdofnrs(0); 
	ARRAY<int> lamf_lo_patchdofnrs(0); 
	ARRAY<int> lamt_lo_patchdofnrs(0); 

	// dof-tables for local patchproblem (i) 
	first_sig_dof[0]=0; 
	first_qt_dof[0]=0;  
	for(int j=0;j<elsonvertex.Size();j++) 
	  {
	    first_sig_dof[j+1] = first_sig_dof[j] + fessig.GetNElemDofs(elsonvertex[j]); 
	    first_qt_dof[j+1] = first_qt_dof[j] + fesl2.GetNElemDofs(elsonvertex[j]); 
	    lam_lo_patchdofnrs.Append(first_qt_dof[j]); 
	    lamt_lo_patchdofnrs.Append(first_qt_dof[j]); 
	  }

	// for facet-dof-nrs 
	first_qf_dof[0] = first_qt_dof[elsonvertex.Size()]; 
	first_qff_dof[0] = 0;
	for(int k=0;k<facetsonvertex.Size();k++)
	  { 
	    first_qf_dof[k+1] = first_qf_dof[k] + fes_facet.GetNFacetDofs(facetsonvertex[k]);  
	    first_qff_dof[k+1] = first_qff_dof[k] + fes_facet.GetNFacetDofs(facetsonvertex[k]);  
	    lam_lo_patchdofnrs.Append(first_qf_dof[k]); 
	    lamf_lo_patchdofnrs.Append(first_qff_dof[k]); 
	  } 

           

	    	
	int ndof_sig = first_sig_dof[elsonvertex.Size()]; 
	int ndof_lam = first_qf_dof[facetsonvertex.Size()];
	int ndof_lamf = first_qff_dof[facetsonvertex.Size()]; 
	
	bool printdoftables = 0; 
	if(printdoftables)
	  {
	    *testout << " ******** DofTables of patch system " << endl; 
	    *testout << " ndof_lam " << ndof_lam << " ndof_sig " << ndof_sig <<  endl; 
	    *testout << "first_qf " << first_qf_dof << endl; 
	    *testout << " first_qt " << first_qt_dof << endl; 
	    *testout << " first_sig " << first_sig_dof << endl;
          *testout << " lam_lo_patchdofnrs " << lam_lo_patchdofnrs << endl;   
          *testout << " lamf_lo_patchdofnrs " << lamf_lo_patchdofnrs << endl;    
	  }
	
	/* *********************************************** */
	/* Set Up Local Problem on Vertex-Patch   
	   
	( A B^T ) (sig) = (0) 
	( B  0  ) (lam)   (res)    
	
	with (sig,tau)_A := (tau,sig)_L2, 
	B^T = (B_T^T, B_F^T) (qt,sig)_B_T := (qt,div(sig))_B_T 
	(qf,sig)_B_F = (qf,[sig])_B_F 
	res = (rt, rf), (rt,qt) := ([(laplace_T(u) + f_T)*phi_V)]_[P_k],qt), (rf,qf) := ([du/dn],qf) */
	/* ******************** *************************** */ 
			     
        FlatVector<> res(ndof_lam, lh);
        res = 0;
        FlatMatrix<> a(ndof_sig, lh),ainv(ndof_sig,lh);
        FlatMatrix<> b(ndof_lam, ndof_sig, lh),aib(ndof_sig,ndof_lam,lh); 
        
        FlatMatrix<> a_hybrid(ndof_sig, lh),ainv_hybrid(ndof_sig,lh);
        FlatMatrix<> b_hybrid(ndof_lam, ndof_sig, lh);
        
	//	FlatMatrix<> c(ndof_lam, ndof_lam, lh); 
	FlatMatrix<> schur(ndof_lam,lh); 
	FlatMatrix<> mat(ndof_sig+ndof_lam, ndof_sig+ndof_lam, lh); 

	// for NEW_ASSEMBLE
	FlatMatrix<> Bf(ndof_lamf,ndof_sig,lh); 
	FlatMatrix<> SS(ndof_sig,lh); 
	FlatVector<> XRt(ndof_sig,lh); 
	FlatVector<> Rf(ndof_lamf,lh); 
       
	XRt=0.0; Rf =0.; Bf=0.; SS=0.;  
	a = 0.; ainv=0.; b=0.; b_hybrid=0.; a_hybrid = 0.;   
      	void * heapj = lh.GetPointer();

	// Loop over all elements of vertex-patch: 
	// Compute element matrices and then assemble into global vertex-patch-system 
	for (int j = 0; j < elsonvertex.Size(); j++)
          {
	    *testout << " elsonvertex  " << j << " : " << elsonvertex[j] << endl; 

	    ma.GetElementTransformation (elsonvertex[j], eltrans, lh);  
	    
	    // Local elementvector of primal variable u  
	    const NodalFiniteElement & fel_u = 
	      dynamic_cast<const NodalFiniteElement&> (fes.GetFE(elsonvertex[j], lh));
	    fes.GetDofNrs (elsonvertex[j], dnums);
	    FlatVector<> elu (dnums.Size(), lh);
            gfu -> GetElementVector (dnums, elu);
	  
	    const HDivHighOrderFiniteElement<DIM> & fel_sig =
	      dynamic_cast<const HDivHighOrderFiniteElement<DIM> &> (fessig.GetFE(elsonvertex[j], lh));
  
	    const L2HighOrderFiniteElement & fel_qt =
	      dynamic_cast<const L2HighOrderFiniteElement &> (fesl2.GetFE(elsonvertex[j], lh));
	    const FacetVolumeFiniteElement & fel_facetvol = 
	      dynamic_cast<const FacetVolumeFiniteElement &> (fes_facet.GetFE(elsonvertex[j],lh));
	    const int nsig_el = fel_sig.GetNDof(); 
	    const int nqt_el = fel_qt.GetNDof(); 
	    const int nfvol_el = fel_facetvol.GetNDof(); 

	    if(nsig_el != first_sig_dof[j+1] - first_sig_dof[j]) 
	      *testout << " ERROR: nsigel elem " << elsonvertex[j] << endl; 
	    if(nqt_el != first_qt_dof[j+1] - first_qt_dof[j]) 
	      *testout << " ERROR: nqt_el elem " << elsonvertex[j] << endl; 
             
          FlatMatrix<> a_elmat_hybrid(nsig_el,lh); 
          FlatMatrix<> bt_elmat_hybrid(nqt_el, nsig_el, lh), bf_elmat_hybrid(nfvol_el, nsig_el, lh); 
          FlatMatrix<> b_elmat_hybrid(nqt_el+nfvol_el,nsig_el,lh); 
           
          
        
          
          bli_hybrid->AssembleElementMatrix (fel_sig,fel_qt,fel_facetvol, eltrans, *coef_eps, a_elmat_hybrid, bt_elmat_hybrid, bf_elmat_hybrid,lh); 
                                        
	    FlatMatrix<> a_elmat(nsig_el,lh), ai_elmat(nsig_el,lh); 
	    FlatMatrix<> bt_elmat(nqt_el, nsig_el, lh),bf_elmat(nfvol_el, nsig_el, lh); 
	    FlatMatrix<> b_elmat(nqt_el+nfvol_el,nsig_el,lh); 
	    // local condensation 
	    FlatVector<> rt_elvec(nqt_el, lh), rf_elvec(nfvol_el,lh), f_elvec(nqt_el,lh), F_elvec(nqt_el,lh), phiv_elvec(nqt_el,lh); 
	    a_elmat = 0.0; 
	    bt_elmat = 0.0; 
	    bf_elmat = 0.0; 
	    //btaibt_e = 0.0; bfaibf_e=0.0; btaibf_e=0.0; aibf_e=0.0; aibt_e=0.0;  
	    rt_elvec = 0.0; 
	    phiv_elvec = 0.0;
	    
	    
	    rf_elvec = 0.0; 
	    f_elvec = 0.0; 
	    b_elmat = 0.0; 

	    // Calculate rhs ft = (Laplace_T(elu),qt) 
	    	    
          FlatMatrix<> shape_sig(nsig_el, DIM, lh);
	    FlatVector<> shape_sign_ref(nsig_el, lh);
	    FlatVector<> div_sig(nsig_el, lh); 
	    FlatVector<> shape_qt(nqt_el, lh);
	    FlatVector<> shape_qfvol(nfvol_el,lh); 
	    FlatVector<> shape_v(dnums.Size(), lh); 
	    FlatMatrix<> dshape_v(dnums.Size(), DIM, lh); 
	    
 
            int vi = 0;
            ma.GetElVertices (elsonvertex[j], vnums);
            for (int k = 0; k < vnums.Size(); k++)
              if (vnums[k] == i) vi = k;

	    // Integration Rule for volume element 
	    const IntegrationRule & ir = 
	      SelectIntegrationRule (fel_sig.ElementType(), (2*fel_sig.Order())); 
	    
	    void * heapp = lh.GetPointer();
	    for (int k = 0; k < ir.GetNIP(); k++)
            {
		SpecificIntegrationPoint<DIM,DIM> sip(ir[k], eltrans, lh);
		double det = fabs(sip.GetJacobiDet());
		double dx = det * ir[k].Weight(); 
			
		// Piola-Trafo of H(div) shapes and div 
		shape_sig = (1.0/det) * fel_sig.GetShape (sip.IP(), lh) * Trans (sip.GetJacobian());
		div_sig = (1.0/det) * fel_sig.GetDivShape (sip.IP(), lh); 
		
		fel_qt.CalcShape(sip.IP(), shape_qt); 
		
		//double eps = coef_eps->Evaluate(sip);
		double eps=1.; 
		// a = (1/eps sig,tau) 
		a_elmat += (dx/eps) * shape_sig * Trans (shape_sig);
		
		// bt = (qt,div(sig))
		bt_elmat += dx * shape_qt * Trans (div_sig);
           
		// (rt,qt) = ((Laplace_T(u) + f_T)*psi_V,qt)
		//         = -(nabla_T u,nabla_T qt)_T + (du/dn,qt)_dT +(f_T,qt)_T 
		
		// fi = f_t * psi_V
		fel_u.CalcShape(ir[k],shape_v);	
		double fi = coef_f -> Evaluate (sip) * shape_v[vi]; 
		
		rt_elvec += dx * fi * shape_qt;
						
		FlatVector<> flux_u(DIM,lh),flux_v(DIM,lh); 
		bli.CalcFlux (fel_u, eltrans, ir[k], elu, flux_u, true, lh);
		FlatVector<> elv(elu.Size(),lh); 
		elv = 0; 
		elv[vi] = 1.; 
		bli.CalcFlux (fel_u,eltrans, ir[k],elv, flux_v,false,lh); 
		

		// Computation of Laplace_T(u)via Integration by parts 
		// rt += - nabla_T(u)*nabla_T(phi_V.qt)  + int(dudn * q * phiV)
		FlatMatrix<> dshape_qt(nqt_el, DIM, lh); 
		
		// dshape_qt = Trans(sip.GetJacobianInverse())* Trans(fel_qt.GetDShape(sip.IP(),lh));   //* sip.GetJacobianInverse(); 
            dshape_qt =  fel_qt.GetDShape(sip.IP(),lh) * sip.GetJacobianInverse();
            
		double gradphiv_dot_gradu = InnerProduct(flux_u,flux_v);
		//	dshape_v = Trans(sip.GetJacobianInverse())* Trans(fel_u.GetDShape(sip.IP(),lh));
		// for(int l=0; l<DIM; l++) 
		//  gradphiv_dot_gradu += dshape_v(l,vi)*flux_u(l);  
		
		rt_elvec -= dx * gradphiv_dot_gradu * shape_qt; 
            
		rt_elvec -= (dx * shape_v[vi]) *(dshape_qt*flux_u);  
            
            f_elvec -= dx * dshape_qt*flux_u; 
            
		lh.CleanUp(heapp);

	      }

	    if(DIM==2) 
	      ma.GetElEdges(elsonvertex[j],fanums); 
	    else 
	      ma.GetElFaces(elsonvertex[j],fanums);
	    
          
           
	    
	    bool printeigsys = 0; 
	    if(printeigsys)
	      {
		Vector<> lami(a_elmat.Height());
		Matrix<> evecs(a_elmat.Height());


		*testout << " patch i : " << i << endl;  
		*testout << " elsonvertex " << elsonvertex << endl; 
		*testout << " elsonvertex.Size() " << elsonvertex.Size() << endl; 
		*testout << " elsonvertex j : " << elsonvertex[j] << endl;  
		*testout << " a_elmat " << a_elmat << endl; 
		CalcEigenSystem (a_elmat, lami, evecs);
		(*testout) << "lami = " << endl << lami << endl; 
	      }

	    // Get patch-position of el-facets (used for dirich edges and facet elmats) 

	    if(DIM==2) 
	      ma.GetElEdges (elsonvertex[j], fanums);
	    else 
	      ma.GetElFaces (elsonvertex[j], fanums); 

	    ARRAY<int> fa_pos(fanums.Size()); 
	    for(int k =0;k<fanums.Size();k++)
	      fa_pos[k]=facetsonvertex.Pos(fanums[k]); 
	    	    
	    for(int k=0;k<fanums.Size();k++) 
	      if (fa_pos[k] > -1 && 
		  fel_facetvol.GetFirstFacetDof(k+1) - fel_facetvol.GetFirstFacetDof(k) !=  
		  first_qf_dof[fa_pos[k]+1] -first_qf_dof[fa_pos[k]]) 
		*testout << " ERROR qf_dofs face " << fanums[k] << endl; 
	    

	    bool printface=0; 
	    ma.GetElVertices (elsonvertex[j], vnums); 
	    const FACE * faces =  ElementTopology::GetFaces (ma.GetElType(elsonvertex[j]));  // DIM == 3
	    const EDGE * edges =  ElementTopology::GetEdges (ma.GetElType(elsonvertex[j])); // DIM == 2 
	    const POINT3D * points = ElementTopology::GetVertices (ma.GetElType(elsonvertex[j]));

	    ELEMENT_TYPE eltype = ma.GetElType(elsonvertex[j]); 

	    // nfa = DIM==2 ? edges.Size() : faces.Size(); // fanums.Size(); 
	    
	    // tranformation of facet integration points (1D / 2D) to element points (2D / 3D):
	    Facet2ElementTrafo transform(eltype); 
		    
	    // normal on reference element with |normal| = length of tangential (in 3D: area of face)
	    const NORMAL * normals = ElementTopology::GetNormals(eltype);

	    int pf=fel_sig.Order(); 
	    for(int k=0; k<fanums.Size();k++) 
	      pf = max(fel_facetvol.GetFacetOrder(k),pf);  
	     
          
	    for(int k =0;k<fanums.Size();k++)
	      {
		if(printface) 
		  *testout << "facenumber " << k << " : " << fanums[k] 
			   << " with facetsonvertex pos " << fa_pos[k] << endl;
		
		// normal on ref_elem on face k
		Vec<DIM> normal_ref;
		for (int i=0; i<DIM; i++)
		  normal_ref(i) = normals[k][i];
		
		// Integrate over facetsonvertex (fa_pos[k] >= 0)for facet entries (bf, rf) entries,  
		// but for rhs rt += int_E( dudn phiv q_t) ds (due to i.b.p. of laplace(u)) also for dirichfacets 
		// (->allfacetsonvertex)
 		// if(fa_pos[k]>=0) 
		//if(allfacetsonvertex.Contains(fanums[k]))   
		  {
		
		    ARRAY<int> pnums(2); 
		    if(DIM==2) 
		      ma.GetEdgePNums(fanums[k], pnums[0], pnums[1]);  
		    else 
		      ma.GetFacePNums(fanums[k], pnums); 
		    
		    
		    /**** Integration over facet fanums[k] ****/ 
		    // Calculate rhs rf_elem = (du/dn phi_V , qf)_F_k
		    // and bf_elem = (qf,sig*n)_F_K
		    
		    
		    
		    ELEMENT_TYPE fatype; 
		    switch(pnums.Size()) 
		      { 
		      case 2: fatype=ET_SEGM; 
			break; 
		      case 3: fatype=ET_TRIG; 
			break; 
		      case 4: fatype=ET_QUAD; 
			break; 
		      }
		    

		    //int pf = fel_sig.Order()+5; 
		    void * heapp = lh.GetPointer();
		    const IntegrationRule & ir_fa = SelectIntegrationRule(fatype, (2*pf));
		    for (int l = 0; l < ir_fa.GetNIP(); l++)  
		      {
			const IntegrationPoint  & ip_fa = ir_fa[l];
			IntegrationPoint ip_vol = transform(k, ip_fa);
			
			SpecificIntegrationPoint<DIM,DIM> sip_vol (ip_vol, eltrans, lh);
			double det = sip_vol.GetJacobiDet(); 
						
			// ***** Transformation of outer normal vector
			FlatVector<> nv(DIM,lh); 
			nv = det * Trans(sip_vol.GetJacobianInverse()) * normal_ref; 
			double area = L2Norm(nv);
			nv /= area; 
					
			fel_u.CalcShape(ip_vol,shape_v);
			FlatVector<> flux_u(DIM,lh); 
			bli.CalcFlux (fel_u, eltrans, ip_vol, elu, flux_u, true, lh);
			double dudn = InnerProduct (flux_u, nv);
			
			double da = area * ir_fa[l].Weight(); 
			double dar = ir_fa[l].Weight();  // for refrence facet 

			// **** rt_elvec: remaining part of laplace_T(u): +(du/dn, psi_V qt) on all facets
			fel_qt.CalcShape(ip_vol,shape_qt); 
			
                  f_elvec += (da*dudn)*shape_qt;  
                  
                  rt_elvec += (da*dudn*shape_v[vi])*shape_qt;  
			
			if(fa_pos[k]<0) // if not facet on vertex 
			  {
			    lh.CleanUp(heapp); 
			    continue;
			  }
			
			// ***** rf_elvec += (du/dn, psi_v*q_f) ***  
			fel_facetvol.CalcFacetShape(k,ip_fa, shape_qfvol); 
			rf_elvec -= (da*dudn *shape_v[vi])*shape_qfvol; 
			
			// ***** bf_elmat += ([sig.n],q_f) 
			// int_A(sig.n) dA = int_Ar (1/det*F*sigr)^T.(F^-T*nr/area) det * area dAr
			// int_A(sig.n) q dA  = int_Ar(sigr.nr) qr dAr  
			shape_sign_ref = fel_sig.GetShape(sip_vol.IP(),lh) * normal_ref;
			
			// if(fa_pos[k]>=0) // add only if no global dirich-edge 
			  bf_elmat -= dar * shape_qfvol * Trans(shape_sign_ref);
			
			lh.CleanUp(heapp); 
		      } // intrule 
		  } // if allfacets 
	      } // fanums 
	    
          
	    for(int k=0;k<fanums.Size();k++) 
	      { 
		// impose BC sigma . n = 0 on local patches: Robin penalty  
		// if(fa_pos[k] < 0 && !dirichletfacets.Test(fanums[k])) // dirich edges !!
		if(ee_robin_penalty) 
		  { 
		    if(patch_bd.Contains(fanums[k]))
		      {  
			ARRAY<int> dtnums; 
			fel_sig.GetFacetDofs(k,dtnums); 
			for(int l=0;l<dtnums.Size();l++) 
			  a_elmat(dtnums[l],dtnums[l]) += 1e5; 
		      }  
		  } 
		else 
 		  {
                   
		    if(patch_bd.Contains(fanums[k])) 
		      {  
			
			ARRAY<int> dtnums; 
			fel_sig.GetFacetDofs(k,dtnums); 
                  
			for(int l=0;l<dtnums.Size();l++)
			  {
                      
                      
			    // dnums_dirich_fa.Append(first_sig_dof[j]+dtnums[l]); 
			    for(int m=0;m<a_elmat.Height();m++) 
			      { 
				a_elmat(dtnums[l],m) = 0.; 
				a_elmat(m,dtnums[l]) = 0.;
                        a_elmat_hybrid(dtnums[l],m) = 0.; 
                        a_elmat_hybrid(m,dtnums[l]) = 0.;
                        }
			    a_elmat(dtnums[l],dtnums[l]) = 1.; 
                      a_elmat_hybrid(dtnums[l],dtnums[l]) = 1.; 
			  }
			
		
			
			for(int l=0;l<dtnums.Size();l++) 
			  for(int m=0;m<bt_elmat.Height();m++) 
                  {
                    bt_elmat(m,dtnums[l]) = 0.;
                    bt_elmat_hybrid(m,dtnums[l]) = 0.;
                  } 
			
			for(int l=0;l<dtnums.Size();l++) 
			  for(int m=0;m<bf_elmat.Height();m++) 
                  { 
                    bf_elmat(m, dtnums[l])=0.;
                    bf_elmat_hybrid(m, dtnums[l])=0.;
                  }
                  }
		  }
	      }
	    
	      if(printeigsys)
	      {
		Vector<> lami(a_elmat.Height());
		Matrix<> evecs(a_elmat.Height());
	
		CalcEigenSystem (a_elmat, lami, evecs);
	      } 

            
            
            
            
	    if(!Equ_EE_Local_Solver)
	      { 
		NgProfiler::StartTimer(timer7); 
		// Assemble element entries to patch-system 
		for (int k = 0; k < nsig_el; k++)
		  for (int l = 0; l < nqt_el; l++)
            { 
              b(first_qt_dof[j]+l, first_sig_dof[j]+k) = bt_elmat(l, k);
              b_hybrid(first_qt_dof[j]+l, first_sig_dof[j]+k) = bt_elmat_hybrid(l, k);
            } 
            
		for (int k = 0; k < nsig_el; k++)
		  for (int l = 0; l < nsig_el; l++)
            { a(first_sig_dof[j]+l, first_sig_dof[j]+k) = a_elmat(l, k);
		    a_hybrid(first_sig_dof[j]+l, first_sig_dof[j]+k) = a_elmat_hybrid(l, k);
            }
            
            
		
		NgProfiler::StopTimer(timer7);
	      }
	   	    
	    
	    ARRAY<int> first_facet_dof(fanums.Size()+1); 
	    for(int k=0;k<=fanums.Size();k++) 
	      first_facet_dof[k]=fel_facetvol.GetFirstFacetDof(k);  

          for(int k=0;k<fanums.Size();k++)
          { 
          if(fa_pos[k]<0) // orig dirich edge has no dofs for lambda_f 
          { 
            
            int nqf = first_facet_dof[k+1] - first_facet_dof[k];  
                  
            for (int m = 0; m < nsig_el; m++)
              for (int l = 0; l < nqf; l++)
            {
                bf_elmat_hybrid(first_facet_dof[k]+l, m) = 0.;
                // bf_elmat(first_facet_dof[k]+l, m) = 0.; this already 0 
            }
          }
          }
          if(print)
          {
            *testout << " elsonvertex[j] " << elsonvertex[j] << endl; 
            *testout << " aelmat " << a_elmat << endl; 
            *testout << " aelmat_hybrid " << a_elmat_hybrid << endl; 
            *testout << " btelmat " << bt_elmat << endl; 
            *testout << " btelmat_hybrid " << bt_elmat_hybrid << endl; 
            *testout << " bfelmat " << bf_elmat << endl; 
            *testout << " bfelmat_hybrid " << bf_elmat_hybrid << endl; 
            
          }
          
          
	    if(!Equ_EE_Local_Solver)
	      {
		NgProfiler::StartTimer(timer7); 
		
		// Assemble the element matrices (of elsonvertex[j]) into local patch system 
		// Get Position of element faces but only for facetsonvertex: 
		for(int k=0;k<fanums.Size();k++)
		  { 
		    if(fa_pos[k] >=0) // no lambda_f dofs on gobal dirich edges (u) // ohnehin schon 0 
		      { 
			res.Range (first_qf_dof[fa_pos[k]],first_qf_dof[fa_pos[k]+1]) += rf_elvec.Range(first_facet_dof[k],first_facet_dof[k+1]); 
			int nqf = first_facet_dof[k+1] - first_facet_dof[k];  
			
			for (int m = 0; m < nsig_el; m++)
			  for (int l = 0; l < nqf; l++)
                  { 
                                             
			    b(first_qf_dof[fa_pos[k]]+l, first_sig_dof[j]+m) += bf_elmat(first_facet_dof[k]+l, m);
                      b_hybrid(first_qf_dof[fa_pos[k]]+l, first_sig_dof[j]+m) += bf_elmat_hybrid(first_facet_dof[k]+l, m);
                  }
                 // b(first_qf_dof[fa_pos[k]]+l, first_sig_dof[j]+m) += bf_elmat(first_facet_dof[k]+l, m);
		      }
		  }
              
              
	    	    
		for(int l=0;l<nqt_el;l++) 
		  for(int m=0;m<nsig_el;m++) 
		    b_elmat(l,m) = bt_elmat(l,m); 
		
		for(int l=0;l<nfvol_el;l++) 
		  for(int m=0;m<nsig_el;m++) 
		    b_elmat(l+nqt_el,m) = bf_elmat(l,m); 

		// Assemble into global(vertex-patch) res-vector and matrices a,b 
		res.Range (first_qt_dof[j], first_qt_dof[j+1]) += rt_elvec; 

		NgProfiler::StopTimer(timer7); 
	      
            
            
	    // Hint: eventuell not used facets (dirichedges,etc) explizit auf 0 setzen!! 
	      }
	    
	    NgProfiler::StartTimer(timer7); 
	    // Assemble inverse of A (discont hdiv-space -> block-diag)
	    NgProfiler::StartTimer(timer6); 
	    CalcInverse(a_elmat_hybrid,ai_elmat); 
	    
	    NgProfiler::StopTimer(timer6); 
	    for (int k = 0; k < nsig_el; k++)
              for (int l = 0; l < nsig_el; l++)
		ainv(first_sig_dof[j]+l, first_sig_dof[j]+k) = ai_elmat(l, k);   
	    NgProfiler::StopTimer(timer7); 
	    
          
          /* fileout2 << "a_elmat " << a_elmat << endl; 
          fileout2 << " bf_elmat " << bf_elmat << endl; 
          fileout2 << " bt_elmat " << bt_elmat << endl; 
          fileout2 << " rt_elvec " << rt_elvec << endl; 
          fileout2 << " rf_elvec " << rf_elvec << endl; 
          
          */
          
	    if(Equ_EE_Local_Solver)
	      {

	    /* ********************************************************************
	       Assembling strategy (elemntwise static condensation of sigma and lambda_T
	       incl. rank1 correction of lambda_F 
	       
	       sigma = SS * BF^T lam_F + X res_T 
	       lam_F = [BF * SS BF^T + 1_lo]^(-1) (res_F - BF X res_T)
	       lam_T = - S_TT^(-1) (res_T + A^(-1) B_F^T lam_F 

	       with block-diagonal matrices (element-wise)  
	       SS = (A^(-1) B_T^T S_TT(-1) B_T A^(-1) - A^(-1))    
	       X = A^(-1) B_T^T S_TT^(-1) 
	       S_TT = B_T A^(-1) B_T^T 

	       with original problem  (  A    B_T^T  B_F^T  ) (sig)    = 0
	                              (  B_T    0      0    ) (lam_T)  = res_T
				      (  B_F    0    1_Flo  ) (lam_F)  = res_F

	       leads to solve system on skeleton
	            [BF * SS * BF^T + 1_lo] lam_F = res_F - BF X res_T ; 
	       
	       -> compute sigma 
	       
	       ************************************************************************* */    
  
		 
	    FlatMatrix<> btaibt_e(nqt_el, lh), inv_btaibt_e(nqt_el, lh), ss_e(nsig_el,lh);
	    FlatMatrix<> aibt_e(nsig_el,nqt_el, lh),aibtsi_e(nsig_el,nqt_el,lh);

	    FlatVector<> xrt_e(nsig_el,lh); 

	    NgProfiler::StartTimer(timer6); 
	    // s_TT 
	    aibt_e = ai_elmat*Trans(bt_elmat); 
	    btaibt_e = bt_elmat*aibt_e; 
            
         
          
	    CalcInverse(btaibt_e,inv_btaibt_e);
         
	    
	    // SS
	    aibtsi_e = aibt_e*inv_btaibt_e;
	    ss_e = aibtsi_e* Trans(aibt_e); 
	    ss_e -= ai_elmat; 
	

	    // xrt
	    xrt_e = aibtsi_e * rt_elvec; 
	    XRt.Range (first_sig_dof[j], first_sig_dof[j+1]) += xrt_e; 

	    for (int k = 0; k < nsig_el; k++)
              for (int l = 0; l < nsig_el; l++)
                SS(first_sig_dof[j]+l, first_sig_dof[j]+k) = ss_e(l, k);
	    
	    // Bf , Rf
	    for(int k=0;k<fanums.Size();k++)
	      { 
		if(fa_pos[k] >=0) // no dirich_edge and facet on vertex 
		  { 
		    int nqf = first_facet_dof[k+1] - first_facet_dof[k];  
		    Rf.Range (first_qff_dof[fa_pos[k]],first_qff_dof[fa_pos[k]+1]) += rf_elvec.Range(first_facet_dof[k],first_facet_dof[k+1]); 
		    for (int m = 0; m < nsig_el; m++)
		      for (int l = 0; l < nqf; l++)
			Bf(first_qff_dof[fa_pos[k]]+l, first_sig_dof[j]+m) += bf_elmat(first_facet_dof[k]+l, m);
		  }
	      }
	    NgProfiler::StopTimer(timer6); 
	    
	      }
	     
	    
	    
	    	    
	    lh.CleanUp(heapj); 

	  } // elsonvertex[j] 

	/*  Full Patch System 
	mat = 0.0; 
	// amat
	for (int l = 0; l < ndof_sig; l++)
	  for(int k=0;k<ndof_sig;k++) 
	    mat(l, k) = a(l, k);

	for (int l = 0; l < ndof_sig; l++)
	  for(int k=0;k<ndof_lam;k++) 
	    mat(l, k+ndof_sig) = b(k, l);
	
	for(int k=0;k<ndof_lam;k++) 
	  for (int l = 0; l < ndof_sig; l++)
	    mat(k+ndof_sig,l) = b(k, l);
	
	if(allfacetsonvertex.Size() == facetsonvertex.Size())
	  for(int l=0;l<lamf_lo_patchdofnrs.Size();l++) 
	    for(int m=0;m<lamf_lo_patchdofnrs.Size();m++) 
	      mat(ndof_sig+lamf_lo_patchdofnrs[l],ndof_sig+lamf_lo_patchdofnrs[m]) = 1.0; 

	fileout << " lamf_lo_patch " << lamf_lo_patchdofnrs << endl; 

	fileout2 << " mat1 " << mat << endl; 

	if(allfacetsonvertex.Size() == facetsonvertex.Size())
	  for(int l=0;l<lamt_lo_patchdofnrs.Size();l++) 
	    for(int m=0;m<lamt_lo_patchdofnrs.Size();m++) 
	      mat(ndof_sig+lamt_lo_patchdofnrs[l],ndof_sig+lamt_lo_patchdofnrs[m]) = 1.0; 

	fileout << " lamt_lo_patch " << lamt_lo_patchdofnrs << endl; 

	fileout2 << " mat2 " << mat << endl;
 
	if(allfacetsonvertex.Size() == facetsonvertex.Size())
	  for(int l=0;l<lam_lo_patchdofnrs.Size();l++) 
	    for(int m=0;m<lam_lo_patchdofnrs.Size();m++) 
	      mat(ndof_sig+lam_lo_patchdofnrs[l],ndof_sig+lam_lo_patchdofnrs[m]) = 1.0; 
	
	fileout << " lam_lo_patch " << lam_lo_patchdofnrs << endl; 

	fileout2 << " mat3 " << mat << endl; 
	
	*/ 



	
	/* **************************************************************** 
	    Solve system on patch-skeleton
	        [BF * SS * BF^T + 1_lo] lam_F = res_F - BF X res_T ; 

	    and recompute sigma_patch 
	        sigma = SS * BF^T lam_F + X res_T 
	    ************************************************************* */

	FlatVector<> lam_f(ndof_lamf,lh), sig_patch(ndof_sig,lh);  	
	if(Equ_EE_Local_Solver)
	  {
	    sig_patch = XRt; 
	    if(ndof_lamf > 0) 
	      { 
		FlatMatrix<> Schur_f(ndof_lamf,lh),invSchur_f(ndof_lamf,lh), TSchur_f(ndof_lamf,lh);  
		FlatMatrix<> SSBf(ndof_sig,ndof_lamf,lh); 
		//FlatMatrix<> Cf(ndof_lamf,lh); 
		//FlatMatrix<> Ct(ndof_lam-ndof_lamf,lh); 
		//FlatMatrix<> C(ndof_lam,lh); 
		
		//Cf=0.0; 
		//Ct=0.0; 
		//C=0.0; 
		NgProfiler::StartTimer(timer6); 
		SSBf=SS*Trans(Bf);  
		Schur_f=Bf*SSBf; // BF * SS * BF_^T 
              
             
		// rank 1 correction 
		if(allfacetsonvertex.Size() == facetsonvertex.Size()) // if local patch has no dirich-edges
            { for(int l=0;l<lamf_lo_patchdofnrs.Size();l++) 
		    for(int m=0;m<lamf_lo_patchdofnrs.Size();m++) 
		      { 
			Schur_f(lamf_lo_patchdofnrs[l],lamf_lo_patchdofnrs[m]) += 1.0; 	
			//Cf(lamf_lo_patchdofnrs[l],lamf_lo_patchdofnrs[m]) = 1.0; 	
                  
		      }
            }   
                   
		/*
		  if(allfacetsonvertex.Size() == facetsonvertex.Size())
		  for(int l=0;l<lamt_lo_patchdofnrs.Size();l++) 
		  for(int m=0;m<lamt_lo_patchdofnrs.Size();m++) 
		  { 
		  Ct(lamt_lo_patchdofnrs[l],lamf_lo_patchdofnrs[m]) = 1.0; 	
	      }
	      
	      if(allfacetsonvertex.Size() == facetsonvertex.Size())
	      for(int l=0;l<lamf_lo_patchdofnrs.Size();l++) 
	      for(int m=0;m<lamf_lo_patchdofnrs.Size();m++) 
	      { 
	      C(lam_lo_patchdofnrs[l],lam_lo_patchdofnrs[m]) = 1.0; 	
	      }
	      fileout2 << " C "  << C << endl; 
	      fileout2 << " Ct "  << Ct << endl; 
	      fileout2 << " Cf "  << Cf << endl; 
		*/
		
		// CholeskyFact + Rückwärtseinsetzen 
				
		//TSchur_f = Trans(TSchur_f);
		Rf -= Bf*XRt;
		
            
		//lam_f = Rf;  
		//LaLinearSolveRHS(ndof_lamf,&Schur_f(0,0),&lam_f(0));
           
	      CalcInverse(Schur_f,invSchur_f); 
		
		/* 
		   Vector<> lami(Schur_f.Height());
		   Matrix<> evecs(Schur_f.Height());
		   
		   
		   CalcEigenSystem (Schur_f, lami, evecs); 
		   fileout2 << " vertex i " << i << endl; 
		   fileout2 << " lami schurf " << lami << endl; 
		*/ 
		
		lam_f = invSchur_f*Rf;
	      

		sig_patch += SSBf*lam_f;
	      } 
	    

	


	NgProfiler::StopTimer(timer6); 
	  }   

	


	bool printres = 0; 
	if(printres)	
	  {
	    *testout << " for vertex " << i << " patch matrices " << endl;  
	    *testout << "has elements " << elsonvertex << endl;
	    *testout << "has facets " << facetsonvertex << endl;
	    *testout << "has bd-facets " << patch_bd << endl; 
	    *testout << "b_all  sz = " << endl << b << endl;
	    *testout << "a_all  sz = " << endl << a << endl;
	    *testout << "res_all sz =" << endl << res << endl;  
	  }
		
	NgProfiler::StopTimer (timer2);
	NgProfiler::StartTimer (timer3);
	
    
	double errpatch = 0;

        double errti = L2Norm2 (res.Range (0, first_qf_dof[0]));
        double errfi = L2Norm2 (res.Range (first_qf_dof[0],res.Size()));
        errt += errti; 
        erre += errfi;
       
	FlatVector<> sig(ndof_sig, lh); 
	if(!Equ_EE_Local_Solver) // Solver full patch system via schur-complement
	  { 
	    NgProfiler::StartTimer(timer7); 
	    // Already computed in assemble routine 
	    // *testout << " CalcInv a " ; 	
	    
	    print2 = 0; 
	    if(print2)
	      {
		fileout2 << " vertex " << i << endl; 
		fileout2 << " a " << endl << a  << endl;  
            fileout2 << " a_hybrid " << endl << a_hybrid  << endl;  
            fileout2 << " a-a_hybrid " << endl << a-a_hybrid  << endl;  
		fileout2 << " b " << endl << b  << endl;  
            fileout2 << " b_hybrid " << endl << b_hybrid  << endl;  
            fileout2 << " b-b_hybrid " << endl << b-b_hybrid  << endl;
		fileout2 << " res " << endl << res  << endl;  
	      }
	    
	    // CalcInverse (a, ainv);
	    
	    if(print2) 
	      {
		fileout2 << " ainv " << endl << ainv << endl;  
		
		Vector<> lami(ainv.Height());
		Matrix<> evecs(ainv.Height());
		
		CalcEigenSystem (ainv, lami, evecs); 
		fileout2 << " lami ainv " << lami << endl; 
	      } 
	    
	    aib = ainv * Trans(b); 
	    
	    
	    if(print2)
	      {

		fileout2 << " aib " << endl << aib << endl;  
	      }
	    
	    FlatMatrix<> invschur(ndof_lam,lh); 
	    FlatMatrix<> schur2(ndof_lam,lh); 
	    NgProfiler::StopTimer (timer3);
	    NgProfiler::StartTimer (timer5);
	    //schur = b * aib;
	    LapackMultAB (b, aib, schur2);
	    
	    // Cholesky should be sufficient 
	    
         // for (int m = 0; m < schur.Height(); m++)
            // schur(m,m) += 1e-10; 
       
            
	    // add rank 1 matrix since constant of lamis is not uniquely determined
	    if(allfacetsonvertex.Size() == facetsonvertex.Size())
	      for(int l=0;l<lam_lo_patchdofnrs.Size();l++) 
		for(int m=0;m<lam_lo_patchdofnrs.Size();m++) 
		  schur2(lam_lo_patchdofnrs[l],lam_lo_patchdofnrs[m]) -=1.;  // original -=1
	    
          
          
	    
	    if(print2)
	      {
		//fileout2 << " b " << endl << b << endl;  
		fileout2 << " vertex " << i << endl; 
		// fileout2 << " schur " << endl << schur << endl;  
		fileout2 << " schur2 cor" << endl << schur2 << endl; 
		
		Vector<> lami(schur.Height());
		Matrix<> evecs(schur.Height());
		
		//CalcEigenSystem (schur, lami, evecs); 
		//fileout2 << " lami schur " << lami << endl;
		CalcEigenSystem (schur2, lami, evecs); 
		fileout2 << " lami schur2 " << lami << endl; 
		
	      }

	    NgProfiler::StopTimer (timer5);
	    NgProfiler::StartTimer (timer4);
	    
	    // Take Choelsky for this and not compute inverse! 
	    //CalcInverse (schur, invschur);
	    invschur = schur2; 
	 
          LapackInverse (invschur); 
	    
	    if(print2) 
          { 
            fileout2 << " invschur " << endl << invschur << endl;  
            fileout2 << " res " << res << endl; 
          } 
	    NgProfiler::StopTimer (timer4);
	    
	    FlatVector<> w(ndof_lam, lh), hv(ndof_sig, lh); 
	    w = invschur * res;
	    sig = aib * w;
	    NgProfiler::StopTimer(timer7); 
          
     

	    if(print2) 
	      {
		fileout2 << " res " << res << endl; 
		fileout2 << " w " << w << endl; 
		fileout2 << " sig " << sig << endl; 
		fileout2 << " i " << i << endl; 
	      }
	    
	  } 

	if(Equ_EE_Local_Solver) sig = sig_patch; 

	for (int j = 0; j < elsonvertex.Size(); j++)
	  { 
	    ARRAY<int> ddnums; 
	    fessig.GetDofNrs(elsonvertex[j],ddnums); 
	    for(int k=0;k<ddnums.Size();k++) 
	      sigma_glob(ddnums[k]) += sig(first_sig_dof[j]+k); 
	  }
	
	if(print2) 
	  fileout2 << " sigma_glob V " << sigma_glob << endl; 

	lh.CleanUp (hp);
	
      } // end loop over vertex-patches  
    
    bool printsigma=0; 
    if(printsigma) 
      *testout << " sigma_glob " << sigma_glob << endl; 
    
    NgProfiler::StopTimer (timer1);


    // Compute error function L2-lowest order with  (sigma,sigma)_T,0 on elem T 
    dnums.SetSize(1);
    Vector<> elerr(1);
    err = 0;
    for (int i = 0; i < nel; i++)
      {
        ma.GetElementTransformation (i, eltrans, lh);

	const HDivHighOrderFiniteElement<DIM> & fel_sig =
	  dynamic_cast<const HDivHighOrderFiniteElement<DIM> &> (fessig.GetFE(i, lh));

	ARRAY<int> ddnums; 
	fessig.GetDofNrs(i,ddnums); 
	int nsig_el = ddnums.Size(); 

	FlatVector<> elsigma (ddnums.Size(), lh);
        for(int k=0;k<ddnums.Size();k++) 
	  elsigma(k) = sigma_glob(ddnums[k]); 

        const IntegrationRule & ir = SelectIntegrationRule (ma.GetElType(i), ((2*fel_sig.Order())));

        FlatMatrix<> shape_sig(nsig_el, dim, lh);
        double erri = 0;
	
	void * heapp = lh.GetPointer();
        for (int k = 0; k < ir.GetNIP(); k++)
          {
            SpecificIntegrationPoint<DIM,DIM> sip(ir[k], eltrans, lh);
	    double det = sip.GetJacobiDet();
	    double dx = det*ir[k].Weight(); 

            shape_sig = (1.0/sip.GetJacobiDet()) * fel_sig.GetShape(sip.IP(),lh) * Trans (sip.GetJacobian());

            Vec<DIM> sigmai = Trans (shape_sig) * elsigma;
           
            erri += dx * L2Norm2 (sigmai);
	    lh.CleanUp(heapp); 
	  }        


        dnums[0] = i;
        elerr(0) = erri;
        gferr -> SetElementVector(dnums, elerr);

        err += erri;
        lh.CleanUp();
      }

 
    cout << "********  error is less than " << sqrt (err) << endl;
    *testout  << "********  error is less than " << sqrt (err) << endl;

    BaseVector & vu = gfu->GetVector();
    BaseVector & hv = *vu.CreateVector();

    hv = bfa -> GetMatrix() * vu;
    double energy = InnerProduct (vu, hv);
    
    delete &hv;
    //*testout << " gferr " << gferr->GetVector() << endl; 
    //*testout << " err " << err << endl; 

    // Visualization of sigma_glob 
    if(gfsigma) 
      {
	BaseVector & vsig = gfsigma->GetVector();
	FlatVector<double> fvsig = vsig.FVDouble(); 
	
	if(sigma_glob.Size() == vsig.Size()) 
	  for( int l=0;l<sigma_glob.Size(); l++)
	    fvsig(l) = sigma_glob(l); 
	else 
	  { 
	    *testout << " ATTENTION sigma glob und gfsigma not of same size " << endl; 
	    *testout << " Sizes: sigma glob " << sigma_glob.Size() << " vsig " << vsig.Size() << endl;   
	  }
      }
        
    if (filename.length())
      {
        fileout << fes.GetNDof() 
                << " " << sqrt(err) 
                << " " << sqrt(fabs ((energy-refenergy)))
                << " " << energy << endl
                << flush;
      }

    NgProfiler::StopTimer (timer); 


    /*   int ss = ma.GetElOrder(2); 
    if(ss<30) 
      { 
	ma.SetElOrder(2,ss+10); 
	} */ 

      delete bli_hybrid;
    
  }

};

template class  NumProcBrSchEESZ<2>;
template class  NumProcBrSchEESZ<3>;


static NumProc * CreateNumProcBrSchEESZ (PDE & pde, const Flags & flags)
{
  const MeshAccess & ma = pde.GetMeshAccess();
  if (ma.GetDimension()==2)
    return new NumProcBrSchEESZ<2> (pde, flags);
  else
    return new NumProcBrSchEESZ<3> (pde, flags);
}

  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("equilibrationeesz", CreateNumProcBrSchEESZ);
  }
  
  Init init;
}
