#include <comp.hpp>
#include <fem.hpp> 
#include "facetfespace_ee.hpp" 
     
namespace ngcomp
{
  using namespace ngfem ;

// ------------------------------------------------------------------------
  FacetFESpaceEE ::  FacetFESpaceEE (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : FacetFESpace(ama, flags,parseflags)
  {
    name="FacetFESpaceEE(facetee)";
    // defined flags
    DefineNumFlag("trig_order");
    DefineNumFlag("quad_order");
    DefineNumFlag("trig_relorder");
    DefineNumFlag("quad_relorder");

    
    if(parseflags) ParseFlags(flags);
  
    quad_relorder = int (flags.GetNumFlag("quad_relorder",rel_order)); 
    trig_relorder = int (flags.GetNumFlag("trig_relorder",rel_order)); 
    quad_order = int (flags.GetNumFlag("quad_order",order)); 
    trig_order = int (flags.GetNumFlag("trig_order",order));
    if(flags.NumFlagDefined("trig_relorder")||flags.NumFlagDefined("quad_relorder")) 
      var_order=1;  
   
  }
  
// ------------------------------------------------------------------------
  FacetFESpaceEE :: ~FacetFESpaceEE ()
  {
    ;
  }

// ------------------------------------------------------------------------
  FESpace * FacetFESpaceEE :: Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FacetFESpaceEE (ma, flags, true);
  }

// ------------------------------------------------------------------------
  void FacetFESpaceEE :: Update()
  {

    if(ma.GetDimension()==3) 
      { 
	FacetFESpace :: Update();
	return;
      }

    if (low_order_space)
      low_order_space -> Update();
    
    // number of facets
    nel = ma.GetNE();
    nfa = (ma.GetDimension() == 2 ? ma.GetNEdges() : ma.GetNFaces()); 
    
    order_facet.SetSize(nfa);
    order_facet = INT<2> (0,0);
    fine_facet.SetSize(nfa);
    fine_facet = 0; 
    
    ARRAY<int> fanums;
    
    
    for (int i = 0; i < nel; i++)
      {
	INT<3> el_orders = ma.GetElOrders(i); 
	
	switch(ma.GetElType(i))
	  {
	  case ET_TRIG:
	    if(var_order)
	      for(int l=0;l<3;l++) el_orders[l] += trig_relorder; 
	    else 
	      for(int l=0;l<3;l++) el_orders[l] = trig_order; 
	    break;
	    
	  case ET_QUAD:
	    if(var_order) 
	      for(int l=0;l<3;l++) 
		{ 
		  el_orders[l] += quad_relorder;  
		  //(el_orders[l]==1) el_orders[l]--; // lowest-order case 
		}
	    else 
	      for(int l=0;l<3;l++) el_orders[l] = quad_order; 
	    break;
	  default: 
	    if(!var_order) 
	      el_orders = INT<3> (order,order,order); 
	  }


	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const POINT3D * points = ElementTopology :: GetVertices (eltype);	
	
	if (ma.GetDimension() == 2)
	  {
	    ma.GetElEdges (i, fanums);
	    for (int j=0;j<fanums.Size();j++) 
	      fine_facet[fanums[j]] = 1; 
	    
	    if(var_order)
	      {
		const EDGE * edges = ElementTopology::GetEdges (eltype);
		for(int j=0; j<fanums.Size(); j++)
		  for(int k=0;k<2;k++)
		    if(points[edges[j][0]][k] != points[edges[j][1]][k])
		      { 
			order_facet[fanums[j]] = INT<2>(max(el_orders[k], order_facet[fanums[j]][0]),0);
			break; 
		      }
	      }
	  }
	else
	  {
	    ARRAY<int> elfaces,vnums;
	    ma.GetElFaces(i,elfaces);
	    for (int j=0;j<elfaces.Size();j++) fine_facet[elfaces[j]] = 1; 
	    
	    if(var_order) 
	      {
		ma.GetElVertices (i, vnums);
		const FACE * faces = ElementTopology::GetFaces (eltype);
		for(int j=0;j<elfaces.Size();j++)
		  {
		    if(faces[j][3]==-1) // trig  
		      {
			order_facet[elfaces[j]][0] = max(order_facet[elfaces[j]][0],el_orders[0]+rel_order);
			order_facet[elfaces[j]][1] = order_facet[elfaces[j]][0]; 
		      }
		    else //quad_face
		      {
			int fmax = 0;
			for(int k = 1; k < 4; k++) 
			  if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
					
			INT<2> f((fmax+3)%4,(fmax+1)%4); 
			if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
			
			// fmax > f[0] > f[1]
			// p[0] for direction fmax,f[0] 
			// p[1] for direction fmax,f[1] 
			for(int l=0;l<2;l++)
			  for(int k=0;k<3;k++)
			    if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			      {
				order_facet[elfaces[j]][l] = max(order_facet[elfaces[j]][l] + rel_order, 
								el_orders[k]);
				break; 
			      } 
			
		      }
		  }
	      }
	    
	  }
	
      }
    
    // distribute dofs
    ncfa = 0; 
    ndof = nfa; // low_order space
    
    first_facet_dof.SetSize(nfa+1); 
    first_facet_dof = nfa;
      
    if (ma.GetDimension() == 2)
    {
      for (int i = 0; i < nfa; i++)
      {
        first_facet_dof[i] = ndof;
        ndof += order_facet[i][0];
      }
      first_facet_dof[nfa] = ndof;
    } // 2D
    else  // 3D
    {
      int inci = 0;
      INT<2> p; 
      ARRAY<int> pnums;
      for (int i=0; i< nfa; i++)
      {
        p = order_facet[i];
        ma.GetFacePNums(i,pnums);

        switch(pnums.Size())
        {
          case 3: 
            inci = ((p[0]+1)*(p[0]+2))/2 - 1;
            break;
          case 4: 
            inci= (p[0]+1)*(p[1]+1) - 1;
            break;
        }
        first_facet_dof[i] = ndof;
        ndof+= inci;
      }
      first_facet_dof[nfa] = ndof;
    } // 3D

      //HERBERT: h1hofe code. is that correct ??
    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;
      
      //HERBERT: no prolongation so far       
      //prol->Update();
    
    if(print) 
      {
	*testout << "*** Update FAcetFESpace: General Information" << endl;
	*testout << " order facet (facet) " << order_facet << endl; 
	*testout << " first_facet_dof (facet)  " << first_facet_dof << endl; 
      }
   /* 
    cout << "    ne = " << nel << ", nfa = " << nfa << ", ncfa=" << ncfa << ", ndof=" << ndof << endl;
    cout << "    order_facet = " << order_facet << endl;*/
  }



  
  
// ------------------------------------------------------------------------
   // register FESpaces
  namespace facefespace_ee_cpp
  {
    class Init
    {
      public:
        Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("facet_ee", FacetFESpaceEE::Create);
    }

    Init init;
  }

}


