/**
   High Order Finite Element Space for H(Div)
*/
/* Continuous and discontinuous version : 
   Discont space has no extra low-order block(!) 
*/ 

#include <comp.hpp> 
#include <fem.hpp>  
#include "hdivhofespacesz.hpp" 
#include "hdivhofespace_ee.hpp" 
 
namespace ngcomp
{


  HDivHighOrderFESpaceEE ::    HDivHighOrderFESpaceEE (const MeshAccess & ama, const Flags & flags)
    : HDivHighOrderFESpaceSZ (ama,flags)
  {
    quad_relorder = int (flags.GetNumFlag("quad_relorder",rel_order)); 
    trig_relorder = int (flags.GetNumFlag("trig_relorder",rel_order)); 
    if(flags.NumFlagDefined("trig_relorder")||flags.NumFlagDefined("quad_relorder")) 
      var_order=1; 
    quad_order = int (flags.GetNumFlag("quad_order",order)); 
    trig_order = int (flags.GetNumFlag("trig_order",order)); 
  }
  
  HDivHighOrderFESpaceEE :: ~HDivHighOrderFESpaceEE ()
  {
    ;
  }

  FESpace * HDivHighOrderFESpaceEE ::
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new HDivHighOrderFESpaceEE (ma, flags);
  }

  void HDivHighOrderFESpaceEE :: Update()
  {
    // In discontinuous spaces the order on edges and faces  
    // are also set to the maximal order of adjacent elements  
    // and not to the element order (Motivation: Equilibrated_EE) 
    
    // gibts net 
    //  if (low_order_space)
    //  low_order_space -> Update();
    
    
    if(ma.GetDimension() == 3)
      {
	HDivHighOrderFESpaceSZ :: Update(); 
	return;
      }

    *testout << " UPDATE HDivHighOrderFESpaceEE with trigrelorder " <<  trig_relorder 
	     << " and quad_relorder " << quad_relorder << endl; 

    

    nv = ma.GetNV();
    nel = ma.GetNE();
    nfa = (ma.GetDimension() == 2 ? ma.GetNEdges() : ma.GetNFaces()); 
       
    order_facet.SetSize(nfa);
    order_inner.SetSize(nel); 
    order_inner_curl.SetSize(nel); 
    fine_facet.SetSize(nfa); 
   
    order_facet = INT<2>(0,0); 
    order_inner = INT<3>(0,0,0); 
    order_inner_curl = INT<3>(0,0,0); 
    fine_facet = 0; 

    int dim = ma.GetDimension();
   
    for(int i=0;i<nel;i++)
      {
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	
	ARRAY<int> elfacets; 
	if(dim==2)
	  ma.GetElEdges(i,elfacets);
	else 
	  ma.GetElFaces(i,elfacets); 
	
	for (int j=0;j<elfacets.Size();j++) fine_facet[elfacets[j]] = 1; 
	
	INT<3> el_orders; 
	if(var_order)
	  el_orders = ma.GetElOrders(i);  
	else 
	  el_orders = INT<3>(order,order,order); 

	if(dim==2)
	  {
	    switch(ma.GetElType(i))
	      {
	      case ET_TRIG:
		if(var_order)
		  for(int l=0;l<2;l++) el_orders[l] += trig_relorder; 
		else 
		  for(int l=0;l<2;l++) el_orders[l] = trig_order; 
		break;
		
	      case ET_QUAD:
		if(var_order) 
		  for(int l=0;l<2;l++) el_orders[l] += quad_relorder;  
		else 
		  for(int l=0;l<2;l++) el_orders[l] = quad_order; 
		break;
	      }
	    
	    for(int k=0;k<dim;k++)
	      {
		order_inner_curl[i][k]= max(el_orders[k],0); // is a testversion :  + rel_curl_order,0);
		order_inner[i][k] = max(el_orders[k],0);
	      }
	    
	    const EDGE * edges = ElementTopology::GetEdges (eltype);
	    for(int j=0; j<elfacets.Size(); j++)
	      for(int k=0;k<2;k++)
		if(points[edges[j][0]][k] != points[edges[j][1]][k])
		  { 
		    order_facet[elfacets[j]][0] = max(el_orders[k], order_facet[elfacets[j]][0]);
		    break; 
		  }
	  }
	else
	  {
	    ARRAY<int> vnums; 
	    ma.GetElVertices (i, vnums);
	    const FACE * faces = ElementTopology::GetFaces (eltype);
	    for(int j=0;j<elfacets.Size();j++)
	      {
		if(faces[j][3]==-1) // trig  
		  {
		    order_facet[elfacets[j]][0] = max(order_facet[elfacets[j]][0],el_orders[0]+rel_curl_order);
		    order_facet[elfacets[j]][1] = order_facet[elfacets[j]][0]; 
		  }
		else //quad_face
		  {
		    int fmax = 0;
		    for(int k = 1; k < 4; k++) 
		      if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		    
		    INT<2> f((fmax+3)%4,(fmax+1)%4); 
		    if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		    
		    // fmax > f[0] > f[1]
		    // p[0] for direction fmax,f[0] 
		    // p[1] for direction fmax,f[1] 
		    for(int l=0;l<2;l++)
		      for(int k=0;k<3;k++)
			    if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			      {
				order_facet[elfacets[j]][l] = max(order_facet[elfacets[j]][l] + rel_curl_order, 
								 el_orders[k]);
				break; 
			      } 
		    
		  }
	      }
	  }
      }
    for(int i=0;i<nfa;i++) if(!fine_facet[i]) order_facet[i] = INT<2> (0,0); 

    
    if(print) 
      {
	*testout << " dicsont EE space  " << discont << endl;
	*testout << " fine_facet[i] (hdiv) " << fine_facet << endl; 
	
	*testout << " order_facet (hdivho) " << order_facet << endl; 
	*testout << " order_inner (hdivho) " << order_inner << endl; 
	*testout << " order_inner_curl (hdivho) " << order_inner_curl << endl; 
      }
    
    if(uniform_order_inner > -1) 
      order_inner = INT<3>(uniform_order_inner,uniform_order_inner,uniform_order_inner);
    if(dim==3 && uniform_order_facet > -1) 
      order_facet = INT<2>(uniform_order_facet,uniform_order_facet); 
    if(dim==2 && uniform_order_facet > -1) 
      order_facet = INT<2>(uniform_order_facet,0); 
    
    UpdateDofTables(); 
  }

 

  // register FESpaces
  namespace
#ifdef MACOS
  hdivhofespace_cpp
#endif
  {
    class Init
    {
    public:
      Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hdivho_ee", HDivHighOrderFESpaceEE::Create);
    }

    Init init;
  }

}



