/**
   High Order Finite Element Space for H(Div)
*/
/* Continuous and discontinuous version : 
   Discont space has no extra low-order block(!) 
*/ 



#include <comp.hpp> 
#include <fem.hpp>  
#include "hdivhofespacesz.hpp" 
#include "hdivhofenew.hpp" 



 
namespace ngcomp 
{
  HDivHighOrderFESpaceSZ ::    HDivHighOrderFESpaceSZ (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama,
	       int (flags.GetNumFlag ("order", 0)),
	       int (flags.GetNumFlag ("dim", 1)),
	       flags.GetDefineFlag ("complex"))
  {
    print = flags.GetDefineFlag("print"); 

    discont = flags.GetDefineFlag("discontinuous"); 
    
    if(!discont) 
      low_order_space = 0; // new RaviartThomasFESpace (ma,dimension, iscomplex);
    else 
      low_order_space = 0; 


    // Variable order space: 
    //      in case of (var_order && order) or (relorder) 
    var_order = flags.GetDefineFlag("variableorder");  
    order =  int (flags.GetNumFlag ("order",0)); 
    curl_order =  int (flags.GetNumFlag ("curlorder",order)); 
    if(flags.NumFlagDefined("relorder") && !flags.NumFlagDefined("order")) 
      var_order = 1; 
    
    rel_order=int(flags.GetNumFlag("relorder",order-1)); 
    rel_curl_order=int(flags.GetNumFlag("rel_curl_order",rel_order)); 
    
    if(flags.NumFlagDefined("order") && flags.NumFlagDefined("relorder")) 
      {
	if(var_order)
	  cerr << " WARNING: HDivHoFeSpace: inconsistent flags: variableorder, order and relorder "
	       << "-> variable order space with order, but rel_order is ingored " << endl; 
	else 
	  { 
	    var_order = 0; 
	      cerr << " WARNING: HDivHoFeSpace: inconsistent flags: order and rel_order "
		   << "-> uniform order space with order " << endl; 
	  }
      }

    if(!var_order && flags.NumFlagDefined("relcurlorder")) 
      throw Exception("HDivHoFeSpace: flag -relcurlorder=. for not variable order space is not valid"); 
        
    uniform_order_inner = int (flags.GetNumFlag ("orderinner", -1));
    
    if(flags.NumFlagDefined("orderedge") || flags.NumFlagDefined ("orderface")) 
      throw Exception("Flag 'orderface' and 'orderedge' for hdivho are obsolete. Use flag 'orderfacet' instead!"); 
    
    uniform_order_facet = int (flags.GetNumFlag ("orderfacet", -1));

    // Evaluator for shape tester 
    if (ma.GetDimension() == 2)
      {
	ARRAY<CoefficientFunction*> coeffs(1);
	coeffs[0] = new ConstantCoefficientFunction(1);
	evaluator = GetIntegrators().CreateBFI("masshdiv", 2, coeffs);
      }
    else
      {
        ARRAY<CoefficientFunction*> coeffs(1);
	coeffs[0] = new ConstantCoefficientFunction(1);
	evaluator = GetIntegrators().CreateBFI("masshdiv", 3, coeffs);
	boundary_evaluator = GetIntegrators().CreateBFI("robinhdiv", 3, coeffs);
      }
     if (dimension > 1)
      {
	evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
	boundary_evaluator =
	  new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
      }
  }
  
  HDivHighOrderFESpaceSZ :: ~HDivHighOrderFESpaceSZ ()
  {
    ;
  }

  FESpace * HDivHighOrderFESpaceSZ ::
  Create (const MeshAccess & ma, const Flags & flags)
  {
    int order = int(flags.GetNumFlag ("order", 0));
    int dim = int(flags.GetNumFlag ("dim", 1));
    bool iscomplex = flags.GetDefineFlag ("complex");
    
    if (order == -1)
      return new RaviartThomasFESpace (ma, dim, iscomplex);
    // Space with MG
    else
      return new   HDivHighOrderFESpaceSZ (ma, flags);
    
  }

  void HDivHighOrderFESpaceSZ :: Update()
  {
    // In discontinuous spaces the order on edges and faces  
    // are also set to the maximal order of adjacent elements  
    // and not to the element order (Motivation: Equilibrated_EE) 
    
    if (low_order_space)
      low_order_space -> Update();
    
    nv = ma.GetNV();
    nel = ma.GetNE();
    nfa = (ma.GetDimension() == 2 ? ma.GetNEdges() : ma.GetNFaces()); 
       
    order_facet.SetSize(nfa);
    order_inner.SetSize(nel); 
    order_inner_curl.SetSize(nel); 
    fine_facet.SetSize(nfa); 
   
     
    int p = 0, pc = 0; 
    
    if(!var_order)
      {
	p = order; pc = curl_order; 
      } 
    
    order_facet = INT<2>(pc,pc); 
    order_inner = INT<3>(p,p,p); 
    order_inner_curl = INT<3>(pc,pc,pc); 
    fine_facet = 0; 

    int dim = ma.GetDimension();
    
    for(int i=0;i<nel;i++)
      {
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	
	ARRAY<int> elfacets; 
	if(dim==2)
	  ma.GetElEdges(i,elfacets);
	else 
	  ma.GetElFaces(i,elfacets); 
	
	for (int j=0;j<elfacets.Size();j++) fine_facet[elfacets[j]] = 1; 
	
	if(!var_order) continue; 
	
	INT<3> el_orders = ma.GetElOrders(i);  
	
	for(int k=0;k<dim;k++)
	  {
	    order_inner_curl[i][k]= max(el_orders[k] + rel_curl_order,0);
	    order_inner[i][k] = max(el_orders[k]+rel_order,0);
	  }
	  
	if(dim==2)
	  {
	    const EDGE * edges = ElementTopology::GetEdges (eltype);
	    for(int j=0; j<elfacets.Size(); j++)
	      for(int k=0;k<2;k++)
		if(points[edges[j][0]][k] != points[edges[j][1]][k])
		  { 
		    order_facet[elfacets[j]][0] = max(el_orders[k]+rel_curl_order, order_facet[elfacets[j]][0]);
		    break; 
		  }
	  }
	else
	  {
	    ARRAY<int> vnums; 
	    ma.GetElVertices (i, vnums);
	    const FACE * faces = ElementTopology::GetFaces (eltype);
	    for(int j=0;j<elfacets.Size();j++)
	      {
		if(faces[j][3]==-1) // trig  
		  {
		    order_facet[elfacets[j]][0] = max(order_facet[elfacets[j]][0],el_orders[0]+rel_curl_order);
		    order_facet[elfacets[j]][1] = order_facet[elfacets[j]][0]; 
		  }
		else //quad_face
		  {
		    int fmax = 0;
		    for(int k = 1; k < 4; k++) 
		      if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		    
		    INT<2> f((fmax+3)%4,(fmax+1)%4); 
		    if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		    
		    // fmax > f[0] > f[1]
		    // p[0] for direction fmax,f[0] 
		    // p[1] for direction fmax,f[1] 
		    for(int l=0;l<2;l++)
		      for(int k=0;k<3;k++)
			    if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			      {
				order_facet[elfacets[j]][l] = max(order_facet[elfacets[j]][l] + rel_curl_order, 
								 el_orders[k]);
				break; 
			      } 
		    
		  }
	      }
	  }
      }
    for(int i=0;i<nfa;i++) if(!fine_facet[i]) order_facet[i] = INT<2> (0,0); 

    
    if(print) 
      {
	*testout << " dicsont " << discont << endl;
	if(!discont) *testout << " fine_facet[i] (hdiv) " << fine_facet << endl; 
	
	*testout << " order_facet (hdivho) " << order_facet << endl; 
	*testout << " order_inner (hdivho) " << order_inner << endl; 
	*testout << " order_inner_curl (hdivho) " << order_inner_curl << endl; 
      }
    
    if(uniform_order_inner > -1) 
      order_inner = INT<3>(uniform_order_inner,uniform_order_inner,uniform_order_inner);
    if(uniform_order_facet > -1) 
      order_facet = INT<2>(uniform_order_facet,uniform_order_facet); 
        
    UpdateDofTables(); 
  }

  void HDivHighOrderFESpaceSZ :: UpdateDofTables()
   {
     nv = ma.GetNV(); 
     nfa = (ma.GetDimension()==3  ? ma.GetNFaces() : ma.GetNEdges());
     nel = ma.GetNE();
     int dim = ma.GetDimension();
     ARRAY<int> pnums; 
     
     first_facet_dof.SetSize(nfa+1); 
     first_inner_dof.SetSize(nel+1); 
     
     ndof = nfa; 
     first_facet_dof = ndof; 

     if(dim==2)
       {
	 for (int i = 0; i < nfa; i++)
	   {
	     first_facet_dof[i] = ndof;
	     if(fine_facet[i])
	       ndof += order_facet[i][0];
	   }

	 first_facet_dof[nfa] = ndof;
       
	 
	 
	 int inci = 0;
	 for(int i=0; i< nel; i++)
	   {
	     INT<3> pc = order_inner_curl[i];
	     INT<3> p = order_inner[i];
	     switch(ma.GetElType(i))
	       {
	       case ET_TRIG:
		 inci = pc[0]*(pc[0]-1)/2 + p[0]*(p[0]-1)/2 + p[0]-1;
		 break;
	       case ET_QUAD:
		 inci = pc[0]*pc[1] + p[0]*p[1]+p[0]+p[1];
		 break;
	       }
	     if (inci < 0) inci = 0;
	     {
	       first_inner_dof[i] = ndof;
	       ndof+=inci;
	     }
	   }
	 first_inner_dof[nel] = ndof;
       }
     else 
       {
	 int inci;
	 for (int i=0; i< nfa; i++) 
	   {
	     inci=0; 
	     if(fine_facet[i])
	       {
		 INT<2> p = order_facet[i]; 
		 ma.GetFacePNums(i,pnums);
		 switch(pnums.Size())
		   {
		   case 3: //Triangle
		     inci= (p[0]*p[0]+3*p[0])/2; 
		     break;
		   case 4: //Quad 
		     inci= p[0]*p[1] + p[0] + p[1];
		     break;
		   }
	       }
	     if (inci < 0) inci = 0;
	     first_facet_dof[i] = ndof;
	     ndof+= inci;
	     
	   }
	 first_facet_dof[nfa] = ndof;
	 
	 for (int i=0; i< nel; i++)
	   {
	     INT<3> p = order_inner[i];
	     INT<3> pc = order_inner_curl[i];
	     int inci = 0;
	     
	     switch(ma.GetElType(i))
	       {
	       case ET_TET:
		 if(p[0]>1)
		   inci = p[0]*(p[0]+1)*(p[0]-1)/6 + p[0]*(p[0]-1)/2 + p[0]-1;
		 if(pc[0]>1)
		   inci += pc[0]*(pc[0]+1)*(pc[0]-1)/3 + pc[0]*(pc[0]-1)/2; ;
		 break;
	       case ET_PRISM:
		 inci = (p[0]+1)*(3*(p[0]-1)+(p[0]-1)*(p[0]-2)) 
		   +(p[0]-1)*(p[0]+1)*(p[0]+2)/2;
		 break;
	       case ET_HEX:
		 inci =  2*pc[0]*pc[1]*pc[2] + pc[0]*pc[1] + pc[1]*pc[2] + pc[0]* pc[2]
		   +  p[0]*(p[1]*p[2] + p[1] + p[2] + 1)  + p[1]*p[2] + p[1] + p[2]; 
		 break;
	       case ET_PYRAMID: 
		 inci=0; 
		 cout << "WARNING: there are hdiv-pyramids (not implemented yet) !! " << endl;
		 break; 
	       default:
		 inci = 0;
		 break;
	       }
	     if (inci < 0) inci = 0;
	     first_inner_dof[i] = ndof;
	     ndof+= inci;
	   }
	 first_inner_dof[nel] = ndof;
       }
     
     if(discont) 
       { 
	 ndof = 0; 
	 for(int i=0;i<nel;i++)
	   {
	     int incii = first_inner_dof[i+1]-first_inner_dof[i]; 

	     ARRAY<int> elvertices; 
	     ARRAY<int> elfacets; 
	     if(dim==2) 
	       ma.GetElEdges(i,elfacets);
	     else 
	       ma.GetElFaces(i,elfacets);
	     
	     ma.GetElVertices(i,elvertices); 
	 
	     for(int j=0; j<elfacets.Size(); j++) 
	       incii+=first_facet_dof[elfacets[j]+1]-first_facet_dof[elfacets[j]]+1; // incl. lowest-order 

	     
	     first_inner_dof[i] = ndof; 
	     ndof += incii;
	     
	   } 
	 first_inner_dof[nel] = ndof; 
	 
	 first_facet_dof = 0; 
       } 
	

     if(print) 
       {
	 (*testout) << "ndof hdivhofespace update = " << endl << ndof << endl;
	 if (!discont)
	   (*testout) << "first_facet_dof (hdiv)  = " << endl << first_facet_dof << endl;
	 (*testout) << "first_inner_dof (hdiv) = " << endl << first_inner_dof << endl;
       }
     
     while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
     ndlevel.Last() = ndof;
     //    prol->Update();
   }
  
  const FiniteElement & HDivHighOrderFESpaceSZ :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe;
    
    // typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
    // typedef TrigExtensionMonomial T_ORTHOPOL;
    

    
    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	{ 

	  fe = new (lh.Alloc (sizeof(HDivHighOrderTetSZ<>)))  HDivHighOrderTetSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
	/*
      case ET_PYRAMID:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderPyramidSZ<IntegratedLegendreMonomialExt>)))  HDivHighOrderPyramidSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
	*/
      case ET_PRISM:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderPrismSZ<IntegratedLegendreMonomialExt>)))  HDivHighOrderPrismSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
      case ET_HEX:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderHexSZ<IntegratedLegendreMonomialExt>)))  HDivHighOrderHexSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
      case ET_TRIG:
	{ 
	  fe = new (lh.Alloc (sizeof(HDivHighOrderTrigSZ<IntegratedLegendreMonomialExt>)))  HDivHighOrderTrigSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
      case ET_QUAD:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderQuadSZ<IntegratedLegendreMonomialExt>)))  HDivHighOrderQuadSZ<IntegratedLegendreMonomialExt> (order);
	  break;
	}
      default:
	fe = 0; 
      }
  
    if (!fe)
      {
	stringstream str;
	str << "HDivHighOrderFESpaceSZ " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }
    
    ARRAY<int> vnums;
    ma.GetElVertices(elnr, vnums);
    if ( ma.GetDimension() == 2)
      {
	HDivHighOrderFiniteElementSZ<2> * hofe =
	  dynamic_cast<HDivHighOrderFiniteElementSZ<2>*> (fe);
	ArrayMem<int, 12> ednums, order_ed;
	
	ma.GetElEdges(elnr, ednums);
	order_ed.SetSize (ednums.Size());
	
	for (int j = 0; j < ednums.Size(); j++)
	  order_ed[j] = order_facet[ednums[j]][0];
      
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (order_ed);
	hofe -> SetOrderInner (order_inner[elnr]);
	hofe -> SetOrderInnerCurl (order_inner_curl[elnr]);
	hofe -> ComputeNDof();
	hofe -> SetDiscontinuous(discont); 
      }
    else
      {

	HDivHighOrderFiniteElementSZ<3> * hofe =
	  dynamic_cast<HDivHighOrderFiniteElementSZ<3>*> (fe);
	ArrayMem<int, 6> fanums; 
	ArrayMem<INT<2>, 6> order_fa;
	ma.GetElFaces(elnr, fanums);
	order_fa.SetSize (fanums.Size());
	
	for (int j = 0; j < fanums.Size(); j++)
	  order_fa[j] = order_facet[fanums[j]];
	
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderFace (order_fa);
	hofe -> SetOrderInner (order_inner[elnr]);
	hofe -> SetOrderInnerCurl (order_inner_curl[elnr]);
	hofe -> ComputeNDof();
	hofe -> SetDiscontinuous(discont); 
      }

    return *fe;
  }

  const FiniteElement & HDivHighOrderFESpaceSZ :: GetSFE (int selnr, LocalHeap & lh) const
  {
    int i, j;

    FiniteElement * fe = 0;

    int porder; 
    if(discont) porder = -1; 
    else porder = order; 

    switch (ma.GetSElType(selnr))
      {
      case ET_SEGM:
	fe = new (lh.Alloc (sizeof(HDivHighOrderNormalSegmSZ<>)))  
     HDivHighOrderNormalSegmSZ<> (porder);
	break;
      case ET_TRIG:
	fe = new (lh.Alloc (sizeof(HDivHighOrderNormalTrigSZ<>)))  
	  HDivHighOrderNormalTrigSZ<> (porder);
	break; 
      case ET_QUAD:
	fe = new (lh.Alloc (sizeof(HDivHighOrderNormalQuadSZ<>)))  
	  HDivHighOrderNormalQuadSZ<> (porder);
	break; 
      }

    if (!fe)
      {
	stringstream str;
	str << "HDivHighOrderFESpaceSZ " << GetClassName()
	    << ", undefined eltype "
	    << ElementTopology::GetElementName(ma.GetSElType(selnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    if(discont) return *fe; 

    ArrayMem<int,4> vnums;
    ArrayMem<int, 4> ednums, order_ed;
    INT<3> order_fa;
    ma.GetSElVertices(selnr, vnums);
    
    if(ma.GetSElType(selnr) == ET_SEGM)
      {
	HDivHighOrderNormalFiniteElementSZ<1> * hofe =
	  dynamic_cast<HDivHighOrderNormalFiniteElementSZ<1>*> (fe);
	
	hofe -> SetVertexNumbers (vnums);
	ma.GetSElEdges(selnr, ednums);
	
	hofe -> SetOrderInner (order_facet[ednums[0]][0]);
	hofe -> ComputeNDof();
      }
    else
      {
	order_fa = INT<3>(order_facet[ma.GetSElFace(selnr)][0],order_facet[ma.GetSElFace(selnr)][1],0);
	
	HDivHighOrderNormalFiniteElementSZ<2> * hofe =
	  dynamic_cast<HDivHighOrderNormalFiniteElementSZ<2>*> (fe);
	
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderInner (order_fa);
	hofe -> ComputeNDof();
      }
    
    return *fe;
  }
  
  int HDivHighOrderFESpaceSZ :: GetNDof () const
  {
    return ndof;
  }

  int HDivHighOrderFESpaceSZ :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }

  void HDivHighOrderFESpaceSZ :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first,next;
    if(discont) 
      {
	// lowest_order included in inner 
      	first = first_inner_dof[elnr];
	next = first_inner_dof[elnr+1];
	for(int j=first; j<next; j++)
	  dnums.Append(j);
	return;
      } 

    
    ARRAY<int> fanums;
    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else
      ma.GetElEdges (elnr, fanums); 
      
    if(order < 0)
      throw Exception(" HDivHighOrderFESpaceSZ :: GetDofNrs() order < 0 ");

    
    
    if(ma.GetDimension() == 3)
      {
	//Ravier-Thomas
	for (int i = 0; i < fanums.Size(); i++)
	  dnums.Append (fanums[i]);
	// faces
	for(int i=0; i<fanums.Size(); i++)
	  {
	    first = first_facet_dof[fanums[i]];
	    next = first_facet_dof[fanums[i]+1];
	    for(int j=first ; j<next; j++)
	      dnums.Append(j);
	  }
	//inner
	first = first_inner_dof[elnr];
	next = first_inner_dof[elnr+1];
	for(int j=first; j<next; j++)
	  dnums.Append(j);
      }
    else
      {
	//Ravier-Thomas
	for (int i = 0; i < fanums.Size(); i++)
	  dnums.Append (fanums[i]);
	
	//edges
	for (int i = 0; i < fanums.Size(); i++)
	  {
	    first = first_facet_dof[fanums[i]];
	    next = first_facet_dof[fanums[i]+1];
	    
	    for (int j = first; j <next; j++)
	      dnums.Append (j);
	  }
       //inner
	first = first_inner_dof[elnr];
	next = first_inner_dof[elnr+1];
	for(int j=first; j<next; j++)
	  dnums.Append(j);
      }
    
  }

  void HDivHighOrderFESpaceSZ :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if(discont) return; 

    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    ARRAY<int> vnums, ednums, fnums;
    int i, j;
    int first,next;

    if(order < 0)
      throw Exception(" HDivHighOrderFESpaceSZ :: GetDofNrs() order < 0 ");

    dnums.SetSize(0);
    
    if(ma.GetDimension() == 3)
      {
	ma.GetElFaces (elnr, fnums);

	//Ravier-Thomas
	for (i = 0; i < fnums.Size(); i++)
	  dnums.Append (fnums[i]);

	// faces
	for(i=0; i<fnums.Size(); i++)
	  {
	    first = first_facet_dof[fnums[i]];
	    next = first_facet_dof[fnums[i]+1];
	    for(j=first ; j<next; j++)
	      dnums.Append(j);
	  }
      }
    else
      {
	ma.GetElEdges (elnr, ednums);
	
	//Ravier-Thomas
	for (i = 0; i < ednums.Size(); i++)
	  dnums.Append (ednums[i]);
	
	//Edges_normal
	for (i = 0; i < ednums.Size(); i++)
	  {
	    first = first_facet_dof[ednums[i]];
	    next = first_facet_dof[ednums[i]+1];
	    
	    for (j = first; j <next; j++)
	      dnums.Append (j);
	  }
      }

    
  }

  void HDivHighOrderFESpaceSZ :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if (discont) return; 

    ARRAY<int> vnums, ednums, eorient;
    int fnum,forient;
    int i, j;
    int first,next;

    ma.GetSElEdges (selnr, ednums, eorient);
    ma.GetSElFace (selnr, fnum, forient);

 

    if(order <0) throw (" HDivHighOrderFESpaceSZ :: GetSDofNrs() order < 0 ");

    if (ma.GetDimension() == 2)
      {
	// RT_0
	for (i = 0; i < ednums.Size(); i++)
          dnums.Append (ednums[i]);
	//edges_normal
	for (i = 0; i < ednums.Size(); i++)
	  {
	    first = first_facet_dof[ednums[i]];
	    next = first_facet_dof[ednums[i]+1];
	    for (j = first; j <next; j++)
	      dnums.Append (j);
	  }
      }
    else
      {
        dnums.Append (fnum);

        if(first_facet_dof.Size()>1)
	  {
	    first = first_facet_dof[fnum];
	    next = first_facet_dof[fnum+1];
	    for(j=first; j<next; j++)
	      dnums.Append(j);
	  }
      }
    // (*testout)<<"SElNr= "<<selnr<<endl;
    //(*testout)<<"SDofNr= "<<dnums<<endl;
  }


  Table<int> * HDivHighOrderFESpaceSZ ::
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, first;
    int ncnt;

    int ii;
    int SmoothingType=1;

    cout << "SmoothingType " << SmoothingType << endl;

     ArrayMem<int,12> vnums, ednums, fnums, orient;

/*
    for(i=0; i< nel ; i++)
      ma.GetElVertices (i, vnums);


    for(i=0; i< nel; i++)
      ma.GetElEdges (i,vnums,orient);

    int pn1,pn2;
    for(i=0; i< ned; i++)
      ma.GetEdgePNums (i,pn1,pn2);*/

    switch(SmoothingType)
      {
      case 1:  // RT_0 - Faeod - Iaecd - block  (RT_0 - Eaeod - Iaecd - block for 2D)
        if (ma.GetDimension() == 3)
        {
	   ncnt = 1 + nfa;
	   if (!eliminate_internal)
	      ncnt += nel;
	}
	else
	{
           ncnt = 1 + ned;
	   if (!eliminate_internal)
	      ncnt += nel;
	}
	break;
	case 2:  // RT_0 - ElFaces -I  -block ( - block for 2D)
        if (ma.GetDimension() == 3)
        {
	   ncnt = 1 + nel;
	   if (!eliminate_internal)
	      ncnt += nel;
	}
	else
	{
           ncnt = 1 + ned;
	   if (!eliminate_internal)
	      ncnt += nel;
	}
	break;
      }

      
    ARRAY<int> cnt(ncnt);
    cnt = 0;

    cout << " ncnt " << ncnt << endl;

    switch(SmoothingType)
    {
      case 1:
      // RT_0 - Faeod - Iaecd - block  (RT_0 - Eaeod - Iaecd - block for 2D)
      if (ma.GetDimension() == 3)
      {
         int i, j, first;
         cnt[0] = nfa;
         for (i = 0; i < nfa; i++)
	   cnt[i+1] = first_facet_dof[i+1]-first_facet_dof[i];
           //cnt[i+1] = 1+first_facet_dof[i+1]-first_facet_dof[i];


	 if (!eliminate_internal)
	   for (i = 0; i < nel; i++)
	      cnt[nfa+1+i] = first_inner_dof[i+1]-first_inner_dof[i];


	 Table<int> & table = *new Table<int> (cnt);
         cnt = 0;
         for (i = 0; i < nfa; i++)
	     table[0][i] = i; //RT_0
         for (i = 0; i < nfa; i++)
	 {
	     //table[i+1][0] = i;
	     //cnt[i+1] = 1;
	     for (j = first_facet_dof[i]; j < first_facet_dof[i+1]; j++)
	        table[i+1][cnt[i+1]++] = j;
	 }
         if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	    {
               for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
                  table[nfa+1+i][cnt[nfa+1+i]++] = j;

	    }
	(*testout) << "smoothingblocks = " << endl << table << endl;
       return &table;

     }
     else
     {
        int i, j, first;
         cnt[0] = ned;
         for (i = 0; i < ned; i++)
	   cnt[i+1] = 1+first_facet_dof[i+1]-first_facet_dof[i];
	 if (!eliminate_internal)
	   for (i = 0; i < nel; i++)
	      cnt[nfa+1+i] = first_inner_dof[i+1]-first_inner_dof[i];

	 Table<int> & table = *new Table<int> (cnt);
         cnt = 0;
         for (i = 0; i < ned; i++)
	     table[0][i] = i; //RT_0
         for (i = 0; i < ned; i++)
	 {
	     table[i+1][0] = i;
	     cnt[i+1] = 1;
	     for (j = first_facet_dof[i]; j < first_facet_dof[i+1]; j++)
	        table[i+1][cnt[i+1]++] = j;
	 }
         if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	    {
               for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	           table[ned+1+i][cnt[ned+1+i]++] = j;
	    }
	    (*testout) << "smoothingblocks = " << endl << table << endl;
	  return &table;
    }

      break;
    case 2:
      // // RT_0 - ElFaces -I  -block ( - block for 2D)
      if (ma.GetDimension() == 3)
      {
         //cout << "cnt = " <<endl;
         int i, j, k, first;
         cnt[0] = nfa;
         for (i = 0; i < nel; i++)
	 {
	    ma.GetElFaces(i,fnums,orient);
	    //cout << "fnums.Size = " <<fnums.Size()<<endl;
	    for (j = 0; j < fnums.Size(); j++)
	    {
               // cout << "firstfacedof = " <<first_facet_dof[fnums[j]]<<endl;
	       cnt[i+1] += first_facet_dof[fnums[j]+1]-first_facet_dof[fnums[j]];
	    }
	 }   
	    //cout << "cnt = " <<endl;
           //cnt[i+1] = 1+first_facet_dof[i+1]-first_facet_dof[i];

         if (!eliminate_internal)
	   for (i = 0; i < nel; i++)
	      cnt[nel+1+i] = first_inner_dof[i+1]-first_inner_dof[i];

         // cout << "cnt = " << endl << cnt << endl;
	 Table<int> & table = *new Table<int> (cnt);
         cnt = 0;
         for (i = 0; i < nfa; i++)
	     table[0][i] = i; //RT_0
         for (i = 0; i < nel; i++)
	 {
             ma.GetElFaces(i,fnums,orient);
	     for (j = 0; j < fnums.Size(); j++)
	        for (k = first_facet_dof[fnums[j]]; k < first_facet_dof[fnums[j]+1]; k++)
	         table[i+1][cnt[i+1]++] = k;
         }
         if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	    {
               for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
                  table[nel+1+i][cnt[nel+1+i]++] = j;

	    }
	(*testout) << "smoothingblocks = " << endl << table << endl;
       return &table;

     }
     else
     {
       ;
     }

      break;

   }
    return 0;
}



  /// 
  void HDivHighOrderFESpaceSZ :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  /// 
  void HDivHighOrderFESpaceSZ :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { 
    dnums.SetSize(0);
    if ( ma.GetDimension() == 2 )
      {
	dnums.Append (ednr);
	
	//Edges_normal
	int first = first_facet_dof[ednr];
	int next = first_facet_dof[ednr+1];
	for (int j = first; j <next; j++)
	  dnums.Append (j);
      }
  }
  /// 
  void HDivHighOrderFESpaceSZ :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    if(ma.GetDimension() == 3)
      {
	//Ravier-Thomas
	dnums.Append (fanr);
	
	// faces
	int first = first_facet_dof[fanr];
	int next = first_facet_dof[fanr+1];
	for(int j=first ; j<next; j++)
	  dnums.Append(j);
      }

  }  
  /// 
  void HDivHighOrderFESpaceSZ :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_inner_dof[elnr];
    int next = first_inner_dof[elnr+1];
    for(int j=first; j<next; j++)
      dnums.Append(j);

  }

  int HDivHighOrderFESpaceSZ :: GetFirstFacetDof(int fanr) const
  {
    int first;  
    // if(!discont) 
    first = (ma.GetDimension() == 2 ? first_facet_dof[fanr] : first_facet_dof[fanr]);  
      // else 
      
    return(first);  
  }; 


  // register FESpaces
  namespace
#ifdef MACOS
  hdivhofespace_cpp
#endif
  {

    class Init
    {
    public:
      Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hdivhosz", HDivHighOrderFESpaceSZ::Create);
    }

    Init init;
  }

}



