/*********************************************************************/
/* File:   l2hofespacesz.cpp                                         */
/* Author: Start                                                     */
/* Date:   24. Feb. 2003                                             */
/*********************************************************************/

/**
   High Order Finite Element Space for L2
*/


#include <comp.hpp>
#include <multigrid.hpp>

using namespace ngmg;

namespace ngcomp
{
  using namespace ngcomp;

  L2HighOrderFESpace ::  
  L2HighOrderFESpace (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama, flags)
  {
    var_order = 0; 
    if (flags.NumFlagDefined("order"))
      order =  int (flags.GetNumFlag("order",0));
    else 
      {
	if(flags.NumFlagDefined("relorder"))
	  {
	    order=0; 
	    var_order = 1; 
	    rel_order = int (flags.GetNumFlag("relorder",0));
	  }
	else 
	  order = 0; 
      }
    
    if(flags.GetDefineFlag("variableorder") )
      {
	throw Exception ("Flag 'variableorder' for l2ho is obsolete. \n  Either choose uniform order by -order= .. \n -relorder=.. for relative mesh order "); 
      }
    
    if (ma.GetDimension() == 2)
      {
	evaluator = 
	  new MassIntegrator<2> (new ConstantCoefficientFunction(1));
      }
    else
      {
	evaluator = 
	  new MassIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      {
	evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
      }

    low_order_space = new  ElementFESpace (ma, 0, dimension, iscomplex);
    prol = new ElementProlongation (*static_cast<ElementFESpace*> (low_order_space));
  }


  L2HighOrderFESpaceSZ :: ~L2HighOrderFESpaceSZ ()
  {
    ;
  }


  FESpace * L2HighOrderFESpaceSZ :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    int order = int(flags.GetNumFlag ("order", 0));
    int dim = int(flags.GetNumFlag ("dim", 1));
    bool iscomplex = flags.GetDefineFlag ("complex");

    if (order == 0)
      return new ElementFESpace (ma, order, dim, iscomplex);
    else
      return new L2HighOrderFESpaceSZ (ma, flags);
  }  

  void L2HighOrderFESpaceSZ :: Update()
  {
    low_order_space -> Update();
                  
    nel = ma.GetNE();
    order_inner.SetSize(nel); 
 
 
    int p  = (var_order ? 0 : order); 
    order_inner = INT<3>(p,p,p); 
    
    int dim = ma.GetDimension(); 

    if(var_order) 
      for(int i=0;i<nel;i++) 
	{
	  INT<3> el_orders = ma.GetElOrders(i);
	  for(int j=0;j<dim;j++)
	    el_orders[j] = int(max2(el_orders[j]+rel_order,0)); 
	  
	  for(int j=0;j<dim;j++)
	    order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
	} 

    // *testout << " order_inner (l2ho) " << order_inner << endl; 
    ndof = nel;
    
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	INT<3> pi = order_inner[i]; 
	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    ndof += (pi[0]+1)*(pi[0]+2)/2 ;
	    break;
	  case ET_QUAD:
	    ndof += (pi[0]+1)*(pi[1]+1);
	    break;
	  case ET_TET:
	    ndof += (pi[0]+1)*(pi[0]+2)*(pi[0]+3)/6;
	    break;
	  case ET_PRISM:
	    ndof += (pi[0]+1)*(pi[0]+2)*(pi[2]+1)/2;
	    break;
	  case ET_PYRAMID:
	    ndof += 5;
	    break;
	  case ET_HEX:
	    ndof += (pi[0]+1)*(pi[1]+1)*(pi[2]+1);
	    break;
	  }
	ndof--; // subtract constant 
      }
    first_element_dof[nel] = ndof;

    // *testout << " first_element dof (l2hofesz) " << first_element_dof << endl;  

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;

    prol->Update();
  }

  const FiniteElement & L2HighOrderFESpaceSZ :: GetFE (int elnr, LocalHeap & lh) const
  {
    try
      { 
	FiniteElement * fe;
	
	typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
	
	switch (ma.GetElType(elnr))
	  {
	  case ET_TET:
	    { 
	      fe = new (lh.Alloc (sizeof(H1HighOrderTet<T_ORTHOPOL>)))  H1HighOrderTet<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_PYRAMID:
	    {
	      fe = new (lh.Alloc (sizeof(H1HighOrderPyramid<T_ORTHOPOL>)))  H1HighOrderPyramid<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_PRISM:
	    {
	      fe = new (lh.Alloc (sizeof(H1HighOrderPrism<T_ORTHOPOL>)))  H1HighOrderPrism<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_HEX:
	    {
	      fe = new (lh.Alloc (sizeof(H1HighOrderHex<T_ORTHOPOL>)))  H1HighOrderHex<T_ORTHOPOL> (order);
	      break;
	    }
	  case ET_TRIG:
	    { 
	      fe = new (lh.Alloc (sizeof(L2HighOrderTrig)))  L2HighOrderTrig(order);
	      break;
	    }
	  case ET_QUAD:
	    {
	      fe = new (lh.Alloc (sizeof(L2HighOrderQuad)))  L2HighOrderQuad(order);
	      break;
	    }
	  default:
	    fe = 0; 
	  }
	
	if (!fe)
	  {
	    stringstream str;
	    str << "L2HighOrderFESpaceSZ " << GetClassName() 
		<< ", undefined eltype " 
		<< ElementTopology::GetElementName(ma.GetElType(elnr))
		<< ", order = " << order << endl;
	    throw Exception (str.str());
	  }
	

	ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
	ma.GetElVertices(elnr, vnums);
	if( ma.GetElType(elnr) == ET_TRIG || ma.GetElType(elnr) == ET_QUAD)
	  { 
	    L2HighOrderFiniteElement * hofe = dynamic_cast<L2HighOrderFiniteElement* > (fe);
	    //hofe-> SetVertexNumbers (vnums); 
	    hofe-> SetOrderInner(order_inner[elnr]); 
	    hofe-> ComputeNDof(); 
	  } 
	else 
	  {
	    if(var_order) 
	      throw Exception("L2HighOrderFESpace not working for 3D and variable order!"); 
	    
	    H1HighOrderFiniteElement * hofe = dynamic_cast<H1HighOrderFiniteElement* > (fe);
	    hofe-> SetVertexNumbers (vnums); 
	    // ATTENTION for var_order this is not correct !!! 
	    // on edges and faces order is still used !!! 
	    hofe-> SetOrderInner(order_inner[elnr]); 
	    hofe-> ComputeNDof(); 
	  } 
	return *fe;
      } 
    catch (Exception & e)
      {
	e.Append ("in L2HoFESpace::GetElement");
	e.Append ("\n");
	throw; 
      }
  }
 
  const FiniteElement & L2HighOrderFESpaceSZ :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
	fe = new FE_SegmDummy; break;
      case ET_TRIG:
	fe = new FE_SegmDummy ; break;
      case ET_QUAD:
	fe = new FE_SegmDummy; break;
      default:
	fe = 0;
      }
    
    ARRAY<int> vnums;
    ma.GetSElVertices(elnr, vnums);

    if (!fe)
      {
	stringstream str;
	str << "FESpace " << GetClassName() 
	    << ", undefined surface eltype " << ma.GetSElType(elnr) 
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }
    /*
    ArrayMem<int, 8> vnums;
    hofe -> SetOrderInner (0); 
    */
    /*
    dynamic_cast<H1HighOrderFiniteElement*> (fe)
      -> SetVertexNumbers (vnums);
    */
    return *fe;
  }
 

  int L2HighOrderFESpaceSZ :: GetNDof () const
  {
    return ndof;
  }

  int L2HighOrderFESpaceSZ :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }


  void L2HighOrderFESpaceSZ :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    dnums.Append (elnr); // lowest_order 
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;

    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;

    // (*testout) << "l2ho, dnums = " << dnums << endl;
  }
  
  void L2HighOrderFESpaceSZ :: 
  GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
  
  Table<int> * L2HighOrderFESpaceSZ :: 
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, first;
    ARRAY<int> cnt(nel);
    cnt = 0;
    for (i = 0; i < nel; i++)
      cnt[i] = first_element_dof[i+1]-first_element_dof[i];
	
    Table<int> & table = *new Table<int> (cnt);
    
    for (i = 0; i < nel; i++)
      {
	first = first_element_dof[i];
	for (j = 0; j < cnt[i]; j++)
	  table[i][j] = first+j;
      }
    return &table;
  }

  void  L2HighOrderFESpaceSZ :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpaceSZ ::GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpaceSZ ::GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2HighOrderFESpaceSZ ::GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs ( elnr, dnums ); return; }

  L2SurfaceHighOrderFESpaceSZ ::  
  L2SurfaceHighOrderFESpaceSZ (const MeshAccess & ama, const Flags & flags)
    : FESpace (ama, flags)
  {
    if(flags.NumFlagDefined("relorder"))
      throw Exception("Variable order not implemented for L2SurfaceHighOrderFESpace"); 
    
    segm = new H1HighOrderSegm<IntegratedLegendreMonomialExt> (order);
    trig = new H1HighOrderTrig<IntegratedLegendreMonomialExt> (order);
    quad = new H1HighOrderQuad<IntegratedLegendreMonomialExt> (order);

    if (ma.GetDimension() == 2)
      {
	boundary_evaluator = 
	  new RobinIntegrator<2> (new ConstantCoefficientFunction(1));
      }
    else
      {
	boundary_evaluator = 
	  new RobinIntegrator<3> (new ConstantCoefficientFunction(1));
      }

    if (dimension > 1)
      {
	boundary_evaluator = 
	  new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
      }
  }

  L2SurfaceHighOrderFESpaceSZ :: ~L2SurfaceHighOrderFESpaceSZ ()
  {
    ;
  }

  FESpace * L2SurfaceHighOrderFESpaceSZ :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    return new L2SurfaceHighOrderFESpaceSZ (ma, flags);
  }  

  void L2SurfaceHighOrderFESpaceSZ :: Update()
  {
    nel = ma.GetNSE();

    ndof = 0;
    first_element_dof.SetSize(nel+1);
    for (int i = 0; i < nel; i++)
      {
	first_element_dof[i] = ndof;
	switch (ma.GetSElType(i))
	  {
	  case ET_SEGM:
	    ndof += order+1;
	    break;
	  case ET_TRIG:
	    ndof += (order+1)*(order+2)/2;
	    break;
	  case ET_QUAD:
	    ndof += (order+1)*(order+1);
	    break;
	  default:
	    ;
	  }
      }
    first_element_dof[nel] = ndof;
  }

  const FiniteElement & L2SurfaceHighOrderFESpaceSZ :: GetSFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(elnr))
      {
      case ET_SEGM:
	fe = segm; break;
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0;
      }
    
    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetSElVertices(elnr, vnums);

    if (!fe)
      {
	stringstream str;
	str << "L2SurfaceHighOrderFESpaceSZ " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    dynamic_cast<H1HighOrderFiniteElement*> (fe) -> SetVertexNumbers (vnums);

    return *fe;
  }
 
  const FiniteElement & L2SurfaceHighOrderFESpaceSZ :: GetFE (int elnr, LocalHeap & lh) const
  {
    throw Exception ("Volume elements not available for L2SurfaceHighOrderFESpaceSZ");
  }
 
  int L2SurfaceHighOrderFESpaceSZ :: GetNDof () const
  {
    return ndof;
  }

  void L2SurfaceHighOrderFESpaceSZ :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_element_dof[elnr];
    int neldofs = first_element_dof[elnr+1] - first;
    for (int j = 0; j < neldofs; j++)
      dnums.Append (first+j);

    if (!DefinedOn (ma.GetSElIndex (elnr)))
      dnums = -1;
  }
  
  void L2SurfaceHighOrderFESpaceSZ :: 
  GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize (0);
  }
  
  Table<int> * L2SurfaceHighOrderFESpaceSZ :: 
  CreateSmoothingBlocks ( int type) const
  {
    int i, j, first;
    ARRAY<int> cnt(nel);
    cnt = 0;
    for (i = 0; i < nel; i++)
      cnt[i] = first_element_dof[i+1]-first_element_dof[i];
	
    Table<int> & table = *new Table<int> (cnt);
    
    for (i = 0; i < nel; i++)
      {
	first = first_element_dof[i];
	for (j = 0; j < cnt[i]; j++)
	  table[i][j] = first+j;
      }
    return &table;
  }


  void  L2SurfaceHighOrderFESpaceSZ :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2SurfaceHighOrderFESpaceSZ ::GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  
  void  L2SurfaceHighOrderFESpaceSZ ::GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { GetDofNrs ( fanr, dnums ); return; }
  
  void  L2SurfaceHighOrderFESpaceSZ ::GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs ( elnr, dnums ); return; }
    
  // register FESpaces
  namespace
#ifdef MACOS
  l2hofespace_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("l2sz", L2HighOrderFESpaceSZ::Create);
      GetFESpaceClasses().AddFESpace ("l2surfsz", L2SurfaceHighOrderFESpaceSZ::Create);
    }
    
    Init init;
  }
}
