
#include <comp.hpp>
#include <fem.hpp> 

  #include "facetfespace.hpp"
  #include "facetfe.hpp"

    
namespace ngcomp
{
  using namespace ngfem ;

// ------------------------------------------------------------------------
  FacetFESpace ::  FacetFESpace (const MeshAccess & ama, const Flags & flags)
  : FESpace(ama, flags)
  {
    ndlevel.SetSize(0);
    Flags loflags;
    loflags.SetFlag("order",0.0);
    if (order != 0)
      low_order_space = new FacetFESpace(ma, loflags);
    else
      low_order_space = 0;
        
    if (flags.NumFlagDefined("order"))
      { 
	var_order=0; 
	order =  int (flags.GetNumFlag ("order",0));
      }
    else 
      {
	if(flags.NumFlagDefined("relorder"))
	  {
	    order=0; 
	    var_order=1; 
	    rel_order = int (flags.GetNumFlag ("relorder",-1));
	  }
	else 
	  {
	    var_order=0;
	    order = 0;  
	  }
      }
    
    // TODO: Evaluator for shape tester 
    static ConstantCoefficientFunction one(1);
    if (ma.GetDimension() == 2)
    {
      evaluator = new MassIntegrator<2> (&one);
      boundary_evaluator = 0;
    }
    else
    {
      evaluator = new MassIntegrator<3> (&one);
      boundary_evaluator = new RobinIntegrator<3> (&one);
    }

    if (dimension > 1)
    {
      evaluator = new BlockBilinearFormIntegrator (*evaluator, dimension);
      boundary_evaluator = 
          new BlockBilinearFormIntegrator (*boundary_evaluator, dimension);
    }
    // Update();
  }
  
// ------------------------------------------------------------------------
  FacetFESpace :: ~FacetFESpace ()
  {
    ;
  }

// ------------------------------------------------------------------------
  FESpace * FacetFESpace :: Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FacetFESpace (ma, flags);
  }

// ------------------------------------------------------------------------
  void FacetFESpace :: Update()
  {

    if (low_order_space)
      low_order_space -> Update();

    // number of facets
    nel = ma.GetNE();
    nfa = (ma.GetDimension() == 2 ? ma.GetNEdges() : ma.GetNFaces()); 
    
    int p = 0; 
    if(!var_order) p = order; 
    
    order_facet.SetSize(nfa);
    order_facet = INT<2> (p,p);
    fine_facet.SetSize(nfa);
    fine_facet = 0; 
    
    ARRAY<int> fanums;
    
    
    for (int i = 0; i < nel; i++)
      {
	INT<3> el_orders = ma.GetElOrders(i); 

	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const POINT3D * points = ElementTopology :: GetVertices (eltype);	
	
	if (ma.GetDimension() == 2)
	  {
	    ma.GetElEdges (i, fanums);
	    for (int j=0;j<fanums.Size();j++) 
	      fine_facet[fanums[j]] = 1; 
	    
	    if(var_order)
	      {
		const EDGE * edges = ElementTopology::GetEdges (eltype);
		for(int j=0; j<fanums.Size(); j++)
		  for(int k=0;k<2;k++)
		    if(points[edges[j][0]][k] != points[edges[j][1]][k])
		      { 
			order_facet[fanums[j]] = INT<2>(max(el_orders[k]+rel_order, order_facet[fanums[j]][0]),0);
			break; 
		      }
	      }
	  }
	else
	  {
	    ARRAY<int> elfaces,vnums;
	    ma.GetElFaces(i,elfaces);
	    for (int j=0;j<elfaces.Size();j++) fine_facet[elfaces[j]] = 1; 
	    
	    if(var_order) 
	      {
		ma.GetElVertices (i, vnums);
		const FACE * faces = ElementTopology::GetFaces (eltype);
		for(int j=0;j<elfaces.Size();j++)
		  {
		    if(faces[j][3]==-1) // trig  
		      {
			order_facet[elfaces[j]][0] = max(order_facet[elfaces[j]][0],el_orders[0]+rel_order);
			order_facet[elfaces[j]][1] = order_facet[elfaces[j]][0]; 
		      }
		    else //quad_face
		      {
			int fmax = 0;
			for(int k = 1; k < 4; k++) 
			  if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
					
			INT<2> f((fmax+3)%4,(fmax+1)%4); 
			if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
			
			// fmax > f[0] > f[1]
			// p[0] for direction fmax,f[0] 
			// p[1] for direction fmax,f[1] 
			for(int l=0;l<2;l++)
			  for(int k=0;k<3;k++)
			    if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			      {
				order_facet[elfaces[j]][l] = max(order_facet[elfaces[j]][l] + rel_order, 
								el_orders[k]);
				break; 
			      } 
			
		      }
		  }
	      }
	    
	  }
	
      }
    
    // distribute dofs 
    ncfa = 0; 
    ndof = nfa - ncfa; // low_order space
    
    first_facet_dof.SetSize(nfa+1); 
    first_facet_dof = nfa - ncfa;
      
    if (ma.GetDimension() == 2)
    {
      for (int i = ncfa; i < nfa; i++)
      {
        first_facet_dof[i] = ndof;
        ndof += order_facet[i][0];
      }
      first_facet_dof[nfa] = ndof;
    } // 2D
    else  // 3D
    {
      int inci = 0;
      INT<2> p; 
      ARRAY<int> pnums;
      for (int i=ncfa; i< nfa; i++)
      {
        p = order_facet[i];
        ma.GetFacePNums(i,pnums);

        switch(pnums.Size())
        {
          case 3: 
            inci = ((p[0]+1)*(p[0]+2))/2 - 1;
            break;
          case 4: 
            inci= (p[0]+1)*(p[1]+1) - 1;
            break;
        }
        first_facet_dof[i] = ndof;
        ndof+= inci;
      }
      first_facet_dof[nfa] = ndof;
    } // 3D

      //HERBERT: h1hofe code. is that correct ??
    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;
      
      //HERBERT: no prolongation so far       
      //prol->Update();
    
   *testout << "*** Update FAcetFESpace: General Information" << endl;
   *testout << " order facet (facet) " << order_facet << endl; 
   *testout << " first_facet_dof (facet)  " << first_facet_dof << endl; 
   /* 
    cout << "    ne = " << nel << ", nfa = " << nfa << ", ncfa=" << ncfa << ", ndof=" << ndof << endl;
    cout << "    order_facet = " << order_facet << endl;*/
  }



// ------------------------------------------------------------------------
  const FiniteElement & FacetFESpace :: GetFE (int elnr, LocalHeap & lh) const
  {
    FacetVolumeFiniteElement * fe;

    switch (ma.GetElType(elnr))
    {
      case ET_TRIG:
        fe = new (lh.Alloc (sizeof(FacetVolumeTrig)))  FacetVolumeTrig ();
        break;
      case ET_QUAD:
	*testout << " new alloc facet quad " << endl; 
        fe = new (lh.Alloc (sizeof(FacetVolumeQuad)))  FacetVolumeQuad ();
        break;
      case ET_TET:
        fe = new (lh.Alloc (sizeof(FacetVolumeTet)))  FacetVolumeTet ();
        break;
      case ET_PYRAMID:
        fe = new (lh.Alloc (sizeof(FacetVolumePyramid)))  FacetVolumePyramid ();
        break;
      case ET_PRISM:
        fe = new (lh.Alloc (sizeof(FacetVolumePrism)))  FacetVolumePrism ();
        break;
      case ET_HEX:
        fe = new (lh.Alloc (sizeof(FacetVolumeHex)))  FacetVolumeHex ();
        break;
    }

    if (!fe)
    {
      stringstream str;
      str << "FacetFESpace " << GetClassName() 
          << ", undefined eltype " 
          << ElementTopology::GetElementName(ma.GetElType(elnr))
          << ", order = " << order << endl;
      throw Exception (str.str());
    }

    ARRAY<int> vnums;
    ArrayMem<int, 6> fanums, order_fa;
    
    ma.GetElVertices(elnr, vnums);

    if (ma.GetDimension() == 2)
      ma.GetElEdges(elnr, fanums);
    else
      ma.GetElFaces(elnr, fanums);
    
    order_fa.SetSize(fanums.Size());
    for (int j = 0; j < fanums.Size(); j++)
      order_fa[j] = order_facet[fanums[j]][0]; //SZ not yet anisotropric
 
    fe -> SetVertexNumbers (vnums);
    fe -> SetOrder (order_fa);
    fe -> ComputeNDof();
    
    return *fe;
  }


// ------------------------------------------------------------------------
  const FiniteElement & FacetFESpace :: GetSFE (int selnr, LocalHeap & lh) const
  {

    FacetFacetFiniteElement * fe = 0;

    switch (ma.GetSElType(selnr))
    {
      case ET_SEGM:
        fe = new (lh.Alloc (sizeof(FacetFacetSegm)))  FacetFacetSegm ();
        break;
      case ET_TRIG:
        fe = new (lh.Alloc (sizeof(FacetFacetTrig)))  FacetFacetTrig ();
        break;
      case ET_QUAD:
        fe = new (lh.Alloc (sizeof(FacetFacetQuad)))  FacetFacetQuad ();
        break;
      default:
        fe = 0;
    }
     
    if (!fe)
    {
      stringstream str;
      str << "FacetFESpace " << GetClassName()
          << ", undefined eltype "
          << ElementTopology::GetElementName(ma.GetSElType(selnr))
          << ", order = " << order << endl;
      throw Exception (str.str());
    }
     
    ArrayMem<int,4> vnums;
    ArrayMem<int, 4> ednums;
    
    ma.GetSElVertices(selnr, vnums);
    switch (ma.GetSElType(selnr))
    {
      case ET_SEGM:
        fe -> SetVertexNumbers (vnums);
        ma.GetSElEdges(selnr, ednums);
        fe -> SetOrder (order_facet[ednums[0]][0]); 
        fe -> ComputeNDof();
        break;
      case ET_TRIG: 
      case ET_QUAD:
        fe -> SetVertexNumbers (vnums);
        fe -> SetOrder (order_facet[ma.GetSElFace(selnr)][0]);// SZ not yet anisotropic order for facet fe !!! 
        fe -> ComputeNDof();
        break;
    }
    return *fe;
  }



// ------------------------------------------------------------------------
  int FacetFESpace :: GetNDof () const
  {
    return ndof;
  }

// ------------------------------------------------------------------------
  int FacetFESpace :: GetNDofLevel (int level) const
  {
    return ndlevel[level];
  }



// ------------------------------------------------------------------------
  void FacetFESpace :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {

    ARRAY<int> fanums; // facet numbers
    int first,next;

    fanums.SetSize(0);
    dnums.SetSize(0);
    

    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fanums);
    else // dim=2
      ma.GetElEdges (elnr, fanums);
     
    for(int i=0; i<fanums.Size(); i++)
    {
      dnums.Append(fanums[i]-ncfa); // low_order
      first = first_facet_dof[fanums[i]];
      next = first_facet_dof[fanums[i]+1];
      for(int j=first ; j<next; j++)
        dnums.Append(j);
    }
    
    // old 
    /*
    if(ma.GetDimension() == 3)
    {
      ma.GetElFaces (elnr, fdnums);
      for(int i=0; i<fdnums.Size(); i++)
      {
        for (int j=0; j<fdnums.Size(); j++
        first = first_facet_dof[fdnums[i]];
        next = first_facet_dof[fdnums[i]+1];
        for(int j=first ; j<next; j++)
          dnums.Append(j);
      }
    }
    else // 2D
    {
      ma.GetElEdges (elnr, ednums); //neu
      for (int i = 0; i < ednums.Size(); i++)
      {
        first = first_facet_dof[ednums[i]];
        next = first_facet_dof[ednums[i]+1];

        for (int j = first; j <next; j++)
          dnums.Append (j);
      }
    }
    */
//     cout << "** GetDofNrs(" << elnr << "): " << dnums << endl;
  }


// ------------------------------------------------------------------------
//   void FacetFESpace :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
//   {
//     GetDofNrs(elnr, dnums);
//   }


// ------------------------------------------------------------------------
  void FacetFESpace :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    int first,next;

    dnums.SetSize(0);
    if (ma.GetDimension() == 2)
    {
      ArrayMem<int, 4> fanums;
      ma.GetSElEdges (selnr, fanums);
      dnums.Append(fanums[0]-ncfa);
      
      first = first_facet_dof[fanums[0]];
      next = first_facet_dof[fanums[0]+1];
      for (int j = first; j <next; j++)
        dnums.Append (j);
      
/*      for (int i = 0; i < fanums.Size(); i++)
      {
        dnums.Append(fanums[i]-ncfa);
      
        first = first_facet_dof[ednums[i]];
        next = first_facet_dof[ednums[i]+1];
        for (int j = first; j <next; j++)
          dnums.Append (j);
      }*/
    }
    else// 3D
    {
        int fnum = ma.GetSElFace(selnr);
        dnums.Append(fnum-ncfa);
        first = first_facet_dof[fnum];
        next = first_facet_dof[fnum+1];
        for(int j=first; j<next; j++)
          dnums.Append(j);
      }

/*      if(first_facet_dof.Size()>1)
      {
        int fnum = ma.GetSElFace(selnr);
        first = first_facet_dof[fnum];
        next = first_facet_dof[fnum+1];
        for(int j=first; j<next; j++)
          dnums.Append(j);
      }*/
    }



// ------------------------------------------------------------------------
  Table<int> * FacetFESpace :: CreateSmoothingBlocks (const Flags & precflags) const
  {
    int ncnt;

    // 1 x low order + faces/edges
    ncnt = nfa - ncfa;
      
    ARRAY<int> cnt(ncnt);
    cnt = 0;

    // setup table
    for (int i = ncfa; i < nfa; i++)
      cnt[i-ncfa] = 1 + first_facet_dof[i+1]-first_facet_dof[i];


    Table<int> & table = *new Table<int> (cnt);
    
    // face/edges
    int ii;
    for (int i = ncfa; i < nfa; i++)
    {
      table[i-ncfa][0] = i-ncfa;
      ii=1;
      for (int j = first_facet_dof[i]; j < first_facet_dof[i+1]; j++, ii++)
        table[i][ii] = j;
    }
      
    cout << "smoothingblocks = " << endl << table << endl;
    return &table;

  }


  
  ARRAY<int> * FacetFESpace :: CreateDirectSolverClusters (const Flags & precflags) const
  {
    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());

    clusters.SetSize(ndof);
    clusters = 0;
    
    return &clusters;
    
    //
  
    for (int i=0; i<nfa-ncfa; i++)
      clusters[i]  = 1;
  
    cout << "direct solver cluster = " << clusters << endl;
    return & clusters;
  }



  
  
// ------------------------------------------------------------------------
   // register FESpaces
  namespace facefespace_cpp
  {
    class Init
    {
      public:
        Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("facet", FacetFESpace::Create);
    }

    Init init;
  }

}


