/**
   High Order Finite Element Space for Facets
 */
 /* facetfespace.*pp
 * An fespace for functions living on the facets (=edges in 2D, faces in 3D)
 *   - the functions on different facets are independent (no continuity accross
 *     vertices ( or edges in 3D))
 *   - the functions on the facets are accessed via the elements
 *     (=FacetVolumeElements), which then can access their facets;
 *     see facetfe.hpp for details
 *   - the ordering of dofs is facet'wise
 *   - no low-order space
 *   - no prolongation so far (will probably never be needed)
 *   - some additional utility functions 
 */

#ifndef FACET_FESPACE_HPP
#define FACET_FESPACE_HPP


#include <fem.hpp>
#include <comp.hpp> 

namespace ngcomp {

class FacetFESpace : public FESpace 
{
public:
  
  // Level
  int level;
  // Number of Facets
  int nfa;
  // Number of coarse facets, number of fine facets;
  int ncfa;
  // Number of Elements
  int nel;
  
  ARRAY<int> first_facet_dof;
  
  // relative order to mesh-order
  int rel_order; 
  
  ARRAY<INT<2> > order_facet;
  ARRAY<bool> fine_facet;
  
  int ndof;
  ARRAY<int> ndlevel;
  bool var_order; 
  
public:
    ///
    FacetFESpace (const MeshAccess & ama, const Flags & flags);
    ///
    virtual ~FacetFESpace ();
    ///
    static FESpace * Create (const MeshAccess & ma, const Flags & flags);
    ///
    virtual string GetClassName () const
    {
      return "FacetFESpace";
    }
    ///
    virtual void Update();
    ///
    virtual int GetNDof () const;
    ///
    virtual int GetNDofLevel (int level) const;
    ///
    virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
    ///
    virtual const FiniteElement & GetSFE (int selnr, LocalHeap & lh) const; 
    ///
    virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
    ///
    virtual void GetFacetDofNrs (int felnr, ARRAY<int> & dnums) const
    {
      dnums.SetSize(0);
      for (int j=first_facet_dof[felnr]; j<first_facet_dof[felnr+1]; j++)
        dnums.Append(j);
    }
    ///
    virtual int GetNFacetDofs (int felnr) const { return (first_facet_dof[felnr+1]-first_facet_dof[felnr]); }
//    ///
//     virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
    ///
    virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
    ///
    virtual Table<int> * CreateSmoothingBlocks (const Flags & precflags) const;
    ///
    virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const;
     
      // some utility functions for convenience
    ///
    virtual void GetVertexNumbers(int elnr, ARRAY<int>& vnums) 
    { ma.GetElVertices(elnr, vnums); };
    ///
    virtual INT<2> GetFacetOrder(int fnr) 
    { return order_facet[fnr]; };
  
  virtual int GetFirstFacetDof(int fanr) const {return (first_facet_dof[fanr]);}; 
  virtual int LowOrderDof() const {return (1);}; 
  
    

};



} // namespace

#endif
