#ifndef FILE_H1HOFETPSZ2
#define FILE_H1HOFETPSZ2



/*********************************************************************/
/* File:   h1hofetp.hpp                                              */
/* Author: Start                                                     */
/* Date:   11 June 2007                                              */
/*********************************************************************/

/**
  High order finite elements for H^1  in full Tensor Product basis
*/


// #define JACOBI

namespace ngfem
{

class H1HighOrderTrigTPSZ : public H1HighOrderTrig<>
{
public:
  H1HighOrderTrigTPSZ (int aorder)
    : H1HighOrderTrig<> (aorder) {;}
  
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename TFA>  
  void T_CalcShape(Tx x, Ty y, TFA & shape) const; 

  void GetTrig2TensorMapping(FlatArray<int[3]> trig2tensor) const; 

  int GetTensor2SegmMapping(FlatArray<int[3] > trig2tensor, FlatArray<int> tensor2segm) const; 
};


class H1HighOrderTetTPSZ : public H1HighOrderTet<>
{
  int sort[4];

  FlatArray<int[4]> tet2tensor;
  FlatArray<int[3]> trig2tensor;
  FlatArray<int[2]> segm2tensor;
  FlatArray<int> map3dto2d;
  FlatArray<int> map2dto1d;

public:
  H1HighOrderTetTPSZ (int aorder) : H1HighOrderTet<> (aorder) { ; }

  /// builds tables 
  virtual void SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh);

  void BuildTables(LocalHeap & lh); 

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  int Sort (int i) const { return sort[i]; }

  FlatArray<int[4]> GetTet2TensorMapping () const { return tet2tensor; }
  FlatArray<int[3]> GetTrig2TensorMapping () const { return trig2tensor; }
  FlatArray<int[2]> GetSegm2TensorMapping () const { return segm2tensor; }
  FlatArray<int> Get3dTo2dMapping () const { return map3dto2d; }
  FlatArray<int> Get2dTo1dMapping () const { return map2dto1d; }



  // shape[i] = facx[i] * facy[3t2[i]] * facz[2t1[3t2[i]]]

  template <typename Tx, typename Tres>
  void CalcXFactor (Tx x, FlatArray<Tres> facx, LocalHeap & lh) const
  {

    FlatArray<Tx> powx(order+1, lh);
    FlatMatrix<Tx> polsx(order+1, order, lh);

    for (int i = 0; i <= order; i++)
      {
        Tx * hp = &polsx(i, 1);
#ifdef JACOBI
        JacobiPolynomial (order-2, 2*x-1, max(0,2*i-2), 0, hp);
#else
        LegendrePolynomial (order-2, 2*x-1, hp);
#endif
        for (int j = 1; j < order; j++)
          polsx(i,j) *= x;
        polsx(i,0) = 1.0;
      }


    Tx hv = 1.0;
    for (int i = 0; i <= order; i++)
      {
        powx[i] = hv;
        hv *= (1-x);
      }
    
    for (int jj = 0; jj < ndof; jj++)
      {
        int power = tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3];
        facx[jj] = polsx(power, tet2tensor[jj][0]) * powx[power];
      }
  }


  template <typename Ty, typename Tres>
  void CalcYFactor (Ty y, FlatArray<Tres> facy, LocalHeap & lh) const
  {
    FlatArray<Ty> powy(order+1, lh);
    FlatMatrix<Ty> polsy(order+1, order, lh);

    for (int i = 0; i <= order; i++)
      {
        Ty * hp = &polsy(i, 1);
#ifdef JACOBI
        JacobiPolynomial (order-2, 2*y-1, max(0,2*i-2), 0, hp);
#else
        LegendrePolynomial (order-2, 2*y-1, hp);
#endif
        for (int j = 1; j < order; j++)
          polsy(i,j) *= y;
        polsy(i,0) = 1.0;
      }

    Ty hv = 1.0;
    for (int i = 0; i <= order; i++)
      {
        powy[i] = hv;
        hv *= (1-y);
      }
    
    int nd2d = trig2tensor.Size();
    for (int jj = 0; jj < nd2d; jj++)
      {
        int power = trig2tensor[jj][1]+trig2tensor[jj][2];
        facy[jj] = polsy(power, trig2tensor[jj][0]) * powy[power];
      }
    
  }

  template <typename Tz, typename Tres>
  void CalcZFactor (Tz z, FlatArray<Tres> facz, LocalHeap & lh) const
  {
    FlatArray<Tz> polz(order+1, lh);
    Vec<2, Tz> polzz;
    
    Tz * hp1 = &polz[1];
    LegendrePolynomialMult (order-2, 2*z-1, z, hp1);
    polz[0] = 1;
    polz[order] = 0;

    polzz(0) = 1;
    polzz(1) = (1-z); 

    int nd1d = segm2tensor.Size();
    for (int j1d = 0; j1d < nd1d; j1d++)
      {
        int j3 = segm2tensor[j1d][0];
        int j4 = segm2tensor[j1d][1];

        facz[j1d] = polz[j3] * polzz(j4);
      }        
  }


};


class H1HighOrderPrismTPSZ : public H1HighOrderPrism<>
{
  int sort[6];

  FlatArray<int[4]> prism2tensor;
  FlatArray<double> fac_pr2tp; 
  FlatArray<int[3]> trig2tensor;

  FlatArray<int> map3dto2d;
  FlatArray<int> map2dto1d;

public:
  H1HighOrderPrismTPSZ (int aorder)
    : H1HighOrderPrism<> (aorder) { ; }

  virtual void SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh);

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  void GetPrism2TensorMapping (FlatArray<int[4]> prism2tensor, FlatArray<double> fac) const;
  int Sort (int i) const { return sort[i]; }

  FlatArray<int[4]> GetPrism2TensorMapping () const { return prism2tensor; }
  FlatArray<double> GetFactorPrism2TensorMapping () const { return fac_pr2tp; }
  FlatArray<int[3]> GetTrig2TensorMapping () const { return trig2tensor; }
  FlatArray<int> Get3dTo2dMapping () const { return map3dto2d; }
  FlatArray<int> Get2dTo1dMapping () const { return map2dto1d; }



  void BuildTables(LocalHeap & lh); 
  void BuildPrism2TensorMapping(); 


  template <typename Tz, typename Tres>
  void CalcZFactor (Tz z, FlatArray<Tres> facz, LocalHeap & lh) const
  {
    Tz xi = 2*z-1; 
    Tz * hp = &facz[2];
    LegendrePolynomialMult (order-2, xi, 0.25 * (1-xi*xi), hp);
    facz[0] = 1-z;
    facz[1] = z; 
  }

 template <typename Ty, typename Tres>
  void CalcYFactor (Ty y, FlatArray<Tres> facy, LocalHeap & lh) const
  {
    FlatArray<Ty> poly(order+1,lh); 
    FlatArray<Ty> powy(order+1,lh); 
        
    AutoDiff<1> * hp1 = &poly[1];
    LegendrePolynomialMult (order-2, 2*y-1, y, hp1);
    poly[0] = 1;
    
    FlatArray<Ty> polyy(2,lh); 
    polyy[0] = 1;
    polyy[1] = (1-y); 
          
    int ndof2d = trig2tensor.Size(); 
    for (int jj = 0; jj < ndof2d; jj++)
      facy[jj] = poly[trig2tensor[jj][0]] * polyy[trig2tensor[jj][1]];
      
  }

  template <typename Tx, typename Tres>
  void CalcXFactor (Tx x, FlatArray<Tres> facx, LocalHeap & lh) const
  {
    FlatArray<Tx> powx(order+1, lh);
    FlatMatrix<Tx> polsx(order+1, order, lh);
    
    for (int i = 0; i <= order; i++)
      {
	Tx * hp = &polsx(i, 1);
#ifdef JACOBI
	JacobiPolynomial (order-2, 2*x-1, max(0,2*i-2), 0, hp);
#else
	LegendrePolynomial (order-2, 2*x-1, hp);
#endif
	for (int j = 1; j < order; j++)
	  polsx(i,j) *= x;
	polsx(i,0) = 1.0;
      }
    
    Tx hv = 1.0;
    for (int i = 0; i <= order; i++)
      {
	powx[i] = hv;
	hv *= (1-x);
      }
    
    for (int jj = 0; jj < ndof; jj++)
      {
	/*
	  AutoDiff<1> fx = 
	  polx[prism2tensor[jj][0]] *
	  powx[prism2tensor[jj][1]+prism2tensor[jj][2]];
	*/
	int power = prism2tensor[jj][1]+prism2tensor[jj][2];
	facx[jj] = polsx(power, prism2tensor[jj][0]) * powx[power];
      }
  }

};



class H1HighOrderHexTPSZ : public H1HighOrderHex<>
{

public:
  H1HighOrderHexTPSZ (int aorder)
    : H1HighOrderHex<> (aorder) { ; }

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  void GetHex2TensorMapping (FlatArray<int[3]> hex2tensor, FlatArray<double> fac) const;
};

}
#endif




