#ifndef TP_TRAFO_HPP
#define TP_TRAFO_HPP

#include <ngstd.hpp> 
#include <fem.hpp> 

namespace tptransformation 
{
  //using namespace netgen; 
  using namespace ngstd;
  using namespace ngfem; 

  void QuickSort (const ARRAY<int> & values, ARRAY<int> & order); 



class TensorProductTransformation3D 
{
private:
  
  ELEMENT_TYPE eltype;
 
  ARRAY<int> sort; 
    
  Mat<3> trans_rot; 
  
  Mat<3> inv_jac;

  double invx, invxy; 

public:
  ///
   
  TensorProductTransformation3D(const ELEMENT_TYPE aeltype, const int * avnums)  
  { 
    eltype = aeltype; 

    int nv = ElementTopology :: GetNVertices(eltype); 
    sort.SetSize(nv); 
    ARRAY<int> vnums(nv); 

    // rotate coord. sys of tet acc. to global vertex-numbers 
    for(int l=0;l<nv; l++) 
      sort[l] = l; 
    for(int l=0;l<nv; l++) 
      vnums[l] = avnums[l]; 

    QuickSort(vnums,sort);

    /*  
       if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
       if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
       if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
       if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
       if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    */
   
    double dlamdx_tet[4][3] = {{1.,0.,0.},{0.,1.,0.},{0.,0.,1.},{-1.,-1.,-1.}}; 
    double dlamdx_prism[4][3] = {{1.,0.,0.},{0.,1.,0.},{-1.,-1.,0.},{0.,0.,1.}}; 
        
    switch(eltype)
      {
      case (ET_TET):
	for (int j=0;j<3;j++) 
	  for (int k=0;k<3;k++)
	    trans_rot(j,k) = dlamdx_tet[sort[j]][k]; 
	break; 
      case (ET_PRISM):
	{ 
	  for (int j=0;j<3;j++) 
	    for (int k=0;k<3;k++)
	      trans_rot(j,k) = dlamdx_prism[sort[j]][k]; 
	  inv_jac(2,0) = 0.; 
	  inv_jac(2,1) = 0.; 
	  inv_jac(2,2) = 1.;
	}
	break; 
      default : ;  
      } 
    
    inv_jac(0,0) = 1.; 
    inv_jac(0,1) = inv_jac(0,2) =  0.; 
    inv_jac(1,2) = 0.; 
    
  }

  inline Mat<3> CalcJacobianInverse(double * ip_hex, bool scaled = 0) 
  {
    
    invx = (scaled) ? 1. : 1./(1-ip_hex[0]); 
    
    switch(eltype)
      {
      case (ET_TET): 
	{
	  invxy = (scaled) ? 1. :  invx/(1-ip_hex[1]); 
	  // inv_jac(0,0) = 1;
	  // inv_jac(0,1) = 0;
	  // inv_jac(0,2) = 0;
	  
	  inv_jac(1,0) = ip_hex[1]*invx;
	  inv_jac(1,1) = invx;
	  // inv_jac(1,2) = 0;
	  
	  inv_jac(2,0) = ip_hex[2]*invxy;
	  inv_jac(2,1) = ip_hex[2]*invxy;
	  inv_jac(2,2) = invxy;
	  break; 
	}
      case (ET_PRISM): 
	{
	  // inv_jac(0,0) = 1;
	  // inv_jac(0,1) = 0;
	  // inv_jac(0,2) = 0;
	  
	  inv_jac(1,0) = ip_hex[1]*invx;
	  inv_jac(1,1) = invx;
	  // inv_jac(1,2) = 0;
	  
	  inv_jac(2,0) = 0.;
	  inv_jac(2,1) = 0.;
	  inv_jac(2,2) = 1;
	  break; 
	}
      }
    return(inv_jac*trans_rot); 
  }

  
  Mat<3> GetTransRot() 
  {
    return(trans_rot); 
  }
  Mat<3> GetInvJac() 
  {
    return(inv_jac); 
  }
  
  inline void SetX(double x, bool scaled = 0)
  { 
    invx = 1./(1-x); 
  } 

  inline void SetY(double y, bool scaled = 0) 
  {
    if(eltype == ET_PRISM || eltype == ET_TET)
      {
	inv_jac(1,0) = y*invx;
	inv_jac(1,1) = invx;
	if(eltype == ET_TET)
	  invxy = (scaled) ? 1. : invx/(1-y);
      }
  }

  inline void SetEtaForJacobianInverse(double * ip_hex, bool scaled = 0) 
  {
    if(eltype == ET_PRISM || eltype == ET_TRIG || eltype == ET_TET)
      {
	invx = (scaled) ? 1. : 1./(1-ip_hex[0]);
	inv_jac(1,0) = ip_hex[1]*invx;
	inv_jac(1,1) = invx;
        inv_jac(1,2) = 0;
      }
  } 

  inline void SetZ(double z) 
  {
    if(eltype == ET_TET) 
      {
	inv_jac(2,0) = z*invxy;
	inv_jac(2,1) = z*invxy;
	inv_jac(2,2) = invxy;
      }
    /* else if (eltype == ET_PRISM)  
      { 
	inv_jac(2,0) = 0.; 
	inv_jac(2,1) = 0.; 
	inv_jac(2,2) = 1.; 
	}*/ 
  } 

  inline Mat<3> GetJacobianInverse()  
  {
    return(inv_jac*trans_rot); 
  }



  double GetJacobiDet(double * ip_hex)
  { 
    switch(eltype)
      { 
      case ET_TET: 
	return (sqr(1-ip_hex[0])*(1-ip_hex[1])); 
      case ET_PRISM: 
	return ((1-ip_hex[0]));
      default: 
	return (1.); 
      } 
  }

  
  void GetElemPointOfTP(double * ip_hex, double * xr)
  {
    double lami[4];   // { x, y, z, 1-x-y-z };

    lami[sort[0]] = ip_hex[0];
    lami[sort[1]] = ip_hex[1] * (1-ip_hex[0]);
    
    switch (eltype) 
      {
      case ET_TET: 
	lami[sort[2]] = ip_hex[2] * (1-ip_hex[1]) * (1-ip_hex[0]);
	lami[sort[3]] = 1 - lami[sort[0]] - lami[sort[1]] - lami[sort[2]]; 
	for (int l=0; l<3; l++) xr[l] = lami[l]; 
	break; 
      case ET_PRISM: 
	lami[sort[2]] = 1 - lami[sort[0]] - lami[sort[1]]; 
	lami[2] = ip_hex[2]; 
	break; 
      } 
      
    for(int l=0;l<3;l++)
      xr[l] = lami[l];  
  }

  Vec<3> ElementPoint(double * ip_hex)
  {
    Vec<3> xr; 
    double lami[4];   // { x, y, z, 1-x-y-z };

    lami[sort[0]] = ip_hex[0];
    lami[sort[1]] = ip_hex[1] * (1-ip_hex[0]);
    
    switch (eltype) 
      {
      case ET_TET: 
	lami[sort[2]] = ip_hex[2] * (1-ip_hex[1]) * (1-ip_hex[0]);
	lami[sort[3]] = 1 - lami[sort[0]] - lami[sort[1]] - lami[sort[2]]; 
	break; 
      case ET_PRISM: 
	lami[sort[2]] = 1 - lami[sort[0]] - lami[sort[1]]; 
	lami[2] = ip_hex[2]; 
	break; 
      } 
      
    for(int l=0;l<3;l++)
      xr(l) = lami[l];  
  }

};


class TensorProductTransformation2D 
{
private:
  
  ELEMENT_TYPE eltype;
 
  ARRAY<int> sort; 
    
  Mat<2> trans_rot; 
  
  Mat<2> inv_jac;

  double invx;

public:
  ///
   
  TensorProductTransformation2D(const ELEMENT_TYPE aeltype, const int * avnums)  
  { 
    eltype = aeltype; 

    int nv = ElementTopology :: GetNVertices(eltype); 
    sort.SetSize(nv); 
    ARRAY<int> vnums(nv); 

    // rotate coord. sys of tet acc. to global vertex-numbers 
    for(int l=0;l<nv; l++) 
      sort[l] = l; 
    for(int l=0;l<nv; l++) 
      vnums[l] = avnums[l]; 

    QuickSort(vnums,sort);

    /* 
       if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
       if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
       if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
       if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
       if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    */
    
    double dlamdx_trig[3][2] = {{1.,0.},{0.,1.},{-1.,-1.}}; 	
    
    if(eltype == ET_TRIG)
      for (int j=0;j<2;j++) 
	for (int k=0;k<2;k++)
	  trans_rot(j,k) = dlamdx_trig[sort[j]][k]; 

    inv_jac(0,0) = 1; 
    inv_jac(0,1) = 0; 
  }

  inline Mat<2> GetJacobianInverse(double * ip_hex, bool scaled = 0) 
  {
    invx = (scaled) ? 1 : 1./(1-ip_hex[0]); 
    
    inv_jac(0,0) = 1;
    inv_jac(0,1) = 0;
    
    inv_jac(1,0) = ip_hex[1]*invx;
    inv_jac(1,1) = invx;
    
    return(inv_jac*trans_rot); 
  }

  
  Mat<2> GetTransRot() 
  {
    return(trans_rot); 
  }
  Mat<2> GetInvJac() 
  {
    return(inv_jac); 
  }
  
  inline void SetX(double * ip_hex, bool scaled = 0)
  { 
    invx = 1./(1-ip_hex[0]); 
    inv_jac(1,1) = invx;
  } 

  inline Mat<3> GetJacobianInverseY(double * ip_hex, bool scaled = 0) 
  {
    inv_jac(1,0) = ip_hex[1]*invx;
    
    return(inv_jac*trans_rot); 
  }

  double GetJacobiDet(double * ip_hex)
  { 
    return ((1-ip_hex[0]));
  }

  Vec<3> ElementPoint(double * ip_hex)
  {
    Vec<3> xr; 
    double lami[3];   // { x, y, z, 1-x-y-z };

    lami[sort[0]] = ip_hex[0];
    lami[sort[1]] = ip_hex[1] * (1-ip_hex[0]);
    lami[sort[2]] = 1 - lami[sort[0]] - lami[sort[1]]; 

    for(int l=0;l<3;l++)
      xr(l) = lami[l];  
  }

};


void QuickSortRec (const ARRAY<int> & values,
		  ARRAY<int> & order, 
		  int left, int right)
{
  int i, j;
  double midval;

  i = left;
  j = right;
  midval = values[order[(i+j)/2]];
  
  do
    {
      while (values[order[i]] < midval) i++;
      while (midval < values[order[j]]) j--;
      
      if (i <= j)
	{
	  Swap (order[i], order[j]);
	  i++; j--;
	}
    }
  while (i <= j);
  if (left < j) QuickSortRec (values, order, left, j);
  if (i < right) QuickSortRec (values, order, i, right);
}

void QuickSort (const ARRAY<int> & values,
	       ARRAY<int> & order)
{
  int i, n = values.Size();
  order.SetSize (n);
  for (i = 0; i < n; i++)
    order[i] = i;

  QuickSortRec (values, order, 0, order.Size()-1);
}

}





#endif
 
