#ifdef LAPACK

#include <solve.hpp>
// #include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;

  ///  
  class NumProcLOBPCG : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    /// 
    BilinearForm * bfm;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    /// 
    string filename, filenameh;



    ///
    int maxsteps;
    /// 

    int nev; 
    ///
    double prec, prec2;
    ///
    bool print;
    
  public:
    ///
    NumProcLOBPCG (PDE & apde, const Flags & flags);
    ///

    virtual ~NumProcLOBPCG();

   static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcLOBPCG (pde, flags);
    }

    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "LOBPCG for Eigenvalue Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl;
    }
  };



  NumProcLOBPCG :: NumProcLOBPCG (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    
    nev =  gfu->GetMultiDim(); 
    filename = flags.GetStringFlag ("filename","lobpcg.out");
    prec = double(flags.GetNumFlag ("prec", 1.e-5));
  }

  NumProcLOBPCG :: ~NumProcLOBPCG()
  {
    ;
  }


  void NumProcLOBPCG :: Do(LocalHeap & lh)
  { 
    cout << "solve EVP via LOBPCG" << endl;
    *testout << "solve EVP via LOBPCG" << endl;
    
    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseMatrix & matm = bfm->GetMatrix();
    BaseVector & vu = gfu->GetVector();
    const BaseMatrix & matpre = pre->GetMatrix();
       
    /*
      BaseMatrix * invmat; 
      invmat = dynamic_cast<const BaseSparseMatrix &> (mata).InverseMatrix(); 
    */  
 
    int dim = bfa->GetFESpace().GetDimension();
    int size = bfa->GetMatrix().Height();
    int ndofs = dim * size; 
    int sdofs = dim*(bfa->GetFESpace().GetMeshAccess().GetNSE())*bfa->GetFESpace().GetOrder(); 
    
    if (3*nev >= ndofs-sdofs)
      throw Exception("Multidim too big vs ndofs");
    
    BaseVector & vw = *mata.CreateVector();
    BaseVector & vd = *mata.CreateVector();
    BaseVector & vv = *mata.CreateVector();
    BaseVector & vmu = *mata.CreateVector();
    BaseVector & vau = *mata.CreateVector();
    
    cout.precision(12);
    (*testout).precision(12); 
    

    Vector<double> Lami(nev); 
    Vector<double> LamiOld(nev); 
     
    // lopcg=2: LOBPCG1  x_n+1 = alpha x_n + tau w_n + gamma p_n mit p_{n+1}^j = x_n+1^j - sum(tau^k x_n^k) 
        
    cout << " Using LOPGCG I" << endl; 

    int N = 3*nev; // but in the first step use std. preconditioned gradient type method  

    ARRAY<BaseVector*> w(N);

    ARRAY<BaseVector*> wh(nev);

    ARRAY<BaseVector*> AX(N), MX(N);          
    // Create vector w with initial vector X and precond. Residuum and p=0;  
    // w = [X,C^[-1](AX- Lami MX),P]
    for(int m=0; m<N;m++)
      {
	w[m]=mata.CreateVector();
	AX[m]=mata.CreateVector();
	MX[m]=mata.CreateVector();
      }
    
    bool complex = bfa->GetFESpace().IsComplex(); 
    for(int m=0;m<nev; m++) 
      {
	wh[m] = mata.CreateVector();
	double norm; 
	if(complex)
	  {

	    FlatVector<Complex> fv = w[m]->FVComplex();
	    FlatVector<Complex> fv2 = wh[m]->FVComplex();
	    
	    time_t start=clock();
	    int sstart = start; 
	    srand(m*m*m*m*m*start); 
	    for (int i = 0; i < fv.Size(); i++)
	      {
		
		double r1 = double(rand())/double(RAND_MAX);
		double r2 = double(rand())/double(RAND_MAX);
		fv(i) = Complex(r1,r2); 
		fv2(i) = Conj(fv(i)); 
	      }
	    
	  
	    norm = sqrt(real(InnerProduct(fv,Conj(fv))));
	  
	    norm = sqrt(real(InnerProduct(fv,fv2)));
	  
	    fv *= norm; 



	  }
	else 
	  {
	    FlatVector<double> fv = w[m]->FVDouble();
	    w[m]->SetRandom(); 
	    norm = sqrt((InnerProduct(fv,fv)));
	    fv *= norm; 
	  } 
	
	

      }
    
    



    // P^0 = 0  // can be skipped 
    for (int m=2*nev;m<N;m++) 
      *w[m]=0.;  
        
    // since first iterate is std. prec. gradient method 
    N -= nev; 
    // Improvement to implement (vg Knyazev Matlab Code) 
    // vmu[nev,nev,nev] und vau[nev,nev,nev] speichern 
    // nach preconditioning nur w-Teil Aw und Mw neu berechnen 
    // in X^T AX keine neue Anwendung von AX,MX 
    // ARRAY<BaseVector*> AV(N), MV(N); 
    

    if(!(bfa->GetFESpace().IsComplex()))
      { 

	for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
	  {
	    cout << endl << "iteration i = " << ii << endl;
	    for(int m=0;m<nev;m++)
	      {
		//	    *MX[m] = matm * (*w[m]); 	  		
		// *AX[m] = mata * (*w[m]);
		
		vmu= matm * (*w[m]); 	  		
		vau = mata * (*w[m]);
		if (ii ==1) 
		LamiOld(m) = InnerProduct(vau,*w[m])/InnerProduct(vmu,*w[m]); 
		// LamiOld(m) = InnerProduct(*AX[m],(*w[m]))/InnerProduct(*MX[m],(*w[m])); 
		//  InnerProduct(*AX[m],(*w[m]))/InnerProduct(*MX[m],(*w[m]));    
		
		//vd = *AX[m]-LamiOld(m)*(*MX[m]);
		vd = (vau-LamiOld(m)*vmu); 
		// *w[nev+m] = matpre*vd; 
		*w[nev+m] = matpre*vd;
	      
	      }
	
	Vector<double> Lami2(N);     
	Matrix<double> asmall(N,N); 
	Matrix<double> msmall(N,N); 
	Matrix<double> asmall1(N,N); 
	Matrix<double> msmall1(N,N); 
	
	

	// [X,W,P]^T A [X,W,P] y = lami [X,W,P]^t M [X,W,P] y 
	/* for(int j=0;j<nev;j++)
	  for(int k=0;k<N;k++) 
	    asmall(j,k) = InnerProduct(*AX[j],*w[k]); 
	
	for(int j=0;j<nev;j++)
	  for(int k=0;k<N;k++) 
	    msmall(j,k) = InnerProduct(*MX[j],*w[k]); 
	*/ 
	for(int j=0;j<N;j++)
	  {
	    vmu = matm * (*w[j]);
	    for(int k=0;k<N;k++) 
	      msmall(j,k) = InnerProduct(vmu,*w[k]);  	  		
	  }
	for(int j=0;j<N;j++)
	  {
	    vau = mata * (*w[j]);  
	    for(int k=0;k<N;k++) 
	      asmall(j,k) = InnerProduct(vau,*w[k]); 
	  }
	
	/*
	  {
	  vmu = matm * (*w[j]);
	  for(int k=0;k<N;k++) 
	  msmall(j,k) = InnerProduct(vmu,*w[k]);  	  		
	  }
	  for(int j=0;j<N;j++)
	  {
	  vau = mata * (*w[j]);  
	  for(int k=0;k<N;k++) 
	  asmall(j,k) = InnerProduct(vau,*w[k]);
	  }
	*/
	  
	

	int info = LapackGHEPEPairs(N, &asmall(0,0), &msmall(0,0), &Lami2(0));
	if(info>0)
	  {
	    throw Exception ("Error in zhegv_ LapackGHEPPairs() ");
	  }


	for(int m=0;m<nev;m++) 
	  Lami(m) = Lami2(m); 
	
	cout << " Lami " << Lami << endl; 


	// use only the first nev eigenvectors and write to gfu
	// gfu = Xy y.. i-th line of a small 
	

	
	for(int m=0;m<nev;m++)
	  {
	    *wh[m]=0.; 
	    for(int k=0;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }

	
	// new p^{j+1}_k = [0,w_j, p_j] v[k]  
	for(int m=0;m<nev;m++)
	  {
	    *w[m] = *wh[m]; 
	    *wh[m] = 0.;  
	    for(int k=nev;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }

	for(int m=0;m<nev;m++)
	  *w[m+2*nev] = *wh[m]; 
	
	// after first iteration use LOBPCG  
	N = 3*nev;
	double err=0.;
	for(int l=0;l<nev/2;l++)
	  err += (LamiOld(l)-Lami(l))*(LamiOld(l)-Lami(l));
	
	*testout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  *testout << " Lami(" << m << ") = " << Lami(m) << endl; 

	cout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  cout << " Lami(" << m << ") = " << Lami(m) << endl; 
		
	cout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	*testout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	if(sqrt(err)/double(nev/2)<prec) break; 

	LamiOld = Lami;

      }
      }
    else
      {
	cout << " complex " << endl;
	for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
	  {
	    cout << endl << "iteration i = " << ii << endl;
	    for(int m=0;m<nev;m++)
	      {
		//		*MX[m] = matm * (*w[m]); 	  		
		// *AX[m] = mata * (*w[m]); 

		vmu = matm * (*w[m]); 	  		
		vau = mata * (*w[m]); 
		
		//FlatVector<Complex> fvmu = MX[m]->FVComplex(); 
		//FlatVector<Complex> fvau = AX[m]->FVComplex(); 
		FlatVector<Complex> fvmu = vmu.FVComplex();
		FlatVector<Complex> fvau = vau.FVComplex(); 
		FlatVector<Complex> v = w[m]->FVComplex(); 
 		
		if(ii == 1) 
		  LamiOld(m) = real(InnerProduct(fvau,Conj(v))/real(InnerProduct(fvmu,Conj(v)))); 
		
		
		// vd = (*AX[m])-LamiOld(m)*(*MX[m]); 
		vd = (vau)-LamiOld(m)*(vmu); 
		*w[nev+m] = matpre*vd; 
		//*w[nev+m] = (*invmat)*vd; 
	      }
	    
	    Vector<double> Lami2(N);     
	    Matrix<Complex> asmall(N,N); 
	    Matrix<Complex> msmall(N,N); 
	    Matrix<Complex> asmall1(N,N); 
	    Matrix<Complex> msmall1(N,N); 
	    
    
	    // [X,W,P]^T A [X,W,P] y = lami [X,W,P]^t M [X,W,P] y 
	    /*	    
	       for(int j=0;j<nev;j++)
	      {
		FlatVector<Complex> fvmu = MX[j]->FVComplex(); 
		FlatVector<Complex> fvau = AX[j]->FVComplex();
		
		
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    msmall(j,k) = InnerProduct(fvmu,Conj(fv));  	  		
		  } 
		
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    asmall(j,k) = InnerProduct(fvau,Conj(fv));  	  		
		  } 
		  } 
	    */ 
		
	    for(int j=0;j<N;j++) 
	      {
		vmu = matm * (*w[j]); 
		FlatVector<Complex> fvmu = vmu.FVComplex(); 
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    msmall(j,k) = InnerProduct(fvmu,Conj(fv));  	  		
		  } 
		
		vau = mata * (*w[j]); 
		FlatVector<Complex> fvau = vau.FVComplex(); 
		for(int k=0;k<N;k++) 
		  { 
		    FlatVector<Complex> fv = w[k]->FVComplex(); 
		    asmall(j,k) = InnerProduct(fvau,Conj(fv));  	  		
		  } 
	      } 
	    
	  
	    int info = LapackGHEPEPairs(N, &asmall(0,0), &msmall(0,0), &Lami2(0)); 
	    
	  
	    
	    if(info>0)
	      {
		cout << " returned info " << info << endl; 
		throw Exception ("Error in zhegv_ LapackGHEPPairs() ");
	      }	
	    
	for(int m=0;m<nev;m++) 
	  Lami(m) = Lami2(m); 
	
	// use only the first nev eigenvectors and write to gfu
	for(int m=0;m<nev;m++)
	  {
	    *wh[m]=0.; 
	    for(int k=0;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }
	
	// new p^{j+1}_k = [0,w_j, p_j] v[k]  
	for(int m=0;m<nev;m++)
	  {
	    *w[m] = *wh[m]; 
	    *wh[m] = 0.;  
	    for(int k=nev;k<N;k++) 
	      *wh[m] += asmall(m,k) * (*w[k]);
	  }

	for(int m=0;m<nev;m++)
	  *w[m+2*nev] = *wh[m]; 
	
	// after first iteration use LOBPCG  
	N = 3*nev;
	double err=0.;
	for(int l=0;l<nev/2;l++)
	  err += (LamiOld(l)-Lami(l))*(LamiOld(l)-Lami(l));
	
	*testout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  *testout << " Lami(" << m << ") = " << Lami(m) << endl; 

	cout << " Iteration i " << ii+1 << endl; 
	for (int m=0;m<nev;m++) 
	  cout << " Lami(" << m << ") = " << Lami(m) << endl; 
		
	cout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	*testout << " Error at iteration " << ii+1 << " : " << sqrt(err) << endl;  
	if(sqrt(err)/double(nev/2)<prec) break; 

	LamiOld = Lami;

      }

      

      } 
    for(int m=0;m<nev;m++)
      {
	BaseVector & vu = gfu->GetVector(m);
	vu = *w[m];
      }

  }


namespace 
{
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("lobpcg", NumProcLOBPCG::Create);
  }
  
  Init init;
  
}
}
#endif 

