#include <solve.hpp>


namespace mainz_cpp
{

using namespace ngsolve;
using namespace ngfem;
using namespace ngcomp;
  ///
  class NumProcAddDipole : public NumProc
  {
    protected:
    
      LinearForm * lff;
      ARRAY<Vec<3> > points;
      Vec<3> pol;
    
    public:
      ///
      NumProcAddDipole (PDE & apde, const Flags & flags);
      ///
      virtual ~NumProcAddDipole();

      static NumProc * Create (PDE & pde, const Flags & flags)
      {
        return new NumProcAddDipole (pde, flags);
      }
      static void PrintDoc (ostream & ost);

      ///
      virtual void Do(LocalHeap & lh);
      ///
      virtual void PrintReport (ostream & ost);

      virtual string GetClassName () const
      {
        return "AddDipole";
      }

  };





  NumProcAddDipole :: NumProcAddDipole (PDE & apde, const Flags & flags)
  : NumProc (apde), points(0)
  {
  
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""), 1);
    
  

    const ARRAY<double> & polarization = flags.GetNumListFlag ("pol");

  

    pol(0)=polarization[0]; pol(1)=polarization[1]; pol(2)=polarization[2];
    
    cerr << "*** polarization is " << pol << endl;
    
    const ARRAY<double> & x = flags.GetNumListFlag ("x");
    const ARRAY<double> & y = flags.GetNumListFlag ("y");
    const ARRAY<double> & z = flags.GetNumListFlag ("z");
    
    
    int nx=x.Size(), ny=y.Size(), nz=z.Size();
    for (int k=0; k<nz; k++)
      for (int j=0; j<ny; j++)
        for (int i=0; i<nx; i++)
        {
          Vec<3> p; p(0)=x[i]; p(1)=y[j]; p(2)=z[k];
          points.Append(p);
        }

    cerr << " AddDipole at " << points  << endl; 
  }

  NumProcAddDipole :: ~NumProcAddDipole()
  {
    ;
  }



  void NumProcAddDipole :: PrintDoc (ostream & ost)

  {
    ost << endl;
  }




  void NumProcAddDipole :: Do(LocalHeap & lh)
  {
    int np=points.Size();
    
    ARRAY<int> elements(np);
    ARRAY<IntegrationPoint> ips(np);
    
    
    
    // first point: have to build tree
    elements[0] = ma.FindElementOfPoint(points[0], ips[0], true);
    
    for (int i=1; i<np; i++)
      elements[i] = ma.FindElementOfPoint(points[i], ips[i], false);
    
    // now make element entries
    const FESpace & fes = lff -> GetFESpace();
    
    void * heapp = lh.GetPointer();
    for (int i=0; i<np; i++)
    {
      int eli = elements[i];
      

   

      const HCurlHighOrderFiniteElement<3> & fel = 
          dynamic_cast<const HCurlHighOrderFiniteElement<3> & >(fes.GetFE(eli, lh));
      
      ElementTransformation  eltrans;
      ma.GetElementTransformation (eli, eltrans, lh);
      
      SpecificIntegrationPoint<3,3> sip(ips[i], eltrans, lh);
      
   

      ARRAY<int> dnums;
      fes.GetDofNrs (eli, dnums);
      int nd = dnums.Size();
      
      if(lff->GetFESpace().IsComplex())
	{ 
      
      FlatVector<Complex> fvec(nd, lh);
      // calculate curl v(xi)
      fvec = (1.0/sip.GetJacobiDet())
          * fel.GetCurlShape(sip.IP(), lh) * Trans(sip.GetJacobian()) * pol;
      fvec *= sip.GetJacobiDet();
      
      //  cerr << "*** pol = " << pol << endl;
      // cerr << "*** fvec = " << fvec << endl;
      
      dynamic_cast<S_LinearForm<Complex> & >(*lff).AddElementVector(dnums,fvec);
	} 
      else // SZ das is fuer delta_X * v 
	{
	  FlatVector<double> fvec(nd, lh);
	  FlatMatrix<double> fm(3,nd,lh); 
	  FlatVector<double> pp(3,lh); 
	  for(int l=0;l<3;l++)  pp(l)=pol(l); 
	  // calculate curl v(xi)
	  //  fvec = fel.GetShape(sip.IP(), lh) * pol;

	  
	  fm = Trans (sip.GetJacobianInverse ()) * Trans (fel.GetShape(sip.IP(),lh)) ; 
	  fvec = Trans(fm)*pp ; 
	  fvec *= sip.GetJacobiDet();

	  cerr << " t " << Trans(Trans (sip.GetJacobianInverse ()) * Trans (fel.GetShape(sip.IP(),lh))) * pp << endl; 
	  cerr << " fm " << fm << endl; 
	  cerr << " pol " << pp << endl; 
	  cerr << " fm*pol " << Trans(fm)*pp << endl; 
	  cerr << " fvec " << fvec << endl; 
	 	  
	  dynamic_cast<S_LinearForm<double> & >(*lff).AddElementVector(dnums,fvec);
	}
      
      lh.CleanUp(heapp);
    }
  }

  void NumProcAddDipole :: PrintReport (ostream & ost)
  {
    ost << "NumProcAddDipole:" << endl;
  }



  class Init
  { 
    public: 
      Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("adddipole", NumProcAddDipole::Create, NumProcAddDipole::PrintDoc);
  }

    
  Init init;
} // namespace

  
