#ifdef LAPACK

#include <solve.hpp>
// #include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;

  ///  
  class NumProcMaxwellEVP : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    /// 
    BilinearForm * bfm;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    Preconditioner * prepot;
    // 
    string filename, filenameh;



    ///
    int maxsteps;
    int proj_steps;
    int nev; 
    ///
    double prec, prec2;
    ///
    bool print;
    bool lshape,cube,fichera; 
    int lopcg; 
    bool fixproj; 

  public:
    ///
    NumProcMaxwellEVP (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcMaxwellEVP();

   static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcMaxwellEVP (pde, flags);
    }

    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "Maxwell Eigenvalue Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl;
    }
  };



  NumProcMaxwellEVP :: NumProcMaxwellEVP (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    prepot = pde.GetPreconditioner (flags.GetStringFlag ("preconditionerpot", ""));
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    proj_steps = int(flags.GetNumFlag ("projsteps", 5));
    nev =  gfu->GetMultiDim(); 
    lshape = flags.GetDefineFlag("lshape"); 
    fichera = flags.GetDefineFlag("fichera"); 
    cube = flags.GetDefineFlag("cube"); 
    filename = flags.GetStringFlag ("filename","maxwellevp.out");
    filenameh = flags.GetStringFlag ("filename2","maxwellevp2.out");
    lopcg = int(flags.GetNumFlag("lopcg",0.)); 
    
    prec = double(flags.GetNumFlag ("prec", 1.e-5));
    prec2 = double(flags.GetNumFlag ("prec2", 1.e-5));
    fixproj =flags.GetDefineFlag("fixproj"); 
    
  }

  NumProcMaxwellEVP :: ~NumProcMaxwellEVP()
  {
    ;
  }

#define LOPCG 
#ifdef LOPCG 
  void NumProcMaxwellEVP :: Do(LocalHeap & lh)
  {
    cout << "solve Maxwell evp" << endl;
    *testout << "solve Maxwell evp" << endl;

    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseMatrix & matm = bfm->GetMatrix();
    BaseVector & vu = gfu->GetVector();
    
    const BaseMatrix & matpre = pre->GetMatrix();
    const BaseMatrix & matprepot = prepot->GetMatrix();
    
    SparseMatrix<double> * grad;
    clock_t start = clock(); 

    if (dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->CreateGradient();
    
    if (dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace())->CreateGradient();

    double thistime = (clock() - start)/double(CLOCKS_PER_SEC); 
    (*testout) << " Time for computing grad     " << thistime << endl;
    (cout) << " Time for computing grad     " << thistime << endl;
    start = clock(); 
    
    int maxorder =-1; 
    int minorder =99; 
    
    if (dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace()))
      {
	maxorder =dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->GetMaxOrder();
	minorder =dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->GetMinOrder();
      }
    cout << " hcurl minorder " << minorder << endl; 
    cout << " hcurl maxorder " << maxorder << endl; 

    int ndofs = gfu->GetFESpace().GetNDof(); 
    
    if ((lopcg+1)*nev >= ndofs)
      { 
	nev = ndofs/(lopcg+2); 
      }
    

    int initproj =  proj_steps; 

    BaseVector & vw = *mata.CreateVector();
    BaseVector & vd = *mata.CreateVector();
    BaseVector & vv = *mata.CreateVector();
    BaseVector & vau = *mata.CreateVector();
    BaseVector & vmu = *mata.CreateVector();

    BaseVector & vwpot = *matprepot.CreateVector();
    BaseVector & vdpot = *matprepot.CreateVector();

    double lam;
    double lam_old = 0.; 

    //vv = 1;
    int it = 0; 
    
    static ofstream evout2 (filename.c_str());
    static ofstream evout (filenameh.c_str());
  
    
    int mnev = int(max2(double(nev), 9.));
    double * solev = new double[mnev];
    if (lshape) 
      {
    solev[0]= 0.963972384472E+01;
    solev[1]= 0.113452262252E+02;
    solev[2]= 0.134036357679E+02; 
    solev[3]= 0.151972519265E+02; 
    solev[4]= 0.195093282458E+02;
    solev[5]= 0.197392088022E+02;
    solev[6]= 0.197392088022E+02;
    solev[7]= 0.197392088022E+02;
    solev[8]= 0.212590837990E+02; 
      }
    else
      if (cube) 
	{
	  solev[0] = 2.; 
	  solev[1] = 2.; 
	  solev[2] = 2.; 
	  solev[3] = 3.; 
	  solev[4] = 3.;
	  solev[5] = 5.; 
	  solev[6] = 5.; 
	  solev[7] = 5.;
	  solev[8] = 5.; 
	  //solev[9] = 5.; 
	 
	}


      else if(fichera) 
	{solev[0] =  3.31380523357;
	  solev[1] = 5.88634994607;
	  solev[2] = 5.88634994619; 
	  solev[3] = 10.6945143272; 
	  solev[4] = 10.6945143276; 
	  solev[5] = 10.7005804029;  
	  solev[6] = 12.3345495948;  
	  solev[7] = 12.3345495949; 
	  solev[8] = 0; 
	} 
      else 
	for(int l=0;l<9;l++) solev[l] = 0.; 
    
    for(int m=10;m<mnev; m++) 
      solev[m] = 0.; 
 
    double compev[nev]; 
    double sum_old=0.; 
  
 
    cout.precision(12);
    (*testout).precision(12); 
	 evout.precision(12); 
	 evout2.precision(12); 
    
 
       Vector<double> Lami(nev); 
    Lami=1.; 
    
    
 
    // lopcg = 0 : Preconditioned inverse iteration x_{n+1} = x_n - C^{-1}(Ax_n - lambda Mx_n) 
    //       = 1 : LOBPG    x_n+1 = alpha x_n + tau C^{-1](Ax_n - lmabda M x_n) = alpha x_n + tau w_n   
    //       = 2 : LOBPCG1  x_n+1 = alpha x_n + tau w_n + gamma p_n mit p_{n+1}^j = x_n+1^j - sum(tau^k x_n^k) 
    //       = 2 : LOBPCG2  x_n+1 = alpha x_n + tau w_n + gamma p_n mit p_{n+1} = x_n+1 - tau x_n 
    
 
    switch(lopcg)
      {
      case 0: 
	{
	  cout << " Using Inexact Inverse Iteration with Inexact Proj " << endl; 
	  for(int m=0;m<nev;m++)
	    {
	      vw.SetRandom();     
	      gfu->GetVector(m)=matpre*vw; 
	    }
	
	  Matrix<double> asmall(nev); 
	  Matrix<double> msmall(nev); 
	  
	  ARRAY<BaseVector*> u1(nev); 
	  for(int m=0;m<nev; m++)
	    u1[m]=mata.CreateVector();
	  
	  
	  
	  
	  for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
	    {
	      it++; 
	      cout << endl << "iteration i = " << it << endl;
	      for(int m=0;m<nev;m++)
		{
	    
		  BaseVector & vu = gfu->GetVector(m); 
		  vmu = matm * gfu->GetVector(m); 	  		
		  vau = mata * gfu->GetVector(m);
		
		  vd = vau - Lami(m) * vmu;	    	    
		
		
		  vw = matpre * vd;
		
		  vu -= vw;
		
		
		  vmu = matm*vu; 
		  double l2 = 1./sqrt(InnerProduct(vu,vmu)); 
		  vu *= l2; 
	
		  // H^1- projection
		  for (int j = 1; j <= proj_steps; j++)
		    {
		      vd = matm * vu;
		      vdpot = Transpose (*grad) * vd;
		      vwpot = matprepot * vdpot;
		      vw = (*grad) * vwpot; 
		      vu -= vw;
		    }

		  vmu = matm*vu; 
		  l2 = 1./sqrt(InnerProduct(vu,vmu)); 
		  vu *= l2; 
		
		// *testout << " after: " << sqrt(InnerProduct(vmu = matm*vu,vu)) << endl;
	      }
	    
	      // X^T A X y = lami X^t M X y 
	      for(int j=0;j<nev;j++)
		{
		  vmu = matm * gfu->GetVector(j); 	  		
		  vau = mata * gfu->GetVector(j);
		  for(int k=0;k<nev;k++) 
		    {
		      asmall(j,k) = InnerProduct(vau,gfu->GetVector(k)); 
		      msmall(j,k) = InnerProduct(vmu,gfu->GetVector(k)); 
		    }
		}
	  	    
	      LapackGHEPEPairs(nev, &asmall(0,0), &msmall(0,0), &Lami(0));


	    // X y = u 
	    for(int m=0;m<nev;m++)
	      *u1[m]=gfu->GetVector(m);
	    
	    for(int m=0;m<nev;m++)
	      {
		BaseVector & vu = gfu->GetVector(m); 
		vu =0.; 
		for(int k=0;k<nev;k++) 
		  //  asmall(m,k) = y_i(k) 
		  vu += asmall(m,k) * (*u1[k]);
	      }
	    
	    for(int m=0; m<nev;m++)
	      {
		// projection
		
		vd = matm * vu;
		vdpot = Transpose (*grad) * vd;
		
		vwpot = matprepot * vdpot;
		vw = (*grad) * vwpot; 
		vu -= vw;
				
		vmu = matm*vu; 
		double l2 = 1./sqrt(InnerProduct(vu,vmu)); 
		vu *= l2; 
	      }
	    
	    lam = Lami[0]; 
	    for(int i=1;i<nev;i++)
	      lam += Lami[i];
	    lam /= double(nev); 
	    
	    
	    if(lshape || cube ) 
	      {
		 
		double sum=0.;
		
		for (int m=0;m<nev; m++)
		  {
		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    
		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
		
		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum) << endl; 
		if(fabs(sum-sum_old)/fabs(sum) < prec) break; // ii=maxsteps+1;  
		sum_old=sum;
		cout << ii <<  " average error " << sum_old/nev << endl; 
		*testout << ii <<  " sum " << sum_old/nev << endl; 
	      }
	    else if(fichera) 
	      {
	 
		double sum=0.;
		
		for (int m=0;m<nev; m++)
		  {
		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    // *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    
		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
		
		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum) << endl; 
		if(fabs(sum-sum_old)/fabs(sum) < prec) break; // ii=maxsteps+1;  
		sum_old=sum;
		cout << ii <<  " average error " << sum_old/nev << endl; 
		*testout << ii <<  " sum " << sum_old/nev << endl; 
	    }
	    /*else 
	      if(cube)
	      { 
	      double sum = 0; 
	      sum =  fabs(Lami(0)-2.)/2.;
	      cout << "cube-err: " << sum << endl;
	      if( fabs(sum-sum_old) < 1.e-5 ) ii = maxsteps+1; 
	      sum_old = sum; 
	      }*/
	    else
	      {
		double sum =0; 
		cout << endl << " ii = " << ii  << " : " << endl;  
		
		sum = Lami[0]; 
		for (int m=0;m<nev; m++)
		  cout << " Lami[ " << m << " ] =  " << Lami[m]  << "\t" ; 
		if(fabs(sum-sum_old)/fabs(sum) < prec) break; // ii=maxsteps+1;  
		sum_old=sum;
		cout << " fabs(sum-sum_old)/fabs(sum) " <<  fabs(sum-sum_old)/fabs(sum) << endl; 
	      }
	    
	    

	    lam_old = lam; 
	    //	if(lshape && (fabs(lam_old-lam)/fabs(lam) < 1.e-4*fabs(solev[0]-lam_old)/fabs(solev[0])))  break; 
	    // lam_old = lam; 
	    
	    
	    
	    
	    //double y = sqrt(InnerProduct(vmu = matm*vu,vu)); 
	    // vu /= sqrt(msmall(m,m));
	    
	  }
	
	  break; 
	}
	
      case 1: // LOPG 
	{
	  cout << " Using LOPG with inexact projection " << endl; 
 	int N = 2*nev; 
	Matrix<double> asmall(N,N); 
 	Matrix<double> msmall(N,N); 
 	Vector<double> Lami2(N); 
	
 	  
  
 	  ARRAY<BaseVector*> w(N); 
	  // w = [X,C^[-1](AX- Lami MX)]
 	  for(int m=0; m<N;m++)
 	    w[m]=mata.CreateVector();	 
 	  for(int m=0;m<nev; m++)
 	    {
 	      vw.SetRandom(); 
 	      *w[m] = matpre * vw; 
 	    }
  
 	  // H^1-projection of initial u 
	  for (int m=0;m<nev;m++)
 	    for (int j = 1; j <= proj_steps; j++)
 	      {
 		vd = matm * (*w[m]);
		vdpot = Transpose (*grad) * vd;
 		vwpot = matprepot * vdpot;
		*w[m] -= (*grad) * vwpot; 
 	      }
	  
 	  for (int ii = 1; ii <= maxsteps; ii++) 
 	    {
 	      it++; 
 	      cout << endl << "iteration i = " << it << endl;
 	      for(int m=0;m<nev;m++)
 		{
 		  vmu = matm * (*w[m]); 	  		
 		  vau = mata * (*w[m]);
 		  if (ii ==1) 
		    Lami(m) = InnerProduct(vau,(*w[m]))/InnerProduct(vmu,(*w[m])); 
		   
		  vd = (vau-Lami(m)*vmu); 
 		  *w[nev+m] = matpre*vd;  
 		  	  
		  // H^1-projection of residuum w 
 		  for (int j = 1; j <= proj_steps; j++)
 		    {
 		      vd = matm * (*w[nev+m]);
 		      vdpot = Transpose (*grad) * vd;
 		      vwpot = matprepot * vdpot;
 		      *w[nev+m] -= (*grad) * vwpot; 
 		    }
	  
 		}
	      
 	      // [X,W]^T A [X,W] y = lami [X,W]^t M [X,W] y 
 	      for(int j=0;j<N;j++)
		{
 		  vmu = matm * (*w[j]); 	  		
 		  vau = mata * (*w[j]); 
 		  for(int k=0;k<N;k++) 
 		    {
 		      asmall(j,k) = InnerProduct(vau,*w[k]); 
 		      msmall(j,k) = InnerProduct(vmu,*w[k]); 
 		    }
 		}

	      *testout << " asmall " << asmall << endl; 
	      *testout << " msmall " << msmall << endl; 
	      
 	    LapackGHEPEPairs(N, &asmall(0,0), &msmall(0,0), &Lami2(0));

	    for(int i=0;i<nev;i++) Lami(i) = Lami2(i); 
	    
	    // 	    *testout << " iteration it " << ii << " Lami2 " << Lami2 << endl << " Lami " << Lami << endl; 
	    
	    // Save only the first nev eigenvectors 
	    for(int m=0;m<nev;m++)
 	      {
 		BaseVector & vu = gfu->GetVector(m); 
 		vu = 0.;  
 		for(int k=0;k<2*nev;k++) 
 		  vu += asmall(m,k) * (*w[k]);
 	      }

 	    for(int m=0;m<nev;m++)
 	      *w[m] = gfu->GetVector(m); 
	      	    
 	    lam = Lami[0]; 
 	    for(int i=1;i<nev;i++)
 	      lam += Lami[i];
 	    lam /= double(nev); 
	    
	    
 	    if(lshape || cube) 
 	      {
 		double sum=0.;
		
 		for (int m=0;m<nev; m++)
 		  {
 		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
 		    *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
	    
 		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
 		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum) << endl; 
 		if(fabs(sum-sum_old)/fabs(sum)/nev < prec) break; // ii=maxsteps+1;  
 		sum_old=sum;
 		cout << ii <<  " average error " << sum_old/nev << endl; 
 		*testout << ii <<  " sum " << sum_old/nev << endl; 
 	      }
	     else if(fichera) 

	       {
		double sum=0.;
		
		for (int m=0;m<nev; m++)
		  {
		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    // *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		    
		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
		
		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum) << endl; 
		if(fabs(sum-sum_old)/fabs(sum) < prec) break; // ii=maxsteps+1;  
		sum_old=sum;
		cout << ii <<  " average error " << sum_old/nev << endl; 
		*testout << ii <<  " sum " << sum_old/nev << endl; 
	    }
 	    else
 	      {
 		double sum =0; 
		cout << endl << " ii = " << ii  << " : " << endl;  
	
 		sum = 0.; 
 		for (int m=0;m<nev/2; m++)
		  {
		    sum += Lami[m]; 
		    cout << " Lami[ " << m << " ] =  " << Lami[m]  << "\t" ; 
		  }
		sum /= nev/2; 
		
		cout << " |sum - sumold|/|sum_old| " << fabs(sum-sum_old)/fabs(sum_old) << endl; 

 		if(fabs(sum-sum_old)/fabs(sum) < prec) break; 
 		sum_old=sum;
 	      }
 	    lam_old = lam; 
 	    }
 	  break;
	
	}
       case 2: // LOBPCG1 
 	{
	  cout << " Using LOBPCG I with inexact projection " << endl; 
	  	 
	  double sum_old = 0; 

	  int N = 3*nev; 
 	  ARRAY<BaseVector*> w(N);
 	  ARRAY<BaseVector*> wh(nev); 

	  // if (gfu->GetFESpace().GetNDof() < 3*nev) break;  
	  
 	  // BaseVector & vu = gfu->GetVector(m);
 	  // w = [X,C^[-1](AX- Lami MX),P]
 	  for(int m=0; m<N;m++)
 	    w[m]=mata.CreateVector();
	  
 	  for(int m=0;m<nev; m++)
 	    {
 	      vw.SetRandom(); 
 	      *w[m] = matpre * vw; 
 	      wh[m] = mata.CreateVector();
 	    }
	  
 	  // H^1-projection of initial u 
 	  for (int m=0;m<nev;m++)
 	    for (int j = 1; j <= proj_steps; j++)
 	      {
 		vd = matm * (*w[m]);
 		vdpot = Transpose (*grad) * vd;
 		vwpot = matprepot * vdpot;
 		*w[m] -= (*grad) * vwpot; 
 	      }
	  
 	  // P^0 = 0 
 	  for (int m=2*nev;m<N;m++) 
 	    *w[m]=0.; 

	  N -= nev; 
	  
	  // Improvement to implement (vg Knyazev Matlab Code) 
	  // vmu[nev,nev,nev] und vau[nev,nev,nev] speichern 
	  // nach preconditioning nur w-Teil Aw und Mw neu berechnen 
	  // in X^T AX keine neue Anwendung von AX,MX 
	  // ARRAY<BaseVector*> AV(N), MV(N); 

	  
 	  for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
 	    {
	      cout << " LOBPCG I : it " << it << endl; 
	      it++; 
 	      *testout << endl << "iteration i = " << it << endl;
 	      for(int m=0;m<nev;m++)
 		{
 		  vmu = matm * (*w[m]); 	  		
 		  vau = mata * (*w[m]);
 		  if (ii ==1) 
 		    Lami(m) = InnerProduct(vau,(*w[m]))/InnerProduct(vmu,(*w[m])); 
		  
 		  vd = (vau-Lami(m)*vmu); 
 		  *w[nev+m] = matpre*vd;  
		  
		  // H^1-projection of prec. residuum w 
 		  for (int j = 1; j <= proj_steps; j++)
 		    { 
 		      vd = matm * (*w[nev+m]);
 		      vdpot = Transpose (*grad) * vd;
 		      vwpot = matprepot * vdpot;
 		      *w[nev+m] -= (*grad) * vwpot; 
 		    }
		}
	      bool reproj=1;   
	      Vector<double> Lami2(3*nev);     
	      //   while(reproj && proj_steps < 10)
		{
		  Matrix<double> asmall(N,N); 
		  Matrix<double> msmall(N,N); 
	
		  reproj=0; 
		  // [X,W,P]^T A [X,W,P] y = lami [X,W,P]^t M [X,W,P] y 
		  for(int j=0;j<N;j++)
		    {
		      vmu = matm * (*w[j]); 	  		
		      vau = mata * (*w[j]);  
		      for(int k=0;k<N;k++) 
			{
			  asmall(j,k) = InnerProduct(vau,*w[k]); 
			  msmall(j,k) = InnerProduct(vmu,*w[k]); 
			}
		    }
		  
		  LapackGHEPEPairs(N, &asmall(0,0), &msmall(0,0), &Lami2(0));
		  
		  /* int sstart=0; 
		  for(int i=0;i<3*nev;i++)
		    if(Lami2(i) < 1e-3) sstart++; 
		    else break; 
		  */
		 
		  // if(sstart > 0)
		  //  cout << "ATTENTION number of zero eigenvalues " << sstart << " last  EV: "  << Lami2(sstart-1)  << endl; 
		  // if(sstart > 0) {sstart = nev +1;}
		  
		  // if(cube || lshape || fichera) sstart =0; 

		  /* if((sstart >= nev || (sstart > 0 && N == 2*nev)) && !fixproj) 
		    {  
		      for (int m=0;m<N;m++)		      // H^1-projection of prec. residuum w 
		      for (int j = 1; j <= proj_steps; j++)
			{
			  vd = matm * (*w[m]);
			  vdpot = Transpose (*grad) * vd;
			  vwpot = matprepot * vdpot;
			  *w[m] -= (*grad) * vwpot; 
			}
		      
		      proj_steps+=5; 
		      cout << " Increase Proj Steps to  " << proj_steps << endl; 
		      reproj= 1; 
		      continue; 
		    }
		  else 
		    if(sstart && !fixproj)
		      {
			proj_steps+=2;
			cout << " Increase Proj Steps to  " << proj_steps << endl; 
			
		      }
		  if (fixproj) sstart =0; 
		  */ 
		  //	  sstart = 0; 
		  // reproj = 0; 

		  for (int i=0; i<nev; i++)
		    Lami(i) = Lami2(i);
		    // Lami(i) = Lami2(i+sstart);
		  
		  // use only the first nev eigenvectors 
		  int sstart =0; 
	      for(int m=sstart;m<sstart+nev;m++)
		{
		  BaseVector & vu = gfu->GetVector(m); 
 		  vu = 0.;  
 		  for(int k=0;k<N;k++) 
 		    vu += asmall(m,k) * (*w[k]);
 		}
	   
	      // H^1-projection of u 
	      for (int m=0;m<nev;m++)
		for (int j = 1; j <= proj_steps; j++)
		  {
		    BaseVector & vu = gfu->GetVector(m);
		    vd = matm * (vu);
		    vdpot = Transpose (*grad) * vd;
		    vwpot = matprepot * vdpot;
		    vu -= (*grad) * vwpot; 
		  }
	  
	      
	      /* if(sstart > 0 && sstart < nev) 
		{
		  for(int m=0;m<nev;m++)
		    *w[m+2*nev] = 0.; 
		  N = 2*nev; 
		} 
		else */
		{
		  // new p^{j+1}_k = [0,w_j, p_j] v[k]  
		  for(int m=0;m<nev;m++)
		    {
		      *wh[m] = 0.;  
		      for(int k=nev;k<N;k++) 
			*wh[m] += asmall(m,k) * (*w[k]);
		    }
		  
		  for(int m=0;m<nev;m++)
		    *w[m+2*nev] = *wh[m]; 
		  
		  N = 3*nev;
		}
	  
 	      // X^{ii+1]
 	      for(int m=0;m<nev;m++)
 		*w[m] = gfu->GetVector(m); 
	  
	      
		}
	      
 	      lam = Lami(0); 
 	      for(int i=1;i<nev;i++)
 		lam += Lami(i);
 	      lam /= double(nev); 


	       
	    
 	    if(lshape || cube ) 
 	      {
 		double sum=0.;
		
 		for (int m=0;m<nev; m++)
 		  {
 		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
 		    *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
	    
 		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
 		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum)/double(nev) << endl; 
 		if(fabs(sum-sum_old)/fabs(sum)/nev < prec) break; // ii=maxsteps+1;  
 		sum_old=sum;
 		cout << ii <<  " average error " << sum_old/nev << endl; 
 		*testout << ii <<  " sum " << sum_old/nev << endl; 
 	      }
 	    else
	      if(fichera)
		{
		  double sum=0.;
		  
		  for (int m=0;m<nev; m++)
		    {
		      cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		      *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		      cout << "Lami[m] = "<< Lami[m] << "\t" << flush;
		      
		      sum +=  Lami[m]; 
		      cout << " lami " << Lami[m] << endl; 
		    }

		  cout << " sum " << sum << endl; 
		  cout << " sumold " << sum_old << endl; 
		  		     
		  cout << " |sum - sumold|/|sum| " << fabs(sum-sum_old)/fabs(sum) << endl;   
		  
		  
		  cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum)/double(nev) << endl; 
		  if(fabs(sum-sum_old)/fabs(sum)/nev < prec) break; // ii=maxsteps+1;  
		  sum_old=sum;
	
		  
		}
	      else 
 	      {
 		double sum =0; 
		cout << endl << " ii = " << ii  << " : " << endl;  
	
 		sum = 0.; 
 		for (int m=0;m<nev/2; m++)
		  {
		    sum += Lami[m]; 
		    cout << " Lami[ " << m << " ] =  " << Lami[m]  << "\t" ; 
		  }
		sum /= nev/2; 
		
		cout << " |sum - sumold|/|sum| " << fabs(sum-sum_old)/fabs(sum) << endl; 

 		if(fabs(sum-sum_old)/fabs(sum) < prec) break; 
 		sum_old=sum;
 	      }
 	    lam_old = lam; 
	  
 	    }
 	  break;
	



	} 

	
      case 3: // LOPCG II  
	{
	  cout << " Using LOPCG II with inexact projection " << endl; 
	  int N = 3*nev; 
	  Matrix<double> asmall(nev,nev); 
	  Matrix<double> msmall(nev,nev); 
	  	
 	  
	  
 	  ARRAY<BaseVector*> w(3*nev); 
	  // w = [X,C^[-1](AX- Lami MX),P]

	  for(int m=0; m<N;m++)
 	    w[m]=mata.CreateVector();
	  
 	  for(int m=0;m<nev; m++)
 	    {
 	      vw.SetRandom(); 
 	      *w[m] = matpre * vw; 
	    }
	  
 	  // H^1-projection of initial u 
 	  for (int m=0;m<nev;m++)
 	    for (int j = 1; j <= proj_steps; j++)
 	      {
 		vd = matm * (*w[m]);
 		vdpot = Transpose (*grad) * vd;
 		vwpot = matprepot * vdpot;
 		*w[m] -= (*grad) * vwpot; 
 	      }
	  
 	  // P^0 = 0 
 	  for (int m=2*nev;m<N;m++) 
 	    *w[m]=0.; 

	  int n=2; 
	  
 	
	  for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
 	    {
	      Matrix<double> asmall(nev,nev); 
	      Matrix<double> msmall(nev,nev); 
	      	  
 	      it++; 
 	      cout << endl << "iteration i = " << it << endl;
 	      for(int m=0;m<nev;m++)
 		{
 		  vmu = matm * (*w[m]); 	  		
 		  vau = mata * (*w[m]);
 		  if (ii ==1) 
 		    Lami(m) = InnerProduct(vau,(*w[m]))/InnerProduct(vmu,(*w[m])); 
		  
 		  vd = (vau-Lami(m)*vmu); 
 		  *w[nev+m] = matpre*vd; 
		  
 		  // H^1-projection of residuum w 
 		  for (int j = 1; j <= proj_steps; j++)
 		    {
 		      vd = matm * (*w[nev+m]);
 		      vdpot = Transpose (*grad) * vd;
 		      vwpot = matprepot * vdpot;
 		      *w[nev+m] -= (*grad) * vwpot; 
 		    }
		  
 		}
	      
	      
	      
	      	      
	      for(int j=0;j<nev;j++)
		{
		  Matrix<double> a(n,n); 
		  Matrix<double> b(n,n); 
		  Vector<double> gamma(n); 
		 
		  for(int l=0;l<n;l++)
		    {
		      vmu = matm * (*w[j+nev*l]); 	  		
		      vau = mata * (*w[j+nev*l]); 
		      for(int k=0;k<n;k++) 
		      {
			a(l,k) = InnerProduct(vau,*w[j+nev*k]); 
			b(l,k) = InnerProduct(vmu,*w[j+nev*k]); 
		      }
		    }

		  *testout << " a-mat " << j << ": " << a << endl; 
		  *testout << " b-mat " << j << ": " << b << endl; 
		  
		  LapackGHEPEPairs(n, &a(0,0), &b(0,0), &gamma(0));
		  
		  *testout << " gamma " << gamma << endl; 

		  // Extract x_j,w_j,p_j
		  ARRAY<BaseVector*> xwp(n); 
		  for(int l=0;l<n;l++) 
		    xwp[l] = mata.CreateVector(); 
	       
		  
		  for(int l=0;l<n;l++)
		    {
		      *xwp[l] = 0.;  
		      for(int k=0;k<n;k++) 
			*xwp[l] += a(l,k) * (*w[k*nev+j]);

		      
		    }

		  		  
		  
		  
		  BaseVector & vu = gfu->GetVector(j); 
		  vu = 0; 
		  for(int l=0;l<n;l++) 
		    vu += gamma[l]*(*xwp[l]);
		  
		  for(int l=0;l<n;l++) 
		    *testout << " xwp " << *xwp[l] << endl; 
 
		  /*	  vu =0; 
		  vu +=(*w[j] -(*w[j+nev]);  // Das wuerde jetzt invere it bedeuten 
		  */
		
		  *w[j] = vu; 
		
		  for(int l=1;l<n;l++) 
		    *w[j+2*nev] += gamma[l]*(*xwp[l]);
		  

		  
		
		}


	      
	      for(int m=0; m<nev; m++)
		*testout << "w[m] " << *w[m] << endl; 
	      *testout << " end wm " << endl; 

	      n=3;
 	      // [X]^T A [X] y = lami [X]^t M [X] y 
 	      for(int j=0;j<nev;j++)
		{
		  *testout << " j " << j  << " nev " << nev << endl;  
 		  vmu = matm * (*w[j]); 	  		
 		  vau = mata * (*w[j]); 
 		  for(int k=0;k<nev;k++) 
 		    {
 		      asmall(j,k) = InnerProduct(vau,*w[k]); 
 		      msmall(j,k) = InnerProduct(vmu,*w[k]); 
 		    }
		 
 		}

	      *testout << " asmall " << asmall << endl; 
	      *testout << " msmall " << msmall << endl; 
	      
	      LapackGHEPEPairs(nev, &asmall(0,0), &msmall(0,0), &Lami(0));


	      // Save only the first nev eigenvectors 
	      for(int m=0;m<nev;m++)
		{
		  BaseVector & vu = gfu->GetVector(m); 
		  vu = 0.;  
		  for(int k=0;k<nev;k++) 
		    vu += asmall(m,k) * (*w[k]);
		}

 	    for(int m=0;m<nev;m++)
 	      *w[m] = gfu->GetVector(m); 
	      	    
 	    lam = Lami[0]; 
 	    for(int i=1;i<nev;i++)
 	      lam += Lami[i];
 	    lam /= double(nev); 
	    
	    
 	    if(lshape || cube || fichera ) 
 	      {
 		double sum=0.;
		
 		for (int m=0;m<nev; m++)
 		  {
 		    cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
 		    *testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
	    
 		    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
		  }	
		
 		cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum_old) << endl; 
 		if(fabs(sum-sum_old)/fabs(sum)/nev < prec) break; // ii=maxsteps+1;  
 		sum_old=sum;
 		cout << ii <<  " average error " << sum_old/nev << endl; 
 		*testout << ii <<  " sum " << sum_old/nev << endl; 
 	      }
 	    else
 	      {
 		double sum =0; 
		cout << endl << " ii = " << ii  << " : " << endl;  
	
 		sum = 0.; 
 		for (int m=0;m<nev/2; m++)
		  {
		    sum += Lami[m]; 
		    cout << " Lami[ " << m << " ] =  " << Lami[m]  << "\t" ; 
		  }
		sum /= nev/2; 
		
		cout << " |sum - sumold|/|sum| " << fabs(sum-sum_old)/fabs(sum) << endl; 

 		if(fabs(sum-sum_old)/fabs(sum) < prec) break; 
 		sum_old=sum;
 	      }
 	    lam_old = lam; 
 	    }
 	  break;
	
	}

	

      }

  

  for (int nn = 0; nn<nev ; nn++) cout << " solev[nn] " <<  solev[nn] << endl; 
    if(lshape)
      cout  << endl << " ********** LSHAPE  ******* " << endl; 
    if(cube) 
      cout << endl << " ********** CUBE    ******* " << endl; if(fichera)
      cout  << endl << " ********** FICHERA  ******* " << endl; 
    
    
    //  if (fichera) nev = 3; 

    if(lshape || cube)   
      {
	double sum=0.;
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	int mm = 0; 
	for (int m=0;m< nev/2; m++,mm++)
	  {
	    cout << "lami ["<< m << "] = " <<  Lami(m) << " \t error[m] = "
		 << fabs((solev[m] - Lami(m))/solev[m]) << "\t" << endl; 
	  	    
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]);
	  }	
	
	evout <<  minorder << " \t " << maxorder  << "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/(mm+1)  << "\t"
	      << nev/2 << "\t"; 

	for (int m=mm;m< nev; m++)
	  {
	    cout << "lami [" << m << "] = " <<  Lami(m) << " \t error[m] = "
		 << fabs((solev[m] - Lami(m))/solev[m]) << "\t" << endl; 
	  	    
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]);
	  }	

	evout2 <<  minorder << " \t" << maxorder <<  "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/nev  << "\t"
	       << nev << "\t" ; 
       
	
	for (int m=0;m<10;m++) 
	  evout2 << fabs((solev[m]-Lami(m))/Lami(m)) << "\t"; 
	for (int m=0;m<5;m++) 
	  evout << fabs((solev[m]-Lami(m))/Lami(m)) << "\t"; 
	
	if(fichera) 
	  for (int m=0; m<nev;m++)
	    {
	      evout << Lami(m) << "\t"; 
	      evout2 << Lami(m) << "\t" ;
	    }
	      
	evout << it << endl; 
	evout2 << it << endl; 
	
	cout << "ndofs " << gfu->GetFESpace().GetNDof()  << "\t sum/nev " << sum/nev << endl; 	       
	

      }
    else if(fichera)   
      {
	double sum=0.;
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	int mm = 0; 
	for (int m=0;m< nev/2; m++,mm++)
	  {
	    cout << "lami ["<< m << "] = " <<  Lami(m) << " \t error[m] = "
		 << fabs((solev[m] - Lami(m))/solev[m]) << "\t" << endl; 
	  	    
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]);
	  }	
	
	evout <<  minorder << " \t " << maxorder  << "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/(mm+1)  << "\t"
	      << nev/2 << "\t"; 

	for (int m=mm;m< nev; m++)
	  {
	    cout << "lami [" << m << "] = " <<  Lami(m) << " \t error[m] = "
		 << fabs((solev[m] - Lami(m))/solev[m]) << "\t" << endl; 
	  	    
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]);
	  }	

	evout2 <<  minorder << " \t" << maxorder <<  "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/nev  << "\t"
	       << nev << "\t" ; 
       
	
	for (int m=0;m<10;m++) 
	  evout2 << fabs((solev[m]-Lami(m))/Lami(m)) << "\t"; 
	for (int m=0;m<5;m++) 
	  evout << fabs((solev[m]-Lami(m))/Lami(m)) << "\t"; 
	
	if(fichera) 
	  { for (int m=0; m<nev;m++)
	    {
	      evout << Lami(m) << "\t"; 
	      evout2 << Lami(m) << "\t" ;
	    } 
	    for (int m=0; m<nev;m++)
	      {
		evout << solev[m] << "\t"; 
		evout2 << solev[m] << "\t" ;
	      }
	  }
	
	      
	evout <<  "\t"  << it << "\t"  << proj_steps << endl; 
	evout2 << "\t"  << it <<  "\t" << proj_steps <<  endl; 
	
	cout << "ndofs " << gfu->GetFESpace().GetNDof()  << "\t sum/nev " << sum/nev << endl; 	       
	
	
      }
    else 
      {
	double sum=0; 
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	sum = 0; 
	
	evout2 <<  minorder << " \t " << maxorder  << "\t" <<  gfu->GetFESpace().GetNDof() << "\t  5 \t" ; 
	int mm = 0; 
	for(int m=0;m<nev/2;m++,mm++)
	  {
	   
	    sum += fabs((solev[m] - Lami(m))/Lami[m]); 
	    cout << " lami[ " << m << " ] =  " << Lami[m]  << "\t" ;
	  } 
	evout <<  minorder << " \t" << maxorder << "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/mm  << "\t" << "10" << "\t"; 

	for(int m=0;m<nev;m++)
	  {
	   
	    sum += fabs((solev[m] - Lami(m))/Lami[m]); 
	    cout << " lami[ " << m << " ] =  " << Lami[m]  << "\t" ;
	  } 
	
	evout2 << sum/nev << endl;  
	for(int m=0;m<nev; m++)
	  evout2 << fabs((solev[m] - Lami(m))/Lami[m]) << "\t";    
	
	

      }
  
    /*  for(int m=0;m<nev; m++)
      {
	BaseVector & vu = gfu->GetVector(m); 
	vmu = matm * gfu->GetVector(m); 	  		
	vu /= InnerProduct(gfu->GetVector(m),vmu); 

	// projection
	
	vd = matm * vu;
	vdpot = Transpose (*grad) * vd;
	
	vwpot = matprepot * vdpot;
	vw = (*grad) * vwpot; 
	vu -= vw;

       
      	
	}*/	
   
    (*testout) << " ********************************************************* " << endl; 

    cout << " with order " << minorder << "  /  " << maxorder <<  endl;  
    proj_steps = initproj;
  }

#else 
  void NumProcMaxwellEVP :: Do(LocalHeap & lh)
  {
    cout << "solve Maxwell evp" << endl;
    *testout << "solve Maxwell evp" << endl;

    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseMatrix & matm = bfm->GetMatrix();
    BaseVector & vu = gfu->GetVector();
    
    const BaseMatrix & matpre = pre->GetMatrix();
    const BaseMatrix & matprepot = prepot->GetMatrix();
    
    SparseMatrix<double> * grad;
    clock_t start = clock(); 

    if (dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->CreateGradient();
    
    if (dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace())->CreateGradient();

    int maxorder =-1; 
    int minorder =99; 
    
    if (dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace()))
      {
	maxorder =dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->GetMaxOrder();
	minorder =dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->GetMinOrder();
      }


    double thistime = (clock() - start)/double(CLOCKS_PER_SEC); 
    (*testout) << " Time for computing grad     " << thistime << endl;
    (cout) << " Time for computing grad     " << thistime << endl;
    start = clock(); 
    
    BaseVector & vw = *mata.CreateVector();
    BaseVector & vd = *mata.CreateVector();
    BaseVector & vv = *mata.CreateVector();
    BaseVector & vau = *mata.CreateVector();
    BaseVector & vmu = *mata.CreateVector();

    BaseVector & vwpot = *matprepot.CreateVector();
    BaseVector & vdpot = *matprepot.CreateVector();

    double lam;
    double lam_old = 0.; 

    //vv = 1;
    int it; 
    
    //  static ofstream evout (filename1.c_str());
    //ofstream evout2 (filename2.c_str());
    
    static ofstream evout ("lshape_error_hpl4a.out");
    static ofstream evout2 ("lshape2_error_hpl4a.out");
    double solev[10]; 
    if (lshape) 
      {
    solev[0]= 0.963972384472E+01;
    solev[1]= 0.113452262252E+02;
    solev[2]= 0.134036357679E+02; 
    solev[3]= 0.151972519265E+02; 
    solev[4]= 0.195093282458E+02;
    solev[5]= 0.197392088022E+02;
    solev[6]= 0.197392088022E+02;
    solev[7]= 0.197392088022E+02;
    solev[9]= 0.212590837990E+02; 
      }
    else
      if (cube) 
	{
	  solev[0] = 2.; 
	  solev[1] = 2.; 
	  solev[2] = 2.; 
	  solev[3] = 3.; 
	  solev[4] = 3.;
	  solev[5] = 5.; 
	  solev[6] = 5.; 
	  solev[7] = 5.;
	  solev[8] = 5.; 
	  solev[9] = 5.; 
	}
      else 
	for(int l=0;l<10;l++) solev[l] = 0.; 
     
	
    double compev[nev]; 
    double sum_old=100; 
  
    for(int m=0;m<nev;m++)
      {
	vw.SetRandom();     
	gfu->GetVector(m)=matpre*vw; 
	
      }
    cout.precision(12);
    (*testout).precision(12); 
    
    //vu = mata * vv;
    
    Vector<double> Lami(nev); 
    Lami=1.; 
    
   

    Matrix<double> asmall(nev); 
    Matrix<double> msmall(nev); 
    
 ARRAY<BaseVector*> u1(nev); 
    for(int m=0;m<nev; m++)
      u1[m]=mata.CreateVector();
    
    for (int ii = 1; ii <= maxsteps; ii++) // maxsteps 
      {
	it++; 
	cout << endl << "iteration i = " << it << endl;
	for(int m=0;m<nev;m++)
	  {
	    
	    BaseVector & vu = gfu->GetVector(m); 
	    vmu = matm * gfu->GetVector(m); 	  		
	    vau = mata * gfu->GetVector(m);
	    
	    vd = vau - Lami(m) * vmu;	    	    
	    
	     
	    vw = matpre * vd;
	    
	    vu -= vw;
	    
	    /* *testout << "m = " << m 
		     << " before: " << sqrt(InnerProduct(vmu = matm*vu,vu)); 
	    */
	    vmu = matm*vu; 
	    double l2 = 1./sqrt(InnerProduct(vu,vmu)); 
	    vu *= l2; 
	    // projection
	    for (int j = 1; j <= proj_steps; j++)
	      {
		
		vd = matm * vu;
		vdpot = Transpose (*grad) * vd;
		//cout << " Projection step " << j << " \t (grad vd , grad vd) " << InnerProduct(vdpot,vdpot) <<endl; 
		vwpot = matprepot * vdpot;
		//cout <<  " l2 " << L2Norm(vwpot) << endl; 
		vw = (*grad) * vwpot; 

		
		vu -= vw;
	      }

	    vmu = matm*vu; 
	    l2 = 1./sqrt(InnerProduct(vu,vmu)); 
	    vu *= l2; 

	    // *testout << " after: " << sqrt(InnerProduct(vmu = matm*vu,vu)) << endl;
	  }

	// X^T A X y = lami X^t M X y 
	for(int j=0;j<nev;j++)
	  {
	    vmu = matm * gfu->GetVector(j); 	  		
	    vau = mata * gfu->GetVector(j);
	    for(int k=0;k<nev;k++) 
	      {
		asmall(j,k) = InnerProduct(vau,gfu->GetVector(k)); 
		msmall(j,k) = InnerProduct(vmu,gfu->GetVector(k)); 
	      }
	  }

	
       LapackGHEPEPairs(nev, &asmall(0,0), &msmall(0,0), &Lami(0));

	for(int m=0;m<nev;m++)
	  *u1[m]=gfu->GetVector(m);
	
	for(int m=0;m<nev;m++)
	  {
	    BaseVector & vu = gfu->GetVector(m); 
	    vu =0.; 
	    for(int k=0;k<nev;k++) 
	      //  asmall(m,k) = y_i(k) 
	      vu += asmall(m,k) * (*u1[k]);
   
	    
	  }
	
	for(int m=0; m<nev;m++)
	  {
	    // projection
	    
	    vd = matm * vu;
	    vdpot = Transpose (*grad) * vd;
	    
	    vwpot = matprepot * vdpot;
	    vw = (*grad) * vwpot; 
	    vu -= vw;
	    
	    

	    vmu = matm*vu; 
	    double l2 = 1./sqrt(InnerProduct(vu,vmu)); 
	    vu *= l2; 
	  }

	lam = Lami[0]; 
	for(int i=1;i<nev;i++)
	  lam += Lami[i];
	lam /= double(nev); 
	

	if(lshape || cube ) 
	  {
	    double sum=0.;
	 	    
	    for (int m=0;m<nev; m++)
	      {
		cout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		*testout << "error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		
		sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
	      }	
	   
	    cout << endl << " average error reduction : " << fabs(sum-sum_old)/fabs(sum) << endl; 
	    if(fabs(sum-sum_old)/fabs(sum)/nev < prec) break; // ii=maxsteps+1;  
	    sum_old=sum;
	    cout << ii <<  " average error " << sum_old/nev << endl; 
	    *testout << ii <<  " sum " << sum_old/nev << endl; 
	  }
	/*else 
	  if(cube)
	    { 
	      double sum = 0; 
	      sum =  fabs(Lami(0)-2.)/2.;
	      cout << "cube-err: " << sum << endl;
	      if( fabs(sum-sum_old) < 1.e-5 ) ii = maxsteps+1; 
	      sum_old = sum; 
	      }*/
	else
	    {
	      double sum =0; 
	      cout << endl << " ii = " << ii  << " : " << endl;  
	      
	      sum = Lami[0]; 
	      for (int m=0;m<nev; m++)
		cout << " Lami[ " << m << " ] =  " << Lami[m]  << "\t" ; 
	      if(fabs(sum-sum_old)/fabs(sum) < prec) break; // ii=maxsteps+1;  
	      sum_old=sum;
	    }
	
	       
	lam_old = lam; 
	//	if(lshape && (fabs(lam_old-lam)/fabs(lam) < 1.e-4*fabs(solev[0]-lam_old)/fabs(solev[0])))  break; 
	// lam_old = lam; 
	

	
	
	//double y = sqrt(InnerProduct(vmu = matm*vu,vu)); 
	// vu /= sqrt(msmall(m,m));
	    

      }
    

    

    for (int nn = 0; nn<nev ; nn++) cout << " solev[nn] " <<  solev[nn] << endl; 
    if(lshape)
      cout  << endl << " ********** LSHAPE  ******* " << endl; 
    if(cube) 
      cout << endl << " ********** CUBE    ******* " << endl; 
    
    if(lshape || cube )   
      {
	double sum=0.;
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
		
	for (int m=0;m< nev; m++)
	  {
	    
	    cout << "lami [m] = " <<  Lami(m) << " \t error[m] = "
		 << fabs((solev[m] - Lami(m))/solev[m]) << "\t" << endl; 
	     
	    evout << minorder << "\t" << maxorder << "\t" <<  gfu->GetFESpace().GetNDof() << "\t" << fabs((solev[m] - Lami(m))/solev[m]) << "\t"; 
	    
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]);
	   
	  }	
	  
	evout2 <<  minorder << " \t" << maxorder << "\t" << gfu->GetFESpace().GetNDof() << "\t" << sum << " \t " << sum/nev  << "\t"; 
     
	
	for (int m=0;m<nev;m++) 
	  evout2 << fabs((solev[m]-Lami(m))/Lami(m)) << "\t"; 
	evout2 << endl; 
	
	cout << "ndofs " << gfu->GetFESpace().GetNDof()  << "\t sum/nev " << sum/nev << endl; 	       
	

      }
    else 
      {
	double sum=0; 
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	sum = 0; 
	evout << minorder << "t" << maxorder  << "\t" <<  gfu->GetFESpace().GetNDof() << "\t" ; 
	evout2 << minorder << "t" << maxorder  << "\t" <<  gfu->GetFESpace().GetNDof() << "\t" ; 
	for(int m=0;m<nev;m++)
	  {
	    evout << fabs((solev[m] - Lami(m))/Lami[m]) << "\t";
	    sum += fabs((solev[m] - Lami(m))/Lami[m]); 
	    cout << " lami[ " << m << " ] =  " << Lami[m]  << "\t" ;
	  } 
	
	evout2 << sum/nev << endl;  
      }
  
    for(int m=0;m<nev; m++)
      {
	BaseVector & vu = gfu->GetVector(m); 
	vmu = matm * gfu->GetVector(m); 	  		
	vu /= InnerProduct(gfu->GetVector(m),vmu); 

	// projection
	
	vd = matm * vu;
	vdpot = Transpose (*grad) * vd;
	
	vwpot = matprepot * vdpot;
	vw = (*grad) * vwpot; 
	vu -= vw;


      	
      }	
    
    /*    for(int m=0;m<nev;m++)
      {
	vmu = matm * gfu->GetVector(m); 	  		
	vau = mata * gfu->GetVector(m);
	
	


        vd = vau - Lami(m) * vmu;
	vd /= InnerProduct(gfu->GetVector(m),vmu); 

      }
    
    */

 

    (*testout) << " ********************************************************* " << endl; 
    delete solev; 


  }
#endif








  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("maxwellevp", NumProcMaxwellEVP::Create);
    }

    
    Init init;
  }
}


#endif
