 #ifdef PARALLEL

#include <mystdlib.h>
#include <parallelngs.hpp>
#include <comp.hpp>

namespace ngparallel
{
  using namespace ngparallel;
  using namespace ngcomp;


  ParallelDofs :: ParallelDofs( const FESpace & afespace) : fespace ( afespace )
  {
    MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
    MPI_Comm_rank(MPI_COMM_WORLD, &id);

    ndof = 0;
    isexchangedof.SetSize(0);;
    localexchangedof = 0;
    distantexchangedof = 0;
    ismasterdof.SetSize(0);

    isghostdof = 0;

    nexdof.SetSize(ntasks);
    nexdof = 0;

    distndof.SetSize(ntasks);
    distndof = 0;



 }


  ParallelDofs :: ~ParallelDofs()
  {
    if ( localexchangedof )
      delete localexchangedof;
    if ( distantexchangedof )
      delete distantexchangedof;
    if ( isghostdof )
      delete isghostdof;
    for ( int dest = 0; dest < ntasks; dest++)
      if ( IsExchangeProc (dest) )
	delete this->mpi_t[dest];
    delete [] this->mpi_t;

  }


  void ParallelDofs :: SetDistantDofs ( const int proc, const ARRAY<int> & dofs )
  {
    for ( int i = 0; i < dofs.Size(); i++)
      SetDistantDof ( proc, i, dofs[i] );
  }


  void ParallelDofs :: SetDistantDof ( const int proc, const int localdof, const int distantdof )
  { 
    // exchangedofs[dest][i] ---- distantdof zu exchangedofs[id][i] = localdof

    // find i s.t. exchangedofs[id][i] == localdof
    int i = 0;
    while ( i < nexdof[proc] )
      {
	if ( (*localexchangedof)[proc][i] == localdof ) break;
	i++;
      }
    if ( i == nexdof[proc] ) 
      {
	cout << "!!!!!!!! error in SetDistantDof - localdof is no exchange dof!!!!!!!!" << endl;
	*testout << "!!!!!!!! error in SetDistantDof - localdof is no exchange dof!!!!!!!!" << endl;
      }
    // set exchangedofs[dest][i] = distantdof
    (*distantexchangedof) [proc][i] = distantdof;

  }


  const int ParallelDofs :: GetDistantDof ( const int proc, const int localdof ) const
  {
    // find i s.t. exchangedof[id][i] == localdof
    int i = 0;
    while ( i < nexdof[proc] )
      {
	if ( (*localexchangedof)[proc][i] == localdof ) break;
	i++;
      }
    if ( i == nexdof[proc] ) 
      {
      cout << "!!!!!!!! error in GetDistantDof (" << id << ") - localdof is no exchange dof!!!!!!!!" << endl;
      *testout << "!!!!!!!! error in GetDistantDof (" << id << ") for " << proc << " - localdof is no exchange dof!!!!!!!!" << endl;
      }
    // set exchangedofs[dest][i] = distantdof
    return (*distantexchangedof) [proc][i];

  }


  void ParallelDofs :: GetExchangeProcs ( const int localdof, ARRAY<int> & procs ) const
  {
    procs.SetSize(0);
    for ( int dest = 0; dest < ntasks; dest++)
      if ( IsExchangeDof( dest, localdof ) && dest != id  )
	procs.Append (dest );
    return;
  }

  void ParallelDofs :: GetExchangeProcs ( ARRAY<int> & procs ) const
  {
    procs.SetSize(0);
    if ( ntasks == 1 ) return;
    for ( int dest = 0; dest < ntasks; dest++)
      if (  (*localexchangedof)[dest].Size() > 0 && dest != id )
	procs.Append (dest );
    return;

  }


  void ParallelDofs :: GetHOExchangeProcs ( const int localdof, ARRAY<int> & procs ) const
  {
    procs.SetSize(0);
    for ( int dest = 1; dest < ntasks; dest++)
      if ( IsExchangeDof( dest, localdof ) && dest != id )
	procs.Append (dest );
    return;
  }

  void ParallelDofs :: GetHOExchangeProcs ( ARRAY<int> & procs ) const
  {
    procs.SetSize(0);
    if ( ntasks == 1 ) return;
    for ( int dest = 1; dest < ntasks; dest++)
      if (  (*localexchangedof)[dest].Size() > 0 && dest != id )
	procs.Append (dest );
    return;

  }

  const bool  ParallelDofs :: IsExchangeProc ( const int proc ) const
  {
    if ( id == proc )
      return false;
    if ( id == 0 || proc == 0 )
      return true;

    if ( (*localexchangedof)[proc].Size() > 0 )
      return true;
    else 
      return false;
  }


  // high-order update for the local procs
  void ParallelDofs :: Update()
  {
    const MeshAccess & ma = fespace . GetMeshAccess();
    ndof = fespace.GetNDof();
    isexchangedof.SetSize( (ntasks+1)*ndof );
    isexchangedof.Clear();


     // ****************************
     // find ghost dofs
     // ****************************

     if ( isghostdof )
       delete isghostdof;
     isghostdof = new BitArray ( ndof );
     isghostdof->Set();

     for ( int el = 0; el < ma.GetNE(); el++ )
       {
	 if ( ma.IsGhostEl ( el ) ) continue;
	 ARRAY<int> dofs;
	 fespace.GetDofNrs ( el, dofs );
	 for ( int i = 0; i < dofs.Size(); i++ )
	   isghostdof -> Clear ( dofs[i] );
       }



    
    
     // **********************
     // decide if dof is masterdof or not
     // **********************


    ismasterdof.SetSize ( ndof );
    ismasterdof.Set();

     for ( int dof = 0; dof < ndof; dof++ )
       {
	 if ( IsGhostDof ( dof ) )
	   ismasterdof.Clear ( dof );
       }

  }

  void ParallelDofs :: UpdateMPIType ()
  {
    this->mpi_t = new MPI_Datatype * [ntasks];

    MPI_Datatype *MPI_T;
    if ( fespace . IsComplex() )
      MPI_T = GetMPIType<Complex>();
    else
      MPI_T = GetMPIType<double>();

    for ( int dest = 0; dest < ntasks; dest++ )
      {
	if ( !IsExchangeProc(dest) ) continue;
	
	FlatArray<int>  localexchangedof = GetLocalExchangeDofs(dest);
	int len_vec = (localexchangedof).Size();


	int * blocklen = new int [len_vec];

	for ( int i = 0; i < len_vec; i++ )
	  {
	    blocklen[i] = fespace.GetDimension();
	  }
	this->mpi_t[dest] = new MPI_Datatype;
	
	MPI_Type_indexed ( len_vec, blocklen, &((localexchangedof)[0]), *MPI_T, this->mpi_t[dest]);    
	MPI_Type_commit ( this->mpi_t[dest] );
	delete [] blocklen; 
      }

    delete MPI_T;
   }


  /************************************
         low order update for local procs
  *************************************/

  /** brauchen wir nimmer --> im fespace **/
//   void ParallelDofs :: UpdateLowOrder( LocalHeap & lh )
//   {
//     (*testout) << "UPDATE LOW ORDER" << endl;

//     const FESpace & low_order_space = fespace . LowOrderFESpace();

//     const MeshAccess & ma = fespace . GetMeshAccess();

//     int ndof_lo = low_order_space.GetNDof();

//     // all dofs are exchange dofs
//     ARRAY<int> nexdof(ntasks);
//     nexdof = ndof_lo;


//     int exdof = 0;
//     BitArray sentdof ( ntasks * ndof );
//     sentdof.Clear();

//     //    localexchangedof = new Table<int> (nexdof);
// //     distantexchangedof = new Table<int> (nexdof);

//     ARRAY<int> owndofs, distantdofs;
 
//     owndofs.SetSize(0);
//     distantdofs.SetSize(0);
    

//     exdof = 0;
//     ARRAY<int> cnt_nexdof(ntasks);
//     cnt_nexdof = 0;


//     // *****************
//     // Parallel Vertex dofs
//     // *****************

//     owndofs.SetSize(1);
//     owndofs[0] = 1;
//     distantdofs.SetSize(0);
    
//      // find local and distant dof numbers for vertex exchange dofs
//     for ( int vert = 0; vert < ma.GetNV(); vert++ )
//       {
// 	ARRAY<int> dnums;
// 	low_order_space.GetVertexDofNrs ( vert, dnums );
	
// 	int dest = 0;
	
// //  	for ( int i=0; i<dnums.Size(); i++ )
// //  	  (*localexchangedof)[id][exdof++] = dnums[i];
	
// 	int distvert = parallelma -> GetDistantPNum ( dest, vert );
// 	owndofs.Append ( distvert );
// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    SetExchangeDof ( dest, dnums[i] );
// 	    // *testout << "set exchange dof " << dnums[i] << " with dest " << dest << endl;
// 	    owndofs.Append ( dnums[i] );
//   	  }
//       }   
    
//     int dest = 0;
//     MyMPI_Send ( owndofs, dest);
//     MyMPI_Recv ( distantdofs, dest);
    



//     int ii = 1;
//     while ( ii < distantdofs.Size() )
//       {
// 	int vnum = distantdofs[ii++];
// 	ARRAY<int> dnums;
// 	low_order_space.GetVertexDofNrs (vnum, dnums);
// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    //		*testout << "receive exdof " << dnums[i] 
// 	    // << " with dest " << dest << endl;
// 	    (*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 	    (*distantexchangedof)[dest][ cnt_nexdof[dest] ] = distantdofs[ii];
// 	    ii++; cnt_nexdof[dest]++;
// 	  }
//       }

    
//     // *****************
//     // Parallel Edge dofs
//     // *****************

//     owndofs.SetSize(1);
//     owndofs[0] = 1;
//     distantdofs.SetSize(0);

//     for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
//       {
// 	ARRAY<int> dnums;
// 	low_order_space.GetEdgeDofNrs ( edge, dnums );
	
//  	for ( int i=0; i<dnums.Size(); i++ )
//  	  (*localexchangedof)[id][exdof++] = dnums[i];

// 	int dest = 0;
// 	int distedge = parallelma -> GetDistantEdgeNum ( dest, edge );

// 	owndofs.Append ( distedge );

// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    SetExchangeDof ( dest, dnums[i] );
// 	    owndofs.Append ( dnums[i] );
//   	  }
//       }   


//     MyMPI_Send ( owndofs, dest);
//     MyMPI_Recv ( distantdofs, dest);

 
//     ii = 1;
//     while ( ii < distantdofs.Size() )
//       {
// 	int ednum = distantdofs[ii++];
// 	ARRAY<int> dnums;
// 	low_order_space.GetEdgeDofNrs (ednum, dnums);
// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    (*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 	    (*distantexchangedof)[dest][ cnt_nexdof[dest] ] = distantdofs[ii];
// 	    ii++; cnt_nexdof[dest]++;
// 	  }
//       }


//     // *****************
//     // Parallel Face dofs
//     // *****************

//     owndofs.SetSize(1);
//     owndofs[0] = 1;
//     distantdofs.SetSize(0);

//     for ( int face = 0; face < ma.GetNFaces(); face++ )
//       {
// 	ARRAY<int> dnums;
// 	low_order_space.GetFaceDofNrs ( face, dnums );

//  	for ( int i=0; i<dnums.Size(); i++ )
//  	  (*localexchangedof)[id][exdof++] = dnums[i];

// 	int dest = 0;
// 	int distface = parallelma -> GetDistantFaceNum ( dest, face );
	
// 	owndofs.Append ( distface );
	
// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    SetExchangeDof ( dest, dnums[i] );
// 	    owndofs.Append ( dnums[i] );
// 	  }
//       }   


//     MyMPI_Send ( owndofs, dest);
//     MyMPI_Recv ( distantdofs, dest);
    
    
//     ii = 1;
//     while ( ii < distantdofs.Size() )
//       {
// 	int fanum = distantdofs[ii++];
// 	ARRAY<int> dnums;
// 	low_order_space.GetFaceDofNrs (fanum, dnums);
// 	for ( int i=0; i<dnums.Size(); i++)
// 	  {
// 	    (*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 	    (*distantexchangedof)[dest][ cnt_nexdof[dest] ] = distantdofs[ii];
// 	    ii++; cnt_nexdof[dest]++;
// 	  }
//       }




//   }

//   // this is what the global proc does
//   void ParallelDofs :: UpdateGlobal( LocalHeap & lh )
//   {
//     (*testout) << "UPDATE GLOBAL" << endl;

//     const MeshAccess & ma = fespace . GetMeshAccess();

//     ndof = fespace.GetNDof();

//     // Find number of exchange dofs
//     nexdof = 0;

//     isexchangedof.SetSize ( (ntasks+1)*ndof );
//     isexchangedof.Clear();


//     // number of vertex exchange dofs
//     for ( int vert = 0; vert < ma.GetNV(); vert++ )
//       {
// 	ARRAY<int> dnums;
// 	fespace.GetVertexDofNrs ( vert, dnums );
// 	      nexdof[id] += dnums.Size();

// 	for ( int dest = 1; dest < ntasks; dest ++ )
// 	  if (  parallelma -> GetDistantPNum ( dest, vert ) >= 0 )
// 	    { 
// 	      nexdof[dest] += dnums.Size(); 
// 	    }
//       }

//     // + number of edge exchange dofs
//     for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
//       {
// 	ARRAY<int> dnums;
// 	fespace.GetEdgeDofNrs ( edge, dnums );
// 	      nexdof[id] += dnums.Size() ;  

// 	for ( int dest = 1; dest < ntasks; dest ++ )
// 	  if (  parallelma -> GetDistantEdgeNum ( dest, edge ) >= 0 )
// 	    { 
// 	      nexdof[dest] += dnums.Size() ;  
// 	    }
//       }

//     // + number of face exchange dofs
//     for ( int face = 0; face < ma.GetNFaces(); face++ )
//       {
	
// 	ARRAY<int> dnums;
// 	fespace.GetFaceDofNrs ( face, dnums );
// 	      nexdof[id] += dnums.Size() ; 

// 	for ( int dest = 1; dest < ntasks; dest ++ )
// 	  if (  parallelma -> GetDistantFaceNum ( dest, face ) >= 0 )
// 	    { 
// 	      nexdof[dest] += dnums.Size() ; 
// 	    }
//       }

//     localexchangedof = new Table<int> (nexdof);
//     distantexchangedof = new Table<int> (nexdof);




//      BitArray sentdof ( ndof * ntasks );
//      sentdof.Clear();


//     ARRAY<int> ** owndofs,** distantdofs;
//     owndofs = new ARRAY<int> * [ntasks];
//     distantdofs = new ARRAY<int> * [ntasks];

//     for ( int i = 0; i < ntasks; i++)
//       {
// 	owndofs[i] = new ARRAY<int> (0);
// 	distantdofs[i] = new ARRAY<int> (0);
//       }

//     int exdof = 0;
//     ARRAY<int> cnt_nexdof(ntasks);
//     cnt_nexdof = 0;

//     // *****************
//     // Parallel Vertex dofs
//     // *****************

//     for ( int dest = 0; dest < ntasks; dest++)
//       {
// 	owndofs[dest]->SetSize(1);
// 	(*owndofs[dest])[0] = 1;
// 	distantdofs[dest]->SetSize(0);
//       }


//     // find local and distant dof numbers for vertex exchange dofs
//     for ( int vert = 0; vert < ma.GetNV(); vert++ )
//       {
// 	ARRAY<int> dnums;
// 	fespace.GetVertexDofNrs ( vert, dnums );
//  	for ( int i=0; i<dnums.Size(); i++ )
//  	  (*localexchangedof)[id][exdof++] = dnums[i];

// 	for ( int dest = 1; dest < ntasks; dest++ )
// 	  {
// 	    int distvert = parallelma -> GetDistantPNum ( dest, vert );
// 	    if( distvert < 0 ) continue;

// 	    owndofs[dest]->Append ( distvert );
// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
//  		SetExchangeDof ( dest, dnums[i] );
//  		SetExchangeDof ( dnums[i] );
// 		// *testout << "set exchange dof " << dnums[i] << " with dest " << dest << endl;
// 		owndofs[dest]->Append ( dnums[i] );
// 	      }
//   	  }
//       }   


//     for ( int dest = 1; dest < ntasks; dest ++ )
//       {
// 	MyMPI_Send ( *owndofs[dest], dest);
// 	MyMPI_Recv ( *distantdofs[dest], dest);
//       }
    


//     int ii = 1;
//     for( int dest=1; dest<ntasks; dest++) 
//       {
// 	ii = 1;
// 	while ( ii < distantdofs[dest]->Size() )
// 	  {
// 	    int vnum = (*distantdofs[dest])[ii++];
// 	    ARRAY<int> dnums;
// 	    fespace.GetVertexDofNrs (vnum, dnums);
// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
// 		//		*testout << "receive exdof " << dnums[i] 
// 		// << " with dest " << dest << endl;
// 		(*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 		(*distantexchangedof)[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
// 		ii++; cnt_nexdof[dest]++;
// 	      }
// 	  }
//       }

    
//     // *****************
//     // Parallel Edge dofs
//     // *****************

//     for ( int dest = 0; dest < ntasks; dest++)
//       {
// 	owndofs[dest]->SetSize(1);
// 	(*owndofs[dest])[0] = 1;
// 	distantdofs[dest]->SetSize(0);
//       }
//     for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
//        {
// 	ARRAY<int> dnums;
// 	fespace.GetEdgeDofNrs ( edge, dnums );

//  	for ( int i=0; i<dnums.Size(); i++ )
//  	  (*localexchangedof)[id][exdof++] = dnums[i];

// 	for ( int dest = 1; dest < ntasks; dest++ )
// 	  {
// 	    int distedge = parallelma -> GetDistantEdgeNum ( dest, edge );
// 	    if( distedge < 0 ) continue;

// 	    owndofs[dest]->Append ( distedge );

// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
// 		SetExchangeDof ( dest, dnums[i] );
// 		SetExchangeDof ( dnums[i] );
// 		owndofs[dest]->Append ( dnums[i] );
// 	      }
//   	  }
//       }   


//     for ( int dest = 1; dest < ntasks; dest ++ )
//       {
// 	MyMPI_Send ( *owndofs[dest], dest);
// 	MyMPI_Recv ( *distantdofs[dest], dest);
//       }

//      for( int dest=1; dest<ntasks; dest++) 
//       {
// 	ii = 1;
// 	while ( ii < distantdofs[dest]->Size() )
// 	  {
// 	    int ednum = (*distantdofs[dest])[ii++];
// 	    ARRAY<int> dnums;
// 	    fespace.GetEdgeDofNrs (ednum, dnums);
// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
// 		if ( dnums[i] < 0 ) continue;
// 		(*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 		(*distantexchangedof)[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
// 		ii++; cnt_nexdof[dest]++;
// 	      }
// 	  }
//       }


//     // *****************
//     // Parallel Face dofs
//     // *****************

//     for ( int dest = 0; dest < ntasks; dest++)
//       {
// 	owndofs[dest]->SetSize(1);
// 	(*owndofs[dest])[0] = 1;
// 	distantdofs[dest]->SetSize(0);
//       }
//     for ( int face = 0; face < ma.GetNFaces(); face++ )
//       {
// 	ARRAY<int> dnums;
// 	fespace.GetFaceDofNrs ( face, dnums );

//  	for ( int i=0; i<dnums.Size(); i++ )
//  	  (*localexchangedof)[id][exdof++] = dnums[i];

// 	for ( int dest = 1; dest < ntasks; dest++ )
// 	  {
// 	    int distface = parallelma -> GetDistantFaceNum ( dest, face );
// 	    if( distface < 0 ) continue;

// 	    owndofs[dest]->Append ( distface );

// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
// 		SetExchangeDof ( dest, dnums[i] );
// 		SetExchangeDof ( dnums[i] );
// 		owndofs[dest]->Append ( dnums[i] );
// 	      }
//   	  }
//       }   


//     for ( int dest = 1; dest < ntasks; dest ++ )
//       {
// 	MyMPI_Send ( *owndofs[dest], dest);
// 	MyMPI_Recv ( *distantdofs[dest], dest);
//       }

//      for( int dest=1; dest<ntasks; dest++) 
//       {
// 	ii = 1;
// 	while ( ii < distantdofs[dest]->Size() )
// 	  {
// 	    int fanum = (*distantdofs[dest])[ii++];
// 	    ARRAY<int> dnums;
// 	    fespace.GetFaceDofNrs (fanum, dnums);
// 	    for ( int i=0; i<dnums.Size(); i++)
// 	      {
// 		if ( dnums[i] < 0 ) continue;
// 		(*localexchangedof)[dest][ cnt_nexdof[dest] ] = dnums[i];
// 		(*distantexchangedof)[dest][ cnt_nexdof[dest] ] = (*distantdofs[dest])[ii];
// 		ii++; cnt_nexdof[dest]++;
// 	      }
// 	  }
//       }

//      // ****************************
//      // find ghost dofs
//      // ****************************

//      if ( isghostdof )
//        delete isghostdof;
//      isghostdof = new BitArray ( ndof );
//      isghostdof->Clear();

//     for ( int i = 0; i < ntasks; i++ )
//       delete owndofs[i], distantdofs[i];
//     delete[] owndofs, distantdofs;
      
//     ismasterdof.SetSize ( ndof );
//     ismasterdof.Set();

//     Print();
//    }




  void ParallelDofs ::  Print() const
  {
    if ( this == 0 ) return;

    (*testout) << endl <<  "DEGREES OF FREEDOM FOR PARALLEL SPACES" << endl << endl;
    (*testout) << "NDOF = "  << ndof << endl;
      for ( int i = 0; i < ndof; i++ )
	{ 
	if ( IsExchangeDof(i) )
	  {

	    (*testout) << "exchange dof  " << i << ": " <<  endl;
	    for ( int dest = 0; dest < ntasks; dest ++)
	      if ( IsExchangeDof( dest, i ) )
		(*testout) << "   p" << dest << ": " << GetDistantDof ( dest, i ) << endl; 

	  }
	}

  }


  void ParallelDofs ::  GetDistantDofs ( const int localdof, ARRAY<int> & distantdofs ) const
  { 
    distantdofs.SetSize(ntasks);
    // no distant dof on proc 0, as this one has only low order fespace
    distantdofs[0] = -1;
    for ( int i=1; i<ntasks; i++)
      distantdofs[i] = GetDistantDof ( i, localdof );

  }


  bool ParallelDofs :: ContainsParallelDof ( const FlatArray<int> & doftable ) const
  {
    if ( this == 0 ) { return false;}
    if ( ntasks == 1 ) return false;
    for ( int i = 0; i < doftable.Size(); i++ )
      if ( IsExchangeDof ( doftable[i] ) ) return true;
    return false;
  }



  void ParallelDofs :: UpdateCompound ( )
  {
    int base = 0;

    const CompoundFESpace & compoundspace = dynamic_cast<const CompoundFESpace & > (fespace);
    int nspaces = compoundspace.GetNSpaces();

    const MeshAccess & ma = fespace . GetMeshAccess();
    ndof = fespace.GetNDof();
    isexchangedof.SetSize( (ntasks+1)*ndof );
    isexchangedof.Clear();
  
    // **************************
    // find exchange dofs
    // **************************
  
    isexchangedof.SetSize ( ndof * (ntasks+1) );
    isexchangedof.Clear();
  
    base = 0;
    for ( int i = 0; i < nspaces; i++ )
      {
	for ( int dof = 0; dof < compoundspace[i]->GetNDof(); dof++ )
	  {
	    if ( compoundspace[i]->GetParallelDofs().IsExchangeDof(dof) ) 
	      SetExchangeDof ( dof + base );
	    for ( int dest = 0; dest < ntasks; dest++ )
	      if ( compoundspace[i]->GetParallelDofs().IsExchangeDof(dest, dof) )
		SetExchangeDof ( dest, dof + base );
	  }
	base += compoundspace[i]->GetNDof();
      }



     // ****************************
     // find ghost dofs
     // ****************************

    if ( isghostdof )
      delete isghostdof;
    isghostdof = new BitArray ( ndof );
    isghostdof->Set();
    
    base = 0;
    for ( int i = 0; i < nspaces; i++ )
      {
	for ( int dof = 0; dof < compoundspace[i]->GetNDof(); dof++ )
	  {
	    if ( ! compoundspace[i]->GetParallelDofs().IsGhostDof(dof) ) 
	      isghostdof -> Clear ( dof + base );
	  }
	base += compoundspace[i]->GetNDof();
      }


    
    
     // **********************
     // decide if dof is masterdof or not
     // **********************


    ismasterdof.SetSize ( ndof );
    ismasterdof.Set();

  base = 0;
    for ( int i = 0; i < nspaces; i++ )
      {
	for ( int dof = 0; dof < compoundspace[i]->GetNDof(); dof++ )
	  {
	    if ( ! compoundspace[i]->GetParallelDofs().IsMasterDof(dof) ) 
	      ismasterdof.Clear ( dof + base );
	  }
	base += compoundspace[i]->GetNDof();
      }

    // ****************************
    // copy localexchangedof, distantexchangedof
    // ****************************


    nexdof = 0;
    for ( int dest = 0; dest < ntasks; dest ++ )
      for ( int i = 0; i < nspaces; i++ )
	nexdof[dest] += compoundspace[i]->GetParallelDofs().GetNExDofs( dest );

    localexchangedof = new Table<int> (nexdof);
    distantexchangedof = new Table<int> (nexdof);
    
    ARRAY<int> distbase(ntasks);
    distbase = 0;
  base = 0;
  ARRAY<int> ii_comp(ntasks);
  ii_comp = 0;
    for ( int i = 0; i < nspaces; i++ )
      {
	for ( int dest = 0; dest < ntasks; dest++ )
	  {
	    for ( int ii = 0; ii < (*compoundspace[i]->GetParallelDofs().localexchangedof)[dest].Size(); ii++ )
	      {
		(*localexchangedof)[dest][ii_comp[dest]] =  
		  (*compoundspace[i]->GetParallelDofs().localexchangedof)[dest][ii] + base;
		(*distantexchangedof)[dest][ii_comp[dest]] =  
		  (*compoundspace[i]->GetParallelDofs().distantexchangedof)[dest][ii] + distbase[dest];
		ii_comp[dest]++;
	      }
	    distbase[dest] += compoundspace[i]->GetParallelDofs() . GetDistNDof (dest);
	  }
	base += compoundspace[i]->GetNDof();
      }


  }




}

#endif
