#include <parallelngs.hpp>

namespace ngla
{
  using namespace ngla;
  using namespace ngparallel;

#ifdef PARALLEL



  template <typename T>
  ParallelVFlatVector<T> :: ParallelVFlatVector () :
    paralleldofs(0)
  { 
    // initialize mpi types needed for sending values to distant procs
    this->entrysize = sizeof(T) / sizeof(double);

    if ( ntasks == 1 || this -> paralleldofs == 0) return;

    this -> SetParallelDofs ( this->paralleldofs );

  }


  template <typename T>
  ParallelVFlatVector<T> :: ParallelVFlatVector (int as, T * adata)
    : VFlatVector<T>(as, adata),
      paralleldofs(0)
  { 

    if ( ntasks == 1 || this -> paralleldofs == 0) return;

    this -> SetParallelDofs ( this->paralleldofs );

  }

  template <typename T>
  ParallelVFlatVector<T> :: ParallelVFlatVector (int as, T * adata, ngparallel::ParallelDofs * aparalleldofs)
    : VFlatVector<T> (as, adata),
      paralleldofs(aparalleldofs)
  { 

    if ( paralleldofs == 0 ) 
      return;

    this -> SetParallelDofs ( aparalleldofs );

  }

  template <typename T>
  ParallelVFlatVector<T> :: ParallelVFlatVector (int as, T * adata,
						 ngparallel::ParallelDofs * aparalleldofs,
						 PARALLEL_STATUS astatus)
    : VFlatVector<T> (as, adata),
      status(astatus),
      paralleldofs(aparalleldofs)
  {
    if ( paralleldofs == 0 ) 
      return;

    this -> SetParallelDofs ( aparalleldofs );

  }


  template <typename T>
  ParallelVVector<T> :: ParallelVVector (int as)
    : VVector<T> (as)
  { 
    this -> paralleldofs  = 0;
  }

  template <typename T>
  ParallelVVector<T> :: ParallelVVector (int as, ngparallel::ParallelDofs * aparalleldofs )
   
    : VVector<T> (as),
      paralleldofs(aparalleldofs)
  {
    if ( aparalleldofs != 0 )
      {
	this -> SetParallelDofs ( aparalleldofs );
	this->status = CUMULATED;
      }
    else
      {
	paralleldofs = 0;
	status = NOT_PARALLEL;
      }

  }

  template <typename T>
  ParallelVVector<T> :: ParallelVVector (int as, ngparallel::ParallelDofs * aparalleldofs,
					 PARALLEL_STATUS astatus ) 
    : VVector<T> (as),
      status(astatus), 
      paralleldofs(aparalleldofs)
  {
    if ( aparalleldofs != 0 )
      this -> SetParallelDofs ( aparalleldofs );
    else
      paralleldofs = 0;

  }

  template <typename T>
  ParallelVVector<T> :: ~ParallelVVector() throw()
  { 
    if ( ntasks == 1 || this -> paralleldofs == 0 ) return;
    delete [] this->recvvalues;
    ;
  }

  template <typename T>
  ParallelVFlatVector<T> :: ~ParallelVFlatVector() throw()
  { 
    ;
    if ( ntasks == 1 || this->paralleldofs == 0 ) return;
    delete [] this->recvvalues;
  }


  template <typename T>
  void ParallelVFlatVector<T> :: SetParallelDofs ( ParallelDofs * aparalleldofs, const ARRAY<int> * procs  )
  {
    this -> paralleldofs = aparalleldofs;
    if ( this -> paralleldofs == 0 ) return;


    if ( ntasks == 1 ) return;

    if ( this->recvvector_size.Size() ) return;

    this->sendvector_size . SetSize(ntasks);
    this->recvvector_size . SetSize(ntasks);

    ARRAY<int> exprocs(0);
    if ( procs == 0 )
      {
	for ( int dest = 0; dest < ntasks; dest++ )
	  if ( this->paralleldofs->IsExchangeProc(dest) )
	    exprocs.Append(dest);
      }
    else
      {
	for ( int idest = 0; idest < procs->Size(); idest++ )
	  if ( this->paralleldofs->IsExchangeProc( (*procs)[idest] ) )
	    exprocs.Append ( (*procs)[idest] );
      }
  
    for ( int idest = 0; idest < exprocs.Size(); idest++)
      {
	int dest = exprocs[idest];

	FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);
	int len_vec = (localexchangedof).Size();
	int len_fvec = len_vec* this->entrysize;
	this->sendvector_size[dest] = len_fvec;

	MPI_Request intrequest;
	MPI_Isend( &len_fvec, 1, MPI_INT, dest, 10, MPI_COMM_WORLD, &intrequest);
      }

    for ( int idest = 0; idest < exprocs.Size(); idest++)
      {
	int dest = exprocs[idest];
	if ( ! this -> paralleldofs -> IsExchangeProc (dest) ) continue;
      
	MPI_Request intrequest;
	MPI_Status status;
	MPI_Irecv( &recvvector_size[dest], 1, MPI_INT, dest, MPI_ANY_TAG, MPI_COMM_WORLD, &intrequest);
	MPI_Wait ( &intrequest, &status);
      }

    this->recvvalues = new ARRAY<T> [ntasks]; 
    for ( int idest = 0; idest < exprocs.Size(); idest++ )
      {
	int dest = exprocs[idest];
	if ( ! this -> paralleldofs -> IsExchangeProc (dest) )
	  { 
	    this->recvvalues[dest].SetSize(0);
	  }
	else
	  {
	    this->recvvalues[dest].SetSize(recvvector_size[dest]);
	  }
      }
  }



  template <typename T>
  void ParallelVVector<T> :: SetParallelDofs ( ParallelDofs * aparalleldofs, const ARRAY<int> * procs )
  {

    this -> paralleldofs = aparalleldofs;
    if ( this -> paralleldofs == 0 ) return;

    if ( this->recvvector_size.Size() ) return;
  
#ifdef SCALASCA
#pragma pomp inst begin (vvector_setparalleldofs)
#endif
  
//     mpi_t = new MPI_Datatype * [ntasks];
    this->sendvector_size . SetSize(ntasks);
    this->recvvector_size . SetSize(ntasks);
  
    MPI_Request * sendintrequest = new MPI_Request[ntasks];

    ARRAY<int> exprocs(0);
    if ( procs == 0 )
      {
	for ( int dest = 0; dest < ntasks; dest++ )
	  if ( this->paralleldofs->IsExchangeProc(dest) )
	    exprocs.Append(dest);
      }
    else
      {
	for ( int idest = 0; idest < procs->Size(); idest++ )
	  if ( this->paralleldofs->IsExchangeProc( (*procs)[idest] ) )
	    exprocs.Append ( (*procs)[idest] );
      }

    for ( int idest = 0; idest < exprocs.Size(); idest++)
      {
	int dest = exprocs[idest];
	if ( ! this -> paralleldofs -> IsExchangeProc (dest) ) continue;
	FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);
	int len_vec = (localexchangedof).Size();
	int len_fvec = len_vec* this->entrysize;
	this->sendvector_size[dest] = len_fvec;
      
	MPI_Isend( &len_fvec, 1, MPI_INT, dest, 100, MPI_COMM_WORLD, sendintrequest + dest);
      }
  
    for ( int idest = 0; idest < exprocs.Size(); idest++)
      {
	int dest = exprocs[idest];
	MPI_Request intrequest;
	MPI_Status status;
	MPI_Irecv( &recvvector_size[dest], 1, MPI_INT, dest, MPI_ANY_TAG, MPI_COMM_WORLD, &intrequest);
	MPI_Wait ( &intrequest, &status);
	MPI_Wait ( sendintrequest + dest, &status );
      }
 
    this -> recvvalues = new ARRAY<T> [ntasks];    // nur einmal anlegen (JS) -- bald :)
    for ( int idest = 0; idest < exprocs.Size(); idest++ )
      {
	int dest = exprocs[idest];
	if ( ! this -> paralleldofs -> IsExchangeProc (dest) )
	  { 
	    this->recvvalues[dest].SetSize(0);
	  }
	else
	  {
	    this->recvvalues[dest].SetSize(recvvector_size[dest]);
	  }
      }
  
#ifdef SCALASCA
#pragma pomp inst end (vvector_setparalleldofs)
#endif
    delete [] sendintrequest;
  }


  /// values from reduceprocs are added up,
  /// vectors in sendtoprocs are set to the cumulated values
  /// default pointer 0 means send to proc 0
  template <typename T>
  void ParallelVVector<T> :: AllReduce ( ARRAY<int> * reduceprocs, ARRAY<int> * sendtoprocs ) const 
  {
    // in case of one process only, return
    if ( status != DISTRIBUTED ) return;
    (*testout) << "ALLREDUCE" << endl;
    
#ifdef SCALASCA
#pragma pomp inst begin (vvector_allreduce)
#endif

    MPI_Status status;

    ARRAY<int> exprocs(0);
    int nexprocs;
    
    // find which processors to communicate with
    for ( int i = 0; i < reduceprocs->Size(); i++)
      if ( paralleldofs->IsExchangeProc((*reduceprocs)[i]) )
	exprocs.Append((*reduceprocs)[i]);
    
    nexprocs = exprocs.Size();
    int cntexproc = 0;
    
    ParallelVVector<T> * constvec = const_cast<ParallelVVector<T> * > (this);
    
    MPI_Request * sendrequest, *recvrequest;
    if ( nexprocs > 0 )
      {
	sendrequest = new MPI_Request[nexprocs];
	recvrequest = new MPI_Request[nexprocs];
      }
    else
      {
	sendrequest = 0; recvrequest = 0;
      }

    ARRAY<int> sendto_exprocs(0);
    if ( sendtoprocs )
      {
	for ( int i = 0; i < sendtoprocs->Size(); i++ )
	  if ( paralleldofs->IsExchangeProc((*sendtoprocs)[i]) 
	       || (*sendtoprocs)[i] == id )
	    sendto_exprocs.Append((*sendtoprocs)[i] );
      }
    else
      sendto_exprocs.Append(0);

    // if the vectors are distributed, reduce
    if ( reduceprocs->Contains(id) && this->Status() == DISTRIBUTED )
      {
	
	// send 
	for ( int idest = 0; idest < nexprocs; idest ++ ) 
	  {
	    int dest = exprocs[idest];
	FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);
// 	FlatArray<int> * localexchangedof = 
// 	      const_cast<FlatArray<int> * > (paralleldofs->GetLocalExchangeDofs(dest));

	    //if ( localexchangedof->Size()<0 || dest == id ) continue; // denen ich was zu sagen hab
	    constvec->ISend ( dest, sendrequest[idest] );
	  }

	// receive	
	for ( int isender=0; isender < nexprocs; isender++)
	  {
	    int sender = exprocs[isender];
	    constvec -> IRecvVec ( sender, recvrequest[isender] );
	  }
    

	// cumulate
    
	// MPI_Waitany --> wait for first receive, not necessarily the one with smallest id

	for ( cntexproc=0; cntexproc < nexprocs; cntexproc++ )
	  {
	    int sender, isender;
	    MPI_Waitany ( nexprocs, &recvrequest[0], &isender, &status); 
	    sender = exprocs[isender];
	    MPI_Wait ( sendrequest+isender, &status );
	    int ii = 0;
	    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof++ )
	      {
		if ( ! paralleldofs->IsExchangeDof(sender, dof) ) continue;
		(*constvec) ( dof ) += this->recvvalues[sender][ii++];
	      } 
	  } 
	
	for ( int isender = 0;  isender < nexprocs; isender ++)
	  MPI_Wait ( sendrequest + isender, &status);
      }

    constvec->SetStatus(CUMULATED);

    // +++++++++++++++
    // 
    // now send vector to the sendto-procs


    if ( reduceprocs->Contains(id) )
      {
	nexprocs = sendto_exprocs.Size();
	if ( sendrequest )
	  delete [] sendrequest; delete [] recvrequest;
	if ( nexprocs > 0 )
	  {
	    sendrequest = new MPI_Request[nexprocs];
	    recvrequest = new MPI_Request[nexprocs];
	  }
	else
	  {
	    sendrequest = 0; recvrequest = 0;
	  }
	for ( int idest = 0; idest < nexprocs; idest++ )
	  {
	    int dest = sendto_exprocs[idest];
	    if ( ! paralleldofs->IsExchangeProc(dest) ) continue;
	    FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);
// 	    FlatArray<int> * localexchangedof = 
// 	      const_cast<FlatArray<int> * > (paralleldofs->GetLocalExchangeDofs(dest));
	    constvec->Send ( dest );
	  }
      }
    else if ( sendto_exprocs.Contains(id) )
      {
	for ( int isender = 0; isender < nexprocs; isender ++)
	  {
	    int sender = exprocs[isender];
	    int ii = 0; 
	    if ( ! paralleldofs->IsExchangeProc ( sender ) ) continue; 
	    constvec -> IRecvVec ( sender, recvrequest[isender] );
	    MPI_Wait(recvrequest+isender, &status);
	    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof++ )
	      {
		if ( ! paralleldofs->IsExchangeDof(sender, dof) ) continue;
		(*constvec) ( dof ) = this->recvvalues[sender][ii++];
	      } 
	  }
      }
#ifdef SCALASCA
#pragma pomp inst end (vvector_horeduce)
#endif
    if ( sendrequest )
      delete [] recvrequest; delete [] sendrequest;

  }
  






  template <typename T>
  void ParallelVVector<T> :: Distribute() const
  {
    if (  status != CUMULATED ) return;

    (*testout ) << "distribute! " << endl;
    //    *testout << "vector before distributing " << *this << endl;
    ParallelVVector<T> * constvec = const_cast<ParallelVVector<T> * > (this);
    constvec->SetStatus(DISTRIBUTED);

    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof ++ )
      if ( ! paralleldofs->IsMasterDof ( dof ) )
	(*constvec)(dof) = 0;

    //   *testout << "distributed vector " << *constvec << endl;
     
  }



  /// values from reduceprocs are added up,
  /// vectors in sendtoprocs are set to the cumulated values
  /// default pointer 0 means send to proc 0
  template <typename T>
  void ParallelVFlatVector<T> :: AllReduce ( ARRAY<int> * reduceprocs, ARRAY<int> * sendtoprocs ) const 
  {
    // in case of one process only, return
    if ( status != DISTRIBUTED ) return;

#ifdef SCALASCA
#pragma pomp inst begin (vvector_allreduce)
#endif

    ARRAY<int> exprocs(0);
    int nexprocs;
    MPI_Status status;

    // find which processors to communicate with
    for ( int i = 0; i < reduceprocs->Size(); i++)
      if ( paralleldofs->IsExchangeProc((*reduceprocs)[i]) )
	exprocs.Append((*reduceprocs)[i]);
    
    nexprocs = exprocs.Size();
    int cntexproc = 0;
    
    ParallelVFlatVector<T> * constvec = const_cast<ParallelVFlatVector<T> * > (this);
    
    MPI_Request * sendrequest, *recvrequest;
    sendrequest = new MPI_Request[nexprocs];
    recvrequest = new MPI_Request[nexprocs];
    
    ARRAY<int> sendto_exprocs(0);
    if ( sendtoprocs )
      {
	for ( int i = 0; i < sendtoprocs->Size(); i++ )
	  if ( paralleldofs->IsExchangeProc((*sendtoprocs)[i]) )
	    sendto_exprocs.Append((*sendtoprocs)[i] );
      }
    else
      sendto_exprocs.Append(0);

    // if the vectors are distributed, reduce
    if ( this->status == DISTRIBUTED && reduceprocs->Contains(id) )
      {
	(*testout) << "reduce! high order" << endl;
	constvec->SetStatus(CUMULATED);
	
	// send 
	for ( int idest = 0; idest < nexprocs; idest ++ ) 
	  {
	    int dest = exprocs[idest];
// 	    FlatArray<int> * localexchangedof = 
// 	      const_cast<FlatArray<int> * > (paralleldofs->GetLocalExchangeDofs(dest));

	    //if ( localexchangedof->Size()<0 || dest == id ) continue; // denen ich was zu sagen hab
	    constvec->ISend ( dest, sendrequest[idest] );
	  }

	// receive	
	for ( int isender=0; isender < nexprocs; isender++)
	  {
	    int sender = exprocs[isender];
	    constvec -> IRecvVec ( sender, recvrequest[isender] );
	  }
    

	// cumulate
    
	// MPI_Waitany --> wait for first receive, not necessarily the one with smallest id

	for ( cntexproc=0; cntexproc < nexprocs; cntexproc++ )
	  {
	    int sender, isender;
	    MPI_Waitany ( nexprocs, &recvrequest[0], &isender, &status); 
	    sender = exprocs[isender];
	    MPI_Wait ( sendrequest+isender, &status );
	    int ii = 0;
	    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof++ )
	      {
		//*testout << "degree of freedom " << dof << endl;
		if ( ! paralleldofs->IsExchangeDof(sender, dof) ) continue;
		constvec->Data()[dof] += this->recvvalues[sender][ii++];
	      } 
	  } 
	
      }


    // +++++++++++++++
    // 
    // now send vector to the sendto-procs


    if ( reduceprocs->Contains(id) )
      {
	nexprocs = sendto_exprocs.Size();
	delete [] sendrequest; delete [] recvrequest;
	sendrequest = new MPI_Request[nexprocs];
	recvrequest = new MPI_Request[nexprocs];
	    
	for ( int idest = 0; idest < nexprocs; idest++ )
	  {
	    int dest = sendto_exprocs[idest];
	    if ( ! paralleldofs->IsExchangeProc(dest) ) continue;
// 	    FlatArray<int> * localexchangedof = 
// 	      const_cast<FlatArray<int> * > (paralleldofs->GetLocalExchangeDofs(dest));
	    //*testout << "localexdof " << endl << *localexchangedof << endl;
	    //MyMPI_Send ( (*this).FVDouble(), this->entrysize, localexchangedof, dest);
	    constvec->Send ( dest );
	  }
      }
    else if ( sendto_exprocs. Contains(id) )
      {
	for ( int isender = 0; isender < nexprocs; isender ++)
	  {
	    int sender = exprocs[isender];
	    int ii = 0; 
	    if ( ! paralleldofs->IsExchangeProc ( sender ) ) continue; 
	    constvec -> IRecvVec ( sender, recvrequest[isender] );
	    MPI_Status status;
	    MPI_Wait(recvrequest+isender, &status);
	    
	    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof++ )
	      {
		if ( ! paralleldofs->IsExchangeDof(sender, dof) ) continue;
		constvec->Data()[dof] = this->recvvalues[sender][ii++];
	      } 
	  }
      }

#ifdef SCALASCA
#pragma pomp inst end (vvector_horeduce)
#endif
    delete [] recvrequest; delete []sendrequest;
  }
  




  template <typename T>
  void ParallelVFlatVector<T> :: Distribute() const
  {
    if ( status != CUMULATED ) return;
    ParallelVFlatVector * constflatvec = const_cast<ParallelVFlatVector<T> * > (this);
    T * constvec = const_cast<T * > (this->Data());
    constflatvec->SetStatus(DISTRIBUTED);
    
    for ( int dof = 0; dof < paralleldofs->GetNDof(); dof ++ )
      if ( ! paralleldofs->IsMasterDof ( dof ) )
	constvec[dof] = 0;
   
  }





  template <typename T>
  void ParallelVVector<T> :: ISend ( const int dest, MPI_Request & request )
  {
    FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);

    int len_vec = (localexchangedof).Size();

    int * blocklen = new int [len_vec];
    for ( int i=0; i<len_vec; i++ ) blocklen[i] = 1;

    MPI_Datatype * mpi_t = this->paralleldofs->GetMPI_Type(dest);
  
    MPI_Isend( const_cast<T*> (& (this->data[0])), 1, *mpi_t, dest, 2000, MPI_COMM_WORLD, &request);
  
  }



  template <typename T>
  void ParallelVFlatVector<T> :: ISend ( const int dest, MPI_Request & request )
  {
    FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);

    int len_vec = (localexchangedof).Size();

    int * blocklen = new int [len_vec];
    for ( int i=0; i<len_vec; i++ ) blocklen[i] = 1;
  
    MPI_Datatype * mpi_t = this->paralleldofs->GetMPI_Type(dest);
    MPI_Isend( const_cast<T*> (& (this->data[0])), 1, *mpi_t, dest, 2000, MPI_COMM_WORLD, &request);
  
  }



  template <typename T>
  void ParallelVVector<T> :: Send ( const int dest )
  {
    FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);

    int len_vec = (localexchangedof).Size();

    int * blocklen = new int [len_vec];
    for ( int i=0; i<len_vec; i++ ) blocklen[i] = 1;
  
    MPI_Datatype * mpi_t = this->paralleldofs->GetMPI_Type(dest);
    MPI_Send( const_cast<T*> (& (this->data[0])), 1, *mpi_t, dest, 2001, MPI_COMM_WORLD);
  
  }



  template <typename T>
  void ParallelVFlatVector<T> :: Send ( const int dest )
  {
    FlatArray<int>  localexchangedof = this->paralleldofs->GetLocalExchangeDofs(dest);

    int len_vec = (localexchangedof).Size();

    int * blocklen = new int [len_vec];
    for ( int i=0; i<len_vec; i++ ) blocklen[i] = 1;
  
    MPI_Datatype * mpi_t = this->paralleldofs->GetMPI_Type(dest);
    MPI_Send( const_cast<T*> (& (this->data[0])), 1, *mpi_t, dest, 2001, MPI_COMM_WORLD);
  
  }


  /*
    template <typename T>
    void VVector<T> :: IRecvVec (  ARRAY<T> & s, const int dest, MPI_Request & request ) const
    {
    MPI_Status status;
    MPI_Datatype * MPI_T = GetMPIType<T> ();
    *testout << "recvvecsize = " << recvvector_size[dest] << endl;
    *testout << "mysize = " << s.Size() << endl;
    *testout << "for recv from " << dest << endl;
    //s.SetSize(this->recvvector_size[dest]);
    MPI_Irecv( &s[0], recvvector_size[dest], *MPI_T, dest, MPI_ANY_TAG, MPI_COMM_WORLD, &request);
    delete MPI_T;
    ;
    }

    template <typename T>
    void VFlatVector<T> :: IRecvVec (  ARRAY<T> & s, const int dest, MPI_Request & request ) const
    {
    MPI_Status status;
    MPI_Datatype * MPI_T = GetMPIType<T> ();
    //s.SetSize(this->recvvector_size[dest]);
    MPI_Irecv( &s[0], recvvector_size[dest], *MPI_T, dest, MPI_ANY_TAG, MPI_COMM_WORLD, &request);
    delete MPI_T;
    ;
    }
  */


  template <typename T>
  void ParallelVVector<T> :: IRecvVec ( const int dest, MPI_Request & request )
  {
    MPI_Datatype * MPI_T = GetMPIType<T> ();
    MPI_Irecv( &(this->recvvalues[dest][0]), recvvector_size[dest], *MPI_T, dest, 
	       MPI_ANY_TAG, MPI_COMM_WORLD, &request);
    delete MPI_T;
    ;
  }

  template <typename T>
  void ParallelVFlatVector<T> :: IRecvVec ( const int dest, MPI_Request & request )
  {
    MPI_Datatype * MPI_T = GetMPIType<T> ();
    MPI_Irecv(&( this->recvvalues[dest][0]), recvvector_size[dest], *MPI_T, 
	      dest, MPI_ANY_TAG, MPI_COMM_WORLD, &request);
    delete MPI_T;
    ;
  }

template <typename T>
void ParallelVFlatVector<T> :: PrintParallelDofs () const
  { paralleldofs->Print(); }

template <typename T>
void ParallelVVector<T> :: PrintParallelDofs () const
  { paralleldofs->Print(); }


template <typename T>
void ParallelVFlatVector<T> :: SetStatus ( PARALLEL_STATUS astatus )
{ 
  status = astatus; 
}

template <typename T>
void ParallelVVector<T> :: SetStatus ( PARALLEL_STATUS astatus ) 
{ status = astatus; }



template <typename T>
BaseVector * ParallelVVector<T> :: CreateVector ( const ARRAY<int> * procs ) const
  {
    ParallelVVector<T> * parvec;
 
    parvec = new ParallelVVector<T> (this->size);
    parvec->SetStatus( this->Status() );

    // *testout << "procs... " << procs << endl;
    if( this->Status() != NOT_PARALLEL )
      {
	ParallelDofs * aparalleldofs = this-> GetParallelDofs();
	parvec->SetParallelDofs ( aparalleldofs, procs );
      }
    return parvec;
    
  }



  template <typename T>
  BaseVector * ParallelVFlatVector<T> :: CreateVector ( const ARRAY<int> * procs ) const
  {
    ParallelVVector<T> * parvec;

    parvec = new ParallelVVector<T> (this->size);
    parvec->SetStatus( this->Status() );
    
    if( this->Status() != NOT_PARALLEL )
      {
	ParallelDofs * aparalleldofs = this-> GetParallelDofs();
	parvec->SetParallelDofs ( aparalleldofs, procs );
      }
    return parvec;
  }


template < class T >
ostream & ParallelVVector<T> :: Print (ostream & ost) const
{
  ost << "addr = " << &this->FV()(0) << endl;
  const PARALLEL_STATUS status = this->Status();
  if ( status == NOT_PARALLEL )
    ost << "NOT PARALLEL" << endl;
  else if ( status == DISTRIBUTED )
    ost << "DISTRIBUTED" <<endl;
  else if ( status == CUMULATED )
    ost << "CUMULATED" << endl;
  return (ost << this->FV() << endl);
}

template < class T >
ostream & ParallelVFlatVector<T> :: Print (ostream & ost) const
{
  ost << "addr = " << &this->FV()(0) << endl;
  const PARALLEL_STATUS status = this->Status();
  if ( status == NOT_PARALLEL )
    ost << "NOT PARALLEL" << endl;
  else if ( status == DISTRIBUTED )
    ost << "DISTRIBUTED" <<endl;
  else if ( status == CUMULATED )
    ost << "CUMULATED" << endl;
  return (ost << this->FV() << endl);
}



#endif

#ifdef PARALLEL
template class ParallelVFlatVector<double>;
template class ParallelVFlatVector<Complex>;
template class ParallelVFlatVector<Vec<1,double> >;
template class ParallelVFlatVector<Vec<1,Complex> >;
template class ParallelVFlatVector<Vec<2,double> >;
template class ParallelVFlatVector<Vec<2,Complex> >;
template class ParallelVFlatVector<Vec<3,double> >;
template class ParallelVFlatVector<Vec<3,Complex> >;
template class ParallelVFlatVector<Vec<4,double> >;
template class ParallelVFlatVector<Vec<4,Complex> >;

template class ParallelVFlatVector<Vec<5,double> >;
template class ParallelVFlatVector<Vec<5,Complex> >;
template class ParallelVFlatVector<Vec<6,double> >;
template class ParallelVFlatVector<Vec<6,Complex> >;
template class ParallelVFlatVector<Vec<7,double> >;
template class ParallelVFlatVector<Vec<7,Complex> >;
template class ParallelVFlatVector<Vec<8,double> >;
template class ParallelVFlatVector<Vec<8,Complex> >;
template class ParallelVFlatVector<Vec<9,double> >;
template class ParallelVFlatVector<Vec<9,Complex> >;

template class ParallelVFlatVector<Vec<12,double> >;
template class ParallelVFlatVector<Vec<12,Complex> >;
template class ParallelVFlatVector<Vec<18,double> >;
template class ParallelVFlatVector<Vec<18,Complex> >;
template class ParallelVFlatVector<Vec<24,double> >;
template class ParallelVFlatVector<Vec<24,Complex> >;


template class ParallelVVector<double>;
template class ParallelVVector<Complex>;
template class ParallelVVector<Vec<1,double> >;
template class ParallelVVector<Vec<1,Complex> >;
template class ParallelVVector<Vec<2,double> >;
template class ParallelVVector<Vec<2,Complex> >;
template class ParallelVVector<Vec<3,double> >;
template class ParallelVVector<Vec<3,Complex> >;
template class ParallelVVector<Vec<4,double> >;
template class ParallelVVector<Vec<4,Complex> >;

template class ParallelVVector<Vec<5,double> >;
template class ParallelVVector<Vec<5,Complex> >;
template class ParallelVVector<Vec<6,double> >;
template class ParallelVVector<Vec<6,Complex> >;
template class ParallelVVector<Vec<7,double> >;
template class ParallelVVector<Vec<7,Complex> >;
template class ParallelVVector<Vec<8,double> >;
template class ParallelVVector<Vec<8,Complex> >;
template class ParallelVVector<Vec<9,double> >;
template class ParallelVVector<Vec<9,Complex> >;
template class ParallelVVector<Vec<10,double> >;
template class ParallelVVector<Vec<10,Complex> >;
template class ParallelVVector<Vec<11,double> >;
template class ParallelVVector<Vec<11,Complex> >;
template class ParallelVVector<Vec<15,double> >;
template class ParallelVVector<Vec<15,Complex> >;
template class ParallelVVector<Vec<13,double> >;
template class ParallelVVector<Vec<13,Complex> >;
template class ParallelVVector<Vec<14,double> >;
template class ParallelVVector<Vec<14,Complex> >;

template class ParallelVVector<Vec<12,double> >;
template class ParallelVVector<Vec<12,Complex> >;
template class ParallelVVector<Vec<18,double> >;
template class ParallelVVector<Vec<18,Complex> >;
template class ParallelVVector<Vec<24,double> >;
template class ParallelVVector<Vec<24,Complex> >;

  
template class ParallelVFlatVector<FlatVector<double> >;
template class ParallelVVector<FlatVector<double> >;
#endif


}
