/* demohybrid.cpp 
 * demonstrate the solution of 2D Poisson problem by mixed-hybrid fem
 * problem: 
 *         -div (eps grad(u)) = f in Omega, u=g on Gamma0
 *
 * notation:
 *         (u,v)_T := int_T u v dx,     (u,v) = sum_T (u,v)_T
 *         <u,v>_dT := int_dT u v ds,   <u,v> = sum_T <u,v>_dT 
 *
 * mixed-hybrid variational form: 
 *  (1)      1/eps (tau, sigma) + (div(tau), u) - <tau.n, lambda> = 0         for all tau
 *  (2)      (v, div(sigma))                                      = -(f,v)    for all v 
 *  (3)     -<mu, sigma.n>                                        = 0         for all mu
 * 
 * note: the lagrange multiplier lambda is a function on the edges and
 *       has the meaning of a (mean value of the ) trace of u
 * 
 * boundary conditions: due to the dirichlet boundary conditions, (3) holds only for
 * mu in the interior. on the boundary set lambda = g by aadding a robin penalty term
 *
 *  (3b)     pen <mu,lambda>_Gamma0 = pen <mu, g>_Gamma0    
 * 
 * spaces: choose the following spaces for a stable discretization:
 *  sigma in discontinuous HDiv, polynomial order p => sigma.n in P_p(E) (if BDM elements as for netgen on trigs)
 *  u in L2(T), polynomial order p-1
 *  lambda in L2(E) = FacetSpace, polynomial order p on edges => enough to ensure continuity of sigma.n
 *
 * Ingredients:
 *    - HybridFESpace -- compound space for (sigma,u,lambda)
 *    - HybridLaplaceIntegrator (1) - (3): has volume contributions (.,.) and facet contributions <.,.>
 *    - HybridRobinIntegrator, HybridNeumannIntegrator (3b): for incorporting the boundary conditions
 *
 */

#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>


namespace demohybrid_cpp
{

  using namespace ngfem;
  using namespace ngsolve;


  /* *********************************************************** */
  /* ********************** HybridFESpace ********************** */
  /* *********************************************************** */
  class HybridFESpace : public CompoundFESpace
  {
    public:
      HybridFESpace (const MeshAccess & ama, const ARRAY<const FESpace*> & aspaces)
      : CompoundFESpace (ama, aspaces)
      { }

      virtual ~HybridFESpace () { ; }

      static FESpace * Create (const MeshAccess & ma, const Flags & flags)
      {
        ARRAY<const FESpace*> spaces(3);

        Flags hdivflags(flags), l2flags(flags), facetflags(flags);

        int order = int (flags.GetNumFlag ("order", 1));

        l2flags.SetFlag ("order", max(order-1,0));
        hdivflags.SetFlag ("order",order);
        hdivflags.SetFlag ("discontinuous");
        facetflags.SetFlag("order", order);

        // a discontinuous HDiv space
        spaces[0] = new HDivHighOrderFESpace (ma, hdivflags);
        
        spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
        
        // a space for functions defined on element interfaces 
        spaces[2] = new FacetFESpace (ma, facetflags);        

        HybridFESpace * fes = new HybridFESpace (ma, spaces);
        return fes;
      }
    
      virtual string GetClassName () const { return "HybridFESpace"; }
  };

  
  /* *********************************************************** */
  /* ************* HybridLaplaceIntegrator ********************* */
  /* *********************************************************** */
  template<int D>
  class HybridLaplaceIntegrator : public BilinearFormIntegrator
  {
    protected:
      CoefficientFunction *coef_eps;
    public:
      HybridLaplaceIntegrator (ARRAY<CoefficientFunction*> & coeffs) 
      : BilinearFormIntegrator()
      { 
        coef_eps  = coeffs[0];
      }

      virtual ~HybridLaplaceIntegrator () { ; }

      static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
      {
        return new HybridLaplaceIntegrator<D> (coeffs);
      }
    
      virtual bool BoundaryForm () const 
      { return 0; }


      virtual void AssembleElementMatrix (const FiniteElement & fel,
                                          const ElementTransformation & eltrans, 
                                          FlatMatrix<double> & elmat,
                                          LocalHeap & lh) const
      {
        const CompoundFiniteElement & cfel = 
            dynamic_cast<const CompoundFiniteElement&> (fel);
    
        const HDivFiniteElement<D> & fel_hdiv = 
            dynamic_cast<const HDivFiniteElement<D>&> (cfel[0]);
        const L2HighOrderFiniteElement & fel_l2 = 
            dynamic_cast<const L2HighOrderFiniteElement&> (cfel[1]);
        const FacetVolumeFiniteElement & fel_facet = 
            dynamic_cast<const FacetVolumeFiniteElement&> (cfel[2]);
    
        ELEMENT_TYPE eltype = cfel.ElementType();
      
        int nd = cfel.GetNDof();
        int nd_l2 = fel_l2.GetNDof();
        int nd_hdiv = fel_hdiv.GetNDof();
        int nd_facet = fel_facet.GetNDof();
      
        elmat.AssignMemory(nd, nd, lh);
        elmat = 0.0;

        
        // ****************************************************************************
        // the volume terms 1/eps (tau,sigma) + (div(tau),u) + (v, div(sigma))
        // ****************************************************************************
        FlatVector<> mat_l2(nd_l2, lh);
        FlatMatrix<> mat_hdiv(nd_hdiv, D, lh); 
        FlatVector<> mat_div_hdiv(nd_hdiv, lh);
        FlatVector<> mat_facet(nd_facet, lh);
      
        FlatMatrix<> mat_tau_sigma(nd_hdiv,nd_hdiv,lh) ;
        FlatMatrix<> mat_tau_u(nd_hdiv,nd_l2,lh); 
          
        const IntegrationRule & volume_ir =
            GetIntegrationRules().SelectIntegrationRule (eltype, 2*fel_hdiv.Order());
        
        void * heapp = lh.GetPointer();
        for (int l=0; l<volume_ir.GetNIP(); l++)
        {
          const SpecificIntegrationPoint<D,D> sip(volume_ir[l], eltrans, lh);
        
          double det = sip.GetJacobiDet();
          fel_l2.CalcShape (sip.IP(), mat_l2) ;
        
          // piola transformation
          mat_hdiv =  (1.0/det) * fel_hdiv.GetShape(sip.IP(), lh) * Trans(sip.GetJacobian());
          mat_div_hdiv = (1.0/det) * fel_hdiv.GetDivShape(sip.IP(), lh);

          double dx = det * volume_ir[l].Weight();
          double eps = coef_eps->Evaluate(sip);
        
          //  1/eps * tau * sigma * dx
          mat_tau_sigma = (dx/eps) * mat_hdiv * Trans( mat_hdiv );
        
          //  div(tau) * u * dx
          mat_tau_u = dx * mat_div_hdiv * Trans(mat_l2);
        
          // copy into the element matrix
          for (int i=0; i<nd_hdiv; i++)
          {
            for (int j=0; j<nd_hdiv; j++)
            {
              elmat(i,j) +=  mat_tau_sigma(i,j);
            }
            double sum=0;
            for (int j=0; j<nd_l2; j++)
            {
              elmat(i,nd_hdiv+j) += mat_tau_u(i,j);     //  (div(tau), u)
              elmat(nd_hdiv+j,i) += mat_tau_u(i,j);     //  (v, div(sigma)
            }
          }
          lh.CleanUp(heapp);
        }
        
        // **********************************************************************************
        // the facet contribution: <tau.n, lambda>, <mu, sigma.n>
        // **********************************************************************************
        int nfa = ElementTopology :: GetNFacets(eltype);
        
        FlatMatrix<> mat_tau_lambda(nd_hdiv, nd_facet, lh);
        FlatVector<> mat_hdiv_n(nd_hdiv, lh);
        
        mat_tau_lambda = 0;
        
        for (int k = 0; k < nfa; k++)
        {
          ELEMENT_TYPE facettype = ElementTopology :: GetFacetType(eltype, k);
          const IntegrationRule & facet_ir =
              GetIntegrationRules().SelectIntegrationRule (facettype, 2*fel_hdiv.Order());
      
          // tranformation of facet integration points (1D) to element points (2D):
          Facet2ElementTrafo transform(eltype); 

          // normal on reference element with |normal| = length of tangential (in 3D: area of face)
          const NORMAL * normals = ElementTopology::GetNormals(eltype);
            
          Vec<D> normal_ref;
          for (int i=0; i<D; i++)
            normal_ref(i) = normals[k][i];
            
          void * heapp = lh.GetPointer();
          

          for (int l = 0; l < facet_ir.GetNIP(); l++)
          {
            const IntegrationPoint  & facet_ip = facet_ir[l];
            IntegrationPoint volume_ip = transform(k, facet_ip);
          
            SpecificIntegrationPoint<D,D> sip (volume_ip, eltrans, lh);
	    
            // piola mapping of the normal: normal = det * F^(-T) normal_ref
            // this mapping has the property that 
            //    length of mapped normal = length of mapped edge (2D)
            //    length of mapped normal = area of mapped surface (3D)
            // area = |normal| and  normal0 = normal/|normal|
            //
            // sigma * normal_0 = 1/det sigma_ref F^T * det F^(-T) normal_ref / |normal|
            //                  = sigma_ref * normal_ref / area
            // the term 1/area cancels out below where ds = area * weight 
            mat_hdiv_n = fel_hdiv.GetShape(sip.IP(), lh) * normal_ref;
         
            // use (d-1) dimensional integration point to evaluate lambda
            fel_facet.CalcFacetShape(k, facet_ir[l], mat_facet); 

            // ds = area * weight: here area cancels out with the 1/area in sigma.n
            double ds = facet_ir[l].Weight();

            mat_tau_lambda += (ds * mat_hdiv_n) * Trans(mat_facet);
            lh.CleanUp(heapp);
          }
        } // loop over facets
        
        // copy to element matrix
        for (int i=0; i<nd_hdiv; i++)
        {
          for (int j=0; j<nd_facet; j++)
          {
            elmat(i,nd_hdiv+nd_l2+j) -= mat_tau_lambda(i,j); // <tau.n, l>
            elmat(j+nd_hdiv+nd_l2,i) -= mat_tau_lambda(i,j); // <mu, sigma.n>
          }
        }
      }
  };


  template class HybridLaplaceIntegrator<2>;
  template class HybridLaplaceIntegrator<3>;


  class Init
  { 
    public: 
      Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hybrid", HybridFESpace::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridlaplace", 2, 1,
    HybridLaplaceIntegrator<2>::Create);
  
    GetIntegrators().AddBFIntegrator ("hybridlaplace", 3, 1,
    HybridLaplaceIntegrator<3>::Create);
  
  }

  Init init;

} // namespace

