/*

Solver for the nonlinear magnetic problem

div [ nu(|grad u|) grad u ] =  f


with  H(B) := nu(|B|) B  piecewise linear

*/


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;




class NonLinMagDMat : public DMatOp<NonLinMagDMat>
{
  CoefficientFunction * mu1;
  CoefficientFunction * mu2;
  CoefficientFunction * bsat;

public:

  enum { DIM_DMAT = 2 };

  NonLinMagDMat (CoefficientFunction * amu1,
		 CoefficientFunction * amu2,
		 CoefficientFunction * absat)
    : mu1(amu1), mu2(amu2), bsat(absat) { ; }


  template <typename FEL, typename SIP, typename VEC, typename MAT>
  void GenerateLinearizedMatrix (const FEL & fel, const SIP & sip, VEC & vlin,
				 MAT & mat, LocalHeap & lh) const 
  {
    double mu1i = mu1 -> Evaluate (sip);
    double mu2i = mu2 -> Evaluate (sip);
    double bsati = bsat -> Evaluate (sip);

    double babs = L2Norm (vlin);

    if (babs < bsati)
      {
	mat = 1.0/mu1i * Id<2> ();
      }
    else
      {
	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);
	double nu = habs / babs;
	double nuprime = (1/mu2i * babs - habs) / (babs*babs);

	mat = nuprime/babs * (vlin * Trans(vlin)) + nu * Id<2>();
      }
  }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    Vec<2> vlin = 0.0;
    GenerateLinearizedMatrix (fel, sip, vlin, mat, lh);
  }


  template <typename FEL, typename SIP, class VECX, class VECY>
  void Apply (const FEL & fel, const SIP & sip,
	      const VECX & x, VECY & y, LocalHeap & lh) const
  {
    typedef typename VECX::TSCAL TSCAL;

    double mu1i = mu1 -> Evaluate (sip);
    double mu2i = mu2 -> Evaluate (sip);
    double bsati = bsat -> Evaluate (sip);

    double babs = L2Norm (x);

    if (babs < bsati)
      {
	y = (1/mu1i) * x;
      }

    else
      
      {
	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);

	y = (habs/babs) * x;
      }
  }


  template <typename FEL, typename SIP, class VECLIN, class VECX, class VECY>
  void ApplyLinearized (const FEL & fel, const SIP & sip, const VECLIN & vlin,
			const VECX & x, VECY & y, ngfem::LocalHeap & lh) const
  {
    typedef typename VECX::TSCAL TSCAL;
    Mat<2,2, TSCAL> mat;

    double mu1i = Evaluate (*mu1, sip);
    double mu2i = Evaluate (*mu2, sip);
    double bsati = Evaluate (*bsat, sip);

    double babs = L2Norm (vlin);

    if (babs < bsati)
      {
	y = 1.0/mu1i * x;
      }

    else
      
      {
	double hsati = bsati / mu1i;
	double habs = hsati + 1/mu2i * (babs-bsati);
	double nu = habs / babs;
	double nuprime = (1/mu2i * babs - habs) / (babs*babs);

	mat = (nuprime/babs) * vlin * Trans(vlin) + nu * Id<2>();
	y = mat * x;
      }
  }




  template <typename FEL, typename SIP, class TVX>
  double Energy (const FEL & fel, const SIP & sip,
		 const TVX & x, LocalHeap & lh) const  
  {
    double mu1i = Evaluate (*mu1, sip);
    double mu2i = Evaluate (*mu2, sip);
    double bsati = Evaluate (*bsat, sip);

    double babs = L2Norm (x);

    if (babs < bsati)
      
      return 0.5/mu1i * babs * babs;

    else
      {
	double hsati = bsati / mu1i;
	double habs = hsati + (babs-bsati) / mu2i;
	return 0.5 * hsati * bsati + 0.5 * (hsati+habs) * (babs-bsati);
      }
  }
  



};



class NonLinMagIntegrator 
  : public T_NonlinearBDBIntegrator<DiffOpGradient<2>, NonLinMagDMat, NodalFiniteElement>
{
public:
  NonLinMagIntegrator (CoefficientFunction * coeff1,
		       CoefficientFunction * coeff2,
		       CoefficientFunction * coeff3)
    :  T_NonlinearBDBIntegrator<DiffOpGradient<2>, NonLinMagDMat, NodalFiniteElement>
  (NonLinMagDMat (coeff1, coeff2, coeff3))
  { 
    ;
  }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new NonLinMagIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }
  
  virtual string Name () const { return "NonLinMag"; }
};





class NumProcNonLinMag : public NumProc
{
protected:
  BilinearForm * bfa;
  LinearForm * lff;
  GridFunction * gfu;

  int maxit;

public:
  NumProcNonLinMag (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { 
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", "a"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    maxit = int ( flags.GetNumFlag ("maxit", 30));
  }

  virtual ~NumProcNonLinMag()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcNonLinMag (pde, flags);
  }

  virtual void Do(LocalHeap & lh)
  {
    cout << "nonlinmag solver called" << endl;

    BaseVector & vecu = gfu->GetVector();
    const BaseVector & vecf = lff->GetVector();

    BaseVector & uold = *vecu.CreateVector();
    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();

    BilinearFormApplication applya(bfa);

    double err, errold, err0;
    double energy, energyold;

    d = vecf - applya * vecu;
    err0 = L2Norm(d);

    for (int i = 1; i <= maxit; i++)
      {
	cout << "newton it " << i << endl;
	
	bfa -> AssembleLinearization (vecu, lh);
	BaseMatrix & inva = 
	  *dynamic_cast<BaseSparseMatrix&> (bfa -> GetMatrix()).InverseMatrix();


	d = vecf - applya * vecu;
	err = L2Norm(d);
	energy = bfa->Energy(vecu) - InnerProduct (vecf, vecu);

	cout << " err = " << err/err0;
	cout << " energy = " << energy << endl;

	errold = err;
	energyold = energy;

	w = inva * d;
	uold = vecu;
	int lin_its = 0;
	double tau = 1;

	do
	  {
	    vecu = uold + tau * w;
	    energy = bfa->Energy(vecu) - InnerProduct (vecf, vecu);

	    cout << "tau = " << tau
		 << " energy = " << energy << endl;

	    tau *= 0.5;
	  }
	while (energy > energyold && lin_its++ < 30 && err > 1e-7*err0);


	delete &inva;

	if (err < 1e-7*err0) break;
      }
  }
};








namespace 
#ifdef MACOS
demo_nonlinear_cpp
#endif 
  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("nonlinmag", NumProcNonLinMag::Create);

    GetIntegrators().AddBFIntegrator ("nonlinmag", 2, 3,
				      NonLinMagIntegrator::Create);
  }
  
  Init init;
}
